#include <math.h>
#include "object.h"
#include "mystdlib.h"
#include "swdef.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"

extern int graphix;
extern Node *LibRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

/*-------------------- eff_lig --------------------*/

int eff_lig(Atom *a)
{
  /*actual ligance of atom = nbr. of non-dummy bonds */

  int i,nbnd;

  if (is_atom(a)) {
    for (nbnd=0,i=0;i<a->nbnd;i++) nbnd+=BT_count(a->bnd[i]->bt);
    return(nbnd);
  }else return(0);
}

/*-------------------- sane_formlig --------------------*/

int sane_formlig(Atom *a)
{
  int i,lig,nbnd,val;

  if (is_atom(a)) {
    if (a->lig=='?') {
      if (a->chrg=='?') {
	for (nbnd=0,val=0,i=0;i<a->nbnd;i++) {
	  nbnd+=BT_count(a->bnd[i]->bt);
	  val+=BT_el(a->bnd[i]->bt);
	}
	val/=2;
	if (mvld(a->atp)) {
	  switch (a->atp->z) {
	  case 1:   /* H  */
	    if (nbnd==0) lig=1;
	    else lig=nbnd;
	    break;
	  case 5:   /* B  */
	    if (nbnd==0) lig=3;
	    else if (nbnd==1) lig=(val==2 ? 2 : val==1 ? 3 : nbnd);
	    else if (nbnd==2) lig=(val==2 ? 3 : nbnd);
	    else lig=nbnd;
	    break;
	  case 6:   /* C  */
	    lig=nbnd+4-val;
	    break;
	  case 7:   /* N  */
	    if (nbnd==0) lig=3;
	    else if (nbnd==1) lig=4-val;
	    else if (nbnd==2) lig=(val==2 ? 3 : 2);
	    else lig=nbnd; 
	    break;
	  case 8:   /* O  */
	    if (nbnd==0) lig=2;
	    else if (nbnd==1) lig=nbnd+(2-val);
	    else lig=nbnd;
	    break;
	  case 9:   /* F  */
	    if (nbnd==0) lig=1;
	    else lig=nbnd;
	    break;
	  case 15:  /* P  */
	    if (nbnd==0) lig=3;
	    else if (nbnd==1) lig=4-val;
	    else if (nbnd==2) lig=(val==2||val==4 ? 3 : nbnd);
	    else if (nbnd==3) lig=(val==4 ? 4 : nbnd );
	    /* else if (nbnd==4) lig=(val==4 ? 5 : nbnd);   prefer charge! */
	    else lig=nbnd;
	    break;
	  case 16:  /* S  */
	    if (nbnd==0) lig=2;
	    else if (nbnd==1) lig=(val==1 ? 2 : nbnd);
	    else if (nbnd==2) lig=(val==3||val==5 ? 3 : nbnd);
	    else if (nbnd==3) lig=(val==3 ? 4 : val==5 ? 4 : nbnd);
	    /* else if (nbnd==4) lig=(val==5 ? 5 : nbnd);   prefer charge! */
	    else lig=nbnd;
	    break;
	  case 17:   /* Cl */
	  case 35:   /* Br */
	  case 53:   /* I  */
	  case 11:   /* Na */
	  case 19:   /* K  */
	  case 12:   /* Mg */
	  case 20:   /* Ca */
	  case 30:   /* Zn */
	    lig=nbnd;
	    break;
	  default:   /* assume zero formal charge */
	    a->chrg=0;
	    lig=asg_formlig(a);
	    a->chrg='?';
	    break;
	  }
	  if (lig<nbnd) lig=nbnd;
	}else lig=nbnd;
      }else lig=asg_formlig(a);
    }else lig=a->lig;
  }else lig=0;
  return(lig);
}

/*-------------------- asg_formlig --------------------*/

int asg_formlig(Atom *a)
{
  /* assign formal ligance */

  int j,nval,val,chrg,nadd,nlp,mlp,noct,lig,nbnd,sv_lig;

  if (!mvld(a)) return(-1);
  
  nval=a->val;
  sv_lig=a->lig;
  a->lig='?';

  for (nbnd=0,val=0,j=0;j<a->nbnd;j++) {
    if (a->bnd[j]->bo!='?') val+=2*a->bnd[j]->bo;
    else val+=BT_el(a->bnd[j]->bt);
    nbnd+=BT_count(a->bnd[j]->bt);
  }
  if (mvld(a->atp)) {
    val/=2;
    chrg=(a->chrg != '?' ? a->chrg : asg_formchrg(a,sane_formlig(a)));  /* if '?', determine */
    nadd=nval-val-chrg;
    nlp=(nval > 4 ? nval - 4 : 0);
    mlp=nlp;
    while (nlp>0 && nadd > 1) {
      nlp--; nadd-=2;
    }
    nlp=mlp-nlp;
    noct=2*val+2*nadd+2*nlp;
    while (noct < 8 && nlp > 0) {
      noct+=2; nlp--; nadd+=2;
    }
    while (noct > 8 && nadd >=2) {
      noct-=2; nlp++; nadd-=2;
    }
    if (nadd>0) lig=nbnd+nadd;
    else lig=nbnd;
  }else lig=nbnd;

  if (lig<nbnd) lig=nbnd;
  a->lig=sv_lig;
  return(lig);
}
/*-------------------- asg_formchrg --------------------*/

int asg_formchrg(Atom *a,int lig)
{
  int j,chrg,nbnd,nlp,nadd,mlp,xtr,val,noct,oct;

  if (is_atom(a)) {
    if (mvld(a->atp)) {
      if (a->atp->z>0) {
	for (nbnd=0,val=0,j=0;j<a->nbnd;j++) {
	  val+=BT_el(a->bnd[j]->bt);
	  nbnd+=BT_count(a->bnd[j]->bt);
	}
      }else{
	val=0;
	nbnd=0;
      }
      if (a->atp->z==5 && val==6 && nbnd==3) {  /* Boron with "Oktett-Luecke" */
	chrg=0;
      }else if (a->atp->z>0){
	val+=2*(lig-nbnd);
	xtr=val%2; val-=xtr;
	if (val==0 && a->atp->z==1) chrg=1;
	else {
	  nadd=a->val-val/2;
	  mlp=(a->val>4 ? a->val-4 : 0); nlp=0;
	  while (nadd>1 && nlp>mlp) {
	    nadd=-2; nlp++;
	  }
	  noct=val+nadd+2*nlp;
	  oct = (a->atp->z > 2 ? 8 : 2);
	  while (noct > oct && nlp<mlp) {
	    noct-=2; nlp++; nadd-=2;
	  }
	  while (noct < oct && nlp > 0) {
	    noct+=2; nlp--; nadd+=2;
	  }
	  chrg=noct-(a->atp->z > 2 ? 8 : 2);
	  if (chrg<-4) chrg+=8;
	  if (chrg>4) chrg-=8;
	}
      }else chrg=0;
    }else chrg=0;
  }else chrg=0;
  return(chrg);
}
/*-------------------- guess_atom_type --------------------*/

extern Node *LibRoot;

Atomdef *guess_atom_type(char *name)
{
  /* attempts to assign atomtype based on name only */

  Node *ResLib;
  Atomdef *t;
  char elt[3],mnem[5];
  int n;

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  if (ResLib&&name) {
    Strcpy(elt,name);
    if (elt[0]) {
      if (elt[0]>='a' && elt[0]<='z') elt[0]=elt[0]+'A'-'a';
      if (elt[1]>='A' && elt[1]<='Z') elt[1]=elt[1]+'a'-'A';
      if (elt[1] < 'a' || elt[1] > 'z') elt[1]='\0';
      strcpy(mnem,elt);strcat(mnem,".?");
      t=find_atomdef_name(ResLib,mnem);
      if (!mvld(t)) {
	t=find_atomdef_name(ResLib,elt);
      }
      if (!mvld(t)) {
	n=ptab_n(elt);
	if (n>=0) {
	  t=(Atomdef *)add_node(ResLib,mnem,OBJ_ATOMDEF);
	  if (mvld(t)) {
	    t->mass=ptab_M(n);
	    t->z=ptab_Z(n);
	    t->val=ptab_val(n);
	    strcpy(t->chs,elt);
	  }
	}
      }
    }else t=(Atomdef *)NULL;
  }else t=(Atomdef *)NULL;
  if (!mvld(t)&&ResLib) t=find_atomdef_name(ResLib,"unk"); 
  return(t);
}
