#include <math.h>
#include "object.h"
#include "mystdlib.h"
#include "swdef.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"

extern int graphix;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

/*========================================================================*/

#define MAX_EL 11
                     /*  C  N  O   S   P  Se  H  F  Cl  Br   I     */
static int Z[MAX_EL] = { 6, 7, 8, 16, 15, 34, 1, 9, 17, 35, 53 };     

static float R0[MAX_EL][MAX_EL] = {
    { 1.52,  1.43,  1.39,  1.80,  1.82, 1.97, 1.08, 1.35, 1.76, 1.90, 2.1},
    { 1.43,  1.42,  1.39,  1.66,  0   , 1.95, 1.05, 1.25, 1.66, 1.80, 2.1},
    { 1.39,  1.39,  0   ,  1.60,  1.61, 1.83, 1.00, 1.25, 1.66, 1.80, 2.1},
    { 1.80,  1.66,  1.60,  0   ,  2.09, 2.18, 1.01, 1.45, 1.86, 2.06, 2.1},
    { 1.82,  0   ,  1.61,  2.09,  0   , 2.25, 1.01, 1.45, 1.86, 2.05, 2.1},
    { 1.97,  1.95,  1.83,  2.18,  2.25, 2.34, 1.01, 1.35, 1.80, 1.80, 2.1},
  };

static float BD[MAX_EL][MAX_EL] = {
    { 0.174, 0.178, 0.185, 0.229, 0.1 , 0.13, 0,    0,    0,    0,    0 },
    { 0.178, 0.177, 0.190, 0.22 , 0   , 0.13, 0,    0,    0,    0,    0 },
    { 0.185, 0.190, 0    , 0.200, 0.15, 0.24, 0,    0,    0,    0,    0 },
    { 0.229, 0.22 , 0.200, 0    , 0.19, 0.5 , 0,    0,    0,    0,    0 },
    { 0.1  , 0    , 0.15 , 0.19 , 0   , 0.14, 0,    0,    0,    0,    0 },
    { 0.13 , 0.13 , 0.24 , 0.5  , 0.14, 0.25, 0,    0,    0,    0,    0 }
  };

/*========================================================================*/

int mark_aromatic_bonds(Bond *b)
{
  int l,k,m,depth;
  Atom *atom[6];
  Bond *bond[5];
  float base[5],bo,bn;
  int done,choice[5];
  Atom *a,*from,*to,*nei;

  if (!mvld(b)) return(0);
  if (b->arom) return(b->arom);

  if (b->f->nbnd<2 || b->t->nbnd<2) return(b->arom);
  if (b->f->nbnd>3 || b->t->nbnd>3) return(b->arom);
  if (!mvld(b->f->atp)||(b->f->atp->z!=6&&b->f->atp->z!=7)) return(b->arom);
  if (!mvld(b->t->atp)||(b->t->atp->z!=6&&b->t->atp->z!=7)) return(b->arom);

  base[0]=b->bo;
  if (base[0]!=2.0 && base[0]!=1.0) return(b->arom);

  depth=0;

  from=b->f;
  to=b->t;
  from->flg=(1<<1);
  to->flg=(1<<1);

  atom[0]=from;
  choice[0]=0;
  bond[0]=b;

  done=FALSE;
  while (!done&&depth>=0) {
    while (!done&&choice[depth]<atom[depth]->nbnd) {
      k=choice[depth];
      choice[depth]++;
      nei=neighbour(k,atom[depth]);
      if (!(nei->flg&(1<<1))) {
	if (nei->nbnd<2||nei->nbnd>3) continue;
	if (!mvld(nei->atp)||(nei->atp->z!=6&&nei->atp->z!=7)) continue;
	bo=3.0-base[depth];
	bn=(atom[depth]->bnd[k]->arom ? bo : atom[depth]->bnd[k]->bo);
	if (bo!=bn) continue;
	if (depth<4) {
	  nei->flg|=(1<<1);
	  depth++;
	  bond[depth]=atom[depth-1]->bnd[k];
	  atom[depth]=nei;
	  choice[depth]=0;
	  base[depth]=bo;
	}
      }else if (depth==4 && nei==to) {
	bo=3.0-base[depth];
	bn=(atom[depth]->bnd[k]->arom ? bo : atom[depth]->bnd[k]->bo);
	if (bo==bn) {
	  for (l=0;l<5;l++) {
	    atom[l]->flg&=~(1<<1);
	    set_bt(bond[l],BT_AROMATIC);
	  }
	  set_bt(atom[depth]->bnd[k],BT_AROMATIC);
	  to->flg&=~(1<<1);
	  for (a=to,l=0;l<5;a=atom[l++]) {  /* recursively expand arom. sys. */
	    if (a->nbnd>2) {
	      for (m=0;m<a->nbnd;m++) {
		if (!a->bnd[m]->arom) mark_aromatic_bonds(a->bnd[m]);
	      }
	    }
	  }
	  done=TRUE;
	}
      }
    }
    if (choice[depth]>=atom[depth]->nbnd) {
      atom[depth]->flg&=~(1<<1);
      depth--;
    }
  }
  to->flg&=~(1<<1);
  return(b->arom);
}
/*-------------------- asg_bondtype --------------------*/

int asg_bondtype(Bond *b)
{
  int j;
  Atom *A,*B,*C;

  if (!mvld(b)) return(-1);

  if (!b->arom&&b->bo==2) mark_aromatic_bonds(b);
  if (b->arom) {
    set_bt(b,BT_AROMATIC);
  }else{
    if (b->bo==1) {
      set_bt(b,BT_SINGLE);
      if (mvld(b->f->atp)&&mvld(b->t->atp)) {     /* amide bond ? */
	A=(Atom *)NULL; B=(Atom *)NULL; C=(Atom *)NULL;
	if (b->f->atp->z == 7 && b->t->atp->z == 6) {A=b->f; B=b->t;}
	else if (b->f->atp->z == 6 && b->t->atp->z == 7) {A=b->t; B=b->f;}
	if (A && B) {
	  if (A->lig==3 && A->chrg==0 && B->lig==3 && B->chrg==0) {
	    for (j=0;j<B->nbnd;j++) {
	      if (B->bnd[j]->bo==2) {
		C=neighbour(j,B);
		if (mvld(C->atp)) {
		  if (C->lig==1 && C->atp->z==8) break;
		}
	      }
	    }
	    if (j<B->nbnd) set_bt(b,BT_AMIDE);
	  }
	}
      }
    }
    else if (b->bo==2) set_bt(b,BT_DOUBLE);
    else if (b->bo==3) set_bt(b,BT_TRIPLE);
  }
  return(b->bt);
}
/*-------------------- asg_bondtype2 --------------------*/

int asg_bondtype2(Bond *b)
{
  int j;
  Atom *A,*B,*C;

  /* variant, supposed to work for atoms with unknown ligance/charge */

  if (!mvld(b)) return(-1);

  if (!b->arom&&b->bo==2) mark_aromatic_bonds(b);
  if (b->arom) {
    set_bt(b,BT_AROMATIC);
  }else{
    if (b->bo==1) {
      set_bt(b,BT_SINGLE);
      if (mvld(b->f->atp)&&mvld(b->t->atp)) {     /* amide bond ? */
	A=(Atom *)NULL; B=(Atom *)NULL; C=(Atom *)NULL;
	if (b->f->atp->z == 7 && b->t->atp->z == 6) {A=b->f; B=b->t;}
	else if (b->f->atp->z == 6 && b->t->atp->z == 7) {A=b->t; B=b->f;}
	if (A && B) {
	  if (A->nbnd<=3 && B->nbnd<=3) {
	    for (j=0;j<A->nbnd;j++) if (A->bnd[j]->bo!=1) break;
	    for (j=(j<A->nbnd ? B->nbnd : 0);j<B->nbnd;j++) {
	      if (B->bnd[j]->bo==2) {
		C=neighbour(j,B);
		if (mvld(C->atp)) {
		  if (C->nbnd==1 && C->atp->z==8) break;
		}
	      }
	    }
	    if (j<B->nbnd) set_bt(b,BT_AMIDE);
	  }
	}
      }
    }
    else if (b->bo==2) set_bt(b,BT_DOUBLE);
    else if (b->bo==3) set_bt(b,BT_TRIPLE);
  }
  return(b->bt);
}
/*-------------------- set_bt --------------------*/

int set_bt(Bond *b,int type) 
{
  int old,i;

  /* set bond type, modify atom flags (arom...) */

  if (mvld(b)) {
    old=b->bt;
    if (b->bt!=type) {
      switch (type) {
      case BT_SINGLE:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_DOUBLE:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_TRIPLE:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_AMIDE:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_AROMATIC:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_NOTCONNECTED:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_UNKNOWN:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      case BT_DUMMY:
	b->bt=type;
	b->bo=BT_order(type);
	break;
      default:
	b->bt=type;
	b->bo='?';
      }
    }
    if (BT_arom(type)) {
      b->arom=TRUE;
      if (mvld(b->f)) b->f->arom=TRUE;
      if (mvld(b->t)) b->t->arom=TRUE;
    }else{
      b->arom=FALSE;
      if (mvld(b->f)&&b->f->arom) {
	for (b->f->arom=FALSE,i=0;i<b->f->nbnd;i++) b->f->arom|=b->f->bnd[i]->arom;
      }
      if (mvld(b->t)&&b->t->arom) {
	for (b->t->arom=FALSE,i=0;i<b->t->nbnd;i++) b->t->arom|=b->t->bnd[i]->arom;
      }
    }
  }else old=BT_UNKNOWN;
  return(old);
}
/*-------------------- est_pi_bondorder --------------------*/

float est_pi_bondorder(Bond *bnd)
{
  /* Returns the estimated pi-bondorder. The estimation is based on an
     empirical linear relationship (PIFF) between bond length 
     and pi-bondorder */

  int i,j;
  float bo;

  bo=-3.0;                               /* the default */
  if (mvld(bnd)) {
    if (mvld(bnd->f)&&mvld(bnd->t)) {
      if (mvld(bnd->f->atp)&&mvld(bnd->t->atp)) {
	for (i=0;i<MAX_EL;i++) if(bnd->f->atp->z==Z[i]) break;
	for (j=0;j<MAX_EL;j++) if(bnd->t->atp->z==Z[j]) break;
	if (i<MAX_EL&&j<MAX_EL) {
	  if (bnd->f->nxyz>0&&bnd->t->nxyz>0) {
	    if (BD[i][j]!=0) {
	      bo=(R0[i][j]-cdist(bnd->f->x,bnd->t->x))/BD[i][j];
	      bo = (bo<-2.0 ? -2.0 : (bo > 3.0 ? 3.0 : bo));
	    }
	  }
	}
      }
    }
  }
  return(bo);
}
/*========================================================================*/

/* flag bits used ... */

#define ATM_FIX     16  /* bit 4 */
#define BND_FIX      2  /* bit 1 */

#define GM_CLASS 0xE    /* bits 1,2,3 */
#define GM_NONE      0  /* missing coordinates... */
#define GM_TERMINAL  2  /* atoms with 1 bond */
#define GM_LINEAR    4  /* two bonds, bond angle > 145 deg. */
#define GM_PLANAR    6  /* three bonds, average torsion(a,n1,n2,n3) < 20 deg. */
#define GM_ISOLATED  8  /* 0 bonds */
#define GM_HIGHER   10  /* more than 3 bonds */
#define GM_OTHER    14  /* 2 or 3 bonds, not linear, not planar */

#define PiBo(x) (((x)->bo36-128.0)/36.0)

/*-------------------- most_likely_dbl --------------------*/

static Bond *most_likely_dbl(Bond *bnd)
{
  Bond *D;
  int k;
  float bo,p;

  D=(Bond *)NULL;
  if (bnd->bt!=BT_UNKNOWN) return(D);
  if (bnd->f->nbnd+bnd->t->nbnd>8) return(D); 

  bo=PiBo(bnd); D=bnd;
  for (k=0;k<bnd->f->nbnd;k++) {
    if (bnd->f->bnd[k]!=bnd) {
      if (bnd->f->bnd[k]->bt==BT_UNKNOWN) {
	if (!(bnd->f->bnd[k]->flg&BND_FIX)) p=-4;	
	else p=PiBo(bnd->f->bnd[k]);
	if (p>bo) if (most_likely_dbl(bnd->f->bnd[k])!=bnd->f->bnd[k]) p=-3;
      }
      else if (bnd->f->bnd[k]->bo > 1) p=4;
      else p=-4;
      if (p>bo) {
	D=bnd->f->bnd[k];
	bo=p;
      }
    }
  }
  for (k=0;k<bnd->t->nbnd;k++) {
    if (bnd->t->bnd[k]!=bnd) {
      if (bnd->t->bnd[k]->bt==BT_UNKNOWN) {
	if (!(bnd->t->bnd[k]->flg&BND_FIX)) p=-4;
	else p=PiBo(bnd->t->bnd[k]);
	if (p>bo) if (most_likely_dbl(bnd->t->bnd[k])!=bnd->t->bnd[k]) p=-3;
      }
      else if (bnd->t->bnd[k]->bo > 1) p=4;
      else p=-4;
      if (p>bo) {
	D=bnd->t->bnd[k];
	bo=p;
      }
    }
  }
  if (D==bnd && bo==-3) D=(Bond *)NULL;
  return(D);
}

/*-------------------- arom_cand_ok --------------------*/

static int arom_cand_ok(Bond *bnd)
{  
  int r,i;
  Bond *b;

  r=bnd->bt==BT_AROMATIC;
  if (!r&&bnd->flg&BND_FIX) {
    r=TRUE;
    if (bnd->f->nbnd>3 || bnd->t->nbnd>3) r=FALSE;
    else if (bnd->f->nbnd<2 || bnd->t->nbnd<2) r=FALSE;
    else if (bnd->f->nbnd==3 && (bnd->f->flg&GM_CLASS)!=GM_PLANAR) r=FALSE;
    else if (bnd->t->nbnd==3 && (bnd->t->flg&GM_CLASS)!=GM_PLANAR) r=FALSE;
    else if (bnd->f->atp->z<6 || bnd->f->atp->z > 7) r=FALSE;
    else if (bnd->t->atp->z<6 || bnd->t->atp->z > 7) r=FALSE;
    else if (bnd->bt==BT_TRIPLE) r=FALSE;
    if (r) {                                    /* exocyclic carbonyl ? */
      if (bnd->f->nbnd>2) {
	for (i=0;i<bnd->f->nbnd;i++) {
	  b=bnd->f->bnd[i];
	  if (b!=bnd&&b->bt==BT_DOUBLE&&(b->f->nbnd==1||b->t->nbnd==1)) r=FALSE;
	}
      }
    }
    if (r) {
      if (bnd->t->nbnd>2) {
	for (i=0;i<bnd->t->nbnd;i++) {
	  b=bnd->t->bnd[i];
	  if (b!=bnd&&b->bt==BT_DOUBLE&&(b->f->nbnd==1||b->t->nbnd==1)) r=FALSE;
	}
      }
    }
  }
  return(r);
}
/*-------------------- follow_arom_cand --------------------*/

static int follow_arom_cand(int level,Atom *a[6],Bond *b[6])
{
  int i,k,r;

  if (level==5) {
    b[5]=find_bond(a[0],a[5]);
    if (mvld(b[5])) {
      r=(b[5]->bt!=BT_SINGLE)||(b[0]->bt!=BT_SINGLE &&b[4]->bt!=BT_SINGLE);
      r= r&& arom_cand_ok(b[5]);
    }else r=FALSE;
  }else{
    for (r=FALSE,i=0;!r&&i<a[level]->nbnd;i++) {
      if ((b[level]=a[level]->bnd[i])!=b[level-1]) {
	if (b[level-1]->bt!=BT_SINGLE || b[level]->bt!=BT_SINGLE) { 
	  a[level+1]=neighbour(i,a[level]);
	  for (k=0;k<=level;k++) if (a[k]==a[level+1]) break; 
	  if (k>level) if (arom_cand_ok(b[level])) r=follow_arom_cand(level+1,a,b);
	}
      }
    }
  }
  return(r);
}
/*-------------------- arom_cand --------------------*/

static int arom_cand(Bond *bnd)
{
  int r;
  Atom *a[6];
  Bond *b[6];

  if ((bnd->flg&BND_FIX) && arom_cand_ok(bnd)) {
    a[0]=bnd->f; a[1]=bnd->t; b[0]=bnd;
    r=follow_arom_cand(1,a,b);
  }else r=0; 
  return(r);
}
/*-------------------- fix_planar_c --------------------*/

static int fix_planar_c(Atom *a)
{
  int k,not_fixed,bo,arom,nunk;
  char buf[1024];
  
  not_fixed=0;
  if (a->atp->z==6 && (a->flg&GM_CLASS)==GM_PLANAR) {
    for (bo=0,arom=FALSE,nunk=0,k=0;k<a->nbnd;k++) {
      if (a->bnd[k]->bo != 63) bo+=a->bnd[k]->bo;
      else{
	if (a->bnd[k]->bt==BT_AROMATIC) arom=TRUE;
	else nunk++;
      }
    }
    if (!arom) {
      if (nunk==0) {
	if (bo<4) {
	  not_fixed=1;
	  for (k=0;k<a->nbnd;k++) if (a->bnd[k]->flg&BND_FIX) set_bt(a->bnd[k],BT_UNKNOWN);
	  for (k=0;k<a->nbnd;k++) {
	    if (a->bnd[k]->flg&BND_FIX) {
	      if (a->bnd[k]==most_likely_dbl(a->bnd[k])) set_bt(a->bnd[k],BT_DOUBLE);
	      else set_bt(a->bnd[k],BT_SINGLE);
	      if (a->bnd[k]->bt==BT_DOUBLE) not_fixed=0;
	    }
	  }
	}else if (bo>4) not_fixed=1;
      }else{
	for (nunk=0,k=0;k<a->nbnd;k++) {
	  if (a->bnd[k]->bt==BT_UNKNOWN) {
	    if (a->bnd[k]->flg&BND_FIX) {
	      if (a->bnd[k]==most_likely_dbl(a->bnd[k])) set_bt(a->bnd[k],BT_DOUBLE);
	      else set_bt(a->bnd[k],BT_SINGLE);
	      bo+=a->bnd[k]->bo;
	    }else nunk++;
	  }
	}
	if (bo+nunk>4 || bo+2*nunk<4) not_fixed=1;
      }
    }else{
      for (arom=0,k=0;k<a->nbnd;k++) {
	if (a->bnd[k]->bt==BT_AROMATIC) arom++;
	else if (a->bnd[k]->bt==BT_UNKNOWN) {
	  if (a->bnd[k]->flg&BND_FIX) set_bt(a->bnd[k],BT_SINGLE);
	}else if (a->bnd[k]->bo>1) {
	  if (a->bnd[k]->flg&BND_FIX) set_bt(a->bnd[k],BT_SINGLE);
	  else not_fixed=1;
	}
      }
      if (arom<2) not_fixed=1;
    }
  }
  if (not_fixed) err_set(1,"planar carbon atom not fixed: %s",nodeptr_to_name(buf,(Node *)a));
  return(not_fixed);
}
/*-------------------- fix_carbon --------------------*/

static int fix_carbon(Atom *a)
{
  if ((a->flg&GM_CLASS)==GM_PLANAR) return(fix_planar_c(a));
  else return(0);
}

/*-------------------- fix_sulphur --------------------*/

static int fix_sulphur(Atom *a)
{
  int k,not_fixed,bo,ox;
  float bm,bx;
  Atom *b;

  not_fixed=0;

  if (a->nbnd==3 && (a->flg&GM_CLASS)!=GM_PLANAR) {
    for (bo=0,k=0;k<a->nbnd;k++) {
      if (a->bnd[k]->bt==BT_UNKNOWN) set_bt(a->bnd[k],BT_SINGLE);
      bo+=a->bnd[k]->bo;
    }
    while (bo<4) {
      for (ox=-1,bm=-4,k=0;k<a->nbnd;k++) {
	b=neighbour(k,a);
	if (b->atp->z==8 && b->nbnd==1 && a->bnd[k]->bo==1) {
	  bx=PiBo(a->bnd[k]);
	  if (bx>bm) {
	    bm=bx;
	    ox=k;
	  }
	}
      }
      if (ox>-1) set_bt(a->bnd[ox],BT_DOUBLE);
      else break;
      bo++;
    }
    if (bo<4) not_fixed=1; 
  }else if (a->nbnd==4) {
    for (bo=0,k=0;k<a->nbnd;k++) {
     if (a->bnd[k]->bt==BT_UNKNOWN) set_bt(a->bnd[k],BT_SINGLE);
      bo+=a->bnd[k]->bo;
    }
    while (bo<6) {
      for (ox=-1,bm=-4,k=0;k<a->nbnd;k++) {
	b=neighbour(k,a);
	if (b->atp->z==8 && b->nbnd==1 && a->bnd[k]->bo==1) {
	  bx=PiBo(a->bnd[k]);
	  if (bx>bm) {
	    bm=bx;
	    ox=k;
	  }
	}
      }
      if (ox>-1) set_bt(a->bnd[ox],BT_DOUBLE);
      else break;
      bo++;
    }
    if (bo<6) not_fixed=1;
  }
  if (not_fixed) err_set(1,"atom not fixed : %s",a->name);
  return(not_fixed);
}
/*-------------------- fix_phosphorus --------------------*/

static int fix_phosphorus(Atom *a)
{
  int k,not_fixed,bo,ox;
  float bm,bx;
  Atom *b;

  not_fixed=0;

  if (a->nbnd==4) {
    for (bo=0,k=0;k<a->nbnd;k++) {
     if (a->bnd[k]->bt==BT_UNKNOWN) set_bt(a->bnd[k],BT_SINGLE);
     bo+=a->bnd[k]->bo;
    }
    while (bo<5) {
      for (ox=-1,bm=-4,k=0;k<a->nbnd;k++) {
	b=neighbour(k,a);
	if ((b->atp->z==8 || b->atp->z==16) && b->nbnd==1 && a->bnd[k]->bo==1) { 
	  bx=PiBo(a->bnd[k]);
	  if (bx>bm) {
	    bm=bx;
	    ox=k;
	  }
	}
      }
      if (ox>-1) set_bt(a->bnd[ox],BT_DOUBLE);
      else break;
      bo++;
    }
    if (bo<5) not_fixed=1;
  }
  if (not_fixed) err_set(1,"atom not fixed : %s",a->name);
  return(not_fixed);
}
/*-------------------- atm_geom_class --------------------*/

#define GM_MIN_LINEAR 145
#define GM_MAX_PLANAR 20

int atm_geom_class(Atom *a)
{
  int class;
  Atom *b,*c,*d;
  float tau;

  if (is_atom(a)) {
    if (a->nxyz>0) {
      switch (a->nbnd) {
      case 0:
	class=GM_ISOLATED;
	break;
      case 1:
	class=GM_TERMINAL;
	break;
      case 2:
	b=neighbour(0,a); c=neighbour(1,a);
	if (b->nxyz>0&&c->nxyz>0) {
	  tau=cangled(b->x,a->x,c->x);
	  if (tau>GM_MIN_LINEAR) class=GM_LINEAR;
	  else class=GM_OTHER;
	}else class=GM_NONE;
	break;
      case 3:
	b=neighbour(0,a); c=neighbour(1,a); d=neighbour(2,a); 
	if (b->nxyz>0&&c->nxyz>0&&d->nxyz>0) {
	  tau=(fabs(ctorsd(a->x,b->x,c->x,d->x))
	       +fabs(ctorsd(a->x,c->x,d->x,b->x))
	       +fabs(ctorsd(a->x,d->x,b->x,c->x)))/3.0;
	  if (tau<GM_MAX_PLANAR) class=GM_PLANAR;
	  else class=GM_OTHER;
	}else class=GM_NONE;
	break;
      default:
	class=GM_HIGHER;
	break;
      }
    }else class=GM_NONE;
  }else class=GM_NONE;
  return(class);
}    
/*-------------------- canonical_bond --------------------*/

static void canonical_bond(Bond *b)
{
  /* reorder (from,to) such that 'from' has higher atomic number, or higher
     connectivity than 'to'. Assign geometry class to atoms, and estimated pi
     bond order to bond.
  */

  Atom *F,*T;

  if (b->f->atp->z > b->t->atp->z) {F=b->f; T=b->t;}
  else if (b->f->atp->z == b->t->atp->z) {
    if (b->f->nbnd > b->t->nbnd) {F=b->f; T=b->t;}
    else {F=b->t; T=b->f;}
  }else {F=b->t; T=b->f;}
  b->f=F; b->t=T;
  if (!(F->flg&GM_CLASS)) F->flg|=atm_geom_class(F);
  if (!(T->flg&GM_CLASS)) T->flg|=atm_geom_class(T);
  b->bo36=128+est_pi_bondorder(b)*36;
}
/*-------------------- set_bt_automatic --------------------*/

#define BO_MIN_DOUBLE  0.3
#define BO_SINGLE      0.725
#define BO_MAX_SINGLE  0.9
#define BO_MIN_TRIPLE  1.35
#define BO_MAX_DOUBLE  1.7
#define BO_12SEP       0.35

void set_bt_automatic(Bond *Bnd)
{
  /* automatic bond type assignment, based on estimated pi-bondorders,
     crude second draft.                                                */

  Bond *b,*B[4],**w,NOK,OK,XT,*nok,*ok,*xt;
  float bo,bm,bx;
  Atom *F,*T,*N[4];
  int i,k,ntot,nunk,nbad,nass,fclass,tclass,nd,no,nf;

  if (!Bnd) return;

  /* set up "canonical" bond direction ( Z(from) >= Z(to), Nbnd(from) >= Nbnd(to), 
     set everything to unknown...                                                  */

  XT.chain=(Bond *)NULL; xt=&XT;            /* chain to hold "xtra" bonds... */
  for (ntot=0,b=Bnd;mvld(b);b=b->chain) {
    canonical_bond(b);
    set_bt(b,BT_UNKNOWN);
    b->flg|=BND_FIX;
    ntot++;
  }
  for (b=Bnd;mvld(b);b=b->chain) {
    F=b->f; T=b->t;
    for (k=0;k<F->nbnd;k++) {
      if (!(F->bnd[k]->flg&BND_FIX)) {
	if (!(F->bnd[k]->chain)&&xt!=F->bnd[k]) {
	  xt->chain=F->bnd[k];
	  xt=F->bnd[k];
	}
      }
    }
    for (k=0;k<T->nbnd;k++) {
      if (!(T->bnd[k]->flg&BND_FIX)) {
	if (!(T->bnd[k]->chain)&&xt!=T->bnd[k]) {
	  xt->chain=T->bnd[k];
	  xt=T->bnd[k];
	}
      }
    }
  }
  xt->chain=(Bond *)NULL;
  for (b=XT.chain;mvld(b);b=b->chain) canonical_bond(b);

  OK.chain=(Bond *)NULL; ok=&OK;
  NOK.chain=Bnd;

  /* first few rounds, do everything that is obvious...    */
  /* should probably sort by pi-bondorder at some point... */

  for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
    F=b->f; T=b->t;
    fclass=F->flg&GM_CLASS; tclass=T->flg&GM_CLASS;
    bo=PiBo(b);
    if (bo<BO_MIN_DOUBLE) set_bt(b,BT_SINGLE);
    else if (bo>BO_MAX_SINGLE&&bo<BO_MIN_TRIPLE) {
      nf=0;
      if (F->atp->z==6 && fclass!=GM_LINEAR) {
	for (k=0;k<F->nbnd;k++) 
	  if (F->bnd[k]!=b&&F->bnd[k]->bt==BT_UNKNOWN&&PiBo(F->bnd[k])>BO_SINGLE) nf++;
      }
      if (T->atp->z==6 && tclass!=GM_LINEAR) {
	for (k=0;k<T->nbnd;k++) 
	  if (T->bnd[k]!=b&&T->bnd[k]->bt==BT_UNKNOWN&&PiBo(T->bnd[k])>BO_SINGLE) nf++;
      }
      if (nf==0) set_bt(b,BT_DOUBLE);
    }
    if ((F->atp->z==8 && F->nbnd>1)||(T->atp->z==8 && T->nbnd>1)) set_bt(b,BT_SINGLE);
    if (b->bt==BT_UNKNOWN) {
      if (bo>BO_MAX_DOUBLE) {
	if (fclass==GM_LINEAR||fclass==GM_TERMINAL) {
	  if (tclass==GM_LINEAR||tclass==GM_TERMINAL) set_bt(b,BT_TRIPLE);
	}
	if (b->bt==BT_UNKNOWN) set_bt(b,BT_DOUBLE);
      }
    }
    if (b->bt==BT_UNKNOWN) {
      if (fclass==GM_HIGHER&&tclass==GM_HIGHER) set_bt(b,BT_SINGLE);
      else if (fclass==GM_HIGHER && b->f->atp->z<10) set_bt(b,BT_SINGLE);
      else if (tclass==GM_HIGHER && b->t->atp->z<10) set_bt(b,BT_SINGLE);
      else if ((F->nbnd==3&&F->atp->z<9&&fclass!=GM_PLANAR)||
	       (T->nbnd==3&&T->atp->z<9&&tclass!=GM_PLANAR)) set_bt(b,BT_SINGLE);
    }
    w=(b->bt==BT_UNKNOWN ? &nok : (nass++,&ok));
    (*w)->chain=b; *w=b;
  }
  ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;

  do {
    for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
      F=b->f; fclass=F->flg&GM_CLASS;
      T=b->t; tclass=T->flg&GM_CLASS;
      if (fclass!=GM_LINEAR&&(F->atp->z==6||F->atp->z==7)) {
	for (k=0;k<F->nbnd;k++) if (F->bnd[k]->bt==BT_DOUBLE) break;
	if (k<F->nbnd) set_bt(b,BT_SINGLE);
      }
      if (b->bt==BT_UNKNOWN && tclass!=GM_LINEAR&&(T->atp->z==6||T->atp->z==7)) {
	for (k=0;k<T->nbnd;k++) if (T->bnd[k]->bt==BT_DOUBLE) break;
	if (k<T->nbnd) set_bt(b,BT_SINGLE);
      }
      if (b->bt==BT_UNKNOWN && (fclass==GM_PLANAR||tclass==GM_PLANAR)) {
	if (fclass==GM_PLANAR) {
	  for (i=0;i<F->nbnd;i++) {
	    if (F->bnd[i]->bt==BT_DOUBLE) {
	      if((neighbour(i,F)->flg&GM_CLASS)==GM_TERMINAL) break;
	    }
	  }
	  if (i<F->nbnd) set_bt(b,BT_SINGLE);
	}
	if (tclass==GM_PLANAR && b->bt==BT_UNKNOWN) {
	  for (i=0;i<T->nbnd;i++) {
	    if (T->bnd[i]->bt==BT_DOUBLE) {
	      if((neighbour(i,T)->flg&GM_CLASS)==GM_TERMINAL) break;
	    }
	  }
	  if (i<T->nbnd) set_bt(b,BT_SINGLE);
	}
      }
      if (b->bt==BT_UNKNOWN && fclass==GM_TERMINAL && tclass==GM_PLANAR) {
	if (F->atp->z==8 && T->atp->z==6) {
	  for (k=0,i=0;i<T->nbnd;i++) {
	    if (T->bnd[i]==b) continue;
	    B[k]=T->bnd[i];
	      N[k++]=neighbour(i,T);
	  }
	  if (k==2) {
	    if (B[0]->bt==BT_DOUBLE || B[1]->bt==BT_DOUBLE) set_bt(b,BT_SINGLE);
	    else if (B[0]->bt==BT_SINGLE && B[1]->bt==BT_SINGLE) set_bt(b,BT_DOUBLE);
	    else if (N[0]->atp->z==8 && B[0]->bt==BT_UNKNOWN) set_bt(b,BT_DOUBLE);
	    else if (N[1]->atp->z==8 && B[1]->bt==BT_UNKNOWN) set_bt(b,BT_DOUBLE);
	    else if (N[0]->atp->z==7 && B[0]->bt!=BT_DOUBLE) set_bt(b,BT_DOUBLE);
	    else if (N[1]->atp->z==7 && B[1]->bt==BT_DOUBLE) set_bt(b,BT_DOUBLE);
	  }
	}
      }
      w=(b->bt==BT_UNKNOWN ? &nok : (nass++,&ok));
      (*w)->chain=b; *w=b;
    }
    ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;
  } while (nass>0);

  /* assign aromatic bonds... but should deal with long exocyclic C=O bonds first. */

  for (b=OK.chain;mvld(b);b=b->chain) if (arom_cand(b)) set_bt(b,BT_AROMATIC);
  for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
    if (arom_cand(b)) {
      set_bt(b,BT_AROMATIC);
      ok->chain=b; ok=b;
    }else{
      nok->chain=b; nok=b;
    }
  }
  ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;

  /* fix some double bonds around planar Carbon atoms */

  do {
    for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
      F=b->f; T=b->t;	bo=PiBo(b);
      fclass=F->flg&GM_CLASS; tclass=T->flg&GM_CLASS;
      if (fclass!=GM_LINEAR&&(F->atp->z==6||F->atp->z==7)) {
	for (k=0;k<F->nbnd;k++) if (F->bnd[k]->bt==BT_DOUBLE) break;
	if (k<F->nbnd) set_bt(b,BT_SINGLE);
      }
      if (b->bt==BT_UNKNOWN && tclass!=GM_LINEAR&&(T->atp->z==6||T->atp->z==7)) {
	for (k=0;k<T->nbnd;k++) if (T->bnd[k]->bt==BT_DOUBLE) break;
	if (k<T->nbnd) set_bt(b,BT_SINGLE);
      }
      if (b->bt==BT_UNKNOWN && F->atp->z==6 && fclass==GM_PLANAR) {
	for (bm=10,bx=-10,k=0;k<F->nbnd;k++) {
	  if (F->bnd[k]!=b && F->bnd[k]->bt==BT_UNKNOWN) {
	    bo=PiBo(F->bnd[k]);
	    bm=(bm > bo ? bo : bm);
	    bx=(bx < bo ? bo : bx);
	  }
	}
	if (bm<=bx) {
	  bo=PiBo(b);
	  if (bo>bx+BO_12SEP) set_bt(b,BT_DOUBLE);
	  else if (bo<bm-BO_12SEP) set_bt(b,BT_SINGLE);
	}
      }
      if (b->bt==BT_UNKNOWN || (T->atp->z==6 && tclass==GM_PLANAR)) {
	for (bm=10,bx=-10,k=0;k<T->nbnd;k++) {
	  if (T->bnd[k]!=b && T->bnd[k]->bt==BT_UNKNOWN) {
	    bo=PiBo(T->bnd[k]);
	    bm=(bm > bo ? bo : bm);
	    bx=(bx < bo ? bo : bx);
	  }
	}
	bo=PiBo(b);
	if (bm<=bx) {
	  if (bo>bx+BO_12SEP) set_bt(b,BT_DOUBLE);
	  else if (bo<bm-BO_12SEP) set_bt(b,BT_SINGLE);
	}
      }
      if (b->bt==BT_UNKNOWN && tclass==GM_OTHER && fclass==GM_OTHER) {
	if (F->nbnd==2 && T->nbnd==2) {
	  if ((F->atp->z==6||F->atp->z==7) && (T->atp->z==6||T->atp->z==7)) {
	    if (b==most_likely_dbl(b)) {
	      for (k=0;k<F->nbnd;k++) if (neighbour(k,F)!=T) N[0]=neighbour(k,F);
	      for (k=0;k<T->nbnd;k++) if (neighbour(k,T)!=F) N[3]=neighbour(k,T);
	      bx=ctorsd(N[0]->x,F->x,T->x,N[3]->x);
	      if (bx<0) bx=-bx;
	      if (bx>90) bx=180-bx;
	      if (bx<20) set_bt(b,BT_DOUBLE);
	      else set_bt(b,BT_SINGLE);
	    }
	  }
	}
      }
      w=(b->bt==BT_UNKNOWN ? &nok : (nass++,&ok));
      (*w)->chain=b; *w=b;
    }
    ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;
  } while (nass>0);

  /* assign aromatic bonds... but should deal with long exocyclic C=O bonds first. */

  for (b=OK.chain;mvld(b);b=b->chain) if (arom_cand(b)) set_bt(b,BT_AROMATIC);
  for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
    if (arom_cand(b)) {
      set_bt(b,BT_AROMATIC);
      ok->chain=b; ok=b;
    }else{
      nok->chain=b; nok=b;
    }
  }
  ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;

  /* non-aromatic bonds connected to aromatic bonds are single... */

  for (nok=&NOK,b=nok->chain;mvld(b);b=b->chain) {
    F=b->f; fclass=F->flg&GM_CLASS;
    T=b->t; tclass=T->flg&GM_CLASS;
    if (fclass==GM_PLANAR) {
      for (k=0;k<F->nbnd;k++) if (F->bnd[k]->bt==BT_AROMATIC) break;
      if (k<F->nbnd) set_bt(b,BT_SINGLE);
    }
    if (b->bt==BT_UNKNOWN && tclass==GM_PLANAR) {
      for (k=0;k<T->nbnd;k++) if (T->bnd[k]->bt==BT_AROMATIC) break;
      if (k<T->nbnd) set_bt(b,BT_SINGLE);
    }
    w=(b->bt==BT_UNKNOWN ? &nok : &ok);
    (*w)->chain=b; *w=b;
  }
  ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;

  /* popular functional groups... */

  do {
    for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
      F=b->f; T=b->t;	bo=PiBo(b);
      fclass=F->flg&GM_CLASS; tclass=T->flg&GM_CLASS;
      if (T->atp->z<6) set_bt(b,BT_SINGLE);
      else if (F->atp->z==6) {                 /* central atom is Carbon */
	if (T->nbnd==1) {
	  if (fclass==GM_LINEAR) {
	    for (k=0;k<F->nbnd;k++) if (F->bnd[k]!=b) break;
	    if (k<F->nbnd) {
	      if (F->bnd[k]->bt==BT_UNKNOWN) {
	      }else if (F->bnd[k]->bt==BT_SINGLE) set_bt(b,BT_TRIPLE);
	      else if (F->bnd[k]->bt==BT_DOUBLE) set_bt(b,BT_DOUBLE);
	      else if (F->bnd[k]->bt==BT_TRIPLE) set_bt(b,BT_SINGLE);
	    }
	  }else if (fclass==GM_PLANAR) {
	    if (b==most_likely_dbl(b)) set_bt(b,BT_DOUBLE);
	  }else{
	    if (b==most_likely_dbl(b)) {
	      if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	      else set_bt(b,BT_SINGLE);
	    }else set_bt(b,BT_SINGLE);
	  }
	}else{
	  if (b==most_likely_dbl(b)) {
	    if (fclass==GM_PLANAR||tclass==GM_PLANAR) set_bt(b,BT_DOUBLE);
	    else if (bo > BO_SINGLE) set_bt(b,BT_DOUBLE);
	    else set_bt(b,BT_SINGLE);
	  }else{
	    set_bt(b,BT_SINGLE);
	  }
	}
      }else if (F->atp->z==7) {                /* central atom is nitrogen */
	if (tclass==GM_LINEAR) {
	  set_bt(b,BT_TRIPLE);                 /* ??? */
	}else{
	  if (b==most_likely_dbl(b)) {
	    if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	    else set_bt(b,BT_SINGLE);         /* ??? guanidinium */
	  }else set_bt(b,BT_SINGLE);
	}
      }else if (F->atp->z==8) {                /* oxygen   */
	if (fclass==GM_TERMINAL) {
	  if (T->atp->z==7) {
	    if (tclass==GM_PLANAR) {           /* nitro ? */
	      for (no=0,nd=0,k=0;k<T->nbnd;k++) {
		if (T->bnd[k]!=b) {
		  if (T->bnd[k]->bt==BT_DOUBLE) nd++;
		  if (neighbour(k,T)->atp->z==8) no++;
		}
	      }
	      if (nd<2 && no >0) set_bt(b,BT_DOUBLE);
	    }
	  }else{
	    if (b==most_likely_dbl(b)) {
	      if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	      else set_bt(b,BT_SINGLE);
	    }else set_bt(b,BT_SINGLE);
	  }
	}else set_bt(b,BT_SINGLE);
      }else if (F->atp->z==15) {
	for (bm=-3,k=0;k<F->nbnd;k++) {
	  if (F->bnd[k]!=b) {
	    if (F->bnd[k]->bt==BT_SINGLE) bx=-3;
	    else if (F->bnd[k]->bt!=BT_UNKNOWN) bx=+3;
	    else bx=PiBo(F->bnd[k]);
	    if (bx>bm) bm=bx;
	  }
	  if (bo<bm) set_bt(b,BT_SINGLE);
	  else if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	  else set_bt(b,BT_SINGLE);
	}
      }else if (F->atp->z==16) {               /* sulphur  */
	if (fclass==GM_TERMINAL) {
	  for (bm=-3,k=0;k<T->nbnd;k++) {
	    if (T->bnd[k]!=b) {
	      if (T->bnd[k]->bt==BT_SINGLE) bx=-3;
	      else if (T->bnd[k]->bt!=BT_UNKNOWN) bx=+3;
	      else bx=PiBo(T->bnd[k]);
	      if (bx>bm) bm=bx;
	    }
	  }
	  if (bo<bm) set_bt(b,BT_SINGLE);
	  else if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	  else set_bt(b,BT_SINGLE);
	}else if (fclass==GM_HIGHER) {
	  if (F->nbnd==4) {
	    for (nd=0,bm=-3,k=0;k<F->nbnd;k++) {
	      if (F->bnd[k]!=b) {
		if (F->bnd[k]->bt==BT_SINGLE) bx=-3;
		else if (F->bnd[k]->bt==BT_DOUBLE) {nd++; bx=-3;}
		else if (F->bnd[k]->bt!=BT_UNKNOWN) bx=+3;
		else bx=PiBo(F->bnd[k]);
		if (bx>bm) bm=bx;
	      }
	    }
	    if (nd<2) {
	      if (bo>=bm) {
		if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	      }
	    }else set_bt(b,BT_SINGLE);
	  }
	}else {
	  if (b==most_likely_dbl(b)) {
	    if (bo>BO_SINGLE) set_bt(b,BT_DOUBLE);
	    else set_bt(b,BT_SINGLE);
	  }else set_bt(b,BT_SINGLE);
	}
      }		
      w=(b->bt==BT_UNKNOWN ? &nok : (nass++,&ok));
      (*w)->chain=b; *w=b;
    }
    ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;
  }while (nass>0);

  /* assign aromatic bonds... but should deal with long exocyclic C=O bonds first. */

  for (b=OK.chain;mvld(b);b=b->chain) if (arom_cand(b)) set_bt(b,BT_AROMATIC);
  for (nok=&NOK,nass=0,b=nok->chain;mvld(b);b=b->chain) {
    if (arom_cand(b)) {
      set_bt(b,BT_AROMATIC);
      ok->chain=b; ok=b;
    }else{
      nok->chain=b; nok=b;
    }
  }
  ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;

  /* non-aromatic bonds connected to aromatic bonds are single... */

  for (nok=&NOK,b=nok->chain;mvld(b);b=b->chain) {
    F=b->f; fclass=F->flg&GM_CLASS;
    T=b->t; tclass=T->flg&GM_CLASS;
    if (fclass==GM_PLANAR) {
      for (k=0;k<F->nbnd;k++) if (F->bnd[k]->bt==BT_AROMATIC) break;
      if (k<F->nbnd) set_bt(b,BT_SINGLE);
    }
    if (b->bt==BT_UNKNOWN && tclass==GM_PLANAR) {
      for (k=0;k<T->nbnd;k++) if (T->bnd[k]->bt==BT_AROMATIC) break;
      if (k<T->nbnd) set_bt(b,BT_SINGLE);
    }
    w=(b->bt==BT_UNKNOWN ? &nok : &ok);
    (*w)->chain=b; *w=b;
  }
  ok->chain=(Bond *)NULL; nok->chain=(Bond *)NULL;

  /* merge the two bond chains and put Bnd at head of merged chain */

  ok->chain=NOK.chain; if (mvld(ok->chain)) ok=nok;
  for (b=OK.chain;mvld(b);b=b->chain) {
    if (b->chain==Bnd) {
      b->chain=(Bond *)NULL;
      ok->chain=OK.chain;
    }
  }

  /* fix some well known problems, assign amide bonds... */

  for (nbad=0,b=Bnd;mvld(b);b=b->chain) {
    if (!(b->f->flg&ATM_FIX)) {
      switch (b->f->atp->z) {
      case 6:
	nbad+=fix_carbon(b->f);
	break;
      case 16:
	nbad+=fix_sulphur(b->f);
	break;
      case 15:
	nbad+=fix_phosphorus(b->f);
	break;
      default:
	break;
      }
      b->f->flg|=ATM_FIX;
    }
    if (!(b->t->flg&ATM_FIX)) {
      switch (b->t->atp->z) {
      case 6:
	nbad+=fix_carbon(b->t);
	break;
      case 16:
	nbad+=fix_sulphur(b->t);
	break;
      case 15:
	nbad+=fix_phosphorus(b->t);
	break;
      default:
	break;
      }
      b->t->flg|=ATM_FIX;
    }
 
    /* should do: review single bonds with high pi-bondorder (e.g. in imidazoles) */

    if (b->f->atp->z==7 && b->t->atp->z==6 && b->bo==1 ) {    /* assign amide bonds */
      for (k=0;k<b->f->nbnd;k++) if (b->f->bnd[k]->bo!=1) break;
      if (k==b->f->nbnd) {
	for (k=0;k<b->t->nbnd;k++) if (b->t->bnd[k]->bo==2 && b->t->bnd[k]->f->atp->z==8) break;
	if (k<b->t->nbnd) set_bt(b,BT_AMIDE);
      } 
    }
  }

  for (b=Bnd;mvld(b);b=b->chain) if (arom_cand(b)) set_bt(b,BT_AROMATIC);

  if (nbad>0) err_set(1,"set_bt_automatic: %d strange atom%s",nbad,(nbad>1 ? "s." : "."));

  /* clear atom class flags and bond flags */

  for (nunk=0,b=Bnd;mvld(b);b=b->chain) {
    if (b->bt==BT_UNKNOWN) {
      nunk++;
      err_set(1,"bond not fixed: %s --- %s",b->f->name,b->t->name);
    }
    b->f->flg&=~(GM_CLASS|ATM_FIX);    
    b->t->flg&=~(GM_CLASS|ATM_FIX);
    b->flg&=~BND_FIX;
  }    
  if (nunk>0) err_set(1,"set_bt_automatic: %d strange bond%s",nunk,(nunk>1 ? "s." : "."));

  for (b=XT.chain;mvld(b);b=xt) {     /* also clear atom flags on "xtra" bonds */
    b->f->flg&=~(GM_CLASS|ATM_FIX);    
    b->t->flg&=~(GM_CLASS|ATM_FIX);
    xt=b->chain;
    b->chain=(Bond *)NULL;
  }
}
/*-------------------- default_bndlength --------------------*/

float default_bndlength(Atom *a,Atom *b,int bt)
{
  /* return default bond length depending on atoms a,b and bond type */

  float r,bo;
  int i,j;

  r=1.5;

  if (bt==BT_SINGLE) bo=0.0;             /* pi-bond order */
  else if (bt==BT_DOUBLE) bo=1.0;
  else if (bt==BT_TRIPLE) bo=2.0;
  else if (bt==BT_AROMATIC) bo=0.667;
  else if (bt==BT_AMIDE)  bo=0.52;
  else bo=0.0;

  if (mvld(a->atp)&&mvld(b->atp)) {
    for (i=0;i<MAX_EL;i++) if(a->atp->z==Z[i]) break;
    for (j=0;j<MAX_EL;j++) if(b->atp->z==Z[j]) break;
    if (i<MAX_EL&&j<MAX_EL) r=R0[i][j]-bo*BD[i][j];
    if (r<=0) r=1.5;
  }
  return(r);
}

/*========================================================================*/

#define DONE      0
#define ADVANCE   1
#define BACKTRACK 2
#define BRANCH    3
#define FIND      4
#define CHECK     5

#define BO_UNK   63
#define BO_BAD    0

static int n_pi_aromatic(Nodechain *A)
{
  int n,nd,i,z;
  Atom *a;

  for (n=0,a=(Atom *)A->s;mvld(a);a=(Atom *)a->chain) {
    z=(mvld(a->atp) ? a->atp->z : 0);
    switch (z) {          
    case 6:               /* C */
      n++;
      break;
    case 8:               /* O */
      n+=2;
      break;
    case 7:               /* N */
      for (nd=0,i=0;i<a->nbnd;i++) {
	if (a->bnd[i]->bo==2&&a->bnd[i]->flg&BND_FIX) nd++;
      }
      if (nd%2==0) n+=2;
      else n+=1;
      break;
    case 15:              /* P */
    case 16:              /* S */
      for (nd=0,i=0;i<a->nbnd;i++) {
	if (a->bnd[i]->bo==2&&a->bnd[i]->flg&BND_FIX) nd++;
      }
      if (nd%2==0) n+=2;
      else n+=1;
      break;
    default:
      break;
    }
  }
  return n;
}

static int guess_aromatic_bo(Atom *a,Bond *bnd)
{
  int i,bo,n1,n2,n2f,no,nu,z;

  z=(mvld(a->atp) ? a->atp->z : 0);
  n1=0; n2=0; no=0; nu=0; n2f=0;
  for (i=0;i<a->nbnd;i++) {
    if (bnd==a->bnd[i]) continue;
    if (a->bnd[i]->bo==BO_UNK) nu++;
    else if (a->bnd[i]->bo==1) n1++;
    else if (a->bnd[i]->bo==2) {
      n2++;
      if (a->bnd[i]->flg&BND_FIX) n2f++;
    }else no++;
  }
  switch (z) {          
  case 6:               /* C */
    if (no!=0) bo=BO_UNK;
    else if (n2>0) bo=1;
    else if (nu==0) bo=2;
    else bo=BO_UNK;
    break;
  case 7:               /* N */
    if (no!=0) bo=BO_UNK;
    else if (n2>0) bo=1;
    else {
      if ((a->lig==3||a->nbnd>2)&&a->chrg==0) bo=1;
      else bo=BO_UNK;
    }
    break;
  case 8:               /* O */
    bo=1;
    break;
  case 15:              /* P */
  case 16:              /* S */
    if (n2f>0) bo=1;
    else bo=BO_UNK;
    break;
  default:
    bo=BO_UNK;
    break;
  }
  return(bo);
}

static int aromatic_bo(Bond *bnd)
{
  /* return formal bond order of aromatic bond, BO_UNK if order
     can not be determined, BO_BAD if any order leads to inconsistency */

  int bo,bf,bt;

  bo=BO_UNK;
  if (mvld(bnd)) {
    bf=guess_aromatic_bo(bnd->f,bnd);
    bt=guess_aromatic_bo(bnd->t,bnd);
    bo=(bf==bt ? bf : ( bf==BO_UNK ? bt : ( bt==BO_UNK ? bf : BO_BAD )));
  }else bo=BO_UNK;
  return(bo);
}

static int localize_first_try(Bond *bnd)
{
  int bo,z;
  Atom *a;

  bo=2;        /* first try double bond... */

  a=bnd->f;    /* except if f=Nitrogen, 3 bonds, neutral or unknown charge */
  z=(mvld(a->atp) ? a->atp->z : 0);
  if (z==7&&(a->lig==3||a->nbnd==3)&&a->chrg!=1) bo=1;

  a=bnd->t;    /* or if t=Nitrogen, 3 bonds, neutral or unknown charge */ 
  z=(mvld(a->atp) ? a->atp->z : 0);
  if (z==7&&(a->lig==3||a->nbnd==3)&&a->chrg!=1) bo=1;

  return(bo);
}

int localize_aromatic_bonds(Node *mol)
{
  /* determine bond order of aromatic bonds in mol,
     return number of bonds for which order could not bee assigned */

  Atom *a;
  Bond *bnd;
  int r=0;

  ForAtoms(a,mol) {
    bnd=aromatic_cluster(a,BND_FIX);
    if (mvld(bnd)) r+=localize_bond_cluster(bnd);
  }

  ForAtoms(a,mol) clear_bond_cluster(a,BND_FIX);
  return(r);
}

Bond *aromatic_cluster(Atom *a,int flag)
{
  Bond *bnd,*b,*e;
  int i;

  bnd=(Bond *)NULL;
  if (is_atom(a)) {
    for (i=0;i<a->nbnd;i++) {
      if (a->bnd[i]->bt==BT_AROMATIC&&!(a->bnd[i]->flg&flag)) {
	bnd=a->bnd[i];
	bnd->flg|=flag;
	break;
      }
    }
    for (e=bnd,b=bnd;mvld(b);b=b->chain) {
      a=b->f;
      for (i=0;i<a->nbnd;i++) {
	if (a->bnd[i]->bt==BT_AROMATIC&&!(a->bnd[i]->flg&flag)) {
	  e->chain=a->bnd[i];
	  e=e->chain;
	  e->flg|=flag;
	}
      }
      a=b->t;
      for (i=0;i<a->nbnd;i++) {
	if (a->bnd[i]->bt==BT_AROMATIC&&!(a->bnd[i]->flg&flag)) {
	  e->chain=a->bnd[i];
	  e=e->chain;
	  e->flg|=flag;
	}
      }
    }
  }
  return(bnd);
}

int localize_bond_cluster(Bond *bnd)
{
  /* determine bond order of aromatic bonds in aromatic bond cluster,
     return number of bonds for which order could not bee assigned */

  Atom *a;
  Nodechain A;
  Bond *b,**bass=(Bond **)NULL;
  int i,k,newo,npi,state,nass,nunk,*alt=(int *)NULL;

  /* count bonds with unknown bond order, and assign temp. workspace */

  init_node_chain(&A);
  for (nunk=0,b=bnd;mvld(b);b=b->chain) {
    add_to_node_chain((Node *)b->f,&A);
    add_to_node_chain((Node *)b->t,&A);
    if (b->bo==BO_UNK) nunk++;
  }
  nass=0;
  if (nunk>0) {
    alt=(int *)malloc(nunk*sizeof(int));
    bass=(Bond **)malloc(nunk*sizeof(Bond *));
    if (alt&&bass) state=FIND;
    else state=DONE;
  }else state=DONE;

  /* use backtrack algorithm to assign bond orders */

  while (state!=DONE) {
    switch (state) {
    case FIND:
      state=CHECK;
      for (b=bnd;mvld(b);b=b->chain) {
	if (b->bo==BO_UNK) {
	  newo=aromatic_bo(b);
	  if (newo==BO_UNK) {
	    if (state==CHECK) {
	      bass[nass]=b;
	      state=BRANCH;
	    }
	  }else if (newo==BO_BAD){
	    state=BACKTRACK;
	    break;
	  }else{
	    bass[nass]=b;
	    alt[nass]=BO_BAD;
	    b->bo=newo;
	    nass++;
	    state=ADVANCE;
	    break;
	  }
	}
      }
      break;
    case BRANCH:
      bass[nass]->bo=localize_first_try(bass[nass]);
      alt[nass]=3-bass[nass]->bo;
      nass++;
      state=ADVANCE;
      break;
    case ADVANCE:
      for (k=nass-1;k<nass&&state==ADVANCE;k++) {
	a=bass[k]->f;
	for (i=0;i<a->nbnd&&state==ADVANCE;i++) {
	  b=a->bnd[i];
	  if (b->flg&BND_FIX&&b->bo==BO_UNK) {
	    newo=aromatic_bo(b);
	    if (newo==BO_BAD) {
	      state=BACKTRACK;
	      break;
	    }else if (newo!=BO_UNK) {
	      b->bo=newo;
	      bass[nass]=b;
	      alt[nass]=BO_BAD;
	      nass++;
	      if (nass==nunk) state=CHECK;
	    }
	  }
	}
	if (state!=ADVANCE) break;
	a=bass[k]->t;
	for (i=0;i<a->nbnd&&state==ADVANCE;i++) {
	  b=a->bnd[i];
	  if (b->flg&BND_FIX&&b->bo==BO_UNK) {
	    newo=aromatic_bo(b);
	    if (newo==BO_BAD) {
	      state=BACKTRACK;
	      break;
	    }else if (newo!=BO_UNK) {
	      b->bo=newo;
	      bass[nass]=b;
	      alt[nass]=BO_BAD;
	      nass++;
	      if (nass==nunk) state=CHECK;
	    }
	  }
	}
      }
      if (state==ADVANCE) state=FIND;
      break;
    case BACKTRACK:
      while (nass>0&&state==BACKTRACK) {
	nass--;
	if (alt[nass]==BO_BAD) bass[nass]->bo=BO_UNK;
	else{
	  bass[nass]->bo=alt[nass];
	  alt[nass]=BO_BAD;
	  nass++;
	  state=ADVANCE;
	}
      }
      if (state==BACKTRACK) state=DONE;
      break;
    case CHECK:
      npi=n_pi_aromatic(&A);
      if (npi%4!=2||npi<6) state=BACKTRACK;
      else state=DONE;
      break;
    }
  }

  /* clean up */

  if (bass) free(bass);
  if (alt) free(alt);
  clear_node_chain(&A);

  return(nunk-nass);
}

void clear_bond_cluster(Atom *a,int flag)
{
  Bond *bnd,*b;
  int i;

  for (i=0;i<a->nbnd;i++) {
    if (a->bnd[i]->flg&flag) {
      bnd=a->bnd[i];
      for (b=bnd;mvld(b);b=bnd) {
	bnd=b->chain;
	b->chain=(Bond *)NULL;
	b->flg&=~flag;
      }
    }
  }
}
