#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <sys/param.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "gcalc.h"
#include "bhtree.h"
#include "graphics.h"
#include "parmset.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define FLG1 (1<<1)
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

static int prmdef=FALSE;
static char *prmfile=(char *)NULL;
static char *modfile=(char *)NULL;
static int  format=PRMTOP_FMT_NEW;

void cmd_amber(void)
{
  Token *key=(Token *)NULL;
  char *topl="frcmod parameters '' read write '' missing sloppy '' format '' end";
  char *fn,cwd[MAXPATHLEN];
  int mode;

  if (!prmdef) {
    if (!prmfile) {
      fn=getenv("WNP_AMBER_PARAM");
      if (fn&&!streq(fn,"-none")) prmfile=strdup(fn);
    }
    if (!modfile) {
      fn=getenv("WNP_AMBER_FRCMOD");
      if (fn&&!streq(fn,"-none")) modfile=strdup(fn);
    }
    prmdef=TRUE;
  }

  if (!prmfile&&!modfile) err_set(1,"Note: no parameter files defined.");

  for (;;) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("Amber> ",";",topl);
    if (key->status==TOK_OK) {
      if (key_is("parameters")) {
	key=free_Token(key);
	if (prmfile) default_token("%s",prmfile);
	*cwd='\0';getCWD(cwd,sizeof(cwd));
	key=Get_File_M(cwd,(char *)NULL,"parameter file: ",";","-none");
	if (key->status==TOK_OK) {
	  if (prmfile) free(prmfile);
	  prmfile=strdup(((UCpipe *)key->ptr)->filename);
	  UCpclose((UCpipe *)key->ptr);
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"-none");
	  if (key->status==TOK_OK) {
	    if (prmfile) free(prmfile);
	    prmfile=(char *)NULL;
	  }else err_set(1,"can't read parameter file %s",key->text);
	}
      }else if (key_is("frcmod")) {
	key=free_Token(key);
	if (modfile) default_token("%s",modfile);
	*cwd='\0';getCWD(cwd,sizeof(cwd));
	key=Get_File_M(cwd,(char *)NULL,"force field modifications file: ",";","-none");
	if (key->status==TOK_OK) {
	  if (modfile) free(modfile);
	  modfile=strdup(((UCpipe *)key->ptr)->filename);
	  UCpclose((UCpipe *)key->ptr);
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"-none");
	  if (key->status==TOK_OK) {
	    if (modfile) free(modfile);
	    modfile=(char *)NULL;
	  }else err_set(1,"can't read modifications file %s",key->text);
	}
      }else if (key_is("read")) {
	key=free_Token(key);
	key=Get_Kwd("read what? ",";","crd");
	if (key->status==TOK_OK) {
	  if (key_is("crd")) cmd_read_amber_crd();
	}else if (key->status==TOK_NOK) err_set(1,"unrecognized Amber file type: %s",key->text);
      }else if (key_is("write")) {
	key=free_Token(key);
	key=Get_Kwd("write what? ",";","crd prep prmtop");
	if (key->status==TOK_OK) {
	  if (key_is("crd")) cmd_write_amber_crd();
	  else if (key_is("prmtop")) cmd_write_amber_prmtop();
	  else if (key_is("prep")) cmd_write_amber_prep();
	}else if (key->status==TOK_NOK) err_set(1,"unrecognized Amber file type: %s",key->text);
      }else if (key_is("sloppy")) {
	key=free_Token(key);
	mode=checkParmSetMode(PRM_SLOPPY_TEST);
	default_token("%s",mode==PRM_SLOPPY_ALWAYS?"always":mode==PRM_SLOPPY_NEVER?"never":"once");
	key=Get_Kwd("write sloppy parmtop files: ",";","always never once");
	if (key->status==TOK_OK) {
	  checkParmSetMode(key->n==0?PRM_SLOPPY_ALWAYS:key->n==1?PRM_SLOPPY_NEVER:PRM_SLOPPY_ONCE);
	}else if (key->status==TOK_NOK) err_set(1,"invalid key: %s",key->text);
      }else if (key_is("missing")) {
	mode=missingParmSetMode(PRM_MISSING_TEST);
	default_token("%s",mode==PRM_MISSING_FIXED?"fixed":mode==PRM_MISSING_ZERO?"zero":"floppy");
	key=Get_Kwd("treatment of missing parameters in sloppy files: ",";","fixed floppy zero");
	if (key->status==TOK_OK) {
	  missingParmSetMode(key->n==0?PRM_MISSING_FIXED:key->n==1?PRM_MISSING_FLOPPY:PRM_MISSING_ZERO);
	}else if (key->status==TOK_NOK) err_set(1,"invalid key: %s",key->text);
      }else if (key_is("format")) {
	key=free_Token(key);
	default_token(format==PRMTOP_FMT_NEW?"new":"old");
	key=Get_Kwd("parmtop format: ",";","new old");
	if (key->status==TOK_OK) {
	  if (key_is("new")) format=PRMTOP_FMT_NEW;
	  else if (key_is("old")) format=PRMTOP_FMT_OLD;
	}else if (key->status==TOK_NOK) err_set(1,"unknown prmtop format: %s\n",key->text);
      }else if (key_is("end")) break;
    }else if (key->status==TOK_NOK) err_set(1,"bad Amber option: %s\n",key->text);
    if (end_parse()!=TOK_OK) err_set(1,"extra characters at end of command ignored.");
    start_parse();
  }
  free_Token(key);
}
/*------------------------------------------------------------------------*/

void cmd_write_amber_crd(void)
{
  Token *key=(Token *)NULL;
  char *fname=(char *)NULL,*menu=(char *)NULL;
  char buf[4096];
  Nodechain *c;
  Node *root;
  Atom *a;

  key=Get_String("File: ",";");
  if (key->status==TOK_OK) {
    fname=Strdup(key->text);
    key=free_Token(key);
    menu=Make_TopMolNameList("",DefaultNode);
    key=Get_Mols_M("molecule: ",";",menu); free_menu(menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n>1) {
	err_set(1,"molecule selection is ambiguous.");
	key->status=TOK_NOK;
      }else if (c->n<1) {
	err_set(1,"no molecule selected.");
	key->status=TOK_OK;
      }else{
	root=c->s;
	clear_node_chain(c);
	ForAtoms(a,root) a->flg|=FLG1;
	if (write_amber_crd(fname,root)) 
	  err_set(1,"failed to write Amber coordinates file %s for %s",
		  fname,nodeptr_to_name(buf,root));
	else prtGraphics("Amber coordinates file written to %s\n",fname);
	ForAtoms(a,root) a->flg&=~FLG1;
      }
      clear_node_chain(c); free_node_chain(c);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad filename: %s",key->text);

  free_Token(key);
  if (mvld(fname)) free(fname);
  if (mvld(menu)) free(menu);
}
/*------------------------------------------------------------------------*/

void cmd_write_amber_prmtop(void)
{
  Token *key=(Token *)NULL;
  char *fname=(char *)NULL,*menu=(char *)NULL;
  char buf[4096];
  Nodechain *c;
  Node *root;
  Atom *a;

  key=Get_String("File: ",";");
  if (key->status==TOK_OK) {
    fname=Strdup(key->text);
    key=free_Token(key);
    menu=Make_TopMolNameList("",DefaultNode);
    key=Get_Mols_M("molecule: ",";",menu); free_menu(menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n>1) {
	err_set(1,"molecule selection is ambiguous.");
	key->status=TOK_NOK;
      }else if (c->n<1) {
	err_set(1,"no molecule selected.");
	key->status=TOK_OK;
      }else{
	root=c->s;
	clear_node_chain(c);
	ForAtoms(a,root) a->flg|=FLG1;
	if (write_amber_prmtop(fname,root,format)) 
	  err_set(1,"failed to write Amber parameter file %s for %s",
		  fname,nodeptr_to_name(buf,root));
	else prtGraphics("Amber parameter file written to %s\n",fname);
	ForAtoms(a,root) a->flg&=~FLG1;
      }
      clear_node_chain(c); free_node_chain(c);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad filename: %s",key->text);

  free_Token(key);
  if (mvld(fname)) free(fname);
  if (mvld(menu)) free(menu);
}
/*------------------------------------------------------------------------*/

void cmd_write_amber_prep(void)
{
  Token *key=(Token *)NULL;
  char *fname=(char *)NULL,*menu=(char *)NULL;
  char buf[4096];
  Nodechain *c;
  Node *root;
  Atom *a;

  key=Get_String("File: ",";");
  if (key->status==TOK_OK) {
    fname=Strdup(key->text);
    key=free_Token(key);
    menu=Make_TopMolNameList("",DefaultNode);
    key=Get_Mols_M("molecule: ",";",menu); free_menu(menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n>1) {
	err_set(1,"molecule selection is ambiguous.");
	key->status=TOK_NOK;
      }else if (c->n<1) {
	err_set(1,"no molecule selected.");
	key->status=TOK_OK;
      }else{
	root=c->s;
	clear_node_chain(c);
	ForAtoms(a,root) a->flg|=FLG1;
	if (write_amber_prep(fname,root)) 
	  err_set(1,"failed to write Amber prep file %s for %s",
		  fname,nodeptr_to_name(buf,root));
	else prtGraphics("Amber prep file written to %s\n",fname);
	ForAtoms(a,root) a->flg&=~FLG1;
      }
      clear_node_chain(c); free_node_chain(c);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad filename: %s",key->text);

  free_Token(key);
  if (mvld(fname)) free(fname);
  if (mvld(menu)) free(menu);
}
/*------------------------------------------------------------------------*/

void cmd_read_amber_crd(void)
{
  Token *key=(Token *)NULL;
  char *menu=(char *)NULL,buf[4096];
  Nodechain *c;
  Node *root;
  Atom *a;
  UCpipe *in;

  key=Get_File((char *)NULL,(char *)NULL,"coordinate file: ",";");
  if (key->status==TOK_OK) {
    in=(UCpipe *)key->ptr;
    key=free_Token(key);
    menu=Make_TopMolNameList("",DefaultNode);
    key=Get_Mols_M("molecule: ",";",menu); free_menu(menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n>1) {
	err_set(1,"molecule selection is ambiguous.");
	key->status=TOK_NOK;
      }else if (c->n<1) {
	err_set(1,"no molecule selected.");
	key->status=TOK_OK;
      }else{
	root=c->s;
	clear_node_chain(c); free_node_chain(c);
	ForAtoms(a,root) a->flg|=FLG1;
	c=collect_by_flags(root,FLG1);
	if (read_amber_crd(in,root)) 
	  err_set(1,"failed to read Amber coordinate file %s for %s",
		  in->filename,nodeptr_to_name(buf,root));
	ForAtoms(a,root) a->flg&=~FLG1;
	send_newx(c);
      }
      clear_node_chain(c); free_node_chain(c);
    }
    UCpclose(in);
  }else if (key->status==TOK_NOK) err_set(1,"bad filename: %s",key->text);

  free_Token(key);
  if (mvld(menu)) free(menu);
}
/*------------------------------------------------------------------------*/

int write_amber_prmtop(char *fname,Node *root,int format)
{
  int status;
  char *fn;

  if (!prmdef) {
    if (!prmfile) {
      fn=getenv("WNP_AMBER_PARAM");
      if (fn&&!streq(fn,"-none")) prmfile=strdup(fn);
    }
    if (!modfile) {
      fn=getenv("WNP_AMBER_FRCMOD");
      if (fn&&!streq(fn,"-none")) modfile=strdup(fn);
    }
    prmdef=TRUE;
  }

  if (!prmfile&&!modfile) err_set(1,"Note: no parameter files defined.");

  if (prmfile||modfile) status=write_amber_parm(fname,root,prmfile,modfile,format);
  else {
    err_set(1,"no Amber parameter files defined, use 'amber parameters ...' or");
    err_set(1,"'amber frcmod ...' command to define parameter files.");
    status=1;
  }
  err_dump();
  return(status);
}
