#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "bhtree.h"
#include "errhand.h"
#include "gcalc.h"

extern int graphix;
extern Node *DefaultNode;

#define key_is(x) (streq(key->key,x))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

#define NATOMS 2

#define DONE 0
#define MAIN 1
#define MAKE 2
#define BREAK 3
#define AUTO 4
#define TWIST 5
#define WIDTH 6
#define CYLINDER 7

#define CYL_ON 1
#define CYL_OFF 0
#define CYL_RAD 2
#define CYL_NOP 3

#define MXLIST 2048
#define MXTREE 2048

static void freeze_rotax(Nodechain *mob);
static Bond *connectUniqueH(Bond *bs,BHtree *tree,float tol);
static Bond *connectOtherH(Bond *bs,BHtree *tree,float tol);
#ifdef CLEANUP_BONDLIST
static Bond *cleanup_bondlist(Bond *chain);
#endif

static char *main_keys="automatic make break cylinder width twist";
static char *mb_keys="make break";
static char *br_keys="make break '' zero";
static char *twist_keys = "reset axis monitor ' ' "
  "180 150 120 90 60 30 15 10 5 4 3 2 1 ' ' "
  "-1 -2 -3 -4 -5 -10 -15 -30 -60 -90 -120 -150 -180";
static char *rad_men="scale 0.02 0.04 0.05 0.06 0.08 0.1 0.2 0.3 0.4 0.5";
static char *scale_men="0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 "
  "1.5 2.0 3.0 4.0 5.0 10.0";

static float twist_angle=0;    /* twist angle in no-graphics mode */
static int AutobondMode = AUTOBOND_OBJECT;

static int cmpchain(const void *a,const void *b)
{
  Atom *A,*B;
  int r;
  
  A=*(Atom **)a;
  B=*(Atom **)b;
  if (A->chain<B->chain) r=-1;
  else if (A->chain>B->chain) r=1;
  else r=0;
  return(r);
}

/*========================================================================*/
void cmd_bond(void)
{
/* Edit bondlist */

  Token *key,*atm[NATOMS];
  Nodechain *mobile;
  Bond *b,*bn,*bs,*be;
  Atom *pre,*root,*an,*am;
  int i,n,nb,lw;
  int state,reparse_atom,ngm,cyl_opt;
  float cyl_rad=0.0;

  mobile=(Nodechain *)NULL;
  state=MAIN;
  key=(Token *)NULL;

  while (state) {
    err_dump();
    if (!key) key=Get_Kwd("option: ",";",main_keys);
    if (key->status!=TOK_OK)  {
      state=DONE;
      if (key->status==TOK_NOK) err_set(1,"unknown bond command: %s",key->text);
    }
    else if (key_is("automatic")) state=AUTO;
    else if (key_is("make"))  state=MAKE;
    else if (key_is("break")) state=BREAK;
    else if (key_is("twist")) state=TWIST;
    else if (key_is("width")) state=WIDTH;
    else if (key_is("cylinder")) state=CYLINDER;
    
    else state=DONE;
    key=free_Token(key);
    reparse_atom=FALSE;
    n=0;
    
    /*   automatic make bond */
    
    while (state==AUTO) {
      key=Get_Atoms_D("atoms: ",";");
      if (key->status==TOK_OK) {
	nb=bond_auto_make((Nodechain *)key->ptr,AutobondMode);
	send_bheader(graphix,"ARMK",0);
	prtGraphics("%d bonds added\n",nb);
	clear_node_chain((Nodechain *)key->ptr);
	free_node_chain((Nodechain *)key->ptr);
	state=DONE;
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
	state=MAIN;
      }else{
	state=DONE;
      }
      key=free_Token(key);
    }
    
    /* set rendering flag for bonds */

    while (state==CYLINDER) {
      cyl_opt=CYL_NOP;
      key=Get_Kwd("cylinder option: ",";","on off radius");
      if (key->status==TOK_OK) {
	if (key_is("on")) cyl_opt=CYL_ON;
	else if (key_is("off")) cyl_opt=CYL_OFF;
	else if (key_is("radius")) {
	  key=free_Token(key);
	  key=Get_Float_M("cylinder radius: ",";",rad_men);
	  if (key->status==TOK_NOK) {
	    key=Prs_Kwd(key,"scale");
	    if (key->status==TOK_OK) {
	      key=free_Token(key);
	      key=Get_Float_M("scale factor: ",";",scale_men);
	      if (key->status==TOK_OK) key->f=-key->f;
	    }
	  }
	  if (key->status==TOK_OK) {
	    cyl_opt=CYL_RAD;
	    cyl_rad=key->f;
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad radius token: %s",key->text);
	    state=MAIN;
	  }else{
	    state=MAIN;
	  }
	}
	if (cyl_opt!=CYL_NOP) {
	  key=free_Token(key);
	  key=Get_Bonds_D("bonds: ",";");
	  if (key->status==TOK_OK) {
	    if (key->n > 0) {
	      n=0; bs=(Bond *)NULL; be=(Bond *)NULL;
	      for (b=(Bond *)key->ptr;mvld(b);b=bn) {
		bn=b->chain;
		if (cyl_opt==CYL_ON||cyl_opt==CYL_OFF) b->cyl=cyl_opt;
		else if (cyl_opt==CYL_RAD) b->r=cyl_rad;
		if (cPtr(b->cross)) {
		  if (mvld(bs)) be->chain=b;
		  else bs=b;
		  be=b;
		  n++;
		}
		b->chain=(Bond *)NULL;
	      }
	      if (n>0) {
		switch (cyl_opt) {
		case CYL_ON:
		case CYL_OFF:
		  send_bheader(graphix,"BSOL",n*cPtrSize()+sizeof(int));
		  send_bint(graphix,cyl_opt);
		  break;
		case CYL_RAD:
		  send_bheader(graphix,"BRAD",n*cPtrSize()+sizeof(int));
		  send_bfloat(graphix,cyl_rad);
		  break;
		}
		for (b=bs;mvld(b);b=bn) {
		  bn=b->chain;
		  send_bcptr(graphix,b->cross);
		  b->chain=(Bond *)NULL;
		}
	      }
	    }
	    state=MAIN;
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad bond selection %s\n",key->text);
	    state=MAIN;
	  }else{
	    state=MAIN;
	  }
	  send_header(graphix,"ARMK",0);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad bond cylinder option: %s",key->text);
	state=MAIN;
      }else{
	state=MAIN;
      }
      key=free_Token(key);
      err_dump();
    }

    /* set linewidth for bonds */

    while (state==WIDTH) {
      key=Get_Int_M("linewidth: ",";","1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20");
      if (key->status==TOK_OK) {
	lw=(key->n < 1 ? 1 : (key->n >100 ? 100 : key->n));
	key=free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  if (key->n > 0) {
	    n=0; bs=(Bond *)NULL; be=(Bond *)NULL;
	    for (b=(Bond *)key->ptr;mvld(b);b=bn) {
	      b->lw=lw;
	      bn=b->chain;
	      if (cPtr(b->cross)) {
		if (mvld(bs)) be->chain=b;
		else bs=b;
		be=b;
		n++;
	      }
	      b->chain=(Bond *)NULL;
	    }
	    if (n>0) {
	      send_bheader(graphix,"SBLW",sizeof(int)+n*cPtrSize());
	      send_bint(graphix,lw);
	      for (b=bs;mvld(b);b=bn) {
		bn=b->chain;
		send_bcptr(graphix,b->cross);
		b->chain=(Bond *)NULL;
	      }
	    }
	  }
	  state=MAIN;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad bond selection %s\n",key->text);
	  state=MAIN;
	}else{
	  state=MAIN;
	}
	key=free_Token(key);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad linewidth token: %s",key->text);
	state=MAIN;
      }else {
	state=MAIN;
      }
    }
    
    /*   make bond   */
    
    while (state==MAKE) {
      n=Get_nAtoms_M(2,atm,"make bond",";",mb_keys);
      if (n==2) {
	b=find_bond(atm[0]->ptr,atm[1]->ptr);
	if (mvld(b)) {
	  err_set(1,"atoms are already bonded.");
	}else{
	  b=add_bond(atm[0]->ptr,atm[1]->ptr);
	  if (!mvld(b)) {
	    err_set(1,"bond allocation failed.");
	  }else{
	    set_bt(b,BT_SINGLE);
	  }
	}
	send_header(graphix,"ARMK",0);
	free_Token(atm[0]);free_Token(atm[1]);
      }else{
	reparse_atom=TRUE;
	break;
      }
    } 
    
    /*   break bond  */
    
    while (state==BREAK) {
      atm[0]=Get_Bond_M("break bond",";",br_keys);
      if (atm[0]->n > 0) send_bheader(graphix,"ARMK",0);
      if (atm[0]->status==TOK_NOK) {
	n=atm[0]->n;
	atm[0]=Prs_Kwd(atm[0],"zero");
	if (atm[0]->status==TOK_OK) {
	  free_Token(atm[0]);
	  atm[0]=Get_Bonds_D("break zero order bonds between atoms ",";");
	  if (atm[0]->status==TOK_OK) {
	    for (nb=0,b=(Bond *)atm[0]->ptr;mvld(b);b=bn) {
	      bn=b->chain;
	      b->chain=NULL;
	      if (b->bo==0) {
		del_bond(b);
		nb++;
	      }
	    }
	    prtGraphics("%d zero-order bond%s deleted\n",nb,nb==1?"":"s");
	  }
	}else{
	  atm[0]->n=n;
	}
      }
      if (atm[0]->status==TOK_OK && atm[0]->n==2) {
	del_bond(atm[0]->ptr);
      }else if (atm[0]->status==TOK_NOK && atm[0]->n==2) {
	err_set(1,"no such bond: %s",atm[0]->text);
      }else{
	n=0;
	reparse_atom=TRUE;
	break;
      }
      free_Token(atm[0]);
    }
    
    /* twist around axis given by atom pair */

    while (state==TWIST) {
      n=Get_nAtoms_M(2,atm,"twist angle or twist axis",";",twist_keys);
      if (n==2) {
	pre=(Atom *)atm[0]->ptr;
	root=(Atom *)atm[1]->ptr;
	atm[0]=free_Token(atm[0]);atm[1]=free_Token(atm[1]);
	if (pre==root) {
	  err_set(1,"two atoms must be different.");
	}else if (graphix>0&&glblobj(assocobj((Node *)root))!=glblobj(assocobj((Node *)pre))) {
	    err_set(1,"two atoms must be in the same graphics object.");
	}else if (graphix>0&&(!cPtr(pre->cross)||!cPtr(root->cross))) {
	    err_set(1,"atoms on axis unknown to graphics.");
	}else{
	  if (mvld(mobile)) {
	    freeze_rotax(mobile);
	    send_bheader(graphix,"DRAX",0);
	    clear_node_chain(mobile);
	    free_node_chain(mobile);
	    mobile=(Nodechain *)NULL;
	  }
	  mobile=connected_to_right(pre,root);
	  if (mvld(mobile)) {
	    if (mobile->n>1) {
	      prtGraphics("%d mobile atoms.\n",mobile->n-1);
	      pre->chain=mobile->s;
	      mobile->s=(Node *)pre;
	      mobile->n++;
	      for (ngm=0,am=(Atom *)mobile->s;mvld(am);am=(Atom *)am->chain) {
		if (cPtr(am->cross)) ngm++;
	      }
	      send_bheader(graphix,"DRAX",ngm*cPtrSize());
	      for (am=(Atom *)mobile->s;mvld(am);am=(Atom *)am->chain)
		if (cPtr(am->cross)) send_bcptr(graphix,am->cross);
	      trflush(graphix);
	    }else{
	      err_set(1,"no mobile atoms.");
	      clear_node_chain(mobile);
	      free_node_chain(mobile); mobile=(Nodechain *)NULL;
	    }
	  }else{
	    err_set(1,"failure in connected_to_right (module graph).");
	  }
	}
      }else{
	key=Prs_Float(atm[n]);
	for (i=0;i<n-1;i++) atm[n]=free_Token(atm[i]);
	if (key->status==TOK_OK) {
	  send_bheader(graphix,"SRAX",sizeof(float));
	  send_float(graphix,key->f);
	  twist_angle+=key->f;
	}else{
	  key=Prs_Kwd(atm[n],"reset axis monitor");
	  if (key->status==TOK_OK) {
	    if (key_is("reset")) {
	      send_header(graphix,"SRAX",0);
	      twist_angle=0;
	    }else if (key_is("monitor")) {
	      n=Get_nAtoms_M(2,atm,"monitor pair",";","clear");
	      if (n==2) {
		am=(Atom *)atm[0]->ptr; an=(Atom *)atm[1]->ptr;
		atm[0]=free_Token(atm[0]); atm[1]=free_Token(atm[1]);
		if (cPtr(am->cross) && cPtr(an->cross)) {
		  send_bheader(graphix,"AMON",2*cPtrSize());
		  send_bcptr(graphix,am->cross);
		  send_bcptr(graphix,an->cross);
		}
	      }else{
		key=Prs_Kwd(atm[0],"clear");
		if (key->status==TOK_OK) {
		  send_header(graphix,"CMON",0);
		}
	      }
	    }
	  }
	}
	if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}else if (key->status!=TOK_OK) {
	  state=DONE;
	  if (mvld(mobile)) {
	    freeze_rotax(mobile);
	    send_bheader(graphix,"DRAX",0);
	    clear_node_chain(mobile);
	    free_node_chain(mobile);
	    mobile=(Nodechain *)NULL;
	  }
	}
      }
      send_bheader(graphix,"ARMK",0);
      err_dump();
      key=free_Token(key);
    }
    
    /*   reparse bad atom token */

    if (reparse_atom) {
      if (n>0) send_header(graphix,"ARMK",0);
      switch (atm[n]->status) {
      case TOK_NOK:
	key=Prs_Kwd(atm[n],mb_keys);
	if (key->status==TOK_OK) state=MAIN;
	else {
	  err_set(1,"no such atom: %s",key->text);
	  key=free_Token(key);
	}
	break;
      case TOK_ABORT:
	state=(n>0 ? state : DONE);
	break;
      case TOK_EOC:
	state=DONE;
	break;  
      default:
	state=MAIN;
      }
      for (i=0;i<n-1;i++) atm[i]=free_Token(atm[i]);
    }
    
  }
}
/*-------------------- bond_auto_make --------------------*/

int bond_auto_make(Nodechain *atoms,int mode)
{
  BHtree *tree;
  Atom *a,*atom[MXLIST],**Atoms,**L;
  Bond *b,*bs,blist;
  float rcov,dist[MXLIST],tol,rmax;
  int i,j,grain,n,na,N,R,nb,error;

  tol=0.40;    /* is this a good value ? */

  nb=0;
  if (atoms->n > 0) {
    Atoms=(Atom **)chain_to_ptr(atoms,FALSE);
    if (mvld(Atoms)) { 
      na=atoms->n;
      clear_node_chain(atoms);
      switch (mode) {     /* chain abuse... chain 'atoms' will be re-ordered... */
      case AUTOBOND_ALL:
	for (i=0;i<na;i++) Atoms[i]->chain=(Node *)NULL;
	qsort(Atoms,na,sizeof(Atom *),cmpchain);
	break;
      case AUTOBOND_OBJECT:
	for (i=0;i<na;i++) Atoms[i]->chain=(Node *)assocobj((Node *)Atoms[i]);
	qsort(Atoms,na,sizeof(Atom *),cmpchain);
	break;
      case AUTOBOND_CHAIN:
	for (i=0;i<na;i++) Atoms[i]->chain=(Atoms[i]->up->up ? Atoms[i]->up->up : Atoms[i]->up);
	qsort(Atoms,na,sizeof(Atom *),cmpchain);
	break;
      case AUTOBOND_RESID:
	for (i=0;i<na;i++) Atoms[i]->chain=Atoms[i]->up;
	qsort(Atoms,na,sizeof(Atom *),cmpchain);
	break;
      default:
	err_set(1,"internal error: unknown autobond mode.");
	return(0);
      }
      bs=&blist; bs->chain=(Bond *)NULL; error=FALSE;
      for (R=na,L=Atoms;L<Atoms+na;L+=N,R-=N) {
	for (N=0;N<R;N++) if (L[0]->chain!=L[N]->chain) break;
	if (N<2) continue;
	grain = ( N < MXTREE ? 1 : N / MXTREE );
	tree=GenerateBHtree(N,L,grain);
	if (mvld(tree)) {
	  bs=connectUniqueH(bs,tree,tol);
	  for (rmax=0,i=0;i<tree->n;i++)if (rmax<tree->atom[i]->atp->rcov) rmax=tree->atom[i]->atp->rcov;
	  for (i=0;i<tree->n;i++) {
	    a=tree->atom[i];
	    if (a->atp->z==1) continue;
	    rcov=a->atp->rcov;
	    n=findBHcloseAtoms(tree,a->x,rmax+rcov+tol,atom,dist,MXLIST);
	    if (n>MXLIST) {
	      err_set(1,"bond_auto_make: overflow, %d neighbours",n);
	      error=TRUE;
	      break;
	    }
	    for (j=0;j<n;j++) {
	      if (atom[j]->atp->z==1) continue;     /* deal with H separately... */
	      if (atom[j]->atp->rcov>rcov||a==atom[j]) continue;
	      if (dist[j]<=(atom[j]->atp->rcov+rcov)+tol&&
		  dist[j]>=(atom[j]->atp->rcov+rcov)-tol) {   /* 20.3.2009: added lower bound */
		if (!find_bond(a,atom[j])) {
		  b=add_bond_raw(a,atom[j]);
		  if (mvld(b)) {
		    bs->chain=b;
		    bs=b;
		  }else{
		    err_set(1,"memalloc failed: add bond");
		    error=TRUE;
		    break;
		  }
		}
	      }
	    }
	  }
	  bs=connectOtherH(bs,tree,tol);
	  freeBHtree(tree);
	}else{
	  err_set(1,"No tree (auto_bond)");
	}
      }
      for (i=0;i<na;i++) {
	Atoms[i]->chain=(Node *)NULL;
	add_to_node_chain((Node *)Atoms[i],atoms);
      }
      bs->chain=(Bond *)NULL;
      if (error) {
	for (bs=blist.chain;mvld(bs);bs=blist.chain) {
	  blist.chain=bs->chain;
	  bs->chain=(Bond *)NULL;
	  del_bond(bs);
	}
	nb=0;
      }else{
#ifdef CLEANUP_BONDLIST
	blist.chain=cleanup_bondlist(blist.chain);   /*** experimental, not working well enough... ***/
#endif
	set_bt_automatic(blist.chain);
	if (mvld(blist.chain)) send_bond_chain(blist.chain);
	for (nb=0,bs=blist.chain;mvld(bs);bs=blist.chain) {
	  blist.chain=bs->chain;
	  bs->chain=(Bond *)NULL;
	  nb++;
	}
      }
      free(Atoms);
    }
  }
  return(nb);
}
/*-------------------- connectUniqueH --------------------*/

static Bond *connectUniqueH(Bond *bs,BHtree *tree,float tol)
{
  int i,j,k,n;
  Atom *a,*atom[MXLIST];
  Bond *b,*bf;
  float rcov,dist[MXLIST],error;

  bf=bs;
  error=FALSE;
  
  for (rcov=0.0,i=0;i<tree->n;i++) {
    a=tree->atom[i];
    rcov=(a->atp->rcov > rcov ? a->atp->rcov : rcov);
  }
  for (i=0;i<tree->n;i++) {
    a=tree->atom[i];
    if (a->atp->z!=1||a->nbnd>0) continue;   /* only H with 0 bonds... */
    n=findBHcloseAtoms(tree,a->x,rcov+a->atp->rcov+tol,atom,dist,MXLIST);
    if (n<=MXLIST) {
      for (k=0,j=0;j<n;j++) { /* purge H and atoms beyond bonding distance */
	if (atom[j]->atp->z>1&&dist[j]<a->atp->rcov+atom[j]->atp->rcov+tol) {
	  if (k>0) {
	   dist[j-k]=dist[j];
	   atom[j-k]=atom[j];
	  }
	}else k++;
      }
      n-=k;
      if (n==1) {
	if (!find_bond(a,atom[0])) {
	  b=add_bond_raw(a,atom[0]);
	  if (mvld(b)) {
	    bs->chain=b;
	    bs=b;
	  }else{
	    error=TRUE;
	    break;
	  }
	}
      }
    }else{
      error=TRUE;
      break;
    }
  }
  if (error) {
    bs=bf;
    while (mvld(bs->chain)) {
      b=bs->chain;
      bs->chain=bs->chain->chain;
      del_bond(bs);
    }
  }
  return(bs);
}

/*-------------------- connectOtherH --------------------*/

static Bond *connectOtherH(Bond *bs,BHtree *tree,float tol)
{
  int i,j,k,l,n;
  Atom *a,*a2,*atom[MXLIST];
  Bond *b,*bf;
  float rcov,dist[MXLIST],w,wmn[MXLIST],wmx[MXLIST],wmin,wmax,error;
  float p,pmax;

  bf=bs;
  error=FALSE;
  
  for (rcov=0.0,i=0;i<tree->n;i++) {
    a=tree->atom[i];
    rcov=(a->atp->rcov > rcov ? a->atp->rcov : rcov);
  }
  for (i=0;i<tree->n;i++) {
    a=tree->atom[i];
    if (a->atp->z!=1||a->nbnd>0) continue;   /* only H with 0 bonds... */
    n=findBHcloseAtoms(tree,a->x,rcov+a->atp->rcov+tol,atom,dist,MXLIST);
    if (n<=MXLIST) {
      for (k=0,j=0;j<n;j++) { /* purge H,Lp,Du and atoms beyond bonding distance */
	if (atom[j]->atp->z<=1) {
	  k++;
	  continue;
	}
	if (dist[j]>a->atp->rcov+atom[j]->atp->rcov+tol) {
	  k++;
	  continue;
	}
	if ((atom[j]->atp->z==6&&atom[j]->nbnd>3)||
	    (atom[j]->atp->z==7&&atom[j]->nbnd>3)||
	    (atom[j]->atp->z==8&&atom[j]->nbnd>1)||
	    (atom[j]->atp->z==15&&atom[j]->nbnd>3)||
	    (atom[j]->atp->z==16&&atom[j]->nbnd>3)) {
	  k++;
	  continue;
	}
	for (wmin=180.0,wmax=0.0,l=0;l<atom[j]->nbnd;l++) {
	  a2=neighbour(l,atom[j]);
	  w=cangled(a->x,atom[j]->x,a2->x);
	  wmin=(wmin>w ? w : wmin);
	  wmax=(wmax<w ? w : wmax);
	}
	wmn[j]=wmin; wmx[j]=wmax;
	if (k>0) {
	  dist[j-k]=dist[j];
	  atom[j-k]=atom[j];
	  wmn[j-k]=wmn[j];
	  wmx[j-k]=wmx[j];
	}
      }
      n-=k;
      if (n>1) {
	for (pmax=-1.e10,l=0,j=0;j<n;j++) {
	  p=(a->up==atom[j]->up ? 1 : 0);
	  if (atom[j]->nbnd==1) {
	    if (wmn[j]<150) {
	      if (wmn[j]<108) p-=1.5*(108-wmn[j])/108;
	      if (wmx[j]>122) p-=1.5*(wmx[j]-122)/30;
	    }else p-=1.5*(180-wmn[j])/30;
	  }else if (atom[j]->nbnd>1) {
	    if (wmn[j]<108) p-=1.5*(108-wmn[j])/108;
	    if (wmx[j]>122) p-=1.5*(wmx[j]-122)/58;
	  }
	  if (p>pmax) {
	    pmax=p;
	    l=j;
	  }
	}
      }else l=0;
      if (n>0) {
	if (!find_bond(a,atom[l])) {
	  b=add_bond_raw(a,atom[l]);
	  if (mvld(b)) {
	    bs->chain=b;
	    bs=b;
	  }else{
	    error=TRUE;
	    break;
	  }
	}
      }
    }else{
      error=TRUE;
      break;
    }
  }
  if (error) {
    bs=bf;
    while (mvld(bs->chain)) {
      b=bs->chain;
      bs->chain=bs->chain->chain;
      del_bond(bs);
    }
  }
  return(bs);
}

/*-------------------- freeze_rotax --------------------*/

static void freeze_rotax(Nodechain *mob)
{
  Atom *a,*p,*r;
  mat_4 rot;
  Qmsg *Msg;
  float angle;
  int size;

  if (!mvld(mob)) return;
  if (mob->n<3) return;
  p=(Atom *)mob->s;
  r=(Atom *)p->chain;

  if (graphix>0) {
    send_header(graphix,"FRAX",0);
    Msg=waitqmsg("SRAX SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }
    }else{
      graphics_down("no response from server.");
    }
    if (Msg) {
      size=(Msg->size)/sizeof(float);
      if (size>0) angle=MY_f(*((int *)Msg->msg));
      else angle=0.0;
      delqmsg(Msg);
    }else{
      angle=0.0;
    }
  }else angle=twist_angle;
  if (angle!=0.0) {
    rot=rotax(p->x,r->x,angle);
    for (a=(Atom *)mob->s->chain->chain;mvld(a);a=(Atom *)a->chain) transform_atom(rot,a);
  }
  twist_angle=0;
}

#ifdef CLEANUP_BONDLIST
/*-------------------- cleanup_bondlist --------------------*/

static Bond *cleanup_bondlist(Bond *chain)
{
  Bond blist,rlist,*lst,*b,*nb;
  Atom *a,*f,*t;
  int i,im,interR,interC,inlist,deleted;
  float w,wm;

  inlist= (1<<0);
  interR= (1<<1);
  interC= (1<<2);
  deleted=(1<<3);

  blist.chain=chain;
  for (b=blist.chain;b;b=b->chain) {
    b->flg&=~(inlist|deleted|interR|interC);
    b->flg|=inlist;
    if (b->f->up!=b->t->up) {
      b->flg|=interR;
      if (b->f->up->up!=b->t->up->up) b->flg|=interC;
    }
  }

  for (b=blist.chain;b;b=b->chain) {
    if (!(b->flg&interC) || b->flg&deleted) continue;
    f=b->f;
    t=b->t;
    wm=500; im=-1;
    for (i=0;i<f->nbnd;i++) {
      if (f->bnd[i]->flg&interC || f->bnd[i]->flg&deleted) continue;
      a=neighbour(i,f);
      if (a!=t) {
	w=cangled(a->x,f->x,t->x);
	if (w<wm) wm=w;
      }
    }
    for (i=0;i<t->nbnd;i++) {
      if (t->bnd[i]->flg&interC || t->bnd[i]->flg&deleted) continue;
      a=neighbour(i,t);
      if (a!=f) {
	w=cangled(a->x,t->x,f->x);
	if (w<wm) wm=w;
      }
    }

    if (wm<45.0) {
	b->flg|=deleted;
    }
  }

  for (b=blist.chain;b;b=b->chain) {
    if (!(b->flg&interR) || b->flg&deleted) continue;
    f=b->f;
    t=b->t;
    wm=500; im=-1;
    for (i=0;i<f->nbnd;i++) {
      if (f->bnd[i]->flg&interR || f->bnd[i]->flg&deleted) continue;
      a=neighbour(i,f);
      if (a!=t) {
	w=cangled(a->x,f->x,t->x);
	if (w<wm) wm=w;
      }
    }
    for (i=0;i<t->nbnd;i++) {
      if (t->bnd[i]->flg&interR || t->bnd[i]->flg&deleted) continue;
      a=neighbour(i,t);
      if (a!=f) {
	w=cangled(a->x,t->x,f->x);
	if (w<wm) wm=w;
      }
    }

    if (wm<45.0) {
	b->flg|=deleted;
    }
  }

  lst=&rlist;
  for (b=blist.chain;b;b=nb) {
    nb=b->chain;
    b->chain=NULL;
    if (b->flg&deleted) {
      del_bond(b);
    }else{
      b->flg&=~(inlist|deleted|interR|interC);
      lst->chain=b;
      lst=b;
    }
  }

  /* should do a 2nd pass to clean up further */

  lst->chain=0;
  return(rlist.chain);
}
#endif

/*-------------------- set_autobond_mode --------------------*/

int set_autobond_mode(int mode)
{
  int r;

  r=AutobondMode;
  AutobondMode=mode;
  return(r);
}
/*-------------------- autobond_mode --------------------*/

int autobond_mode(void)
{
  return(AutobondMode);
}
