#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *MolRoot,*DefaultNode;

#define key_is(a) (streq(a,key->key))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define READ_TYPES FT_DOTS|FT_MOL_TYPES|FT_WLB

static char *kwdl = "Bond Construct Copy Delete Discover Fragment Invert Join Loop Maccs "
                    "Measure Merge Minimax Modify Mopac Move PIFF Sketch Split End "
                    "' ' command color cpk display fit freeze label read reset "
                    "select set write";

static char *xtra_stuff = "'list'";

static void Construct(void);

/*========================================================================*/

void cmd_build(void)
{
  Token *key;

  key=(Token *)NULL;

  for(;;) {
    start_parse(); err_dump(); key=free_Token(key);
    key=Get_Kwd("Build> ",";",kwdl);
    if (key->status==TOK_NOK) key=Prs_Kwd(key,xtra_stuff);
    if (key->status == TOK_OK) {
      if (key_is("Delete"))         cmd_delete();
      else if (key_is("Discover"))  cmd_discover();
      else if (key_is("Bond"))      cmd_bond();
      else if (key_is("Copy"))      cmd_copy(FALSE);
      else if (key_is("Move"))      cmd_copy(TRUE);
      else if (key_is("Merge"))     cmd_merge();
      else if (key_is("Construct")) Construct();
      else if (key_is("Invert"))    cmd_invert();
      else if (key_is("Fragment"))  cmd_fragment();
      else if (key_is("Join"))      cmd_join();
      else if (key_is("Loop"))      cmd_loop();
      else if (key_is("Maccs"))     cmd_maccs();
      else if (key_is("Measure"))   cmd_measure();
      else if (key_is("Modify"))    cmd_modify();
      else if (key_is("Sketch"))    cmd_sketch();
      else if (key_is("Minimax"))   cmd_minimax();
      else if (key_is("PIFF"))      cmd_piff();
      else if (key_is("Mopac"))     cmd_mopac();
      else if (key_is("Split"))     cmd_split();
      else if (key_is("End"))       break;
      else if (key_is("command"))   command_loop("Wit!P(Build)> ");
      else if (key_is("color"))     cmd_color();
      else if (key_is("cpk"))       cmd_cpk();
      else if (key_is("label"))     cmd_label();
      else if (key_is("display"))   cmd_display();
      else if (key_is("reset"))     cmd_display_reset();
      else if (key_is("fit"))       cmd_fit("fit");
      else if (key_is("freeze"))    cmd_freeze();
      else if (key_is("select"))    cmd_select();
      else if (key_is("set"))       cmd_set();
      else if (key_is("read"))      cmd_read((Node *)NULL,READ_TYPES);
      else if (key_is("write"))     cmd_write((Node *)NULL,FT_CAN_WRITE);
      else if (key_is("list"))      cmd_list();
    }else if (key->status == TOK_NOK) {
      err_set(0,"unknown command %s\n",key->text);
    }else if (key->status == TOK_NULL) {
      err_set(1,"null token at top level, aborted.");
      err_dump();
      exit(1);
    }
    if (end_parse()!=TOK_OK) {
      err_set(0,"extra characters at end of command ignored.");
    }
  }
  free_Token(key);
}
/*-------------------- cmd_split --------------------*/

void cmd_split(void)
{
  Token *atm[2];
  int i,n;
  Atom *pre,*root;
  Nodechain *c;
  Node *o,*p;
  
  for (i=0;i<2;i++) atm[i]=(Token *)NULL;

  n=Get_nAtoms(2,atm,"split bond ",";");
  send_bheader(graphix,"ARMK",0);
  if (n==2) {
    pre=(Atom *)atm[0]->ptr;
    root=(Atom *)atm[1]->ptr;
    c=connected_to_right(pre,root);
    if (mvld(c)) {
      if (c->n>0) {
	for (o=c->s;mvld(o);o=o->chain) {
	  p=o->up; unlink_node(o);
	  if (mvld(p)) {
	    if (!mvld(p->fst)) add_to_node_chain(p,c);
	  }
	}
	del_node_chain(c);
      }else{
	err_set(1,"first atom is right-connected to second atom, no split.");
      }
      clear_node_chain(c); free_node_chain(c);
    }else{
      err_set(1,"failed to find right-connected fragment\n");
    }
  }else{
    if (atm[n]->status==TOK_NOK) {
      err_set(1,"bad atom token: %s",atm[n]->text);
    }
    n++;
  }
  for (i=0;i<2;i++) free_Token(atm[i]);
}
/*-------------------- cmd_invert --------------------*/

void cmd_invert(void)
{
  Token *key;

  key=(Token *)NULL;
  key=Get_Kwd("invert what: ",";","center molecule");
  if (key->status==TOK_OK) {
    if (key_is("center")) cmd_invert_center();
    else if (key_is("molecule")) cmd_invert_molecule();
  }else if (key->status==TOK_NOK) err_set(1,"bad invert option: %s",key->text);
  free_Token(key);
}
/*-------------------- cmd_invert_center --------------------*/

void cmd_invert_center(void)
{
  Token *Atm[3];
  int i,k,l,n;
  Atom *a,*atm[3];
  Nodechain c;
  vec_3 xa,xb,xc;
  mat_4 rot;
  float pt[3];
  
  for (i=0;i<2;i++) Atm[i]=(Token *)NULL;

  n=Get_nAtoms(1,Atm,"central atom: ",";");
  send_bheader(graphix,"ARMK",0);
  if (n==1) {
    n+=Get_nAtoms(2,Atm+n,"exchange atoms ",";");
    send_bheader(graphix,"ARMK",0);
    if (n==3) {
      for (i=0;i<3;i++) atm[i]=Atm[i]->ptr;
      for (i=1;i<3;i++) {
	for(k=0;k<atm[i]->nbnd;k++) if (neighbour(k,atm[i])==atm[0]) break;
	if (k==atm[i]->nbnd) break;
      }
      if (i<3) {
	err_set(1,"both exchange atoms must be bonded to central atom.");
      }else{
	atm[0]->chain=(Node *)atm[0];
	init_node_chain(&c);
	add_to_node_chain((Node *)atm[1],&c);
	add_to_node_chain((Node *)atm[2],&c);
	connected_atoms(&c);
	atm[0]->chain=(Node *)NULL;
	for (i=0,k=0;k<atm[0]->nbnd;k++) if (node_in_chain((Node *)neighbour(k,atm[0]),&c)) i++;
	if (atm[1]==atm[2]) i++;
	if (i==2) {
	  if (atm[1]==atm[2]){
	    if (atm[0]->nbnd==3) {
	      for (l=0,k=0,a=atm[2];k<atm[0]->nbnd;k++) {
		if (neighbour(k,atm[0])!=a) atm[++l]=neighbour(k,atm[0]);
	      }
	      xa=v3n(v3mv3(vec3(atm[0]->x),vec3(atm[1]->x)));
	      xb=v3n(v3mv3(vec3(atm[0]->x),vec3(atm[2]->x)));
	      xc=v3pv3(v3pv3(vec3(atm[0]->x),xa),xb);
	      v32arr(pt,xc);
	      rot=rotax(atm[0]->x,pt,180.0);
	      for (a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) transform_atom(rot,a);
	      send_newx(&c);
	    }else{
	      err_set(1,"sorry, I don't check this...");
	    }
	  }else{
	    xa=v3n(v3mv3(vec3(atm[1]->x),vec3(atm[0]->x)));
	    xb=v3n(v3mv3(vec3(atm[2]->x),vec3(atm[0]->x)));
	    xc=v3pv3(v3pv3(vec3(atm[0]->x),xa),xb);
	    v32arr(pt,xc);
	    rot=rotax(atm[0]->x,pt,180.0);
	    for (a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) transform_atom(rot,a);
	    send_newx(&c);
	  }
	}else{
	  err_set(1,"exchange atoms are connected to precursor atoms.");
	}
	clear_node_chain(&c);
      }
    }
  }
  if (n<3) {
    if (Atm[n]->status==TOK_NOK) {
      err_set(1,"bad atom token: %s",Atm[n]->text);
    }
    n++;
  }
  for (i=0;i<n;i++) free_Token(Atm[i]);
}
/*-------------------- cmd_invert_molecule --------------------*/

void cmd_invert_molecule(void)
{
  Token *key;
  Node *n;
  Nodechain C;
  Atom *a;
  Xcn *xcn;
  float com[3];
  int i,j;
  mat_4 T;

  key=(Token *)NULL;
  key=Get_Mol_D("invert molecule: ",";");
  if (key->status==TOK_OK) {
    n=(Node *)key->ptr;
    C=tree2chain(n,OBJ_ATOM);
    if (C.n>0) {
      xcn=(Xcn *)find_cousin(C.s,OBJ_XCN);
      if (is_xcn(xcn)) prtGraphics("Warning: non cryst. inversion !\n");
      for (i=0;i<3;i++) com[i]=0;
      for (a=(Atom *)C.s;mvld(a);a=(Atom *)a->chain) {
	for (i=0;i<3;i++) com[i]+=a->x[i];
      }
      for (i=0;i<3;i++) com[i]/=C.n;
      for (i=0;i<4;i++) for(j=0;j<3;j++) T.r[i].c[j]=(i==j ? -1.0 : 0.0);
      for (i=0;i<3;i++) {T.r[i].c[3]=2*com[i];} T.r[3].c[3]=1.0;
      for (a=(Atom *)C.s;mvld(a);a=(Atom *)a->chain) transform_atom(T,a);
      send_newx(&C);
    }
    clear_node_chain(&C);
  }else if (key->status==TOK_NOK) err_set(1,"bad molecule selection %s",key->text);
  free_Token(key);
}
/*-------------------- cmd_join --------------------*/

void cmd_join(void)
{
  Token *key;
  Atom *a[2],*p[2]={NULL,NULL},*atm;
  Xatom *xatm;
  Xcn *xcn;
  Node *r[2],*up,*cp;
  char buf[256];
  char *menu;
  Nodechain *c,C;
  Dspobj *proot,*mdl;

  float d,x[2][3],y[2][3];
  mat_4 trf;
  int i;
  
  a[0]=(Atom *)NULL; a[1]=a[0];
  r[0]=(Node *)NULL; r[1]=r[0];
  d=1.5;

  key=(Token *)NULL;
  key=free_Token(key);
  menu=Make_TopMolNameList("* /*",DefaultNode);
  key=Get_Mols_M("target molecule: ",";",menu);
  send_bheader(graphix,"ARMK",0);
  free_menu(menu);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n==1) {
      r[0]=c->s;
    }else{
      err_set(1,"bad molecule selection: %s",key->text);
      key->status=TOK_ABORT;
    }
    clear_node_chain(c); free_node_chain(c);
  }

  if (key->status==TOK_OK) {
    proot=pick_root(assocobj(r[0]));
    key=free_Token(key);
    key=Get_Atom("target atom: ",";");
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      a[0]=(Atom *)key->ptr;
      for (up=a[0]->up;up!=r[0];up=up->up) {}
      if (up!=r[0]) {
	err_set(1,"atom not found in target molecule.");
	key->status=TOK_ABORT;
      }
    }
    pick_root(proot);
    if (key->status==TOK_NOK) {
      err_set(1,"bad atom selection: %s",key->text);
      key->status=TOK_ABORT;
    }
  }

  if (key->status==TOK_OK) {
    menu=Make_TopMolNameList("* /*",DefaultNode);
    key=Get_Mols_M("source molecule: ",";",menu);
    send_bheader(graphix,"ARMK",0);
    free_menu(menu);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n==1) {
	r[1]=c->s;
	for (up=r[0];mvld(up);up=up->up) if (up==r[1]) break;
	if (mvld(up)) {
	  err_set(1,"source contains target...");
	  key->status=TOK_ABORT;
	}
      }else{
	err_set(1,"bad source molecule selection: %s",key->text);
	key->status=TOK_ABORT;
      }
      clear_node_chain(c); free_node_chain(c);
    }
  }

  if (key->status==TOK_OK) {
    proot=pick_root(assocobj(r[1]));
    key=free_Token(key);
    key=Get_Atom("source atom: ",";");
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      a[1]=(Atom *)key->ptr;
      for (up=a[1]->up;up!=r[1];up=up->up) {}
      if (up!=r[1]) {
	err_set(1,"atom not found in source molecule.");
	key->status=TOK_ABORT;
      }
    }
    pick_root(proot);
    if (key->status==TOK_NOK) {
      err_set(1,"bad source atom selection: %s",key->text);
      key->status=TOK_ABORT;
    }
  }

  if (key->status==TOK_OK) {
    if (a[0]->nbnd!=1) {
      err_set(1,"%s is not a terminal atom.",nodeptr_to_name(buf,(Node *)a[0]));
      key->status=TOK_ABORT;
    }else{
      p[0]=neighbour(0,a[0]);
    }
  }
  if (key->status==TOK_OK) {
    if (a[1]->nbnd!=1) {
      err_set(1,"%s is not a terminal atom.",nodeptr_to_name(buf,(Node *)a[1]));
      key->status=TOK_ABORT;
    }else{
      p[1]=neighbour(0,a[1]);
    }
  }

  if (key->status==TOK_OK) {
    key=free_Token(key);
    key=Get_Float_M("bondlength: ",";",
		    "1.15 1.20 1.25 1.30 1.35 1.40 1.45 1.50 1.55 1.60 1.65 1.70 1.75 1.80");
    if (key->status==TOK_OK) {
      d=key->f;
    }else{
      if (key->status==TOK_NOK) {
	err_set(1,"bad floating point value: %s",key->text);
	key->status=TOK_ABORT;
      }
    }
  }

  if (key->status==TOK_OK) {
    for (i=0;i<3;i++) {
      x[0][i]=p[0]->x[i];
      y[1][i]=p[1]->x[i];
    }
    v32arr(x[1],v3pv3(vec3(x[0]),sxv3(d,v3n(v3mv3(vec3(a[0]->x),vec3(x[0]))))));
    v32arr(y[0],v3pv3(vec3(y[1]),sxv3(d,v3n(v3mv3(vec3(a[1]->x),vec3(y[1]))))));
    rms_best_mat_yx(2,(float *)y,(float *)x,&trf);
    a[1]->flg|=(1<<1); p[1]->flg|=(1<<2);             /* mark join atom and precursor */
    cp=tree_copy(r[1]);
    a[1]->flg&=~(1<<1); p[1]->flg&=~(1<<2);
    if (mvld(cp)) {
      cp->nxt=r[0];
      link_node(cp);              /* renumbers atomsets as necessary... */
      copy_set_bits(r[1],cp);
      unlink_node(cp);
      renumber_atomsets(cp,r[0]);
      ForAtoms(atm,cp) {
	if (atm->flg&(1<<1)) a[1]=atm; 
	if (atm->flg&(1<<2)) p[1]=atm;
	transform_atom(trf,atm);
	while(mvld(xatm=first_xatom(atm))) del_node((Node *)xatm);
	atm->xtal=FALSE;
      }
      if (a[1]->flg&(1<<1)) del_node((Node *)a[1]);
      p[1]->flg&=~(1<<2); 
      C=tree2chain(cp,OBJ_ATOM);
      while(mvld(xcn=rfirst_xcn(cp))) del_node((Node *)xcn);
      for (up=cp->fst;mvld(up);up=cp->fst) {
	unlink_node(up);
	up->up=r[0]; link_node(up);
	if (is_atomset(up)) register_atomset((Atomset *)up);
      }
      merge_atomsets(r[0]);
      set_bt(add_bond_ng(p[0],p[1]),BT_SINGLE);
      lock_evtq();
      mdl=assocmdl((Node *)a[0]);
      del_node((Node *)a[0]);
      del_node(r[1]);                                  /* zap source molecule ...   */
      mdl_add_atom_chain(mdl,&C);                      /* and send copy to graphics */
      unlock_evtq();
      clear_node_chain(&C);
      del_node(cp);
    }else{
      err_set(1,"failed to copy source tree.");
    }
  }

  if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- Construct --------------------*/

static void Construct(void)
{
  Token *key=(Token *)NULL;
  Nodechain *c;
  Atom *a;
  int i,n;
  float x[3];

  key=Get_Kwd("construct what: ",";","atom centroid");
  if (key->status==TOK_OK) {
    if (key_is("atom")) cmd_construct();
    else if (key_is("centroid")) {
      key=free_Token(key);
      key=Get_Atoms_D("atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (i=0;i<3;i++) x[i]=0;
	for (n=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  if (a->x&&a->nxyz>0) {
	    x[0]+=a->x[0];
	    x[1]+=a->x[1];
	    x[2]+=a->x[2];
	    n++;
	  }
	}
	if (n>0) for (i=0;i<3;i++) x[i]/=n;
	key=free_Token(key);
	key=Get_String("name: ",";");
	if (key->status==TOK_OK) {
	  if (c->n>0) {
	    a=(Atom *)make_node(OBJ_ATOM);
	    set_node_name((Node *)a,key->text);
	    set_atom_type(a,guess_atom_type("du"));
	    a->nxyz=1;
	    a->x[0]=x[0];
	    a->x[1]=x[1];
	    a->x[2]=x[2];
	    a->up=c->s->up;
	    link_node((Node *)a);
	    send_one_atom(a);
	  }
	}else if (key->status==TOK_NOK) err_set(1,"bad name: %s",key->text);
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad construct option: %s",key->text);
  key=free_Token(key);
}
