#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <math.h>
#include "swdef.h"
#include "transpor.h"
#include "parser.h"
#include "mystdlib.h"
#include "object.h"
#include "errhand.h"
#include "bhtree.h"
#include "graphics.h"
#include "strutil.h"

extern Node *MolRoot,*DefaultNode;
extern int graphix;

#define TRUE 1
#define FALSE 0

#define Pi 3.141592653
#define LPMRK 20
#define MRKWID 0.1

#define MAXVEC 20
#define MXTREE 1024

#define key_is(x) (streq(x,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static char *cav_menu = "browse graft template vectors end ' ' ";

static char *browse_menu = "next previous goto accept reject save ' ' ";

static char *add_menu = "command bond color delete display label measure "
                        "modify reset select set sketch split";

static char *vector_menu = "add clear color delete display tolerance read write ' ' end";

static int InUse = FALSE;

static int NV;
static Atom *Vec[2][MAXVEC];
static Dspobj *Mrk[MAXVEC];
static int vec_dsp,vec_color=3;
static Node *Template;
static float Tol[2]={0.15,15.0};
static UCpipe *Hits=(UCpipe *)NULL;
static FileCache *cache;
static int nseq;
static Mol *prev_mol;

static void cmd_caveat_vector(void);
static void cmd_caveat_vector_add(void);
static void cmd_caveat_vector_delete(void);
static void cmd_caveat_vector_color(void);
static void cmd_caveat_vector_display(void);
static void cmd_caveat_vector_tolerance(void);
static void cmd_caveat_vector_read(void);
static void cmd_caveat_vector_write(void);
static void caveat_vector_clear(void);
static void caveat_vector_add(Atom *a,Atom *b);
static void caveat_vector_delete(int n);
static void open_browser(void);
static void get_template(void);
static int  sanity_check(void);
static void read_hit(int nr);
static int  graft_sidechain(void);
static int  graft_one_sidechain(Node *core,int nr);
static Dspobj *vector_mark(int nr);
static void vector_unmark(int nr);

/*==================== Caveat Interface ====================*/

void cmd_caveat(void)
{
  Token *key;
  char main_menu[4096];
  int next=0,read_it,reject,sane;
  Mol *mol;

  if (InUse) {
    err_set(1,"Caveat interface may not be called recursively.");
    return;
  }

  key=(Token *)NULL;

  Template=(Node *)NULL;
  Hits=(UCpipe *)NULL;
  prev_mol=(Mol *)NULL;
  reject=TRUE;

  get_template();
  if (!mvld(Template)) return;

  InUse=TRUE;
  NV=0; vec_dsp=TRUE;
  for(sane=TRUE;sane;) {
    sprintf(main_menu,"%s %s %s",cav_menu,(mvld(Hits) ? browse_menu : ""),add_menu);
    key=free_Token(key); err_dump();
    key=Get_Kwd("Caveat> ",";",main_menu);
    if (key->status==TOK_OK) {
      read_it=FALSE;
      if (key_is("browse")) open_browser();
      else if (key_is("template")) get_template();
      else if (key_is("vectors")) cmd_caveat_vector();
      else if (key_is("graft")) reject=!graft_sidechain();
      else if (key_is("end")) break;
      else if (key_is("next")) {
	next=nseq;
	read_it=TRUE;
      }else if (key_is("previous")) {
	next=nseq-2;
	read_it=TRUE;
      }else if (key_is("goto")) {
	key=free_Token(key);
	default_token("%d",nseq);
	key=Get_Int_M("molecule number: ",";","1 +1 +10 +100 -1 -10 -100");
	if (key->status==TOK_OK) {
	  next=key->n;
	  if (*key->text=='+' || *key->text=='-') next+=nseq-1;
	  if (!mvld(cache) && next<nseq) {
	    err_set(1,"sorry, can't backspace on a pipe.");
	    read_it=FALSE;
	  }else{
	    read_it=TRUE;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad molecule number: %s",key->text);
	  read_it=FALSE;
	}
      }else if (key_is("accept")) {
	reject=FALSE;
      }else if (key_is("reject")) {
	reject=TRUE;
      }else if (key_is("save")) {
	if (is_mol(prev_mol)) {
	  cmd_write((Node *)prev_mol,FT_CAN_WRITE&FT_MOL_TYPES);
	}else{
	  err_set(1,"no molecule to save.");
	}
      }else{
	if (key_is("command"))        {command_loop("Wit!P(Caveat)> ");sane=sanity_check();}
	else if (key_is("bond"))      cmd_bond();
	else if (key_is("color"))     cmd_color();
	else if (key_is("display"))   cmd_display();
	else if (key_is("delete"))    {cmd_delete();sane=sanity_check();}
	else if (key_is("label"))     cmd_label();
	else if (key_is("modify"))    cmd_modify();
	else if (key_is("measure"))   cmd_measure();
	else if (key_is("reset"))     cmd_display_reset();
	else if (key_is("select"))    cmd_select();
	else if (key_is("set"))       cmd_set();
	else if (key_is("sketch"))    cmd_sketch();
	else if (key_is("split"))     {cmd_split();sane=sanity_check();}
      }
      if (read_it && next >0) {
	if (Hits&&Hits->pipe!=UCP_PIPE) lock_evtq();
	if (reject) {
	  forMols(mol,MolRoot) if (mol==prev_mol) break;
	  if (mvld(mol)) del_node((Node *)mol);
	}
	read_hit(next);
	if (Hits&&Hits->pipe!=UCP_PIPE) unlock_evtq();
      }
      if (next<0) next=1;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad Caveat keyword: %s",key->text);
    }else if (key->status==TOK_EOC) break;
  }

  caveat_vector_clear();
  free_Token(key);
  InUse=FALSE;
  if (mvld(Hits)) UCpclose(Hits);
  Hits=(UCpipe *)NULL;
  cache=freeFileCache(cache);
}
/*-------------------- get_template --------------------*/

static void get_template(void)
{
  Token *key;
  char *menu;
  Nodechain *c;
  Node *new;

  key=(Token *)NULL;
  menu=Make_TopMolNameList("",DefaultNode);

  new=(Node *)NULL;
  while (!mvld(new)) {
    err_dump(); key=free_Token(key);
    menu=Make_TopMolNameList("",DefaultNode);
    key=Get_Mols_M("template: ",";",menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n==1) {
	new=c->s;
	if (mvld(Template)&&Template!=new) caveat_vector_clear();
	Template=new;
      }else if (c->n>1) {
	err_set(1,"ambiguous choice: %s",key->text);
      }else{
	err_set(1,"no template selected.");
      }
      clear_node_chain(c);
      free_node_chain(c);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
    }else{
      break;
    }
  }
  free_menu(menu);
  free_Token(key);
}
/*-------------------- cmd_caveat_vector --------------------*/

static void cmd_caveat_vector(void)
{
  Token *key;

  key=(Token *)NULL;

  if (!is_mol(Template)) get_template();

  if (is_mol(Template)) {
    for (;;) {
      err_dump();
      key=free_Token(key);
      key=Get_Kwd("vector option: ",";",vector_menu);
      if (key->status==TOK_OK) {
	if (key_is("add")) cmd_caveat_vector_add();
	else if (key_is("clear")) {
	  caveat_vector_clear();
	  prtGraphics("vector list is empty.\n");
	}
	else if (key_is("delete")) cmd_caveat_vector_delete();
	else if (key_is("display")) cmd_caveat_vector_display();
	else if (key_is("tolerance")) cmd_caveat_vector_tolerance();
	else if (key_is("read")) cmd_caveat_vector_read();
	else if (key_is("write")) cmd_caveat_vector_write();
	else if (key_is("color")) cmd_caveat_vector_color();
	else if (key_is("end"))  break;
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad vector option: %s",key->text);
      }else{
	break;
      }
    }
  }

  free_Token(key);
}
/*-------------------- cmd_caveat_vector_add --------------------*/

static void cmd_caveat_vector_add(void)
{
  int i;
  Token *key;
  Dspobj *proot;
  Atom *base,*tip;

  key=(Token *)NULL;

  proot=pick_root(assocmdl(Template));
  while (NV<MAXVEC) {
    err_dump(); send_bheader(graphix,"ARMK",0);
    key=free_Token(key);
    key=Get_Atom("base-point of new vector: ",";");
    if (key->status==TOK_OK) {
      base=(Atom *)key->ptr;
      free_Token(key);
      key=Get_Atom("tip of new vector: ",";");
      if (key->status==TOK_OK) {
	tip=(Atom *)key->ptr;
	for (i=0;i<NV;i++) {
	  if (Vec[0][i]==base && Vec[1][i]==tip) break;
	  if (Vec[0][i]==tip  && Vec[1][i]==base) break;
	}
	if (i<NV) {
	  if (Vec[0][i]==tip) {
	    prtGraphics("vector %d turned around.",i+1);
	    Vec[0][i]=base; Vec[1][i]=tip;
	    if (is_dsp(Mrk[i])) {
	      lock_evtq();
	      vector_unmark(i);
	      Mrk[i]=vector_mark(i);
	      unlock_evtq();
	    }
	  }else{
	    err_set(1,"vector is already in list.");
	  }
	}else if (base==tip) {
	  err_set(1,"base and tip atoms are the same... vector ignored.");
	}else{
	  caveat_vector_add(base,tip);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad tip: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad base point: %s",key->text);
    }else{
      break;
    }
  }
  pick_root(proot);
  send_bheader(graphix,"ARMK",0);

  prtGraphics("vector list contains %d vectors.\n",NV);
  if (NV==MAXVEC) err_set(1,"maximum number of vectors reached.");
  free_Token(key);
}
/*-------------------- caveat_vector_add --------------------*/

static void caveat_vector_add(Atom *a,Atom *b)
{
  if (NV<MAXVEC) {
    if (is_atom(a)&&is_atom(b)) {
      Vec[0][NV]=a;
      Vec[1][NV]=b;
      NV++;
      Mrk[NV-1]=vector_mark(NV-1);
    }
  }
}
/*-------------------- vector_mark --------------------*/

static Dspobj* vector_mark(int nr)
{
  int i,k;
  float x[3],dm,w;
  Gfxobj *G;
  vec_3 n1,n2,A,B;
  Dspobj *r;

  if (nr<0||nr>NV-1) {
    r=(Dspobj *)NULL;
  }else if (is_dsp(Mrk[nr])) {
    r=Mrk[nr];
  }else{
    G=OpenGfxObject(glblobj(assocobj(Template)),1000);
    if (mvld(G)) {
      r=G->Obj;
      dm=fabs(Vec[0][nr]->x[0]-Vec[1][nr]->x[0]);
      for (k=0,i=2;i<3;i++) {
	if (fabs(Vec[0][nr]->x[i]-Vec[1][nr]->x[i])>dm) {
	  dm=fabs(Vec[0][nr]->x[i]-Vec[1][nr]->x[i]);
	  k=i;
	}
      }
      if (dm>0.001) {
	for (i=0;i<3;i++) n2.c[i]=0.0;
	n2.c[k]=1.0;
	A=vec3(Vec[0][nr]->x); B=vec3(Vec[1][nr]->x);
	n1=v3n(v3xv3(v3mv3(B,A),n2));
	n2=v3n(v3xv3(n1,v3mv3(B,A)));
	for (w=0;w<2*Pi;w+=2*Pi/LPMRK) {
	  v32arr(x,v3pv3(A,v3pv3(sxv3(MRKWID*cos(w),n1),sxv3(MRKWID*sin(w),n2))));
	  AddMD(G,x,Vec[1][nr]->x,vec_color);
	}
      }
      CloseGfxObject(G);
    }else r=(Dspobj *)NULL;
  }
  return(r);
}
/*-------------------- vector_unmark --------------------*/

static void vector_unmark(int nr)
{
  if (nr<0||nr>NV-1) {
    return;
  }else{
    if (is_dsp(Mrk[nr])) del_node((Node *)Mrk[nr]);
    Mrk[nr]=(Dspobj *)NULL;
  }
}
/*-------------------- cmd_caveat_vector_display --------------------*/

static void cmd_caveat_vector_display(void)
{
  Token *key;
  int i;

  key=Get_Kwd("vector display status: ",";","on off");
  if (key->status==TOK_OK) {
    vec_dsp=key_is("on");
    if (NV>0) {
      lock_evtq();
      if (vec_dsp) {
	for (i=0;i<NV;i++) Mrk[i]=vector_mark(i);
      }else{
	for (i=0;i<NV;i++) vector_unmark(i);
      }
      unlock_evtq();
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_caveat_vector_color --------------------*/

static void cmd_caveat_vector_color(void)
{
  Token *key;
  int i,new;

  key=Get_Color("vector color: ",";");
  if (key->status == TOK_OK) {
    new=key->n;
    if (new!=vec_color) {
      vec_color=new;
      lock_evtq();
      for (i=0;i<NV;i++) {
	if (is_dsp(Mrk[i])) {
	  vector_unmark(i);
	  Mrk[i]=vector_mark(i);
	}
      }
      unlock_evtq();
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad color token: %s",key->text);
  }
}
/*-------------------- cmd_caveat_vector_delete --------------------*/

static void cmd_caveat_vector_delete(void)
{
  int i;
  Token *key;
  Dspobj *proot;
  Atom *base,*tip;

  key=(Token *)NULL;

  proot=pick_root(assocmdl(Template));
  while (NV>0) {
    err_dump(); send_bheader(graphix,"ARMK",0);
    key=free_Token(key);
    key=Get_Atom("base-point of vector: ",";");
    if (key->status==TOK_OK) {
      base=(Atom *)key->ptr;
      key=free_Token(key);
      key=Get_Atom("tip of vector: ",";");
      if (key->status==TOK_OK) {
	tip=(Atom *)key->ptr;
	for (i=0;i<NV;i++) {
	  if (Vec[0][i]==base && Vec[1][i]==tip) break;
	}
	if (i<NV) {
	  caveat_vector_delete(i);
	}else{
	  err_set(1,"vector is not in list.");
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad tip: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad base point: %s",key->text);
    }else{
      break;
    }
  }
  pick_root(proot);
  send_bheader(graphix,"ARMK",0);

  prtGraphics("vector list contains %d vectors.\n",NV);
  if (NV==0) err_set(1,"vector list is empty.");
  free_Token(key);
}
/*-------------------- caveat_vector_delete --------------------*/

static void caveat_vector_delete(int n)
{
  int i;

  if (n>=0&&n<NV) {
    vector_unmark(n);
    for (i=n;i<NV-1;i++) {
      Vec[0][i]=Vec[0][i+1];
      Vec[1][i]=Vec[1][i+1];
      Mrk[i]=Mrk[i+1];
    }
    NV--;
  }
}
/*-------------------- cmd_caveat_vector_tolerance --------------------*/

static void cmd_caveat_vector_tolerance(void)
{
  int i;
  Token *key;
  char *prmt[2]={"base-base distance (A) ","angle at base (deg) "};
  char *tol_menu[2]={
    "0.00 0.01 0.02 0.03 0.04 0.05 0.10 0.15 0.20 0.25 0.30 0.35 0.40 0.45 0.50 0.75",
    "0 1 2 3 4 5 10 15 20 25 30 35 40 45 50 55 60"
  };

  key=(Token *)NULL;
  for (i=0;i<2;i++) {
    err_dump();
    key=free_Token(key);
    default_token("%0.3f",Tol[i]);
    key=Get_Float_M(prmt[i],";",tol_menu[i]);
    if (key->status==TOK_OK) {
      if (key->f>=0) {
	Tol[i]=key->f;
      }else{
	err_set(1,"tolerance must be a non-negative");
	i--;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad value for tolerance: %s",key->text);
      i--;
    }else{
      break;
    }
  }
  free_Token(key);
}
/*-------------------- cmd_caveat_vector_write --------------------*/

static void cmd_caveat_vector_write(void)
{
  Token *key;
  FILE *out;
  int i,j,k;
  char buff[4096];

  key=(Token *)NULL;

  if (NV>0) {
    key=Get_String("vector file: ",";");
    if (key->status==TOK_OK) {
      out=fopen(extfil(buff,key->text,".vec",FALSE),"w");
      if (out) {
	fprintf(out,"# Caveat vector list  written by Wit!P");
	for (i=0;i<NV;i++) {
	  fprintf(out,"\n%db   %8.3f  %8.3f  %8.3f",i+1,
		  Vec[0][i]->x[0],Vec[0][i]->x[1],Vec[0][i]->x[2]);
	  fprintf(out,"\n%dt   %8.3f  %8.3f  %8.3f",i+1,
		  Vec[1][i]->x[0],Vec[1][i]->x[1],Vec[1][i]->x[2]);
	}
	for (k=0,i=0;i<NV-1;i++) {
	  for (j=1;j<4;j++) {
	    if (i+j < NV) {
	      if (k++%5 == 0) fprintf(out,"\n(%0.3f %0.3f %0.3f) ",Tol[0],Tol[1]*Pi/180,Tol[1]*Pi/180);
	      fprintf(out,"[%d %d] ",i+1,i+j+1);
	    }else break;
	  }
	}
	fprintf(out,"\n");
	fclose(out);
	prtGraphics("%d vectors written to %s\n",NV,key->text);
      }else{
	err_set(1,"too bad, could not write file %s",buff);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad filename for vector file: %s",key->text);
    }
  }else{
    err_set(1,"vector list is empty.");
  }

  free_Token(key);

}
/*-------------------- cmd_caveat_vector_read --------------------*/

static void cmd_caveat_vector_read(void)
{
  Token *key;
  FILE *in;
  UCpipe *In;
  int i,s,m,n,grain;
  char buff[1025],tok[1025];
  Atom *cand;
  float x[3],dist;
  Parse p;
  BHtree *T;
  Nodechain c;

  key=(Token *)NULL;

  c=tree2chain(Template,OBJ_ATOM);
  grain = ( c.n < MXTREE ? 1 : c.n / MXTREE );
  T=generateBHtree(&c,grain);
  clear_node_chain(&c);
  if (!mvld(T)) return;

  key=Get_File((char *)NULL,".vec","vector file: ",";");
  if (key->status==TOK_OK) {
    In=(UCpipe *)key->ptr;
    in=In->fd;
    if (in) {
      caveat_vector_clear();
      for (i=0;i<MAXVEC;i++) {
	Vec[0][i]=(Atom *)NULL;
	Vec[1][i]=(Atom *)NULL;
      }

      p.tok=tok;p.buf=buff;p.sep="#()bt";p.mts=1024;
      
      for(;;) {
	if (!fgets(buff,1024,in)) break;;
	p.pos=0;
	if (!get_token(&p)) continue;
	if (p.term=='b' || p.term=='t') {
	  if (p.term=='b') s=0;
	  else s=1;
	  if (p.ltok<1) {
	    err_set(1,"bad vector definition read from file:");
	    err_set(1,"%s",buff);
	    break;
	  }
	  n=str2i(p.tok);
	  if (n<1||n>MAXVEC) {
	    err_set(1,"vector number out of range:");
	    err_set(1,"%s",buff);
	    break;
	  }
	  for (i=0;i<3;) {
	    if (!get_token(&p)) break;
	    x[i++]=str2f(p.tok);
	    if (p.term=='#') break;
	  }
	  if (i<3) {
	    err_set(1,"missing coordinate in vector definition:");
	    err_set(1,"%s",buff);
	  }else{
	    m=findBHcloseAtoms(T,x,0.1,&cand,&dist,1);
	    if (m>0) {
	      Vec[s][n-1]=cand;
	    }else{
	      err_set(1,"atom not found in template:");
	      err_set(1,"%s",buff);
	    }
	  }
	}
      }
      for (i=0;i<MAXVEC;i++) {
	if (is_atom(Vec[0][i])&&is_atom(Vec[0][i])) {
	  caveat_vector_add(Vec[0][i],Vec[1][i]);
	}
      }
      prtGraphics("%d vectors read from %s\n",NV,key->text);
    }
    UCpclose(In);
  }else if (key->status==TOK_NOK) {
    err_set(1,"too bad, could not read file %s",key->text);
  }

  free_Token(key);
  freeBHtree(T);
}
/*-------------------- caveat_vector_clear --------------------*/

static void caveat_vector_clear(void)
{
  int i;

  if (NV>0) {
    lock_evtq();
    for (i=NV-1;i>=0;i--) caveat_vector_delete(i);
    unlock_evtq();
  }
}
/*------------------------------------------------------------------------*/

static void open_browser(void)
{
  Token *key;
  char buf[512];

  key=(Token *)NULL;
  
  if (mvld(Hits)) {
    sprintf(buf,"close file %s ? ",Hits->filename);
    default_token("yes");
    for (;;) {
      key=free_Token(key); err_dump();
      key=Get_Kwd(buf,";","no yes");
      if (key->status==TOK_OK) {
	if (key_is("yes")) {
	  UCpclose(Hits); Hits=(UCpipe *)NULL;
	  if (mvld(cache)) cache=freeFileCache(cache);
	}
	break;
      }else if (key->status==TOK_NOK) {
	err_set(1,"please answer yes or no.");
      }else break;
    }
  }
  if (!mvld(Hits)) {
    key=free_Token(key);
    key=Get_File((char *)NULL,".sdf","File: ",";");
    if (key->status==TOK_OK) {
      Hits=(UCpipe *)key->ptr;
      prev_mol=(Mol *)NULL;
      if (Hits->pipe==UCP_FILE) cache=makeFileCache(250,25);
      nseq=1;
    }else if (key->status==TOK_NOK) err_set(1,"could nor read hits from %s",key->text);
  }
  free_Token(key);
}
/*-------------------- read_hit --------------------*/

static void read_hit(int nr)
{
  char mol_name[512];
  Node *mol;
  Dspobj *global,*droot;

  if (!mvld(Hits)) return;
  nseq=skip_maccs(Hits,cache,nr,nseq);
  if (feof(Hits->fd)) {
    if (!mvld(cache)) {
      err_set(1,"reading past end of file.");
      UCpclose(Hits); 
      Hits=(UCpipe *)NULL;
    }else{
      err_set(1,"reading past end of file.");
      err_set(1,"last entry is #%d",nseq-1);
    }
    return;
  }

  sprintf(mol_name,"!MOL%d",nr);
  mol=add_node(MolRoot,mol_name,OBJ_MOL);
  if (mvld(mol)) {
    if (read_maccs_mol(Hits,mol)>0) {
      del_node(mol); mol=(Node *)NULL;
      err_set(1,"Failed to read molecule. I'm confused...");
    }
    if (feof(Hits->fd)) {
      if (!mvld(cache)) {
	UCpclose(Hits); 
	Hits=(UCpipe *)NULL;
      }
      del_node(mol); mol=(Node *)NULL;
    }
    if (mvld(mol)) {
      prtGraphics("molecule %s read from %s\n",mol->name,Hits->filename);
      droot=glblobj(assocobj(MolRoot));
      global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
      if (mvld(global)) {
	send_mdl(global,(Mol *)mol);
	global->root=mol;
      }
      ((Mol *)mol)->dsptr=global;
      nseq=nseq+1;
      prev_mol=(Mol *)mol;
    }
  }
}
/*-------------------- graft_sidechain --------------------*/

static int graft_sidechain(void)
{
  Token *key;
  Dspobj *proot;
  Node *core=(Node *)NULL;
  Nodechain *nc;
  Atom *a;
  char *menu,vecmenu[1024];
  int i,nr,r;

  key=(Token *)NULL;
  menu=(char *)NULL;
  r=FALSE;

  if (NV<1) {
    err_set(1,"you must define vectors before grafting.");
  }else if (!mvld(Template)) {
    err_set(1,"you have no template.");
  }else{
    sprintf(vecmenu,"all ");
    for (i=0;i<NV;i++) sprintf(vecmenu+strlen(vecmenu)," vec%d",i+1);
    proot=pick_root(assocmdl(Template)); pick_mode("PATM");
    key=Get_Kwd("vector: ",";",vecmenu);
    send_bheader(graphix,"ARMK",0); pick_root(proot); pick_mode("POFF");
    if (key->status==TOK_NOK) {
      key=Prs_Atom(key);
      if (key->status==TOK_OK) {
	a=(Atom *)key->ptr;
	for (i=0;i<NV;i++) if (Vec[0][i]==a||Vec[1][i]==a) break;
	if (i<NV) {
	  key->n=i+1;
	}
      }
    }else if (key->status==TOK_OK) {
      if (key_is("all")) key->n=0;
      else key->n=str2i(key->key+3);
    }
    if (key->status==TOK_NOK) {
      err_set(1,"not a known vector: %s",key->text);
    }
    if (key->status==TOK_OK) {
      nr=key->n;
      key=free_Token(key); free_menu(menu);
      menu=Make_TopMolNameList("",DefaultNode);
      key=Get_Mols_M("core: ",";",menu);
      send_bheader(graphix,"ARMK",0);
      free_menu(menu);
      if (key->status==TOK_OK) {
	nc=(Nodechain *)key->ptr;
	if (nc->n!=1) key->status=TOK_NOK;
	else core=nc->s;
	clear_node_chain(nc); free_node_chain(nc);
      }
      if (key->status==TOK_NOK) {
	err_set(1,"bad core selection: %s",key->text);
      }else if (key->status==TOK_OK) {
	if (nr==0) for(i=0;i<NV;i++) r|=graft_one_sidechain(core,i);
	else r|=graft_one_sidechain(core,nr-1);
      }
    }
  }
  return(r);
}
/*-------------------- graft_one_sidechain --------------------*/

static int graft_one_sidechain(Node *core,int nr)
{
  Token *key;
  Atom **new,*a,*b;
  Bond *bnd,*nb,*ob;
  Nodechain *sc,*nc;
  int i,k,r;
  float d,dm;

  sc=(Nodechain *)NULL;
  r=FALSE;
  key=(Token *)NULL;
  
  if (nr<0 || nr>=NV) {
    err_set(1,"vector number out of range: %d",nr);
  }else{
    sc=connected_to_right(Vec[0][nr],Vec[1][nr]);
    if (mvld(sc)) {
      if (sc->n>0) {
	nc=make_node_chain();
	if (nc) {
	  nc->n=0; nc->s=(Node *)NULL; nc->p=nc->s; nc->e=nc->s;
	  for (i=0,a=(Atom *)sc->s;mvld(a);a=(Atom *)a->chain,i++) {
	    a->seqnr=i; a->flg|=(1<<1);
	    b=(Atom *)copy_node((Node *)a);
	    if (mvld(b)) {
	      add_to_node_chain((Node *)b,nc);
	    }else{
	      err_set(1,"failed to copy atoms");
	    }
	  }
	  if (nc->n==sc->n) {
	    new=(Atom **)chain_to_ptr(nc,TRUE); nc=(Nodechain *)NULL;
	    if (mvld(new)) {
	      for (i=0;i<sc->n;i++) {
		new[i]->up=core;
		link_node((Node *)new[i]);
		new[i]->flg&=~(1<<1);
	      }
	      for (a=(Atom *)sc->s;mvld(a);a=(Atom *)a->chain) {
		for (i=0;i<a->nbnd;i++) {
		  b=neighbour(i,a);
		  if (b->flg&(1<<1)) {
		    nb=add_bond_ng(new[a->seqnr],new[b->seqnr]);
		    if (mvld(nb)) {
		      nb->bt=a->bnd[i]->bt;
		      nb->bo=a->bnd[i]->bo;
		      nb->arom=a->bnd[i]->arom;
		    }
		  }
		}
	      }
	      bnd=find_bond(Vec[0][nr],Vec[1][nr]); ob=(Bond *)NULL;
	      if (mvld(bnd)) {                 /* if group is connected in template */
		dm=10*Tol[0]; a=(Atom *)NULL;  /* try to connect group to core...   */
		ForAtoms(b,core) {
		  if (b->nxyz>0) {
		    d=cdist(b->x,Vec[0][nr]->x);
		    if (d<dm) {
		      dm=d;
		      a=b;
		    }
		  }
		}
		if (mvld(a)&&dm<Tol[0]) {      /* ...but only if core base atom found */
		  dm=10*Tol[1]; k=-1;          /* check for near-parallel bond  */
		  for (i=0;i<a->nbnd;i++) {
		    b=neighbour(i,a);
		    if (b->nxyz>0) {
		      d=cangle(Vec[1][nr]->x,a->x,b->x);   /* in radians */
		      if (d<dm) {
			dm=d;
			k=i;
		      }
		    }
		  }
		  if (dm<Tol[1] && k>=0) {     /* if core has bond ~parallel to new bond */
		    b=neighbour(k,a);          /* remove sidechain (if bond is not in ring) */
		    nc=connected_to_right(a,b);
		    if (mvld(nc)) {
		      if (nc->n>0) {
			del_node_chain(nc);
			ob=find_bond(a,b);
		      }
		      free_node_chain(nc); nc=(Nodechain *)NULL;
		    }
		  }else{
		    err_set(1,"warning: sidechain vec%d grafted, but not connected to core.",nr+1);
		  }
		  nb=add_bond_ng(a,new[0]);
		  if (mvld(nb)) {
		    nb->bt=bnd->bt;
		    nb->bo=bnd->bo;
		    nb->arom=bnd->arom;
		    if (mvld(ob)) {
		      if (nb->bt!=ob->bt) {
			err_set(1,"warning: bond type in core differs from type in template (vec%d).",nr+1);
		      }
		    }
		  }		  
		}
	      }
	      nc=ptr_to_chain((Node **)new,sc->n); free(new);
	      mdl_add_atom_chain(assocmdl(core),nc);
	      clear_node_chain(nc); free_node_chain(nc);
	      r=TRUE;
	    }else{
	      err_set(1,"memory allocation failure.");
	    }
	  }else{
	    del_node_chain(nc);
	    free_node_chain(nc);
	  }
	  for (a=(Atom *)sc->s;mvld(a);a=(Atom *)a->chain) a->flg&=~(1<<1);
	}else{
	  err_set(1,"memory allocation failure.");
	}
      }else{
	err_set(1,"sorry, vector does not determine sidechain.");
      }
    }else{
      err_set(1,"could not determine right-connected part");
    }
  }
  if (mvld(sc)) {
    clear_node_chain(sc);
    free_node_chain(sc);
    sc=(Nodechain *)NULL;
  }
  free_Token(key);
  return(r);
}
/*-------------------- sanity_check --------------------*/

static int sanity_check(void)
{
  int i,r,nf[MAXVEC],nt;
  Mol *m;
  Atom *a;

  r=TRUE;

  ForMols(m,MolRoot) if (m==(Mol *)Template) break;
  if (is_mol(m)) {
    if (NV>0) {
      for (i=0;i<NV;i++) {nf[i]=0;} nt=0;
      ForAtoms(a,m) {
	for (i=0;i<NV;i++) {
	  if (a==Vec[0][i]) {nf[i]++; nt++;}
	  if (a==Vec[1][i]) {nf[i]++; nt++;}
	}
      }
      if (nt<2*NV) {
	lock_evtq();
	for (i=NV-1;i>=0;i--) if (nf[i]<2) caveat_vector_delete(i);
	unlock_evtq();
      }
    }
  }else{
    caveat_vector_clear();
    Template=(Node *)NULL;
    err_set(1,"sorry, lost the template.");
    get_template();
    r=is_mol(Template);
  }
  return(r);
}
