#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;

#define key_is(x) (streq(key->key,x))

#define DONE  0
#define BASE  101
#define PARA  102
#define TARG  103
#define START 1
#define FIRST 2

#define CHAIN 0
#define PREV  1
#define THREE 2

static char *base_menu = {"~chain *previous"};
static char *base_default[2] = {
  "~","*" 
  };
static char *prmpt[3] = {"distance = ","angle = ","dihedral = "};
static char *pmen[3] = {
  "'1.00' '1.05' '1.08' '1.32' '1.39' '1.54'",
  "'109.5' '120.0'",
  "'    0' ' 180', '+120' '-120' ' +60' ' -60' ' +30' ' -30'"
  };

void cmd_construct(void)
{
/* add atom via cnstr ... */

  Token *key;
  float p[3],u[3],D,d;
  Node *ctxt,*root;
  Atom *a[3],*atom,*last;
  char prompt[128],*name;
  int i,na=0,np=0;
  int state,mode;

  state=START;
  key=(Token *)NULL;
  ctxt=(Node *)NULL;
  for (i=0;i<4-1;i++) a[i]=(Atom *)NULL;
  last=(Atom *)NULL;
  for (i=0;i<3;i++) u[i]=0.0;
  p[0]=1.08; p[1]=109.5; p[2]=180.0;
  mode=THREE;

  while (state!=DONE) {
    err_dump();
    if (state==START) {
      na=0; np=0; send_bheader(graphix,"ARMK",0);
      if (mvld(key)) key=free_Token(key);
      state=FIRST;
    }
    if (!(mvld(key))) {
      switch(state) {
      case FIRST:
	if (mode!=THREE) {
	  default_token(base_default[mode]);
	  sprintf(prompt,"atom 1");
	}else{
	  sprintf(prompt,"atom 1: ");
	}
	pick_mode("PATM");
	key=Get_String_M(prompt,";",base_menu);
	pick_mode("POFF");
	break;
      case BASE:
	pick_mode("PATM");
	if (na==0&&mode!=THREE&&mvld(last)) {
	  default_token(base_default[mode]);
	  sprintf(prompt,"atom %d",na+1);
	}else{
	  sprintf(prompt,"atom %d: ",na+1);
	}
	key=Get_String(prompt,";");
	pick_mode("POFF");
	break;
      case PARA:
	if (np==0) default_token("%6.3f",p[np]);
	else default_token("%6.1f",p[np]);
	key=Get_String_M(prmpt[np],";",pmen[np]);
	break;
      case TARG:
	key=Get_String("new name: ",";");
	break;
      }
    }
    if (mvld(key)) {
      switch (state) {
      case FIRST:
	key=Prs_Kwd(key,"~chain *previous");
	if (key->status==TOK_OK) {
	  if (key_is("~chain")) {
	    mode=CHAIN;
	    if (mvld(last)) {
	      a[0]=a[1]; a[1]=a[2]; a[2]=last;
	      ctxt=(Node *)last;
	      na=3;
	      state=PARA;
	      key=free_Token(key);
	    }else{
	      state=BASE;
	      key=free_Token(key);
	    }
	  }else if (key_is("*previous")) {
	    mode=PREV;
	    if (mvld(last)) {
	      na=3;
	      ctxt=(Node *)a[2];
	      state=PARA;
	      key=free_Token(key);
	    }else{
	      state=BASE;
	      key=free_Token(key);
	    }
	  }else{
	    state=BASE;
	  }
	}else{
	  state=BASE;
	}
	break;
      case BASE:
	key=Prs_AtomCtxt(key,ctxt);
	if (key->status==TOK_OK) {
	  a[0]=a[1]; a[1]=a[2]; a[2]=(Atom *)key->ptr;
	  ctxt=(Node *)a[2];
	  if ( ++na > 2) state=PARA;
	  key=free_Token(key);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"atom not found: %s",key->text);
	  key=free_Token(key);
	  end_parse();
	}else{
	  state=DONE;
	}
	break;
      case PARA:
	key=Prs_Float(key);
	if (key->status==TOK_OK) {
	  u[np]=key->f;
	  if ( ++np > 2) state=TARG;
	  key=free_Token(key);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad floating point number: %s",key->text);
	  key=free_Token(key);
	  end_parse();
	}else{
	  state=DONE;
	}
	break;
      case TARG:
	end_parse();
	if (np==0) {
	  err_set(1,"FUGU: no parameters");
	  state=START;
	}else{
	  root=(Node *)NULL;
	  name=strrchr(key->text,'/');
	  if (name) {
	    *name++='\0';
	    if (mvld(ctxt)) {
	      root=node_in_ctxt(key->text,ctxt->up,strcasecmp);
	    }else{
	      root=node_in_ctxt(key->text,ctxt,strcasecmp);
	    }
	  }else{
	    name=key->text;
	    if (mvld(ctxt)) root=ctxt->up;
	  }
	  /*Note: should check the root ! */
	  for (i=0;i<np;i++) p[2-i]=u[np-i-1];
	  atom=(Atom *)make_node(OBJ_ATOM);
	  if (mvld(atom)) {
	    set_node_name((Node *)atom,name);
	    if (!mvld(atom->atp)) set_atom_type(atom,guess_atom_type(name));
	    atom->up=root;
	    link_node((Node *)atom);
	    cnstr(atom->x,a[0]->x,a[1]->x,a[2]->x,p[0],p[1],p[2]);
	    atom->nxyz=1;
	    if (mvld(atom->atp)&&mvld(a[2]->atp)) {
	      D=cdist(atom->x,a[2]->x);
	      d=atom->atp->rcov+a[2]->atp->rcov;
	      if ((D-0.4 < d) && (d < D+0.4)) set_bt(add_bond_raw(a[2],atom),BT_SINGLE);
	    }
	    if (a[2]->color<0) atom->color=-atom->color;
	    send_one_atom(atom);
	    last=atom;
	  }
	  send_bheader(graphix,"ARMK",0);
	  key=free_Token(key);
	  state=START;
	}
	break;
      default:
	break;
      }
    }
  }
  send_bheader(graphix,"ARMK",0);
  if (mvld(key)) free_Token(key);
}
