#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "swdef.h"
#include "transpor.h"
#include "parser.h"
#include "mystdlib.h"
#include "object.h"
#include "errhand.h"
#include "bhtree.h"
#include "graphics.h"
#include "gcalc.h"
#include "list.h"
#include "agglom.h"

#define tFLG (1<<3)

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

static int pipenr=0;

extern Node *MolRoot,*DefaultNode;
extern Dspobj *DspRoot;
extern int graphix;
extern int server;

#define TRUE 1
#define FALSE 0

#define UNDEFINED 0
#define FORWARD  1
#define BACKWARD 2
#define LOOP     3

#define CL_METH_OTHER  0
#define CL_METH_AGGLOM 1


#define CONFA_SORT_ENERGY 1
#define CONFA_SORT_SEQNR  2

#define MAXBUF 2048

#define key_is(x) (streq(x,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

static char *main_menu = "add animate average backtransform browse cluster cotransform dcd fit "
                         "mmread priax purge read restore sort "
                         "status template trace variable write ' ' "
                         "Command ' ' Color CPK Delete Display Label Measure reset select Set ' '"
                         "End";

static char *browse_menu = "first last previous next goto ' ' "
                         "accept copy remove restore write end ' ' "
                         "Command ' ' Color CPK Delete Display Label Measure reset select Set";

static char *cluster_menu = "find info list method metric reduce reset '' end";

static char *cl_method_name[] = {"leader","","average","weighted","flexible","","single","complete","","centroid", "median","ward"};
static int cl_method_procedure[] = {CL_METH_OTHER,0,CL_METH_AGGLOM,CL_METH_AGGLOM,CL_METH_AGGLOM,0,CL_METH_AGGLOM,CL_METH_AGGLOM,0,CL_METH_AGGLOM,CL_METH_AGGLOM,CL_METH_AGGLOM};
static char *cl_methods="leader '' average weighted flexible '' single complete '' centroid median ward";
static int  method=0;

static char *cl_metric_name[] = {"inplace","fitted"};
static char *cl_metrics="inplace fitted";
static int  metric=1;

static char *cl_metopt_name[] = {"gseal","rmsd"};
static char *cl_metopts="gseal rmsd";
static int metopt=1;

static int *SubsetIdx=NULL;
static int nSubset=0;

int confa_sort_cmp(const void *a,const void *b);
static int cmp_int_asc(const void *A,const void *B);

static void cmd_confa_add(void);
static void cmd_confa_accept(void);
static void cmd_confa_template(void);
static void cmd_confa_mmread(void);
static void cmd_confa_dcd(void);
static void cmd_confa_read(void);
static void cmd_confa_copy(void);
static void cmd_confa_write_conformations(void);
static void cmd_confa_write_variables(void);
static void cmd_confa_write(void);
static void cmd_confa_animate(void);
static void cmd_confa_average(void);
static void cmd_confa_browse(void);
static void cmd_confa_fit(void);
static void cmd_confa_variable(void);
static void cmd_confa_trace(void);
static void cmd_confa_variable_add(void);
static void cmd_confa_variable_delete(void);
static void confa_set_conf(int seqnr,CnfNode *c);
static void cmd_confa_purge(void);
static void cmd_confa_cotransform(char *mode);
static void confa_sort(int key);
static void confa_reduce(void);
static void confa_priax(void);
static void confa_status(void);
static void cmd_confa_cluster(void);
static void cmd_confa_cluster_find(void);
static void confa_cluster_list(void);
static void confa_cluster_info(void);
static void confa_cluster_leader(int N,CnfNode **C,float rho);
static void confa_cluster_agglom(int N,CnfNode **C);
static CnfNode *save_conformation(int N,Atom **A);
static void init_confa_var(void *data);
static void del_confa_var(void *data);
static int  sanity_check(void);
int    get_template_atom(char *prmt);
static Nodechain *getAtoms(CnfNode *c);
static int getCoords(CnfNode *c,float *x);
static float *getAtomCoords(CnfNode *c,int n);
static Token *get_trace(char *prmt);
static CnfNode *newCnf(char *name,int size);
static void confa_dcd_read(UCpipe *trj,int from,int to,int step);
static int nTemplateOK(int n,Atom **T);
static int cmp_addr(const void *A,const void *B);
static int cmp_addrp(const void *A,const void *B);
static float Distance(CnfNode *a,CnfNode *b);
static float fittedDistance(CnfNode *a,CnfNode *b);
static float inplaceDistance(CnfNode *a,CnfNode *b);
static float cl_cluster_gseal(float *x, float *y);

static int ptr2seq(CnfNode *cnf);
static CnfNode *seq2ptr(int nr);

static int in_use=FALSE;
static Node *ConfaRoot=(Node *)NULL;
static Atom **Template;
static int  NTemplate;
static int *Mobile;
static int NMobile;
static int confa_sort_key;
static CnfNode *svdCnf=(CnfNode *)NULL;
static CnfNode *setCnf=(CnfNode *)NULL;
static int nxtSeq=0;
static float sv_inter=0.1;
static int sv_dir=FORWARD;
static int autoreverse=1;
static float gseal_alpha=0.7;

#define VAR_NONE      0
#define VAR_ENERGY    1
#define VAR_TIME      2
#define VAR_DISTANCE  3
#define VAR_ANGLE     4
#define VAR_TORSION   5
#define VAR_FITRMSD   6

typedef struct CnfVar {
  char *name;
  void *data;
  int type;
}CnfVar;

static ListEntry Vars;

static int ptr2seq(CnfNode *cnf)
{
  int nr;
  CnfNode *c;

  nr=0;
  forCnfs(c,ConfaRoot) {
    nr++;
    if (c==cnf) break;
  }
  if (!c) nr=0;
  return(nr);
}

static CnfNode *seqt2ptr(char *text)
{
  CnfNode *cnf=NULL;
  int n,since;

  if (text && text[0]=='#') text=text+1;
  since=err_since();
  n=str2i(text);
  if (since==err_since()) {
    cnf=seq2ptr(n);
  }else{
    err_pop(err_since()-since);
  }
  return(cnf);
}

static CnfNode *seq2ptr(int nr)
{
  CnfNode *cnf=NULL;
  
  if (nr>0) {
    forCnfs(cnf,ConfaRoot) if (--nr==0) break;
  }
  return(cnf);
}

/*==================== Conformational Analysis ====================*/

void cmd_confa(void)
{
  Token *key;

  if (in_use) {
    err_set(1,"sorry, ConfA may not be called recursiveley.");
    return;
  }

  key=(Token *)NULL;

  Vars.prv=(ListEntry *)NULL;
  Vars.nxt=(ListEntry *)NULL;
  Template=(Atom **)NULL;
  NTemplate=0;
  Mobile=(int *)NULL;
  NMobile=0;
  svdCnf=(CnfNode *)NULL;
  setCnf=svdCnf;
  in_use=TRUE;
  nxtSeq=0;

  for (;;) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("Confa> ",";",main_menu);
    if (key->status==TOK_OK) {
      if (key_is("template")) cmd_confa_template();
      else if (key_is("mmread")) cmd_confa_mmread();
      else if (key_is("dcd")) cmd_confa_dcd();
      else if (key_is("read")) cmd_confa_read();
      else if (key_is("add")) cmd_confa_add();
      else if (key_is("animate")) cmd_confa_animate();
      else if (key_is("average")) cmd_confa_average();
      else if (key_is("trace")) cmd_confa_trace();
      else if (key_is("browse")) cmd_confa_browse();
      else if (key_is("cluster")) cmd_confa_cluster();
      else if (key_is("fit")) cmd_confa_fit();
      else if (key_is("cotransform")) cmd_confa_cotransform("co");
      else if (key_is("backtransform")) cmd_confa_cotransform("back");
      else if (key_is("status")) confa_status();
      else if (key_is("priax")) confa_priax();
      else if (key_is("purge")) cmd_confa_purge();
      else if (key_is("restore")) {
	confa_set_conf(0,svdCnf);
	setCnf=(CnfNode *)NULL;
      }else if (key_is("sort")) confa_sort(CONFA_SORT_ENERGY);
      else if (key_is("variable")) cmd_confa_variable();
      else if (key_is("write")) cmd_confa_write();
      else if (key_is("CPK")) cmd_cpk();
      else if (key_is("Color")) cmd_color();
      else if (key_is("Delete")) {cmd_delete();if (!sanity_check()) break;}
      else if (key_is("Display")) cmd_display();
      else if (key_is("Label")) cmd_label();
      else if (key_is("Measure")) cmd_measure();
      else if (key_is("reset")) cmd_display_reset();
      else if (key_is("select")) cmd_select();
      else if (key_is("Command")) {
	command_loop("Wit!P(Confa)> ");
	if (!sanity_check()) break;
      }else if (key_is("Set")) cmd_set();
      else if (key_is("End")) break;              /* should we ask to confirm ? */
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad Confa keyword: %s",key->text);
    }
    if (end_parse()!=TOK_OK) err_set(1,"extra characters a end of command, ignored.\n");
    err_dump();
    start_parse();
  }

  del_node(ConfaRoot);
  delete_list_entries(&Vars,del_confa_var);
  ConfaRoot=(Node *)NULL;
  svdCnf=(CnfNode *)del_node((Node *)svdCnf);
  setCnf=(CnfNode *)NULL;
  in_use=FALSE;
  if (mvld(Template)) free(Template);
  if (mvld(Mobile)) free(Mobile);
  free_Token(key);
}
/*-------------------- cmd_confa_template --------------------*/

static void cmd_confa_template(void)
{
  Token *key;
  Nodechain *c;
  Atom *a;
  int i;

  key=(Token *)NULL;

  key=Get_Atoms_D("template atoms: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (mvld(svdCnf)) del_node((Node *)svdCnf);
    if (mvld(Template)) free(Template);
    Template=(Atom **)NULL;
    NTemplate=c->n;
    if (NTemplate>0) {
      Template=(Atom **)malloc((NTemplate+1)*sizeof(Atom *));
      if (mvld(Template)) {
	for (i=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain,i++) Template[i]=a;
	Template[NTemplate]=(Atom *)NULL;
	walk_order(MolRoot,NTemplate,Template,(tFLG));
	svdCnf=save_conformation(NTemplate,Template);
      }else{
	err_set(1,"failed to allocate memory for template atoms.");
	NTemplate=0;
      }
    }else{
      err_set(1,"zero template atoms.");
    }
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad template atom selection: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_confa_add --------------------*/

static void cmd_confa_add(void)
{
  Token *key=(Token *)NULL;
  Nodechain *c;
  float *x;
  int i,j,nc=0;
  CnfNode *new,*sv;

  if (!mvld(ConfaRoot)) ConfaRoot=make_node(OBJ_CNF);
  if (!mvld(ConfaRoot)) {
    err_set(1,"Failed to generate root for Confa.");
    return;
  }
  set_node_name(ConfaRoot,"ConfaRoot");

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)) {
    err_set(1,"no template atoms selected...");
    return;
  }
  c=ptr_to_chain((Node **)Template,NTemplate);
  if (c) {
    if (cmd_read_coordinates(c)) {
      clear_node_chain(c);
      sv=setCnf; setCnf=(CnfNode *)NULL;
      forCnfs(new,ConfaRoot) nc++;
      new=(CnfNode *)add_node(ConfaRoot,(char *)NULL,OBJ_CNF);
      if (new) {
	new->x=(float *)malloc(3*NTemplate*sizeof(float));
	if (new->x) {
	  for (x=new->x,i=0;i<NTemplate;i++) {
	    for (j=0;j<3;j++) *x++=Template[i]->x[j];
	  }
	  new->n=NTemplate;
	  key=free_Token(key);
	  default_token("%d",nc+1);
	  key=Get_String("conformation id: ",";");
	  if (key->status==TOK_OK) {
	    set_node_name((Node *)new,key->text);
	    key=free_Token(key);
	    default_token("0.0");
	    key=Get_Float("energy: ",";");
	    if (key->status==TOK_OK) {
	      new->energy=key->f;
	      new->cseqnr=++nxtSeq;
	      confa_set_conf(0,new);
	    }else{
	      if (key->status==TOK_NOK) err_set(1,"bad energy: %s\n",key->text);
	      new=(CnfNode *)del_node((Node *)new);
	    }
	  }else{
	    if (key->status==TOK_NOK) err_set(1,"bad conformation id: %s\n",key->text);
	    new=(CnfNode *)del_node((Node *)new);
	  }
	}else{
	  err_set(1,"failed to allocate memory for new coordinates\n");
	  new=(CnfNode *)del_node((Node *)new);
	}
      }else err_set(1,"failed to create node for new conformation.\n");
      if (!setCnf) {
	if (sv) confa_set_conf(0,sv);
	else confa_set_conf(0,svdCnf);
	setCnf=sv;
      }
    }
    clear_node_chain(c);
    free_node_chain(c);
  }
  key=free_Token(key);
}
/*-------------------- cmd_confa_mmread --------------------*/

static void cmd_confa_mmread(void)
{
  char buff[MAXBUF];
  Token *key=(Token *)NULL;
  UCpipe *In=(UCpipe *)NULL;
  CnfNode *c;
  Atom **Atms=(Atom **)NULL,**atms=(Atom **)NULL;
  Node *mol;
  int status,nat,i,j;
  char *fn=(char *)NULL,pipename[120];
  float *p;

  key=Get_File((char *)NULL,".out","MMOD multi conf. file: ",";");
  if (key->status==TOK_OK) {

    In=(UCpipe *)key->ptr;
    fn=strdup(strrchr(key->text,'/')?strrchr(key->text,'/')+1:key->text);
    if (*In->filename=='|') {
      sprintf(pipename,"confa%d",++pipenr);
      fn=pipename;
    }

    /* do a reset of everything... */

    ConfaRoot=del_node(ConfaRoot);
    delete_list_entries(&Vars,del_confa_var); 
    svdCnf=(CnfNode *)del_node((Node *)svdCnf);
    setCnf=(CnfNode *)NULL;
    if (mvld(Template)) free(Template);
    Template=(Atom **)NULL;
    NTemplate=0;
    if (Mobile) free(Mobile);
    Mobile=(int *)NULL;
    NMobile=0;

    /* get a new confa root */

    ConfaRoot=make_node(OBJ_CNF);
    if (mvld(ConfaRoot)) {
      set_node_name(ConfaRoot,"ConfaRoot");
      mol=add_node(MolRoot,fn,OBJ_MOL);
      if (is_mol(mol)) {
	if ((status=read_mmod1(In,mol,&nat,&Atms))==0) {
	  send_molecule(MolRoot,(Mol *)mol); /* and post it for display.*/
      	  c=read_mmodcnf(In,nat,Atms,&atms);
	  if (mvld(c)) {
	    Template=atms; NTemplate=c->n;
	    c->up=ConfaRoot; link_node((Node *)c);       /* this is the 2nd conformation... */
	    c=(CnfNode *)make_node(OBJ_CNF);             /* generate first cnf node from molecule */
	    if (mvld(c)) {
	      c->n=NTemplate;
	      c->x=(float *)malloc(3*NTemplate*sizeof(float));
	      if (mvld(c->x)) {
		for (p=c->x,i=0;i<NTemplate;i++) {
		  for (j=0;j<3;j++) *p++ = Template[i]->x[j];
		}
		set_node_name((Node *)c,strsqz(buff,strfld(buff,mol->name,1,10)));
		c->energy=str2f(strfld(buff,mol->name+14,1,8))/4.186;
		c->nxt=ConfaRoot->fst; link_node((Node *)c);
		svdCnf=save_conformation(NTemplate,Template);
		svdCnf->energy=((CnfNode *)ConfaRoot->fst)->energy;
		while ((c=read_mmodcnf(In,nat,Atms,&atms))) {
		  c->up=ConfaRoot; link_node((Node *)c);
		  if (mvld(atms)) free(atms);    /* assuming always the same atoms... */
		  c->cseqnr=++nxtSeq;
		}
		confa_set_conf(0,(CnfNode *)ConfaRoot->fst);
	      }else {
		err_set(1,"failed to add memory for coordinates.");
		del_node((Node *)c);
		if (mvld(Template)) free(Template);
		Template=(Atom **)NULL; NTemplate=0;
		ConfaRoot=del_node(ConfaRoot);
	      }
	    }else {
	      err_set(1,"failed to add conformation.");
	      if (mvld(Template)) free(Template);
	      Template=(Atom **)NULL; NTemplate=0;
	      ConfaRoot=del_node(ConfaRoot);
	    }
	  }else err_set(1,"%s doesn't seem te be a multi conformer file.",In->filename);
	  set_node_name(mol,fn);  /* forget about the name in the file... */
	}else{
	  err_set(1,"failed to read macromode file %s, status=%d",In->filename,status);
	  del_node(mol);
	}
      }else err_set(1,"failed to add molecule.");
    }else{
      err_set(1,"Failed to generate root for Confa.");
    }
    UCpclose(In);
  }else if (key->status==TOK_NOK) {
    err_set(1,"too bad, could not read file %s",key->text);
  }

  if (Atms) free(Atms);
  if (fn) free(fn);
  key=free_Token(key);
  confa_status();
}
/*-------------------- cmd_confa_read --------------------*/

static void cmd_confa_read(void)
{
  int *mob,nmob;
  CnfNode *c;

  if (!mvld(ConfaRoot)) ConfaRoot=make_node(OBJ_CNF);
  if (!mvld(ConfaRoot)) {
    err_set(1,"Failed to generate root for Confa.");
    return;
  }

  set_node_name(ConfaRoot,"ConfaRoot");
  if ((mob=get_mobile_cnf(&nmob))) free(mob);
  cmd_read(ConfaRoot,FT_CNF);
  ForCnfs(c,ConfaRoot) if (c->cseqnr==0) c->cseqnr=++nxtSeq;
  mob=get_mobile_cnf(&nmob);
  if (mob) {
    if (Mobile) {
      if (nmob!=NMobile) {
	err_set(1,"new mobile atom list is inconsistent with old, new list ignored");
	free(mob);
      }
    }else{
      Mobile=mob;
      NMobile=nmob;
    }
  }
  err_dump();
  confa_status();
}
/*-------------------- cmd_confa_animate --------------------*/

static void cmd_confa_animate(void)
{
  Token *key;
  int dir,newdir,now,svnow,running;
  CnfNode *c,*svc;
  float inter;
  Qmsg *Msg;
  char *msgmask;
  char *topmenu="forward backward stop continue '' autoreverse speed '' goto first last previous next '' end";

  key=(Token *)NULL;
  
  if (!mvld(ConfaRoot)) {
    err_set(1,"no conformations have been read in yet...");
    return;
  }

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)) {
    err_set(1,"no template atoms selected...");
    return;
  }

  if (!mvld(ConfaRoot->fst)) {
    err_set(1,"no conformations found.");
    return;
  }

  if (!mvld(ConfaRoot->fst->nxt)) {
    err_set(1,"less than two conformations.");
    return;
  }

  now=0;
  forCnfs(c,ConfaRoot) {
    now++;
    if (c==setCnf) break;
  }
  if (!c) now=0;
  dir=FORWARD;
  Msg=(Qmsg *)NULL;
  running=FALSE;
  inter=0;

  for (;;) {
    if (running) {
      if (!c) {
	c=first_cnf(ConfaRoot);
	now=1;
      }else{
	newdir=(dir==FORWARD ? (mvld(c->nxt) ? FORWARD : BACKWARD) : (mvld(c->prv) ? BACKWARD : FORWARD));
	if (newdir!=dir) {
	  if (graphix&&autoreverse) dir=newdir;
	  else running=FALSE;
	}
	if (running) {
	  c=(CnfNode *)(dir==FORWARD ? c->nxt : c->prv);
	  now+=(dir==FORWARD?1:-1);
	}
      }
    }
    if (running) {
      confa_set_conf(now,c);
      send_header(graphix,"EVTQ",0);
      if (inter>=0) {
	syncqmsg((char *)NULL);
	if (inter>0) uSleep(1000000*inter);      /* minus time elapsed already */
      }
      sv_inter=inter;
      sv_dir=dir;
      fillmsgq(MSGQNOWAIT);
      msgmask="KEYB MENU SCCN USAP";
      Msg=nextqmsg(msgmask);
    }
    if (!running||Msg) {
      Msg=(Qmsg *)NULL;
      key=free_Token(key);
      key=Get_Kwd("animate> ",";",topmenu);
      if (key->status==TOK_OK) {
	if (key_is("forward")) {
	  dir=FORWARD;
	  running=TRUE;
	}else if (key_is("backward")) {
	  dir=BACKWARD;
	  running=TRUE;
	}else if (key_is("stop")) {
	  running=FALSE;
	}else if (key_is("continue")) {
	  inter=sv_inter;
	  dir=sv_dir;
	  running=TRUE;
	}else if (key_is("autoreverse")) {
	  key=free_Token(key);
	  default_token("%s",autoreverse?"on":"off");
	  key=Get_Kwd("mode: ",";","on off");
	  if (key->status==TOK_OK) autoreverse=key_is("on");
	  else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
	  key->status=TOK_OK;
	}else if (key_is("speed")) {
	  key=free_Token(key);
	  if (inter>=0) default_token("%0.3f",inter);
	  else default_token("async");
	  key=Get_Float_M("delay between frames (sec): ",";","async 0 0.1 0.2 0.3 0.4 0.5 1.0 2.0 3.0 4.0 5.0");
	  if (key->status==TOK_OK) inter=(key->f>=0?key->f:0);
	  else if (key->status==TOK_NOK) {
	    key=Prs_Kwd(key,"async");
	    if (key->status==TOK_OK) inter=-1;
	    else err_set(1,"bad token: %s",key->text);
	  }
	  key->status=TOK_OK;
	}else if (key_is("goto")) {
	  key=free_Token(key);
	  key=Get_String("goto (name or number): ",";");
	  if (key->status==TOK_OK) {
	    svnow=now; svc=c;
	    if (!key->quoted) c=seqt2ptr(key->text);
	    else c=(CnfNode *)NULL;
	    if (c) {
	      now=ptr2seq(c);
	      confa_set_conf(now,c);
	    }else{
	      now=0;
	      forCnfs(c,ConfaRoot) {
		now++;
		if (wild_xmatch(c->name,key->text)) break;
	      }
	      if (c) confa_set_conf(now,c);
	      else{
		err_set(1,"conformation %s not found.",key->text);
		now=svnow;
		c=svc;
	      }
	    }
	  }else if (key->status==TOK_NOK) err_set(1,"bad token %s",key->text);
	  key->status=TOK_OK;
	}else if (key_is("first")) {
	  c=(CnfNode *)ConfaRoot->fst;
	  now=1;
	  if (c) confa_set_conf(now,c);
	}else if (key_is("last")) {
	  now=0;
	  forCnfs(c,ConfaRoot) {
	    now++;
	    if (!c->nxt) break;
	  }
	  if (c) confa_set_conf(now,c);
	}else if (key_is("previous")) {
	  if (c&&c->prv) {
	    c=(CnfNode *)c->prv;
	    now--;
	    confa_set_conf(now,c);
	  }
	}else if (key_is("next")) {
	  if (c&&c->nxt) {
	    c=(CnfNode *)c->nxt;
	    now++;
	    confa_set_conf(now,c);
	  }
	}else if (key_is("end")) {
	  key->status=TOK_ABORT;
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
	running=FALSE;
	key->status=TOK_OK;
      }else running=FALSE;
      send_bmessage(graphix,"PRMT",strlen("animate> "),"animate> ");
      load_menu_buttons(0,topmenu);
    }
    if (key->status!=TOK_OK) break;
  }
  free_Token(key);
}
/*-------------------- cmd_confa_average --------------------*/

static void cmd_confa_average(void)
{
  Token *key;
  int i,j,k,from=0,to=0,n,m;
  CnfNode *c;
  float *x,*y,*q,*p,*b;
  Atom *a;
  Nodechain *A;

  key=(Token *)NULL;
  y=(float *)NULL;
  x=(float *)NULL;
  b=(float *)NULL;

  if (!mvld(ConfaRoot)) {
    err_set(1,"no conformations have been read in yet...");
    return;
  }

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)||NTemplate==0) {
    err_set(1,"you have no template...");
    return;
  }

  n=NTemplate;

  n=NTemplate;
  m=0;  /* count conformations */
  forCnfs(c,ConfaRoot) m++;

  default_token("1");
  key=Get_Int("first conformation in average: ",";");
  if (key->status==TOK_OK) {
    from=key->n;
    if (from<1||from>m) {
      err_set(1,"out of range, must be between %d and %d.",1,m);
      key->status=TOK_ABORT;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad integer token: %s",key->text);
    key->status=TOK_ABORT;
  }
  if (key->status==TOK_OK) {
    default_token("%d",m);
    key=free_Token(key);
    key=Get_Int("last conformation in average: ",";");
    if (key->status==TOK_OK) {
      to=key->n;
      if (to<from||to>m) {
	err_set(1,"out of range, must be between %d and %d.",1,m);
	key->status=TOK_ABORT;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad integer token: %s",key->text);
      key->status=TOK_ABORT;
    }
  }
  if (key->status==TOK_OK) {
    x=(float *)malloc(3*n*sizeof(float));
    b=(float *)malloc(n*sizeof(float));
    y=(float *)malloc(3*n*sizeof(float));
    if (mvld(x)&&mvld(b)&&mvld(y)) {
      for (p=x,i=0;i<3*n;i++) *p++ = 0.0;
      i=0;k=0;
      forCnfs(c,ConfaRoot) {
	i++;
	if (i>=from && i<=to) {
	  getCoords(c,y);
	  for (p=x,q=y,j=0;j<3*n;j++) *p++ += *q++;
	  k++;
	}
      }
      if (k>1) {
	for (p=x,i=0;i<3*n;i++) *p++ /= k;
      }
      i=0;k=0;
      for (j=0;j<n;j++) b[j]=0.0;
      forCnfs(c,ConfaRoot) {      /* compute average structure */
	i++;
	if (i>=from && i<=to) {
	  getCoords(c,y);
	  for (p=x,q=y,j=0;j<n;j++) {
	    b[j]+=(*p-*q)*(*p-*q)+(*(p+1)-*(q+1))*(*(p+1)-*(q+1))
	      +(*(p+2)-*(q+2))*(*(p+2)-*(q+2));
	    p+=3; q+=3;
	  }
	  k++;
	}
      }
      if (k>0) for (j=0;j<n;j++) b[j]=8*M_PI*M_PI*(b[j]/(3*k));
      p=x;
      for(j=0,a=Template[0];j<n;a=Template[++j]) {/* compute rmsd from average */
	a->x[ATM_BISO]=b[j];
	a->x[0]=*p++; a->x[1]=*p++; a->x[2]=*p++;
      }
      A=ptr_to_chain((Node **)Template,NTemplate);
      send_newx(A);
      clear_node_chain(A);       /* send to graphix box */
      free_node_chain(A);
    }else{
      err_set(1,"memory allocation error.");
    }
  }
  free_Token(key);
  if (mvld(x)) free(x);
  if (mvld(b)) free(b);
  if (mvld(y)) free(y);
}
/*-------------------- cmd_confa_trace --------------------*/

static void cmd_confa_trace(void)
{
  Token *key;
  int i,j,k,on,newcolor,color=2,mdstype=0,lw;
  Nodechain *c;
  Node *m;
  Gfxobj *T;
  Atom *a;
  char *name;
  CnfNode *C;
  float *x=(float *)NULL,*y=(float *)NULL;
  Dspobj *t;

  key=(Token *)NULL;

  if (!mvld(ConfaRoot)) {
    err_set(1,"no conformations have been read in yet...");
    return;
  }

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)) {
    err_set(1,"you have no template...");
    return;
  }

  key=Get_Kwd("trace option: ",";","on off ' ' add delete ' ' color lines points width");
  if (key->status==TOK_OK) {
    if (key_is("add")) {
      name=(char *)NULL;
      key=free_Token(key);
      key=Get_String("name: ",";");
      if (key->status==TOK_OK) {
	name=key->text;
	key->text=(char *)NULL;
	key=free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (c->n>0) T=OpenGfxObject(glblobj(assocmdl(common_root(c))),4096);
	  else T=(Gfxobj *)NULL;
	  if (mvld(T)) {
	    t=T->Obj;
	    t->dsptype=DSP_TRCE;
	    send_bheader(graphix,"MDST",sizeof(int)+sizeof(T->Obj));
	    send_bint(graphix,WNP_CSO_POINTS);
	    send_bptr(graphix,T->Obj);
	    for (i=0,j=0,a=Template[0];j<NTemplate;a=Template[++j]) {
	      if (in_chain(a,c)) {
		k=0;
		forCnfs(C,ConfaRoot) {
		  y=getAtomCoords(C,j);
		  if (k>0) AddMD(T,x,y,color);
		  x=y;
		  k++;
		}
	      }
	    }
	    prtGraphics("%d template atoms selected.\n",i);
	    prtGraphics("%d vector%s added to graphics object.\n",
			T->Nvec,(T->Nvec==1 ? "" : "s"));
	    set_node_name((Node *)t,name);
	    CloseGfxObject(T);
	  }else{
	    err_set(1,"failed to add graphics object.");
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token (trace name): %s",key->text);
      }
    }else if (key_is("color")) {
      key=free_Token(key);
      key=Get_Color("color: ",";");
      if (key->status==TOK_OK) {
	newcolor=key->n;
	key=free_Token(key);
	key=get_trace("trace: ");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (m=c->s;mvld(m);m=m->chain) {
	    if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	      send_bheader(graphix,"DSPC",sizeof(int)+sizeof(m));
	      send_bint(graphix,newcolor);
	      send_bptr(graphix,m);
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad color token: %s",key->text);
    }else if (key_is("width")) {
      key=free_Token(key);
      key=Get_Int_M("linewidth: ",";","1 2 3 4 5 6 7 8 9 10");
      if (key->status==TOK_OK) {
	lw=(key->n < 1 ? 1 : (key->n >33 ?  32 : key->n));
	key=free_Token(key);
	key=get_trace("trace: ");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (m=c->s;mvld(m);m=m->chain) {
	    if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	      send_bheader(graphix,"SOLW",sizeof(int)+sizeof(m));
	      send_bint(graphix,lw);
	      send_bptr(graphix,m);
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad linewidth token: %s",key->text);
    }else if (key_is("lines")||key_is("points")) {
      if (key_is("lines")) mdstype=WNP_CSO_LINES;
      else if (key_is("points")) mdstype=WNP_CSO_POINTS;
      key=free_Token(key);
      key=get_trace("trace: ");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (m=c->s;mvld(m);m=m->chain) {
	  if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	    send_bheader(graphix,"MDST",sizeof(int)+sizeof(m));
	    send_bint(graphix,mdstype);
	    send_bptr(graphix,m);
	  }
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
    }else if (key_is("delete")) {
      key=free_Token(key);
      key=get_trace("trace: ");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	del_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
    }else if (key_is("on")||key_is("off")) {
      on=key_is("on");
      key=free_Token(key);
      key=get_trace("trace: ");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (m=c->s;mvld(m);m=m->chain) {
	  if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	    send_bheader(graphix,"DSPO",sizeof(int)+sizeof(m));
	    send_bint(graphix,on);
	    send_bptr(graphix,m);
	  }
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad trace option: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_confa_browse --------------------*/

static void cmd_confa_browse(void)
{
  Token *key;
  int now,n;
  CnfNode *c;
  Node *rm;

  key=(Token *)NULL;
  
  if (!mvld(ConfaRoot)) {
    err_set(1,"no conformations have been read in yet...");
    return;
  }

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)) {
    err_set(1,"you have no template...");
    return;
  }

  now=ptr2seq(setCnf);

  for (;;) {
    c=setCnf;
    err_dump(); key=free_Token(key);
    key=Get_Kwd("browse option: ",";",browse_menu);
    if (key->status==TOK_OK) {
      if (key_is("first")) {
	c=(CnfNode *)ConfaRoot->fst;
	confa_set_conf(1,c);
	now=1;
      }else if (key_is("last")) {
	now=0;
	forCnfs(c,ConfaRoot) now++;
	c=(CnfNode *)ConfaRoot->lst;
	confa_set_conf(now,c);
      }else if (key_is("next")) {
	if (!mvld(c)) {
	  c=first_cnf(ConfaRoot);
	  now=1;
	}else if (is_cnf(c->up)&&c->up!=ConfaRoot) {
	  c=(CnfNode *)c->up;
	  now=ptr2seq(c);
	}else{
	  if (c->nxt) {
	    c=(CnfNode *)c->nxt;
	    now++;
	  }
	}
	confa_set_conf(now,c);
      }else if (key_is("previous")) {
	if (!mvld(c)) {
	  c=last_cnf(ConfaRoot);
	  now=ptr2seq(c);
	}else if (is_cnf(c->up)&&c->up!=ConfaRoot) {
	  c=(CnfNode *)c->up;
	  now=ptr2seq(c);
	}else{
	  if (c->prv) {
	    c=(CnfNode *)c->prv;
	    now--;
	  }
	}
	confa_set_conf(now,c);
      }else if (key_is("goto")) {
	key=free_Token(key);
	key=Get_Int("conformation number: ",";");
	if (key->status==TOK_OK) {
	  n=key->n;
	  if (n<1) n=1;
	  now=0;
	  c=seq2ptr(n);
	  if (!mvld(c)) err_set(1,"conformation #%d not found.",n);
	  else {
	    now=n;
	    confa_set_conf(now,c);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad number: %s",key->text);
	}
      }else if (key_is("accept")) {
	cmd_confa_accept();
      }else if (key_is("copy")) {
	cmd_confa_copy();
      }else if (key_is("remove")) {
	if (mvld(c)) {
	  rm=(Node *)c;
	  if (is_cnf(c->up)&&c->up!=ConfaRoot) {
	    c=(CnfNode *)c->up;
	    now=ptr2seq(c);
	  }else if (c->nxt) {
	    c=(CnfNode *)c->nxt;
	  }else{
	    c=(CnfNode *)c->prv;
	    now--;
	  }
	  confa_set_conf(now,c);
	  del_node(rm);
	}
      }else if (key_is("restore")) {
	confa_set_conf(0,svdCnf);
	setCnf=(CnfNode *)NULL;
	c=setCnf;
      }else if (key_is("write")) { 
	cmd_write((Node *)NULL,FT_CAN_WRITE&FT_MOL_TYPES);
      }else if (key_is("CPK")) cmd_cpk();
      else if (key_is("Color")) cmd_color();
      else if (key_is("Delete")) {cmd_delete();if (!sanity_check()) break;}
      else if (key_is("Display")) cmd_display();
      else if (key_is("Label")) cmd_label();
      else if (key_is("Measure")) cmd_measure();
      else if (key_is("reset")) cmd_display_reset();
      else if (key_is("select")) cmd_select();
      else if (key_is("Set")) cmd_set();
      else if (key_is("Command")) {
	command_loop("Wit!P(Confa,browse)> ");
	if (!sanity_check()) break;
      }else if (key_is("end")) break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad browse option: %s",key->text);
    }
  }

  free_Token(key);
}
/*-------------------- cmd_confa_accept --------------------*/

static void cmd_confa_accept(void)
{
  /* replaces coordinates in current conformation by template */
  /* coordinates. Application: modify coord, eg. after minimization, */
  /* then accept new coords, write new Confa file... Note: new */
  /* coordinates will be accepted for mobile atoms only. */

  Token *key=(Token *)NULL;
  Nodechain *A;
  Atom *a;
  int i,j;

  if (!setCnf) {
    err_set(1,"no current conformation.\n");
    return;
  }
  default_token("%0.3f",setCnf->energy);
  key=Get_Float("new energy: ",";");
  if (key->status==TOK_OK) {
    if (!key->def) setCnf->energy=key->f;
    A=getAtoms(setCnf);
    if (mvld(A)) {
      for (j=0,a=(Atom *)A->s;
	   j<setCnf->n&&mvld(a);
	   j++,a=(Atom *)a->chain) {
	for (i=0;i<3;i++) setCnf->x[3*j+i]=a->x[i];
      }
      clear_node_chain(A);
      free_node_chain(A);
    }else err_set(1,"accept failed (memory allocation failure)");
  }else if (key->status==TOK_NOK) err_set(1,"bad floating point number: %s\n",key->text);

  free_Token(key);
}
/*-------------------- cmd_confa_copy --------------------*/

static void cmd_confa_copy(void)
{
  Token *key=(Token *)NULL;
  CnfNode *L,*cnf;
  Atom *a;
  Mol *mol,*new;
  Dspobj *global;
  float *x,*X=(float *)NULL;
  int j,how;
  
  how=OBJ_CPYINTRABNDS|OBJ_CPYSETS|OBJ_CPYXCN|OBJ_CPYNOE;

  if (!setCnf) {
    err_set(1,"no current conformation.\n");
    return;
  }
  L=setCnf;
  if (L&&L->up&&L->up!=ConfaRoot) L=(CnfNode *)L->up;

  if (L->fst||1>0) key=Get_Kwd("copy what: ",";","conformation cluster");
  else key=Get_Kwd("copy what: ",";","conformation");
  if (key->status==TOK_OK) {
    if (key_is("conformation")) {
      key=free_Token(key);
      default_token("Conf_%s",L->name);
      key=Get_String("give new molecule name: ",";");
      if (key->status==TOK_OK) {
	new=TreeCopyAtoms(NTemplate,Template,tFLG,how,(Node *)NULL,(Node *)NULL);
	if (new) {
	  set_node_name((Node *)new,key->text);
	  new->up=MolRoot; link_node((Node *)new);
	  send_molecule(MolRoot,new);
	}else err_set(1,"failed to generate copy of template...\n");
      }else if (key->status==TOK_NOK) err_set(1,"bad token: %s\n",key->text);
    }else if (key_is("cluster")) {
      key=free_Token(key);
      j=0;forCnfs(cnf,ConfaRoot) {j++;if (cnf==L) break;}
      if (cnf==L) default_token("Cluster_%s",L->name);
      else default_token("Cluster_x");
      key=Get_String("give new molecule name: ",";");
      if (key->status==TOK_OK) {
	mol=(Mol *)add_node(MolRoot,key->text,OBJ_MOL);
	if (mvld(mol)) {
	  global=new_dsp_obj(DspRoot,(char *)NULL,DSP_GLBL);
	  if (mvld(global)) {
	    global->root=(Node *)mol;
	    mol->dsptr=global;
	  }
	  new=TreeCopyAtoms(NTemplate,Template,tFLG,how,(Node *)NULL,(Node *)NULL);
	  if (new) {
	    X=(float *)malloc(3*NTemplate*sizeof(float));
	    if (mvld(X)) {
	      getCoords(L,X);
	      set_node_name((Node *)new,L->name);
	      new->up=(Node *)mol; link_node((Node *)new);
	      x=X; ForAtoms(a,new) for (j=0;j<3;j++) a->x[j]=*x++;
	      forCnfs(cnf,L) {
		new=TreeCopyAtoms(NTemplate,Template,tFLG,how,(Node *)NULL,(Node *)NULL);
		if (new) {
		  getCoords(cnf,X);
		  set_node_name((Node *)new,cnf->name);
		  x=X; ForAtoms(a,new) for (j=0;j<3;j++) a->x[j]=*x++;
		  new->up=(Node *)mol; link_node((Node *)new);
		}else{
		  err_set(1,"failed to generate copy of template...\n");
		  break;
		}
	      }
	      free(X); X=(float *)NULL;
	    }else err_set(1,"copy failed (memory allocation failure");
	    if (!cnf) {
	      lock_evtq();
	      j=0;
	      forMols(new,mol) {
		send_molecule((Node *)mol,new);
		if (j++%10==0) {
		  unlock_evtq();
		  send_bheader(graphix,"EVTQ",0);
		  lock_evtq();
		}
	      }
	      unlock_evtq();
	    }else del_node((Node *)mol);
	  }else{
	    err_set(1,"failed to generate copy of template...\n");
	    del_node((Node *)mol);
	  }
	}else err_set(1,"failed to generate copy of cluster...\n");
      }else if (key->status==TOK_NOK) err_set(1,"bad token: %s\n",key->text);
    }else err_set(1,"unknown option: %s\n",key->text);
  }else if (key->status==TOK_NOK) err_set(1,"bad token: %s\n",key->text);

  free_Token(key);
}

/*-------------------- confa_priax --------------------*/

static void confa_priax(void)
{
  int i,n;
  mat_4 R;
  float **x,*m;
  CnfNode *C;
  Token *key=(Token *)NULL;

  m=(float *)NULL; n=NTemplate;
  default_token("yes");
  key=Get_Kwd("original coordinates will be lost, proceed? ",";","yes no");
  if (key->status==TOK_OK) {
    ForCnfs(C,ConfaRoot) n=(C->n>n ? C->n : n);
    if (n>0) {
      m=malloc(n*sizeof(float));
      if (mvld(m)) {
	for(i=0;i<NTemplate;i++) m[i]=(mvld(Template[i]->atp) ? Template[i]->atp->mass : 0.0);
	for (;i<n;i++) m[i]=0.0;
      }else{
	err_set(1,"memory allocation failure, all masses assumed = 1.0");
      }
    }
    if (key_is("yes")) {
      ForCnfs(C,ConfaRoot) {
	if (C->n>0) {
	  x=(float **)malloc(C->n*sizeof(float *));
	  if (mvld(x)) {
	    for (i=0;i<C->n;i++) x[i]=C->x+3*i;
	    R=PMofI(C->n,x,m);
	    prtGraphics("%s, moments = %0.4E, %0.4E, %0.4E\n",
			C->name,R.r[3].c[0],R.r[3].c[1],R.r[3].c[2]);
	    ApplyTransf(R,C->n,x,x);
	    free(x);
	  }else{
	    err_set(1,"memory allocation failure");
	  }
	}
      }
      if (setCnf) confa_set_conf(0,setCnf);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad response: %s\n",key->text);
  if (m) free(m);
  free_Token(key);
}

/*-------------------- cmd_confa_fit --------------------*/

static void cmd_confa_fit(void)
{
  Atom **A,*a;
  Nodechain *c;
  Token *key;
  int i,j,k,n;
  mat_4 R;
  float *x,*y,*p,d,D;
  CnfNode *C;

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)) {
    err_set(1,"you have no template...");
    return;
  }

  if (Mobile) {
    ForCnfs(C,ConfaRoot) if (C->n!=NTemplate) break;
    if (C) {
      err_set(1,"sorry, confa fitting does not work with fixed atoms.");
      return;
    }
  }

  key=(Token *)NULL;
  key=Get_Atoms_D("select atoms to be used in fit: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    n=0; A=(Atom **)NULL;
    c=(Nodechain *)key->ptr;
    if (c->n>0) {
      for (n=0,i=0,a=Template[0];i<NTemplate;a=Template[++i]) {
	if (in_chain(a,c)) n++;
	a->seqnr=i;
      }
    }
    if (n>0) {
      A=(Atom **)malloc(n*sizeof(Atom *));
      if (!mvld(A)) err_set(1,"memory allocation failed.");
      else{
	for (n=0,i=0,a=Template[0];i<NTemplate;a=Template[++i]) if (in_chain(a,c)) A[n++]=a;
      }
    }else err_set(1,"zero template atoms selected.");
    clear_node_chain(c);
    free_node_chain(c);
    if (mvld(A)) {
      x=(float *)malloc(3*n*sizeof(float));
      y=(float *)malloc(3*NTemplate*sizeof(float));
      if (mvld(x)&&mvld(y)) {
	k=0; D=0.0;
	ForCnfs(C,ConfaRoot) {
	  getCoords(C,y);
	  for (p=x,i=0;i<n;i++) {
	    for (j=0;j<3;j++) *p++ = y[3*A[i]->seqnr+j];
	  }
	  if (i<n) {
	    err_set(1,"template mismatch.");
	  }else{
	    d=rms_best_mat_x(n,A,x,&R);
	    if (d>=0) {
	      R=mmxinv(R);
	      for (p=C->x,i=0;i<C->n;i++,p+=3) v32arr(p,m4xv3(R,vec3(p)));
	      k++; D+=d;
	    }else{
	      err_set(1,"fit failed.");
	    }
	  }
	}
	prtGraphics("%d conformations fitted to template.\n",k);
	if (k>0) {
	  D/=k;
	  prtGraphics("average r.m.s.d = %3.3f\n",D);
	}
	free(x); free(y);
      }else{
	err_set(1,"memory allocation failed.");
	if (x) free(x);
	if (y) free(y);
      }
      free(A);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad atom selection: %s",key->text);
  }

  free_Token(key);
}

/*-------------------- confa_set_conf --------------------*/

static void confa_set_conf(int seqnr,CnfNode *c)
{
  Nodechain *A;
  Atom *a;
  int i,j,n;
  float *x;
  CnfNode *m;
  
  if (!mvld(c)) {
    err_set(1,"specified conformation not found.");
    setCnf=c;
  }else if (!mvld(Template)) {
    err_set(1,"no template defined yet.");
  }else{
    A=getAtoms(c); x=c->x;
    if (mvld(A)) {
      for (a=(Atom *)A->s,i=0;i<c->n&&mvld(a);i++,a=(Atom *)a->chain) {
	for (j=0;j<3;j++) a->x[j] = *x++;
      }
      if (i<c->n || mvld(a)) {
	err_set(1,"warning: conf-node does not match template.");
      }
      if (i>0) send_newx(A);
      setCnf=c;
      err_dump();
      n=0;forCnfs(m,c) n++;
      if (n==0) {
	if (seqnr>0) {
	  prtGraphics("Nr. %d, Conformation %s, Energy = %3.3f Kcal/mole\n",
		      seqnr,c->name,c->energy);
	}else{
	  prtGraphics("Conformation %s, Energy = %3.3f Kcal/mole\n",
		      c->name,c->energy);
	}	  
      }else{
	if (seqnr>0) {
	  prtGraphics("Nr. %d, Cluster %s, Energy = %3.3f Kcal/mole, %d members\n",
		      seqnr,c->name,c->energy,n+1);
	}else{
	  prtGraphics("Cluster %s, Energy = %3.3f Kcal/mole, %d members\n",
		      c->name,c->energy,n+1);
	}
      }
      clear_node_chain(A);
      free_node_chain(A);
    }else{
      err_set(1,"memory allocation failed.");
    }
  }
}
/*-------------------- confa_status --------------------*/

static void confa_status(void)
{
  int n,nmin,nmax,mis,nclu;
  float emin,emax;
  CnfNode *c;

  n=0; mis=FALSE; nclu=0;
  nmin=10000000; nmax=-1; 
  emin=9999999.0; emax= -99999999.0;
  ForCnfs(c,ConfaRoot) {
    n++;
    if (c->n<nmin) nmin=c->n;
    if (c->n>nmax) nmax=c->n;
    if (c->energy<emin) emin=c->energy;
    if (c->energy>emax) emax=c->energy;
    if (Template) {
      if (c->n!=NTemplate) {
	if (Mobile) {
	  if (c->n!=NMobile) mis=TRUE;
	}else mis=TRUE;
      }
    }
    if (c->up==ConfaRoot) nclu++;
  }
    
  if (Template) {
    if (Mobile) {
      prtGraphics("%s%d template atoms (%d fixed)\n",
		  mis?"template mismatch, ":"",
		  NTemplate,NTemplate-NMobile);
    }else{
      prtGraphics("%s%d template atoms\n",
		  mis?"template mismatch, ":"",
		  NTemplate);
    }
  }else prtGraphics("no template selected.\n");

  if (nclu<n) prtGraphics("%d conformations in %d clusters.\n",n,nclu);
  else prtGraphics("%d conformations.\n",n);
  if (n>0) {
    if (nmin==nmax) {
      prtGraphics("number of atoms = %d\n",nmin);
    }else{
      prtGraphics("min. number of atoms = %d, max = %d\n",nmin,nmax);
    }
    prtGraphics("min. energy  = %3.3f, max = %3.3f\n",emin,emax);
  }
  
}
/*-------------------- cmd_confa_purge --------------------*/

static void cmd_confa_purge(void)
{
  Token *key;
  float e;
  int n,now;
  Nodechain c;
  CnfNode *C,*newCnf=NULL;

  key=(Token *)NULL;

  key=Get_Float("max. energy = ",";");
  if (key->status==TOK_OK) {
    e=key->f;
    c.n=0; c.s=(Node *)NULL; c.e=c.s; c.p=c.s;
    forCnfs(C,ConfaRoot) if (C->energy>=e) add_to_node_chain((Node *)C,&c);
    if (setCnf&&setCnf->energy>=e) {    /* find node to replcae setCnf if setCnf is purged */
      for(newCnf=setCnf;newCnf;newCnf=(CnfNode *)newCnf->nxt) if (newCnf->energy<e) break;
      if (!newCnf) {
	for(newCnf=setCnf;newCnf;newCnf=(CnfNode *)newCnf->prv) if (newCnf->energy<e) break;
      }
      setCnf=(CnfNode *)NULL;
    }
    n=c.n; del_node_chain(&c);
    prtGraphics("%d conformations deleted.\n",n);
    if (newCnf) {
      now=ptr2seq(newCnf);
      confa_set_conf(now,newCnf);
    }else if (setCnf) {
      now=ptr2seq(setCnf);
      confa_set_conf(now,setCnf);
    }else{
      confa_set_conf(0,svdCnf);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad float: %s",key->text);
  }

  confa_status();
  free_Token(key);
}
/*-------------------- confa_reduce --------------------*/

static void confa_reduce(void)
{
  CnfNode *c,*d;

  /* remove all cluster memmber, keeping only leaders */

  forCnfs(c,ConfaRoot) {
    while ((d=first_cnf(c))) del_node((Node *)d);
  }
}


/*-------------------- confa_sort --------------------*/

static void confa_sort(int key)
{
  void **rec;
  CnfNode *c;
  int n,i;

  n=0; ForCnfs(c,ConfaRoot) n++;
  if (n>0) {
    rec=(void **)malloc(n*sizeof(void *));
    if (mvld(rec)) {
      n=0; ForCnfs(c,ConfaRoot) rec[n++]=(void *)c;
      confa_sort_key=key;
      qsort(rec,n,sizeof(void *),confa_sort_cmp);
      for (i=0;i<n;i++) unlink_node((Node *)rec[i]);
      for (i=0;i<n;i++) {
	((Node *)rec[i])->up=ConfaRoot;
	link_node((Node *)rec[i]);
      }
      free(rec);
    }else{
      err_set(1,"confa_sort: memory allocation failed.");
    }
  }
}

/*-------------------- confa_sort_cmp --------------------*/

int confa_sort_cmp(const void *a,const void *b)
{
  int r;
  CnfNode *A,*B;

  r=0;

  switch (confa_sort_key) {
  case CONFA_SORT_ENERGY:
    A=*(CnfNode **)a;
    B=*(CnfNode **)b;
    if (A->energy < B->energy) r=-1;
    else if (A->energy > B->energy) r=1;
    else {
      if (A->cseqnr < B->cseqnr) r=-1;
      else if (A->cseqnr > B->cseqnr) r=1;
    }
    break;
  case CONFA_SORT_SEQNR:
    A=*(CnfNode **)a;
    B=*(CnfNode **)b;
    if (A->cseqnr < B->cseqnr) r=-1;
    else if (A->cseqnr > B->cseqnr) r=1;
    break;
  default:
    break;
  }
  return r;
}
/*-------------------- cmd_confa_cluster --------------------*/

static void cmd_confa_cluster(void)
{
  Token *key;
  
  key=(Token *)NULL;
  
  for(;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("cluster option: ",";",cluster_menu);
    if (key->status==TOK_OK) {
      if (key_is("reset")) {
	confa_sort(CONFA_SORT_SEQNR);
	confa_status();
      }else if (key_is("find")) {
	cmd_confa_cluster_find();
	confa_status();
      }else if (key_is("info")) {
	confa_cluster_info();
      }else if (key_is("list")) {
	confa_cluster_list();
      }else if (key_is("method")) {
	key=free_Token(key); 
	default_token("%s",cl_method_name[method]);
	key=Get_Kwd("method: ",";",cl_methods);
	if (key->status==TOK_OK) method=key->n;
	else if (key->status==TOK_NOK) err_set(1,"unknown cluster method: %s",key->text);
      }else if (key_is("metric")) {
	key=free_Token(key); 
	default_token("%s",cl_metric_name[metric]);
	key=Get_Kwd("metric: ",";",cl_metrics);
	if (key->status==TOK_OK) {
	  metric=key->n;
	  free_Token(key);
	  default_token("%s",cl_metopt_name[metopt]);
	  key = Get_Kwd("gseal or rmsd ?: ",";",cl_metopts);
	  if (key->status == TOK_OK) {
	    metopt=key->n;
	  }else if (key->status==TOK_NOK) err_set(1,"unknown cluster metric option: %s",key->text);
	}else if (key->status==TOK_NOK) err_set(1,"unknown cluster metric: %s",key->text);
      }else if (key_is("reduce")) {
	confa_reduce();
	confa_status();
      }else if (key_is("end")) break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad cluster option: %s",key->text);
    }
  }
  free_Token(key);
}
/*-------------------- cmd_confa_cluster_find --------------------*/

static void cmd_confa_cluster_find(void)
{
  Token *key;
  int N,n,i,status=TOK_OK;
  float rho=0.0;
  Nodechain *c;
  Atom *a;
  CnfNode *C,**Confs;
  char *rho_menu;

  if (!mvld(Template)) cmd_confa_template();
  if (!mvld(Template)) {
    err_set(1,"you have no template...");
    return;
  }

  key=(Token *)NULL;
  
  if (streq(cl_method_name[method],"leader")) {
    rho_menu="0.01 0.05 0.10 0.20 0.25 0.30 0.35 0.40 0.45 0.50 0.75 1.0 1.5 2.0";
    key=Get_Float_M("rho = ",";",rho_menu);
    if (key->status==TOK_OK) rho=key->f;
    else if (key->status==TOK_NOK) err_set(1,"bad value for rho: %s",key->text);
    status=key->status;
  }

  if (status==TOK_OK) {
    if (streq(cl_metric_name[metric],"fitted")||
	streq(cl_metric_name[metric],"inplace")) {
      key=Get_Atoms_D("atoms to be used for comparison: ",";");
      send_bheader(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (n=0,i=0,a=Template[0];i<NTemplate;a=Template[++i]) {
	  if (in_chain(a,c)) n++;
	  a->seqnr=i;
	}
	if (n>0) {
	  SubsetIdx=(int *)malloc(n*sizeof(int));
	  if (mvld(SubsetIdx)) {
	    for (n=0,i=0,a=Template[0];i<NTemplate;a=Template[++i]) {
	      if (in_chain(a,c)) SubsetIdx[n++]=a->seqnr;
	    }
	    nSubset=n;
	  }else{
	    err_set(1,"memory allocationfailed.");
	    key->status=TOK_NOK;
	  }
	}else{
	  err_set(1,"no atomselected, can't cluster");
	  key->status=TOK_NOK;
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
      status=key->status;
    }
  }

  if (status==TOK_OK) {
    N=0; ForCnfs(C,ConfaRoot) N++;
    Confs=(CnfNode **)malloc(N*sizeof(CnfNode *));
    if (mvld(Confs)) {
      N=0; ForCnfs(C,ConfaRoot) Confs[N++]=C;
      if (streq(cl_method_name[method],"leader")) {
	confa_cluster_leader(N,Confs,rho);
      }else if (cl_method_procedure[method]==CL_METH_AGGLOM) {
	confa_cluster_agglom(N,Confs);
      }else{
	prtGraphics("sorry, %s not yet implemented\n",cl_method_name[method]);
      }
      err_dump();
      free(Confs);
    }else{
      err_set(1,"memory allocation failed.");
    }
  }
    
  if (SubsetIdx) free(SubsetIdx);
  nSubset=0; SubsetIdx=NULL;
      
  free_Token(key);
}
/*-------------------- confa_cluster_list --------------------*/

static void confa_cluster_list(void)
{
  int nc=0,ic=0,n,sz;
  CnfNode *C,*c;

  forCnfs(C,ConfaRoot) nc++;
  prtGraphics("#-----------------------------------\n"
	      "# Total number of clusters =%8d\n"
	      "#\n",
	      nc);
  if (nc>0) {
    prtGraphics("#-----------------------------------\n"
		"# Center     Number\n"
		"#   of         of\n"
		"# cluster    conformations\n"
		"#-----------------------------------\n");
    forCnfs(C,ConfaRoot) {
      sz=1;
      forCnfs(c,C) sz++;
      prtGraphics("%8d %7d\n",C->cseqnr,sz);
    }
    prtGraphics("#-----------------------------------\n");
    ic=0;
    forCnfs(C,ConfaRoot) {
      prtGraphics(" Cluster Number = %7d\n",++ic);
      prtGraphics(" %7d",C->cseqnr); n=1;
      forCnfs(c,C) {
	prtGraphics(" %7d%s",c->cseqnr,(++n%10==0?"\n":""));
      }
      if (n%10!=0) prtGraphics("\n");
    }
  }
}
/*-------------------- confa_cluster_info --------------------*/

static void confa_cluster_info(void)
{
  int nc=0,i,*sz,s,n;
  CnfNode *C,*c;

  forCnfs(C,ConfaRoot) nc++;
  prtGraphics("Cluster-info\n");
  prtGraphics("    size  clusters\n");
  if (nc>0) {
    if ((sz=(int *)malloc(nc*sizeof(int)))) {
      i=0;
      forCnfs(C,ConfaRoot) {
	sz[i]=1;
	forCnfs(c,C) sz[i]++;
	i++;
      }
      qsort(sz,nc,sizeof(int),cmp_int_asc);
      for (n=1,s=sz[0],i=1;i<nc;i++) {
	if (sz[i]==s) n++;
	else {
	  prtGraphics("%8d  %8d\n",s,n);
	  n=1; s=sz[i];
	}
      }
      prtGraphics("%8d  %8d\n",s,n);
      free(sz);
    }else err_set(1,"memory allocation failure in confa_cluster_info");
    prtGraphics("Total number of clusters: %d\n",nc);
  }
}
/*-------------------- confa_cluster_leader --------------------*/

static void confa_cluster_leader(int N,CnfNode **C,float rho)
{
  /* Trivial "leader" cluster method... */

  float d,dmin;
  int i;
  CnfNode *c,*cmin;

  /* sanity check... */

  if (N<1) return;
  for(i=0;i<N;i++) unlink_node((Node *)C[i]);
  
  C[0]->up=(Node *)ConfaRoot;
  link_node((Node *)C[0]);

  for(i=1;i<N;i++) {
    dmin=1.0e32; cmin=(CnfNode *)NULL;
    forCnfs(c,ConfaRoot) {
      d=Distance(C[i],c);
      if (d<dmin) {
	dmin=d;
	cmin=c;
      }
    }
    if (dmin<rho && mvld(cmin)) C[i]->up=(Node *)cmin;
    else C[i]->up=(Node *)ConfaRoot;
    link_node((Node *)C[i]);
  }

}
/*-------------------- confa_cluster_agglom --------------------*/

static void confa_cluster_agglom(int N,CnfNode **C)
{
  float **D=NULL;
  float alpha=0.6;
  int status,i,j,ncl;
  agglom_clhist_t *hist=NULL;
  Token *key=(Token *)NULL;
  CnfNode *c;

  D=malloc(N*sizeof(float *));
  hist=malloc((N-1)*sizeof(agglom_clhist_t));
  if (D&&hist) {
    for (i=0;i<N;i++) {
      D[i]=malloc((i+1)*sizeof(float));
      if (!D[i]) break;
      for (j=0;j<i;j++) D[i][j]=Distance(C[i],C[j]);
      D[i][i]=0.0;
    }
    if (i<N) {
      for (j=0;j<i;j++) free(D[j]);
    }else{
      status=agglom(N,D,cl_method_name[method],alpha,FALSE,hist);
      if (!status) {
	for (i=N-21<0?0:N-21;i<N-1;i++) {
	  prtGraphics("%5d clusters at %0.3f\n",N-i-1,hist[i].d);
	}
	key=free_Token(key);
	key=Get_Int_M("number of clusters: ",";","1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 20");
	if (key->status==TOK_OK) {
	  ncl=key->n;
	  for (i=0;i<N;i++) unlink_node((Node *)C[i]);
	  for (i=0;i<N-ncl;i++) {
	    C[hist[i].i]->up=(Node *)C[hist[i].j];
	    link_node((Node *)C[hist[i].i]);
	    while (C[hist[i].i]->fst) {
	      c=(CnfNode *)C[hist[i].i]->fst;
	      unlink_node((Node *)c);
	      c->up=(Node *)C[hist[i].j];
	      link_node((Node *)c);
	    }
	  }
	  for (i=0;i<N;i++) {
	    if (!C[i]->up) {
	      C[i]->up=(Node *)ConfaRoot;
	      link_node((Node *)C[i]);
	    }
	  }
	}else if (key->status==TOK_NOK) err_set(1,"bad number: %s",key->text);
      }else if (status==1) {
	err_set(1,"Failed to allocate memory in agglom");
      }else if (status==2) {
	err_set(1,"Error in agglom: unknown method %s",cl_method_name[method]);
      }
      for (j=0;j<N;j++) free(D[j]);
    }
  }
  if (D) free(D);
  if (hist) free(hist);
  free_Token(key);
}
/*-------------------- save_conformation --------------------*/

static CnfNode *save_conformation(int N,Atom **A)
{
  int n,i;
  CnfNode *r;
  float *x;
  Atom *a;

  r=(CnfNode *)NULL;
  if (N>0 && mvld(A)) {
    r=(CnfNode *)make_node(OBJ_CNF);
    if (mvld(r)) {
      set_node_name((Node *)r,"Saved_Conformation");
      r->energy=-9999.999;
      r->x=(float *)malloc(3*N*sizeof(float));
      if (mvld(r->x)) {
	r->n=N; x=r->x;
	for (n=0,a=A[0];n<N;a=A[++n]) for (i=0;i<3;i++) *x++ = a->x[i];
      }else{
	r=(CnfNode *)del_node((Node *)r);
      }
    }
  }
  return(r);
}
/*-------------------- cmd_confa_write --------------------*/

static void cmd_confa_write(void)
{
  Token *key;
  
  key=Get_Kwd("write what ? ",";","conformations variables");
  if (key->status==TOK_OK) {
    if (key_is("conformations")) cmd_confa_write_conformations();
    else if (key_is("variables")) cmd_confa_write_variables();
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad keyword: %s",key->text);
  }
}
/*-------------------- cmd_confa_write_conformations --------------------*/

static void cmd_confa_write_conformations(void)
{
  Token *key;
  FILE *out;
  char buf[4096];
  int n,i,k;
  CnfNode *c;
  float *x;

  key=(Token *)NULL;
  key=Get_String("filename: ",";");
  if (key->status==TOK_OK) {
    out=fopen(extfil(buf,key->text,".cnf",FALSE),"w");
    if (out) {
      n=0;
      if (Mobile&&NMobile>0) {
	fprintf(out,"Mobile %d\n",NMobile);
	for (i=0;i<NMobile;) {
	  for (k=0;k<10&&i<NMobile;k++,i++) fprintf(out," %d",Mobile[i]+1);
	  if (k>0) fprintf(out,"\n");
	}
      }
      ForCnfs(c,ConfaRoot) {
	if (c->n>0 && mvld(c->x)) {
	  fprintf(out,"Conformation %s Energy = %0.4f Status = unknown\n",c->name,c->energy);
	  for (x=c->x,i=0;i<c->n;i++,x+=3) fprintf(out,"%0.4f %0.4f %0.4f\n",x[0],x[1],x[2]);
	  if (ferror(out)) {
	    err_set(1,"error writing conformation %d (%s)\n",n+1,c->name);
	    break;
	  }
	  n++;
	}
      }
      err_dump();
      prtGraphics("%d conformations written to file %s\n",n,buf);
      fclose(out);
    }else{
      err_set(1,"could not open %s for write.",buf);
    }
  }else{
    if (key->status==TOK_NOK) err_set(1,"bad filename token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- init_confa_var --------------------*/

static void init_confa_var(void *data)
{
  CnfVar *v;

  if (data) {
    v=(CnfVar *)data;
    v->name=(char *)NULL;
    v->type=VAR_NONE;
    v->data=(void *)NULL;
  }
}
/*-------------------- del_confa_var --------------------*/
static void del_confa_var(void *data)
{
  CnfVar *v;

  if (data) {
    v=(CnfVar *)data;
    switch (v->type) {
    case VAR_DISTANCE:
    case VAR_ANGLE:
    case VAR_TORSION:
    case VAR_FITRMSD:
      if (v->data) free(v->data);
      break;
    default:
      break;
    }
    if (v->name) free(v->name);
    free(data);
  }    
}
/*-------------------- cmd_confa_variable --------------------*/

static void cmd_confa_variable(void)
{
  Token *key;
  char *vmenu = "add delete write";

  key=Get_Kwd("variable option: ",";",vmenu);
  if (key->status==TOK_OK) {
    if (key_is("add")) cmd_confa_variable_add();
    else if (key_is("delete")) cmd_confa_variable_delete();
    else if (key_is("write")) cmd_confa_write_variables();
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown option: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_confa_variable_add --------------------*/

static void cmd_confa_variable_add(void)
{
  Token *key=(Token *)NULL,*atoms[4];
  ListEntry *new,*l;
  CnfVar *v;
  Atom **nr,*a;
  Nodechain *c;
  int n,nd=0,na=0,nt=0,nf=0,i;
  char *menu="angle distance torsion ' ' fitrmsd ' ' energy time";

  for (l=Vars.nxt;mvld(l);l=l->nxt) {
    v=(CnfVar *)l->data;
    if (v->type==VAR_DISTANCE) nd++;
    else if (v->type==VAR_ANGLE) na++;
    else if (v->type==VAR_TORSION) nt++;
    else if (v->type==VAR_FITRMSD) nf++;
  }

  atoms[0]=(Token *)NULL; atoms[1]=atoms[0];
  new=make_list_entry(sizeof(CnfVar),init_confa_var);
  if (mvld(new)) {
    v=(CnfVar *)new->data;
    key=Get_Kwd("variable type: ",";",menu);
    if (key->status==TOK_OK) {
      if (key_is("distance")) {
	n=Get_nAtoms(2,atoms,"distance",";");
	send_bheader(graphix,"ARMK",0);
	if (n==2) {
	  v->type=VAR_DISTANCE;
	  nr=(Atom **)malloc(2*sizeof(Atom *));
	  v->data=(void *)nr;
	  if (nr) {
	    nr[0]=atoms[0]->ptr;
	    nr[1]=atoms[1]->ptr;
	    if (graphix>0) {
	      if (cPtr(nr[0]->cross) && cPtr(nr[1]->cross)) {
		send_bheader(graphix,"AMON",2*cPtrSize());
		send_bcptr(graphix,nr[0]->cross);
		send_bcptr(graphix,nr[1]->cross);
	      }
	    }
	    default_token("D%d",nd+1);
	  }else{
	    key->status=TOK_ABORT;
	    err_set(1,"memory allocation failed.");
	  }
	}else if (atoms[n]->status==TOK_NOK) {
	  key->status=TOK_ABORT;
	  err_set(1,"bad atom token: %s",atoms[n]->text);
	}
	for (n=0;n<2;n++) free_Token(atoms[n]);
      }else if (key_is("angle")) {
	n=Get_nAtoms(3,atoms,"angle",";");
	send_bheader(graphix,"ARMK",0);
	if (n==3) {
	  v->type=VAR_ANGLE;
	  nr=(Atom **)malloc(3*sizeof(Atom *));
	  v->data=(void *)nr;
	  if (nr) {
	    nr[0]=atoms[0]->ptr;
	    nr[1]=atoms[1]->ptr;
	    nr[2]=atoms[2]->ptr;
	    default_token("A%d",na+1);
	  }else{
	    key->status=TOK_ABORT;
	    err_set(1,"memory allocation failed.");
	  }
	}else if (atoms[n]->status==TOK_NOK) {
	  key->status=TOK_ABORT;
	  err_set(1,"bad atom token: %s",atoms[n]->text);
	}
	for (n=0;n<3;n++) free_Token(atoms[n]);
      }else if (key_is("torsion")) {
	n=Get_nAtoms(4,atoms,"torsion",";");
	send_bheader(graphix,"ARMK",0);
	if (n==4) {
	  v->type=VAR_TORSION;
	  nr=(Atom **)malloc(4*sizeof(Atom *));
	  v->data=(void *)nr;
	  if (nr) {
	    nr[0]=atoms[0]->ptr;
	    nr[1]=atoms[1]->ptr;
	    nr[2]=atoms[2]->ptr;
	    nr[3]=atoms[3]->ptr;
	    default_token("T%d",nt+1);
	  }else{
	    key->status=TOK_ABORT;
	    err_set(1,"memory allocation failed.");
	  }
	}else if (atoms[n]->status==TOK_NOK) {
	  key->status=TOK_ABORT;
	  err_set(1,"bad atom token: %s",atoms[n]->text);
	}
	for (n=0;n<4;n++) free_Token(atoms[n]);
      }else if (key_is("fitrmsd")) {
	if (mvld(Template)) {
	  v->type=VAR_FITRMSD;
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms to be used in rmsd: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (n=0,i=0,a=Template[0];i<NTemplate;a=Template[++i]) {
	      if (in_chain(a,c)) n++;
	    }
	    v->data=malloc((n+1)*sizeof(int));
	    ((int *)v->data)[0]=n;
	    c=(Nodechain *)key->ptr;
	    for (n=0,i=0,a=Template[0];i<NTemplate;a=Template[++i]) {
	      if (in_chain(a,c)) ((int *)v->data)[++n]=i;
	    }
	    prtGraphics("%d atoms.\n",n);
	    default_token("frmsd%d",nf+1);
	    clear_node_chain(c);
	    free_node_chain(c);
	  }
	}else{
	  key->status=TOK_ABORT;
	  err_set(1,"you must select the template first.");
	}
      }else if (key_is("energy")) {
	v->type=VAR_ENERGY;
	default_token("Energy");
      }else if (key_is("time")) {
	v->type=VAR_TIME;
	default_token("Time");
      }
      if (key->status==TOK_OK) {
	key=free_Token(key);
	key=Get_String("variable name: ",";");
	if (key->status==TOK_OK) {
	  v->name=(char *)key->text;
	  key->text=(char *)NULL;
	  append_entry_to_list(new,&Vars);
	  new=(ListEntry *)NULL;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad varibale name: %s",key->text);
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad variable type: %s",key->text);
    }
  }else{
    err_set(1,"memory allocation failed.");
  }
  if (mvld(new)) delete_list_entry(new,del_confa_var);
  free_Token(key);
}
/*-------------------- cmd_confa_variable_delete --------------------*/

static void cmd_confa_variable_delete(void)
{
  Token *key;
  char *menu;
  int len;
  ListEntry *l;
  CnfVar *v;

  menu=(char *)NULL;
  key=(Token *)NULL;

  for (len=0,l=Vars.nxt;mvld(l);l=l->nxt) {
    if (mvld(l->data)) {
      v=(CnfVar *)l->data;
      if (mvld(v->name)) len+=strlen(v->name)+1;
    }
  }
  if (len>0) {
    menu=(char *)malloc(len+1);
    for (len=0,l=Vars.nxt;mvld(l);l=l->nxt) {
      if (mvld(l->data)) {
	v=(CnfVar *)l->data;
	if (mvld(v->name)) len+=sprintf(menu+len,"%s ",v->name);
      }
    }
    menu[len]='\0';
    key=Get_Kwd("variable: ",";",menu);
    if (key->status==TOK_OK) {
      for (l=Vars.nxt;mvld(l);l=l->nxt) {
	if (mvld(l->data)) {
	  v=(CnfVar *)l->data;
	  if (mvld(v->name)) {
	    if (streq(v->name,key->key)) {
	      prtGraphics("ConfA varibale %s deleted.\n",v->name);
	      delete_list_entry(l,del_confa_var);
	      break;
	    }
	  }
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"ConfA variable %s not found.",key->text);
    }
  }else{
    err_set(1,"no ConfA variables defined.");
  }
  if (mvld(menu)) free(menu);
  free_Token(key);
}
/*-------------------- cmd_confa_write_variables --------------------*/

static void cmd_confa_write_variables(void)
{
  Token *key;
  FILE *out;
  float Time,d;
  CnfNode *c;
  int i,k,n,nf,m,nr,nval,*Nr;
  ListEntry *l;
  CnfVar *v;
  Atom **atm,*a;
  float *x[4],*y,*X,*Y;
  mat_4 R;
  
  out=(FILE *)NULL;
  key=(Token *)NULL;

  if (n_entries(&Vars)>0) {
    key=Get_String("filename: ",";");
    if (key->status==TOK_OK) {
      out=fopen(key->text,"w");
      if (out) {
	for (m=0,l=Vars.nxt;mvld(l);l=l->nxt) {
	  v=(CnfVar *)l->data;
	  if (mvld(v)) {
	    switch (v->type) {
	    case VAR_DISTANCE:
	      if (mvld(v->data)) {
		atm=(Atom **)v->data;
		atm[0]->seqnr=-1;
		atm[1]->seqnr=-1;
	      }
	      break;
	    case VAR_ANGLE:
	      if (mvld(v->data)) {
		atm=(Atom **)v->data;
		atm[0]->seqnr=-1;
		atm[1]->seqnr=-1;
		atm[2]->seqnr=-1;
	      }
	      break;
	    case VAR_TORSION:
	      if (mvld(v->data)) {
		atm=(Atom **)v->data;
		atm[0]->seqnr=-1;
		atm[1]->seqnr=-1;
		atm[2]->seqnr=-1;
		atm[3]->seqnr=-1;
	      }
	      break;
	    default: break;
	    }
	  }
	}
	if (mvld(Template)) for (n=0,a=Template[0];n<NTemplate;a=Template[++n]) a->seqnr=n;
	c=(mvld(ConfaRoot) ? (CnfNode *)ConfaRoot->fst : (CnfNode *)NULL);
	y=(float *)malloc(3*NTemplate*sizeof(float)+1);
	X=(float *)malloc(3*NTemplate*sizeof(float)+1);
	Y=(float *)malloc(3*NTemplate*sizeof(float)+1);
	for (n=0;mvld(c);c=(CnfNode *)c->nxt) {
	  if (is_cnf(c)) {
	    n++;
	    Time=n;
	    getCoords(c,y);
	    for (m=0,l=Vars.nxt;mvld(l);l=l->nxt) {
	      v=(CnfVar *)l->data;
	      if (mvld(v)) {
		switch (v->type) {
		case VAR_ENERGY:
		  m++;
		  fprintf(out,"%0.2f ",c->energy);
		  break;
		case VAR_TIME:
		  fprintf(out,"%0.3f ",Time);
		  m++;
		  break;
		case VAR_DISTANCE:
		  if (mvld(v->data)) {
		    atm=(Atom **)v->data;
		    for (i=0;i<2;i++) {
		      if (atm[i]->seqnr<0) x[i]=atm[i]->x;
		      else {
			nr=atm[i]->seqnr;
			if (y&&nr>=0&&nr<NTemplate) x[i]=y+3*nr;
			else x[i]=atm[i]->x;
		      }
		    }
		    d=cdist(x[0],x[1]);
		    fprintf(out,"%0.3f ",d);
		    m++;
		  }
		  break;
		case VAR_ANGLE:
		  if (mvld(v->data)) {
		    atm=(Atom **)v->data;
		    for (i=0;i<3;i++) {
		      if (atm[i]->seqnr<0) x[i]=atm[i]->x;
		      else {
			nr=atm[i]->seqnr;
			if (y&&nr>=0&&nr<NTemplate) x[i]=y+3*nr;
			else x[i]=atm[i]->x;
		      }
		    }
		    d=cangled(x[0],x[1],x[2]);
		    fprintf(out,"%0.3f ",d);
		    m++;
		  }
		  break;
		case VAR_TORSION:
		  if (mvld(v->data)) {
		    atm=(Atom **)v->data;
		    for (i=0;i<4;i++) {
		      if (atm[i]->seqnr<0) x[i]=atm[i]->x;
		      else {
			nr=atm[i]->seqnr;
			if (y&&nr>=0&&nr<NTemplate) x[i]=y+3*nr;
			else x[i]=atm[i]->x;
		      }
		    }
		    d=ctorsd(x[0],x[1],x[2],x[3]);
		    fprintf(out,"%0.3f ",d);
		    m++;
		  }
		  break;
		case VAR_FITRMSD:
		  if (mvld(v->data)) {
		    nval=((int *)v->data)[0];
		    Nr=((int *)v->data)+1;
		    if (n>0&&mvld(Template)) {
		      for (nf=0,i=0;i<nval;i++) {
			if (Nr[i]<NTemplate) {
			  for (k=0;k<3;k++) {
			    X[3*nf+k]=Template[Nr[i]]->x[k]; 
			    Y[3*nf+k]=y[3*Nr[i]+k];
			  }
			  nf++;
			}
		      }
		      d=rms_best_mat_yx(nf,X,Y,&R);
		    }else d=0;
		  }else d=0;
		  fprintf(out,"%0.3f ",d);
		  m++;
		  break;
		default:
		  break;
		}
	      }
	    }
	    fprintf(out,"\n");
	  }
	}
	if (mvld(y)) free(y);
	if (mvld(X)) free(X);
	if (mvld(Y)) free(Y);
	prtGraphics("%d variables for %d conformations written to %s.\n",
		    m,n,key->text);
      }else{
	err_set(1,"could not open file %s for write.",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad filename: %s",key->text);
    }
  }else{
    err_set(1,"no ConfA variables defined.");
  }
  if (out) fclose(out);
  free_Token(key);
}
/*-------------------- get_template_atom --------------------*/

int get_template_atom(char *prmt)
{
  Token *key=(Token *)NULL;
  int i,n;
  Atom *a;

  n=-1;

  if (mvld(Template)) {
    key=Get_Atom(prmt,";");
    if (key->status==TOK_OK) {
      for (i=0,a=Template[0];i<NTemplate;a=Template[++i]) if (a==(Atom *)key->ptr) break;
      if (i<NTemplate) n=i;
      else err_set(1,"atom not found in template.");
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad atom token: %s",key->text);
    }
  }else{
    err_set(1,"no template defined.");
  }

  free_Token(key);
  return(n);
}
/*-------------------- get_trace --------------------*/

static Token *get_trace(char *prmt)
{
  Nodechain c,*f;
  Dspobj *o;
  Node *z,*nxt;
  int l;
  char *names;
  Token *key;

  key=make_Token(); key->status=TOK_NOK;
  names=(char *)NULL;
  l=4; c.n=0;c.s=(Node *)NULL; c.e=c.s; c.p=c.s;

  ForDspobjs(o,DspRoot) {
    if (o->dsptype==DSP_TRCE) {
      if (mvld(o->name)) {
	l+=strlen(o->name)+1;
	add_to_node_chain((Node *)o,&c);
      }
    }
  }
  if (c.n>0) {
    names=(char *)malloc(l*sizeof(char)+1);
    if (mvld(names)) {
      sprintf(names,"* ");
      for (z=c.s;mvld(z);z=z->chain) {
	if (mvld(z->name)) sprintf(names+strlen(names),"%s ",z->name);
      }
      free_Token(key);
      key=Get_String_M(prmt,";",names);
      if (key->status==TOK_OK) {
	f=(void *)make_node_chain();
	key->ptr=f;
	if (mvld(f)) {
	  for (z=c.s;mvld(z);z=nxt) {
	    nxt=z->chain; z->chain=(Node *)NULL;
	    if (wild_imatch(z->name,key->text)) add_to_node_chain(z,f);
	  }
	  if (f->n==0) {
	    key->status=TOK_NOK;
	    key->ptr=(void *)NULL;
	    free_node_chain(f);
	  }
	}else{
	  err_set(1,"memory allocation problem in 'get_trace'.");
	  clear_node_chain(&c);
	  key->status=TOK_NOK;
	}
      }else{
	clear_node_chain(&c);
      }
    }else{
      err_set(1,"memory allocation problem in 'get_trace'.");
      clear_node_chain(&c);
    }
  }else{
    err_set(1,"no trace objects.");
    key->status=TOK_EOC;
  }

  if (mvld(names)) free(names);
  return(key);
}
/*----------------------------- cmd_confa_cotransform --------------------*/

static void cmd_confa_cotransform(char *mode)
{
  mat_4 Tra;
  CnfNode *C;
  int i;
  float *p;

  Tra=GetFitCoTransform();
  if (streq(mode,"back")) Tra=mmxinv(Tra);
  if (mvld(ConfaRoot)) {
    ForCnfs(C,ConfaRoot) {
      for (p=C->x,i=0;i<C->n;i++,p+=3) v32arr(p,m4xv3(Tra,vec3(p)));
    }
  }
}
/*-------------------- getAtoms --------------------*/

static Nodechain *getAtoms(CnfNode *c)
{
   Nodechain *r;
   int i;

   if ((r=make_node_chain())) {
     if (is_cnf(c)&&mvld(Template)) {
       if (c->n==NTemplate) {
	for(i=0;i<c->n;i++) add_to_node_chain((Node *)Template[i],r);
      }else if (mvld(Mobile)&&c->n==NMobile) {
	for (i=0;i<c->n;i++) {
	  if (Mobile[i]>=0&&Mobile[i]<NTemplate) 
	    add_to_node_chain((Node *)Template[Mobile[i]],r);
	}
      }else{
	for (i=0;i<NTemplate&&i<c->n;i++) add_to_node_chain((Node *)Template[i],r);
      }
    }
  }
  return(r);
}
/*-------------------- getAtomCoords --------------------*/
static float *getAtomCoords(CnfNode *c,int n)
{
  /* return pointer to conf. coordinates for n-th Template atom */

  static float zero[3]={0,0,0},*r;
  int k;

  if (is_cnf(c)&&mvld(Template)&&n>=0&&n<NTemplate) {
    if (c->n==NTemplate) r=c->x+3*n;
    else if (c->n==NMobile&&mvld(Mobile)) {
      for (k=0;k<NMobile;k++) if (n==Mobile[k]) break;
      if (k<NMobile) r=c->x+3*k;
      else r=Template[n]->x;
    }else if (c->n>=n) r=c->x+3*n;
    else r=Template[n]->x;
  }else r=zero;
  return(r);
}
/*-------------------- getCoords --------------------*/

static int getCoords(CnfNode *c,float *x)
{
  int r=1,i,k;

  if (is_cnf(c)&&mvld(Template)&&x) {
    if (c->n>=NTemplate) {
      for (i=0;i<3*NTemplate;i++) x[i]=c->x[i];
    }else{
      if (mvld(Mobile)&&c->n==NMobile) {
	for (i=0;i<NTemplate;i++) {
	  for (k=0;k<3;k++) x[3*i+k]=Template[i]->x[k];
	}
	for (i=0;i<c->n;i++) {
	  if (Mobile[i]>=0&&Mobile[i]<NTemplate) {
	    k=Mobile[i];
	    x[3*k]=c->x[3*i];
	    x[3*k+1]=c->x[3*i+1];
	    x[3*k+2]=c->x[3*i+2];
	  }
	}
      }else{
	for(i=0;i<NTemplate;i++) {
	  for (k=0;k<3;k++) {
	    if (i<c->n) x[3*i+k]=c->x[3*i+k];
	    else x[3*i+k]=Template[i]->x[k];
	  }
	}
      }
    }
  }else r=0;
  return(r);
}
/*-------------------- sanity_check --------------------*/

static int sanity_check(void)
{
  /* return 0 on fatal errors */

  int r,k,n,i,m,nval,*Nr;
  Atom *a,**atm;
  float *p,*q;
  CnfVar *v;
  CnfNode *C,*nxtC;
  ListEntry *l,*nxt;

  r=1;
  if (mvld(Template)) {
    if (NTemplate>0) {
      k=0;
      ForAtoms(a,MolRoot) {        /* this is just a quick test... */
	if (k>=NTemplate) break;
	if (a==Template[k]) k++;
      }
      if (k!=NTemplate) {          /* quick test failed, so... */
	k=nTemplateOK(NTemplate,Template);
	if (k==0) {
	  NTemplate=0;
	  free(Template);
	  Template=(Atom **)NULL;
	  err_set(1,"lost the template.");
	}else{
	  if ((ConfaRoot&&ConfaRoot->fst)||svdCnf) 
	    prtGraphics("found deleted atoms, updating trajectory\n");
	  if (svdCnf) {  /* delete coords in saved conformation */
	    n=0;
	    for (p=svdCnf->x,q=svdCnf->x,k=0;k<svdCnf->n&&k<NTemplate;k++,q+=3) {
	      if (Template[k]) {
		p[0]=q[0]; p[1]=q[1]; p[2]=q[2];
		p+=3; 
		n++;
	      }
	    }
	    if (n==0) {
	      del_node((Node *)svdCnf);
	      svdCnf=(CnfNode *)NULL;
	    }
	  }
	  for (C=rfirst_cnf(ConfaRoot);C;C=nxtC) {
	    nxtC=rnext_cnf(ConfaRoot,C);
	    n=0;
	    if (mvld(Mobile)&&C->n==NMobile) {
	      for (p=C->x,q=C->x,k=0;k<NMobile;k++,q+=3) {
		if (Mobile[k]>=0&&Mobile[k]<NTemplate&&Template[Mobile[k]]) {
		  p[0]=q[0]; p[1]=q[1]; p[2]=q[2];
		  p+=3; 
		  n++;
		}
	      }
	    }else{
	      for (p=C->x,q=C->x,k=0;k<C->n&&k<NTemplate;k++,q+=3) {
		if (Template[k]) {
		  p[0]=q[0]; p[1]=q[1]; p[2]=q[2];
		  p+=3; 
		  n++;
		}
	      }
	    }
	    if (n==0) del_node((Node *)C);
	    else{
	      C->x=realloc(C->x,3*n*sizeof(float));
	      C->n=n;
	    }
	  }
	  if (mvld(Mobile)) {
	    for (k=0,i=0,m=0,n=0;k<NMobile;k++) {
	      while(i<Mobile[k]&&i<NTemplate) if (Template[i++]) m++;
	      if (i<NTemplate&&Template[i]) Mobile[n++]=m;
	    }
	    NMobile=n;
	    if (n==0) {
	      free(Mobile);
	      Mobile=(int *)NULL;
	    }
	  }
	}
      }
    }
  }
  for (l=Vars.nxt;mvld(l);l=nxt) {
    nxt=l->nxt;
    v=(CnfVar *)l->data;
    if (mvld(v)) {
      switch (v->type) {
      case VAR_DISTANCE:
	if (mvld(v->data)) {
	  atm=(Atom **)v->data;
	  if (!find_node(MolRoot,atm[0],OBJ_ATOM)||!find_node(MolRoot,atm[1],OBJ_ATOM)) {
	    delete_list_entry(l,del_confa_var);
	  }
	}
	break;
      case VAR_ANGLE:
	if (mvld(v->data)) {
	  atm=(Atom **)v->data;
	  if (!find_node(MolRoot,atm[0],OBJ_ATOM)||!find_node(MolRoot,atm[1],OBJ_ATOM)||
	      !find_node(MolRoot,atm[2],OBJ_ATOM)) {
	    delete_list_entry(l,del_confa_var);
	  }
	}
	break;
      case VAR_TORSION:
	if (mvld(v->data)) {
	  atm=(Atom **)v->data;
	  if (!find_node(MolRoot,atm[0],OBJ_ATOM)||!find_node(MolRoot,atm[1],OBJ_ATOM)||
	      !find_node(MolRoot,atm[2],OBJ_ATOM)||!find_node(MolRoot,atm[3],OBJ_ATOM)) {
	    delete_list_entry(l,del_confa_var);
	  }
	}
	break;
      case VAR_FITRMSD:
	if (mvld(Template)) {
	  nval=((int *)v->data)[0];
	  Nr=((int *)v->data)+1;
	  for (k=0,i=0,m=0,n=0;k<nval;k++) {
	    while(i<Nr[k]&&i<NTemplate) if (Template[i++]) m++;
	    if (i<NTemplate&&Template[i]) Nr[n++]=m;
	  }
	  ((int *)v->data)[0]=n;
	  if (m==0) delete_list_entry(l,del_confa_var);
	}else delete_list_entry(l,del_confa_var);
	break;
      }
    }
  }
  if (Template) {
    for (n=0,k=0;k<NTemplate;k++) {
      if (Template[k]) Template[n++]=Template[k];
    }
    NTemplate=n;
  }
  return(r);
}
/*-------------------- cmd_confa_dcd --------------------*/

static void cmd_confa_dcd(void)
{
  Token *key=(Token *)NULL;
  UCpipe *trj;
  int from=1,step=1,to=-1;

  for (;;) {
    key=free_Token(key);
    key=Get_Kwd("DCD> ",";","read ' ' from to step");
    if (key->status==TOK_OK) {
      if (key_is("from")) {
	key=Get_Int_M("from frame: ",";","1 2 3 4 5 6 7 8 9 10 20 30 40 50 60 70 80 90 100");
	if (key->status==TOK_OK) from=key->n;
	else {
	  if (key->status==TOK_NOK) err_set(1,"bad value for from: %s",key->text);
	  break;
	}
      }else if (key_is("to")) {
	key=Get_Int_M("to frame: ",";","eof 1 2 3 4 5 6 7 8 9 10 20 30 40 50 60 70 80 90 100");
	if (key->status==TOK_OK) to=key->n;
	else {
	  if (key->status==TOK_NOK) {
	    key=Prs_Kwd(key,"eof");
	    if (key->status==TOK_OK) to=-1;
	    else {
	      err_set(1,"bad value for to: %s",key->text);
	      break;
	    }
	  }else break;
	}
      }else if (key_is("step")) {
	key=Get_Int_M("step: ",";","1 2 3 4 5 6 7 8 9 10 20 30 40 50 60 70 80 90 100");
	if (key->status==TOK_OK) step=key->n;
	else {
	  if (key->status==TOK_NOK) err_set(1,"bad value for step: %s",key->text);
	  break;
	}
      }else if (key_is("read")) {
	key=free_Token(key);
	key=Get_File(NULL,".dcd","DCD trajectory file: ",";");
	if (key->status==TOK_OK) {
	  trj=key->ptr;
	  confa_dcd_read(trj,from,to,step);
	  UCpclose(trj);
	}else if (key->status==TOK_NOK) err_set(1,"DCD can't read trajectory file %s",key->text);
	break;
      }
    }else{
      if (key->status==TOK_NOK) err_set(1,"bad DCD keyword: %s",key->text);
      break;
    }
  }
  free_Token(key);
}
/*-------------------- Distance --------------------*/

static float Distance(CnfNode *a,CnfNode *b)
{
  float r;

  if (streq(cl_metric_name[metric],"fitted")) r=fittedDistance(a,b);
  else if (streq(cl_metric_name[metric],"inplace")) r=inplaceDistance(a,b);
  else r=0.0;

  return(r);
}

static float fittedDistance(CnfNode *a,CnfNode *b)
{
  int j,k;
  mat_4 R;
  float r=0.0;
  float *x=NULL,*y=NULL,*z=NULL,*px;

  x=malloc(3*nSubset*sizeof(float));
  y=malloc(3*nSubset*sizeof(float));
  z=malloc(3*NTemplate*sizeof(float));

  if (x&&y&&z) {
    getCoords(a,z);
    for (px=x,j=0;j<nSubset;j++) for (k=0;k<3;k++) *px++ = z[3*SubsetIdx[j]+k];

    getCoords(b,z);
    for (px=y,j=0;j<nSubset;j++) for (k=0;k<3;k++) *px++ = z[3*SubsetIdx[j]+k];

    if (streq(cl_metopt_name[metopt],"rmsd")) r=rms_best_mat_yx(nSubset,x,y,&R);
    else if (streq(cl_metopt_name[metopt],"gseal")) err_set(1,"Can't use gseal with fitted yet.");

  }else{
    err_set(1,"fittedDistance: memory allocation failed");
  }
  if (x) free(x);
  if (y) free(y);
  if (z) free(z); 

  return(r);
}

static float inplaceDistance(CnfNode *a,CnfNode *b)
{
  int j,k;
  float r=0.0;
  float *x=NULL,*y=NULL,*z=NULL,*px;

  x=malloc(3*nSubset*sizeof(float));
  y=malloc(3*nSubset*sizeof(float));
  z=malloc(3*NTemplate*sizeof(float));

  if (x&&y&&z) {
    getCoords(a,z);
    for (px=x,j=0;j<nSubset;j++) for (k=0;k<3;k++) *px++ = z[3*SubsetIdx[j]+k];

    getCoords(b,z);
    for (px=y,j=0;j<nSubset;j++) for (k=0;k<3;k++) *px++ = z[3*SubsetIdx[j]+k];

    if (streq(cl_metopt_name[metopt],"rmsd")) {
      for (r=0.0,j=0;j<nSubset;j++) r+=cdistsq(x+j*3,y+j*3);
      r=sqrt(r/nSubset);
    }else if (streq(cl_metopt_name[metopt],"gseal")) r=cl_cluster_gseal(x,y);
  }else{
    err_set(1,"inplaceDistance: memory allocation failed");
  }
  if (x) free(x);
  if (y) free(y);
  if (z) free(z); 

  return(r);
}

/*--------------- cl_cluster_gseal ---------------*/
static float cl_cluster_gseal(float *x, float *y)
{
  float r=0.0;
  int i,j;
  Atom *a, *b;

  for (i=0; i<nSubset; i++) {
    a = Template[SubsetIdx[i]];
    for (j=0; j<nSubset; j++) {
      b = Template[SubsetIdx[j]];
      if (a->atp->z == b->atp->z) {
	r += exp(-gseal_alpha*cdistsq(x+i*3,y+j*3));
      }
    }
  }

  return(-r);
}


/*-------------------- newCnf --------------------*/

static CnfNode *newCnf(char *name,int size)
{
  CnfNode *r=(CnfNode *)NULL;
  if (mvld(ConfaRoot)) {
    if ((r=(CnfNode *)add_node(ConfaRoot,name,OBJ_CNF))) {
      r->energy=0.0; r->n=size;
      r->x=(float *)malloc(3*size*sizeof(float));
      if (!mvld(r->x)) r=(CnfNode *)del_node((Node *)r);
      else r->cseqnr=++nxtSeq;
    }
  }
  return(r);
}
/*========================================================================*/

#define BYTESEX_UNKNOWN   0
#define BYTESEX_HOMO   1
#define BYTESEX_HETERO 2
#define TIMFAC 0.04888821 /* conversion factor from AKMA time to picoseconds.*/

static int read_header(FILE *F,int *sex,char *type,int *version,int *qcrys,
		       int *nat,int *nfix,int **mob,int *nfrm,
		       int *first,int *nsteps,float *dt,char ***cmnt);
static int read_dvec(FILE *F,int sex,int n,double *v);
static int read_fvec(FILE *F,int sex,int n,float *v);

static char *swap(int n,char *byte);

/*-------------------- confa_dcd_read --------------------*/

static void confa_dcd_read(UCpipe *trj,int from,int to,int step)
{ 
  int status;
  int k,n,nat=0,nfix=0,*mob=(int *)NULL,nfrm=0,first=0,nsteps=0,nfree=0,sex;
  int version=0,qcrys;
  float dt=0;
  char type[5],**cmnt=(char **)NULL,**c,name[256];
  float *x=(float *)NULL,*y=(float *)NULL,*z=(float *)NULL;
  double crys[6];
  CnfNode *new=(CnfNode *)NULL;

  if (!mvld(ConfaRoot)) ConfaRoot=make_node(OBJ_CNF);
  if (!mvld(ConfaRoot)) {
    err_set(1,"Failed to generate root for Confa.");
    return;
  }

  set_node_name(ConfaRoot,"ConfaRoot");

  status=read_header(trj->fd,&sex,type,&version,&qcrys,&nat,
		     &nfix,&mob,&nfrm,&first,&nsteps,&dt,&cmnt);
  if (status) {
    err_set(1,"DCD, failed to read trajectory header record.\n");
  }else{
    prtGraphics("Filetype: %s, version %d, %d atoms (%d fixed), timestep %0.4f ps\n", 
		type,version,nat,nfix,dt);
    if (nfrm==0) {
      prtGraphics("Frames: first = %d, %0.4f ps per frame, number of frames unknown\n",
		  first,nsteps*dt);
    }else{
      prtGraphics("Frames: first = %d, %0.4f ps per frame, number of frames = %d\n",
		  first,nsteps*dt,nfrm);
    }
    if (cmnt) {
      for (c=cmnt;*c;c++) {
	prtGraphics("%s\n",*c[0]==' '?*c+1:*c);
      }
    }
    nfree=nat-nfix;
    if (mob) {
      Mobile=mob;
      NMobile=nfree;
    }
    if (strcmp(type,"CORD")) {
      err_set(1,"DCD, trajectory type must be CORD (coordinates)");
      status=1;
    }
  }

  if (!status) {
    x=(float *)malloc(nat*sizeof(float));
    y=(float *)malloc(nat*sizeof(float));
    z=(float *)malloc(nat*sizeof(float));
    if (!x||!y||!z) {
      err_set(1,"DCD: memory allocation failure\n");
      if (x) free(x);
      if (y) free(y);
      if (z) free(z); 
      status=1;
    }
  }
  if (!status) {
    for(n=0;to<0||n<to;n++) {
      /* read next frame */
      if (qcrys&&(status=read_dvec(trj->fd,sex,6,crys))!=6) break; /* skip crysinf */
      if ((status=read_fvec(trj->fd,sex,nat,x))<=0) break;
      if (status!=read_fvec(trj->fd,sex,nat,y)) break;
      if (status!=read_fvec(trj->fd,sex,nat,z)) break;
      if (n+1-from>=0&&((n+1-from)%step==0)) {
	/* store frame in cnf node */
	sprintf(name,"%d",first+n*nsteps);
	if ((new=newCnf(name,status))) {
	  for (k=0;k<status;k++) {
	    new->x[3*k]=x[k];
	    new->x[3*k+1]=y[k];
	    new->x[3*k+2]=z[k];
	  }
	}else{
	  err_set(1,"failed to create Cnf node.");
	  break;
	}
      }
      status=0;
    }

    if (status) err_set(1,"DCD: error reading file after %d frames\n",n);
  }
  
  if (cmnt) {
    for (c=cmnt;*c;c++) free(*c);
    free (cmnt);
  }
  if (status&&mob) {
    Mobile=(int *)NULL;
    free(mob);
    mob=Mobile;
  }
  if (x) free(x);
  if (y) free(y);
  if (z) free(z); 

  err_dump();
  confa_status();
}

/*-------------------- read_header --------------------*/

static int read_header(FILE *F,int *sex,char *type,int *version,int *qcrys,
		       int *nat,int *nfix,int **mob,int *nfrm,
		       int *first,int *nsteps,float *dt,char ***cmnt)
{
  int r=0,n,i,nfree,nc;
  char buff[256],*c,**p;

  *cmnt=(char **)NULL;
  *sex=BYTESEX_UNKNOWN;
  *qcrys=0;

  if (fread(buff,1,92,F)==92) { /* (rl)(84 bytes)(rl) ... */
    *sex=BYTESEX_HOMO;
    memcpy(&n,buff,4);
    if (n!=84) {
      *sex=BYTESEX_HETERO;
      swap(4,buff);
      memcpy(&n,buff,4);
    }
    if (n!=84) r=1;
  }else r=2;
  if (r==0) {
    strncpy(type,buff+4,4);type[4]='\0';
    if (*sex==BYTESEX_HETERO) swap(4,buff+84);
    memcpy(version,buff+84,4);
    if (*sex==BYTESEX_HETERO) {
      swap(4,buff+8);
      swap(4,buff+40);
      swap(4,buff+12);
      swap(4,buff+16);
      if (version==0) swap(8,buff+44);  /* X-Plor and CNS, CNX, maye NAMD */
      else{
	swap(4,buff+44);                /* CHARMM */
	swap(4,buff+48);
      }
    }
    memcpy(nfrm,buff+8,4);
    memcpy(nfix,buff+40,4);
    memcpy(first,buff+12,4);
    memcpy(nsteps,buff+16,4);
    if (*version==0) {
      *dt=*(double *)(buff+44)*TIMFAC;
      *qcrys=0;
    }else{
      *dt=*(float *)(buff+44)*TIMFAC;
      *qcrys=*(int *)(buff+48);
    }
  }

  if (r==0) {                     /* REMARKS block */
    if (fread(buff,1,8,F)==8) {
      if (*sex==BYTESEX_HETERO) {
	swap(4,buff);
	swap(4,buff+4);
      }
      memcpy(&n,buff+4,4);
      memcpy(&nc,buff,4);
      if (n*80+4==nc) {
	*cmnt=(char **)malloc((n+1)*sizeof(char *));
	if (*cmnt) for (i=0;i<=n;i++) (*cmnt)[i]=(char *)NULL;
	for (i=0;i<n;i++) {
	  if (fread(buff,1,80,F)==80) {
	    if (*cmnt) {
	      buff[81]='\0';
	      for (c=buff+strlen(buff)-1;c>=buff;c--) if (*c!=' ') break;
	      c[1]='\0';
	      (*cmnt)[i]=strdup(buff);
	    }
	  }else r=3;
	}
      }else r=4;
      if (r==0&&fread(buff,1,4,F)!=4) r=5; 
    }else r=6;
  }

  if (r==0) {                       /* Number of Atoms */
    if (fread(buff,1,12,F)==12) { 
      if (*sex==BYTESEX_HETERO) {
	swap(4,buff);
	swap(4,buff+4);
      }
      *nat=*(int *)(buff+4);
      memcpy(&nc,buff,4);
      if (nc!=4) r=7;
    }else r=8;
  }

  if (r==0&&*nfix>0) {              /* free atom list */
    if (fread(buff,1,4,F)==4) {
      if (*sex==BYTESEX_HETERO) swap(4,buff);
      nfree=*nat-*nfix;
      memcpy(&nc,buff,4);
      if (nc==nfree*4) {
	*mob=(int *)malloc(nfree*sizeof(int));
	if (*mob) {
	  if (fread(*mob,4,nfree,F)==nfree) {
	    if (*sex==BYTESEX_HETERO) for (i=0;i<nfree;i++) swap(4,(char *)((*mob)+i));
	    for (i=0;i<nfree;i++) (*mob)[i]-=1;
	  }else r=9;
	}else r=10;
      }else r=11;
      if (r==0&&fread(buff,1,4,F)!=4) r=12;
    }else r=0;
  }

  if (r) {                          /* clean up */
    if (*cmnt) {
      for (p=*cmnt;*p;p++) free(*p);
      free(*cmnt);
      *cmnt=(char **)NULL;
    }
    if (*mob) {free(*mob);} *mob=(int *)NULL;
  }
  return(r);
}
/*---------------------read_dvec --------------------*/

static int read_dvec(FILE *F,int sex,int n,double *v)
{
  /* read double-vector, return n-read (success), 0 (eof), -1 (error) */

  char buff[4];
  int i,m;
  
  if (n<1) return(-1);
  if (fread(buff,1,4,F)==4) {
    if (sex==BYTESEX_HETERO) swap(4,buff);
    memcpy(&m,buff,4);
    if (m%8==0&&m/8<=n) {
      m/=8;
      if (fread(v,8,m,F)==m) {
	if (sex==BYTESEX_HETERO) {
	  for (i=0;i<m;i++) swap(8,(char *)(v+i));
	}
	if (fread(buff,1,4,F)!=4) return(-1);
      }else return(-1);
    }else return(-1);
  }else if (feof(F)) return(0);
  else return(-1);
  return (m);
}
/*---------------------read_fvec --------------------*/

static int read_fvec(FILE *F,int sex,int n,float *v)
{
  /* read float-vector, return n-read (success), 0 (eof), -1 (error) */

  char buff[4];
  int i,m;
  
  if (n<1) return(-1);
  if (fread(buff,1,4,F)==4) {
    if (sex==BYTESEX_HETERO) swap(4,buff);
    memcpy(&m,buff,4);
    if (m%4==0&&m/4<=n) {
      m/=4;
      if (fread(v,4,m,F)==m) {
	if (sex==BYTESEX_HETERO) {
	  for (i=0;i<m;i++) swap(4,(char *)(v+i));
	}
	if (fread(buff,1,4,F)!=4) return(-1);
      }else return(-1);
    }else return(-1);
  }else if (feof(F)) return(0);
  else return(-1);
  return (m);
}
/*-------------------- swap --------------------*/

static char *swap(int n,char *byte)
{
  int i,m;
  char c;

  switch (n) {
  case 2:
    c=byte[0]; 
    byte[0]=byte[1]; 
    byte[1]=c;
    break;
  case 4:
    c=byte[0]; 
    byte[0]=byte[3]; 
    byte[3]=c;
    c=byte[1]; 
    byte[1]=byte[2]; 
    byte[2]=c;
    break;
  case 8:
    c=byte[0]; 
    byte[0]=byte[7]; 
    byte[7]=c;
    c=byte[1]; 
    byte[1]=byte[6]; 
    byte[6]=c;
    c=byte[2]; 
    byte[2]=byte[5]; 
    byte[5]=c;
    c=byte[3]; 
    byte[3]=byte[4]; 
    byte[4]=c;
    break;
  default:
    m=n/2;
    for (i=0;i<m;i++) {
      c=byte[i];
      byte[i]=byte[n-i-1];
      byte[n-i-1]=c;
    }
    break;
  }
  return(byte);
}
/*-------------------- cmp_int_asc --------------------*/

static int cmp_int_asc(const void *A,const void *B)
{
  int a,b;

  a=*(int *)A;
  b=*(int *)B;
  return(a<b?-1:a>b?1:0);
}
/*-------------------- nTemplateOK --------------------*/

static int nTemplateOK(int n,Atom **T)
{
  /* set pointers in array T to NULL if atom is not in MolRoot tree,
     return number of good pointers */

  Nodechain allc;
  Atom **all,***p;
  int nall,k,na,np,nok;

  if (n>0&&T) {
    allc=tree2chain(MolRoot,OBJ_ATOM);
    nall=allc.n;
    if (nall>0) {
      all=(Atom **)chain_to_ptr(&allc,FALSE);
      clear_node_chain(&allc);
      if (all) qsort(all,nall,sizeof(Atom *),cmp_addr);
      p=(Atom ***)malloc(n*sizeof(Atom **));
      if (all&&p) {
	for (k=0;k<n;k++) p[k]=T+k;
	qsort(p,n,sizeof(Atom **),cmp_addrp);
	for (na=0,np=0;na<nall&&np<n;) {
	  if (all[na]==*p[np]) {
	    na++;
	    np++;
	  }else{
	    while(np<n&&*p[np]<all[na]) {
	      *p[np]=(Atom *)NULL;
	      np++;
	    }
	    if (np<n) while(na<nall&&all[na]<*p[np]) na++;
	  }
	}
	while (np<n) *p[np++]=(Atom *)NULL;
	for (nok=0,k=0;k<n;k++) if (T[k]) nok++;
      }else{
	err_set(1,"memory allocation failure (nTemplateOK).");
	nok=0;
      }
      if (all) free(all);
      if (p) free(p);
    }else{
      for (k=0;k<n;k++) T[k]=(Atom *)NULL;
      nok=0;
    }
  }else nok=0;
  return(nok);
}
/*-------------------- cmp_addr --------------------*/

static int cmp_addr(const void *A,const void *B)
{
  Atom *a,*b;

  a=*(Atom **)A;
  b=*(Atom **)B;
  return(a<b?-1:a>b?1:0);
}
/*-------------------- cmp_int_asc --------------------*/

static int cmp_addrp(const void *A,const void *B)
{
  Atom *a,*b;

  a=**(Atom ***)A;
  b=**(Atom ***)B;
  return(a<b?-1:a>b?1:0);
}
