#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "graphics.h"
#include "xtal.h"
#include "isosurface.h"

extern int graphix;
extern Node *DefaultNode;
extern Dspobj *DspRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define TRUE 1
#define FALSE 0

#define ABOVE   1
#define BELOW  -1

#define MXTREE 2048
#define MXLIST 1024

#define LOOP 1
#define DONE 2

#define X_AXIS (1<<0)
#define Y_AXIS (1<<1)
#define Z_AXIS (1<<2)

#define SEL_ABOVE 1
#define SEL_BELOW 2
#define SEL_RANGE 3
#define SEL_MIN   4
#define SEL_MAX   5

#define CONT_LINES 1
#define CONT_SOLID 2

static void cmd_contour_add(int type);
static void cmd_contour_read(int own_g);
static void cmd_contour_inside(void);
static void cmd_contour_sites(void);
static Node *new_molecule(char *name);
static Nodechain *select_site_points(int mode,float minf,float maxf);

static char *main_keys = "add color inside style transparency width "
                         "delete on off toggle sites render read ' ' end";

static int  inSide=ABOVE;
static float Missing=1.E30;

static Grid *G=(Grid *)NULL;

/*======================================================================*/

void cmd_contour(Grid *g)
{
  Token *key=(Token *)NULL;
  int own_g;

  G=g;
  own_g=(mvld(g) ? FALSE : TRUE);

  for (;;) {
    err_dump(); key=free_Token(key);
    key=Get_Kwd("Contour> ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("add")) cmd_contour_add(CONT_LINES);
      else if (key_is("render")) cmd_contour_add(CONT_SOLID);
      else if (key_is("delete")) cmd_surface_delete();
      else if (key_is("on")) cmd_surface_status(key->key);
      else if (key_is("off")) cmd_surface_status(key->key);
      else if (key_is("toggle"))cmd_surface_status(key->key); 
      else if (key_is("color")) cmd_surface_color();
      else if (key_is("read")) {cmd_contour_read(own_g); own_g=TRUE;}
      else if (key_is("inside")) cmd_contour_inside(); 
      else if (key_is("style")) cmd_surface_style();
      else if (key_is("transparency")) cmd_surface_transparency();
      else if (key_is("width")) cmd_surface_linewidth();
      else if (key_is("sites")) cmd_contour_sites();
      else if (key_is("end")) break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad contour option: %s",key->text);
    }else break;
  }

  if (own_g) {
    if (mvld(G)) prtGraphics("closing contour file %s\n",
			     (G->name ? G->name : " (unknown)"));
    G=freeGrid(G);
  }else G=(Grid *)NULL;
  free_Token(key);
}
/*-------------------- cmd_contour_read --------------------*/

static void cmd_contour_read(int own_g)
{
  if (own_g) {
    if (mvld(G)) prtGraphics("closing old contour file %s\n",
			   (G->name ? G->name : " (unknown)"));
    G=freeGrid(G);
  }else G=(Grid *)NULL;
  cmd_read(&G,FT_CNT);          /* note: cmd_read wants a Node* */
  if (!mvld(G)) err_set(1,"failed to read contour file.");
  else{
    prtGraphics("Grid read from file %s\n",
		(G->name ? G->name : " (unknown)"));
    prtGraphics("nx = %d, ny = %d, nz = %d\n",G->n[0],G->n[1],G->n[2]);
  }
}

/*-------------------- cmd_contour_sites --------------------*/

static void cmd_contour_sites(void)
{
  Token *key;
  char *name;
  Node *Sites,*a;
  Nodechain *c;
  float min_r,max_r;
  int sel_mode=0,select;
  char *p;

  if (!mvld(G)) cmd_contour_read(TRUE);
  if (!mvld(G)) return;

  c=(Nodechain *)NULL;
  key=(Token *)NULL;
  Sites=(Node *)NULL;
  select=FALSE; min_r=0.0;max_r=0.0;

  key=Get_Mol_D("Sites: ",";");
  if (key->status==TOK_OK) {
    Sites=(Node *)key->ptr;
  }else if (key->status==TOK_NOK) {
    name=key->text;
    if (!(Sites=new_molecule(name))) {
      err_set(1,"failed to create molecule %s.",name);
    }
  }

  if (Sites) {
    key=free_Token(key);
    key=Get_Kwd("site selection mode: ",";","above below maxima minima range");
    if (key->status==TOK_OK) {
      if (key_is("minima")) sel_mode=SEL_MIN;
      else if (key_is("maxima")) sel_mode=SEL_MAX;
      else if (key_is("above")) sel_mode=SEL_ABOVE;
      else if (key_is("below")) sel_mode=SEL_BELOW;
      else if (key_is("range")) sel_mode=SEL_RANGE;
      if (sel_mode==SEL_ABOVE||sel_mode==SEL_BELOW||sel_mode==SEL_RANGE) {
	p=(sel_mode==SEL_RANGE ? "minimum: " : sel_mode==SEL_ABOVE ? "select above: " : "select below: ");
	key=Get_Float(p,";");
	if (key->status==TOK_OK) {
	  if (sel_mode==SEL_BELOW) max_r=key->f;
	  else min_r=key->f;
	  if (sel_mode==SEL_RANGE) {
	    key=free_Token(key);
	    key=Get_Float("maximum: ",";");
	    if (key->status==TOK_OK) {
	      max_r=key->f;
	      select=TRUE;
	    }else if (key->status==TOK_NOK) err_set(1,"bad float %s",key->text);
	  }else select=TRUE;
	}else if (key->status==TOK_NOK) err_set(1,"bad float %s",key->text);
      }else select=TRUE;
    }else if (key->status==TOK_NOK) err_set(1,"bad selection mode: %d",key->text);
    if (select) {
      c=select_site_points(sel_mode,min_r,max_r);
      if (c) {
	if (c->n>0) {
	  for (a=c->s;mvld(a);a=a->chain) {
	    a->up=Sites;
	    link_node(a);
	  }
	  send_atom_chain(c);
	}
	clear_node_chain(c);
	free_node_chain(c);
      }
    }
  }

  free_Token(key);
}

/*-------------------- cmd_contour_add --------------------*/

static void cmd_contour_add(int ctype)
{
  Atom *a;
  Nodechain *c;
  Surface *S;
  Token *key;
  int i,status,color;
  float level,marg,surf,vol;
  Dspobj *glob,*d;
  char *menu;
  mat_4 btr;
  vec_3 v;

  if (!mvld(G)) cmd_contour_read(TRUE);
  if (!mvld(G)) return;

  key=(Token *)NULL;
  menu=(char *)NULL;

  key=free_Token(key);
  menu=Make_GlblNameList("/",DspRoot);
  key=Get_DspGlbl_M("object: ",";",menu);
  if (key->status==TOK_OK) glob=(Dspobj *)key->ptr;
  else {
    glob=(Dspobj *)NULL;
    if (key->status==TOK_NOK) err_set(1,"object not found");
  }
  send_bheader(graphix,"ARMK",0);
  free_menu(menu);
  if (mvld(glob)) {
    status=LOOP;
    while (status==LOOP) {
      key=free_Token(key);
      key=Get_Float_M("level = ",";","around ' ' -10 -5 -4 -3 -2 -1 0 1 2 3 4 5 10");
      if (key->status==TOK_OK) {
	level=key->f;
	key=free_Token(key);
	key=Get_Color("color: ",";");
	if (key->status == TOK_OK) {
	  color=key->n;
	  key=free_Token(key);
	  key=Get_String("unique contour name: ",";");
	  if (key->status==TOK_OK) {
	    switch (ctype) {
	    case CONT_LINES:
	      S=isoMesh(G,level,inSide,Missing,ISOS_PDB,X_AXIS+Y_AXIS+Z_AXIS,TRUE);
	      if (S) S->color=color;
	      surface_info(S);
	      addSurfaceToDisplay(S,glob,key->text);
	      S=freeSurface(S);
	      break;
	    case CONT_SOLID:
	      S=isoSurface(G,level,inSide,Missing,ISOS_PDB,TRUE);
	      if (mvld(S)) {
		S->color=color;
		SurfAndVol(S,&surf,&vol);
		prtGraphics("isosurface: Surface area = %0.3f,"
			    "Volume inside surface = %0.3f\n",surf,vol);
		addSurfaceToDisplay(S,glob,key->text);
		S=freeSurface(S);
	      }
	      break;
	    }
	  }else if (key->status==TOK_NOK) err_set(1,"bad name token: %s",key->text);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad color specification: %s",key->text);
	}else{
	  status=DONE;
	}
      }else if (key->status==TOK_NOK) {
	key=Prs_Kwd(key,"around");
	if (key->status==TOK_OK) {
	  key=free_Token(key);
	  key=Get_Atoms_D("contour around atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    key=free_Token(key);
	    key=Get_Float_M("margin: ",";",
			    "0 1 2 3 4 5 6 7 8 9 10 15 20 25 30 40 50 75 100 ' ' infinity");
	    if (key->status==TOK_NOK) {
	      key=Prs_Kwd(key,"infinity");
	      if (key->status==TOK_OK) {
		key->f=1.e30;
		key->status=TOK_OK;
	      }
	    }
	    if (key->status==TOK_OK) {
	      marg=key->f;
	      get_MVT(DspRoot);
	      btr=inv_4(glob->mtrx);
	      ForDspobjs(d,DspRoot) {
		if (d->dsptype==DSP_GLBL||d->dsptype==DSP_ROOT) d->mtrx=m4xm4(btr,d->mtrx);
	      }
	      if (c->n>0) {
		for (i=0;i<3;i++) {
		  G->xl[i]=+1.e30;
		  G->xh[i]=-1.e30;
		}
		for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
		  d=glblobj(assocobj(a->up));
		  if (is_dsp(d)) {
		    v=m4xv3(d->mtrx,vec3(a->x));
		    for (i=0;i<3;i++) {
		      G->xl[i]=(G->xl[i]>v.c[i] ? v.c[i] : G->xl[i]);
		      G->xh[i]=(G->xh[i]<v.c[i] ? v.c[i] : G->xh[i]);
		    }
		  }else{
		    for (i=0;i<3;i++) {
		      G->xl[i]=(G->xl[i]>a->x[i] ? a->x[i] : G->xl[i]);
		      G->xh[i]=(G->xh[i]<a->x[i] ? a->x[i] : G->xh[i]);
		    }
		  }
		}
	      }else{
		for (i=0;i<3;i++) {
		  G->xl[i]=0;
		  G->xh[i]=0;
		}
	      }
	      for (i=0;i<3;i++) {
		G->xl[i]-=marg;
		G->xh[i]+=marg;
	      }
	      btr=inv_4(PDB_fta_matrix(G->cl,G->ca));
	      v32arr(G->xl,m4xv3(btr,vec3(G->xl)));
	      v32arr(G->xh,m4xv3(btr,vec3(G->xh)));
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad float: %s",key->status);
	    }
	    clear_node_chain(c);
	    free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s",key->text);
	  }
	}else{
	  err_set(1,"illegal contour level: %s",key->text);
	}
      }else{
	status=DONE;
      }
      err_dump();
      if (end_parse()!=TOK_OK) err_set(0,"extra characters at end of command ignored.");
      start_parse();
    }
  }
  free_Token(key);
  free_menu(menu);
}
/*-------------------- cmd_contour_inside --------------------*/

static void cmd_contour_inside(void)
{
  int color,n;
  Token *key;
  Nodechain *c;
  Node *o;

  if (inSide==ABOVE) default_token("above");
  else default_token("below");
  key=Get_Kwd("option: ",";","above below color");
  if (key->status==TOK_OK) {
    if (key_is("above")) inSide=ABOVE;
    else if (key_is("below")) inSide=BELOW;
    else if (key_is("color")) {
      key=free_Token(key);
      key=Get_Color_M("color: ",";","frontface none ' '");
      if (key->status == TOK_OK) {
	color=key->n; if (color<0) color++;
	key=free_Token(key);
	key=Get_Objects(DSP_CNT|DSP_TRG,"change backface color for contour object(s):",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (n=0,o=c->s;mvld(o);o=o->chain) {
	    if (is_dsp(o)) if (cPtr(((Dspobj *)o)->cross)) n++;
	  }
	  if (n>0) {
	    send_bheader(graphix,"SBFC",sizeof(int)+n*sizeof(o));
	    send_bint(graphix,color);
	    for (n=0,o=c->s;mvld(o);o=o->chain) {
	      if (is_dsp(o)) {
		if (cPtr(((Dspobj *)o)->cross)) send_bptr(graphix,o);
	      }
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"contour object not found: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad color selection: %s",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad response: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- new_molecule --------------------*/

static Node *new_molecule(char *name)
{
  Node *r;
  char *p;
  Token path;
  Nodechain *c;
  Dspobj *mdl;

  /* adds a mol node with given name (incl. intermed. levels) to mol-tree,
     if node exists, returns pointer to existing node. */

  r=(Node *)NULL;

  path.status=TOK_NOK;
  path.text=name;

  if (!name||strlen(name)==0) name="sites";
  Prs_Mols(&path);
  if (path.status==TOK_OK) {
    c=(Nodechain *)path.ptr;
    if (c->n==1) {
      r=c->s;
    }else{
      err_set(1,"ambiguous molecule selection: %s",name);
    }
    clear_node_chain(c);
    free_node_chain(c);
  }else{
    if ((p=strrchr(name,'/'))) {
      *p='\0';
      r=new_molecule(name); 
      *p='/';
    }else r=DefaultNode;
    if (r) {
      r=add_node(r,(p?p+1:name),OBJ_MOL);
      if (is_mol(r)) {
	mdl=assocmdl(r);
	if (mdl) mdl_add_atoms(mdl,r);
	else send_molecule(r->up,(Mol *)r);
      }else{
	err_set(1,"failed to add mol %s to %s",p+1,name);
	r=(Node *)NULL;
      }
    }
   
  }
  return(r);
}

static Nodechain *select_site_points(int mode,float minf,float maxf)
{
  Nodechain *c;
  float *v;
  int i,j,k,l,d[6];
  Atom *a;
  mat_4 btr;

  /* returns nodechain with selected sitepoints (dummy atoms), xyz A/cart,
     atoms are not linked, not displayed... */

  c=make_node_chain();
  if (c) {
    if (mode==SEL_ABOVE) maxf=Missing;
    if (mode==SEL_BELOW) minf=-Missing;
    switch (mode) {
    case SEL_ABOVE:
    case SEL_BELOW:
    case SEL_RANGE:
      for (v=G->val,k=0;k<G->n[2];k++) {
	for (j=0;j<G->n[1];j++) {
	  for (i=0;i<G->n[0];i++,v++) {
	    if (*v>=minf&&*v<=maxf) {
	      a=(Atom *)make_node(OBJ_ATOM);
	      if (is_atom(a)) {
		set_node_name((Node *)a,"s");
		add_to_node_chain((Node *)a,c);
		a->x[0]=G->xmin[0]+i*G->dx[0];
		a->x[1]=G->xmin[1]+j*G->dx[1];
		a->x[2]=G->xmin[2]+k*G->dx[2];
		set_atom_type(a,guess_atom_type("Du"));
		a->nxyz=1;
		a->q=*v;
	      }
	    }
	  }
	}
      }
      break;
    case SEL_MIN:
      d[0]=-1; d[1]=1; d[2]=-G->n[0]; d[3]=G->n[0];
      d[4]=-G->n[0]*G->n[1]; d[5]=G->n[0]*G->n[1];
      for (k=1;k<G->n[2]-1;k++) {
	v=G->val+k*G->n[0]*G->n[1]+G->n[0];
	for (j=1;j<G->n[1]-1;j++,v+=G->n[0]) {
	  for (i=1;i<G->n[0]-1;i++) {
	    if (v[i]>=Missing||v[i]<=-Missing) continue;
	    for (l=0;l<6;l++) {
	      if (v[i+d[l]]>=Missing||v[i+d[l]]<=-Missing) break;
	      if (v[i+d[l]]<=v[i]) break;
	    }
	    if (l==6) {
	      a=(Atom *)make_node(OBJ_ATOM);
	      if (is_atom(a)) {
		set_node_name((Node *)a,"s");
		add_to_node_chain((Node *)a,c);
		a->x[0]=G->xmin[0]+i*G->dx[0];
		a->x[1]=G->xmin[1]+j*G->dx[1];
		a->x[2]=G->xmin[2]+k*G->dx[2];
		set_atom_type(a,guess_atom_type("Du"));
		a->nxyz=1;
		a->q=v[i];
	      }
	    }
	  }
	}
      }
      break;
    case SEL_MAX:
      d[0]=-1; d[1]=1; d[2]=-G->n[0]; d[3]=G->n[0];
      d[4]=-G->n[0]*G->n[1]; d[5]=G->n[0]*G->n[1];
      for (k=1;k<G->n[2]-1;k++) {
	v=G->val+k*G->n[0]*G->n[1]+G->n[0];
	for (j=1;j<G->n[1]-1;j++,v+=G->n[0]) {
	  for (i=1;i<G->n[0]-1;i++) {
	    if (v[i]>=Missing||v[i]<=-Missing) continue;
	    for (l=0;l<6;l++) {
	      if (v[i+d[l]]>=Missing||v[i+d[l]]<=-Missing) break;
	      if (v[i+d[l]]>=v[i]) break;
	    }
	    if (l==6) {
	      a=(Atom *)make_node(OBJ_ATOM);
	      if (is_atom(a)) {
		set_node_name((Node *)a,"s");
		add_to_node_chain((Node *)a,c);
		a->x[0]=G->xmin[0]+i*G->dx[0];
		a->x[1]=G->xmin[1]+j*G->dx[1];
		a->x[2]=G->xmin[2]+k*G->dx[2];
		set_atom_type(a,guess_atom_type("Du"));
		a->nxyz=1;
		a->q=v[i];
	      }
	    }
	  }
	}
      }
      break;
    }
    if (c->n>0) {
      prtGraphics("found %d site points.\n",c->n);
      btr=PDB_fta_matrix(G->cl,G->ca);
      for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) transform_atom(btr,a);
    }
  }
  return(c);
}
