#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *MolRoot,*DefaultNode;

#define key_is(a) (streq(a,key->key))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define COPY_AFTER  0
#define COPY_BEFORE 1
#define COPY_INTO   2

static void cmd_copy_atoms(int move);
static void cmd_copy_flat(int move);
static Node  *createNode(char *name,Node **root);
static Node  *GetTarget(char *prompt,Node **root);
static void copy_dots(int ns,Atom **src,int nc,Atom **cpy);

#define tFLG (1<<3)
#define MAXBUF 4096

/*========================================================================*/

void cmd_copy(int move)
{
  Token *key;
  Node *s,*t,*r,*u,*sv_default;
  Mol *new;
  Dspobj *d;
  Nodechain A;
  Atom **atoms,**copy;
  int mode,nat,nc,how,level;

  how=(move ? OBJ_CPYALLBNDS : OBJ_CPYINTRABNDS);
  how|=(OBJ_CPYSETS|OBJ_CPYXCN|OBJ_CPYNOE);

  key=(Token *)NULL;
  key=Get_Kwd("mode: ",";","-atoms -flat -molecule");
  if (key->status==TOK_OK) {
    if (key_is("-atoms")) {
      cmd_copy_atoms(move);
      key->status=TOK_EOC;
    }else if (key_is("-flat")) {
      cmd_copy_flat(move);
      key->status=TOK_EOC;
    }else{
      key=free_Token(key);
      key=Get_Mol_DM("source molecule: ",";","/");
    }
  }else if (key->status==TOK_NOK) {
    key=Prs_Mol(key);
  }
  if (key->status==TOK_OK) {
    s=(Node *)key->ptr;
    t=GetTarget("target molecule: ",&r);
    if (!mvld(t)) {
      key->status=TOK_ABORT;
      end_parse();
    }
    if (key->status==TOK_OK) {
      key=free_Token(key);
      key=Get_Kwd_W("insertion mode: ",";","after before into",FALSE);  /*compatibility with  */
      if (key->status==TOK_NOK) key=Prs_Kwd(key,"-afer -before -into"); /*earlier versions... */
      if (key->status==TOK_OK) {
	mode=key->n+1;
	if (mode==1||mode==2) {
	  sv_default=DefaultNode;
	  DefaultNode=r->up; unlink_node(r);
	  key=free_Token(key);
	  key=Get_Mol_DM("insertion point: ",";","");
	  if (key->status==TOK_OK) {
	    if (((Node *)key->ptr)->up != DefaultNode) {
	      err_set(1,"error: insertion point must be direct descendant of target.");
	      key->status=TOK_NOK;
	      del_node(t);
	    }else{
	      if (mode==1) r->prv=(Node *)key->ptr;
	      else if (mode==2) r->nxt=(Node *)key->ptr;
	      link_node(r);
	    }
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad molecule specification: %s",key->text);
	    del_node(r);
	  }
	  DefaultNode=sv_default;
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown copy/move insertion mode: %s",key->text);
	del_node(r);
      }else if (key->status!=TOK_ABORT) key->status=TOK_OK;
    }
    if (key->status==TOK_OK) {
      A=tree2chain(s,OBJ_ATOM);
      nat=A.n;
      atoms=(Atom **)chain_to_ptr(&A,FALSE);
      clear_node_chain(&A);
      for (level=1,u=(Node *)atoms[0];u&&u!=s;u=u->up) level++;
      how|=level*OBJ_CPYLEV;
      new=TreeCopyAtoms(nat,atoms,tFLG,how,t,r);
      if (new) { 
	d=assocmdl(t);
	if (mvld(d)) mdl_add_atoms(d,t);
	else send_molecule(MolRoot,(Mol *)r);
	copy=(Atom **)tree2ptr(r,OBJ_ATOM,&nc);
	copy_dots(nat,atoms,nc,copy);
	if (copy) free(copy);
	if (move) RDeleteAtomsInArray(nat,atoms);
      }else{
	err_set(1,"copy: failed to generate copy of selected atoms.");
	del_node(r);
      }
      if (atoms) free(atoms);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad source molecule specification: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_copy_atom --------------------*/

static void cmd_copy_atoms(int move)
{
  /* new version (July 2000):
         allows recursive copy,
         copies atom selections instead of molecules.
  */

  Token *key;
  Dspobj *d;
  Nodechain *A;
  Node *root,*target,*rtarget;
  Mol *new;
  int nat,nc,how;
  Atom **atoms=(Atom **)NULL,**copy;
  char buf[MAXBUF],prompt[MAXBUF];
  
  how=(move ? OBJ_CPYALLBNDS : OBJ_CPYINTRABNDS);
  how|=(OBJ_CPYSETS|OBJ_CPYXCN|OBJ_CPYNOE);
  how|=OBJ_CPYMINTREE;

  key=Get_Atoms_D((move ? "atoms to move: " : "atoms to copy: "),";");
  if (key->status==TOK_OK) {
    A=(Nodechain *)key->ptr;
    nat=A->n;
    if (nat>0) {
      root=common_root(A); if (!is_mol(root)&&root->up) root=root->up;
      sprintf(prompt,"%d atoms with root at %s. Target: \n",
		  nat,nodeptr_to_name(buf,root));
      target=GetTarget(prompt,&rtarget);
      if (target) {
	atoms=(Atom **)chain_to_ptr(A,TRUE);
	walk_order(MolRoot,nat,atoms,tFLG);
	new=TreeCopyAtoms(nat,atoms,tFLG,how,target,rtarget);
	if (new) { 
	  d=assocmdl(target);
	  if (mvld(d)) mdl_add_atoms(d,target);
	  else send_molecule(MolRoot,(Mol *)rtarget);
	  copy=(Atom **)tree2ptr(rtarget,OBJ_ATOM,&nc);
	  copy_dots(nat,atoms,nc,copy);
	  if (copy) free(copy);
	  /* copy dots here */
	  if (move) RDeleteAtomsInArray(nat,atoms);
	}else{
	  err_set(1,"copy: failed to generate copy of selected atoms.");
	  del_node(rtarget);
	}
      }else{
	clear_node_chain(A);
	free_node_chain(A);
      }
    }else {
      err_set(1,"copy: 0 atoms selected, nothing copied.");
      clear_node_chain(A);
      free_node_chain(A);
    }
  }else if (key->status==TOK_NOK) err_set(1,"invalid atom selection: %s\n.",key->text); 
  if (mvld(atoms)) free(atoms);
  key=free_Token(key);
}

/*-------------------- cmd_copy_flat --------------------*/

static void cmd_copy_flat(int move)
{
  /* copies or moves atoms into a 'flat' structure
     'move' copies all bonds, 'copy' only bonds within set
     of copied atoms. xtal and noe info not copied.
     set membership of copied atoms will be lost.
  */
 
  Token *key;
  Dspobj *d;
  Nodechain *A;
  Node *target,*rtarget,*n;
  int nat,nc,i,j;
  Atom **atoms=(Atom **)NULL,**copy,*a,*b;
  
  key=Get_Atoms_D((move ? "atoms to move: " : "atoms to copy: "),";");
  if (key->status==TOK_OK) {
    A=(Nodechain *)key->ptr;
    nat=A->n;
    if (nat>0) {
      target=GetTarget("Target: ",&rtarget);
      if (target&&!target->name&&target->up==MolRoot) {
	err_set(1,"flat copy into root node / is not allowed.");
	del_node(rtarget);
	target=(Node *)NULL;
      }
      if (target) {
	atoms=(Atom **)chain_to_ptr(A,TRUE);
	walk_order(MolRoot,nat,atoms,tFLG);
	for (i=0;i<nat;i++) {         /* make copies of atoms */
	  a=(Atom *)copy_node((Node *)atoms[i]);
	  if (!a) break;
	  a->chain=(Node *)atoms[i];
	  atoms[i]->chain=(Node *)a;
	  a->up=target; 
	  link_node((Node *)a);
	}
	if (i==nat) {
	  for (i=0;i<nat;i++) {       /* add bonds */
	    a=(Atom *)atoms[i]->chain;
	    for (j=0;j<atoms[i]->nbnd;j++) {
	      b=neighbour(j,atoms[i]);
	      if (b->chain) b=(Atom *)b->chain;
	      else if (!move) b=(Atom *)NULL;   /* only 'move' copies this bond */
	      add_bond_copy(a,b,atoms[i]->bnd[j]);
	    }
	  }
	  for (i=0;i<nat;i++) {
	    atoms[i]->chain->chain=(Node *)NULL;
	    atoms[i]->chain=(Node *)NULL;
	  }
	  d=assocmdl(target);
	  if (mvld(d)) {
	    mdl_add_atoms(d,target);
	    if (!target->name) {  /* a dummy target, re-link atoms...*/
	      while (target->fst) {
		n=target->fst;
		unlink_node(n);
		n->nxt=target;
		link_node(n);
	      }
	      del_node(target);   /* ...and delete target */
	    }
	  }else send_molecule(MolRoot,(Mol *)rtarget);
	  copy=(Atom **)tree2ptr(rtarget,OBJ_ATOM,&nc);
	  copy_dots(nat,atoms,nc,copy);
	  if (copy) free(copy);
	  /* copy dots here... */
	  if (move) RDeleteAtomsInArray(nat,atoms);
	}else{
	  err_set(1,"copy/move: failed to generate copy of selected atoms.");
	  for (i=0;i>nat;i++) atoms[i]->chain=(Node *)NULL;
	  del_node(rtarget);
	}
      }else{
	clear_node_chain(A);
	free_node_chain(A);
      }
    }else {
      err_set(1,"copy: 0 atoms selected, nothing copied.");
      clear_node_chain(A);
      free_node_chain(A);
    }
  }else if (key->status==TOK_NOK) err_set(1,"invalid atom selection: %s\n.",key->text); 
  if (mvld(atoms)) free(atoms);
  key=free_Token(key);
}

/*-------------------- cmd_merge --------------------*/

void cmd_merge(void)
{
  Token *key;
  Node *s=(Node *)NULL,*t,*c,*b,*o,*sv_default,*ins=(Node *)NULL;
  Atom **copy,**atoms;
  Nodechain A;
  char *pmt[3]={"after: ","before: ","into: "};
  int mode,nat;

  mode=-1;
  key=(Token *)NULL;
  key=Get_Mol_D("target molecule: ",";");
  if (key->status==TOK_OK) {
    t=(Node *)key->ptr;
    key=free_Token(key);
    key=Get_Mol_DM("molecule to merge with target: ",";","-after -before ' '");
    if (key->status==TOK_NOK) {
      key=Prs_Kwd(key,"-after -before");
      if (key->status==TOK_OK) {
	mode=key->n+1;
	sv_default=DefaultNode;
	DefaultNode=t;
	key=free_Token(key);
	key=Get_Mol_DM(pmt[mode-1],";","");
	DefaultNode=sv_default;
	if (key->status==TOK_OK) {
	  ins=(Node *)key->ptr;
	  if (ins->up!=t) {
	    err_set(1,"merge: insertion point not direct descendant of target.");
	    key->status=TOK_ABORT;
	    end_parse();
	  }else{
	    key=free_Token(key);
	    key=Get_Mol_DM("molecule to merge with target: ",";","");
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"molecule %s not found.",key->text);
	  key->status=TOK_ABORT;
	  end_parse();
	}
      }
    }
    if (key->status==TOK_OK) {
      s=(Node *)key->ptr;
      if (!mvld(t->up)||!mvld(s->up)) {
	err_set(1,"root node is not a valid target for merge.");
	key->status=TOK_NOK;
      }
      for (b=t;mvld(b);b=b->up) if (b==s) break;
      if (mvld(b)) {
	err_set(1,"merge: target is descendant of source, bad.");
	key->status=TOK_NOK;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"merge bad molecule selection: %s",key->text);
    }
    if (mode<0&&key->status==TOK_OK) {
      mode=3;
      key=free_Token(key);
      key=Get_Kwd_W("insertion  mode: ",";","after before into",FALSE);
      if (key->status==TOK_OK) {
	mode=key->n+1;
	if (mode==1||mode==2) {
	  sv_default=DefaultNode;
	  DefaultNode=t;
	  key=free_Token(key);
	  key=Get_Mol_D("insertion point: ",";");
	  DefaultNode=sv_default;
	  if (key->status==TOK_OK) {
	    ins=(Node *)key->ptr;
	    if (ins->up!=t) {
	      err_set(1,"error: insertion point must be direct descendant of target.");
	    key->status=TOK_NOK;
	    }
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad molecule specification: %s",key->text);
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown copy/move insertion mode: %s",key->text);
      }else if (key->status!=TOK_ABORT) key->status=TOK_OK;
    }
    if (key->status==TOK_OK) {
      c=tree_copy(s);
      if (mvld(c)) {
	c->up=t; link_node(c);
	copy_set_bits(s,c);
	unlink_node(c);
	A=tree2chain(c,OBJ_ATOM);
	switch(mode) {
	case 1:
	  while (mvld(c->lst)) {
	    o=c->lst;
	    unlink_node(o);
	    o->prv=ins;
	    link_node(o);
	  }
	  break;
	case 2:
	  while (mvld(c->fst)) {
	    o=c->fst;
	    unlink_node(o);
	    o->nxt=ins;
	    link_node(o);
	  }
	  break;
	case 3:
	  while (mvld(c->fst)) {
	    o=c->fst;
	    unlink_node(o);
	    o->up=t;
	    link_node(o);
	  }
	  break;
	}
	merge_atomsets(t);
	send_atom_chain(&A);
	atoms=(Atom **)tree2ptr(s,OBJ_ATOM,&nat);
	copy=(Atom **)chain_to_ptr(&A,FALSE);
	copy_dots(nat,atoms,A.n,copy);
	if (atoms) free(atoms);
	if (copy) free(copy);
	clear_node_chain(&A);
	rdel_obj(s,OBJ_MOL|OBJ_ATOM);
	del_node(c);
      }else err_set(1,"merge failed to copy tree.");
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad merge target specificaton: %s",key->text);
  free_Token(key);
}

/*-------------------- GetTarget --------------------*/

static Node *GetTarget(char *prompt,Node **root)
{
  Node *t=(Node *)NULL,*ins,*sv_def;
  Token *key=(Token *)NULL;
  int mode;
  char *pmt[3]={"after: ","before: ","into: "};

  if (root) *root=(Node *)NULL;
  key=Get_Mol_M(prompt,";","-after -before -into");
  if (key->status==TOK_OK) {
    t=add_node((Node *)key->ptr,(char *)NULL,OBJ_MOL);
    if (root) *root=t;
    if (!mvld(t)) err_set(1,"failed to create target node.\n");
  }else if (key->status==TOK_NOK) {
    key=Prs_Kwd(key,"-after -before -into");
    if (key->status==TOK_OK) {
      mode=key->n;
      key=free_Token(key);
      key=Get_Mol_DM(pmt[mode],";","");
      if (key->status==TOK_OK) {
	ins=(Node *)key->ptr;
	sv_def=DefaultNode;
	DefaultNode=(mode!=2&&mvld(ins->up) ? ins->up : ins);
	key=free_Token(key);
	key=Get_String(prompt,";");
	if (key->status==TOK_OK) {
	  t=createNode(key->text,root);
	  if (mvld(t)) {
	    if (root) {
	      if (mode!=2&&(*root)->up==DefaultNode) {
		unlink_node(*root);
		if (mode==0) (*root)->prv=ins;
		else (*root)->nxt=ins;
		link_node(*root);
	      }
	    }else{
	      if (mode!=2&&t->up==DefaultNode) {
		unlink_node(t);
		if (mode==0) t->prv=ins;
		else t->nxt=ins;
		link_node(t);
	      }
	    }
	  }else err_set(1,"failed to create target node.\n");
	}else if (key->status==TOK_NOK) err_set(1,"bad target specification: %s",key->text);
	DefaultNode=sv_def;
      }else err_set(1,"molecule %s not found.",key->text);
    }else{
      t=createNode(key->text,root);
      if (!mvld(t)) err_set(1,"failed to create target node.\n");
    }
  }
  free_Token(key);
  return(t);
}
/*-------------------- createNode --------------------*/

static Node *createNode(char *name,Node **root)
{
  char *nn,*hd,*r;
  Nmatch *nmatch;
  Node *t,*m=(Node *)NULL;
  int hit;

  if (root) *root=(Node *)NULL;
  nn=Strdup(name);
  if (mvld(nn)) {
    r=strrchr((*nn=='/' ? nn+1 : nn),'/');
    if (r) {
      *r++ = '\0';
      hd=nn;
    }else {
      r=(*nn=='/' ? nn+1 : nn);
      hd=(*nn=='/' ? "/" : "./");
    }
    if (!strpbrk(r,"*?+#%")) { /* wild cards are not allowd in 'r' */
      nmatch=open_nmatch(hd);
      t=next_nmatch(nmatch);
      hit=(is_mol(t) ? (next_nmatch(nmatch) ? 2 : 1 ) : 0);
      close_nmatch(nmatch);	
      switch (hit) {
      case 0:
	if (!streq(hd,"/") && !streq(hd,"./")) {
	  t=createNode(nn,root);
	  m=add_node(t,(strlen(r)>0?r:NULL),OBJ_MOL);
	}else m=(Node *)NULL;
	break;
      case 1:
	m=add_node(t,(strlen(r)>0?r:NULL),OBJ_MOL);
	if (root) *root=m;
	break;
      case 2:
	m=(Node *)NULL;
	break;
      }
    }else{
      m=(Node *)NULL;
    }
    free(nn);
  }else m=(Node *)NULL;
  return(m);
}
/*-------------------- copy_dots --------------------*/

static void copy_dots(int ns,Atom **src,int nc,Atom **cpy)
{
  int i,n;

  if (src&&cpy&&ns>0&&nc>0) {
    for (n=0,i=0;i<ns&&i<nc;i++) {
      if (is_atom(src[i])&&cPtr(src[i]->cross)&&
	  is_atom(cpy[i])&&cPtr(cpy[i]->cross)) n++;
    }
    if (n>0) {
      send_bheader(graphix,"CDOT",2*n*cPtrSize());
      for (n=0,i=0;i<ns&&i<nc;i++) {
	if (is_atom(src[i])&&cPtr(src[i]->cross)&&
	    is_atom(cpy[i])&&cPtr(cpy[i]->cross)) {
	  send_bcptr(graphix,src[i]->cross);
	  send_bcptr(graphix,cpy[i]->cross);
	}
      }
    }
  }
}
