#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "fileutil.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;

#define TRUE 1
#define FALSE 0

#define key_is(x) (streq(x,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

static void cmd_define_atomset(void);
static void cmd_define_fragment(void);
static void cmd_define_macros(void);


/*========================================================================*/

void cmd_define(void)
{
  Token *key;

  key=(Token *)NULL;

  key=Get_Kwd("object type: ",";","fragments macros set");
  if (key->status==TOK_OK) {
    if (key_is("set")) cmd_define_atomset();
    else if (key_is("fragments")) cmd_define_fragment();
    else if (key_is("macros")) cmd_define_macros();
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown object type: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_define_atomset --------------------*/

static void cmd_define_atomset(void)
{
  Token *key;
  char *name,*sname=(char *)NULL;
  Nodechain *c;
  Node *root=(Node *)NULL,*save_default;

  key=(Token *)NULL;
  name=(char *)NULL;

  key=Get_String("name of set: ",";");
  if (key->status==TOK_OK) {
    name=key->text; key->text=(char *)NULL;
    sname=strrchr(name,'/');
    if (sname) {
      *sname++='\0';
      if (strlen(sname)>0) {
	if (strlen(name)==0) {
	  root=MolRoot;
	}else{
	  key->text=name;
	  key=Prs_Mols(key);
	  key->text=(char *)NULL;
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    if (c->n==1) {
	      root=(Node *)c->s;
	    }else{
	      err_set(1,"ambiguous root for set %s: %s.",sname,name);
	      key->status=TOK_ABORT;
	    }
	    clear_node_chain(c);
	    free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"root %s not found for set %s.",name,sname);
	    key->status=TOK_ABORT;
	  }
	}
      }else{
	*sname='/';
	err_set(1,"must specify a set name (%s)",name);
	key->status=TOK_ABORT;
      }
    }else{
      root=DefaultNode;
      sname=name;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad name for atomset: %s",key->text);
    key->status=TOK_ABORT;
  }
  if (key->status==TOK_OK) {
    save_default=DefaultNode;
    DefaultNode=root;
    key=free_Token(key);
    key=Get_Atoms_D("atoms: ",";");
    if (key->status==TOK_OK) {
      define_atomset((Mol *)root,sname,key->ptr);
      clear_node_chain(key->ptr); free_node_chain(key->ptr);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad atom selection: %s",key->text);
    }
    DefaultNode=save_default;
  }
  free_Token(key);
  if (mvld(name)) free(name);
}
/*-------------------- cmd_define_fragment --------------------*/

static void cmd_define_fragment(void)
{
  /* associate filename with fragment library */

  Token *key=NULL;
  char *name;
  UCpipe *L;

  key=(Token *)NULL; name=(char *)NULL;
  key=Get_String("library name: ",";");
  if (key->status==TOK_OK) {
    name=key->text; key->text=(char *)NULL;
    key=free_Token(key);
    key=Get_File((char *)NULL,".frglib","library file: ",";");
    if (key->status==TOK_OK) {
      L=(UCpipe *)key->ptr;
      define_fragment_lib(name,L);
      UCpclose(L);
    }else if (key->status==TOK_NOK) err_set(1,"bad library file: %s",key->text);
  }else if (key->status==TOK_NOK) err_set(1,"bad library name: %s",key->text);
  free_Token(key);
  if (mvld(name)) free(name);
}
/*-------------------- cmd_define_macros --------------------*/

static void cmd_define_macros(void)
{
  Token *key=NULL;
  wnpmacro_t *macros;

  key=Get_Kwd("macro type: ",";","maccs");
  if (key->status==TOK_OK) {
    if (key_is("maccs")) {
      macros=getMACCSmacros();
      cmd_define_macro(macros);
    }
  }else if (key->status==TOK_NOK) err_set(1,"define_macros: %s is not a known macro type",key->text);

  free_Token(key);
}
