#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *DefaultNode;

#define TRUE 1
#define FALSE 0

#define key_is(x) (streq(x,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

void cmd_delete(void)
{
  Token *atoms,*key,*obj_name;
  Node *obj,*m,*sv_default;
  Atom **atms,*neigh;
  Nodechain *c,O;
  Nmatch *nmatch;
  int i,j,ndel,natms,wait_flag;
  char *menu,*p;
  
  key=(Token *)NULL;
  menu=(char *)NULL;

  key=Get_Kwd("object type: ",";","atoms bonds fragments molecules objects sets");
  if (key->status==TOK_OK) {
    if (key_is("atoms")) {
      atoms=Get_Atoms_D("atoms: ",";");
      ndel=0;
      if (atoms->status==TOK_OK) {
	c=(Nodechain *)atoms->ptr;
	ndel=c->n;
	atms=(Atom **)chain_to_ptr(c,TRUE);
	RDeleteAtomsInArray(ndel,atms);
	if (atms) free(atms);
	else if (ndel>0) err_set(1,"memory allocation failure.");
      }
      free_Token(atoms);
      send_bheader(graphix,"ARMK",0);
      prtGraphics("%d atoms deleted.\n",ndel);
    }else if (key_is("molecules")) {
      ndel=0;wait_flag=TRUE;
      for (;;) {
	err_dump();
	key=free_Token(key);
	key=Get_Mols_DMW("molecules: ",";","* */",wait_flag);
	wait_flag=FALSE;
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  ndel+=RDeleteNodesInChain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad mol token: %s.",key->text);
	}else break;
      }
      prtGraphics("%d mol-nodes deleted.\n",ndel);
    }else if (key_is("fragments")) {
      ndel=0;wait_flag=TRUE;
      sv_default=DefaultNode;
      DefaultNode=fragment_root();
      if (is_mol(DefaultNode)) {
	for (;;) {
	  err_dump();
	  key=free_Token(key);free_menu(menu);
	  menu=Make_TopMolNameList("*",DefaultNode);
	  if (mvld(menu)) for (p=menu;*p;p++) *p=(*p=='/' ? ' ' : *p);
	  key=Get_Mols_MW("fragments: ",";",menu,wait_flag);
	  send_bheader(graphix,"ARMK",0);
	  wait_flag=FALSE;
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (m=c->s;mvld(m);) {
	      obj=m;
	      m=m->chain;
	      obj->chain=(Node *)NULL;
	      ndel+=del_fragment(obj);
	    }
	    free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad mol token: %s.",key->text);
	  }else break;
	}
	prtGraphics("%d fragment libraries deleted.\n",ndel);
      }else{
	err_set(1,"no fragment libraries found.");
      }
      DefaultNode=sv_default;
    }else if (key_is("sets")) {
      ndel=0;wait_flag=TRUE;
      err_dump();
      key=free_Token(key);
      key=Get_Sets_D("atomsets: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (m=c->s;mvld(m);) {
	  obj=m;
	  m=m->chain;
	  del_node(obj);
	  ndel++;
	}
	free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad set token: %s.",key->text);
      }
      prtGraphics("%d atomsets deleted.\n",ndel);
    }else if (key_is("objects")) {
      obj_name=Get_String_M("object: ",";","* /*"); /* should do better menu */
      if (obj_name->status == TOK_OK) {
	nmatch=open_nmatch(obj_name->text);
	init_node_chain(&O);
	while((obj=next_nmatch(nmatch))) add_to_node_chain(obj,&O);
	ndel=RDeleteNodesInChain(&O);
	close_nmatch(nmatch);	
	prtGraphics("%d objects deleted.\n",ndel);
      }
      free_Token(obj_name);
    }else if (key_is("bonds")) {
      key=free_Token(key);
      key=Get_Atoms_D("from atoms: ",";");
      if (key->status==TOK_OK) {
	natms=((Nodechain *)key->ptr)->n;
	atms=(Atom **)chain_to_ptr((Nodechain *)key->ptr,TRUE);
	key=free_Token(key);
	key=Get_Atoms_D("to atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  lock_evtq();
	  for (i=0;i<natms;i++) {
	    for (j=atms[i]->nbnd-1;j>=0;j--) {
	      neigh=neighbour(j,atms[i]);
	      if (node_in_chain((Node *)neigh,c)) del_bond(atms[i]->bnd[j]);
	    }
	  }
	  unlock_evtq();
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	if (atms) free(atms);
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown object type: %s\n",key->text);
  }
  free_Token(key);
  if (menu) free(menu);
  send_bheader(graphix,"ARMK",0);
}
