#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "list.h"

extern int graphix;
extern Node *DefaultNode;

#define TRUE 1
#define FALSE 0

#define MAXBUF 1024

#define FIX_FLG (1<<6)

#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define key_is(a) (streq(a,key->key))

typedef struct DRdata {
  Atom *from,*to;
  float lower,upper;
  float klower,kupper,maxforce;
} DRdata;

static DRdata default_dr_data = {
  (Atom *)NULL,(Atom *)NULL,
  -1.0, -1.0,
  50.0, 50.0,
  500.0
};

static int write_restraints(FILE *out,Node *root,ListEntry *head);
static void init_dr_data(void *data);
static void del_dr_data(void *data);
static void add_restraint(ListEntry *restr,DRdata *r);
static void clear_restraints(ListEntry *restr);

static int discover_end(void);
static void discover_run(Node *n);
static int discover_script(FILE *out,Node *n);
static void discover_restraints(Node *n,ListEntry *restr);
static void write_fixed_list(FILE *out,Node *root,char *qual,Nodechain *atoms);
static int discover_parse_line(char *buf,FILE *out,Node *root);
static int discover_parse_token(char what,char *in,FILE *out,Node *root);
static int MolNr(Node *m,Node *n);
static void cmd_add_restraints(Node *n,ListEntry *restr);
static void cmd_write_restraints(Node *n,ListEntry *restr);
static void cmd_clear_restraints(ListEntry *restr);
static void discover_peek(void);

static char *main_keys = "Peek Restraints Run ' ' End";

/*========================================================================*/

void cmd_discover(void)
{
  Token *key;
  FILE *out;
  Node *root=(Node *)NULL;
  int loop;
  ListEntry *restr;
  
  key=(Token *)NULL;
  out=(FILE *)NULL;
  restr=make_list_entry(sizeof(DRdata),init_dr_data);
  
  key=Get_Mol_D("molecule: ",";");
  if (key->status==TOK_OK) {
    root=(Node *)key->ptr;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad molecule selection: %s",key->text);
  }

  if (key->status==TOK_OK) {
    for (loop=TRUE;loop;) {
      err_dump();
      key=free_Token(key);
      key=Get_Kwd("discover option: ",";",main_keys);
      if (key->status==TOK_OK) {
	if (key_is("Restraints")) discover_restraints(root,restr);
	else if (key_is("Run")) discover_run(root);
	else if (key_is("Peek")) discover_peek();
	else if (key_is("End")) loop=discover_end();
      }else{
	if (key->status==TOK_NOK) err_set(1,"bad discover option: %s",key->text);
      }
    }
  }
  
  delete_list(restr,del_dr_data);
  if (out) fclose(out);
  free_Token(key);
}
/*-------------------- discover_peek --------------------*/

static void discover_peek(void)
{
  Token *key;
  UCpipe *P;
  FILE *p;
  char rec[MAXBUF];

  key=Get_File((char *)NULL,".pek","peek file: ",";");
  if (key->status==TOK_OK) {
    P=(UCpipe *)key->ptr;
    p=P->fd;
    while (fgets(rec,MAXBUF-1,p)) {
      if (strneq(rec+1,"Time used",9)||
	  strneq(rec+1,"Executing",9)||
	  strneq(rec+1,"Last step",9)||
	  strneq(rec+1,"Average a",9)||
	  strneq(rec+1,"Stand. de",9)||
	  strneq(rec+1,"RMS deriv",9)||
	  strneq(rec+26,"=total en",9)) prtGraphics("%s",rec);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad peek file token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- discover_run --------------------*/

static void discover_run(Node *n)
{
  FILE *out;
  Token *key;
  char cmd[MAXBUF],rec[MAXBUF],*srv,*jobname,*resource=(char *)NULL;
  UCpipe *Srv;
  int status;

  out=(FILE *)NULL;
  key=(Token *)NULL;
  jobname=(char *)NULL;

  key=Get_String("jobname: ",";");
  if (key->status==TOK_OK) {
    jobname=Strdup(key->text);
    extfil(rec,jobname,".inp",TRUE);
    out=fopen(rec,"w");
    if (!out) {
      key->status=TOK_ABORT;
      err_set(1,"could not open file %s for write.",rec);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad jobname: %s",key->text);
    key->status=TOK_ABORT;
  }

  if (key->status==TOK_OK) {
    if (!discover_script(out,n)) key->status=TOK_ABORT;
    if (out) fclose(out);
  }

  if (key->status==TOK_OK) {
    key=free_Token(key);
    key=Get_Kwd("run ? ",";","yes no");
    if (key->status==TOK_OK) {
      if (key_is("yes")) {
	key=free_Token(key);
	default_token("1");
	key=Get_Kwd("nprocs: ",";","1 2 3 4 5 6 7 8 9 10 11 12");
	if (key->status==TOK_OK) {
	  resource=Strdup(key->key);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"not a valid resource: %s",key->text);
	}
	if (key->status==TOK_OK) {
	  WriteMolecule(n,jobname,FT_CAR,FALSE);   /* should check error ! */
	  srv=getenv("DISCOVERserver");
	  if (srv) {
	    sprintf(cmd,"%s %s %s",srv,jobname,resource);
	    if (*cmd=='|') {
	      prtGraphics("running %s (foreground)...\n",cmd);
	      Srv=UCpopen(cmd);
	      if (Srv) {
		for(;;) {
		  if (checkabort()) {
		    prtGraphics("Killed...\n");
		    break;
		  }
		  if (!fgets(rec,sizeof(rec)-1,Srv->fd)) break;
		  else prtGraphics(rec);
		}
		UCpclose(Srv);
	      }else{
		err_set(1,"could not start server in foreground.");
	      }
	    }else{
	      sprintf(cmd+strlen(cmd)," 1> %s.log 2>&1 &",jobname);
	      prtGraphics("starting %s\n",cmd);
	      status=system(cmd);
	      if (status) {
		err_set(1,"shell returns with error status %d",status);
	      }
	    }
	  }else err_set(1,"sorry, DISCOVERserver is not defined.");
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s, expected 'yes' or 'no'.",key->text);
    }
  }

  if (jobname) free(jobname);
  free_Token(key);
}
/*-------------------- discover_script --------------------*/

static int discover_script(FILE *out,Node *n)
{
  Token *key;
  UCpipe *S;
  char buf[MAXBUF];
  int loop,ok;
  
  ok=FALSE;
  S=(UCpipe *)NULL;
  key=(Token *)NULL;
  key=Get_File((char *)NULL,".tpl","script template file: ",";");
  if (key->status==TOK_OK) {
    S=(UCpipe *)key->ptr;
  }else if (key->status==TOK_NOK) {
    err_set(1,"failed to open script template for read: %s",key->text);
    key->status=TOK_ABORT;
  }

  if (key->status==TOK_OK) {
    for (ok=loop=TRUE;loop&&ok;) {
      if (!fgets(buf,MAXBUF-1,S->fd)) loop=FALSE;
      else ok=discover_parse_line(buf,out,n);
    }
  }

  free_Token(key);
  UCpclose(S);
  return(ok);
}
/*-------------------- discover_parse_line --------------------*/

static int discover_parse_line(char *in,FILE *out,Node *root)
{
  int cont;
  char *pl,*pr,c;

  cont=TRUE;

  pr=(char *)in;
  pl=strchr(in,'{');
  while (pl&&cont) {
    if (pl[1]=='?'||pl[1]=='!') {
      c=pl[1];
      *pl++='\0'; *pl++='\0';
      fprintf(out,"%s",pr);
      pr=strchr(pl,'}');
      if (pr) {
	*pr++='\0';
	cont=discover_parse_token(c,pl,out,root);
	pl=strchr(pr,'{');
      }else{
	err_set(1,"missing right brace in script template.");
	*--pl=c; *--pl='{';
	fprintf(out,"%s",pl);
	cont=FALSE;
      }
    }else{
      *pl++='\0';
      fprintf(out,"%s",pr);
      pr=pl;
      pl=strchr(pr,'{');
    }
  }
  fprintf(out,"%s",pr);
  return(cont);
}
/*-------------------- discover_parse_token --------------------*/

static int discover_parse_token(char what,char *in,FILE *out,Node *root)
{
  Parse p;
  int cont=FALSE;
  char buf[MAXBUF],*kwd,*prompt,*def;
  Token *key=(Token *)NULL;
  Nodechain *c;

  switch (what) {
  case '?' :
    
    cont=TRUE; key=(Token *)NULL;
    p.buf=in; p.sep=","; p.mts=MAXBUF; p.pos=0;
    
    p.tok=buf;
    kwd=token(&p); if (kwd) p.tok+=strlen(kwd)+1;
    prompt=token(&p); if (prompt) p.tok+=strlen(prompt)+1;
    def=token(&p);
    
    if (strceq(kwd,"int")) {
      default_token(def);
      key=Get_Int((prompt ? prompt : "int: "),"");
      if (key->status==TOK_OK) {
	fprintf(out,"%d",key->n);
      }else{
	if (key->status==TOK_NOK) err_set(1,"bad integer token: %s",key->text);
	cont=FALSE;
      }
      key=free_Token(key);
    }else if (strceq(kwd,"float")) {
      default_token(def);
      key=Get_Float((prompt ? prompt : "float: "),"");
      if (key->status==TOK_OK) {
	fprintf(out,"%0.4f",key->f);
      }else{
	if (key->status==TOK_NOK) err_set(1,"bad floating point  token: %s",key->text);
	cont=FALSE;
      }
      key=free_Token(key);
    }else if (strceq(kwd,"fixed")) {
      key=Get_Atoms_D((prompt ? prompt : "fix atoms: "),"");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	chain_reduce_subtree(root,c);
	key->status=TOK_NOK;
	while (key->status==TOK_NOK) {
	  key=free_Token(key);
	  default_token("all");
	  key=Get_Kwd("selection qualifier: ",";","all calpha heavy hydrogen main side");
	  if (key->status==TOK_OK) {
	    write_fixed_list(out,root,key->key,c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad qualifier: %s",key->text);
	  }
	}
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NULL) {
	write_fixed_list(out,root,"all",(Nodechain *)NULL);
	prtGraphics("no fixed atoms.\n");
      }else{
	if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	cont=FALSE;
      }
      free_Token(key);
    }else{
      err_set(1,"unrecognizable token type: %s",kwd);
      cont=FALSE;
    }
    break;

  case '!' :
    cont=TRUE;
    prtGraphics("%s\n",in);
    break;

  default:
    err_set(1,"internal error: fugu 1 in discover_parse_token.");
    err_set(1,"token is {%c%s}",what,in);
    break;
  }
  return(cont);
}
/*-------------------- write_fixed_list --------------------*/

static void write_fixed_list(FILE *out,Node *n,char *qual,Nodechain *c)
{
  Node *r,*ra,*re;
  Atom *a,*s,*e;
  int molnr,whole;
  char Ra[20],Re[20],Ran[20],Ren[20];
  
  if (!mvld(c)||c->n==0) {
    fprintf(out,"      fixed atom list generation brief output\n");
    fprintf(out,"     *   clear\n");
  }else{
    fprintf(out,"      fixed atom list generation brief output\n");
    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg|=FIX_FLG;
    clear_node_chain(c);
    ForAtoms(a,n) {
      if (a->flg&FIX_FLG) {
	a->flg&=~FIX_FLG;
	add_to_node_chain((Node *)a,c);
      }
    }  /* atom chain now sorted by walk-through order */
    
    ra=(Node *)NULL; re=(Node *)NULL;
    for (a=(Atom *)c->s;mvld(a);) {
      r=a->up;
      whole=(a==first_atom(r));
      s=a; e=a;
      while(a->up==r) {
	if (e!=a&&next_atom(e)!=a) break;
	e=a;
	a=(Atom *)a->chain;
	if (!mvld(a)) break;
      }
      whole&=next_atom(e)==(Atom *)NULL;
      if (whole) {
	if (re) {
	  if (r!=(Node *)next_mol(re)) {
	    molnr=MolNr(re->up,n);
	    if (ra==re) {
	      make_bio_resnam(re->name,Re,Ren);
	      fprintf(out,"     *   add %s\n",qual);
	      fprintf(out,"     *       molecule %d residue %s %s\n",molnr,Re,Ren);
	    }else{
	      make_bio_resnam(ra->name,Ra,Ran);
	      make_bio_resnam(re->name,Re,Ren);
	      fprintf(out,"     *   add %s\n",qual);
	      fprintf(out,"     *       molecule %d residue %s %s to %s %s\n",molnr,Ra,Ran,Re,Ren);
	    }
	    ra=r;
	  }
	}else{
	  ra=r;
	}
	re=r;
	if (!mvld(a)) {
	  molnr=MolNr(re->up,n);
	  if (ra==re) {
	    make_bio_resnam(re->name,Re,Ren);
	    fprintf(out,"     *   add %s\n",qual);
	    fprintf(out,"     *       molecule %d residue %s %s\n",molnr,Re,Ren);
	  }else{
	    make_bio_resnam(ra->name,Ra,Ran);
	    make_bio_resnam(re->name,Re,Ren);
	    fprintf(out,"     *   add %s\n",qual);
	    fprintf(out,"     *       molecule %d residue %s %s to %s %s\n",molnr,Ra,Ran,Re,Ren);
	  }
	}
      }else{
	if (re) {
	  molnr=MolNr(re->up,n);
	  if (ra==re) {
	    make_bio_resnam(re->name,Re,Ren);
	    fprintf(out,"     *   add %s\n",qual);
	    fprintf(out,"     *       molecule %d residue %s %s\n",molnr,Re,Ren);
	  }else{
	    make_bio_resnam(ra->name,Ra,Ran);
	    make_bio_resnam(re->name,Re,Ren);
	    fprintf(out,"     *   add %s\n",qual);
	    fprintf(out,"     *       molecule %d residue %s %s to %s %s\n",molnr,Ra,Ran,Re,Ren);
	  }
	  re=(Node *)NULL;
	  ra=(Node *)NULL;
	}
	molnr=MolNr(r->up,n);
	make_bio_resnam(r->name,Re,Ren);
	if (e==s) {
	  fprintf(out,"     *   add %s\n",qual);
	  fprintf(out,"     *       molecule %d residue %s %s atom %s\n",molnr,Re,Ren,e->name);
	}else{
	  fprintf(out,"     *   add %s\n",qual);
	  fprintf(out,"     *       molecule %d residue %s %s atom %s to %s\n",molnr,Re,Ren,s->name,e->name);
	}
      }
    }
  }
}
/*-------------------- MolNr --------------------*/

static int MolNr(Node *m,Node *n)
{
  /* Molecule number of node m in root n */

  int nr;
  Node *t;

  if (m==n) nr=1;
  else if (m->up!=n) nr=0;    /* works only for level 3 and 4 structures */
  else{
    for (nr=0,t=n->fst;mvld(t);t=t->nxt) {
      if (is_mol(t)) nr++;
      if (m==t) break;
    }
  }
  return(nr);
}
/*-------------------- discover_restraints --------------------*/

static void discover_restraints(Node *root,ListEntry *restr)
{
  char *menu = "add clear  ' ' write ' ' end";
  Token *key;
  
  key=(Token *)NULL;

  for (;;) {
    start_parse();
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("restraints option: ",";",menu);
    if (key->status==TOK_OK) {
      if (key_is("add"))          cmd_add_restraints(root,restr);
      else if (key_is("clear"))   cmd_clear_restraints(restr);
      else if (key_is("write"))   cmd_write_restraints(root,restr);
      else if (key_is("end"))     break;
    }else{
      if (key->status==TOK_NOK) err_set(1,"bad discover restraints option: %s",key->text);
    }
    if (end_parse()!=TOK_OK) {
      err_set(0,"extra characters at end of command ignored.");
    }
  }
  free_Token(key);
}
/*-------------------- cmd_add_restraints --------------------*/

static void cmd_add_restraints(Node *root,ListEntry *restr)
{
  Token *key;
  DRdata r;
  Dspobj *proot;
  char buf[1024],dmenu[1024];
  float defup,dnow=0,xf[3],xt[3];

  proot=pick_root(glblobj(assocobj(root)));
  r=default_dr_data;
  key=(Token *)NULL;
  defup = default_dr_data.upper;

  do {
    err_dump();
    key=free_Token(key);
    key=Get_Atom("first atom: ",";");
    if (key->status==TOK_OK) {
      r.from=(Atom *)key->ptr;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad atom selection: %s",key->text);
    }else{
      key->status=TOK_ABORT;
    }
    if (key->status==TOK_OK) {
      key=free_Token(key);
      key=Get_Atom("second atom: ",";");
      if (key->status==TOK_OK) {
	r.to=(Atom *)key->ptr;
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad atom selection: %s",key->text);
      }else{
	key->status=TOK_ABORT;
      }
    }
    if (key->status==TOK_OK) {
      if (!mvld(find_node(root,(void *)r.from,OBJ_ATOM))) {
	err_set(1,"first atom not found in %s",nodeptr_to_name(buf,root));
	key->status=TOK_NOK;
      }
      if (!mvld(find_node(root,(void *)r.to,OBJ_ATOM))) {
	err_set(1,"second atom not found in %s",nodeptr_to_name(buf,root));
	key->status=TOK_NOK;
      }
    }
    if (key->status==TOK_OK) {
      get_graph_xyz(1,xf,&r.from);
      get_graph_xyz(1,xt,&r.to);
      dnow=cdist(xf,xt);
      if (cPtr(r.to->cross) && cPtr(r.from->cross)) {
	send_bheader(graphix,"AMON",2*cPtrSize());
	send_bcptr(graphix,r.from->cross);
	send_bcptr(graphix,r.to->cross);
      }
    }
    if (key->status==TOK_OK) {
      key=free_Token(key);
      default_token("%0.3f",dnow);
      sprintf(dmenu,"%0.3f 1.0 2.0 3.0 4.0 5.0 7.5 10.0 20.0 30.0 40.0 50.0",defup);
      key=Get_Float_M("upper limit: ",";",dmenu);
      if (key->status==TOK_OK) {
	r.upper=key->f;
	defup=key->f;
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad floating point value: %s",key->text);
      }else {
	key->status=TOK_ABORT;
      }
    }
    if (key->status==TOK_OK) {
      add_restraint(restr,&r);
    }
    send_bheader(graphix,"ARMK",0);
  } while (key->status!=TOK_ABORT);
  
  if (mvld(restr)) {
    prtGraphics("%d restraints\n",n_entries(restr));
  }

  pick_root(proot);
  free_Token(key);
}

static void add_restraint(ListEntry *restr,DRdata *r)
{
  ListEntry *new;

  if (mvld(restr)) {
    new=make_list_entry(sizeof(DRdata),init_dr_data);
    if (mvld(new)) {
      *(DRdata *)new->data= *r;
      append_entry_to_list(new,restr);
    }
  }
}

/*-------------------- cmd_clear_restraints --------------------*/

static void cmd_clear_restraints(ListEntry *restr)
{
  clear_restraints(restr);
  prtGraphics("all restraints removed.\n");
}

static void clear_restraints(ListEntry *restr)
{
  delete_list_entries(restr,del_dr_data);
}
/*-------------------- cmd_write_restraints --------------------*/

static void cmd_write_restraints(Node *root,ListEntry *restr)
{
  Token *key;
  char *filename;
  FILE *out;
  int n;

  key=(Token *)NULL;
  filename=(char *)NULL;

  key=Get_String("filename: ",";");
  if (key->status==TOK_OK) {
    filename=key->text; key->text=(char *)NULL;
    out=fopen(filename,"w");
    if (out) {
      n=write_restraints(out,root,restr);
      fclose(out);
      prtGraphics("%d restraints written to %s.\n",n,filename);
    }else{
      err_set(1,"sorry, could not open file %s for write.",filename);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad filename token: %s",key->text);
  }

  if (mvld(filename)) free(filename);
  free_Token(key);
}
/*-------------------- write_restraints --------------------*/

static int write_restraints(FILE *out,Node *root,ListEntry *h)
{
  int r,mf,mt;
  ListEntry *e;
  DRdata *d;
  Node *from,*to;
  Node *nf,*nt;
  
  r=0;
  fprintf(out,"!BIOSYM restraint 1\n");
  fprintf(out,"\n#distance\n");
  if (mvld(h)) {
    for(e=h->nxt;mvld(e);e=e->nxt) {
      d=(DRdata *)e->data;
      if (mvld(d)) {
	from=find_node(root,(void *)d->from,OBJ_ATOM);
	to=find_node(root,(void *)d->to,OBJ_ATOM);
	if (mvld(from) && mvld(to)) {
	  nf=(from->up->up ? from->up->up : from->up );
	  nt=(to->up->up ? to->up->up : to->up );
	  mf=MolNr(nf,root);
	  mt=MolNr(nt,root);
	  fprintf(out,"%d:%s:%s\t %d:%s:%s\t %9.3f %9.3f %6.2f %6.2f %9.2f\n",
		  mf,from->up->name,from->name,
		  mt,to->up->name,to->name,
		  (d->lower >=0 ? d->lower : 0.0),
		  (d->upper >= 0 && d->upper >=d->lower ? d->upper : 0.0),
		  (d->lower>0 ? d->klower : 0.0),
		  (d->upper>0 && d->upper >= d->lower ? d->kupper : 0.0),
		  d->maxforce);
	  r++;
	}
      }
    }
  }
  return(r);
}
/*-------------------- discover_end --------------------*/

static int discover_end(void)
{
  return(FALSE);
}
/*-------------------- init_dr_data --------------------*/

static void init_dr_data(void *data)
{
  *(DRdata *)data=default_dr_data;
}
static void del_dr_data(void *data)
{
  if (data) free(data);
}
