#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define key_is(a) (streq(a,key->key))

#define TRUE 1
#define FALSE 0

PrjTS DefPrj;

/*========================================================================*/

void cmd_display(void)
{
  int i,op;
  float dx[3],axis[3];
  char *tr_prmpt[3] = {"dx: ","dy: ","dz: "};
  char *angle_menu = { "180 90 60 45 40 30 20 15 10 5 4 3 2 1 ' ' "
		       "-1 -2 -3 -4 -5 -10 -15 -20 -30 -40 -45 -60 -90"};

  char *mainkeys = "on only off ' ' global molecule ' ' save"
    " ' ' find rotate scale translate ' ' reset set";
  Atom *a;
  Dspobj *obj;
  Nodechain *c,modify;
  Token *key;
  
  key=Get_Kwd("status: ",";",mainkeys);
  if (key->status == TOK_OK) {
    if (key_is("on")||key_is("off")||key_is("only")) {
      op=key->n;
      free_Token(key);
      key=Get_Atoms_D("Atoms: ",";");
      if (key->status == TOK_OK) {
	c=(Nodechain *)key->ptr;
	switch (op) {
	case ATM_DSP_ON:
	case ATM_DSP_OFF:
	  display_atom_chain(c,op);
	  break;
	case ATM_DSP_ONLY:
	  modify.s=(Node *)NULL; modify.e=modify.s;
	  modify.p=modify.s; modify.n=0;
	  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg|=(1<<1);
	  ForAtoms(a,DefaultNode) {
	    if (!(a->flg&(1<<1))) {
	      if (a->dsp && cPtr(a->cross)) add_to_node_chain((Node *)a,&modify);
	      a->dsp=0;
	    }
	  }
	  if (modify.n>0) {
	    send_bheader(graphix,"ATDS",modify.n*cPtrSize()+sizeof(int));
	    send_bint(graphix,0);
	    for (a=(Atom *)modify.s;mvld(a);a=(Atom *)a->chain) {
	      send_bcptr(graphix,a->cross);
	    }
	    clear_node_chain(&modify);
	  }
	  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg&=~(1<<1);
	  break;
	default:
	  break;
	}
	clear_node_chain(key->ptr);
	free_node_chain(key->ptr);
      }
    }else if (key_is("global")||key_is("molecule")) {
      cmd_display_obj(key->key);
    }else if (key_is("save")) {
      key=free_Token(key);
      key=Get_String("name of screen dump file: ",";");
      if (key->status==TOK_OK) {
	send_message(graphix,"SCDP",strlen(key->text),key->text);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad filename token: %s",key->text);
      }
    }else if (key_is("find")) {
      key=free_Token(key);
      key=Get_DspGlbl_D("object: ",";","/");
      if (key->status==TOK_OK) {
	obj=(Dspobj *)(key->ptr);
	key=free_Token(key);
	key=Get_Atoms_D("which atoms? ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  displayFind(obj,c);
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
      }else err_set(1,"bad object selection: %s",key->text);
    }else if (key_is("rotate")) {
      key=free_Token(key);
      key=Get_DspGlbl_D("object: ",";","/");
      if (key->status==TOK_OK) {
	obj=(Dspobj *)(key->ptr);
	key=free_Token(key);
	key=Get_Kwd("axis of rotation: ",";","X Y Z");
	if (key->status==TOK_OK) {
	  for (i=0;i<3;i++) axis[i]=0;
	  axis[key->n]=1;
	  key=free_Token(key);
	  key=Get_Float_M("angle: ",";",angle_menu);
	  if (key->status==TOK_OK) {
	    rotate_object(obj,axis,key->f);
	  }else if (key->status==TOK_NOK) err_set(1,"bad rotation angle: %s",key->text);
	}else if (key->status==TOK_NOK) err_set(1,"bad rotation axis: %s",key->text);
      }else err_set(1,"bad object selection: %s",key->text);
    }else if (key_is("scale")) {
      key=free_Token(key);
      key=Get_Float("scale factor: ",";");
      if (key->status==TOK_OK) {
	key->f=(key->f>0 ? 1.0/key->f : 0);
	send_bheader(graphix,"SCDR",sizeof(float));
	send_bfloat(graphix,key->f);
      }else if (key->status==TOK_NOK) err_set(1,"bad scale factor: %s\n",key->text);
    }else if (key_is("translate")) {
      key=free_Token(key);
      key=Get_DspGlbl_D("object: ",";","/");
      if (key->status==TOK_OK) {
	obj=(Dspobj *)(key->ptr);
	for (i=0;i<3&&key->status==TOK_OK;i++) {
	  key=free_Token(key);
	  key=Get_Float(tr_prmpt[i],";");
	  if (key->status==TOK_OK) dx[i]=key->f;
	  else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
	}
	if (i==3) translate_object(obj,dx);
      }else if (key->status==TOK_NOK) err_set(1,"bad object selection: %s",key->text);
    }else if (key_is("reset")) {
      cmd_display_reset();
    }else if (key_is("set")) {
      cmd_display_set();
    }
  }else if (key->status == TOK_NOK) {
    err_set(1,"unknown status: %s",key->text);
  }
  free_Token(key);
  send_header(graphix,"ARMK",0);
}
/*-------------------- cmd_display_obj --------------------*/

void cmd_display_obj(char *type)
{
  Token *key;
  int status=0,n,wait_flg,loop;
  unsigned long typ=0;
  Nodechain *c;
  Node *o;

  key=(Token *)NULL;
  key=Get_Kwd("status: ",";","on off toggle");
  if (key->status==TOK_OK) {
    status = 1-key->n;
    if (streq(type,"molecule")) typ=DSP_MOL;
    else if (streq(type,"global")) typ=DSP_GLBL;
    else typ=0;
  }else if (key->status==TOK_NOK) err_set(1,"bad display status: %s",key->text);
  loop=key->status==TOK_OK;
  wait_flg=TRUE;
  while(loop) {
    err_dump();
    key=free_Token(key);
    key=Get_Objects_MW(typ,"objects: ",";","on off toggle ' '",wait_flg);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      for (n=0,o=c->s;mvld(o);o=o->chain) {
	if (is_dsp(o)) {
	  if (cPtr(((Dspobj *)o)->cross)) n++;
	  switch (status) {
	  case 1:    /* on */
	    ((Dspobj *)o)->status|=OBJ_FLG_DSP;
	    break;
	  case 0:    /* off */
	    ((Dspobj *)o)->status&=~OBJ_FLG_DSP;
	    break;
	  case -1:   /* toggle */
	    if (((Dspobj *)o)->status&OBJ_FLG_DSP) ((Dspobj *)o)->status&=~OBJ_FLG_DSP;
	    else ((Dspobj *)o)->status|=OBJ_FLG_DSP;
	    break;
	  }
	}
      }
      if (n>0) {
	send_bheader(graphix,"DSPO",sizeof(int)+n*sizeof(o));
	send_bint(graphix,status);
	for (n=0,o=c->s;mvld(o);o=o->chain) {
	  if (is_dsp(o)) {
	    if (cPtr(((Dspobj *)o)->cross)) send_bptr(graphix,o);
	  }
	}
      }
      clear_node_chain(c);
      free_node_chain(c);
      wait_flg=FALSE;
    }else if (key->status==TOK_NOK) {
      key=Prs_Kwd(key,"on off toggle");
      if (key->status==TOK_OK) {
	status = 1-key->n;
	wait_flg=TRUE;
      }else {
	err_set(1,"bad object selection: %s",key->text);
	loop=FALSE;
      }
    }else loop=FALSE;
  }  
  key=free_Token(key);
}
/*-------------------- initThreeDPrj --------------------*/

void initThreeDPrj(void)
{
  DefPrj=getThreeDPrj();
  setThreeDPrj(&DefPrj);
}

/*-------------------- cmd_display_reset --------------------*/

void cmd_display_reset(void)
{
  PrjTS prj;
  Token *key=(Token *)NULL;

  key=Get_DspGlbl_D("reset what: ",";","projection ' ' perspective scale slabposition "
	      "slabwidth stereoshift ' ' /");
  if (key->status==TOK_OK) {
    reset_object((Dspobj *)key->ptr);
  }else if (key->status==TOK_NOK) {
    key=Prs_Kwd(key,"projection perspective scale slabposition "
		"slabwidth stereoshift");
    if (key->status==TOK_OK) {
      prj=getThreeDPrj();
      if (key_is("perspective")) prj.viewdist=DefPrj.viewdist;
      else if (key_is("scale")) prj.ysize=DefPrj.ysize;
      else if (key_is("slabposition")) prj.znear=DefPrj.znear;
      else if (key_is("slabwidth")) prj.slabwidth=DefPrj.slabwidth;
      else if (key_is("stereoshift")) prj.sep=DefPrj.sep;
      else if (key_is("projection")) {
	prj.viewdist=DefPrj.viewdist;
	prj.ysize=DefPrj.ysize;
	prj.znear=DefPrj.znear;
	prj.slabwidth=DefPrj.slabwidth;
	prj.sep=DefPrj.sep;
      }
      setThreeDPrj(&prj);
    }
  }
  if (key->status==TOK_NOK) err_set(1,"not a display parameter: %s\n",key->text);
  free_Token(key);
}
/*-------------------- cmd_display_set --------------------*/

void cmd_display_set(void)
{
  PrjTS prj;
  Token *key=(Token *)NULL;
  Token *val=(Token *)NULL;

  key=Get_Kwd("set what: ",";","perspective scale slabposition "
	      "slabwidth stereoshift");
  if (key->status==TOK_OK) {
    prj=getThreeDPrj();
    if (key_is("perspective")) {
      default_token("%0.1f",prj.viewdist);
      val=Get_Float("viewing distance:",";");
      if (val->status==TOK_OK) prj.viewdist=(val->f>0.1 ? val->f : 0.1);
    }else if (key_is("scale")) {
      default_token("%0.2f",prj.ysize);
      val=Get_Float("screensize (height, in A):",";");
      if (val->status==TOK_OK) prj.ysize=(val->f>0.01 ? val->f : 0.01);
    }else if (key_is("slabposition")) {
      default_token("%0.2f",prj.znear);
      val=Get_Float("slab position (hither plane):",";");
      if (val->status==TOK_OK) {
	if (prj.viewmode==PERSPECTIVE) 
	  prj.ysize*=1.0-(val->f-prj.znear)/(prj.viewdist*prj.scrysize);
	prj.znear=val->f;
      }
    }else if (key_is("slabwidth")) {
      default_token("%0.3f",prj.slabwidth);
      val=Get_Float("slab width (distance between hither and yon planes):",";");
      if (val->status==TOK_OK) prj.slabwidth=(val->f>0.001 ? val->f : 0.001);
    }else if (key_is("stereoshift")) {
      default_token("%0.3f",prj.sep);
      val=Get_Float("stereo separation:",";");
      if (val->status==TOK_OK) prj.sep=val->f;
    }
    if (val->status==TOK_OK) setThreeDPrj(&prj);
    else if (val->status==TOK_NOK) err_set(1,"bad floating point value: %s",val->text);
  }else if (key->status==TOK_NOK) err_set(1,"not a display parameter: %s\n",key->text);
  free_Token(key);
  free_Token(val);
}
/*-------------------- display_atom_chain --------------------*/

int display_atom_chain(Nodechain *c,int op)
{
  int nchgd=0,display,size;
  Atom *a;

  switch (op) {
  case ATM_DSP_ON:
  case ATM_DSP_OFF:
    display=(op==ATM_DSP_ON);
    for (nchgd=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (display!=a->dsp && cPtr(a->cross)) nchgd++;
    }
    if (nchgd>0) {
      size=sizeof(int)+nchgd*cPtrSize();
      send_bheader(graphix,"ATDS",size);
      send_bint(graphix,display);
    }
    for (nchgd=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (a->dsp != display) {
	if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
	a->dsp=display;
	nchgd++;
      }
    }
    break;
  default:
    break;
  }
  return(nchgd);
}

void displayFind(Dspobj *obj,Nodechain *c)
{
  Atom *a;
  Dspobj *glbl;
  PrjTS prj;
  mat_4 ma;
  int i,n;
  float dx[3];
  vec_3 pv;

  if (c->n>0) {
    prj=getThreeDPrj();
    get_MVT(DspRoot);
    for (i=0;i<3;i++) pv.c[i]=0.0;
    for (n=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain,n++) {
      glbl=glblobj(assocobj((Node *)a));
      ma=(mvld(glbl)?glbl->mtrx:ident_4());
      if (a->nxyz>0) pv=v3pv3(pv,m4xv3(ma,vec3(a->x)));
      else n--;
    }
    dx[0]=-pv.c[0]/n; dx[1]=-pv.c[1]/n; dx[2]=prj.znear-0.5*prj.slabwidth-pv.c[2]/n;
    translate_object(obj,dx);
    ma=mmxinv(dsp_transform(obj));  /* now set object pivot (optional ?) */
    pv.c[0]=0; pv.c[1]=0; pv.c[2]=prj.znear-0.5*prj.slabwidth;
    v32arr(dx,m4xv3(ma,pv));
    set_pivot(obj,dx);
  }
}
