#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "bhtree.h"
#ifdef mingw
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#ifndef M_PI
#define M_PI  3.14159265358979323846
#endif

extern int graphix;
extern Node *DefaultNode,*MolRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))

#define DOTS_DSP_OFF 0
#define DOTS_DSP_COND_ON 1
#define DOTS_DSP_ON  2
#define DOTS_CLIP_NO 0
#define DOTS_CLIP_BONDED 1

#define COL_BY_ATOM -1

#define TRUE 1
#define FALSE 0

#define MAX_TEMPLATES 250
#define MAX_T  50
#define MXTREE 2048
#define MXLIST 1024

/* main-level keyword: */

static char *dotkwd="on off clip color compute delete density map probe style";
static char *sur_type_kwd = "connolly buried contact Lee-Richards LR-buried LR-Contact vdW";
static char *density_menu = "0.10 0.15 0.20 0.25 0.30 0.35 0.40 0.45 0.50 "
"0.60 0.70 0.80 0.90 1.00";

static char *dist_men = "1.0 2.0 3.0 4.0 5.0 6.0 7.0 8.0 9.0 10.0";

static float dot_density=0.25;
static float probe_radius=1.5;
static float dot_pmin=-10.0,dot_pmax=+10.0;
static int dot_clip_mode=DOTS_CLIP_BONDED;

#if defined(mingw)
static int mkstemp(char *template)
{
  mktemp(template);
  return(open(template,O_RDWR));
}
#endif

/*-------------------- dots_get_atoms --------------------*/

static Nodechain *dots_get_atoms(char *prmt)
{
  Token *key;
  Nodechain *c;
  
  key=(Token *)NULL;
  key=Get_Atoms_D(prmt,";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    key=free_Token(key);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad atom selection: %s",key->text);
    c=(Nodechain *)NULL;
  }else if (key->status==TOK_NULL) {
    c=make_node_chain();               /* yes, but empty... */
  }else{
    c=(Nodechain *)NULL;
  }
  free_Token(key);
  return(c);
}
/*======================================================================*/

void cmd_dots(void)
{
  Token *key;
  Nodechain *c;
  int dsp,color;
  
  key=Get_Kwd("option: ",";",dotkwd);
  if (key->status == TOK_OK) {                  

    if (key_is("on")||key_is("off")) {         
      if (key_is("on")) dsp=DOTS_DSP_ON;         /* modify display status */
      else if (key_is("off")) dsp=DOTS_DSP_OFF;
      else dsp=DOTS_DSP_COND_ON;
      c=dots_get_atoms("Atoms: ");
      set_dots_dsp(c,dsp);
      clear_node_chain(c); free_node_chain(c);

    }else if (key_is("probe")) {
      default_token("%2.2f",probe_radius);
      key=free_Token(key);
      key=Get_Float_M("probe radius: ",";","1.0 1.5 2.0 2.5");
      if (key->status==TOK_OK) {
	if (key->f>=0.0) probe_radius=key->f;
	else err_set(1,"illegal probe radius: %s",key->text);
      }else if (key->status==TOK_NOK) {
	err_set(1,"illegal probe radius: %s",key->text);
      }
      
    }else if (key_is("delete")) {
      c=dots_get_atoms("Atoms: ");
      delete_dots(c);
      clear_node_chain(c); free_node_chain(c);

    }else if (key_is("clip")) {
      key=free_Token(key);
      key=Get_Kwd("clip mode for dots: ",";","on off");
      if (key->status==TOK_OK) {
	if (key_is("on"))       dot_clip_mode=DOTS_CLIP_BONDED;
	else if (key_is("off")) dot_clip_mode=DOTS_CLIP_NO;
      }else if (key->status==TOK_NOK) {
	err_set(1,"dots: bad clip mode (%s)",key->text);
      }

    }else if (key_is("map")) {
      key=free_Token(key);
      default_token("%0.4f",dot_pmin);
      key=Get_Float("minimum = ",";");
      if (key->status==TOK_OK) {
	dot_pmin=key->f;
	key=free_Token(key);
	default_token("%0.4f",dot_pmax);
	key=Get_Float("maximum = ",";");
	if (key->status==TOK_OK) {
	  dot_pmax=key->f;
	  setPropertyMap(dot_pmin,dot_pmax);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad float: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad float: %s",key->text);
      }

    }else if (key_is("color")) {
      key=free_Token(key);
      key=Get_Color_M("color: ",";","atomcolor property ' '");
      if (key->status == TOK_OK) {
	color=key->n;
	c=dots_get_atoms("Atoms: ");
	set_dots_color(c,color);
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown color: %s",key->text);
      }
     
    }else if (key_is("compute")) {
      cmd_compute_dots();

    }else if (key_is("style")) {
      key=free_Token(key);
      key=Get_Kwd("dot style: ",";","small medium large xtralarge");
      if (key->status==TOK_OK) {
	c=dots_get_atoms("Atoms: ");
	set_dots_style(c,key->n);
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown dot style: %s",key->text);
      }

    }else if (key_is("density")) {
      default_token("%5.2f",dot_density);
      key=free_Token(key);
      key=Get_Float_M("dot density (A): ",";",density_menu);
      if (key->status==TOK_OK) {
	if (key->f>0.001 && key->f <100.0) dot_density=key->f;
	else key->status=TOK_NOK;
      }
      if (key->status==TOK_NOK) {
	err_set(1,"bad dot density: %s",key->text);
      }
    }

  }else if (key->status==TOK_NOK) {
    err_set(1,"dots: bad keyword: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- set_dots_dsp --------------------*/

void set_dots_dsp(Nodechain *c,int dsp)
{
  int n,size;
  Atom *a;

  if (!c) return;

  /* count atoms... */

  for (n=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    if (a->type==OBJ_ATOM) {
      if (cPtr(a->cross)) n++;
    }
  }

  if (n>0) {
    size=n*cPtrSize()+sizeof(int);
    send_bheader(graphix,"DOTD",size);
    send_bint(graphix,dsp);
    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (a->type==OBJ_ATOM) {
	if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
      }
    }                           /* flush ? */
  }
}
/*-------------------- set_dots_color --------------------*/

void set_dots_color(Nodechain *c,int col)
{
  int n,size;
  Atom *a;

  if (!c) return;

  /* count atoms... */

  for (n=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    if (a->type==OBJ_ATOM) {
      if (cPtr(a->cross)) n++;
    }
  }

  if (n>0) {
    size=n*cPtrSize()+sizeof(int);
    send_bheader(graphix,"DOTC",size);
    send_bint(graphix,col);
    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (a->type==OBJ_ATOM) {
	if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
      }
    }                           /* flush ? */
  }
}
/*-------------------- set_dots_style --------------------*/

void set_dots_style(Nodechain *c,int style)
{
  int n,size;
  Atom *a;

  if (!c) return;

  /* count atoms... */

  for (n=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    if (a->type==OBJ_ATOM) {
      if (cPtr(a->cross)) n++;
    }
  }

  if (n>0) {
    size=n*cPtrSize()+sizeof(int);
    send_bheader(graphix,"DOTS",size);
    send_bint(graphix,style);
    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (a->type==OBJ_ATOM) {
	if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
      }
    }                           /* flush ? */
  }
}
/*-------------------- delete_dots --------------------*/

void delete_dots(Nodechain *c)
{
  int n,size;
  Atom *a;

  if (!c) return;

  /* count atoms... */

  for (n=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    if (a->type==OBJ_ATOM) {
      if (cPtr(a->cross)) n++;
    }
  }

  if (n>0) {
    size=n*cPtrSize();;
    send_bheader(graphix,"DOTR",size);
    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (a->type==OBJ_ATOM) {
	if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
      }
    }                           /* flush ? */
  }
}

/*-------------------- cmd_compute_dots --------------------*/

#define MXDATM 65536

void cmd_compute_dots(void)
{
  /* uses atom flags 1 and 2 */

  Token *key;
  Nodechain *c=(Nodechain *)NULL;
  Atom **atoms,*a,*candidate[MXLIST];
  float dist[MXLIST],cutoff,xyz[MXDATM][4];
  int m,n,i,grain,foreground,tmpfd,LR,BU,CO,VW,status=0,nd,nr,now,done;
  char *filename;
  char buff[1024],*cmd,command[1024],fld[20];
  BHtree *tree;
  FILE *dots=(FILE *)NULL;
  int havetmp=FALSE;
  dot_template *D;

  atoms=(Atom **)NULL; n=0;
  filename=(char *)NULL;
  key=Get_Kwd("surface type: ",";",sur_type_kwd);

  foreground=FALSE;
  filename=NULL;
  done=FALSE;

  if (key->status==TOK_OK) {
    LR = key_is("Lee-Richards")||key_is("LR-Contact")||key_is("LR-buried");
    BU = key_is("buried")||key_is("LR-buried");
    CO = key_is("contact")||key_is("LR-Contact");
    VW = key_is("vdW");

    if (VW) {
      c=dots_get_atoms("Atoms: ");
      compute_dots_vdw(c);
      clear_node_chain(c); free_node_chain(c);
      done=TRUE;

    }else if (key_is("connolly")||key_is("Lee-Richards")||BU) {
      c=dots_get_atoms("Atoms: ");
      if (mvld(c)) {
	n=c->n;
	if (n>0) {
	  atoms=(Atom **)chain_to_ptr(c,TRUE);
	  if (atoms) {
	    for (i=0;i<n;i++) atoms[i]->flg|=(1<<1);
	    c=dots_get_atoms("context: ");
	    if (mvld(c)) {
	      for (i=0;i<n;i++) add_to_node_chain((Node *)atoms[i],c);
	      walk_order_chain(MolRoot,c,1<<2); n=c->n;
	      free(atoms); atoms=(Atom **)chain_to_ptr(c,TRUE);
	    }else{
	      free(atoms); atoms=NULL;
	      done=TRUE;
	    }
	  }
	}else{
	  clear_node_chain(c); free_node_chain(c);
	}
      }else{
	done=TRUE;
      }

    }else if (CO) {
      c=dots_get_atoms("Atoms: ");
      if (mvld(c)) {
	n=c->n;
	if (n>0) {
	  grain = ( c->n < MXTREE ? 1 : c->n / MXTREE );
	  tree=generateBHtree(c,grain);
	  clear_node_chain(c); free_node_chain(c); c=(Nodechain *)NULL;
	  if (mvld(tree)) {
	    c=dots_get_atoms("center atoms: ");
	    if (mvld(c)) {
	      key=free_Token(key);
	      key=Get_Float_M("max. distance from center = ",";",dist_men);
	      if (key->status==TOK_OK) {
		cutoff=key->f;
		for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
		  if (a->nxyz>0 && mvld(a->x)) {
		    m=findBHcloseAtoms(tree,a->x,cutoff,candidate,dist,MXLIST);
		    m=(m>MXLIST ? MXLIST : m);
		    for (i=0;i<m;i++) candidate[i]->flg|=(1<<1);
		  }
		}
		atoms=tree->atom;
		tree->atom=(Atom **)NULL;
		walk_order(MolRoot,n,atoms,1<<2);
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad float: %s.",key->text);
		done=TRUE;
	      }else done=TRUE;
	    }else done=TRUE;
	    freeBHtree(tree);
	  }
	}else{
	  done=TRUE;
	}
	clear_node_chain(c); free_node_chain(c);
      }else{
	done=TRUE;
      }
    }else{
      err_set(1,"unknown dotted surface type: %s",key->text);
      done=TRUE;
    }

    if (!atoms && !done) {
      err_set(1,"failed to allocate scratch array for atom pointers");
      done=TRUE;
    }

    if (!done) {
      key=free_Token(key);
      default_token("dots.dot");
      key=Get_String_M("filename: ",";","-foreground");
      if (key->status==TOK_OK) {
	foreground=(kwd_match(key->text,"-foreground",(char *)NULL)>=0);
	if (!foreground) {
	  filename=key->text;
	  key->text=(char *)NULL;
	}
      }else done=TRUE;
    }

    if (!done && LR && foreground) {
      for (i=0;i<n;i++) {  /* swap flag: in LR context atoms ar flagged, normal atomes are not */
	if (atoms[i]->flg&(1<<1)) atoms[i]->flg&=~(1<<1);
	else atoms[i]->flg|=(1<<1);
      }
      if (!LRinit(atoms,n,probe_radius,dot_density,BU?1<<1:0)) {
	for (i=0;i<n;i++) {
	  if (!atoms[i]->flg&&(1<<1)) {
	    if ((D=LRdots(atoms[i]))) {
	      send_dots(atoms[i],D->n,D->x);
	      free_dot_template(D);
	    }
	  }
	}
	LRclose();
      }
      for (i=0;i<n;i++) atoms[i]->flg&=~(1<<1);
      if (atoms) {free(atoms);} atoms=NULL;
      done=TRUE;
    }

    if (!done) {
      c=ptr_to_chain((Node **)atoms,n);
      if (foreground) {
	if (filename) free(filename);
	filename=Strdup("connolly.tmpXXXXXX");
      }
      tmpfd=mkstemp(filename);
      if (tmpfd!=-1) {
	close(tmpfd);
	havetmp=TRUE;
      }
      if (compute_dots_connolly(c,filename,LR,BU)) {
	cmd=getenv("CONNOLLYserver");
	if (cmd) {
	  if (foreground) {
	    sprintf(command,"%s %s -",cmd,filename);
	    prtGraphics("running %s in foreground...\n",cmd);
	    /* obsolete: status=startCmd(command,"","",TRUE,"belownormal"); */
	    dots=PipeOpen(command,"r");
	    if (!dots) status=1;
	  }else{
	    sprintf(command,"%s %s",cmd,filename);
	    status=startCmd(command,"-null",extfil(buff,filename,".log",TRUE),
			    FALSE,"belownormal");
	    prtGraphics("running %s in background...\n",cmd);
	  }
	  if (!status) {
	    if (foreground) {
	      for (nd=0,now=-1;;) {
		if (!fgets(buff,sizeof(buff),dots)) break;
		StrCpy(fld,buff,6);nr=str2i(fld);
		if (nr>0&&nr<=n) {
		  nr--;
		  if (nr!=now) {
		    if (nd>0) send_dots(atoms[now],nd,(float *)xyz);
		    nd=0;
		    now=nr;
		  }
		  if (nd<MXDATM) {
		    StrCpy(fld,buff+17,10);xyz[nd][0]=str2f(fld);
		    StrCpy(fld,buff+26,10);xyz[nd][1]=str2f(fld);
		    StrCpy(fld,buff+35,10);xyz[nd][2]=str2f(fld);
		    xyz[nd][3]=0;
		    nd++;
		  }
		}
	      }
	      if (nd>0) send_dots(atoms[now],nd,(float *)xyz);
	      status=PipeClose(dots);
	    }
	  }
	  if (status) {
	    err_set(1,"system: %s\n",command);
	    err_set(1,"failed: error status (%d,%d)\n",status>>8,status&0xff);
	    err_dump();
	    switch (status>>8) {
	    case 3:
	      prtGraphics("\n");
	      prtGraphics("*** Could not find program to compute Connolly dots.\n");
	      prtGraphics("*** The QCPE program ms.exe (#429) or an equivalent\n");
	      prtGraphics("*** program is needed to compute Connolly dots.\n");
	      prtGraphics("\n");
	      break;
	    case 101:
	      prtGraphics("*** Failed to create working directory for ms.exe (OCPE 429)\n");
	      break;
	    case 102:
	      prtGraphics("*** Can't find dot files generated by ms.exe (OCPE 429)\n");
	      break;
	    }
	  }
	}
      }
      if (filename) {
	if (havetmp) unlink(filename);
	free(filename); filename=(char *)NULL;
      }
      clear_node_chain(c); free_node_chain(c);
      for (i=0;i<n;i++) atoms[i]->flg&=~(1<<1);
      if (atoms) {free(atoms);} atoms=NULL;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown surface type: %s",key->text);
  }
  if (atoms) free(atoms);
  if (filename) free(filename);
  free_Token(key);
}

/*-------------------- compute_dots_connolly --------------------*/

int compute_dots_connolly(Nodechain *c,char *filename,int LR,int BU)
{
  /* This is not actually computing the Connolly surface, but is
     just an interface to the MS program. If LR then compute Lee-Ricards
     type surface  */

  FILE *out;
  Atom *a;
  int ns,nc,i,k,nt=0,f,m;
  float r,R[MAX_T],lr;
  char full_name[1024];
  int *T,status;

  status=0;

  /*  generate atomtype table for Connolly program */

  lr=(LR ? probe_radius : 0);

  T=(int *)malloc(c->n*sizeof(int));
  if (!mvld(T)) {
    err_set(1,"memory allocation failure in compute_dots_connolly.");
  }else{
    for (a=(Atom *)c->s,i=0,nt=0;a;i++,a=(Atom *)a->chain) {
      r=(mvld(a->atp) ? a->atp->rvdw : 0.0) + lr;
      for (k=0;k<nt;k++) if (r==R[k]) break;
      if (k<nt) {
	T[i]=k+1;
      }else{
	if (nt<MAX_T) {
	  R[nt++]=r;
	  T[i]=nt;
	}else{
	  err_set(1,"too bad: more than %d atomtypes.",MAX_T);
	  free(T); T=(int *)NULL;
	}
      }
      if (!mvld(T)) break;
    }
  }

  if (mvld(T)) {
    
    /* write parameter file */
    
    sprintf(full_name,"%s_connolly_param",filename);
    out=fopen(full_name,"w");
    if (out) {
      fprintf(out,"%10.3f%10.3f%5d%5d\n",
	      1.0/(dot_density*dot_density),(LR ? 0.0 : probe_radius),(BU ? 1 : 0),0);
      fclose(out);
    }else{
      err_set(1,"could not open %s for write.",full_name);
      free(T);
      return(0);
    }
    
    /* write atomtypes f */
       
    sprintf(full_name,"%s_connolly_types",filename);
    out=fopen(full_name,"w");
    if (out) {
      for (k=0;k<nt;k++) {
	fprintf(out,"%5d%10.5f\n",k+1,R[k]);
      }
      fclose(out);
    }else{
      err_set(1,"compute_dots_connolly: could not open file %s\n",full_name);
      free(T);
      return(0);
    }
       
    /* write coordinates file */
    
    sprintf(full_name,"%s_connolly_atoms",filename);
    out=fopen(full_name,"w");
    if (out) {
      for (ns=0,nc=0,i=0,a=(Atom *)c->s;a;a=(Atom *)a->chain,i++) {
	r=(mvld(a->atp) ? a->atp->rvdw : 0.0);
	f=(r<0.001 ? 0 : ( a->flg&(1<<1) ? 2 : 1)); 
        m = ( BU ? ( a->flg&(1<<1) ? 2 : 1 ) : 1);
	fprintf(out,"%10.3f%10.3f%10.3f%5d%5d%5d\n",
		a->x[0],a->x[1],a->x[2],T[i],f,m);
	nc+=(f==1);
	ns+=(f==2);
      }
      fclose(out);
      prtGraphics("%d surface generating atoms, %d context atoms.\n",
		  ns,nc);
      status=1;
    }else{
      err_set(1,"compute_dots_connolly: could not open file %s\n",full_name);
      free(T);
      return(0);
    }
    free(T);
  }
  return(status);
}

/*-------------------- compute_dots_vdw --------------------*/

void compute_dots_vdw(Nodechain *c)
{
  
  Atom *a;
  int i,n,n_templates,size;
  int nx=0;
  float R,*px,*tx,*x=(float *)NULL;
  dot_template *T[MAX_TEMPLATES],*t;

  if (graphix<1) {
    err_set(1,"dots: graphics not running...");
    /*    return; */
  }

  if (!c) return;

  T[0]=make_dot_template(1.7,dot_density);
  n_templates=(mvld(T[0]) ? 1 : 0);

  for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    if (!cPtr(a->cross)||!mvld(a->x)||a->nxyz<1) continue;
    R=(mvld(a->atp) ? a->atp->rvdw : 0.0);
    if (R<0.0001) continue;
    t=find_best_template(R,n_templates,T);
    if (t) {
      if (fabs(t->R - R)>0.025 && n_templates<MAX_TEMPLATES) {
	T[n_templates++]=t=make_dot_template(R,dot_density);
	if (!t)  n_templates--;
      }
    }
    if (t) {
      if (t->n > 0) {
	if (t->n>nx) {
	  if (mvld(x)) free(x);
	  x=(float *)malloc(4*t->n*sizeof(float));
	  if (mvld(x)) nx=t->n;
	  else nx=0;
	}
	if (mvld(x)) {
	  for (i=0,px=x,tx=t->x;i<t->n;i++,tx+=3,px+=4) {
	    px[0]=a->x[0]+tx[0];
	    px[1]=a->x[1]+tx[1];
	    px[2]=a->x[2]+tx[2];
	    px[3]=0.0;
	  }
	  n=clip_dots(t->n,x,a);
	  if (n>0) {
	    size=4*n*sizeof(float)+cPtrSize();
	    send_bheader(graphix,"DOTA",size);
	    send_bcptr(graphix,a->cross);
	    send_bfloata(graphix,4*n,(float *)x);
	  }
	}
      }
    }
  }

  if (mvld(x)) free(x);
  for (i=0;i<n_templates;i++) free_dot_template(T[i]);
}

/*-------------------- make_dot_template --------------------*/

dot_template *make_dot_template(float R,float dens)
{
  float dl,dz,z,*tx;
  int i,N;
  dot_template *t;

  /* determine number of points... */
  /* "golden" spiral method        */

  N=4*M_PI*R*R/(dens*dens);
  t=(dot_template *)malloc(sizeof(dot_template));  /* generate template */
  if (mvld(t)) {
    t->R=R;
    t->n=N;
    t->x=(float *)malloc(3*N*sizeof(float));
    if (mvld(t->x)) { 
      dl=M_PI*(3-sqrt(5));
      dz=2.0*R/N;
      for (i=0,tx=t->x;i<N;i++,tx+=3) {
	z=-R+(i+0.5)*dz;
	tx[0]=sqrt(R*R-z*z)*cos(i*dl);
	tx[1]=sqrt(R*R-z*z)*sin(i*dl);
	tx[2]=z;
      }
    }else{
      t->n=0;
    }
  }
  return(t);
}
/*-------------------- find_best_template --------------------*/

dot_template *find_best_template(float R,int n,dot_template *t[])
{
  dot_template *r;
  int i;
  float d,dbest;

  if (n>0) {
    r=t[0]; dbest=fabs(R-r->R);
    for (i=1;i<n;i++) {
      d=fabs(R-t[i]->R);
      if (d<dbest) {
	dbest=d;
	r=t[i];
      }
    }
  }else{
    r=(dot_template *)NULL;
  }
  return(r);
}
/*-------------------- clip_dots --------------------*/

int clip_dots(int n,float *x,Atom *a)
{
  int i,j,r;
  Atom *b;
  float R2,R2S,d,*pt,*qt;
  int *clipped=(int *)NULL;

  switch (dot_clip_mode) {
  case DOTS_CLIP_NO:
    r=n;
    break;
  case DOTS_CLIP_BONDED:
    if (n>0) {
      clipped=(int *)malloc(n*sizeof(int));
      if (mvld(clipped)) {
	for(i=0;i<n;i++) clipped[i]=FALSE;
	for (j=0;j<a->nbnd;j++) {
	  b=neighbour(j,a);
	  if (b->nxyz>0) {
	    R2=(mvld(b->atp) ? b->atp->rvdw : 0.0);
	    R2S=R2*R2;
	    for (pt=x,i=0;i<n;i++,pt+=4) {
	      if (!clipped[i]) {
		d=(pt[0]-b->x[0])*(pt[0]-b->x[0])+
		  (pt[1]-b->x[1])*(pt[1]-b->x[1])+
		  (pt[2]-b->x[2])*(pt[2]-b->x[2]);
		clipped[i]=d<R2S;
	      }
	    }
	  }
	}
	for (qt=x,pt=x,r=0,i=0;i<n;i++,pt+=4) {
	  if (!clipped[i]) {
	    qt[0]=pt[0]; qt[1]=pt[1]; qt[2]=pt[2];
	    qt+=4;
	    r++;
	  }
	}
	free(clipped);
      }else r=n;
    }else r=0;
    break;
  default:
    r=n;
    break;
  }
  return(r);
}

/*-------------------- free_dot_template --------------------*/

void free_dot_template(dot_template *t)
{
  if (mvld(t)) {
    if (mvld(t->x)) free(t->x);
    free(t);
  }
}
