#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"
#include "bhtree.h"
#include "structal.h"
#include "gseal.h"

extern int graphix;
extern Node *MolRoot;
extern Node *DefaultNode;

#define key_is(x) (streq(key->key,x))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

static void cmd_fit_exact(char *op);
static void cmd_fit_sequence(char *op);
static void cmd_fit_pairs(char *op);
static void cmd_fit_autopairs(char *op);
static void cmd_fit_average(char *op);
static void cmd_fit_shape(char *op);
static void cmd_fit_cotransform(char *mode);

static Token *Get_SourceMols(char *pr);
static Token *Get_SourceMol(char *pr,Node *tmol);
static Token *Get_TemplateMol(void);
static Token *Get_SourceAtoms(int nat,Node *smol);
static Token *Get_AtomPairlist(Node *smol,Atom ***satoms,\
			       Node *tmol,Atom ***tatoms);
static Token *Get_TemplateAtoms(Node *tmol);
static Token *Get_AutoSourceAtoms(Node *tmol);
static void   SetFitCoTransform(mat_4 Tra);
static int node_is_sub(Node *a,Node *b);
static Atom **exact_partners(Node *smol,int nat,Node *tmol,Atom **tatoms);
static void clear_amark(Atom *a);

static float gsealFit(int nats,Node *smol,Atom **satoms,int natt,Atom **tatoms,mat_4 *Tra);

static mat_4 FitCoTransform = { 
  {{{1.0,0.0,0.0,0.0}},
   {{0.0,1.0,0.0,0.0}},
   {{0.0,0.0,1.0,0.0}},
   {{0.0,0.0,0.0,1.0}}
  }
};

static float alpha=1.0;  /* default value for alpha in gseal */

#define MAX(a,b) ((a)>=(b)?(a):(b))

/*========================================================================*/

void cmd_fit(char *op)
{
/* rigid body least-squares fitting... */

  Token *key;
  char *measkeys="autopairs average exact pairs sequence";
  char *fitkeys="autopairs average exact pairs sequence shape ' ' backtransform cotransform";

  key=Get_Kwd("method: ",";",streq(op,"measure")?measkeys:fitkeys);
  if (key->status==TOK_OK) {
    if (key_is("exact"))          cmd_fit_exact(op);
    else if (key_is("average"))   cmd_fit_average(op);
    else if (key_is("pairs"))     cmd_fit_pairs(op);
    else if (key_is("autopairs")) cmd_fit_autopairs(op);
    else if (key_is("sequence"))  cmd_fit_sequence(op);
    else if (key_is("shape"))     cmd_fit_shape(op);
    else if (key_is("backtransform")) cmd_fit_cotransform("back");
    else if (key_is("cotransform")) cmd_fit_cotransform("co");
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token %s",key->text);
  }

  free_Token(key);
}
/*-------------------- cmd_fit_exact --------------------*/

static void cmd_fit_exact(char *op)
{
  Token *key;
  Node *tmol,*smol;
  int nat,nmol,i;
  Atom **tatoms,**satoms;
  Node **smols;
  float rms;
  char nm1[1024],nm2[1024],*prmt;
  mat_4 Tra;

  key=Get_TemplateMol();
  if (key->status==TOK_OK) {
    tmol=(Node *)key->ptr;
    key=free_Token(key);
    key=Get_TemplateAtoms(tmol);
    if (key->status==TOK_OK) {
      nat=key->n;
      prtGraphics("%d atoms in template %s.\n",
		  nat,nodeptr_to_name(nm1,tmol));
      tatoms=(Atom **)key->ptr;
      walk_order(tmol,nat,tatoms,(1<<3));
      key=free_Token(key);
      prmt=(streq(op,"fit") ? "molecules to fit to template: " : 
	    "measure rms-distance to molecules: ");
      key=Get_SourceMols(prmt);      if (key->status==TOK_OK) {
	nmol=key->n;
	smols=chain_to_ptr((Nodechain *)key->ptr,TRUE);
	if (mvld(smols)) {
	  for (i=0;i<nmol;i++) {
	    smol=smols[i];
	    satoms=exact_partners(smol,nat,tmol,tatoms);
	    if (mvld(satoms)) {
	      if (streq(op,"fit")) {
		rms=RmsFit(nat,smol,satoms,tatoms,&Tra);
		if (rms>=0) SetFitCoTransform(Tra);
	      }else rms=rms_dist(nat,satoms,tatoms);        /*measure only */
	      prtGraphics("r.m.s. ( %s , %s ) = %6.3f\n",
			  nodeptr_to_name(nm1,tmol),
			  nodeptr_to_name(nm2,smol),
			  rms);
	      free(satoms); satoms=(Atom **)NULL;
	    }
	    err_dump();
	  }
	  free(smols); smols=(Node **)NULL;
	}
      }
      free(tatoms); tatoms=(Atom **)NULL;
    }
  }
  free_Token(key);
}
/*-------------------- cmd_fit_average --------------------*/

static void cmd_fit_average(char *op)
{
  Token *key;
  int nat,nmol=0,i,j,k;
  Atom ***atoms;
  Node **mols;
  float rms,r,rav,*p,*x;
  char *prmt;

  mols=(Node **)NULL;
  atoms=(Atom ***)NULL;

  prmt=(streq(op,"fit") ? "molecules to fit to average: " :
	"measure rms-distance to average from molecules: ");
  key=Get_SourceMols(prmt);
  if (key->status==TOK_OK) {
    nmol=key->n;
    mols=chain_to_ptr((Nodechain *)key->ptr,TRUE);
    if (mvld(mols)) {
      atoms=(Atom ***)malloc(nmol*sizeof(Atom **));
      if (mvld(atoms)) {
	for (k=0;k<nmol;k++) atoms[k]=(Atom **)NULL;
	key=free_Token(key);
	key=Get_TemplateAtoms(mols[0]);
	if (key->status==TOK_OK) {
	  nat=key->n;
	  prtGraphics("%d atoms in template.\n",key->n);
	  atoms[0]=(Atom **)key->ptr;
	  walk_order(mols[0],nat,atoms[0],(1<<3));
	  for (k=1;k<nmol;k++) {
	    atoms[k]=exact_partners(mols[k],nat,mols[0],atoms[0]);
	    if (!mvld(atoms[k])) break;
	  }
	  if (k==nmol) {
	    if (streq(op,"fit")) {
	      rms=RmsMultiFit(nat,nmol,mols,atoms);
	      err_dump();
	    }else{
	      x=(float *)malloc(3*nat*sizeof(float));
	      if (mvld(x)) {
		for (j=0;j<3*nat;j++) x[j]=0.0;
		for (k=0;k<nmol;k++) {
		  for (p=x,i=0;i<nat;i++) {
		    for (j=0;j<3;j++) *p++ += atoms[k][i]->x[j];
		  }
		}
		if (nmol>0) {
		  for (j=0;j<3*nat;j++) x[j]/=nmol;
		  for (rav=0,rms=0,k=0;k<nmol;k++) {
		    r=rms_dist_x(nat,atoms[k],x); 
		    rms+=r*r;
		    rav+=r;
		  }
		  rav/=nmol;
		  rms=sqrt(rms/nmol);
		}else{
		  rav=0;
		  rms=0;
		}
		prtGraphics("average r.m.s.d. to average structure = %6.3f\n",rav);
		free(x);
	      }else{
		err_set(1,"memory allocation failed.");
		rms=-1;
	      }
	    }
	    if (rms<-0.1) {
	      err_set(1,"rms_mult_fit error.");
	    }else{
	      prtGraphics("r.m.s.d. to average = %6.3f\n",rms);
	    }
	  }
	}
      }
    }
  }

  err_dump();
  free_Token(key);
  if (mvld(mols)) free(mols);
  if (mvld(atoms)) {
    for (k=0;k<nmol;k++) if (mvld(atoms[k])) free(atoms[k]);
    free(atoms);
  }
}
/*-------------------- cmd_fit_sequence --------------------*/

static void cmd_fit_sequence(char *op)
{
  Token *key;
  Node *tmol,*smol;
  int nat;
  Atom **tatoms,**satoms;
  float rms;
  char nm1[1024],nm2[1024],*prmt;
  mat_4 Tra;

  tatoms=(Atom **)NULL;
  satoms=(Atom **)NULL;

  key=Get_TemplateMol();
  if (key->status==TOK_OK) {
    tmol=(Node *)key->ptr;
    key=free_Token(key);
    key=Get_TemplateAtoms(tmol);
    if (key->status==TOK_OK) {
      nat=key->n;
      tatoms=(Atom **)key->ptr;
      walk_order(tmol,nat,tatoms,(1<<3));
      prtGraphics("%d atoms selected in template %s\n",
		  nat,nodeptr_to_name(nm1,tmol));
      key=free_Token(key);
      prmt=(streq(op,"fit") ? "molecules to fit to template: " : 
	    "measure rms-distance to molecules: ");
      key=Get_SourceMol(prmt,tmol);
      if (key->status==TOK_OK) {
	smol=(Node *)key->ptr;
	key=free_Token(key);
	key=Get_SourceAtoms(nat,smol);
	if (key->status==TOK_OK) {
	  satoms=(Atom **)key->ptr;
	  walk_order(smol,nat,satoms,(1<<3));
	  if (streq(op,"fit")) {
	    rms=RmsFit(nat,smol,satoms,tatoms,&Tra);
	    if (rms>=0) SetFitCoTransform(Tra);
	  }else rms=rms_dist(nat,satoms,tatoms);        /*measure only */
	  prtGraphics("r.m.s. ( %s , %s ) = %6.3f\n",
		      nodeptr_to_name(nm1,tmol),
		      nodeptr_to_name(nm2,smol),
		      rms);
	  free(satoms); satoms=(Atom **)NULL;
	}
      }
      free(tatoms); tatoms=(Atom **)NULL;
    }
  }
  free_Token(key);
}
/*-------------------- cmd_fit_autopairs --------------------*/

static void cmd_fit_autopairs(char *op)
{
  Token *key;
  Node *tmol,*smol;
  int natt,nats,npairs,method=0,i,onp,conv,nit;
  Atom **tatoms,**satoms,**sp,**tp;
  float rms,maxd=0,orms;
  char nm1[1024],nm2[1024],*prmt,*maxdM;
  mat_4 Tra,Tot;

  key=(Token *)NULL;
  tatoms=(Atom **)NULL; tp=(Atom **)NULL;
  satoms=(Atom **)NULL; sp=(Atom **)NULL;

  key=Get_TemplateMol();
  if (key->status==TOK_OK) {
    tmol=(Node *)key->ptr;
    key=free_Token(key);
    key=Get_TemplateAtoms(tmol);
    if (key->status==TOK_OK) {
      natt=key->n;
      tatoms=(Atom **)key->ptr;
      prtGraphics("%d atoms selected in template %s\n",
		  natt,nodeptr_to_name(nm1,tmol));
      key=free_Token(key);
      prmt=(streq(op,"fit") ? "molecules to fit to template: " : 
	    "measure rms-distance to molecule: ");
      key=Get_SourceMol(prmt,tmol);
      if (key->status==TOK_OK) {
	smol=(Node *)key->ptr;
	key=free_Token(key);
	key=Get_AutoSourceAtoms(smol);
	if (key->status==TOK_OK) {
	  satoms=(Atom **)key->ptr;
	  nats=key->n;
	  prtGraphics("%d atoms selected in source %s\n",
		  nats,nodeptr_to_name(nm2,smol));
	  key=free_Token(key);
	  key=Get_Kwd("method: ",";","nearest stable structal");
	  if (key->status==TOK_OK) {
	    if (key_is("nearest")) method=AUTO_PAIR_NEAR;
	    else if (key_is("stable")) method=AUTO_PAIR_STABLE;
	    else if (key_is("structal")) method=AUTO_PAIR_STRUCTAL;
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"unknown auto pair method: %s",key->text);
	    key->status=TOK_ABORT;
	  }
	  if (key->status==TOK_OK) {
	    key=free_Token(key);
	    default_token(method==AUTO_PAIR_STRUCTAL?"-3":"3");
	    maxdM=(method==AUTO_PAIR_STRUCTAL?
		   "-10 -9 -8 -7 -6 -5 -4 -3 -2 -1 ' ' 0 ' ' 1 2 3 4 5 6 7 8 9 10" :
		   "0.00 0.05 0.10 0.15 0.20 0.25 0.30 0.35 0.40 0.45 0.50 1.0 2.0 3.0 4.0 5.0 10.0");
	    key=Get_Float_M("max.distance for pairs: ",";",maxdM);
	    if (key->status==TOK_OK) {
	      if (method==AUTO_PAIR_STRUCTAL) maxd=key->f;
	      else maxd=(key->f < 0 ? 0 : key->f);
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad floating point value: %s",key->text);
	      key->status=TOK_ABORT;
	    }
	  }
	  if (key->status==TOK_OK) {
	    if (method==AUTO_PAIR_STRUCTAL) {
	      walk_order(tmol,natt,tatoms,(1<<3));
	      walk_order(smol,nats,satoms,(1<<3));
	    }
	    conv=!streq(op,"fit")||method==AUTO_PAIR_STRUCTAL; 
	    orms=-1.0; onp=0; nit=100;
	    sp=(Atom **)NULL; tp=(Atom **)NULL;
	    if (streq(op,"fit")) lock_evtq();
	    Tot=ident_4();
	    do {
	      if (mvld(sp)) {free(sp);} sp=(Atom **)NULL;
	      if (mvld(tp)) {free(tp);} tp=(Atom **)NULL;
	      npairs=auto_pairs(natt,tatoms,nats,satoms,maxd,method);
	      if (npairs>0) {
		sp=(Atom **)malloc(npairs*sizeof(Atom *));
		tp=(Atom **)malloc(npairs*sizeof(Atom *));
		if (!mvld(sp)||!mvld(tp)) {
		  err_set(1,"memeory allocation failed.");
		  npairs=0;
		  break;
		}
		for (npairs=0,i=0;i<nats;i++) {
		  if (satoms[i]->chain) {
		    sp[npairs]=satoms[i];
		    tp[npairs]=(Atom *)satoms[i]->chain;
		    satoms[i]->chain=(Node *)NULL;
		    npairs++;
		  }
		}
		if (streq(op,"fit")) {
		  rms=RmsFit(npairs,smol,sp,tp,&Tra);
		  if (rms>=0) Tot=m4xm4(Tra,Tot);
		}else rms=rms_dist(npairs,sp,tp);        /*measure only */
	      }else break;
	      nit--;
	      conv=conv||(npairs==onp && rms-orms < 0.001 && orms-rms < 0.01)||nit<0;
	      orms=rms; onp=npairs;
	    }while(!conv);
	    if (streq(op,"fit")) unlock_evtq();
	    if (onp>0) {
	      prtGraphics("r.m.s. ( %s , %s ) = %6.3f\n",
			  nodeptr_to_name(nm1,tmol),
			  nodeptr_to_name(nm2,smol),
			  orms);
	      if (nit<0) prtGraphics("warning: auto-fit did not converge.\n");
	      prtGraphics("%d automatic pairs.\n",onp);
	      if (streq(op,"fit")) SetFitCoTransform(Tot);
	    }else{
	      prtGraphics("r.m.s. ( %s , %s ) = 999999\n",
			  nodeptr_to_name(nm1,tmol),
			  nodeptr_to_name(nm2,smol));
	      prtGraphics("0 automatic pairs.\n");
	    }
	  }
	}
      }
    }
  }
  if (satoms) free(satoms);
  if (sp) free(sp);
  if (tatoms) free(tatoms);
  if (tp) free(tp);
  free_Token(key);
}
/*-------------------- cmd_fit_pairs --------------------*/

static void cmd_fit_pairs(char *op)
{
  Token *key;
  Node *tmol,*smol;
  int nat;
  Atom **tatoms,**satoms;
  float rms;
  char nm1[1024],nm2[1024],*prmt;
  mat_4 Tra;

  key=Get_TemplateMol();
  if (key->status==TOK_OK) {
    tmol=(Node *)key->ptr;
    key=free_Token(key);
    prmt=(streq(op,"fit") ? "molecule to fit to template: " : 
	  "measure rms-distance to molecules: ");
    key=Get_SourceMol(prmt,tmol);
    if (key->status==TOK_OK) {
      smol=(Node *)key->ptr;
      key=free_Token(key);
      key=Get_AtomPairlist(smol,&satoms,tmol,&tatoms);
      if (key->status==TOK_OK) {
	nat=key->n;
	if (streq(op,"fit")) {
	  rms=RmsFit(nat,smol,satoms,tatoms,&Tra);
	  if (rms>=0) SetFitCoTransform(Tra);
	}else rms=rms_dist(nat,satoms,tatoms);        /*measure only */
	prtGraphics("r.m.s. ( %s , %s ) = %6.3f\n",
		    nodeptr_to_name(nm1,tmol),
		    nodeptr_to_name(nm2,smol),
		    rms);
	free(satoms); free(tatoms);
      }
    }
  }
  free_Token(key);
}
/*-------------------- cmd_fit_shape --------------------*/

static void cmd_fit_shape(char *op)
{
  Token *key;
  Node *tmol,*smol;
  int natt,nats;
  Atom **tatoms,**satoms;
  char nm1[1024],nm2[1024],*prmt;
  float g;
  mat_4 Tra;

  key=(Token *)NULL;
  tatoms=(Atom **)NULL;
  satoms=(Atom **)NULL;

  key=Get_TemplateMol();
  if (key->status==TOK_OK) {
    tmol=(Node *)key->ptr;
    key=free_Token(key);
    key=Get_TemplateAtoms(tmol);
    if (key->status==TOK_OK) {
      natt=key->n;
      tatoms=(Atom **)key->ptr;
      prtGraphics("%d atoms selected in template %s\n",
		  natt,nodeptr_to_name(nm1,tmol));
      key=free_Token(key);
      prmt=(streq(op,"fit") ? "molecules to fit to template: " : 
	    "measure rms-distance to molecule: ");
      key=Get_SourceMol(prmt,tmol);
      if (key->status==TOK_OK) {
	smol=(Node *)key->ptr;
	key=free_Token(key);
	key=Get_AutoSourceAtoms(smol);
	if (key->status==TOK_OK) {
	  satoms=(Atom **)key->ptr;
	  nats=key->n;
	  prtGraphics("%d atoms selected in source %s\n",
		  nats,nodeptr_to_name(nm2,smol));
	  g=gsealFit(nats,smol,satoms,natt,tatoms,&Tra);
	  prtGraphics("%s %s %0.3f\n",nm1,nm2,g);
	  SetFitCoTransform(Tra);
	  free(satoms);
	}
	free(tatoms);
      }
    }
  }
}
/*-------------------- cmd_fit_cotransform --------------------*/

static void cmd_fit_cotransform(char *mode)
{
  Token *key=(Token *)NULL;
  Nodechain *c;
  Node *mol;
  mat_4 Tra;

  key=Get_SourceMols("molecules to co-transform: ");
  send_bheader(graphix,"ARMK",0);  
  if (key->status==TOK_OK) {
    Tra=GetFitCoTransform();
    if (streq(mode,"back")) Tra=mmxinv(Tra);
    c=(Nodechain *)key->ptr;
    for (mol=c->s;mvld(mol);mol=mol->chain) {
      if (is_mol(mol)) cotransform(Tra,mol);
    }
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) err_set(1,"bad molecule selection: %s",key->text);
  key=free_Token(key);
}
/*-------------------- Get_TemplateMol --------------------*/

static Token *Get_TemplateMol(void)
{
  Token *key;
  char *menu;
  Nodechain *c;

  menu=Make_TopMolNameList("",DefaultNode);
  key=Get_Mols_M("template molecule: ",";",menu);
  free_menu(menu);
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n<0) {
      err_set(1,"no template molecule selected.");
      key->status=TOK_NOK;
    }else if (c->n>1) {
      err_set(1,"ambiguous selection: %s",key->text);
      key->status=TOK_NOK;
    }else{
      key->n=1;
      key->ptr=(void *)c->s;
    } 
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  return(key);
}

/*-------------------- Get_TemplateAtoms --------------------*/

static Token *Get_TemplateAtoms(Node *root)
{
  Token *key;
  Nodechain *c;
  Dspobj *proot;
  Node *save_default;
  char nm[1024];

  proot=pick_root(assocmdl(root));
  save_default=DefaultNode;
  DefaultNode=root;

  key=Get_Atoms_D("template atoms: ",";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    chain_reduce_subtree(root,c);
    if (c->n<1) {
      err_set(1,"zero atoms in template %s",nodeptr_to_name(nm,root));
      key->status=TOK_NOK;
      clear_node_chain(c); 
      free_node_chain(c);
    }else{
      key->n=c->n;
      key->ptr=(void*)chain_to_ptr((Nodechain *)key->ptr,TRUE);
      if (!mvld(key->ptr)) key->status=TOK_NOK;
    }
  }
  pick_root(proot);
  DefaultNode=save_default;
  return(key);
}
/*-------------------- Get_AutoSourceAtoms --------------------*/

static Token *Get_AutoSourceAtoms(Node *root)
{
  Token *key;
  Nodechain *c;
  Dspobj *proot;
  Node *save_default;
  char nm[1024];

  proot=pick_root(assocmdl(root));
  save_default=DefaultNode;
  DefaultNode=root;

  key=Get_Atoms_D("source atoms: ",";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    chain_reduce_subtree(root,c);
    if (c->n<1) {
      err_set(1,"zero atoms in source %s",nodeptr_to_name(nm,root));
      key->status=TOK_NOK;
      clear_node_chain(c); 
      free_node_chain(c);
    }else{
      key->n=c->n;
      key->ptr=(void*)chain_to_ptr((Nodechain *)key->ptr,TRUE);
      if (!mvld(key->ptr)) key->status=TOK_NOK;
    }
  }
  pick_root(proot);
  DefaultNode=save_default;
  return(key);
}
/*-------------------- Get_SourceMols --------------------*/

static Token *Get_SourceMols(char *pr)
{
  Token *key;
  char *menu;
  Nodechain *c;

  menu=Make_TopMolNameList("",DefaultNode);
  key=Get_Mols_M(pr,";",menu);
  free_menu(menu);
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n<=0) {
      err_set(1,"no molecule selected.");
      key->status=TOK_NOK;
      free_node_chain(c);
    }else{
      key->n=c->n;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  return(key);
}
/*-------------------- Get_SourceMol --------------------*/

static Token *Get_SourceMol(char *pr,Node *tmol)
{
  Token *key;
  Nodechain *c;
  Node *smol;
  char nm1[1024],nm2[1024];

  key=Get_SourceMols(pr);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n>1) {
      err_set(1,"ambiguous selection: %s",key->text);
      key->status=TOK_NOK;
    }else{
      smol=c->s;
      if (smol!=tmol) {
	if (node_is_sub(smol,tmol)) {
	  err_set(1,"molecule %s is substructure of %s, not fitted.",
		  nodeptr_to_name(nm1,smol),nodeptr_to_name(nm2,tmol));
	  key->status=TOK_NOK;
	}else if (node_is_sub(tmol,smol)) {
	  err_set(1,"molecule %s is substructure of %s, not fitted.",
		  nodeptr_to_name(nm1,tmol),nodeptr_to_name(nm2,smol));
	  key->status=TOK_NOK;
	}else{
	  key->n=1;
	  key->ptr=smol;
	}
      }else{
	err_set(1,"Can't fit %s to itself.",nodeptr_to_name(nm2,tmol));
	key->status=TOK_NOK;
      }
    } 
    clear_node_chain(c);
    free_node_chain(c);
  }
  return(key);
}
/*-------------------- Get_SourceAtoms --------------------*/

static Token *Get_SourceAtoms(int nat,Node *root)
{
  Token *key;
  Nodechain *c;
  Dspobj *proot;
  Node *save_default;
  char nm[1024];

  proot=pick_root(assocmdl(root));
  save_default=DefaultNode;
  DefaultNode=root;

  key=Get_Atoms_D("atoms to be used in fit: ",";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    chain_reduce_subtree(root,c);
    if (c->n<1) {
      err_set(1,"zero atoms selected in %s",nodeptr_to_name(nm,root));
      key->status=TOK_NOK;
      free_node_chain(c);
    }else if (nat>0 && c->n != nat) {
      err_set(1,"number of atoms in source (%d) is different from",c->n);
      err_set(1,"number of atoms in template (%d).",nat);
      clear_node_chain(c); free_node_chain(c);
      key->status=TOK_NOK;
    }else{
      key->n=c->n;
      key->ptr=(void *)chain_to_ptr((Nodechain *)key->ptr,TRUE);
      if (!mvld(key->ptr)) key->status=TOK_NOK;
    }
  }
  pick_root(proot);
  DefaultNode=save_default;
  return(key);
}
/*-------------------- node_is_sub --------------------*/

static int node_is_sub(Node *a,Node *b)
{
  /* return TRUE iff a is subtree of b */

  for (;mvld(a);a=a->up) if (a==b) break;
  return(mvld(a)!=(void *)NULL);
}
/*-------------------- clear_amark --------------------*/

static void clear_amark(Atom *a)
{
  if (is_atom(a)) {
    if (cPtr(a->cross)) {
      send_bheader(graphix,"ARMK",cPtrSize());
      send_bcptr(graphix,a->cross);
      trflush(graphix);
    }
  }
}

/*-------------------- exact_partners --------------------*/

static Atom **exact_partners(Node *smol,int nat,Node *tmol,Atom **tatoms)
{
  /* Find atoms in smol that correspond to atoms in tmol(tatoms) */
  /* if smol==tmol or one a subtree of the other ---> error.     */

  Atom **r;
  Atom *a,*b;
  int i,status;
  char nm1[1024],nm2[1024];

  r=(Atom **)NULL; status=0;

  if (smol!=tmol) {
    if (node_is_sub(smol,tmol)) {
      err_set(1,"molecule %s is substructure of %s, not fitted.",
	      nodeptr_to_name(nm1,smol),nodeptr_to_name(nm2,tmol));
      status=1;
    }else if (node_is_sub(tmol,smol)) {
      err_set(1,"molecule %s is substructure of %s, not fitted.",
	      nodeptr_to_name(nm1,tmol),nodeptr_to_name(nm2,smol));
      status=2;;
    }
  }else{
    err_set(1,"Can't fit %s to itself.",nodeptr_to_name(nm2,tmol));
    status=1;
  }
  
  if (nat<1) status=1;

  if (!status) {
    r=(Atom **)malloc(nat*sizeof(Atom *));
    if (mvld(r)) {
      i=0; b=rfirst_atom(smol);
      ForAtoms(a,tmol) {
	if (!mvld(b)) break;              /* smol has less atoms than tmol */
	if (i<nat) if (tatoms[i]==a) r[i++]=b;
	b=rnext_atom(smol,b);
      }
      if (mvld(a)||mvld(b)||i!=nat) {
	err_set(1,"molecules are different ( %s , %s )",
		nodeptr_to_name(nm1,tmol),
		nodeptr_to_name(nm2,smol));
	free(r); r=(Atom **)NULL;
      }
    }
  }
  return(r);
}
/*-------------------- Get_AtomPairlist --------------------*/

static Token *Get_AtomPairlist(Node *s,Atom ***sa,Node *t,Atom ***ta)
{
  Token *key;
  Nodechain cs,ct;
  Node *save_default,*p;
  Dspobj *proot,*sd,*td;
  int state,n;
  Atom *a;
  char buf1[512];
  

  cs.n=0; cs.s=(Node *)NULL; cs.e=cs.s;
  ct=cs;

  save_default=DefaultNode;
  sd=assocobj(s);
  td=assocobj(t);
  proot=pick_root((Dspobj *)NULL);
  key=(Token *)NULL;
  state=TOK_OK;
  *sa=(Atom **)NULL;
  *ta=(Atom **)NULL;
  a=(Atom *)NULL;

  while (state==TOK_OK) {
    err_dump(); clear_amark(a);
    free_Token(key);
    DefaultNode=t; pick_root(td);
    key=Get_Atom("template atom: ",";");
    state=key->status;
    if (key->status==TOK_OK) {
      a=(Atom *)key->ptr;
      for (p=a->up;mvld(p);p=p->up) if (p==t) break;
      if (mvld(p)) {
	if (!(a->flg&(1<<1))) {
	  add_to_node_chain((Node *)a,&ct);
	  a->flg|=(1<<1); a=(Atom *)NULL;
	}else{
	  err_set(1,"template atom %s already selected.",
		  nodeptr_to_name(buf1,(Node *)a));
	  continue;
	}
      }else{
	err_set(1,"atom %s not in template.",
		nodeptr_to_name(buf1,(Node *)a));
	continue;
      }
    }
    while (state==TOK_OK) {
      err_dump(); clear_amark(a);
      free_Token(key);
      DefaultNode=s; pick_root(sd);
      key=Get_Atom("source atom: ",";");
      state=key->status;
      if (key->status==TOK_OK) {
	a=(Atom *)key->ptr;
	for (p=a->up;mvld(p);p=p->up) if (p==s) break;
	if (mvld(p)) {
	  if (!(a->flg&(1<<1))) {
	    add_to_node_chain((Node *)a,&cs);
	    a->flg|=(1<<1); a=(Atom *)NULL;
	    break;
	  }else{
	    err_set(1,"source atom %s already selected.",
		    nodeptr_to_name(buf1,(Node *)a));
	    continue;
	  }
	}else{
	  err_set(1,"atom %s not in source.",
		  nodeptr_to_name(buf1,(Node *)a));
	  continue;
	}
      }
    }
  }

  send_bheader(graphix,"ARMK",0);

  if (key->status==TOK_EOL||key->status==TOK_EOC) key->status=TOK_OK;

  if (cs.n!=ct.n) {
    err_set(1,"source/template atom mismatch.");
    key->status=TOK_NOK;
  }
  if (cs.n<1) {
    err_set(1,"zero atom pairs selected.");
    key->status=TOK_NOK;
  }

  if (key->status==TOK_OK) {
    *sa=(Atom **)malloc(cs.n*sizeof(Atom *));
    *ta=(Atom **)malloc(ct.n*sizeof(Atom *));
    if (!mvld(*sa)||!mvld(*ta)) {
      err_set(1,"memory allocation failure.");
      if (mvld(*sa)) free(*sa);
      *sa=(Atom **)NULL; 
      if (mvld(*ta)) free(*ta);
      *ta=(Atom **)NULL; 
      key->status=TOK_NOK;
    }else{
      for (n=0,a=(Atom *)cs.s;mvld(a);a=(Atom *)a->chain) (*sa)[n++]=a;
      for (n=0,a=(Atom *)ct.s;mvld(a);a=(Atom *)a->chain) (*ta)[n++]=a;
      key->n=n;
    }
  }

  for (a=(Atom *)cs.s;mvld(a);a=(Atom *)a->chain) a->flg&=~(1<<1);
  for (a=(Atom *)ct.s;mvld(a);a=(Atom *)a->chain) a->flg&=~(1<<1);
  clear_node_chain(&cs); clear_node_chain(&ct);

  DefaultNode=save_default;
  pick_root(proot);
  return(key);
}
/*-------------------- --------------------*/

static float gsealFit(int nats,Node *smol,Atom **satoms,int natt,Atom **tatoms,mat_4 *Tra)
{
  gseal_par_t P;
  float ct[3],cs[3],gs,gt,g,T[3],dT[3],E[6],dE[3];
  mat_4 GT;
  vec_3 V;
  int i,j;

  if (natt==0||nats==0) {
    if (Tra) *Tra=ident_4();
    g=0.0; gs=1.0; gt=1.0;
  }else{
    P.alpha=alpha;
    P.n1=natt;
    P.n2=nats;
    P.X1=(float *)malloc(3*natt*sizeof(float));
    P.X2=(float *)malloc(3*nats*sizeof(float));
    P.T1=NULL;
    P.T2=NULL;

    for (j=0;j<3;j++) ct[j]=0.0;

    for (i=0;i<natt;i++) {
      for (j=0;j<3;j++) {
	P.X1[3*i+j]=tatoms[i]->x[j];
	ct[j]+=tatoms[i]->x[j];
      }
    }

    for (j=0;j<3;j++) ct[j]/=natt;
    for (i=0;i<natt;i++) {
      for (j=0;j<3;j++) {
	P.X1[3*i+j]-=ct[j];
      }
    }
    
    for (j=0;j<3;j++) cs[j]=0.0;
    for (i=0;i<nats;i++) {
      for (j=0;j<3;j++) {
	P.X2[3*i+j]=satoms[i]->x[j];
	cs[j]+=satoms[i]->x[j];
      }
    }
    for (j=0;j<3;j++) cs[j]/=nats;
    for (i=0;i<nats;i++) {
      for (j=0;j<3;j++) {
	P.X2[3*i+j]-=cs[j];
      }
    }

    gt=gsealValue(P.n1,P.X1,P.T1,P.n1,P.X1,P.T1,P.alpha);
    gs=gsealValue(P.n2,P.X2,P.T2,P.n2,P.X2,P.T2,P.alpha);

    for (i=0;i<3;i++) {
      T[i]=0.0;
      dT[i]=3.0;
      E[i]=0.0; E[i+3]=0.0;   /* E[i+3] 0-translation, for EulerRot4 */
      dE[i]=2*3.141592653;
    }

    g=gsealBestFit(50,T,dT,E,dE,&P);
    if (Tra) {
      GT=EulerRot4(E);
      V=m4xv3(GT,vec3(cs));
      for (i=0;i<3;i++) GT.r[i].c[3]=T[i]+ct[i]-V.c[i];
      cotransform(GT,smol);
      *Tra=GT;
    }
  }
  /*** return(g*g/(gs*gt)); ***/
  return(g/MAX(gt,gs));
}

/*-------------------- SetFitCoTransform --------------------*/

static void SetFitCoTransform(mat_4 t)
{
  FitCoTransform=t;
}

/*-------------------- GetFitCoTransform --------------------*/

mat_4 GetFitCoTransform(void)
{
  return(FitCoTransform);
}
