#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "fileutil.h"
#include "errhand.h"

extern int graphix;
extern Node *MolRoot,*DefaultNode;

#define key_is(a) (streq(a,key->key))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define DOM_MOL   0
#define DOM_FRG   1

#define LIB_UNDEF  0
#define LIB_DEF    1
#define LIB_OPEN   2

#define COPY_UL      1
#define COPY_CENTER  2
#define COPY_ORIGIN  3

#define BLD_FLAT     1
#define BLD_RESIDUES 2

#define MAXLIB 100                   /* maximum number of defined libraries     */
static Node* FrgRoot = (Node *)NULL; /* root of fragment library                */
static int lib_state[MAXLIB];        /* state (undefined, defined, opened       */
static char *lib_file[MAXLIB];       /* name of file associated with library    */
static char *lib_name[MAXLIB];       /* official name of library                */
static char *lib_title[MAXLIB];      /* library title (name given in .mol file) */

static int node_nr(Node *n);
static int mol_nr(Node *n);
static int open_frag_lib(Node *n);
static Node *find_fragment(Node *lib,char *name);
static Node *switch_active_lib(Node *old,char *new);
static Node *active_frag(void);
static Node *switch_active_frag(Node *lib,char *name);
static char *make_menu(char *libs,char *kwd,Node *active);
static void del_active_copy(void);
static Node *fragment_copy(Node *f,int mode);
static Node *molecule(Atom *a);
static Bond *join(Atom *frg,Atom *mol,int dom);
static void split(void);
static Bond *rotate_bond(Bond *b);
static void freeze_bond(Bond *b);
static void get_fragment(Node *F);
static void adjust_torsion(Atom *a,Atom *b,Nodechain *c,int bt);

static int bld_mode=BLD_RESIDUES;

/*========================================================================*/

void cmd_fragment(void)
{
  char *libs,*menu,*p,*prmt;
  char *kwd = "close flip get mode split End";
  Token *key;
  int loop;
  Node *ActiveLib,*ActiveFrag,*m;
  Atom *picked_atom,*FrgAtom,*MolAtom,*from,*to;
  Bond *Rot,*bnd;

  libs=(char *)NULL; menu=(char *)NULL;
  key=(Token *)NULL;
  loop=TRUE;
  ActiveLib=(Node *)NULL;
  ActiveFrag=(Node *)NULL;
  FrgAtom=MolAtom=(Atom *)NULL;
  Rot=(Bond *)NULL;
  
  if (!is_mol(FrgRoot)) {
    err_set(1,"No fragments. Define fragments first.");
    return;
  }
  if (!mvld(FrgRoot->fst)) {
    err_set(1,"No fragments. Define fragments first.");
    return;
  }

  libs=Make_TopMolNameList("",FrgRoot);
  if (mvld(libs)) for (p=libs;*p;p++) *p=(*p=='/' ? ' ' : *p);

  while(loop) {
    err_dump(); start_parse();
    free_menu(menu); key=free_Token(key);
    menu=make_menu(libs,kwd,ActiveLib);
    if (mvld(ActiveFrag)) prmt="join fragment> ";
    else if (mvld(ActiveLib)) prmt="select fragment> ";
    else prmt="select library> ";
    picked_atom=(Atom *)NULL;
    key=Get_Atom_M(prmt,";",menu);
    if (key->status==TOK_OK) {
      picked_atom=(Atom *)key->ptr;
      if (picked_atom->nbnd!=1) {    /* this might be twist... */
	from=picked_atom; to=(Atom *)NULL;
	picked_atom=(Atom *)NULL;
	key=free_Token(key);
	key=Get_Atom("twist bond (2nd Atom): ",";");
	if (key->status==TOK_OK) {
	  to=(Atom *)key->ptr;
	  bnd=find_bond(from,to);
	  if (mvld(bnd)) {
	    if (assocmdl((Node *)bnd->f)==assocmdl((Node *)bnd->t)) {
	      if (from!=bnd->f) {bnd->f=from;bnd->t=to;}   /* swap (f,t), if necessary... */
	      freeze_bond(Rot);
	      Rot=rotate_bond(bnd);
	    }else err_set(1,"intermolecular bond, no twist.");
	  }else err_set(1,"2nd atom not bonded to first atom.");
	}else if (key->status==TOK_NOK) err_set(1,"bad atom selection (twist): %s",key->text);
	key->status=TOK_NULL;
	if (from!=FrgAtom&&from!=MolAtom) clear_atom_mark(from);
	if (to!=FrgAtom&&to!=MolAtom) clear_atom_mark(to);
      }
    }else key=Prs_Kwd(key,menu);
    if (picked_atom) {
      m=molecule(picked_atom);
      if (mvld(ActiveFrag)&&m==ActiveFrag->chain) {
	if (mvld(FrgAtom)) if (FrgAtom!=picked_atom) clear_atom_mark(FrgAtom);
	FrgAtom=picked_atom;
	if (mvld(MolAtom)) {
	  freeze_bond(Rot);
	  Rot=rotate_bond(join(MolAtom,FrgAtom,DOM_MOL));
	  clear_atom_mark(FrgAtom);
	  MolAtom=FrgAtom=(Atom *)NULL;
	}
      }else{
	if (mvld(MolAtom)) if (MolAtom!=picked_atom) clear_atom_mark(MolAtom);
	MolAtom=picked_atom;
	if (mvld(FrgAtom)) {
	  freeze_bond(Rot);
	  Rot=rotate_bond(join(MolAtom,FrgAtom,DOM_FRG));
	  MolAtom=(Atom *)NULL;
	}
      }
    }else if (key->status==TOK_OK) {
      if (find_fragment(ActiveLib,key->key)) {
	FrgAtom=(Atom *)NULL;
	ActiveFrag=switch_active_frag(ActiveLib,key->key);
      }else if ((key=Prs_Kwd(key,libs))->status==TOK_OK) {
	if (mvld(ActiveFrag)) del_active_copy();
	ActiveFrag=(Node *)NULL;
	FrgAtom=(Atom *)NULL;
	ActiveLib=switch_active_lib(ActiveLib,key->key);
      }else if ((key=Prs_Kwd(key,kwd))->status==TOK_OK) {
	if (key_is("End")) loop=FALSE;
	else if (key_is("mode")) {
	  key=free_Token(key);
	  default_token(bld_mode==BLD_FLAT?"flat":"residues");
	  key=Get_Kwd("build mode: ",";","flat residues");
	  if (key->status==TOK_OK) {
	    if (key_is("flat")) bld_mode=BLD_FLAT;
	    else if (key_is("residues")) bld_mode=BLD_RESIDUES;
	  }else if (key->status==TOK_NOK) err_set(1,"invalid build mode: %s",key->text);
	}else if (key_is("get")) {
	  if (mvld(ActiveFrag)) get_fragment(ActiveFrag->chain);
	  else err_set(1,"you must select a fragment first.");
	}else if (key_is("close")) {
	  if (MolAtom) {clear_atom_mark(MolAtom);MolAtom=(Atom *)NULL;}
	  if (FrgAtom) {clear_atom_mark(FrgAtom);FrgAtom=(Atom *)NULL;}
	  key=free_Token(key);
	  key=Get_Atom("first atom",";");
	  if (key->status==TOK_OK) {
	    from=key->ptr;
	    if (from->nbnd==1) {
	      key=free_Token(key);
	      key=Get_Atom("second atom",";");
	      if (key->status==TOK_OK) {
		to=key->ptr;
		if (to->nbnd==1) {
		  if (Rot) {freeze_bond(Rot);Rot=(Bond *)NULL;}
		  bnd=add_bond(neighbour(0,from),neighbour(0,to));
		  set_bt(bnd,from->bnd[0]->bt);
		  del_node((Node *)from);
		  del_node((Node *)to);
		}else err_set(1,"sorry, ring clouse atom must be terminal atom");
	      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	    }else err_set(1,"sorry, ring clouse atom must be terminal atom");
	  }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	}else if (key_is("flip")) {
	  if (Rot) {
	    key=free_Token(key);
	    key=Get_Float_M("flip angle: ",";","180 150 120 90 60 30 ' ' -30 -60 -90 -120 -150");
	    if (key->status==TOK_OK) {
	      send_bheader(graphix,"SRAX",sizeof(float));
	      send_float(graphix,key->f);
	    }else if (key->status==TOK_NOK) err_set(1,"bad flip angle: %s",key->text);
	  }else err_set(1,"axis of rotation is not defined.");
	}else if (key_is("split")) {
	  if (Rot) {freeze_bond(Rot);Rot=(Bond *)NULL;}
	  if (MolAtom) {clear_atom_mark(MolAtom);MolAtom=(Atom *)NULL;}
	  if (FrgAtom) {clear_atom_mark(FrgAtom);FrgAtom=(Atom *)NULL;}
	  split();
	}
      }
    }
    if (key->status==TOK_NOK) {
      err_set(1,"bad fragment command: %s",key->text);
    }else if (key->status==TOK_ABORT||key->status==TOK_EOL) loop=FALSE;  
    if (end_parse()!=TOK_OK) {
      err_set(0,"extra characters at end of command ignored.");
    }
  }

  send_bheader(graphix,"ARMK",0);
  freeze_bond(Rot);
  del_active_copy();
  free_menu(libs);
  free_menu(menu);  
  free_Token(key);

}

/*-------------------- define_fragment_lib --------------------*/

void define_fragment_lib(char *name,UCpipe *lib)
{
  Node *r,*o;
  int n;

  /* Associate fragment library with file... */

  if (!mvld(FrgRoot)) {
    FrgRoot=make_node(OBJ_MOL);
    if (mvld(FrgRoot)) set_node_name(FrgRoot,"FragmentLibrary");
    else err_set(1,"failed to initialize fragment library.");
  }
  if (is_mol(FrgRoot)) {
    r=find_name(FrgRoot,name,OBJ_MOL);
    if (mvld(r)) {
      n=node_nr(r);
      if (lib_state[n]==LIB_OPEN) while(mvld(r->fst)) del_node(r->fst);
      if (mvld(lib_file[n])) free(lib_file[n]);
      if (mvld(lib_name[n])) free(lib_name[n]);
      if (mvld(lib_title[n])) free(lib_title[n]);
      lib_file[n]=Strdup(lib->filename);
      lib_name[n]=Strdup(name);
      lib_title[n]=(char *)NULL;
      lib_state[n]=(mvld(lib_file[n]) ? LIB_DEF : LIB_UNDEF);
    }else{
      o=add_node(FrgRoot,name,OBJ_MOL);
      if (mvld(o)) {
	n=node_nr(o);
	if (n<MAXLIB) {
	  lib_file[n]=Strdup(lib->filename);
	  lib_name[n]=Strdup(name);
	  lib_title[n]=(char *)NULL;
	  lib_state[n]=(mvld(lib_file[n]) ? LIB_DEF : LIB_UNDEF);
	}else{
	  err_set(1,"too many defined fragment libraries.");
	  del_node(o);
	}
      }else err_set(1,"failed to add fragment library definition.");
    }
  }else err_set(1,"fragment library root not found.");
}
/*-------------------- del_fragment --------------------*/

int del_fragment(Node *l)
{
  int n,r;
  Node *o;

  r=0;
  if (is_mol(l)) {
    if (is_mol(FrgRoot)&&l->up==FrgRoot) {
      n=node_nr(l);
      lib_state[n]=LIB_UNDEF;
      if (mvld(lib_file[n])) free(lib_file[n]);
      if (mvld(lib_name[n])) free(lib_name[n]);
      if (mvld(lib_title[n])) free(lib_title[n]);
      for (o=l->nxt;mvld(o);n++) {
	lib_state[n]=lib_state[n+1];
	lib_file[n]=lib_file[n+1];
	lib_name[n]=lib_name[n+1];
	lib_title[n]=lib_title[n+1];
	o=o->nxt;
      }
      del_node(l);
      r++;
    }
  }
  return(r);
}
/*-------------------- switch_active_lib --------------------*/

static Node *switch_active_lib(Node *old,char *name)
{
  Node *o;

  o=find_name(FrgRoot,name,OBJ_MOL);
  if (open_frag_lib(o)) {
    if (o!=old) del_active_copy();
  }else {
    err_set(1,"failed to open fragment library %s",name);
    o=old;;
  }
  return(o);
}
/*-------------------- find_fragment --------------------*/

static Node *find_fragment(Node *lib,char *name)
{
  Node *o,*a;

  if (is_mol(lib)&&lib->up==FrgRoot) {
    o=find_name(lib,name,OBJ_MOL);
    a=first_type(o,OBJ_MOL);
    o=(is_mol(a) ? a : o);
  }else o=(Node *)NULL;
  return(o);
}

/*-------------------- switch_active_frag --------------------*/

static Node *switch_active_frag(Node *lib,char *name)
{
  Node *o,*a;
  Atom *x;
  Bond *b;
  int i;

  if (is_mol(lib)&&lib->up==FrgRoot) {
    o=find_fragment(lib,name);
    if (is_mol(o)) {
      a=active_frag();
      lock_evtq();
      if (mvld(a)) {
	del_node(a->chain);
	a->chain=(Node *)NULL;
      }
      o->chain=fragment_copy(o,COPY_UL);
      if (mvld(o->chain)) {
	ForAtoms(x,o->chain) {
	  for (i=0;i<x->nbnd;i++) {
	    b=x->bnd[i];
	    if (b->bt!=BT_SINGLE) set_bond_label(b,BND_LAB_TYPE,NULL);
	  }
	}
      }
      unlock_evtq();
    }
  }else{
    o=(Node *)NULL;
    err_set(1,"fragment %s not found in current library.",name);
  }
  return(o);
}
/*-------------------- fragment_copy --------------------*/

static Node *fragment_copy(Node *frg,int mode)
{
  Node *c,*m;
  Atom *a;
  int i;
  mat_4 M,R;
  PrjTS prj;
  vec_3 ctr,t;
  float l[3],r[3];

  c=tree_copy(frg);
  if (mvld(c)) {
    m=add_node(MolRoot,frg->up->name,OBJ_MOL);
    if (mvld(frg->up->up)&&frg->up->up->up==FrgRoot) set_node_name(m,frg->up->up->name);
    if (mvld(m)) {
      c->up=m; link_node(c); c=m;
      a=rfirst_atom(c);
      if (mvld(a)) for (i=0;i<3;i++) l[i]=r[i]=a->x[i];
      else for (i=0;i<3;i++) l[i]=r[i]=0.0;
      ForAtoms(a,c) {
	a->flg|=(1<<1);
	for (i=0;i<3;i++) {
	  l[i]=(l[i]>a->x[i]?a->x[i]:l[i]);
	  r[i]=(r[i]<a->x[i]?a->x[i]:r[i]);
	}
      }
      ForAtoms(a,c) {
	for (i=0;i<3;i++) a->x[i]-=0.5*(r[i]+l[i]);
	for (i=a->nbnd;i>0;i--) if (!(neighbour(i-1,a)->flg&(1<<1))) del_bond(a->bnd[i-1]);
      }
      prj=getThreeDPrj(); M=obj_transform(MolRoot); R=mmxinv(M);
      switch (mode) {
      case COPY_UL:
	ctr.c[0]=-0.5*prj.ysize*prj.aspect+0.5*(r[0]-l[0])+1.0;
	ctr.c[1]= 0.5*prj.ysize-0.5*(r[1]-l[1])-1.0;
	ctr.c[2]= prj.znear-0.5*(r[2]-l[2])-1.0;
	break;
      case COPY_CENTER:
	ctr.c[0]=0.0;
	ctr.c[1]=0.0;
	ctr.c[2]=prj.znear-0.5*prj.slabwidth;
	break;
      case COPY_ORIGIN:
	ctr.c[0]=0.0;
	ctr.c[1]=0.0;
	ctr.c[2]=0.0;
	break;
      default:
	ctr.c[0]=0.0;
	ctr.c[1]=0.0;
	ctr.c[2]=prj.znear-0.5*prj.slabwidth;
	break;
      }
      ctr=m4xv3(R,ctr);for (i=0;i<3;i++) t.c[i]=ctr.c[i]-R.r[i].c[3];
      ForAtoms(a,c) {
	a->flg&=~(1<<1);
	v32arr(a->x,v3pv3(m4xv3(R,(vec3(a->x))),t));
      }
      lock_evtq();
      send_molecule(MolRoot,(Mol *)m);
      set_pivot(glblobj(assocobj(m)),ctr.c);
      send_bheader(graphix,"SALT",sizeof(Dspobj *));
      send_bptr(graphix,glblobj(assocobj(m)));
      unlock_evtq();
    }else err_set(1,"memory allocation failed.");
  }else err_set(1,"failed to copy fragment %s",frg->name);
  return(c);
}

/*-------------------- open_frag_lib --------------------*/

static int open_frag_lib(Node *l)
{
  int r,n;
  Node *m,*z;
  Atom *a;
  char *cn,*rn;
  
  r=0;
  if (is_mol(l)&&l->up==FrgRoot) {
    n=node_nr(l);
    if (lib_state[n]==LIB_OPEN) {
      r++;
      if (mvld(lib_title[n])) prtGraphics("%s\n",lib_title[n]);
    }else if (lib_state[n]==LIB_DEF) {
      if (mvld(lib_file[n])) {
	if (!read_syb((UCpipe *)NULL,lib_file[n],l,FT_MOL2)) {
	  lib_title[n]=Strdup(l->name);
	  set_node_name(l,lib_name[n]);
	  lib_state[n]=LIB_OPEN;
	  forObjs(m,l,OBJ_MOL) {         /* swap chain and residue name */
	    z=first_type(m,OBJ_MOL);
	    if (mvld(z)) {
	      rn=m->name; cn=z->name;
	      m->name=cn; z->name=rn;
	    }
	  }
	  ForAtoms(a,l) {
	    if (mvld(a->atp)) a->color=-a->atp->color;
	    a->lig=eff_lig(a);
	    a->chrg=asg_formchrg(a,a->lig);
	  }
	  r++;
	}else{
	  err_set(1,"failed to read fragment library file %s",lib_file[n]);
	}
      }
    }
  }
  return(r);
}
/*-------------------- make_menu --------------------*/

static char *make_menu(char *libs,char *kwd,Node *active)
{
  Mol *m;
  char *r;
  int l1,l2,l3,l;

  r=(char *)NULL;
  
  if (libs) l1=strlen(libs); else l1=0;
  if (kwd) l2=strlen(kwd); else l2=0;
  l3=0;
  forMols(m,active) if (mvld(m->name)) l3+=strlen(m->name)+1;
  l=l1+l2+l3+(l1>0 ? strlen(" ' ' ") : 0)+(l2>0 ? strlen(" ' ' "):0)+1;
  if (l>0) {
    r=(char *)malloc(l*sizeof(char));
    if (mvld(r)) {
      *r='\0';
      if (l2>0) sprintf(r+strlen(r),"%s ' ' ",kwd);
      if (l1>0) sprintf(r+strlen(r),"%s ' ' ",libs);
      if (l3>0) forMols(m,active) if (mvld(m->name)) sprintf(r+strlen(r),"%s ",m->name);
    }
  }
  return(r);
}
/*-------------------- join --------------------*/

static Bond *join(Atom *mol,Atom *frg,int dom)
{
  /* join fragment with molecule */

  Bond *bnd;
  float x[2][3],y[2][3],D;
  Node *cp,*F,*M,*up;
  mat_4 trf;
  Atom *a[2],*p[2],*atm;
  Xatom *xatm;
  Dspobj *mdl;
  int i,bt,flat;
  Nodechain C;
  Xcn *xcn;
  char buff[4096];

  bnd=(Bond *)NULL;

  F=molecule(frg);
  M=molecule(mol);
  flat = (mol->up==M);

  if (!is_atom(frg)||!is_atom(mol)) return(bnd);
  if (frg->nbnd!=1 || mol->nbnd!=1) return(bnd);


  a[0]=mol; p[0]=neighbour(0,a[0]);
  a[1]=frg; p[1]=neighbour(0,a[1]);
  for (i=0;i<3;i++) {
    x[0][i]=p[0]->x[i]; y[0][i]=a[0]->x[i];
    x[1][i]=a[1]->x[i]; y[1][i]=p[1]->x[i];
  }
  bt=a[dom]->bnd[0]->bt;
  D=default_bndlength(p[0],p[1],bt);
  v32arr(x[1],v3pv3(vec3(x[0]),sxv3(D,v3n(v3mv3(vec3(a[0]->x),vec3(x[0]))))));
  v32arr(y[0],v3pv3(vec3(y[1]),sxv3(D,v3n(v3mv3(vec3(a[1]->x),vec3(y[1]))))));
  rms_best_mat_yx(2,(float *)y,(float *)x,&trf);
  a[1]->flg|=(1<<1); p[1]->flg|=(1<<2);   /* mark join atom and precursor */
  cp=tree_copy(F);
  a[1]->flg&=~(1<<1); p[1]->flg&=~(1<<2); /* clear marks in source, keep in copy */
  if (mvld(cp)) {
    cp->nxt=M;
    link_node(cp);                        /* renumbers atomsets as necessary... */
    copy_set_bits(F,cp);
    unlink_node(cp);
    renumber_atomsets(cp,M);
    ForAtoms(atm,cp) {
      if (atm->flg&(1<<1)) a[1]=atm; 
      if (atm->flg&(1<<2)) p[1]=atm;
      transform_atom(trf,atm);
      while(mvld(xatm=first_xatom(atm))) del_node((Node *)xatm);
      atm->xtal=FALSE;
    }
    if (a[1]->flg&(1<<1)) del_node((Node *)a[1]);
    p[1]->flg&=~(1<<2); 
    C=tree2chain(cp,OBJ_ATOM);
    while(mvld(xcn=rfirst_xcn(cp))) del_node((Node *)xcn);
    if (bld_mode==BLD_FLAT) {
      for(up=C.s;mvld(up);up=up->chain) {
	unlink_node(up);
	up->up=mol->up; link_node(up);
      }
    }else{      
      if (flat) {
	for(up=C.s;mvld(up);up=up->chain) {
	  unlink_node(up);
	  up->up=M; link_node(up);
	}
      }else{
	for (up=cp->fst;mvld(up);up=cp->fst) {   
	  unlink_node(up);
	  up->up=M; link_node(up);
	  sprintf(buff,"%s_%d",(up->name?up->name:"ANON"),mol_nr(up));
	  set_node_name(up,buff);
	  if (is_atomset(up)) register_atomset((Atomset *)up);
	}
	merge_atomsets(M);
      }
    }
    lock_evtq();
    del_node((Node *)a[0]);
    adjust_torsion(p[0],p[1],&C,bt);
    bnd=add_bond_ng(p[0],p[1]);
    set_bt(bnd,bt);
    mdl=assocmdl((Node *)p[0]);
    mdl_add_atom_chain(mdl,&C);      /* and send copy to graphics */
    unlock_evtq();
    clear_node_chain(&C);
    del_node(cp);
  }else{
    err_set(1,"failed to copy source tree.");
  }
  return(bnd);
}
/*-------------------- rotate_bond --------------------*/

static Bond *rotate_bond(Bond *B)
{
  /* set up rotatable bond B */

  Nodechain *m;
  int ngm,size;
  Atom *am;

  if (mvld(B)) {
    m=connected_to_right(B->f,B->t);
    if (mvld(m)) {
      if (m->n>1) {
	B->f->chain=m->s;
	m->s=(Node *)B->f;
	m->n++;
	for (ngm=0,am=(Atom *)m->s;mvld(am);am=(Atom *)am->chain) if (cPtr(am->cross)) ngm++;
	size=(ngm)*cPtrSize();
	send_bheader(graphix,"DRAX",size);
	for (am=(Atom *)m->s;mvld(am);am=(Atom *)am->chain)
	  if (cPtr(am->cross)) send_bcptr(graphix,am->cross);
	trflush(graphix);
      }else B=(Bond *)NULL;
      clear_node_chain(m);
      free_node_chain(m);
    }else B=(Bond *)NULL;
  }
  return(B);
}
/*-------------------- freeze_bond --------------------*/

static void freeze_bond(Bond *B)
{
  /* freeze rotation around rotatable bond B */

  Atom *a;
  mat_4 rot;
  Qmsg *Msg;
  float angle;
  int size;
  Nodechain *m;
  
  if (mvld(B)&&graphix>0) {
    send_header(graphix,"FRAX",0);
    Msg=waitqmsg("SRAX SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }
    }
    if (Msg) {
      size=(Msg->size)/sizeof(float);
      if (size>0) angle=MY_f(*((int *)Msg->msg));
      else angle=0.0;
      delqmsg(Msg);
    }else angle=0.0;
    send_bheader(graphix,"DRAX",0);
    if (angle<-0.001||angle>0.001) {
      m=connected_to_right(B->f,B->t);
      if (mvld(m)) {
	if (m->n>1) {
	  rot=rotax(B->f->x,B->t->x,angle);
	  for (a=(Atom *)m->s;mvld(a);a=(Atom *)a->chain) transform_atom(rot,a);
	}
	clear_node_chain(m);
	free_node_chain(m);
      }
    }
  }
}
/*-------------------- split --------------------*/

static void split(void)
{
  /* split along bond, add Hydrogen to fill free valence */

  Token *key;
  Bond *bnd;
  Nodechain *c;
  Node *o,*p;
  Atom *a,*b,*n;
  char buff[1024];
  float d,s;
  int bt;

  key=(Token *)NULL;
  key=Get_Bond("split bond ",";");
  if (key->status==TOK_OK) {
    bnd=(Bond *)key->ptr;
    c=connected_to_right(bnd->f,bnd->t);
    if (mvld(c)) {
      if (c->n>0) {
	a=(Atom *)c->s;
	c->s=c->s->chain; c->n--;       /* don't delete first atom... */
	a->chain=(Node *)NULL;
	for (o=c->s;mvld(o);o=o->chain) {
	  p=o->up; unlink_node(o);
	  if (mvld(p)) {
	    if (!mvld(p->fst)) add_to_node_chain(p,c);
	  }
	}
	lock_evtq();
	del_node_chain(c);
	if (a->nbnd==1) {       /* convert a to hydrogen */
	  b=neighbour(0,a);
	  o=a->up; unlink_node((Node *)a); if (!mvld(o->fst)) del_node(o);
	  a->prv=(Node *)b; link_node((Node *)a);
	  d=cdist(a->x,b->x); d=(d>0 ? d : 1.0);
	  s=(bnd->bt==BT_SINGLE ? 1.05/d : 1.0/d);
	  v32arr(a->x,v3pv3(vec3(b->x),sxv3(s,(v3mv3(vec3(a->x),vec3(b->x))))));
	  set_atom_type(a,guess_atom_type(bnd->bt==BT_SINGLE ? "H" : "Du"));
	  a->lig=1; a->q=0.0; a->chrg=0;
	  if (mvld(a->atp)) a->color=a->atp->color;
	  set_node_name((Node *)a,auto_atom_name(a,buff,AUTONM_SEQ));
	  if (cPtr(a->cross)) {
	    send_bheader(graphix,"ALRM",cPtrSize());
	    send_bcptr(graphix,a->cross);
	    send_bheader(graphix,"NEWX",3*sizeof(float)+cPtrSize());
	    send_bcptr(graphix,a->cross); send_bfloata(graphix,3,a->x);
	    send_bheader(graphix,"ATCO",sizeof(int)+cPtrSize());
	    send_bint(graphix,a->color); send_bcptr(graphix,a->cross);
	  }
	}
	unlock_evtq();
	clear_node_chain(c);
      }else{                              /* remove ring bond, fill free valencies */
	lock_evtq();
	a=bnd->f; b=bnd->t;
	bt=bnd->bt;
	del_bond(bnd);
	n=(Atom *)add_node(b->up,(char *)NULL,OBJ_ATOM);
	if (is_atom(n)) {
	  d=cdist(a->x,b->x); d=(d>0 ? d : 1.0);
	  s=(bt==BT_SINGLE ? 1.05/d : 1.0/d);
	  v32arr(n->x,v3pv3(vec3(b->x),sxv3(s,(v3mv3(vec3(a->x),vec3(b->x))))));
	  n->nxyz=1;
	  set_atom_type(n,guess_atom_type(bt==BT_SINGLE ? "H" : "Du"));
	  n->lig=1; n->q=0.0; n->chrg=0;
	  if (mvld(n->atp)) n->color=n->atp->color;
	  set_node_name((Node *)n,auto_atom_name(n,buff,AUTONM_SEQ));
	  bnd=add_bond_raw(b,n);
	  set_bt(bnd,bt);
	  send_one_atom(n);
	}else err_set(1,"failed to add atom.");
	if (is_atom(n)) {
	  n=(Atom *)add_node(a->up,(char *)NULL,OBJ_ATOM);
	  d=cdist(a->x,b->x); d=(d>0 ? d : 1.0);
	  s=(bt==BT_SINGLE ? 1.05/d : 1.0/d);
	  v32arr(n->x,v3pv3(vec3(a->x),sxv3(s,(v3mv3(vec3(b->x),vec3(a->x))))));
	  n->nxyz=1;
	  set_atom_type(n,guess_atom_type(bt==BT_SINGLE ? "H" : "Du"));
	  n->lig=1; n->q=0.0; n->chrg=0;
	  if (mvld(n->atp)) n->color=n->atp->color;
	  set_node_name((Node *)n,auto_atom_name(n,buff,AUTONM_SEQ));
	  bnd=add_bond_raw(a,n);
	  set_bt(bnd,bt);
	  send_one_atom(n);
	}else err_set(1,"failed to add atom.");
	unlock_evtq();
      }
    }else err_set(1,"right_connected: failed.");
  }else if (key->status==TOK_NOK) err_set(1,"bad bond selection: %s",key->text);
  send_bheader(graphix,"ARMK",0);
  free_Token(key);
}
/*-------------------- get_fragment --------------------*/

static void get_fragment(Node *F)
{
  /* copy active fragment into molecule tree */

  Token *key;
  Node *m,*o;
  char buff[4096];

  key=(Token *)NULL;
  if (mvld(F)) {
    key=Get_String("name of new molecule: ",";");
    if (key->status==TOK_OK) {
      m=fragment_copy(F,COPY_ORIGIN);
      if (mvld(m)) {
	set_node_name(m,key->text);
	forObjs(o,m->fst,OBJ_MOL) {
	  sprintf(buff,"%s_%d",(o->name?o->name:"ANON"),mol_nr(o));
	  set_node_name(o,buff);
	}
      }
    }else if (key->status==TOK_NOK) err_set(1,"bad name: %s",key->text);
  }else err_set(1,"you must select a fragment first.");
  free_Token(key);
}
/*-------------------- adjust_torsion --------------------*/

static void adjust_torsion(Atom *a,Atom *b,Nodechain *c,int bt)
{
  Atom *na,*nb,*d;
  int wa,wb,i,j,z;
  float tau,target;
  mat_4 tra;

  if (c->n<1) return;
  if (a->nbnd<1||b->nbnd<1) return;
  if (bt==BT_AROMATIC) {
    target=0.0;
    for (i=0;i<a->nbnd;i++) if (a->bnd[i]->bt==BT_AROMATIC) break;
    na=(i<a->nbnd ? neighbour(i,a) : (Atom *)NULL);
    for (i=0;i<b->nbnd;i++) if (b->bnd[i]->bt==BT_AROMATIC) break;
    nb=(i<b->nbnd ? neighbour(i,b) : (Atom *)NULL);
  }else{
    target=180.0;
    for (wa=0,na=(Atom *)NULL,i=0;i<a->nbnd;i++) {
      d=neighbour(i,a);
      for (z=0,j=0;j<d->nbnd;j++) z+=(mvld(neighbour(j,d)->atp) ? neighbour(j,d)->atp->z : 1);
      if (z>wa) {
	wa=z;
	na=d;
      }
    }
    for (wb=0,nb=(Atom *)NULL,i=0;i<b->nbnd;i++) {
      d=neighbour(i,b);
      for (z=0,j=0;j<d->nbnd;j++) z+=(mvld(neighbour(j,d)->atp) ? neighbour(j,d)->atp->z : 1);
      if (z>wb) {
	wb=z;
	nb=d;
      }
    }
  }
  if (is_atom(na)&&is_atom(nb)) {
    tau=ctorsd(na->x,a->x,b->x,nb->x);
    tra=rotax(a->x,b->x,target-tau);
    for (d=(Atom *)c->s;mvld(d);d=(Atom *)d->chain) transform_atom(tra,d);
  }
}
  
/*-------------------- node_nr --------------------*/

static int node_nr(Node *n)
{
  int r;

  for (r=0;mvld(n->prv);n=n->prv) r++;
  return(r);
}
/*-------------------- mol_nr --------------------*/

static int mol_nr(Node *n)
{
  int r;

  for (r=1;mvld(n->prv);n=n->prv) if (is_mol(n)) r++;
  if (!is_mol(n)) r=0;
  return(r);
}
/*-------------------- active_frag --------------------*/

static Node *active_frag(void)
{
  Node *m;
  
  ForObjs(m,FrgRoot,OBJ_MOL) if (is_mol(m->chain)) break;
  return(m);
}
/*-------------------- del_active_copy --------------------*/

static void del_active_copy(void)
{
  Node *m;
  
  ForObjs(m,FrgRoot,OBJ_MOL) {
    if (is_mol(m->chain)) del_node(m->chain);
    m->chain=(Node *)NULL;
  }
}
/*-------------------- molecule --------------------*/

static Node *molecule(Atom *a)
{
  /* returns molecule containing a (assocmdl->root, a->up->up, or a->up if mol is "flat") */

  Node *r;

  r=assocmdl((Node *)a)->root;
  if (!mvld(r)) {
    r=(a->up->up);
    if (r==MolRoot) r=a->up;
  }else{
    r=(a->up==r ? r : (a->up->up == MolRoot ? a->up : a->up->up));
  }
  return(r);
}
/*-------------------- fragment_root --------------------*/

Node *fragment_root(void)
{
  return(FrgRoot);
}
