#include <stdio.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "parser.h"
#include "transpor.h"
#include "bhtree.h"
#include "errhand.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define MAXTREE 1024
#define MAXCLS  1024

/*-------------------- grid_create --------------------*/

Grid *grid_create(Grid *G)
{
  Token *key;
  Nodechain *c;
  Atom *a;
  char *prx[]={"x: ","y: ","z: "};
  char *prm[]={"x-extent: ","y-extent: ","z-extent: "};
  char *margmv="rvdw 0.0 0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 6.0 7.0 8.0 9.0 10.0 15.0 20.0 30.0";
  char *margm="0.0 0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 6.0 7.0 8.0 9.0 10.0 15.0 20.0 30.0";
  char *resm="0.20 0.25 0.30 0.33 0.35 0.40 0.50 0.67 0.750 1.0 1.25 1.50 2.0";
  float xmin[3],xmax[3],res=0.0,xmrg;
  int i,mrgvdw;

  key=(Token *)NULL;

  key=Get_Kwd("center: ",";","around xyz");
  if (key->status==TOK_OK) {
    if (key_is("around")) {
      key=free_Token(key);
      key=Get_Atoms_D("box around atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	if (c->n>0) {
	  default_token("0.5");
	  key=Get_Float_M("resolution: ",";",resm);
	  if (key->status==TOK_OK) {
	    res=key->f;
	    key=free_Token(key);
	    default_token("rvdw");
	    key=Get_Float_M("margin: ",";",margmv);
	    mrgvdw=FALSE;
	    if (key->status==TOK_NOK) {
	      key=Prs_Kwd(key,margmv);
	      if (key->status==TOK_OK) {
		if (key_is("rvdw")) {
		  mrgvdw=TRUE;
		  key->f=0.0;
		}else key->status=TOK_NOK;
	      }
	    }
	    if (key->status==TOK_OK) {
	      a=(Atom *)c->s;
	      xmrg=(mrgvdw&&a->atp?a->atp->rvdw:key->f);
	      for (i=0;i<3;i++) {
		xmin[i]=(a->nxyz>0 ? a->x[i] : 0.0)-xmrg;
		xmax[i]=xmin[i]+2*xmrg;
	      }
	      for (a=(Atom *)a->chain;mvld(a);a=(Atom *)a->chain) {
		xmrg=(mrgvdw&&a->atp?a->atp->rvdw:key->f);
		for (i=0;i<3;i++) {
		  if (a->nxyz>0) {
		    if (xmin[i]>a->x[i]-xmrg) xmin[i]=a->x[i]-xmrg;
		    if (xmax[i]<a->x[i]+xmrg) xmax[i]=a->x[i]+xmrg;
		  }
		}
	      }
	      if (mrgvdw) {
		for (i=0;i<3;i++) {
		  xmin[i]-=0.5*res;
		  xmax[i]+=0.5*res;
		}
	      }
	    }else if (key->status==TOK_NOK) err_set(1,"bad margin: %s",key->text);;
	  }else if (key->status==TOK_NOK) err_set(1,"bad resolution: %s",key->text);
	}else err_set(1,"zero atoms selected.");
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_OK) err_set(1,"bad atom selection: %s",key->text);
    }else if (key_is("xyz")) {
      for (i=0;i<3;i++) {
	key=free_Token(key);
	default_token("0.0");
	key=Get_Float(prx[i],";");
	if (key->status==TOK_OK) xmin[i]=key->f;
	else break;
      }
      if (key->status==TOK_OK) {
	default_token("5.0");
	for (i=0;i<3;i++) {
	  key=free_Token(key);
	  key=Get_Float_M(prm[i],";",margm);
	  if (key->status==TOK_OK) xmax[i]=key->f;
	  else break;
	  if (i<2) default_token(key->text);
	}
	if (key->status==TOK_OK) {
	  for (i=0;i<3;i++) {
	    xmin[i]=(xmax[i]<0 ? xmin[i]+xmax[i] : xmin[i]-xmax[i]);
	    xmax[i]=(xmax[i]<0 ? xmin[i]-2*xmax[i] : xmin[i]+2*xmax[i]);
	  }
	  key=free_Token(key);
	  default_token("0.5");
	  key=Get_Float_M("resolution: ",";",resm);
	  if (key->status==TOK_OK) res=key->f;
	  else if (key->status==TOK_NOK) err_set(1,"bad resolution: %s",key->text);
	}else err_set(1,"bad extent: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad coordinate: %s",key->text);
    }
    if (key->status==TOK_OK) {
      if (res<0.05) res=0.05;
      G=cond_del_grid(G);
      if (!mvld(G)) {
	G=emptyGrid();
	if (mvld(G)) {
	  G->name=strdup("(no file)");
	  for (i=0;i<3;i++) {
	    G->dx[i]=res;
	    G->n[i]=(xmax[i]-xmin[i])/res+1;
	    if (G->n[i]<2) G->n[i]=2;
	    if ((G->n[i]-1)*res<xmax[i]-xmin[i]) G->n[i]++;
	    G->xmin[i]=0.5*((xmax[i]+xmin[i])-(G->n[i]-1)*res);
	    G->xmax[i]=G->xmin[i]+(G->n[i]-1)*res;
	  }
	  G->val=(float *)calloc(G->n[0]*G->n[1]*G->n[2],sizeof(float));
	  if (!G->val) {
	    err_set(1,"sorry, failed to allocate space for grid data.");
	    G=freeGrid(G);
	  }else grid_report(G);
	}else err_set(1,"failed to allocate new grid.");
      }else prtGraphics("kept old grid.");
      
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad grid center option: %s",key->text);
  key=free_Token(key);
  return(G);
}
/*-------------------- grid_compute --------------------*/

int grid_compute(Grid *G)
{
  int r;
  Token *key;
  char *fnmenu="'t!grid function:' 't! ' contact Coulomb EDFDist SaSDist vdwdist *const truncate zero h!dotdist h!oldSaS";

  key=(Token *)NULL;
  r=0;
  if (mvld(G)) {
    key=Get_Kwd("grid function: ",";",fnmenu);
    if (key->status==TOK_OK) {
      if (key_is("vdwdist")) r=grid_FN_vdwdist(G);
      else if (key_is("SaSDist")) r=grid_FN_ConnollyDist(G);
      else if (key_is("EDFDist")) r=grid_FN_EDFDist(G);
      else if (key_is("dotdist")) r=grid_FN_dotdist(G);
      else if (key_is("oldSaS")) r=grid_FN_SaSDist(G);
      else if (key_is("contact")) r=grid_FN_Contact(G);
      else if (key_is("truncate")) r=grid_FN_truncate(G);
      else if (key_is("*const")) r=grid_FN_mult(G);
      else if (key_is("zero")) r=grid_fn_zero(G);
      else if (key_is("Coulomb")) r=grid_FN_Coulomb(G);
    }else if (key->status==TOK_NOK) err_set(1,"not a valid grid function: %s",key->text);
  }else err_set(1,"must create grid, before you can compute.");
  key=free_Token(key);
  return(r);
}
/*-------------------- grid_report --------------------*/

void grid_report(Grid *G)
{
  int i;

  if (mvld(G)) {
    prtGraphics("Grid read from file %s\n",
		(G->name ? G->name : " (unknown)"));
    prtGraphics("nx = %d, ny = %d, nz = %d\n",G->n[0],G->n[1],G->n[2]);
    if (G->dx[0]==G->dx[1]&&G->dx[1]==G->dx[2]) {
      prtGraphics("resolution = %0.3f\n",G->dx[0]);
    }else{
      prtGraphics("resolution = (%0.3f,%0.3f,%0.3f)\n",G->dx[0],G->dx[1],G->dx[2]);
    }
    prtGraphics("total of %d gridpoints.\n",G->n[0]*G->n[1]*G->n[2]);
    for (i=0;i<3;i++) if (G->cl[i]!=1.0||G->ca[i]!=90.0) break;
    if (i<3) prtGraphics("Warning: box is non-cartesian.\n");
  }else{
    prtGraphics("no grid.\n");
  }
}
/*-------------------- grid_read --------------------*/

Grid *grid_read(Grid *G)
{
  G=cond_del_grid(G);
  if (!mvld(G)) {
    cmd_read(&G,FT_CNT);          /* note: cmd_read wants a Node* */
    if (!mvld(G)) err_set(1,"failed to read grid file.");
    else grid_report(G);
  }
  return(G);
}
/*-------------------- grid_write --------------------*/

void grid_write(Grid *G)
{
  Token *key;
  FILE *out;
  char buff[4096];

  key=(Token *)NULL;

  if (mvld(G)) {
    key=Get_String("write grid to file: ",";");
    if (key->status==TOK_OK) {
      out=fopen(extfil(buff,key->text,".cnt",FALSE),"w");
      if (out) {
	if (write_grid(out,G)) err_set(1,"error writing grid to %s",buff);
	if (fclose(out)==EOF) err_set(1,"error closing file %s",buff);
	else prtGraphics("grid written to %s\n",buff);
      }else err_set(1,"error opening file %s for write.",buff);
    }else if (key->status==TOK_NOK) err_set(1,"bad filename: %s",key->text);
  }else err_set(1,"grid has not been defined.");
  key=free_Token(key);
}

/*-------------------- grid_delete --------------------*/

Grid *grid_delete(Grid *G)
{
  if (mvld(G)) {
    G=freeGrid(G);
    prtGraphics("old grid deleted.\n");
  }else err_set(1,"no grid.");
  return(G);
}

/*-------------------- cond_del_grid --------------------*/

Grid *cond_del_grid(Grid *G)
{
  Token *key;

  key=(Token *)NULL;

  if (mvld(G)) {
    key=Get_Kwd("delete old grid ? ",";","no yes");
    if (key->status==TOK_OK) {
      if (key_is("yes")) G=freeGrid(G);
    }else if (key->status==TOK_NOK) err_set(1,"sorry, %s ?",key->text);
  }
  key=free_Token(key);
  return(G);
}
/*-------------------- grid_FN_vdwdist --------------------*/

int grid_FN_vdwdist(Grid *G)
{
  Token *key=(Token *)(NULL);
  Nodechain *c=(Nodechain *)NULL;
  int r=0;

  if (mvld(G)&&mvld(G->val)) {
    key=Get_Atoms_D("compute vdw distance from atoms: ",";");
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      r=grid_fn_vdwdist(G,c);
      clear_node_chain(c);
      free_node_chain(c); 
    }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    key=free_Token(key);
  }
  return(r);
}
/*-------------------- grid_FN_Coulomb --------------------*/

int grid_FN_Coulomb(Grid *G)
{
  Token *key=(Token *)(NULL);
  Nodechain *c=(Nodechain *)NULL;
  int r=0,diel;
  float eps;

  if (mvld(G)&&mvld(G->val)) {
    key=Get_Kwd("dielectric: ",";","constant distance_dependent");
    if (key->status==TOK_OK) {
      diel=key->n;
      key=free_Token(key);
      key=Get_Float_M("factor: ",";","0.5 1.0 2.0 3.0 4.0 5.0 10.0 80.0");
      if (key->status==TOK_OK) {
	eps=(key->f>0?key->f:1.0);
	key=Get_Atoms_D("use charges from atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  r=grid_fn_Coulomb(G,c,diel,eps);
	  clear_node_chain(c);
	  free_node_chain(c); 
	}else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad vaule for factor: %s",key->text);
    }else if (key->status==TOK_NOK) err_set(1,"illegal choice of dielectric: %s",key->text);
    key=free_Token(key);
  }
  return(r);
}
/*-------------------- grid_FN_mult --------------------*/

int grid_FN_mult(Grid *G)
{
  int r=0;
  Token *key;
  char *mena="-100 -10 -5 -1 -0.5 -0.1 ' ' 0 ' ' 0.1 0.5 5 10 100";

  if (mvld(G)&&mvld(G->val)) {
    key=(Token *)NULL;
    key=Get_Float_M("scale factor for grid values: ",";",mena);
    if (key->status==TOK_OK) r=grid_fn_mult(G,key->f);
    else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
    key=free_Token(key);
  }
  return(r);
}
/*-------------------- grid_FN_truncate --------------------*/

int grid_FN_truncate(Grid *G)
{
  int r=0;
  Token *key;
  char *mena="-1000 -100 -50 -10 -5 -4 -3 -2 -1 "
             " ' ' 0 ' ' 1 2 3 4 5 10 10 50 100 1000";
  float a=1.e30,b=-1.e30,x;

  if (mvld(G)&&mvld(G->val)) {
    key=(Token *)NULL;
    key=Get_Kwd("truncate: ",";","above below between");
    if (key->status==TOK_OK) {
      if (key_is("above")) {
	key=free_Token(key);
	key=Get_Float_M("truncate above> ",";",mena);
	if (key->status==TOK_OK) a=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
      }else if (key_is("below")) {
	if (key_is("below")) {
	  key=free_Token(key);
	  key=Get_Float_M("truncate below> ",";",mena);
	  if (key->status==TOK_OK) b=key->f;
	}else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
      }else if (key_is("between")) {
	key=free_Token(key);
	key=Get_Float_M("truncate between (lower limit)> ",";",mena);
	if (key->status==TOK_OK) {
	  b=key->f;
	  if (key_is("above")) {
	    key=free_Token(key);
	    key=Get_Float_M("and truncate above (upper limit)> ",";",mena);
	    if (key->status==TOK_OK) a=key->f;
	    else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
      }
    }else if (key->status==TOK_NOK) err_set(1,"unknown truncation option: %s",key->text);
    if (key->status==TOK_OK) {
      if (a<b) {
	x=a;
	a=b;
	b=x;
      }
      r=grid_fn_truncate(G,b,a);
    }
    key=free_Token(key);
  }
  return(r);
}
/*-------------------- grid_FN_dotdist --------------------*/

int grid_FN_dotdist(Grid *G)
{
  int s=0;
  Token *key;
  UCpipe *in;
  char rec[256],fld[20];
  float **dot=(float **)NULL,**new;
  int mxdot=0,ndot=0,i;
  
  key=Get_File("",".ms","MS dot file: ",";");
  if (key->status==TOK_OK) {
    in=key->ptr;
    while(fgets(rec,sizeof(rec),in->fd)) {
      if (strlen(rec)>74) {
	if (ndot==mxdot) {
	  new=realloc(dot,(mxdot+32768)*sizeof(float *));
	  if (new) {
	    mxdot+=32768;
	    dot=new;
	  }else{
	    err_set(1,"re-allocation to %d dots failed\n",mxdot);
	    s=1;
	    break;
	  }
	}
	dot[ndot]=malloc(6*sizeof(float));
	if (dot[ndot]) {
	  for (i=0;i<3;i++) dot[ndot][i]=str2f(strfld(fld,rec+17+i*9,1,9));
	  for (i=0;i<3;i++) dot[ndot][i+3]=str2f(strfld(fld,rec+51+i*7,1,7));
	  ndot++;
	}else{
	  s=1;
	  err_set(1,"allocation of dotfailed\n");
	  break;
	}
      }
    }
    if (!s) s=grid_fn_dotdist(G,ndot,dot);
  }else if (key->status==TOK_NOK) err_set(1,"bad file specicifvction: %s\n",key->text);
  free_Token(key);
  if (dot) {
    for (i=0;i<ndot;i++) free(dot[ndot]);
    free(dot);
  }
  return(s);
}
/*-------------------- grid_FN_ConnollyDist --------------------*/

int grid_FN_ConnollyDist(Grid *G)
{
  int s=0;
  Token *key;
  Nodechain *c;
  float r;
  char *menup="0.5 1.0 1.1 1.2 1.3 1.4 1.5 1.6 1.7 1.8 1.9 2.0 2.5 3.0";

  /* approx. distance from solvent accessible surface */

  if (mvld(G)&&mvld(G->val)) {
    c=(Nodechain *)NULL;
    key=(Token *)NULL;
  
    default_token("1.4");
    key=Get_Float_M("probe radius: ",";",menup);
    if (key->status==TOK_OK) {
      r=key->f;
      if (r<0) r=0;
      key=free_Token(key);
      key=Get_Atoms_D("compute SaS distance from atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	s=grid_fn_ConnollyDist(G,c,r);
	clear_node_chain(c);
	free_node_chain(c); 
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    }else err_set(1,"bad probe radius: %s",key->text);
    key=free_Token(key);
  }
  return(s);
}

/*-------------------- grid_FN_SaSDist --------------------*/

int grid_FN_SaSDist(Grid *G)
{
  int s=0;
  Token *key;
  Nodechain *c;
  float r;
  char *menup="0.5 1.0 1.1 1.2 1.3 1.4 1.5 1.6 1.7 1.8 1.9 2.0 2.5 3.0";

  /* OBSOLETE (use grid_FN_ConnollyDist instead) */
  /* approx. distance from solvent accessible surface */

  if (mvld(G)&&mvld(G->val)) {
    c=(Nodechain *)NULL;
    key=(Token *)NULL;
  
    default_token("1.4");
    key=Get_Float_M("probe radius: ",";",menup);
    if (key->status==TOK_OK) {
      r=key->f;
      if (r<0) r=0;
      key=free_Token(key);
      key=Get_Atoms_D("compute SaS distance from atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	s=grid_fn_SaSDist(G,c,r);
	clear_node_chain(c);
	free_node_chain(c); 
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    }else err_set(1,"bad probe radius: %s",key->text);
    key=free_Token(key);
  }
  return(s);
}
/*-------------------- grid_FN_EDFDist --------------------*/

int grid_FN_EDFDist(Grid *G)
{
  /* "Analytically defined surfaces to analyze molecular interaction properties"
     R.R. Gabdoulline, R.C. Wade, J. Mol. Graphics 14:341-353, 1996 
  */

  int s=0;
  Token *key;
  Nodechain *c;
  float d;
  char *menup="0.05 0.1 0.2 0.25 0.5 0.75 1.0";

  if (mvld(G)&&mvld(G->val)) {
    default_token("0.5");
    key=Get_Float_M("d: ",";",menup);
    if (key->status==TOK_OK) {
      d=key->f;
      if (d<0.001) d=0.001;
      key=free_Token(key);
      key=Get_Atoms_D("compute contact with atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	s=grid_fn_EDFDist(G,c,d);
	clear_node_chain(c);
	free_node_chain(c); 
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    }else err_set(1,"d value: %s",key->text);
    key=free_Token(key);
  }
  return(s);
}

/*-------------------- grid_FN_Contact --------------------*/

int grid_FN_Contact(Grid *G)
{
  int s=0;
  Token *key=(Token *)NULL;
  Nodechain *c=(Nodechain *)NULL;
  float r;
  char *menup="0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 5.5 6.0 6.5 7.0 7.5 8.0 8.5 9.0 9.5 10.0";

  /* Number of atoms within given distance... */

  if (mvld(G)&&mvld(G->val)) {
    default_token("5.0");
    key=Get_Float_M("cutoff: ",";",menup);
    if (key->status==TOK_OK) {
      r=key->f;
      if (r<0) r=0;
      key=free_Token(key);
      key=Get_Atoms_D("compute contact with atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	s=grid_fn_Contact(G,c,r);
	clear_node_chain(c);
	free_node_chain(c); 
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
    }else err_set(1,"bad cutoff value: %s",key->text);
    key=free_Token(key);
  }
  return(s);
}
