#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "wnpconst.h"

extern int graphix;
extern Node *DefaultNode;

#define key_is(a) (streq(a,key->key))

#define TRUE 1
#define FALSE 0

static int auto_label_flag=0;      /* auto-labelling flag (used e.g. by modify) */

static char *labkwd="bond ' ' aromaticity biso charge fullname geometry group "
                    "ligance locations molecule name oopflag q residue seqnr shift sof switch symbol text type "
                    "valence ' ' on off auto color remove size ";

static char *char_size_menu = { "0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5" };

extern Mol *MolRoot;

void cmd_label(void)
{
  int dsp=0,option=0,color=0,size,ng,labt;
  int n;
  Token *key;
  Nodechain *c;
  Atom *a;
  Bond *b,*bn;
  
  key=Get_Kwd("option: ",";",labkwd);
  if (key->status == TOK_OK) {
    labt=ATM_LAB_NONE;
    if (key_is("bond")) {
      option=0;
      free_Token(key);
      key=Get_Kwd("bond label option: ",";","pi-order order type remove");
      if (key->status==TOK_OK) {
	if (key_is("order")) labt=BND_LAB_ORDER;
	else if (key_is("type")) labt=BND_LAB_TYPE;
	else if (key_is("remove")) labt=BND_LAB_NONE;
	else if (key_is("pi-order")) labt=BND_LAB_PI;
	free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  for (b=(Bond *)key->ptr;mvld(b);b=bn) {
	    bn=b->chain;
	    set_bond_label(b,labt,NULL);
	    b->chain=(Bond *)NULL;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }
    }else if (key_is("on")) {
      option=1;
      dsp=TRUE;
    }else if (key_is("off")) {
      option=1;
      dsp=FALSE;
    }else if (key_is("size")) {
      option=0;
      free_Token(key);
      key=Get_Float_M("character size (A): ",";",char_size_menu);
      if (key->status==TOK_OK) {
	send_bheader(graphix,"ALSZ",sizeof(float));
	send_float(graphix,key->f);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad float : %s",key->text);
      }
    }else if (key_is("auto")) {
      option=0;
      if (auto_label_flag) default_token("on");
      else default_token("off");
      free_Token(key);
      key=Get_Kwd("auto-label on/off: ",";","on off");
      if (key->status==TOK_OK) auto_label_flag=key_is("on");
      else if (key->status==TOK_NOK) {
	err_set(1,"bad label-auto token: %s",key->text);
      }
    }else if (key_is("color")) {
      key=free_Token(key);
      option=0;
      key=Get_Color_M("color: ",";","by_atomcolor ' '");
      if (key->status == TOK_OK) {
	if (key_is("by_atomcolor")) color=WNP_COLOR_BYATOM;
	else color=key->n;
	option=2;	
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown color: %s",key->text);
      }
    }else if (key_is("remove")) {
      option=3;
    }else if (key_is("name")) {
      option=4;
      labt=ATM_LAB_NAME;
    }else if (key_is("residue")) {
      option=4;
      labt=ATM_LAB_RESIDUE;
    }else if (key_is("type")) {
      option=4;
      labt=ATM_LAB_TYPE;
    }else if (key_is("q")) {
      option=4;
      labt=ATM_LAB_Q;
    }else if (key_is("geometry")) {
      option=4;
      labt=ATM_LAB_GEOMETRY;
    }else if (key_is("symbol")) {
      option=4;
      labt=ATM_LAB_SYMBOL;
    }else if (key_is("valence")) {
      option=4;
      labt=ATM_LAB_VALENCE;
    }else if (key_is("ligance")) {
      option=4;
      labt=ATM_LAB_LIGANCE;
    }else if (key_is("charge")) {
      option=4;
      labt=ATM_LAB_CHARGE;
    }else if (key_is("aromaticity")) {
      option=4;
      labt=ATM_LAB_AROMATICITY;
    }else if (key_is("group")) {
      option=4;
      labt=ATM_LAB_GROUP;
    }else if (key_is("switch")) {
      option=4;
      labt=ATM_LAB_SWITCH;
    }else if (key_is("oopflag")) {
      option=4;
      labt=ATM_LAB_OOPFLAG;
    }else if (key_is("text")) {
      option=5;
      labt=ATM_LAB_TEXT;
    }else if (key_is("biso")) {
      option=4;
      labt=ATM_LAB_BISO;
    }else if (key_is("sof")) {
      option=4;
      labt=ATM_LAB_SOF;
    }else if (key_is("seqnr")) {
      option=6;
      labt=ATM_LAB_SEQNR;
    }else if (key_is("shift")) {
      option=4;
      labt=ATM_LAB_SHIFT;
    }else if (key_is("fullname")) {
      option=4;
      labt=ATM_LAB_FULLNAME;
    }else if (key_is("molecule")) {
      option=4;
      labt=ATM_LAB_MOLECULE;
    }else if (key_is("locations")) {
      option=4;
      labt=ATM_LAB_LOCATIONS;
    }
    
    if (option) {
      free_Token(key);
      key=Get_Atoms_D("Atoms: ",";");
      if (key->status == TOK_OK) {
	c=(Nodechain *)key->ptr;
	if (c->n<=0) option=0;
	for (ng=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) if (cPtr(a->cross)) ng++;
	if (ng==0) option=0;
	switch (option) {
	case 1 :                                         /* status on/off */
	  size=sizeof(int)+ng*cPtrSize();
	  send_bheader(graphix,"ALST",size);
	  send_bint(graphix,dsp);
	  for(a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	    a->ldsp=dsp;
	    if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
	  }
	  break;
	case 2:                                          /* color */
	  size=sizeof(int)+ng*cPtrSize();
	  send_bheader(graphix,"ALCL",size);
	  send_bint(graphix,color);
	  for(a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	    a->lcolor=color;
	    if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
	  }
	  break;
	case 3:                                          /* remove */
	  size=ng*cPtrSize();
	  send_bheader(graphix,"ALRM",size);
	  for(a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	    a->ltype=ATM_LAB_NONE;
	    if (a->label) free(a->label);
	    a->label=NULL;
	    if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
	  }
	  break;
	case 4:                                          /* atom property */
	  for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,labt,NULL,FALSE);
	  break;
	case 5:                                          /* text */
	  key=free_Token(key);
	  key=Get_String("label text: ",";");
	  if (key->status==TOK_OK) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,labt,key->text,FALSE);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad label text token %s",key->text);
	  }
	  break;
	case 6:                                          /* seqnr */
	  for (n=1,a=(Atom *)c->s;a;a=(Atom *)a->chain,n++) {
	    a->seqnr=n;
	    set_atom_label(a,labt,NULL,FALSE);
	  }
	  break;
	}
	trflush(graphix);
	clear_node_chain(c);
	free_node_chain(c);
      }
    }
  }
  send_header(graphix,"ARMK",0);
  free_Token(key);
}
/*-------------------- auto_label --------------------*/

int auto_label(void)
{
  return(auto_label_flag);
}
/*-------------------- set_auto_label --------------------*/

int set_auto_label(int new)
{
  int old;
  
  old=auto_label_flag;
  auto_label_flag=new;
  return(old);
}
/*-------------------- set_atom_label --------------------*/

void set_atom_label(Atom *a,int type,char *text,int forceon)
{
  char buff[WNP_MAXREC];
  Node *up;

  *buff=0;
  
  switch (type) {
  case ATM_LAB_NONE:
    break;
  case ATM_LAB_NAME:
    if (a->name) sprintf(buff,"%s",a->name);
    break;
  case ATM_LAB_RESIDUE:
    if (a->up&&a->up->name) sprintf(buff,"%s",a->up->name);
    break;
  case ATM_LAB_TYPE:
    if (a->atp) sprintf(buff,"%s",a->atp->name); 
    break;
  case ATM_LAB_Q:
    sprintf(buff,"%6.2f",a->q);strsqz(buff,buff);
    break;
  case ATM_LAB_GEOMETRY:
    if (a->atp) sprintf(buff,"%s",a->atp->geom); 
    break;
  case ATM_LAB_SYMBOL:
    if (a->atp) sprintf(buff,"%s",a->atp->chs); 
    break;
  case ATM_LAB_VALENCE:
    if (a->val!=0) sprintf(buff,"%d",(int)a->val);
    break;
  case ATM_LAB_LIGANCE:
    if (a->lig<0||a->lig==63) sprintf(buff,"??");
    else if (a->lig>0) sprintf(buff,"%d",(int)a->lig);
    break;
  case ATM_LAB_CHARGE:
    if (a->chrg!=0) {
      if (a->chrg==63) sprintf(buff,"%s","?");
      else sprintf(buff,"%+d",a->chrg);
    }
    break;
  case ATM_LAB_AROMATICITY:
    if (a->arom) sprintf(buff,"*");
    break;
  case ATM_LAB_GROUP:
    sprintf(buff,"%s%s",a->cosw?"*":"",a->grp);
    break;
  case ATM_LAB_SWITCH:
    if (a->cosw) sprintf(buff,"*%s",a->grp);
    break;
  case ATM_LAB_OOPFLAG:
    if (a->oop) sprintf(buff,"%s",a->oop==1?"1":a->oop==2?"2":"?");
    break;
  case ATM_LAB_TEXT:
    if (text) sprintf(buff,"%s",text);
    break;
  case ATM_LAB_BISO:
    sprintf(buff,"%6.1f",a->x[ATM_BISO]);strsqz(buff,buff);
    break;
  case ATM_LAB_SOF:
    sprintf(buff,"%6.2f",a->x[ATM_SOF]);strsqz(buff,buff);
    break;
  case ATM_LAB_SEQNR:
    sprintf(buff,"%d",a->seqnr);
    break;
  case ATM_LAB_SHIFT:
    if (a->shift>WNP_NO_SHIFT) {
      if (a->stass==ATM_STASS_NO)         sprintf(buff,"%0.2f",a->shift);
      else if (a->stass==ATM_STASS_YES)   sprintf(buff,"%0.2f",a->shift);
      else if (a->stass==ATM_STASS_GUESS) sprintf(buff,"%0.2f*",a->shift);
      else if (a->stass==ATM_STASS_UNK)   sprintf(buff,"%0.2f?",a->shift);
    }
    break;
  case ATM_LAB_FULLNAME:
    nodeptr_to_name(buff,(Node *)a);
    break;
  case ATM_LAB_MOLECULE:
    for (up=(Node *)a;up;up=up->up) if (up->up==(Node *)MolRoot) break;
    if (up) sprintf(buff,"%s",up->name);
    break;
  case ATM_LAB_LOCATIONS:
    if (a->nxyz>1 || (a->alt && a->alt[0]!=' ')) {
      strncpy(buff,a->alt,a->nxyz);
      buff[a->nxyz]='\0';
    }
    break;
  default:
    break;
  }

  if (a->label) free(a->label);
  a->label=NULL;

  if (strlen(buff)>0) {
    a->ltype=type;
    if (forceon) a->ldsp=TRUE;
    if (a->ltype==ATM_LAB_TEXT||a->ltype==ATM_LAB_SEQNR) a->label=strdup(buff);
    if (cPtr(a->cross)) {
      send_bheader(graphix,forceon?"ALTN":"ALTX",cPtrSize()+strlen(buff));
      send_bcptr(graphix,a->cross);
      send_bdata(graphix,strlen(buff),buff);
    }
  }else{
    if (a->type!=ATM_LAB_NONE) {
      if (cPtr(a->cross)) {
	send_bheader(graphix,"ALRM",cPtrSize());
	send_bcptr(graphix,a->cross);
      }
      a->ltype=ATM_LAB_NONE;
    }
  }
}

void set_bond_label(Bond *b,int type,char *text)
{
  char buff[256],*s;
  float p;
  int l;

  if (b) {
    *buff='\0';
    switch (type) {
    case BND_LAB_TYPE:
      b->labt=BND_LAB_TYPE;
      if (b->bt!=1) sprintf(buff,"%s",BT_mnem(b->bt));
      break;
    case BND_LAB_ORDER:
      b->labt=BND_LAB_ORDER;
      if (b->bo==63) sprintf(buff,"%s","?");
      else if (b->bo!=1) sprintf(buff,"%d",b->bo);
      break;
    case BND_LAB_PI:
      b->labt=BND_LAB_PI;
      p=est_pi_bondorder(b);
      if (p>0.1) sprintf(buff,"%0.2f",p);
      break;
    case BND_LAB_TEXT:
      b->labt=BND_LAB_TEXT;
      if (text) sprintf(buff,"%s",text);
      if (strchr(buff,'"')&&strchr(buff,'\'')) {
	for (s=buff;*s;s++) if (*s=='\'') *s='`';
      }
      break;
    case BND_LAB_NONE:
    default:
      b->labt=BND_LAB_NONE;
      break;
    }

    if (b->label) free(b->label);
    b->label=NULL;

    if (cPtr(b->cross)) {
      if (*buff) {
	l=strlen(buff);
	send_bheader(graphix,"BLTN",cPtrSize()+l);
	send_bcptr(graphix,b->cross);
	send_bdata(graphix,l,buff);
	if (b->labt==BND_LAB_TEXT) b->label=Strdup(buff);
      }else{
	b->labt=BND_LAB_NONE;
	send_bheader(graphix,"BLRM",cPtrSize());
	send_bcptr(graphix,b->cross);
      }
    }
  }
}
