#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *MolRoot;

#define TRUE 1
#define FALSE 0
#define key_is(a) (streq(a,key->key))

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static void list_objects(char *objects);
static void list_atoms(Nodechain *c);
static void list_bonds(Nodechain *c);
static void list_residues(Nodechain *c);

void cmd_list(void)
{
  Token *key=(Token *)NULL;
  char *objtypes = "-atoms -bonds -objects -residues";
  
  key=Get_Kwd("object: ",";",objtypes);
  if (key->status == TOK_OK) {

    if (key_is("-atoms")) {
      key=free_Token(key);
      key=Get_Atoms_D("atoms: ",";");
      if (key->status==TOK_OK) {
	list_atoms(walk_order_chain(MolRoot,key->ptr,1<<1));
	clear_node_chain(key->ptr);
	free_node_chain(key->ptr);
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s\n",key->text);

    }else if (key_is("-residues")) {
      key=free_Token(key);
      key=Get_Atoms_D("atoms: ",";");
      if (key->status==TOK_OK) {
	list_residues(walk_order_chain(MolRoot,key->ptr,1<<1));
	clear_node_chain(key->ptr);
	free_node_chain(key->ptr);
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s\n",key->text);

    }else if (key_is("-objects")) {
      key=free_Token(key);
      key=Get_String("object: ",";");
      if (key->status==TOK_OK) {
	list_objects(key->text);
      }else if (key->status==TOK_NOK) err_set(1,"invalid token: %s",key->text);

    }else if (key_is("-bonds")) {
      key=free_Token(key);
      key=Get_Atoms_D("list bond between atoms: ",";");
      if (key->status==TOK_OK) {
	list_bonds(walk_order_chain(MolRoot,key->ptr,1<<1));
	clear_node_chain(key->ptr);
	free_node_chain(key->ptr);
      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s\n",key->text);
    }
  }else if (key->status==TOK_NOK) {
    list_objects(key->text);
  }
  free_Token(key);
}

/*------------------------------------------------------------------------*/

static void list_atoms(Nodechain *c)
{
  Node *a;
  char name[4096];
  
  for (a=c->s;a;a=a->chain) {
    prtGraphics("%s\n",nodeptr_to_name(name,a));
  }
}

static void list_bonds(Nodechain *c)
{
  int i;
  Atom *a,*b;
  char nm1[1024],nm2[1024];

  for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    for (i=0;i<a->nbnd;i++) {
      b=neighbour(i,a);
      if (b>a&&node_in_chain((Node *)b,c)) {
	prtGraphics("%s %s %s\n",
		    nodeptr_to_name(nm1,(Node *)a),
		    nodeptr_to_name(nm2,(Node *)b),
		    BT_mnem(a->bnd[i]->bt)
		    );
      }
    }
  }
}

static void list_residues(Nodechain *c)
{
  Node *a,*r=NULL;
  char name[4096];

  for (a=c->s;a;a=a->chain) {
    if (a->up!=r) {
      prtGraphics("%s\n",nodeptr_to_name(name,a->up));
      r=a->up;
    }
  }
}

static void list_objects(char *objects)
{
  Node *obj;
  char obj_name[256];
  Nmatch *nmatch;
  nmatch=open_nmatch(objects);
  obj=next_nmatch(nmatch);
  while (obj) {
    nodeptr_to_name(obj_name,obj);
    switch (obj->type) {
    case OBJ_DATA:
      prtGraphics("%s (data:%s)\n",obj_name,data_typename(((Data *)obj)->dtype));
      break;
    default:
      prtGraphics("%s (%s)\n",obj_name,obj_typename(obj->type));
      break;
    }
    obj=next_nmatch(nmatch);
  }
  close_nmatch(nmatch);
}
