#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "gcalc.h"
#include "bhtree.h"
#include "graphics.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define MAX_DIST_CA 4.1

static int in_use=FALSE;
static int pre=2,post=1,length=0;
static Node *start,*end;

static Node *get_pdb_entry(char *text);
static void loop_extract(void);
static void loop_define(void);
static void loop_replace(void);
static Node *get_residue(char *prmt);
static float fit_loop(Node *a,Node *b);
static int sanity_check(void);

/*======================================================================*/

void cmd_loop(void)
{
  Token *key;
  int sane;
  char *main_menu = "define extract replace ' ' "
                    "command color delete display fit freeze measure reset select "
                    "' ' end";

  if (in_use) {
    err_set(1,"sorry, Loop may not be called recursively.");
    return;
  }
  in_use=TRUE;
  key=(Token *)NULL;

  for(sane=FALSE;;) {
    if (!sane) {length=0; start=(Node *)NULL; end=(Node *)NULL;sane=TRUE;}
    err_dump();start_parse();key=free_Token(key);
    key=Get_Kwd("Loop> ",";",main_menu);
    if (key->status==TOK_OK) {
      if (key_is("extract"))       loop_extract();
      else if (key_is("define"))   loop_define();
      else if (key_is("replace"))  loop_replace();
      else if (key_is("end"))      break;
      else if (key_is("color"))    cmd_color();
      else if (key_is("display"))  cmd_display();
      else if (key_is("measure"))  cmd_measure();
      else if (key_is("fit"))      cmd_fit("fit");
      else if (key_is("select"))   cmd_select();
      else if (key_is("reset"))    cmd_display_reset();
      else if (key_is("freeze"))   cmd_freeze();
      else if (key_is("delete")) { cmd_delete();sane=sanity_check();}
      else if (key_is("command")){ command_loop("Wit!P(Loop)> ");sane=sanity_check();}
      if (end_parse()!=TOK_OK) {
	err_set(0,"extra characters at end of command ignored.");
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad Loop command: %s",key->text);
    }
    end_parse();
  }
  free_Token(key);
  in_use=FALSE;
}

/*-------------------- loop_extract --------------------*/

static void loop_extract(void)
{
  Token *key;
  Node *n,*m,*r,*t;
  int l;
  Dspobj *global,*droot;
  char *c;
  float rmsd;

  n=(Node *)NULL;
  key=(Token *)NULL;
  if (!mvld(start)) loop_define();
  if (mvld(start)) {
    key=Get_String("PDB residue code: ",";");
    c=strchr(key->text+1,'/');
    if (c) c++;
    else key->status=TOK_NOK;
    if (key->status==TOK_OK) {
      n=get_pdb_entry(key->text);
      if (mvld(n)) {
	r=node_in_root(c,n,strcasecmp);
	if (is_mol(r)) {
	  m=r->up;
	  if (mvld(m->up)) {
	    unlink_node(m);
	    set_node_name(m,n->name);
	    del_node(n);
	  }
	  for (n=r->up->fst,m=n;mvld(n);n=m) {
	    m=m->nxt;
	    if (!is_mol(n)) del_node(n);
	  }
	  n=r->up;
	  while(mvld(r->prv)) del_node(r->prv);
	  for (l=0,t=r;l<pre+post+length-1&&mvld(t->nxt);l++) t=t->nxt;
	  while(mvld(t->nxt)) del_node(t->nxt);
	  if (mvld(t)) {
	    for (l=0;l<pre-1&&mvld(r);l++)  r=r->nxt;
	    for (l=0;l<post-1&&mvld(t);l++) t=t->prv;
	    rmsd=fit_loop(r,t);
	    prtGraphics("%s, rmsd = %0.3f\n",key->text,rmsd);
	  }else{
	    err_set(1,"loop %s is too short, not fitted.",key->text);
	  }
	  n->up=MolRoot;
	  link_node(n);
	  droot=glblobj(assocobj(MolRoot));
	  global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	  if (mvld(global)) {
	    send_mdl(global,(Mol *)n);
	    ((Mol *)n)->dsptr=global;
	    global->root=n;
	  }	  
	}else{
	  err_set(1,"could not find residue %s",key->text);
	}
      }else err_set(1,"could not retriveve %s from database.\n");
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad PDB residue code: %s\n",key->text);
    }
  }
  free_Token(key);
}
/*-------------------- get_pdb_entry --------------------*/

static Node *get_pdb_entry(char *text)
{
  char filename[512],code[512],*c,*dir;
  Node *n,*m,*p;
  Xcn *xcn;
  int status,l;

  dir=getenv("WNP_BROOKH");
  if (!dir) dir="/brookh";
  if (*text!='/') sprintf(filename,"%s/%s",dir,text);
  else sprintf(filename,"%s%s",dir,text);
  l=strlen(dir);if (filename[l]=='/') l++;
  sprintf(code,"%s",filename+l);
  for (c=code;*c;c++) *c=(*c=='/'?'_':*c);
  if (*strchr(filename+l,'/')) *strchr(filename+l,'/')=0;
  n=make_node(OBJ_MOL);
  if (mvld(n)) {
    set_node_name(n,code);
    status=read_pdb((UCpipe *)NULL,filename,n,FT_PDB);
    if (status) {
      err_set(1,"could not read %s",filename);
      del_node(n);
      n=(Node *)NULL;
    }else{
      xcn=first_xcn(n);          /* remove extra level for cryst. entry */
      if (mvld(xcn)) {
	del_node((Node *)xcn);
	m=(Node *)first_mol(n);
	if (mvld(m)) {
	  unlink_node(m);
	  while(mvld(m->fst)) {
	    p=m->fst;
	    unlink_node(p);
	    p->up=n;
	    link_node(p);
	  }
	  del_node(m);
	}
      }
    }
  }else{
    err_set(1,"failed to add mol-node.\n");
  }
  return(n);
}
/*-------------------- loop_define --------------------*/

static void loop_define(void)
{
  Token *key;
  Node *s,*e=(Node *)NULL,*t;
  int a,b;

  key=(Token *)NULL;
  s=get_residue("last pre-loop residue: ");
  if (mvld(s)) e=get_residue("first post-loop residue: ");
  if (mvld(s)&&mvld(e)) {
    if (s->up!=e->up) {
      err_set(1,"Error: pre and post residue are not in the same chain.");
    }else{
      for (t=s;mvld(t);t=t->nxt) if (t==e) break;
      if (!mvld(t)) {
	t=s;s=e;e=t;
      }
      key=free_Token(key);
      default_token("%d",length);
      key=Get_Int_M("number of loop residues: ",";","0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16");
      if (key->status==TOK_OK && key->n<0) {
	err_set(1,"error: loop length must be non-negative");
	key->status=TOK_NOK;
      }
      if (key->status==TOK_OK) {
	a=key->n;
	key=free_Token(key);
	default_token("%d",pre);
	key=Get_Int_M("number of pre-loop residues: ",";","1 2 3 4 5 6 7 8 9 10");
	if (key->status==TOK_OK) {
	  b=(key->n >0 ? key->n : 1);
	  key=free_Token(key);
	  default_token("%d",post);
	  key=Get_Int_M("number of post-loop residues: ",";","1 2 3 4 5 6 7 8 9 10");
	  if (key->status==TOK_OK) {
	    start=s;
	    end=e;
	    pre=b;
	    post=(key->n >0 ? key->n : 1);
	    length=a;
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad integer: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad integer: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad integer: %s",key->text);
      }
    }
  }
  free_Token(key);
}
/*-------------------- get_residue --------------------*/

static Node *get_residue(char *prmt)
{

  Token *key;
  Nodechain *c;
  Node *r=(Node *)NULL;

  key=(Token *)NULL;
  key=Get_Mols(prmt,";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n==1) {
      r=c->s;
    }else{
      err_set(1,"ambiguous residue selection: %s",key->text);
    }
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) {
    key=Prs_Atom(key);
    if (key->status==TOK_OK) {
      r=((Atom *)key->ptr)->up;
      if (!is_mol(r)) {
	err_set(1,"not a residue: %s",key->text);
	r=(Node *)NULL;
      }
    }else{
      err_set(1,"bad residue token: %s\n",key->text);
    }
  }
  else r=(Node *)NULL;;

  send_bheader(graphix,"ARMK",0);
  free_Token(key);
  return(r);
}
/*-------------------- fit_loop --------------------*/

static float fit_loop(Node *a,Node *b)
{
  Atom **S,**T,*s,*t;
  int nex,n=0,l;
  Node *p,*q;
  mat_4 tra;
  float rmsd;

  nex=3*(pre+post);
  S=(Atom **)malloc(nex*sizeof(Atom *));
  T=(Atom **)malloc(nex*sizeof(Atom *));
  if (mvld(S)&&mvld(T)) {
    n=0;
    for (q=start,p=a,l=0;mvld(p)&&mvld(q)&&l<pre;l++,p=p->prv) {
      forAtoms(s,p) if (strceq(s->name,"N")) break;
      forAtoms(t,q) if (strceq(t->name,"N")) break;
      if (mvld(s)&&mvld(t)) {
	S[n]=s; T[n]=t;
	n++;
      }
      forAtoms(s,p) if (strceq(s->name,"CA")) break;
      forAtoms(t,q) if (strceq(t->name,"CA")) break;
      if (mvld(s)&&mvld(t)) {
	S[n]=s; T[n]=t;
	n++;
      }
      forAtoms(s,p) if (strceq(s->name,"C")) break;
      forAtoms(t,q) if (strceq(t->name,"C")) break;
      if (mvld(s)&&mvld(t)) {
	S[n]=s; T[n]=t;
	n++;
      }
      for (q=q->prv;mvld(q);q=q->prv) if (is_mol(q)) break;
    }
    for (q=end,p=b,l=0;mvld(p)&&mvld(q)&&l<post;l++,p=p->nxt) {
      forAtoms(s,p) if (strceq(s->name,"N")) break;
      forAtoms(t,q) if (strceq(t->name,"N")) break;
      if (mvld(s)&&mvld(t)) {
	S[n]=s; T[n]=t;
	n++;
      }
      forAtoms(s,p) if (strceq(s->name,"CA")) break;
      forAtoms(t,q) if (strceq(t->name,"CA")) break;
      if (mvld(s)&&mvld(t)) {
	S[n]=s; T[n]=t;
	n++;
      }
      forAtoms(s,p) if (strceq(s->name,"C")) break;
      forAtoms(t,q) if (strceq(t->name,"C")) break;
      if (mvld(s)&&mvld(t)) {
	S[n]=s; T[n]=t;
	n++;
      }
      for (q=q->nxt;mvld(q);q=q->nxt) if (is_mol(q)) break;
    }
  }
  if (n>0) {
    rmsd=rms_best_mat(n,S,T,&tra);
    ForAtoms(s,a->up) transform_atom(tra,s);
    return(rmsd);
  }else{
    err_set(1,"loop_fit: found zero atoms.");
    rmsd=10000.0;
  }
  if (mvld(S)) free(S);
  if (mvld(T)) free(T);
  return(rmsd);
}
/*-------------------- loop_replace --------------------*/

static void loop_replace(void)
{
  Token *key;
  Nodechain L;
  Node *m,*l1=(Node *)NULL,*l2=(Node *)NULL,*l,**R,*loop;
  Atom *a,*s,*e;
  float d,d1,d2;
  int i;

  key=(Token *)NULL;

  if (!is_mol(start)) loop_define();
  if (mvld(start)) {
    if (length<1) {
      err_set(1,"sorry, loops of length < 1 can not be replaced.");
    }else{
      key=Get_Mol_D("select loop: ",";");
      if (key->status==TOK_OK) {
	m=(Node *)key->ptr;
	ForAtoms(s,start) if (strceq(s->name,"CA")) break;
	ForAtoms(e,end)   if (strceq(e->name,"CA")) break;
	if (is_atom(s)&&is_atom(e)) {
	  d1=1.e10; d2=1.e10;
	  ForAtoms(a,m) {
	    if (strceq(s->name,"CA")) {
	      d=cdist(a->x,s->x);
	      if (d<d1) {
		d1=d;
		l1=a->up;
	      }
	      d=cdist(a->x,e->x);
	      if (d<d2) {
		d2=d;
		l2=a->up;
	      }
	    }
	  }
	  if (d1<1.5 && d2<1.5) {
	    if (l1->up==l2->up) {
	      loop=l1->up;
	      init_node_chain(&L);
	      add_to_node_chain(l1,&L);
	      for (l=l1->nxt;mvld(l);l=l->nxt) {
		if (is_mol(l)) add_to_node_chain(l,&L);
		if (L.n==length+2) break;
	      }
	      if (L.e==l2&&L.n==length+2) {
		R=chain_to_ptr(&L,FALSE);
		clear_node_chain(&L);
		if (mvld(R)) {
		  R[0]=start; R[length+1]=end;
		  for (i=1;i<length+1;i++) {
		    R[i]=tree_copy(R[i]);
		    if (mvld(R[i])) {
		      R[i]->prv=R[i-1];
		      if (mvld(R[i-1])) {
			ForAtoms(s,R[i-1]) if (strceq(s->name,"C")) break;
			ForAtoms(e,R[i]) if (strceq(e->name,"N")) break;
			if (is_atom(s)&&is_atom(e)) {
			  if (cdist(s->x,e->x)<1.95) set_bt(add_bond_ng(s,e),BT_AMIDE);
			}
		      }
		    }else{
		      err_set(1,"failed to copy residues.");
		      break;
		    }
		  }
		  if (i==length+1) {
		    ForAtoms(s,R[i-1]) if (strceq(s->name,"C")) break;
		    ForAtoms(e,R[i]) if (strceq(e->name,"N")) break;
		    if (is_atom(s)&&is_atom(e)) {
		      if (cdist(s->x,e->x)<1.95) set_bt(add_bond_ng(s,e),BT_AMIDE);
		    }
		    if (start!=end) {
		      for(l=start->nxt;mvld(l)&&l!=end;l=m) {
			m=l->nxt;
			if (is_mol(l)) del_node(l);
		      }
		    }
		    init_node_chain(&L);
		    for (i=1;i<length+1;i++) {
		      link_node(R[i]);
		      ForAtoms(a,R[i]) add_to_node_chain((Node *)a,&L);
		    }
		    for (a=(Atom *)L.s;mvld(a);a=(Atom *)a->chain) {
		      for (i=a->nbnd-1;i>=0;i--) if (neighbour(i,a)->up->up==loop) del_bond(a->bnd[i]);
		    }
		    send_atom_chain(&L);
		    clear_node_chain(&L);
		  }else{
		    for (i=1;i<length+1;i++) del_node(R[i]);
		  }
		  free(R);
		}else err_set(1,"memory allocation failed,");
	      }else err_set(1,"sorry, new loop has incorrect length.");
	    }else err_set(1,"sorry, first and last residue of new loop must be in singe chain.");
	  }else err_set(1,"sorry, bad rms fit at join");
	}else err_set(1,"sorry, could not find C-alphas in loop to replace");
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad loop selection: %s",key->text);
      }
    }
  }
  free_Token(key);
}
/*-------------------- sanity_check --------------------*/

static int sanity_check(void)
{
  int r;

  if (mvld(start)&&mvld(end)) {
    if (find_node(MolRoot,start,OBJ_MOL)) {
      if (find_node(start->up,end,OBJ_MOL)) r=1;
      else r=0;
    }else r=0;
  }else r=0;
  return(r);
}
