#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *DefaultNode;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

static char *ludikwd = "Run host list radius L-density P-density "
                       "rms-max preselect ang-max ali-arom "
                       "vdw-cut es-check es-cut bifurcation min-contact "
                       "w-lipo w-hbond w-link" ;

static float radius=7.5;
static int ldens=25;
static int pdens=25;
static float rmsmax=0.5;
static float presel=2.5;
static float angmax=15.0;
static int aliarom=1;
static float vdwc=3.3;
static int esc=1;
static float escut=2.5;
static int bifur=1;
static int mincont=0;
static float wlipo=1.0;
static float whbond=1.0;
static float wlink=1.0;
static char *host=(char *)NULL;

static void ludi_run(void);

/*======================================================================*/

void cmd_ludi(void)
{
  Token *key;
  int loop;
  char *txt;
  
  loop=TRUE;
  key=(Token *)NULL;

  while (loop) {
    free_Token(key);
    key=Get_Kwd("Ludi option: ",";",ludikwd);
    if (key->status==TOK_OK) {
      if (key_is("Run")) {
	if (mvld(host)) {
	  ludi_run();
	}else{
	  err_set(1,"ludi host is not set; can't run.");
	}
      }else if (key_is("list")) {
	txt=(mvld(host) ? host : "not set");
	prtGraphics("ludi host = %s\n",txt);
	prtGraphics("radius = %5.2f\n",radius);
	prtGraphics("points per lipohilic site: %d\n",ldens);
	prtGraphics("points per H-bonding site: %d\n",pdens);
	prtGraphics("max. rmsd = %5.2f\n",rmsmax);
	prtGraphics("max. angl = %5.2f\n",angmax);
	txt=(aliarom ? "no" : "yes");
	prtGraphics("distinguish aliphatic/aromatic sites : %s\n",txt);
	prtGraphics("min. C-C distance = %5.2f\n",vdwc);
	prtGraphics("min. ES distance = %5.2f\n",escut);
	txt=(esc ? "yes" : "no");
	prtGraphics("do electrostatic check : %s\n",txt);
	txt=(bifur ? "no" : "yes");
	prtGraphics("allow bifurcated H-bonds : %s\n",txt);
	prtGraphics("min. percentage contact = %d\n",mincont);
	prtGraphics("lipo weight = %8.2f\n",wlipo);
	prtGraphics("H-bond weight = %8.2f\n",whbond);
	prtGraphics("ink weight = %8.2f\n",wlink);
      }else if (key_is("host")) {
	key=free_Token(key);
	key=Get_String_M("ludi host: ",";",
			 "opus jumbo bigred turbo hpw robin iris piris phf malcolm aw");
	if (key->status==TOK_OK) {
	  if (mvld(host)) free(host);
	  host=(char *)key->text;
	  key->text=(char *)NULL;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad ludi host: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("radius")) {
	key=free_Token(key);
	key=Get_Float_M("radius of binding site: ",";",
			"4.5 5.0 5.5 6.0 6.5 7.0 7.5 8.0 8.5 9.0 9.5 10.0");
	if (key->status==TOK_OK) {
	  radius=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad radius: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("L-density")) {
	key=free_Token(key);
	key=Get_Int_M("number of points per lipohilic site: ",";",
		      "5 10 15 20 25 30");
	if (key->status==TOK_OK) {
	  ldens=key->n;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad lipophilic site density: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("P-density")) {
	key=free_Token(key);
	key=Get_Int_M("number of points per HB site: ",";",
		      "5 10 15 20 25 30");
	if (key->status==TOK_OK) {
	  pdens=key->n;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad HB site density: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("rms-max")) {
	key=free_Token(key);
	key=Get_Float_M("max rmsd of fit to interaction sites: ",";",
			"0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0");
	if (key->status==TOK_OK) {
	  rmsmax=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad rmsd value: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("preselect")) {
	key=free_Token(key);
	key=Get_Float_M("preselection factor: ",";",
			"1.5 2.0 2.5 3.0");
	if (key->status==TOK_OK) {
	  presel=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad preselection factor: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("ang-max")) {
	key=free_Token(key);
	key=Get_Float_M("max linkage angle: ",";",
			"10.0 12.5 15.0 17.5 20.0 22.5 25.0 27.5 30.0");
	if (key->status==TOK_OK) {
	  angmax=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad angle tolerance: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("ali-arom")) {
	key=free_Token(key);
	key=Get_Kwd("distinguish between aliphatic/aromatic sites ? ",";",
		"yes no");
	if (key->status==TOK_OK) {
	  aliarom=key->n;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad response: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("vdw-cut")) {
	key=free_Token(key);
	key=Get_Float_M("min. C-C distance: ",";",
			"2.75 3.0 3.25 3.5");
	if (key->status==TOK_OK) {
	  vdwc=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad vdw tolerance: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("es-check")) {
	key=free_Token(key);
	key=Get_Kwd("do electrostatic check  ? ",";",
		"yes no");
	if (key->status==TOK_OK) {
	  esc=1-key->n;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad response: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("es-cut")) {
	key=free_Token(key);
	key=Get_Float_M("threshold for electrostatic check: ",";",
			"2.0 2.2 2.4 2.6 2.8 3.0");
	if (key->status==TOK_OK) {
	  escut=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad electrostatic threshold: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("bifurcation")) {
	key=free_Token(key);
	key=Get_Kwd("allow bifurcated H-bonds ? ",";",
		"yes no");
	if (key->status==TOK_OK) {
	  bifur=key->n;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad response: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("min-contact")) {
	key=free_Token(key);
	key=Get_Float_M("min. % of surface contact: ",";",
			"0 10 20 30 40 50 60 70 80 90 100");
	if (key->status==TOK_OK) {
	  mincont=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad surface contact value: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("w-lipo")) {
	key=free_Token(key);
	key=Get_Float_M("lipo weight: ",";",
			"0.5 1.0 1.5 2.0 2.5 3.0");
	if (key->status==TOK_OK) {
	  wlipo=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad lipo weight: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("w-hbond")) {
	key=free_Token(key);
	key=Get_Float_M("H-bond weight: ",";",
			"0.5 1.0 1.5 2.0 2.5 3.0");
	if (key->status==TOK_OK) {
	  whbond=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad H-bond weight: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("w-link")) {
	key=free_Token(key);
	key=Get_Float_M("link weight: ",";",
			"0.5 1.0 1.5 2.0 2.5 3.0");
	if (key->status==TOK_OK) {
	  wlink=key->f;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad link weight: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad Ludi option: %s",key->text);
    }else{
      loop=FALSE;
    }
    err_dump();
  }
  free_Token(key);
}
/*-------------------- ludi_run --------------------*/

static void ludi_run(void)
{
  Token *key;
  char *jobname,*menu,file[512],cmd[512],dir[512];
  FILE *inp;
  Node *receptor,*inhibitor=(Node *)NULL;
  int nr,with_inhibitor;
  Atom *linkatom=(Atom *)NULL,*a;
  vec_3 cen;
  Dspobj *drec,*dcen,*proot;
  mat_4 mrec,mcen;

  key=(Token *)NULL;
  jobname=(char *)NULL;
  
  key=free_Token(key);
  menu=Make_TopMolNameList("* /*",DefaultNode);
  key=Get_Mol_M("receptor: ",";",menu);
  if (key->status==TOK_OK) {
    receptor=(Node *)key->ptr;
  }else {
    if (key->status==TOK_NOK) err_set(1,"bad receptor molecule: %s",key->text);
    free_menu(menu);
    return;
  }

  key=free_Token(key);
  key=Get_Kwd("add fragments to inhibitor ? ",";","yes no");
  if (key->status==TOK_OK) {
    with_inhibitor=1-key->n;
  }else{
    if (key->status==TOK_NOK) err_set(1,"bad response: %s",key->text);
    free_menu(menu);
    return;
  } 

  if (with_inhibitor) {
    key=free_Token(key);
    key=Get_Mol_M("inhibitor: ",";",menu);
    if (key->status==TOK_OK) {
      inhibitor=(Node *)key->ptr;
    }else {
      if (key->status==TOK_NOK) err_set(1,"bad inhibitor molecule: %s",key->text);
      free_menu(menu);
      return;
    }

    proot=pick_root(assocmdl(inhibitor));
    key=free_Token(key);
    key=Get_Atom("linkage point: ",";");
    pick_root(proot);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      linkatom=(Atom *)key->ptr;
      ForAtoms(a,inhibitor) if (a==linkatom) break;
      if (!mvld(a)) {
	err_set(1,"link atom is not an inhibitor atom\n");
	key->status=TOK_NOK;
      }
      if (linkatom->nbnd!=1) {
	err_set(1,"link atom must have exactly one bond.\n");
	key->status=TOK_NOK;
      }
    }
    if (key->status!=TOK_OK) {
      if (key->status==TOK_NOK) err_set(1,"bad link atom: %s",key->text);
      free_menu(menu);
      return;
    }
  }

  key=free_Token(key);
  key=Get_Atom("atom at center of binding site: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    a=(Atom *)key->ptr;
    drec=glblobj(assocobj(receptor));
    dcen=glblobj(assocobj((Node *)a));
    mrec=dsp_transform(drec);
    mcen=dsp_transform(dcen);
    mcen=m4xm4(mmxinv(mrec),mcen);
    cen=m4xv3(mcen,vec3(a->x));
  }else{
    if (key->status!=TOK_OK) {
      if (key->status==TOK_NOK) err_set(1,"bad center atom: %s",key->text);
      free_menu(menu);
      return;
    }
  }

  key=Get_String("jobname: ",";");
  if (key->status==TOK_OK) {
    jobname=key->text;
    key->text=(char *)NULL;
  }else  {
    if (key->status==TOK_NOK) err_set(1,"bad jobname: %s",key->text);
    free_menu(menu); return;
  }

  sprintf(file,"%s.receptor",jobname);
  WriteMolecule(receptor,file,FT_XPDB,FALSE);

  if (with_inhibitor) {
    nr=0;
    sprintf(file,"%s.inhibitor",jobname);
    WriteMolecule(inhibitor,file,FT_XPDB,FALSE);
    ForAtoms(a,inhibitor) a->seqnr = ++nr;
  }
  
  sprintf(file,"%s.inp",jobname);
  inp=fopen(file,"w");
  if (inp) {
    fprintf(inp,"TITLE Ludi job %s, generated by Wit!P\n",jobname);
    fprintf(inp,"CUTOFF %7.2f\n",radius);
    fprintf(inp,"XSITE  %7.2f\n",cen.c[0]);
    fprintf(inp,"YSITE  %7.2f\n",cen.c[1]);
    fprintf(inp,"ZSITE  %7.2f\n",cen.c[2]);
    fprintf(inp,"RMSMAX %7.2f\n",rmsmax);
    fprintf(inp,"PRESEL %7.2f\n",presel);
    fprintf(inp,"VDWCUT %7.2f\n",vdwc);
    fprintf(inp,"ESCUT  %7.2f\n",escut);
    fprintf(inp,"ANGMAX %7.2f\n",angmax);
    fprintf(inp,"IOUT   %3d\n",0);
    fprintf(inp,"IELEC  %3d\n",(esc ? 1 : 0));
    fprintf(inp,"IDENSL %3d\n",ldens);
    fprintf(inp,"IDENSP %3d\n",pdens);
    fprintf(inp,"IFLAGV %3d\n",(aliarom ? 1 : 0));
    if (with_inhibitor) {
      fprintf(inp,"ILINK  %3d\n",1);
      fprintf(inp,"%2d %2d\n",linkatom->seqnr,neighbour(0,linkatom)->seqnr);
    }else{
      fprintf(inp,"ILINK  %3d\n",0);
    }
    fprintf(inp,"IANALG %3d\n",0);
    fprintf(inp,"IBIFUR %3d\n",(bifur ? 1 : 0));
    fprintf(inp,"ICONMI %3d\n",mincont);
    fprintf(inp,"WLINK  %7.2f\n",wlink);
    fprintf(inp,"WLIPO  %7.2f\n",wlipo);
    fprintf(inp,"WHBOND %7.2f\n",whbond);
    fclose(inp);
    getCWD(dir,sizeof(dir));
    if (with_inhibitor) {
      sprintf(cmd,"rsh %s -n ludi %s %s %s.receptor link %s.inhibitor",
	      host,jobname,dir,jobname,jobname);
    }else{
      sprintf(cmd,"rsh %s -n ludi %s %s %s.receptor all",
	      host,jobname,dir,jobname);
    }
    prtGraphics(cmd);
    startCmd(cmd,"","",FALSE,"belownormal");
    prtGraphics("ludi job %s started on %s\n",jobname,host);
  }else{
    err_set(1,"could not write file %s\n",file);
  }

  free_Token(key);
  free_menu(menu);
  free_menu(jobname);
}
    
    
