#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <math.h>

#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "fileutil.h"
#include "errhand.h"

extern Node *MolRoot;

extern int graphix;

#define TRUE 1
#define FALSE 0

#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

/*------------------------------------------------------------------------*/

void cmd_maccs(void)
{
  Token *key;
  UCpipe *fpipe;
  char cmd[4096],buf[4096];
  Node *mol,*ms;
  Atom *a;
  Dspobj *global,*droot;
  int nat,nbnd,nmol,m;

  if (!getenv("MACCSserver")) {
    err_set(1,"MACCSserver is not defined.");
    return;
  }

  key=(Token *)NULL;

  key=Get_String("compound number(s) = ",";");
  if (key->status==TOK_OK) {
    sprintf(cmd,"|%s %s",getenv("MACCSserver"),key->text);
    nmol=1;
    while (key->status==TOK_OK) {
      key=free_Token(key);
      key=Get_String_W("compound number(s) = ",";",FALSE);
      if (key->status==TOK_OK) {
	nmol++;
	sprintf(cmd+strlen(cmd)," %s",key->text);
      }
    }
    fpipe=UCpopen(cmd);
    if (!fpipe) {
      err_set(1,"read_maccs: could not open '%s' for read.\n",cmd);
    }else{
      ms=MolRoot->lst;
      for (;;) {
	mol=add_node(MolRoot,key->text,OBJ_MOL);
	if (!mol) {
	  err_set(1,"maccs: could not add molecule");
	}
	if (is_mol(mol)) {
	  ((Mol *)mol)->moltype=MOL_FILE;
	  if (read_maccs_mol(fpipe,mol)==0) {
	    nat=0; nbnd=0; ForAtoms(a,mol) {nat++; nbnd+=a->nbnd;}
	    nbnd/=2;
	    if (nat==0) {
	      del_node(mol);
	      err_set(1,"zero atoms, molecule deleted...");
	    }
	  }else{
	    del_node(mol);
	    break;
	  }
	}
      }
      UCpclose(fpipe);
      lock_evtq();
      for (m=0,mol=(mvld(ms) ? ms->nxt : MolRoot->fst);mvld(mol);mol=mol->nxt) {
      	if (!mvld(mol->nxt)||!streq(mol->name,mol->nxt->name)) {
	  if (m>0) {
	    m++;
	    sprintf(buf,"%s_%d",mol->name,m);
	    set_node_name(mol,buf);
	    m=0;
	  }
	}else{
	  m++;
	  sprintf(buf,"%s_%d",mol->name,m);
	  set_node_name(mol,buf);
	}
	prtGraphics("%s\n",mol->name);
	droot=glblobj(assocobj(MolRoot));
	global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	if (mvld(global)) {
	  prtGraphics("sending vectorlist...\n");
	  send_mdl(global,(Mol *)mol);
	  ((Mol *)mol)->dsptr=global;
	  global->root=mol;
	}
      }
      unlock_evtq();
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad compound number token: %s",key->text);
  }

  free_Token(key);
}
