#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"
#include "math.h"
#include "graphics.h"
#include "xtal.h"

extern int graphix;
extern Node *MolRoot;
extern Dspobj *DspRoot;

#define key_is(a) (streq(a,key->key))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static void match_apply(void);
static void match_check(void);
static void match_list(void);
static void match_fix3d(void);
static void apply_property(char *property,int na,Atom **A);


void cmd_match(void)
{
  Token *key=(Token *)NULL;
  char *match_keys;

  match_keys="apply check list h!fix3d";
  key=Get_Kwd("match: ",";",match_keys);
  if (key->status==TOK_OK) {
    if (key_is("check")) {
      match_check();
    }else if (key_is("apply")) {
      match_apply();
    }else if (key_is("list")) {
      match_list();
    }else if (key_is("fix3d")) {
      match_fix3d();
    }else{
      err_set(1,"match option not implemented: %s",key->key);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad match option: %s",key->text);
  }
  free_Token(key);
  return;
}

static void match_check(void)
{
  Token *key=(Token *)NULL;
  Atom **A=NULL,**B=NULL,*a;
  Nodechain *c;
  int i,j,na,nb,nm,namb,nbmb;

  key=Get_Atoms_D("source: ",";");
  if (key->status==TOK_OK) {
    c=key->ptr;
    na=c->n;
    A=(Atom **)chain_to_ptr(c,TRUE);
    key=free_Token(key);
    key=Get_Atoms_D("template: ",";");
    if (key->status==TOK_OK) {
      c=key->ptr;
      nb=c->n;
      B=(Atom **)chain_to_ptr(c,TRUE);
      nm=findSubstructure(na,A,nb,B);
      namb=0;
      for (i=0;i<na;i++) {
	if (A[i]->chain) {
	  for (j=0;j<A[i]->nbnd;j++) {
	    a=neighbour(j,A[i]);
	    if (a->chain && !find_bond((Atom *)a->chain,(Atom *)A[i]->chain)) namb++;
	  }
	}
      }
      nbmb=0;
      for (i=0;i<nb;i++) {
	if (B[i]->chain) {
	  for (j=0;j<B[i]->nbnd;j++) {
	    a=neighbour(j,B[i]);
	    if (a->chain && !find_bond((Atom *)a->chain,(Atom *)B[i]->chain)) nbmb++;
	  }
	}
      }
      prtGraphics("unmatched atoms: %d/%d %d/%d\n",na-nm,na,nb-nm,nb);
      prtGraphics("unmatched bonds: %d %d\n",namb,nbmb);
      clearSubstructureMatch(na,A);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad template selection.");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad source selection.");
  }

  free_Token(key);
}

static void match_list(void)
{
  Token *key=(Token *)NULL;
  Atom **A=NULL,**B=NULL;
  Nodechain *c;
  int i,na,nb,nm;
  char namA[1024],namB[1024];

  key=Get_Atoms_D("source: ",";");
  if (key->status==TOK_OK) {
    c=key->ptr;
    na=c->n;
    A=(Atom **)chain_to_ptr(c,TRUE);
    key=free_Token(key);
    key=Get_Atoms_D("template: ",";");
    if (key->status==TOK_OK) {
      c=key->ptr;
      nb=c->n;
      B=(Atom **)chain_to_ptr(c,TRUE);
      nm=findSubstructure(na,A,nb,B);
      if (nm>0) {
	for (i=0;i<na;i++) {
	  if (A[i]->chain) {
	    prtGraphics("%s --> %s\n",nodeptr_to_name(namA,A[i]->chain->chain),nodeptr_to_name(namB,A[i]->chain));
	  }
	}
	clearSubstructureMatch(na,A);
      }else{
	err_set(1,"findSubstructure: no match");
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad template selection.");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad source selection.");
  }

  free_Token(key);
}


static void match_apply(void)
{
  Token *key=(Token *)NULL;
  Atom **A=NULL,**B=NULL;
  Nodechain *c;
  int na,nb;
  char *properties;

  properties="atom bond '' aromaticity charge ligance name planarity q type xyz";

  key=Get_Atoms_D("source: ",";");
  if (key->status==TOK_OK) {
    c=key->ptr;
    na=c->n;
    A=(Atom **)chain_to_ptr(c,TRUE);
    key=free_Token(key);
    key=Get_Atoms_D("template: ",";");
    if (key->status==TOK_OK) {
      c=key->ptr;
      nb=c->n;
      B=(Atom **)chain_to_ptr(c,TRUE);
      findSubstructure(na,A,nb,B);
      while (key->status==TOK_OK) {
	key=free_Token(key);
	key=Get_Kwd("what: ",";",properties);
	if (key->status==TOK_OK) {
	  apply_property(key->key,na,A);
	}
      }
      if (key->status==TOK_NOK) {
	err_set(1,"unknown property: %s",key->text);
      }
      clearSubstructureMatch(na,A);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad template selection.");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad source selection.");
  }

  free_Token(key);

}


static void apply_property(char *property,int na,Atom **A)
{
  int i,j,n,nup; 
  Bond *b;

  if (streq(property,"atom")) {
    apply_property("q",na,A);
    apply_property("aromatic",na,A);
    apply_property("planarity",na,A);
    apply_property("ligance",na,A);
    apply_property("charge",na,A);
    apply_property("type",na,A);
    apply_property("name",na,A);
  }else if (streq(property,"bond")) {
    for (i=0;i<na;i++) {
      if (A[i]->chain) {
	for (j=0;j<A[i]->nbnd;j++) {
	  if (A[i]==A[i]->bnd[j]->f && A[i]->bnd[j]->t->chain) {
	    b=find_bond((Atom *)A[i]->chain,(Atom *)A[i]->bnd[j]->t->chain);
	    if (b) {
	      A[i]->bnd[j]->arom=b->arom;
	      A[i]->bnd[j]->bo36=b->bo36;
	      A[i]->bnd[j]->bo=b->bo;
	      A[i]->bnd[j]->bt=b->bt;
	    }
	  }
	}
      }
    }
  }else if (streq(property,"xyz")) {
    nup=0;
    for (i=0;i<na;i++) {
      if (A[i]->chain) {
	n=((Atom *)A[i]->chain)->nxyz;
	reall_atom_x(A[i],n);
	strncpy(A[i]->alt,((Atom *)A[i]->chain)->alt,n);
	for (j=0;j<ATM_LOCSZ*n;j++) A[i]->x[j]=((Atom *)A[i]->chain)->x[j];
	if (cPtr(A[i]->cross)) nup++;
      }
    }
    if (nup>0) {
      send_bheader(graphix,"NEWX",nup*(3*sizeof(float)+cPtrSize()));
      for (i=0;i<na;i++) {
	if (cPtr(A[i]->cross)) {
	  send_bcptr(graphix,A[i]->cross);
	  send_bfloata(graphix,3,A[i]->x);
	}
      }
      trflush(graphix);
    }
  }else if (streq(property,"name")) {
    for (i=0;i<na;i++) {
      if (A[i]->chain) {
	set_node_name((Node *)A[i],A[i]->chain->name);
      }
    }
  }else if (streq(property,"ligance")) {
      for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->lig=((Atom *)A[i]->chain)->lig;
      }
    }
  }else if (streq(property,"aromatic")) {
      for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->arom=((Atom *)A[i]->chain)->arom;
      }
    }
  }else if (streq(property,"planarity")) {
      for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->oop=((Atom *)A[i]->chain)->oop;
      }
    }
  }else if (streq(property,"charge")) {
      for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->chrg=((Atom *)A[i]->chain)->chrg;
      }
    }
  }else if (streq(property,"q")) {
      for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->q=((Atom *)A[i]->chain)->q;
      }
    }
  }else if (streq(property,"type")) {
      for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->atp=((Atom *)A[i]->chain)->atp;
	A[i]->arom=((Atom *)A[i]->chain)->arom;
	A[i]->oop=((Atom *)A[i]->chain)->oop;
      }
    }
  }
}

/* for compatibility with fix3Dligand script (Proasis, RELIBASE) */
static void match_fix3d(void)
{
  Token *key=(Token *)NULL;
  Atom **A=NULL,**B=NULL;
  Nodechain *c;
  int i,na,nb,nm;
  char namA[1024],namB[1024];

  key=Get_Atoms_D("source: ",";");
  if (key->status==TOK_OK) {
    c=key->ptr;
    na=c->n;
    A=(Atom **)chain_to_ptr(c,TRUE);
    key=free_Token(key);
    key=Get_Atoms_D("target: ",";");
    if (key->status==TOK_OK) {
      c=key->ptr;
      nb=c->n;
      B=(Atom **)chain_to_ptr(c,TRUE);
      nm=findSubstructure(na,A,nb,B);
      if (nm==na) {
	nm=applySubstructureMatch(na,A);
	if (na<nb) prtGraphics("findSubstructure: partial match\n"); 
	if (nm>0) prtGraphics("findSubstructure: unmatched bonds\n");
	for (i=0;i<na;i++) {
	  if (FALSE && A[i]->chain) {
	    prtGraphics("%s --> %s\n",nodeptr_to_name(namA,A[i]->chain->chain),nodeptr_to_name(namB,A[i]->chain));
	  }
	}
	clearSubstructureMatch(na,A);
      }else{
	err_set(1,"findSubstructure: no match");
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad target selection.");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad source selection.");
  }

  free_Token(key);
}

