#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"
#include "bhtree.h"

extern int graphix;
extern Node *MolRoot,*DefaultNode;

#define key_is(a) (streq(a,key->key))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

#define NATOMS 4
#define NTRA   4

#define NONE -1
#define DONE 0
#define MAIN 1
#define DISTANCE 2
#define ANGLE 3
#define TORSION 4
#define MEASURE 5
#define MONITOR 6
#define MONITOR_ADD 7
#define RMSDISTANCE 8
#define COORDINATES 9
#define PMOI 10

#define MON_MODE_NONE       0
#define MON_MODE_VDW        1
#define MON_MODE_CLOSE      2
#define MON_MODE_HB         3
#define MON_MODE_COVALENT   4

#define MON_LIST_OFF        0
#define MON_LIST_TERMINAL   1
#define MON_LIST_FILE       2

#define MXLIST 2048
#define MXTREE 2048

static void cmd_monitor_pairs(char *op);
static void cmd_monitor_hbonds(void);
static void cmd_monitor_close(char *op);
static void cmd_measure_pmoi(void);

static int mon_list=MON_LIST_OFF;
static int redundant=TRUE;
static char *mon_file_name=(char *)NULL;
static FILE *mon_file=(FILE *)NULL;
static char *max_d_menu="1.0 2.0 3.0 4.0 5.0 6.0 7.0 8.0 9.0 10.0";
static char *vdw_f_menu="0.50 0.55 0.60 0.65 0.70 0.75 0.80 0.85 0.90 0.95 1.00";
static int shortname=TRUE;

/*-------------------- get_atoms --------------------*/

static Nodechain *get_atoms(char *prmt)
{
  Token *key;
  Nodechain *c;
  
  key=(Token *)NULL;
  key=Get_Atoms_D(prmt,";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad atom selection: %s",key->text);
    c=(Nodechain *)NULL;
  }else{
    c=(Nodechain *)NULL;
  }
  free_Token(key);
  return(c);
}
/*========================================================================*/
void cmd_measure(void)
{
  Token *atm[NATOMS],*key;
  int state,measure,reparse_atom,m=0,n=0,i,erro;
  float D,A,T,x[NTRA][3];
  Atom *a[NTRA];
  char *prompt=(char *)NULL,*mon_options,*meas_keys,sn[NTRA][512];
  
  state=MAIN;
  key=(Token *)NULL;
  meas_keys="coordinates distance angle torsion PMOI rmsdistance monitor";
  mon_options="add clear ' ' autopairs close covalent hbonds h!near h!stable vdw ' ' color label list width ' ' on off";
  while (state) {
    err_dump();
    if (!key) key=Get_Kwd("measure: ",";",meas_keys);
    if (key->status==TOK_NOK)  {
      state=NONE;
      err_set(1,"measure: unknown option %s\n",key->text);
    }else if (key->status!=TOK_OK) state=DONE;
    else if (key_is("coordinates")) state=COORDINATES;
    else if (key_is("distance"))  state=DISTANCE;
    else if (key_is("rmsdistance")) state=RMSDISTANCE;
    else if (key_is("angle"))     state=ANGLE;
    else if (key_is("torsion"))   state=TORSION;
    else if (key_is("monitor"))   state=MONITOR;
    else if (key_is("PMOI"))      state=PMOI;
    else state=NONE;
    key=free_Token(key);
    reparse_atom=FALSE;

    switch (state) {
    case COORDINATES:
      m=1; prompt="coordinates";
      measure=TRUE;
      break;
    case DISTANCE:
      m=2; prompt="distance";
      measure=TRUE;
      break;
    case ANGLE:
      m=3; prompt="angle";
      measure=TRUE;
      break;
    case TORSION:
      m=4; prompt="torsion";
      measure=TRUE;
      break;
    case RMSDISTANCE:
      measure=FALSE;
      cmd_fit("measure");
      state=MAIN;
      break;
    case PMOI:
      measure=FALSE;
      cmd_measure_pmoi();
      state=MAIN;
      break;
    case MONITOR:
      key=free_Token(key);
      key=Get_Kwd("monitor option: ",";",mon_options);
      erro=err_seqnr();
      if (key->status==TOK_OK) {
	measure=FALSE; state=MAIN;
	if (key_is("add")) {
	  m=2; prompt="add monitor distance";
	  measure=TRUE;
	  state=MONITOR_ADD;
	}else if (key_is("hbonds")) {
	  cmd_monitor_hbonds();
	}else if (key_is("close")||key_is("vdw")||key_is("covalent")) {
	  cmd_monitor_close(key->key);
	}else if (key_is("autopairs")) {
	  cmd_monitor_pairs((char *)NULL);
	}else if (key_is("near")||key_is("stable")) {   /* compatibiliy with old versions... */
	  cmd_monitor_pairs(key->key);
	}else if (key_is("clear")) {
	  send_header(graphix,"CMON",0);
	}else if (key_is("label")) {
	  key=free_Token(key);
	  key=Get_Kwd("monitor label option: ",";","on off");
	  if (key->status==TOK_OK) {
	    if (key_is("on")) {
	      send_bheader(graphix,"MONL",sizeof(int));
	      send_int(graphix,1);
	    }else if (key_is("off")) {
	      send_bheader(graphix,"MONL",sizeof(int));
	      send_int(graphix,0);
	    }
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad monitor label option: %s",key->text);
	  }
	}else if (key_is("list")) {
	  key=free_Token(key);
	  key=Get_Kwd("list mode: ",";","terminal file off '' fullnames shortnames '' redundant unique");
	  if (key->status==TOK_OK) {
	    if (key_is("terminal"))  {
	      if (mvld(mon_file_name)) {
		prtGraphics("closing file %s\n",mon_file_name);
		fclose(mon_file); mon_file=(FILE *)NULL;
		free (mon_file_name);mon_file_name=(char *)NULL;
	      }
	      mon_list=MON_LIST_TERMINAL;
	    }else if (key_is("off")) {
	      if (mvld(mon_file_name)) {
		prtGraphics("closing file %s\n",mon_file_name);
		fclose(mon_file); mon_file=(FILE *)NULL;
		free (mon_file_name);mon_file_name=(char *)NULL;
	      }
	      mon_list=MON_LIST_OFF;
	    }else if (key_is("file")) {
	      if (mvld(mon_file_name)) {
		prtGraphics("closing old file %s\n",mon_file_name);
		fclose(mon_file); mon_file=(FILE *)NULL;
		free (mon_file_name);mon_file_name=(char *)NULL;
	      }
	      mon_list=MON_LIST_OFF;
	      key=free_Token(key);
	      key=Get_String("monitor file name = ",";");
	      if (key->status==TOK_OK) {
		mon_file=fopen(key->text,"w");
		if (mon_file) {
		  mon_file_name=Strdup(key->text);
		  mon_list=MON_LIST_FILE;
		}else{
		  err_set(1,"could not open %s for write.",key->text);
		}
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad file-name token: %s",key->text);
	      }
	    }else if (key_is("fullnames")) {
	      shortname=FALSE;
	    }else if (key_is("shortnames")) {
	      shortname=TRUE;
	    }else if (key_is("redundant")) {
	      redundant=TRUE;
	    }else if (key_is("unique")) {
	      redundant=FALSE;
	    }
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad monitor list option: %s",key->text);
	  }
	}else if (key_is("on")) {
	  send_bheader(graphix,"MONS",sizeof(int));
	  send_bint(graphix,1);
	  prtGraphics("distance monitoring turned on.\n");
	}else if (key_is("off")) {
	  send_bheader(graphix,"MONS",sizeof(int));
	  send_bint(graphix,0);
	  prtGraphics("distance monitoring turned of.\n");
	}else if (key_is("color")) {
	  key=free_Token(key);
	  key=Get_Color("monitor line color: ",";");
	  if (key->status==TOK_OK) {
	    send_bheader(graphix,"MONC",sizeof(int));
	    send_bint(graphix,key->n);
    	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad color token: %s",key->text);
	  }
	}else if (key_is("width")) {
	  key=free_Token(key);
	  key=Get_Int_M("monitor linewidth: ",";","1 2 3 4 5 6 7 8 9 10");
	  if (key->status==TOK_OK) {
	    send_bheader(graphix,"MONW",sizeof(int));
	    send_bint(graphix,key->n);
    	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad integer token: %s",key->text);
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"measure monitor: bad token: %s",key->text);
	state=MAIN;
	measure=FALSE;
      }else{
	state=NONE;
	measure=FALSE;
      }
      key=free_Token(key);
      if (erro<err_seqnr()) end_parse();
      break;
    default:
      measure=FALSE;
      if (state==NONE) end_parse();
      break;
    }

/*  measure distance, angle, torsion  */

    while(measure) {
      n=Get_nAtoms_M(m,atm,prompt,";",meas_keys);
      if (n==m) {
	send_bheader(graphix,"ARMK",0);
	for (i=0;i<m;i++) {
	  a[i]=atm[i]->ptr; 
	  free_Token(atm[i]);
	}
	get_graph_xyz(m,x[0],a);
	nodeptr_to_name(sn[0],(Node *)a[0]);
	for (i=1;i<m;i++) {
	  if (shortname) nodeptr_to_name_ctxt(sn[i],(Node *)a[i],(Node *)a[i-1]);
	  else nodeptr_to_name(sn[i],(Node *)a[i]);
	}
	switch (state) {
	case COORDINATES:
	  if (a[0]->nxyz>0&&mvld(a[0]->x)) {
	    prtGraphics(
			"%s at %0.4f,%0.4f,%0.4f, "
			"displayed at %0.4f,%0.4f,%0.4f\n",
			sn[0],a[0]->x[0],a[0]->x[1],a[0]->x[2],
			x[0][0],x[0][1],x[0][2]);
	  }else{
	    prtGraphics("no coordinates for %s\n",
			sn[0]);
	  }
	  break;
	case DISTANCE:
	  D=cdist(x[0],x[1]);
	  prtGraphics(
		      "distance %s -- %s = %0.3f\n",
		      sn[0],sn[1],D);
	  break;
	case ANGLE:
	  A=cangled(x[0],x[1],x[2]);
	  prtGraphics(
		      "angle %s -- %s -- %s = %0.2f\n",
		      sn[0],sn[1],sn[2],A);
	  break;
	case TORSION:
	  T=ctorsd(x[0],x[1],x[2],x[3]);
	  prtGraphics(
		      "dihedral %s -- %s -- %s -- %s = %0.1f\n",
		      sn[0],sn[1],sn[2],sn[3],T);
	  break;
	case MONITOR_ADD:
	  if (cPtr(a[0]->cross) && cPtr(a[1]->cross)) {
	    send_bheader(graphix,"AMON",2*cPtrSize());
	    send_bcptr(graphix,a[0]->cross);
	    send_bcptr(graphix,a[1]->cross);
	  }
	  D=cdist(x[0],x[1]);
	  prtGraphics(
		      "monitor distance %s -- %s = %0.3f\n",
		      sn[0],sn[1],D);
	  break;
	}
      }else{
	reparse_atom=TRUE;
	break;
      }
    }

/*   reparse bad atom token */

    if (reparse_atom) {
      if (n>0) send_header(graphix,"ARMK",0);
      switch (atm[n]->status) {
      case TOK_NOK:
	key=Prs_Kwd(atm[n],meas_keys);
	if (key->status==TOK_OK) state=MAIN;
	else {
	  prtGraphics("no such atom: %s\n",key->text);
	  key=free_Token(key);
	}
	break;
      case TOK_ABORT:
	state=(n>0 ? state : NONE);
	break;
      case TOK_EOC:
	state=NONE;
	break;  
      default:
	state=MAIN;
      }
      for (i=0;i<n-1;i++) atm[i]=free_Token(atm[i]);
    }

  }
}
/*-------------------- cmd_measure_pmoi --------------------*/

static void cmd_measure_pmoi(void)
{
  int i;
  mat_4 R;
  float **x,*m=(float *)NULL;
  Token *key=(Token *)NULL;
  Nodechain *c;
  Atom *a;

  key=Get_Atoms_D("Atoms: ",";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n>0) {
      m=malloc(c->n*sizeof(float));
      if (!mvld(m)) err_set(1,"memory allocation failure, all masses assumed = 1.0");
      x=(float **)malloc(c->n*sizeof(float *));
      if (mvld(x)) {
	for(i=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain,i++) {
	  if (mvld(m)) m[i]=(mvld(a->atp) ? a->atp->mass : 0.0);
	  x[i]=a->x;
	}
	R=PMofI(c->n,x,m);
	prtGraphics("principal moments of inertia = %0.4E, %0.4E, %0.4E\n",
		    R.r[3].c[0],R.r[3].c[1],R.r[3].c[2]);
	free(x);
      }else{
	err_set(1,"memory allocation failure");
      }
      if (mvld(m)) free(m);
    }
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) err_set(1,"bad selection of atoms: %s\n",key->text);
  free_Token(key);
}

/*-------------------- cmd_monitor_hbonds --------------------*/

static void cmd_monitor_hbonds(void)
{
  Atom **atms,**from,*a,*cand[MXLIST];
  float max_dist=0,dist[MXLIST];
  Nodechain *c;
  BHtree *t;
  int nat,i,j,k,n,grain,mode,nout,n_from;
  Token *key;
  char sn[512],tn[512];

  key=(Token *)NULL;
  mode=MON_MODE_HB;

  default_token("2.5");
  key=Get_Float_M("max. distance (H -- A) = ",";",
		  "2.0 2.1 2.2 2.3 2.4 2.5 2.6 2.7 2.8 2.9 3.0");
  if (key->status==TOK_OK) {
    max_dist=key->f;
    if (max_dist<=0) mode=MON_MODE_NONE;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad float: %s\n",key->text);
    mode=MON_MODE_NONE;
  }else{
    mode=MON_MODE_NONE;
  }
  
  free_Token(key);
  if (mode==MON_MODE_NONE) return;

  c=get_atoms("monitor HB from atoms: ");
  if (!mvld(c)) return;
  
  /* remove atoms that are neither potential acceptors nor donors */

  if (c->n>0) {
    nat=c->n;
    atms=(Atom **)chain_to_ptr(c,FALSE);
    clear_node_chain(c);
    if (mvld(atms)) {
      for (k=0,i=0;i<nat;i++) {
	if (mvld(atms[i]->atp)) {
	  if (atms[i]->atp->hbacc) {
	    atms[k++]=atms[i];
	    add_to_node_chain((Node *)atms[i],c);
	  }
	  if (atms[i]->atp->hbdon) {
	    for (j=0;j<atms[i]->nbnd;j++) {
	      a=neighbour(j,atms[i]);
	      if (mvld(a->atp)) {
		if (a->atp->z==1) add_to_node_chain((Node *)a,c);
	      }
	    }
	  }
	  if (atms[i]->atp->z==1 && atms[i]->nbnd==1) {
	    a=neighbour(0,atms[i]);
	    if (mvld(a->atp)) {
	      if (a->atp->hbdon) add_to_node_chain((Node *)atms[i],c);
	    }
	  }
	}
      }
      free(atms);
    }
  }
  
  if (c->n==0) {
    err_set(1,"no acceptor/donor atoms found in specified list.");
    free_node_chain (c);
    return;
  }
  
  n_from=c->n;
  from=(Atom **)chain_to_ptr(c,TRUE);
  if (!mvld(from)) return;

  c=get_atoms("monitor HB to atoms: ");
  if (!mvld(c)) {
    free(from);
    return;
  }


  /* remove atoms that are neither potential acceptors nor donors */

  if (c->n>0) {
    nat=c->n;
    atms=(Atom **)chain_to_ptr(c,FALSE);
    clear_node_chain(c);
    if (mvld(atms)) {
      for (k=0,i=0;i<nat;i++) {
	if (mvld(atms[i]->atp)) {
	  if (atms[i]->atp->hbacc) {
	    atms[k++]=atms[i];
	    add_to_node_chain((Node *)atms[i],c);
	  }
	  if (atms[i]->atp->hbdon) {
	    for (j=0;j<atms[i]->nbnd;j++) {
	      a=neighbour(j,atms[i]);
	      if (mvld(a->atp)) {
		if (a->atp->z==1) add_to_node_chain((Node *)a,c);
	      }
	    }
	  }
	  if (atms[i]->atp->z==1 && atms[i]->nbnd==1) {
	    a=neighbour(0,atms[i]);
	    if (mvld(a->atp)) {
	      if (a->atp->hbdon) add_to_node_chain((Node *)atms[i],c);
	    }
	  }
	}
      }
      free(atms);
    }
  }
  
  if (c->n==0) {
    err_set(1,"no acceptor/donor atoms found in specified list.");
    free_node_chain (c);
    free(from);
    return;
  }

  grain = ( c->n < MXTREE ? 1 : c->n / MXTREE );
  t=generateBHtree(c,grain);
  clear_node_chain(c);
  free_node_chain(c);
  if (!mvld(t)) {
    err_set(1,"failed to generate BH-tree...");
    free(from);
    return;
  }
  for (i=0;i<t->n;i++) t->atom[i]->flg|=(1<<2);
  
  nout=0;

  for (k=0;k<n_from;k++) {
    a=from[k]; a->flg|=(1<<1);
    if (a->nxyz>0) n=findBHcloseAtoms(t,a->x,max_dist,cand,dist,MXLIST);
    else n=0;
    if (n>MXLIST) {
      err_set(1,"neighbour list overflow. aborted.");
      break;
    }else if (n>0) {
      for (i=0;i<n;i++) {
	if ((a->atp->hbacc && !cand[i]->atp->hbacc) ||
	    (!a->atp->hbacc && cand[i]->atp->hbacc)) {
	  if (!near_neighbours(a,cand[i],3)) {
	    if (!(a->flg&(1<<2) && cand[i]->flg&(1<<1))) {   /* display unique */
	      nout++;
	      if (cPtr(a->cross) && cPtr(cand[i]->cross)) {
		if (!near_neighbours(a,cand[i],3)) {
		  send_bheader(graphix,"AMON",2*cPtrSize());
		  send_bcptr(graphix,a->cross);
		  send_bcptr(graphix,cand[i]->cross);
		}
	      }
	    }
	    if (mon_list) {                            /* list redundant */
	      nodeptr_to_name(sn,(Node *)a);
	      if (shortname) nodeptr_to_name_ctxt(tn,(Node *)cand[i],(Node *)a);
	      else nodeptr_to_name(tn,(Node *)cand[i]);
	      if (mon_list==MON_LIST_TERMINAL) 
		prtGraphics(
			    "%s -- %s = %0.3f\n",sn,tn,dist[i]);
	      else if (mon_list==MON_LIST_FILE && mon_file)
		fprintf(mon_file,
			"%s -- %s = %0.3f\n",sn,tn,dist[i]);
	    }
	  }
	}
      }	  
    }
  }

  prtGraphics("found %d hydrogen bonds (H -- A < %0.3f).\n",
	      nout,max_dist);
  if (mon_list==MON_LIST_FILE && mon_file) {
    fprintf(mon_file,"found %d hydrogen bonds (H -- A < %0.3f).\n\n",
	    nout,max_dist);
    fflush(mon_file);
  }

  for (i=0;i<t->n;i++) t->atom[i]->flg&=~(1<<2);
  freeBHtree(t);
  for (i=0;i<n_from;i++) from[i]->flg&=~(1<<1);
  free(from);
}
/*-------------------- cmd_monitor_close --------------------*/

static void cmd_monitor_close(char *op)
{
  Atom **from,*a,*cand[MXLIST];
  float dist[MXLIST],TOL=0,ra=0;
  char sn[512],tn[512];
  Nodechain *c;
  BHtree *t;
  float min_dist=0,max_dist=0,max_r;
  int n_from,i,l,n,grain,mode=0,nout;
  Token *key;

  key=(Token *)NULL;

  if (streq(op,"vdw")) mode=MON_MODE_VDW;
  else if (streq(op,"close")) mode=MON_MODE_CLOSE;
  else if (streq(op,"covalent")) mode=MON_MODE_COVALENT;

  nout=0;

  switch (mode) {
  case MON_MODE_VDW:
    default_token("0.90");
    key=Get_Float_M("vdW-factor = ",";",vdw_f_menu);
    if (key->status==TOK_OK) {
      TOL=key->f;
      if (TOL<=0) mode=MON_MODE_NONE;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad float: %s\n",key->text);
      mode=MON_MODE_NONE;
    }else{
      mode=MON_MODE_NONE;
    }
    max_dist=6.0;
    break;
  case MON_MODE_CLOSE:
    default_token("4.5");
    key=Get_Float_M("max. distance = ",";",max_d_menu);
    if (key->status==TOK_OK) {
      max_dist=key->f;
      if (max_dist<=0) mode=MON_MODE_NONE;
      TOL=1.0;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad float: %s\n",key->text);
      mode=MON_MODE_NONE;
    }else{
      mode=MON_MODE_NONE;
    }
    break;
  case MON_MODE_COVALENT:
    default_token("0.40");
    key=Get_Float_M("tolerance = ",";","0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0");
    if (key->status==TOK_OK) {
      TOL=key->f;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad float: %s\n",key->text);
      mode=MON_MODE_NONE;
    }else{
      mode=MON_MODE_NONE;
    }
    max_dist=2.5+TOL;
    break;
  default:
    err_set(1,"unknown monitor mode: %s",op);
    mode=MON_MODE_NONE;
    break;
  }

  key=free_Token(key);
  if (mode==MON_MODE_NONE) return;

  c=get_atoms("monitor close contacts from atoms: ");
  if (!mvld(c)) return;
  if (c->n<1) {
    err_set(1,"no atoms selected...");
    free_node_chain(c);
    return;
  }

  n_from=c->n;
  from=(Atom **)chain_to_ptr(c,TRUE);
  if (!mvld(from)) return;

  c=get_atoms("monitor close contacts to atoms: ");
  if (!mvld(c)) {
    free(from);
    return;
  }

  if (c->n<1) {
    err_set(1,"no atoms selected...");
    free_node_chain(c); free(from);
    return;
  }
  
  grain = ( c->n < MXTREE ? 1 : c->n / MXTREE );
  t=generateBHtree(c,grain);
  clear_node_chain(c);
  free_node_chain(c);
  if (!mvld(t)) {
    err_set(1,"failed to generate BH-tree");
    free(from);
    return;
  }
  for (max_r=0,i=0;i<t->n;i++) {
    t->atom[i]->flg|=(1<<2);
    t->atom[i]->seqnr=i;
    if (mode==MON_MODE_VDW) {
      if (mvld(t->atom[i]->atp)) {
	ra=t->atom[i]->atp->rvdw;
	max_r=(max_r<ra ? ra : max_r);
      }
    }else if (mode==MON_MODE_COVALENT) {
      if (mvld(t->atom[i]->atp)) {
	ra=t->atom[i]->atp->rcov;
	max_r=(max_r<ra ? ra : max_r);
      }
    }
  }

  for (l=0;l<n_from;l++) {
    a=from[l]; a->flg|=(1<<1);
    if (mode==MON_MODE_VDW) {
      if (!mvld(a->atp)) continue;
      ra=a->atp->rvdw;
      max_dist=(ra+max_r)*TOL;
    }else if (mode==MON_MODE_COVALENT) {
      if (!mvld(a->atp)) continue;
      ra=a->atp->rcov;
      max_dist=(ra+max_r)+TOL;
    }      
    if (a->nxyz>0) n=findBHcloseAtoms(t,a->x,max_dist,cand,dist,MXLIST);
    else n=0;
    if (n>MXLIST) {
      err_set(1,"neighbour list overflow. aborted.");
      break;
    }else if (n>0) {
      for (i=0;i<n;i++) {
	if (mode==MON_MODE_VDW) {
	  if (mvld(cand[i]->atp)) max_dist=TOL*(ra+cand[i]->atp->rvdw);
	  else continue;
	}else if (mode==MON_MODE_COVALENT) {
	  if (mvld(cand[i]->atp)) {
	    max_dist=(ra+cand[i]->atp->rcov)+TOL;
	    min_dist=(ra+cand[i]->atp->rcov)-TOL;
	  }
	  else continue;
	}
	if (dist[i]<max_dist && dist[i]>=min_dist) {
	  if (!near_neighbours(a,cand[i],3)) {
	    if (!(a->flg&(1<<2) && cand[i]->flg&(1<<1))) { /* add unique */
	      nout++;
	      if (cPtr(a->cross) && cPtr(cand[i]->cross)) {
		send_bheader(graphix,"AMON",2*cPtrSize());
		send_bcptr(graphix,a->cross);
		send_bcptr(graphix,cand[i]->cross);
	      }
	    }
	    if (mon_list && (redundant || !(a->flg&(1<<2) && cand[i]->flg&(1<<1)))) {
	      nodeptr_to_name(sn,(Node *)a);
	      if (shortname) nodeptr_to_name_ctxt(tn,(Node *)cand[i],(Node *)a);
	      else nodeptr_to_name(tn,(Node *)cand[i]);
	      if (mon_list==MON_LIST_TERMINAL) 
		prtGraphics(
			    "%s -- %s = %0.3f\n",sn,tn,dist[i]);
	      else if (mon_list==MON_LIST_FILE && mon_file)
		fprintf(mon_file,
			"%s -- %s = %0.3f\n",sn,tn,dist[i]);

	    }
	  }
	}
      }
    }	  
  }

  switch (mode) {
  case MON_MODE_VDW:
    prtGraphics("found %d short (<%0.3f vdW) non-bonded distances.\n",
		nout,TOL);
    if (mon_list==MON_LIST_FILE && mon_file) {
      fprintf(mon_file,"found %d short (<%0.3f vdW) non-bonded distances.\n\n",
	      nout,TOL);
      fflush(mon_file);
    }
    break;
  case MON_MODE_COVALENT:
    prtGraphics("found %d short (< rcov + %0.3f) non-bonded distances.\n",
		nout,TOL);
    if (mon_list==MON_LIST_FILE && mon_file) {
      fprintf(mon_file,"found %d short (< rcov + %0.3f) non-bonded distances.\n\n",
	      nout,TOL);
      fflush(mon_file);
    }
    break;
  case MON_MODE_CLOSE:
    prtGraphics("found %d non-bonded distances within %0.3f\n",
		nout,max_dist);
    if (mon_list==MON_LIST_FILE && mon_file) {
      fprintf(mon_file,"found %d non-bonded distances within %0.3f\n",
	      nout,max_dist);
      fflush(mon_file);
    }
    break;
  }

  for (i=0;i<t->n;i++) t->atom[i]->flg&=~(1<<2);
  freeBHtree(t);
  for (i=0;i<n_from;i++) from[i]->flg&=~(1<<1);
  free(from);
}
/*-------------------- cmd_monitor_pairs --------------------*/

static void cmd_monitor_pairs(char *op)
{
  Atom **from,**to,*a;
  char sn[512],tn[512];
  Nodechain *c;
  float dist,TOL=0;
  int i,n_from=0,n_to=0,npairs,mode=0;
  Token *key;
  char *maxdM;
  
  key=(Token *)NULL;
  from=(Atom **)NULL; to=(Atom **)NULL;
  
  if (!op) {
    key=Get_Kwd("autopair method: ",";","near stable structal");
    if (key->status==TOK_OK) op=key->key;
    else if (key->status==TOK_NOK) err_set(1,"unknown autopair method: %s",key->text);
  }

  if (op) {
    if (streq(op,"near")) mode=AUTO_PAIR_NEAR;
    else if (streq(op,"stable")) mode=AUTO_PAIR_STABLE;
    else if (streq(op,"structal")) mode=AUTO_PAIR_STRUCTAL;
  
    key=free_Token(key);
    default_token(mode==AUTO_PAIR_STRUCTAL?"-3":"3");
    maxdM=(mode==AUTO_PAIR_STRUCTAL?
	   "-10 -9 -8 -7 -6 -5 -4 -3 -2 -1 ' ' 0 ' ' 1 2 3 4 5 6 7 8 9 10" :
	   "0.00 0.05 0.10 0.15 0.20 0.25 0.30 0.35 0.40 0.45 0.50 1.0 2.0 3.0 4.0 5.0 10.0");
    key=Get_Float_M("max. dist.:",";",maxdM);
    if (key->status==TOK_OK) {
      TOL=key->f;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad float: %s\n",key->text);
      key->status=TOK_ABORT;
    }
    if (key->status==TOK_OK) {
      c=get_atoms("monitor close contacts from atoms: ");
      if (!mvld(c)) key->status=TOK_ABORT;
      else if (c->n<1) {
	err_set(1,"no atoms selected...");
	clear_node_chain(c);
	free_node_chain(c);
	key->status=TOK_ABORT;
      }else{
	n_from=c->n;
	from=(Atom **)chain_to_ptr(c,TRUE);
	if (!mvld(from)) {
	  err_set(1,"memory allocation failed.");
	  key->status=TOK_ABORT;
	}
      }
    }
    if (key->status==TOK_OK) {
      c=get_atoms("monitor close contacts to atoms: ");
      if (!mvld(c)) key->status=TOK_ABORT;
      else if (c->n<1) {
	err_set(1,"no atoms selected...");
	clear_node_chain(c);
	free_node_chain(c);
	key->status=TOK_ABORT;
      }else{
	n_to=c->n;
	to=(Atom **)chain_to_ptr(c,TRUE);
	if (!mvld(to)) {
	  err_set(1,"memory allocation failed.");
	  key->status=TOK_ABORT;
	}
      }
    }
    if (key->status==TOK_OK) {
      if (mode==AUTO_PAIR_STRUCTAL) {
	walk_order(MolRoot,n_to,to,(1<<3));
	walk_order(MolRoot,n_from,from,(1<<3));
      }
      npairs=auto_pairs(n_to,to,n_from,from,TOL,mode);
      for (i=0;i<n_from;i++) {
	if (is_atom(from[i]->chain)) {
	  a=(Atom *)from[i]->chain;
	  if (cPtr(from[i]->cross) && cPtr(a->cross)) {
	    send_bheader(graphix,"AMON",2*cPtrSize());
	    send_bcptr(graphix,from[i]->cross);
	    send_bcptr(graphix,a->cross);
	  }
	  if (mon_list) {                            /* list redundant */
	    nodeptr_to_name(sn,(Node *)from[i]);
	    if (shortname) nodeptr_to_name_ctxt(tn,(Node *)a,(Node *)from[i]);
	    else nodeptr_to_name(tn,(Node *)a);
	    dist=cdist(from[i]->x,a->x);
	    if (mon_list==MON_LIST_TERMINAL) 
	      prtGraphics(
			  "%s -- %s = %0.3f\n",sn,tn,dist);
	    else if (mon_list==MON_LIST_FILE && mon_file)
	      fprintf(mon_file,
		      "%s -- %s = %0.3f\n",sn,tn,dist);
	    
	  }
	}
	from[i]->chain=(Node *)NULL;
      }
      prtGraphics("found %d pairs.\n",(npairs>0?npairs:0));
      if (mon_list==MON_LIST_FILE && mon_file) {
	fprintf(mon_file,"found %d pairs.\n\n",npairs);
	fflush(mon_file);
      }
    }
    if (mvld(from)) free(from);
    if (mvld(to)) free(to);
  }
  free_Token(key);
}
/*-------------------- near_neighbours --------------------*/

int near_neighbours(Atom *a,Atom *b,int nbonds)
{
  /* return true iff there is a path from a to b of length <= nbonds */
  /* works for nbonds < 4 only...                                     */

  Atom *c;
  int i,j,k;

  if (!mvld(a) || !mvld(b)) return(FALSE);

  if (a==b) return(TRUE);

  for (i=0;i<a->nbnd;i++) {   /* bonded */
    c=neighbour(i,a);
    if (b==c) return(TRUE);
  }
  
  if (nbonds>1) {             /* 1-3 */
    for (i=0;i<a->nbnd;i++) {
      c=neighbour(i,a);
      for (j=0;j<c->nbnd;j++) {
	if (b==neighbour(j,c)) return(TRUE);
      }
      if (nbonds>2) {
	for (k=0;k<b->nbnd;k++) {
	  for (j=0;j<c->nbnd;j++) {
	    if (neighbour(k,b)==neighbour(j,c)) return(TRUE);
	  }
	}
      }
    }
  }
  return(FALSE);
}
