#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "parmset.h"

extern int graphix;
extern Node *DefaultNode;

#define TRUE 1
#define FALSE 0

#define NORMAL_COMPLETION        99
#define INPUT_ERROR               1
#define NO_ATOMS                  2
#define TOPLOLOGY_ERROR           3
#define MISSING_FF_PARAMETERS     4
#define NB_LIST_ERROR             5
#define UNKNOWN_ERROR             6
#define NO_CONVERGENCE            7
#define MAX_ITER_EXCEEDED         8

#define MFLG                  (1<<6)

#define FF_AMBER    0
#define FF_CHARMM   1     /* must match order of keywords in FF menu... */
#define FF_TAFF     2

#define ESW_DEFAULT 0     /* must match order of keywords in ecutoff menu... */
#define ESW_SHIFT   1
#define ESW_SWITCH  2

#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define key_is(a) (streq(a,key->key))

static void cmd_run_minimax(Node *root,int fgrd);
static void cmd_minimax_constraints(Node **root);

static char *taff_keys = "molecule ' ' run submit ' '"
   "ForceField FFTables charges chirality constraints cutoff dielectric "
   "h!e14fac h!ecutoff h!vcutoff hbscale h!missing  h!nbxmod h!prmfile h!scee h!scnb h!sloppy h!topfile xrqfile ' '"
   "bad processors report save ' ' method gradient iterations ' ' list reset";

static char *amber_keys = "molecule ' ' run submit ' '"
   "ForceField h!FFTables charges chirality constraints cutoff dielectric "
   "h!e14fac ecutoff vcutoff h!hbscale missing h!nbxmod h!prmfile scee scnb sloppy topfile xrqfile ' '"
   "bad processors report save ' ' method gradient iterations ' ' list reset";

static char *charmm_keys = "molecule ' ' run submit ' '"
   "ForceField FFTables charges chirality constraints cutoff dielectric "
   "e14fac ecutoff vcutoff h!hbscale h!missing nbxmod prmfile h!scee h!scnb h!sloppy h!topfile xrqfile ' '"
   "bad processors report save ' ' method gradient iterations ' ' list reset";

static char *ffmnem[] = {"Amber","CHARMM","TAFF"};
static int Iterations=-1;
static int Processors=0;
static char *Method=(char *)NULL;
static char *FFTables=(char *)NULL;
static float rmsgrad=0.00035;
static int ignore_charges=0;
static float diel=1.0;
static int dielc=FALSE;
static float hbscale=0.70;
static float cutoff=0;
static float cutsw=1.0;
static float cutmarg=1.0;
static int bad=0;
static int report=10;
static int save=FALSE;
static int pres_chir=FALSE;
static char *constr_file=(char *)NULL;
static int FField=FF_TAFF;
static float e14fac=0.5;
static int nbxmod=5;
static int eswCH=ESW_DEFAULT;
static int vswCH=ESW_DEFAULT;
static int eswA=ESW_DEFAULT;
static int vswA=ESW_DEFAULT;
static float scnb=2.0,scee=1.2;
static char *prmfile=(char *)NULL;
static char *prmtop=(char *)NULL;
static int  xrq=FALSE;

void cmd_minimax(void)
{
  Token *key;
  Node *root;
  char *mol_menu,*main_keys[3],buff[4096];
  int loop,mode;
  Nodechain *c;
    
  main_keys[FF_AMBER]=amber_keys;
  main_keys[FF_CHARMM]=charmm_keys;
  main_keys[FF_TAFF]=taff_keys;

  mol_menu=Make_TopMolNameList("./ /",DefaultNode);
  key=(Token *)NULL;
  
  root=(Node *)NULL;
  
  for (loop=TRUE;loop;) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("minimax option: ",";",main_keys[FField]);
    if (key->status==TOK_OK) {
      if (key_is("run")) {
	cmd_run_minimax(root,TRUE);
	loop=FALSE;
      }else if (key_is("submit")) {
	cmd_run_minimax(root,FALSE);
	loop=FALSE;
      }else if (key_is("molecule")) {
	key=free_Token(key);
	key=Get_Mols_M("molecule: ",";",mol_menu);
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (c->n==1) {
	    root=c->s;
	  }else{
	    root=(Node *)NULL;
	    err_set(1,"molecule selection %s is ambiguous.",key->text);
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad molecule selection: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("ForceField")) {
	key=free_Token(key);
	default_token(ffmnem[FField]);
	key=Get_Kwd("force field: ",";","Amber CHARMM TAFF");
	if (key->status==TOK_OK) FField=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad force field: %s",key->text);
      }else if (key_is("e14fac")) {
	default_token("%0.3f",e14fac);
	key=free_Token(key);
	key=Get_Float_M("scaling factor for 1-4 electrostatics: ",":","0.0 0.25 0.5 0.75 1.0");
	if (key->status==TOK_OK) {
	  e14fac=key->f;
	}else if (key->status==TOK_NOK) err_set(1,"bad 1-4 electrostatic scale factor: %s",key->text);
      }else if (key_is("ecutoff")) {
	default_token(eswCH==0?"default":eswCH==1?"shift":"switch");
	key=free_Token(key);
	key=Get_Kwd("electrostatic cutoff switching method: ",";",
		    "default shift switch");
	if (key->status==TOK_OK) eswCH=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad switching method: %s",key->text);
      }else if (key_is("vcutoff")) {
	default_token(vswCH==0?"default":vswCH==1?"shift":"switch");
	key=free_Token(key);
	key=Get_Kwd("van der Waals cutoff switching method: ",";",
		    "default shift switch");
	if (key->status==TOK_OK) vswCH=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad switching method: %s",key->text);
      }else if (key_is("nbxmod")) {
	default_token("%d",nbxmod);
	key=free_Token(key);
	key=Get_Int_M("automatic non-bonded generation method: ",";",
		      "-5 -4 -3 -2 -1 '' 0 '' 1 2 3 4 5");
	if (key->status==TOK_OK) {
	  if (key->n>-6&&key->n<6) nbxmod=key->n;
	  else key->status=TOK_NOK;
	}
	if (key->status==TOK_NOK) err_set(1,"bad value for nbxmod: %s",key->text);
      }else if (key_is("prmfile")) {
	key=free_Token(key);
	default_token(prmfile?prmfile:"PARM.PRM");
	key=Get_String("CHARMm parameter file: ",";");
	if (key->status==TOK_OK) {
	  if (prmfile) free(prmfile);
	  if (!streq(key->text,"PARM.PRM")) prmfile=strdup(key->text);
	  else prmfile=(char *)NULL;
	}else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }else if (key_is("scee")) {
	key=free_Token(key);
	default_token("%0.4f",scee);
	key=Get_Float_M("Amber 1-4 electrostatics scale: ",";","1.0 1.2 2.0");
	if (key->status==TOK_OK) scee=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }else if (key_is("scnb")) {
	key=free_Token(key);
	default_token("%0.4f",scnb);
	key=Get_Float_M("Amber 1-4 vdW scale: ",";","1.0 2.0");
	if (key->status==TOK_OK) scnb=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }else if (key_is("sloppy")) {
	key=free_Token(key);
	mode=checkParmSetMode(PRM_SLOPPY_TEST);
	default_token("%s",mode==PRM_SLOPPY_ALWAYS?"always":mode==PRM_SLOPPY_NEVER?"never":"once");
	key=Get_Kwd("write sloppy parmtop files: ",";","always never once");
	if (key->status==TOK_OK) {
	  checkParmSetMode(key->n==0?PRM_SLOPPY_ALWAYS:key->n==1?PRM_SLOPPY_NEVER:PRM_SLOPPY_ONCE);
	}else if (key->status==TOK_NOK) err_set(1,"invalid key: %s",key->text);
      }else if (key_is("missing")) {
	mode=missingParmSetMode(PRM_MISSING_TEST);
	default_token("%s",mode==PRM_MISSING_FIXED?"fixed":mode==PRM_MISSING_ZERO?"zero":"floppy");
	key=Get_Kwd("treatment of missing parameters in sloppy files: ",";","fixed floppy zero");
	if (key->status==TOK_OK) {
	  missingParmSetMode(key->n==0?PRM_MISSING_FIXED:key->n==1?PRM_MISSING_FLOPPY:PRM_MISSING_ZERO);
	}else if (key->status==TOK_NOK) err_set(1,"invalid key: %s",key->text);
      }else if (key_is("topfile")) {
	key=free_Token(key);
	default_token(prmtop?prmtop:"-automatic");
	key=Get_File_M(".",".prm","Amber parameter/topology file: ",";","-automatic");
	if (key->status==TOK_OK) {
	  if (prmtop) free(prmtop);
	  prmtop=strdup(key->text);
	  UCpclose(key->ptr);
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"-automatic");
	  if (key->status==TOK_OK) {
	    if (prmtop) free(prmtop);
	    prmtop=(char *)NULL;
	    err_pop(1);
	  }else{
	    if (!FileExists("",key->text,buff,sizeof(buff))) {
	      prmtop=strdup(key->text);
	      prtGraphics("%sParameter file %s will be created at run time.\n",err_text(),key->text);
	      err_pop(1);
	    }else{
	      err_set(1,"can't read par/top file %s\n",key->text);
	    }
	  }
	}
      }else if (key_is("xrqfile")) {
	key=free_Token(key);
	default_token(xrq?"yes":"no");
	key=Get_Kwd("produce xrq-file? ",";","yes no");
	if (key->status==TOK_OK) xrq=key_is("yes");
	else if (key->status==TOK_NOK) err_set(1,"bad response to simple question: %s\n",key->text);
      }else if (key_is("FFTables")) {
	key=free_Token(key);
	default_token("default");
	key=Get_String("directory containing FF-parameter tables: ",";");
	if (key->status==TOK_OK) {
	  if (mvld(FFTables)) free(FFTables);
	  if (!streq(key->text,"default")) FFTables=strdup(key->text);
	  else FFTables=(char *)NULL;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}
      }else if (key_is("gradient")) {
	key=free_Token(key);
	default_token("%0.5f",rmsgrad);
	key=Get_Float_M("rms. gradient (mdyn): ",";","0.00035 ' ' 0.00001 0.0001 0.001 0.01 0.1");
	if (key->status==TOK_OK) rmsgrad=key->f;
	else if (key->status==TOK_NOK) {
	  if (streq(key->text,"default")) rmsgrad=0.00035;
	  else err_set(1,"bad rms value: %s",key->text);
	}
      }else if (key_is("charges")) {
	key=free_Token(key);
	default_token("yes");
	key=Get_Kwd("use charges ? ",";","yes no");
	if (key->status==TOK_OK) ignore_charges=key->n;
	else err_set(1,"bad reply: %s",key->text);
      }else if (key_is("dielectric")) {
	key=free_Token(key);
	default_token(dielc?"constant":"distance-dependent");
	key=Get_Kwd("type of dielectric: ",";","constant distance-dependent");
	if (key->status==TOK_OK) {
	  dielc=key_is("constant");
	  key=free_Token(key);
	  default_token("%0.3f",diel);
	  key=Get_Float_M((dielc ? "constant: ":"factor: "),";","0.5 1.0 2.0 3.0 4.0 5.0 10.0 80.0");
	  if (key->status==TOK_OK) diel=key->f;  
	  else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
	}else if (key->status==TOK_NOK) err_set(1,"bad dielectric keyword: %s",key->text);
      }else if (key_is("hbscale")) {
	default_token("0.70");
	key=free_Token(key);
	key=Get_Float_M("scale for vdw in HBonds: ",";","0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0");
	if (key->status==TOK_OK) hbscale=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
      }else if key_is("cutoff") {
	key=free_Token(key);
	default_token("%0.3f",cutoff);
	key=Get_Float_M("non-bonded cutoff (0 for no cutoff): ",";","0 10 11 12 13 14 15 16 17 18 19 20");
	if (key->status==TOK_OK) {
	  cutoff=key->f;
	  if (cutoff>0) {
	    key=free_Token(key);
	    default_token("%0.3f",cutsw);
	    key=Get_Float_M("switching distance: ",";","0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0");
	    if (key->status==TOK_OK) {
	      cutsw=key->f;
	      default_token("%0.3f",cutmarg);
	      key=free_Token(key);
	      key=Get_Float_M("nb-list margin: ",";","0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0");
	      if (key->status==TOK_OK) cutmarg=key->f;
	      else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
	    }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
      }else if (key_is("bad")) {
	key=free_Token(key);
	if (bad==0||bad==3) default_token(bad?"yes":"no");
	else default_token("%d",bad);
	key=Get_Int_M("bad starting geometry ? ",";","yes no '' 0 1 2 3 4 5 6 7 8 9 10");
	if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"yes no");
	  if (key->status==TOK_OK) {
	    if (key_is("yes")) key->n=3;
	    else if (key_is("no")) key->n=0;
	  }
	}
	if (key->status==TOK_OK) bad=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }else if (key_is("report")) {
	key=free_Token(key);
	default_token("%d",report);
	key=Get_Int_M("report frequency (iterations): ",";","1 2 3 4 5 10 15 20 25 50 100 250 500 1000");
	if (key->status==TOK_OK) report=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad integer: %s",key->text);
      }else if (key_is("save")) {
	key=free_Token(key);
	default_token(save?"yes":"no");
	key=Get_Kwd("save intermediate conformations ? ",";","yes no");
	if (key->status==TOK_OK) save=key_is("yes");
	else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }else if (key_is("iterations")) {
	key=free_Token(key);
	if (Iterations<0) default_token("default");
	else default_token("%d",Iterations);
	key=Get_Int_M("max. number of iterations: ",";",
		      "default 0 10 50 100 200 500 1000");
	if (key->status==TOK_OK) {
	  if (key->n>=0) {
	    Iterations=key->n;
	  }else{
	    err_set(1,"number of iterations must be a non-negative integer.");
	  }
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"default");
	  if (key->status==TOK_OK) {
	    Iterations=-1;
	  }else{
	    err_set(1,"bad max. iter count: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad max. iter count: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("method")) {
	key=free_Token(key);
	default_token(Method?Method:"default");
	key=Get_Kwd("minimizer: ",";","default bfgs conjugate_gradient otherCG");
	if (key->status==TOK_OK) {
	  if (mvld(Method)) free(Method);
	  Method=strdup(key->key);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"unknown minimizer: %s",key->text);
	}else{
	  loop=FALSE;
	}
      }else if (key_is("processors")) {
	if (Processors) default_token("%d",Processors);
	else default_token("default");
	key=Get_Int_M("number of processor to use in parallel: ",";",
		      "default 1 2 3 4 5 6 7 8");
	if (key->status==TOK_OK) {
	  if (key->n>0) {
	    Processors=key->n;
	  }else{
	    err_set(1,"number of processors must be a positive integer.");
	  }
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"default");
	  if (key->status==TOK_OK) {
	    Processors=0;
	  }else{
	    err_set(1,"bad number of processors: %s",key->text);
	  }
	}else{
	  loop=FALSE;
	}
      }else if (key_is("constraints")) {
	cmd_minimax_constraints(&root);
      }else if (key_is("chirality")) {
	key=free_Token(key);
	default_token(pres_chir?"yes":"no");
	key=Get_Kwd("preserve stereochemistry of 4-bond centers ? ",";","yes no");
	if (key->status==TOK_OK) pres_chir=key_is("yes");
	else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }else if (key_is("list")) {
	prtGraphics("Minimax options, current settings\n");
	prtGraphics("---------------------------------\n");
	prtGraphics("Force Field: %s\n",ffmnem[FField]);
	prtGraphics("Directory with parameter files: %s\n",FFTables?FFTables:"$MINIMAX_FF_TABLES");
	prtGraphics("Parameter file (CHARMM): %s\n",prmfile?prmfile:"PARM.PRM");
	prtGraphics("Param/top file (Amber): %s\n",prmtop?prmtop:"-automatic");
	prtGraphics("TAFF HB scale: %0.3f\n",hbscale);
	prtGraphics("%s charges, dielectric %0.3f%s\n",ignore_charges?"ignore":"use",
		    diel,dielc?"":"*R");
	if (cutoff>0) prtGraphics("cutoff: %0.3f, switching distance: %0.3f, margin: %0.3f\n",
		    cutoff,cutsw,cutmarg);
	else prtGraphics("no cutoff, switching distance: %0.3f, margin: %0.3f\n",
		    cutsw,cutmarg);
	prtGraphics("CHARMM nbond parameters: nbxmod=%d, e14fac=%0.3f%s\n",nbxmod,e14fac,
		    eswCH==ESW_DEFAULT?"":eswCH==ESW_SHIFT?", eshift":", eswitch",
		    vswCH==ESW_DEFAULT?"":vswCH==ESW_SHIFT?", vshift":", vswitch");
	prtGraphics("Amber nbond parameters: SCEE=%0.4f, SCNB=%0.4f\n",scee,scnb,
		    eswA==ESW_DEFAULT?"":eswA==ESW_SHIFT?", eshift":", eswitch",
		    vswA==ESW_DEFAULT?"":vswA==ESW_SHIFT?", vshift":", vswitch");
	prtGraphics("constraints: %s%s\n",constr_file?constr_file:"none",
		    pres_chir?", preserve chirality":"");
	if (Iterations>0) prtGraphics("%d iterations, gradient=%0.5f, %s method, %d processors.\n",
		    Iterations,rmsgrad,Method?Method:"default",Processors);
	else prtGraphics("default iterations, gradient=%0.5f, %s method, %d processors.\n",
		    rmsgrad,Method?Method:"default",Processors);
	if (bad) prtGraphics("geometry bad(%s),%s save intermediate structures, report every %d iterations\n",
		    bad,save?"":" don't",report);
	else prtGraphics("geometry OK,%s save intermediate structures, report every %d iterations\n",
			 save?"":" don't",report);
	prtGraphics(xrq?"xrq file will be written.\n":"xrq file will not be written.\n");
      }else if (key_is("reset")) {
	if (mvld(constr_file)) {free(constr_file);} constr_file=(char *)NULL;
	if (mvld(Method)) {free(Method);} Method=(char *)NULL;
	if (mvld(FFTables)) {free(FFTables);} FFTables=(char *)NULL;
	if (mvld(prmfile)) {free(prmfile);} prmfile=(char *)NULL;
	if (mvld(prmtop)) {free(prmtop);} prmtop=(char *)NULL;
	e14fac=0.5; nbxmod=5; eswCH=ESW_DEFAULT; vswCH=ESW_DEFAULT; 
	scnb=2.0; scee=1.2; eswA=ESW_DEFAULT; vswA=ESW_DEFAULT; 
	FField=FF_TAFF;
	Iterations=-1;	Processors=0;
	rmsgrad=0.00035;
	ignore_charges=0; diel=1.0; dielc=FALSE;
	hbscale=0.70; cutoff=0;	cutsw=1.0; cutmarg=1.0;
	bad=0; report=10;
	save=FALSE; pres_chir=FALSE; xrq=FALSE;
	prtGraphics("Minimax options reset to default...\n");
      }else err_set(1,"unknown minimax keyword: %s",key->text);
    }else if (key->status==TOK_NOK) {
      if (!mvld(root)) {
	key=Prs_Mols_M(key,(char *)NULL);
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (c->n==1) {
	    root=c->s;
	  }else{
	    root=(Node *)NULL;
	    err_set(1,"molecule selection %s is ambiguous.",key->text);
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else{
	  err_set(1,"unknown minimiax option: %s.",key->text);
	}
      }
    }else{
      if (key->status==TOK_EOC) cmd_run_minimax(root,TRUE);
      loop=FALSE;
    }
  }
  
  free_menu(mol_menu);
  free_Token(key);
}
/*-------------------- cmd_run_minimax --------------------*/

static void cmd_run_minimax(Node *root,int fgrd)
{
  Token *key;
  char *menu,*srv;
  char cmd[2048],rec[256],jobname[1024],prmfn[4096],logfile[1024];
  Nodechain *c,atms;
  Atom *a;
  int status,iterations,i;
  UCpipe *Srv;

  key=(Token *)NULL;
  menu=(char *)NULL;
  init_node_chain(&atms);
  status=0;

  if (!is_mol(root)) {
    menu=Make_TopMolNameList("./ /",DefaultNode);
    key=Get_Mols_M("molecule: ",";",menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n==1) {
	root=c->s;
      }else{
	err_set(1,"ambiguous molecule specification: %s",key->text);
      }
      clear_node_chain(c);
      free_node_chain(c);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad molecule specification: %s",key->text);
    }
  }

  if (is_mol(root)) {
    if (mvld(root->name)) sprintf(jobname,"%s",root->name);
    else sprintf(jobname,"NoName");
    key=free_Token(key);
    default_token(jobname);
    key=Get_String("jobname: ",";");
    if (key->status==TOK_OK) {
      sprintf(jobname,"%s",key->text);
      WriteMolecule(root,jobname,FT_MOL2,FALSE);
      srv=getenv("MINIMAXserver");
      if (!srv) srv="minimax";
      if (*srv=='|') {
	srv++;
	/* fgrd=TRUE;   always use parameter passed in by caller... */
      }
      iterations=Iterations;
      if (FField==FF_CHARMM) {  /* CHARMM needs a PSF */
	ForAtoms(a,root) a->flg|=(1<<1);
	write_charmm_psf(jobname,root,CHRM_PSF_CHARMM);
	ForAtoms(a,root) a->flg&=~(1<<1);
      }
      if (FField==FF_AMBER) { /* Amber needs a .prm (parm-top) file... */
	if (iterations==-1) {
	  if (checkParmSetMode(PRM_SLOPPY_TEST)!=PRM_SLOPPY_NEVER &&
	      missingParmSetMode(PRM_MISSING_TEST)==PRM_MISSING_ZERO) iterations=0;
	}
	if (!prmtop || !FileExists("",prmtop,prmfn,sizeof(prmfn))) {
	  if (prmtop) sprintf(prmfn,"%s",prmtop);
	  else extfil(prmfn,jobname,".prm",TRUE);
	  ForAtoms(a,root) a->flg|=(1<<1);
	  status=write_amber_prmtop(prmfn,root,PRMTOP_FMT_NEW);
	  ForAtoms(a,root) a->flg&=~(1<<1);
	}
      }	
      sprintf(cmd,"%s%s%s",(fgrd?"|":""),srv,(fgrd?" -X":""));
      if (FField==FF_CHARMM) {  /* do CHARMM specific options, aware of defaults */ 
	sprintf(cmd+strlen(cmd)," -charmm");
	if (xrq) sprintf(cmd+strlen(cmd)," -xrq %s.xrq",jobname);
	if (nbxmod!=5) sprintf(cmd+strlen(cmd)," -nbx %d",nbxmod);
	else {
	  if ((e14fac<0.49999||e14fac>0.50001)&&!ignore_charges) {
	    sprintf(cmd+strlen(cmd)," -e14 %0.4f",e14fac);
	  }
	}
	if (!ignore_charges) {
	  if (eswCH!=ESW_DEFAULT) {
	    sprintf(cmd+strlen(cmd),"%s",eswCH==ESW_SHIFT?" -esh":eswCH==ESW_SWITCH?" -esw":"");
	  }
	}
	if (vswCH!=ESW_DEFAULT) {
	  sprintf(cmd+strlen(cmd),"%s",vswCH==ESW_SHIFT?" -vsh":vswCH==ESW_SWITCH?" -vsw":"");
	}
	if (prmfile&&!streq(prmfile,"PARM.PRM")) sprintf(cmd+strlen(cmd)," -prm %s",prmfile); 
      }else if (FField==FF_AMBER) {
	sprintf(cmd+strlen(cmd)," -amber");
	if (prmtop) sprintf(cmd+strlen(cmd)," -prm %s",prmtop);
	if (xrq) sprintf(cmd+strlen(cmd)," -xrq %s.xrq",jobname);
	if (scnb<1.9999||scnb>2.0001)sprintf(cmd+strlen(cmd)," -scnb %0.4f",scnb); 
	if (scee<1.1999||scee>1.2001)sprintf(cmd+strlen(cmd)," -scee %0.4f",scee);
	if (vswA!=ESW_DEFAULT) {
	  sprintf(cmd+strlen(cmd),"%s",vswA==ESW_SHIFT?" -vsh":vswA==ESW_SWITCH?" -vsw":"");
	}
	if (!ignore_charges) {
	  if (eswA!=ESW_DEFAULT) {
	    sprintf(cmd+strlen(cmd),"%s",eswA==ESW_SHIFT?" -esh":eswA==ESW_SWITCH?" -esw":"");
	  }
	}
      }else if (FField==FF_TAFF) {
	sprintf(cmd+strlen(cmd)," -taff");  /* redundant... */
	if (xrq) sprintf(cmd+strlen(cmd)," -xrq %s.xrq",jobname);
	if ((hbscale<0.699||hbscale>0.701)) sprintf(cmd+strlen(cmd)," -h%0.3f",hbscale);
      }
      if (mvld(FFTables)) sprintf(cmd+strlen(cmd)," -t%s",FFTables);
      if (iterations>=0) sprintf(cmd+strlen(cmd)," -i%d",iterations);
      if (ignore_charges) sprintf(cmd+strlen(cmd)," -c-");
      if (pres_chir) sprintf(cmd+strlen(cmd)," -Cs");
      if (mvld(constr_file)) sprintf(cmd+strlen(cmd)," -Cf %s",constr_file);
      if (!ignore_charges) sprintf(cmd+strlen(cmd)," %s%0.1f",(dielc?"-d":"-D"),diel);
      if (cutoff>0) sprintf(cmd+strlen(cmd)," -n%0.1f-%0.3f+%0.3f",cutoff,cutsw,cutmarg);
      else sprintf(cmd+strlen(cmd)," -n0");
      if (rmsgrad<0.000349||rmsgrad>0.000351) sprintf(cmd+strlen(cmd)," -g%f",rmsgrad);
      if (bad>0) {
	sprintf(cmd+strlen(cmd)," -");
	for (i=0;i<bad;i++) sprintf(cmd+strlen(cmd),"b");
      }
      if (save) sprintf(cmd+strlen(cmd)," -s");
      if (report!=10) sprintf(cmd+strlen(cmd)," -r%d",report);
      if (Processors>0) sprintf(cmd+strlen(cmd)," -p%d",Processors);
      if (mvld(Method)&&!streq(Method,"default")) sprintf(cmd+strlen(cmd)," -m%s",Method);
      sprintf(cmd+strlen(cmd)," %s.mol",jobname);
      if (!status) {          
	if (*cmd=='|') {      /* a pipe, run in foreground with automatic update */
	  prtGraphics("running Minimax in foreground:\n");
	  prtGraphics("%s\n",cmd);
	  Srv=UCpopen(cmd);
	  if (Srv) {
	    atms=tree2chain(root,OBJ_ATOM);	
	    for(;;) {
	      if (checkabort()) {
		prtGraphics("Killed...\n");
		break;
	      }
	      if (!fgets(rec,sizeof(rec)-1,Srv->fd)) {
		break;
	      }
	      if (feof(Srv->fd)) break;
	      if (strneq(rec,"X> BEGIN",8)) {
		a=(Atom *)atms.s;
		for(;;) {
		  if (!fgets(rec,sizeof(rec)-1,Srv->fd)) break;
		  if (!strneq(rec,"X> ",3)) {
		    prtGraphics(rec);
		    err_set(1,"coordinate set is incomplete.");
		    err_dump();
		    break;
		  }
		  if (strneq(rec+3,"END",3)) {
		    send_newx(&atms);
		    break;
		  }else if (is_atom(a)) {
		    sscanf(rec+3,"%f %f %f",a->x,a->x+1,a->x+2);
		    a=(Atom *)a->chain;
		  }
		}
	      }else{
		prtGraphics(rec);
	      }
	    }
	    UCpclose(Srv);
	  }else{
	    err_set(1,"could not start minimax server.");
	  }
	}else{                /* a regular command, run in background without update */
	  sprintf(logfile,"%s.log",jobname);
	  prtGraphics("running Minimax in background:\n");
	  prtGraphics("%s\n",cmd);
	  status=startCmd(cmd,logfile,"-join",FALSE,"belownormal");
	  if (status) {
	    err_set(1,"Minimax: shell returns with error status %d",status);
	  }
	}
      }else err_set(1,"Minimax: failed to run %s",jobname);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad jobname token: %s",key->text);
    }
  }

  clear_node_chain(&atms);
  free_menu(menu);
  free_Token(key);

}
/*-------------------- cmd_minimax_constraints --------------------*/

static void cmd_minimax_constraints(Node **root)
{
  Token *key,*A[4];
  char *mol_menu,buf[4096];
  Nodechain *c;
  Atom *a,*atm[4];
  int i,n,noffd,noffa,nofft;
  float offd[]={2.0,1.5,1.0,0.5,0.25,0.1,0.05};
  float offa[]={180.0,120.0,90.0,60.0,30.0,15.0,10.0,5.0,1.0,0.5};
  float offt[]={180.0,120.0,90.0,60.0,30.0,15.0,10.0,5.0,1.0,0.5};
  float sw,tol,kh;
  FILE *cf;
  Dspobj *proot;
  Node *sv_def;
  float f,dmin,dmax,dfc,amin,amax,afc,tmin,tmax,tfc;

  noffd=sizeof(offd)/sizeof(float);
  noffa=sizeof(offa)/sizeof(float);
  nofft=sizeof(offt)/sizeof(float);

  key=(Token *)NULL;
  mol_menu=(char *)NULL;
  default_token(constr_file?constr_file:"none");
  key=Get_Kwd("minimax constraints> ",";","none new old");
  if (key->status==TOK_OK) {
    if (key_is("none")) {
      if (mvld(constr_file)) free(constr_file);
      constr_file=(char *)NULL;
    }else if (key_is("new")) {
      cf=(FILE *)NULL;
      key=free_Token(key);
      key=Get_String("name of constraints file> ",";");
      if (key->status==TOK_OK) {
	cf=fopen(key->text,"w");
	if (!cf) err_set(1,"could not open constraints file for write: %s",key->text);
	else {
	  if (mvld(constr_file)) free(constr_file);
	  constr_file=strdup(key->text);
	}
      }else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      if (!mvld(*root)&&cf) {
	mol_menu=Make_TopMolNameList("./ /",DefaultNode);
	key=free_Token(key);
	key=Get_Mols_M("molecule: ",";",mol_menu);
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (c->n==1) {
	    *root=c->s;
	  }else err_set(1,"molecule selection %s is ambiguous.",key->text);
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad molecule selection: %s",key->text);
	}
      }
      if (mvld(*root)&&cf) {
	n=0;
	ForAtoms(a,*root) {
	  a->seqnr=++n;
	  a->flg|=MFLG;
	}
	sv_def=DefaultNode;
	DefaultNode=*root;
	proot=pick_root(glblobj(assocobj(*root)));
	dfc=50.0; afc=1.0, tfc=1.0;
	for (;;) {
	  err_dump();
	  key=free_Token(key);
	  key=Get_Kwd("add constraint> ",";","fixed position '' angle dihedral distance");
	  if (key->status==TOK_OK) {
	    if (key_is("fixed")) {
	      key=free_Token(key);
	      key=Get_Atoms_D("fix atoms> ",";");
	      if (key->status==TOK_OK) {
		c=(Nodechain *)key->ptr;
		if (c->n>0) {
		  fprintf(cf,"Fix\n"); n=0;
		  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
		    if (a->flg&MFLG) {
		      n++; fprintf(cf," %d",a->seqnr);
		      if (n%10==0) fprintf(cf,"\n");
		    }
		  }
		  if (n%10!=0) fprintf(cf,"\n");
		}
		clear_node_chain(c);
		free_node_chain(c);
	      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	    }else if (key_is("position")) {
	      key=free_Token(key);
	      key=Get_Atoms_D("constrain atoms> ",";");
	      if (key->status==TOK_OK) {
		c=(Nodechain *)key->ptr;
		default_token("1.0");
		key=free_Token(key);
		key=Get_Float_M("harmonic force constant: ",";",
				"0.1 0.5 1.0 5.0 10.0 50.0 100.0 500.0 1000.0");
		if (key->status==TOK_OK) {
		  kh=key->f;
		  key=free_Token(key);
		  default_token("0.0");
		  key=Get_Float_M("tolerance: ",";",
				  "0.0 0.05 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0");
		  if (key->status==TOK_OK) {
		    tol=key->f;
		    key=free_Token(key);
		    default_token("never");
		    key=Get_Float_M("harmonic/linear switch at viol.= ",";",
				    "never 0.5 1.0 1.5 2.0 3.0 4.0 5.0 7.0 8.0 9.0 10.0");
		    if (key->status==TOK_NOK) {
		      key=Prs_Kwd(key,"never");
		      if (key->status==TOK_OK) key->f=0.0;
		    }
		    if (key->status==TOK_OK) {
		      sw=key->f;
		      if (c->n>0) {
			fprintf(cf,"XYZ\n"); n=0;
			for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
			  if (a->flg&MFLG) {
			    fprintf(cf," %6d",a->seqnr);
			    for (n=0;n<3;n++) fprintf(cf," %10.5f",a->x[n]);
			    fprintf(cf," %8.2f",kh);
			    if (tol>0.000001||sw>0.000001) fprintf(cf," %8.4f",tol);
			    if (sw>0.000001) fprintf(cf," %8.2f",sw);
			    fprintf(cf,"\n");
			  }
			}
		      }
		    }else if (key->status==TOK_NOK) err_set(1,"bad value: %s",key->text);
		  }else if (key->status==TOK_NOK) err_set(1,"bad tolerance: %s",key->text);
		}else if (key->status==TOK_NOK) err_set(1,"bad force constant: %s",key->text);
		clear_node_chain(c);
		free_node_chain(c);
	      }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	    }else if (key_is("angle")) {
	      n=Get_nAtoms(3,A,"angle constraint ",";");
	      if (n==3) {
		atm[0]=A[0]->ptr; atm[1]=A[1]->ptr; atm[2]=A[2]->ptr;
		free_Token(A[0]); free_Token(A[1]); free_Token(A[2]);
		if ((atm[0]->flg&MFLG)&&(atm[1]->flg&MFLG)&&(atm[2]->flg&MFLG)) {
		  f=cangled(atm[0]->x,atm[1]->x,atm[2]->x);
		  default_token("%0.2f",f);
		  *buf='\0';
		  for (i=0;i<noffa;i++) 
		    if (f+offa[i]>=0) sprintf(buf+strlen(buf)," %0.3f",f-offa[i]);
		  sprintf(buf+strlen(buf)," ' ' %0.2f ' '",f);
		  for (i=noffa-1;i>=0;i--) sprintf(buf+strlen(buf)," %0.3f",f+offa[i]);
		  key=free_Token(key); key=Get_Float_M("omega-min: ",";",buf);
		  if (key->status==TOK_OK) {
		    amin=key->f;
		    f=(amin < f ? f : amin);
		    default_token("%0.2f",f);
		    *buf='\0';
		    for (i=0;i<noffa;i++) 
		      if (f+offa[i]>=amin) sprintf(buf+strlen(buf)," %0.3f",f-offa[i]);
		    sprintf(buf+strlen(buf)," ' ' %0.3f ' '",f);
		    for (i=noffa-1;i>=0;i--) sprintf(buf+strlen(buf)," %0.3f",f+offa[i]);
		    key=free_Token(key); key=Get_Float_M("omega-max: ",";",buf);
		    if (key->status==TOK_OK) {
		      amax=key->f;
		      default_token("%0.1f",afc); key=free_Token(key); 
		      key=Get_Float_M("force-constant: ",";","0.10 0.25 0.50 1.0 2.0 5.0 10.0 25.0 50.0");
		      if (key->status==TOK_OK) {
			afc=key->f;
			fprintf(cf,"Angle %4d %4d %4d   %0.2f %0.2f   %0.3f\n",
				atm[0]->seqnr,atm[1]->seqnr,atm[2]->seqnr,amin,amax,afc);
		      }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		    }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		  }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		}else err_set(1,"bad atom selection (atom not in current molecule)");
	      }else{
		if (n>0) send_header(graphix,"ARMK",0);
		if (A[n]->status==TOK_NOK) 
		  err_set(1,"no such atom: %s in molecule %s",A[n]->text,nodeptr_to_name(buf,*root));
		for (i=0;i<n-1;i++) A[i]=free_Token(A[i]);
	      }
	    }else if (key_is("dihedral")) {
	      n=Get_nAtoms(4,A,"dihedral constraint ",";");
	      if (n==4) {
		atm[0]=A[0]->ptr; atm[1]=A[1]->ptr; atm[2]=A[2]->ptr; atm[3]=A[3]->ptr;
		free_Token(A[0]); free_Token(A[1]); free_Token(A[2]); free_Token(A[3]);
		if ((atm[0]->flg&MFLG)&&(atm[1]->flg&MFLG)&&(atm[2]->flg&MFLG)&&(atm[3]->flg&MFLG)) {
		  f=ctorsd(atm[0]->x,atm[1]->x,atm[2]->x,atm[3]->x);
		  default_token("%0.3f",f);
		  *buf='\0';
		  for (i=0;i<nofft;i++) sprintf(buf+strlen(buf)," %0.3f",f-offt[i]);
		  sprintf(buf+strlen(buf)," ' ' %0.3f ' '",f);
		  for (i=nofft-1;i>=0;i--) sprintf(buf+strlen(buf)," %0.3f",f+offt[i]);
		  key=free_Token(key); key=Get_Float_M("tau-min: ",";",buf);
		  if (key->status==TOK_OK) {
		    tmin=key->f;
		    default_token("%0.3f",f);
		    *buf='\0';
		    for (i=0;i<nofft;i++) sprintf(buf+strlen(buf)," %0.3f",f-offt[i]);
		    sprintf(buf+strlen(buf)," ' ' %0.3f ' '",f);
		    for (i=nofft-1;i>=0;i--) sprintf(buf+strlen(buf)," %0.3f",f+offt[i]);
		    key=free_Token(key); key=Get_Float_M("tau-max: ",";",buf);
		    if (key->status==TOK_OK) {
		      tmax=key->f;
		      default_token("%0.1f",tfc); key=free_Token(key); 
		      key=Get_Float_M("force-constant: ",";","0.10 0.25 0.50 1.0 2.0 5.0 10.0 25.0 50.0");
		      if (key->status==TOK_OK) {
			tfc=key->f;
			fprintf(cf,"Torsion %4d %4d %4d %4d   %0.2f %0.2f   %0.3f\n",
				atm[0]->seqnr,atm[1]->seqnr,atm[2]->seqnr,atm[3]->seqnr,tmin,tmax,tfc);
		      }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		    }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		  }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		}else err_set(1,"bad atom selection (atom not in current molecule)");
	      }else{
		if (n>0) send_header(graphix,"ARMK",0);
		if (A[n]->status==TOK_NOK) 
		  err_set(1,"no such atom: %s in molecule %s",A[n]->text,nodeptr_to_name(buf,*root));
		for (i=0;i<n-1;i++) A[i]=free_Token(A[i]);
	      }
	    }else if (key_is("distance")) {
	      n=Get_nAtoms(2,A,"distance constraint ",";");
	      if (n==2) {
		atm[0]=A[0]->ptr; atm[1]=A[1]->ptr;
		free_Token(A[0]); free_Token(A[1]);
		if ((atm[0]->flg&MFLG)&&(atm[1]->flg&MFLG)) {
		  f=cdist(atm[0]->x,atm[1]->x);
		  default_token("%0.3f",f);
		  *buf='\0';
		  for (i=0;i<noffd;i++) 
		    if (f+offd[i]>=0) sprintf(buf+strlen(buf)," %0.3f",f-offd[i]);
		  sprintf(buf+strlen(buf)," ' ' %0.3f ' '",f);
		  for (i=noffd-1;i>=0;i--) sprintf(buf+strlen(buf)," %0.3f",f+offd[i]);
		  key=free_Token(key); key=Get_Float_M("d-min: ",";",buf);
		  if (key->status==TOK_OK) {
		    dmin=key->f;
		    f=(dmin < f ? f : dmin);
		    default_token("%0.3f",f);
		    *buf='\0';
		    for (i=0;i<noffd;i++) 
		      if (f+offd[i]>=dmin) sprintf(buf+strlen(buf)," %0.3f",f-offd[i]);
		    sprintf(buf+strlen(buf)," ' ' %0.3f ' '",f);
		    for (i=noffd-1;i>=0;i--) sprintf(buf+strlen(buf)," %0.3f",f+offd[i]);
		    key=free_Token(key); key=Get_Float_M("d-max: ",";",buf);
		    if (key->status==TOK_OK) {
		      dmax=key->f;
		      default_token("%0.1f",dfc); key=free_Token(key); 
		      key=Get_Float_M("force-constant: ",";","10.0 20.0 30.0 40.0 50.0 75.0 100.0");
		      if (key->status==TOK_OK) {
			dfc=key->f;
			fprintf(cf,"Dist %4d %4d   %0.3f %0.3f   %0.3f\n",
				atm[0]->seqnr,atm[1]->seqnr,dmin,dmax,dfc);
		      }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		    }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		  }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
		}else err_set(1,"bad atom selection (atom not in current molecule)");
	      }else{
		if (n>0) send_header(graphix,"ARMK",0);
		if (A[n]->status==TOK_NOK) 
		  err_set(1,"no such atom: %s in molecule %s",A[n]->text,nodeptr_to_name(buf,*root));
		for (i=0;i<n-1;i++) A[i]=free_Token(A[i]);
	      }
	    }
	    send_bheader(graphix,"ARMK",0);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad constraint type: %s",key->text);
	    end_parse();
	  }else break;
	  if (end_parse()!=TOK_OK) {
	    err_set(0,"extra characters at end of command ignored.");
	  }
	}
	pick_root(proot);
	DefaultNode=sv_def;
	ForAtoms(a,*root) a->flg&=~MFLG;
	fclose(cf);
      }
    }else if (key_is("old")) {
      key=free_Token(key);
      key=Get_String("name of constraints file> ",";");
      if (key->status==TOK_OK) {
	if (mvld(constr_file)) free(constr_file);
	constr_file=strdup(key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad constraints option: %s",key->text);

  key=free_Token(NULL);
  free_menu(mol_menu);
}
