#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"

extern Node *LibRoot;
extern Node *DefaultNode;
extern int graphix;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define key_is(a) (streq(a,key->key))

#define TRUE 1
#define FALSE 0

#define SEQ_AFTER 1
#define SEQ_FIRST 2
#define SEQ_BEFORE 3
#define SEQ_NEXT 4

#define BLAB_TYPE 1
#define BLAB_ORDER 2

#define LAB_TYPE 1
#define LAB_NAME 2

#define MAXBUF 4096

#define GRP_GREEDY 1
#define GRP_WOOLY  2
#define GRP_HETERO 3

static int   grpMethod=GRP_WOOLY;
static float CTOL=0.02;

static Atomdef *AtpH=(Atomdef *)NULL;

static void cmd_modify_atom(void);
static void cmd_modify_molecule(void);
static void cmd_modify_bond(void);
static void new_atom_labels(Nodechain *c,int mode);
static void new_bond_labels(Bond *B,int mode);
static void cmd_modify_atom_type(void);
static void cmd_modify_atom_position(void);
static void cmd_modify_atom_location(void);
static void cmd_modify_torsion(void);
static void cmd_modify_angle(void);
static void cmd_modify_distance(void);
static void cmd_modify_hydrogens(void);
static void cmd_modify_resnum(void);
static int  cmd_all_charges(Nodechain *c,int label_flg);
static int  cmd_modify_all_sofs(Nodechain *c,int label_flag);
static int  cmd_modify_all_bisos(Nodechain *c,int label_flag);
static Atom *make_new_h(Nodechain *hc,Atom *prv,Atom *root,
			float *a,float *b,float *c,
			float dist,float ang,float tors);
static int   automatic_names(Atom *a,int mode);

static char *mod_kwd = "angle atom bond distance hydrogens molecule mutate "
                       "resnum torsion ' ' label";
static char *atom_kwd = "aromaticity biso charge group ligance location name oopflag "
                        "position q sequence shift sof switch type";
static char *atom_seq_kwd = "first before after";
static char *bond_kwd = "order type";
static char *mol_kwd = "name";

static char *bo_mnem = "0 1 2 3";

static char *grp_menu = "check method select sort switch tolerance automatic "
                        "g1 g2 g3 g4 g5 g6 g7 g8 g9 g10 "
                        "g11 g12 g13 g14 g15 g16 g17 g18 g19 g20 "
                        "g21 g22 g23 g24 g25 g26 g27 g28 g29 g30 "
                        "g31 g32 g33 g34 g35 g36 g37 g38 g39 g40 "
                        "g41 g42 g43 g44 g45 g46 g47 g48 g49 g50 ";

static int status;

/*========================================================================*/


void cmd_modify(void)
{
  Token *key;
  
  status=TRUE;
  
  while (status) {
    start_parse(); err_dump();
    key=Get_Kwd("what: ",";",mod_kwd);
    if (key->status==TOK_OK) {
      if (key_is("label")) {
	cmd_label();
      }else if (key_is("molecule")) {
	cmd_modify_molecule();
      }else if (key_is("mutate")) {
	cmd_mutate();
      }else if (key_is("atom")) {
	cmd_modify_atom();
      }else if (key_is("bond")) {
	cmd_modify_bond();
      }else if (key_is("angle")) {
	cmd_modify_angle();
      }else if (key_is("distance")) {
	cmd_modify_distance();
      }else if (key_is("torsion")) {
	cmd_modify_torsion();
      }else if (key_is("hydrogens")) {
	cmd_modify_hydrogens();
      }else if (key_is("resnum")) {
	cmd_modify_resnum();
      }
    }else if (key->status==TOK_NOK) {
      status=FALSE;
      err_set(1,"bad token: %s",key->text);
      end_parse();
    }else status=FALSE;
    free_Token(key);
    if (end_parse()!=TOK_OK) {
      err_set(1,"extra characters at end of command ignored.");
    }
  }
}
/*-------------------- cmd_modify_atom --------------------*/

static void cmd_modify_atom(void)
{
  char npr[256],buf[MAXBUF],*prmt;
  Nodechain *c;
  Atom *a,*b,*prv,*nxt,*old;
  Node *CruleLib;
  Node *subst;
  int nchgd,mode=0,autom,auto_label_flag,lig,chrg,nsw,oop,arom;
  int k,reparse,nok;
  Token *key;
  char *grpn;
  float Q,ppm;
  char *allmenu="all '' 0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0";
  char *nummenu="0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0";
  char *allbiso="all '' 1.0 2.0 3.0 5.0 10.0 20.0 30.0 40.0 50.0 60.0 70.0 80.0 90.0 100.0"; 
  char *bisomnu="1.0 2.0 3.0 5.0 10.0 20.0 30.0 40.0 50.0 60.0 70.0 80.0 90.0 100.0";

  key=Get_Kwd("what: ",";",atom_kwd);
  if (key->status == TOK_OK) {	    
    if (key_is("type")) {
      cmd_modify_atom_type();
    }else if (key_is("position")) {
      cmd_modify_atom_position();

    }else if (key_is("location")) {
      cmd_modify_atom_location();

    }else if (key_is("ligance")) {
      
      /* modify formal ligance */

      free_Token(key);
      default_token("current");
      key=Get_Kwd("formal ligance: ",";",
		  "'0' '1' '2' '3' '4' '5' '6' '7' '8' '9' '10' ? ' ' "
		  "automatic current");
      if (key->status==TOK_OK) {
	lig=key->n;
	if (key_is("automatic")) lig=-1;
	else if (key_is("current")) lig=-2;
	else if (key_is("?")) lig='?';
	free_Token(key);
	key=Get_Atoms_D("Atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (lig==-1) {
	    for (nok=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	      for (k=0;k<a->nbnd;k++) if (a->bnd[k]->bt==BT_UNKNOWN) break;
	      if (k<a->nbnd||a->chrg=='?') nok++;
	      lig=asg_formlig(a);
	      if (lig>=0) a->lig=lig;
	    }
	    if (nok>0) err_set(0,"Warning: %d atoms have unknown charge or bondtypes.",nok);
	  }else if (lig==-2) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	      for (a->lig=0,k=0;k<a->nbnd;k++) a->lig+=BT_count(a->bnd[k]->bt);
	    }
	  }else{
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->lig=lig;
	  }
	  auto_label_flag=auto_label();
	  send_bheader(graphix,"ARMK",0);
	  if (auto_label_flag) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_LIGANCE,NULL,TRUE);
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }

    }else if (key_is("oopflag")) {
      
      /* modify oop flag */

      free_Token(key);
      key=Get_Kwd("oop flag : ",";","1 2 off");
      if (key->status==TOK_OK) {
	oop=(key->n < 2 ? key->n+1 : 0);
	free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->oop=oop;
	  auto_label_flag=auto_label();
	  send_bheader(graphix,"ARMK",0);
	  if (auto_label_flag) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_OOPFLAG,NULL,TRUE);
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }

    }else if (key_is("aromaticity")) {
      
      /* modify aromaticity flag */

      free_Token(key);
      key=Get_Kwd("aromaticity flag : ",";","on off");
      if (key->status==TOK_OK) {
	arom=(1-key->n);
	free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->arom=arom;
	  auto_label_flag=auto_label();
	  send_bheader(graphix,"ARMK",0);
	  if (auto_label_flag) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_AROMATICITY,NULL,TRUE);
	      if (cPtr(a->cross)) {
	    }
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }

    }else if (key_is("charge")) {
      
      /* modify formal charge */

      autom=FALSE;
      free_Token(key);
      key=Get_Int_M("formal charge : ",";","0 1 2 3 -1 -2 -3 ? ' ' automatic");
      if (key->status==TOK_NOK) {
	key=Prs_Kwd(key,"? automatic");
	if (key->status==TOK_OK) {
	  if (key_is("automatic")) autom=TRUE;
	  else if (key_is("?")) key->n='?';
	  else key->status=TOK_NOK;
	}
      }
      if (key->status==TOK_OK) {
	chrg=key->n;
	free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (autom) {
	    for (nok=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	      for (k=0;k<a->nbnd;k++) if (a->bnd[k]->bt==BT_UNKNOWN) break;
	      if (k<a->nbnd) nok++;
	      a->chrg='?';               /* NEEDED, since asg_formchrg calls sane_formlig */
	      a->chrg=asg_formchrg(a,sane_formlig(a));
	    }
	    if (nok>0) err_set(0,"Warning: %d atoms have unknown bondtypes.",nok);
	  }else for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->chrg=chrg;
	  auto_label_flag=auto_label();
	  send_bheader(graphix,"ARMK",0);
	  if (auto_label_flag) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_CHARGE,NULL,TRUE);
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }

    }else if (key_is("switch")) {
      
      /* select switching atoms...  */

      free_Token(key);
      key=Get_Atoms_D("atoms: ",";");
      send_bheader(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  if (!(a->flg&(1<<1))) {
	    ForAtoms(b,a->up) if (streq(a->grp,b->grp)) {
	      b->cosw=0;
	      b->flg|=(1<<1);
	    }
	    a->cosw=1;
	  }
	}
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  if (a->flg&(1<<1)) {
	    ForAtoms(b,a->up) if (streq(a->grp,b->grp)) {
	      b->flg&=~(1<<1);
	      if (auto_label()) set_atom_label(b,ATM_LAB_SWITCH,NULL,TRUE);
	    }
	  }
	}
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }

    }else if (key_is("group")) {
      
      /* define charge groups */

      grpn=(char *)NULL;
      auto_label_flag=auto_label();
      reparse=FALSE;
      do {
	if (!reparse || !mvld(key)) {
	  key=free_Token(key);
	  if (mvld(grpn)) default_token(grpn);
	  key=Get_Atom_M("group (or atom): ",";",grp_menu);
	  send_bheader(graphix,"ARMK",0);
	}else{
	  key=Prs_Atom(key);
	}
	reparse=FALSE;
	if (key->status==TOK_OK) {
	  a=(Atom *)key->ptr;
	  if (mvld(grpn)) {
	    Strcpy(a->grp,grpn);
	  }else{
	    grpn=Strdup(a->grp);
	  }
	  a->cosw=0;
	  if (auto_label_flag) {
	    set_atom_label(a,ATM_LAB_GROUP,NULL,TRUE);
	  }
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"check select sort switch tolerance automatic method");
	  if (key->status==TOK_OK) {
	    if (key_is("sort")) {
	      key=free_Token(key);
	      key=Get_Atoms_D("sort charge groups for atoms: ",";");
	      send_bheader(graphix,"ARMK",0);
	      if (key->status==TOK_OK) {
		c=(Nodechain *)key->ptr;
		for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
		  if (!(a->flg&(1<<1))) {
		    forAtoms(prv,a->up) if (streq(a->grp,prv->grp)) break;
		    if (mvld(prv)) {
		      prv->flg|=(1<<1); add_to_node_chain((Node *)prv,c);
		      for (b=(Atom *)prv->nxt;mvld(b);b=nxt) {
			nxt=(Atom *)b->nxt;
			if (is_atom(b)) {
			  if (streq(a->grp,b->grp)) {
			    b->flg|=(1<<1); add_to_node_chain((Node *)b,c);
			    unlink_node((Node *)b); b->prv=(Node *)prv;
			    link_node((Node *)b); prv=b;
			  }
			}
		      }
		    }
		  }
		}
		for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->flg&=~(1<<1);
		clear_node_chain(c); free_node_chain(c);
	      }
	    }else if (key_is("tolerance")) {
	      key=free_Token(key);
	      default_token("%0.4f\n",CTOL);
	      key=Get_Float_M("charge tolerance for groups",";","0.0010 0.0025 0.0050 0.0075 0.010 0.020 0.030 0.040 0.050 0.060 0.070 0.080 0.090 0.10");
	      if (key->status==TOK_OK) {
		CTOL=key->f;
	      }else if (key->status==TOK_NOK) err_set(1,"bad float: %s",key->text);
	    }else if (key_is("method")) {
	      key=free_Token(key);
	      default_token("%s",
			    grpMethod==GRP_GREEDY?"greedy"
			    :grpMethod==GRP_WOOLY?"wooly"
			    :grpMethod==GRP_HETERO?"hetero"
			    :"wooly");
	      key=Get_Kwd("method: ",";","greedy hetero wooly");
	      if (key->status==TOK_OK) {
		if (key_is("greedy")) grpMethod=GRP_GREEDY;
		else if (key_is("wooly")) grpMethod=GRP_WOOLY;
		else if (key_is("hetero")) grpMethod=GRP_HETERO;
		else grpMethod=GRP_WOOLY;
	      }else if (key->status==TOK_NOK) err_set(1,"unknown method: %s",key->text);

	    }else if (key_is("automatic")) {
	      key=free_Token(key);
	      key=Get_Atoms_D("assign charge groups to atoms: ",";");
	      send_bheader(graphix,"ARMK",0);
	      if (key->status==TOK_OK) {
		c=(Nodechain *)key->ptr;
		for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
		  if (!(a->flg&(1<<1))) {
		    if (grpMethod==GRP_WOOLY) auto_atom_group_wooly(a->up);
		    else if (grpMethod==GRP_GREEDY) auto_atom_group_greedy(a->up);
		    else if (grpMethod==GRP_HETERO) auto_atom_group_hetero(a->up);
		    else auto_atom_group_greedy(a->up);
		    forAtoms(b,a->up) {
		      b->flg|=(1<<1);
		      add_to_node_chain((Node *)b,c);
		      if (auto_label_flag) {
			set_atom_label(b,ATM_LAB_GROUP,NULL,TRUE);
		      }
		    }
		  }
		}
		for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->flg&=~(1<<1);
		clear_node_chain(c); free_node_chain(c);
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad token: %s\n",key->text);
		key->status=TOK_OK;
	      }
	      
	    }else if (key_is("check")) {
	      key=free_Token(key);
	      key=Get_Atoms_D("check charge groups for atoms: ",";");
	      send_bheader(graphix,"ARMK",0);
	      if (key->status==TOK_OK) {
		c=(Nodechain *)key->ptr;
		for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
		  if (!(a->flg&(1<<1))) {
		    Q=0.0;nsw=0;
		    ForAtoms(b,a->up) {
		      if (streq(a->grp,b->grp)) {
			Q+=b->q;
			if (b->cosw) {
			  nsw++; 
			}
			b->flg|=(1<<1); add_to_node_chain((Node *)b,c);
			if (auto_label_flag) {
			  set_atom_label(b,ATM_LAB_GROUP,NULL,TRUE);
			}
		      }
		    }
		    if (nsw!=1)  {
		      if (nsw>1) {
			prtGraphics(
				    "charge group %4s (%s) : multiple switching atoms\n",
				    a->grp,nodeptr_to_name(buf,a->up));
		      }else if (nsw<1) {
			prtGraphics(
				    "charge group %4s (%s) : no switching atom\n",
				    a->grp,nodeptr_to_name(buf,a->up));
		      }
		    }
		    if (Q < -CTOL || Q > CTOL) {
		      prtGraphics(
				  "charge group %4s (%s) : charge = %6.3f\n",
				  a->grp,nodeptr_to_name(buf,a->up),Q);
		    }
		  }
		}
		for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->flg&=~(1<<1);
		clear_node_chain(c); free_node_chain(c);
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad token: %s\n",key->text);
		key->status=TOK_OK;
	      }
	    }else if (key_is("select")) {
	      key=free_Token(key);
	      key=Get_Atom("select group (pick atom in group): ",";");
	      if (key->status==TOK_OK) {
		if (mvld(grpn)) free(grpn);
		grpn=Strdup(((Atom *)(key->ptr))->grp);
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad token: %s\n",key->text);
	      }
	      key->status=TOK_OK;
	    }else if (key_is("switch")) {
	      while(key->status==TOK_OK) {
		key=free_Token(key);
		key=Get_Atom_M("switch atom: ",";",grp_menu);
		send_bheader(graphix,"ARMK",0);
		if (key->status==TOK_OK) {
		  a=(Atom *)key->ptr;
		  ForAtoms(b,a->up) if (streq(a->grp,b->grp)) {
		    b->cosw=(b==a ? 1 : 0);
		    if (cPtr(b->cross)) {
		      if (auto_label()) {
			set_atom_label(b,ATM_LAB_GROUP,NULL,TRUE);
		      }
		    }
		  }
		}else if (key->status==TOK_NOK) {
		  reparse=TRUE;
		}else{
		  reparse=TRUE;
		}
	      }
	      err_dump();
	      key->status=TOK_OK;
	    }
	  }else{
	    if (mvld(key->text)) {
	      if (strlen(key->text)<sizeof(a->grp)) {
		if (mvld(grpn)) free(grpn);
		grpn=key->text;
		key->text=(char *)NULL;
		key->status=TOK_OK;
	      }else{
		err_set(1,"group name %s too long (max. chars is 4)",key->text);
		key->status=TOK_OK;
	      }
	    }else{
	      err_set(1,"bad token.");
	    }
	  }
	}
	err_dump();
      } while (key->status==TOK_OK);
      if (mvld(grpn)) free(grpn);
      
    }else if (key_is("name")) {
      
      /* modify atom name */
      
      free_Token(key);
      key=Get_Atoms_D("Atoms: ",";");
      send_header(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	nchgd=0;
	auto_label_flag=auto_label();
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  free_Token(key);
	  if (mvld(a->name)) default_token("%s",a->name);
	  sprintf(npr,"new name for %s : ",nodeptr_to_name(buf,(Node *)a));
	  if (cPtr(a->cross)&&a->dsp) {
	    send_bheader(graphix,"ASMK",cPtrSize());
	    send_bcptr(graphix,a->cross);
	  }
	  key=Get_String_M(npr,";","-all -automatic");
	  send_bheader(graphix,"ARMK",0);
	  if (key->status==TOK_EOC) break;
	  if (key->status==TOK_ABORT) break;
	  if (key->status==TOK_OK) {
	    if (kwd_match(key->text,"-automatic",buf)>=0) {
	      key=free_Token(key);
	      key=Get_Kwd("mode: ",";","alarm hseq hheav seq");
	      if (key->status==TOK_OK) {
		if (key_is("hseq")) mode=AUTONM_HSEQ;
		else if (key_is("hheav")) mode=AUTONM_HHEAV;
		else if (key_is("seq")) mode=AUTONM_SEQ;
		else if (key_is("alarm")) mode=AUTONM_ALARM;
		nchgd+=automatic_names(a,mode);
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad automatic name mode: %s",key->text);
	      }
	      a=(Atom *)NULL;
	      break;
	    }else if (kwd_match(key->text,"-all",buf)>=0) {
	      key=free_Token(key);
	      if (mvld(a->name)) default_token("%s",a->name);
	      sprintf(npr,"new name for %s : ",nodeptr_to_name(buf,(Node *)a));
	      key=Get_String(npr,";");
	      if (key->status==TOK_OK) {
		for (;mvld(a);a=(Atom *)a->chain) {
		  set_node_name((Node *)a,key->text);
		  if (auto_label_flag) {
		    set_atom_label(a,ATM_LAB_NAME,NULL,TRUE);
		  }
		  nchgd++;
		}
		break;
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad atom name: %s",key->text);
		break;
	      }else break;
	    }else{
	      set_node_name((Node *)a,key->text);
	      if (auto_label_flag) {
		set_atom_label(a,ATM_LAB_NAME,NULL,TRUE);
	      }
	      nchgd++;
	    }
	  }else{
	    err_set(1,"bad atom name");
	    break;
	  }
	}
	prtGraphics("%d atom names changed.\n",nchgd);
	for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg&=~(1<<1);
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad atomlist specification.");
      }
      
    }else if (key_is("q")) {
      
      /* modify partial atomic charge */
      
      free_Token(key);
      key=Get_Atoms_D("Atoms: ",";");
      send_header(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	nchgd=0;
	auto_label_flag=auto_label();
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  free_Token(key);
	  default_token("%6.2f",a->q);
	  sprintf(npr,"new partial charge for %s : ",nodeptr_to_name(buf,(Node *)a));
	  if (cPtr(a->cross)&&a->dsp) {
	    send_bheader(graphix,"ASMK",cPtrSize());
	    send_bcptr(graphix,a->cross);
	  }
	  key=Get_Float_M(npr,";",(nchgd ? (char *)NULL : "all bondinc fragment MPEOE"));
	  send_bheader(graphix,"ARMK",0);
	  if (key->status==TOK_EOC) break;
	  if (key->status==TOK_ABORT) break;
	  if (key->status==TOK_OK) {
	    a->q=key->f;
	    if (auto_label_flag) {
	      set_atom_label(a,ATM_LAB_Q,NULL,TRUE);
	    }
	    nchgd++;
	  }else {
	    if (nchgd==0) key=Prs_Kwd(key,"-automatic all bondinc fragment MPEOE");
	    if (key->status==TOK_OK) {
	      if (key_is("all")) nchgd=cmd_all_charges(c,auto_label_flag);
	      else if (key_is("-automatic")) nchgd=bondinc_charges(c,auto_label_flag); /*compatibility with old syntax */
	      else if (key_is("MPEOE")) nchgd=MPEOEcharges(c,auto_label_flag);
	      else if (key_is("bondinc")) nchgd=bondinc_charges(c,auto_label_flag);
	      else if (key_is("fragment")) {
		key=free_Token(key);
		CruleLib=find_name(LibRoot,"CruleLib",OBJ_LIB);
		key=Get_RuleLib(CruleLib,"charge fragment set: ",";");
		if (key->status==TOK_OK) {
		  nchgd=fragment_charges(c,(Lib *)key->ptr,auto_label_flag);
		}else if (key->status==TOK_NOK) {
		  err_set(1,"bad fragment set name: %s",key->text);
		}
	      }
	      break;
	    }else{
	      err_set(1,"bad partial charge: %s",key->text);
	      break;
	    }
	  }
	}
	err_dump();
	prtGraphics("%d partial charge%s modified.\n",nchgd,nchgd==1?"":"s");
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	prtGraphics("bad atomlist specification.\n");
      }
      
    }else if (key_is("sequence")) {
      
      /* modify atom sequence within a substructure (e.g. residue) */
      
      key->status=TOK_OK;
      subst=(Node *)NULL;
      prv=nxt=old=(Atom *)NULL;
      mode=SEQ_FIRST; prmt="first atom: ";
      
      do {
	err_dump();
	free_Token(key);
	key=Get_Atom_M(prmt,";",atom_seq_kwd);
	if (key->status==TOK_OK) {
	  nxt=(Atom *)key->ptr;
	  if (old) {
	    if (old!=nxt&&cPtr(old->cross)&&old->dsp) {
	      send_bheader(graphix,"ARMK",cPtrSize());
	      send_bcptr(graphix,old->cross);
	    }
	  }
	  old=nxt;
	  switch(mode) {
	  case SEQ_AFTER:
	    prv=nxt;
	    subst=prv->up;
	    prmt="next atom: ";
	    mode=SEQ_NEXT;
	    break;
	  case SEQ_BEFORE:
	    prv=(Atom *)nxt->prv;
	    subst=nxt->up;
	    prmt="next atom: ";
	    mode=SEQ_NEXT;
	    break;
	  case SEQ_FIRST:
	    subst=nxt->up;
	    unlink_node((Node *)nxt);
	    if (subst->fst) nxt->nxt=subst->fst;
	    else nxt->up=subst;
	    link_node((Node *)nxt);
	    prv=nxt;
	    prmt="next atom: ";
	    mode=SEQ_NEXT;
	    break;
	  case SEQ_NEXT:
	    if (subst!=nxt->up) {
	      prv=(Atom *)NULL;
	      subst=nxt->up;
	    }
	    if (nxt!=prv) {
	      unlink_node((Node *)nxt);
	      if (mvld(prv)) {
		nxt->prv=(Node *)prv;
	      }else{
		if (mvld(subst->fst)) nxt->nxt=subst->fst;
		else nxt->up=subst;
	      }
	      link_node((Node *)nxt);
	      prv=nxt;
	    }
	    prmt="next atom: ";
	    mode=SEQ_NEXT;
	    break;
	  default:
	    break;
	  }
	}else if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,atom_seq_kwd);
	  if (key->status==TOK_OK) {
	    if (key_is("first")) {
	      mode=SEQ_FIRST;
	      prmt="first atom: ";
	    }else if (key_is("after")) {
	      mode=SEQ_AFTER;
	      prmt="insert after atom: ";
	    }else if (key_is("before")) {
	      mode=SEQ_BEFORE;
	      prmt="insert before atom: ";
	    }
	    if (key->status==TOK_NOK) {
	      err_set(1,"bad seq-token.");
	    }
	  }
	}
      } while(key->status==TOK_OK);
      send_bheader(graphix,"ARMK",0);

    }else if (key_is("shift")) {
      
      /* modify chemical shift */

      free_Token(key);
      key=Get_Float("ppm : ",";");
      if (key->status==TOK_OK) {
	ppm=key->f;
	free_Token(key);
	key=Get_Atoms_D("atom(s): ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (a=(Atom *)c->s;a;a=(Atom *)a->chain) a->shift=ppm;
	  auto_label_flag=auto_label();
	  send_bheader(graphix,"ARMK",0);
	  if (auto_label_flag) {
	    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_SHIFT,NULL,TRUE);
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  prtGraphics("bad token: %s\n",key->text);
	}
      }else if (key->status==TOK_NOK) {
	prtGraphics("bad token: %s\n",key->text);
      }

    }else if (key_is("sof")) {

      /* modify atomic site occupancy factor */
      
      free_Token(key);
      key=Get_Atoms_D("Atoms: ",";");
      send_header(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	nchgd=0;
	auto_label_flag=auto_label();
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  free_Token(key);
	  default_token("%0.2f",a->x[ATM_SOF]);
	  sprintf(npr,"new site occupancy factor for %s : ",nodeptr_to_name(buf,(Node *)a));
	  if (cPtr(a->cross)&&a->dsp) {
	    send_bheader(graphix,"ASMK",cPtrSize());
	    send_bcptr(graphix,a->cross);
	  }
	  key=Get_Float_M(npr,";",(nchgd ? nummenu : allmenu));
	  send_bheader(graphix,"ARMK",0);
	  if (key->status==TOK_EOC) break;
	  if (key->status==TOK_ABORT) break;
	  if (key->status==TOK_OK) {
	    a->x[ATM_SOF]=key->f;
	    if (auto_label_flag) {
	      set_atom_label(a,ATM_LAB_SOF,NULL,TRUE);
	    }
	    nchgd++;
	  }else {
	    if (nchgd==0) key=Prs_Kwd(key,"all");
	    if (key->status==TOK_OK) {
	      if (key_is("all")) nchgd=cmd_modify_all_sofs(c,auto_label_flag);
	      break;
	    }else{
	      prtGraphics("bad atomic site occupancy factor: %s\n",key->text);
	      break;
	    }
	  }
	}
	err_dump();
	prtGraphics("%d site occupancy factors modified.\n",nchgd);
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	prtGraphics("bad atomlist specification.\n");
      }
      
    }else if (key_is("biso")) {

      /* modify atomic isotropic temperature factor */
      
      free_Token(key);
      key=Get_Atoms_D("Atoms: ",";");
      send_header(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	nchgd=0;
	auto_label_flag=auto_label();
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  free_Token(key);
	  default_token("%0.2f",a->x[ATM_BISO]);
	  sprintf(npr,"new isotropic temperature factor for %s : ",nodeptr_to_name(buf,(Node *)a));
	  if (cPtr(a->cross)&&a->dsp) {
	    send_bheader(graphix,"ASMK",cPtrSize());
	    send_bcptr(graphix,a->cross);
	  }
	  key=Get_Float_M(npr,";",(nchgd ? bisomnu : allbiso));
	  send_bheader(graphix,"ARMK",0);
	  if (key->status==TOK_EOC) break;
	  if (key->status==TOK_ABORT) break;
	  if (key->status==TOK_OK) {
	    set_atom_biso(a,key->f);
	    if (auto_label_flag) {
	      set_atom_label(a,ATM_LAB_BISO,NULL,TRUE);
	    }
	    nchgd++;
	  }else {
	    if (nchgd==0) key=Prs_Kwd(key,"all");
	    if (key->status==TOK_OK) {
	      if (key_is("all")) nchgd=cmd_modify_all_bisos(c,auto_label_flag);
	      break;
	    }else{
	      prtGraphics("bad isotropic temperature factor: %s\n",key->text);
	      break;
	    }
	  }
	}
	err_dump();
	prtGraphics("%d isotropic temperature factors modified.\n",nchgd);
	clear_node_chain(c); free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	prtGraphics("bad atomlist specification.\n");
      }
    }
      
  }else if (key->status==TOK_NOK) {
    prtGraphics("unknown option: %s\n",key->text);
  }

  free_Token(key);
}
/*-------------------- new_bond_labels --------------------*/

static void new_bond_labels(Bond *B,int mode)
{
  Bond *b;
  int n;

  for (n=0,b=B;mvld(b);b=b->chain) n++;

  if (n>0) {
    lock_evtq();
    switch (mode) {
    case BLAB_ORDER:
      for (b=B;mvld(b);b=b->chain) set_bond_label(b,BND_LAB_ORDER,NULL);
      break;
    case BLAB_TYPE:
      for (b=B;mvld(b);b=b->chain) set_bond_label(b,BND_LAB_TYPE,NULL);
      break;
    }
    unlock_evtq();
  }
}
/*-------------------- new_atom_labels --------------------*/

static void new_atom_labels(Nodechain *c,int mode)
{
  Atom *a;

  if (c->n>0) {
    lock_evtq();
    switch (mode) {
    case LAB_NAME:
      for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_NAME,NULL,TRUE);
      break;
    case LAB_TYPE:
      for (a=(Atom *)c->s;a;a=(Atom *)a->chain) set_atom_label(a,ATM_LAB_TYPE,NULL,TRUE);
      break;
    }
    unlock_evtq();
  }
}
/*-------------------- cmd_modify_atom_type --------------------*/

static void cmd_modify_atom_type(void)
{
  Nodechain *c,chgd;
  Node *ResLib,*PruleLib;
  Atom *a,*next_atom;
  Atomdef *new_type;
  int nmnem,loop,nunk;
  Token *key;
      
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  PruleLib=find_name(LibRoot,"PruleLib",OBJ_LIB);

  loop=TRUE;

  while (loop) {
    nmnem=0;
    if (mvld(PruleLib)) load_menu_buttons(nmnem++,"automatic");
    forAtomdefs(new_type,ResLib) {
      load_menu_buttons(nmnem++,new_type->name);
    }
    key=Get_String("type: ",";");
    clear_menu();
    if (key->status==TOK_OK) {
      new_type=find_atomdef_name(ResLib,key->text);
      if (mvld(new_type)) {
	free_Token(key);
	key=Get_Atoms_D("Atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  chgd.s=chgd.e=(Node *)NULL; chgd.n=0;
	  for (a=(Atom *)c->s;a;a=next_atom) {
	    next_atom=(Atom *)a->chain;
	    a->chain=(Node *)NULL;
	    if (a->atp!=new_type && cPtr(a->cross)) {
	      if (chgd.e) chgd.e->chain=(Node *)a;
	      else chgd.s=(Node *)a;
	      chgd.e=(Node *)a;
	      chgd.n++;
	      a->chain=(Node *)NULL;
	    }
	    if (a->color < 0) a->color = -new_type->color;
	    set_atom_type(a,new_type);
	  }
	  mdl_repl_atom_chain(&chgd);
	  if (auto_label()) new_atom_labels(&chgd,LAB_TYPE);
	  prtGraphics("%d atom%s changed type.\n",c->n,(c->n==1?"":"s"));
	  free_node_chain(c); clear_node_chain(&chgd);
	}else if (key->status==TOK_NOK) {
	  prtGraphics("bad atom list.\n");
	  end_parse();
	}
	send_bheader(graphix,"ARMK",0);
      }else{
	key=Prs_Kwd(key,"automatic");
	if (key->status==TOK_OK) {
	  nunk=automatic_types((Nodechain *)NULL);
	  if (nunk>0) err_set(1,"%d atom%s could not be assigned.",nunk,(nunk==1?"":"s"));
	}else err_set(1,"type definition not found: %s\n",key->text);
	end_parse(); err_dump();
      }
    }else if (key->status==TOK_NOK) {
      prtGraphics("bad type specification: %s\n",key->text);
      loop=FALSE;
    }else{
      loop=FALSE;
    }
  }

  free_Token(key);

}
/*-------------------- cmd_modify_bond --------------------*/

static void cmd_modify_bond(void)
{
  Bond *b,*bn,*B;
  Token *key;
  int i,l,bt,bo;
  char bt_menu[1024];

  key=Get_Kwd("what: ",";",bond_kwd);
  if (key->status==TOK_OK) {
    if (key_is("order")) {
      free_Token(key);
      key=Get_Kwd("bond order: ",";",bo_mnem);
      if (key->status==TOK_OK) {
	bo=key->n;
	free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  B=(Bond *)key->ptr;
	  for (b=B;b;b=b->chain) b->bo=bo;
	  if (auto_label()) new_bond_labels(B,BLAB_ORDER);
	  for (b=B;b;b=bn) {
	    bn=b->chain;
	    b->chain=(Bond *)NULL;
	  }
	}else if (key->status==TOK_NOK) {
	  prtGraphics("bad token: %s\n",key->text);
	}
      }else if (key->status==TOK_NOK) {
	prtGraphics("bad token: %s\n",key->text);
      }
    }else if (key_is("type")) {
      free_Token(key);
      for (l=0,i=0;i<MX_OBJ_BTYPES;i++) {
	sprintf(bt_menu+l,"%s ",BT_mnem(i));
	l=strlen(bt_menu);
      }
      sprintf(bt_menu+l," ' ' automatic");
      key=Get_Kwd("bond type: ",";",bt_menu);
      if (key->status==TOK_OK) {
	bt=BT_number(key->key);
	free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  B=(Bond *)key->ptr;
	  if (bt<0) set_bt_automatic(B);
	  else for (b=B;b;b=b->chain) set_bt(b,bt);
	  if (auto_label()) new_bond_labels(B,BLAB_TYPE);
	  for (b=B;b;b=bn) {
	    bn=b->chain;
	    b->chain=(Bond *)NULL;
	  }
	}else if (key->status==TOK_NOK) {
	  prtGraphics("bad token: %s\n",key->text);
	}
      }else if (key->status==TOK_NOK) {
	prtGraphics("bad token: %s\n",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    prtGraphics("bad token %s\n",key->text);
  }
  free_Token(key);
  send_header(graphix,"ARMK",0);
}
/*-------------------- cmd_modify_angle --------------------*/

static void cmd_modify_angle(void)
{
  Token *atm[3],*key;
  Atom *a[3],*an;
  vec_3 v[3];
  int i,n;
  Dspobj *gob;
  float told,tnew;
  Nodechain c;
  mat_4 rot;
  float ep[3];

  key=(Token *)NULL;
  for (i=0;i<3;i++) atm[i]=(Token *)NULL;
  n=Get_nAtoms(3,atm,"modify angle",";");
  send_bheader(graphix,"ARMK",0);
  if (n==3) {
    for (i=0;i<3;i++) {
      a[i]=atm[i]->ptr; 
      atm[i]=free_Token(atm[i]);
    }
    gob=glblobj(assocobj((Node *)a[0]));
    for (i=1;i<3;i++) if (gob!=glblobj(assocobj((Node *)a[i]))) break;
    if (i<3) {
      err_set(1,"three atoms must be in same global object.\n");
    }else{
      init_node_chain(&c);
      a[1]->chain=(Node *)a[1];
      add_to_node_chain((Node *)a[2],&c);
      connected_atoms(&c);
      a[1]->chain=(Node *)NULL;
      if (!in_chain(a[0],&c)) {
	told=cangled(a[0]->x,a[1]->x,a[2]->x);
	default_token("%0.3f",told);
	key=Get_Float_M("new angle: ",";","15 30 45 60 90 105 109.5 120.0 180.0");
	if (key->status==TOK_OK) {
	  if (key->def) tnew=told;
	  else tnew=key->f;
	  if (tnew==told) {
	    prtGraphics("angle kept at old value.\n");
	  }else{
	    for (i=0;i<3;i++) v[i]=vec3(a[i]->x);
	    v[0]=v3mv3(v[0],v[1]); v[2]=v3mv3(v[2],v[1]);
	    v32arr(ep,v3pv3(v[1],v3xv3(v[0],v[2])));
	    rot=rotax(a[1]->x,ep,tnew-told);
	    for (an=(Atom *)c.s;an;an=(Atom *)an->chain) transform_atom(rot,an);
	    send_newx(&c);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad floating point value: %s\n",key->text);
	}
	key=free_Token(key);
      }else{
	err_set(1,"first atom is connected to second atom.");
      }
      clear_node_chain(&c);
    }
  }else if (atm[n]->status==TOK_NOK) {
    err_set(1,"bad token: %s\n",atm[n]->text);
  }
  for (i=0;i<3;i++) free_Token(atm[i]);
}
/*-------------------- cmd_modify_distance --------------------*/

static void cmd_modify_distance(void)
{
  Token *atm[2],*key;
  Atom *a[2],*an;
  int i,n;
  Dspobj *gob;
  float told,tnew;
  Nodechain c;
  vec_3 v;

  key=(Token *)NULL;
  for (i=0;i<2;i++) atm[i]=(Token *)NULL;
  n=Get_nAtoms(2,atm,"modify distance",";");
  send_bheader(graphix,"ARMK",0);
  if (n==2) {
    a[0]=atm[0]->ptr; a[1]=atm[1]->ptr;
    gob=glblobj(assocobj((Node *)a[0]));
    if (gob==glblobj(assocobj((Node *)a[1]))) {
      init_node_chain(&c);
      a[0]->chain=(Node *)a[0];
      add_to_node_chain((Node *)a[1],&c);
      connected_atoms(&c);
      a[0]->chain=(Node *)NULL;
      if (!in_chain(a[0],&c)) {
	told=cdist(a[0]->x,a[1]->x);
	default_token("%0.3f",told);
	key=Get_Float_M("new distance: ",";","0.5 1.0 1.08 1.2 1.3 1.4 1.5 1.54 1.6 1.7 1.8 1.9 2.0");
	if (key->status==TOK_OK) {
	  if (key->def) tnew=told;
	  else tnew=key->f;
	  if (tnew==told) {
	    prtGraphics("distance kept at old value.\n");
	  }else{
	    v=sxv3(tnew-told,v3n(v3mv3(vec3(a[1]->x),vec3(a[0]->x))));
	    for (an=(Atom *)c.s;an;an=(Atom *)an->chain) v32arr(an->x,v3pv3(v,vec3(an->x)));
	    send_newx(&c);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad floating point value: %s\n",key->text);
	}
	key=free_Token(key);
      }else{
	err_set(1,"first atom is connected to second atom.");
      }
      clear_node_chain(&c);
    }else{
      err_set(1,"both atoms must be in same global object.\n");
    }
  }else if (atm[n]->status==TOK_NOK) {
    err_set(1,"bad token: %s\n",atm[n]->text);
  }
  for (i=0;i<2;i++) free_Token(atm[i]);
}
/*-------------------- cmd_modify_torsion --------------------*/

static void cmd_modify_torsion(void)
{
  Token *atm[4],*key;
  Atom *a[4],*an;
  int i,n;
  Dspobj *gob;
  float told,tnew;
  Nodechain c;
  mat_4 rot;

  for (i=0;i<4;i++) atm[i]=(Token *)NULL;
  n=Get_nAtoms(4,atm,"modify torsion",";");
  send_bheader(graphix,"ARMK",0);
  if (n==4) {
    for (i=0;i<4;i++) a[i]=atm[i]->ptr; 
    gob=glblobj(assocobj((Node *)a[0]));
    for (i=1;i<4;i++) if (gob!=glblobj(assocobj((Node *)a[i]))) break;
    if (i<4) {
      err_set(1,"four atoms must be in same global object.\n");
    }else{
      init_node_chain(&c);
      a[1]->chain=(Node *)a[1];
      add_to_node_chain((Node *)a[2],&c);
      add_to_node_chain((Node *)a[3],&c);
      connected_atoms(&c);
      a[1]->chain=(Node *)NULL;
      if (in_chain(a[0],&c)) {
	err_set(1,"precursor connected to third or fourth atom.");
      }else{
	for (i=0;i<a[1]->nbnd;i++) {
	  an=neighbour(i,a[1]);
	  if (in_chain(an,&c)&&an!=a[2]) break;
	}
	if (i<a[1]->nbnd) {
	  err_set(1,"nogrok topology for modify torsion.");
	}else{
	  told=ctorsd(a[0]->x,a[1]->x,a[2]->x,a[3]->x);
	  default_token("%7.2f",told);
	  key=Get_Float("new torsion angle: ",";");
	  if (key->status==TOK_OK) {
	    if (key->def) tnew=told;
	    else tnew=key->f;
	    if (tnew==told) {
	      prtGraphics("torsion angle kept at old value.\n");
	    }else{
	      rot=rotax(a[1]->x,a[2]->x,tnew-told);
	      for (an=(Atom *)c.s;an;an=(Atom *)an->chain) transform_atom(rot,an);
	      send_newx(&c);
	    }
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad floating point token: %s",key->text);
	  }
	  key=free_Token(key);
	}
      }
      clear_node_chain(&c);
    }
  }else if (atm[n]->status==TOK_NOK) {
    err_set(1,"bad token: %s",atm[n]->text);
  }
  for (i=0;i<4;i++) free_Token(atm[i]);
}
/*-------------------- cmd_modify_hydrogens --------------------*/

static void cmd_modify_hydrogens(void)
{
  Token *key;
  int k,nok;
  Nodechain *ca,ch;
  Atom *a;
  Lib  *reslib;
  Node *ResLib;
  Atomdef *default_type;

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open residue library");
    return;
  }

  reslib=(Lib *)ResLib;
  default_type=find_atomdef_name((Node *)reslib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  AtpH=(Atomdef *)NULL;
  key=Get_Atoms_D("add hydrogens to atoms: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    if (!mvld(AtpH)) AtpH=find_atomdef_name((Node *)reslib,"H.?");
    if (!mvld(AtpH)) AtpH=find_atomdef_name((Node *)reslib,"unk");
    ca=key->ptr;
    ch.n=0; ch.s=(Node *)NULL; ch.e=(Node *)NULL;
    for (nok=0,a=(Atom *)ca->s;mvld(a);a=(Atom *)ca->s) {
      if (a->lig=='?') {
	for (k=0;k<a->nbnd;k++) if (a->bnd[k]->bt==BT_UNKNOWN) break;
	if (k<a->nbnd||a->chrg=='?') nok++;
      }
      ca->s=a->chain;
      a->chain=(Node *)NULL;
      add_hydrogens(a,&ch);
    }
    if (ch.n>0) send_atom_chain(&ch);
    prtGraphics("%d hydrogens added.\n",ch.n);
    clear_node_chain(&ch);
    free_node_chain(ca);
    if (nok>0) err_set(0,"Warning: %d atoms have unknown charge or bondtypes.",nok);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- add_hydrogens --------------------*/

#define UNKNOWN 0
#define SP1  1
#define SP2  2
#define SP3  3
#define CIS  4

#define AH_NELT 10

#define ELT_C  0
#define ELT_H  1
#define ELT_N  2
#define ELT_O  3
#define ELT_S  4
#define ELT_Br 5
#define ELT_Cl 6
#define ELT_F  7
#define ELT_P  8
#define ELT_Si 9

static char *elt[AH_NELT] = { "C", "H", "N", "O", "S", "Br", "Cl", "F", "P", "Si" };

void add_hydrogens(Atom *a,Nodechain *hc)
{
  int lig,chs,nh,nb,j,k,mode,bn[10];
  float dist=0,ang=0,tors,pt[3];
  vec_3 va,vb,vc;
  Atom *an[10],*b,*c=(Atom *)NULL,*prv;

  /* determine number of hydrogens to add */

  lig=sane_formlig(a);
  if (lig=='?') return;

  for (nb=0,j=0;nb<10&&j<a->nbnd;j++) {
    if (BT_count(a->bnd[j]->bt) > 0) bn[nb++]=j;
  }
  nh=lig-nb;

  if (nh<=0) return;        /* nothing to add */
  for (j=0;j<nb;j++) an[j]=neighbour(bn[j],a);

  /* determine geometry */

  for (chs=0;chs<AH_NELT;chs++) if (streq(a->atp->chs,elt[chs])) break;
  mode=UNKNOWN;
  switch (chs) {
  case ELT_C:
    if (lig==4) {dist=1.08; ang=109.5; mode=SP3;}
    else if (lig==3) {dist=1.05; ang=120.0; mode=SP2;}
    else if (lig==2) {dist=1.05; ang=180.0; mode=SP1;}
    else {dist=1.05; ang=180.0; mode=SP1;}
    break;
  case ELT_H:
    dist=0.85; ang=180.0; mode=SP1;
    break;
  case ELT_N:
    if (lig==4) {dist=1.06; ang=109.5; mode=SP3;}
    else if (lig==3) {
      for (j=0;j<nb;j++) {
	for (k=0;k<an[j]->nbnd;k++) {
	  if (an[j]->bnd[k]->bt==BT_DOUBLE||an[j]->bnd[k]->bt==BT_AROMATIC) break;
	}
	if (k<an[j]->nbnd) break;
      }
      if (j<nb) {
	dist=1.02;
	ang=120.0;
	mode=SP2;
      }else{
	dist=1.04;
	ang=109.5;
	mode=SP3;
      }
    }else if (lig==2) {
      if (a->bnd[bn[0]]->bt==BT_TRIPLE) {
	dist=1.02;
	ang=180.0;
	mode=SP1;
      }else{
	dist=1.02;
	ang=120.0;
	mode=SP2;
      }
    }else{
      dist=1.02;
      ang=180;
      mode=SP1;
    }
    break;
  case ELT_O:
    if (nh==2) {
      dist=1.0;
      ang=100.0;
      mode=SP3;
    }else if (nh==1) {
      if (nb==1) {
	for (k=0;k<an[0]->nbnd;k++) {
	  if (an[0]->bnd[k]==a->bnd[bn[0]]) continue;
	  if (an[0]->bnd[k]->bt==BT_DOUBLE||an[0]->bnd[k]->bt==BT_AROMATIC) break;
	}
	if (k<an[0]->nbnd) {
	  an[1]=neighbour(k,an[0]);
	  mode=CIS;
	  ang=120.0;
	  dist=1.0;
	}else{
	  dist=1.0;
	  ang=100.0;
	  mode=SP3;
	}
      }
    }
    break;
  case ELT_S:
    dist=1.2;
    ang=95;
    mode=SP3;
    break;
  case ELT_Br:
    dist=1.2;
    ang=180;
    mode=SP1;
    break;
  case ELT_Cl:
    dist=1.2;
    ang=180;
    mode=SP1;
    break;
  case ELT_F:
    dist=1.0;
    ang=100.0;
    mode=SP1;
    break;
  case ELT_P:
    dist=1.2;
    ang=109.5;
    mode=SP3;
    break;
  case ELT_Si:
    dist=1.475;
    if (lig>3) { mode=SP3; ang=109.5; }
    else if (lig>2) {mode=SP2; ang=120.0;}
    else {mode=SP1; ang=180.0;}
    break;
  default:
    dist=1.1;
    ang=109.5;
    mode=SP3;
    break;
  }

  /* add hydrogens */

  prv=a;
  switch (mode) {

  case SP1:
    tors=0.0; 
    if (nb==0) {
      c=a; b=a;
      prv=make_new_h(hc,prv,a,a->x,a->x,a->x,dist,ang,tors);
      if (nh<2) break;
      prv=make_new_h(hc,prv,a,prv->x,prv->x,a->x,dist,ang,tors);
    }else{
      c=an[0]; b=an[0];
      prv=make_new_h(hc,prv,a,an[0]->x,an[0]->x,a->x,dist,ang,tors);
    } 
    break;

  case SP2:
    tors=180.0;
    if (nb==0) {
      prv=make_new_h(hc,prv,a,a->x,a->x,a->x,dist,ang,tors);
      if (nh<2) break;
      c=prv;
      prv=make_new_h(hc,prv,a,prv->x,prv->x,a->x,dist,ang,tors);
      if (nh<3) break;
      prv=make_new_h(hc,prv,a,c->x,prv->x,a->x,dist,ang,tors);
    }else if (nb==1) {
      b=an[0];
      for (j=0;j<b->nbnd;j++) if ((c=neighbour(j,b))!=a) break;
      if (j>=b->nbnd) c=b;
      prv=make_new_h(hc,prv,a,c->x,b->x,a->x,dist,ang,tors);
      if (nh<2) break;
      prv=make_new_h(hc,prv,a,prv->x,b->x,a->x,dist,ang,tors);
    }else if (nb==2) {
      ang=0.5*(360.0-cangled(an[0]->x,a->x,an[1]->x));
      prv=make_new_h(hc,prv,a,an[0]->x,an[1]->x,a->x,dist,ang,tors);
    }
    break;

  case SP3:
    if (nb==0) {
      tors=120;
      prv=make_new_h(hc,prv,a,a->x,a->x,a->x,dist,ang,tors);
      if (nh<2) break;
      c=prv;
      prv=make_new_h(hc,prv,a,prv->x,prv->x,a->x,dist,ang,tors);
      if (nh<3) break;
      b=prv;
      prv=make_new_h(hc,prv,a,c->x,prv->x,a->x,dist,ang,tors);
      if (nh<4) break;
      prv=make_new_h(hc,prv,a,c->x,prv->x,a->x,dist,ang,tors);
    }else if (nb==1) {
      tors=60;
      b=an[0];
      for (j=0;j<b->nbnd;j++) if ((c=neighbour(j,b))!=a) break;
      if (j>=b->nbnd) c=b;
      prv=make_new_h(hc,prv,a,c->x,b->x,a->x,dist,ang,tors);
      if (nh<2) break;
      prv=make_new_h(hc,prv,a,c->x,b->x,a->x,dist,ang,tors+120.0);
      if (nh<3) break;
      prv=make_new_h(hc,prv,a,c->x,b->x,a->x,dist,ang,tors-120.0);
    }else if (nb==2) {
      tors=90;
      va=v3n(v3mv3(vec3(an[0]->x),vec3(a->x)));
      vb=v3n(v3mv3(vec3(an[1]->x),vec3(a->x)));
      v32arr(pt,v3mv3(v3mv3(vec3(a->x),va),vb));
      prv=make_new_h(hc,prv,a,an[0]->x,pt,a->x,dist,0.5*ang,tors);
      if (nh<2) break;
      prv=make_new_h(hc,prv,a,an[1]->x,pt,a->x,dist,0.5*ang,tors);
    }else if (nb==3) {
      tors=0.0;
      va=v3n(v3mv3(vec3(an[0]->x),vec3(a->x)));
      vb=v3n(v3mv3(vec3(an[1]->x),vec3(a->x)));
      vc=v3n(v3mv3(vec3(an[2]->x),vec3(a->x)));
      va=v3pv3(v3pv3(va,vb),vc);
      if (v3l(va)<0.00005) va=v3xv3(vc,vb);
      ang=180.; 
      v32arr(pt,v3pv3(vec3(a->x),va));
      prv=make_new_h(hc,prv,a,pt,pt,a->x,dist,ang,tors);
    }
    break;

  case CIS:
    prv=make_new_h(hc,prv,a,an[1]->x,an[0]->x,a->x,dist,ang,0.0);
    if (nh<2) break;
    prv=make_new_h(hc,prv,a,an[1]->x,an[0]->x,a->x,dist,ang,180.0);
    break;

  case UNKNOWN:
  default:
    break;
  }
}
/*-------------------- make_new_h --------------------*/

static Atom *make_new_h(Nodechain *hc,Atom *prv,Atom *root,
			float *a,float *b,float *c,
			float dist,float ang,float tors)
{
  Atom *n;
  Bond *nb;
  
  if (!is_atom(root)) return((Atom *)NULL);

  n=(Atom *)make_node(OBJ_ATOM);
  if (!mvld(n)) return(n);

  if (n->mxyz>0) {
    cnstr(n->x,a,b,c,dist,ang,tors);
    n->nxyz++;
  }
  set_node_name((Node *)n,"H");     /* should make unique name based on k,N */
  n->prv=(Node *)prv;
  link_node((Node *)n);
  set_atom_type(n,AtpH);
  n->dsp=root->dsp;
  n->lig=1;
  if (mvld(n->atp)) {
    if (root->color < 0) n->color=-n->atp->color;
    else if (root->color==root->atp->color) n->color=n->atp->color;
    else n->color=root->color;
  }
  nb=add_bond_raw(root,n);
  if (mvld(nb)) set_bt(nb,BT_SINGLE);
  add_to_node_chain((Node *)n,hc);
  return(n);
}
/*-------------------- cmd_modify_atom_position --------------------*/

static void cmd_modify_atom_position(void)
{
  Token *key;
  Atom *a;
  mat_4 mvm;
  float *pt1,*pt2;
  vec_3 newx;
  int loop;

  loop=TRUE;

  while(loop) {
    key=Get_Atom("atom: ",";");
    if (key->status==TOK_OK) {
      a=(Atom *)key->ptr;
      free_Token(key);
      key=Get_XYZ("new position: ",";");
      if (key->status==TOK_OK) {
	pt1=(float *)key->ptr; pt2=(float *)key->ptr+3;
	mvm=obj_transform((Node *)a);
	newx=m4xv3(mvm,vec3(a->x));
	mvm=mmxinv(mvm);
	newx.c[0]=pt1[0]+(pt2[0]-pt1[0])*(pt1[2]-newx.c[2])/(pt1[2]-pt2[2]);
	newx.c[1]=pt1[1]+(pt2[1]-pt1[1])*(pt1[2]-newx.c[2])/(pt1[2]-pt2[2]);
	v32arr(a->x,m4xv3(mvm,newx));
	if (cPtr(a->cross)) {
	  send_bheader(graphix,"NEWX",cPtrSize()+3*sizeof(float));
	  send_bcptr(graphix,a->cross);
	  send_bfloata(graphix,3,a->x);
	}
	free(key->ptr);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }else{
	loop=FALSE;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
    }else{
      loop=FALSE;
    }
    send_header(graphix,"ARMK",0);
    free_Token(key);
  }
}

static void cmd_modify_atom_location(void)
{
  Token *key=(Token *)NULL;
  Nodechain *c=NULL;
  Atom *a;
  int nchgd=0;

  key=Get_Atoms_D("atoms: ",";");
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    key=free_Token(key);
    key=Get_String_M("location indicator: ",";","");
    if (key->status==TOK_OK) {
      nchgd=0;
      for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	if (a->nxyz>1) {
	  if (a->alt[0]!=key->text[0]) {
	    set_atom_location(a,key->text[0]);
	    if (a->alt[0]==key->text[0]&&cPtr(a->cross)) {
	      a->flg|=(1<<7);
	      nchgd++;
	    }
	  }
	}
      }
      if (nchgd>0) {
	send_bheader(graphix,"NEWX",nchgd*(3*sizeof(float)+cPtrSize()));
	for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	  if (a->flg&(1<<7)) {
	    send_bcptr(graphix,a->cross);
	    send_bfloata(graphix,3,a->x);
	    a->flg&=~(1<<7);
	  }
	}
	trflush(graphix);
      }
    }else if (key->status==TOK_NOK) err_set(1,"bad atom location indicatr: %s",key->text);
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) err_set(1,"bad atom selection %s",key->text);
  free_Token(key);
}

static void cmd_modify_resnum(void)
{
  Token *key;
  char *menu,*u;
  Nodechain *c;
  Node *m,*r,*ch;
  char buff[4096];
  int n,i;

  key=(Token *)NULL;
  menu=(char *)NULL;

  key=free_Token(key);free_menu(menu);
  menu=Make_TopMolNameList("* /*",DefaultNode);
  key=Get_Mols_M("chain: ",";",menu);
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n == 1) {
      key=free_Token(key);
      key=Get_Int_M("initial: ",";","1 2 3 4 5 6 7 8 9");
      if (key->status==TOK_OK) {
	n=key->n; i=n;
	m=c->s;
	r=rfirst_type(m,OBJ_MOL);
	if (mvld(r)) {
	  ch=r->up;
	  if (mvld(ch)) {
	    if (r != m) {
	      prtGraphics("renumbering chain %s of %s\n",
			  ch->name,m->name);
	      forObjs(r,ch,OBJ_MOL) {
		Strcpy(buff,r->name);
		u=strrchr(buff,'_');
		if (u) {
		  u++;
		  sprintf(u,"%d",i++);
		}else{
		  u=buff+strlen(buff);
		  sprintf(u,"_%d",i++);
		}
		set_node_name(r,buff);
	      }
	      if (i>n) {
		prtGraphics("residues renumbered: %s to %s\n",
			    ch->fst->name,ch->lst->name);
	      }else{
		prtGraphics("no residues found to renumber\n");
	      }
	    }else{
	      Strcpy(buff,r->name);
	      u=strrchr(buff,'_');
	      if (u) {
		u++;
		sprintf(u,"%d",i++);
	      }else{
		u=buff+strlen(buff);
		sprintf(u,"_%d",i++);
	      }
	      set_node_name(r,buff);
	      prtGraphics("new residue name is %s\n",r->name);
	    }
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad integer: %s",key->text);
      }
    }else{
      err_set(1,"multiple selection of molecules not allowed.");
    }
    clear_node_chain(c);
    free_node_chain(c);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad mol token: %s.",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_modify_molecule --------------------*/

static void cmd_modify_molecule(void)
{
  Token *key;
  char *menu;
  Nodechain *c;
  Node *m;
  char buf[2048];

  key=(Token *)NULL;
  menu=(char *)NULL;

  key=Get_Kwd("what: ",";",mol_kwd);
  if (key->status==TOK_OK) {
    if (key_is("name")) {
      key=free_Token(key);
      menu=Make_TopMolNameList("",DefaultNode);
      key=Get_Mols_M("molecule(s): ",";",menu);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (m=c->s;mvld(m);m=m->chain) {
	  if (is_mol(m)) {
	    default_token(m->name);
	    sprintf(buf,"new name for ");
	    nodeptr_to_name(buf+strlen(buf),m);
	    sprintf(buf+strlen(buf),": ");
	    key=free_Token(key);
	    key=Get_String(buf,";");
	    if (key->status==TOK_OK) {
	      set_node_name(m,key->text);
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad name string: %s",key->text);
	      break;
	    }else{
	      break;
	    }
	  }
	}
	clear_node_chain(c);free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad molecule selection: %s",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown option: %s",key->text);
  }

  free_Token(key);
  free_menu(menu);
  err_dump();

}
/*-------------------- automatic_types --------------------*/

#define tFlg 0

int automatic_types(Nodechain *atms)
{
  /* automatic assignment of atomtypes, using rules */

  Atom *a,*as,*e,**Q=NULL;
  Prule *pr;
  Frule *fr;
  Atomdef *t;
  int i,n,k,nunk,nq,seqnr;
  Node *ruleset,*frules;
  Token *key;
  char buff[4096];
  Nodechain *c=(Nodechain*)NULL;
  Node *PruleLib,*ResLib,*FruleLib;
  
  key=(Token *)NULL;
  frules=(Node *)NULL;
  nunk=(mvld(atms) ? atms->n : 0);

  PruleLib=find_name(LibRoot,"PruleLib",OBJ_LIB);
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  FruleLib=find_name(LibRoot,"FruleLib",OBJ_LIB);

  if (!mvld(PruleLib)) {
    err_set(1,"no automatic rules available.");
    return(nunk);
  }

  buff[0]='\0';
  for (i=0,ruleset=PruleLib->fst;mvld(ruleset);ruleset=ruleset->nxt) {
    if (is_lib(ruleset)) {
      if (mvld(ruleset->name)) {
	sprintf(buff+strlen(buff)," %s",ruleset->name);
	i++;
      }
    }
  }

  if (i==0) {
    err_set(1,"no automatic rules available.");
    return (nunk);
  }


  key=Get_Kwd("ruleset: ",";",buff);
  if (key->status==TOK_OK) {
    ruleset=find_name(PruleLib,key->text,OBJ_LIB);
    if (!is_lib(ruleset)) {
      err_set(1,"ruleset %s not found.",key->text);
      key->status=TOK_NOK;
    }else frules=find_name(FruleLib,key->text,OBJ_LIB);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s\n",key->text);
  }  
  
  if (key->status==TOK_OK) {
    if (!mvld(atms)) {
      key=free_Token(key);
      key=Get_Atoms_D("atoms: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad atom selection: %s",key->text);
	key->status=TOK_ABORT;
      }
    }else{
      c=atms;
    }
  }
  if (key->status==TOK_OK) {
    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {     /* apply Prules ... */
      t=(Atomdef *)NULL;
      ForPrules(pr,ruleset) {
	if (matchPruleAtom(a,pr,0,0)) {
	  t=find_atomdef_name(ResLib,pr->name);
	  if (mvld(t)) break;
	}
      }
      if (!mvld(t)) {
	sprintf(buff,"%s.?",a->atp->chs);
	t=find_atomdef_name(ResLib,buff);
	a->flg|=(1<<7);
      }
      if (!mvld(t)) {
	t=find_atomdef_name(ResLib,"unk");
      }
      if (mvld(t)) set_atom_type(a,t);
    }
    n=c->n; e=(Atom *)c->e;
    if (is_lib(frules)) {
      for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg|=(1<<tFlg);
      Q=(Atom **)chain_to_ptr(c,FALSE);
      for (k=0;k<n;k++) Q[k]->seqnr=k;
      nq=1;
      for (k=0;k<n;k++) { /* apply Frules (if any)... */
	a=Q[k];
	if (a->flg&(1<<tFlg)) {
	  a->flg&=~(1<<tFlg);
	  ForFrules(fr,frules) {
	    if (matchFrule(a,fr)) {
	      for (i=0;i<fr->nat;i++) {
		if (is_atomdef(fr->oatp[i])) {
		  if (fr->oatp[i]!=fr->match[i]->atp) {
		    if (!node_in_chain((Node *)fr->match[i],c)) {
		      add_to_node_chain((Node *)fr->match[i],c);
		      fr->match[i]->flg&=~(1<<tFlg);
		    }
		    set_atom_type(fr->match[i],fr->oatp[i]);
		    fr->match[i]->flg&=~(1<<7);
		  }
		}
	      }
	      for (i=0;i<a->nbnd;i++) {  /* topological sort */
		if (nq<=k) nq=k+1;
		as=neighbour(i,a);
		if (as->flg&(1<<tFlg)) {
		  seqnr=as->seqnr;
		  Q[seqnr]=Q[nq];
		  Q[nq]=as;
		  Q[seqnr]->seqnr=seqnr;
		  Q[nq]->seqnr=nq;
		  nq++;
		}
	      }
	      break;
	    }
	  }
	}
      }
      free(Q);
    }
    for (nunk=0,k=0,a=(Atom *)c->s;mvld(a)&&k<n;a=(Atom *)a->chain,k++) {  /* count unassigned atoms */
      if (a->flg&(1<<7)) {
	err_set(1,"automatic assignment failed: %s",nodeptr_to_name(buff,(Node *)a));
	a->flg&=~(1<<7);
	nunk++;
      }
    }
    if (auto_label()) new_atom_labels(c,LAB_TYPE);
    err_dump();
      
    if (!mvld(atms)) {
      prtGraphics("%d atoms changed type.\n",c->n);
      clear_node_chain(c);
      free_node_chain(c);
    }else if (is_atom(e)) {    /* get rid of extra atoms added by frules */
      c->e=(Node *)e; c->n=n;
      for (a=(Atom *)e->chain;mvld(a);a=e) {
	e=(Atom *)a->chain;
	a->chain=(Node *)NULL;
      }
    }
  }
  free_Token(key);
  return(nunk);
}
/*-------------------- auto_atom_name --------------------*/

static int automatic_names(Atom *a,int mode)
{
  Atom *C;
  int nchgd=0;
  char *txt,buf[MAXBUF];

  C=a;

  for (a=C;mvld(a);a=(Atom *)a->chain) {
    if (a->atp->z!=1) {
      txt=auto_atom_name(a,buf,mode);
      if (txt) {
	if (!streq(a->name,txt)&&strlen(txt)>0) {
	  set_node_name((Node *)a,txt);
	  if (auto_label()) {
	    set_atom_label(a,ATM_LAB_NAME,NULL,TRUE);
	  }
	  nchgd++;
	}
      }
    }
  }
  for (a=C;mvld(a);a=(Atom *)a->chain) {
    if (a->atp->z==1) {
      txt=auto_atom_name(a,buf,mode);
      if (txt) {
	if (!streq(a->name,txt)&&strlen(txt)>0) {
	  set_node_name((Node *)a,txt);
	  if (auto_label()) {
	    set_atom_label(a,ATM_LAB_NAME,NULL,TRUE);
	  }
	  nchgd++;
	}
      }
    }
  }
  return nchgd;
}

char *auto_atom_name(Atom *a,char *name,int mode)
{
  /* generate atom name depending on mode */

  char *r;
  int n,nh,nok;
  Atom *b,*c;

  r=(char *)NULL;
  if (!name) return(r);
  name[0]='\0';

  if (!mvld(a)) return(r);

  if (mvld(a->name)) {
    r=name;
    strcpy(name,a->name);   /* note: Strcpy would be the  W R O N G  thing */
  }

  if (!mvld(a->atp)) return(r);

  switch (mode) {
  case AUTONM_ALARM:
    if (a->atp->z!=1) {
      if (is_mol(a->up)) {
	n=0;
	forAtoms(b,a->up) {
	  if (b->atp->z!=1) n++;
	  if (b==a) break;
	}
	if (mvld(b)) {
	  sprintf(name,"%s%d",a->atp->chs,n);
	  r=name;
	}
      }
    }else{
      if (a->nbnd==1) {
	b=neighbour(0,a);
	if (mvld(b->atp)&&b->atp->z!=1) {
	  nok=FALSE;
	  for (nh=0,n=0;n<b->nbnd;n++) {
	    c=neighbour(n,b);
	    if (mvld(c->atp)) {
	      if (c->atp->z==1) nh++;
	    }
	    if (a==c) {
	      strncpy(name,b->name,strlen(b->atp->chs));
	      name[strlen(b->atp->chs)]='\0';
	      if (strcasecmp(name,b->atp->chs)==0) {
		sprintf(name,"%s%s%c",
			a->atp->chs,b->name+strlen(b->atp->chs)," ABCDEFGHIJKLMNOPQRSTUVWXYZ"[nh>26?26:nh]);
		r=name; nok=TRUE;
	      }else{
		strcpy(name,a->name); /* note: Strcpy would be the  W R O N G  thing */
	      }
	    }
	  }
	  if (nh==1&&nok) name[strlen(name)-1]='\0';
	}
      }
    }
    break;
    
  case AUTONM_HSEQ:             /* H+seq# , not terribly efficient... */
    if (a->atp->z==1) {
      if (is_mol(a->up)) {
	n=0;
	forAtoms(b,a->up) {
	  n++;
	  if (b==a) break;
	}
	if (mvld(b)) {
	  sprintf(name,"%s%d",a->atp->chs,n);
	  r=name;
	}
      }
    }
    break;
  
  case AUTONM_HHEAV:
    if (a->atp->z==1) {
      if (a->nbnd==1) {
	b=neighbour(0,a);
	if (mvld(b->atp)&&mvld(b->name)) {
	  nok=FALSE;
	  for (nh=0,n=0;n<b->nbnd;n++) {
	    c=neighbour(n,b);
	    if (mvld(c->atp)) {
	      if (c->atp->z==1) nh++;
	    }
	    if (a==c) {
	      strncpy(name,b->name,strlen(b->atp->chs));
	      name[strlen(b->atp->chs)]='\0';
	      if (strcasecmp(name,b->atp->chs)==0) {
		sprintf(name,"%s%s%d",
			a->atp->chs,b->name+strlen(b->atp->chs),nh);
		r=name; nok=TRUE;
	      }else{
		strcpy(name,a->name); /* note: Strcpy would be the  W R O N G  thing */
	      }
	    }
	  }
	  if (nh==1&&nok) name[strlen(name)-1]='\0';
	}
      }
    }
    break;
  case AUTONM_SEQ:
    if (is_mol(a->up)) {
      n=0;
      forAtoms(b,a->up) {
	n++;
	if (b==a) break;
      }
      if (mvld(b)) {
	sprintf(name,"%s%d",a->atp->chs,n);
	r=name;
      }
    }
    break;
  default:
    break;
  }

  return(r);
}
/*-------------------- auto_atom_group --------------------*/

void auto_atom_group_greedy(Node *res)
{
  /* generate automatic charge groups for a given residue...      */
  /* this will work in a very limited number of situations only,  */
  /* it uses a naive, and not very efficient heuristic...         */

  int i,j,Gi,Gj,nat,ngrp,nit=0;
  int *grp;
  float *chg,D,d,di,dj;
  Atom **atm,*a,*b;
  char buff[MAXBUF];

  if (!is_mol(res)) return;

  nat=0;
  forAtoms(a,res) nat++;

  if (nat<1) return;         /* no atoms in this residue... */

  /* dynamically allocate some scratch space... */

  grp=(int *)malloc(nat*sizeof(int));
  chg=(float *)malloc(nat*sizeof(float));
  atm=(Atom **)malloc(nat*sizeof(Atom *));

  if (!mvld(grp)||!mvld(chg)||!mvld(atm)) {
    if (mvld(grp)) free(grp);
    if (mvld(chg)) free(chg);
    if (mvld(atm)) free(atm);
    return;
  }

  /* put each atom in a separate group */

  i=0; forAtoms(a,res) atm[i++]=a;

  for (i=0;i<nat;i++) {
    grp[i]=i;     
    chg[i]=atm[i]->q;
    atm[i]->seqnr=i;                 /* sequence number within residue */
  }
 
  for (D=CTOL+1;D>=CTOL;) {
    D=0; Gi=0; Gj=0;
    for (i=0;i<nat;i++) {
      if (chg[grp[i]]<CTOL&&chg[grp[i]]>-CTOL) continue;
      for (j=0;j<atm[i]->nbnd;j++) {
	b=neighbour(j,atm[i]);
	if (b->up==res&&grp[b->seqnr]!=grp[i]) {
	  if (chg[grp[i]]*chg[grp[b->seqnr]]<0) {
	    di=chg[grp[i]]>0?chg[grp[i]]:-chg[grp[i]];
	    dj=chg[grp[b->seqnr]]>0?chg[grp[b->seqnr]]:-chg[grp[b->seqnr]];
	    d=(di<dj?di:dj);
	    if (d>D) {
	      D=d;
	      Gi=grp[i];
	      Gj=grp[b->seqnr];
	    }
	  }
	}
      }
    }
    if (D>=CTOL) {
      d=chg[Gi]+chg[Gj];
      err_set(0,"merge %d,%d : %0.4f,%0.4f --> %0.4f\n",Gi,Gj,chg[Gi],chg[Gj],d);
      for (i=0;i<nat;i++) {
	if (grp[i]==Gi||grp[i]==Gj) {
	  grp[i]=Gi;
	  chg[i]=d;
	}
      }
      nit++;
      if (nit>nat) break;
    }
  }	

  if (nit>nat) err_set(1,"note: charge groups not converged.");

  /* generate the group names ... */

  for (ngrp=0,i=0;i<nat;i++) {
    if (grp[i]==i) {
      ngrp++;
      atm[i]->cosw=1;
      sprintf(buff,"g%d",ngrp);
      Strcpy(atm[i]->grp,buff);
    }
  }
  for (i=0;i<nat;i++) {
    if (grp[i]!=i) {
      atm[i]->cosw=0;
      Strcpy(atm[i]->grp,atm[grp[i]]->grp);
    }
  }

  if (mvld(grp)) free(grp);
  if (mvld(atm)) free(atm);
  if (mvld(chg)) free(chg);

  return;
}

void auto_atom_group_hetero(Node *res)
{
  /* Topological method, does not depend on partial atomic charges   */
  /* Jan. 6 2005, for GROMOS input */

  int i,j,Gi,Gj,nat,ngrp;
  int *grp;
  Atom **atm,*a,*b;
  char buff[MAXBUF];

  if (!is_mol(res)) return;

  nat=0;
  forAtoms(a,res) nat++;

  if (nat<1) return;         /* no atoms in this residue... */

  /* dynamically allocate some scratch space... */

  grp=(int *)malloc(nat*sizeof(int));
  atm=(Atom **)malloc(nat*sizeof(Atom *));

  if (!mvld(grp)||!mvld(atm)) {
    if (mvld(grp)) free(grp);
    if (mvld(atm)) free(atm);
    return;
  }

  /* put each atom in a separate group */

  i=0; forAtoms(a,res) atm[i++]=a;

  for (i=0;i<nat;i++) {
    grp[i]=i;     
    atm[i]->seqnr=i;                 /* sequence number within residue */
  }
 
  /* merge groups that are connected via bond to non-C atom */

  for (i=0;i<nat;i++) {
    if (atm[i]->atp->z!=6) {
      for (j=0;j<atm[i]->nbnd;j++) {
	b=neighbour(j,atm[i]);
	if (b->up==res&&grp[b->seqnr]!=grp[i]) break;
      }
      if (j<atm[i]->nbnd) {
	Gi=grp[i];
	Gj=grp[neighbour(j,atm[i])->seqnr];
	for (i=0;i<nat;i++) if (grp[i]==Gj) grp[i]=Gi;
	i=0;
      }
    }
  }

  /* generate the group names ... */

  for (ngrp=0,i=0;i<nat;i++) {
    if (grp[i]==i) {
      ngrp++;
      atm[i]->cosw=1;
      sprintf(buff,"g%d",ngrp);
      Strcpy(atm[i]->grp,buff);
    }
  }
  for (i=0;i<nat;i++) {
    if (grp[i]!=i) {
      atm[i]->cosw=0;
      Strcpy(atm[i]->grp,atm[grp[i]]->grp);
    }
  }

  if (mvld(grp)) free(grp);
  if (mvld(atm)) free(atm);

  return;
}

void auto_atom_group_wooly(Node *res)
{
  /* generate automatic charge groups for a given residue...      */
  /* this will work in a very limited number of situations only,  */
  /* it uses a naive, and not very efficient heuristic...         */

  int i,j,k,Gmin,Gmax,nat,ngrp,loop;
  int *grp;
  float *chg,Cmax,Cmin;
  Atom **atm,*a,*b;
  char buff[MAXBUF];

  if (!is_mol(res)) return;

  nat=0;
  forAtoms(a,res) nat++;

  if (nat<1) return;         /* no atoms in this residue... */

  /* dynamically allocate some scratch space... */

  grp=(int *)malloc(nat*sizeof(int));
  chg=(float *)malloc(nat*sizeof(float));
  atm=(Atom **)malloc(nat*sizeof(Atom *));

  if (!mvld(grp)||!mvld(chg)||!mvld(atm)) {
    if (mvld(grp)) free(grp);
    if (mvld(chg)) free(chg);
    if (mvld(atm)) free(atm);
    return;
  }

  /* put each atom in a separate group */

  i=0; forAtoms(a,res) atm[i++]=a;

  for (i=0;i<nat;i++) {
    grp[i]=i;     
    chg[i]=atm[i]->q;
    atm[i]->seqnr=i;                 /* sequence number within residue */
  }
    
  /* merge non-neutral charge groups with single neighbours... */

  loop=TRUE;

  while(loop) {
    loop=FALSE;
    for (i=0;i<nat;i++) {
      if (grp[i]==i && (chg[i]>CTOL || chg[i] < -CTOL)) {
	Gmin=nat; Gmax=-1;
	for (k=0;k<nat;k++) {
	  if (grp[k]==i) {
	    for (j=0;j<atm[i]->nbnd;j++) {
	      b=neighbour(j,atm[i]);
	      if (b->up==res) {
		if (grp[b->seqnr]!=i) {
		  Gmin=(Gmin < grp[b->seqnr] ? Gmin : grp[b->seqnr]);
		  Gmax=(Gmax > grp[b->seqnr] ? Gmax : grp[b->seqnr]);
		}
	      }
	    }
	  }
	}
	if (Gmin==Gmax) { /* Gmin is single neighbour... --> merge */
	  Cmax=chg[i];
	  Cmin=chg[Gmin];
	  for (k=0;k<nat;k++) {
	    if (grp[k]==i) {
	      grp[k]=Gmin;
	      chg[k]=Cmin+Cmax;
	    }else if (grp[k]==Gmin) {
	      chg[k]=Cmin+Cmax;
	    }
	  }
	  loop=TRUE;
	}
      }
    }
  }
	

  /* try to reduce group-charge by merging groups */

  loop=TRUE;
  while (loop) {

    /* find group with min and max net-charge */

    Gmin=grp[0]; Gmax=grp[0];
    Cmin=chg[Gmin]; Cmax=chg[Gmax];
    for (i=0;i<nat;i++) {
      if (grp[i]==i) {
	if (Cmin>chg[i]) {
	  Cmin=chg[i];
	  Gmin=i;
	}
	if (Cmax<chg[i]) {
	  Cmax=chg[i];
	  Gmax=i;
	}
      }
    }
    if (Cmin>=-CTOL && Cmax <=CTOL) break;   /* no further reduction */
    
    loop=FALSE;                       /* stop, if no further reduction... */ 

    if (-Cmax < Cmin) {
      /* find most positive neighbour*/

      Gmin=Gmax; Cmin=Cmax;
      for (i=0;i<nat;i++) {
	if (grp[i]==Gmax) {
	  for (j=0;j<atm[i]->nbnd;j++) {
	    b=neighbour(j,atm[i]);
	    if (b->up==res) {
	      if (chg[b->seqnr]<Cmin) {
		Cmin=chg[b->seqnr];
		Gmin=grp[b->seqnr];
	      }
	    }
	  }
	}
      }
      if (Cmin<0) {
	/* merge the two... */
	
	Cmax=Cmax+Cmin;
	for (i=0;i<nat;i++) {
	  if (grp[i]==Gmin || grp[i]==Gmax) {
	    grp[i]=Gmax;
	    chg[i]=Cmax;
	  }
	}
	loop=TRUE;
      }
      
    }else{
      /* find most negative neighbour*/

      Gmax=Gmin; Cmax=Cmin;
      for (i=0;i<nat;i++) {
	if (grp[i]==Gmin) {
	  for (j=0;j<atm[i]->nbnd;j++) {
	    b=neighbour(j,atm[i]);
	    if (b->up==res) {
	      if (chg[b->seqnr]>Cmax) {
		Cmax=chg[b->seqnr];
		Gmax=grp[b->seqnr];
	      }
	    }
	  }
	}
      }
      if (Cmax>0) {
	/* merge the two... */
	
	Cmin=Cmax+Cmin;
	for (i=0;i<nat;i++) {
	  if (grp[i]==Gmin || grp[i]==Gmax) {
	    grp[i]=Gmin;
	    chg[i]=Cmin;
	  }
	}
	loop=TRUE;
      }
    }
  }

  /* generate the group names ... */

  for (ngrp=0,i=0;i<nat;i++) {
    if (grp[i]==i) {
      ngrp++;
      atm[i]->cosw=1;
      sprintf(buff,"g%d",ngrp);
      Strcpy(atm[i]->grp,buff);
    }
  }
  for (i=0;i<nat;i++) {
    if (grp[i]!=i) {
      atm[i]->cosw=0;
      Strcpy(atm[i]->grp,atm[grp[i]]->grp);
    }
  }

  if (mvld(grp)) free(grp);
  if (mvld(atm)) free(atm);
  if (mvld(chg)) free(chg);

  return;
}
/*----------------------------- cmd_all_charges --------------------*/

static int cmd_all_charges(Nodechain *c,int label_flag)
{
  Token *key;
  int nchgd;

  nchgd=0;
  default_token("0.0");
  key=Get_Float("partial charge to be assigned to selected atoms: ",";");
  if (key->status==TOK_OK) nchgd=all_charges(c,label_flag,key->f);
  else if (key->status==TOK_NOK) err_set(1,"bad floating point value: %s",key->text);
  return(nchgd);
}
/*----------------------------- cmd_modify_all_sofs --------------------*/

static int cmd_modify_all_sofs(Nodechain *c,int label_flag)
{
  Token *key=(Token *)NULL;
  Atom *a;
  char *nummenu="0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0";
  int nchgd;

  nchgd=0;
  default_token("1.0");
  key=Get_Float_M("new site occupancy factor: ",";",nummenu);
  if (key->status==TOK_OK) {
    for (a=(Atom *)c->s,nchgd=0;a;a=(Atom *)a->chain) {
      if (is_atom(a)) {
	a->x[ATM_SOF]=key->f;
	if (label_flag) {
	  set_atom_label(a,ATM_LAB_SOF,NULL,TRUE);
	}
	nchgd++;
      }
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad site occupancy factor: %s",key->text);
  free_Token(key);
  return(nchgd);
}
/*----------------------------- cmd_modify_all_bisos --------------------*/

static int cmd_modify_all_bisos(Nodechain *c,int label_flag)
{
  Token *key=(Token *)NULL;
  Atom *a;
  char *bisomnu="1.0 2.0 3.0 5.0 10.0 20.0 30.0 40.0 50.0 60.0 70.0 80.0 90.0 100.0";
  int nchgd;

  nchgd=0;
  default_token("1.0");
  key=Get_Float_M("new isotropic temperature factor: ",";",bisomnu);
  if (key->status==TOK_OK) {
    for (a=(Atom *)c->s,nchgd=0;a;a=(Atom *)a->chain) {
      if (is_atom(a)) {
	a->x[ATM_BISO]=key->f;
	if (label_flag) {
	  set_atom_label(a,ATM_LAB_BISO,NULL,TRUE);
	}
	nchgd++;
      }
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad isotropic temperature factor: %s",key->text);
  free_Token(key);
  return(nchgd);
}
