#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"
#ifdef mingw
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

extern Node *LibRoot;
extern Node *DefaultNode;
extern int graphix;

#define TRUE 1
#define FALSE 0
#define SHOW_FLG (1<<2)
#define MAXTOKENSIZE 1024

#define key_is(x) (streq(x,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define OPTF(x) (x ? "*" : "")

static void cmd_mpeoe_show(MPEOEopt opt);

#if defined(mingw)
static int mkstemp(char *template)
{
  mktemp(template);
  return(open(template,O_RDWR));
}
#endif

/*========================================================================*/

void cmd_mpeoe(void) 
{
  Token *key;
  char *keys,buff[4096],tmpfn[12];
  Nodechain *A;
  float rms;
  Stream *in;
  MPEOEopt opt;
  FILE *out;
  int tmpfd,status=0;
  Lib *l;
  int loop;

  keys="accuracy compute nsteps ' ' "
    "dChi dF iterations maxF minF minocc report tolerance ' ' "
    "edit list read refine show write ' ' "
    "Command display label modify reset select ' ' "
    "end";

  key=(Token *)NULL;
  opt=MPEOEgetOpt();
  for(loop=TRUE;loop;) { 
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("MPEOE> ",";",keys);
    if (key->status==TOK_OK) {
      if (key_is("Command")) command_loop("Wit!P(MPEOE)> ");
      else if (key_is("display")) {
	cmd_display();
      }else if (key_is("label")) {
	cmd_label();
      }else if (key_is("modify")) {
	cmd_modify();
      }else if (key_is("reset")) {
	cmd_display_reset();
      }else if (key_is("select")) {
	cmd_select();
      }else if (key_is("refine")) {
	key=free_Token(key);
	key=Get_Atoms_D("fit to atoms: ",";");
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  prtGraphics("Number of atoms: %d\n",A->n);
	  MPEOEsetOpt(opt);
	  rms=MPEOErefine(A);
	  prtGraphics("rms=%0.6f\n",rms);
	  clear_node_chain(A); free_node_chain(A);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad mol token: %s.",key->text);
	}
      }else if (key_is("dChi")) {
	key=free_Token(key);
	key=Get_Float_M("initial delta-Chi: ",";","1.0 0.5 0.25 0.1 0.05 0.01");
	if (key->status==TOK_OK) opt.dChi=key->f;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad dChi value: %s",key->text);
	  end_parse();
	}
      }else if (key_is("dF")) {
	key=free_Token(key);
	key=Get_Float_M("initial delta-f: ",";",
			"0.10 0.050 0.025 0.001 0.005 0.0025 0.0001");
	if (key->status==TOK_OK) opt.dF=key->f;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad dF value: %s",key->text);
	  end_parse();
	}
      }else if (key_is("iterations")) {
	key=free_Token(key);
	key=Get_Int_M("maximum number of par. ref. iterations: ",";",
		      "1 10 100 250 500 1000 2500 5000 10000");
	if (key->status==TOK_OK) opt.mxref=key->n;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad maxiter value: %s",key->text);
	  end_parse();
	}
      }else if (key_is("minocc")) {
	key=free_Token(key);
	key=Get_Int_M("min. number of occurances for opt. param.: ",";","1 2 3 4 5 10");
	if (key->status==TOK_OK) opt.minocc=key->n;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad minocc value: %s",key->text);
	  end_parse();
	}
      }else if (key_is("tolerance")) {
	key=free_Token(key);
	key=Get_Float_M("refinement tolerance (delta-rms): ",";",
			"0.10 0.01 0.001 0.0001 0.00001 0.000001 0.0000001");
	if (key->status==TOK_OK) opt.tolref=key->f;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad rms tolerance: %s",key->text);
	  end_parse();
	}
      }else if (key_is("minF")) {
	key=free_Token(key);
	key=Get_Float_M("minimum allowed value for f: ",";",
			"0.0 0.05 0.10 0.15 0.20 0.25");
	if (key->status==TOK_OK) opt.minf=key->f;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad min-F token: %s",key->text);
	  end_parse();
	}
      }else if (key_is("maxF")) {
	key=free_Token(key);
	key=Get_Float_M("maximum allowed value for f: ",";",
			"0.40 0.50 0.60 0.70 0.75 0.80 0.85 0.90 0.95 0.99");
	if (key->status==TOK_OK) opt.maxf=key->f;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad max-F token: %s",key->text);
	  end_parse();
	}
      }else if (key_is("accuracy")) {
	key=free_Token(key);
	key=Get_Float_M("accuracy (dq in MPEOE): ",";",
			"0.1 0.01 0.001 0.0001 0.00001 0.000001 0.0000001");
	if (key->status==TOK_OK) opt.tolcomp=key->f;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad accuracy: %s",key->text);
	  end_parse();
	}
      }else if (key_is("show")) {
	cmd_mpeoe_show(opt);
      }else if (key_is("compute")) {
	key=free_Token(key);
	key=Get_Atoms_D("compute MPEOE charges for atoms: ",";");
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  MPEOEsetOpt(opt);
	  MPEOEcharges(A,TRUE);
	  clear_node_chain(A); free_node_chain(A);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",key->text);
	  end_parse();
	}
      }else if (key_is("nsteps")) {
	key=free_Token(key);
	key=Get_Int_M("max. number of MPEOE steps: ",";","10 100 1000 10000");
	if (key->status==TOK_OK) opt.mxcomp=key->n;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad value for nsteps: %s",key->text);
	  end_parse();
	}
      }else if (key_is("report")) {
	key=free_Token(key);
	key=Get_Int_M("Simplex, report frequency: ",";",
		      "0 1 5 10 25 50 100 1000 10000");
	if (key->status==TOK_OK) opt.repf=key->n;
	else if (key->status==TOK_NOK) {
	  err_set(1,"bad report frequency: %s",key->text);
	  end_parse();
	}
      }else if (key_is("edit")) {
	sprintf(tmpfn,"MPEOEXXXXXX");
	tmpfd=mkstemp(tmpfn);
	if (tmpfd!=-1) close(tmpfd);
	out=fopen(tmpfn,"w");
	if (out) {
	  MPEOEdump(out);
	  fclose(out);
	  sprintf(buff,"wnp_edit %s",tmpfn);
	  startCmd(buff,"","",TRUE,"normal");
	  l=(Lib *)LibRoot; /* was: (Lib *)find_name(LibRoot,"ResLib",OBJ_LIB) */
	  if (is_lib(l)) {
	    in=open_stream(tmpfn,MAXTOKENSIZE);
	    if (!in||(status=read_lib_stream(in,l))) {
	      err_set(1,"sorry, (%d) failed to read parameters from %s",status,tmpfn);
	    }
	  }else{
	    err_set(1,"sorry, can't find internal library root.");
	  }
	  unlink(tmpfn);
	}else err_set(1,"failed to create temporary file %s, edit failed.",tmpfn);
      }else if (key_is("list")) {
	MPEOEdump((FILE *)NULL);
      }else if (key_is("read")) {  /* a bit complicated, because 'read_lib' wants filename...*/ /*NOTE use read_lib_ucp!*/
	l=(Lib *)LibRoot; /* was: (Lib *)find_name(LibRoot,"ResLib",OBJ_LIB) */
	if (is_lib(l)) {
	  key=free_Token(key);
	  key=Get_File((char *)NULL,".wlb","MPEOE library file: ",";");
	  if (key->status==TOK_OK) {
	    if (((UCpipe *)(key->ptr))->pipe) {
	      err_set(1,"sorry, cant't read this from a pipe.");
	      UCpclose((UCpipe *)(key->ptr));
	    }else{
	      if (key->text) free(key->text);
	      key->text=Strdup(((UCpipe *)(key->ptr))->filename);
	      UCpclose((UCpipe *)(key->ptr));
	      prtGraphics("reading library from from %s\n",key->text);
	      if (read_lib(key->text,l)) {
		err_set(1,"sorry, failed to read parameters from %s",key->text);
	      }
	    }
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"sorry, could not read library from file %s",key->text);
	    end_parse();
	  }
	}else{
	  err_set(1,"sorry, can't find internal library root.");
	  end_parse();
	}
      }else if (key_is("write")) {
	key=free_Token(key);
	key=Get_String_M("write MPEOE parameters to file: ",";","mpeoe.wlb");
	if (key->status==TOK_OK) {
	  out=fopen(extfil(buff,key->text,".wlb",FALSE),"w");
	  if (out) {
	    fprintf(out,"# refined MPEOE parameters\n\n");
	    MPEOEdump(out);
	    fclose(out);
	    prtGraphics("refined MPEOE parameters written to %s\n",key->text);
	  }else{
	    err_set(1,"sorry, failed to open file %s for write",key->text);
	    end_parse();
	  }
	}else{
	  err_set(1,"bad token: %s",key->text);
	  end_parse();
	}
      }else if (key_is("end")) loop=FALSE;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad MPEOE keyword: %s",key->text);
    }else break;
  }
  MPEOEsetOpt(opt);
  free_Token(key);
}
/*-------------------- cmd_mpeoe_show --------------------*/

static void cmd_mpeoe_show(MPEOEopt opt)
{
  Token *key=(Token *)NULL;
  Nodechain *A;
  Atom *a,*b;
  MPEOEinfo G;
  MPEOEatom *mpa;
  MPEOEbond *mpb;
  Prule *pr;
  int nbad,no;
  int i,fr,to,r;
  char buf1[4096],buf2[4096],tag[3],*t;
  char *menu;
  char *menuprt="-print h!-noprint '' atoms bonds errors missing qcalc usage h!use";
  char *menunoprt="h!-print -noprint '' atoms bonds errors missing qcalc usage h!use";
  float tol,dq;
  int loop,prt=FALSE;

  for (loop=TRUE;loop;) {
    loop=FALSE;
    menu=(prt?menunoprt:menuprt);
    key=Get_Kwd("show what: ",";",menu);
    if (key->status==TOK_OK) {
      if (key_is("-print")) {
	prt=TRUE;
	loop=TRUE;
      }else if (key_is("-noprint")) {
	prt=FALSE;
	loop=TRUE;
      }else if (key_is("qcalc")) {
	key=free_Token(key);
	key=Get_Atoms_D("assignment for atoms: ",";");
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  no=A->n;
	  MPEOEsetOpt(opt);
	  G=MPEOEsetup(A);
	  MPEOEreport_missing(G,TRUE);
	  for (i=0;i<G.nat;i++) G.Q[i]=G.iQ[i];
	  dq=MPEOEiterate(G,opt.mxcomp,opt.tolcomp);
	  if (dq>opt.tolcomp) err_set(1,"warning: MPEOE charges not converged, dq=%0.5f",dq);
	  for (i=0;i<no;i++) {
	    if (prt) {
	      nodeptr_to_name(buf1,(Node *)G.A[i]);
	      if (strlen(buf1)<32) {
		sprintf(buf1+strlen(buf1),"                                ");
		buf1[32]='\0';
	      }
	      sprintf(buf1+strlen(buf1),"%6.3f = %6.3f %s%0.3f",
		      G.Q[i],G.A[i]->q,G.Q[i]-G.A[i]->q>0?"+":"",G.Q[i]-G.A[i]->q);
	      prtGraphics("%s\n",buf1);
	    }
	    if (cPtr(G.A[i]->cross)) {
	      sprintf(buf1,"%0.2f",G.Q[i]);
	      send_bheader(graphix,"ALTN",cPtrSize()+strlen(buf1));
	      send_bcptr(graphix,G.A[i]->cross);
	      send_bdata(graphix,strlen(buf1),buf1);
	    }
	  }
	  MPEOEfree_scratch(G);
	  clear_node_chain(A); free_node_chain(A);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s.",key->text);
	}
      }else if (key_is("errors")) {
	key=free_Token(key);
	key=Get_Float_M("show if error is larger than:  ",";",
			"0.50 0.25 0.20 0.15 0.10 0.05 0.025 0.010 0.005 0.001 0");
	if (key->status==TOK_OK) {
	  tol=key->f;
	  if (tol<0) tol=0;
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms: ",";");
	  send_bheader(graphix,"ARMK",0);
	  if (key->status==TOK_OK) {
	    A=(Nodechain *)key->ptr;
	    no=A->n;
	    MPEOEsetOpt(opt);
	    G=MPEOEsetup(A);
	    r=MPEOEreport_missing(G,FALSE);
	    if (r>0) err_set(1,"%d missing MPEOE parameter%s.",r,r==1?"":"s");
	    for (i=0;i<G.nat;i++) G.Q[i]=G.iQ[i];
	    dq=MPEOEiterate(G,opt.mxcomp,opt.tolcomp);
	    if (dq>opt.tolcomp) err_set(1,"warning: MPEOE charges not converged, dq=%0.5f",dq);
	    for (nbad=0,i=0;i<no;i++) {
	      dq=G.Q[i]-G.A[i]->q;
	      if (dq>tol||-dq>tol) {
		G.A[i]->flg|=SHOW_FLG;
		nbad++;
		if (prt) {
		  nodeptr_to_name(buf1,(Node *)G.A[i]);
		  if (strlen(buf1)<32) {
		    sprintf(buf1+strlen(buf1),"                                ");
		    buf1[32]='\0';
		  }
		  sprintf(buf1+strlen(buf1),"%6.3f = %6.3f %s%0.3f",
			  G.Q[i],G.A[i]->q,G.Q[i]-G.A[i]->q>0?"+":"",G.Q[i]-G.A[i]->q);
		  prtGraphics("%s\n",buf1);
		}
		if (cPtr(G.A[i]->cross)) {
		  sprintf(buf1,"%0.3f",dq);
		  send_bheader(graphix,"ALTN",cPtrSize()+strlen(buf1));
		  send_bcptr(graphix,G.A[i]->cross);
		  send_bdata(graphix,strlen(buf1),buf1);
		}
	      }
	    }
	    prtGraphics("%d atoms have abs(dq)>%0.4f\n",nbad,tol);
	    display_atom_chain(A,ATM_DSP_OFF);
	    a=(Atom *)A->s; 
	    A->s=(Node *)NULL; 
	    A->p=(Node *)NULL; 
	    A->e=(Node *)NULL; A->n=0;
	    for (;mvld(a);a=b) {
	      b=(Atom *)a->chain;
	      a->chain=(Node *)NULL;
	      if (a->flg&SHOW_FLG) {
		a->flg&=~SHOW_FLG;
		add_to_node_chain((Node *)a,A);
	      }
	    }
	    if (A->n>0) {
	      connected_atoms(A);
	      display_atom_chain(A,ATM_DSP_ON);
	    }
	    clear_node_chain(A); free_node_chain(A);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s.",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad dq value: %s",key->text);
	  end_parse();
	}
      }else if (key_is("atoms")) {
	key=free_Token(key);
	key=Get_Atoms_D("assignment for atoms: ",";");
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  no=A->n;
	  MPEOEsetOpt(opt);
	  G=MPEOEsetup(A);
	  for (i=0;i<no;i++) {
	    if (prt||cPtr(G.A[i]->cross)) {
	      for (pr=G.Mpa[i]->p;pr&&!matchPruleAtom(G.A[i],pr,0,0);pr=(Prule *)pr->nxt);
	      if (pr) {
		t=(pr->interp==PR_HARD?"":pr->interp==PR_SOFT?"~":"?");
		sprintf(buf1,"%s%s",t,pr->def);
	      }else if (G.Mpa[i]->def) {
		sprintf(buf1,"%s",G.Mpa[i]->def);
	      }else sprintf(buf1,"??");
	      if (cPtr(G.A[i]->cross)) {
		send_bheader(graphix,"ALTN",cPtrSize()+strlen(buf1));
		send_bcptr(graphix,G.A[i]->cross);
		send_bdata(graphix,strlen(buf1),buf1);
	      }
	      if (prt) {
		nodeptr_to_name(buf2,(Node *)G.A[i]);
		if (strlen(buf2)<32) {
		  sprintf(buf2+strlen(buf2),"                                ");
		  buf2[32]='\0';
		}
		prtGraphics("%s %s\n",buf2,buf1);
	      }
	    }
	  }
	  MPEOEfree_scratch(G);
	  clear_node_chain(A); free_node_chain(A);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s.",key->text);
	}
      }else if (key_is("bonds")) {
	key=free_Token(key);
	key=Get_Atoms_D("show damping factor for bonds between atoms: ",";");
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  no=A->n;
	  MPEOEsetOpt(opt);
	  G=MPEOEsetup(A);
	  for (i=0;i<no;i++) G.A[i]->flg|=SHOW_FLG;
	  for (i=0;i<G.nbnd;i++) {
	    if ((G.B[i]->f->flg&SHOW_FLG)&&
		(G.B[i]->t->flg&SHOW_FLG)&&
		(cPtr(G.B[i]->cross)||prt)) {
	      for (pr=G.Mpb[i]->p;pr&&!matchPruleBond(G.B[i],pr,0,0);pr=(Prule *)pr->nxt);
	      if (pr) {
		t=(pr->interp==PR_HARD?"":pr->interp==PR_SOFT?"~":"?");
		sprintf(buf1,"%s%s",t,pr->def);
	      }else if (G.Mpb[i]->def) {
		sprintf(buf1,"%s",G.Mpb[i]->def);
	      }else sprintf(buf1,"??");
	      set_bond_label(G.B[i],BND_LAB_TEXT,buf1);
	      if (prt) {
		nodeptr_to_name(buf2,(Node *)G.B[i]->f);
		sprintf(buf2+strlen(buf2)," -- ");
		nodeptr_to_name_ctxt(buf2+strlen(buf2),(Node *)G.B[i]->t,(Node *)G.B[i]->f);
		if (strlen(buf2)<32) {
		  sprintf(buf2+strlen(buf2),"                                ");
		  buf2[32]='\0';
		}else if (strlen(buf2)<45) {
		  sprintf(buf2+strlen(buf2),"                                ");
		  buf2[45]='\0';
		}
		prtGraphics("%s %s\n",buf2,buf1);
	      }
	    }
	  }
	  for (i=0;i<no;i++) G.A[i]->flg&=~SHOW_FLG;
	  MPEOEfree_scratch(G);
	  clear_node_chain(A); free_node_chain(A);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s.",key->text);
	}
      }else if (key_is("missing")) {
	key=free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  MPEOEsetOpt(opt);
	  G=MPEOEsetup(A);
	  r=0;
	  sprintf(tag,"??");
	  for (i=0;i<G.nat;i++) {
	    if (!G.Mpa[i]->def) {
	      r++;
	      if (prt) {
		prtGraphics("MPEOE parameters missing for %s\n",
			    nodeptr_to_name(buf1,(Node *)G.A[i]));
	      }
	      set_atom_label(G.A[i],ATM_LAB_TEXT,tag,TRUE);
	    }
	  }
	  for (i=0;i<G.nbnd;i++) {
	    if (!G.Mpb[i]->def) {
	      r++;
	      if (prt) {
		fr=G.B[i]->f->seqnr; to=G.B[i]->t->seqnr;
		prtGraphics("MPEOE parameters missing for %s -- %s\n",
			    nodeptr_to_name(buf1,(Node *)G.A[fr]),
			    nodeptr_to_name_ctxt(buf2,(Node *)G.A[to],(Node *)G.A[fr]));
	      }
	      set_bond_label(G.B[i],BND_LAB_TEXT,tag);
	    }
	  }
	  prtGraphics("%d missing MPEOE parameter%s.\n",r,r>0?"s":"");
	  MPEOEfree_scratch(G);
	  clear_node_chain(A); free_node_chain(A);
	}else if (key->status==TOK_NOK) err_set(1,"bad atom selcetion: %s",key->text);
      }else if (key_is("usage")||key_is("use")) {
	key=free_Token(key);
	key=Get_Atoms_D("show parameter usage for atoms: ",";");
	if (key->status==TOK_OK) {
	  A=(Nodechain *)key->ptr;
	  G=MPEOEsetup(A);
	  for (mpa=G.Latom;mpa;mpa=mpa->nxt) {
	    if (mpa->def) {
	      sprintf(buf1,"MPEOE atom (used:%d) %s {%0.3f%s,%0.3f%s",mpa->nu,
		      mpa->def,mpa->P[0],OPTF(mpa->opt[0]),mpa->P[1],OPTF(mpa->opt[1]));
	      if (mpa->indX) sprintf(buf1+strlen(buf1),",%0.3f%s}",
				     mpa->P[2],OPTF(mpa->opt[2]));
	      else sprintf(buf1+strlen(buf1),"}");
	      prtGraphics("%s\n",buf1);
	    }
	  }
	  for (mpb=G.Lbond;mpb;mpb=mpb->nxt) {
	    if (mpb->def) {
	      prtGraphics("MPEOE bond (used:%d) %s %0.3f%s\n",mpb->nu,mpb->def,mpb->f,OPTF(mpb->opt));
	    }
	  }
	  r=MPEOEreport_missing(G,FALSE);
	  if (r>0) err_set(1,"%d missing MPEOE parameter%s.",r,r==1?"":"s");
	  MPEOEfree_scratch(G);
	  clear_node_chain(A);free_node_chain(A);
	}else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
      }
    }else if (key->status==TOK_NOK) err_set(1,"bad mpeoe show option: %s",key->text);
  }
  free_Token(key);
}
