#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"

extern Node *DefaultNode;
extern int graphix;

#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define key_is(a) (streq(a,key->key))

#define DONE 0
#define NEXT 1
#define RESIDUE 2

#define TRUE 1
#define FALSE 0

#define N_BACKBONE 7
static char *backbone[N_BACKBONE] = 
{ "C", "N", "CA", "O", "OT1","OT2","OXT" };

#define N_MATCHED 3
static char *matched[N_MATCHED] = { "C", "N", "CA" };

static Node *template;

static char *keys = "'next' "
  "'ALA' 'ARG' 'ASN' 'ASP' 'CYS' 'GLN' 'GLU' 'GLY' 'HIS' 'ILE' "
  "'LEU' 'LYS' 'MET' 'PHE' 'PRO' 'SER' 'THR' 'TRP' 'TYR' 'VAL'";

static void do_mutation(Mol *r,char *new);

void cmd_mutate(void)
{
  Token *key;
  Mol *r;
  char buff[4096],prmt[4096],*u,*menu;
  int mode;
  Nodechain *c;
  Atom *a;

  template=(Node *)NULL;

  key=(Token *)NULL;
  menu=(char *)NULL;

  menu=Make_TopMolNameList("* /*",DefaultNode);
  key=Get_Mol_M("residue templates: ",";",menu);
  send_bheader(graphix,"ARMK",0);
  if (key->status==TOK_OK) {
    template=(Node *)key->ptr;
    if (key->source==SRC_PICK) {
      a=rfirst_atom(template);
      if (mvld(a)) {
	if (a->up==template) {
	  if (mvld(template->up)) template=template->up;
	}
      }
    }
    prtGraphics("using template residues from %s\n",
		nodeptr_to_name(buff,template));
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad mol token: %s.",key->text);
  }
  if (mvld(menu)) free_menu(menu);
  key=free_Token(key);

  r=(Mol *)NULL;
  mode=(mvld(template) ? NEXT : DONE);

  while (mode) {
    send_bheader(graphix,"ARMK",0);
    err_dump();
    switch (mode) {
    case NEXT:
      if (!mvld(key)) key=Get_Mols("residue: ",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	if (c->n==1) {
	  r=(Mol *)c->s;
	  mode=RESIDUE;
	}else{
	  err_set(1,"ambiguous residue selection: %s",key->text);
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	key=Prs_Atom(key);
	if (key->status==TOK_OK) {
	  r=(Mol *)((Atom *)key->ptr)->up;
	  mode=RESIDUE;
	}else{
	  err_set(1,"bad residue token: %s\n",key->text);
	}
      }else{
	mode=DONE;
      }
      key=free_Token(key);
      break;
    case RESIDUE:
      sprintf(prmt,"new type for %s : ",nodeptr_to_name(buff,(Node *)r));
      Strcpy(buff,r->name);
      *(buff+strlen(buff)+1)='\0';
      u=strrchr(buff,'_');
      if (u) *u='\0';
      default_token(buff);
      pick_mode("PATM");
      key=Get_Kwd(prmt,";",keys);
      pick_mode("POFF");
      if (key->status==TOK_NOK) {
	mode=NEXT;
	key=Prs_Mols(key);
      }else if (key->status==TOK_OK) {
	if (key_is("next")) {
	  mode=NEXT;
	  r=(Mol *)NULL;
	}else{
	  for (u=buff;*u;u++) *u=(*u >= 'a' && *u <= 'z' ? *u+'A'-'a' : *u); 
	  if (!streq(buff,key->key)) {
	    do_mutation(r,key->key);
	  }else{
	    prtGraphics("%s left unmodified.\n",r->name);
	  }
	  r=next_mol(r);
	  if (!mvld(r)) mode=NEXT;
	}
	key=free_Token(key);
      }else{
	mode=DONE;
      }
      break;
    default:
      mode=DONE;
      break;
    }
  }
  free_Token(key);
  send_bheader(graphix,"ARMK",0);
}
/*-------------------- do_mutation --------------------*/

static void do_mutation(Mol *r,char *new)
{
  char new_name[512],buff[512],*u;
  Nodechain *c;
  Atom *cpr,*cn,*ca,*hn,*a,*b,*satom[N_MATCHED],*tatom[N_MATCHED];
  Atomset *s;
  Bond *bnd;
  Node *tr,*nr,*pr;
  int i,dsp_flag;
  float rms,dist,ang;
  mat_4 tra;

  if (!new) new="UNK";
  if (!mvld(r)) {
    err_set(1,"do_mutation: no residue");
  }else if (!mvld(template)) {
    err_set(1,"do_mutation: residue template is not defined.");
  }else{
    /* generate new residue name */
    sprintf(new_name,"%s",new);
    if (mvld(r->name)) {
      u=strrchr(r->name,'_');
      if (u) sprintf(new_name+strlen(new_name),"%s",u);
    }
    /* find template residue */
    tr=(Node *)NULL; nr=(Node *)NULL;
    ForAtoms(a,template) {
      tr=a->up;
      a=(Atom *)tr->lst;
      Strcpy(buff,tr->name);
      if (strrchr(buff,'_')) *strrchr(buff,'_')='\0';
      if (strcasecmp(buff,new)==0) break;
      tr=(Node *)NULL;
    }
    if (mvld(tr)) {
      /* make a copy of the template residue */
      nr=tree_copy(tr);
      if (mvld(nr)) {
	forAtomsets(s,nr) del_node((Node *)s);
	nr->nxt=(Node *)r;              /* shuffle set bits... */
	link_node(nr);
	copy_set_bits(tr,nr);
	unlink_node(nr);
	/* remove inter-residue bonds in copy */
	ForAtoms(a,nr) {
	  for (i=a->nbnd-1;i>=0;i--) {
	    if (a->bnd[i]->f->up!=a->bnd[i]->t->up) del_bond(a->bnd[i]);
	  }
	}
	/* fit template residue to target residue */
	for (i=0;i<N_MATCHED;i++) tatom[i]=(Atom *)NULL;
	ForAtoms(a,r) {
	  for (i=0;i<N_MATCHED;i++) {
	    if (!strcasecmp(a->name,matched[i])) tatom[i]=a;
	  }
	}
	for (i=0;i<N_MATCHED;i++) satom[i]=(Atom *)NULL;
	ForAtoms(a,nr) {
	  for (i=0;i<N_MATCHED;i++) {
	    if (!strcasecmp(a->name,matched[i])) satom[i]=a;
	  }
	}
	for (i=0;i<N_MATCHED;i++) {
	  if (!mvld(tatom[i]) || !mvld(satom[i])) break;
	}
	if (i>=N_MATCHED) {
	  /* use atoms in satom/tatom arrays for best_fist */
	  rms=rms_best_mat(N_MATCHED,satom,tatom,&tra);
	  if (rms>=0) {
	    for (i=0;i<N_MATCHED;i++) {
	      satom[i]->flg|=(1<<1);
	      satom[i]->chain=(Node *)tatom[i];
	    }
	    /* transform atoms in template */
	    ForAtoms(a,nr) {
	      transform_atom(tra,a);
	      if (a->flg&(1<<1)) {
		/* link new atoms to target backbone */
		for (i=0;i<a->nbnd;i++) {
		  if (!(a->bnd[i]->f->flg&(1<<1))) {
		    bnd=add_bond_ng(a->bnd[i]->f,(Atom *)a->chain);
		    set_bt(bnd,BT_SINGLE);
		  }else if (!(a->bnd[i]->t->flg&(1<<1))) {
		    bnd=add_bond_ng((Atom *)a->chain,a->bnd[i]->t);
		    set_bt(bnd,BT_SINGLE);
		  }
		}
	      }
	    }
	  }else{
	    err_set(1,"do_mutation: rms fit of template residue %s failed.",
		    nr->name);
	    del_node(nr); nr=(Node *)NULL;
	  }
	}else{
	  err_set(1,"do_mutation: template residue %s is incomplete.",nr->name);
	  del_node(nr); nr=(Node *)NULL;
	}
      }else{
	err_set(1,"do_mutation: memory allocation failure.");
      }
    }else{
      err_set(1,"residue %s not found in template %s\n",
	      new,nodeptr_to_name(buff,template));
    }
	  
    if (mvld(nr)) {
      /* flag backbone atoms in template ... */ 
      ForAtoms(a,nr) {
	a->chain=(Node *)NULL;
	if (mvld(a->name)) {
	  for (i=0;i<N_BACKBONE;i++) {
	    if (!strcasecmp(a->name,backbone[i])) a->flg|=(1<<1);
	  }
	}
      }
      /* ...and sidechain atoms in target */
      dsp_flag=0;
      ForAtoms(a,r) {
	a->flg|=(1<<1);
	if (mvld(a->name)) {
	  for (i=0;i<N_BACKBONE;i++) {
	    if (!strcasecmp(a->name,backbone[i])) a->flg&=~(1<<1);
	  }
	  dsp_flag=dsp_flag||a->dsp;
	}
      }
      /* move atoms from template to target */
      for (a=rfirst_atom(nr);a;) {
	b=a; a=rnext_atom(r,a);
	unlink_node((Node *)b);
	b->up=(Node *)r;
	link_node((Node *)b);
	b->dsp=dsp_flag;
      }
      /* fix HN  position (except for N-terminus, and PRO) */
      if (strcasecmp(new,"PRO")) {
	for (pr=r->prv;mvld(pr);pr=pr->prv) if (is_mol(pr)) break;
	if (mvld(pr)) {
	  forAtoms(cpr,pr) {
	    if (mvld(cpr->name)) {
	      if (strcasecmp(cpr->name,"C")==0) break;
	    }
	  }
	  cn=ca=hn=(Atom *)NULL;
	  forAtoms(a,r) {
	    if (!(a->flg&(1<<1))&&mvld(a->name)) {
	      if (strcasecmp(a->name,"N")==0)  cn=a;
	      if (strcasecmp(a->name,"CA")==0) ca=a;
	      if (strcasecmp(a->name,"HN")==0) hn=a;
	      if (strcasecmp(a->name,"H")==0)  hn=a;
	    }
	  }
	  if (mvld(cpr)&&mvld(cn)&&mvld(ca)&&mvld(hn)) {
	    if (find_bond(cn,hn)) {
	      dist=cdist(cn->x,hn->x);
	      ang=cangled(ca->x,cn->x,hn->x);
	      cnstr(hn->x,cpr->x,ca->x,cn->x,dist,ang,180.0);
	    }
	  }
	}
      }
      /* delete flagged atoms from target */
      for (a=rfirst_atom(r);a;) {
	b=a; a=rnext_atom(r,a);
	if (b->flg&(1<<1)) del_node((Node *)b);
      }
      /* send new atoms to graphics box */
      c=make_node_chain();
      if (mvld(c)) {
	ForAtoms(a,r) {
	  if (!cPtr(a->cross)) add_to_node_chain((Node *)a,c);
	}
	if (c->n>0) send_atom_chain(c);
	clear_node_chain(c);
	free_node_chain(c);
      }else{
	err_set(1,"memory allocation failed...");
      }
      prtGraphics("%s modified to %s\n",r->name,new_name);
      set_node_name((Node *)r,new_name);
      /* delete template copy */
      del_node(nr);
    }
  }
  err_dump();
}
