#include <math.h>
#include "strutil.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "object.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"
#include "nmr.h"
#include "insane.h"
#include "peaks.h"

extern int graphix;
extern Node *MolRoot;
extern Dspobj *DspRoot;
extern Node *DefaultNode;

/* ==================== NMR module ==================== */

#define tFLG (1<<3)

#define key_is(x) (streq(key->key,x))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define avtxt(t) (t==NOE_AVTYPE_CENTER ? "center" :\
                  t==NOE_AVTYPE_R6 ? "R6" :\
                  t==NOE_AVTYPE_R3 ? "R3" :\
                  t==NOE_AVTYPE_SUM ? "sum" :\
                  "unknown")

#define MAXBUF 4096

#define DONE 0
#define LOOP 1

#define ADD_PSEUDO    1
#define REMOVE_PSEUDO 2

static void cmd_nmr_noe_write(Mol *root);
static void cmd_nmr_noe(Mol *root);
static void cmd_nmr_noe_add(Mol *root);
static void cmd_nmr_noe_average(Mol *root);
static void cmd_nmr_noe_read(Mol *root);
static void cmd_nmr_noe_color(void);      
static void cmd_nmr_noe_label(void);      
static void cmd_nmr_noe_style(void);      
static void cmd_nmr_noe_threshold(void);  
static void cmd_nmr_noe_list(Mol *root);
static void cmd_nmr_noe_summary(Mol *root);
static void cmd_nmr_noe_missing(Mol *root);
static void cmd_nmr_noe_display(Mol *root);
static void cmd_nmr_noe_delete(Mol *root);
static void cmd_nmr_noe_width(void);
static void cmd_nmr_noe_activate(Mol *root);
static void cmd_nmr_noe_pseudoatoms(Mol *root);
static void cmd_nmr_noe_peaklist(Mol *root);
static Mol *cmd_nmr_new(void);
static void cmd_nmr_status(Mol *root);
static void cmd_nmr_structure(Mol *root);
static void cmd_nmr_confa(Mol *root);
static void cmd_nmr_structure_read(Mol *root);
static void cmd_nmr_structure_delete(Mol *root);
static void cmd_nmr_structure_average(Mol *root);
static void add_pseudo(Mol *mol,char *seg,char *res,char *def,char *atom);
static void remove_pseudo(Mol *mol,char *res,char *atom);
static int  find_all_noes(Mol *root,int active);
static int  find_1noe(NOEdata *noe,Mol *mol);
static void send_bnoe(NOEdata *noe,int col,float viol);
static void delete_noe_display(Mol *mol);
static void set_noe_display(Mol *mol,int label,int style,int width);
static float NoeAvDist(NOEdata *noe,Mol *m);
static Nodechain NoeAtoms(Mol *m,char *seg,char *res,char *def);

/*-------------------------------------------------------------*/

#define NOE_LAB_NONE 0
#define NOE_LAB_VIOL 1
#define NOE_LAB_LIM  2

#define NOE_STYLE_FULL   0
#define NOE_STYLE_DOTTED 1
#define NOE_STYLE_BROKEN 2
#define NOE_STYLE_MIXED  3

static float noe_thresh=0.5;
static int noe_color=2;          /* green, by convention */
static int noe_label=NOE_LAB_NONE; 
static int noe_style=NOE_STYLE_BROKEN;
static int noe_width=1;

/* ============================================================ */

void cmd_nmr(void)
{
  Token *key;
  int state;
  Node *oldDefault;
  char *main_keys;
  Mol *NmrRoot=(Mol *)NULL;
  
  oldDefault=DefaultNode;
  NmrRoot=cmd_nmr_new();
  if (end_parse()!=TOK_OK) 
    err_set(1,"extra characters at end of command ignored.");
  err_dump();

  state=LOOP;
  key=(Token *)NULL;
  main_keys = "h!new noes spectrum shifts status structure ' ' Command ' '"
    "bond color CPK define delete display freeze label list measure modify reset select set ' ' End";

  if (insane_sanity_check(NmrRoot,TRUE)<1) state=DONE;
  while(state!=DONE && mvld(NmrRoot)) {
    start_parse();
    key=free_Token(key);
    key=Get_Kwd("NMR> ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("noes"))            cmd_nmr_noe(NmrRoot);
      else if (key_is("spectrum"))   {
	insane_spectrum(NmrRoot);
	if (insane_sanity_check(NmrRoot,TRUE)<1) state=DONE;
      }
      else if (key_is("shifts"))     insane_shifts(NmrRoot);
      else if (key_is("new"))        NmrRoot=cmd_nmr_new();
      else if (key_is("status"))     cmd_nmr_status(NmrRoot);
      else if (key_is("structure"))  cmd_nmr_structure(NmrRoot);
      else if (key_is("Command"))    {
	command_loop("Wit!P(NMR)> ");
	if (insane_sanity_check(NmrRoot,TRUE)<1) state=DONE;
      }
      else if (key_is("bond"))       cmd_bond();
      else if (key_is("color"))      cmd_color();
      else if (key_is("CPK"))        cmd_cpk();
      else if (key_is("define"))     cmd_define();
      else if (key_is("delete"))     {
	cmd_delete();
	if (insane_sanity_check(NmrRoot,TRUE)<1) state=DONE;
      }
      else if (key_is("display"))    cmd_display();
      else if (key_is("freeze"))     cmd_freeze();
      else if (key_is("label"))      cmd_label();
      else if (key_is("list"))       cmd_list();
      else if (key_is("measure"))    cmd_measure();
      else if (key_is("modify"))     cmd_modify();
      else if (key_is("reset"))      cmd_display_reset();
      else if (key_is("select"))     cmd_select();
      else if (key_is("set"))        cmd_set();
      else if (key_is("End"))        state=DONE;
    }else if (key->status==TOK_NOK) {
      err_set(1,"NMR: bad option: %s",key->text);
    }
    if (end_parse()!=TOK_OK) 
      err_set(1,"extra characters at end of command ignored.");
    err_dump();
  }

  key=free_Token(key);
  DefaultNode=oldDefault;
}
/*-------------------- cmd_nmr_noe --------------------*/

static void cmd_nmr_noe(Mol *NmrRoot)
{
  Token *key;
  char *menu;

  menu="add average color delete display label list missing peaklist "
       "pseudoatoms read status style threshold width write 'end'";

  for(;;) {
    key=Get_Kwd("NOE option> ",";",menu);
    if (key->status==TOK_OK) {
      if (key_is("read"))            cmd_nmr_noe_read(NmrRoot);
      else if (key_is("add"))        cmd_nmr_noe_add(NmrRoot);
      else if (key_is("average"))    cmd_nmr_noe_average(NmrRoot);
      else if (key_is("write"))      cmd_nmr_noe_write(NmrRoot);
      else if (key_is("color"))      cmd_nmr_noe_color();      
      else if (key_is("label"))      cmd_nmr_noe_label();      
      else if (key_is("style"))      cmd_nmr_noe_style();      
      else if (key_is("threshold"))  cmd_nmr_noe_threshold();  
      else if (key_is("delete"))     cmd_nmr_noe_delete(NmrRoot); 
      else if (key_is("list"))       cmd_nmr_noe_list(NmrRoot);   
      else if (key_is("missing"))    cmd_nmr_noe_missing(NmrRoot); 
      else if (key_is("display"))    cmd_nmr_noe_display(NmrRoot);
      else if (key_is("width"))      cmd_nmr_noe_width();
      else if (key_is("status"))     cmd_nmr_noe_activate(NmrRoot);
      else if (key_is("pseudoatoms"))cmd_nmr_noe_pseudoatoms(NmrRoot);
      else if (key_is("peaklist"))   cmd_nmr_noe_peaklist(NmrRoot);
      else if (key_is("end"))        break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad NOE option: %s",key->text);
    }else{
      break;
    }
    end_parse();
    err_dump();
  }
}
/*-------------------- cmd_nmr_noe_read --------------------*/

static void cmd_nmr_noe_read(Mol *NmrRoot)
{
  cmd_read((Node *)NmrRoot,FT_DUPPER|FT_NOE);
}
/*-------------------- cmd_nmr_noe_add --------------------*/

static void cmd_nmr_noe_add(Mol *NmrRoot)
{
  Token *key,*atm[2];
  NOElist *noes;
  NOEdata *noe;
  char *menu,*ulimit_menu;
  Nodechain c;
  Dspobj *proot;
  Atom *from,*to;
  Node *o;
  int n,i,lvf,lvt;
  float limit;

  if (!is_mol(NmrRoot)) {
    err_set(1,"no root defined.");
    return;
  }
  if (NmrRoot->moltype!=MOL_NMRSET) {
    err_set(1,"NMR root is fugu.");
    return;
  }

  key=(Token *)NULL;
  menu=(char *)NULL;
  ulimit_menu="0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 6.0 7.0 8.0 9.0 10.0";

  c.n=0; c.s=(Node *)NULL; c.e=c.s;

  forNOElists(noes,NmrRoot) add_to_node_chain((Node *)noes,&c);
  noes=(NOElist *)NULL;
  if (c.n > 0) {
    free_menu(menu);
    menu=Make_NodeNameList("",&c); clear_node_chain(&c);
    key=Get_String_M("NOE list to edit: ",";",menu);
    if (key->status==TOK_OK) {
      forNOElists(noes,NmrRoot) {
	if (wild_match(noes->name,key->text)) break;
      }
      if (!is_NOElist(noes)) {
	err_set(1,"list %s not found.",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
      return;
    }
  }
  if (!is_NOElist(noes)) {
    err_dump();
    key=free_Token(key);
    key=Get_String("new list, enter name: ",";");
    if (key->status==TOK_OK) {
      noes=(NOElist *)add_node((Node *)NmrRoot,key->text,OBJ_NOELIST);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad list name: %s",key->text);
    }
  }

  if (is_NOElist(noes)) {
    for (i=0;i<2;i++) atm[i]=(Token *)NULL;
    proot=pick_root(assocmdl((Node *)NmrRoot));
    for(n=0;;) {
      for (i=0;i<n;i++) free_Token(atm[i]);
      err_dump();
      n=Get_nAtoms(2,atm,"pick atom pair: ",";");
      send_bheader(graphix,"ARMK",0);
      if (n==2) {
	from=(Atom *)atm[0]->ptr;
	to=(Atom *)atm[1]->ptr;
	key=free_Token(key);
	key=Get_Float_M("upper limit = ",";",ulimit_menu);
	if (key->status==TOK_OK) {
	  limit=key->f;
	  for (o=(Node *)from,lvf=0;mvld(o);o=o->up,lvf++) if (o->up==(Node *)NmrRoot) break;
	  if (!mvld(o)) {
	    err_set(1,"first atom is not in NMR set.");
	    continue;
	  }
	  for (o=(Node *)to,lvt=0;mvld(o);o=o->up,lvt++) if (o->up==(Node *)NmrRoot) break;
	  if (!mvld(o)) {
	    err_set(1,"second atom is not in NMR set.");
	    continue;
	  }
	  noe=add_noedata(noes);
	  if (mvld(noe)) {
	    Strcpy(noe->sres,from->up->name);
	    Strcpy(noe->tres,to->up->name);
	    Strcpy(noe->satom,from->name);
	    Strcpy(noe->tatom,to->name);
	    if (lvf>2) noe->sseg=Strdup(from->up->up->name);
	    if (lvt>2) noe->tseg=Strdup(to->up->up->name);
	    noe->dupper=limit;
	    noe->weight=1.0;
	  }else{
	    err_set(1,"failed to add NOE to list.");
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad float: %s",key->text);
	}
      }else{ 
	if (atm[0]->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",atm[n]->text);
	}else{
	  break;
	}
      }
    }
    for (i=0;i<n;i++) free_Token(atm[i]);
    pick_root(proot);
  }else{
    err_set(1,"no NOE list.");
  }

  free_menu(menu);
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_average --------------------*/

static void cmd_nmr_noe_average(Mol *NmrRoot)
{
  char *menu;
  Token *key;
  Nodechain c;
  NOElist *noes;
  int wait_flag,nset,avtype=0;

  menu=(char *)NULL;
  key=(Token *)NULL;
  c.n=0; c.s=(Node *)NULL; c.e=c.s;

  key=Get_Kwd("distance averaging = ",";","center R3 R6 sum");
  if (key->status==TOK_OK) {
    if (key_is("center")) avtype=NOE_AVTYPE_CENTER;
    else if (key_is("R3")) avtype=NOE_AVTYPE_R3;
    else if (key_is("R6")) avtype=NOE_AVTYPE_R6;
    else if (key_is("sum")) avtype=NOE_AVTYPE_SUM; 
    for (wait_flag=TRUE;;wait_flag=FALSE) {
      forNOElists(noes,NmrRoot) add_to_node_chain((Node *)noes,&c);
      if (c.n > 0) {
	free_menu(menu);
	menu=Make_NodeNameList("* ' ' ",&c);
	key=free_Token(key);
	key=Get_String_MW("NOE list to modifiy: ",";",menu,wait_flag);
	if (key->status==TOK_OK) {
	  nset=0;
	  for (noes=(NOElist *)c.s;mvld(noes);noes=(NOElist *)c.s) {
	    c.n--; c.s=noes->chain; if (!mvld(c.s)) c.e=c.s; 
	    noes->chain=(Node *)NULL;
	    if (wild_match(noes->name,key->text)) {
	      if (avtype!=noes->avtype) {
		noes->avtype=avtype;
		nset++;
	      }
	    }
	  }
	  prtGraphics("%d NOE set(s) modified.\n",nset);
	  if (streq(key->text,"*")) break;
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token: %s",key->text);
	}else break;
      }else{
	prtGraphics("No NOEs in %s\n",NmrRoot->name);
	break;
      }
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad averaging option: %s",key->text);

  free_menu(menu);
  if (c.n>0) clear_node_chain(&c);
  key=free_Token(key);
}

/*-------------------- cmd_nmr_noe_activate --------------------*/

static void cmd_nmr_noe_activate(Mol *NmrRoot)
{
  Token *key;
  int state,non,noff;
  Atom *a;
  NOElist *noes;
  NOEdata *noe;
  Node *sv_default;
  Mol *m;
  Nodechain *c;
  Dspobj *proot;

  m=first_mol(NmrRoot);
  if (!mvld(m)) {
    err_set(1,"must read in structures first...");
    return;
  }

  key=Get_Kwd("new state: ",";","on off");
  if (key->status==TOK_OK) {
    state=key->n==0;
    err_dump();
    key=free_Token(key);
    
    sv_default=DefaultNode; DefaultNode=(Node *)m;
    proot=pick_root(assocmdl((Node *)m));
    key=Get_Atoms_D("noe atoms: ",";");
    pick_root(proot);
    DefaultNode=sv_default;
    if (key->status==TOK_OK) {
      non=0;noff=0;
      c=(Nodechain *)key->ptr;
      if (c->n>0) {
	for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg|=tFLG;
	find_all_noes(m,3);
	forNOElists(noes,NmrRoot) {
	  for (noe=noes->first;mvld(noe);noe=noe->nxt) {
	    if (noe->sptr) {
	      if (noe->sptr->flg&tFLG) {
		noe->active=state;
		if (noe->active) non++;
		else noff++;
		continue;
	      }
	    }
	    if (noe->tptr) {
	      if (noe->tptr->flg&tFLG) {
		noe->active=state;
		if (noe->active) non++;
		else noff++;
		continue;
	      }
	    }
	  }
	}
	for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg&=~tFLG;
      }
      clear_node_chain(c);
      free_node_chain(c);
      prtGraphics("on = %d, off = %d\n",non,noff);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
    }
    err_dump();
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad state token: %s",key->text);
  }

  free_Token(key);
}
/*-------------------- cmd_nmr_noe_pseudoatoms --------------------*/

static void cmd_nmr_noe_pseudoatoms(Mol *NmrRoot)
{
  Token *key;
  char *menu,*prmt=(char *)NULL;
  int wait_flag,loop,option=0,i,M;
  Mol *mol;
  Node **m;
  Atom *a;
  Nodechain *c;
  NOElist *noes;
  NOEdata *noe;

  menu="add remove";

  key=Get_Kwd("add or remove ? ",";",menu);
  if (key->status==TOK_OK) {
    if (key_is("add")) {
      prmt="add pseudoatoms to: ";
      option=ADD_PSEUDO;
    }else if (key_is("remove")) {
      prmt="remove pseudoatoms from: ";
      option=REMOVE_PSEUDO;
    }
    for (loop=TRUE,wait_flag=TRUE;loop;wait_flag=FALSE) {
      err_dump();
      key=free_Token(key);
      key=Get_NMRStructures(NmrRoot,prmt,"* ' ' ",wait_flag);
      if (key->status==TOK_OK) {
	if (streq(key->text,"*")) loop=FALSE;
	M=((Nodechain *)key->ptr)->n;
	m=chain_to_ptr((Nodechain *)key->ptr,TRUE);
	for (i=0;i<M;i++) {
	  mol=(Mol *)m[i];
	  forNOElists(noes,NmrRoot) {
	    switch (option) {
	    case ADD_PSEUDO:
	      for(noe=noes->first;mvld(noe);noe=noe->nxt) {
		add_pseudo(mol,noe->sseg,noe->sres,noe->def1,noe->satom);
		add_pseudo(mol,noe->tseg,noe->tres,noe->def2,noe->tatom);
	      }
	      break;
	    case REMOVE_PSEUDO:
	      for(noe=noes->first;mvld(noe);noe=noe->nxt) {
		if (noe->ps1) remove_pseudo(mol,noe->sres,noe->satom);
		if (noe->ps2) remove_pseudo(mol,noe->tres,noe->tatom);
	      }
	      break;
	    }
	  }
	  c=collect_by_flags((Node *)mol,tFLG);
	  if (mvld(c)) {
	    switch (option) {
	    case ADD_PSEUDO:
	      for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) a->flg&=~tFLG;
	      send_atom_chain(c);
	      clear_node_chain(c);
	      free_node_chain(c);
	      break;
	    case REMOVE_PSEUDO:
	      del_node_chain(c);
	      free_node_chain(c);
	      break;
	    }
	  }else{
	    err_set(1,"memory allocation failed...");
	  }
	}
	if (m) {
	  free(m);
	  m=(Node **)NULL;
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad structure selection: %s.",key->text);
      }else{
	break;
      }
    }

  }else if (key->status==TOK_NOK) {
    err_set(1,"bad pseudoatom option: %s",key->text);
  }

  free_Token(key);
  err_dump();
}

/*-------------------- cmd_nmr_noe_write --------------------*/

       /*     K L U D G E      */

int write_discover_noe(char *filename,NOElist *noes);
int write_sodgeom_noe(char *filename,NOElist *noes);

static void cmd_nmr_noe_write(Mol *NmrRoot)
{
  Token *key;
  int s;
  NOElist *noes;
  Mol *m;

  prtGraphics("THIS IS A KLUDGE ! Use at your own risk.\n");
  key=(Token *)NULL;

  noes=first_NOElist(NmrRoot);
  if (!mvld(noes)) {
    prtGraphics("there are no NOEs in %s.\n",NmrRoot->name);
    return;
  }

  key=free_Token(key);
  key=Get_Kwd("filetype: ",";","Xplor Discover Sodgeom");
  if (key->status==TOK_OK) {
    if (key_is("Xplor")) {
      key=free_Token(key);
      key=Get_String("NOE  K L U D G E  file to write: ",";");
      if (key->status==TOK_OK) {
	s=write_xplor_noe(key->text,noes);
	if (!s) prtGraphics("%s written.\n",key->text);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token (NOE filename): %s",key->text);
      }
    }else if (key_is("Discover")) {
      key=free_Token(key);
      key=Get_String("NOE  K L U D G E  file to write: ",";");
      if (key->status==TOK_OK) {
	s=write_discover_noe(key->text,noes);
	if (!s) prtGraphics("%s written.\n",key->text);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token (NOE filename): %s",key->text);
      }
    }else if (key_is("Sodgeom")) {
      key=free_Token(key);
      key=Get_String("NOE  K L U D G E  file to write: ",";");
      if (key->status==TOK_OK) {
	m=first_mol(NmrRoot);
	if (m) {
	  find_all_noes(m,3);
	  s=write_sodgeom_noe(key->text,noes);
	  if (!s) prtGraphics("%s written.\n",key->text);
	}else err_set(1,"need a molecule to write SODGEOM file\n");
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token (NOE filename): %s",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"unrecognizable filetype: %s\n",key->text);
  }

  key=free_Token(key);
}
/*-------------------- cmd_nmr_noe_label --------------------*/

static void cmd_nmr_noe_label(void)
{
  Token *key;
  
  if (noe_label==NOE_LAB_NONE)       default_token("none");
  else if (noe_label==NOE_LAB_VIOL)  default_token("violation");
  else if (noe_label==NOE_LAB_LIM)   default_token("limit");

  key=Get_Kwd("NOE label type: ",";","none limit violation");
  if (key->status==TOK_OK) {
    if (key_is("none")) noe_label=NOE_LAB_NONE;
    else if (key_is("limit")) noe_label=NOE_LAB_LIM;
    else if (key_is("violation")) noe_label=NOE_LAB_VIOL;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad label type option: %s.",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_width --------------------*/

static void cmd_nmr_noe_width(void)
{
  Token *key;

  default_token("%d",noe_width);
  key=Get_Int_M("NOE linewidth: ",";","1 2 3 4 5 6 7 8 9 10");
  if (key->status==TOK_OK) {
    if (key->n>0 && key->n<32) {
      noe_width=key->n;
    }else{
      err_set(1,"linewidth parameter out of range: %s",key->text);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad NOE linewidth token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_color --------------------*/

static void cmd_nmr_noe_color(void)
{
  Token *key;

  default_token(ColMnem(noe_color));
  key=Get_Color("NOE display color: ",";");
  if (key->status==TOK_OK) {
    noe_color=key->n;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad color token: %s.",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_style --------------------*/

static void cmd_nmr_noe_style(void)
{
  Token *key;

  if (noe_style==NOE_STYLE_BROKEN)      default_token("broken");
  else if (noe_style==NOE_STYLE_DOTTED) default_token("dashed");
  else if (noe_style==NOE_STYLE_FULL)   default_token("full");
  else if (noe_style==NOE_STYLE_MIXED)  default_token("mixed");

  key=Get_Kwd("NOE display style: ",";","broken dashed full mixed");
  if (key->status==TOK_OK) {
    if (key_is("broken"))      noe_style=NOE_STYLE_BROKEN;
    else if (key_is("dashed")) noe_style=NOE_STYLE_DOTTED;
    else if (key_is("full"))   noe_style=NOE_STYLE_FULL;
    else if (key_is("mixed"))   noe_style=NOE_STYLE_MIXED;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad NOE display style token: %s.",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_threshold --------------------*/

static void cmd_nmr_noe_threshold(void)
{
  Token *key;

  default_token("%5.2f",noe_thresh);
  key=Get_Float_M("NOE list/display threshold: ",";",
		"'0.00' '0.05' '0.10' '0.15' '0.20' '0.25' '0.3'"
		"'0.4' '0.5' '0.6' '0.7' '0.8' '0.9' '1.0'");

  if (key->status==TOK_OK) {
    noe_thresh=key->f;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad NOE threshold token: %s.",key->text);
  }
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_list --------------------*/

static void cmd_nmr_noe_list(Mol *NmrRoot)
{
  NOElist *noes;
  NOEdata *noe;
  Mol *mol;
  Token *key;
  Nodechain *m;
  float D,d,SV,SV2;
  char buff2[512],*hdr2,*pref;
  int listed,ntfnd,wait_flag,NV,nnoe;

  key=(Token *)NULL;
  pref="-summary ' ' * ' '";
  for (wait_flag=TRUE;;wait_flag=FALSE) {
    err_dump();
    key=Get_NMRStructures(NmrRoot,"list NOE for: ",pref,wait_flag);
    if (key->status==TOK_OK) {
      m=(Nodechain *)key->ptr;
      for (mol=(Mol *)m->s;mvld(mol);mol=(Mol *)mol->chain) {
	prtGraphics("\n");
	prtGraphics("NOE Violations > %0.3f for %s: ",noe_thresh,mol->name);
	ntfnd=0;listed=0;
	SV=0;SV2=0;NV=0; nnoe=0;
	forNOElists(noes,NmrRoot) {
	  sprintf(buff2,"(NOE list %s, averaging = %s)",noes->name,avtxt(noes->avtype)); hdr2=buff2;
	  for(noe=noes->first;mvld(noe);noe=noe->nxt) {
	    if (!noe->active) continue;
	    D=NoeAvDist(noe,mol); nnoe++;
	    if (D>-0.5) {
	      d=(D > noe->dupper ? D-noe->dupper : 0.0);
	      SV+=d;
	      SV2+=d*d;
	      if (d >= noe_thresh) { 
		NV++;
		if (hdr2) {
		  prtGraphics("%s\n",hdr2);
		  hdr2=(char *)NULL;
		}
		prtGraphics(
			    "%-10s %-5s ---   %-10s %-5s = "
			    "%5.2f ( %5.2f + %4.2f ) %s\n",
			    noe->sres,noe->satom,noe->tres,noe->tatom,
			    D,noe->dupper,d,noe->cmnt?noe->cmnt:"");
		listed++;
	      }
	    }else{
	      ntfnd++;
	    }
	  }
	}
	if (ntfnd>0) {
	  err_set(1,"%d NOEs not found in structure %s",ntfnd,mol->name);
	}
	if (listed==0) {
	  prtGraphics("no NOE violations > %5.2f found for %s.\n",
		      noe_thresh,mol->name);
	}else{
	  prtGraphics("rms(viol) = %0.3f, mean(viol) = %0.4f, n(viol) = %2d\n",
			sqrt(SV2/nnoe),SV/nnoe,NV);
	}
	err_dump();
      }
      clear_node_chain(m);
      free_node_chain(m);
    }else if (key->status==TOK_NOK) {
      key=Prs_Kwd(key,"-summary");
      if (key->status==TOK_OK) {
	if (key_is("-summary")) {
	  cmd_nmr_noe_summary(NmrRoot);
	  break;
	}
      }else err_set(1,"bad structure selection: %s.",key->text);
    }else{
      break;
    }
    key=free_Token(key);
    err_dump();
    pref="* ' ' ";
  }
  err_dump();
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_summary --------------------*/

static void cmd_nmr_noe_summary(Mol *NmrRoot)
{
  NOElist *noes;
  NOEdata *noe;
  Mol *mol;
  Token *key;
  Nodechain *m;
  float D,d,*sv,*sv2,*SV,*SV2;
  int i,k,*nv,nnoe,*NV;
  char buff2[512],*hdr2;

  nv=(int *)NULL; sv=(float *)NULL; sv2=(float *)NULL;
  NV=(int *)NULL; SV=(float *)NULL; SV2=(float *)NULL;
  key=(Token *)Get_NMRStructures(NmrRoot,"list NOEs for: ","* ' ' ",TRUE);
  if (key->status==TOK_OK) {
    m=(Nodechain *)key->ptr;
    if (m->n>0) {
      if (SV)  {free(SV); }  SV=(float *)malloc(m->n*sizeof(float));
      if (SV2) {free(SV2);}  SV2=(float *)malloc(m->n*sizeof(float));
      if (NV)  {free(NV); }  NV=(int *)malloc(m->n*sizeof(float));
    }
    nnoe=0; 
    forNOElists(noes,NmrRoot) {
      for(noe=noes->first;mvld(noe);noe=noe->nxt) if (noe->active) nnoe++;
    }
    if (nnoe>0) {
      sv=(float *)malloc(nnoe*sizeof(float));
      sv2=(float *)malloc(nnoe*sizeof(float));
      nv=(int *)malloc(nnoe*sizeof(int));
    }
    if (mvld(sv)&&mvld(nv)&&mvld(sv2)) {
      for (i=0;i<nnoe;i++) {
	nv[i]=0;
	sv[i]=0.0;
	sv2[i]=0.0;
      }
      for (k=0,mol=(Mol *)m->s;mvld(mol);mol=(Mol *)mol->chain,k++) {
	i=0; if (SV) SV[k]=0; if (SV2) SV2[k]=0; if (NV) NV[k]=0;
	forNOElists(noes,NmrRoot) {
	  for(noe=noes->first;mvld(noe);noe=noe->nxt) {
	    if (noe->active) {
	      D=NoeAvDist(noe,mol);
	      d=(D > noe->dupper ? D-noe->dupper : 0.0);
	      if (d >= noe_thresh) {
		nv[i]++;
		if (NV) NV[k]++;
	      }
	      sv2[i]+=d*d; if (SV2) SV2[k]+=d*d;
	      sv[i]+=d; if (SV) SV[k]+=d;
	      i++;
	    }
	  }
	}
      }
      mol=(Mol *)m->s;
      if (mol) {
	i=0;
	prtGraphics("\n");prtGraphics("Summary of NOE Violations > %0.3f:\n",noe_thresh);
	forNOElists(noes,NmrRoot) {
	  sprintf(buff2,"(NOE list %s, averaging = %s)",noes->name,avtxt(noes->avtype)); hdr2=buff2;
	  for(noe=noes->first;mvld(noe);noe=noe->nxt) {
	    if (noe->active) {
	      if (nv[i]>0) {
		if (hdr2) prtGraphics("%s\n",hdr2);
		prtGraphics(
			    "%-10s %-5s ---   %-10s %-5s < "
			    "%5.2f ( %5.3f, %5.3f, %2d ) %s\n",
			    noe->sres,noe->def1,noe->tres,noe->def2,
			    noe->dupper,sqrt(sv2[i]/m->n),sv[i]/m->n,nv[i],
			    noe->cmnt?noe->cmnt:"");
		hdr2=(char *)NULL;
	      }
	      i++;
	    }
	  }
	}
	if (SV&&SV2&&NV) {
	  prtGraphics("\n");
	  prtGraphics("Summary of Violations > %0.3f per Structure:\n",noe_thresh);
	  for (k=0,mol=(Mol *)m->s;mvld(mol);mol=(Mol *)mol->chain,k++) {
	    prtGraphics("%-20s: rms(viol) = %0.3f, mean(viol) = %0.4f, n(viol) = %2d\n",
			mol->name,sqrt(SV2[k]/nnoe),SV[k]/nnoe,NV[k]);
	  }
	}
      }
    }
    clear_node_chain(m);
    free_node_chain(m);
  }else err_set(1,"bad structure selection: %s.",key->text);

  free_Token(key);
  if (nv) free(nv);
  if (sv) free(sv);
  if (sv2) free(sv2);
  if (SV) free(SV);
  if (SV2) free(SV2);
  if (NV) free(NV);
}
/*-------------------- cmd_nmr_noe_peaklist --------------------*/

static void cmd_nmr_noe_peaklist(Mol *NmrRoot)
{
  NOElist *noes;
  NOEdata *noe;
  Mol *mol,*res;
  Atom *a;
  FILE *out=(FILE *)NULL;
  int i,nat,nprot,nf,*pknr,pk,nr,n1,n2;
  Token *key;
  UCpipe *in=(UCpipe *)NULL;
  float *shifts,shft;
  char nm[80],cdummy[120],buf[MAXBUF];

  key=(Token *)NULL;
  shifts=(float *)NULL;
  pknr=(int *)NULL;

  mol=first_mol(NmrRoot);
  if (!is_mol(mol)) {
    err_set(1,"no structure in current NMR-root.");
    return;
  }

  nat=0; 
  ForAtoms(a,mol) {
    a->seqnr=nat;
    nat++;
  }
  if (nat==0) {
    err_set(1,"no atoms found in current structure.");
    return;
  }
    
  shifts=(float *)malloc(nat*sizeof(float));
  pknr=(int *)malloc(nat*sizeof(int));
  if (!mvld(shifts)||!mvld(pknr)) {
    err_set(1,"internal error: memory allocation failed.");
    if (mvld(shifts)) free(shifts);
    shifts=(float *)NULL;
    if (mvld(pknr)) free(pknr);
    pknr=(int *)NULL;
    return;
  }
  for (i=0;i<nat;i++) {
    pknr[i]=0;
    shifts[i]=0.0;
  }

  key=Get_File(".",".prot","list of protons: ",";");
  if (key->status==TOK_OK) {
    in=(UCpipe *)key->ptr;
    prtGraphics("file is %s\n",in->filename);
  }else if (key->status==TOK_OK) {
    err_set(1,"could not read list of protons: %s",key->text);
  }
  if (key->status==TOK_OK) {
    nprot=0; nf=0;
    for(;;) {
      if (!fgets(buf,MAXBUF-1,in->fd)) break;
      sscanf(buf,"%d %f %s %s %d",&pk,&shft,cdummy,nm,&nr);
      sprintf(cdummy,"*_%d",nr);
      a=(Atom *)NULL;
      forMols(res,mol) {
	if (wild_xmatch(res->name,cdummy)) {
	  forAtoms(a,res) {
	    if (strceq(a->name,nm)) break;
	  }
	  if (is_atom(a)) break;
	}
      }
      if (is_atom(a)) {
	nprot++;
	pknr[a->seqnr]=pk;
	shifts[a->seqnr]=shft;
      }else{
	nf++;
	err_set(1,"warning: proton %d ( *_%s/%s ) not found in structure.",pk,nr,nm);
      }
    }
    err_dump();
    if (nf>0) prtGraphics("%d protons read, %d not found in structure.\n",nprot+nf,nf);
    else prtGraphics("%d protons read.\n",nprot);
    UCpclose(in);
    if (nprot>0) {
      key=free_Token(key);
      key=Get_String("name of peaklist file: ",";");
      if (key->status==TOK_OK) {
	out=fopen(key->text,"w");
	if (!out) {
	  err_set(1,"could not open file %s for write.",key->text);
	  key->status=TOK_ABORT;
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad filename token: %s",key->text);
      }
    }else key->status=TOK_ABORT;
  }
  if (key->status==TOK_OK) {
    pk=0; nf=0; nr=0;
    find_all_noes(mol,1);          /* only active NOE's */
    forNOElists(noes,mol->up) {
      for (noe=noes->first;mvld(noe);noe=noe->nxt) {
	if (noe->active) {
	  if (is_atom(noe->sptr) && is_atom(noe->tptr)) {
	    n1=noe->sptr->seqnr;
	    n2=noe->tptr->seqnr;
	    if (pknr[n1]>0 && pknr[n2]>0) {
	      fprintf(out,"%4d %7.3f %7.3f %6d %6d\n",
		      ++pk,shifts[n1],shifts[n2],pknr[n1],pknr[n2]);
	    }else nr++;
	  }else nf++;
	}
      }
    }
    if (nf>0) err_set(1,"%d NOEs not found in structure.",nf);
    if (nr>0) err_set(1,"%d NOEs have missing atom in protons list.",nr);
    err_dump();
    prtGraphics("%d peaks written to peaklist file.\n",pk);
    fclose(out);
  }
  free_Token(key);
  if (mvld(shifts)) free(shifts);
  if (mvld(pknr)) free(pknr);
}
/*-------------------- cmd_nmr_noe_missing --------------------*/

static void cmd_nmr_noe_missing(Mol *NmrRoot)
{
  NOElist *noes;
  NOEdata *noe;
  Mol *mol;
  Token *key;
  Nodechain *m;
  char buff2[512],buff1[512],*hdr2,*hdr1;
  int ntfnd,ntfnd2,wait_flag;

  key=(Token *)NULL;
  for (wait_flag=TRUE;;wait_flag=FALSE) {
    err_dump();
    key=Get_NMRStructures(NmrRoot,"list missing NOEs for: ","*",wait_flag);
    if (key->status==TOK_OK) {
      m=(Nodechain *)key->ptr;
      for (mol=(Mol *)m->s;mvld(mol);mol=(Mol *)mol->chain) {
	sprintf(buff1,"NOEs with missing atoms in %s: ",mol->name);
	hdr1=buff1;ntfnd=0; ntfnd2=0;
	forNOElists(noes,NmrRoot) {
	  sprintf(buff2,"(NOE list %s)",noes->name); hdr2=buff2;
	  for(noe=noes->first;mvld(noe);noe=noe->nxt) {
	    if (!noe->active) continue;
	    if (NoeAvDist(noe,mol)<-0.5) {
	      if (hdr1) {
		prtGraphics("%s\n",hdr1);
		hdr1=(char *)NULL;
	      }
	      if (hdr2) {
		prtGraphics("%s\n",hdr2);
		hdr2=(char *)NULL;
	      }
	      prtGraphics(
			  "%-10s %-5s ---   %-10s %-5s < %5.2f\n",
			  noe->sres,noe->def1,noe->tres,noe->def2,noe->dupper);
	      ntfnd2++;
	    }
	    if (!find_1noe(noe,mol)) {
	      if (hdr1) {
		prtGraphics("%s\n",hdr1);
		hdr1=(char *)NULL;
	      }
	      if (hdr2) {
		prtGraphics("%s\n",hdr2);
		hdr2=(char *)NULL;
	      }
	      prtGraphics(
			  "%-10s %-5s ---   %-10s %-5s < %5.2f (pseudoatoms missing)\n",
			  noe->sres,noe->satom,noe->tres,noe->tatom,noe->dupper);
	      ntfnd++;
	    }
	  }
	}
	if (ntfnd+ntfnd2>0) {
	  err_set(1,"%d NOEs not found, %d have missing pseudoatoms in structure %s",
		  ntfnd2,ntfnd,mol->name);
	}else{
	  err_set(1,"all NOEs found in structure %s",mol->name);
	} 
      }
      clear_node_chain(m);
      free_node_chain(m);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad structure selection: %s.",key->text);
    }else{
      break;
    }
    key=free_Token(key);
    err_dump();
  }
  err_dump();
  free_Token(key);
}
/*-------------------- cmd_nmr_noe_display --------------------*/

static void cmd_nmr_noe_display(Mol *NmrRoot)
{
  NOElist *noes;
  NOEdata *noe;
  Mol *mol;
  Token *key;
  Nodechain *m;
  float D,d;
  int op,ntfnd,wait_flag,loop;

  key=(Token *)NULL;

  key=Get_Kwd("NOE display option: ",";","on off");
  op=0;
  if (key->status==TOK_OK) {
    if (key_is("on")) op=1;
    else if (key_is("off")) op=2;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad NOE display option: %s",key->text);
  }

  key=free_Token(key);

  switch (op) {

  case 1:                                  /* turn NOE display on */

    for (loop=TRUE,wait_flag=TRUE;loop;wait_flag=FALSE) {
      err_dump();
      key=Get_NMRStructures(NmrRoot,"display NOEs for: ","* ' ' ",wait_flag);
      if (key->status==TOK_OK) {
	if (streq(key->text,"*")) loop=FALSE;
	m=(Nodechain *)key->ptr;
	for (mol=(Mol *)m->s;mvld(mol);mol=(Mol *)mol->chain) {
	  delete_noe_display(mol);
	  set_noe_display(mol,noe_label,noe_style,noe_width);
	  ntfnd=0;
	  forNOElists(noes,NmrRoot) {
	    for(noe=noes->first;mvld(noe);noe=noe->nxt) {
	      if (!noe->active) continue;
	      if (find_1noe(noe,mol)) {
		D=NoeAvDist(noe,mol);
		d=(D > noe->dupper ? D-noe->dupper : 0.0);
		if (d >= noe_thresh) send_bnoe(noe,noe_color,d);
	      }else{
		ntfnd++;
	      }
	    }
	  }
	  if (ntfnd>0) {
	    err_set(1,"%d NOEs not found in structure %s",ntfnd,mol->name);
	  }
	}
	clear_node_chain(m);
	free_node_chain(m);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad structure selection: %s.",key->text);
      }else break;
      key=free_Token(key);
      err_dump();
    }
    key=free_Token(key);
    break;

  case 2:                                /* turn NOE display off */
    for(loop=TRUE,wait_flag=TRUE;loop;wait_flag=FALSE) {
      err_dump();
      key=Get_NMRStructures(NmrRoot,"remove NOE display for: ","*",wait_flag);
      if (key->status==TOK_OK) {
	if (streq(key->text,"*")) loop=FALSE;
	m=(Nodechain *)key->ptr;
	for (mol=(Mol *)m->s;mvld(mol);mol=(Mol *)mol->chain) {
	  delete_noe_display(mol);
	}
	clear_node_chain(m);
	free_node_chain(m);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad structure selection: %s.",key->text);
      }else{
	break;
      }
      key=free_Token(key);
      err_dump();
    }
    key=free_Token(key);
    break;

  default:
    break;
  }
  err_dump();
  free_Token(key); /* for good measure... */
}
/*-------------------- cmd_nmr_noe_delete --------------------*/

static void cmd_nmr_noe_delete(Mol *NmrRoot)
{
  char *menu;
  Token *key;
  Nodechain c;
  NOElist *noes;
  int wait_flag,ndel,loop;

  menu=(char *)NULL;
  key=(Token *)NULL;
  c.n=0; c.s=(Node *)NULL; c.e=c.s;
  ndel=0;

  for (loop=TRUE,wait_flag=TRUE;loop;wait_flag=FALSE) {
    forNOElists(noes,NmrRoot) add_to_node_chain((Node *)noes,&c);
    if (c.n > 0) {
      free_menu(menu);
      menu=Make_NodeNameList("* ' ' ",&c);
      key=free_Token(key);
      key=Get_String_MW("NOE list to delete: ",";",menu,wait_flag);
      if (key->status==TOK_OK) {
	if (streq(key->text,"*")) loop=FALSE;
	for (noes=(NOElist *)c.s;mvld(noes);noes=(NOElist *)c.s) {
	  c.n--; c.s=noes->chain; if (!mvld(c.s)) c.e=c.s; 
	  noes->chain=(Node *)NULL;
	  if (wild_match(noes->name,key->text)) {
	    del_node((Node *)noes);
	    ndel++;
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }else break;
    }else{
      prtGraphics("No NOEs in %s\n",NmrRoot->name);
      break;
    }
  }
  if (ndel>0) prtGraphics("%d NOE set(s) deleted.\n",ndel);
  free_menu(menu);
  if (c.n>0) clear_node_chain(&c);
  key=free_Token(key);
}
/*-------------------- cmd_nmr_status --------------------*/

static void cmd_nmr_status(Mol *NmrRoot)
{
  int nnoelists,nnoes,nmols;
  Mol *mol;
  NOElist *noes;
  NOEdata *noe;

  forMols(mol,MolRoot) {
    if (mol->moltype==MOL_NMRSET && mol==NmrRoot) break;
  }

  if (!mvld(mol)) NmrRoot=(Mol *)NULL;

  if (!mvld(NmrRoot)) {
    err_set(1,"NMR Root is invalid");
    return;
  }

  prtGraphics("Status of NMR set %s:\n",NmrRoot->name);

  /* count molecules */

  nmols=0;
  forMols(mol,NmrRoot) nmols++;
  if (nmols>0) prtGraphics("%d molecules in NMR set\n",nmols);
  else prtGraphics("no molecules in NMR set.\n");

  /* count NOE lists and NOEs */

  nnoelists=0; nnoes=0;
  forNOElists(noes,NmrRoot) {
    nnoelists++;
    for (noe=noes->first;mvld(noe);noe=noe->nxt) nnoes++;
  }
  if (nnoes > 0 || nnoelists>0) {
    prtGraphics("%d NOEs in %d lists.\n",nnoes,nnoelists);
  }
}
/*-------------------- cmd_nmr_new --------------------*/

static Mol *cmd_nmr_new(void)
{
  Token *key;
  char *menu; 
  Nodechain c;
  Mol *mol;
  Dspobj *global;
  Mol *NmrRoot;

  NmrRoot=(Mol *)NULL;
  key=(Token *)NULL;
  menu=(char *)NULL;
  init_node_chain(&c);

  forMols(mol,MolRoot) {
    if (mol->moltype==MOL_NMRSET) {
      add_to_node_chain((Node *)mol,&c);
    }
  }
  if (c.n > 0)  menu=Make_NodeNameList("",&c);

  if (c.n==1) {
    default_token(c.s->name);
    menu=strdup("");
  }else {
    menu=Make_NodeNameList("",&c);
  }
  key=Get_String_M("NMR set: ",";",menu);
  if (key->status==TOK_OK) {
    for (mol=(Mol *)c.s;mvld(mol);mol=(Mol *)mol->chain) {
      if (mvld(mol->name)) if (strcasecmp(key->text,mol->name)==0) break;
    }
    clear_node_chain(&c);
    if (!mvld(mol)) {
      mol=(Mol *)add_node(MolRoot,key->text,OBJ_MOL);
      if (mvld(mol)) {
	mol->moltype=MOL_NMRSET;
	global=new_dsp_obj(DspRoot,(char *)NULL,DSP_GLBL);
	if (mvld(global)) {
	  global->root=(Node *)mol;
	  mol->dsptr=global;
	}
	prtGraphics("created new NMR set %s.\n",key->text);
      }else{
	err_set(1,"failed to create new NMR set %s.",key->text);
      }
    }else{
      prtGraphics("using existing NMR set %s.\n",mol->name);
    }
    
    if (mvld(mol)) {
      NmrRoot=mol;
      DefaultNode=(Node *)NmrRoot;
    }else{
      NmrRoot=(Mol *)NULL;
    }
 
  }else if (key->status==TOK_NOK) {
    err_set(1,"NMR set, bad token : %s",key->text);
  }

  clear_node_chain(&c);
  free_menu(menu);
  key=free_Token(key);
  return(NmrRoot);
}
/*-------------------- cmd_nmr_structure --------------------*/

static void cmd_nmr_structure(Mol *NmrRoot)
{
  Token *key;

  key=(Token *)NULL;

  for (;;) {
    key=Get_Kwd("NMR structure option: ",";","average confa delete fit read 'end'");
    if (key->status==TOK_OK) {
      if (key_is("read"))         cmd_nmr_structure_read(NmrRoot);
      else if (key_is("delete"))  cmd_nmr_structure_delete(NmrRoot);
      else if (key_is("average")) cmd_nmr_structure_average(NmrRoot);
      else if (key_is("confa"))   cmd_nmr_confa(NmrRoot);
      else if (key_is("fit"))     cmd_fit("fit");
      else if (key_is("end"))     break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad structure option: %s",key->text);
    }else break;
    key=free_Token(key);
    end_parse();
    err_dump();
  }
  free_Token(key);
}
/*-------------------- cmd_nmr_structure_read --------------------*/

static void cmd_nmr_structure_read(Mol *NmrRoot)
{
  cmd_read((Node *)NmrRoot,FT_MOL_TYPES);
}
/*-------------------- cmd_nmr_structure_delete --------------------*/

static void cmd_nmr_structure_delete(Mol *NmrRoot)
{
  Token *key;
  Mol *mol;
  Node *obj;
  Nodechain *c;
  int ndel,wait_flag,loop;

  key=(Token *)NULL;
  ndel=0;

  for (loop=TRUE,wait_flag=TRUE;loop;wait_flag=FALSE) {
    key=Get_NMRStructures(NmrRoot,"structures to delete: ","* ' ' ",wait_flag);
    if (key->status==TOK_OK) {
      if (streq(key->text,"*")) loop=FALSE;
      c=(Nodechain *)key->ptr;
      if (c->n > 0) lock_evtq();
      for (mol=(Mol *)c->s;mvld(mol);) {
	obj=(Node *)mol;
	mol=(Mol *)mol->chain;
	del_node(obj); 
	ndel++;
      }
      if (c->n > 0) unlock_evtq();
      free_node_chain(c);
      key=free_Token(key);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
      key=free_Token(key);
    }else{
      key=free_Token(key);
      break;
    }
    err_dump();
  }
  if (ndel>0) {
    prtGraphics("%d structure(s) deleted.\n",ndel);
    insane_sanity_check(NmrRoot,FALSE);
  }
  key=free_Token(key);
}
/*-------------------- cmd_nmr_structure_average --------------------*/

static void cmd_nmr_structure_average(Mol *NmrRoot)
{
  Token *key;
  Node *mol,*new,**mols;
  Nodechain *c;
  int j,n,nsel,wait_flag;
  char *new_name;
  Atom *a,*b;
  Mol *m;

  new_name=(char *)NULL;
  key=(Token *)NULL;
  mols=(Node **)NULL;

  n=0;forMols(m,NmrRoot) n++;
  if (n<1) {
    err_set(1,"no molecules in NMR set %s",NmrRoot->name);
    return;
  }
  mols=(Node **)malloc(n*sizeof(Node *));
  if (!mvld(mols)) {
    err_set(1,"memory allocation failed...");
    return;
  }

  key=Get_String("name of average structure: ",";");
  if (key->status==TOK_OK) {
    new_name=Strdup(key->text); key->text=(char*)NULL;
    nsel=0;
    for (wait_flag=TRUE;;wait_flag=FALSE) {
      key=free_Token(key);
      key=Get_NMRStructures(NmrRoot,"structures used in average: ","",wait_flag);
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (n=0;n<nsel;n++) add_to_node_chain(mols[n],c);
	for (n=0,mol=c->s;mvld(mol);mol=mol->chain) mols[n++]=mol;
	nsel=n;
	clear_node_chain(c);free_node_chain(c);
	key=free_Token(key);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }else if (key->status==TOK_EOC||key->status==TOK_EOL) {
	if (nsel>0) {
	  new=tree_copy(mols[0]);
	  if (mvld(new)) {
	    set_node_name(new,new_name);
	    ForAtoms(a,new) {
	      for (j=0;j<3;j++) a->x[j]=0.0;
	      a->x[ATM_BISO]=0;
	    }
	    for (n=0;n<nsel;n++) {
	      b=rfirst_atom(mols[n]);
	      ForAtoms(a,new) {
		if (!mvld(b)) break;
		for (j=0;j<3;j++) a->x[j]+=b->x[j]/nsel;
		b=rnext_atom(mols[n],b);
	      }
	      if (mvld(a)||mvld(b)) {
		err_set(1,"mismatch for structure %s",mols[n]->name);
		break;
	      }
	    }
	    for (n=0;n<nsel;n++) {
	      b=rfirst_atom(mols[n]);
	      ForAtoms(a,new) {
		if (!mvld(b)) break;
		a->x[ATM_BISO]+=cdistsq(b->x,a->x);
		b=rnext_atom(mols[n],b);
	      }
	      if (mvld(a)||mvld(b)) break;
	    }
	    if (n>0) ForAtoms(a,new) a->x[ATM_BISO]=sqrt(a->x[ATM_BISO]/n); 
	    if (n==nsel) {
	      new->up=(Node *)NmrRoot;
	      link_node(new);
	      send_molecule(new->up,(Mol *)new);
	    }else{
	      del_node(new);
	    }
	  }else{
	    err_set(1,"failed to create node for averaged structure.");
	  }
	}else{
	  err_set(1,"no structure selected.");
	}
	break;
      }else break;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad name token: %s",key->text);
  }

  free_Token(key);
  if (new_name) free(new_name);
  if (mvld(mols)) free(mols);
}
/*-------------------- cmd_nmr_confa --------------------*/

static void cmd_nmr_confa(Mol *NmrRoot)
{
  Node *cnf,*str,*nxt;
  CnfNode *c;
  int j,n,m,t,*M,nm;
  Token *key=(Token *)NULL;
  Node **T;
  Atom *a;

  str=(Node *)first_mol(NmrRoot);
  if (is_mol(str)) {
    cnf=make_node(OBJ_CNF);
    if (mvld(cnf)) {
      if ((M=get_mobile_cnf(&nm))) free(M);   /* to clean out leftovers... */
      cmd_read(cnf,FT_CNF);
      M=get_mobile_cnf(&nm);
      if (M) {
	err_set(1,"error: mobile atom record not allowd in cnf file for NMR");
	free(M);
      }
      m=0; forCnfs(c,cnf) m++;
      c=(CnfNode *)cnf->fst;
      n=(is_cnf(c) ? c->n : 0);
      prtGraphics("%d conformations for %d atoms.\n",m,n);
      key=Get_Atoms_D("template atoms: ",";");
      if (key->status==TOK_OK) {
	t=((Nodechain *)key->ptr)->n;
	T=chain_to_ptr((Nodechain *)key->ptr,TRUE);
	if (mvld(T)) {
	  t=walk_order(str,t,(Atom **)T,tFLG);
	  if (t!=n) err_set(1,"template mismatch: %d atoms in template, %d atoms in file.",t,n);
	  if (t>n) t=n;
	  if (n>t) n=t;
	  if (n>0) {
	    while (next_mol(str)) del_node((Node *)next_mol(str));
	    ForAtoms(a,str) a->seqnr=-1;
	    for (t=0;t<n;t++) ((Atom *)T[t])->seqnr=t;
	    free(T);
	    lock_evtq();
	    forCnfs(c,cnf) {
	      nxt=tree_copy(str);
	      if (is_mol(nxt)) {
		nxt->up=(Node *)NmrRoot; link_node(nxt);
		set_node_name(nxt,c->name);
		ForAtoms(a,nxt)
		  if (a->seqnr>-1)
		    if (a->nxyz>0) for (j=0;j<3;j++) a->x[j]=c->x[3*a->seqnr+j];
		send_molecule(nxt->up,(Mol *)nxt);
	      }else{
		err_set(1,"failed to copy conformation.");
		break;
	      }
	    }
	    str=del_node(str);
	    insane_sanity_check(NmrRoot,FALSE);
	    unlock_evtq();
	  }
	}else err_set(1,"no template atoms.");
      }else if (key->status==TOK_NOK) err_set(1,"bad selection of template atoms: %s",key->text);
      cnf=del_node(cnf);
    }else err_set(1,"failed to allocate CNF node.");
  }else err_set(1,"no structure in current NMR set.");
  key=free_Token(key);
}

/*-------------------- find_all_noes --------------------*/

static int find_all_noes(Mol *root,int active)
{
  NOElist *noes;
  NOEdata *noe;
  int count;

  /* active = 1 : find only active NOEs, = 2 : only inactive NOEs, 3 a;; NOEs */

  count=0;
  forNOElists(noes,root->up) {
    for (noe=noes->first;mvld(noe);noe=noe->nxt) {
      if ((active==1||active==3) && noe->active) {
	if (find_1noe(noe,(Mol *)root)) count++;
	continue;
      }
      if ((active==2||active==3) && !noe->active) {
	if (find_1noe(noe,(Mol *)root)) count++;
	continue;
      }
    }
  }
  return(count);
}

/*-------------------- find_1noe --------------------*/

static int find_1noe(NOEdata *noe,Mol *mol)
{
  /* finds source and target atoms of noe in mol */

  Atom *a;
  Node *res;

  noe->sptr=(Atom *)NULL;
  noe->tptr=noe->sptr;

  if (is_mol(mol) && mvld(noe)) {
    ForAtoms(a,mol) {
      res=a->up;
      if (!noe->sptr) {
	if (!mvld(noe->sseg)||wild_imatch(res->up->name,noe->sseg)) {
	  if (wild_imatch(res->name,noe->sres)) {/* residue may contain wild */
	    ForAtoms(a,res) {
	      if (strcasecmp(a->name,noe->satom)==0) {
		noe->sptr=a;
		break;
	      }
	    }
	    if (!mvld(noe->sptr)||mvld(noe->tptr)) break;
	  }
	}
      }
      if (!noe->tptr) {
	if (!mvld(noe->tseg)||wild_imatch(res->up->name,noe->tseg)) {
	  if (wild_imatch(res->name,noe->tres)) {
	    ForAtoms(a,res) {
	      if (strcasecmp(a->name,noe->tatom)==0) {
		noe->tptr=a;
		break;
	      }
	    }
	    if (!mvld(noe->tptr)||mvld(noe->sptr)) break;
	  }
	}
      }
      a=(Atom *)res->lst;
    }
  }
  return(mvld(noe->tptr)&&mvld(noe->sptr));
}
/*-------------------- send_bnoe --------------------*/

static void send_bnoe(NOEdata *noe,int col,float viol)
{
  /* send NOE to graphics box */

  int size;
  float D,v;

  if (graphix<1) return;

  v=SrvVersion();
  size=2*cPtrSize()+sizeof(float)+sizeof(int);
  if (v>2.001) size+=sizeof(float);
  if (is_atom(noe->sptr)&&is_atom(noe->tptr)) {
    if (cPtr(noe->sptr->cross)&&cPtr(noe->tptr->cross)) {
      send_bheader(graphix,"ANOE",size);
      send_bcptr(graphix,noe->sptr->cross);
      send_bcptr(graphix,noe->tptr->cross);
      if (noe_label==NOE_LAB_VIOL&&v<2.001) D=cdist(noe->sptr->x,noe->tptr->x)-viol;
      else D=noe->dupper;
      send_bfloat(graphix,D);
      send_bint(graphix,col);
      if (v>2.001) send_bfloat(graphix,viol);
    }
  }
}
/*-------------------- delete_noe_display --------------------*/

static void delete_noe_display(Mol *mol)
{
  Dspobj *obj;

  obj=assocobj((Node *)mol);
  if (mvld(obj)) {
    if (cPtr(obj->cross)) {
      send_bheader(graphix,"DNOE",cPtrSize());
      send_bcptr(graphix,obj->cross);
    }
  }
}
/*-------------------- set_noe_display --------------------*/

static void set_noe_display(Mol *mol,int label,int style,int width)
{
  Dspobj *obj;

  obj=assocobj((Node *)mol);
  if (mvld(obj)) {
    if (cPtr(obj->cross)) {
      send_bheader(graphix,"SNOE",cPtrSize()+3*sizeof(int));
      send_bcptr(graphix,obj->cross);
      send_bint(graphix,label);
      send_bint(graphix,style);
      send_bint(graphix,width);
    }
  }
}
/*-------------------- remove_pseudo --------------------*/

static void remove_pseudo(Mol *m,char *res,char *name)
{
  Atom *a;
  Node *Res;

  if (!is_mol(m)) return;
  
  ForAtoms(a,m) {
    Res=a->up;
    if (wild_imatch(Res->name,res)) {/* residue may contain wild */
      ForAtoms(a,Res) if (strcasecmp(a->name,name)==0) break;
      if (mvld(a)) {
	a->flg|=tFLG;
	break;
      }
    }
    a=(Atom *)Res->lst;
  }
}
/*-------------------- add_pseudo --------------------*/

static void add_pseudo(Mol *m,char *seg,char *res,char *def,char *name)
{
  Nodechain c;
  Atom *a,*b=(Atom *)NULL;
  Node *oldDefault,*Res;
  int n,i,dsp;
  float x[3];

  if (!mvld(def)) return;
  if (!mvld(m)) return;

  c.n=0; c.s=(Node *)NULL; c.e=c.s; c.p=c.s;

  ForAtoms(a,m) {
    Res=a->up;
    if (!seg||wild_imatch(Res->up->name,seg)) {
      if (wild_imatch(Res->name,res)) {/* residue may contain wild */
	ForAtoms(a,Res) if (strcasecmp(a->name,name)==0) break;
	if (!mvld(a)) {
	  oldDefault=DefaultNode; DefaultNode=Res;
	  atom_chain(&c,def);
	  DefaultNode=oldDefault;
	  for (i=0;i<3;i++) x[i]=0.0;
	  for (dsp=FALSE,n=0,a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) {
	    if (is_atom(a)) {
	      b=a;
	      dsp|=a->dsp;
	      if (a->nxyz>0) {
		for (i=0;i<3;i++) x[i]+=a->x[i];
		n++;
	      }
	    }
	  }
	  clear_node_chain(&c);
	  if (n>0) {
	    for (i=0;i<3;i++) x[i]/=n;
	    a=(Atom *)make_node(OBJ_ATOM);
	    if (mvld(a)) {
	      set_node_name((Node *)a,name);
	      a->prv=(Node *)b;
	      link_node((Node *)a);
	      a->flg|=tFLG;
	      if (mvld(a->x)) {
		a->nxyz=1;
		for (i=0;i<3;i++) a->x[i]=x[i];
	      }
	      a->dsp=dsp;
	      set_atom_type(a,guess_atom_type("Du"));
	    }
	  }
	}
	break;
      }
    }
    a=(Atom *)Res->lst;
  }
}
/*-------------------- NoeAvDist --------------------*/

static float NoeAvDist(NOEdata *noe,Mol *m)
{
  /* compute averaged distance for given NOE */
  /* return -1 if distance is undefined.     */

  float r,xa[3],xb[3],d,ex;
  int type,na,nb,n,i,j;
  Nodechain c;
  Atom **A,**B;

  if (mvld(noe)&&mvld(m)) {
    type=(mvld(noe->root)?noe->root->avtype:NOE_AVTYPE_CENTER);
    c=NoeAtoms(m,noe->sseg,noe->sres,noe->def1);
    A=(Atom **)chain_to_ptr(&c,FALSE); na=c.n; clear_node_chain(&c);
    c=NoeAtoms(m,noe->tseg,noe->tres,noe->def2);
    B=(Atom **)chain_to_ptr(&c,FALSE); nb=c.n; clear_node_chain(&c);
    if (mvld(B)) for (i=0;i<nb;i++) B[i]->flg|=tFLG;
    if (mvld(A)&&mvld(B)) {
      switch (type) {
      case NOE_AVTYPE_CENTER:
	for (j=0;j<3;j++) xa[j]=xb[j]=0;
	for (i=0;i<na;i++) for (j=0;j<3;j++) xa[j]+=A[i]->x[j]/na;
	for (i=0;i<nb;i++) for (j=0;j<3;j++) xb[j]+=B[i]->x[j]/nb;
	r=cdist(xa,xb);
	break;
      case NOE_AVTYPE_R3:
      case NOE_AVTYPE_R6:
      case NOE_AVTYPE_SUM:
	ex=(type==NOE_AVTYPE_R3 ? 3.0 : 6.0);
	r=0; n=0;
	for (i=0;i<na;i++) {
	  for (j=0;j<nb;j++) {
	    if (!((A[i]->flg&tFLG)&&(B[j]->flg&tFLG))||B[j]>A[i])
	    if ((d=cdist(A[i]->x,B[j]->x))>0) {
	      r+=exp(-ex*log(d));
	      n++;
	    }
	  }
	}
	r=(type==NOE_AVTYPE_SUM ? r : (n>0 ? r/n : 0));
	r=(r>0 ? exp(-log(r)/ex) : 0);
	break;
      default:
	r=-1;
	break;
      }
    }else r=-1;
    if (mvld(B)) for (i=0;i<nb;i++) B[i]->flg&=~tFLG;
    if (mvld(A)) free(A);
    if (mvld(B)) free(B);
  }else r=-1;
  return(r);
}

/*-------------------- NoeAtoms --------------------*/

static Nodechain NoeAtoms(Mol *m,char *seg,char *res,char *def)
{
  /* collect atoms in seg/res/def NOE definition */

  Nodechain c;
  Atom *a;
  Node *Res;

  c.n=0; c.s=(Node *)NULL; c.e=c.s; c.p=c.s;

  if (mvld(def)&&mvld(m)) {
    ForAtoms(a,m) {
      Res=a->up;
      if (!seg||wild_imatch(Res->up->name,seg)) {
	if (wild_imatch(Res->name,res)) AtomsInNode(&c,def,Res);
      }
      a=(Atom *)Res->lst;
    }
  }
  return(c);
}
