#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "fileutil.h"
#include "xtal.h"
#include "gcalc.h"
#include "errhand.h"

extern int graphix;
extern Mol *MolRoot;
extern Dspobj *DspRoot;

typedef struct {               /* ORTEP Atom structure */
  Atom  *atom; 
  Xatom *xatom;
  Dspobj *dsp;
  int   xtal;
  int   tfc;
  int   ftp;
  float radius;
  float xloff;
  float yloff;
  float lsize;
  int   label;
}OAtom;

typedef struct {               /* ORTEP Bond structure */
  Bond *bnd;
  float rad;
  int   from;
  int   to;
  int   type;
}OBond;

#define TRUE 1
#define FALSE 0

#define EPS (78.956835)

#define LOOP 0
#define DONE 1
#define CLOSE 2
#define MAX_ORTEP_ATOMS 2000    /* should not exceed dimension in ORTEP */
#define MXOPEN 10

#define key_is(x) (streq(x,key->key))
#define newline(f) fprintf(f,"\n")

#define RAD(x) (((x)*3.141592653)/180.0)
#define DEG(x) (((x)*180.0)/3.141592653)

static void  ortep_get_atoms(void);
static void  ortep_newpic(void);
static FILE *ortep_open(void);
static void  ortep_close(void);
static void  ortep_write_atom(FILE *o,mat_4 DX,int i);
static void  ortep_reset_defaults(void);
static void  ortep_free_tmp_data(void);
static void  ortep_plot(void);
static void  ortep_set_command(void);

static char *ortep_command=(char *)NULL;
static Mol *ortepMolRoot;
static int nopen=0;
static FILE *open_pipe[MXOPEN];
static int needs_reset=TRUE;
static OAtom *Atoms;
static OAtom defaultAtom;
static int NAtoms=0;
static OBond *Bonds;
static OBond defaultBond;
static int NBonds=0;
static FILE *ortep;
static Xcn *Xinfo=(Xcn*)NULL;
static float defaultProb,prob,defaultRetra,retra;
static int defaultTfc,defaultFtp,defaultBondtype;
static float defaultAtomradius,defaultLsize,defaultBondrad,
       defaultXloff,defaultYloff;

/* football definitions ... */

static int nplane[4]={1,4,4,4};
static int ndot[4]  ={0,0,0,0};
static int nline[4] ={0,0,1,6};
static int ndash[4] ={0,0,0,0};

static int uo[6] = {0,1,2,5,4,3};              /* Permutation of XCN to Ortep uij order */

static char *FMTxyz;
static char *XTLxyz = "%9.4f%9.4f%9.4f\n";
static char *ANGxyz = "%9.3f%9.3f%9.3f\n";

static PrjTS prj;

#define LPtab 11

static int PtabP[LPtab] = {
  10,20,30,40,50,60,70,80,85,90,95};

static float PtabC[LPtab] = {
  0.7644,1.0026,1.1932,1.3672,1.5382,1.7164,1.9144,2.1544,
  2.3059,2.5003,2.7955};

static char *mainkeys = 
   "display newpic plot pipe default footballs radius label bond "
   "probability retrace uij-order end";
static char *par_keys="footballs radius label bond";
static char *bond_keys="radius type";
static char *retrace_keys="0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 2.0";
static char *label_keys="on off position offset size";

static char *atom_rad="0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5";
static char *bond_rad="0.01 0.02 0.03 0.04 0.05 0.06 0.07 0.08 0.09 0.1";
static char *bond_typ="0 1 2 3 4 5";
static char *lab_sizes="0.05 0.1 0.15 0.2 0.25 0.3 0.35 0.4 0.45 0.5";

static char *footballTypes =  
  "anisotropic isotropic constant simple solid cutout fancy";

#define NONE 0
#define RADIUS 1
#define FOOTBALL_TYPE 2
#define TYPE 3
#define ONOFF 5
#define LSIZE 6
#define LOFFS 7
#define BRADIUS 8
#define BTYPE 9
#define FOOTBALL_STYLE 10

/*========================================================================*/
void cmd_ortep(void)
{
  Token *key;
  int a,i,do_default,apar,bpar,tfc=0,ftp=0,onoff=0,status;
  char buff[1024];
  char *txt;
  float rad=0,type=0,lsz=0,ox=0,oy=0,x[3],labxyz[3];
  Nodechain *c;
  Atom *atm;
  Bond *bnd;

  if (needs_reset) ortep_reset_defaults();
  ortep_get_atoms();
  status=LOOP;

  while(status==LOOP) {
    err_dump();
    start_parse();
    apar=bpar=NONE;
    do_default=FALSE;
    key=Get_Kwd("Ortep> ",";",mainkeys);
    if (key->status==TOK_OK) {
      if (key_is("default")) {
	free_Token(key);
	key=Get_Kwd("what ? ",";",par_keys);
	do_default=TRUE;
      }
    }
    if (key->status==TOK_OK) {
      if (key_is("display")) {
	cmd_display();
	ortep_get_atoms();
      }else if (key_is("newpic")) {
	ortep_newpic();
	while ((txt=err_text_bot())) {
	  prtGraphics("%s\n",txt);
	  err_dequeue(1);
	}
      }else if (key_is("plot")) {
	ortep_plot();
      }else if (key_is("pipe")) {
	ortep_set_command();
      }else if (key_is("end")) {
	status=CLOSE;
      }else if (key_is("footballs")) {
	free_Token(key);
	key=Get_Kwd("football type: ",";",footballTypes);
	if (key->status==TOK_OK) {
	  if (key_is("anisotropic")||key_is("isotropic")||key_is("constant")){
	    tfc=key->n;
	    apar=FOOTBALL_TYPE;
	  }else{
	    ftp=key->n-3;
	    apar=FOOTBALL_STYLE;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"unknown football type: %s",key->text);
	  end_parse();
	}
      }else if (key_is("radius")) {
	free_Token(key);
	key=Get_Float_M("atom radius: ",";",atom_rad);
	if (key->status==TOK_OK) {
	  rad=key->f;
	  if (rad>0.001) apar=RADIUS;
	  else {
	    err_set(1,"radius is too small...");
	    end_parse();
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad float: %s",key->text);
	  end_parse();
	}
      }else if (key_is("label")) {
	free_Token(key);
	key=Get_Kwd("label option: ",";",label_keys);
	if (key->status==TOK_OK) {
	  if (key_is("on") || key_is("off")) {
	    apar=ONOFF;
	    onoff=key_is("on");
	  }else if (key_is("size")) {
	    free_Token(key);
	    key=Get_Float_M("label size (A): ",";",lab_sizes);
	    if (key->status==TOK_OK) {
	      if (key->f>=0) {
		apar=LSIZE;
		lsz=key->f;
	      }else{
		err_set(1,"bad value of label size: %s",key->text);
		end_parse();
	      }
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad label size token: %s",key->text);
	      end_parse();
	    }
	  }else if (key_is("offset")) {
	    free_Token(key);
	    sprintf(buff,"0.0");
	    for (ox=0.25;ox<=4;ox+=0.25) {
	      sprintf(buff+strlen(buff)," +%4.2f -%4.2f",ox,ox);
	    }
	    key=Get_Float_M("label x offset (A): ",";",buff);
	    if (key->status==TOK_OK) {
	      ox=key->f;
	      free_Token(key);
	      key=Get_Float_M("label y offset (A): ",";",buff);
	      if (key->status==TOK_OK) {
		apar=LOFFS;
		oy=key->f;
	      }
	    }
	    if (key->status==TOK_NOK) {
	      err_set(1,"bad label offset token : %s",key->text);
	      end_parse();
	    }
	  }else if (key_is("position")) {
	    while (key->status==TOK_OK) {
	      key=free_Token(key);
	      key=Get_Atom("atom: ",";");
	      if (key->status==TOK_OK) {
		atm=key->ptr;
		for (a=0;a<NAtoms;a++) if (Atoms[a].atom==atm) break;
		if (a>=NAtoms) {
		  key->status=TOK_NOK;
		  err_set(1,"atom not in ORTEP atom list");
		  end_parse();
		  break;
		}
		key=free_Token(key);
		key=Get_XYZ("pick label position...\n",";");
		if (key->status==TOK_OK) {
		  get_graph_xyz(1,x,&atm);
		  wline2xyz((float *)key->ptr,x[2],labxyz);
		  Atoms[a].xloff=labxyz[0]-x[0];
		  Atoms[a].yloff=labxyz[1]-x[1];
		  free(key->ptr);
		}else{
		  err_set(1,"bad locate response: %s",key->text);
		  end_parse();
		}
	      }else if (key->status==TOK_NOK) {
		err_set(1,"bad atom token: %s",key->text);
		end_parse();
	      }
	      send_header(graphix,"ARMK",0);
	    }
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"unknown label option: %s",key->text);
	  end_parse();
	}
      }else if (key_is("bond")) {
	free_Token(key);
	key=Get_Kwd("bond parameter: ",";",bond_keys); 
	if (key->status==TOK_OK) {
	  if (key_is("radius")) {
	    free_Token(key);
	    key=Get_Float_M("radius (A): ",";",bond_rad);
	    if (key->status==TOK_OK) {
	      rad=key->f;
	      if (rad>0.0001) {
		bpar=BRADIUS;
	      }else{
		err_set(1,"bond radius too small...");
	      }
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad format of bond radius: %s",key->text);
	      end_parse();
	    }
	  }else if (key_is("type")) {
	    free_Token(key);
	    key=Get_Kwd("bond type: ",";",bond_typ);
	    if (key->status==TOK_OK) {
	      bpar=BTYPE;
	      type=key->n;
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"unknown bond type: %s",key->text);
	      end_parse();
	    }
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"unknown bond parameter: %s",key->text);
	  end_parse();
	}
      }else if (key_is("probability")) {
	free_Token(key);
	buff[0]='\0';
	for (i=0;i<LPtab;i++) sprintf(buff+strlen(buff),"%d ",PtabP[i]);
	key=Get_Float_M("probability (%): ",";",buff);
	if (key->status==TOK_OK) {
	  if (key->f < PtabP[0] || key->f >PtabP[LPtab-1]) {
	    err_set(1,"probability value out of range");
	    end_parse();
	  }else{
	    prob=key->f;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad probability token: %s",key->text);
	  end_parse();
	}
      }else if (key_is("retrace")) {
	key=free_Token(key);
	key=Get_Float_M("retrace distance (mm): ",";",retrace_keys);
	if (key->status==TOK_OK) {
	  if (key->f<0 || key->f >5) {
	    err_set(1,"retrace distance out of range");
	    end_parse();
	  }else{
	    retra=key->f;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad retrace token: %s",key->text);
	  end_parse();
	}
      }else if (key_is("uij-order")) {
	key=free_Token(key);
	key=Get_Kwd("uij-order> ",";","normal swapped");
	if (key->status==TOK_OK) {
	  if (key_is("normal")) {     /* note XCN: 11, 22, 33, 23, 13, 12 */
	    prtGraphics("uij-order: 11, 22, 33, 12, 13, 23\n");
	    uo[0]=0;uo[1]=1;uo[2]=2;uo[3]=5;uo[4]=4;uo[5]=3;
	  }else{
	    prtGraphics("uij-order: 11, 22, 33, 23, 13, 12\n");
	    uo[0]=0;uo[1]=1;uo[2]=2;uo[3]=3;uo[4]=4;uo[5]=5;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad uij-order: %s",key->text);
	  end_parse();
	}
      }
      
    }else if (key->status==TOK_NOK) {
      err_set(1,"unknown Ortep option: %s",key->text);
      end_parse();
    }else if (!do_default) status=DONE;
    
    if (bpar!=NONE) {
      if (do_default) {
	switch (bpar) {
	case BRADIUS:
	  defaultBond.rad=rad;
	  break;
	case BTYPE:
	  defaultBond.type=type;
	  break;
	default: break;
	}
      }else{
	free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  for (i=0;i<NBonds;i++) Bonds[i].bnd->flg|=(1<<1);
	  for (bnd=(Bond *)key->ptr;mvld(bnd);bnd=(Bond *)key->ptr) {
	    key->ptr=(Bond *)bnd->chain;
	    bnd->chain=(Bond *)NULL;
	    bnd->flg&=~(1<<1);
	  }
	  for (i=0;i<NBonds;i++) {
	    if (!(Bonds[i].bnd->flg&(1<<1))) {
	      switch(bpar) {
	      case BRADIUS:
		Bonds[i].rad=rad;
		break;
	      case BTYPE:
		Bonds[i].type=type;
		break;
	      default: break;
	      }
	    }else{
	      Bonds[i].bnd->flg&=~(1<<1);
	    }
	  }
	}
	send_header(graphix,"ARMK",0);
      }
    }
    
    if (apar!=NONE) {
      if (do_default!=NONE) {
	switch(apar) {
	case FOOTBALL_TYPE:
	  defaultAtom.ftp=ftp;
	  break;
	case FOOTBALL_STYLE:
	  defaultAtom.tfc=tfc;
	  break;
	case RADIUS:
	  defaultAtom.radius=rad;
	  break;
	case LSIZE:
	  defaultAtom.lsize=lsz;
	  break;
	case LOFFS:
	  defaultAtom.xloff=ox;
	  defaultAtom.yloff=oy;
	  break;
	case ONOFF:
	  defaultAtom.label=onoff;
	  break;
	default:
	  break;
	}
      }else{
	free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (i=0;i<NAtoms;i++) {
	    if (in_chain(Atoms[i].atom,c)) {
	      switch(apar) {
	      case FOOTBALL_TYPE:
		Atoms[i].tfc=tfc;
		break;
	      case FOOTBALL_STYLE:
		Atoms[i].ftp=ftp;
		break;
	      case RADIUS:
		Atoms[i].radius=rad;
		break;
	      case LSIZE:
		Atoms[i].lsize=lsz;
		break;
	      case LOFFS:
		Atoms[i].xloff=ox;
		Atoms[i].yloff=oy;
		break;
	      case ONOFF:
		Atoms[i].label=onoff;
		break;
	      default:
		break;
	      }
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}
	send_header(graphix,"ARMK",0);
      }
    }
    
    if (end_parse()!=TOK_OK) {
      err_set(1,"extra characters at end of command ignored.");
    }
    free_Token(key);
  }
  
  ortep_free_tmp_data();
  if (status==CLOSE) ortep_close();
} 
/*-------------------- ortep_get_atoms --------------------*/

static void ortep_get_atoms(void)
{
  /* put displayed atoms into ORTEP atom list */

  Atom *atm;
  int i,j,k,a,b;
  Nodechain *c;
  Xcn *xcn;
  OAtom *oldAtoms;
  OBond *oldBonds;
  int multi_xcn,bad_xcn,oldNAtoms,oldNBonds;

  c=make_node_chain();
  ForAtoms(atm,ortepMolRoot) {
    if (atm->dsp) add_to_node_chain((Node *)atm,c);
  }

  /* allocate space for ortep atoms */

  if (c->n>MAX_ORTEP_ATOMS-1) {
    err_set(1,"too many atoms, max is %d",MAX_ORTEP_ATOMS-1);
    clear_node_chain(c);
    free_node_chain(c);
    return;
  }

  oldNAtoms=NAtoms;oldAtoms=Atoms;
  oldNBonds=NBonds;oldBonds=Bonds;
  NAtoms=c->n;
  

  if (NAtoms > 0) {
    Atoms=(OAtom *)malloc(NAtoms*(sizeof(OAtom)));
    if (!mvld(Atoms)) {
      err_set(1,"memory allocation failed");
      clear_node_chain(c);
      free_node_chain(c);
      NAtoms=0;
      return;
    }
  }else{
    Atoms=(OAtom *)NULL;
  }

  for (i=0,atm=(Atom *)c->s;atm;i++,atm=(Atom *)atm->chain) {
    Atoms[i]=defaultAtom;
    Atoms[i].atom=atm;
  }
  clear_node_chain(c);
  for (i=0;i<NAtoms;i++) Atoms[i].atom->chain=(Node *)NULL+(i+1);

  for (i=0;i<oldNAtoms;i++) {            /* copy old Atoms info... */
    if (oldAtoms[i].atom->dsp) {
      k=(oldAtoms[i].atom->chain-(Node *)NULL)-1;
      Atoms[k]=oldAtoms[i];
    }
  }

  /* find xtal info and check consistency */

  Xinfo=(Xcn*)NULL;
  multi_xcn=FALSE;

  bad_xcn=FALSE;
  for (i=0;i<NAtoms;i++) {
    if (Atoms[i].atom->xtal) {
      Atoms[i].xatom=first_xatom(Atoms[i].atom);
      xcn=(Xcn*)find_cousin((Node *)Atoms[i].atom,OBJ_XCN);
      if (!mvld(xcn)) {
	bad_xcn=TRUE;
      }else{
	if (Xinfo && Xinfo != xcn) {
	  multi_xcn=TRUE;
	}else{
	  Xinfo=xcn;
	}
      }
    }else{
      Atoms[i].xatom=(Xatom*)NULL;
    }
  }

  /* find associated display objects */

  for (i=0;i<NAtoms;i++) {
    Atoms[i].dsp=glblobj(assocobj((Node*)Atoms[i].atom));
  }

  if (Xinfo) for (i=0;i<NAtoms;i++) Atoms[i].xtal=(Atoms[i].xatom ? 1 : 0);
  else for (i=0;i<NAtoms;i++) Atoms[i].xtal=0; 

  /* if xtal atom without xcn then we have a mess ! */

  if (bad_xcn) {
    err_set(1,"messed up object tree !!!");
    Xinfo=(Xcn*)NULL;
  }

  /* if more than one xcn in tree, we can't handle it as xtal ortep */

  if (multi_xcn) {
    err_set(1,"multi Xtal Ortep not supported.");
    err_set(1,"non-Xtal Ortep will be used...");
    Xinfo=(Xcn*)NULL;
  }

  /* generate bond list with default parameters (abusing 'chain') */

  for (i=0,NBonds=0;i<NAtoms;i++) {    /* just counting... */
    for (j=0;j<Atoms[i].atom->nbnd;j++) {
      a=(Atoms[i].atom->bnd[j]->f->chain-(Node *)NULL)-1;
      b=(Atoms[i].atom->bnd[j]->t->chain-(Node *)NULL)-1;
      if (a==i && a <NAtoms && b>=0 && b<NAtoms) NBonds++;
    }
  }

  if (NBonds>0) Bonds=(OBond *)malloc(NBonds*sizeof(OBond));
  else Bonds=(OBond *)NULL;

  if (Bonds) {
    for (i=0,k=0;i<NAtoms;i++) {       /* store bond info in temp. structure */
      for (j=0;j<Atoms[i].atom->nbnd;j++) {
	a=(Atoms[i].atom->bnd[j]->f->chain-(Node *)NULL)-1;
	b=(Atoms[i].atom->bnd[j]->t->chain-(Node *)NULL)-1;
	if (a==i && a <NAtoms && b>=0 && b<NAtoms) {
	  Bonds[k]=defaultBond;
	  Bonds[k].bnd=Atoms[i].atom->bnd[j];
	  Bonds[k].bnd->chain=(Bond *)NULL+(k+1);       /* bond chain abuse */
	  Bonds[k].from=a;
	  Bonds[k].to=b;
	  k++;
	}
      }
    }
    for (i=0;i<oldNBonds;i++) {                /* copy old bond info */
      a=oldBonds[i].from; b=oldBonds[i].to;
      if (oldAtoms[a].atom->dsp && oldAtoms[b].atom->dsp) {
	if (mvld(oldBonds[i].bnd)) {
	  k=(oldBonds[i].bnd->chain-(Bond *)NULL);
	  if (k>0 && k <=NBonds) {
	    Bonds[k-1].rad=oldBonds[i].rad;
	    Bonds[k-1].type=oldBonds[i].type;
	  }
	}
      }
    }
  }else if (NBonds>0) {
    err_set(1,"memory allocation failed for %d bonds",
		NBonds);
    NBonds=0;
  }
  
  for(i=0;i<NAtoms;i++) Atoms[i].atom->chain=(Node *)NULL; 
  for (i=0;i<NBonds;i++) Bonds[i].bnd->chain=(Bond *)NULL;
  if (oldNBonds>0) free(oldBonds);
  if (oldNAtoms>0) free(oldAtoms);
  free_node_chain(c);
}
/*-------------------- ortep_newpic --------------------*/

static void ortep_newpic(void)
{
  /* redraw molecule in ortep box... */

  FILE *o;
  int sgpdevice;
  int i,j,m1,m2,m3,ifrom,ito,nstereo,s;
  float t,ox,oy,rx,ry,rz,vwpxsh,vwpysh,srot,
    vwpxo,vwpxs,vwpyo,vwpys,cmarg,sizex,sizey,viewd,lsz,
    probC,inchpA,StereoSep=0;
  Symop sym;
  char buffer[256],rec_ifrom[90],rec_i[90],c;
  Dspobj *dspr;
  mat_4 DO,DX;
  vec_3 orgn,znear;

  if (NAtoms==0) {
    prtGraphics("0 Atoms, nothing to display in ORTEP\n");
    return;
  }

  o=ortep_open();
  if (ortep) {
    
    prj=getThreeDPrj();                   /* 3D projection */
    get_MVT(DspRoot);                     /* object transformations */
    
    /* set up transformation matrices ... */

    if (mvld(Xinfo)) {
      dspr=glblobj(assocobj((Node *)Xinfo));
      DO = (mvld(dspr) ? dspr->mtrx : ident_4());
      DX = m4xm4(Xinfo->atf,mmxinv(DO));
      FMTxyz = XTLxyz;
    }else{
      dspr = DspRoot;
      DO = dspr->mtrx;
      DX = mmxinv(DO);
      FMTxyz = ANGxyz;
    }
    znear.c[0]=0.0;znear.c[1]=0.0;znear.c[2]=prj.znear;
    orgn = m4xv3(DX,znear);

    sgpdevice=21;  /* sgp X graphics device */
    
    /* figure out window and viewport parameters */

    sizey=prj.ysize;
    sizex=prj.aspect*prj.ysize;
    cmarg=2.0;                  /* margin for proper xy-clipping */
    vwpysh=100;
    vwpxsh=130.0;
    srot=0.5*atan(prj.eyesep/prj.viewdist)*180.0/3.141502653;
    nstereo=1;
    switch (prj.stereomode) {
    case STEREO_NO:
      srot=0.0;
      break;
    case STEREO_LEFT:
      StereoSep=-prj.sep;
      srot=1.0*srot;
      break;
    case STEREO_RIGHT:
      StereoSep=prj.sep;
      srot=-1.0*srot;
      break;
    case STEREO_CROSSED:
      StereoSep=prj.sep;
      srot=-1.0*srot;
      nstereo=2;
      break;
    case STEREO_RELAXED:
      StereoSep=-prj.sep;
      srot=1.0*srot;
      nstereo=2;
      break;
    default:
      srot=0.0;
      break;
    }
    if (prj.stereomode==STEREO_RELAXED) vwpxsh=vwpxsh*0.5;
    if (prj.stereomode==STEREO_CROSSED) vwpxsh=vwpxsh*0.5;
    vwpxs=vwpxsh; vwpys=vwpysh;
    if (vwpxs<prj.aspect*vwpys) vwpys=vwpxs/prj.aspect;
    if (vwpys<vwpxs/prj.aspect) vwpxs=vwpys*prj.aspect;
    vwpxo=0.5*(vwpxsh-vwpxs);
    vwpyo=0.5*(vwpysh-vwpys);
    
    switch (prj.viewmode) {
    case ORTHO:
      viewd=0.0;
      break;
    case PERSPECTIVE:
      viewd=prj.viewdist*prj.scrysize;
      break;
    default:
      viewd=0.0;
    }

    /* so far, everything was in A, but ORTEP wants it in inches, so... */

    inchpA = 19.0*(vwpys/100.0)/(sizey*2.54); /* assumes device y = 19 cm */
    viewd*=inchpA;
    sizex*=inchpA; sizey*=inchpA;
    cmarg*=inchpA;
    StereoSep*=inchpA;

    eqrxyz(&rx,&ry,&rz,DO);                   /* display rotation of root */

    if (prob<PtabP[0]) prob=PtabP[0];
    if (prob>PtabP[LPtab-1]) prob=PtabP[LPtab-1];
    probC=PtabC[LPtab-1];
    for (i=1;i<LPtab;i++) {
      if (prob<PtabP[i]) {
	probC=PtabC[i-1]
	  +(PtabC[i]-PtabC[i-1])*(prob-PtabP[i-1])/(PtabP[i]-PtabP[i-1]);
	break;
      }
    }

    /* Title (comment) line */
    
    fprintf(o,"ORTEP-II input generated by WITNOTP\n");
    
    /* unit cell */
    
    if (mvld(Xinfo)) {            /* Cryst. Ortep */
      
      /* unit cell geometry */
      
      for (i=0;i<3;i++) fprintf(o,"%9.4f",Xinfo->cl[i]);
      for (i=0;i<3;i++) fprintf(o,"%9.4f",Xinfo->ca[i]);
      newline(o);
      
      /* space group symmetry */
      
      for (i=0;i<Xinfo->nsymx;i++) {
	sym=str2sym(sym2str(buffer,symF(Xinfo,Xinfo->sym[i])));
	c=(i+1<Xinfo->nsymx ? ' ' : '1') ;
	for (j=0;j<3;j++,c=' ') {
	  t=sym.r[j].c[3]; m1=sym.r[j].c[0]; 
	  m2=sym.r[j].c[1]; m3=sym.r[j].c[2];
	  fprintf(o,"%c%14.8f%3d%3d%3d",c,t,m1,m2,m3);
	}
	newline(o);
      }
      
    }else{                        /* non-cryst. Ortep */
      
      /* dummy unit cell for non-cryst. Ortep */
      
      for (i=0;i<3;i++) fprintf(o,"%9.4f",1.0);
      for (i=0;i<3;i++) fprintf(o,"%9.4f",90.0);
      newline(o);
      
      /* only symmop is Id */
      
      fprintf(o,"1%14.8f%3d%3d%3d",0.0,1,0,0);
      fprintf(o," %14.8f%3d%3d%3d",0.0,0,1,0);
      fprintf(o," %14.8f%3d%3d%3d",0.0,0,0,1);
      newline(o);
      
    }
    
    /* output atoms */
    
    for (i=0;i<NAtoms;i++) ortep_write_atom(o,DX,i);
    
    /* add dummy atom at origin */
    
    fprintf(o,"%6s"," ORGN ");
    fprintf(o,"                     ");   /* 21x */
    fprintf(o,"%9.4f%9.4f%9.4f\n",orgn.c[0],orgn.c[1],orgn.c[2]);
    fprintf(o,"1\n");   /* dummy at end of list */
    
    /* atom list instruction */
    
    ifrom=1*100000+55501;
    ito = -NAtoms*100000-55501;
    fprintf(o,"%10d%10d%10d\n",401,ifrom,ito);
    
    /* explicit bond list instruction */
    
    fprintf(o,"%10d\n",1202);                /* reset bond list */
    for (i=0;i<NBonds;i++) {                 /* add bonds to explicit list */
      fprintf(o,"%10d%10d%10d%10.3f%10d\n",
	      1201,
	      (Bonds[i].from+1)*100000+55501,
	      (Bonds[i].to+1)*100000+55501,
	      Bonds[i].rad,
	      Bonds[i].type);
    }

    /* graphics start, erase screen, window/viewport etc... */

    fprintf(o,"%10d%10d\n%10d\n",201,sgpdevice,202);

    fprintf(o,"%10d%10.2f%10.2f%10.2f%10.2f\n",301,
	    sizex+2*cmarg,sizey+2*cmarg,viewd,0.0);

    /* viewing/orienting instructions come here... */

    fprintf(o,"%10d%10d%10d%10d%10d%10d\n",  /* set reference coord system */
	    501,          
	    (NAtoms+1)*100000+55501,     
	    55501,55601,  
	    55501,54501); 

    fprintf(o,"%10d%10d%10.2f\n",              /* permute axes */
	    502,
	    -2,0.0);

    fprintf(o,"%10d%10d%10.2f%10d%10.2f%10d%10.2f\n",
	    502,
	    1,DEG(-rx),
	    2,DEG(-ry),
	    3,DEG(-rz));

    fprintf(o,"%10d%10.3f%10.3f%10.3f%10.5f\n", /* position/scale */
	    601,
	    0.5*sizex+cmarg,
	    0.5*sizey+cmarg,
	    inchpA,probC);

    for (s=0;s<nstereo;s++) {

      fprintf(o,"%10d%10.1f%10.2f%10.2f%10.1f\n",209,          /* SGP window */
	      cmarg+(1-2*s)*StereoSep,sizex+cmarg+(1-2*s)*StereoSep,cmarg,sizey+cmarg);

      fprintf(o,"%10d%10.2f%10.2f%10.2f%10.2f\n",210,          /* SGP viewport */
	      vwpxo,vwpxo+vwpxs,vwpyo,vwpyo+vwpys);

      fprintf(o,"%10d%10d%10.3f\n",503,2,srot);
      
      fprintf(o,"%10d%10.5f\n",303,retra/25.4); /* retrace */
      fprintf(o,"%10d\n",511);            /* hidden line removal */
      
      /* draw bonds */ 
      
      fprintf(o,"%10d\n",815);            /* bonds from explicit list */
      
      /* footballs */
      
      for (i=0,ifrom=0;i<NAtoms;i++) {
	lsz=(Atoms[i].label ? Atoms[i].lsize : 0.0)*inchpA; /* A --> inches */
	ox=Atoms[i].xloff*inchpA; oy=Atoms[i].yloff*inchpA;
	sprintf(rec_i,"%3d%7d%10d%10d%10d%10d%10.3f%10.3f%10.3f\n",
		1,715,
		nplane[Atoms[i].ftp],
		ndot[Atoms[i].ftp],
		nline[Atoms[i].ftp],
		ndash[Atoms[i].ftp],
		lsz,ox,oy);
	if (i>0&&strcmp(rec_i,rec_ifrom)!=0) {
	  fprintf(o,"%s",rec_ifrom);
	  fprintf(o,"%10d%10d%10d%10d%10d\n",
		  0,0,0,
		  (ifrom+1),i);
	  ifrom=i;
	}
	strcpy(rec_ifrom,rec_i);
      }
      if (ifrom<NAtoms) {
	fprintf(o,"%s",rec_ifrom);
	fprintf(o,"%10d%10d%10d%10d%10d\n",
		0,0,0,
		(ifrom+1),NAtoms);
      }	
      
      /* get ready for stereo redraw ... */

      srot=-srot;
      vwpxo+=vwpxsh;

    }
    /* end of graphics ... */
    
    fprintf(o,"%10d\n",207);
    fflush(o);
  }
}
/*-------------------- ortep_write_atom --------------------*/

static void ortep_write_atom(FILE *o,mat_4 DX,int n)
{
/* write ortep atom record to outputfile o */

  Atom *a;
  Xatom *xa;
  vec_3 xyz;
  mat_4 Di;
  float biso;
  int tfc;

  if (n<0 || n>=NAtoms) return;

  a=Atoms[n].atom;
  xa=Atoms[n].xatom;

  if (!mvld(a)) return;

  /* coordinates record ... */
  
  Di = m4xm4(DX,Atoms[n].dsp->mtrx);
  xyz = m4xv3(Di,vec3(a->x));
  if (mvld(a->name))fprintf(o,"%6s",a->name);
  else fprintf(o,"      ");
  fprintf(o,"                     ");   /* 21x */
  fprintf(o,FMTxyz,xyz.c[0],xyz.c[1],xyz.c[2]);

  /* temperature factor record ... */

  if (Atoms[n].xtal) {
    if (xa->iso) {                              /* isotropic atom */
      tfc=(Atoms[n].tfc==0 ? 1 : Atoms[n].tfc);
      if (tfc==2) {
	fprintf(o," %8.3f%9d%9d%9d%9d%9d%9d\n",
		Atoms[n].radius,0,0,0,0,0,7);
      }else{
	fprintf(o," %8.3f%9d%9d%9d%9d%9d%9d\n",
	      xa->bfac[0]*EPS,0,0,0,0,0,6);
      }
    }else{                                      /* anisotropic atom */
      tfc=Atoms[n].tfc;
      if (tfc==2) {
	fprintf(o," %8.3f%9d%9d%9d%9d%9d%9d\n",
		Atoms[n].radius,0,0,0,0,0,7);
      }else if (tfc==1) {
	biso=xa->bfac[0]*xa->bfac[1]*xa->bfac[2]
	  +2*xa->bfac[3]*xa->bfac[4]*xa->bfac[5]
	  -xa->bfac[3]*xa->bfac[3]*xa->bfac[0]
	  -xa->bfac[4]*xa->bfac[4]*xa->bfac[1]
	  -xa->bfac[5]*xa->bfac[5]*xa->bfac[2];
	biso=(biso > 0 ? pow(biso,1.0/3.0) : 0.0001)*EPS;
	fprintf(o," %8.3f%9d%9d%9d%9d%9d%9d\n",
	      biso,0,0,0,0,0,6);
      }else{
	fprintf(o," %8.4f%9.4f%9.4f%9.4f%9.4f%9.4f%9d\n",
		xa->bfac[uo[0]]*EPS,xa->bfac[uo[1]]*EPS,xa->bfac[uo[2]]*EPS,
		xa->bfac[uo[3]]*EPS,xa->bfac[uo[4]]*EPS,xa->bfac[uo[5]]*EPS,   
		4);
      }
    }
  }else{                                       /* non-crystallographic atom */
    fprintf(o," %8.3f%9d%9d%9d%9d%9d%9d\n",
	    Atoms[n].radius,0,0,0,0,0,7);
  }
}

/*-------------------- file utilities --------------------*/

static FILE *ortep_open(void)
{
  char *cmd;

  if (!ortep) {
    if (mvld(ortep_command)) {
      cmd=ortep_command;
    }
    else cmd=getenv("ORTEPserver");
    if (cmd) {
      ortep=PipeOpen(cmd,"w");
      if (!ortep) {
	err_set(1,"could not connect to ORTEP command");
      }
    }else{
      err_set(1,"ORTEPserver is not defined.");
    }
  }else{
    fprintf(ortep,"%10d\n",-2);  /* restart */
  }
  return(ortep);
}

static void ortep_close(void)
{
  if (ortep) {
    fprintf(ortep,"%10d\n",-1);
    fflush(ortep);
    PipeClose(ortep);
    ortep=(FILE *)NULL;
  }

  while (nopen>0) {
    nopen--;
    fprintf(open_pipe[nopen],"%10d\n",-1);
    fflush(open_pipe[nopen]);
    open_pipe[nopen]=(FILE *)NULL;
  }
}

static void ortep_set_command(void)
{
  Token *key;

  default_token(getenv("ORTEPserver"));
  key=Get_String("setenv ORTEPserver = ",";");
  if (key->status==TOK_OK) {
    if (mvld(ortep_command)) free(ortep_command);
    ortep_command=Strdup(key->text);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad_token: %s",key->text);
    end_parse();
  }

  if (ortep) {
    free_Token(key);
    default_token("yes");
    key=Get_Kwd("close old ORTEP ? ",";","yes no");
    if (key->status==TOK_OK) {
      if (key_is("yes")) ortep_close();
      else {
	if (nopen>=MXOPEN) {
	  err_set(1,"too many open ortep pipes...");
	  ortep_close();
	}else{
	  open_pipe[nopen++]=ortep;
	  ortep=(FILE *)NULL;
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
    }
  }

  free_Token(key);
}

static void ortep_plot(void)
{
  Token *key;
  char *dev=(char *)NULL,*cmd;

  if (!ortep) {
    prtGraphics("Nothing to plot...\n");
    end_parse();
    return;
  }

  cmd=(char *)NULL;
  key=Get_Kwd("device: ",";","HP7500 EPScript EPScolor");
  if (key->status==TOK_OK) {
    if (key_is("HP7500")) {
      cmd=getenv("HP7500plot");
      dev="HP7500";
    }else if (key_is("EPScript")) {
      cmd=getenv("EPSplot");
      dev="EPScript";
    }else if (key_is("EPScolor")) {
      cmd=getenv("EPSplot");
      dev="EPScript";
    }
    if (cmd) {
      fprintf(ortep,"3 %8d\nPLOT %s %s\n",211,dev,cmd);
      fflush(ortep);
    }else{
      err_set(1,"no plot command defined for device %s",dev);
    }
  }

  free_Token(key);
}
/*-------------------- ortep_defaults --------------------*/

static void ortep_reset_defaults(void)
{
  /* set static ortep parameters to defaults */
  
  defaultRetra=0.0;
  defaultProb=50.0;
  defaultTfc=0;
  defaultFtp=1;
  defaultAtomradius=0.075;
  defaultXloff=0.0;
  defaultYloff=0.0;
  defaultLsize=0.25;
  defaultBondrad=0.04;
  defaultBondtype=1;

  ortepMolRoot=MolRoot;
  prob=defaultProb;
  retra=defaultRetra;

  defaultAtom.atom=(Atom *)NULL;
  defaultAtom.xatom=(Xatom *)NULL;
  defaultAtom.dsp=(Dspobj *)NULL;
  defaultAtom.xtal=FALSE;
  defaultAtom.tfc=defaultTfc;
  defaultAtom.ftp=defaultFtp;
  defaultAtom.radius=defaultAtomradius;
  defaultAtom.xloff=defaultXloff;
  defaultAtom.yloff=defaultYloff;
  defaultAtom.lsize=defaultLsize;
  defaultAtom.label=FALSE;

  defaultBond.bnd=(Bond *)NULL;
  defaultBond.from=0;
  defaultBond.to=0;
  defaultBond.rad=defaultBondrad;
  defaultBond.type=defaultBondtype;

  needs_reset=FALSE;
}
/*-------------------- ortep_free_tmp_data --------------------*/

static void ortep_free_tmp_data(void)
{
  /* free memory for temporary data structures */

  NAtoms=0;
  if (mvld(Atoms)) free(Atoms);
  Atoms=(OAtom *)NULL;

  NBonds=0;
  if (mvld(Bonds)) free(Bonds);
  Bonds=(OBond *)NULL;
}
