#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *DefaultNode;

#define MAXBUF 2048

#define TRUE 1
#define FALSE 0

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define key_is(a) (streq(a,key->key))

static void cmd_run_piff(Node *root,int fgrd);
static Nodechain wr_inter(Node *root,char *job);
static int asg_piff_types(Node *root);
static int piff_type(Atom *a);

static char *main_keys = "molecule ' ' "
     "charges cycles deltaE gradient iterations SCF ' ' run h!submit";

static int IgnoreCharge = TRUE;
static int Cycles = 10;
static float deltaE = 0.01;
static float grad = 0.00035;
static int  Iter = -1;
static int SCFtype = 1;

void cmd_piff(void)
{
  Token *key;
  Node *root;
  char *mol_menu;
  int loop;
  Nodechain *c;
  
  mol_menu=Make_TopMolNameList("./ /",DefaultNode);
  key=(Token *)NULL;
  
  root=(Node *)NULL;
  
  for (loop=TRUE;loop;) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("piff option: ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("run")) {
	cmd_run_piff(root,TRUE);
	loop=FALSE;
      }else if (key_is("submit")) {
	cmd_run_piff(root,FALSE);
	loop=FALSE;
      }else if (key_is("molecule")) {
	key=free_Token(key);
	key=Get_Mols_M("molecule: ",";",mol_menu);
	send_bheader(graphix,"ARMK",0);
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (c->n==1) {
	    root=c->s;
	  }else{
	    root=(Node *)NULL;
	    err_set(1,"molecule selection %s is ambiguous.",key->text);
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad molecule selection: %s",key->text);
	}else {
	  loop=FALSE;
	}
      }else if (key_is("charges")) {
	if (IgnoreCharge) default_token("ignore");
	else default_token("use");
	key=free_Token(key);
	key=Get_Kwd("use/ignore partial charges ? ",";","use ignore");
	if (key->status==TOK_OK) IgnoreCharge=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad response: %s",key->text);
      }else if (key_is("cycles")) {
	key=free_Token(key); default_token("%d",Cycles);
	key=Get_Int_M("max. number of SCF/MM cycles: ",";","0 5 10 15 20 25 30 40 50 100");
	if (key->status==TOK_OK) Cycles=(key->n>=0 ? key->n : 10);
	else if (key->status==TOK_NOK) err_set(1,"bad numer of cycles: %s",key->text);
      }else if (key_is("deltaE")) {
	key=free_Token(key); default_token("%0.4f",deltaE);
	key=Get_Float_M("max. delta E (kJ/mol) between cycles: ",";",
		      "1.0 0.5 0.1 0.05 0.01 0.005 0.001");
	if (key->status==TOK_OK) deltaE=(key->f>0 ? key->f : 0.01);
	else if (key->status==TOK_NOK) err_set(1,"bad value for deltaE: %s",key->text);
      }else if (key_is("gradient")) {
	key=free_Token(key); default_token("%0.6f",grad);
	key=Get_Float_M("max. rms. gradient (mdyn) in MM: ",";",
		      "0.01 0.001 0.00035 0.00001 0.000005 0.000001");
	if (key->status==TOK_OK) grad=(key->f>=0 ? key->f : 0.00035);
	else if (key->status==TOK_NOK) err_set(1,"bad value for rms. gradient: %s",key->text);
      }else if (key_is("iterations")) {
	key=free_Token(key);
	if (Iter>0) default_token("%d",Iter);
	else default_token("15*NAtoms");
	key=Get_Int_M("max. numer of MM iterations per cycle: ",";",
		      "10 20 30 40 50 60 70 80 90 100 ' ' 15*NAtoms");
	if (key->status==TOK_NOK) {
	  key=Prs_Kwd(key,"15*NAtoms");
	  if (key->status==TOK_OK) key->n=-1;
	}
	if (key->status==TOK_OK) Iter=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad numer of iterations: %s",key->text);
      }else if (key_is("SCF")) {
	key=free_Token(key);
	if (SCFtype==1) default_token("PPP");
	else default_token("Huckel");
	key=Get_Kwd("pi-system treatment: ",";","Huckel PPP-SCF");
	if (key->status==TOK_OK) SCFtype=key->n;
	else if (key->status==TOK_NOK) err_set(1,"bad selection: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      if (!mvld(root)) {
	key=Prs_Mols_M(key,(char *)NULL);
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  if (c->n==1) {
	    root=c->s;
	  }else{
	    root=(Node *)NULL;
	    err_set(1,"molecule selection %s is ambiguous.",key->text);
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else{
	  err_set(1,"unknown piff option: %s.",key->text);
	}
      }
    }else{
      if (key->status==TOK_EOC) cmd_run_piff(root,TRUE);
      loop=FALSE;
    }
  }
  
  free_menu(mol_menu);
  free_Token(key);
}
/*-------------------- cmd_run_piff --------------------*/

static void cmd_run_piff(Node *root,int fgrd)
{
  Token *key;
  char *menu,*srv;
  char cmd[2048],rec[256],jobname[1024],buf[MAXBUF];
  Nodechain *c,atms;
  Atom *a;
  int status,i;
  UCpipe *Srv;

  key=(Token *)NULL;
  menu=(char *)NULL;
  init_node_chain(&atms);

  if (!is_mol(root)) {
    menu=Make_TopMolNameList("./ /",DefaultNode);
    key=Get_Mols_M("molecule: ",";",menu);
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      if (c->n==1) {
	root=c->s;
      }else{
	err_set(1,"ambiguous molecule specification: %s",key->text);
      }
      clear_node_chain(c);
      free_node_chain(c);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad molecule specification: %s",key->text);
    }
  }

  if (is_mol(root)) {
    if (mvld(root->name)) sprintf(jobname,"%s",root->name);
    else sprintf(jobname,"NoName");
    key=free_Token(key);
    default_token(jobname);
    key=Get_String("jobname: ",";");
    if (key->status==TOK_OK) {
      sprintf(jobname,"%s.piff",key->text);
      srv=getenv("PIFFserver");
      if (!srv) srv="piff";
      if (*srv=='|') srv++;
      sprintf(cmd,"%s%s%s",fgrd?"|":"",srv,fgrd?" -X":"");
      sprintf(cmd+strlen(cmd)," %s",jobname);
      sprintf(cmd+strlen(cmd)," -c%c ",(IgnoreCharge?'-':'+'));
      sprintf(cmd+strlen(cmd)," -I %d",Cycles);
      if (SCFtype==0) sprintf(cmd+strlen(cmd)," -H");
      sprintf(cmd+strlen(cmd)," -G %0.6f",deltaE);
      if (Iter>=0) sprintf(cmd+strlen(cmd)," -i %d",Iter);
      sprintf(cmd+strlen(cmd)," -g %0.6f",grad);
      if (fgrd) {      /* a pipe, run in foreground with automatic update */
	prtGraphics("running PIFF in foreground:\n");
	prtGraphics("%s\n",cmd);
	atms=wr_inter(root,jobname);
	if (atms.n>0) Srv=UCpopen(cmd);
	else Srv=(UCpipe *)NULL;
	if (Srv) {
	  for(;;) {
	    if (checkabort()) {
	      prtGraphics("Killed...\n");
	      break;
	    }
	    if (!fgets(rec,sizeof(rec)-1,Srv->fd)) break;
	    if (strneq(rec,"X> BEGIN",8)) {
	      a=(Atom *)atms.s;
	      while (fgets(rec,sizeof(rec)-1,Srv->fd)) {
		if (!strneq(rec,"X> ",3)) {
		  err_set(1,"%s",rec);
		  err_set(1,"coordinate set is incomplete.");
		  err_dump();
		  break;
		}
		if (strneq(rec+3,"END",3)) {
		  send_newx(&atms);
		  break;
		}else if (is_atom(a)) {
		  for (i=0;i<3;i++) a->x[i]=atof(rec+3+i*9);
		  a=(Atom *)a->chain;
		}
	      }
	    }else{
	      prtGraphics(rec);
	    }
	  }
	  UCpclose(Srv);
	}else{
	  if (atms.n>0) err_set(1,"could not start piff server.");
	}
	unlink(jobname);
      }else{                /* a regular command, run in background without update */
	WriteMolecule(root,jobname,FT_MOL,FALSE);
	status=startCmd(cmd,extfil(buf,jobname,".log",TRUE),"-join",TRUE,"belownormal");
	prtGraphics("running PIFF in background:\n");
	prtGraphics("%s\n",cmd);
	if (status) {
	  err_set(1,"PIFF: shell returns with error status %d",status);
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad jobname token: %s",key->text);
    }
  }

  clear_node_chain(&atms);
  if (menu) free_menu(menu);
  free_Token(key);

}
/*-------------------- wr_inter --------------------*/

static Nodechain wr_inter(Node *root,char *job) 
{
  Nodechain r;
  Atom *a,*b;
  char buff[4096];
  int clr,tp,i;
  FILE *out;

  init_node_chain(&r);
  clr=FALSE;

  if (asg_piff_types(root)) {
    out=fopen(job,"w");
    if (!out) {
      err_set(1,"sorry, could not open file %s for write.",buff);
      return(r);
    }
    ForAtoms(a,root) {
      tp=a->seqnr;
      if (tp>0) {
	add_to_node_chain((Node *)a,&r);
	a->seqnr=r.n;
	a->flg|=(1<<1);
	if (mvld(a->name)) sprintf(buff,"%-8s",a->name);
	else strncpy(buff,"Atom    ",8);
	buff[8]='\0'; fprintf(out,"%s %2d",buff,tp);
	for(i=0;i<3;i++) fprintf(out,"%12.4f",a->x[i]);
	fprintf(out,"%6.3f",a->q);
	for (i=0;i<a->nbnd;i++) {
	  b=neighbour(i,a);
	  if (b->flg&(1<<1)) fprintf(out,"%5d",b->seqnr);
	}
	fprintf(out,"\n");
      }else if (tp==0) {
	err_set(1,"failed to assign PIFF atom type to %s\n",
		nodeptr_to_name(buff,(Node *)a));
	clr=TRUE;
      }
    }
    for(a=(Atom *)r.s;mvld(a);a=(Atom *)a->chain) a->flg&=~(1<<1);
    if (clr) clear_node_chain(&r);
    fclose(out);
  }
  return(r);
}
/*-------------------- asg_piff_types --------------------*/

static int asg_piff_types(Node *root)
{
  /* sets a->seqnr to PIFF type (1...14) for atom in root,
                      0, if type could not be assigned,
                     -1, if a->atp->z==0 (dummy atoms).
  */

  Atom *a;
  int new,tp;

  ForAtoms(a,root) a->seqnr=(mvld(a->atp) ? (a->atp->z>0 ? 0 : -1) : 0);
  for (new=1;new>0;) {
    new=0;
    ForAtoms(a,root) {
      if (a->seqnr>=0) {
	tp=piff_type(a);
	if (tp>0) {
	  if (tp!=a->seqnr) {
	    a->seqnr=tp;
	    new++;
	  }
	}
      }
    }
  }
  return(1);
}
/*-------------------- piff_type --------------------*/

static int piff_type(Atom *a)
{
  Atom *b;
  int r,z,i;
  int npi[15]  ={0,1,0,0,2,1,0,1,2,0,2,0,2,0,0};
  int potpi[15]={0,0,0,0,0,0,1,0,0,1,0,1,0,1,0};

  if (a->seqnr==0) {
    r=0;
    if (mvld(a)&&mvld(a->atp)) {
      z=a->atp->z;
      switch (z) {
      case 1:
	for(r=3,i=0;i<a->nbnd;i++) {
	  b=neighbour(i,a);
	  if (b->atp) if (b->atp->z==7||b->atp->z==8) r=14;  /* Polar H */
	}
	break;
      case 6:
	for(r=2,i=0;i<a->nbnd;i++) {
	  if (a->bnd[i]->bt==BT_DOUBLE || a->bnd[i]->bt==BT_AROMATIC) r=1;
	  if (a->bnd[i]->bt==BT_TRIPLE) {
	    r=0;
	    break;
	  }
	}
	break;
      case 7:
	for(r=6,i=0;i<a->nbnd;i++) {
	  if (a->bnd[i]->bt==BT_DOUBLE || a->bnd[i]->bt==BT_AROMATIC ) r=5;
	  if (a->bnd[i]->bt==BT_TRIPLE) {
	    r=0;
	    break;
	  }
	}
	break;
      case 8:
	for(r=9,i=0;i<a->nbnd;i++) {
	  if (a->bnd[i]->bt==BT_DOUBLE) r=7;
	}     
	break;
      case 16:
	if (a->nbnd==2) {
	  if (a->bnd[0]->bt==BT_SINGLE && a->bnd[1]->bt==BT_SINGLE) r=11;
	}
	break;
      case 17:
	r=13;
	break;
      }
    }
  }else r=a->seqnr;

  if (potpi[r]) {
    switch (r) {
    case 6:
      for (i=0;i<a->nbnd&&r==6;i++) {
	b=neighbour(i,a);
	if (b->seqnr>0&&b->seqnr<15) {
	  if (npi[b->seqnr]>0) r=4;
	}
      }
      break;
    case 9:
      for (i=0;i<a->nbnd&&r==9;i++) {
	b=neighbour(i,a);
	if (b->seqnr>0&&b->seqnr<15) {
	  if (npi[b->seqnr]>0) r=8;
	}
      }
      break;
    case 11:
      for (i=0;i<a->nbnd&&r==10;i++) {
	b=neighbour(i,a);
	if (b->seqnr>0&&b->seqnr<15) {
	  if (npi[b->seqnr]>0) r=10;
	}
      }
      break;
    case 13:
      for (i=0;i<a->nbnd&&r==13;i++) {
	b=neighbour(i,a);
	if (b->seqnr>0&&b->seqnr<15) {
	  if (npi[b->seqnr]>0) r=12;
	}
      }
      break;
    }
  }
  return(r);
}
