#include <stdio.h>
#include "swdef.h"
#include "transpor.h"
#include "parser.h"
#include "mystdlib.h"
#include "object.h"
#include "errhand.h"
#include "strutil.h"

extern Node *MolRoot,*DefaultNode;
extern int graphix;

#define key_is(x) (streq(x,key->key))
#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

#define TRUE 1
#define FALSE 0

#define MAX_CODES   2000
#define MAX_TESTS    200
#define MAX_SCREENS  683
#define MAX_ATOMS    250
#define MAX_BONDS    300
#define MAX_ELDEFS    30
#define MAX_PARAM    500

typedef struct {
  char *mnem;
  int nr;
} testdef;

typedef struct {
  char *mnem;
  int nr;              /* nr: true -> set nr, false set -nr */
  int *excl;           /* list of incompatible screens  */
  char *txt;
} screendef;

int excl_sigma[]={73,74,75,76,77,0};  /* 0, to terminate the list */
int excl_R[]={87,88,89,90,91,92,0};

#define MAX_KNOWN_SCREENS 28

static screendef screen[MAX_KNOWN_SCREENS]={
  {"Metals",28,(int *)NULL,"Li,Na,K,Rb,Cs,Fr,Be,Mg,Ca,Sr,Ba,Ra,ln,An,Tr,metal present"},
  {"Organic",57,(int *)NULL,"Entry is an 'organic' compound"},
  {"Errors",34,(int *)NULL,"Entry contains uncorrected error(s)"},
  {"Disorder",-35,(int *)NULL,"Entry contains disorder"},    /* nr<0 --> FALSE = no disorder */
  {"SuppressedAtoms",36,(int *)NULL,"Entry contains suppressed atoms"},
  {"ValenceErrors",37,(int *)NULL,"Entry contains valence errors"},
  {"''",0,(int *)NULL,NULL},            /* just a spacer in menu */
  {"AbsConf",48,(int *)NULL,"The absolute configuration is reported"},
  {"Drugs",52,(int *)NULL,"'drug,'agnet' or 'activity' in qualifier text"},
  {"charged",55,(int *)NULL,"charged residues are present"},
  {"Diffractometer",81,(int *)NULL,"intensity measured with diffractometer"},
  {"''",0,(int *)NULL,NULL},              
  {"all-sigma(C-C)",0,excl_sigma,NULL},
  {"sigma(C-C)unkown",73,excl_sigma,"average sigma(C-C) not recorded in CSD"},
  {"sigma(C-C)0.001-0.005",74,excl_sigma,"average sigma(C-C) 0.001-0.005"},
  {"sigma(C-C)0.006-0.010",75,excl_sigma,"average sigma(C-C) 0.006-0.010"},
  {"sigma(C-C)0.011-0.030",76,excl_sigma,"average sigma(C-C) 0.011-0.030"},
  {"sigma(C-C)>0.031",77,excl_sigma,"average sigma(C-C) > 0.031 A"},
  {"''",0,(int *)NULL,NULL},
  {"all-R",0,excl_R,NULL},
  {"R.le.0.120",87,excl_R,"R-factor <= 0.120"},
  {"R.le.0.100",88,excl_R,"R-factor <= 0.100"},
  {"R.le.0.075",89,excl_R,"R-factor <= 0.075"},
  {"R.le.0.050",90,excl_R,"R-factor <= 0.050"},
  {"R.le.0.030",91,excl_R,"R-factor <= 0.030"},
  {"R_unknown", 92,excl_R,"R-factor not recorded"},
  {"''",0,(int *)NULL,NULL},
  {"Coords",153,(int *)NULL,"atom coordinates present"}
};

#define TST_NOP  0
#define TST_TEXT 1
#define TST_NBR  2
#define TST_CONN 4

typedef struct text_test {
  char *keyword;
  char *text;
} text_test;

typedef struct number_test {
  char *keyword;
  char *op,*val;
} number_test; 

typedef struct connser_test {
  Atom *atm[MAX_ATOMS];
  Bond *bnd[MAX_BONDS];
  char *eldef[MAX_ELDEFS],*pname[MAX_PARAM];
  int  mca[MAX_ATOMS],chg[MAX_ATOMS],
       lig[MAX_ATOMS];
  int  from[MAX_BONDS],to[MAX_BONDS],bt[MAX_BONDS];
  int  pnr[MAX_PARAM][4];
  int  npar,nbnd,nat,nel,noln,nocr;
  char newel[MAX_ELDEFS][3],exct[MAX_ATOMS],cyca[MAX_ATOMS],cycb[MAX_BONDS],
       el[MAX_ATOMS][3],nh[MAX_ATOMS][20];
} connser_test;

#define MAX_KNOWN_TESTS 22

static testdef test[MAX_KNOWN_TESTS] = {
   {"Connser",TST_CONN}, {"' '",TST_NOP},{"Author",TST_TEXT}, {"Compound",TST_TEXT},
   {"Formula",TST_TEXT}, {"Name",TST_TEXT},
   {"Page",TST_TEXT}, {"Refcode",TST_TEXT}, {"Spacegroup",TST_TEXT}, {"Surname",TST_TEXT},
   {"Synonym",TST_TEXT}, {"Text",TST_TEXT}, {"Volume",TST_TEXT}, {"Xname",TST_TEXT},
   {"' '",TST_NOP},
   {"Class",TST_NBR}, {"Coden",TST_NBR}, {"Coords",TST_NBR}, {"RFactor",TST_NBR},
   {"Spgnumber",TST_NBR}, {"Year",TST_NBR}, {"Zvalue",TST_NBR} 
   
 };
  
#define CSD_BT_SINGLE      1
#define CSD_BT_DOUBLE      2
#define CSD_BT_TRIPLE      3
#define CSD_BT_QUADRUPLE   4
#define CSD_BT_AROMATIC    5
#define CSD_BT_POLYMERIC   6
#define CSD_BT_DELOCAL     7
#define CSD_BT_PIBOND      9
#define CSD_BT_ANY        99

#define N_SAVE_TYPES       6
#define SAVE_REFCODES     (0)
#define SAVE_ASER         (1)
#define SAVE_FDAT         (2)
#define SAVE_FCON         (3)
#define SAVE_FBIB         (4)
#define SAVE_MOL2         (5)

static char *save_mnem[N_SAVE_TYPES]={"REFCODES","ASER","FDAT","FCON","FBIB","MOL2"};

typedef struct Quest {
  char *jobname;
  char *Question;
  void *Test[MAX_TESTS];
  char *Ret[MAX_CODES];
  int Type[MAX_TESTS];
  int Screen[MAX_SCREENS],retmode,nRet,Stop,Save;
}Quest;

static char *main_keys = "clear define delete edit question retrieve save screen stop ' ' "
  "browse list write run ' ' command color cpk display label sketch ' ' end";

static void init_quest(Quest *Q);
static void clear_quest(Quest *Q);
static void *free_Test(void *tst,int type);
static void quest_browse(Quest *Q);
static void quest_clear(Quest *Q);
static void quest_define(Quest *Q);
static void quest_clear(Quest *Q);
static void quest_delete(Quest *Q);
static void quest_edit(Quest *Q);
static void edit_connser(connser_test *t);
static void edit_connser_atom(connser_test *t);
static void edit_connser_bond(connser_test *t);
static void quest_eldef(connser_test *t);
static void quest_parameter(connser_test *t);
static void quest_parameter_add(connser_test *t);
static void quest_parameter_delete(connser_test *t);
static void quest_parameter_list(connser_test *t);
static void quest_question(Quest *Q);
static void quest_retrieve(Quest *Q);
static void quest_save(Quest *Q);
static void quest_screen(Quest *Q);
static void quest_stop(Quest *Q);
static void quest_file(Quest *Q);
static void quest_run(Quest *Q);
static void write_quest(FILE *out,Quest *Q);
static void define_test_text(Quest *Q,int nr,char *mnem);
static void define_test_number(Quest *Q,int nr,char *mnem);
static void define_test_connser(Quest *Q,int nr);
static void write_test_text(FILE *out,int nr,text_test *t);
static void write_test_number(FILE *out,int nr,number_test *t);
static void write_test_connser(FILE *out,int nr,connser_test *t);
static number_test *make_number_test(void);
static text_test *make_text_test(void);
static connser_test *make_connser_test(void);
static int sane_connser(connser_test *q);
static void label_connser(connser_test *q);
static void label_connser_atom(connser_test *q,int nr);
static void label_connser_bond(connser_test *q,int nr);
static int csd_bt(int bt);
static void list_screens(Quest *Q);

/*========================================================================*/

void cmd_quest(void)
{
  
  Quest Q;

  Token *key;
  int loop;

  init_quest(&Q);

  key=(Token *)NULL; loop=TRUE;
  while (loop) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("Quest> ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("define"))          quest_define(&Q);
      else if (key_is("clear"))      quest_clear(&Q);
      else if (key_is("delete"))     quest_delete(&Q);
      else if (key_is("edit"))       quest_edit(&Q);
      else if (key_is("question"))   quest_question(&Q);
      else if (key_is("retrieve"))   quest_retrieve(&Q);
      else if (key_is("save"))       quest_save(&Q);
      else if (key_is("screen"))     quest_screen(&Q);
      else if (key_is("stop"))       quest_stop(&Q);
      else if (key_is("browse"))     quest_browse(&Q);
      else if (key_is("write"))      quest_file(&Q);
      else if (key_is("list"))       write_quest((FILE *)NULL,&Q);
      else if (key_is("run"))        quest_run(&Q);
      else if (key_is("command"))    command_loop("Wit!P(Quest)> ");
      else if (key_is("color"))      cmd_color();
      else if (key_is("cpk"))        cmd_cpk();
      else if (key_is("display"))    cmd_display();
      else if (key_is("label"))      cmd_label();
      else if (key_is("sketch"))     cmd_sketch();
      else if (key_is("end"))        loop=FALSE;
    }else if (key->status==TOK_NOK) {
      err_set(1,"not a Quest keyword: %s",key->text);
    }else if (key->status==TOK_EOC) loop=FALSE;
  }
  
  key=free_Token(key);
  clear_quest(&Q);
}
/*-------------------- init_quest --------------------*/

static void init_quest(Quest *Q)
{
  int i;

  for (i=0;i<MAX_TESTS;i++) Q->Test[i]=(void *)NULL;
  for (i=0;i<MAX_SCREENS;i++) Q->Screen[i]=0;
  for (i=0;i<MAX_CODES;i++) Q->Ret[i]=(char *)NULL;
  Q->nRet=0; Q->retmode=1;
  Q->Question=(char *)NULL;
  Q->Stop=1000;
  Q->Save=(1<<SAVE_FDAT);
  Q->jobname=(char *)NULL;
}

/*-------------------- clear_question --------------------*/

static void clear_quest(Quest *Q)
{
  int i;

  for (i=0;i<MAX_TESTS;i++) {
    if (mvld(Q->Test[i])) Q->Test[i]=free_Test(Q->Test[i],Q->Type[i]);
  }
  for(i=0;i<MAX_CODES;i++) if (mvld(Q->Ret[i])) free(Q->Ret[i]);
  if (mvld(Q->Question)) free(Q->Question);
  Q->Question=(char *)NULL;
  if (mvld(Q->jobname)) free(Q->jobname);
  Q->jobname=(char *)NULL;
  init_quest(Q);
}
/*-------------------- free_Test --------------------*/

static void *free_Test(void *tst,int type) 
{
  int i;
  text_test *tx;
  number_test *tn;
  connser_test *tc;

  if (mvld(tst)) {
    switch (type) {
    case TST_TEXT: 
      tx=(text_test *)tst;
      if (mvld(tx->keyword)) free(tx->keyword);
      if (mvld(tx->text)) free(tx->text);
      break;
    case TST_NBR:
      tn=(number_test *)tst;
      if (mvld(tn->keyword)) free(tn->keyword);
      break;
    case TST_CONN:
      tc=(connser_test *)tst;
      for (i=0;i<tc->nel;i++) if (mvld(tc->eldef[i])) free(tc->eldef[i]);
      for (i=0;i<tc->npar;i++) if (mvld(tc->pname[i])) free(tc->pname[i]);
      tc->nel=0; tc->npar=0;
      break;
    default:
      err_set(1,"unknown test type %d in free_test.",type);
      break;
    }
    free(tst);
  }
  return((void *)NULL);
}
/*-------------------- quest_edit --------------------*/

static void quest_edit(Quest *Q)
{
  Token *key;
  char menu[4096];
  int i,k;

  for (*menu='\0',k=0,i=1;i<MAX_TESTS;i++) {
    if (mvld(Q->Test[i])) {
      if (Q->Type[i]==TST_CONN) {
	sprintf(menu+strlen(menu)," T%d",i);
	k++;
      }
    }
  }
  if (k==0) {
    err_set(1,"no connser test to edit...");
    return;
  }
  key=Get_Kwd("edit connser test> ",";",menu);
  if (key->status==TOK_OK) {
    k=str2i(key->text+1);
    if (k>0&&k<MAX_TESTS) {
      if (Q->Type[k]==TST_CONN) edit_connser((connser_test *)Q->Test[k]);
    }else{
      err_set(1,"T%d is not a connser test.",k);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"no such connser test: %s",key->text);
  }
  free_Token(key);
}

/*-------------------- quest_delete --------------------*/

static void quest_delete(Quest *Q)
{
  Token *key;
  char menu[4096];
  int i,k;

  for (*menu='\0',k=0,i=1;i<MAX_TESTS;i++) {
    if (mvld(Q->Test[i])) {
      sprintf(menu+strlen(menu)," T%d",i);
      k++;
    }
  }
  if (k==0) {
    err_set(1,"nothing to delete...");
    return;
  }
  key=Get_Kwd("delete test> ",";",menu);
  if (key->status==TOK_OK) {
    k=str2i(key->text+1);
    if (k>0&&k<MAX_TESTS) {
      Q->Test[k]=free_Test(Q->Test[k],Q->Type[k]);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"no such test: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- quest_define --------------------*/

static void quest_define(Quest *Q)
{
  Token *key;
  char menu[4096];
  int i,k;

  key=(Token *)NULL;
  for (*menu='\0',k=0,i=1;i<MAX_TESTS;i++) {
    if (!mvld(Q->Test[i])) {
      sprintf(menu+strlen(menu)," T%d",i);
      if (k==0) default_token("T%d",i);
      k++;
    }
  }
  if (k==0) {
    err_set(1,"Can't add new tests...");
    default_token((char *)NULL);
    return;
  }
  key=Get_Kwd("define test> ",";",menu);
  if (key->status==TOK_OK) {
    k=str2i(key->text+1);
    if (k>0&&k<MAX_TESTS) {
      if (!mvld(Q->Test[k])) {
	key=free_Token(key);
	for (*menu='\0',i=0;i<MAX_KNOWN_TESTS;i++) sprintf(menu+strlen(menu)," %s",test[i].mnem);
	key=Get_Kwd("type> ",";",menu);
	if (key->status==TOK_OK) {
	  switch (test[key->n].nr) {
	  case TST_TEXT:
	    define_test_text(Q,k,test[key->n].mnem);
	    break;
	  case TST_NBR:
	    define_test_number(Q,k,test[key->n].mnem);
	    break;
	  case TST_CONN:
	    define_test_connser(Q,k);
	    break;
	  case TST_NOP:
	    break;
	  default:
	    err_set(1,"internal error: unknown test %s",test[key->n]);
	    break;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"sorry, tests of type %s are not supported...",key->text);
	}
      }else{
	err_set(1,"sorry, internal error in quest_define (used).");
      }
    }else{
      err_set(1,"sorry, internal error in quest_define (range).");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"Test %s is not available.",key->text);
  }
  free_Token(key);
}
/*-------------------- quest_parameter --------------------*/

static void quest_parameter(connser_test *t)
{
  Token *key=(Token *)NULL;

  key=Get_Kwd("parameter> ",";","add delete list");
  if (key->status==TOK_OK) {
    if (key_is("add")) quest_parameter_add(t);
    else if (key_is("delete")) quest_parameter_delete(t);
    else if (key_is("list")) quest_parameter_list(t);
  }else if (key->status==TOK_NOK) err_set(1,"bad parameter command: %s\n",key->text);
  free_Token(key);
}

static void quest_parameter_add(connser_test *t)
{
  Token *key=(Token *)NULL;
  Token *atm[4];
  char buf[4096];
  int loop=TRUE;
  int nat[4]={3,2,4,0};
  int ng,n,i,k;

  if (t->npar>=MAX_PARAM) {
    err_set(1,"parameter table is full.\n");
    loop=FALSE;
  }

  while (loop) {
    key=Get_Kwd("type> ",";","angle distance torsion h!end");
    if (key->status==TOK_OK) {
      ng=nat[key->n];
      key=free_Token(key);
      if (ng==0) break;
      n=Get_nAtoms(ng,atm,"",";");
      if (n==ng) {
	send_header(graphix,"ARMK",0);
	default_token("%c%d",(n==2?'B':(n==3?'A':'T')),t->npar+1);
	key=Get_String("paramater name: ",";");
	if (key->status==TOK_OK) {
	  for (i=0;i<4;i++) t->pnr[t->npar][i]=-1;
	  for (i=0;i<ng;i++) {
	    for (k=0;k<t->nat;k++) if (t->atm[k]==(Atom *)(atm[i]->ptr)) break;
	    if (k<t->nat) t->pnr[t->npar][i]=k;
	    else {
	      err_set(1,"not a query atom: %s\n",nodeptr_to_name(buf,(Node *)(atm[i]->ptr)));
	      break;
	    }
	  }
	  if (i==ng) {
	    t->pname[t->npar]=key->text;
	    key->text=(char *)NULL;
	    t->npar++;
	  }
	}else if (key->status==TOK_NOK) err_set(1,"bad parameter name: %s\n",key->text);
	for (i=0;i<ng;i++) free_Token(atm[i]);
      }else{
	if (n>0) send_header(graphix,"ARMK",0);
	for (i=0;i<n;i++) free_Token(atm[i]);
      }
    }else if (key->status==TOK_NOK) err_set(1,"bad parameter add command: %s\n",key->text);
    else loop=FALSE;
    err_dump();
    start_parse();
    free_Token(key);
  }
}

static void quest_parameter_delete(connser_test *t)
{
  Token *key=(Token *)NULL;
  char *menu=(char *)NULL;
  int i,n,l;
  
  if (t->npar==0) {
    err_set(1,"no parameters defined for this test,\n");
  }
  while (t->npar>0) {
    for (l=strlen("* ''")+t->npar,i=0;i<t->npar;i++) l+=strlen(t->pname[i]);
    menu=(char *)malloc(l*sizeof(char)+1);
    if (menu) {
      l=sprintf(menu,"* ''");
      for (i=0;i<t->npar;i++) l+=sprintf(menu+l," %s",t->pname[i]);
      key=Get_String_M("delete: ",";",menu);
      free(menu);
      if (key->status==TOK_OK) {
	for (n=0,i=0;i<t->npar;i++) {
	  if (!wild_xmatch(t->pname[i],key->text)) {
	    if (n>0) {
	      t->pname[i-n]=t->pname[i];
	      for (l=0;l<4;l++) t->pnr[i-n][l]=t->pnr[i][l];
	    }
	  }else{
	    free(t->pname[i]);
	    n++;
	  }
	}
	t->npar-=n;
	prtGraphics("%d parameters deleted\n",n);
	if (n==0) {
	  key->status=TOK_NOK;
	  err_set(1,"bad parameter name: %s\n",key->text);
	}
      }else if (key->status==TOK_NOK) err_set(1,"bad paramater name: %s\n",key->text);
    }else{
      err_set(1,"memory allocation failure.\n");
      break;
    }
    if (key->status!=TOK_OK) break;
    key=free_Token(key);
  }

  free_Token(key);
}

static void quest_parameter_list(connser_test *t)
{
  Token *key=(Token *)NULL;
  Node *ctxt;
  char buf[4096],name[4096];
  int l,i,j;

  if (t->npar==0) {
    prtGraphics("no paramters defined for this test\n");
  }else{
    prtGraphics("The following paramters are defined for this test:\n");
    for (i=0;i<t->npar;i++) {
      l=sprintf(buf,"%-6s ",t->pname[i]);
      for (ctxt=(Node *)NULL,j=0;j<4;j++) {
	if (t->pnr[i][j]>-1) {
	  l+=sprintf(buf+l,"%s -- ",
		     nodeptr_to_name_ctxt(name,(Node *)(t->atm[t->pnr[i][j]]),ctxt));
	  ctxt=(Node *)(t->atm[t->pnr[i][j]]);
	}
      }
      buf[l-4]='\0';
      prtGraphics("%s\n",buf);
    }
  }
  free_Token(key);
}

/*-------------------- quest_question --------------------*/

static void quest_question(Quest *Q)
{
  Token *key;

  key=Get_String("question> ",";");
  if (key->status==TOK_OK) {
    if (mvld(Q->Question)) free(Q->Question);
    Q->Question=key->text;
    key->text=(char *)NULL;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad question: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- quest_retrieve --------------------*/

static void quest_retrieve(Quest *Q)
{
  Token *key;
  int i;
  
  key=(Token *)NULL;
  key=Get_Kwd("retrieve what ? ",";","codes families nothing");
  if (key->status==TOK_OK) {
    if (key_is("codes")) Q->retmode=2;
    else if (key_is("families")) Q->retmode=1;
    else if (key_is("nothing")) {
      for (i=0;i<Q->nRet;i++) {
	if (mvld(Q->Ret[i])) free(Q->Ret[i]);
	Q->Ret[i]=(char *)NULL;
      }
      Q->nRet=0;
      key=free_Token(key);
      return;
    }
  }else{
    if (key->status==TOK_NOK) err_set(1,"bad retrieve option: %s",key->text);
    key=free_Token(key);
    return;
  }

  if (Q->nRet>0) prtGraphics("current list contains %d codes.",Q->nRet);

  for (;Q->nRet<MAX_CODES;) {
    key=free_Token(key); err_dump();
    key=Get_String_M("refcodes> ",";","end");
    if (key->status==TOK_OK) {
      key=Prs_Kwd(key,"end");
      if (key->status==TOK_OK) break;
      if (strlen(key->text)<9) {
	Q->Ret[Q->nRet++]=key->text;
	key->text=(char *)NULL;
      }else{
	err_set(1,"refocde %s is too long, rejected.",key->text);
      }
    }else{
      if (key->status==TOK_NOK) err_set(1,"bad refcode token: %s",key->text);
      break;
    }
  }

  free_Token(key);
}
/*-------------------- quest_save --------------------*/

static void quest_save(Quest *Q)
{
  Token *key;
  int i;
  char buff[512];
  char *menu="+REFCODES -REFCODES '' +ASER -ASER ' ' +FDAT -FDAT '' "
             "+FCON -FCON '' +FBIB -FBIB '' +MOL2 -MOL2 '' end";

  *buff='\0';
  for (i=1;i<N_SAVE_TYPES;i++) if (Q->Save&(1<<i)) sprintf(buff+strlen(buff)," %s",save_mnem[i]);
  prtGraphics("current setting: Save %s\n",buff);
  key=(Token *)NULL;
  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("save> ",";",menu);
    if (key->status==TOK_OK) {
      if (key_is("+REFCODES")) Q->Save|=(1<<SAVE_REFCODES);
      else if (key_is("+ASER")) Q->Save|=(1<<SAVE_ASER);
      else if (key_is("+FDAT")) Q->Save|=(1<<SAVE_FDAT);
      else if (key_is("+FCON")) Q->Save|=(1<<SAVE_FCON);
      else if (key_is("+FBIB")) Q->Save|=(1<<SAVE_FBIB);
      else if (key_is("+MOL2")) Q->Save|=(1<<SAVE_MOL2);
      else if (key_is("-REFCODES")) Q->Save&=~(1<<SAVE_REFCODES);
      else if (key_is("-ASER")) Q->Save&=~(1<<SAVE_ASER);
      else if (key_is("-FDAT")) Q->Save&=~(1<<SAVE_FDAT);
      else if (key_is("-FCON")) Q->Save&=~(1<<SAVE_FCON);
      else if (key_is("-FBIB")) Q->Save&=~(1<<SAVE_FBIB);
      else if (key_is("-MOL2")) Q->Save&=~(1<<SAVE_MOL2);
      else if (key_is("end")) break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad save option: %s",key->text);
    }else break;
  }
  
  *buff='\0';
  for (i=1;i<N_SAVE_TYPES;i++) if (Q->Save&(1<<i)) sprintf(buff+strlen(buff)," %s",save_mnem[i]);
  prtGraphics("new setting: Save %s\n",buff);
		
  free_Token(key);
}
/*-------------------- quest_screen --------------------*/

static void quest_screen(Quest *Q)
{
  Token *key;
  char menu[4096],prompt[256];        /* let's pray that this is long enough... */
  int i,mode,prefix,*s;

  key=(Token *)NULL;
  mode=1;
  sprintf(menu,"set h!true negate h!false ignore list '' end '' ");
  prefix=strlen(menu);
  for (i=0;i<MAX_KNOWN_SCREENS;i++) sprintf(menu+strlen(menu)," %s",screen[i].mnem);
  for (;;) {
    key=free_Token(key); err_dump();
    sprintf(prompt,"%s screen: ",(mode==1?"set":mode==0?"ignore":"negate"));
    key=Get_Kwd(prompt,";",menu);
    if (key->status==TOK_NOK) {
      key=Prs_Int(key);
      if (key->status==TOK_OK) {
	if (key->n>0 && key->n <MAX_SCREENS) Q->Screen[key->n]=mode;
	else err_set(1,"screen %d is out of range (1...%d)",key->n,MAX_SCREENS-1);
      }else{
	err_set(1,"bad token: %s",key->text);
      }
    }else if (key->status==TOK_OK) {
      if (key_is("end")) break;
      else if (key_is("true")||key_is("set"))   mode=1;
      else if (key_is("false")||key_is("negate"))  mode=-1;
      else if (key_is("ignore")) mode=0;
      else if (key_is("list")) list_screens(Q);
      else {
	key=Prs_Kwd(key,menu+prefix);
	if (key->n>=0 && key->n < MAX_KNOWN_SCREENS) {
	  if (screen[key->n].excl) {
	    if (mode==1||screen[key->n].nr==0) {
	      for (s=screen[key->n].excl;*s;s++) {
		if (*s>0 && *s<MAX_SCREENS) Q->Screen[*s]=0;
	      }
	    }
	  }
	  if (screen[key->n].nr>0) {
	    Q->Screen[screen[key->n].nr]=mode;
	  }else if (screen[key->n].nr<0){
	    Q->Screen[-screen[key->n].nr]=-mode;
	  }
	}
      }
    }else break;
  }
  free_Token(key);
}
/*-------------------- quest_stop --------------------*/

static void quest_stop(Quest *Q)
{
  Token *key;

  if (Q->Stop==0) default_token("infinity");
  else default_token("%d",Q->Stop);
  key=Get_Int_M("maximum number of hits: ",";","1 5 10 50 100 1000 2000 infinity");
  if (key->status==TOK_NOK) {
    key=Prs_Kwd(key,"infinity");
    if (key->status==TOK_OK) key->n=0;
  }
  if (key->status==TOK_OK) {
    if (key->n<0) {
      err_set(1,"number must be non-negative: %d",key->n);
    }else{
      Q->Stop=key->n;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- quest_file --------------------*/

static void quest_file(Quest *Q)
{
  Token *key;
  FILE *out;
  char buf[4096];

  key=Get_String("filename: ",";");
  if (key->status==TOK_OK) {
    out=fopen(extfil(buf,key->text,".que",FALSE),"w");
    if (out) {
      prtGraphics("writing question to %s",buf);
      write_quest(out,Q);
      fclose(out);
    }else err_set(1,"could not open %s for write.",buf);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad filename: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- quest_run --------------------*/

static void quest_run(Quest *Q)
{
  Token *key;
  FILE *out;
  char buf[4096],*srv;
  int status;

  srv=getenv("QUESTserver");
  if (!srv) {
    err_set(1,"sorry, QUESTserver is not defined.");
    return;
  }

  default_token(Q->jobname);
  key=Get_String("jobname: ",";");
  if (key->status==TOK_OK) {
    out=fopen(extfil(buf,key->text,".que",FALSE),"w");
    if (out) {
      write_quest(out,Q);
      fclose(out);
      sprintf(buf,"%s %s",srv,key->text);
      status=startCmd(buf,"","",TRUE,"normal");
      if (mvld(Q->jobname)) free(Q->jobname);
      Q->jobname=Strdup(key->text);
      if (status&0xff) {
	err_set(1,"Quest: shell returns with error status %d",status&0xff);
      }else if (status>>8) {
	err_set(1,"Quest: exit status is %d",status>>8);
      }
    }else err_set(1,"could not open %s for write.");
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  key=free_Token(key);
}
/*-------------------- write_quest --------------------*/

static void write_quest(FILE *out,Quest *Q)
{
  int i,n,ntests;
  char default_question[1024],*question;
  char buff[4096];

  *default_question='\0';

  if (out) fprintf(out,"Comment QUEST input file written by Wit!P\n");
  sprintf(buff,"SAVE ");
  for (i=1;i<N_SAVE_TYPES;i++) if (Q->Save&(1<<i)) sprintf(buff+strlen(buff)," %s",save_mnem[i]);
  if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
  if (Q->Stop>0) {
    if (out) fprintf(out,"STOP %d\n",Q->Stop);else prtGraphics("STOP %d\n",Q->Stop);
  }
  for (n=0,i=0;i<MAX_SCREENS;i++) {
    if (Q->Screen[i]) {
      if (n++==0) sprintf(buff,"SCREEN ");
      sprintf(buff+strlen(buff)," %d",(Q->Screen[i] > 0 ? i : -i));
      if (n==10) {
	if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
	n=0;
      }
    }
  }
  if (n>0) {
    if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
  }
  if (Q->nRet>0) {
    sprintf(buff,"RETRIEVE %d\n",Q->retmode);
    if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
    for (n=0,i=0;i<Q->nRet;i++) {
      if (mvld(Q->Ret[i])) {
	if (n++==0) sprintf(buff,"   ");
	sprintf(buff+strlen(buff)," %-8s",Q->Ret[i]);
	if (n==5) {
	  if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
	  n=0;
	}
      }
    }
    if (n>0) {
      if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
    }
    if (out) fprintf(out,"END\n"); else prtGraphics("END\n");
  }
  for(ntests=0,i=1;i<MAX_TESTS;i++) {
    if (mvld(Q->Test[i])) {
      if (ntests++>0) sprintf(default_question+strlen(default_question),"+");
      sprintf(default_question+strlen(default_question),"T%d",i);
      switch (Q->Type[i]) {
      case TST_TEXT:
	write_test_text(out,i,(text_test *)Q->Test[i]);
	break;
      case TST_NBR:
	write_test_number(out,i,(number_test *)Q->Test[i]);
	break;
      case TST_CONN:
	write_test_connser(out,i,(connser_test *)Q->Test[i]);
	break;
      default:
	err_set(1,"test type %d not implemeted.",Q->Type[i]);
      }
    }
  }
  question=(mvld(Q->Question) ? Q->Question : default_question);
  if (strlen(question)>0) {
    if (out) fprintf(out,"Question %s\n",question);else prtGraphics("Question %s\n",question);
  }
}
/*-------------------- define_test_text --------------------*/
static void define_test_text(Quest *Q,int nr,char *mnem)
{
  Token *key;
  text_test *t;

  key=Get_String("text> ",";");
  if (key->status==TOK_OK) {
    t=make_text_test();
    if (mvld(t)) {
      t->keyword=Strdup(mnem);
      if (!mvld(t->keyword)) {
	err_set(1,"memory allocation failure.");
	free(t);
      }else{
	t->text=key->text;
	key->text=(char *)NULL;
	Q->Test[nr]=(void *)t;
	Q->Type[nr]=TST_TEXT;
      }
    }else{
      err_set(1,"memory allocation failure.");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- quest_browse --------------------*/

static void quest_browse(Quest *Q)
{
  Node *root,*sv_default;
  
  if (mvld(Q->jobname)) {
    sv_default=DefaultNode;
    root=ReadNode(MolRoot,Q->jobname,FT_CSD);
    if (mvld(root)) del_node(root);   /* don't even ask... */
    DefaultNode=sv_default;
  }else{
    err_set(1,"must run before browse.");
  }
}
/*-------------------- quest_clear --------------------*/

static void quest_clear(Quest *Q)
{
  Token *key;
  
  key=Get_Kwd("confirm: ",";","yes no");
  if (key->status==TOK_OK) {
    if (key_is("yes")) clear_quest(Q);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
}
/*-------------------- make_text_test --------------------*/

static text_test *make_text_test(void)
{
  text_test *r;

  r=(text_test *)malloc(sizeof(text_test));
  if (mvld(r)) {
    r->keyword=(char *)NULL;
    r->text=(char *)NULL;
  }
  return(r);
}
/*-------------------- write_test_text --------------------*/

static void write_test_text(FILE *out,int nr,text_test *t)
{
  if (out) fprintf(out,"DEFINE T%d *%s '%s'\n",nr,t->keyword,t->text);
  else prtGraphics("DEFINE T%d *%s '%s'\n",nr,t->keyword,t->text);
}
/*-------------------- define_test_number --------------------*/

static void define_test_number(Quest *Q,int nr,char *mnem)
{
  Token *key;
  number_test *t;
  char *op;

  key=Get_Kwd("operator> ",";",".eq. .ne. .lt. .le. .gt. .ge.");
  if (key->status==TOK_NOK) {
    key=Prs_Kwd(key,"eq ne lt le gt ge");
    if (key->status==TOK_OK) {
      if (key_is("eq")) {
	free(key->key);key->key=Strdup(".eq.");
      }else if (key_is("ne")) {
	free(key->key);key->key=Strdup(".ne.");
      }else if (key_is("lt")) {
	free(key->key);key->key=Strdup(".lt.");
      }else if (key_is("le")) {
	free(key->key);key->key=Strdup(".le.");
      }else if (key_is("gt")) {
	free(key->key);key->key=Strdup(".gt.");
      }else if (key_is("ge")) {
	free(key->key);key->key=Strdup(".ge.");
      }
      if (!mvld(key)) key->status=TOK_NOK;
    }
  }
  if (key->status==TOK_OK) {
    op=key->key; key->key=(char *)NULL;
    key=free_Token(key);
    key=Get_String("value> ",";");
    if (key->status==TOK_OK) {
      t=make_number_test();
      if (mvld(t)) {
	t->keyword=Strdup(mnem);
	if (!mvld(t->keyword)) {
	  err_set(1,"memory allocation failure.");
	  free(op);
	  free(t);
	}else{
	  t->op=op;
	  t->val=key->text;
	  key->text=(char *)NULL;
	  Q->Test[nr]=(void *)t;
	  Q->Type[nr]=TST_NBR;
	}
      }else{
	err_set(1,"memory allocation failure.");
	free(op);
      }
    }else{
      if (key->status==TOK_NOK) err_set(1,"bad value token: %s",key->text);
      free(op);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- make_number_test --------------------*/

static number_test *make_number_test(void)
{
  number_test *r;

  r=(number_test *)malloc(sizeof(number_test));
  if (mvld(r)) {
    r->keyword=(char *)NULL;
    r->op=(char *)NULL;
    r->val=(char *)NULL;
  }
  return(r);
}
/*-------------------- write_test_number --------------------*/

static void write_test_number(FILE *out,int nr,number_test *t)
{
  if (out) fprintf(out,"DEFINE T%d *%s %s %s\n",nr,t->keyword,t->op,t->val);
  else prtGraphics("DEFINE T%d *%s %s %s\n",nr,t->keyword,t->op,t->val);
}
/*-------------------- define_test_connser --------------------*/

static void define_test_connser(Quest *Q,int nr)
{
  Token *key;
  Nodechain *c=(Nodechain *)NULL;
  Atom *a,*aa,*A[MAX_ATOMS];
  connser_test *q;
  char buff[4096];
  int i,j,nat=0,nbnd=0;
  
  key=(Token *)NULL;

  key=Get_Atoms_D("atoms: ",";");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n<1) {
      err_set(1,"no atoms selected.");
      clear_node_chain(c);
      free_node_chain(c);
      key->status=TOK_ABORT;
    }
  }
  if (key->status==TOK_OK) {
    q=make_connser_test();
    if (!mvld(q)) {
      err_set(1,"failed to allocate connser data structure.");
    }else{
      nat=0; nbnd=0;
      for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	if (mvld(a->atp)) {
	  if (!streq(a->atp->chs,"H")) {    /* ignore hydrogens */
	    if (nat<MAX_ATOMS) A[nat++]=a;
	    else{
	      err_set(1,"too many connser atoms, max. is %d",MAX_ATOMS);
	      q=free_Test(q,TST_CONN);
	      break;
	    }
	  }
	}else{
	  err_set(1,"no atomtype defined for connser atom %s",nodeptr_to_name(buff,(Node *)a));
	  q=free_Test(q,TST_CONN);
	  break;
	}
      }
      if (nat<1) {
	err_set(1,"no non-H atoms in connser fragment.");
	q=free_Test(q,TST_CONN);
      }
    }
    if (mvld(q)) {
      for (i=0;i<nat;i++) {
	A[i]->seqnr=i;
	A[i]->flg|=(1<<1);
      }
      for (i=0;i<nat;i++) {
	a=A[i];
	q->atm[i]=a;
	Strcpy(q->el[i],A[i]->atp->chs);
	if (streq(q->el[i],"Du")) Strcpy(q->el[i],"AA");
	q->mca[i]=0; sprintf(q->nh[i],"any"); q->lig[i]=99;
	q->chg[i]=99;
	q->exct[i]=' ';
	q->cyca[i]=' ';
	for (j=0;j<a->nbnd;j++) {
	  aa=neighbour(j,a);
	  if (aa->flg&(1<<1)) {
	    q->mca[i]++;
	    if (a==a->bnd[j]->f) {
	      if (nbnd<MAX_BONDS) {
		q->from[nbnd]=a->seqnr+1;
		q->to[nbnd]=aa->seqnr+1;
		q->bt[nbnd]=csd_bt(a->bnd[j]->bt);
		q->cycb[nbnd]=' ';
		q->bnd[nbnd]=a->bnd[j];
		nbnd++;
	      }else nbnd++;
	    }
	  }
	}
      }
      if (nbnd>MAX_BONDS) {
	err_set(1,"too many bonds in connser fragment, max. is %d",MAX_BONDS);
	q=free_Test(q,TST_CONN);
      }
      for (i=0;i<nat;i++) A[i]->flg&=~(1<<1);
    }
    if (mvld(q)) {
      q->nat=nat; q->nbnd=nbnd;
      Q->Test[nr]=(void *)q;
      Q->Type[nr]=TST_CONN;
      label_connser(q);
      prtGraphics("Fragment has %d atom%s and %d bond%s,\n",
	      q->nat,(q->nat==1?"":"s"),q->nbnd,(q->nbnd==1?"":"s"));
      prtGraphics("links %sallowed, cyclic routes %sallowed\n",
	      (q->noln?"dis":""),(q->nocr?"dis":""));
    }
    clear_node_chain(c); free_node_chain(c);
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }

  free_Token(key);
}
/*-------------------- make_connser_test --------------------*/

static connser_test *make_connser_test(void)
{
  connser_test *r;

  r=(connser_test *)malloc(sizeof(connser_test));
  if (mvld(r)) {
    r->nel=0; r->nat=0; r->nbnd=0;
    r->noln=FALSE; r->nocr=FALSE;
    r->npar=0;
  }
  return(r);
}
/*-------------------- write_test_connser --------------------*/

static void write_test_connser(FILE *out,int nr,connser_test *t)
{
  int i;
  char buff[4096];

  if (out) fprintf(out,"DEFINE T%d *Connser\n",nr);
  else prtGraphics("DEFINE T%d *Connser\n",nr);
  for (i=0;i<t->nel;i++) {
    if (out) fprintf(out,"  ELDEF  %s = %s\n",t->newel[i],t->eldef[i]);
    else prtGraphics("  ELDEF  %s = %s\n",t->newel[i],t->eldef[i]);
  }
  for (i=0;i<t->nat;i++) {
    sprintf(buff,"  AT%-3d  %s  %d  ",i+1,t->el[i],t->mca[i]);
    if (strcmp(t->nh[i],"any")) sprintf(buff+strlen(buff),"%s  ",t->nh[i]);
    if (t->chg[i]!=99) {
      if (strcmp(t->nh[i],"any")) sprintf(buff+strlen(buff),"%d ",t->chg[i]); 
      else sprintf(buff+strlen(buff)," 99  %d ",t->chg[i]); 
    }
    if (t->lig[i]!=99) sprintf(buff+strlen(buff)," T%d ",t->lig[i]);
    sprintf(buff+strlen(buff)," %c %c ",t->exct[i],t->cyca[i]);
    if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
  }
  for (i=0;i<t->nbnd;i++) {
    if (out) fprintf(out,"  BO  %-3d  %-3d  %-2d  %c\n",t->from[i],t->to[i],t->bt[i],t->cycb[i]);
    else prtGraphics("  BO  %-3d  %-3d  %-2d  %c\n",t->from[i],t->to[i],t->bt[i],t->cycb[i]);
  }
  if (t->noln) {if (out) fprintf(out,"  NOLN\n"); else prtGraphics("  NOLN\n");}
  if (t->nocr) {if (out) fprintf(out,"  NOCR\n"); else prtGraphics("  NOCR\n");}
  if (t->npar>0) {
    if (out) fprintf(out,"GEOM\n"); else prtGraphics("GEOM\n");
    for (i=0;i<t->npar;i++) {
      sprintf(buff,"DEFINE %s %2d %2d",t->pname[i],t->pnr[i][0]+1,t->pnr[i][1]+1);
      if (t->pnr[i][2]>-1) sprintf(buff+strlen(buff)," %2d",t->pnr[i][2]+1);
      if (t->pnr[i][3]>-1) sprintf(buff+strlen(buff)," %2d",t->pnr[i][3]+1);
      if (out) fprintf(out,"%s\n",buff); else prtGraphics("%s\n",buff);
    }
    if (out) fprintf(out,"NFRAG -99\nSYMCHK ON\nENANT NOIN\n");
    else {
      prtGraphics("NFRAG -99\n");
      prtGraphics("SYMCHK ON\n");
      prtGraphics("ENANT NOIN\n");
    }
  }
  if (out) fprintf(out,"END\n"); else prtGraphics("END\n");
}
/*-------------------- edit_connser --------------------*/

static void edit_connser(connser_test *q)
{
  Token *key;
  int ok;
  char *connser_edit_menu = "atom bond eldef parameter ' ' +NOLN -NOLN ' ' +NOCR -NOCR ' ' end";

  ok=sane_connser(q);
  if (ok<1) {
    err_set(1,"Connser fragment is empty. No edit.");
    return;
  }

  label_connser(q);
  prtGraphics("Fragment has %d atom%s and %d bond%s,\n",
	      q->nat,(q->nat==1?"":"s"),q->nbnd,(q->nbnd==1?"":"s"));
  prtGraphics("links %sallowed, cyclic routes %sallowed\n",
	      (q->noln?"dis":""),(q->nocr?"dis":""));

  key=(Token *)NULL;
  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("connser edit> ",";",connser_edit_menu);
    if (key->status==TOK_OK) {
      if (key_is("atom")) {
	edit_connser_atom(q);
      }else if (key_is("bond")) {
	edit_connser_bond(q);
      }else if (key_is("eldef")) {
	quest_eldef(q);
      }else if (key_is("parameter")) {
	quest_parameter(q);
      }else if (key_is("+NOLN")) {
	q->noln=TRUE;
      }else if (key_is("-NOLN")) {
	q->noln=FALSE;
      }else if (key_is("+NOCR")) {
	q->nocr=TRUE;
      }else if (key_is("-NOCR")) {
	q->nocr=FALSE;
      }else if (key_is("end")) {
	break;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad connser edit option: %s",key->text);
    }else break;
  }

  prtGraphics("Fragment has %d atom%s and %d bond%s,\n",
	      q->nat,(q->nat==1?"":"s"),q->nbnd,(q->nbnd==1?"":"s"));
  prtGraphics("links %sallowed, cyclic routes %sallowed\n",
	      (q->noln?"dis":""),(q->nocr?"dis":""));

  free_Token(key);
}
/*-------------------- edit_connser_atom --------------------*/

static void edit_connser_atom(connser_test *q)
{
  Token *key;
  Atom *a;
  Nodechain *c;
  char flg,menu[4096],elt[3],hc[20];
  int k,chg;

  key=(Token *)NULL;
  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("field: ",";","acyclic cyclic either ' ' charge element exact hydrogens ligance mca");
    if (key->status==TOK_OK) {
      if (key_is("charge")) {
	key=free_Token(key);
	key=Get_Kwd("value: ",";","+1 -1 +2 -2 +3 -3 positive negative charged neutral indifferent");
	if (key->status==TOK_OK) {
	  if (key_is("+1")) chg=1;
	  else if (key_is("-1")) chg=-1;
	  else if (key_is("+2")) chg=2;
	  else if (key_is("-2")) chg=-2;
	  else if (key_is("+3")) chg=3;
	  else if (key_is("-3")) chg=-3;
	  else if (key_is("positive")) chg=51;
	  else if (key_is("negative")) chg=49;
	  else if (key_is("charged")) chg=50;
	  else if (key_is("neutral")) chg=0;
	  else chg=99;
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	      if (k<q->nat) {
		q->chg[k]=chg;
		label_connser_atom(q,k);
	      }
	    }
	    clear_node_chain(c); free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad charge value: %s",key->text);
	}
      }else if (key_is("hydrogens")) {
	key=free_Token(key);
	key=Get_Kwd("h-count: ",";","0 1 2 3 4 '' 0,1 1,2 2,3 '' 0,1,2 1,2,3 '' any");
	if (key->status==TOK_OK) {
	  strcpy(hc,key->key);
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	      if (k<q->nat) {
		strcpy(q->nh[k],hc);
		label_connser_atom(q,k);
	      }
	    }
	    clear_node_chain(c); free_node_chain(c);
	  }
	}else if (key->status==TOK_NOK) err_set(1,"bad hydrogen count: %s\n",key->text);
      }else if (key_is("element")) {
	for(*menu='\0',k=0;k<q->nel;k++) sprintf(menu+strlen(menu)," %s",q->newel[k]);
	if (q->nel > 0) sprintf(menu+strlen(menu)," ' '");
	sprintf(menu+strlen(menu)," AA 1A 2A 3B 4B 5B 6B 7B 8X 8Y 8Z 8B 1B 2B 3A 4A");
	sprintf(menu+strlen(menu)," 5A 6A 7A 8A Ln An 1R 2R 3R Tr 1M 2M 3M 4M HD");
	key=Get_String_M("symbol: ",";",menu);
	if (key->status==TOK_OK) {
	  Strcpy(elt,key->text);
	  key=Get_Atoms_D("atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	      if (k<q->nat) {
		Strcpy(q->el[k],elt);
		label_connser_atom(q,k);
	      }
	    }
	    clear_node_chain(c); free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad element symbol: %s",key->text);
	}
      }else if (key_is("exact")) {
	key=free_Token(key);
	key=Get_Kwd("state: ",";","on off");
	if (key->status==TOK_OK) {
	  flg=(key_is("on") ? 'E' : ' ');
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	      if (k<q->nat) {
		q->exct[k]=flg;
		label_connser_atom(q,k);
	      }
	    }
	    clear_node_chain(c); free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad option: %s",key->text);
	}
      }else if (key_is("cyclic")) {
	key=free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	    for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	    if (k<q->nat) {
	      q->cyca[k]='C';
	      label_connser_atom(q,k);
	    }
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",key->text);
	}
      }else if (key_is("acyclic")) {
	key=free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	    for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	    if (k<q->nat) {
	      q->cyca[k]='A';
	      label_connser_atom(q,k);
	    }
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",key->text);
	}
      }else if (key_is("either")) {
	key=free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	    for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	    if (k<q->nat) {
	      q->cyca[k]=' ';
	      label_connser_atom(q,k);
	    }
	  }
	  clear_node_chain(c); free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",key->text);
	}
      }else if (key_is("ligance")) {
	key=free_Token(key);
	key=Get_Kwd("ligance: ",";","0 1 2 3 4 5 6 7 8 any");
	if (key->status==TOK_OK) {
	  chg=(key->n > 8 ? 99 : key->n);
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	      if (k<q->nat) {
		q->lig[k]=chg;
		label_connser_atom(q,k);
	      }
	    }
	    clear_node_chain(c); free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad ligance: %s",key->text);
	}
      }else if (key_is("mca")) {
	key=free_Token(key);
	key=Get_Int_M("minimum number of connected non-H atoms: ",";",
		    "0 1 2 3 4 5 6 7 8");
	if (key->status==TOK_OK) {
	  chg=(key->n < 0 ? 0 : key->n);
	  key=free_Token(key);
	  key=Get_Atoms_D("atoms: ",";");
	  if (key->status==TOK_OK) {
	    c=(Nodechain *)key->ptr;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      for (k=0;k<q->nat;k++) if (a==q->atm[k]) break;
	      if (k<q->nat) {
		q->mca[k]=chg;
		label_connser_atom(q,k);
	      }
	    }
	    clear_node_chain(c); free_node_chain(c);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad atom selection: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad integer: %s",key->text);
	}
      }
      break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad field name: %s",key->text);
    }else break;
  }
  free_Token(key);
}
/*-------------------- edit_connser_bond --------------------*/

static void edit_connser_bond(connser_test *q)
{
  Token *key;
  int k,bt=0,opt;
  char cyc='\0';
  Bond *bnd,*bn;

  key=(Token *)NULL;
  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("field: ",";",
		"acyclic cyclic either ' '"
		"any single double triple quadruple aromatic polymeric delocalized pi-bond");
    if (key->status==TOK_OK) {
      if (key_is("acyclic")) {
	opt=1; cyc='A';
      }else if (key_is("cyclic")) {
	opt=1; cyc='C';
      }else if (key_is("either")) {
	opt=1; cyc=' ';
      }else if (key_is("any")) {
	opt=2; bt=99;
      }else if (key_is("single")) {
	opt=2; bt=1;
      }else if (key_is("double")) {
	opt=2; bt=2;
      }else if (key_is("triple")) {
	opt=2; bt=3;
      }else if (key_is("quadruple")) {
	opt=2; bt=4;
      }else if (key_is("aromatic")) {
	opt=2; bt=5;
      }else if (key_is("polymeric")) {
	opt=2; bt=6;
      }else if (key_is("delocalized")) {
	opt=2; bt=7;
      }else if (key_is("pi-bind")) {
	opt=2; bt=9;
      }else opt=0;
      key=free_Token(key);
      key=Get_Bonds_D("bonds: ",";");
      if (key->status==TOK_OK) {
	for (bnd=(Bond *)key->ptr;mvld(bnd);bnd=bn) {
	  bn=bnd->chain; bnd->chain=(Bond *)NULL;
	  for (k=0;k<q->nbnd;k++) if (bnd==q->bnd[k]) break;
	  if (k<q->nbnd) {
	    switch (opt) {
	    case 1:
	      q->cycb[k]=cyc;
	      break;
	    case 2:
	      q->bt[k]=bt;
	      break;
	    }
	    label_connser_bond(q,k);
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad bond selection: %s",key->text);
      }
      break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad field name: %s",key->text);
      break;
    }else break;



  }
}
/*-------------------- label_connser --------------------*/

static void label_connser(connser_test *q)
{
  int i;

  if (q) {
    for (i=0;i<q->nat;i++)  label_connser_atom(q,i);
    for (i=0;i<q->nbnd;i++) label_connser_bond(q,i);
  }
}
/*-------------------- label_connser_atom --------------------*/

static void label_connser_atom(connser_test *q,int nr)
{
  Atom *a;
  char buf[4096];

  if (nr<0 || nr>=q->nat) return;
  a=q->atm[nr];
  if (cPtr(a->cross)) {
    sprintf(buf,"%s,%d",q->el[nr],q->mca[nr]);
    if (strcmp(q->nh[nr],"any")) sprintf(buf+strlen(buf),"{%s}",q->nh[nr]);
    if (q->chg[nr]!=99) {
      if (q->chg[nr]<49) sprintf(buf+strlen(buf),",%+d",q->chg[nr]);
      else if (q->chg[nr]==49) sprintf(buf+strlen(buf),",-");
      else if (q->chg[nr]==50) sprintf(buf+strlen(buf),",+-");
      else if (q->chg[nr]==51) sprintf(buf+strlen(buf),",+");
    }
    if (q->exct[nr]=='E') sprintf(buf+strlen(buf),",E");
    if (q->lig[nr]!=99)   sprintf(buf+strlen(buf),",T%d",q->lig[nr]);
    if (q->cyca[nr]=='A') sprintf(buf+strlen(buf),",A");
    if (q->cyca[nr]=='C') sprintf(buf+strlen(buf),",C");
    set_atom_label(a,ATM_LAB_TEXT,buf,TRUE);
  }
}
/*-------------------- label_connser_bond --------------------*/

static void label_connser_bond(connser_test *q,int nr)
{
  Bond *b;
  char buf[4096];
  char *csd_bt_lab[10] = {
    "","","2","3","4","ar","ply","dl","X","pi"
    };

  if (nr<0 || nr>=q->nbnd) return;
  b=q->bnd[nr];
  if (q->bt[nr]>=0 && q->bt[nr]<10) sprintf(buf,"%s",csd_bt_lab[q->bt[nr]]);
  else if (q->bt[nr]==99) sprintf(buf,"any");
  if (q->cycb[nr]=='A'||q->cycb[nr]=='C') {
    if (*buf) sprintf(buf+strlen(buf),",");
    sprintf(buf+strlen(buf),"%c",q->cycb[nr]);
  }
  set_bond_label(b,BND_LAB_TEXT,buf);
}

/*-------------------- csd_bt --------------------*/

static int csd_bt(int bt)
{
  int r;

  switch (bt) {
  case BT_SINGLE:
  case BT_AMIDE:
    r=CSD_BT_SINGLE;
    break;
  case BT_DOUBLE:
    r=CSD_BT_DOUBLE;
    break;
  case BT_TRIPLE:
    r=CSD_BT_TRIPLE;
    break;
  case BT_AROMATIC:
    r=CSD_BT_AROMATIC;
    break;
  default:
    r=CSD_BT_ANY;
    break;
  }
  return(r);
}
/*-------------------- sane_connser --------------------*/

static int sane_connser(connser_test *q)
{
  int i,j,k,nat,nbnd,ok[MAX_ATOMS],inc,npar;
  Atom *a;

  nat=q->nat;
  for (j=0;j<nat;j++) ok[j]=FALSE;
  ForAtoms(a,MolRoot) {
    if (nat==0) break;
    for (j=0;j<q->nat;j++) if (a==q->atm[j]) break;
    if (j<q->nat) {
      ok[j]=TRUE;
      nat--;
    }
  }

  for (nat=0,j=0;j<q->nat;j++) {
    if (ok[j]) q->atm[j]->seqnr=nat++;
  }

  if (nat>0) {
    for (nbnd=0,j=0;j<q->nbnd;j++) {
      i=q->from[j]-1; k=q->to[j]-1;
      if (i>=0&&i<q->nat&&k>=0&&k<q->nat&&ok[i]&&ok[k]) {
	if (q->bnd[j]==find_bond(q->atm[i],q->atm[k])) inc=1;
	else inc=0;
      }else inc=0;
      if (inc>0) {
	if (nbnd<j) {
	  q->bnd[nbnd]=q->bnd[j];
	  q->from[nbnd]=q->atm[q->from[j]-1]->seqnr+1;
	  q->to[nbnd]=q->atm[q->to[j]-1]->seqnr+1;
	  q->bt[nbnd]=q->bt[j];
	  q->cycb[nbnd]=q->cycb[j];
	}
	nbnd++;
      }
    }
    
    for (npar=0,j=0;j<q->npar;j++) {
      for (k=0;k<4;k++) {
	if (q->pnr[j][k]>-1) {
	  if (ok[q->pnr[j][k]]) q->pnr[j][k]=q->atm[q->pnr[j][k]]->seqnr;
	  else break;
	}
      }
      if (k==4) {
	if (npar<j) {
	  for (k=0;k<4;k++) q->pnr[npar][k]=q->pnr[j][k];
	  q->pname[npar]=q->pname[j];
	  q->pname[j]=(char *)NULL;
	}
	npar++;
      }else{
	free(q->pname[j]);
	q->pname[j]=(char *)NULL;
      }
    }

    for (nat=0,j=0;j<q->nat;j++) {
      if (ok[j]) {
	if (nat<j) {
	  q->atm[nat]=q->atm[j];
	  q->mca[nat]=q->mca[j];
	  strcpy(q->nh[nat],q->nh[j]);
	  q->chg[nat]=q->chg[j];
	  q->lig[nat]=q->lig[j];
	  q->exct[nat]=q->exct[j];
	  q->cyca[nat]=q->cyca[j];
	  strcpy(q->el[nat],q->el[j]);
	}
	nat++;
      }
    }

  }else{
    nbnd=0;
    npar=0;
  }

  q->nat=nat; q->nbnd=nbnd; q->npar=npar;

  return(q->nat);
}
/*-------------------- quest_eldef --------------------*/

static void quest_eldef(connser_test *q)
{
  char newel[3];
  Token *key;
  int i;

  key=(Token *)NULL;

  key=Get_String("new symbol: ",";");
  if (key->status==TOK_OK) {
    Strcpy(newel,key->text);
    for (i=0;i<q->nel;i++) if (strcasecmp(q->newel[i],newel)==0) break;
    if (i<MAX_ELDEFS) {
      if (i<q->nel) prtGraphics("element symbol %s exists,",q->newel[i]);
      key=free_Token(key);
      key=Get_String("new definition: ",";");
      if (key->status==TOK_OK) {
	if (i<q->nel) free(q->eldef[i]);
	else Strcpy(q->newel[i],newel);
	q->eldef[i]=key->text;
	key->text=(char *)NULL;
	if (i==q->nel) q->nel++;
      }else{
	if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
      }
    }else{
      err_set(1,"sorry, element definition table is full...");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s.",key->text);
  }
  free_Token(key);
}
/*-------------------- list_screens --------------------*/

static void list_screens(Quest *Q)
{
  int i,k;
  char buff[4096];

  for (i=0;i<MAX_SCREENS;i++) {
    if (Q->Screen[i]!=0) {
      for (k=0;k<MAX_KNOWN_SCREENS;k++) if (screen[k].nr==i||screen[k].nr==-i) break;
      if (k<MAX_KNOWN_SCREENS) {
	sprintf(buff,"%s ",Q->Screen[i]*screen[k].nr<0?"-":"+");
	if (screen[k].txt) sprintf(buff+strlen(buff),"%s",screen[k].txt);
	else sprintf(buff+strlen(buff),"%s",screen[k].mnem);
      }else{
	sprintf(buff,"%s ",Q->Screen[i]<0?"-":"+");
	sprintf(buff+strlen(buff),"screen %d\n",i);
      }
      prtGraphics("%s\n",buff);
    }
  }
}
