#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

extern int graphix;

#define key_is(a) (streq(a,key->key))

static char *rad_keys = "covalent vdW biso scale "
  "0.05 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 2.0 3.0 4.0 5.0";
static char *scale_M = "0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 1.5 2.0";

#define RADIUS 0
#define RCOV 1
#define RVDW 2
#define BISO 3
#define SCALE 4

#define TRUE 1
#define FALSE 0

void cmd_radii(void)
{
  int nchgd,size,rtype;
  float r=0;
  Atom *a;
  Xatom *xa;
  Nodechain *c;
  Token *key;
  
  rtype=RADIUS;
  key=Get_Float_M("atom radius: ",";",rad_keys);
  if (key->status == TOK_NOK) {
    key=Prs_Kwd(key,rad_keys);
    if (key->status==TOK_OK) {
      rtype=key->n+1;
      if (key_is("scale")) {
	free_Token(key);
	key=Get_Float_M("scale: ",";",scale_M);
	if (key->status==TOK_OK) {
	  r=(key->f >0 ? key->f : 0);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad option %s",key->text);
	}
      }
    }
  }else if (key->status==TOK_OK) {
    r=(key->f >0 ? key->f : 0);
    rtype=RADIUS;
  }

  if (key->status==TOK_OK) {
    free_Token(key);
    key=Get_Atoms_D("Atoms: ",";");
    if (key->status == TOK_OK) {
      c=(Nodechain *)key->ptr;
      for (nchgd=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	switch (rtype) {
	case RCOV:
	  a->r=(mvld(a->atp) ? a->atp->rcov : 0.001);
	  break;

	case RVDW:
	  a->r=(mvld(a->atp) ? a->atp->rvdw : 0.001);
	  break;

	case BISO:
	  r=0.05;
	  if (a->xtal) {
	    xa=first_xatom(a);
	    if (mvld(xa)) {
	      if (xa->iso && xa->bfac[0] > 0) r=sqrt(xa->bfac[0]);
	      else {
		r=xa->bfac[0]*xa->bfac[1]*xa->bfac[2]
		  +2*xa->bfac[3]*xa->bfac[4]*xa->bfac[5]
		  -xa->bfac[3]*xa->bfac[3]*xa->bfac[0]
		  -xa->bfac[4]*xa->bfac[4]*xa->bfac[1]
		  -xa->bfac[5]*xa->bfac[5]*xa->bfac[2];
		r=(r > 0 ? pow(r,1.0/6.0) : 0.01);
	      }
	    }
	  }else{
	    r=(a->x[ATM_BISO]<0?0.01:a->x[ATM_BISO]);
	    r=sqrt(r/8)/M_PI;
	  }
	  a->r=r;
	  break;
	  
	case RADIUS:
	  a->r=r;
	  break;

	case SCALE:
	  a->r*=r;
	  break;
	}
	if (cPtr(a->cross)) nchgd++;
      }
      size=nchgd*(cPtrSize()+sizeof(int));
      send_bheader(graphix,"ASTR",size);
      for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
	if (cPtr(a->cross)) {
	  send_bcptr(graphix,a->cross);
	  send_bfloat(graphix,a->r);
	}
      }
      trflush(graphix);
      clear_node_chain(key->ptr);
      free_node_chain(key->ptr);
    }
  }else if (key->status == TOK_NOK) {
    err_set(1,"bad option: %s",key->text);
  }
  send_header(graphix,"ARMK",0);
  free_Token(key);
}
