#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <math.h>

#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

extern Node *MolRoot;
extern int graphix;

#define TRUE 1
#define FALSE 0

#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL


/*------------------------------------------------------------------------*/

void cmd_read(void *vroot,int types)
{
  Token *key;
  char *p,*menu,*file,*srch,*prefix,*flist,*wild;
  char twd[MAXPATHLEN];
  char ffn[MAXPATHLEN];
  int loop,is_wild,status,wait_flag,ft,have_file,have_dir,save_char;
  int do_dir,one_file_only,nread,i,k,km;
  struct stat sbuff;
  Node *root;
  Parse P;
  
  root=(Node *)vroot;
  do_dir=FALSE;
  file=(char *)NULL;
  wild=(char *)NULL;
  flist=(char *)NULL;
  prefix="-searchmask -directories ' '";
  srch=(char *)NULL;
  if (!getCWD(twd,sizeof(twd))) strcpy(twd,".");
  if (twd[strlen(twd)-1]!='/') strcat(twd,"/");

  key=(Token *)NULL;
  menu=(char *)NULL;
  ft=FT_NONE;
  one_file_only=FALSE;

  for (k=0,km=0,i=0;i<MAX_FT;i++) {
    if (types&FT_CAN_READ&(1<<i)) {
      k++;
      if (FT_MOL_TYPES&(1<<i)) km++;
    }
  }
  if (k==1) {
    ft=types;
  }else{
    prefix=(km>0 ? "coordinates ' ' " : "");
    menu=Make_FileTypeList(prefix,types&FT_CAN_READ);
    key=Get_Kwd("Filetype: ",";",menu);
    if (key->status==TOK_OK) {
      if (key_is("coordinates")) {
	cmd_read_coordinates((Nodechain *)NULL);
	ft=FT_NONE;
      }else{
	ft=ft_number(key->key);
	if (ft==FT_NONE) key->status=TOK_NOK;
      }
    }
  }

  if (ft!=FT_NONE) {
    prefix="-searchmask -directories ' '";
    free_menu(menu); 
    menu=Strdup(prefix);
    loop=TRUE;
    if (ft==FT_MACCS || ft==FT_CSD) one_file_only=TRUE;
    if (ft==FT_WLB) {
      if (mvld(root)) if (!is_lib(root)) {
	err_set(1,"Can't read library into %s.",root->name);
	loop=FALSE;
      }
    }else if (ft==FT_CNF) {     /* Conformational Analysis File */
      if (!mvld(root)) {
	err_set(1,"bad root in read confa.");
	loop=FALSE;
      }
      one_file_only=TRUE;
    }else if (ft==FT_CNT) {     /* Contour File */
      one_file_only=TRUE;

    }else if (ft==FT_NOE) {     /* XPLOR NOE-list */
      if (!is_mol(root)) {
	err_set(1,"Can't read upper limits file into %s.",MolRoot->name);
	loop=FALSE;
      }else if (((Mol *)root)->moltype!=MOL_NMRSET) {
	err_set(1,"%s is not an NMR set.",root->name);
	loop=FALSE;
      }
    }else if (ft==FT_DUPPER) {  /* DIANA upper limits */
      if (!is_mol(root)) {
	err_set(1,"Can't read upper limits file into %s.",MolRoot->name);
	loop=FALSE;
      }else if (((Mol *)root)->moltype!=MOL_NMRSET) {
	err_set(1,"Can't read upper limits file into %s.",root->name);
	loop=FALSE;
      }
    }else if (ft==FT_SESS) {   /* session file */
      one_file_only=TRUE;

    }else if (ft&FT_MOL_TYPES) {
      if (mvld(root)) if(!is_mol(root)) {
	err_set(1,"Can't read molecule into %s.",root->name);
	loop=FALSE;
      }
    }else if (ft==FT_DOTS) {
       one_file_only=TRUE;
       
    }else{
      err_set(1,"cmd_read: internal error, filetype (%d) is fugu.",ft);
      loop=FALSE;
    }
    have_file=FALSE;
    have_dir=FALSE;
    srch=Strdup(ft_searchmask(ft));
    wait_flag=TRUE;
    while(loop) {
      err_dump();
      key=free_Token(key);
      key=Get_String_MW("File: ",";",menu,wait_flag);
      if (key->status==TOK_OK) {
	key=Prs_Kwd(key,"-searchmask -directories");
	if (key_is("-searchmask")) {
	  key=free_Token(key);
	  default_token(ft_searchmask(ft));
	  key=Get_String_M("searchmask: ",";","* */");
	  if (key->status==TOK_OK) {
	    if (srch) free(srch);
	    srch=key->text;
	    key->text=(char *)NULL;
	    free_menu(menu);
	    menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
	    have_dir=TRUE;
	    wait_flag=TRUE;
	    continue;
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"read: bad searchmask token (%s).",key->text);
	    break;
	  }else break;
	}else if (key_is("-directories")) {
	  do_dir=!do_dir;
	  free_menu(menu);
	  menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
	  have_dir=TRUE;
	  wait_flag=TRUE;
	  continue;
	}else{
	  if (*key->text=='|') {
	    strcpy(ffn,key->text);
	    have_file=TRUE;
	    is_wild=FALSE;
	  }else{
	    p=strrchr(key->text,'/'); if (!p) p=key->text;
	    is_wild=strchr(p,'*')||strchr(p,'?');
	    strcpy(ffn,twd);newtwd(ffn,key->text);
	    if (is_wild) {
	      have_file=TRUE;
	    }else{
	      status=stat(ffn,&sbuff);
	      if (status<0) {
		have_file=TRUE;
	      }else{
		if (sbuff.st_mode&S_IFDIR) {
		  strcpy(twd,ffn);
		  free_menu(menu);
		  menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
		  have_dir=TRUE;
		}else{
		  have_file=TRUE;
		}
	      }
	    }
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"read : bad filename token: %s",key->text);
	break;
      }else break;
   
      if (have_file) {
	wait_flag=FALSE;
	if (strlen(ffn)>0&&ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	if (is_wild) {
	  file=strrchr(ffn,'/');*file++='\0'; 
	  P.pos=0; P.sep=""; P.tok=file; 
	  p=twd+strlen(twd)-1; save_char=*p; if (*p=='/') *p='\0';
	  if (!streq(twd,ffn)||!have_dir) {
	    if (flist) free(flist);
	    flist=Make_FileNameList("",ffn,"*");
	    P.buf=flist;
	  }else{
	    P.buf=menu+strlen(prefix);
	  } 
	  P.mts=sizeof(ffn)-strlen(ffn)-1;
	  *p=save_char;
	  if ( ft != FT_DOTS) lock_evtq();
	  *(file-1)='/';
	  nread=0;
	  if (wild) free(wild);
	  wild=Strdup(file);
	  while(loop&&get_token(&P)) {
	    if (wild_match(file,wild)) {
	      if (!isDirectory(ffn)) {
		ReadNode(vroot,ffn,ft);
		err_dump();
		nread++;
		if (one_file_only) loop=FALSE;
	      }
	    }
	  }
	  if (wild) free(wild);
	  wild=(char *)NULL;
	  if (flist) free(flist);
	  flist=(char *)NULL;
	  if (!one_file_only) prtGraphics("%d file(s) read.\n",nread);
	  if ( ft != FT_DOTS) unlock_evtq();
	}else{
	  if (!isDirectory(ffn)) {
	    ReadNode(vroot,ffn,ft);
	    err_dump();
	  }
	}
	if (one_file_only) loop=FALSE;
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"read: unrecognizable filetype (%s)",key->text);
  }

  if (srch)  free(srch);
  if (menu)  free(menu);
  if (wild)  free(wild);
  if (flist) free(flist);

  free_Token(key);
}
/*-------------------- cmd_read_coordinates --------------------*/

int cmd_read_coordinates(Nodechain *template)
{
  char *menu;
  int i,j,NTemplate,ft,nat,ownt,status;
  Node *mol;
  Token *key;
  UCpipe *C;
  Atom **Template,*a;
  Nodechain *c;

  status=0;
  c=template;
  ownt=(c==(Nodechain *)NULL);
  menu=(char *)NULL;

  menu=Make_FileTypeList("",FT_MOL_TYPES&FT_CAN_READ);
  ft=FT_NONE;
  key=Get_Kwd("Filetype: ",";",menu);
  if (key->status==TOK_OK) {
    ft=ft_number(key->key);
    if (ft==FT_NONE) key->status=TOK_NOK;
  }else if (key->status==TOK_NOK) {
    err_set(1,"unknown filetype: %s",key->text);
  }

  if (ft!=FT_NONE) {
    key=free_Token(key);
    key=Get_File((char *)NULL,ft_defext(ft),"file containing coordinates: ",";");
    if (key->status==TOK_OK) {
      C=(UCpipe *)(key->ptr);
      mol=readmoltypes(C,(char *)NULL,ft);
      if (is_mol(mol)) {
	nat=0; ForAtoms(a,mol) nat++;
	if (ownt) {
	  key=free_Token(key);
	  key=Get_Atoms_D("template atoms: ",";");
	  send_bheader(graphix,"ARMK",0);
	  if (key->status==TOK_OK) c=(Nodechain *)key->ptr;
	  else if (key->status==TOK_NOK) err_set(1,"bad template atom selection: %s",key->text);
	}
	if (c) {
	  NTemplate=c->n;
	  if (NTemplate==nat) {
	    Template=(Atom **)malloc(NTemplate*sizeof(Atom *));
	    if (mvld(Template)) {
	      for (i=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain,i++) Template[i]=a;
	      walk_order(MolRoot,NTemplate,Template,(1<<1));
	      i=0;
	      ForAtoms(a,mol) {
		for(j=0;j<3;j++) Template[i]->x[j]=a->x[j];
		i++;
	      }
	      send_newx(c);
	      prtGraphics("coordinates updated for %d atoms\n",nat);
	      status=1;
	    }else err_set(1,"failed to allocate memory for template atoms.");
	    free(Template);
	  }else{
	    err_set(1,"template mismatch: %d (template), %d (file)",NTemplate,nat);
	  }
	}
	if (ownt) {
	  clear_node_chain(c);
	  free_node_chain(c);
	}
	del_node(mol);
      }else err_set(1,"could not read coordinates from %s",C->filename);
      UCpclose(C);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad file specification: %s\n",key->text);
    }
  }

  free_Token(key);
  free_menu(menu);
  return(status);
}
