#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "xtal.h"
#include "gcalc.h"
#include "errhand.h"

extern int graphix;
extern Mol *MolRoot;
extern Dspobj *DspRoot;

#define key_is(x) (streq(x,key->key))

static void cmd_render_cpk(void);
static void cmd_render_ribbon(void);

/*========================================================================*/

void cmd_render(void)
{
  Token *key;

  key=(Token *)NULL;
  for (;;) {
    err_dump(); key=free_Token(key);
    key=Get_Kwd("render option: ",";","CPK Ribbon ' ' End");
    if (key->status==TOK_OK) {
      if (key_is("CPK")) cmd_render_cpk();
      else if (key_is("ribbon")) cmd_render_ribbon();
      else if key_is("End") break;
    }else if (key->status==TOK_NOK) {
      err_set(1,"unknown render option: %d",key->text);
    }
    if (end_parse()!=TOK_OK) {
      err_set(0,"extra characters at end of command ignored.");
    }
  }
  free_Token(key);
}
/*-------------------- cmd_render_cpk --------------------*/

static void cmd_render_cpk(void)
{
  /* Interface to "render" of the Raster3D suite of programs */

  Token *key;
  FILE *out;
  char *filename;
  PrjTS prj;
  Dspobj *dsp;
  mat_4 mtrx;
  vec_3 x;
  Atom *a;
  int i,cidx,nx,ny;
  float r,c[3],sizex,sizey,scale,ztra;

  filename=(char *)NULL;
  default_token("cpk.inp");
  key=Get_String("filename: ",";");
  if (key->status==TOK_OK) {
    out=fopen(key->text,"w");
    filename=key->text;
    key->text=(char *)NULL;
    if (out) {
      prj=getThreeDPrj();
      sizey=prj.ysize;
      sizex=prj.aspect*sizey;
      scale=(sizex<sizey ? sizex : sizey);
      if (prj.aspect > 1.25) {
	nx=160;
	ny=nx/prj.aspect;
      }else{
	ny=128;
	nx=prj.aspect*ny;
      }
      ztra=0.0; /* prj.znear-prj.slabwidth;*/
      get_MVT(DspRoot);
      fprintf(out,"Render input file generated by Wit!P\n");
      fprintf(out,"%3d %3d    tiles in x,y\n",nx,ny);
      fprintf(out,"12 12      pixels (x,y) per tile\n");
      fprintf(out,"3          3x3 pixels -> 2x2 pixels\n");
      for (i=0;i<3;i++) {
	c[i]=ColVal(0,i)/255.0;
	c[i]=c[i]*c[i];
      }
      fprintf(out,"%6.4f %6.4f %6.4f  background color\n",c[0],c[1],c[2]);
      fprintf(out,"T          yes, I LIKE shadows!\n");
      fprintf(out,"25         Phong power\n");
      fprintf(out,"0.15       secondary light contribution\n");
      fprintf(out,"0.05       ambient light contribution\n");
      fprintf(out,"0.25       specular reflection component\n");
      fprintf(out,"%5.1f   eye position\n",prj.viewdist);
      fprintf(out,"1 1 1      main light source position\n");
      fprintf(out,"1 0 0 0    input co-ordinate + radius transformation\n");
      fprintf(out,"0 1 0 0\n");
      fprintf(out,"0 0 1 0\n");
      fprintf(out,"  0.00  0.00  %9.3f  %9.3f\n",ztra,scale);
      fprintf(out,"3          mixed object types\n");
      fprintf(out,"*\n*\n*\n");
      ForAtoms(a,MolRoot) {
	if (a->dsp) {
	  dsp=glblobj(assocobj(a->up));
	  mtrx=(mvld(dsp) ? dsp->mtrx : ident_4());
	  x=m4xv3(mtrx,vec3(a->x));
	  r=(mvld(a->atp) ? a->atp->rvdw : 0.01);
	  cidx=(a->color < 0 ? -a->color : a->color);
	  for (i=0;i<3;i++) {
	    c[i]=ColVal(cidx,i)/255.0;
	    c[i]=c[i]*c[i];
	  }
	  fprintf(out,
		  "2\n"
		  "%9.4f %9.4f %9.4f %7.4f %6.4f %6.4f %6.4f\n",
		  x.c[0],x.c[1],x.c[2],r,c[0],c[1],c[2]); 
	}
      }
      fclose(out);
      prtGraphics("render input written to %s\n",filename);
    }else{
      err_set(1,"failed to open %s for write.",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad filename token: %s",key->text);

  free_Token(key);
  if (filename) free(filename);
}
/*-------------------- cmd_render_ribbon --------------------*/

static void cmd_render_ribbon(void)
{
  /* Interface to "render" of the Raster3D suite of programs */

  err_set(1,"sorry, option is not implemented.");
}
