#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include <sys/param.h>
#include "errhand.h"
#include "parmset.h"
#include "structal.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

int chdir(const char *path);
int use_readline(int s);

extern Node *DefaultNode;
extern Dspobj *DspRoot;
extern int graphix;
extern Node *LibRoot;

#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define TRUE 1
#define FALSE 0

#define MXRAMP 500
#define COL_RGB -1

static char *col_pmt[3] = {"R: ","G: ","B : "};
static char *col_cval = "'0' '15' '31' '47' '63' '79' '95' '111' '127' "
                       "'143' '159' '175' '191' '207' '225' '255'"; 
static char *set_kwd = "amber antialiasing autocenter autobond altobjkey background biosym bond color "
                       "cwd cylinder default depthcue display dither env fog gnuread linedef "
                       "menu mouse mpeoe pivot radii ramp RGB rock sphere stereo structal TAFF text view";

static char *TAFF_wmenu = "0.0 0.5 1.0 2.0 3.0 4.0 5.0 6.0 7.0 8.0 9.0 10.0";

/* note: cyl_mode and sph_mode are correlated in server */
static int sph_poly=9,sph_mode=0,cyl_poly=6,cyl_mode=-1;

extern int autocenter_on_read;
extern int autobond_on_read;

extern int Setenv(char *name,char *value);

static void set_gap_penalty(void);
static void set_gerstXcompare5Factor(void);
static void setResCompare(void);

void cmd_set(void)
{
  char path[MAXPATHLEN],buffer[4096],*menu;
  Node *PruleLib=NULL;
  Dspobj *dobj;
  Nodechain *c;
  Atom *atm;
  Bond *b,*bn,*bs,*be;
  float lwmap[4],pt[3],rock_amplitude,rock_step;
  int cldf[3],lw,i,j,n,nat,nr,mode,function,ramp[MXRAMP];
  int st_mode[STEREO_MAX_MODES]={0,4,5,1,2,3,6};
  Token *key;
  mat_4 mo,ma;
  vec_3 pv;
  Node *ruleset;
  
  key=(Token *)NULL;
  menu=(char *)NULL;

  key=Get_Kwd("what: ",";",set_kwd);
  if (key->status == TOK_OK) {
    if (key_is("amber")) {
      key=free_Token(key);
      key=Get_Kwd("mode: ",";","missing sloppy");
      if (key->status==TOK_OK) {
	if (key_is("sloppy")) {
	  mode=checkParmSetMode(PRM_SLOPPY_TEST);
	  default_token("%s",mode==PRM_SLOPPY_ALWAYS?"always":mode==PRM_SLOPPY_NEVER?"never":"once");
	  key=Get_Kwd("write sloppy parmtop files: ",";","always never once");
	  if (key->status==TOK_OK) {
	    checkParmSetMode(key->n==0?PRM_SLOPPY_ALWAYS:key->n==1?PRM_SLOPPY_NEVER:PRM_SLOPPY_ONCE);
	  }else if (key->status==TOK_NOK) err_set(1,"invalid key: %s",key->text);
	}else if (key_is("missing")) {
	  mode=missingParmSetMode(PRM_MISSING_TEST);
	  default_token("%s",mode==PRM_MISSING_FIXED?"fixed":mode==PRM_MISSING_ZERO?"zero":"floppy");
	  key=Get_Kwd("treatment of missing parameters in sloppy files: ",";","fixed floppy zero");
	  if (key->status==TOK_OK) {
	    missingParmSetMode(key->n==0?PRM_MISSING_FIXED:key->n==1?PRM_MISSING_FLOPPY:PRM_MISSING_ZERO);
	  }else if (key->status==TOK_NOK) err_set(1,"invalid key: %s",key->text);
	}
      }else if (key->status==TOK_NOK) err_set(1,"invalid amber mode: %s",key->text);
    }else if (key_is("altobjkey")) {
      key=free_Token(key);
      key=Get_Kwd("key: ",";","altkey shiftkey");
      if (key->status==TOK_OK) {
	set_altobj_key(key->key);
      }else{
	err_set(1,"invalid key: %s",key->text);
      }
    }else if (key_is("autocenter")) {
      key=free_Token(key);
      key=Get_Kwd("state: ",";","on off");
      if (key->status==TOK_OK) {
	autocenter_on_read=key_is("on");
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }
    }else if (key_is("autobond")) {
      key=free_Token(key);
      key=Get_Kwd("autobond option> ",";","mode on off");
      if (key->status==TOK_OK) {
	if (key_is("mode")) {
	  key=free_Token(key);
	  if (autobond_mode()==AUTOBOND_ALL) default_token("all");
	  else if (autobond_mode()==AUTOBOND_RESID) default_token("residue");
	  else if (autobond_mode()==AUTOBOND_CHAIN) default_token("chain");
	  else if (autobond_mode()==AUTOBOND_OBJECT) default_token("object");
	  key=Get_Kwd("automatic bond mode: ",";","all chain object residue");
	  if (key->status==TOK_OK) {
	    if (key_is("all")) set_autobond_mode(AUTOBOND_ALL);
	    else if (key_is("chain")) set_autobond_mode(AUTOBOND_CHAIN);
	    else if (key_is("residue")) set_autobond_mode(AUTOBOND_RESID); 
	    else if (key_is("object")) set_autobond_mode(AUTOBOND_OBJECT); 
	  }else if (key->status==TOK_NOK) err_set(1,"bad autobond mode: %s\n",key->text);
	}else autobond_on_read=key_is("on");
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }
    }else if (key_is("biosym")) {
      key=free_Token(key);
      sprintf(buffer,"1 2 3");
      PruleLib=find_name(LibRoot,"PruleLib",OBJ_LIB);
      if (mvld(PruleLib)) {
	for (i=0,ruleset=PruleLib->fst;mvld(ruleset);ruleset=ruleset->nxt) {
	  if (is_lib(ruleset)) {
	    if (mvld(ruleset->name)) {
	      sprintf(buffer+strlen(buffer)," %s",ruleset->name);
	      i++;
	    }
	  }
	}
      }
      key=Get_Kwd("car/mdf file version: ",";",buffer);
      if (key->status==TOK_OK) {
	set_biosym_version(key->key);
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown biosym car/mdf file version: %s",key->text);
      }
    }else if (key_is("bond")) {
      key=free_Token(key);
      key=Get_Int_M("linewidth: ",";","1 2 3 4 5 6 7 8 9 10");
      if (key->status==TOK_OK) {
	lw=clamp(key->n,1,100);
	free_Token(key);
	key=Get_Bonds_D("bonds: ",";");
	if (key->status==TOK_OK) {
	  if (key->n > 0) {
	    n=0; bs=(Bond *)NULL; be=(Bond *)NULL;
	    for (b=(Bond *)key->ptr;mvld(b);b=bn) {
	      b->lw=lw;
	      bn=b->chain;
	      if (cPtr(b->cross)) {
		if (mvld(bs)) be->chain=b;
		else bs=b;
		be=b;
		n++;
	      }
	      b->chain=(Bond *)NULL;
	    }
	    if (n>0) {
	      send_bheader(graphix,"SBLW",sizeof(int)+n*cPtrSize());
	      send_bint(graphix,lw);
	      for (b=bs;mvld(b);b=bn) {
		bn=b->chain;
		send_bcptr(graphix,b->cross);
		b->chain=(Bond *)NULL;
	      }
	    }
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad token %s\n",key->text);
	}
	send_header(graphix,"ARMK",0);
      }else if (key->status==TOK_NOK) err_set(1,"bad bond width: %s\n",key->text);
    }else if (key_is("cwd")) {
      free_Token(key);
      key=Get_Directory("directory: ",";");
      if (key->status==TOK_OK) {
	if (chdir(key->text)) {
	  err_set(1,"failed to change dircetory.");
	}
	if (!getCWD(path,sizeof(path))) {
	  err_set(1,"failed to determine current directory.");
	}else{
	  prtGraphics("current directory: %s\n",path);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }
    }else if (key_is("env")) {
      cmd_setenv();
    }else if (key_is("cylinder")) {
      free_Token(key);
      key=Get_Kwd("mode: ",";","half full quality");
      if (key->status == TOK_OK) {
	mode = key->n;
	if (key_is("quality")) {
	  free_Token(key);
	  if (cyl_poly>0) default_token("%d",cyl_poly);
	  key=Get_Int_M("n-poly: ",";","2 3 4 5 6 7 8 9 10 11 12 13 14 15");
	  if (key->status == TOK_OK) cyl_poly=key->n;
	}else{
	  sph_mode=mode;  /* sph_mode and cyl_mode are correlated */
	}
	if (key->status == TOK_OK) {
	  send_bheader(graphix,"SCYM",2*sizeof(int));
	  send_bint(graphix,mode);
	  send_bint(graphix,cyl_poly);
	}
      }else if (key->status == TOK_NOK) {
	err_set(1,"unkown option: %s",key->text);
      }
    }else if (key_is("sphere")) {
      free_Token(key);
      key=Get_Kwd("mode: ",";","half full quality");
      if (key->status == TOK_OK) {
	mode = key->n;
	if (key_is("quality")) {
	  free_Token(key);
	  if (sph_poly>0) default_token("%d",sph_poly);
	  key=Get_Int_M("tesselation depth: ",";","2 3 4 5 6 7 8 9 10 11 12 13 14 15");
	  if (key->status == TOK_OK) sph_poly=key->n;
	}else{
	  sph_mode=mode;
	}
	if (key->status == TOK_OK) {
	  send_bheader(graphix,"SPHM",2*sizeof(int));
	  send_bint(graphix,mode);
	  send_int(graphix,sph_poly);
	}
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown option: %s",key->text);
      }
    }else if (key_is("gnuread")) {
      free_Token(key);
      key=Get_Kwd("on/off ? ",";","on off");
      if (key->status==TOK_OK) {
	use_readline(1-key->n);
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad gnuread state: %s\n",key->text);
      }
    }else if (key_is("linedef")) {
      key=free_Token(key);
      key=Get_Float_M("minimum linewidth: ",";","1.00 1.25 1.50 1.75 2.0 2.25 2.5 2.75 3.0");
      if (key->status==TOK_OK) lwmap[0]=key->f;
      else if (key->status==TOK_NOK) err_set(1,"bad linewidth value: %s",key->text);
      if (key->status==TOK_OK) {
	key=free_Token(key);
	key=Get_Float_M("granularity : ",";","0.125 0.250 0.375 0.500 0.625 0.750 0.875 1.000 1.500 2.000");
	if (key->status==TOK_OK) lwmap[2]=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad linewidth value: %s",key->text);
      }
      if (key->status==TOK_OK) {
	key=free_Token(key);
	key=Get_Float_M("maximum linewidth: ",";","1.0 2.0 3.0 4.0 5.0 6.0 7.0 8.0 9.0 10.0");
	if (key->status==TOK_OK) lwmap[1]=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad linewidth value: %s",key->text);
      }
      if (key->status==TOK_OK) {
	key=free_Token(key);
	key=Get_Float_M("maximum width of antialiased line: ",";","1.0 2.0 3.0 4.0 5.0 6.0 7.0 8.0 9.0 10.0");
	if (key->status==TOK_OK) lwmap[3]=key->f;
	else if (key->status==TOK_NOK) err_set(1,"bad linewidth value: %s",key->text);
      }
      if (key->status==TOK_OK) {
	setAttributeFA("SLWM",lwmap);
      }
    }else if (key_is("menu")) {
      free_Token(key);
      key=Get_Kwd("mode: ",";","on off");
      if (key->status == TOK_OK) {
	mode=key->n;
	send_bheader(graphix,"SMEN",sizeof(int));
	send_bint(graphix,mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown menu mode: %s",key->text);
      }
    }else if (key_is("mouse")) {
      free_Token(key);
      sprintf(buffer,"2Buttons 3Buttons left right ' ' %s",mouse_func_mnems());
      sprintf(buffer+strlen(buffer),"' ' show_map");
      key=Get_Kwd("function: ",";",buffer);
      if (key->status == TOK_OK) {
	if (key_is("left")) {
	  set_mouse("left");
	}else if (key_is("right")) {
	  set_mouse("right");
	}else if (key_is("3Buttons")) {
	  set_mouse("3Buttons");
	}else if (key_is("2Buttons")) {
	  set_mouse("2Buttons");
	}else if (key_is("show_map")) {
	  show_mouse_map();
	}else{
	  function = key->n-5;
	  default_token(get_mouse_keys(function));
	  key=free_Token(key);
	  key=Get_String("keys (combination of S,A,C,L,M,R): ",";");
	  if (key->status==TOK_OK) {
	    set_mouse_function(function,key->text);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad token: %s",key->text);
	  }
	}
      }else if (key->status == TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
      }
    }else if (key_is("mpeoe")) {
      key=free_Token(key);
      key=Get_Kwd("mpeoe parameter to set: ",";","granularity");
      if (key->status==TOK_OK) {
	key=free_Token(key);
	key=Get_Kwd("granularity: ",";","0 1 2 3 4 5 6");
	if (key->status==TOK_OK) {
	  MPEOEsetGranularity(key->n);
	}else if (key->status==TOK_NOK) err_set(1,"bad value for mpeoe parameter: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad mpeoe parameter: %s",key->text);
    }else if (key_is("RGB")) {		    /* RGB mode on/off */
      free_Token(key);
      key=Get_Kwd("mode: ",";","on off");
      if (key->status == TOK_OK) {
	mode=key->n;
	send_bheader(graphix,"SRGB",sizeof(int));
	send_bint(graphix,mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown RGB mode: %s",key->text);
      }
    }else if (key_is("antialiasing")) {		/* mode on/off */
      free_Token(key);
      key=Get_Kwd("mode: ",";","on off");
      if (key->status == TOK_OK) {
	mode=1-key->n;
	setAttribute("SAAL",mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown mode: %s",key->text);
      }
    }else if (key_is("default")) {
      cmd_setdefault();
    }else if (key_is("depthcue")) {		 /* mode on/off */
      free_Token(key);
      key=Get_Kwd("mode: ",";","on off");
      if (key->status == TOK_OK) {
	mode=1-key->n;
	setAttribute("SDPC",mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown mode: %s",key->text);
      }
    }else if (key_is("dither")) {		 /* mode on/off */
      free_Token(key);
      key=Get_Kwd("mode: ",";","on off");
      if (key->status == TOK_OK) {
	mode=1-key->n;
	setAttribute("SDTH",mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown mode: %s",key->text);
      }
    }else if (key_is("display")) {
      cmd_display_set();
    }else if (key_is("radii")) {
      cmd_radii();
    }else if (key_is("ramp")) {
      key=free_Token(key);
      key=Get_Int("length of ramp: ",";");
      if (key->status==TOK_OK) {
	n=key->n; if (n>MXRAMP) n=MXRAMP;
	for (i=0;i<n&&key->status==TOK_OK;i++) {
	  key=free_Token(key);
	  sprintf(buffer,"color for ramp entry %d: ",i+1);
	  key=Get_Color(buffer,";");
	  if (key->status==TOK_OK) ramp[i]=key->n;
	  else if (key->status==TOK_NOK) err_set(1,"bad color: %s",key->text);
	}
	if (key->status==TOK_OK) DefineColorRamp(n,ramp);
      }
    }else if (key_is("rock")) {             /* set pars for rocking */
      key=free_Token(key);
      key=Get_Float_M("amplitude: ",";","5 10 15 20 25 30 35 40 45 50 55 60");
      if (key->status==TOK_OK) {
	rock_amplitude=key->f;
	key=free_Token(key);
	key=Get_Float_M("step: ",";","0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0");
	if (key->status==TOK_OK) {
	  rock_step=key->f;
	  send_bheader(graphix,"ROCK",2*sizeof(float));
	  send_bfloat(graphix,rock_amplitude);
	  send_bfloat(graphix,rock_step);
	}else if  (key->status==TOK_NOK) {
	  err_set(1,"bad rocking step: %s",key->text);
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad rocking amplitude: %s",key->text);
      }
    }else if (key_is("stereo")) {
      key=free_Token(key);
      key=Get_Kwd("mode: ",";",
		  SrvVersion()<2.109 ? "off crossed glasses relaxed left right" :
		  "off crossed glasses relaxed left right quadbuffered" );
      if (key->status == TOK_OK) {
	mode=st_mode[key->n];
	send_bheader(graphix,"SSTM",sizeof(int));
	send_bint(graphix,mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown stereo mode: %s",key->text);
      }
    }else if (key_is("structal")) {
      key=free_Token(key);
      key=Get_Kwd("structal parameter: ",";","compare gap mode SSslope");
      if (key->status==TOK_OK) {
	if (key_is("compare")) {
	  setResCompare();
	}else if (key_is("gap")) {
	  set_gap_penalty();
	}else if (key_is("mode")) {
	  key=free_Token(key);
	  key=Get_Kwd("mode: ",";","full partial simple");
	  if (key->status==TOK_OK) {
	    if (key_is("partial")) {
	      key=free_Token(key);
	      key=Get_Int_M("minimum block size: ",";","5 6 7 8 9 10 15 20 25 30 40 50");
	      if (key->status==TOK_OK) {
		SetStructalMode("partial",key->n);
	      }else if (key->status==TOK_NOK) err_set(1,"invalid block size: %s",key->text);
	    }else SetStructalMode(key->key,0);
	  }else if (key->status==TOK_NOK) err_set(1,"invalid structal mode: %s",key->text);
	}else if (key_is("SSslope")) {
	  set_gerstXcompare5Factor();
	}
      }else if (key->status==TOK_NOK) err_set(1,"bad structal paramater: %s\n",key->text);
    }else if (key_is("TAFF")) {
      key=free_Token(key);
      key=Get_Kwd("energy term: ",";","bond angle plane torsion '' mode");
      if (key->status==TOK_OK) {
	if (key_is("bond")) {
	  key=free_Token(key);
	  key=Get_Float_M("weight for bond stretch> ",";",TAFF_wmenu);
	  if (key->status==TOK_OK) setTAFFweight("bond",key->f);
	  else if (key->status==TOK_NOK) err_set(1,"bad TAFF weight: %s",key->text);
	}else if (key_is("angle")) {
	  key=free_Token(key);
	  key=Get_Float_M("weight for angle bend> ",";",TAFF_wmenu);
	  if (key->status==TOK_OK)  setTAFFweight("angle",key->f);
	  else if (key->status==TOK_NOK) err_set(1,"bad TAFF weight: %s",key->text);
	}else if (key_is("plane"))  {
	  key=free_Token(key);
	  key=Get_Float_M("weight for out-of-plane bend> ",";",TAFF_wmenu);
	  if (key->status==TOK_OK)  setTAFFweight("plane",key->f);
	  else if (key->status==TOK_NOK) err_set(1,"bad TAFF weight: %s",key->text);
	}else if (key_is("torsion")) {
	  key=free_Token(key);
	  key=Get_Float_M("weight for torsion> ",";",TAFF_wmenu);
	  if (key->status==TOK_OK)  setTAFFweight("torsion",key->f);
	  else if (key->status==TOK_NOK) err_set(1,"bad TAFF weight: %s",key->text);
	}else if (key_is("mode")) {
	  key=free_Token(key);
	  key=Get_Kwd("file format> ",";","NAMD X-Plor");
	  if (key->status==TOK_OK)  setTAFFformat(key->key);
	  else if (key->status==TOK_NOK) err_set(1,"bad TAFF file format: %s",key->text);
	}
      }else if (key->status==TOK_NOK) err_set(1,"unknown TAFF energy term: %s",key->text);

    }else if (key_is("text")) {
      free_Token(key);
      key=Get_Kwd("mode: ",";","hardware stroke library");
      if (key->status == TOK_OK) {
	mode=key->n+1;
	setAttribute("SLTM",mode);
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown text mode: %s",key->text);
      }
    }else if (key_is("background")) {
      key=free_Token(key);
      key=Get_Color_M("background color: ",";","RGB ' '");
      if (key->status==TOK_OK) {
	if (key->n==COL_RGB) {
	  for (i=0;i<3;i++) {
	    key=free_Token(key);
	    key=Get_Int_M(col_pmt[i],",;",col_cval);
	    if (key->status==TOK_OK) 
	      cldf[i]=(key->n > 255 ? 255 : (key->n < 0 ? 0 : key->n));
	    else break;
	  }
	  if (i==3) {
	    DefineColor(0,"background",cldf[0],cldf[1],cldf[2]);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad integer token: %s",key->text);
	  }
	}else{
	  DefineColor(0,"background",
		      ColVal(key->n,0),ColVal(key->n,1),ColVal(key->n,2));
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad color keyword: %s",key->text);
      }
    }else if (key_is("fog")) {
      key=free_Token(key);
      key=Get_Color_M("fog color: ",";","RGB ' '");
      if (key->status==TOK_OK) {
	if (key->n==COL_RGB) {
	  for (i=0;i<3;i++) {
	    key=free_Token(key);
	    key=Get_Int_M(col_pmt[i],",;",col_cval);
	    if (key->status==TOK_OK) 
	      cldf[i]=(key->n > 255 ? 255 : (key->n < 0 ? 0 : key->n));
	    else break;
	  }
	  if (i==3) {
	    DefineColor(-1,"fog",cldf[0],cldf[1],cldf[2]);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad integer token: %s",key->text);
	  }
	}else{
	  DefineColor(-1,"fog",
		      ColVal(key->n,0),ColVal(key->n,1),ColVal(key->n,2));
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad color keyword: %s",key->text);
      }
    }else if (key_is("color")) {
      free_Token(key);
      key=Get_Color("color: ",";");
      if (key->status == TOK_OK) {
	nr=key->n;
	for (i=0;i<3;i++) {
	  key=free_Token(key);
	  default_token("%d",ColVal(nr,i));
	  key=Get_Int_M(col_pmt[i],",;",col_cval);
	  if (key->status==TOK_OK) 
	    cldf[i]=(key->n > 255 ? 255 : (key->n < 0 ? 0 : key->n));
	  else break;
	}
	if (i==3) {
	  free_Token(key);
	  default_token(ColMnem(nr));
	  key=Get_String("mnemonic: ",";");
	  if (key->status==TOK_OK) {
	    DefineColor(nr,key->text,cldf[0],cldf[1],cldf[2]);
	  }
	}
      }else if (key->status==TOK_NOK) {
	err_set(1,"unknown option : %s",key->text);
      }	
    }else if (key_is("pivot")) {
      free_Token(key); free_menu(menu);
      menu=Make_GlblNameList("/",DspRoot);
      key=Get_DspGlbl_M("object: ",";",menu);
      if (key->status==TOK_OK) dobj=(Dspobj *)key->ptr;
      else {
	dobj=(Dspobj *)NULL;
	if (key->status==TOK_NOK) err_set(1,"object not found");
      }
      send_bheader(graphix,"ARMK",0);
      if (mvld(dobj)) {
	get_MVT(DspRoot);
	mo=mmxinv(glblobj(dobj)->mtrx);
	free_Token(key);
	key=Get_Kwd("option: ",";","atom COM");
	if (key->status==TOK_OK) {
	  if (key_is("atom")) {
	    free_Token(key);
	    key=Get_Atom("pivot atom: ",";");
	    if (key->status==TOK_OK) {
	      atm=(Atom *)key->ptr;
	      ma=glblobj(assocobj((Node *)atm))->mtrx;
	      v32arr(pt,m4xv3(mo,m4xv3(ma,vec3(atm->x))));
	      set_pivot(dobj,pt);
	    }else if (key->status==TOK_NOK) {
	      err_set(1,"bad atom token.");
	    }
	  }else if (key_is("COM")) {
	    free_Token(key);
	    key=Get_Atoms_D("atoms: ",";");
	    if (key->status==TOK_OK) {
	      c=(Nodechain *)key->ptr;
	      for (j=0;j<3;j++) pv.c[j]=0.0;
	      for (nat=0,atm=(Atom *)c->s;atm;atm=(Atom *)atm->chain) {
		if (is_atom(atm)) {
		  nat++;
		  ma=glblobj(assocobj((Node *)atm))->mtrx;
		  pv=v3pv3(pv,m4xv3(mo,m4xv3(ma,vec3(atm->x))));
		}
	      }
	      clear_node_chain(c); free_node_chain(c);
	      if (nat>0) {
		for (j=0;j<3;j++) pt[j]=pv.c[j]/nat;
		set_pivot(dobj,pt);
	      }else err_set(1,"0 atoms selected, pivot unchanged.");
	    }else if (key->status==TOK_NOK) err_set(1,"bad atom selection: %s",key->text);
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"pivot: bad token");
	}
        send_bheader(graphix,"ARMK",0);
      }else{
	err_set(1,"object not valid: %s",key->text);
      }
    }else if (key_is("view")) {
      free_Token(key);
      key=Get_Kwd("mode: ",";","ortho perspective");
      if (key->status == TOK_OK) {
	if (key_is("perspective")) send_message(graphix,"SVWM",4,"PERS");
	else if (key_is("ortho"))  send_message(graphix,"SVWM",4,"ORTH");
      }else if (key->status == TOK_NOK) {
	err_set(1,"unknown viewing mode : %s",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"unkown option : %s",key->text);
  }
  free_Token(key);
  free_menu(menu);
}
/*-------------------- cmd_setenv --------------------*/

void cmd_setenv(void)
{
  Token *key=(Token *)NULL;
  char obj_name[4096],buffer[4096],*prmpt=(char *)NULL;
  int echo=1,old=1;

  do {                           /* handle options: -prompt and -noecho */
    key=free_Token(key);
    key=Get_String("variable: ",";");
    if (key->status==TOK_OK) {
      key=Prs_Kwd(key,"-prompt -noecho");
      if (key->status==TOK_OK) {
	if (key_is("-noecho")) {
	  echo=0;
	  key->status=TOK_NOK;
	}else if (key_is("-prompt")) {
	  if (prmpt) free(prmpt);
	  prmpt=(char *)NULL;
	  key=free_Token(key);
	  key=Get_String("prompt: ",";");
	  if (key->status==TOK_OK) {
	    prmpt=key->text;
	    key->text=(char *)NULL;
	    key->status=TOK_NOK;
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad token: %s",key->text);
	    key->status=TOK_ABORT;
	    break;
	  }
	}
      }else if (key->status==TOK_NOK) {
	key->status=TOK_OK;
	break;
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
      break;
    }else break;
  } while (key->status==TOK_NOK);

  if (key->status==TOK_OK) {
    Strcpy(obj_name,key->text);
    free_Token(key);
    if (!echo) old=set_kbd_echo(echo);
    key=Get_String(prmpt? prmpt:"definition: ",";");
    if (!echo) set_kbd_echo(old);
    if (key->status==TOK_OK) {
      sprintf(buffer,"%s=%s",obj_name,key->text);
      send_message(graphix,"SENV",strlen(buffer),buffer);
      Setenv(obj_name,key->text);
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad token: %s",key->text);
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }

  key=free_Token(key);
  if (prmpt) free(prmpt);
}
/*-------------------- cmd_pwd --------------------*/

void cmd_pwd(void)
{
  char path[MAXPATHLEN];

  if (getCWD(path,sizeof(path))) {
    prtGraphics("current directory: %s\n",path);
  }else{
    err_set(1,"sorry, can't determine current working directory.");
    err_dump();
  }
}

/*-------------------- get_I_parm --------------------*/

int get_I_parm(char *what)
{
  int r;

  if (streq(what,"cyl_poly")) r=cyl_poly;
  else if (streq(what,"cyl_mode")) r=cyl_mode;
  else if (streq(what,"sph_poly")) r=sph_poly;
  else if (streq(what,"sph_mode")) r=sph_mode;
  else r=-1;

  return(r);
}

void set_cyl_mode(int m,int n)
{
  if (m==WNP_CYLINDER_FULL||m==WNP_CYLINDER_HALF) {
    send_bheader(graphix,"SCYM",2*sizeof(int));
    send_bint(graphix,m);
    send_int(graphix,0);
    sph_mode=m;           /* nore sph/cyl mode correlation */
  }
  if (n>-1) {
    send_bheader(graphix,"SCYM",2*sizeof(int));
    send_bint(graphix,WNP_CYLINDER_QUALITY);
    send_int(graphix,n);
    cyl_poly=n;
  }
}
void set_sph_mode(int m,int n)
{
  if (m==WNP_SPHERE_FULL||m==WNP_SPHERE_HALF) {
    send_bheader(graphix,"SPHM",2*sizeof(int));
    send_bint(graphix,m);
    send_int(graphix,0);
    sph_mode=m;
  }
  if (n>-1) {
    send_bheader(graphix,"SPHM",2*sizeof(int));
    send_bint(graphix,WNP_SPHERE_QUALITY);
    send_int(graphix,n);
    sph_poly=n;
  }
}
/*-------------------- set_gerstXcompare5Factor --------------------*/

static void set_gerstXcompare5Factor(void)
{
  Token *key=(Token *)NULL;
  float old;

  old=GetGersteinCompare5Factor();
  default_token("%0.2f",old);
  key=Get_Float_M("slope for SS mismatch penalty: ",";",
		  "0.0 0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 6.0 7.0 8.0 9.0 10.0");
  if (key->status==TOK_OK) {
    SetGersteinCompare5Factor(key->f<0?0:key->f);
  }else if (key->status==TOK_NOK) err_set(1,": %s, bad token.",key->text);
  free_Token(key);
}

/*-------------------- setResCompare --------------------*/

static void setResCompare(void)
{
  Token *key=(Token *)NULL;

  key=Get_Kwd("comparison function to use: ",";","ResCompare5 ResCompare6");
  if (key->status==TOK_OK) {
    SetStructalResCompare(key->key);
  }else if (key->status==TOK_NOK) err_set(1,"unknown comparison function: %s",key->text);
  free_Token(key);
}
/*-------------------- set_gap_penalty --------------------*/

static void set_gap_penalty(void)
{
  Token *key=(Token *)NULL;
  float gapo,gape;

  GetStructalGapPenalty(&gapo,&gape);
  default_token("%0.2f",gapo);
  key=Get_Float_M("gap penalty (open): ",";",
		  "0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 6.0 7.0 8.0 9.0 10.0");
  if (key->status==TOK_OK) {
    gapo=key->f;
    key=free_Token(key);
    default_token("%0.3f",gape);
    key=Get_Float_M("gap penalty (extend): ",";",
		    "0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.6 0.9 1.0");
    if (key->status==TOK_OK) {
      gape=key->f;
      SetStructalGapPenalty(gapo,gape);
    }else if (key->status==TOK_NOK) err_set(1,"gap penalty (extend): %s, bad token.",key->text);
  }else if (key->status==TOK_NOK) err_set(1,"gap penalty (open): %s, bad token.",key->text);
  if (key->status==TOK_OK) {
    key=free_Token(key);
    default_token("%0.3f",GetGersteinGapPenalty());
    key=Get_Float_M("Gerstein's gap penalty factor: ",";",
		    "0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.6 0.9 1.0 10.0 100.0 1000.0 10000.0");
    if (key->status==TOK_OK) {
      SetGersteinGapPenalty(key->f);
    }else if (key->status==TOK_NOK) err_set(1,"Gerstein's gap penalty factor: %s, bad token.",key->text);
  }
  free_Token(key);
}
