#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include <sys/param.h>
#include "errhand.h"

extern int graphix;

#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define TRUE 1
#define FALSE 0

static void set_default_atom(void);
static void set_default_bond(void);
static void set_default_label(void);

/*------------------------------------------------------------------------*/
void cmd_setdefault(void)
{
  static char *kwdlist = "atom bond label";
  Token *key=(Token *)NULL;

  key=Get_Kwd("set defaults for: ",";",kwdlist);
  if (key->status==TOK_OK) {
    if (key_is("atom")) set_default_atom();
    else if (key_is("bond")) set_default_bond();
    else if (key_is("label")) set_default_label();
  }else if (key->status==TOK_NOK) err_set(1,"set default: unknown option (%s)\n",key->text);
  key=free_Token(key);
}
/*-------------------- set default atom --------------------*/

static void set_default_atom(void)
{
  static char *keys="color";
  Token *key=(Token *)NULL;
  int color;

  key=Get_Kwd("atom option: ",";",keys);
  if (key->status==TOK_OK) {
    if (key_is("color")) {
      key=free_Token(key);
      key=Get_Color_M("default atom color: ",";","by_atomtype ' '");
      if (key->status==TOK_OK) {
	if (key_is("by_atomtype")) color=WNP_COLOR_BYATOM;
	else color=key->n;
	SetDefaultAtomColor(color);
      }else if (key->status==TOK_NOK) err_set(1,"bad color: %s\n",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"unknown label option: %s\n",key->text);
  key=free_Token(key);
}
/*-------------------- set default bond --------------------*/

static void set_default_bond(void)
{
  static char *keys="width";
  Token *key=(Token *)NULL;

  key=Get_Kwd("bind option: ",";",keys);
  if (key->status==TOK_OK) {
    if (key_is("width")) {
      key=free_Token(key);
      key=Get_Int_M("default bond width: ",";","1 2 3 4 5 6 7 8 9 10");
      if (key->status==TOK_OK) {
	send_bheader(graphix,"SBDW",sizeof(int));
	send_bint(graphix,clamp(key->n,1,100));
	SetDefaultBondWidth(key->n);
      }else if (key->status==TOK_NOK) err_set(1,"bad bond width: %s\n",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"unknown bond option: %s\n",key->text);
  key=free_Token(key);
}
/*-------------------- set default label --------------------*/

static void set_default_label(void)
{
  static char *keys="color";
  Token *key=(Token *)NULL;
  int color;

  key=Get_Kwd("label option: ",";",keys);
  if (key->status==TOK_OK) {
    if (key_is("color")) {
      key=free_Token(key);
      key=Get_Color_M("default label color: ",";","by_atomcolor ' '");
      if (key->status==TOK_OK) {
	if (key_is("by_atomcolor")) color=WNP_COLOR_BYATOM;
	else color=key->n;
	send_bheader(graphix,"SLDC",sizeof(int));
	send_bint(graphix,color);
      }else if (key->status==TOK_NOK) err_set(1,"bad color: %s\n",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"unknown label option: %s\n",key->text);
  key=free_Token(key);
}

