#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Dspobj *DspRoot;
extern Node *MolRoot;
extern Node *LibRoot;

#define key_is(x) (streq(key->key,x))

#define NONE 0
#define DONE 1
#define SKETCH 2

#define ATM_LAB_COLOR 2   /* green ? */

static Node *cmd_sketch_new_mol(void);
static void get_sketch_defaults(Atomdef **atp);
static void sketch_atom(void);
static void sketch_bond(void);
static void sketch_addh(Node *root);
static void sketch_striph(Node *root);
static void sketch_autotypes(Node *root);

void cmd_sketch(void)
{
  Node *root;
  Node *ResLib;
  Dspobj *droot,*proot,*svp;
  int mode,default_bt;
  Atom *new,*prv;
  Token *key;
  Lib *reslib;
  Atomdef *default_type;
  mat_4 mmv;
  vec_3 dx;
  char prmt[1024],buff[1024];
  float *pn,*pf;

  if (!graphix) {
    err_set(1,"cmd_sketch: available only in graphics mode.\n");
    return;
  }

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    return;
  }
  reslib=(Lib *)ResLib;
  default_type=find_atomdef_name((Node *)reslib,"C.?");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }
  default_bt=BT_SINGLE;

  root=(Node *)NULL;;
  droot=(Dspobj *)NULL;
  prv=(Atom *)NULL;
  proot=pick_root(DspRoot);
  key=(Token *)NULL;
  sprintf(prmt,"sketch...");

  mode = SKETCH;
  while(mode!=DONE) {
    switch (mode) {
    case SKETCH:
      err_dump();
      pick_mode("PXYA");
      key=Get_Kwd(prmt,";","AddH Atom AutoTypes Bond Default Molecule StripH End");
      pick_mode("POFF");
      if (key->status==TOK_OK) {
	if (is_atom(prv)) send_bheader(graphix,"ARMK",0);
	prv=(Atom *)NULL;                  /* break draw sequence... */
	if (key_is("Atom"))            sketch_atom();
	else if (key_is("Bond"))       sketch_bond();
	else if (key_is("Default"))    get_sketch_defaults(&default_type);
	else if (key_is("AddH"))       sketch_addh(root);
	else if (key_is("StripH"))     sketch_striph(root);
	else if (key_is("AutoTypes"))  sketch_autotypes(root);
	else if (key_is("Molecule")) {
	  svp=pick_root(proot);
	  key=Get_Mol_D("select molecule: ",";");
	  pick_root(svp);
	  if (key->status==TOK_OK) {
	    root=(Node *)key->ptr;
	    droot=assocmdl(root);
	    pick_root(droot);
	    sprintf(prmt,"sketch %s ...",nodeptr_to_name(buff,root));
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad molecule selection: %s",key->text);
	  }
	}else if (key_is("End")) mode=DONE;
      }else{
	if (key->status==TOK_NOK) {
	  key=Prs_AtomCtxt(key,(Node *)prv);
	  if (key->status==TOK_OK) {
	    if (mvld(prv)) {
	      send_bheader(graphix,"ARMK",0);
	      set_bt(add_bond(prv,(Atom *)key->ptr),default_bt);
	      prv=(Atom *)NULL;
	    }else{
	      prv=(Atom *)key->ptr;
	      if (!mvld(root)) {
		droot=assocmdl((Node *)prv);
		pick_root(droot);
		root=prv->up;
		sprintf(prmt,"sketch %s ...",nodeptr_to_name(buff,root));
	      }
	    }
	  }else{
	    key=Prs_XYZ(key);
	    if (key->status==TOK_OK) {
	      if (!mvld(root)) {
		root=cmd_sketch_new_mol();
		sprintf(prmt,"sketch %s ...",nodeptr_to_name(buff,root));
		droot=assocmdl(root);
		pick_root(droot);
	      }
	      new=(Atom *)add_node(root,(char *)NULL,OBJ_ATOM);
	      if (mvld(new)) {
		if (new->mxyz<1) reall_atom_x(new,1);
		mmv=obj_transform(root);
		if (mvld(prv)) dx=m4xv3(mmv,vec3(prv->x));
		else {
		  pn=(float *)key->ptr; pf=pn+3;
		  dx.c[2]=0.25*pf[2]+0.75*pn[2];
		}
		dx=wline2v3((float *)key->ptr,dx.c[2]);
		v32arr(new->x,m4xv3(mmxinv(mmv),dx));
		new->nxyz=1;
		set_atom_type(new,default_type);
		new->color=-new->atp->color;
		set_node_name((Node *)new,new->atp->chs);
		/* new->chrg=0;    no longer needed, thanks to improved ligance guesser */
		if (mvld(prv)) set_bt(add_bond_raw(prv,new),default_bt);
		send_bheader(graphix,"ARMK",0);
		send_one_atom(new);
		if (cPtr(new->cross)) {
		  /* use default WNP_COLOR_BYATOM
		  send_bheader(graphix,"ALCL",sizeof(new)+sizeof(int));
				  send_bint(graphix,ATM_LAB_COLOR);
		  send_bptr(graphix,new->cross);
		  */
		  send_bheader(graphix,"ASMK",cPtrSize());
		  send_bcptr(graphix,new->cross);
		  trflush(graphix);
		}
		prv=new;
	      }
	      free(key->ptr);
	    }
	  }
	}
      }
      if (key->status==TOK_NOK) {
	err_set(1,"bad token: %s",key->text);
	send_bheader(graphix,"ARMK",0);
      }
      key=free_Token(key);
    }
  }
  send_header(graphix,"ARMK",0);
  pick_root(proot);
}
/*-------------------- cmd_sketch_new_mol --------------------*/

static Node *cmd_sketch_new_mol(void)
{
  Node *r;
  Dspobj *d;

  r=add_node(MolRoot,"sketch",OBJ_MOL);
  if (mvld(r) && graphix>0) {
    d=new_dsp_obj(DspRoot,(char *)NULL,DSP_GLBL);
    if (mvld(d)) {
      send_mdl(d,(Mol *)r);
      d->root=r;
      ((Mol *)r)->dsptr=d;
    }
  }
  return(r);
}
/*-------------------- get_sketch_defaults --------------------*/

static void get_sketch_defaults(Atomdef **atom)
{
  Token *key;

  key=(Token *)NULL;
  key=Get_Kwd("default for: ",";","atom");
  if (key->status==TOK_OK) {
    if (key_is("atom")) {
      key=free_Token(key);
      key=Get_AtomType("default atom type: ",";");
      if (key->status==TOK_OK) {
	*atom = (Atomdef *)key->ptr;
      }else if (key->status==TOK_NOK) {
	err_set(1,"bad atomtype: %s",key->text);
      }
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s",key->text);
  }
  free_Token(key);
}
/*-------------------- get_sketch_atom_pars --------------------*/

#define SKETCH_ATMOP_DONE     0
#define SKETCH_ATMOP_CHARGE   1
#define SKETCH_ATMOP_LIGANCE  2
#define SKETCH_ATMOP_TYPE     3
#define SKETCH_ATMOP_DELETE   4
#define SKETCH_ATMOP_POSITION 5

static int get_sketch_atom_pars(Token *K,int *ligance,int *charge,Atomdef **atp)
{
  Token *key;
  int op;
  
  key=(Token *)NULL;
  if (streq(K->key,"ligance")) {
    op=SKETCH_ATMOP_LIGANCE;
    key=Get_Kwd("formal ligance: ",";","0 1 2 3 4 5 6 7 8 9 10 automatic");
    if (key->status==TOK_OK) {
      if (key_is("automatic")) *ligance='?';
      else *ligance=key->n;
    }else if (key->status==TOK_NOK) {
      key=Prs_Int(key);
      if (key->status==TOK_OK) *ligance=key->n;
      else {
	err_set(1,"bad ligance: %s",key->text);
	op=SKETCH_ATMOP_DONE;
      }
    }else op=SKETCH_ATMOP_DONE;
  }else if (streq(K->key,"charge")) {
    op=SKETCH_ATMOP_CHARGE;
    key=Get_Int_M("formal charge : ",";","0 1 2 3 -1 -2 -3 ' ' automatic");
    if (key->status==TOK_OK) {
      *charge=key->n;
    }else if (key->status==TOK_NOK){
      key=Prs_Kwd(key,"automatic");
      if (key->status==TOK_OK) *charge='?';
      else {
	err_set(1,"bad ligance: %s",key->text);
	op=SKETCH_ATMOP_DONE;
      }
    }else op=SKETCH_ATMOP_DONE;
  }else if (streq(K->key,"type")) {
    op=SKETCH_ATMOP_TYPE;
    key=Get_AtomType("atom type: ",";");
    if (key->status==TOK_OK) *atp=(Atomdef *)key->ptr;
    else{
      if (key->status==TOK_NOK) err_set(1,"bad atom type: %s",key->text);
      op=SKETCH_ATMOP_DONE;
    }
  }else if (streq(K->key,"delete")) {
    op=SKETCH_ATMOP_DELETE;
  }else if (streq(K->key,"position")) {
    op=SKETCH_ATMOP_POSITION;
  }else op=SKETCH_ATMOP_DONE;
  free_Token(key);
  return(op);
}
/*-------------------- sketch_atom --------------------*/

static void sketch_atom(void)
{
  Token *key;
  int op,ligance,charge,llen=0;
  Atomdef *atp;
  Atom *a;
  char *menu="charge ligance position type ' ' delete";
  char label[256];
  char prmt[256];
  char *ptxt[] = {
    "nop","modify charge","modify ligance","modify type","delete atom","move atom"};
  mat_4 mvm;
  float *pt1,*pt2;
  vec_3 newx;

  key=(Token *)NULL;
  key=Get_Kwd("atom option: ",";",menu);
  if (key->status==TOK_OK) {
    op=get_sketch_atom_pars(key,&ligance,&charge,&atp);
    while (op!=SKETCH_ATMOP_DONE) {
      key=free_Token(key);
      sprintf(prmt,"%s (pick atom or select option) ...",ptxt[op]);
      key=Get_Atom_M(prmt,";",menu);
      if (key->status==TOK_OK) {
	a=(Atom *)key->ptr;
	switch (op) {
	case SKETCH_ATMOP_DELETE:
	  del_node((Node *)a);
	  a=(Atom *)NULL;
	  break;
	case SKETCH_ATMOP_CHARGE:
	  a->chrg=charge;
	  if (a->chrg!='?') llen=sprintf(label,"%+d",a->chrg);
	  else llen=sprintf(label,"??");
	  if (a->chrg==0) {
	    llen=0;
	    if (cPtr(a->cross)) {
	      send_bheader(graphix,"ALRM",cPtrSize());
	      send_bcptr(graphix,a->cross);
	    }
	  }
	  break;
	case SKETCH_ATMOP_LIGANCE:
	  a->lig=ligance;
	  if (a->lig!='?') llen=sprintf(label,"%+d",a->lig);
	  else llen=sprintf(label,"??");
	  break;
	case SKETCH_ATMOP_TYPE:
	  if (atp->color != a->color && atp->color != -a->color) {
	    if (cPtr(a->cross)) {
	      send_bheader(graphix,"ATCO",cPtrSize()+sizeof(int));
	      send_bint(graphix,atp->color);
	      send_bcptr(graphix,a->cross);
	    }
	  }
	  set_atom_type(a,atp);
	  a->color=-a->atp->color;
	  a->lig='?';
	  /* a->chrg=0; dont want to do this... */
	  llen=sprintf(label,"%s",a->atp->name);
	  break;
	case SKETCH_ATMOP_POSITION:
	  key=free_Token(key);
	  key=Get_XYZ("new position: ",";");
	  if (key->status==TOK_OK) {
	    pt1=(float *)key->ptr; pt2=(float *)key->ptr+3;
	    mvm=obj_transform((Node *)a);
	    newx=m4xv3(mvm,vec3(a->x));
	    mvm=mmxinv(mvm);
	    newx.c[0]=pt1[0]+(pt2[0]-pt1[0])*(pt1[2]-newx.c[2])/(pt1[2]-pt2[2]);
	    newx.c[1]=pt1[1]+(pt2[1]-pt1[1])*(pt1[2]-newx.c[2])/(pt1[2]-pt2[2]);
	    v32arr(a->x,m4xv3(mvm,newx));
	    if (cPtr(a->cross)) {
	      send_bheader(graphix,"NEWX",cPtrSize()+3*sizeof(float));
	      send_bcptr(graphix,a->cross);
	      send_bfloata(graphix,3,a->x);
	    }
	    free(key->ptr);
	  }else if (key->status==TOK_NOK) {
	    err_set(1,"bad token: %s",key->text);
	    op=SKETCH_ATMOP_DONE;
	  }
	  llen=0;
	  break;
	}
	if (is_atom(a)&&llen>0) {
	  if (auto_label()) set_atom_label(a,ATM_LAB_TEXT,label,TRUE);
	}
      }else if (key->status==TOK_NOK) {
	key=Prs_Kwd(key,menu);
	if (key->status==TOK_OK) {
	  op=get_sketch_atom_pars(key,&ligance,&charge,&atp);
	}else{
	  err_set(1,"bad atom option: %s",key->text);
	  key->status=TOK_ABORT;
	  op=SKETCH_ATMOP_DONE;
	}
      }else op=SKETCH_ATMOP_DONE;
      send_bheader(graphix,"ARMK",0);
    }
  }
  free_Token (key);
}
/*-------------------- get_sketch_bond_pars --------------------*/

#define SKETCH_BNDOP_DONE     0
#define SKETCH_BNDOP_TYPE     1
#define SKETCH_BNDOP_DELETE   2

static int get_sketch_bond_pars(Token *K,int *bt)
{
  Token *key;
  int op,l,i;
  char bt_menu[1024];
  
  key=(Token *)NULL;
  if (streq(K->key,"type")) {
    op=SKETCH_BNDOP_TYPE;
    for (l=0,i=0;i<MX_OBJ_BTYPES;i++) {
      sprintf(bt_menu+l,"%s ",BT_mnem(i));
      l=strlen(bt_menu);
    }
    key=Get_Kwd("bond type: ",";",bt_menu);
    if (key->status==TOK_OK) *bt=BT_number(key->key);
    else if (key->status==TOK_NOK) {
      err_set(1,"bad bond type token: %s",key->text);
      op=SKETCH_BNDOP_DONE;
    }else op=SKETCH_BNDOP_DONE;
  }else if (streq(K->key,"delete")) {
    op=SKETCH_BNDOP_DELETE;
  }else op=SKETCH_BNDOP_DONE;
  free_Token(key);
  return(op);
}
/*-------------------- sketch_bond --------------------*/

static void sketch_bond(void)
{
  Token *key;
  int op,bt;
  char *menu="type ' ' delete";
  char prmt[256];
  char *ptxt[] = {
    "nop","modify type","delete bond"};
  Bond *bnd;

  key=(Token *)NULL;
  key=Get_Kwd("bond option: ",";",menu);
  if (key->status==TOK_OK) {
    op=get_sketch_bond_pars(key,&bt);
    while (op!=SKETCH_BNDOP_DONE) {
      key=free_Token(key);
      sprintf(prmt,"%s (pick bond or select option) ...",ptxt[op]);
      key=Get_Bond_M(prmt,";",menu);
      send_bheader(graphix,"ARMK",0);
      if (key->status==TOK_OK) {
	bnd=(Bond *)key->ptr;
	switch (op) {
	case SKETCH_BNDOP_DELETE:
	  del_bond(bnd);
	  bnd=(Bond *)NULL;
	  break;
	case SKETCH_BNDOP_TYPE:
	  set_bt(bnd,bt);
	  set_bond_label(bnd,BND_LAB_TYPE,NULL);
	  break;
	}
      }else if (key->status==TOK_NOK) {
	key=Prs_Kwd(key,menu);
	if (key->status==TOK_OK) {
	  op=get_sketch_bond_pars(key,&bt);
	}else{
	  err_set(1,"bad bond option: %s",key->text);
	  key->status=TOK_ABORT;
	  op=SKETCH_BNDOP_DONE;
	}
      }else op=SKETCH_BNDOP_DONE;
    }
  }
  free_Token (key);
}
/*-------------------- sketch_addh --------------------*/

static void sketch_addh(Node *root)
{
  Nodechain ch;
  Atom *a;
  Atomdef *atp;
  Node *ResLib;
  char buf[2048];

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (mvld(root)) {
    init_node_chain(&ch);
    ForAtoms(a,root) add_hydrogens(a,&ch);
    atp=(Atomdef *)rfind_name(ResLib,"H.?",OBJ_ATOMDEF);
    if (mvld(atp)) {
      for (a=(Atom *)ch.s;mvld(a);a=(Atom *)a->chain) {
	set_atom_type(a,atp);
	a->color=-a->atp->color;
      }
    }
    send_atom_chain(&ch);
    prtGraphics("%d hydrogen%s added to %s\n",
		ch.n,(ch.n == 1 ? "" : "s"),nodeptr_to_name(buf,root));
    clear_node_chain(&ch);
  }else{
    err_set(1,"invalid sketch-root.");
  }
}
/*-------------------- sketch_striph --------------------*/

static void sketch_striph(Node *root)
{
  Nodechain ch;
  Atom *a;
  char buf[1024];

  if (mvld(root)) {
    init_node_chain(&ch);
    ForAtoms(a,root) if (mvld(a->atp)&&a->atp->z==1) add_to_node_chain((Node *)a,&ch);
    prtGraphics("%d hydrogen%s deleted from %s\n",
		ch.n,(ch.n == 1 ? "" : "s"),nodeptr_to_name(buf,root));
    del_node_chain(&ch);
  }else{
    err_set(1,"invalid sketch-root.");
  }
}
/*-------------------- sketch_autotypes --------------------*/

#define tCFLG 2
#define tLFLG 3

static void sketch_autotypes(Node *root)
{
  Nodechain ca;
  Atom *a;
  int nunk;

  if (mvld(root)) {
    ca=tree2chain(root,OBJ_ATOM);
    for (a=(Atom *)ca.s;mvld(a);a=(Atom *)a->chain) { /* automatic_types messes up chrg and lig */
      if (a->chrg=='?') a->flg|=1<<tCFLG;
      if (a->lig=='?')  a->flg|=1<<tLFLG;
    }
    nunk=automatic_types(&ca);
    for (a=(Atom *)ca.s;mvld(a);a=(Atom *)a->chain) {
      if (a->flg&(1<<tCFLG)) a->chrg='?';
      a->flg&=~(1<<tCFLG);
      if (a->flg&(1<<tLFLG)) a->lig='?';
      a->flg&=~(1<<tLFLG);
    }
    prtGraphics("%d atoms assigned.\n",ca.n-nunk);
    if (nunk>0) err_set(1,"%d atoms (out of %d) could not be assigned.",nunk,ca.n);
    clear_node_chain(&ca);
  }else{
    err_set(1,"invalid sketch-root.");
  }
}
