#include <stdio.h>
#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "isosurface.h"
#include "surface.h"
#include "graphics.h"

extern int graphix;
extern Node *DefaultNode;
extern Dspobj *DspRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define ABOVE   1
#define BELOW  -1
#define XYZPLANES 7
#define NO_COLOR 0

static Grid *surface_grid(Grid *G,Surface **S);
static Surface *surface_contour(Grid *G);
static void box_attributes(void);
static void update_box(Grid *G);

static int box_width=1;
static int box_color=3;
static int box_on=FALSE;

static int inSide=ABOVE;
static int contColor=3;
static int contType=SURF_TYPE_TRG;
static float contLevel=0.0;
static float contTrans=ISOS_DFLT;

/*========================================================================*/

void cmd_surface(void)
{
  Surface *S=(Surface *)NULL;
  Grid *G=(Grid *)NULL;
  char *main_keys=
    "T!Surface "
    "on off toggle delete ' ' color style transparency width ' '"
    "clip grid read write ' ' command end";
  Token *key;

  key=(Token *)NULL;
  for(;;) {
    key=free_Token(key);
    key=Get_Kwd("Surface> ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("end")) {
	break;
      }else if (key_is("on")) {
	cmd_surface_status("on");
      }else if (key_is("off")) {
	cmd_surface_status("off");
      }else if (key_is("toggle")) {
	cmd_surface_status("toggle");
      }else if (key_is("delete")) {
	cmd_surface_delete();
      }else if (key_is("color")) {
	cmd_surface_color();
      }else if (key_is("style")) {
	cmd_surface_style();
      }else if (key_is("transparency")) {
	cmd_surface_transparency();
      }else if (key_is("width")) {
	cmd_surface_linewidth();
      }else if (key_is("clip")) {
	cmd_surface_clip();
      }else if (key_is("grid")) {
	G=surface_grid(G,&S);
      }else if (key_is("read")) {
	S=freeSurface(S);
	S=cmd_read_surface();
	cmd_add_surface(S);
      }else if (key_is("write")) {
	cmd_write_surface(S);
      }else if (key_is("command")) {
	command_loop("Wit!P(Surface)> ");
      }
    }else if (key->status==TOK_NOK) err_set(1,"bad surface option: %s",key->text);
    /*    if (key->status==TOK_EOC||key->status==TOK_ABORT) break; */
    if (end_parse()!=TOK_OK) err_set(1,"extra characters at end of command ignored.");
    start_parse();
    err_dump();
  }
  key=free_Token(key);
  if (mvld(S)) freeSurface(S);
  G=freeGrid(G);
}
/*-------------------- cmd_contour_delete -----------------------------*/

void cmd_surface_delete(void)
{
  Token *key;

  key=(Token *)NULL;

  for(;;) {
    key=free_Token(key); err_dump();
    key=Get_Objects_M(DSP_CNT|DSP_TRG,"delete surface object(s): ",";","'t!delete surface:' 't! '");
    if (key->status==TOK_OK) {
      del_node_chain((Nodechain *)key->ptr);
      free_node_chain((Nodechain *)key->ptr);
    }else if (key->status==TOK_NOK) continue;
    else break;
  }
  free_Token(key);
}
/*-------------------- cmd_contour_delete -----------------------------*/

void cmd_surface_clip(void)
{
  Token *key=(Token *)NULL;
  Nodechain *c;
  Dspobj *obj=NULL;
  Surface *S;
  int flags;
  float cutoff;

  key=free_Token(key); err_dump();
  key=Get_Objects_M(DSP_TRG,"clip surface object: ",";","'t!clip surface:' 't! '");
  if (key->status==TOK_OK) {
    c=(Nodechain *)key->ptr;
    if (c->n!=1) {
      key->status=TOK_NOK;
      err_set(1,"clip surface: object selection for clipping must be unique.");
    }else obj=(Dspobj *)c->s;
    clear_node_chain(c);
    free_node_chain(c);
    if (key->status==TOK_OK) {
      S=getSurfaceFromServer(obj,&flags);
      if (S) {
	key=free_Token(key);
	key=Get_Kwd("clipping method: ",";","atoms largest");
	if (key->status==TOK_OK) {
	  if (key_is("atoms")) {
	    key=free_Token(key);
	    key=Get_Float_M("clipping at cutoff ",";","1.0 2.0 3.0 4.0 5.0 6.0 7.0 8.0 9.0 10 15 20 30 40 50");
	    if (key->status==TOK_OK) {
	      cutoff=key->f;
	      key=free_Token(key);
	      key=Get_Atoms_D("atoms: ",";");
	      if (key->status==TOK_OK) {
		c=(Nodechain *)key->ptr;
		clipSurfaceAtoms(S,c,cutoff);
		replaceSurface(obj,S);
		clear_node_chain(c);
		free_node_chain(c);
	      }else if (key->status==TOK_NOK) err_set(1,"invalid atom selection: %s",key->text);
	    }else if (key->status==TOK_NOK) err_set(1,"bad floating point number: %s",key->text);
	  }else if (key_is("largest")) {
	    clipSurfaceLargest(S);
	    replaceSurface(obj,S);
	  }
	}else if (key->status==TOK_NOK)  err_set(1,"unknown clipping method: %s",key->text);

	S=freeSurface(S);
      }else err_set(0,"failed to retrieve surface %s from server",key->text);
    }
  }
  free_Token(key);
}
/*-------------------- cmd_contour_color -----------------------------*/

void cmd_surface_color(void)
{
  int color,type;
  Token *key;
  Nodechain *c;
  Node *o;
  char *menu1="'T!face:' inside outside ' ' 'T!outside color:'";
  char *menu2="'T!face:' inside outside ' ' 'T!inside color:' none";
  char *prmt1="color (outside): ";
  char *prmt2="color (inside): ";
  char *prmt,*menu,*face;

  menu=menu1;
  prmt=prmt1;
  face="DSPC";
  type=DSP_TRG|DSP_CNT;
  key=(Token *)NULL;
  
  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Color_M(prmt,";",menu);
    if (key->status==TOK_OK) {
      color=key->n;
      if (color<0) {
	key=Prs_Kwd(key,"inside outside none");
	if (key->status==TOK_OK) {
	  if (key_is("outside")) {
	    menu=menu1;
	    prmt=prmt1;
	    face="DSPC";
	    type=DSP_TRG|DSP_CNT;
	    color=-2;
	  }else if (key_is("inside")) {
	    menu=menu2;
	    prmt=prmt2;
	    face="SBFC";
	    type=DSP_TRG;
	    color=-2;
	  }else if (key_is("none")) {
	    color=-1;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad color selection: %s",key->text);
	}
      }
      if (color>-2) {
	key=free_Token(key);
	key=Get_Objects(type,"change color for surface object(s):",";");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (o=c->s;mvld(o);o=o->chain) {
	    if (is_dsp(o)&&cPtr(((Dspobj *)o)->cross)) {
	      if (((Dspobj *)o)->dsptype==DSP_CNT) {
		if (color>0) {
		  send_bheader(graphix,"DSPC",sizeof(int)+sizeof(o));
		  send_bint(graphix,color);
		  send_bptr(graphix,o);
		}
	      }else if (((Dspobj *)o)->dsptype==DSP_TRG) {
		send_bheader(graphix,face,sizeof(int)+sizeof(o));
		send_bint(graphix,color);
		send_bptr(graphix,o);
	      }
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"contour object not found: %s",key->text);
	}
      }
    }else if (key->status==TOK_NOK) err_set(1,"bad color selection: %s",key->text);
    else break;
  }
  free_Token(key);
}
/*-------------------- cmd_contour_surface -----------------------------*/

void cmd_surface_linewidth(void)
{
  int lw,n;
  Token *key;
  Nodechain *c;
  Node *o;

  key=(Token *)NULL;

  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Int_M("linewidth: ",";","'t!linewidth:' 't! ' 1 2 3 4 5 6 7 8 9 10");
    if (key->status == TOK_OK) {
      lw=(key->n < 1 ? 1 : (key->n >33 ?  32 : key->n));
      key=free_Token(key);
      key=Get_Objects(DSP_TRG|DSP_CNT,"change linewidth for contour object(s):",";");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (n=0,o=c->s;mvld(o);o=o->chain) {
	  if (is_dsp(o)) if (cPtr(((Dspobj *)o)->cross)) n++;
	}
	if (n>0) {
	  send_bheader(graphix,"SOLW",sizeof(int)+n*sizeof(o));
	  send_bint(graphix,lw);
	  for (n=0,o=c->s;mvld(o);o=o->chain) {
	    if (is_dsp(o)) {
	      if (cPtr(((Dspobj *)o)->cross)) send_bptr(graphix,o);
	    }
	  }
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"contour object not found: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad linewidth selection: %s",key->text);
    }else break;
  }
  free_Token(key);
}
/*-------------------- cmd_surface_style -----------------------------*/

void cmd_surface_style(void)
{
  int style=0,n,wrn;
  Token *key;
  Nodechain *c;
  Node *o;

  key=(Token *)NULL;

  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("style: ",";","'t!surface style:' 't! ' rendered mesh normals dots");
    if (key->status == TOK_OK) {
      if (key_is("rendered")) style=1;
      else if (key_is("mesh")) style=2;
      else if (key_is("normals")) style=3;
      else if (key_is("dots")) style=4;
      key=free_Token(key);
      key=Get_Objects_M(DSP_TRG,"change style for surface object(s):",";","'t!surface:' 't! '");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (n=0,o=c->s;mvld(o);o=o->chain) {
	  if (is_dsp(o)) if (cPtr(((Dspobj *)o)->cross)) n++;
	}
	if (n>0) {
	  send_bheader(graphix,"TRGS",sizeof(int)+n*sizeof(o));
	  send_bint(graphix,style);
	  for (wrn=0,n=0,o=c->s;mvld(o);o=o->chain) {
	    if (is_dsp(o)) {
	      if (cPtr(((Dspobj *)o)->cross)) {
		send_bptr(graphix,o);
		if (((Dspobj *)o)->dsptype!=DSP_TRG) wrn++;
	      }
	    }
	  }
	  if (wrn>0) err_set(1,"note: only triangulated surfaces have style.");
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"contour object not found: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad style selection: %s",key->text);
    }else break;
  }
  free_Token(key);
}
/*-------------------- cmd_surface_transparency -----------------------------*/

void cmd_surface_transparency(void)
{
  int style=0,n,wrn;
  Token *key;
  Nodechain *c;
  Node *o;

  key=(Token *)NULL;

  for (;;) {
    key=free_Token(key); err_dump();
    key=Get_Kwd("transparency level: ",";","'t!transparency:' 't! ' opaque low medium high");
    if (key->status == TOK_OK) {
      if (key_is("opaque")) style=0;
      else if (key_is("low")) style=1;
      else if (key_is("medium")) style=2;
      else if (key_is("high")) style=3;
      key=free_Token(key);
      key=Get_Objects_M(DSP_TRG,"change transparency for contour object(s):",";","'t!surface:' 't! '");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (n=0,o=c->s;mvld(o);o=o->chain) {
	  if (is_dsp(o)) if (cPtr(((Dspobj *)o)->cross)) n++;
	}
	if (n>0) {
	  send_bheader(graphix,"TRGT",sizeof(int)+n*sizeof(o));
	  send_bint(graphix,style);
	  for (wrn=0,n=0,o=c->s;mvld(o);o=o->chain) {
	    if (is_dsp(o)) {
	      if (cPtr(((Dspobj *)o)->cross)) {
		send_bptr(graphix,o);
		if (((Dspobj *)o)->dsptype!=DSP_TRG) wrn++;
	      }
	    }
	  }
	  if (wrn>0) err_set(1,"note: only triangulated surfaces have transparency.");
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) {
	err_set(1,"contour object not found: %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad transparency level selection: %s",key->text);
    }else break;
  }
  free_Token(key);
}
	
/*-------------------- cmd_contour_status -----------------------------*/

void cmd_surface_status(char *op)
{
  int status,n;
  Token *key;
  Nodechain *c;
  Node *o;
  char *menu="'T!surface status' on off toggle 'h!end' ' ' T!surface:";

  key=(Token *)NULL;
  status=(streq("on",op)?1:(streq("off",op)?0:-1));
  for(;;) {
    key=free_Token(key); err_dump();
    key=Get_Objects_M(DSP_CNT|DSP_TRG,"change display status for surface object(s): ",";",menu);
    if (key->status==TOK_NOK) {
      key=Prs_Kwd(key,"on off toggle end") ;
      if (key->status==TOK_OK) {
	if (key_is("end")) break;
	status=1-key->n;
	continue;
      }
    }
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      for (n=0,o=c->s;mvld(o);o=o->chain) {
	if (is_dsp(o)) {
	  if (cPtr(((Dspobj *)o)->cross)) n++;
	  switch (status) {
	  case 1:    /* on */
	    ((Dspobj *)o)->status|=OBJ_FLG_DSP;
	    break;
	  case 0:    /* off */
	    ((Dspobj *)o)->status&=~OBJ_FLG_DSP;
	    break;
	  case -1:   /* toggle */
	    if (((Dspobj *)o)->status&OBJ_FLG_DSP) ((Dspobj *)o)->status&=~OBJ_FLG_DSP;
	    else ((Dspobj *)o)->status|=OBJ_FLG_DSP;
	    break;
	  }
	}
      }
      if (n>0) {
	send_bheader(graphix,"DSPO",sizeof(int)+n*sizeof(o));
	send_bint(graphix,status);
	for (n=0,o=c->s;mvld(o);o=o->chain) {
	  if (is_dsp(o)) {
	    if (cPtr(((Dspobj *)o)->cross)) send_bptr(graphix,o);
	  }
	}
      }
      clear_node_chain(c);
      free_node_chain(c);
    }else if (key->status==TOK_NOK) {
      err_set(1,"no such surface object: %s",key->text);
    }else break;
  }
  free_Token(key);
}
/*-------------------- cmd_read_surface --------------------*/

Surface *cmd_read_surface(void)
{
  Token *key=(Token *)NULL;
  UCpipe *F=(UCpipe *)NULL;
  Surface *S=(Surface *)NULL;;

  key=Get_File((char *)NULL,".trg","file: ",";");
  if (key->status==TOK_OK) {
    F=(UCpipe *)key->ptr;
    S=readSurface(F->fd);
    UCpclose(F);
  }else if (key->status==TOK_NOK) err_set(1,"failed to open %s for read.",key->text);
  return(S);
}
/*-------------------- cmd_write_surface --------------------*/

int  cmd_write_surface(Surface *S)
{
  Token *key=(Token *)NULL;
  char filename[4096],*ext;
  FILE *out=(FILE *)NULL;
  int status=1;

  if (!mvld(S)||
      (S->type==SURF_TYPE_MSH&&(S->nv==0||!mvld(S->v)))||
      (S->type==SURF_TYPE_TRG&&(S->nv==0||!mvld(S->v)||S->nt==0||!mvld(S->t)))) {
    err_set(1,"no surface to write.");
    return(status);
  }

  surface_info(S);
  key=Get_String("write surface to file: ",";");
  if (key->status==TOK_OK) {
    ext=".trg";
    extfil(filename,key->text,ext,FALSE);
    out=fopen(filename,"w");
    if (out) {
      status=writeSurface(out,S);
      if (status) err_set(1,"error writing surface to file %s, status=%d.",filename,status);
      else prtGraphics("surface written to file %s.\n",filename);
      fclose(out);
    }else err_set(1,"could not open surface file %s for write.",filename);
  }else if (key->status==TOK_NOK) err_set(1,"bad filename token: %s",key->text);
  free_Token(key);
  return(status);
}
/*-------------------- cmd_add_surface --------------------*/

Dspobj *cmd_add_surface(Surface *S)
{
  Token *key=(Token *)NULL;
  char *menu=(char *)NULL;
  Dspobj *glob,*surf=(Dspobj *)NULL;

  if (!mvld(S)||
      (S->type==SURF_TYPE_MSH&&(S->nv==0||!mvld(S->v)))||
      (S->type==SURF_TYPE_TRG&&(S->nv==0||!mvld(S->v)||S->nt==0||!mvld(S->t)))) {
    err_set(1,"no surface to add.");
    return(surf);
  }

  key=free_Token(key);
  menu=Make_GlblNameList("/",DspRoot);

  glob=(Dspobj *)NULL;
  key=Get_DspGlbl_M("object: ",";",menu);
  if (key->status==TOK_OK) {
    glob=(Dspobj *)key->ptr;
    if (S->color==NO_COLOR) {
      key=free_Token(key);
      key=Get_Color("color: ",";");
      if (key->status==TOK_OK) S->color=key->n;
      else if (key->status==TOK_NOK) err_set(1,"bad color: %s",key->text);
    }
    if (key->status==TOK_OK) {
      key=free_Token(key);
      key=Get_String("name: ",";");
      if (key->status==TOK_OK) surf=addSurfaceToDisplay(S,glob,key->text);
      else if (key->status==TOK_NOK) err_set(1,"bad name: %s",key->text); 
    }
  }else if (key->status==TOK_NOK) err_set(1,"object not found");
  key=free_Token(key);
  free_menu(menu);
  return(surf);
}
/*-------------------- surface_grid --------------------*/

static Grid *surface_grid(Grid *G,Surface **S)
{
  char *keys=
    "T!Grid "
    "h!dump create delete compute contour ' ' box ' ' "
    "read write ' ' end";
  Token *key;
  Surface *s=(Surface *)NULL;

  key=(Token *)NULL;
  grid_report(G);
  for (;;) {
    err_dump();
    key=Get_Kwd("grid> ",";",keys);
    if (key->status==TOK_OK) {
      if (key_is("box")) box_attributes();
      else if (key_is("dump"))       grid_dump(G);
      else if (key_is("create"))   G=grid_create(G);
      else if (key_is("compute"))    grid_compute(G);
      else if (key_is("delete"))   G=grid_delete(G);
      else if (key_is("read"))     G=grid_read(G);
      else if (key_is("write"))      grid_write(G);
      else if (key_is("contour"))  s=surface_contour(G);
      else if (key_is("end")) break;
      update_box(G);
    }else if (key->status==TOK_NOK) err_set(1,"bad grid keyword: %s",key->text);
    if (key->status==TOK_EOC||key->status==TOK_ABORT) break;
    if (end_parse()!=TOK_OK) {
      err_set(1,"extra characters at end of command ignored.");
    }
    start_parse();
  }
  if (mvld(s)) {
    if (S) {
      freeSurface(*S);
      *S=s;
    }else freeSurface(s);
  }
  key=free_Token(key);
  return(G);
}
/*-------------------- surface_contour --------------------*/

Surface *surface_contour(Grid *G)
{
  char *keys=
    "'T!contour grid:' "
    "color inside level type ' ' add h!end";
  Surface *S=(Surface *)NULL;
  Token *key=(Token *)NULL;
  int add=FALSE;
  float surf,vol;
  
  if (mvld(G)) {
    for (;;) {
      key=Get_Kwd("grid contour> ",";",keys);
      if (key->status==TOK_OK) {
	if (key_is("color")) {
	  default_token("%s",ColMnem(contColor));
	  key=free_Token(key);
	  key=Get_Color("contour color: ",";");
	  if (key->status==TOK_OK) contColor=key->n;
	  else if (key->status==TOK_NOK) err_set(1,"bad color: %s",key->text);
	}else if (key_is("inside")) {
	  default_token((inSide==ABOVE?"above":"below"));
	  key=free_Token(key);
	  key=Get_Kwd("inside above/below contour level? ",";","above below");
	  if (key->status==TOK_OK) inSide=(key->n==0?ABOVE:BELOW);
	  else if (key->status==TOK_NOK) err_set(1,"bad option: %s",key->text);
	}else if (key_is("level")) {
	  default_token("%0.3f",contLevel);
	  key=free_Token(key);
	  key=Get_Float_M("contour level: ",";","-10 -5 -4 -3 -2 -1 0 1 2 3 4 5 10");
	  if (key->status==TOK_OK) contLevel=key->f;
	  else if (key->status==TOK_NOK) err_set(1,"bad contour level: %s",key->text);
	}else if (key_is("type")) {
	  default_token((contType==SURF_TYPE_TRG?"solid":"lines"));
	  key=free_Token(key);
	  key=Get_Kwd("surface type: ",";","lines h!mesh solid");
	  if (key->status==TOK_OK) contType=(key->n<2?SURF_TYPE_MSH:SURF_TYPE_TRG);
	  else if (key->status==TOK_NOK) err_set(1,"bad type: %s",key->text);
	}else if (key_is("add")) {
	  add=TRUE;
	  break;
	}else if (key_is("end")) break;
      }else if (key->status==TOK_NOK) err_set(1,"bad contour option: %s",key->text);
      if (key->status==TOK_EOC||key->status==TOK_ABORT) break;
    }
    if (add) {
      switch (contType) {
      case SURF_TYPE_MSH:
	S=isoMesh(G,contLevel,inSide,WNP_GRD_MISSING,contTrans,XYZPLANES,TRUE);
	if (mvld(S)) {
	  S->color=contColor;
	  surface_info(S);
	}
	cmd_add_surface(S);
	break;
      case SURF_TYPE_TRG:
	S=isoSurface(G,contLevel,inSide,WNP_GRD_MISSING,contTrans,TRUE);
	if (mvld(S)) {
	  S->color=contColor;
	  SurfAndVol(S,&surf,&vol);
	  prtGraphics("isosurface: Surface area = %0.3f,"
		      "Volume inside surface = %0.3f\n",surf,vol);
	}
	cmd_add_surface(S);
	break;
      }
    }
  }else err_set(1,"no grid to contour.");
  free_Token(key);
  return(S);
}

/*-------------------- box_attributes --------------------*/

static void box_attributes(void)
{
  Token *key=(Token *)NULL;

  key=Get_Kwd("box attribute: ",";","'t!grid box:' 't! ' on off ' ' color width");
  if (key->status==TOK_OK) {
    if (key_is("on")) box_on=TRUE;
    else if (key_is("off")) box_on=FALSE;
    else if (key_is("color")) {
      key=(Token *)NULL;
      key=Get_Color("box color: ",";");
      if (key->status==TOK_OK) box_color=key->n;
      else if (key->status==TOK_NOK) err_set(1,"bad color token: %s",key->text);
    }else if (key_is("width")) {
      key=(Token *)NULL;
      key=Get_Int_M("box line width: ",";","1 2 3 4 5 6 7 8");
      if (key->status==TOK_OK) box_width=key->n;
      else if (key->status==TOK_NOK) err_set(1,"bad linewidth : %s",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad box attribute: %s",key->text);
  free_Token(key);
}
/*-------------------- update_box --------------------*/

static void update_box(Grid *G)
{
  if (G) {
    if (box_on!=G->boxon||
	box_color!=G->boxcolor||
	box_width!=G->boxwidth) {
      G->boxon=box_on;
      G->boxcolor=box_color;
      G->boxwidth=box_width;
      grid_box_make(G);
    }
  }
}
