#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *DefaultNode;

#define TRUE 1
#define FALSE 0

#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL
#define key_is(a) (streq(a,key->key))

static int biosym_file_version=3;   /* default Biosym file version */

/*========================================================================*/
void cmd_write(Node *root,int types)
{
  Token *key,*filename,*r;
  char *menu;
  int ft=0;
  int append=FALSE;
  Nodechain *c;

  menu=Make_FileTypeList("append ''",types&FT_CAN_WRITE);
  key=Get_Kwd("Filetype: ",";",menu);
  if (menu) free_menu(menu);
  ft=FT_NONE;
  if (key->status == TOK_OK) {
    if (key_is("append")) {
      append=TRUE;
      menu=Make_FileTypeList("",types&FT_CAN_APPEND);
      key=free_Token(key);
      key=Get_Kwd("Filetype: ",";",menu);
      if (menu) free_menu(menu);
      if (key->status==TOK_NOK) {
	menu=Make_FileTypeList("",types&FT_CAN_WRITE);
	key=Prs_Kwd(key,menu);
	if (menu) free(menu);
	if (key->status==TOK_OK) {
	  ft=ft_number(key->key);
	  key->status=TOK_NOK;
	}
      }
    }
    if (key->status==TOK_OK) {
      ft=ft_number(key->key);
      if (ft==FT_NONE) key->status=TOK_NOK;
    }
  }
  if (key->status==TOK_OK) {
    filename=Get_String("File: ",";");
    if (filename->status == TOK_OK) { 
      if (!mvld(root)&&ft!=FT_SESS) {
	if (mvld(menu)) free(menu);
	menu=Make_TopMolNameList("./ /",DefaultNode);
	r=Get_Mols_M("molecule: ",";",menu); free_menu(menu);
	send_bheader(graphix,"ARMK",0);
	if (r->status==TOK_OK) {
	  c=(Nodechain *)r->ptr;
	  if (c->n>1) {
	    err_set(1,"molecule selection is ambiguous.");
	    r->status=TOK_NOK;
	  }else if (c->n<1) {
	    err_set(1,"no molecule selected.");
	    r->status=TOK_OK;
	  }else{
	    root=c->s;
	  }
	  clear_node_chain(c); free_node_chain(c);
	}
	if (r->status==TOK_NOK) {
	  err_set(1,"no such molecule: %s",r->text);
	}
	r=free_Token(r);
      }
      if (ft==FT_SESS) saveSession(filename->text);
      else if (is_mol(root)) WriteMolecule(root,filename->text,ft,append);
    }
    free_Token(filename);
  }else if (key->status == TOK_NOK) {
    if (ft==FT_NONE) err_set(1,"unknown filetype: %s",key->text);
    else err_set(1,"can't append to files of type %s",key->text);
  }
  free_Token(key);
}

int  WriteMolecule(Node *m,char *filename,int type,int append)
{
  int status;
  char hdr1[256],hdr2[256],buff[1024];
  Atom *a;
  
  err_clear();
  status=0;
  if (mvld(m)) ForAtoms(a,m) a->flg|=(1<<1);
  switch (type) {
  case FT_CAR:
    sprintf(hdr1,"!BIOSYM archive %d",biosym_file_version);
    sprintf(hdr2,"Generated by WITNOTP");
    status=write_bio(filename,hdr1,hdr2,m);
    break;
  case FT_MACCS:
    status=write_maccs(filename,m,append);
    break;
  case FT_MMOD:
    status=write_mmod(filename,m,append);
    break;
  case FT_MOL:
    status=write_mol(filename,m,append);
    break;
  case FT_MOL2:
    status=write_mol2(filename,m,append);
    break;
  case FT_PDB:
    status=write_pdb(filename,m);
    break;
  case FT_SHX:
    status=write_shelx(filename,m);
    break;
  case FT_XPLOR:
    status=write_xplor(filename,m);
    break;
  case FT_XPDB:
    status=write_xplor_pdb(filename,m,append);
    break;
  default: break;
  }
  if (mvld(m)) ForAtoms(a,m) a->flg&=~(1<<1);
  if (status) {
    nodeptr_to_name(buff,m);
    err_set(1,"error writing molecule %s to file %s,status = %d\n",buff,filename,status);
  }else if (!streq(filename,"-")) {
    nodeptr_to_name(buff,m);
    prtGraphics("molecule %s written to %s\n",buff,filename);
  }
  return(status);
}

/*-------------------- set_biosym_version --------------------*/

void set_biosym_version(char *v)
{
  if (streq(v,"1")) biosym_file_version=1;
  else if (streq(v,"2")) biosym_file_version=2;
  else if (streq(v,"3")) biosym_file_version=3;
  else setBiosymFF(v);
}
