#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "strutil.h"
#include "errhand.h"

#define YES 1
#define NO  0

#define MAXBUF 256

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define ring_cand(a) (mvld(a->atp)&&a->atp->oop&&(a->flg&(1<<1))&&a->up==n&&a->nbnd>1)
#define psf_r_cand(a) (a->atp->oop&&a->nbnd>1&&a->flg&(1<<1))

static int linAtom(Atom *a,Bond *b,Bond *c);
static void chrm_type(char *type,Atomdef *t,int mode);


int write_charmm_rtf(char *file,Node *mol)
{
  FILE *out;
  Atom *a,*b,*p[3],*at[10],*bt[10];
  Node *n;
  int status,i,j,k,ki,l,li,nb,ni,Z,nt[6],R;
  char filename[1024],rname[5],rnum[5],*grp;
  float q,tau;

  /* open output file */

  extfil(filename,file,".rtf",NO);
  if ((out=fopen(filename,"w"))==NULL) {
    err_set(1,"write_charmm_rtf: could not open %s for write\n",filename);
    return(1);
  }else status=0;
  

  /* write header records... */

  fprintf(out,"* CHARMM RTF file generated by WITNOTP\n");
  fprintf(out,"*\n");
  fprintf(out,"   22    0\n\n");
  fprintf(out,"AUTOGENERATE ANGLES DIHEDRALS\n");
  fprintf(out,"DEFA FIRST NONE LAST NONE\n");
  fprintf(out,"\n");

  /* write each residue... */

  ForAtoms(a,mol) {
    n=a->up;
    if (!(a->flg&(1<<1))) continue;
    make_xpl_resnam(n->name,rname,rnum);
    q=0; ForAtoms(a,n) if (a->flg&(1<<1)) q+=a->q;
    fprintf(out,"RESI %-8s %0.3f\n",rname,q);
    grp=(char *)NULL;

    /* GROUP and ATOM records... */

    ForAtoms(a,n) {
      if (!(a->flg&(1<<1))||!mvld(a->atp)) continue;
      if (!streq(grp,a->grp)) {
	fprintf(out,"  GROUP\n");
	grp=a->grp;
      }
      fprintf(out,"    ATOM %-6s %-6s %8.3f\n",a->name,(mvld(a->atp)?a->atp->name:"unk"),a->q);
    }

    /* BOND records */

    nb=0;
    ForAtoms(a,n) {
      if (!(a->flg&(1<<1))||!mvld(a->atp)) continue;
      for (i=0;i<a->nbnd;i++) {
	b=neighbour(i,a);
	if (b==a->bnd[i]->f || b->up!=n || !(b->flg&(1<<1))) continue;
	if (nb%5==0) fprintf(out,"  BOND %-6s %-6s",a->name,b->name);
	else fprintf(out,"  %-6s %-6s",a->name,b->name);
	if (++nb%5==0) fprintf(out,"\n");
      }
    }
    if (nb%5!=0) fprintf(out,"\n");

    /* IMPH improper records: oop, extended sp3, planar rings */

    ni=0;
    ForAtoms(a,n) {
      if (!(a->flg&(1<<1))||!mvld(a->atp)) continue;
      if (a->oop && a->nbnd==3) {   /* add improper if oop flag is set */
	for (i=0;i<3;i++) {
	  p[i]=neighbour(i,a);
	  if (p[i]->up!=n || !(p[i]->flg&(1<<1))) break;
	}
	if (i==3) {
	  if (ni%2==0) fprintf(out,"  IMPH %-6s %-6s %-6s %-6s",
			       a->name,p[0]->name,p[1]->name,p[2]->name);
	  else fprintf(out,"      %-6s %-6s %-6s %-6s",
		       a->name,p[0]->name,p[1]->name,p[2]->name);
	  if (++ni%2==0) fprintf(out,"\n");
	}
      }else if (a->atp->z==6&&a->nbnd==3&&sane_formlig(a)==4) { /* add improper for extended Csp3 */
	for (i=0;i<3;i++) {
	  p[i]=neighbour(i,a);
	  if (p[i]->up!=n || !(p[i]->flg&(1<<1))) break;
	}
	if (i==3) {
	  tau=ctorsd(a->x,p[0]->x,p[1]->x,p[2]->x);
	  if (tau>0) {   /* assume that equilibrium angle is positive */
	    if (ni%2==0) fprintf(out,"  IMPH %-6s %-6s %-6s %-6s",
			       a->name,p[0]->name,p[1]->name,p[2]->name);
	    else fprintf(out,"      %-6s %-6s %-6s %-6s",
		       a->name,p[0]->name,p[1]->name,p[2]->name);
	  }else{
	    if (ni%2==0) fprintf(out,"  IMPH %-6s %-6s %-6s %-6s",
			       a->name,p[1]->name,p[0]->name,p[2]->name);
	    else fprintf(out,"      %-6s %-6s %-6s %-6s",
		       a->name,p[1]->name,p[0]->name,p[2]->name);
	  }
	  if (++ni%2==0) fprintf(out,"\n");
	}	
      }
    }
    ForAtoms(at[0],n) {                   /* ring planarity */
      if (!ring_cand(at[0])) continue;
      for (j=0;j<at[0]->nbnd;j++) {
	if (at[0]==at[0]->bnd[j]->t) continue;
	if (!ring_cand(at[0]->bnd[j]->t)) continue;
	at[1]=at[0]->bnd[j]->t; nt[1]=0;
	for (R=1;R>0;R--) {
	  for (;nt[R]<at[R]->nbnd;) {
	    k=nt[R];nt[R]+=1;
	    if (at[R]->bnd[k]<=at[0]->bnd[j]) continue;
	    b=neighbour(k,at[R]);
	    if (!ring_cand(b)) continue;
	    if (b==at[0]) {           /* found ring of size R+1 */
	      if (R+1>4) {
		for (i=0;i<=R;i++) {
		  if (ni%2==0) fprintf(out,"  IMPH %-6s %-6s %-6s %-6s",
				       at[i%(R+1)]->name,at[(i+1)%(R+1)]->name,
				       at[(i+2)%(R+1)]->name,at[(i+3)%(R+1)]->name);
		  else fprintf(out,"      %-6s %-6s %-6s %-6s",
			       at[i%(R+1)]->name,at[(i+1)%(R+1)]->name,
			       at[(i+2)%(R+1)]->name,at[(i+3)%(R+1)]->name);
		  if (++ni%2==0) fprintf(out,"\n");
		}
	      }
	    }else if (R+1<6) {
	      for (i=1;i<R;i++) if (b==at[i]) break;
	      if (i==R) {
		at[++R]=b;
		nt[R]=0;
	      }
	    }
	  }
	}
      }
    }
    if (ni%2!=0) fprintf(out,"\n");
    
    /* HB donor and acceptor records (needs refinement...) */

    ForAtoms(a,n) {
      if (!(a->flg&(1<<1))||!mvld(a->atp)) continue;
      if (a->atp->hbdon) {
	for (i=0;i<a->nbnd;i++) {
	  b=neighbour(i,a);
	  if (b->up!=n || !(b->flg&(1<<1))) continue;
	  if (!mvld(b->atp) || b->atp->z!=1) continue;
	  fprintf(out,"  DONO %-6s %-6s\n",b->name,a->name);
	}
      }
      if (a->atp->hbacc) {
	if (a->atp->z==7&&((a->oop&&eff_lig(a)==3)||eff_lig(a)==4)) continue;
	for (Z=0,i=0,j=0;i<a->nbnd;i++) {
	  b=neighbour(i,a);
	  if (b->up!=n || !(b->flg&(1<<1))) continue;
	  if (!mvld(b->atp)) continue;
	  if (Z<b->atp->z) {
	    Z=b->atp->z;
	    j=i;
	  }
	}
	if (Z>0) fprintf(out,"  ACCE %-6s %-6s\n",a->name,neighbour(j,a)->name);
      }
    }
 
    /* internal coordinate records... */

    ForAtoms (a,n) {
      if (!(a->flg&(1<<1))||!mvld(a->atp)) continue;
      for (i=0;i<a->nbnd;i++) {
	b=neighbour(i,a);
	if (b==a->bnd[i]->f||b->up!=n||!(b->flg&1<<1)) continue;
	for (k=0,j=0;j<a->nbnd;j++) {
	  at[k]=neighbour(j,a);
	  if (at[k]==b||at[k]->up!=n||!(at[k]->flg&1<<1)) continue;
	  if (k<10) k++;
	  else break;
	}
	if (k==0) continue;
	for (l=0,j=0;j<b->nbnd;j++) {
	  bt[l]=neighbour(j,b);
	  if (bt[l]==a||bt[l]->up!=n||!(bt[l]->flg&1<<1)) continue;
	  if (l<10) l++;
	  else break;
	}
	if (l==0) continue;
	fprintf(out,"  IC   %-7s %-7s  %-7s %-7s %5.3f %7.2f %6.1f %7.2f %5.3f\n",
		at[0]->name,a->name,b->name,bt[0]->name,
		cdist(at[0]->x,a->x),cangled(at[0]->x,a->x,b->x),
		ctorsd(at[0]->x,a->x,b->x,bt[0]->x),
		cangled(a->x,b->x,bt[0]->x),cdist(b->x,bt[0]->x));
	for (ki=0;ki<k-1;ki++) {
	  for (li=ki+1;li<k;li++) {
	    fprintf(out,"  IC   %-7s %-7s *%-7s %-7s %5.3f %7.2f %6.1f %7.2f %5.3f\n",
		    b->name,at[ki]->name,a->name,at[li]->name,
		    cdist(b->x,a->x),cangled(b->x,a->x,at[ki]->x),
		    ctorsd(b->x,at[ki]->x,a->x,at[li]->x),
		    cangled(at[ki]->x,a->x,at[li]->x),cdist(a->x,at[li]->x));
	  }
	}
	for (ki=0;ki<l-1;ki++) {
	  for (li=ki+1;li<l;li++) {
	    fprintf(out,"  IC   %-7s %-7s *%-7s %-7s %5.3f %7.2f %6.1f %7.2f %5.3f\n",
		    a->name,bt[ki]->name,b->name,bt[li]->name,
		    cdist(a->x,b->x),cangled(a->x,b->x,bt[ki]->x),
		    ctorsd(a->x,bt[ki]->x,b->x,bt[li]->x),
		    cangled(bt[ki]->x,b->x,bt[li]->x),cdist(b->x,bt[li]->x));
	  }
	}
      }
    }

    /* all done for this residue... */
    a=(Atom *)n->lst;
  }

  /* done... */

  fprintf(out,"\nEND\n");
  fclose(out);

  return(status);
}
/*-------------------- write_charmm_psf --------------------*/

int write_charmm_psf(char *file,Node *m,int mode)
{
  /* write CharMM/PSF file (only "selected" atoms (i.e. flg_1 = 1) ) */

  FILE *out;
  int status,lev;
  long where;
  int i,j,k,l,ntit,nat,nbnd,nth,nphi,nimphi,ndon,nacc,nnb,ngrp,nt[6],R;
  char filename[256],chain_id[5],res_nam[5],res_nr[5],type[5],aname[5],buf[4096],*grp;
  Atom *a,*b,*n1,*n2,*t1,*t2,*t3,*t4,*at[6];
  Node *n;

  status=0;
  nat=0; nbnd=0; nth=0; nphi=0; nimphi=0; ndon=0; nacc=0; nnb=0;
  ngrp=0; ntit=0;

  /* count atoms, check atomnames, potential names etc... */

  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      nat++;
      a->seqnr=nat;
      if (!mvld(a->atp)) {
	err_set(1,"atom %s has no type assigned.",nodeptr_to_name(buf,(Node *)a));
	status=2;
      }
    }
  }
  if (status) return(status);

  /* open output file */

  extfil(filename,file,".psf",YES);
  if ((out=fopen(filename,"w"))==NULL) {
    err_set(1,"write_charmm_psf: could not open %s for write\n",filename);
    return(1);
  }

  /* write header records */

  fprintf(out,"PSF\n");
  ntit=1;
  fprintf(out,"\n%8d !NTITLE title records\n",ntit);
  fprintf(out," REMARKS CharMM structure file generated by WITNOTP\n");
  if (ntit==0) fprintf(out,"\n");

  /* write atom records */

  fprintf(out,"\n%8d !NATOM atoms\n",nat);

  ForAtoms(a,m) {
    for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
    if (lev>2) {
      Strcpy(chain_id,a->up->up->name);
      n=a->up;
    }else{
      sprintf(chain_id," ");
      n=a->up;
    }
    ForAtoms(b,n) {
      a=b;
      if (a->flg&(1<<1)) {
	make_xpl_resnam(a->up->name,res_nam,res_nr);
	Strcpy(aname,a->name); 
	chrm_type(type,a->atp,mode);
	fprintf(out,"%8d %-4s %-4s %-4s %-4s %-4s %14.6f %9.4f %11d\n",
		a->seqnr,chain_id,res_nr,res_nam,aname,type,a->q,a->atp->mass,0);
	for (k=0,j=0;j<a->nbnd;j++) {
	  if (a->bnd[j]->bo!=0) {
	    if (a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) nbnd++;
	    if (a->bnd[j]->f->flg&(1<<1)&&a->bnd[j]->t->flg&(1<<1)) k++;
	  }
	}
	nth+=(k*(k-1))/2;
      }
    }
  }
  if (nat==0) fprintf(out,"\n");

  /* bonds... */

  fprintf(out,"\n%8d !NBOND bonds\n",nbnd); k=0;
  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (j=0;j<a->nbnd;j++) {
	if (a->bnd[j]->bo!=0 && a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) {
	  fprintf(out,"%8d%8d",a->seqnr,a->bnd[j]->t->seqnr);
	  if ((++k)%4==0) fprintf(out,"\n");
	}
      }
    }
  }
  if (k%4!=0) fprintf(out,"\n");
  if (nbnd==0) fprintf(out,"\n");

  /* angles... */

  fprintf(out,"\n%8d !NTHETA angles\n",nth); k=0;
  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (j=1;j<a->nbnd;j++) {
	if (a->bnd[j-1]->bo==0) continue;
	n1=neighbour(j-1,a);
	if (n1->flg&(1<<1)) {
	  for (l=j;l<a->nbnd;l++) {
	    if (a->bnd[l]->bo==0) continue;
	    n2=neighbour(l,a);
	    if (n2->flg&(1<<1)) {
	      fprintf(out,"%8d%8d%8d",n1->seqnr,a->seqnr,n2->seqnr);
	      if ((++k)%3==0) fprintf(out,"\n");
	    }
	  }
	}
      }
    }
  }

  if (k%3!=0) fprintf(out,"\n");
  if (nth==0) fprintf(out,"\n");

  /* torsions, play dirty trick with ftell/fseek ... */

  where=ftell(out);
  fprintf(out,"\n%8d !NPHI torsions\n",nphi);  /* dymmy only */
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1))||a->nbnd<2) continue;
    for (j=0;j<a->nbnd;j++) {
      if (a->bnd[j]->bo!=0 && a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) {
	if (a->bnd[j]->t->nbnd<2) continue;
	t2=a; t3=a->bnd[j]->t;
	for (k=0;k<t2->nbnd;k++) {
	  if (t2->bnd[k]->bo==0 || t2->bnd[k]==a->bnd[j]) continue;
	  if (linAtom(a,a->bnd[j],a->bnd[k])) continue;    /* 1.12.97: avoid ill-defined dihedrals... */
	  t1=neighbour(k,t2); if (!(t1->flg&(1<<1))) continue;
	  for (l=0;l<t3->nbnd;l++) {
	    if (t3->bnd[l]->bo==0 || t3->bnd[l]==a->bnd[j]) continue;
	    if (linAtom(t3,a->bnd[j],t3->bnd[l])) continue;/* 1.12.97: avoid ill-defined dihedrals... */
	    t4=neighbour(l,t3);
	    if (!(t4->flg&(1<<1)) || t1==t4) continue;
	    fprintf(out,"%8d%8d%8d%8d",
		    t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr);
	    if ((++nphi)%2==0) fprintf(out,"\n");
	  }
	}
      }
    }
  }
  if (nphi%2 != 0) fprintf(out,"\n");
  if (nphi==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NPHI torsions\n",nphi);  /* the real number */
  fseek(out,0,2);
	    
  /* improper torsions (oop, planar rings and chiral [nyi]) */

  where=ftell(out);
  fprintf(out,"\n%8d !NIMPHI improper torsions\n",nimphi); /* only a dummy... */
  ForAtoms(a,m) {       /* oop */
    if (!(a->flg&(1<<1)) || a->oop==0 || a->nbnd<3) continue;
    at[0]=a;
    for (j=0,k=1;j<a->nbnd&&k<6;j++) {
      if (a->bnd[j]->bo!=0) {
	at[k]=neighbour(j,a);
	if (at[k]->flg&(1<<1)) k++;
      }
    }
    if (k==4) {
      fprintf(out,"%8d%8d%8d%8d",
	    at[0]->seqnr,at[1]->seqnr,at[2]->seqnr,at[3]->seqnr);
      if ((++nimphi)%2==0) fprintf(out,"\n");
    }
  }   
  ForAtoms(at[0],m) {   /* ring planarity for 5 and 6 membered rings */
    if (!psf_r_cand(at[0])) continue;
    for (j=0;j<at[0]->nbnd;j++) {
      if (at[0]->bnd[j]->bo==0) continue;
      at[1]=neighbour(j,at[0]); nt[1]=0;
      if (!psf_r_cand(at[1])) continue;
      if (at[1]->seqnr<at[0]->seqnr) continue;
      for (R=1;R>0;R--) {
	for (;nt[R]<at[R]->nbnd;) {
	  k=nt[R];nt[R]+=1;
	  if (at[R]->bnd[k]->bo==0) continue;
	  b=neighbour(k,at[R]);
	  if (!psf_r_cand(b)||b->seqnr<at[0]->seqnr) continue;
	  if (b==at[0]) {           /* found ring of size R+1 */
	    if (R+1>4&&at[1]->seqnr<at[R]->seqnr) { /* only do rings of size 5 and 6 */
	      for (i=0;i<=R;i++) {
		fprintf(out,"%8d%8d%8d%8d",
			at[i%(R+1)]->seqnr,at[(i+1)%(R+1)]->seqnr,
			at[(i+2)%(R+1)]->seqnr,at[(i+3)%(R+1)]->seqnr);
		if ((++nimphi)%2==0) fprintf(out,"\n");
	      }
	    }
	  }else if (R+1<6) {
	    for (i=1;i<R;i++) if (b==at[i]) break;
	    if (i==R) {
	      at[++R]=b;
	      nt[R]=0;
	    }
	  }
	}
      }
    }
  }
  if (nimphi%2 !=0 ) fprintf(out,"\n");
  if (nimphi==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NIMPHI improper torsions\n",nimphi);  /* the real number */
  fseek(out,0,2);

  /* hydrogen bond donors */

  where=ftell(out);
  fprintf(out,"\n%8d !NDON hb donors\n",ndon);
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1)) || !a->atp->hbdon) continue;
    for (k=0,j=0;j<a->nbnd;j++) {
      t2=neighbour(j,a); if (!(t2->flg&(1<<1))) continue;
      if (t2->atp->z==1) {
	fprintf(out,"%8d%8d",a->seqnr,t2->seqnr);
	if ((++ndon)%4==0) fprintf(out,"\n");
	k++;
      }
    }
  }
  if (ndon%4 !=0 ) fprintf(out,"\n");
  if (ndon==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NDON hb donors\n",ndon);  /* the real number */
  fseek(out,0,2);

  /* hydrogen bond acceptors */

  where=ftell(out);
  fprintf(out,"\n%8d !NACC hb acceptors\n",nacc);
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1)) || !a->atp->hbacc) continue;
    if (a->atp->z==7&&(a->nbnd==4||(a->atp->oop&&a->nbnd==3))) continue; /* nitrogen, no lone pair */
    k=(a->nbnd==1 ? neighbour(0,a)->seqnr : 0);
    fprintf(out,"%8d%8d",a->seqnr,k);
    if ((++nacc)%4==0) fprintf(out,"\n");
  }
  if (nacc%4 !=0 ) fprintf(out,"\n");
  if (nacc==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NACC hb acceptors\n",nacc);  /* the real number */
  fseek(out,0,2);

  /* etc... */

  fprintf(out,"\n%8d !NNB non-bonded exclusions\n",nnb);
  if (nnb==0) fprintf(out,"\n");              /* no exclusions... */

  for (i=0;i<nat;i++) {                       /* pointers into exclusions list */
    fprintf(out,"%8d",0);
    if ((i+1)%8==0) fprintf(out,"\n");
  }
  if (i%8!=0&&i>0) fprintf(out,"\n");

  ngrp=0; 
  b=(Atom *)NULL; grp=(char *)NULL;
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1))) continue;
    if (mvld(b)&&a->up!=b->up) grp=(char *)NULL;
    if (!streq(grp,a->grp)) ngrp++;
    b=a; grp=a->grp;
  }
  fprintf(out,"\n%8d%8d !NGRP charge groups",ngrp,0); 
  if (ngrp==1) {
    fprintf(out,"\n\n");
  }else{
    ngrp=0;
    b=(Atom *)NULL; grp=(char *)NULL;
    k=0;
    ForAtoms(a,m) {
      if (!(a->flg&(1<<1))) continue;
      if (mvld(b)&&a->up!=b->up) grp=(char *)NULL;
      if (!streq(grp,a->grp)) {
	if (ngrp%3 == 0) fprintf(out,"\n");
	fprintf(out,"%8d%8d%8d",k,0,0);
	ngrp++;
      }
      b=a; grp=a->grp;
      k++;
    }
    fprintf(out,"\n");
  }
  fclose(out);
  return(status);
}
/*-------------------- linAtom ----------------------*/

static int linAtom(Atom *a,Bond *b,Bond *c)
{
  return(a->nbnd==2&&
	 (b->bt==BT_TRIPLE||c->bt==BT_TRIPLE
	  ||(b->bt==BT_DOUBLE&&c->bt==BT_DOUBLE)
	  )
	 );
}

/*-------------------- chrm_type --------------------*/

static void chrm_type(char *type,Atomdef *t,int mode)
{
  switch (mode) {
  case CHRM_PSF_XPLOR:
    type[0]='\0';
    if (is_atomdef(t)) StrCpy(type,t->name,5);
    break;
  case CHRM_PSF_CHARMM:
  default:
    type[0]='\0';
    if (is_atomdef(t)) {
      if (t->numtp<10000&&t->numtp>0) sprintf(type,"%4d",t->numtp);
      else strcpy(type,"****");
    }
    break;
  }
}
