#include <stdio.h>
#include <stdlib.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

extern int graphix;
extern Node *MolRoot;
extern Dspobj *DspRoot;

int command_level=0;

#define READ_TYPES FT_DOTS|FT_MOL_TYPES|FT_WLB

#define key_is(a) (streq(a,key->key))

static char *kwdl = 

/* regular commands */
"Read Write Amber Bond Build Caveat CHARMm CPK Color Confa Contour Crystal Cyana Define "
"Delete Discover Display Dots freeze Fit Label List Maccs Match Measure "
"Minimax Modify Mopac NMR PIFF h!Pluto Quest select Surface reset "
"Set TAFF XRay"
" 'Exit' "

/* hidden commands */
"'h!chkdat' 'h!Pause' 'h!Dock' 'h!Copy' 'h!Move' 'h!Merge' 'h!Loop' 'h!Ludi' h!Ortep h!Pluto "
"'h!pdb' 'h!pwd' 'h!render' 'h!setenv' h!Fragment 'h!mpeoe' 'h!junk' 'h!erad' 'h!xmlrpc' 'h!smcsa'";

void foobar(void);  /* junky-stuff */
void junk(void);
void cmd_taff(void);

/*------------------------------------------------------------------------*/
void command_loop(char *prompt)
{
  Token *key;

  command_level++;

  key=(Token *)NULL;
  while(command_level>0) {
    start_parse();
    key=Get_Kwd(prompt,";",kwdl);
    if (key->status == TOK_OK) {
      if (key_is("Read")||key_is("erad"))  cmd_read((Node *)NULL,READ_TYPES);
      else if (key_is("TAFF"))      cmd_taff();
      else if (key_is("junk"))      junk();
      else if (key_is("foobar"))    foobar();
      else if (key_is("pdb"))       cmd_pdb();
      else if (key_is("Dock"))      cmd_dock();
      else if (key_is("CHARMm"))    cmd_charmm();
      else if (key_is("Amber"))     cmd_amber();
      else if (key_is("Cyana"))     cmd_cyana();
      else if (key_is("Contour"))   cmd_contour((Grid *)NULL);
      else if (key_is("Write"))	    cmd_write((Node *)NULL,FT_CAN_WRITE);
      else if (key_is("Crystal"))   cmd_crystal((Xcn *)NULL);
      else if (key_is("Define"))    cmd_define();
      else if (key_is("Delete"))    cmd_delete();
      else if (key_is("Label"))     cmd_label();
      else if (key_is("List")) 	    cmd_list();
      else if (key_is("Set")) 	    cmd_set();
      else if (key_is("reset"))     cmd_display_reset();
      else if (key_is("freeze"))    cmd_freeze();
      else if (key_is("Pluto"))     cmd_pluto();
      else if (key_is("Quest"))     cmd_quest();
      else if (key_is("Color"))     cmd_color();
      else if (key_is("Discover"))  cmd_discover();
      else if (key_is("Display"))   cmd_display();
      else if (key_is("Dots"))      cmd_dots();
      else if (key_is("Bond"))      cmd_bond();
      else if (key_is("CPK"))	    cmd_cpk();
      else if (key_is("select"))    cmd_select();
      else if (key_is("Ortep"))     cmd_ortep();
      else if (key_is("Match"))     cmd_match();
      else if (key_is("Measure"))   cmd_measure();
      else if (key_is("Modify"))    cmd_modify();
      else if (key_is("Pause"))     cmd_pause();
      else if (key_is("Fit"))       cmd_fit("fit");
      else if (key_is("NMR"))       cmd_nmr();
      else if (key_is("Ludi"))      cmd_ludi();
      else if (key_is("Maccs"))     cmd_maccs();
      else if (key_is("Minimax"))   cmd_minimax();
      else if (key_is("PIFF"))      cmd_piff();
      else if (key_is("Mopac"))     cmd_mopac();
      else if (key_is("Confa"))     cmd_confa();
      else if (key_is("Caveat"))    cmd_caveat();
      else if (key_is("Build"))     cmd_build();
      else if (key_is("setenv"))    cmd_setenv();
      else if (key_is("pwd"))       cmd_pwd();
      else if (key_is("render"))    cmd_render();
      else if (key_is("Loop"))      cmd_loop();
      else if (key_is("Copy"))      cmd_copy(FALSE);
      else if (key_is("Move"))      cmd_copy(TRUE);
      else if (key_is("Merge"))     cmd_merge();
      else if (key_is("Fragment"))  cmd_fragment();
      else if (key_is("mpeoe"))     cmd_mpeoe();
      else if (key_is("Surface"))   cmd_surface();
      else if (key_is("XRay"))      cmd_xray();
      else if (key_is("xmlrpc"))    cmd_xmlrpc();
      else if (key_is("smcsa"))     cmd_smcsa();
      else if (key_is("Exit")) {
	if (cmd_exit()) {key=free_Token(key);break;}
      }else if (key_is("chkdat"))   {InternalDataCheck(CHKDAT_ASK);end_parse();}
    }else if (key->status == TOK_NOK) {
      err_set(0,"unknown command %s",key->text);
      
    }else if (key->status == TOK_NULL) {
      err_set(1,"null token at top level, aborted.");
      err_dump();
      exit(1);
    }
    key=free_Token(key);
    if (end_parse()!=TOK_OK) {
      err_set(0,"extra characters at end of command ignored.");
    }
    err_dump();
  }
}
/*-------------------- InternalDataCheck --------------------*/

#define pl(x) ((x)!=1?"s":"")

int InternalDataCheck(int fix)
{
  Node *obj;
  Dspobj *dsp;
  Atom *a;
  Token *key;
  int i,k,n_chain,b_chain,n_flg[8],b_flg[8],nflg,bflg,nat,nbnd,nobj;
  int rep,fixit,ok;
  char buf[1024];

  rep=(fix==CHKDAT_ASK||fix==CHKDAT_REPORT);
  fixit=(fix==CHKDAT_FIX);
  ok=TRUE;

  key=(Token *)NULL;
  n_chain=0; b_chain=0;
  for (i=0;i<8;i++) {
    n_flg[i]=0;
    b_flg[i]=0;
  }

  nbnd=0; nobj=0; nat=0;
  ForObjs(obj,MolRoot,OBJ_ALL) {
    nobj++;
    if (obj->chain) n_chain++;
    if (is_atom(obj)) {
      nat++;
      a=(Atom *)obj;
      for (i=0;i<8;i++) if (a->flg&(1<<i)) n_flg[i]++;
      for (k=0;k<a->nbnd;k++) {
	if (a==a->bnd[k]->f) {
	  nbnd++;
	  if (a->bnd[k]->chain) b_chain++;
	  for (i=0;i<8;i++) if (a->bnd[k]->flg&(1<<i)) b_flg[i]++;
	}
      }
    }
  }
  nobj-=nat;
  if (rep) prtGraphics("%d atom%s, %d bond%s in %d node%s.\n",
		       nat,pl(nat),nbnd,pl(nbnd),nobj,pl(nobj));
  for (nflg=0,i=0;i<8;i++) nflg+=n_flg[i];
  for (bflg=0,i=0;i<8;i++) bflg+=b_flg[i];
  if (n_chain>0) if (rep) prtGraphics("%d object%s in chain.\n",n_chain,pl(n_chain));
  if (nflg>0) {
    if (rep) prtGraphics("%d flagged atom%s: ",nflg,pl(nflg));
    for (*buf='\0',i=0;i<8;i++) sprintf(buf+strlen(buf),"%d ",n_flg[i]);
    if (rep) prtGraphics("%s\n",buf);
  }
  if (b_chain>0) if (rep) prtGraphics("%d bond%s in chain.\n",b_chain,pl(b_chain));
  if (bflg>0) {
    if (rep) prtGraphics("%d flagged bond%s: ",bflg,pl(bflg));
    for (*buf='\0',i=0;i<8;i++) sprintf(buf+strlen(buf),"%d ",b_flg[i]);
    if (rep) prtGraphics("%s\n",buf);
  }

  if (nflg>0||bflg>0||n_chain>0||b_chain>0) {
    ok=FALSE;
    if (fix==CHKDAT_ASK) {
      default_token("yes");
      key=Get_Kwd("fix ? ",";","no yes");
      if (key->status==TOK_OK) fixit=key_is("yes");
      else if (key->status==TOK_NOK) {
	err_set(1,"data check: bad token %s",key->text);
	fixit=FALSE;
      }
    }
    if (fixit) {
      ForObjs(obj,MolRoot,OBJ_ALL) {
	obj->chain=(Node *)NULL;
	if (is_atom(obj)) {
	  a=(Atom *)obj;
	  a->flg=0;
	  for (k=0;k<a->nbnd;k++) {
	    if (a==a->bnd[k]->f) {
	      a->bnd[k]->chain=(Bond *)NULL;
	      a->bnd[k]->flg=0;
	    }
	  }
	}
      }
    }
  }
  n_chain=0; nobj=0;
  key=free_Token(key);
  ForDspobjs(dsp,DspRoot) {
    if (dsp->chain) n_chain++;
    nobj++;
  }
  if (rep) prtGraphics("%d nodes in display tree.\n",nobj);
  if (n_chain>0) {
    ok=FALSE;
    if (rep) prtGraphics("%d dsp nodes in chains.\n",n_chain);
    if (fix==CHKDAT_ASK) {
      default_token("yes");
      key=Get_Kwd("fix ? ",";","no yes");
      if (key->status==TOK_OK) fixit=key_is("yes");
      else {
	if (key->status==TOK_NOK) err_set(1,"sanity check: bad token %s",key->text);
	fixit=FALSE;
      }
    }
    if (fixit) ForDspobjs(dsp,DspRoot) dsp->chain=(Node *)NULL;
  }
  key=free_Token(key);
  return(ok);
}
