#ifndef CYANADEF
#define CYANADEF

#include <stdio.h>
#include "object.h"

#define CYAMAXNB       4   /* maximum number of bonds per atom */

typedef struct cyaResidue_t {
  struct cyaAtom_t **A;
  struct cyaBond_t **B;
  struct cyaTorDef_t **tor;
  struct cyaRing_t **ring;
  Node *root;
  int nat,nbnd,ntor,nrings,nc;
} cyaResidue_t;

typedef struct cyaAtom_t {
  struct cyaBond_t *nb[CYAMAXNB];
  struct cyaAtom_t *n[CYAMAXNB];
  struct cyaAtom_t *up;
  struct cyaAtom_t *lst;
  struct cyaAtom_t *psatm;
  struct cyaTorDef_t *ft;
  Atom *atm;
  char name[8],type[6];
  float x[3];
  int nn;
  int c,flg,utp,z,seqnr,lb,dp,oseq;
} cyaAtom_t;

typedef struct cyaBond_t {
  struct cyaAtom_t *f;
  struct cyaAtom_t *t;
  Bond *bnd;
  int flg,rc,rs,rigid;
} cyaBond_t;

typedef struct cyaTorDef_t {
  struct cyaAtom_t *A[5];
} cyaTorDef_t;

typedef struct cyaRing_t {
  cyaBond_t **bnd;
  cyaAtom_t **atm;
  int rs;
} cyaRing_t;

typedef struct cyaOptions_t {
  int overlapAtoms;
  int useNames;
  int rigidRings;
  int bridgedRings;
  int addPseudoAtoms;
  int chsUpper;
  int nameUpper;
} cyaOptions_t;

cyaResidue_t *newCyaResidue(void);
cyaResidue_t *freeCyaResidue(cyaResidue_t *L);
cyaAtom_t    *newCyaAtom(Atom *a);
cyaAtom_t    *freeCyaAtom(cyaAtom_t *A);
cyaBond_t    *newCyaBond(cyaAtom_t *f,cyaAtom_t *t);
cyaBond_t    *freeCyaBond(cyaBond_t *B);
cyaTorDef_t  *newCyaTorDef(cyaAtom_t **A,cyaAtom_t *lst);
cyaTorDef_t  *freeCyaTorDef(cyaTorDef_t *T);
cyaRing_t    *newCyaRing(int n,cyaBond_t **B);
cyaRing_t    *freeCyaRing(cyaRing_t *R);

void          cyaMarkRigidBonds(cyaResidue_t *L,cyaOptions_t Options);
void          cyaFindTorsions(cyaResidue_t *L);

cyaResidue_t *cyanaResidue(Node *root, cyaOptions_t Options);
cyaResidue_t *collectAtomsAndBonds(Node *root);
void          cyaResidueRings(cyaResidue_t *M);
void          cyaSortRings(cyaResidue_t *L);

int           cyaBFTree(cyaResidue_t *L);
int           cyaSortTree(cyaResidue_t *L);
cyaAtom_t    *cyaBestRoot(cyaResidue_t *L);
void          cyaNewRoot(cyaResidue_t *L,cyaAtom_t *A);
int           cyaCheckAtomNames(cyaResidue_t *L,int verbose);
int           writeCyanaResidue(FILE *out, cyaResidue_t *L, char *code, cyaOptions_t Options);
int           writeCyanaRestraints(FILE *out, cyaResidue_t *L, char *code, cyaOptions_t Options);
int           cyaAddPseudoAtoms(cyaResidue_t *L);
void          cyaRemovePseudoAtoms(cyaResidue_t *L);

void          cyaAssignUniqueTypes(cyaResidue_t *L);


#endif
