#include <stdlib.h>
#include <string.h>
#include "object.h"
#include "cyana.h"

/*-------------------- Residue --------------------*/

cyaResidue_t *newCyaResidue(void)
{
  cyaResidue_t *L;

  L=malloc(sizeof(cyaResidue_t));
  if (L) {
    L->A=NULL;
    L->B=NULL;
    L->ring=NULL;
    L->tor=NULL;
    L->nat=0;
    L->nbnd=0;
    L->ntor=0;
    L->nc=0;
    L->nrings=0;
  }
  return L;
}

cyaResidue_t *freeCyaResidue(cyaResidue_t *L)
{
  int i;

  if (L) {
    if (L->A) {
      for (i=0;i<L->nat;i++) {
	freeCyaAtom(L->A[i]);
      }
      free(L->A);
    }
    if (L->B) {
      for (i=0;i<L->nbnd;i++) {
	freeCyaBond(L->B[i]);
      }
      free(L->B);
    }
    if (L->tor) {
      for (i=0;i<L->ntor;i++) {
	freeCyaTorDef(L->tor[i]);
      }
      free(L->tor);
    }
    if (L->ring) {
      for (i=0;i<L->nrings;i++) {
	freeCyaRing(L->ring[i]);
      }
      free(L->ring);
    }
    free(L);
  }
  return(NULL);
}

/*-------------------- Atom --------------------*/

cyaAtom_t *newCyaAtom(Atom *a)
{
  cyaAtom_t *A;
  int i;

  A=malloc(sizeof(cyaAtom_t));
  if (A) {
    A->nn=0;
    A->utp=-1;
    A->seqnr=-1;
    A->oseq=-1;
    A->atm=NULL;
    A->psatm=NULL;
    A->c=0;
    if (a) {
      A->z=a->atp->z;
      strncpy(A->name,a->name,7); A->name[7]='\0';
      strcpy(A->type,a->atp->chs);
      for (i=0;i<3;i++) A->x[i]=a->x[i];
      A->atm=a;
    }else{
      A->z=0;
      A->name[0]='\0';
      for (i=0;i<3;i++) A->x[i]=0.0;
    }
  }
  return(A);
}

cyaAtom_t *freeCyaAtom(cyaAtom_t *A)
{
  if (A) {
    free(A);
  }
  return(NULL);
}

/*-------------------- Bond --------------------*/

cyaBond_t *newCyaBond(cyaAtom_t *f,cyaAtom_t *t)
{
  cyaBond_t *B;

  B=malloc(sizeof(cyaBond_t));
  if (B) {
    if (f) {
      B->f=f;
      B->t=t;
      B->rigid=FALSE;
      B->rc=FALSE;
      B->rs=0;
      B->bnd=find_bond(f->atm,t->atm);
    }
  }
  return(B);
}

cyaBond_t *freeCyaBond(cyaBond_t *B)
{
  if (B) {
    free(B);
  }
  return(NULL);
}

/*-------------------- Torsion definitions --------------------*/

cyaTorDef_t *newCyaTorDef(cyaAtom_t **A,cyaAtom_t *lst)
{
  cyaTorDef_t *T;
  int i;

  T=malloc(sizeof(cyaTorDef_t));
  if (T) {
    if (A) {
      for (i=0;i<4;i++) T->A[i]=A[i];
    }else{
      for (i=0;i<4;i++) T->A[i]=NULL;
    }
    T->A[4]=lst;
  }
  return(T);
}

cyaTorDef_t *freeCyaTorDef(cyaTorDef_t *T)
{
  if (T) {
    free(T);
  }
  return(NULL);
}

/*-------------------- Rings --------------------*/

cyaRing_t *newCyaRing(int n,cyaBond_t **B)
{
  cyaRing_t *R;
  int i;

  R=malloc(sizeof(cyaRing_t));
  if (R) {
    if (n>0) {
      R->rs=n;
      R->bnd=malloc(n*sizeof(cyaBond_t *));
      R->atm=malloc(n*sizeof(cyaAtom_t *));
      if (R->bnd && R->atm) {
	if (B) {
	  for (i=0;i<n;i++) R->bnd[i]=B[i];
	  R->atm[0]=R->bnd[0]->f;
	  for (i=1;i<n;i++) {
	    if (R->bnd[i]->f==R->atm[i-1]) R->atm[i]=R->bnd[i]->t;
	    else  R->atm[i]=R->bnd[i]->f;
	  }
	}else{
	  for (i=0;i<n;i++) R->bnd[i]=NULL;
	  for (i=0;i<n;i++) R->atm[i]=NULL;
	}
      }else R=freeCyaRing(R);
    }
  }
  return R;
}

cyaRing_t *freeCyaRing(cyaRing_t *R)
{
  if (R) {
    if (R->bnd) free(R->bnd);
    if (R->atm) free(R->atm);
    free(R);
  }
  return(NULL);
}
