#include <stdlib.h>
#include <string.h>
#include "cyana.h"

static int nH(cyaAtom_t *A)
{
  int i,n=0;
  
  for (i=0;i<A->nn;i++) if (A->n[i]->z==1) n++;
  return n;
}

static cyaAtom_t *H(cyaAtom_t *A,int i)
{
  int k,j;
  
  for (k=0,j=0;j<A->nn;j++) {
    if (A->n[j]->z==1) {
      if (k==i) return(A->n[j]);
      else k++;
    }
  }
  return NULL;
}
  
static cyaAtom_t *pseudoAtom(cyaAtom_t **A,int n,cyaAtom_t *up,cyaTorDef_t *ft)
{
  int i,j,l;
  cyaAtom_t *P;

  P=newCyaAtom(NULL);
  if (P) {
    for (j=0;j<3;j++) P->x[j]=0.0;
    for (i=0;i<n;i++) {
      for (j=0;j<3;j++) P->x[j]+=A[i]->x[j]/n;
      A[i]->psatm=P;
    }
    P->up=up;
    P->utp=A[0]->utp;
    P->ft=ft;
    if (up->atm) {
      l=strlen(up->atm->atp->chs);
      sprintf(P->name,"Q%s",up->name+l);
    }else{
      sprintf(P->name,"Q%s",A[0]->name);
    }
    strcpy(P->type,"PSEUD");
  }
  return(P);
}

static int insertPseudoAtom(cyaResidue_t *L,int pos,cyaAtom_t **A,int n,cyaAtom_t *up,cyaTorDef_t *ft)
{
  int i,r=FALSE;
  cyaAtom_t *P;

  if (n>1) {
    P=pseudoAtom(A,n,up,ft);
    if (P) {
      for (i=L->nat;i>pos+1;i--) {
	L->A[i]=L->A[i-1];
	L->A[i]->seqnr=i;
      }
      L->A[pos+1]=P;
      L->nat++;
      P->seqnr=pos+1;
      r=TRUE;
    }
  }
  return r;
}

/*------------------------------------------------------------------------*/

int cyaAddPseudoAtoms(cyaResidue_t *L)
{
  int i,k,n,nh,utp;
  cyaAtom_t **A=NULL;
  cyaTorDef_t *ft;

  A=malloc(L->nat*sizeof(cyaAtom_t *));
  if (A) {
    /* first level pseudo atoms: H2 and H3, and groups of H1 in same rigid fragment */
    for (i=0;i<L->nat;i++) {
      if (L->A[i]->z>1) {
	nh=nH(L->A[i]);
	if (nh==1) {      /* single hydrogen, only one level of pseudo atoms */
	  A[0]=H(L->A[i],0);
	  ft=A[0]->ft;
	  utp=A[0]->utp;
	  for (n=0,k=i+1;k<L->nat;k++) {
	    if (L->A[k]->z==1 && L->A[k]->utp==utp && !L->A[k]->psatm && L->A[k]->ft==ft) {
	      A[n++]=L->A[k];
	    }
	  }
	  if (insertPseudoAtom(L,i,A,n,L->A[i],A[0]->ft)) i++;
	}else if (nh>1) {
	  for (k=0;k<nh;k++) A[k]=H(L->A[i],k);
	  if (nh==3) {
	    if (insertPseudoAtom(L,i-1,A,nh,L->A[i],L->A[i]->ft)) i++;
	  }else{
	    if (insertPseudoAtom(L,i,A,nh,L->A[i],A[0]->ft)) i++;
	  }
	}
      }
    }
    /* second level: groups of equivalent (utp) pseudo atoms in same rigid fragment */
    for (i=0;i<L->nat;i++) {
      if (!L->A[i]->atm && !L->A[i]->psatm) {
	utp=L->A[i]->utp;
	ft=L->A[i]->ft;
	for (n=0,k=i;k<L->nat;k++) {
	  if (!L->A[k]->atm && !L->A[k]->psatm && L->A[k]->utp==utp &&  L->A[k]->ft==ft) {
	     A[n++]=L->A[k];
	  }
	}
	if (insertPseudoAtom(L,i-1,A,n,L->A[i],ft)) i+=2;
      }
    }
    free(A);
  }
  for (n=0,i=0;i<L->nat;i++) if (!L->A[i]->atm) n++;
  return n;
}

void cyaRemovePseudoAtoms(cyaResidue_t *L)
{
  int i,nat=0;

  for (i=0;i<L->nat;i++) {
    if (L->A[i]->atm) L->A[nat++]=L->A[i];
    else freeCyaAtom(L->A[i]);
  }
  for (i=0;i<nat;i++) {
    L->A[i]->seqnr=i;
    L->A[i]->psatm=NULL;
  }
  L->nat=nat;
}
