#include <stdlib.h>
#include "cyana.h"

#define TRUE 1
#define FALSE 0

static int compareBondRS(const void *a,const void *b)
{
  cyaBond_t *A,*B;

  A=*(cyaBond_t **)a;
  B=*(cyaBond_t **)b;

  return(A->rs-B->rs);
}

static int compareRings(const void *a,const void *b)
{
  cyaRing_t *A,*B;

  A=*(cyaRing_t **)a;
  B=*(cyaRing_t **)b;

  return(A->rs-B->rs);
}


static void resetRings(cyaResidue_t *M)
{
  int i,nr;

  for (i=0;i<M->nbnd;i++) {
    M->B[i]->flg=0;
    M->B[i]->rc=FALSE;
    M->B[i]->rs=0;
  }
  for (i=0;i<M->nrings;i++) {
    freeCyaRing(M->ring[i]);
  }
  M->nrings=0;
  nr=M->nbnd-M->nat+M->nc;
  M->ring=realloc(M->ring,nr*sizeof(cyaRing_t *));
}

static cyaRing_t *smallestRing(cyaResidue_t *M,cyaBond_t *bnd,int useRS)
{
  cyaRing_t *R=NULL;
  cyaAtom_t **A=NULL;
  cyaBond_t **B=NULL,*rc=NULL;
  int *from;
  int i,j,k,l,m=0,n=0,nr,rs;

  A=malloc(M->nat*sizeof(cyaAtom_t *));
  B=malloc(M->nat*sizeof(cyaBond_t *));
  from=malloc(M->nat*sizeof(int));

  if (A && B && from) {
    A[0]=bnd->f; A[0]->flg=+1; B[0]=bnd; from[0]=0;
    A[1]=bnd->t; A[1]->flg=-1; B[1]=bnd; from[1]=0;
    n=2;
    for (k=0,l=-1;k<n&&l<0;k++) {
      for (j=0;j<A[k]->nn&&l<0;j++) {
	if (A[k]->nb[j]!=B[k] && (!useRS || A[k]->nb[j]->rs>0)) {
	  if (A[k]->flg*A[k]->n[j]->flg<0) {
	    for (l=0;l<n;l++) if (A[l]==A[k]->n[j]) break;
	    rc=A[k]->nb[j];
	    m=k;
	    break;
	  }else if (A[k]->n[j]->flg==0) {
	    from[n]=k;
	    B[n]=A[k]->nb[j];
	    A[n]=A[k]->n[j];
	    A[n]->flg=A[k]->flg>0 ? A[k]->flg+1 : A[k]->flg-1;
	    n++;
	  }
	}
      }
    }
    if (l<n && l>-1) {
      rs=A[m]->flg>0 ? A[m]->flg-A[l]->flg : A[l]->flg-A[m]->flg;
      R=newCyaRing(rs,NULL);
      for (nr=0;m>1;m=from[m],nr++) {
	R->bnd[nr]=B[m];
	R->atm[nr]=A[m];
      }
      R->bnd[nr]=B[m];
      R->atm[nr++]=A[m];
      for (nr=rs-1;l>1;l=from[l],nr--) {
	R->bnd[nr]=rc;
	R->atm[nr]=A[l];
	rc=B[l];
      }
      R->bnd[nr]=rc;
      R->atm[nr]=A[l];
    }
    for (i=0;i<n;i++) A[i]->flg=0;
  }

  if (A) free(A);
  if (B) free(B);
  if (from) free(from);

  return(R);
}

static int cyaResidueRingSizes(cyaResidue_t *M)
{
  /* Foreach bond in Molecule M: determine size of smallest ring containing
     bond (0 if bond is not in a ring). Return: error status, 0 if no error. 
   
     Ring size of j-th bond will be stored in M->B[j]->rs.
  */

  int i,j,k,l,n,status=0;
  cyaAtom_t **AL=(cyaAtom_t **)NULL;
  cyaAtom_t *m;

  if (M&&M->nat>0&&M->nbnd>0) {

    /* Allocate scratch array for explore list: */

    AL=(cyaAtom_t **)malloc(M->nat*sizeof(cyaAtom_t *));

    if (AL) {
      for (j=0;j<M->nbnd;j++) M->B[j]->rs=M->nbnd+1; /* set ring status 'unknown'                 */
      for (i=0;i<M->nat;i++)  M->A[i]->flg=0;        /* clear all atomflags                       */
      for (j=0;j<M->nbnd;j++) {                      /* for each bond...                          */
	M->B[j]->rs=0; l=0;
	if (M->B[j]->f->nn>1&&M->B[j]->t->nn>1) {    /* ...which is not terminal...               */
	  M->B[j]->f->flg=-1;                        /* set left atom flag                        */
	  M->B[j]->t->flg=+1;                        /* set right atom flag                       */
	  AL[l++]=M->B[j]->f;                        /* add left atom to explore list             */
	  AL[l++]=M->B[j]->t;                        /* add right atom to explore list            */
	  for (k=0;k<l;k++) {                        /* while list has unexplored atoms...        */
	    if (AL[k]->flg<0) {                      /* if the atom is on the left of bond j:     */
	      for (n=0;n<AL[k]->nn;n++) {            /* examine each neighbor atom                */
		m=AL[k]->n[n];
		if (m->flg==0) {                     /* if neighbor has no flag                   */
		  if (AL[k]->nb[n]->rs!=0) {         /* and if not bonded via known chain bond    */
		    m->flg=AL[k]->flg-1;             /* flag it (parent flag - 1)                 */
		    AL[l++]=m;                       /* and add it to list of atoms to explore    */
		  }                                  /* else ignore it                            */
		}else{                               /* atom has a flag already...                */
		  if (m->flg>1) {                    /* flag indicates ring closure               */
		    M->B[j]->rs=m->flg-AL[k]->flg;   /* ring size of bond j is sum of abs(flags)  */
		    k=l;                             /* no need to continue with explore list     */
		    break;
		  }
		}
	      }
	    }else{                                   /* if the atom is on the right of bond j:    */
	      for (n=0;n<AL[k]->nn;n++) {            /* examine each neighbor atom                */
		m=AL[k]->n[n];
		if (m->flg==0) {                     /* if neighbor has no flag                   */
		  if (AL[k]->nb[n]->rs!=0) {          /* and if not bonded via known chain bond   */
		    m->flg=AL[k]->flg+1;             /* flag it (parent flag + 1)                 */
		    AL[l++]=m;                       /* and add it to list of atoms to explore    */
		  }                                  /* else ignore it                            */
		}else{                               /* atom has a flag already...                */
		  if (m->flg<-1) {                   /* flag indicates ring closure               */
		    M->B[j]->rs=-m->flg+AL[k]->flg;  /* ring size of bond j is sum of abs(flags)  */
		    k=l;                             /* no need to continue with explore list     */
		    break;
		  }
		}
	      }
	    }
	  }
	}else M->B[j]->rs=0;                         /* terminal bonds have rs 0                  */
	for (k=0;k<l;k++) AL[k]->flg=0;              /* reset atom flags for next bonds analysis  */
      }
      status=0;                                      /* no error.                                 */
    }else{
      status=1;                                      /* error status: malloc failed.              */
      for (j=0;j<M->nbnd;j++) M->B[j]->rs=M->nbnd+1; /* undefined ring size...                    */
    }
  }

  if (AL) free(AL);

  return(status);
}
/*-------------------- Rings --------------------*/

void cyaSortRings(cyaResidue_t *M)
{
  if (M->nrings>0) {
    qsort(M->ring,M->nrings,sizeof(cyaRing_t *),compareRings);
  }
}

void cyaResidueRings(cyaResidue_t *M)
{
  int i,j;
  cyaRing_t *R=NULL;

  resetRings(M);
  cyaResidueRingSizes(M);
  for (i=0;i<M->nrings;i++) freeCyaRing(M->ring[i]);
  M->nrings=0;
  qsort(M->B,M->nbnd,sizeof(cyaBond_t *),compareBondRS);
  for (i=0;i<M->nat;i++) M->A[i]->flg=0;
  for (i=0;i<M->nbnd;i++) {
    if (M->B[i]->rs>0 && !M->B[i]->flg) {
      R=smallestRing(M,M->B[i],TRUE);
      if (R){
	M->B[i]->rc=TRUE;
	M->ring[M->nrings++]=R;
	for (j=0;j<R->rs;j++) {
	  R->bnd[j]->flg=1;
	}
      }
    }
  }
  for (j=0;j<M->nbnd;j++) M->B[j]->flg=0;
}

