#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cyana.h"

static char *rcmac1=
  "## 7MACROS: CYANA macro\n"
  "###\n"
  "### Parameters: weight=<float> (default: 1.0)\n"
  "###             dtol=<float>   (default: 0.0)\n"
  "###             atol=<float>   (default: 0.0)\n"
  "###             ttol=<float>   (default: 0.0)\n"   
  "###\n"
  "### Adds constraints (upper and lower limits on distances) for ring-closure\n"
  "### bonds in all residues of type %s. The optional parameter 'weight' may\n";

static char *rcmac2=
  "### be used to define the weight of the ring-closure restraints in the target\n"
  "### function. 'dtol', 'atol', and 'ttol' are tolerances (in Angstrom) for the\n"
  "###  distance restraints generated for ring-closure bondlengths (dtol),\n"
  "###  angles (atol), and torsions.\n"
  "\n"
  "var info echo i res\n"
  "\n"
  "syntax weight=@r=1.0 dtol=@r=0.0 atol=@r=0.0 ttol=@r=0.0\n"
  "\n"
  "echo:=off\n"
  "info:=none\n"
  "\n";

static float distance(float x[3],float y[3])
{
  float d;
  
  d=(x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]);
  return(d>0?sqrt(d):0.0);
}

static int torsEquiv(cyaResidue_t *L,cyaAtom_t *A,cyaAtom_t *B)
{
  int r=TRUE,i,a,b;

  for (i=0;r&&i<L->ntor;i++) {
    a=A->seqnr>=L->tor[i]->A[2]->seqnr && A->seqnr<=L->tor[i]->A[4]->seqnr?1:0;
    b=B->seqnr>=L->tor[i]->A[2]->seqnr && B->seqnr<=L->tor[i]->A[4]->seqnr?1:0;
    r=a==b;
  }
  return r;
}

static char *upper(char *buff,char *txt)
{
  int i;

  for (i=0;txt[i];i++) buff[i]=txt[i]>='a'&&txt[i]<='z' ? 'A'+txt[i]-'a' : txt[i];
  buff[i]='\0';
  return buff;
}

/*------------------------------------------------------------------------*/

int writeCyanaResidue(FILE *out, cyaResidue_t *L, char *code, cyaOptions_t Options)
{
  int i,j,nt,ovOffset,status=0;
  cyaAtom_t **A;
  char buff[20],name[20];
  float x[3];
  
  if (L) {
    for (j=0;j<3;j++) x[j]=0;
    for (i=0;i<L->nat;i++) {
      for (j=0;j<3;j++) x[j]+=L->A[i]->atm->x[j];
    }
    if (L->nat>0) {
      for (j=0;j<3;j++) x[j]/=L->nat;
    }

    if (Options.addPseudoAtoms) cyaAddPseudoAtoms(L);
    if (cyaCheckAtomNames(L,TRUE)>0) status=1;

    if (Options.overlapAtoms) {
      ovOffset=4;
      fprintf(out,"%-10s%-5s%5d%5d%5d%5d\n","RESIDUE",code,L->ntor+1,L->nat+6,3,L->nat+5);
    }else{
      ovOffset=1;
      fprintf(out,"%-10s%-5s%5d%5d%5d%5d\n","RESIDUE",code,L->ntor,L->nat,1,L->nat);
    }
    
    if (Options.useNames) {
      if (Options.overlapAtoms) {
	fprintf(out,"%4d LNK  %5d%5d%10.4f %s\n",1,0,0,0.0,
		"-Q2   QL3  QL1  QL2");
	nt=1;
      }else{
	nt=0;
      }
      for (i=0;i<L->ntor;i++) {
	A=L->tor[i]->A;
	fprintf(out,"%4d T%03d %5d%5d%10.4f  ",i+1+nt,i+1,0,0,0.0);
	for (j=0;j<5;j++) fprintf(out," %-4s",Options.nameUpper? upper(name,A[j]->name) : A[j]->name);
	fprintf(out,"\n");
      }
    }else{
      if (Options.overlapAtoms) {
	fprintf(out,"%4d LNK  %5d%5d%10.4f%5d%5d%5d%5d%5d\n",1,0,0,0.0,2,3,L->nat+4,L->nat+5,0);
	nt=1;
      }else{
	nt=0;
      }
      for (i=0;i<L->ntor;i++) {
	A=L->tor[i]->A;
	fprintf(out,"%4d T%03d %5d%5d%10.4f%5d%5d%5d%5d%5d\n",i+1+nt,i+1,0,0,0.0,
		A[0]->seqnr+ovOffset,A[1]->seqnr+ovOffset,
		A[2]->seqnr+ovOffset,A[3]->seqnr+ovOffset,
		A[4]->seqnr+ovOffset);
      }
    }

    if (Options.overlapAtoms) {
      if (Options.useNames) {
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f\n",
		1,"Q1","DUMMY",0,0.0,0.0,0.0,0.0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f\n",
		2,"Q2","DUMMY",0,0.0,1.0,0.0,0.0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f\n",
		3,"QL3","DUMMY",0,0.0,1.0,1.0,0.0);
      }else{
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
		1,"Q1","DUMMY",0,0.0,0.0,0.0,0.0,0,0,0,0,0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
		2,"Q2","DUMMY",0,0.0,1.0,0.0,0.0,0,0,0,0,0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
		3,"QL3","DUMMY",0,0.0,1.0,1.0,0.0,0,0,0,0,0);
      }
    }
    A=L->A;
    for (i=0;i<L->nat;i++) {
      fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f",
	      i+ovOffset,
	      Options.nameUpper? upper(name,A[i]->name) : A[i]->name,
	      Options.chsUpper? upper(buff,A[i]->type) : A[i]->type,
	      0,0.0,A[i]->x[0]-x[0],A[i]->x[1]-x[1],A[i]->x[2]-x[2]);
      if (Options.useNames) {
	fprintf(out," ");
	for (j=0;j<A[i]->nn;j++) fprintf(out," %-4s",Options.nameUpper? upper(name,A[i]->n[j]->name) : A[i]->n[j]->name) ;
	if (A[i]->psatm) {
	  for (;j<4;j++) fprintf(out," %-4s","-");
	  fprintf(out," %-4s",Options.nameUpper? upper(name,A[i]->psatm->name) : A[i]->psatm->name);
	}
      }else{
	for (j=0;j<A[i]->nn;j++) fprintf(out,"%5d",A[i]->n[j]->seqnr+ovOffset);
	for (;j<4;j++) fprintf(out,"%5d",0);
	if (A[i]->psatm) fprintf(out,"%5d",A[i]->psatm->seqnr+ovOffset);
	else fprintf(out,"%5d",0);
      }
      fprintf(out,"\n");
    }

    if (Options.overlapAtoms) {
      if (Options.useNames) {
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f\n",
		L->nat+ovOffset,"QL1","DUMMY",0,0.0,2.0,1.0,0.0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f\n",
		L->nat+ovOffset+1,"QL2","DUMMY",0,0.0,2.0,0.0,0.0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f\n",
		L->nat+ovOffset+2,"Q3","DUMMY",0,0.0,3.0,0.0,0.0);
      }else{
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
		L->nat+ovOffset,"QL1","DUMMY",0,0.0,2.0,1.0,0.0,0,0,0,0,0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
		L->nat+ovOffset+1,"QL2","DUMMY",0,0.0,2.0,0.0,0.0,0,0,0,0,0);
	fprintf(out,"%4d %-5s%-5s%5d%10.4f%10.4f%10.4f%10.4f%5d%5d%5d%5d%5d\n",
		L->nat+ovOffset+2,"Q3","DUMMY",0,0.0,3.0,0.0,0.0,0,0,0,0,0);
      }
    }
    
    cyaRemovePseudoAtoms(L);
  }
  return status;
}

int writeCyanaRestraints(FILE *out, cyaResidue_t *L, char *code, cyaOptions_t Options)
{
  int i,j,k,nc=0,status=0;
  float d;
  float dtol=0.000,atol=0.00,ttol=0.00;   /* 0: user should use dtol and atol macro parameters to set tolerances */
  char name1[20],name2[20];

  if (L) {
    fprintf(out,rcmac1,code);
    fprintf(out,"%s",rcmac2);

    for (i=0;i<L->nbnd;i++) {
      if (L->B[i]->rc && !torsEquiv(L,L->B[i]->f,L->B[i]->t)) nc++;
    }
    if (nc>0) {
      fprintf(out,"do i 1 $nr\n");
      fprintf(out,"   if (rnam(i).eq.'%s') then\n",code);
      fprintf(out,"      res=rnum(i)\n");

      for (i=0;i<L->nbnd;i++) {
	if (L->B[i]->rc && !torsEquiv(L,L->B[i]->f,L->B[i]->t)) {
	  fprintf(out,"\n");
	  /* ring closure distance */
	  d=distance(L->B[i]->f->x,L->B[i]->t->x);
	  fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.3f-$dtol upl=%0.3f+$dtol weight=$weight\n",
		  Options.nameUpper?upper(name1,L->B[i]->f->name):L->B[i]->f->name,
		  Options.nameUpper?upper(name2,L->B[i]->t->name):L->B[i]->t->name,
		  d-dtol,d+dtol);
	  /* angles at first ring-closure atom */
	  for (k=0;k<L->B[i]->f->nn;k++) {
	    if (L->B[i]->f->n[k]!=L->B[i]->t) {
	      d=distance(L->B[i]->f->n[k]->x,L->B[i]->t->x);
	      fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.3f-$atol upl=%0.3f+$atol weight=$weight\n",
		      Options.nameUpper?upper(name1,L->B[i]->f->n[k]->name):L->B[i]->f->n[k]->name,
		      Options.nameUpper?upper(name2,L->B[i]->t->name):L->B[i]->t->name,
		      d-atol,d+atol);
	    }
	  }
	  /* angles at second ring-closure atom */
	  for (k=0;k<L->B[i]->t->nn;k++) {
	    if (L->B[i]->t->n[k]!=L->B[i]->f) {
	      d=distance(L->B[i]->t->n[k]->x,L->B[i]->f->x);
	      fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.3f-$atol upl=%0.3f+$atol weight=$weight\n",
		      Options.nameUpper?upper(name1,L->B[i]->t->n[k]->name):L->B[i]->t->n[k]->name,
		      Options.nameUpper?upper(name2,L->B[i]->f->name):L->B[i]->f->name,
		      d-atol,d+atol);
	    }
	  }
	  /* dihedrals for stiff torsions */
	  if (L->B[i]->rigid) {
	    for (k=0;k<L->B[i]->t->nn;k++) {
	      if (L->B[i]->t->n[k]!=L->B[i]->f) {
		for (j=0;j<L->B[i]->f->nn;j++) {
		  if (L->B[i]->f->n[j]!=L->B[i]->t) {
		    d=distance(L->B[i]->t->n[k]->x,L->B[i]->f->n[j]->x);
		    fprintf(out,"      distance make \"%-4s $res\" \"%-4s $res\" lol=%0.3f-$ttol upl=%0.3f+$ttol weight=$weight\n",
			    Options.nameUpper?upper(name1,L->B[i]->t->n[k]->name):L->B[i]->t->n[k]->name,
			    Options.nameUpper?upper(name2,L->B[i]->f->n[j]->name):L->B[i]->f->n[j]->name,
			    d-ttol,d+ttol);
		  }
		}
	      }
	    }
	  }
	}
      }
      fprintf(out,"\n      print \"    Added ring-closure restraints for residue $rnam(i) $rnum(i)\"\n\n");
      fprintf(out,"   end if\n");
      fprintf(out,"end do\n\n");
    }

    fprintf(out,"unset info\n");
    fprintf(out,"write upl %s\n",code);
    fprintf(out,"write lol %s\n",code);
  }
  return status;
}
