#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/param.h>
#include <sys/stat.h>
#include "mystdlib.h"
#include "strutil.h"
#include "token.h"
#include "errhand.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

typedef struct stringlist {
  struct stringlist *nxt;
  char *str;
} stringlist;

#define MX_WLD 64

int dir_alpha_asc(const void *a, const void *b);

/*-------------------- DirList --------------------*/

int DirList(char *dir,char *sel,int mode,char **names[])
{
  stringlist Head,*nxt,*prv;
  int i,n=0,status=0;
  DIR *directory;
  struct dirent *de;
  char *nm;
  Parse p;
  int nwld,err;
  char *wld[MX_WLD];
  char ffn[MAXPATHLEN];
  struct stat sbuf;

  err=0;

  *names=(char **)NULL;
  
  p.buf=sel;
  p.tok=(char *)NULL;
  p.sep="|,";
  p.pos=0;
  nwld=0;
  while (get_token(&p)) {
    if (nwld>=MX_WLD) {
      if (!err) err_set(0,"DirList: MX_WLD exceeded.\n");
      err=1;
    }else{
      if (p.ltok<1) continue;
      wld[nwld]=(char *)malloc((p.ltok+1)*sizeof(char));
      strncpy(wld[nwld],p.tks,p.ltok);
      wld[nwld][p.ltok]='\0';
      nwld++;
    }
  }

  if (nwld>0) {
    Head.nxt=(stringlist *)NULL; Head.str=(char *)NULL; prv=&Head; n=0;
    directory=opendir(dir);
    if (directory) {
      while((de=readdir(directory))) {
	for (i=0;i<nwld;i++) if (wild_match(de->d_name,wld[i])) break;
	if (i<nwld) {
	  if (mode != 0) {
	    sprintf(ffn,"%s/%s",dir,de->d_name);
	    status=stat(ffn,&sbuf);
	  }
	  if (mode == 0 || (status>=0 && (mode&sbuf.st_mode))) {
	    nm=Strdup(de->d_name);
	    if (nm) {
	      nxt=(stringlist *)malloc(sizeof(stringlist));
	      if (nxt) {
		prv->nxt=nxt;
		nxt->str=nm;
		prv=nxt;
		n++;
	      }
	    }
	  }
	}
      }
      closedir(directory);
      prv->nxt=(stringlist *)NULL;
      if (n>0) {
	*names=(char **)malloc(n*sizeof(char *));
	for (prv=Head.nxt,i=0;i<n&&mvld(prv);i++,prv=Head.nxt) {
	  if (*names) (*names)[i]=prv->str;
	  Head.nxt=prv->nxt;
	  free(prv);
	}
	if (i<n) err_set(1,"dir_list, internal error: i<n.");
	if (mvld(prv)) {
	  err_set(1,"dir_list, internal error: mvld(prv).");
	  for (;mvld(prv);prv=Head.nxt) {
	    Head.nxt=prv->nxt;
	    free(prv);
	  }
	}
	if (*names&&n>1) qsort(*names,n,sizeof(char *),dir_alpha_asc);
      }
    }
  }
  for (i=0;i<nwld;i++) free(wld[i]);
  if (!(*names)) n=0;
  return(n);
}

/*-------------------- isDirectory --------------------*/

int isDirectory(char *name)
{
  int r;
  struct stat sbuff;

  if (stat(name,&sbuff)==0) {
    r=(S_IFDIR&sbuff.st_mode?1:0);
  }else r=0;
  return(r);
}
/*-------------------- isRegularFile --------------------*/

int isRegularFile(char *name)
{
  int r;
  struct stat sbuff;

  if (stat(name,&sbuff)==0) {
    r=(S_IFREG&sbuff.st_mode?1:0);
  }else r=0;
  return(r);
}
/*-------------------- dir_alpha_asc --------------------*/

int dir_alpha_asc(const void *str1,const void *str2)
{
  return(strcmp(*(char **)str1,*(char **)str2));
}
