#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "transpor.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define TRUE  1
#define FALSE 0

static mat_4 rot_parent(Dspobj *obj);
static mat_4 rtp_matrix(Dspobj *obj);
static void  buildMVT(Dspobj *obj);
static mat_4 dsp_transform_nogrfx(Dspobj *obj);
static mat_4 local_transform(Dspobj *obj);

/*-------------------- reset_object --------------------*/

void reset_object(Dspobj *obj)
{
  int i;

  obj=glblobj(obj);
  if (!is_dsp(obj)) return;

  if (graphix<1) {
    obj->rot=ident_4();
    for (i=0;i<3;i++) obj->tra[i]=0;
  }else{
    send_bheader(graphix,"ROBT",sizeof(obj));
    send_ptr(graphix,obj);
  }
}

/*-------------------- rotate_object --------------------*/

void rotate_object(Dspobj *obj, float *axis, float angle)
{
  mat_4 u,r;
  float zero[3] = {0,0,0};

  obj=glblobj(obj);
  if (!is_dsp(obj)) return;
  
  if (graphix<1) {
    u=rot_parent(obj);
    r=m4xm4(u,obj->rot);
    r=m4xm4(rotax(zero,axis,angle),r);
    obj->rot=m4xm4(mmxinv(u),r);
  }else{
    send_bheader(graphix,"RODO",sizeof(obj)+4*sizeof(float));
    send_bptr(graphix,obj);
    send_bfloat(graphix,angle);
    send_bfloata(graphix,3,axis);
  }
}

/*-------------------- set_pivot --------------------*/

void set_pivot(Dspobj *obj,float *piv)
{
  int i,j;

  obj=glblobj(obj);
  if (!is_dsp(obj)) return;

  if (graphix<1) {
    for (i=0;i<3;i++) {
      obj->tra[i]+=obj->pivot[i]-piv[i];
      for (j=0;j<3;j++) obj->tra[i]+=obj->rot.r[i].c[j]*(piv[j]-obj->pivot[j]);
    }
    for (i=0;i<3;i++) obj->pivot[i]=piv[i];
  }else{
    send_bheader(graphix,"NPIV",sizeof(obj)+3*sizeof(float));
    send_bptr(graphix,obj);
    send_bfloata(graphix,3,piv);
  }
}
/*-------------------- translate_object --------------------*/

void translate_object(Dspobj *obj,float *dx)
{
  mat_4 u;
  vec_3 t;

  obj=glblobj(obj);
  if (!is_dsp(obj)) return;

  if (graphix<1) {
    u=mmxinv(rot_parent(obj));
    t=v3pv3(vec3(obj->tra),m4xv3(u,vec3(dx)));
    v32arr(obj->tra,t);
  }else{
    send_bheader(graphix,"TRDO",sizeof(obj)+3*sizeof(float));
    send_bptr(graphix,obj);
    send_bfloata(graphix,3,dx);
  }
}

/*-------------------- get_RTP --------------*/

void get_RTP(Dspobj *root)
{
  Dspobj *obj;

  root=glblobj(root);
  if (!is_dsp(root)) return;
  
  if (graphix<1) {
    ForDspobjs(obj,root) obj->mtrx=rtp_matrix(obj);
    root->mtrx=rtp_matrix(root);
  }else{
    get_tree_matrices(root,"GRTP");
  }
}

/*-------------------- put_RTP --------------------*/

void put_RTP(Dspobj *obj,mat_4 *mtrx)
{
  int i;

  obj=glblobj(obj);
  if (!is_dsp(obj)) return;

  if (graphix<1) {
    obj->rot=*mtrx;
    for (i=0;i<3;i++) {
      obj->pivot[i]=obj->rot.r[i].c[3];
      obj->tra[i]=obj->rot.r[3].c[i];
      obj->rot.r[i].c[3]=0;
      obj->rot.r[3].c[i]=0;
    }
    obj->rot.r[3].c[3]=1;
  }else{
    send_dsp_matrix_type(obj,&obj->mtrx,"SRTP");
  }
}

/*-------------------- dsp_transform --------------------*/

mat_4 dsp_transform(Dspobj *obj)
{
  mat_4 mtrx;

  obj=glblobj(obj);
  if (is_dsp(obj)) {

    if (graphix<1) {
      mtrx=dsp_transform_nogrfx(obj);
    }else{
      mtrx=dsp_transform_grfx(obj);
    }

  }else mtrx=ident_4();

  return(mtrx);
}

/*-------------------- get_MVT --------------------*/

void get_MVT(Dspobj *obj)
{
  Dspobj *o;

  obj=glblobj(obj);
  if (!is_dsp(obj)) return;

  if (graphix<1) {
    obj->mtrx=dsp_transform_nogrfx(obj);
    forDspobjs(o,obj) buildMVT(o);
  }else{
    get_tree_matrices(obj,"GMMx");
  }
}

/*========================================================================*/

static mat_4 rot_parent(Dspobj *obj)
{
  mat_4 u;
  Dspobj *p;

  /* get cumulative rotation matrix of parent of obj */

  u=ident_4();
  for (p=(Dspobj *)obj->up;p;p=(Dspobj *)p->up) {
    if (is_dsp(p)&&p->dsptype==DSP_GLBL) u=m4xm4(p->rot,u);
  }
 
  return(u);
}

static mat_4 rtp_matrix(Dspobj *obj)
{
  mat_4 m;
  int i;

  /* pack rot, tra and pivot into a 4x4 matrix suitable for SRTP request */

  if (is_dsp(obj) && obj->dsptype==DSP_GLBL) {
    m=obj->rot;
    for (i=0;i<3;i++) {
      m.r[i].c[3]=obj->pivot[i];
      m.r[3].c[i]=obj->tra[i];
    }
  }else m=ident_4();
  
  return(m);
}

static void buildMVT(Dspobj *obj)
{
  Dspobj *o;

  /* recursively build MVT of global display object */

  if (is_dsp(obj)) {
    if (obj->dsptype==DSP_GLBL) {
      obj->mtrx=local_transform(obj);
      if (obj->up) obj->mtrx=m4xm4(((Dspobj *)obj->up)->mtrx,obj->mtrx);
    }
    forDspobjs(o,obj) buildMVT(o);
  }  
}

static mat_4 dsp_transform_nogrfx(Dspobj *obj)
{
  mat_4 mtrx;

  if (is_dsp(obj)) {
    if (obj->dsptype==DSP_GLBL) {
      if (obj->up) mtrx=dsp_transform_nogrfx((Dspobj *)obj->up);
      else mtrx=ident_4();
      mtrx=m4xm4(mtrx,local_transform(obj));
    }
  }
  return(mtrx);
}

static mat_4 local_transform(Dspobj *obj)
{
  mat_4 mtrx;
  vec_3 t;
  int i;

  if (is_dsp(obj)) {
    if (obj->dsptype==DSP_GLBL) {
      mtrx=obj->rot;
      t=m4xv3(mtrx,vec3(obj->pivot));
      t=v3pv3(vec3(obj->tra),v3mv3(vec3(obj->pivot),t));
      for (i=0;i<3;i++) mtrx.r[i].c[3]=t.c[i];
    }else mtrx=ident_4();
  }else mtrx=ident_4();
  return(mtrx);
}
