#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"

extern int graphix;
extern Dspobj *DspRoot;
extern Dspobj *PickRoot;

static char current_pick_mode[5]="????";
static char pending_pick_mode[5]="POFF";

#define TRUE 1
#define FALSE 0

/*---------- delete display structures associated with object ----------*/

void del_assoc_dsp(Node *obj)
{
  Atom *a;
  int i,n,size;
  
  if (obj) {
    switch (obj->type) {
    case OBJ_MOL:
      if (is_dsp(((Mol *)obj)->dsptr)) {
	if (((Mol *)obj)->moltype!=MOL_ROOT) 
	  del_node((Node *)((Mol *)obj)->dsptr);
      }else{
	n=0; ForAtoms(a,obj) if (cPtr(a->cross)) n++;
	if (n>0) {
	  size=n*cPtrSize();
	  send_bheader(graphix,"DATM",size);
	  ForAtoms(a,obj) {
	    if (cPtr(a->cross)) {
	      send_bcptr(graphix,a->cross);
	      a->cross=cPtrNULL();
	      for (i=0;i<a->nbnd;i++) a->bnd[i]->cross=cPtrNULL();
	    }
	  }
	}
      }
      break;
    case OBJ_ATOM:
      a=(Atom *)obj;
      if (cPtr(a->cross)) {
	send_bheader(graphix,"DATM",cPtrSize());
	send_bcptr(graphix,a->cross);
	trflush(graphix);
	a->cross=cPtrNULL();
      }
      for (i=0;i<a->nbnd;i++) a->bnd[i]->cross=cPtrNULL();
      break;
    default:
      break;
    }
  }
}
/*----------  inform graphics box about deleted objects ----------*/

void del_graph(Dspobj *d)
{
  if (d) {
    if (d->type==OBJ_DSP) {
      if (cPtr(d->cross)) {
	switch (d->dsptype) {    /* all fall through, but... */
	case DSP_GLBL:
	case DSP_MOL:
	case DSP_CSO:
	case DSP_CNT:
	case DSP_TRCE:
	default:
	  send_bheader(graphix,"DOBJ",sizeof(d));
	  send_bptr(graphix,d);
	  d->cross=cPtrNULL();
	  break;
	}
      }
    }
  }
}
/*-------------------- add bond to graphics --------------------*/

cptr_t add_bond_graph(Bond *b)
{
  Qmsg *msg;
  cptr_t cross;

  if (mvld(b)) {
    cross=cPtrNULL();
    if (!cPtr(b->cross)) {
      if (cPtr(b->f->cross) && cPtr(b->t->cross)) {
	send_bheader(graphix,"ABND",2*cPtrSize());
	send_bcptr(graphix,b->f->cross);
	send_bcptr(graphix,b->t->cross);
	msg=waitqmsg("BND*");
	if (msg) {
	  if (msg->size>=cPtrSize())  memcpy(&cross,msg->msg,cPtrSize());
	  b->cross=cross;
	  delqmsg(msg);
	}
      }
    }
  }else{
    cross=cPtrNULL();
  }
  return(cross);
}

/*-------------------- delete bond in graphics box ----------*/

void del_bond_graph(Bond *b)
{
  if (b) {
    if (cPtr(b->cross)) {
      send_bheader(graphix,"DBND",cPtrSize());
      send_bcptr(graphix,b->cross);
      b->cross=cPtrNULL();
    }
  }
}
/*-------------------- get_graph_xyz --------------------*/

void get_graph_xyz(int n,float *x,Atom **a)
{
/*  return graphics transformed (scale=1) coordinates for n atoms a[] */
/*  should be optimized to minimize number of GMMx requests !!!       */ 

  float rot[3][3],tra[3],sca;
  int i,j,k;

  for (k=0;k<n;k++) {
    if (mvld(a[k])) {
      m4factor(rot,tra,&sca,obj_transform((Node *)a[k]));
      for (i=0;i<3;i++) {
	x[i]=tra[i];
	for (j=0;j<3;j++) x[i]+=rot[i][j]*a[k]->x[j];
      }
    }else{
      for (i=0;i<3;i++) x[i]=0.0;
    }
    x+=3;
  }
}
/*-------------------- pick_root --------------------*/

Dspobj *pick_root(Dspobj *new)
{
  Dspobj *old;
  
  old=PickRoot;
  if (new) PickRoot=new;
  return(old);
}

/*-------------------- pick_mode --------------------*/

void pick_mode(char *mode)
{
  if (strneq(mode,"EUSA",4)||strneq(mode,"DUSA",4)) {
    send_bheader(graphix,mode,sizeof(PickRoot));
    send_ptr(graphix,PickRoot);
  }else{
    Strcpy(pending_pick_mode,mode);
  }
}
/*-------------------- do_pick --------------------*/

void do_pick(char *mode)
{
/*  send pick token to graphix box */

  if (streq(mode,"default")) {
    Strcpy(current_pick_mode,pending_pick_mode);
  }else{
    Strcpy(current_pick_mode,mode);
    Strcpy(pending_pick_mode,mode);
  }
  send_bheader(graphix,current_pick_mode,sizeof(PickRoot));
  send_ptr(graphix,PickRoot);
}
