#include <stdarg.h>
#include <stdio.h>
#include "mystdlib.h"
#include "errhand.h"
#include "strutil.h"

#define MAX_DEPTH 50
#define MAX_EDIT 8196

typedef struct msg {
  struct msg *prv,*nxt;
  char *t;
  unsigned long sq;
  int n;
} msg;

/*-------------------- static variables --------------------*/

static msg *Top=(msg *)NULL;         /* most recent message */
static msg *Bot=(msg *)NULL;         /* least recent message */
static int depth=0;                  /* number of kept messages */
static int max_depth=MAX_DEPTH;      /* max number of kept messages */
static int since=0;                  /* number of errors since last clear */
static int seqnr=0;

/*-------------------- err_clear --------------------*/

int err_clear(void)
{
/* delete all pending messages, return number of errors since last clear */

  msg *d,*n;
  int r;

  for (d=Bot;d;d=n) {
    n=d->nxt;
    if (mvld(d->t)) free(d->t);
    free(d);
  }
  r=since;
  Bot=(msg *)NULL;
  Top=Bot;
  depth=0;
  since=0;
  return(r);
}
/*-------------------- err_set --------------------*/

void err_set(int n,char *format,...)
{
  msg *new;
  va_list args;
  char edit_buffer[MAX_EDIT];
  int k;

  if (format) {
    va_start(args,format);
    vsprintf(edit_buffer,format,args);
    va_end(args);
    k=strlen(edit_buffer);
    /*AWI 18-Feb-2009: if (k>0&&edit_buffer[k-1]=='\n') edit_buffer[k-1]='\0'; */
    if (k<1||edit_buffer[k-1]!='\n') strcpy(edit_buffer+k,"\n");
  }
  new=(msg *)malloc(sizeof(msg));
  if (mvld(new)) {
    if (n!=0) {
      seqnr++;
      since++;
    }
    new->prv=Top;
    new->nxt=(msg *)NULL;
    new->sq=seqnr;
    new->n=n;
    if (format) new->t=Strdup(edit_buffer);
    else new->t=(char *)NULL;
    depth++;
  }
  if (mvld(Top)) Top->nxt=new;
  else Bot=new;
  Top=new;
  if (depth>max_depth) err_set_maxdepth(max_depth);
}
/*-------------------- err_set_maxdepth --------------------*/

void err_set_maxdepth(int new_max_depth)
{
  max_depth=new_max_depth;
  if (depth>max_depth) err_dequeue(depth-max_depth);
}
/*-------------------- err_pop --------------------*/

void err_pop(int n)
{
  int i;
  msg *new;

  for (i=0;i<n&&depth>0;i++) {
    if (mvld(Top)) {
      new=Top->prv;
      if (mvld(Top->t)) free(Top->t);
      if (Top->n!=0) {
	since--;
	seqnr--;
      }
      free(Top);
      Top=new;
      if (mvld(Top)) Top->nxt=(msg *)NULL;
      else Bot=Top;
      depth--;
    }
  }
}
/*-------------------- err_dequeue --------------------*/

void err_dequeue(int n)
{
  int i;
  msg *new;

  for (i=0;i<n&&depth>0;i++) {
    if (mvld(Bot)) {
      new=Bot->nxt;
      if (mvld(Bot->t)) free(Bot->t);
      if (mvld(Bot)) free(Bot);
      Bot=new;
      if (mvld(Bot)) Bot->prv=(msg *)NULL;
      else Top=Bot;
      depth--;
    }
  }
}
/*-------------------- err_nr --------------------*/

int err_nr(void)
{
  return(Top ? Top->n : 0);
}

int err_nr_bot(void)
{
  return(mvld(Bot) ? Bot->n : 0);
}
/*-------------------- err_text --------------------*/

char *err_text(void)
{
  return(mvld(Top) ? (mvld(Top->t) ? Top->t : "Error ???") : (char *)NULL);
}
char *err_text_bot(void)
{
  return(mvld(Bot) ? (mvld(Bot->t) ? Bot->t : "Error ???") : (char *)NULL);
}
/*-------------------- err_since --------------------*/

int err_since(void)
{
  return(since);
}
/*-------------------- err_seqnr --------------------*/

int err_seqnr(void)
{
  return(seqnr);
}
/*-------------------- err_depth --------------------*/

int err_depth(void)
{
  return(depth);
}
/*-------------------- err_maxdepth --------------------*/

int err_maxdepth(void)
{
  return(max_depth);
}
