#ifndef FELIXDEF
#define FELIXDEF

#include <stdio.h>
#include "fileutil.h"
#include "nmr.h"

/* Constants: */

#define FELIXMDIM      3        /* maximum dimension that we can handle... */
#define BYTESEX_NONE   0        /* relative bytesex is unknown */
#define BYTESEX_HOMO   1        /* processor has same bytesex as file */
#define BYTESEX_HETERO 2        /* processor's bytesex differs from file's */
#define FELIX_BAD_VALUE -1.e33  /* undef grid value... */
#define FELIX_ACC_CURRENT 0
#define FELIX_ACC_NEAREST 1     /* access: nearest gridpoint */
#define FELIX_ACC_TRILIN  2     /*   "   : tr-linear interpolation */
#define FELIX_ACC_MAX     3     /*   "   : max-abs value 0f 8 surrounding points */


/* Type Definitions: */

typedef struct {            /* a word is 4 unsigned chars */
  unsigned char c[4];
} WORD;

typedef struct {            /* a brick is a block of 4096 words */
  WORD w[4096];
} BRICK;

typedef struct {       /* structure to hold header infor of mat file */
  int   ok;                   /* valid header ? */
  int   dim;                  /* matrix dimension */
  int   sex;                  /* files's relative bytesex (1:homo,2:hetero) */
  int   type;                 /* matrix type ("real" or "complex") */
  int   nbricks;              /* total number of bricks */
  int   mbricks;              /* largest no. of bricks in one dimension */
  int   sqz;                  /* squeezed or not */
  int   start[FELIXMDIM];     /* starting points */
  int   npnts[FELIXMDIM];     /* total number of points */
  int   nbrck[FELIXMDIM];     /* total number of bricks */
  int   boff[FELIXMDIM];      /* brick offset in file records */
  int   nppb[FELIXMDIM];      /* points per brick */
  int   poff[FELIXMDIM];      /* point offset in brick */
  float refsf[FELIXMDIM];     /* reference SF */
  float refsw[FELIXMDIM];     /* reference SW */
  float refpnt[FELIXMDIM];    /* reference point */
  float refppm[FELIXMDIM];    /* reference ppm */
  int   refaxt[FELIXMDIM];    /* reference axis type */
  float appm[FELIXMDIM];      /* a-coefficient for pts to ppm conversion */
  float bppm[FELIXMDIM];      /* b-coefficient for pts to ppm conversion */
  float apts[FELIXMDIM];      /* a-coefficient for ppm to pts conversion */
  float bpts[FELIXMDIM];      /* b-coefficient for ppm to pts conversion */
  float as[2];                /* convert elt. specific ppm to spectral ppm */
  float bs[2];                /* convert elt. specific ppm to spectral ppm */
  int   nfiles;               /* no. of physical files */
  int   brcksps;              /* bricks per segment */
} FELIXMATHEADER;

typedef struct FELIXMATRIX {
  int fildes;
  int bricknr;
  float noise;                /* noise level */
  float av,sd;                /* average intensity, std. dev. intensity */
  int autofold;               /* automatically 'fold' shifts into window ? */
  int accm;                   /* access method (1: nearest, 2: interpolate) */
  char *name;                 /* a name, usually derived from file name */
  FELIXMATHEADER *hdr;
  nmrPeakList *P;
  BRICK *brick;
} FELIXMATRIX;

typedef FELIXMATRIX *FELIXMATRIXP;

/* Prototypes: */

FELIXMATRIXP FelixOpenMatR(int fildes);
FELIXMATRIXP FelixCloseMat(FELIXMATRIXP matrix);
int   FelixReadBrick(int fildes,BRICK *brick);
int   FelixReadBrickN(int fildes,int no,BRICK *brick);
float FelixWord2Float(int sex,WORD w);
int   FelixWord2Int(int sex,WORD w);
FELIXMATHEADER FelixReadMatHeader(int filedes);
void  FelixPrintMatHeader(FILE *out,FELIXMATHEADER *hdr);
float FelixMatValue(FELIXMATRIXP mat,int *p);
float FelixMatValuePPM(FELIXMATRIXP mat,float *p);
float FelixMatValuePPMr(FELIXMATRIXP mat,float *p);
float FelixMatValuePPMi(FELIXMATRIXP mat,float *p);
float FelixMatValuePPMm(FELIXMATRIXP mat,float *p);
float FelixMatRValuePNT(FELIXMATRIXP mat,float *p);
float FelixMatValuePNT(FELIXMATRIXP mat,float *p);
float FelixMatValuePNTr(FELIXMATRIXP mat,float *p);
float FelixMatValuePNTi(FELIXMATRIXP mat,float *p);
float FelixMatValuePNTm(FELIXMATRIXP mat,float *p);
void  FelixPPM2Pnt(FELIXMATRIXP mat,float *ppm,int *ip,float *pt);
int   FelixMapPPM2Point(FELIXMATRIXP mat,float *ppm,int z,float *pnt);
float FelixPPM2ppm(FELIXMATRIXP mat,int dim,int z,float ppm);
float FelixInterpolate3D(float *pt,float *v);
void  FelixPoint2BW(FELIXMATRIXP mat,int *p,int *bnr,int *wnr);
void  FelixGetBrickN(FELIXMATRIXP mat,int bnr);
float FelixSetNoiseLevel(FELIXMATRIXP mat,float noise,float prob);
float FelixVShift(FELIXMATRIXP mat,int z,float ppm);

nmrPeakList      read_nmrPeak_file(UCpipe *in);
void             FelixWrite_nmrPeak_file(FILE *out,char *name,nmrPeakList *L);

#endif
