/*  fileutil.c  */

#include <stdio.h>
#include "mystdlib.h"
#include "strutil.h"
#include "fileutil.h"
#include <errno.h>
#include "errhand.h"
#include <unistd.h>
#include <sys/stat.h>
#ifndef mingw
#include <sys/wait.h>
#include <sys/types.h>
#endif
#include "errhand.h"

#define NO 0
#define YES 1

#define TRUE 1
#define FALSE 0

#define MAXBUF 4096

/*------------------------------------------------------------------------*/

UCpipe *UCpopen(char *file)
{
/* Open file for read . If file does not exist, try compressed file
   (extension .Z) and open through pipe. Return NULL if open fails */

  char longfn[2048];
  char pipe_command[2048];
  UCpipe *ucp;
  int pipe,err;
  FILE *f;

  ucp=(UCpipe *)NULL;
  pipe=UCP_FILE; err=FALSE;
  errno=0;

  if (*file=='|') {                              /* a pipe... */
    f=PipeOpen(file+1,"r");
    if (!f) {
      err_set(errno,"UCpopen(%s): %s",file,strerror(errno));
      err=TRUE;
    }else if (ferror(f)||feof(f)) {
      err_set(errno,"UCpopen(%s): %s",file,strerror(errno));
      err=TRUE;
      fclose(f);
      f=(FILE *)NULL;
    }else{
      sprintf(longfn,"%s",file);
      pipe=UCP_PIPE;
    }
  }else{
    if (strlen(file)>2) {
      if (streq(file+strlen(file)-2,".Z")) pipe=UCP_PIPE; 
    }
    sprintf(longfn,"%s",file);
    f=fopen(longfn,"r");
    if (!f) {
      err_set(errno,"UCpopen(%s): %s",longfn,strerror(errno));
      err=TRUE;
      if (pipe==UCP_FILE && errno==ENOENT) {
	pipe=UCP_PIPE;
	sprintf(longfn,"%s.Z",file);
	f=fopen(longfn,"r");
      }
    }
    if (f) {
      if (pipe==UCP_PIPE) {   /* should check if file is really compressed (how ?) */
	if (err) err_pop(1);
	err=FALSE;
	fclose(f);
	sprintf(pipe_command,"compress -d -c %s",longfn);
	f=PipeOpen(pipe_command,"r");
	if (!f) {
	  err_set(errno,"UCpopen(%s): %s",longfn,strerror(errno));
	  err=TRUE;
	}
      }
    }
  }

  if (f) {
    ucp=(UCpipe *)malloc(sizeof(UCpipe));
    if (ucp) {
      ucp->pipe=pipe;
      ucp->filename=Strdup(longfn);
      ucp->block=NULL;
      ucp->bptr=NULL;
      ucp->fd=f;
      ucp->eof=FALSE;
      if (!(ucp->filename)) {
	free(ucp);
	ucp=(UCpipe *)NULL;
      }
    }
  }
  return(ucp);
}
 
char *UCpgets(char *s,int size,UCpipe *p)
{
  char *r=NULL;
  int sz;

  if (p) {
    if (p->pipe==UCP_FILE || p->pipe==UCP_PIPE) {
      r=fgets(s,size,p->fd);
    }else if (p->pipe==UCP_BLOCK && p->bptr) {
      if (p->bptr && strlen(p->bptr)>0) {
	r=strchr(p->bptr,'\n');
	if (r) {
	  r++;
	  sz=r-p->bptr>size?size:r-p->bptr;
	  strncpy(s,p->bptr,sz);
	  p->bptr+=sz;
	}else{
	  sz=strlen(p->bptr)>size?size:strlen(p->bptr);
	  strncpy(s,p->bptr,sz);
	  p->bptr+=sz;
	}
	r=s;
	r[sz]='\0';
	if (strrchr(r,'\r')) *strrchr(r,'\r')='\0';
	p->eof=strlen(p->bptr)==0;
      }else{
	p->eof=TRUE;
	r=NULL;
      }
    }
  }
  return r;
}

UCpipe *UCpblock(char *block)
{
  UCpipe *ucp=NULL;

  /* create UCpipe from data blcok */

  ucp=(UCpipe *)malloc(sizeof(UCpipe));
  if (ucp) {
    ucp->filename=strdup("textblock");
    ucp->block=block;
    ucp->bptr=ucp->block;
    ucp->pipe=UCP_BLOCK;
    ucp->eof=FALSE;
  }
  return(ucp);
}

int UCpclose(UCpipe *ucp)
{
  int s=0;

  if (ucp) {
    if (ucp->pipe==UCP_PIPE) {
      s=PipeClose(ucp->fd);
      /* if (s) err_set(1,"UCpclose: %s, exit status %d",ucp->filename,s); */
      if (ucp->filename) free(ucp->filename);
    }else if (ucp->pipe==UCP_FILE) {
      s=fclose(ucp->fd);
      if (ucp->filename) free(ucp->filename);
    }else if (ucp->pipe==UCP_BLOCK) {
      if (ucp->filename) free(ucp->filename);
    }
    free(ucp);
  }
  return(s);
}

int UCpeof(UCpipe *ucp)
{
  int eof=0;

  if (ucp->pipe==UCP_PIPE || ucp->pipe==UCP_FILE) {
    eof=feof(ucp->fd);
  }else if (ucp->pipe==UCP_BLOCK && ucp->bptr) {
    eof=ucp->eof;
  }else{
    eof=TRUE;
  }
  return(eof);
}
 
int UCperror(UCpipe *ucp)
{
  int error=0;

  if (ucp->pipe==UCP_PIPE || ucp->pipe==UCP_FILE) {
    error=ferror(ucp->fd);
  }else if (ucp->pipe==UCP_BLOCK && ucp->bptr) {
    error=FALSE;
  }else{
    error=TRUE;
  }
  return(error);
}
 
int UCpgetc(UCpipe *ucp)
{
  int r=EOF;
  
  if (ucp->pipe==UCP_FILE || ucp->pipe==UCP_PIPE) {
    r=fgetc(ucp->fd);
  }else if (ucp->pipe==UCP_BLOCK && ucp->bptr) {
    if (ucp->bptr && strlen(ucp->bptr)>0) {
      r=*ucp->bptr;
      ucp->bptr++;
    }else{
      r=EOF;
      ucp->eof=TRUE;
    }
  }
  return r;
}

/*------------------------------------------------------------------------*/

FileCache *makeFileCache(int size,int limit)
{
  FileCache *cache;

  if (size<1) size=1;
  if (limit<1) limit=1;

  cache=(FileCache *)malloc(sizeof(FileCache));
  if (mvld(cache)) {
    cache->N=size; cache->n=0;
    cache->limit=limit;
    cache->nr=(int *)malloc(size*sizeof(int));
    cache->pos=(long *)malloc(size*sizeof(long));
    if (!mvld(cache->nr)||!mvld(cache->pos)) cache=freeFileCache(cache);
  }
  return(cache);
}

FileCache *freeFileCache(FileCache *cache)
{
  if (mvld(cache)) {
    if (mvld(cache->nr))  free(cache->nr);
    if (mvld(cache->pos)) free(cache->pos);
    free(cache);
  }
  return((FileCache *)NULL);
}

int findFileCachePos(FileCache *cache,int target)
{
  int k;
  
  for (k=0;k<cache->n;k++) if (cache->nr[k]>target) break;
  k--;
  return(k);
}

void addFileCachePos(FileCache *cache,int now,long pos)
{
  int k,l,delta;

  if (cache->n==0) {
    cache->pos[cache->n]=pos;
    cache->nr[cache->n]=now;
    cache->n++;
  }else{
    k=findFileCachePos(cache,now);
    delta=now-cache->nr[k];
    if (delta>cache->limit) {
      if (cache->n>k+1) {
	if (cache->nr[k+1]-now>cache->limit) {
	  if (cache->n<cache->N) {
	    for (l=cache->n-1;l>k;l--) {
	      cache->nr[l+1]=cache->nr[l];
	      cache->pos[l+1]=cache->pos[l];
	    }
	    cache->n++;
	    cache->nr[k+1]=now;
	    cache->pos[k+1]=pos;
	  }else{
	    purgeFileCache(cache);
	    addFileCachePos(cache,now,pos);
	  }
	}
      }else if (cache->n<cache->N) {
	cache->pos[cache->n]=pos;
	cache->nr[cache->n]=now;
	cache->n++;
      }else{
	purgeFileCache(cache);
	cache->pos[cache->n]=pos;
	cache->nr[cache->n]=now;
	cache->n++;
      }
    }
  }
}
/*-------------------- purgeFileCache --------------------*/

void purgeFileCache(FileCache *cache)
{
  int k,l,delta;

  if (cache->n<2) return;
  delta=cache->nr[1]-cache->nr[0];
  for (l=0,k=1;k<cache->n-1;k++) {
    if (delta<cache->nr[k+1]-cache->nr[k]) {
      delta=cache->nr[k+1]-cache->nr[k]; l=k;
    }
  }
  for (k=l+1;k<cache->n;k++) {
    cache->nr[k-1]=cache->nr[k];
    cache->pos[k-1]=cache->pos[k];
  }
  cache->n--;
}

/*-------------------- FindFileName --------------------*/

int FindFileName(char *path,char *name,char *fullname,int nl)
{
  /* Find file with given name in path, return full name in fullname  */
  /* nl is size of 'fullname' (input). Function returns TRUE if file  */
  /* with read access is found, FALSE otherwise. 'path' is a list     */
  /* of colon separated directories, or an environment variable       */
  /* if path[0]=='$'. path[0]='|' name[0]=='|' is special: a pipe.    */
  /* For pipes, the first word (sep: "()|;<>&") is used to search an  */
  /* executable file with that name in path. If such a file is found, */
  /* fullname is set to a "pipe" command with the first word expanded */
  /* to the full name (|path/...), otherwise fullename will be set to */
  /* name without modification. Note: for pipes, only the first word  */
  /* will be expanded by path.                                        */
  
  Parse p,q;
  char cmd[MAXBUF];
  char *dirs=(char *)NULL;
  int r=FALSE;

  if (name) {
    if (name[0]=='|') {                       /*special 'name' is a pipe*/
      q.buf=name+1; q.pos=0; 
      q.tok=cmd; q.mts=sizeof(cmd);
      q.sep="()|;<>&";
      if (get_token(&q)&&q.ltok>0) {
	if (strchr(cmd,'/')) {
	  if (FileExecutable("",cmd,cmd,MAXBUF)&&nl>strlen(name)) {
	    strcpy(fullname,name);
	    r=TRUE;
	  }
	}else{
	  fullname[0]='|';
	  if (path[0]=='$') dirs=Strdup(getenv(path+1));
	  else dirs=Strdup(path);
	  if (dirs&&strlen(dirs)==0) {
	    free(dirs);
	    dirs=(char *)NULL;
	  }
	  if (dirs) {
	    p.sep=":"; p.tok=fullname+1; p.mts=nl;
	    p.buf=dirs; p.pos=0;
	    while (!r&&token(&p)) r=FileExecutable(fullname+1,cmd,fullname+1,nl-1);
	    if (r) {
	      if (strlen(fullname)+strlen(name+q.pos)+1<nl) {
		sprintf(fullname+strlen(fullname),"%c%s",q.term,q.buf+q.pos);
		r=TRUE;
	      }
	    }else if (nl>strlen(name)) {
	      strcpy(fullname,name);
	      r=TRUE;
	    }
	  }else{
	    if (nl>strlen(name)) {
	      strcpy(fullname,name);
	      r=TRUE;
	    }
	  }
	}
      }else{
	if (nl>strlen(name)) {
	  strcpy(fullname,name);
	  r=TRUE;
	}
      }

    }else if (strchr(name,'/')) {             /* 'name' is a file */
      r=FileReadable("",name,fullname,nl);
    }else{
      if (path) {
	if (path[0]=='$') dirs=Strdup(getenv(path+1));
	else dirs=Strdup(path);
	if (dirs&&strlen(dirs)==0) {
	  free(dirs);
	  dirs=(char *)NULL;
	}
      }
      if (dirs) {
	p.sep=":"; p.tok=fullname; p.mts=nl;
	p.buf=dirs; p.pos=0;
	while (!r&&token(&p)) r=FileReadable(fullname,name,fullname,nl);
      }else{
	r=FileReadable("",name,fullname,nl);
      }
    }
  }else r=FALSE;
  
  if (mvld(dirs)) free(dirs);
  return(r);
}
/*-------------------- FileReadable --------------------*/

int FileReadable(char *dir,char *name,char *fullname,int n)
{
  /* Check if dir/name is a reg. file to which we have read-access */
  /* Copy full name to fullname and return TRUE, if yes.           */

  return(FileAccessible(R_OK,dir,name,fullname,n));
}
/*-------------------- FileReadable --------------------*/

int FileExists(char *dir,char *name,char *fullname,int n)
{
  /* Check if dir/name is a reg. file to which we have read-access */
  /* Copy full name to fullname and return TRUE, if yes.           */

  return(FileAccessible(F_OK,dir,name,fullname,n));
}
/*-------------------- FileExecutable --------------------*/

int FileExecutable(char *dir,char *name,char *fullname,int n)
{
  /* Check if dir/name is a reg. file to which we have read-access */
  /* Copy full name to fullname and return TRUE, if yes.           */

  return(FileAccessible(X_OK,dir,name,fullname,n));
}
/*-------------------- FileAccessible --------------------*/

int FileAccessible(int amode,char *dir,char *name,char *fullname,int n)
{
  /* Check if dir/name is a reg. file to which we have 'amode' 
     access (c.f. access(2)). Copy full dir/name to fullname and 
     return TRUE, if yes. Return FALSE else, and on all errors.   */

  int r;
  struct stat info;

  r=FALSE;
  if (dir&&name) {
    if (strlen(dir)+strlen(name)+1<n) {
      if (strlen(dir)>0) {
	strcpy(fullname,dir);
	if (fullname[strlen(fullname)-1]!='/') strcat(fullname,"/");
	strcat(fullname,name);
      }else strcpy(fullname,name);
      if (!access(fullname,amode)) {
	if (!stat(fullname,&info)) {
	  if (S_ISREG(info.st_mode)) r=TRUE;
	}
      }
    }
  }
  return(r);
}

/*------------------------------------------------------------------------*/

/* Replacements for popen and pclose functions */
/* PipeOpen finds shell in /bin, or in $PATH  */

static pid_t *openpid = NULL;
static int maxopen=0;

FILE *PipeOpen(char *cmd,char *type)
{
#if defined(mingw)
  return(popen(cmd,type));
#else
  int	i,fd[2];
  pid_t	pid;
  FILE	*fp=(FILE *)NULL;

  if (*type=='r' || *type=='w') {
    if (!openpid) {            /* need to allocate memory for pids ... */
      maxopen=sysconf(_SC_OPEN_MAX);
      if (maxopen<=0) maxopen=512;
      openpid=(pid_t *)calloc(maxopen,sizeof(pid_t));
    }
    if (openpid) {
      if (!pipe(fd)) {

	pid=fork();

	if (pid==0) {          /* subprocess */
	  if (*type=='r') {
	    close(fd[0]);
	    if (fd[1] != STDOUT_FILENO) {
	      dup2(fd[1], STDOUT_FILENO);
	      close(fd[1]);
	    }
	  }else if (*type=='w') {
	    close(fd[1]);
	    if (fd[0] != STDIN_FILENO) {
	      dup2(fd[0], STDIN_FILENO);
	      close(fd[0]);
	    }
	  }
	  /* close open pipes in child */
	  for (i=0;i<maxopen;i++) if (openpid[i]>0) close(i);
	  /* execute shell from /bin, or whatever is found in path */
	  if (!access("/bin/sh",X_OK)) execl("/bin/sh","sh","-c",cmd,NULL);
	  else if (!access("/bin/sh.exe",X_OK)) execl("/bin/sh","sh","-c",cmd,NULL);
	  else execlp("sh","sh","-c",cmd,NULL);
	  exit(127);           /* here only if execl failed */

	}else if (pid>0){      /* parent */
	  if (*type=='r') {
	    close(fd[1]);
	    fp=fdopen(fd[0],type);
	  }else if (*type=='w') {
	    close(fd[0]);
	    fp=fdopen(fd[1],type);
	  }
	  if (fp) openpid[fileno(fp)]=pid;  /* store pid of child for use by PipeClose */
	  else err_set(1,"PipeOpen: failed to open stream\n");

	}else err_set(1,"PipeOpen: failed to fork subprocess\n");

      }else err_set(1,"PipeOpen: failed to create pipe\n");
    }else err_set(1,"PipeOpen: memory allocation failure.\n");
  }else err_set(1,"PipeOpen: invalid access type\n");
  return(fp);
#endif
}

int PipeClose(FILE *fp)
{
#if defined(mingw)
  return(pclose(fp));
#else
  int	fd,status=-1;
  pid_t	pid;

  if (openpid) {
    fd=fileno(fp);
    if ((fd=fileno(fp))>=0 && (pid=openpid[fd])!=0) {
      openpid[fd]=0;
      if (!fclose(fp)) {
	while (waitpid(pid, &status,0)<0) {
	  if (errno!=EINTR) {
	    err_set(1,"PipeClose: error waiting for child");
	    status=-1;
	    break;
	  }
	}
      }else err_set(1,"PipeClose: invalid stream or pipe not open");
    }else err_set(1,"PipeClose: invalid stream or pipe not open");
  }else err_set(1,"PipeClose: no open pipes");
  return(status);
#endif
}
