#include "mystdlib.h"
#include "object.h"
#include "errhand.h"
#include "strutil.h"
#include "swdef.h"

#define TRUE  1
#define FALSE 0

#define MX_FRULE_NAT     250
#define MX_FRULE_NITP    500
#define MX_FRULE_NBND    250
#define MX_FRULE_RINGC    10
#define MX_FRULE_LPREN    20
#define MX_FRULE_BTYPES    5
#define MX_FRULE_TXT     512

#define DONE       0
#define ATOM       1
#define ATOMLIST   2
#define ERROR     99

#define ADV        1
#define RET        2

/*-------------------- parseFrule --------------------*/

Frule *parseFrule(char *txt,Lib *lib)
{
  /* parse frule string and convert to Frule structure */

  Frule *r;
  int btp[MX_FRULE_NBND],from[MX_FRULE_NBND],to[MX_FRULE_NBND];
  int nitp[MX_FRULE_NAT];
  Atomdef *iatp[MX_FRULE_NITP];
  int nat,nbnd;

  int i,l,fatom,ntitp,ra,pa,ratom[MX_FRULE_RINGC],patom[MX_FRULE_LPREN],mode;
  char buf[MX_FRULE_TXT];
  Atomdef *atp;
  Parse p;

  r=(Frule *)NULL;
  if (!txt||strlen(txt)<1||strlen(txt)>MX_FRULE_TXT||!is_lib(lib)) return(r);

  p.buf=txt; p.pos=0; p.mts=MX_FRULE_TXT; 
  p.tok=buf; p.sep="-=#~,()@";

  nat=0; nbnd=0; ntitp=0; nitp[0]=0;
  for (i=0;i<MX_FRULE_RINGC;i++) ratom[i]=-1;
  for (i=0;i<MX_FRULE_LPREN;i++) patom[i]=-1;
  pa=0; fatom=-1;

  for(mode=ATOM;mode!=DONE&&mode!=ERROR;) {
    if (get_token(&p)) {
      if (p.ltok>0) {
	atp=find_atomdef_name((Node *)lib,p.tok);
	if (mvld(atp)) {
	  if (nat<MX_FRULE_NAT) {
	    if (nbnd>0) {
	      if (to[nbnd-1]<0) to[nbnd-1]=nat;
	      else if (mode==ATOM) {
		mode=ERROR;
		break;
	      }
	    }
	    if (ntitp<MX_FRULE_NITP) {
	      iatp[ntitp++]=atp;
	      nitp[nat]++;
	      if (p.term!=',') {
		fatom=nat; nat++;
		if (nat<MX_FRULE_NAT) nitp[nat]=0;
	      }
	    }else mode=ERROR;
	  }else mode=ERROR;
	}else mode=ERROR;
      }else if (mode==ATOMLIST) mode=ERROR;
      switch (p.term) {
      case '-':
      case '=':
      case '#':
      case '~':
	mode=ATOM;
	if (nbnd<MX_FRULE_NBND&&fatom>=0) {
	  if (p.term=='-')      btp[nbnd]=BT_SINGLE;
	  else if (p.term=='=') btp[nbnd]=BT_DOUBLE;
	  else if (p.term=='#') btp[nbnd]=BT_TRIPLE;
	  else if (p.term=='~') btp[nbnd]=BT_AROMATIC;
	  from[nbnd]=fatom; to[nbnd]=-1;
	  nbnd++; fatom=-1;
	}else mode=ERROR;
	break;
      case ',':
	if (p.ltok>0) mode=ATOMLIST;
	else mode=ERROR;
	break;
      case '(':
	mode=ATOM;
	if (pa<MX_FRULE_LPREN) patom[pa++]=fatom;
	else mode=ERROR;
	break;
      case ')':
	mode=ATOM;
	if (pa>0) fatom=patom[--pa];
	else mode=ERROR;
	break;
      case '@':
	mode=ATOM;
	ra=p.buf[p.pos++]-'0';
	if (ra>=0&&ra<='9') {
	  if (p.ltok!=0) {
	    ratom[ra]=nat-1;
	  }else{
	    if (nbnd>0&&to[nbnd-1]==-1&&ratom[ra]>=0) {
	      to[nbnd-1]=ratom[ra];
	      fatom=from[nbnd-1];
	    }else mode=ERROR;
	  }
	}else mode=ERROR;
	break;
      }
    }else mode=DONE;
  }

  if (mode==DONE) {
    if (pa>0) mode=ERROR;
    if (nbnd>0&&to[nbnd-1]<0) mode=ERROR;
  }
    
  if (mode!=ERROR) {
    r=(Frule *)make_node(OBJ_FRULE);
    if (is_frule(r)) {
      if (nat>0) {
	r->oatp=(Atomdef **)malloc(nat*sizeof(Atomdef *));
	r->iatp=(Atomdef **)malloc(ntitp*sizeof(Atomdef *));
	r->nitp=(int *)malloc(ntitp*sizeof(int));
	r->nat=nat; r->match=(Atom **)NULL;
      }
      if (nbnd>0){
	r->btp=(int *)malloc(nbnd*sizeof(int));
	r->from=(int *)malloc(nbnd*sizeof(int));
	r->to=(int *)malloc(nbnd*sizeof(int));
	r->nbnd=nbnd;
      }
      if ((nat>0&&(!mvld(r->oatp)||!mvld(r->iatp)||!mvld(r->nitp)))||
	  (nbnd>0&&(!mvld(r->btp)||!mvld(r->from)||!mvld(r->to)))) {
	if (mvld(r->oatp)) free(r->oatp);
	if (mvld(r->iatp)) free(r->iatp);
	if (mvld(r->nitp)) free(r->nitp);
	if (mvld(r->btp))  free(r->btp);
	if (mvld(r->from)) free(r->from);
	if (mvld(r->to))   free(r->to);
	r->nat=0; r->nbnd=0;
	del_node((Node *)r);
	r=(Frule *)NULL;
      }else{
	r->nat=nat; r->nbnd=nbnd;
	for (i=0;i<nat;i++) {
	  r->oatp[i]=(Atomdef *)NULL;
	  r->nitp[i]=nitp[i];
	}
	for (i=0;i<ntitp;i++) {
	  r->iatp[i]=iatp[i];
	  if (is_atomdef(r->iatp[i])) r->iatp[i]->nlnks++;
	}
	for (i=0;i<nbnd;i++) {
	  r->btp[i]=btp[i];
	  r->from[i]=from[i];
	  r->to[i]=to[i];
	}
      }
    }
  }else{
    err_set(1,"bad frule: %s",txt);
    l=strlen("bad frule: ");
    for (i=0;i<p.pos+l-1;i++) buf[i]='-';
    buf[i]='^'; buf[i+1]='\0';
    err_set(1,"%s",buf);
  }
  return(r);
}
/*-------------------- matchFrule --------------------*/

int matchFrule(Atom *a,Frule *p)
{
  int r;
  int i,j,k,b,af,at,c[MX_FRULE_NBND],mode;
  Atomdef **iatp[MX_FRULE_NAT];
  Bond *bnd;
 
  r=FALSE;

  if (!is_frule(p)) return(r);

  for (i=0,k=0;i<p->nat;k+=p->nitp[i++]) iatp[i]=p->iatp+k;
  
  for (j=0;j<p->nitp[0];j++) if (iatp[0][j]==a->atp) break;
  if (j<p->nitp[0]) {
    if (!mvld(p->match)) {
      p->match=(Atom **)malloc(p->nat*sizeof(Atom *));
      if (!p->match) return(FALSE);
    }
    for (j=1;j<p->nat;j++) p->match[j]=(Atom *)NULL; /* 17-02-2002: move from inside prev. if */
    p->match[0]=a; a->flg|=(1<<1);
    for (b=0,c[b]=0;b>=0&&b<p->nbnd;) {
      af=p->from[b]; at=p->to[b];
      if (!mvld(p->match[af])) {
	err_set(1,"matchFrule: internal error...");
	break;
      }
      if (mvld(p->match[at])) {
	bnd=find_bond(p->match[af],p->match[at]);
	if (mvld(bnd)) {
	  if (bnd->bt==p->btp[b]||(bnd->bt==BT_AMIDE&&p->btp[b]==BT_SINGLE)) mode=ADV;
	  else mode=RET;
	}else mode=RET;
      }else{
	while (c[b]<p->match[af]->nbnd) {
	  bnd=p->match[af]->bnd[c[b]]; c[b]++;
	  a=(bnd->f==p->match[af] ? bnd->t : bnd->f);
	  if (a->flg&(1<<1)) continue;
	  if (bnd->bt==p->btp[b]||(bnd->bt==BT_AMIDE&&p->btp[b]==BT_SINGLE)) {
	    for (j=0;j<p->nitp[at];j++) if (iatp[at][j]==a->atp) break;
	    if (j<p->nitp[at]) {
	      p->match[at]=a; 
	      a->flg|=(1<<1);
	      break;
	    }
	  }
	}
	if (p->match[at]) mode=ADV;
	else mode=RET;
      }
      switch (mode) {
      case ADV:                /* current bond matched... advance */
	b++;
	if (b<p->nbnd) c[b]=0;
	break;
      case RET:                /* current bond does not match... retreat */
	do {
	  if (p->to[b]>p->from[b]) {
	    if (is_atom(p->match[p->to[b]])) p->match[p->to[b]]->flg&=~(1<<1);
	    p->match[p->to[b]]=(Atom *)NULL;
	    if (c[b]>=p->match[p->from[b]]->nbnd) b--;
	    else break;
	  }else b--;
	}while(b>=0);
	break;
      }
    }
    for (r=TRUE,i=0;i<p->nat;i++) {
      if (is_atom(p->match[i])) p->match[i]->flg&=~(1<<1);
      else r=FALSE;
    }
  }
  return(r);
}
