#ifndef GENLIGDEF
#define GENLIGDEF

#include "object.h"

#define GENL_DIH_UNK     -2
#define GENL_DIH_IGNORE  -1
#define GENL_DIH_ZERO     0
#define GENL_DIH_DOUBLE   1
#define GENL_DIH_AMIDE    2
#define GENL_DIH_AROMATIC 3
#define GENL_DIH_RING     4


#define GENL_IMP_UNK     -2
#define GENL_IMP_IGNORE  -1
#define GENL_IMP_ZERO     0
#define GENL_IMP_PLANAR   1
#define GENL_IMP_CHIRAL   2

#define MAXLIGATMNB 10

typedef struct glAtom_t {
  Atom *a;
  struct glAtom_t *n[MAXLIGATMNB];
  struct glBond_t *b[MAXLIGATMNB];
  Atomdef *atp;
  float mass;
  float charge;
  int nb[MAXLIGATMNB];
  int nn;
  int flg,frg,planar,z;
  char name[8];
  char type[8];
} glAtom_t;

typedef struct glBond_t {
  struct glAtom_t *l,*r;
  float d,d0,k,sigma;
  int a[2],rs,bt,M;
} glBond_t;

typedef struct glAngle_t {
  float w,w0,k,sigma;
  int a[3];
} glAngle_t;

typedef struct glDihedral_t {
  float w,k,sigma,p;
  int type;
  int n,b;
  int a[4];
} glDihedral_t;

typedef struct glImproper_t {
  float w,k,sigma,p;
  int type;
  int n;
  int a[4];
} glImproper_t;

typedef struct glPlane_t {
  int *a;
  float *d;
  float rmsd;
  int n;
} glPlane_t;

typedef struct glLigand_t {
  struct glAtom_t *A;
  struct glBond_t *B;
  struct glAngle_t *angle;
  struct glDihedral_t *dihedral;
  struct glImproper_t *improper;
  struct glPlane_t *plane;
  struct glLigandTree_t *T;
  int nat,nbnd,nang,ndih,nimp,nonh,nplan;
  int use_charges,hyd_mode;
  char code[8];
  char segid[4];
} glLigand_t;

typedef struct glLigandTree_t {
  int *nr,         /* nr[i] points to atom in L->A[] */
    *up,           /* up[i] points to parent in tree */
    *fw,           /* fw[i] points to forward atom in tree */
    *new;          /* new[i] translates L->A[i] index into tree index */
  int size;
} glLigandTree_t;

typedef struct glCNXvdwpar_t {
  char type[8];
  float eps,sig,eps14,sig14;
} glCNXvdwpar_t;

int   genligCNX(char *filename,glLigand_t *L);
int   genligDYANA(char *filename,glLigand_t *L);
int   genligREFMAC(char *filename,glLigand_t *L);

#endif

