#include "swdef.h"
#include "mystdlib.h"

void pconnected_atoms(Nodechain *c,int p)
{
  /* add all atoms connected to atoms in c by path of max length p */

  Node *n;
  Atom *a,*b;
  int i,k,m;

  if (c) {
    if (p<0) clear_node_chain(c);
    for (k=0,m=c->n,n=c->s;p>0;p--,m=c->n) {
      for (;k<m;k++,n=n->chain) {
	if (is_atom(n)) {
	  a=(Atom *)n;
	  for (i=0;i<a->nbnd;i++) {
	    b=(a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f);
	    if (!mvld(b->chain)&&(Node *)b!=c->e) {
	      c->e->chain=(Node *)b;
	      c->e=(Node *)b;
	      b->chain=(Node *)NULL;
	      c->n++;
	    }
	  }
	}
      }
      if (m==c->n) break;
    }
  }
}

void connected_atoms(Nodechain *c)
{
  /* put atoms connected (directly or indirectly) to atoms in c into
     chain c */

  Node *n;
  Atom *a,*b;
  int i;

  if (c) {
    for (n=c->s;mvld(n);n=n->chain) {
      if (is_atom(n)) {
	a=(Atom *)n;
	for (i=0;i<a->nbnd;i++) {
	  b=(a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f);
	  if (!mvld(b->chain)&&(Node *)b!=c->e) {
	    c->e->chain=(Node *)b;
	    c->e=(Node *)b;
	    b->chain=(Node *)NULL;
	    c->n++;
	  }
	}
      }
    }
  }
}

Nodechain *connected_to_right(Atom *left,Atom *right)
{
  /* return nodechain containing all atoms in right-connected fragment. return
     empty chain if left is in right-connected fragment, NULL on error */

  Nodechain *c;
  Atom *b;
  int i;

  c=(Nodechain *)NULL;
  if (!is_atom(left)||!is_atom(right)) return(c);
  c=make_node_chain();
  if (mvld(c)) {
    if (left!=right) {
      left->chain=(Node *)left;
      add_to_node_chain((Node *)right,c);
      connected_atoms(c);
      left->chain=(Node *)NULL;
      for (i=0;i<left->nbnd;i++) {
	b=(left->bnd[i]->f == left ? left->bnd[i]->t : left->bnd[i]->f);
	if (b!=right) {
	  if (mvld(b->chain)||(Node *)b==c->e) {
	    clear_node_chain(c);
	    break;
	  }
	}
      }
    }
  }
  return(c);
}
/*-------------------- ringsz --------------------*/

#define is_used(a) (a->seqnr >=0 && a->seqnr < K && ratom[a->seqnr] == a)

static int ext_ringsz(int K,Atom **ratom,int *sq,int maxsz)
{
  int i,r,R;
  Atom *a,*b;

  r=maxsz+1;
  if (ratom[K-1]->nbnd>1) {
    for (a=ratom[K-1],i=0;i<a->nbnd;i++) {
      b=(a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f);
      if (b!=ratom[K-2]) {
	if (b==ratom[0]) {
	  r=K;
	  break;
	}else if (K<maxsz&&!is_used(b)) {
	  ratom[K]=b; sq[K]=b->seqnr; b->seqnr=K;
	  R=ext_ringsz(K+1,ratom,sq,r-1); 
	  r=( R<r ? R : r );
	  b->seqnr=sq[K]; ratom[K]=(Atom *)NULL;
	}
      }
    }
  }
  return(r);
}

int ringsz(Bond *bnd,int maxsz)
{
  /* find minimum ring size <= maxsz for bond bnd, 0 on error, maxsz+1 if
     no ring found with size <= maxsz */

  Atom **ratom;
  int i,*sq,r;

  if (maxsz<3) return(0);
  if (bnd->f->nbnd<=1 && bnd->t->nbnd<=1) return(maxsz+1);
  ratom=(Atom **)malloc(maxsz*sizeof(Atom *));
  sq=(int *)malloc(maxsz*sizeof(int));
  if (mvld(ratom)&&mvld(sq)) {
    for (i=2;i<maxsz;i++) ratom[i]=(Atom *)NULL;
    ratom[0]=bnd->f; sq[0]=ratom[0]->seqnr; ratom[0]->seqnr=0;
    ratom[1]=bnd->t; sq[1]=ratom[1]->seqnr; ratom[1]->seqnr=1;
    r=ext_ringsz(2,ratom,sq,maxsz);
    ratom[0]->seqnr=sq[0]; ratom[1]->seqnr=sq[1];
  }else r=0;
  if (mvld(sq)) free(sq);
  if (mvld(ratom)) free(ratom);
  return(r);
}

