#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "graphics.h"

extern int graphix;

static int n_ramp_colors=0;
static int *ramp_colors=(int *)NULL;
static float dot_pmin=-1,dot_pmax=1;

/* Icosahedron data (ref.: OpenGL Programming Guide) */

#define X 0.525731112119133606
#define Z 0.850650808352039932

static float IV[12][3] = {                         /* vertices */
  {X,0,Z}, {-X,0,Z}, {X,0,-Z}, {-X,0,-Z},
  {-0,Z,X}, {0,Z,-X}, {0,-Z,X}, {0,-Z,-X},
  {-Z,X,0}, {Z,X,0}, {-Z,-X,0}, {Z,-X,0}
};

static int IT[20][3] = {                           /* triangles */
  {0,4,1},  {0,9,4},  {9,5,4}, {4,5,8}, {4,8,1},
  {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},
  {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6},
  {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11}
};

static void TrgSubdiv(Gfxobj *G,float *v1,float *v2,float *v3,
		      float *center,float radius,int div,int Color);

/*------------------------------------------------------------------------*/

Gfxobj *OpenGfxObject(Dspobj *global,int buffsize)
{
  Gfxobj *G=(Gfxobj *)NULL;

  if (is_dsp(global)) {
    G=(Gfxobj *)malloc(sizeof(Gfxobj));
    if (mvld(G)) {
      G->Vecbuff=(float *)NULL;
      G->Nvec=0;
      G->Nbuff=0;
      G->Buffsize=buffsize;
      G->Color=-99;
      G->Vecbuff=(float *)malloc(6*buffsize*sizeof(float));
      G->Obj=new_dsp_obj(global,(char *)NULL,DSP_CSO);
      if (!mvld(G->Obj)||!mvld(G->Vecbuff)) G=CloseGfxObject(G);
    }
  }
  return(G);
}

Gfxobj *CloseGfxObject(Gfxobj *G)
{
  if (mvld(G)) {
    FlushGfxBuffer(G);
    if (G->Nvec==0) del_node((Node *)G->Obj);
    if (G->Vecbuff) free(G->Vecbuff);
    free(G);
    G=(Gfxobj *)NULL;
  }
  return(G);
}

void AddMD(Gfxobj *G,float *x,float *y,int Color)
{
  int i; 

  if (!mvld(G)) return;
  if (!mvld(G->Obj)) return;
  if (Color!=G->Color || G->Nbuff>=G->Buffsize) FlushGfxBuffer(G);
  G->Nvec++;
  for (i=0;i<3;i++) {
    G->Vecbuff[6*G->Nbuff+i]=x[i];
    G->Vecbuff[6*G->Nbuff+3+i]=y[i];
  }
  G->Nbuff++;
  G->Color=Color;
}

void AddTMD(Gfxobj *G,mat_4 T,float *x,float *y,int Color)
{
  int i; 

  if (!mvld(G)) return;
  if (!mvld(G->Obj)) return;
  if (Color!=G->Color || G->Nbuff>=G->Buffsize) FlushGfxBuffer(G);
  G->Nvec++;
  for (i=0;i<3;i++) {
    G->Vecbuff[6*G->Nbuff+i]=T.r[i].c[3]+
      T.r[i].c[0]*x[0]+T.r[i].c[1]*x[1]+T.r[i].c[2]*x[2];
    G->Vecbuff[6*G->Nbuff+3+i]=T.r[i].c[3]+
      T.r[i].c[0]*y[0]+T.r[i].c[1]*y[1]+T.r[i].c[2]*y[2];
  }
  G->Nbuff++;
  G->Color=Color;
}

void FlushGfxBuffer(Gfxobj *G)
{
  int size;

  if (!mvld(G)) return;
  if (!mvld(G->Obj)) return;
  if (G->Nbuff>0) {
    size=sizeof(G->Obj)+sizeof(int)+G->Nbuff*6*sizeof(float);
    send_bheader(graphix,"AMDS",size);
    send_bptr(graphix,G->Obj);
    send_bint(graphix,G->Color);
    send_bfloata(graphix,6*G->Nbuff,G->Vecbuff);
    G->Nbuff=0;
  }
}
/*------------------------------------------------------------------------*/

void AddIcoSphere(Gfxobj *G,float *center,float radius,int div,int Color)
{
  /* Draw "sphere" (icosahedron, recursive subdivision of faces) */

  int i;

  /* subdivision of faces */

  for (i=0;i<20;i++) 
    TrgSubdiv(G,IV[IT[i][0]],IV[IT[i][1]],IV[IT[i][2]],center,radius,div,Color);
  return;
}

static void TrgSubdiv(Gfxobj *G,float *v1,float *v2,float *v3,
		      float *center,float radius,int div,int Color)
{
  int j;
  float m1[3],m2[3],m3[3],s1,s2,s3,p1[3],p2[3],p3[3];

  s1=0;s2=0;s3=0;
  for (j=0;j<3;j++) {
    m1[j]=0.5*(v1[j]+v2[j]);s1+=m1[j]*m1[j];
    m2[j]=0.5*(v2[j]+v3[j]);s2+=m2[j]*m2[j];
    m3[j]=0.5*(v3[j]+v1[j]);s3+=m3[j]*m3[j];
  }
  s1=sqrt(s1);s2=sqrt(s2);s3=sqrt(s3);
  for (j=0;j<3;j++) {
    m1[j]/=s1;
    m2[j]/=s2;
    m3[j]/=s3;
  }
  s1=0;s2=0;s3=0;
  for (j=0;j<3;j++) {
    p1[j]=v1[j]; s1+=p1[j]*p1[j];
    p2[j]=v2[j]; s2+=p2[j]*p2[j];
    p3[j]=v3[j]; s3+=p3[j]*p3[j];
  }
  s1=sqrt(s1);s2=sqrt(s2);s3=sqrt(s3);
  for (j=0;j<3;j++) {
    p1[j]/=s1;
    p2[j]/=s2;
    p3[j]/=s3;
  }
  if (div>0) {
    TrgSubdiv(G,m1,m2,m3,center,radius,div-1,Color);
    TrgSubdiv(G,v1,m1,m3,center,radius,div-1,Color);
    TrgSubdiv(G,v2,m2,m1,center,radius,div-1,Color);
    TrgSubdiv(G,v3,m3,m2,center,radius,div-1,Color);
  }else{
    for (j=0;j<3;j++) {
      p1[j]=p1[j]*radius+center[j];
      p2[j]=p2[j]*radius+center[j];
      p3[j]=p3[j]*radius+center[j];
      m1[j]=m1[j]*radius+center[j];
      m2[j]=m2[j]*radius+center[j];
      m3[j]=m3[j]*radius+center[j];
    }
    AddMD(G,p1,m1,Color); AddMD(G,p2,m2,Color); AddMD(G,p3,m3,Color);
    AddMD(G,m1,m2,Color), AddMD(G,m2,m3,Color); AddMD(G,m3,m1,Color);
  }
}
/*------------------------------------------------------------------------*/

void DefineColorRamp(int ncol,int *cols)
{
  int *new,i;

  if (ncol>0) {
    new=(int *)malloc(ncol*sizeof(int));
    if (mvld(new)) {
      DefineColorRamp(0,(int *)NULL);
      ramp_colors=new;
      n_ramp_colors=ncol;
      for (i=0;i<ncol;i++) ramp_colors[i]=cols[i];
    }
  }else{
    if (mvld(ramp_colors)) free(ramp_colors);
    n_ramp_colors=0;
    ramp_colors=(int *)NULL;
  }
  setPropertyMap(dot_pmin,dot_pmax);
}

int RampColor(float val,float mn,float mx)
{
  int r;
  float x;

  if (n_ramp_colors<1||mn>=mx) {
    r=7;
  }else{
    x=(val-mn)/(mn-mx);
    r=(n_ramp_colors-2)*x+1;
    r=(r<0 ? 0 : (r>n_ramp_colors-1 ? n_ramp_colors-1 : r));
    r=ramp_colors[r];
  }
  return(r);
}

int *ColorRamp(void)
{
  return(ramp_colors);
}

int NRampColors(void)
{
  return(n_ramp_colors);
}

void setPropertyMap(float min,float max)
{
  float dp;
  int i,n,*ramp;

  n=NRampColors();
  ramp=ColorRamp();
  dot_pmin=min;
  dot_pmax=max;
  if (n>1) {
    dp=(n>2 ? (dot_pmax-dot_pmin)/(n-2) : 0);
    send_bheader(graphix,"DOTM",(2*n-1)*sizeof(int));
    send_bint(graphix,ramp[0]);
    for (i=1;i<n;i++) {
      send_bfloat(graphix,dot_pmin+(i-1)*dp);
      send_bint(graphix,ramp[i]);
    }
  }
}

void getPropertyMap(float *min,float *max)
{
  *min=dot_pmin;
  *max=dot_pmax;
}
