#include <stdio.h>
#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "transpor.h"
#include "graphics.h"
#include "bhtree.h"
#include "xbhtree.h"
#include "errhand.h"

extern int graphix;
extern int server;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define MAXTREE 1024
#define MAXCLS  1024
#define MAXCLSV 50   /* must not exceed MAXCLS */

static float *findWitness(xBHnode *node,float *x,float mxr,float cutoff,float *r);

/*-------------------- grid_value --------------------*/

float grid_value(Grid *g,vec_3 pt)
{
  /* interpolated grid value at point pt. Uses trilinear interpolation */
  /* point pt is a vec_3 vector */

  float r,x[3],x0[3],c[3],V[8],*v;
  int ix[3],k;

  r=0;

  if (mvld(g)) {
    if (mvld(g->val)) {
      for (k=0;k<3;k++) {
	x[k]=clamp(pt.c[k],g->xmin[k],g->xmin[k]+(g->n[k]-1.001)*g->dx[k]);
	ix[k]=clamp((x[k]-g->xmin[k])/g->dx[k],0,g->n[k]-2);
	x0[k]=g->xmin[k]+ix[k]*g->dx[k];
	c[k]=(x[k]-x0[k])/g->dx[k];
      }
      v=g->val+ix[2]*(g->n[0]*g->n[1])+ix[1]*g->n[0]+ix[0];
      V[0]=v[0];
      V[1]=v[1]-V[0];
      V[2]=v[g->n[0]]-V[0];
      V[4]=v[g->n[0]*g->n[1]]-V[0];
      V[3]=v[g->n[0]+1]-V[0]-V[1]-V[2];
      V[5]=v[g->n[0]*g->n[1]+1]-V[0]-V[1]-V[4];
      V[6]=v[g->n[0]+g->n[0]*g->n[1]]-V[0]-V[2]-V[4];
      V[7]=v[g->n[0]+g->n[0]*g->n[1]+1]-V[0]-V[1]-V[2]-V[3]-V[4]-V[5]-V[6];
      r=V[0]+
	c[0]*(c[1]*(V[7]*c[2]+V[3])+c[2]*V[5]+V[1])+
	c[1]*(c[2]*V[6]+V[2])+
	c[2]*V[4];
    }
  }
  return(r);
}
/*-------------------- grid_valuex --------------------*/

float grid_valuex(Grid *g,float *pt)
{
  /* interpolated grid value at point pt. Uses trilinear interpolation */
  /* point pt is an array */

  float r,x[3],x0[3],c[3],V[8],*v;
  int ix[3],k;

  r=0;

  if (mvld(g)) {
    if (mvld(g->val)) {
      for (k=0;k<3;k++) {
	x[k]=clamp(pt[k],g->xmin[k],g->xmin[k]+(g->n[k]-1.001)*g->dx[k]);
	ix[k]=clamp((x[k]-g->xmin[k])/g->dx[k],0,g->n[k]-2);
	x0[k]=g->xmin[k]+ix[k]*g->dx[k];
	c[k]=(x[k]-x0[k])/g->dx[k];
      }
      v=g->val+ix[2]*(g->n[0]*g->n[1])+ix[1]*g->n[0]+ix[0];
      V[0]=v[0];
      V[1]=v[1]-V[0];
      V[2]=v[g->n[0]]-V[0];
      V[4]=v[g->n[0]*g->n[1]]-V[0];
      V[3]=v[g->n[0]+1]-V[0]-V[1]-V[2];
      V[5]=v[g->n[0]*g->n[1]+1]-V[0]-V[1]-V[4];
      V[6]=v[g->n[0]+g->n[0]*g->n[1]]-V[0]-V[2]-V[4];
      V[7]=v[g->n[0]+g->n[0]*g->n[1]+1]-V[0]-V[1]-V[2]-V[3]-V[4]-V[5]-V[6];
      r=V[0]+
	c[0]*(c[1]*(V[7]*c[2]+V[3])+c[2]*V[5]+V[1])+
	c[1]*(c[2]*V[6]+V[2])+
	c[2]*V[4];
    }
  }
  return(r);
}
/*-------------------- grid_grad --------------------*/

vec_3 grid_grad(Grid *g,vec_3 pt)
{
  /* gradient of grid at point pt */

  vec_3 r;
  float d;
  int k;

  for (k=0;k<3;k++) r.c[k]=0.0;
  if (mvld(g) && mvld(g->val)) {
    for (k=0;k<3;k++) {
      d=0.005*g->dx[k];
      pt.c[k]+=d; r.c[k]=grid_value(g,pt);
      pt.c[k]-=2*d; r.c[k]=(r.c[k]-grid_value(g,pt))/(2*d);
      pt.c[k]+=d;
    }
  }
  return(r);
}
/*-------------------- grid_gradx --------------------*/

vec_3 grid_gradx(Grid *g,float* pt)
{
  /* gradient of grid at point pt */

  vec_3 r;
  float d;
  int k;

  for (k=0;k<3;k++) r.c[k]=0.0;
  if (mvld(g) && mvld(g->val)) {
    for (k=0;k<3;k++) {
      d=0.005*g->dx[k];
      pt[k]+=d; r.c[k]=grid_valuex(g,pt);
      pt[k]-=2*d; r.c[k]=(r.c[k]-grid_valuex(g,pt))/(2*d);
      pt[k]+=d;
    }
  }
  return(r);
}
/*-------------------- free_grid --------------------*/

Grid *freeGrid(Grid *g)
{
  if (mvld(g)) {
    g->Box=grid_box_delete(g);
    if (mvld(g->val)) free(g->val);
    if (mvld(g->name)) free(g->name);
    free(g);
  }
  return(Grid *)NULL;
}
/*-------------------- grid_box_make --------------------*/

void grid_box_make(Grid *G)
{
  Gfxobj *B;
  float x[3],y[3];
  int i,j;

  if (mvld(G)) {
    grid_box_delete(G);
    if (G->boxon) {
      B=OpenGfxObject(DspRoot,1024);
      if (mvld(B)) {
	G->Box=B->Obj;
	if (cPtr(G->Box->cross)) {
	  send_bheader(graphix,"SOLW",sizeof(int)+sizeof(G->Box));
	  send_bint(graphix,G->boxwidth);
	  send_bptr(graphix,G->Box);
	}
	for (i=0;i<3;i++) {         /* draw rectangular box... */
	  for (j=0;j<4;j++) {
	    x[i]=G->xmin[i];  y[i]=G->xmax[i];
	    x[(i+1)%3]=(j%2 ? G->xmin[(i+1)%3] : G->xmax[(i+1)%3]);
	    y[(i+1)%3]=x[(i+1)%3];
	    x[(i+2)%3]=(j<2 ? G->xmin[(i+2)%3] : G->xmax[(i+2)%3]);
	    y[(i+2)%3]=x[(i+2)%3];
	    AddMD(B,x,y,G->boxcolor);
	  }
	}
	set_node_name((Node *)G->Box,".GridBox");
	CloseGfxObject(B);
      }
    }
  }
}
/*-------------------- grid_box_delete --------------------*/

Dspobj *grid_box_delete(Grid *G)
{
  Dspobj *r=(Dspobj *)NULL;;

  if (G) {
    if (is_dsp(G->Box)) r=(Dspobj *)del_node((Node *)G->Box);
    G->Box=r;
  }
  return(r);
}
/*-------------------- emptyGrid --------------------*/

Grid *emptyGrid(void)
{
  Grid *g;
  int i;

  g=(Grid *)malloc(sizeof(Grid));
  if (mvld(g)) {
    g->val=(float *)NULL;
    g->name=(char *)NULL;
    for (i=0;i<3;i++) {
      g->xmin[i]=0;
      g->xmax[i]=0;
      g->dx[i]=1;
      g->cl[i]=1.0;
      g->ca[i]=90.0;
      g->n[i]=0;
      g->xl[i]=-1.e30;
      g->xh[i]=+1.e30;
    }
    g->Box=(Dspobj *)NULL;
    g->boxon=FALSE;
    g->boxcolor=3;
    g->boxwidth=1;
  }
  return(g);
}
/*-------------------- write_grid --------------------*/

int write_grid(FILE *f,Grid *g)
{
  int xmin[3],dx[3],nx[3],cl[3],ca[3],val,status,i,n,m;
  float *v;

  /* slight complication, because file should use SGI float and int... */

  status=0;
  if (f&&mvld(g)) {
    for(i=0;i<3;i++) {
      xmin[i]=SGI_f(g->xmin[i]);
      dx[i]=SGI_f(g->dx[i]);
      cl[i]=SGI_f(g->cl[i]);
      ca[i]=SGI_f(g->ca[i]);
      nx[i]=SGI_i(g->n[i]);
    }
    n=0;
    n+=fwrite(xmin,sizeof(int),3,f);
    n+=fwrite(dx,sizeof(int),3,f);
    n+=fwrite(nx,sizeof(int),3,f);
    if (mvld(g->val)) m=g->n[0]*g->n[1]*g->n[2];
    else m=0;
    for (v=g->val,i=0;i<m;i++,v++) {
      val=SGI_f(*v);
      n+=fwrite(&val,sizeof(float),1,f);
    }
    n+=fwrite(cl,sizeof(int),3,f);
    n+=fwrite(ca,sizeof(int),3,f);
    status=(n-m==15 ? 0 : 2);
  }else status=1;
  return(status);
}
/*-------------------- grid_dump --------------------*/

void grid_dump(Grid *g)
{
  float *v,val;
  int i,j,k;
  FILE *out;

  /* temprorary kludge to look at grids... */

  out=fopen("wnp_grid.dump","w");
  if (out&&g) {
    fprintf(out,"axis: XYZ\n");
    fprintf(out,"origin:     %10.4f %10.4f %10.4f\n",g->xmin[0],g->xmin[1],g->xmin[2]);
    fprintf(out,"dx,dy,dz:   %10.4f %10.4f %10.4f\n",g->dx[0],g->dx[1],g->dx[2]);
    fprintf(out,"nx,ny,nz:   %10d %10d %10d\n",g->n[0],g->n[1],g->n[2]);
    fprintf(out,"a,b,c:      %10.4f %10.4f %10.4f\n",g->cl[0],g->cl[1],g->cl[2]);
    fprintf(out,"angles:     %10.4f %10.4f %10.4f\n",g->ca[0],g->ca[1],g->ca[2]);
    for (v=g->val,i=0;i<g->n[2];i++) {
      for (j=0;j<g->n[1];j++) {
    	for (k=0;k<g->n[0];k++) {
	  val=*v++; 
	  if (val>999.999) val=999.999;
	  if (val<-999.999) val=-999.999;
	  fprintf(out,"%8.3f",val);
	  if (k%10==9) fprintf(out,"\n");
	}
	if (k%10) fprintf(out,"\n"); 
      }
    }
    fclose(out);
  }
}

/*-------------------- grid_fn_vdwdist --------------------*/

int grid_fn_vdwdist(Grid *g,Nodechain *c)
{
  int i,j,k,l,n,status;
  float *v,x[3],dist[MAXCLS],D,d,ds,rmin,rmax;
  Atom *cls[MAXCLS],*a;
  BHtree *bht;

  /* compute distance from vdw surface. Points inside have pos. distance. */

  status=1;
  if (!(mvld(g))||!mvld(c)) return(status);
  bht=generateBHtree(c,( c->n < MAXTREE ? 1 : c->n / MAXTREE ));
  if (mvld(bht)) {
    for (rmax=0,rmin=100000,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      d=(a->atp ? a->atp->rvdw : 1.5);
      rmin=(d<rmin?d:rmin);
      rmax=(d>rmax?d:rmax);
    }
    D=sqrt((g->xmax[0]-g->xmin[0])*(g->xmax[0]-g->xmin[0])+
	   (g->xmax[1]-g->xmin[1])*(g->xmax[1]-g->xmin[1])+
	   (g->xmax[2]-g->xmin[2])*(g->xmax[2]-g->xmin[2]));
    n=g->n[0]*g->n[1]*g->n[2];
    for (i=0,v=g->val;i<n;i++,v++) *v=-D-1;
    for (x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
      dotdotdot("vdw",k,g->n[2],60);
      for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
	v=g->val+g->n[0]*(g->n[1]*k+j);
	d=D;  	                 /*  initial cutoff ...*/
	for (x[0]=g->xmin[0],i=0;i<g->n[0];i++,x[0]+=g->dx[0],v++) {
	  n=findBHcloseAtoms(bht,x,d,cls,dist,MAXCLSV);
	  if (n>MAXCLSV||n==0) { /* overflow, have to reduce cutoff... */
	    a=findBHclosest(bht,x,(n==0 ? D : d),&ds);
	    if (mvld(a)) {
	      d=(a->atp ? ds+rmax-a->atp->rvdw+0.01 : ds+rmax-1.5); /* 0.01: tol. needed if rmax==a->atp->rvdw */
	      n=findBHcloseAtoms(bht,x,d,cls,dist,MAXCLS);
	      if (n>MAXCLS) n=MAXCLS;
	    }else n=0;
	  }
	  if (n>0) {
	    for (l=0;l<n;l++) {
	      ds=dist[l]-(cls[l]->atp ? cls[l]->atp->rvdw : 1.5);
	      if (-ds>*v) *v=-ds;
	    }
	    d=-*v+rmax+g->dx[0];
	  }else d=D;
	}
      }
    }
    dotdotdot("",0,0,0);
    freeBHtree(bht);
    status=0;
  }else err_set(1,"failed to generate BH-tree.");
  return(status);
}
/*-------------------- grid_fn_Coulomb --------------------*/

int grid_fn_Coulomb(Grid *g,Nodechain *c,int dielt,float eps)
{
  /* Coulomb potential, no cutoff */

  int i,j,k;
  float *v,x[3],d;
  Atom *a;

  eps=332.0716/eps;
  for (x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
    dotdotdot("Coulomb",k,g->n[2],60);
    for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
      v=g->val+g->n[0]*(g->n[1]*k+j);
      for (x[0]=g->xmin[0],i=0;i<g->n[0];i++,x[0]+=g->dx[0],v++) {
	for (*v=0,a=(Atom *)c->s;is_atom(a);a=(Atom *)a->chain) {
	  if (a->x&&a->nxyz>0&&fabs(a->q)>1.0e-9) {
	    d=(a->x[0]-x[0])*(a->x[0]-x[0])
	      +(a->x[1]-x[1])*(a->x[1]-x[1])
	      +(a->x[2]-x[2])*(a->x[2]-x[2]);
	    if (dielt==0) {
	      *v+=eps*a->q/sqrt(d);
	    }else{
	      *v+=eps*a->q/d;
	    }
	  }
	}
      }
    }
  }
  return(0);
}

/*-------------------- grid_fn_SaSDist --------------------*/

int grid_fn_SaSDist(Grid *g,Nodechain *c,float r)
{
  /* OBSOLETE (use grid_fn_ConnollyDist instead) */
  /* compute distance from Connolly surface generated by atoms in c */
  /* probe radius is r. */

  int status,i,j,k,n,np,quick;
  float *v,**p,x[3],*X,T,d,D,*cp,mxr;
  xBHtree *bh;

  /* compute distance from Connolly surface. Points inside have pos. distance. */

  status=1;

  if (!mvld(g)||!mvld(g->val)) return(status);
  if (!mvld(c)||c->n<1) return(status);
  if (r<0) r=0;
  p=(float **)NULL;
  X=(float *)NULL;

  status=grid_fn_vdwdist(g,c);       /* do distance to vdW surface first... */
  if (!status) {

    /* count points close to LR(r) surface ... */

    n=g->n[0]*g->n[1]*g->n[2];
    mxr=r+g->dx[0];
    for (np=0,i=0,v=g->val;i<n;i++,v++) if (-*v>=r&&-*v<=mxr) np++;
    prtGraphics("%d np points close to LR surface...\n",np);

    if (np>0) {
      p=(float **)malloc(np*sizeof(float *));
      X=(float *)malloc(4*np*sizeof(float));
    }
     
    if (p&&X) {

      /* collect the points ... */

      for (np=0,x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
	for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
	  v=g->val+g->n[0]*(g->n[1]*k+j);
	  for (i=0;i<g->n[0];i++,v++) {
	    if (-*v>=r&&-*v<=mxr) {
	      p[np]=X+np*4;
	      p[np][0]=g->xmin[0]+i*g->dx[0];
	      p[np][1]=x[1];
	      p[np][2]=x[2];
	      p[np][3]=-*v;
	      np++;
	    }
	  }
	}
      }
      
      /* and generate xBH tree */

      bh=GeneratexBHtree(np,p,(np < MAXTREE ? 1 : np / MAXTREE ));
      if (bh) {
	
	/* compute distance from SaS for points within LR(r) */

	D=sqrt((g->xmax[0]-g->xmin[0])*(g->xmax[0]-g->xmin[0])+
	       (g->xmax[1]-g->xmin[1])*(g->xmax[1]-g->xmin[1])+
	       (g->xmax[2]-g->xmin[2])*(g->xmax[2]-g->xmin[2]));
	T=sqrt(3)*g->dx[0];
	for (x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
	  dotdotdot("sas",k,g->n[2],60);
	  for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
	    v=g->val+g->n[0]*(g->n[1]*k+j);
	    for (quick=FALSE,d=D,x[0]=g->xmin[0],i=0;i<g->n[0];i++,v++,x[0]+=g->dx[0]) {
	      if (-*v<r) {
		*v=D;
		if (quick) {
		  cp=findxBHclosest(bh,x,d,&d);
		  if (cp) {
		    *v=d-cp[3];
		    d=*v+mxr;
		  }
		}else{
		  cp=findWitness(bh->root,x,mxr,d,v);
		  if (!cp) cp=findWitness(bh->root,x,mxr,D,v);
		  if (cp) {
		    d=*v+mxr;
		  }else{
		    *v=D;
		    d=D;
		  }
		}
	      }
	      d+=T;
	      quick=(d>3*r&&d>6*g->dx[0]);
	    }
	  }
	}
	dotdotdot("",0,0,0);
	freexBHtree(bh);
      }else err_set(1,"failed to generate xBH tree");
      
    }else if (np>0) err_set(1,"memory allocation failed...");
  }

  if (mvld(p)) free(p);
  if (mvld(X)) free(X);
  return(status);
}
/*-------------------- grid_fn_EDFDist --------------------*/

int grid_fn_EDFDist(Grid *g,Nodechain *c,float d)
{
  int i,j,k,l,n,status;
  float *v,x[3],dist[MAXCLS],ds,rmax,cutoff=3;
  Atom *cls[MAXCLS],*a;
  BHtree *bht;

  /* compute distance from exponential density function surface. Points inside have pos. distance. */

  status=1;
  if (!(mvld(g))||!mvld(c)) return(status);
  bht=generateBHtree(c,( c->n < MAXTREE ? 1 : c->n / MAXTREE ));
  if (mvld(bht)) {
    for (rmax=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      ds=(a->atp ? a->atp->rvdw : 1.5);
      rmax=(ds>rmax?d:rmax);
    }
    rmax=rmax+cutoff;

    for (x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
      dotdotdot("EDFdist",k,g->n[2],60);
      for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
	v=g->val+g->n[0]*(g->n[1]*k+j);
	for (x[0]=g->xmin[0],i=0;i<g->n[0];i++,x[0]+=g->dx[0],v++) {
	  *v=exp(-10/d);
	  n=findBHcloseAtoms(bht,x,rmax,cls,dist,MAXCLS);
	  if (n>MAXCLS) n--;
	  for (l=0;l<n;l++) {
	    ds=dist[l]-(cls[l]->atp ? cls[l]->atp->rvdw : 1.5);
	    if (ds<cutoff) *v+=exp(-ds/d);
	  }
	  *v=d*log(*v);
	}
      }
    }
    dotdotdot("",0,0,0);
    freeBHtree(bht);
    status=0;
  }else err_set(1,"failed to generate BH-tree.");
  return(status);
}

/*-------------------- grid_fn_dotdist --------------------*/

int grid_fn_dotdist(Grid *g,int ndot,float **dots)
{
  int i,j,k,n,status;
  float *v,x[3],D,d;
  float *dot;
  xBHtree *bht;

  /* compute distance from MS dots */

  status=1;
  if (!(mvld(g))||!mvld(dots)) return(status);
  bht=GeneratexBHtree(ndot,dots,( ndot < MAXTREE ? 1 : ndot / MAXTREE ));
  if (mvld(bht)) {
    D=sqrt((g->xmax[0]-g->xmin[0])*(g->xmax[0]-g->xmin[0])+
	   (g->xmax[1]-g->xmin[1])*(g->xmax[1]-g->xmin[1])+
	   (g->xmax[2]-g->xmin[2])*(g->xmax[2]-g->xmin[2]));
    n=g->n[0]*g->n[1]*g->n[2];
    for (i=0,v=g->val;i<n;i++,v++) *v=-D-1;

    for (x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
      dotdotdot("dotdist",k,g->n[2],60);
      for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
	v=g->val+g->n[0]*(g->n[1]*k+j);
	for (x[0]=g->xmin[0],i=0;i<g->n[0];i++,x[0]+=g->dx[0],v++) {
	  dot=findxBHclosest(bht,x,D,&d);
	  if (dot) {
	    *v=(dot[0]-x[0])*dot[3]+(dot[1]-x[1])*dot[4]+(dot[2]-x[2])*dot[5];
	  }
	}
      }
    }
    dotdotdot("",0,0,0);
    freexBHtree(bht);
    status=0;
  }else err_set(1,"failed to generate BH-tree.");
  return(status);
}

/*-------------------- grid_fn_Contact --------------------*/

int grid_fn_Contact(Grid *g,Nodechain *c,float r)
{
  int i,j,k,status;
  float *v,x[3],dist[MAXCLS];
  Atom *cls[MAXCLS];
  BHtree *bht;

  /* count number of atoms within cutoff r */

  status=1;
  if (!(mvld(g))||!mvld(c)) return(status);
  bht=generateBHtree(c,( c->n < MAXTREE ? 1 : c->n / MAXTREE ));
  if (mvld(bht)) {
    for (x[2]=g->xmin[2],k=0;k<g->n[2];k++,x[2]+=g->dx[2]) {
      prtGraphics("contact plane %d\n",k);
      for (x[1]=g->xmin[1],j=0;j<g->n[1];j++,x[1]+=g->dx[1]) {
	v=g->val+g->n[0]*(g->n[1]*k+j);
	for (x[0]=g->xmin[0],i=0;i<g->n[0];i++,x[0]+=g->dx[0],v++)
	  *v=findBHcloseAtoms(bht,x,r,cls,dist,MAXCLS);
      }
    }
    freeBHtree(bht);
    status=0;
  }else err_set(1,"failed to generate BH-tree.");
  return(status);
}

/*-------------------- grid_fn_truncate --------------------*/

int grid_fn_truncate(Grid *g,float b,float a)
{
  float *v;
  int i,n,r=0;

  if (mvld(g)&&mvld(g->val)) {
    r=1;
    n=g->n[0]*g->n[1]*g->n[2];
    for (v=g->val,i=0;i<n;i++,v++) {
      if (*v >= -1.e30 && *v <= 1.e30) {
	if (*v<b) *v=b;
	if (*v>a) *v=a;
      }
    }
  }
  return(r);
}
/*-------------------- grid_fn_zero --------------------*/

int grid_fn_zero(Grid *g)
{
  float *v;
  int i,n,r=0;

  if (mvld(g)&&mvld(g->val)) {
    r=1;
    n=g->n[0]*g->n[1]*g->n[2];
    for (v=g->val,i=0;i<n;i++) *v++=0.0;
  }
  return(r);
}
/*-------------------- grid_fn_mult --------------------*/
int grid_fn_mult(Grid *g,float f)
{
  float *v;
  int i,n,r=0;

  if (mvld(g)&&mvld(g->val)) {
    r=1;
    n=g->n[0]*g->n[1]*g->n[2];
    for (v=g->val,i=0;i<n;i++) *v++*=f;
  }
  return(r);
}
/*========================================================================*/

/*-------------------- findWitness --------------------*/

static float *findWitness(xBHnode *node,float *x,float mxr,float cutoff,float *r)
{
  int j;
  float d[3],v;
  float *a,*b;

  a=(float *)NULL;
  if (node->dim<0) {
    for (j=0;j<node->n;j++) {
      d[0]=x[0]-node->atom[j][0];
      d[1]=x[1]-node->atom[j][1];
      d[2]=x[2]-node->atom[j][2];
      v=sqrt(d[0]*d[0]+d[1]*d[1]+d[2]*d[2]);
      if (v>cutoff) continue;
      if (v-node->atom[j][3]<*r) {
	*r=v-node->atom[j][3];
	cutoff=mxr+*r;
	a=node->atom[j];
      }
    }
  }else{
    if (x[node->dim]<node->cut) {
      a=findWitness(node->left,x,mxr,cutoff,r);
      if (a) cutoff= *r+mxr;
      if (x[node->dim]>=node->cut-cutoff) {
	b=findWitness(node->right,x,mxr,cutoff,r);
	if (b) a=b;
      }
    }else{
      a=findWitness(node->right,x,mxr,cutoff,r);
      if (a) cutoff= *r+mxr;
      if (x[node->dim]<node->cut+cutoff) {
	b=findWitness(node->left,x,mxr,cutoff,r);
	if (b) a=b;
      }
    }
  }
  return(a);
}
