#ifdef mingw
#include <winsock.h>
#endif
#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "parser.h"
#include "errhand.h"
#if defined(sgi) || defined (linux) || defined (darwin) || defined(altix)
#include <unistd.h>
#endif
#if defined(mingw)
#include <io.h>
#endif
#include "transpor.h"

#define TRUE 1
#define FALSE 0

#ifdef USESOCKETS
#include "utlsock.h"
#endif

extern int SGI;
extern Stream *IOS; /* yuck */
extern Mol *MolRoot,*DefaultNode;
extern Dspobj *DspRoot;
extern int server;
extern char *srvString;
extern float srvVersion;
extern int graphix;
extern Node *LibRoot;

extern int autocenter_on_read;
extern int autobond_on_read;
static void init_remote_env(void);

static int background=FALSE;

int init_all(int argc,char **argv)
{
  /* initialize everything, depending on arguments on command line  */
  
  int i,ntry,ptrsize;
  char host[121],*service_name,*host_name,*ini_file;
  char msg[256],*p,*port=NULL;
  Qmsg *qm;
  
  init_defaults();
  
  server=0; graphix=0;
  host_name="localhost";
  service_name="port_19945";
  ini_file=(char *)NULL;
  /*prj=getThreeDPrj();*/
  
  /* parse command line ... */
  
  for (i=1;i<argc;i++) {
    if (strncmp(argv[i],"-rpc",4)==0) {
      if (strlen(argv[i])>4) port=argv[i]+4;
      else if (i<argc-1) port=argv[++i];
    }else if (strcmp(argv[i],"-B")==0) {
      background=TRUE;
    }else if (strncmp(argv[i],"-I",2)==0) {
      if (strlen(argv[i])>2) ini_file=argv[i]+2;
      else ini_file=argv[++i];
    }else if (strncmp(argv[i],"-s",2)==0) {
      if (strlen(argv[i])>2) service_name=argv[i]+2;
      else service_name=argv[++i];
    }else if (strncmp(argv[i],"-h",2)==0) {
      host_name=argv[i]+2;
      if (host_name[0]=='=') host_name++;
      if (strlen(host_name)==0) {
	i++;
	if (i<argc) {
	  host_name=argv[i];
	}else{
	  fprintf(stderr,"Illegal -h option\n");
	  return(1);
	}
      }
    }
  }

#ifdef USESOCKETS
  if (strcmp(host_name,"localhost")==0) {
    gethostname(host,120);
  }else{
    strcpy(host,host_name);
  }
  if (strcmp(host,"none")!=0) {
    for (ntry=0;ntry<50 && !server;ntry++) {
      server=conn_to_service(host,service_name);
      if (!server) {
	if (ntry%5==0) 
	  fprintf(stderr,
		  "Waiting to make connection to server %s on %s...\n",
		  service_name,host);
	fflush(stderr);
	uSleep(2000000);
      }
    }
    if (!server) {
      fprintf(stderr,
	      "Could not establish connection to server %s on %s...\n",
	      service_name,host);
    }else{
      graphix=open_transport(server,8192,8192);
      for (p=msg;p-msg < 256;p++) {
	read_data_raw(graphix,1,p);
	if (*p == '\n') break;
      }
      if (p>msg) p-=1;
      *p='\0';
      fprintf(stderr,"%s\n",msg+1);
      if (strncmp(msg,"+WITNOTP:",9)) {
	close_transport(graphix);
	close_service(server);
	server=0;graphix=0;
	srvString=strdup("no server");
	fprintf(stderr,"Invalid response from server.\n");
      }else{
	srvString=strdup(msg+9);
      }
    }
  }
#endif
  initmsgq();
  if (!init_parser()) {
    if (graphix) {
      send_header(graphix,"EXIT",0);
      close_transport(graphix);
      graphix=0;
    }
    return(4);
  }
  if (graphix) {
    if (srvVersion>=2.02) {
      send_bheader(graphix,"PTSZ",sizeof(int));
      send_int(graphix,sizeof(void *));
      qm=waitqmsg("PTSZ");
      if (qm&&qm->size==sizeof(int)) {
	readqmsg(qm,(char *)&ptrsize);
	ptrsize=MY_i(ptrsize);
	setcPtrSize(ptrsize);
      }else {
	if (qm) send_header(graphix,"EXIT",0);
	close_transport(graphix);
	graphix=0;
	if (qm) fprintf(stderr,"*** FATAL ERROR server architecture mismatch.\n");
	else fprintf(stderr,"Server closed connection\n");
	return(2);
      }
    }else{
      if (sizeof(void *)!=cPtrSize()) {
	send_header(graphix,"EXIT",0);
	close_transport(graphix);
	graphix=0;
	fprintf(stderr,"*** FATAL ERROR server version mismatch.\n");
	fprintf(stderr,"    Upgrade server to Version >= 2.02\n");
	return(2);
      }
    }
  }else use_readline(TRUE);
  if (!init_root()) {
    if (graphix) {
      send_header(graphix,"EXIT",0);
      close_transport(graphix);
      graphix=0;
    }
    return(5);
  }
  
  pick_mode("EUSA");                             /* turn it on... */

  init_remote_env();
  init_mousemap();
  /*  setThreeDPrj(&prj); */
  initThreeDPrj();
  
  set_sph_mode(get_I_parm("sph_mode"),get_I_parm("sph_poly"));
  set_cyl_mode(get_I_parm("cyl_mode"),get_I_parm("cyl_poly"));

  fprintf(stderr,"\n%s\n%s\n%s\n%s\n%s\n\n",
	  "Copyright (C) 2000 Novartis Pharma AG, All Rights Reserved.",
	  "This program contains confidential and proprietary information of",
	  "Novartis Pharma AG, furnished under license agreement. This program,",
          "including supporting material and manuals, may be used or copied",
          "only in accordance with the terms of the agreement."
	  );

  Setenv("WNP_RPC_PORT","");

  if (port) {
    if (wnpHaveRPC()) {
      wnpRPCstart(atoi(port));    /* should handle more complicated syntax (i.e. -rpc 1900-2000) */
      if (!wnpRPCport()) {
	if (graphix) {
	  send_header(graphix,"EXIT",0);
	  close_transport(graphix);
	  graphix=0;
	}
	err_set(1,"Session aborted.");
	err_dump();
	return(3);
      }
    }else{
      if (graphix) {
	send_header(graphix,"EXIT",0);
	close_transport(graphix);
	graphix=0;
      }
      err_set(1,"RPC is not supported in this version of Wit!P. Session aborted.");
      err_dump();
      return(3);
    }
  }

  if (ini_file) {  /* multiple -I files: push in reverse order... (NYI) */
    if (access(ini_file,R_OK)) err_set(1,"Can't open %s",ini_file);
    else push_stream(IOS,ini_file);
    err_dump();
  }

  return(0);
}

void init_defaults(void)
{
  PrjTS prj;
  int colramp[]={1,10,3,11,2,12,6,13,4};
  
  setcPtrSize(4);                /* assumed size of server's pointers    */

  LibRoot=newLib("MainLibrary");
  prj.viewmode=PERSPECTIVE;
  prj.stereomode=STEREO_NO;
  prj.ysize=15;
  prj.znear=10;
  prj.slabwidth=20;
  prj.viewdist=2.0;
  prj.eyesep=0.2;
  prj.aspect=1.3;
  prj.scrysize=15;
  prj.sep=0.0;
  setThreeDPrj(&prj);

  autocenter_on_read=FALSE;
  autobond_on_read=FALSE;

  DefineColorRamp(9,colramp);
}

/*  Allocate and initialize  root of dynamic data structures   */

int init_root(void)
{   
  MolRoot=(Mol *)make_node(OBJ_MOL);
  MolRoot->moltype=MOL_ROOT;
  DefaultNode=MolRoot;
  DspRoot = (Dspobj *)make_node(OBJ_DSP);
  if (DspRoot) {
    DspRoot->dsptype=DSP_GLBL;
    MolRoot->dsptr=DspRoot;
    DspRoot->root=(Node *)MolRoot;
  }
  new_dsp_obj(DspRoot,"DisplayRoot",DSP_ROOT); /* make it known to graphics */
  pick_root(DspRoot);
  if (DspRoot && MolRoot) return(1);
  else return(0);
}

/* are we running in background mode? */

int inBackground(void)
{
  return(background);
}

/*  Initialize environment of server process */

static void init_remote_env(void)
{

  char buff[512],*value;

  if ((value=getenv("PLOTserver"))) {
    sprintf(buff,"%s=%s","PLOTserver",value);
    send_bmessage(graphix,"SENV",strlen(buff),buff);
  }

  if ((value=getenv("ORTEPserver"))) {
    sprintf(buff,"%s=%s","ORTEPserver",value);
    send_bmessage(graphix,"SENV",strlen(buff),buff);
  }

  if ((value=getenv("PLUTOserver"))) {
    sprintf(buff,"%s=%s","PLUTOserver",value);
    send_bmessage(graphix,"SENV",strlen(buff),buff);
  }

  if ((value=getenv("HP7500plot"))) {
    sprintf(buff,"%s=%s","HP7500plot",value);
    send_bmessage(graphix,"SENV",strlen(buff),buff);
  }

  if ((value=getenv("EPSplot"))) {
    sprintf(buff,"%s=%s","EPSplot",value);
    send_bmessage(graphix,"SENV",strlen(buff),buff);
  }

  if ((value=getenv("WNP_FONTPATH"))) {
    sprintf(buff,"%s=%s","WNP_FONTPATH",value);
    send_bmessage(graphix,"SENV",strlen(buff),buff);
  }
}
