#include <stdio.h>
#include <math.h>
#include <string.h>
#include "mystdlib.h"
#include "swdef.h"
#include "transpor.h"
#include "gcalc.h"
#include "isosurface.h"
#include "cubetrig.h"
#include "errhand.h"
#include "xtal.h"

extern int graphix;

#define TRUE  1
#define FALSE 0

#define ISOS_LEVFORW 1
#define ISOS_LEVBACK 2

static float C[8][3] =        /* corners of the cube                            */
{
  {0,0,0}, {1,0,0}, {1,1,0}, {0,1,0},
  {0,0,1}, {1,0,1}, {1,1,1}, {0,1,1}
};

static int E[12][2] =         /* edges of the cube (pair of vertices)           */
{
  {0,1},{1,2},{2,3},{3,0},    /* the bottom face                                */
  {4,5},{5,6},{6,7},{7,4},    /* the top face                                   */
  {0,4},{1,5},{2,6},{3,7}     /* edges connecting bottom to top                 */          
};
  
static float xmin[3],dx[3];   /* local copy of grid (just for convenience)      */
static float cl[3],ca[3];     
static float *val;
static int nx[3];

static int *bs;               /* bs[k]:number of points generated before cube k */
static unsigned char *ct;     /* ct[k]: cube type (index into CTbl)             */
static int npts;              /* total number of points in grid                 */

static int nv,nt;             /* local surface description                      */
static int mxpt,mxt;
static STriangle *Trg;
static SPoint    *Pts;

static int VAcc[256][12];  /* VAcc[t][j] = offset of vertex on edge j for cube type t */
static int Neigh[12];      /* cube# that added vertex                                 */
static int Edge[12]={0,3,0,3,0,3,0,3,8,8,8,8}; /* # of edge in cube that added vertex */
static int off[8];         /* offset of current cube corners into val                 */

static void make_acclist(void);
static void transform_points(int mtype);
static void compute_normals(void);
static float grd_val(float *pt);
static void grd_grd(float *pt,float *g);
static void shiftV(float *v,int n,float lev,int inv,float miss,int mode);

/*-------------------- isoMesh --------------------*/

Surface *isoMesh(Grid *g,float lev,int inSide,float missing,int Mtype,int axis,int dots)
{
  Surface *S=(Surface *)NULL;
  float x[3],y[3],z[3],*v,vll,vul,vlu,vuu;
  int dk=0,dj=0,di=0,dul=0,dlu=0,duu=0,ik=0,ij=0,ii=0;
  int i,j,k,pat,p;
  mat_4 T;

  if (!mvld(g)) return(S);
  else S=newSurface(SURF_TYPE_MSH);

  if (Mtype==ISOS_PDB||Mtype==ISOS_DFLT) {
    T=PDB_fta_matrix(g->cl,g->ca);
  }else if (Mtype==ISOS_WNP) {
    T=fract_to_angstrom(g->cl,g->ca);
  }else T=ident_4();

  for (p=0;p<3;p++) {
    if (axis & (1<<p)) {
      switch (p) {
      case 0:                      /* constant x contours */
	ii=1; ij=2; ik=0;
	di=g->n[0]; dj=g->n[0]; dk=1-(g->n[2]-1)*g->n[0]*g->n[1];
	dul=g->n[0]; dlu=g->n[0]*g->n[1]; duu=dul+dlu;
	break;
      case 1:                      /* constant y contours */
	ii=0; ij=2; ik=1;
	di=1; dj=g->n[0]*(g->n[1]-1)+1; dk=g->n[0]-(g->n[2]-1)*g->n[0]*g->n[1];
	dul=1; dlu=g->n[0]*g->n[1]; duu=dul+dlu;
	break;
      case 2:                      /* constant z contours */
	ii=0; ij=1; ik=2;
	di=1; dj=1; dk=g->n[0];
	dul=1; dlu=g->n[0]; duu=dul+dlu;
      default:
	break;
      }
      
      for (v=g->val,x[ik]=g->xmin[ik],k=0;k<g->n[ik];k++,x[ik]+=g->dx[ik],v+=dk) {
	if (dots) dotdotdot((p==0?"X":(p==1?"Y":"Z")),k,g->n[ik],60);
	y[ik]=z[ik]=x[ik];
	for (x[ij]=g->xmin[ij],j=0;j<g->n[ij]-1;j++,x[ij]+=g->dx[ij],v+=dj) {
	  for (x[ii]=g->xmin[ii],i=0;i<g->n[ii]-1;i++,x[ii]+=g->dx[ii],v+=di) {
	    if (x[0]<g->xl[0]||x[0]>g->xh[0]||
		x[1]<g->xl[1]||x[1]>g->xh[1]||
		x[2]<g->xl[2]||x[2]>g->xh[2] ) continue;     /* expensive and dirty... */
	    pat=0; 
	    vll=*v; vul=*(v+dul);
	    vlu=*(v+dlu); vuu=*(v+duu);
	    if (vll<=-missing || vul<=-missing || vlu<=-missing ||vuu<=-missing) continue;
	    if (vll>=missing || vul>=missing || vlu>=missing ||vuu>=missing) continue;
	    if (lev < vll) {
	      if (lev >= vul) pat|=(1<<0);
	      if (lev >= vlu) pat|=(1<<1);
	      if (lev >= vuu) pat|=(1<<2);
	    }
	    if (lev >= vll) {
	      if (lev < vul) pat|=(1<<0);
	      if (lev < vlu) pat|=(1<<1);
	      if (lev < vuu) pat|=(1<<2);
	    }
	    switch (pat) {
	    case 1:
	      y[ii]=x[ii]+g->dx[ii]*(vll-lev)/(vll-vul);y[ij]=x[ij];
	      z[ii]=x[ii]+g->dx[ii];z[ij]=x[ij]+g->dx[ij]*(vul-lev)/(vul-vuu);
	      AddTMDtoS(S,T,y,z);
	      break;
	    case 2:
	      y[ii]=x[ii];y[ij]=x[ij]+g->dx[ij]*(vll-lev)/(vll-vlu);
	      z[ii]=x[ii]+g->dx[ii]*(vlu-lev)/(vlu-vuu);z[ij]=x[ij]+g->dx[ij];
	      AddTMDtoS(S,T,y,z);
	      break;
	    case 3:              
	      if (inSide*(vll-vlu)>0) {    /* O.K. for points below = inside ? */
		y[ii]=x[ii]+g->dx[ii]*(vll-lev)/(vll-vul);y[ij]=x[ij];
		z[ii]=x[ii];z[ij]=x[ij]+g->dx[ij]*(vll-lev)/(vll-vlu);
		AddTMDtoS(S,T,y,z);
		y[ii]=x[ii]+g->dx[ii];y[ij]=x[ij]+g->dx[ij]*(vul-lev)/(vul-vuu);
		z[ii]=x[ii]+g->dx[ii]*(vlu-lev)/(vlu-vuu);z[ij]=x[ij]+g->dx[ij];
		AddTMDtoS(S,T,y,z);
	      }else{
		y[ii]=x[ii]+g->dx[ii]*(vll-lev)/(vll-vul);y[ij]=x[ij];
		z[ii]=x[ii]+g->dx[ii];z[ij]=x[ij]+g->dx[ij]*(vul-lev)/(vul-vuu);
		AddTMDtoS(S,T,y,z);
		y[ii]=x[ii];y[ij]=x[ij]+g->dx[ij]*(vll-lev)/(vll-vlu);
		z[ii]=x[ii]+g->dx[ii]*(vlu-lev)/(vlu-vuu);z[ij]=x[ij]+g->dx[ij];
		AddTMDtoS(S,T,y,z);
	      }
	      break;
	    case 4:
	      y[ii]=x[ii]+g->dx[ii]*(vlu-lev)/(vlu-vuu);y[ij]=x[ij]+g->dx[ij];
	      z[ii]=x[ii]+g->dx[ii];z[ij]=x[ij]+g->dx[ij]*(vul-lev)/(vul-vuu);
	      AddTMDtoS(S,T,y,z);
	      break;
	    case 5:
	      y[ii]=x[ii]+g->dx[ii]*(vll-lev)/(vll-vul);y[ij]=x[ij];
	      z[ii]=x[ii]+g->dx[ii]*(vlu-lev)/(vlu-vuu);z[ij]=x[ij]+g->dx[ij];
	      AddTMDtoS(S,T,y,z);
	      break;
	    case 6:
	      y[ii]=x[ii];y[ij]=x[ij]+g->dx[ij]*(vll-lev)/(vll-vlu);
	      z[ii]=x[ii]+g->dx[ii];z[ij]=x[ij]+g->dx[ij]*(vul-lev)/(vul-vuu);
	      AddTMDtoS(S,T,y,z);
	      break;
	    case 7:
	      y[ii]=x[ii];y[ij]=x[ij]+g->dx[ij]*(vll-lev)/(vll-vlu);
	      z[ii]=x[ii]+g->dx[ii]*(vll-lev)/(vll-vul);z[ij]=x[ij];
	      AddTMDtoS(S,T,y,z) ;
	      break;
	    default:
	      break;
	    }
	  }
	}
      }
      if (dots) dotdotdot("",0,0,0);
    }
  }
  return(S);
}

/*------------------------------------------------------------------------*/

Surface *isoSurface(Grid *g,float lev,int inv,float missing,int Mtype,int dots)
{
  Surface *S=(Surface *)NULL;
  int h,i,j,k,l,m,a,b,q;
  float *v,x[3];
  float s,t;

  if (!mvld(g)) return(S);

  /* copy grid data to local variables for easy ref. */

  for (j=0;j<3;j++) {
    nx[j]=g->n[j];
    xmin[j]=g->xmin[j];
    dx[j]=g->dx[j];
    cl[j]=g->cl[j];
    ca[j]=g->ca[j];
  }
  val=g->val;

  if (missing<0) missing=-missing;

  npts=nx[0]*nx[1]*nx[2];
  if (npts<1) {
    err_set(1,"isosurface: no points, nop.");
    return(S);
  }

  prtGraphics("isosurface: %d grid points, nx=%d, ny=%d, nz=%d\n",
	      npts,nx[0],nx[1],nx[2]);

  /* set up pointer arrays and stuff... */

  make_acclist();
  off[0]=0; off[1]=1; off[2]=nx[0]+1; off[3]=nx[0];
  for (k=0;k<4;k++) off[4+k]=off[k]+nx[0]*nx[1];
  /*  for (k=0;k<8;k++) for (l=0;l<3;l++) C[k][l]*=dx[l];*/

  /* assign ancillary storage */
  
  ct=(unsigned char *)malloc(npts*sizeof(unsigned char));
  bs=(int *)malloc((npts+1)*sizeof(int));
  if (!ct ||!bs) {
    err_set(1,"isosurface: failed to allocate cubes list");
    if (ct) free(ct);
    if (bs) free(bs);
    return(S);
  }

  /* shift contour level to zero, turn inside above... */

  shiftV(val,npts,lev,inv,missing,ISOS_LEVFORW);

  /* count loop: count number of vertices and number of triangles */

  nv=0; nt=0;
  for (m=0,v=val,k=0;k<nx[2];k++) {
    if (dots) dotdotdot("counting",k,nx[2],60);
    if (k==nx[2]-1) {
      off[4]-=nx[0]*nx[1]; off[5]-=nx[0]*nx[1]; 
      off[6]-=nx[0]*nx[1]; off[7]-=nx[0]*nx[1];
    }
    for (j=0;j<nx[1];j++) {
      if (j==nx[1]-1) {
	off[2]-=nx[0]; off[3]-=nx[0];off[6]-=nx[0]; off[7]-=nx[0];
      }
      for (i=0;i<nx[0];i++,v++,m++) {
	if (i==nx[0]-1) {
	  off[1]-=1; off[2]-=1; off[5]-=1; off[6]-=1;
	}
	bs[m]=nv; ct[m]=0;
	for (l=0;l<8;l++) if (*(v+off[l])<-missing||*(v+off[l])>missing) break;
	if (l==8) {
	  for (l=0;l<8;l++) if (*(v+off[l])>0) ct[m]+=(1<<l);
	  for (l=0;l<CTbl[ct[m]].nv;l++) if (Neigh[CTbl[ct[m]].v[l]]==0) nv++;
	  nt+=CTbl[ct[m]].nf;
	}else ct[m]=0;
      }
      off[1]+=1; off[2]+=1; off[5]+=1; off[6]+=1;
    }
    off[2]+=nx[0]; off[3]+=nx[0]; off[6]+=nx[0]; off[7]+=nx[0];
  }
  if (dots) dotdotdot("",0,0,0);
  prtGraphics("isosurface: %d vertices, %d triangles.\n",nv,nt);
  off[4]+=nx[0]*nx[1]; off[5]+=nx[0]*nx[1]; off[6]+=nx[0]*nx[1]; off[7]+=nx[0]*nx[1];
  bs[m]=nv;

  mxpt=nv;
  mxt=1.2*nt;
  Pts=(SPoint *)malloc(mxpt*sizeof(SPoint));
  Trg=(STriangle *)malloc(mxt*sizeof(STriangle));

  if (!Pts || !Trg) {
    shiftV(val,npts,lev,inv,missing,ISOS_LEVBACK);
    err_set(1,"isosurface: memory allocation failed for points and triangles.");
    free(bs);
    free(ct);
    if (Pts) free(Pts);
    if (Trg) free(Trg);
    return(S);
  }

  /* Compute vertex positions */

  for (x[2]=xmin[2],m=0,nv=0,v=val,k=0;k<nx[2];k++,x[2]+=dx[2]) {
    if (dots) dotdotdot("vertices",k,nx[2],60);
    if (k==nx[2]-1) {
      off[4]-=nx[0]*nx[1]; off[5]-=nx[0]*nx[1]; 
      off[6]-=nx[0]*nx[1]; off[7]-=nx[0]*nx[1];
    }
    for (x[1]=xmin[1],j=0;j<nx[1];j++,x[1]+=dx[1]) {
      if (j==nx[1]-1) {
	off[2]-=nx[0]; off[3]-=nx[0];off[6]-=nx[0]; off[7]-=nx[0];
      }
      for (x[0]=xmin[0],i=0;i<nx[0];i++,v++,m++,x[0]+=dx[0]) {
	if (i==nx[0]-1) {
	  off[1]-=1; off[2]-=1; off[5]-=1; off[6]-=1;
	}
	if (bs[m]!=bs[m+1]) {
	  for (l=0;l<CTbl[ct[m]].nv;l++) {
	    if (Neigh[CTbl[ct[m]].v[l]]==0) {
	      q=CTbl[ct[m]].v[l];
	      a=E[q][0]; b=E[q][1];
	      t= *(v+off[a])/( *(v+off[a]) - *(v+off[b])); s=1-t;
	      for (h=0;h<3;h++) Pts[nv].x[h]=x[h]+dx[h]*(s*C[a][h]+t*C[b][h]);
	      nv++;
	    }
	  }
	}
      }
      off[1]+=1; off[2]+=1; off[5]+=1; off[6]+=1;
    }
    off[2]+=nx[0]; off[3]+=nx[0]; off[6]+=nx[0]; off[7]+=nx[0];
  }
  if (dots) dotdotdot("",0,0,0);
  off[4]+=nx[0]*nx[1]; off[5]+=nx[0]*nx[1]; off[6]+=nx[0]*nx[1]; off[7]+=nx[0]*nx[1];

  /* generate list of triangles */

  for (nt=0,m=0,k=0;k<nx[2]-1;k++,m+=nx[0]) {
    if (dots) dotdotdot("triangles",k,nx[2],60);
    for (j=0;j<nx[1]-1;j++,m++) {
      for (i=0;i<nx[0]-1;i++,m++) {
	for (l=0;l<CTbl[ct[m]].nf;l++,nt++) {
	  for (q=0;q<3;q++) {
	    a=m+Neigh[CTbl[ct[m]].f[l][q]];
	    b=Edge[CTbl[ct[m]].f[l][q]];
	    Trg[nt].v[q]=bs[a]+VAcc[ct[a]][b];
	  }
	}
      }
    }
  }
  if (dots) dotdotdot("",0,0,0);

  prtGraphics("isosurface: found %d points and %d triangles...\n",nv,nt);
  compute_normals();
  transform_points(Mtype);

  /* revert zero shift and inversion ... */

  shiftV(val,npts,lev,inv,missing,ISOS_LEVBACK);

  /* prepare return package... */

  S=newSurface(SURF_TYPE_TRG);
  if (mvld(S)) {
    S->v=Pts;
    S->t=Trg;
    S->nv=nv;
    S->nt=nt;
  }else{
    free(Pts);
    free(Trg);
  }
  if (mvld(bs)) free(bs);
  if (mvld(ct)) free(ct);

  return(S);
}
/*-------------------- make_acclist --------------------*/

static void make_acclist(void)
{
  int i,j,k;

  /* Neigh[e] = offset to cube that generated edge e */

  Neigh[0]=0; Neigh[1]=1; Neigh[2]=nx[0]; Neigh[3]=0;
  Neigh[4]=nx[0]*nx[1]; Neigh[5]=Neigh[4]+1; Neigh[6]=Neigh[4]+nx[0]; Neigh[7]=Neigh[4];
  Neigh[8]=0; Neigh[9]=1; Neigh[10]=nx[0]+1; Neigh[11]=nx[0];

  /* VAcc[type][j] = offset of vertex on edge j for this type of cube */

  for (i=0;i<256;i++) {
    for (j=0;j<12;j++) VAcc[i][j]=0;
    for (k=0,j=0;j<CTbl[i].nv;j++) {
      if (Neigh[CTbl[i].v[j]]==0) VAcc[i][CTbl[i].v[j]]=k++;
    }
  }
   
}
/*-------------------- transform_points --------------------*/

void transform_points(int Mtype)
{
  mat_4 M,N;
  float x[3],s;
  int i,j,k;

  if (Mtype==ISOS_PDB||Mtype==ISOS_DFLT) {
    M=PDB_fta_matrix(cl,ca);
    prtGraphics("isosurface: transforming coordinates using PDB convention...\n");
  }else if (Mtype==ISOS_WNP) {
    M=fract_to_angstrom(cl,ca);
    prtGraphics("isosurface: transforming coordinates using Wit!P convention...\n");
  }else return;

  N=inv_4(M);
  for (k=0;k<nv;k++) {
    for(i=0;i<3;i++) for (x[i]=M.r[i].c[3],j=0;j<3;j++) x[i]+=M.r[i].c[j]*Pts[k].x[j];
    for(i=0;i<3;i++) Pts[k].x[i]=x[i];
    for (s=0,i=0;i<3;i++) {
      for (x[i]=0,j=0;j<3;j++) x[i]+=Pts[k].n[j]*N.r[j].c[i];
      s+=x[i]*x[i];
    }
    for (i=0;i<3;i++) Pts[k].n[i]=x[i]/sqrt(s);
  }
}
/*-------------------- compute_normals --------------------*/

static void compute_normals(void)
{
  int i,j;
  float s;
  
  /* Set normals at vertices to gradient */

  for (i=0;i<nv;i++) {
    grd_grd(Pts[i].x,Pts[i].n);
    s=sqrt(Pts[i].n[0]*Pts[i].n[0]+Pts[i].n[1]*Pts[i].n[1]+Pts[i].n[2]*Pts[i].n[2]);
    if (s>0) {
      s=-1/s;
      for (j=0;j<3;j++) Pts[i].n[j]*=s;
    }else{
      Pts[i].n[0]=1.0;
      Pts[i].n[1]=0.0;
      Pts[i].n[2]=0.0;
    }
  }    
}
/*-------------------- grid_value --------------------*/

static float grd_val(float *pt)
{
  /* interpolated grid value at point x */

  float r,x[3],x0[3],c[3],V[8],*v;
  int ix[3],k;

  r=0;

  for (k=0;k<3;k++) {
    x[k]=clamp(pt[k],xmin[k],xmin[k]+(nx[k]-1.001)*dx[k]);
    ix[k]=clamp((x[k]-xmin[k])/dx[k],0,nx[k]-2);
    x0[k]=xmin[k]+ix[k]*dx[k];
    c[k]=(x[k]-x0[k])/dx[k];
  }
  v=val+ix[2]*(nx[0]*nx[1])+ix[1]*nx[0]+ix[0];
  V[0]=v[0];
  V[1]=v[1]-V[0];
  V[2]=v[nx[0]]-V[0];
  V[4]=v[nx[0]*nx[1]]-V[0];
  V[3]=v[nx[0]+1]-V[0]-V[1]-V[2];
  V[5]=v[nx[0]*nx[1]+1]-V[0]-V[1]-V[4];
  V[6]=v[nx[0]+nx[0]*nx[1]]-V[0]-V[2]-V[4];
  V[7]=v[nx[0]+nx[0]*nx[1]+1]-V[0]-V[1]-V[2]-V[3]-V[4]-V[5]-V[6];
  r=V[0]+
    c[0]*(c[1]*(V[7]*c[2]+V[3])+c[2]*V[5]+V[1])+
    c[1]*(c[2]*V[6]+V[2])+
    c[2]*V[4];
  return(r);
}
/*-------------------- grid_grad --------------------*/

static void grd_grd(float *pt,float *g)
{
  /* gradient of grid at point pt */

  float d;
  int k;

  for (k=0;k<3;k++) g[k]=0.0;
  for (k=0;k<3;k++) {
    d=0.33333*dx[k];
    pt[k]+=d; g[k]=grd_val(pt);
    pt[k]-=2*d; g[k]=(g[k]-grd_val(pt))/(2*d);
    pt[k]+=d;
  }
}
/*-------------------- shiftV --------------------*/

static void shiftV(float *v,int n,float lev,int inv,float miss,int mode)
{
  int i;

  switch (mode) {
  case ISOS_LEVFORW:
    /* shift contour level to zero and invert */
    for (i=0;i<n;i++)
      if (v[i]>-miss&&v[i]<miss) v[i]=(v[i]-lev)/inv;
    break;
  case ISOS_LEVBACK:
    /* invert and shift zero to contour level */
    for (i=0;i<n;i++) 
      if (v[i]>-miss&&v[i]<miss) v[i]=v[i]/inv+lev;
    break;
  }
}
