#include "list.h"
#include "mystdlib.h"

/*-------------------- make_list_entry --------------------*/

ListEntry *make_list_entry(int size,void (*init_data)(void *data))
{
  /* create a new list entry, allocate data-chunk of requested size */
  /* if size > 0, init_data will be called to initialize the data   */

  ListEntry *r;

  r=(ListEntry *)malloc(sizeof(ListEntry));
  if (mvld(r)) {
    r->prv=(ListEntry *)NULL;
    r->nxt=(ListEntry *)NULL;
    if (size>0) {
      r->data=(void *)malloc(size);
      if (!mvld(r->data)) {
	free(r);
	r=(ListEntry *)NULL;
      }else{
	if (init_data) init_data(r->data);
      }
    }else{
      r->data=(void *)NULL;
    }
  }
  return(r);
}
/*-------------------- delete_list_entry --------------------*/

ListEntry *delete_list_entry(ListEntry *e,void (*del_data)(void *data))
{
  /* remove entry e from list, free associated data */

  if (mvld(e)) {
    remove_list_entry(e);
    if (mvld(e->data)) {
      if (del_data) del_data(e->data);
    }  
    free(e);
  }
  return((ListEntry *)NULL);
}
/*-------------------- insert_list_entry --------------------*/

ListEntry *insert_list_entry(ListEntry *e)
{
  /* insert entry e into list. if e->prv, insert after e->prv,
     if e->nxt insert after e->nxt. NOTE : e->prv && e->nxt is bad ! */

  if (mvld(e)) {
    if (mvld(e->prv)) {
      if (mvld(e->prv->nxt)) {
	e->prv->nxt->prv=e;
	e->nxt=e->prv->nxt;
      }
      e->prv->nxt=e;
    }else if (mvld(e->nxt)) {
      if (mvld(e->nxt->prv)) {
	e->nxt->prv->nxt=e;
	e->prv=e->nxt->prv;
      }
      e->nxt->prv=e;
    }
  }
  return(e);
}
/*-------------------- remove_list_entry --------------------*/

ListEntry *remove_list_entry(ListEntry *e)
{
  /* remove entry e from list, return pointer to orphaned entry e */

  if (mvld(e)) {
    if (mvld(e->prv)) e->prv->nxt=e->nxt;
    if (mvld(e->nxt)) e->nxt->prv=e->prv;
    e->prv=(ListEntry *)NULL;
    e->nxt=(ListEntry *)NULL;
  }
  return(e);
}
/*-------------------- delete_list_entries --------------------*/

ListEntry *delete_list_entries(ListEntry *e,void (*del_data)(void *data))
{
  /* delete all entries except head in list containing e, return pointer to
     head of now empty list.   */

  if (mvld(e)) {
    while(mvld(e->prv)) e=e->prv;
    while(mvld(e->nxt)) delete_list_entry(e->nxt,del_data);
  }
  return(e);
}
/*-------------------- delete_list --------------------*/

ListEntry *delete_list(ListEntry *e,void (*del_data)(void *data))
{
  /* delete list, incl. head. */

  e=delete_list_entries(e,del_data);
  if (mvld(e)) e=delete_list_entry(e,del_data);
  return(e);
}
/*-------------------- append_entry_to_list --------------------*/

ListEntry *append_entry_to_list(ListEntry *e,ListEntry *h)
{
  /* remove entry e from it's list (if any), then add to end of list h */

  e=remove_list_entry(e);
  if (mvld(h)) {
    while(mvld(h->nxt)) h=h->nxt;
    e->prv=h;
    insert_list_entry(e);
  }
  return(e);
}
/*-------------------- n_entries --------------------*/

int n_entries(ListEntry *e)
{
  /* return number of entries following e */

  int r;

  r=0;
  if (mvld(e)) {
    for (;mvld(e->nxt);e=e->nxt) r++;
  }
  return(r);
}
