#include <math.h>
#include <unistd.h>
#include <ctype.h>
#include "strutil.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "object.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"
#ifdef mingw
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#define TRUE  1
#define FALSE 0

#define CHUNKSIZE 10

#define key_is(x) (streq(key->key,x))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

extern int graphix;

typedef struct {
  char *name;
  wnpmacro_t *(*get)(void);
} macNamTab_t;

#if defined(mingw)
static int mkstemp(char *template)
{
  mktemp(template);
  return(open(template,O_RDWR));
}
#endif

static macNamTab_t macTab[] = {
  { "maccs" , getMACCSmacros }
};

static wnpmacro_t *get_macro_text(wnpmacro_t *m)
{
  Token *key=NULL;
  int loop=TRUE;
  Stream *sv;
  char *p;

  sv=new_stream(0);
  if (sv) set_stream_props(sv,CurrStream()); /* save current stream attributes */
  else err_set(1,"failed to set input stream attributes");

  set_stream(CurrStream(),                   /* turn all translations, variable expansions etc. off */
	     "eol=yes,redirection=off,continuation,spctok=off,expvar=off");

  while (loop && m) {
    key=Get_RawLine_M("macro line >","ORCAM");
    if (key->status==TOK_OK) {
      for (p=key->text;isspace(*p);) p++;
      if (strneq(p,"ORCAM",5)) loop=FALSE;
      else if (strneq(p,"abort",5)) m=unlink_macro(m);
      else add_macro_line(m,key->text);
    }
  }

  if (sv) {                                  /* restore "normal" stream attributes */
    set_stream_props(CurrStream(),sv);
    del_stream(sv);
  }

  free_Token(key);
  return(m);
}

static void list_macro(wnpmacro_t *m)
{
  int i;

  if (m) {
    prtGraphics("Macro: %s\n",m->name);
    for (i=0;i<m->n;i++) {
      prtGraphics("%s",m->txt[i]);
    }
  }
}

/*========================================================================*/

#define MAXBUFF 65535

void cmd_macro(wnpmacro_t *macros)
{
  Token *key=NULL;
  int loop=TRUE;
  int i,tmpfd,oldn;
  wnpmacro_t *m;
  char *menu=NULL,**oldtxt;
  char buff[MAXBUFF],tmpfn[4096];
  FILE *file;

  if (!macros) {
    err_set(1,"No root for macros, macro command is invalis in this context.");
    loop=FALSE;
  }

  while (loop) {
    err_dump();
    key=free_Token(key);
    key=Get_Kwd("macro command: ",";","define delete edit list ' ' end");
    if (key->status==TOK_OK) {

      if (key_is("define")) {
	m=cmd_define_macro(macros);

      }else if (key_is("delete")) {
        for (;;) {
	  key=free_Token(key); free_menu(menu);
	  menu=make_macro_menu("* ''",macros);
	  key=Get_Kwd("macro name: ",";",menu);
	  if (key->status==TOK_OK) {
	    if (key_is("*")&&macros) {
	      while(macros->nxt) unlink_macro(macros->nxt);
	    }else {
	      m=get_macro_by_name(macros,key->key);
	      unlink_macro(m);
	    }
	  }else if (key->status==TOK_NOK) err_set(1,"invalid macro name token: %s",key->text);
	  else break;
	}
      }else if (key_is("list")) {
	key=free_Token(key); free_menu(menu);
	menu=make_macro_menu(NULL,macros);
	key=Get_Kwd("macro name: ",";",menu);
	if (key->status==TOK_OK) {
	  m=get_macro_by_name(macros,key->key);
	  list_macro(m);
	}else err_set(1,"invalid macro name: %s",key->text);
	
      }else if (key_is("edit")) {
	m=NULL;
	key=free_Token(key); free_menu(menu);
	menu=make_macro_menu(NULL,macros);
	key=Get_Kwd("macro name: ",";",menu);
	if (key->status==TOK_OK) {
	  m=get_macro_by_name(macros,key->key);
	  sprintf(tmpfn,"macro_%s_XXXXXX",key->key);
	}else if (key->status==TOK_NOK) {
	  m=new_macro();
	  if (m) {
	    set_macro_name(m,key->text);
	    insert_macro(m,macros);
	    key->status=TOK_OK;
	    sprintf(tmpfn,"macro_%s_XXXXXX",key->text);
	  }else err_set(1,"failed to create new macro %s",key->text);
	}
	if (key->status==TOK_OK) {
	  tmpfd=mkstemp(tmpfn);
	  if (tmpfd!=-1) {
	    if (m) {
	      for (i=0;i<m->n;i++) {
		if (write(tmpfd,m->txt[i],strlen(m->txt[i]))<0) break;
	      }
	    }
	    close(tmpfd);
	    sprintf(buff,"wnp_edit %s",tmpfn);
	    startCmd(buff,"","",TRUE,"normal");
	    file=fopen(tmpfn,"r");
	    if (file) {
	      oldtxt=m->txt; oldn=m->n;
	      m->txt=NULL; m->n=0;
	      for (i=0;fgets(buff,MAXBUFF,file);i++) add_macro_line(m,buff);
	      if (ferror(file)) {  /* ignore changes... */
		for (i=0;i<m->n;i++) free(m->txt[i]);
		free(m->txt);
		m->txt=oldtxt;
		m->n=oldn;
	      }else{  /* accept, delete old text */
		for (i=0;i<oldn;i++) free(oldtxt[i]);
		free(oldtxt);
	      }
	      fclose(file);
	    }
	    unlink(tmpfn);
	  }else err_set(1,"cant edit file %s",tmpfn);
	}

     }else if (key_is("end")) {
	loop=FALSE;
      }
    }else if (key->status==TOK_NOK) err_set(1,"invalid macro keyword: %s",key->text);
    else if (key->status==TOK_ABORT||key->status==TOK_EOC||key->status==TOK_EOL) loop=FALSE;
    if (end_parse()!=TOK_OK) {
      err_set(0,"extra characters at end of command ignored.");
    }
    start_parse();
  }

  free_Token(key);
  free_menu(menu);
}

wnpmacro_t *cmd_define_macro(wnpmacro_t *macros)
{
  Token *key=NULL;
  wnpmacro_t *m=NULL, *old;

  key=free_Token(key);
  key=Get_String("macro name: ",";");
  if (key->status==TOK_OK) {
    if ((old=get_macro_by_name(macros, key->text))) {
      err_set(1,"Warning the macro name %s is in use.", key->text);
      err_set(1,"New macro will replace previous one");
      err_dump();
    }
    m=new_macro();
    if (m) {
      set_macro_name(m,key->text);
      m=get_macro_text(m);
      if (old && m) {
	replace_macro(old,m);
	unlink_macro(old);
      }else if (m) insert_macro(m,macros);
    }else err_set(1,"failed to allocate memory for new macro.");
  }else if (key->status==TOK_NOK) err_set(1,"invalid macro name token: %s",key->text);
  
  free_Token(key);
  return(m);
}

wnpmacro_t *new_macro(void)
{
  wnpmacro_t *r;

  r=malloc(sizeof(wnpmacro_t));
  if (r) {
    r->prv=NULL;
    r->nxt=NULL;
    r->name=NULL;
    r->n=0;
    r->txt=NULL;
    r->n_open=0;
  }
  return(r);
}

void set_macro_name(wnpmacro_t *m,char *name)
{
  if (m) {
    if (m->name) free(m->name);
    if (name) m->name=strdup(name);
    else m->name=NULL;
  }
}

void add_macro_line(wnpmacro_t *m,char *line)
{
  if (m&&line) {
    if (m->n%CHUNKSIZE==0) {
      m->txt=realloc(m->txt,(m->n+CHUNKSIZE)*sizeof(wnpmacro_t));
    }
    if (m->txt) {
      m->txt[m->n]=strdup(line);
      if (m->txt[m->n]) m->n++;
    }
  }
}

void insert_macro(wnpmacro_t *m,wnpmacro_t *list)
{
  if (list&&m) {
    while (list->nxt) list=list->nxt;
    list->nxt=m;
    m->prv=list;
    m->nxt=NULL;
  }
}

void replace_macro(wnpmacro_t *old, wnpmacro_t *new)
{
  if (old && new) {
    new->prv = old->prv;
    new->nxt = old->nxt;
    if (old->prv) old->prv->nxt = new;
    if (old->nxt) old->nxt->prv = new;
    old->prv = NULL;
    old->nxt = NULL;
  }
}

wnpmacro_t *unlink_macro(wnpmacro_t *m)
{
  if (m) {
    if (m->prv) m->prv->nxt=m->nxt;
    if (m->nxt) m->nxt->prv=m->prv;
    m=delete_macro(m);
  }
  return(m);
}

wnpmacro_t *delete_macro(wnpmacro_t *m)
{
  if (m) {
    if (!m->prv&&!m->nxt&&m->n_open<1) {
      if (m->name) free(m->name);
      if (m->txt) {
	while (m->n>0) {
	  m->n--;
	  if (m->txt[m->n]) free(m->txt[m->n]);
	}
	free(m->txt);
      }
      free(m);
      m=NULL;
    }
  }
  return(m);
}

wnpmacro_t *get_macro_by_name(wnpmacro_t *macros,char *name)
{
  wnpmacro_t *m=NULL;

  if (name&&macros) {
    for (m=macros->nxt;m;m=m->nxt) if (streq(m->name,name)) break;
  }
  return(m);
}

char *make_macro_menu(char *prfx,wnpmacro_t *macros)
{
  char *menu=NULL;
  wnpmacro_t *m;
  int l=0;

  if (macros) {
    for (m=macros->nxt;m;m=m->nxt) {
      l+=m->name&&strlen(m->name)>0?strlen(m->name)+1:0;
    }
    l++;
  }else l=1;
  if (prfx) l+=strlen(prfx)+1;
  menu=malloc(l);
  if (menu) {
    *menu='\0';
    if (prfx) sprintf(menu,"%s ",prfx);
    if (macros) {
      for (m=macros->nxt;m;m=m->nxt) 
	if (m->name&&strlen(m->name)>0) sprintf(menu+strlen(menu),"%s ",m->name);
    }
  }
  return(menu);
}

void execute_macro(wnpmacro_t *m)
{
  Stream *s;

  if (m) {
    s=push_stream(CurrStream(),NULL);
    if (s) set_stream_macro(s,m);
    else err_set(1,"failed to execute macro %s.",m->name);
  }
}

void open_macro(wnpmacro_t *m)
{
  if (m) m->n_open++;
}

void close_macro(wnpmacro_t *m)
{
  if (m) m->n_open--;
  delete_macro(m);
}

char *read_macro(wnpmacro_t *m,int lnr)
{
  char *r;

  if (m&&lnr<m->n) r=m->txt[lnr];
  else r=NULL;

  return(r);
}

int nMacroTypes(void)
{
  return(sizeof(macTab)/sizeof(macNamTab_t));
}

char *getMacroType(int i)
{
  char *type;

  if (i<sizeof(macTab)/sizeof(macNamTab_t)) type=macTab[i].name;
  else type=NULL;
  return(type);
}

wnpmacro_t *getMacros(char *type)
{
  int n,i;
  wnpmacro_t *macros;

  n=sizeof(macTab)/sizeof(macNamTab_t);
  for (i=0;i<n;i++) if (streq(type,macTab[i].name)) break;
  if (i<n) macros=macTab[i].get();
  else macros=NULL;
  return(macros);
}

void clearMacros(void)
{
  int i;
  wnpmacro_t *macros;

  for (i=0;i<sizeof(macTab)/sizeof(macNamTab_t);i++) {
    if ((macros=macTab[i].get())) {
      while (macros->nxt) unlink_macro(macros->nxt);
    }
  }
}
