#include <string.h>
#include "maestream.h"
#include "strutil.h"
#include "mystdlib.h"
#include "errhand.h"
#include <ctype.h>

#define TRUE  1
#define FALSE 0
#define TT_REGULAR   1
#define TT_COMMENT   2
#define TT_QUOTED    3
#define FOREVER while(TRUE)

maestream_t *maeStream_open(char *filename,UCpipe *pipe,int maxbuff,char *token,int mts)
{
  maestream_t *s;
  
  s=malloc(sizeof(maestream_t));
  if (s) {
    s->own_p=FALSE;
    s->eof=FALSE;
    s->own_t=FALSE;
    if (maxbuff<100) maxbuff=100;
    s->buff=malloc(maxbuff*sizeof(char));
    if (!token) {
      if (mts>1) s->token=malloc(mts*sizeof(char));
      else s->token=NULL;
      s->own_t=TRUE;
    }else{
      s->token=token;
    }
    if (s->buff && s->token) {
      s->buffsz=maxbuff;
      s->buff[0]='\0';
      s->token[0]='\0';
      s->pos=0;
      s->tpos=0;
      s->mts=mts;
      if (pipe) {
	s->in=pipe;
      }else if (filename) {
	s->in=UCpopen(filename);
	s->own_p=TRUE;
      }
      if (!s->in) {
	if (s->token && s->own_t) free(s->token);
	free(s->buff);
	free(s);
	s=NULL;
      }
    }else{
      if (s->token && s->own_t) free(s->token);
      if (s->buff) free(s->buff);
      free(s);
      s=NULL;
    }
  }
  return(s);
}

maestream_t *maeStream_close(maestream_t *s)
{
  if (s) {
    if (s->own_p) UCpclose(s->in);
    if (s->token && s->own_t) free(s->token);
    if (s->buff) free(s->buff);
    free(s);
    s=NULL;
  }
  return(s);
}

char *mae_token(maestream_t *s)
{
  char *r=NULL;
  int ttype=0;

  /* skip over white-space */
  do {
    while (s->buff[s->pos]&&isspace(s->buff[s->pos])) s->pos++;
    if (!s->buff[s->pos]) {
      if (UCpgets(s->buff,s->buffsz,s->in)) {
	s->pos=0;
      }else{
	return(NULL);
      }
    }
  } while(!s->buff[s->pos] || isspace(s->buff[s->pos]));
  
  /* at start of token, initialize token */
  s->token[0]='\0';
  s->tpos=0;

  /* determine token type */
  if (s->buff[s->pos]=='\\' && strchr("\"'",s->buff[s->pos])) {
    ttype=TT_REGULAR;
    s->pos++;
  }else if (s->buff[s->pos]=='"') {
    ttype=TT_QUOTED;
    s->pos++;
  }else if (s->buff[s->pos]=='#') {
    ttype=TT_COMMENT;
    s->pos++;
  }else{
    ttype=TT_REGULAR;
  }
    
  /* parse and store token */
  switch (ttype) {

  case TT_REGULAR: 
    do {
      while (s->buff[s->pos]&&!isspace(s->buff[s->pos])) {
	if (s->tpos<s->mts-1) {
	  s->token[s->tpos]=s->buff[s->pos];
	  s->tpos++;
	}
	s->pos++;
      }
      if (!s->buff[s->pos]) {
	s->buff[0]='\0';
	if (!UCpgets(s->buff,s->buffsz,s->in)) break;
	s->pos=0;
      }else break;
    } FOREVER;
    s->token[s->tpos]='\0';
    r=strlen(s->token)>0?s->token:(char *)NULL;
    break;

  case TT_COMMENT:
    do {
      while (s->buff[s->pos]&&s->buff[s->pos]!='#') {
	s->pos++;
      }
      if (!s->buff[s->pos]) {
	s->buff[0]='\0';
	if (!UCpgets(s->buff,s->buffsz,s->in)) break;
	s->pos=0;
      }else{
	s->pos++;
	break;
      }
    } FOREVER;
    r=mae_token(s);
    break;

  case TT_QUOTED:
    do {
      while (s->buff[s->pos]&&s->buff[s->pos]!='"') {
	if (s->buff[s->pos]=='\\'&&s->buff[s->pos+1]=='"') s->pos++;
	if (s->tpos<s->mts-1) {
	  s->token[s->tpos]=s->buff[s->pos];
	  s->tpos++;
	}
	s->pos++;
      }
      if (!s->buff[s->pos]) {
	s->buff[0]='\0';
	if (!UCpgets(s->buff,s->buffsz,s->in)) break;
	s->pos=0;
      }else{
	s->pos++;
	break;
      }
    } FOREVER;
    s->token[s->tpos]='\0';
    r=s->token;
    break;
  }
  return(r);
}
