#include <stdio.h>
#include <math.h>
#include <sys/param.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "transpor.h"
#include "errhand.h"
#include "maestream.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

#define RECSZ 1024
#define TRUE 1
#define FALSE 0
#define YES 1
#define NO 0

typedef struct fts_ { int *idx; char *mnem; } fts_;
typedef struct maeBond_t { int f; int t; int bo; } maeBond_t;

extern int graphix;
extern Node *LibRoot;
extern Node *MolRoot;

static int MM2WnpBT(int bt);
static Atomdef *MM2WnpAT(int tp);
static int MM2WnpColor(int color);
static int Wnp2MMColor(int c);
static char *MM2PDB(char *one);
static int MMNumType(Atom *a);
static int MMBondOrder(Bond *b);
static void fixCnfOrder(Node *root,CnfNode *c,Atom **atms);
static void MMNames(Atom *a,Node *node,char *resnum,char *one,char *chain,char *resnam);
static char *MMpdbName(Atom *a,char *buff,int size);

static char **getMAEvalues(maestream_t *in,char **tags);
static char **getMAEtags(maestream_t *in,int seqnr);
static char **freeMAEvalues(char **values);
static char **freeMAEtags(char **tags);
static int findMAEtag(char *tag,char **tags);
static int countMAEblock(char *tag);
static Atom **readMAEatoms(maestream_t *in,int nat,Node *node);
static maeBond_t *readMAEbonds(maestream_t *in,int nbnd);

/* MMods's single letter residue codes... */

static char *MM_CNV[] = {
  "A","B","C","D","E",
  "F","G","H","I","J",
  "K","L","M","N","O",
  "P","Q","R","S","T",
  "U","V","W","X","Y",
  "Z","1","2","3","4",
  "5","6","7","8","9",
  "0","a","b","c","d",
  "e","f","g","h","i",
  "j","k","l","m","n",
  "o","p"," "
};

/* ...and corresponding  PDB "3" letter codes */

static char *PDB_CNV[] = {
  "ALA","ASX","CYS","ASP","GLU",
  "PHE","GLY","HIS","ILE","HSE",
  "LYS","LEU","MET","ASN","HYP",
  "PRO","GLN","ARG","SER","THR",
  "HYL","VAL","TRP","UNK","TYR",
  "GLX","ORN","SAR","TAU","THY",
  "ALB","PHO","PCA","ACE","FOR",
  "HOH","A"  ,"C"  ,"G"  ,"U"  ,
  "1MA","5MC","OMC","2MG","M2G",
  "7MG","OMG","5MU","H2U","PSU",
  "YG" ,"T"  ,"UNK",
  ""                              /* one more than MM ! (or else...) */
};

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

#define MAXTOKENSIZE 1024

#define Return(x,y) \
   {err_set(x,"read_mae: %s",y);\
   maeStream_close(in);\
   if (atoms) free(atoms);\
   if (bonds) free(bonds);\
   return(x);\
   }

/*********************************************************************/
int read_mmod(UCpipe *in,char *file,Node *node)
{
  /* read Macromodel structure file (just the first frame...) */

  Atom **atms=(Atom **)NULL;
  int nat,own_p,status;
  char rec[RECSZ],*oname;

  own_p=(in ? FALSE : TRUE);
  if (own_p) in=UCpopen(extfil(rec,file,".mmd",NO));
  if (!in) return(1);

  oname=Strdup(node->name);
  status=read_mmod1(in,node,&nat,&atms);
  set_node_name(node,oname);
  if (oname) free(oname);

  if (mvld(atms)) free(atms);
  if (own_p) UCpclose(in); 
  return(status);
}  

int read_mmod1(UCpipe *in,Node *node,int *Nat,Atom ***Atms)
     
     /*   Read 1st frame of compressed Macromodel coordinate file.
	  Returns non-zero status if read fails.  */
{
  Node *resid,*chain;
  char rec[RECSZ],buff[RECSZ];
  Node *ResLib;
  Lib  *reslib;
  Atom *a,**atms;
  Bond *bnd;
  Atomdef *default_type;
  int n,nat=0,i,status,tp,nb[6],bt[6],resnr,color,nnegb,nnegq;
  float q,biso;
  char rest[2],chainn[2],pdbr[5],atnm[5],oldres[128],newres[128],resnam[128];
  
  status=0;
  *Atms=(Atom **)NULL;

  if (!in) return(1);
  
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    return(-1);
  }
  reslib=(Lib *)ResLib;
  default_type=find_atomdef_name((Node *)reslib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }
  
  atms=(Atom **)NULL;
  if (!UCpgets(rec,RECSZ,in)) rec[0]='\0';
  if (strrchr(rec,'\n')) *strrchr(rec,'\n')='\0';
  if (!UCpeof((in))) {
    nat=str2i(strfld(buff,rec+1,1,5));
    set_node_name(node,strfld(buff,rec+8,1,70));
    if (nat>0) {
      atms=(Atom **)malloc(nat*sizeof(Atom *));
      for (i=0;i<nat;i++) atms[i]=(Atom *)NULL;
    }else{
      err_set(1,"no atoms.");
      status=1;
    }
  }else{
    err_set(1,"readmmod: empty file.");
    status=1;
  }

  resid=node;
  oldres[0]='\0';
  for (n=0;n<nat&&!status;n++) {
    if (!UCpgets(rec,RECSZ,in)) break;
    if (strrchr(rec,'\n')) *strrchr(rec,'\n')='\0';
    for (i=strlen(rec);i<132;i++) rec[i]=' ';
    rec[i]='\0';
    a=(Atom *)make_node(OBJ_ATOM);
    if (is_atom(a)) {

      /* decode atom record... */
      
      tp=str2i(strfld(buff,rec+1,1,3));
      for (i=0;i<6;i++) {
	nb[i]=str2i(strfld(buff,rec+4+i*8,2,6));
	bt[i]=str2i(strfld(buff,rec+4+i*8,8,8));
      }
      for (i=0;i<3;i++) a->x[i]=str2f(strfld(buff,rec+52+i*12,1,12));
      a->nxyz=1;
      resnr=str2i(strfld(buff,rec+89,1,5));
      rest[0]=rec[94]; rest[1]='\0';
      strsqz(chainn,strfld(buff,rec+95,1,1));
      color=str2i(strfld(buff,rec+96,1,4));
      q=str2f(strfld(buff,rec+100,1,9));
      biso=str2f(strfld(buff,rec+109,1,9));
      strsqz(pdbr,strfld(buff,rec+119,1,4));
      strsqz(buff,strfld(buff,rec+124,1,4));
      if (buff[0]<='9'&&buff[0]>='0') sprintf(atnm,"%s%c",buff+1,buff[0]);
      else strcpy(atnm,buff);
	
      atms[n]=a;
      
      /* add bonds... */

      for (i=0;i<6;i++) {
	if (nb[i]>0&&nb[i]<=n) {
	  bnd=add_bond(atms[nb[i]-1],a);
	  set_bt(bnd,MM2WnpBT(bt[i]));
	}
      }
	  
      /* modify atom properties... */

      set_atom_type(a,MM2WnpAT(tp));
      a->x[ATM_SOF]=q;                          /* may be moved to q in clean-up */
      a->x[ATM_BISO]=biso;
      color=MM2WnpColor(color);
      if (color) a->color=color;
      set_node_name((Node *)a,atnm);

      /* find out where this atom should go... */

      if (strlen(pdbr)==0) sprintf(pdbr,"%s",MM2PDB(rest));
      sprintf(resnam,"%s_%d",pdbr,resnr);
      sprintf(newres,"%s/%s",chainn,resnam);
      if (!streq(oldres,newres)) {
	if (strlen(chainn)>0) {
	  forObjs(chain,node,OBJ_MOL) if (streq(chain->name,chainn)) break;
	  if (!is_mol(chain)) chain=add_node(node,chainn,OBJ_MOL);
	  if (!is_mol(chain)) {
	    err_set(1,"failed to add chain %s to molecule %s.",chainn,node->name);
	    chain=node;
	    status=2;
	  }
	}else chain=node;
	forObjs(resid,chain,OBJ_MOL) if (streq(resid->name,resnam)) break;
	if (!is_mol(resid)) resid=add_node(chain,resnam,OBJ_MOL);
	if (!is_mol(resid)) {
	  err_set(1,"failed to add residue %s to chain %s.",resnam,chain->name);
	  resid=chain;
	  status=3;
	}
	strcpy(oldres,newres);
      }

      /* ...and link it. */

      a->up=resid;
      link_node((Node *)a);
    }else{
      err_set(1,"failed to add atom.");
      status=2;
    }
  }
  if (n<nat&&!status) {
    err_set(1,"premature end of file, expected %d atoms, found %d",nat,n);
  }

  if (!status) {            /* clean-up... */
    nnegq=0; nnegb=0;
    ForAtoms(a,node) {
      for (i=0;i<a->nbnd;i++) if (a->bnd[i]->f==a) asg_bondtype2(a->bnd[i]);
      if (a->x[ATM_BISO]<0) nnegb++;
      if (a->x[ATM_SOF]<0)  nnegq++;
    }
    if (nnegb>0||nnegq>0) {
      ForAtoms(a,node) {
	if (nnegq>0) {
	  a->q=a->x[ATM_SOF];
	  a->x[ATM_SOF]=1.0;
	}
	if (nnegb>0) {
	  if (nnegq==0) a->q=a->x[ATM_BISO];
	  a->x[ATM_BISO]=0;
	}
      }
    }
  }

  *Nat=nat;                  /* pass back number of atoms */
  *Atms=atms;               /* caller is expected to free this */
  return(status);
}
/*-------------------- read_mmodcnf --------------------*/

CnfNode *read_mmodcnf(UCpipe *in,int nat,Atom **Atms,Atom ***atms)
{
  /* 
     returns CnfNode for next conformation in Macromdel multi conformer file.
     *atms will be allocated and filled in with pointers to atoms in this frame.
     Caller is expected to free *atms, when no longer needed.
     Atoms is the atom pointer array in Macromodel order for the complete system.

     Returns NULL on errors and EOF.
  */

  CnfNode *c=(CnfNode *)NULL;
  char rec[RECSZ],buff[RECSZ];
  float *p;
  int i,k,n,compr;

  *atms=(Atom **)NULL;
  if (in&&nat>0) {
    if (UCpgets(rec,RECSZ,in)) {
      if (strrchr(rec,'\n')) *strrchr(rec,'\n')='\0';
      n=str2i(strfld(buff,rec+1,1,5));
      compr=n<0; n=(n<0?-n:n);
      if (n>0&&((compr&&n<=nat)||(!compr&&n==nat))) c=(CnfNode *)make_node(OBJ_CNF);
      else {
	if (n==0) err_set(1,"read_mmodcnf: 0 atoms in frame");
	else if (compr) err_set(1,"read_mmodcnf: too many atoms in compressed frame");
	else err_set(1,"read_mmodcnf: incorrect number of atoms in frame");
	c=(CnfNode *)NULL;
      }
      if (c) {
	set_node_name((Node *)c,strsqz(buff,strfld(buff,rec+9,1,9)));
	c->energy=str2f(strfld(buff,rec+22,1,8));
	c->energy/=4.186; /* Macromodel has energies in kJ/mol */
	c->n=n;
	if (n>0) {
	  c->x=(float *)malloc(3*c->n*sizeof(float));
	  *atms=(Atom **)malloc(c->n*sizeof(Atom *));
	  if (mvld(atms)&&c->x) {
	    if (compr) {
	      for(i=0;i<c->n;i++) (*atms)[i]=(Atom *)NULL;
	      for (p=c->x,i=0;i<n;i++) {
		if (UCpgets(rec,RECSZ,in)) {
		  if (i<c->n) {
		    k=str2i(strfld(buff,rec,1,5));
		    if (k>0&&k<=nat) {
		      k--; /* since our numbers are zero-based */
		      *p++=str2f(strfld(buff,rec+5,1,12));
		      *p++=str2f(strfld(buff,rec+17,1,12));
		      *p++=str2f(strfld(buff,rec+29,1,12));
		      (*atms)[i]=Atms[k];
		    }else{
		      err_set(1,"atom number out of range: %d",k);
		      (*atms)[i]=Atms[0];  /* to prevent disaster, if frame gets used...*/
		    }
		  }
		}else {
		  err_set(1,"premature EOF detected in mmod multi conformer file."); 
		  c=(CnfNode *)del_node((Node *)c);
		  break;
		}
	      }
	    }else{
	      for(i=0;i<c->n&&i<nat;i++) (*atms)[i]=Atms[i];
	      for(;i<c->n;i++) (*atms)[i]=Atms[i];  /* to prevent disaster, if n>nat */
	      for (p=c->x,i=0;i<c->n;i++) {
		if (UCpgets(rec,RECSZ,in)) {
		  *p++=str2f(strfld(buff,rec+52,1,12));
		  *p++=str2f(strfld(buff,rec+64,1,12));
		  *p++=str2f(strfld(buff,rec+76,1,12));
		}else {
		  err_set(1,"premature EOF detected in mmod multi conformer file."); 
		  c=(CnfNode *)del_node((Node *)c);
		  break;
		}
	      }
	    }
	  }else err_set(1,"failed to create memory for conformation.");
	}
      }else err_set(1,"failed to create conf. node.");
    }
  }else err_set(1,"read_mmodcnf: file not open.");
  if (mvld(c)) fixCnfOrder(MolRoot,c,*atms);
  return(c);
}

/*-------------------- write_mmod --------------------*/

int write_mmod(char *file,Node *node,int append)
{
  FILE *out;
  char rec[RECSZ];
  Atom *a,*b;
  Node *resid;
  int i,k,nat,nb,status;
  char resnum[6],one[2],chain[5],resnam[5];

  status=0;
  extfil(rec,file,".mmd",NO);
  if ((out=fopen(rec,append?"a":"w"))==NULL) {
    err_set(1,"failed to open %s for write.",rec);
    return(1);
  }

  nat=0; 
  ForAtoms(a,node) a->seqnr=++nat;

  k=localize_aromatic_bonds(node);
  if (k>0) err_set(1,"warning: failed to localize aromatic %d bonds.\n",k);

  fprintf(out,"%6d  %s\n",nat,node->name);
  resid=(Node *)NULL;
  ForAtoms(a,node) {
    fprintf(out,"%4d",MMNumType(a));
    for (nb=0,i=0;i<a->nbnd&&nb<6;i++) {
      b=neighbour(i,a);
      if (b->flg&(1<<1)) {
	fprintf(out,"%6d%2d",b->seqnr,MMBondOrder(a->bnd[i]));
	nb++;
      }
    }
    for (;nb<6;nb++) fprintf(out,"     0 0");
    for (i=0;i<3;i++) {
      if (a->nxyz>0) fprintf(out," %11.6f",a->x[i]);
      else fprintf(out," %11.6f",0.0);
    }
    if (a->up!=resid) {
      MMNames(a,node,resnum,one,chain,resnam);
      resid=a->up;
    }

    fprintf(out," %5s%1s%1s%4d%9.5f%9.5f %-4s %-4s",
      resnum,one,chain,Wnp2MMColor(a->color),a->q,a->q,resnam,MMpdbName(a,rec,sizeof(rec)));
    if (fprintf(out,"\n")<0) break;
  }

  if (a) err_set(1,"error writing macromodel file %s",file); 
  if (fclose(out)==EOF) err_set(1,"error writing macromodel file %s",file);
  return(status);
}

/*========================================================================*/

int read_mae(UCpipe *pipe,char *file,Node *node)
{
  int status=0,i,lev=0,nat=0,nbnd=0,have_atoms=FALSE,have_bonds=FALSE,have_ft=FALSE;
  maestream_t *in;
  char buf[MAXPATHLEN],**tags=NULL,**values=NULL;
  Atom **atoms=NULL,*a;
  Bond *bnd;
  maeBond_t *bonds=NULL;

  if (pipe) {
    in=maeStream_open(NULL,pipe,2*MAXTOKENSIZE,NULL,MAXTOKENSIZE);
  }else{
    in=maeStream_open(extfil(buf,file,".mae",NO),(UCpipe *)NULL,2*MAXTOKENSIZE,NULL,MAXTOKENSIZE);
  }
  if (!in) Return(1,"could not open input file");

  while (mae_token(in)) {
    if (streq(in->token,"f_m_ct")) {
      mae_token(in);
      if (streq(in->token,"{")) lev++;
      have_ft=TRUE;
      tags=getMAEtags(in,FALSE);
      if (tags) {
	values=getMAEvalues(in,tags);
	if (values) {
	  i=findMAEtag("s_m_title",tags);
	  if (i>=0) prtGraphics("%s \n",values[i]);
	  i=findMAEtag("s_m_entry_name",tags);
	  if (i>=0) prtGraphics("%s \n",values[i]);
	  i=findMAEtag("s_pdb_PDB_TITLE",tags);
	  if (i>=0) prtGraphics("%s \n",values[i]);
	  values=freeMAEvalues(values);
	}
	tags=freeMAEtags(tags);
      }
      while (!(have_atoms && have_bonds) && lev>0 && mae_token(in)) {
	if (strlen(in->token)>6&&strneq(in->token,"m_atom[",7)) {
	  if (have_atoms) {
	    mae_token(in);
	    if (streq(in->token,"{")) lev++;
	  }else{
	    nat=countMAEblock(in->token);
	    mae_token(in);
	    if (!streq(in->token,"{")) Return(1,"read_mae: atoms block syntax error");
	    atoms=readMAEatoms(in,nat,node);
	    if (nat>0 && !atoms) Return(1,"read_mae: failed to read atoms block");
	    have_atoms=TRUE;
	  }
	}else if (strlen(in->token)>6&&strneq(in->token,"m_bond[",7)) {
	  if (have_bonds) {
	    mae_token(in);
	    if (streq(in->token,"{")) lev++;
	  }else{
	    nbnd=countMAEblock(in->token);
	    mae_token(in);
	    if (!streq(in->token,"{")) Return(1,"read_mae: bonds block syntax error");
	    bonds=readMAEbonds(in,nbnd);
	    if (nbnd>0 && !bonds) Return(1,"read_mae: failed to read bonds block");
	    have_bonds=TRUE;
	  }
	}else{
	  if (streq(in->token,"{")) lev++;
	  if (streq(in->token,"}")) lev--;
	}
      }
      while (lev>0&&mae_token(in)) {
      	if (streq(in->token,"{")) lev++;
      	if (streq(in->token,"}")) lev--;
      }
      break;
    }
  }
  /* add bonds */
  if (atoms && bonds) {
    for (i=0;i<nbnd;i++) {
      if (bonds[i].f>0&&bonds[i].t>0&&bonds[i].f<=nat&&bonds[i].t<=nat&&atoms[bonds[i].f-1]&&atoms[bonds[i].t-1]) {
	bnd=add_bond(atoms[bonds[i].f-1],atoms[bonds[i].t-1]);
	set_bt(bnd,MM2WnpBT(bonds[i].bo));
      }
    }
    /* aromatic types, amide types... */
    ForAtoms(a,node) {
      for (i=0;i<a->nbnd;i++) {
	if (a->bnd[i]->f==a) asg_bondtype(a->bnd[i]);
      }
    }
  }
  /* Epilogue */
  if (lev!=0) {
    err_set(1,"read_mae: detected parenthesis mismatch in maestro file");
  }
  if (!have_ft) {
    err_set(1,"read_mae: did not find full ct section in file");
    status=-1;
  }else{
    if (!have_atoms) {
      err_set(1,"read_mae: did not find atoms block");
      status=-1;
    }else if (!have_bonds) {
      err_set(1,"read_mae: did not find bonds block");
    }
  }
  if (atoms) free(atoms);
  if (bonds) free(bonds);
  maeStream_close(in);
  return(status);
}  

static Atom **readMAEatoms(maestream_t *in,int nat,Node *node)
{
  Node *ResLib,*c,*r;
  Lib  *reslib;
  Atomdef *default_type;
  Atom **atoms=NULL,*a;
  char **tags=NULL,**V=NULL,buff[MAXTOKENSIZE];
  int i,n,status=0;
  static int immt,rx,ry,rz,resnu,insc,mmrest,chain,q,pdbresn,pdbatname,atnum,charge,atname,
    sof,biso;
  fts_ FTS[] = {
    {&immt, "i_m_mmod_type"},
    {&rx, "r_m_x_coord"},
    {&ry, "r_m_y_coord"},
    {&rz, "r_m_z_coord"},
    {&resnu,"i_m_residue_number"},
    {&insc,"s_m_insertion_code"},
    {&mmrest,"s_m_mmod_res"},
    {&chain,"s_m_chain_name"},
    {&q,"r_m_charge1"},
    {&pdbresn,"s_m_pdb_residue_name"},
    {&pdbatname,"s_m_pdb_atom_name"},
    {&atnum,"i_m_atomic_number"},
    {&charge,"i_m_formal_charge"},
    {&atname,"s_m_atom_name"},
    {&sof,"r_m_pdb_occupancy"},
    {&biso,"r_m_pdb_tfactor"},
    {NULL,NULL}
  };

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    return(NULL);
  }
  reslib=(Lib *)ResLib;
  default_type=find_atomdef_name((Node *)reslib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }
  
  if (nat>0) {
    atoms=malloc(nat*sizeof(Atom *));
    if (atoms) {
      for (i=0;i<nat;i++) atoms[i]=NULL;
    }else{
      err_set(1,"readMAEatoms: failed to allocate atom pointer list");
    }
  }

  if (atoms) {
    tags=getMAEtags(in,TRUE);
    if (tags) {
      for (i=0;FTS[i].idx;i++) *FTS[i].idx=findMAEtag(FTS[i].mnem,tags);
    }
  }else nat=0;
    
  for (n=0;n<nat;n++) {
    V=getMAEvalues(in,tags);
    if (V) {
      a=(Atom *)make_node(OBJ_ATOM);
      if (is_atom(a)) {
	/* generate chain name, create chain if needed */
	if (chain>=0) {
	  strsqz(buff,V[chain]);
	  if (strlen(buff)>0) {
	    forObjs(c,node,OBJ_MOL) if (streq(c->name,buff)) break;
	    if (!is_mol(c)) c=add_node(node,buff,OBJ_MOL);
	    if (!is_mol(c)) {
	      err_set(1,"readMAEatoms: failed to add chain %s to molecule %s.",buff,node->name);
	      c=node;
	      status=2;
	    }
	  }else c=node;
	}else c=node;
	/* generate residue name, create residue if needed */
	r=NULL;
	*buff='\0';
	if (pdbresn>=0) {
	  strsqz(buff,V[pdbresn]);
	}
	if (strlen(buff)==0) {
	  if (mmrest>=0) {
	    strsqz(V[mmrest],V[mmrest]);
	    if (strlen(V[mmrest])==1) {
	      strcpy(buff,MM2PDB(V[mmrest]));
	    }
	  }
	}
	if (resnu>=0&&!streq(V[resnu],"<>")) {
	  i=str2i(strsqz(V[resnu],V[resnu]));
	  sprintf(buff+strlen(buff),"_%d",i);
	}
	if (insc>=0&&!streq(V[insc],"<>")) {
	  sprintf(buff+strlen(buff),"%s",strsqz(V[insc],V[insc]));
	}
	if (strlen(buff)>0) {
	  if (n>0) {
	    if (streq(atoms[n-1]->up->name,buff) && streq(atoms[n-1]->up->up->name,c->name)) r=atoms[n-1]->up;
	  }
	  if (!r) {
	    forObjs(r,c,OBJ_MOL) if (streq(r->name,buff)) break;
	    if (!is_mol(r)) r=add_node(c,buff,OBJ_MOL); 
	    if (!is_mol(r)) {
	      err_set(1,"readMAEatoms: failed to add residue %s to chain %s.",buff,c->name);
	      r=c;
	      status=2;
	    }
	  }
	}else r=c;
	/* insert new atom into molecular data tree */
	a->up=r;
	link_node((Node *)a);
	/* set atom properties */
	set_node_name((Node *)a,"");
	if (strlen(a->name)==0 && pdbatname>=0&& !streq(V[pdbatname],"<>")) { /* precedence pdbatname over atname */
	  strsqz(V[pdbatname],V[pdbatname]);
	  set_node_name((Node *)a,V[pdbatname]);
	}
	if (strlen(a->name)==0 && atname>=0 && !streq(V[atname],"<>")) {
	  strsqz(V[atname],V[atname]);
	  set_node_name((Node *)a,V[atname]);
	}
	set_atom_type(a,immt>=0?MM2WnpAT(str2i(V[immt])):default_type);
	if ((!a->atp||(a->atp&&streq(a->atp->name,"unk")))&&atnum>=0) {
	  i=str2i(strsqz(V[atnum],V[atnum]));
	  if (ptab_chs(i)) set_atom_type(a,guess_atom_type(ptab_chs(i)));
	}
	if (charge>=0&&!streq(V[charge],"<>")) a->chrg=str2i(V[charge]);
	if (q>=0&&!streq(V[q],"<>")) a->q=str2f(V[q]);
	a->x[0]=rx>=0&&!streq(V[rx],"<>")?str2f(V[rx]):0.0;
	a->x[1]=ry>=0&&!streq(V[ry],"<>")?str2f(V[ry]):0.0;
	a->x[2]=rz>=0&&!streq(V[rz],"<>")?str2f(V[rz]):0.0;
	a->nxyz=1;
	if (sof>=0&&!streq(V[sof],"<>")) a->x[ATM_SOF]=str2f(V[sof]);
	if (biso>=0&&!streq(V[biso],"<>")) a->x[ATM_BISO]=str2f(V[biso]);
	atoms[n]=a;
      }
      freeMAEvalues(V);
    }
  }
  if (tags) freeMAEtags(tags);

  while (mae_token(in)) {
      if (streq(in->token,"}")) break;
  }
  if (status) {
    if (atoms) {
      free(atoms);
      atoms=NULL;
    }
  }
  return atoms;
}

static maeBond_t *readMAEbonds(maestream_t *in,int nbnd)
{
  maeBond_t *bonds=NULL;
  char **tags=NULL,**V;
  int i,n;
  static int order,from,to;
  fts_ FTS[] = {
    {&from, "i_m_from"},
    {&to, "i_m_to"},
    {&order, "i_m_order"},
    {NULL,NULL}
  };

  if (nbnd>0) {
    bonds=malloc(nbnd*sizeof(maeBond_t));
    if (!bonds) {
      err_set(1,"readMAEbonds: failed to allocate mempry for bond table");
      nbnd=0;
    }
  }

  if (bonds) {
    tags=getMAEtags(in,TRUE);
    if (tags) {
      for (i=0;FTS[i].idx;i++) *FTS[i].idx=findMAEtag(FTS[i].mnem,tags);
    }
  }else nbnd=0;

  for (n=0;n<nbnd;n++) {
    V=getMAEvalues(in,tags);
    if (V) {
      bonds[n].f=from>=0?str2i(V[from]):-1;
      bonds[n].t=to>=0?str2i(V[to]):-1;
      bonds[n].bo=order>=0?str2i(V[order]):-1;
      freeMAEvalues(V);
    }else{
      bonds[n].f=-1;
    }
  }
  while (mae_token(in)) {
    if (streq(in->token,"}")) break;
  }
  if (tags) freeMAEtags(tags);
  return(bonds);
}

static char **getMAEtags(maestream_t *in,int seqnr)
{
  char **tags;
  int n=0,m=50;

  tags=malloc(m*sizeof(char *));
  if (tags) {
    if (seqnr) {
      tags[0]=strdup("i_dummy_seqnr");
      n=1;
    }
    while (mae_token(in)) {
      if (n>=m) {
	m+=50;
	tags=realloc(tags,m*sizeof(char *));
	if (!tags) break;
      }
      if (streq(":::",in->token)) {
	tags[n]=NULL;
	break;
      }else{
	tags[n]=strdup(in->token);
      }
      n++;
    }
  }
  return(tags);
}

static int findMAEtag(char *tag,char **tags)
{
  int i,n=-1;
  for (i=0;tags[i]&&n<0;i++) {
    if (streq(tag,tags[i])) n=i;
  }
  return(n);
}

static char **getMAEvalues(maestream_t *in,char **tags)
{
  char **values=NULL;
  int i,n=0;

  for (n=0;;n++) if (!tags[n]) break;
  if (n>0) {
    values=malloc(((n+1)*sizeof(char *)));
    if (values) {
      for (i=0;i<n;i++) {
	if (mae_token(in)) {
	  values[i]=strdup(in->token);
	}else{
	  values[i]=strdup("");
	}
      }
      values[n]=NULL;
    }
  }
  return(values);
}

static char **freeMAEtags(char **tags)
{
  return(freeMAEvalues(tags));
}

static char **freeMAEvalues(char **values)
{
  int i;

  if (values) {
    for (i=0;values[i];i++) free(values[i]);
    free(values);
    values=(char **)NULL;
  }
  return(values);
}
/*========================================================================*/

static int countMAEblock(char *tag)
{
  int n;
  char *start,*end;
  start=strrchr(tag,'[');
  end=strrchr(tag,']');
  if (start&&end&&start<end) {
    *end='\0';
    n=str2i(start+1);
    *end=']';
  }else n=0;
  return(n);
}

/*-------------------- MM2WnpBT --------------------*/

static int MM2WnpBT(int bt)
{
  int r;

  switch (bt) {
  case 0:
    r=BT_DUMMY;
    break;
  case 1:
    r=BT_SINGLE;
    break;
  case 2:
    r=BT_DOUBLE;
    break;
  case 3:
    r=BT_TRIPLE;
    break;
  case 4:
    r=BT_UNKNOWN;
    break;
  default:
    r=BT_DUMMY;
    break;
  }
  return(r);
}
/*-------------------- MM2WnpAT --------------------*/

static Atomdef *MM2WnpAT(int tp)
{
  /* return pointer to atom type definition for MacModel type tp */

  Node *ResLib;
  Atomdef *t;
  int l;

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  ForAtomdefs(t,ResLib) {
    if (t->numtp==tp) {
      l=strlen(t->name);
      if (l>2&&streq(".m",t->name+l-2)) break;  /* .m is s kludge. Better: group Atomdefs in library by FF */
    }
  }
  if (!is_atomdef(t)) {
    t=find_atomdef_name((Node *)ResLib,"unk");
  }

  return(t);
}
/*-------------------- MM2WnpColor --------------------*/

static int MM2WnpColor(int c)
{
  /* should use close match... */

  int colmap[] = {
    0,15,7,4,4,13,6,12,6,2,2,
    2,3,3,9,1,1,8,5,5,5,7 };

  int r;

  r=(c<0||c>sizeof(colmap)-1 ? 7 : colmap[c]);
  return(r);
}

/*-------------------- Wnp2MMColor --------------------*/

static int Wnp2MMColor(int c)
{
  /* should use close match... */

  int colmap[] = {
    0,16,10,13,4,19,8,21,17,14,14,14,7,5,9,1};

  int r;

  r=(c<0||c>sizeof(colmap)-1 ? 0 : colmap[c]);
  return(r);
}

/*-------------------- MM2PDB --------------------*/

static char* MM2PDB(char *one)
{
  /* convert from MMod single letter to "pdb" code */

  int i,m;

  m=sizeof(MM_CNV)/sizeof(char *);
  for (i=0;i<m;i++) if (streq(MM_CNV[i],one)) break;
  return(PDB_CNV[i]);
}
/*-------------------- MMNumType --------------------*/

static int MMNumType(Atom *a)
{
  int l,r=0,z;

  if (is_atom(a)) {
    if (mvld(a->atp)) {
      l=strlen(a->atp->name);
      if (l>2&&streq(".m",a->atp->name+l-2)) r=a->atp->numtp;
      else {
	z=a->atp->z;
	switch (z) {
	case  1: r=48; break;
	case  5: r=55; break;
	case  6: r=14; break;
	case  7: r=40; break;
	case  8: r=23; break;
	case  9: r=56; break;
	case 14: r=60; break;
	case 15: r=53; break;
	case 16: r=52; break;
	case 17: r=57; break;
	case 35: r=58; break;
	case 53: r=59; break;
	case  0: if (streq(a->atp->chs,"Lp"))r=63;else r=61; break;
	default: r=62; break;
	}
      }
    }else r=62;           /* Z0, special atom, to be defined... */
  }
  return(r);
}
/*-------------------- MMNames --------------------*/

static void MMNames(Atom *a,Node *node,char *resnum,char *one,char *chain,char *resnam)
{
  Node *n;
  int k;
  char B[4096];

  /* the other names need only be generated, if node has changed... */
  n=a->up;
  if (n==node) {
    sprintf(resnum,"     ");
    sprintf(one," ");
    sprintf(chain," ");
    sprintf(resnam,"    ");
  }else{
    if (n->up==node) sprintf(B," ");
    else sprintf(B,"%1s",n->up->name);
    B[1]='\0'; strcpy(chain,B);
    if (strrchr(n->name,'_')) sprintf(B,"%5d",atoi(strrchr(n->name,'_')+1));
    else sprintf(B,"     ");
    B[5]='\0'; strcpy(resnum,B);
    sprintf(B,"%4s",n->name);
    if (strrchr(B,'_')) *strrchr(B,'_')='\0';
    B[4]='\0'; strcpy(resnam,B);
    for (k=0;strlen(PDB_CNV[k])>0;k++) if (streq(resnam,PDB_CNV[k])) break;
    if (strlen(PDB_CNV[k])>0) sprintf(one,"%s",MM_CNV[k]);
    else sprintf(one," ");
  }
}
/*-------------------- MMpdbName --------------------*/

static char *MMpdbName(Atom *a,char *buff,int size)
{
  /* convert atom name to PDB type, if possible */
  /* buff is a buffer of length >= size         */

  int k;

  if (a->name) {
    if (a->atp&&strlen(a->atp->chs)==1&&strlen(a->name)<4) {
      sprintf(buff," %s",a->name);
    }else{
      k=strlen(a->name); if (k>size-1) k=size-1;
      strncpy(buff,a->name,k);
      buff[k]='\0';
    }
  }else{
    buff[0]='\0';
  }
  if (strlen(buff)>4) buff[4]='\0';   /* truncate to the max. allowed by PDB */
  return(buff);
}

/*-------------------- MMBondOrder --------------------*/

static int MMBondOrder(Bond *b)
{
  int r;

  if (mvld(b)) {
    r=b->bo;
    if (r>3) r=0;   /* have to deal with aromatic bonds here... */
  }else r=0;
  return(r);
}

/*-------------------- fixCnfOrder --------------------*/

static void fixCnfOrder(Node *root,CnfNode *c,Atom **atms)
{
  int i,j,k;
  float *x;

  /*
    Confa needs things in walk-oder, so sort template atom pointers
    and coordinates...
  */

  if (c&&c->n>0&&atms) {
    for (i=0;i<c->n;i++) atms[i]->seqnr=i;
    walk_order(root,c->n,atms,(1<<0));
    if (mvld(c->x)) {
      x=(float *)malloc(3*c->n*sizeof(float));
      if (mvld(x)) {
	for (i=0;i<c->n;i++) {
	  k=atms[i]->seqnr;
	  for (j=0;j<3;j++) x[3*i+j]=c->x[3*k+j];
	}
	free(c->x);
	c->x=x;
      }
    }
  }
}
