#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include <sys/param.h>
#include "errhand.h"

extern int graphix;

static int altobj_key;
static int mouse_right;
static int map[64];
static int objtra[2][4];

#define XYROT 0
#define ZROT  1
#define XYTRA 2
#define ZTRA  3

static char *func_mnem =
"none Pick XYTrans ZTrans XYRot ZRot Scale ViewDist BondRot SlabWidth "
"SlabPos AtomInfo StereoSep SetPivot SetAltPivot SelectAlt SelectMAlt "
"LightXYTrans LightZTrans LightRot LightIntensity LightAttenuation ";

typedef struct MouseFunction {
  int opnr;
  char *opcode;
} MouseFunction;

/* Default mouse mapping for 3-button mice: */

static MouseFunction Mouse3[WNP_DEF_MOUSE_FUNC] = {
  {WNP_MOUSE_TRACK        ,"" },
  {WNP_MOUSE_PICK         ,"R" },
  {WNP_MOUSE_XYTRANS      ,"M" },
  {WNP_MOUSE_ZTRANS       ,"MR" },
  {WNP_MOUSE_XYROT        ,"L" },
  {WNP_MOUSE_ZROT         ,"LR" },
  {WNP_MOUSE_SCALE        ,"LM" },
  {WNP_MOUSE_VIEWDIST     ,"CLMR" },
  {WNP_MOUSE_AXROT        ,"CLR" },
  {WNP_MOUSE_SLABWIDTH    ,"CLM" },
  {WNP_MOUSE_SLABPOS      ,"CMR" },
  {WNP_MOUSE_USAP         ,"SR" },
  {WNP_MOUSE_STEREOSEP    ,"CASM" },
  {WNP_MOUSE_SETPIVOT     ,"ASR" },
  {WNP_MOUSE_SETALTPIVOT  ,"ASM" },
  {WNP_MOUSE_SELALT       ,"CAR" },
  {WNP_MOUSE_SELMALT      ,"CAM" },
  {WNP_MOUSE_LIGHTXY      ,"CASL" },
  {WNP_MOUSE_LIGHTZ       ,"CASRM" },
  {WNP_MOUSE_LIGHTROT     ,"CASLR" },
  {WNP_MOUSE_LIGHTINT     ,"CASLM" },
  {WNP_MOUSE_LIGHTATTEN   ,"CASLMR"}
};

/* Default mouse mapping for 2-button mice: */

static MouseFunction Mouse2[WNP_DEF_MOUSE_FUNC] = {
  {WNP_MOUSE_TRACK        ,"" },
  {WNP_MOUSE_PICK         ,"R" },
  {WNP_MOUSE_XYTRANS      ,"A" },
  {WNP_MOUSE_ZTRANS       ,"AR" },
  {WNP_MOUSE_XYROT        ,"L" },
  {WNP_MOUSE_ZROT         ,"LR" },
  {WNP_MOUSE_SCALE        ,"AL" },
  {WNP_MOUSE_VIEWDIST     ,"CALR" },
  {WNP_MOUSE_AXROT        ,"CLR" },
  {WNP_MOUSE_SLABWIDTH    ,"CAL" },
  {WNP_MOUSE_SLABPOS      ,"CAR" },
  {WNP_MOUSE_USAP         ,"SR" },
  {WNP_MOUSE_STEREOSEP    ,"CSLR" },
  {WNP_MOUSE_SETPIVOT     ,"ASR" },
  {WNP_MOUSE_SETALTPIVOT  ,"ASL" },
  {WNP_MOUSE_SELALT       ,"CSL" },
  {WNP_MOUSE_SELMALT      ,"CSR" },
  {WNP_MOUSE_LIGHTXY      ,"CASL" },
  {WNP_MOUSE_LIGHTZ       ,"CAS"},
  {WNP_MOUSE_LIGHTROT     ,"CASLR" },
  {WNP_MOUSE_LIGHTINT     ,"CASR"},
  {WNP_MOUSE_LIGHTATTEN   ,"CASLRM"}
};

static char keys_string_buffer[10];

static void send_mouse_map(void);
static void set_mouse_map(MouseFunction *mmap);

/*-------------------- set_altobj_key --------------------*/

void set_altobj_key(char *key)
{
  int new,i;

  if (!key) return;
  new=altobj_key;
  if (streq(key,"altkey")) new=WNP_MOD_ALT;
  else if (streq(key,"shiftkey")) new=WNP_MOD_SHIFT;

  /* remap the obj-tra functions and send new map */

  if (new!=altobj_key) {
    for (i=0;i<4;i++ ) {
      map[altobj_key+objtra[0][i]]=WNP_MOUSE_TRACK;
      map[objtra[0][i]]=WNP_MOUSE_TRACK;
    }
    altobj_key=new;
    for (i=0;i<4;i++ ) {
      map[altobj_key|objtra[0][i]]=objtra[1][i];
      map[objtra[0][i]]=objtra[1][i];
    }
    send_bheader(graphix,"MMAP",(64+1)*sizeof(int));
    send_bint(graphix,altobj_key);
    for (i=0;i<64;i++) {
      send_bint(graphix,map[i]);
    }
    trflush(graphix);
  }
}

/*-------------------- init_mousemap --------------------*/

void init_mousemap(void)
{
  int i;
  
  altobj_key=WNP_MOD_ALT;
  mouse_right=1;

  objtra[1][XYROT]=WNP_MOUSE_XYROT;
  objtra[1][ZROT]=WNP_MOUSE_ZROT;
  objtra[1][XYTRA]=WNP_MOUSE_XYTRANS;
  objtra[1][ZTRA]=WNP_MOUSE_ZTRANS;

  for (i=0;i<64;i++) map[i]=WNP_MOUSE_TRACK;
  
  set_mouse("right");
  set_mouse_map(Mouse3);
  send_mouse_map();
}
  
/*-------------------- get_mouse_keys --------------------*/

char *get_mouse_keys(int func)
{
  int i;

 *keys_string_buffer='\0';

  for (i=0;i<64;i++) if (func==map[i]) break;
  if (i<64) {
    func=i;
    if (func&WNP_MOD_ALT) 
      sprintf(keys_string_buffer+strlen(keys_string_buffer),"%s","A");
    if (func&WNP_MOD_CTRL) 
      sprintf(keys_string_buffer+strlen(keys_string_buffer),"%s","C");
    if (func&WNP_MOD_SHIFT) 
      sprintf(keys_string_buffer+strlen(keys_string_buffer),"%s","S");
    if (func&WNP_MOD_LEFT) 
      sprintf(keys_string_buffer+strlen(keys_string_buffer),"%s",(mouse_right?"L":"R"));
    if (func&WNP_MOD_MIDDLE) 
      sprintf(keys_string_buffer+strlen(keys_string_buffer),"%s","M");
    if (func&WNP_MOD_RIGHT) 
      sprintf(keys_string_buffer+strlen(keys_string_buffer),"%s",(mouse_right?"R":"L"));
    return(keys_string_buffer);
  }else{
    return((char *)NULL);
  }
}

/*-------------------- set_mouse_function --------------------*/

void set_mouse_function(int func,char *keys)
{
  int i,n;

  /* NOTE: "keys" mapping depends on state of "mouse_right" */

  n=0;
  for (i=0;i<strlen(keys);i++) {
    if (keys[i]=='A'||keys[i]=='a') n|=WNP_MOD_ALT;
    else if (keys[i]=='S'||keys[i]=='s') n|=WNP_MOD_SHIFT;
    else if (keys[i]=='C'||keys[i]=='c') n|=WNP_MOD_CTRL;
    else if (keys[i]=='L'||keys[i]=='l') n|=(mouse_right?WNP_MOD_LEFT:WNP_MOD_RIGHT);
    else if (keys[i]=='M'||keys[i]=='m') n|=WNP_MOD_MIDDLE;
    else if (keys[i]=='R'||keys[i]=='r') n|=(mouse_right?WNP_MOD_RIGHT:WNP_MOD_LEFT);
    else {
      err_set(1,"bad mouse key string: %s",keys);
      break;
    }
  }
  if (!keys[i]&&n>0) {
    for (i=0;i<64;i++) if (map[i]==func) map[i]=WNP_MOUSE_TRACK; /* undo old mapping */
    for (i=0;i<4;i++) if (func==objtra[1][i]) break;
    if (i<4) {
      n &= ~altobj_key;
      objtra[0][i]=n;
      map[n]=func;
      map[n|altobj_key]=func;
    }else{
      map[n]=func;
    }
    send_mouse_map();
  }
}
/*-------------------- set_mouse --------------------*/

void set_mouse(char *left_or_right)
{
  int n;

  if (streq(left_or_right,"left")) n=0;
  else if (streq(left_or_right,"right")) n=1;
  else if (streq(left_or_right,"3Buttons")) n=-3;
  else if (streq(left_or_right,"2Buttons")) n=-2;
  else n=-1;

  if (n>=0) {
    send_bheader(graphix,"SMLR",sizeof(int));
    send_int(graphix,n);
    mouse_right=n;
  }else if (n==-3) {
    set_mouse_map(Mouse3);
    send_mouse_map();
  }else if (n==-2) {
    set_mouse_map(Mouse2);
    send_mouse_map();
  }else{
    err_set(1,"bad argument in set_mouse: %s",left_or_right);
  }
}

/*-------------------- show_mouse_map --------------------*/

void show_mouse_map(void)
{
  int i;
  Parse p;
  char buf[40],*key,*alt;

  p.sep=""; p.buf=func_mnem; p.pos=0;
  p.tok=buf; p.ltok=sizeof(buf);

  alt=(altobj_key==WNP_MOD_ALT?"A":"S");
  get_token(&p);   /* throw away first token */
  for (i=1;i<WNP_DEF_MOUSE_FUNC;i++) {
    get_token(&p);
    key=get_mouse_keys(i);
    if (i==WNP_MOUSE_XYROT||i==WNP_MOUSE_ZROT||i==WNP_MOUSE_XYTRANS||i==WNP_MOUSE_ZTRANS) {
      prtGraphics("%25s  %-6s (selected object: %s%s)\n",buf,key,alt,key);
    }else{
      prtGraphics("%25s  %s\n",buf,key);
    }
  }
}

/*-------------------- mouse_func_mnems --------------------*/

char *mouse_func_mnems(void)
{
  return(func_mnem);
}
/*-------------------- set_mouse_map --------------------*/

static void set_mouse_map(MouseFunction *mmap)
{
  int i,k,n,func;
  char *keys;
  
  for (i=0;i<64;i++) map[i]=WNP_MOUSE_TRACK;

  /* NOTE: mmap tables are for a "right" mouse, i.e. buttons
     (1,2,3) map to (R,M,L)
  */

  for (k=0;k<WNP_DEF_MOUSE_FUNC;k++) {
    keys=mmap[k].opcode; func=mmap[k].opnr;
    for (i=0,n=0;i<strlen(keys);i++) {
      if (keys[i]=='A'||keys[i]=='a') n|=WNP_MOD_ALT;
      else if (keys[i]=='S'||keys[i]=='s') n|=WNP_MOD_SHIFT;
      else if (keys[i]=='C'||keys[i]=='c') n|=WNP_MOD_CTRL;
      else if (keys[i]=='L'||keys[i]=='l') n|=WNP_MOD_LEFT;
      else if (keys[i]=='M'||keys[i]=='m') n|=WNP_MOD_MIDDLE;
      else if (keys[i]=='R'||keys[i]=='r') n|=WNP_MOD_RIGHT;
      else {
	err_set(1,"bad mouse key string: %s",keys);
	break;
      }
    }
    if (!keys[i]&&n>0) {
      for (i=0;i<4;i++) if (func==objtra[1][i]) break;
      if (i<4) {
	n &= ~altobj_key;
	objtra[0][i]=n;
	map[n]=func;
	map[n|altobj_key]=func;
      }else{
	map[n]=func;
      }
    }
  }
}
/*-------------------- send_mouse_map --------------------*/

static void send_mouse_map()
{
  int i;

  send_bheader(graphix,"MMAP",(64+1)*sizeof(int));
  send_bint(graphix,altobj_key);
  for (i=0;i<64;i++) {
    send_bint(graphix,map[i]);
  }
  trflush(graphix);
}
