/* namenptr.c */

#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"

#define MTS 4096
#define TRUE 1
#define FALSE 0

extern int graphix;
extern Mol *MolRoot,*DefaultNode;

/* Search node with given name and return pointer (NULL if not found) */

Node *name_to_nodeptr(char *name,int (*compare)(const char *a,const char *b))
{
  Node *node,*root;

  if (!(node=pick_to_node(name))) {
    root=(Node *)(DefaultNode ? DefaultNode : MolRoot);
    if (name) if (name[0]=='/') root=(Node *)MolRoot;
    node=node_in_root(name,root,compare);
  }
  return(node);
}
/*-------------------- node_in_ctxt --------------------*/

Node *node_in_ctxt(char *name,Node *ctxt,
		   int (*compare)(const char *reg,const char *wild))
{
  /* Return pointer to node with given name in context ctxt,
     (NULL if node not found). 

  Example: name = CB, name(ctxt) = /crn/ala_17/ca ---> /crn/ala_17/cb,
           name = pro_19/cd , name(ctxt) = /crn/ala_17/ca --> /crn/pro_19/cd
  */
  
  int nlev;
  Node *root,*node;
  char *rest;
  
  if (!name) {
    node=(Node *)NULL;
  }else if (name[0]=='.' || name[0]=='/' || !ctxt) {
    node=name_to_nodeptr(name,compare);
  }else{
    if (!(node=pick_to_node(name))) {
      rest=name;
      nlev=(strlen(name) == 0 ? 0 : 1);
      while ((rest=strchr(++rest,'/'))) nlev++;
      for (root=ctxt;nlev>0&&mvld(root);nlev--) root=root->up;
      node=node_in_root(name,root,compare);
    }
  }
  return(node);
}
/*-------------------- node_in_root --------------------*/

Node *node_in_root(char *name,Node *root,\
		   int (*compare)(const char *reg,const char *wild))
{
  /* Return pointer to node with given name in subtree with root,
     (NULL if node not found). 
  */

  
  Node *now,*next;
  Parse p;
  int save;
  char tkn[MTS];
  
  p.buf=name;
  p.pos=0;
  p.sep="/";
  p.tok=tkn;
  p.mts=MTS;
  
  if (!name || !root) return((Node *)NULL);
  now=root;
  if (name[0]=='/') {
    p.pos++;
  }else{
    save=p.pos;
    if (get_token(&p)) {
      if (streq(tkn,".")) {
	save=p.pos;
      }else{
	while(compare(tkn,"..")==0) {
	  now=now->up;
	  if (!now) return((Node *)NULL);
	  save=p.pos;
	  if (!get_token(&p)) break;
	}
      }
    }
    p.pos=save;
  }
    
  while (get_token(&p)) {
    if (!p.ltok) return((Node *)NULL);
    if (is_atom(now)) return((Node *)NULL);
    forObjs(next,now,OBJ_MOL|OBJ_ATOM) {
      if (next->name) {
	if (!compare(tkn,next->name)) break;
      }
    }
    if (!next) return((Node *)NULL);
    now=next;
  }

  return(now);
}  

/*-------------------- nodeptr_to_name --------------------*/

/* Return full node name of node pointed to by "n" */

char *nodeptr_to_name(char *name,Node *n)
{
  if (!n) return((char *)NULL);
  if (n->up) {
    nodeptr_to_name(name,n->up);
    if (strcmp(name,"/")!=0) strcat(name,"/");
    if (n->name)strcat(name,n->name);
  }else{
    strcpy(name,"/");
  }
  return(name);
}

/*-------------------- nodeptr_to_name_ctxt --------------------*/

/* Return nodename of node pointed to by n in context ctxt */

char *nodeptr_to_name_ctxt(char *name,Node *n,Node *ctxt)
{
  Node *x;
  int nd,nc;

  if (!n) return((char *)NULL);
  if (!ctxt) {
    name=nodeptr_to_name(name,n);
  }else{
    for(nd=0,x=n;x;x=x->up) nd++;
    for(nc=0,x=ctxt;x;x=x->up) nc++;
    if (nd==nc) {
      for (x=n->up,ctxt=ctxt->up;x!=ctxt;x=x->up) ctxt=ctxt->up;
      if (ctxt && ctxt->up) {
	name=nodeptr_to_name(name,ctxt);
	if (name) {
	  nc=strlen(name)+1;
	  name=nodeptr_to_name(name,n);
	  if (name) strcpy(name,name+nc);
	}
      }else{
	name=nodeptr_to_name(name,n);
      }
    }else{
      name=nodeptr_to_name(name,n);
    }
  }
  return(name);
}
  
/*-------------------- open_nmatch --------------------*/

Nmatch *open_nmatch(char *name)
{
  int save,loop;
  char tkn[MTS];
  Parse p;
  Nmatch *s;
  Node *now;
  
  if (!name) return((Nmatch *)NULL);
  s=(Nmatch *)malloc(sizeof(Nmatch));
  if (!s) return(s);
  s->nlev=0;
  s->level=0;
  s->nlev=1;
  s->start[1]=0;
  s->nprim[1]=0;

  /* handle pick response */
  
  s->node=pick_to_node(name);
  if (s->node) {
    s->nlev=1;
  }else{

  /* handle wild-string */

    p.buf=name;p.pos=0;p.sep="/,-";p.tok=tkn;p.mts=MTS;
    
    now=(Node *)DefaultNode;
    if (!now || name[0]=='/') {
      now=(Node *)MolRoot;
      p.pos++;
    }else{
      save=p.pos;
      if (get_token(&p)) {
	if (strcmp(tkn,".")==0) {
	  save=p.pos;
	}else{
	  while(strcmp(tkn,"..")==0) {
	    now=now->up;
	    if (!now) return((Nmatch *)NULL);
	    save=p.pos;
	    if (!get_token(&p)) break;
	  }
	}
      }
      p.pos=save;
    }
    
    s->node=now;
    
    while (get_token(&p)) {
      if (s->nlev==MAX_NLEV) {
	prtGraphics("MAX_NLEV exceeded in open_nmatch.\n");
	close_nmatch(s);
	return((Nmatch *)NULL);
      }
      s->start[s->nlev]=s->level;
      s->nprim[s->nlev]=0;
      do {
	if (p.ltok==0 && p.term != '/') {
	  err_set(1,"Syntax error in atom-expression.");
	  close_nmatch(s);
	  return((Nmatch *)NULL);
	}
	switch (p.term) {
	case '/' :
	  if (s->level==MAX_NPRIM) {
	    err_set(1,"MAX_NPRIM exceeded in open_nmatch.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  if (p.ltok > 0) s->str1[s->level]=Strdup(tkn);
	  else s->str1[s->level]=Strdup("*");
	  s->type[s->level]=1;
	  s->nprim[s->nlev]++;
	  s->level++;
	  loop=FALSE;
	  break;
	case ',' :
	  if (s->level==MAX_NPRIM) {
	    err_set(1,"MAX_NPRIM exceeded in open_nmatch.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  s->str1[s->level]=Strdup(tkn);
	  s->type[s->level]=1;
	  s->nprim[s->nlev]++;
	  s->level++;
	  loop=TRUE;
	  if (!get_token(&p)) {
	    err_set(1,"Syntax error in atom-expression.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  break;
	case '-' :
	  if (s->level>MAX_NPRIM-2) {
	    err_set(1,"MAX_NPRIM exceeded in open_nmatch.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  s->str1[s->level]=Strdup(tkn);
	  s->type[s->level]=2;
	  if (!get_token(&p)) {
	    err_set(1,"Syntax error in atom-expression.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  if (p.term=='-') {
	    err_set(1,"Syntax error in atom-expression.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  s->str2[s->level]=Strdup(tkn);
	  s->nprim[s->nlev]++;
	  s->level++;
	  if (p.term==',') {
	    loop=TRUE;
	    if (!get_token(&p)) {
	      err_set(1,"Syntax error in atom-expression.");
	      close_nmatch(s);
	      return((Nmatch *)NULL);
	    }
	  }else{
	    loop=FALSE;
	  }
	  break;
	default:
	  if (s->level==MAX_NPRIM) {
	    err_set(1,"MAX_NPRIM exceeded in open_nmatch.");
	    close_nmatch(s);
	    return((Nmatch *)NULL);
	  }
	  s->str1[s->level]=Strdup(tkn);
	  s->type[s->level]=1;
	  s->nprim[s->nlev]++;
	  s->level++;
	  loop=FALSE;
	  break;
	}
      } while (loop);
      s->nlev++;
    }
    
    s->level=0;
  }
  return(s);
}

static int match_current_node(Node *node,Nmatch *s)
{
  int k,l,match;
  
  match=FALSE;
  if (!node || !s) return(match);
  for (k=0;k<s->nprim[s->level];k++) {
    l=s->start[s->level]+k;
    if (s->type[l]==2) {
      if (node==s->first[l]) s->active[l]=TRUE;
      if (s->active[l]) {
	match=TRUE;
	if (node==s->last[l]) s->active[l]=FALSE;
      }
    }
  }
  if (match) return(TRUE);
  for (k=0;k<s->nprim[s->level];k++) {
    l=s->start[s->level]+k;
    if (s->type[l]!=2) {
      if (wild_xmatch(node->name,s->str1[l])) return(TRUE);
    }
  }
  return(FALSE);
}

static void enter_new_level(Nmatch *s)
{
  int k,l;
  Node *x;
  
  s->level++;
  for (k=0;k<s->nprim[s->level];k++) {
    l=s->start[s->level]+k;	
    if (s->type[l]==2) {
      s->first[l]=(Node *)NULL;
      s->last[l]=(Node *)NULL;
      x=s->node->fst;
      for (x=s->node->fst;x;x=x->nxt) {
	if (wild_xmatch(x->name,s->str1[l])) break;
      }
      s->first[l]=x;
      if (x) {
	for (x=s->node->lst;x!=s->first[l]->prv;x=x->prv) {
	  if (wild_xmatch(x->name,s->str2[l])) break;
	}
	if (x!=s->first[l]->prv) {
	  s->last[l]=x;
	}else{
	  s->first[l]=(Node *)NULL;
	  s->last[l]=(Node *)NULL;
	}
      }else{
	s->last[l]=(Node *)NULL;
      }
      s->active[l]=FALSE;
    }
  }
  s->up=s->node;
  s->node=s->node->fst;
}

Node *next_nmatch(Nmatch *s)
{
  Node *match;
  
  if (!s) return((Node *)NULL);
  if (s->nlev==0) return((Node *)NULL);
  
  if (s->level==0) {
    if (s->nlev==1) {
      match=s->node;
      s->node=(Node *)NULL;
      return(match);
    }else{
      enter_new_level(s);
    }
  }
  
  while(s->level > 0) {
    if (s->node) {
      match=s->node;
      s->node=s->node->nxt;
      if (match_current_node(match,s)) {
	if (s->level==s->nlev-1) return(match);
	if (is_mol(match)) {
	  s->node=match;
	  enter_new_level(s);
	}
      }
    }else{
      s->level--;
      s->node=s->up->nxt;
      s->up=s->up->up;
    }
  }
  return((Node *)NULL);
}

void close_nmatch(Nmatch *s)
{
  int i,l,k;
  
  if (!s) return;
  for (i=1;i<s->nlev;i++) {
    for(k=0;k<s->nprim[i];k++) {
      l=s->start[i]+k;
      if (mvld(s->str1[l])) free(s->str1[l]);
      if (s->type[l]==2) {
	if (mvld(s->str2[l])) free(s->str2[l]);
      }
    }
  }
  free(s);
}
/*-------------------- MatchNodeName --------------------*/

Node *MatchNodeName(Node *a,char *name)
{
  /* Return pointer to node containing a, if name matches
     a->name, else return NULL.

     name may contain X-Plor wildcards, ',' is 'or', not case sensitive */

  Node *r=(Node *)NULL;
  char *p;
  
  if (!a||!name) return((Node *)NULL);
  if (strlen(name)==0) return(a);
  if (!a->name) return((Node *)NULL);
  for (p=name;*p;) {
    if (*p=='/') {
      r=MatchNodeName(a,p+1);
      if (r) {
	*p='\0';
	r=r->up;
	if (r) r=(mwild_xmatch(r->name,name) ? r : (Node *)NULL);
	*p='/';
      }
      break;
    }else if (*p=='\\') p++;
    if (*p) p++;
  }
  if (*p!='/') r=(mwild_xmatch(a->name,name) ? a : (Node *)NULL);
  return(r);
}
/*-------------------- pick_to_node --------------------*/

Node *pick_to_node(char *s)
{
/*   Convert pick string to node pointer   */

  char buff[1024];
  Node *r;
  Atom *x;
  cptr_t cross;
  int i,n;
  unsigned int b[8];

  r=(Node *)NULL;
  if (!s) return(r);
  if (strneq(s,"#PATM(",6)) {
    Strcpy(buff,s+6);
    if (strchr(buff,')')) {
      *strchr(buff,')')='\0';
      n=sscanf(buff,"%x %x %x %x %x %x %x %x",b,b+1,b+2,b+3,b+4,b+5,b+6,b+7);
      if (n>8) n=8;
      for (i=0;i<n;i++) cross.b[i]=b[i];
      for (;i<sizeof(cross);i++) cross.b[i]=0;
      ForAtoms (x,MolRoot) {
	if (!memcmp(&(x->cross),&cross,sizeof(cross))) {
	  r=(Node *)x;
	  break;
	}
      }
    }
  }
  return(r);
}
