#include "strutil.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "object.h"
#include "gcalc.h"
#include "transpor.h"
#include "errhand.h"
#include "nmr.h"

#define key_is(x) (streq(key->key,x))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

extern int graphix;
extern Node *MolRoot;
extern Dspobj *DspRoot;
extern Node *DefaultNode;

/*-------------------- Get_NMRStructures --------------------*/

Token *Get_NMRStructures(Mol *NmrRoot,char *prmt,char *prefix,int wait_flag)
{
  Token *key;
  Nodechain *c,m;
  Node *mol;
  Dspobj *proot;
  char *menu;

  c=make_node_chain();
  menu=(char *)NULL;

  if (!mvld(c)) {
    key=make_Token();
    key->status=TOK_FAIL;
    default_token((char *)NULL);
  }else{

    m.n=0; m.s=(Node *)NULL; m.e=m.s;
    forObjs(mol,(Node *)NmrRoot,OBJ_MOL) add_to_node_chain(mol,&m);
    if (m.n>0) {
      menu=Make_NodeNameList(prefix,&m);
      if (!mvld(menu)) menu=Strdup("done");
      if (mvld(menu)) {
	if (strlen(menu)==0) {
	  free(menu);
	  menu=Strdup("done");
	}
      }
      
      clear_node_chain(&m);
      
      proot=pick_root(assocobj((Node *)NmrRoot));
      key=Get_Mol_MW(prmt,";",menu,wait_flag);
      pick_root(proot);
      send_bheader(graphix,"ARMK",0);
      
      if (key->status==TOK_OK) {
	if (streq(key->text,"*")) wait_flag=FALSE;
	for (mol=(Node *)key->ptr;mvld(mol);mol=mol->up) {
	  if (mol->up==(Node *)NmrRoot) break;
	}
	if (mvld(mol)) add_to_node_chain(mol,c);
      }
      if (key->status==TOK_NOK) {
	forObjs(mol,(Node *)NmrRoot,OBJ_MOL) {
	  if (wild_imatch(mol->name,key->text))add_to_node_chain(mol,c);
	}
	if (c->n >0) key->status=TOK_OK;
      }
      
      if (key->status==TOK_OK) {
	key->ptr=(void *)c;
      }else{
	clear_node_chain(c);
	free_node_chain(c);
	key->ptr=(void *)NULL;
      }
    }else{
      prtGraphics("no structures in NMR set %s.\n",NmrRoot->name);
      key=make_Token();
      key->status=TOK_EOL;
      default_token((char *)NULL);
    }
  }
  free_menu(menu);
  return(key);
}
/*-------------------- new_nmrPeakList --------------------*/

nmrPeakList *new_nmrPeakList(void)
{
  nmrPeakList *r;

  r=(nmrPeakList *)malloc(sizeof(nmrPeakList));
  if (r) *r=EmptyNmrPeakList();
  return(r);
}
/*-------------------- release_nmrPeakList --------------------*/

nmrPeakList release_nmrPeakList(nmrPeakList L)
{
  int i;

  if (L.npeaks>0&&L.peak) {
    for (i=0;i<L.npeaks;i++) free_nmrPeak(L.peak[i],L.dim);
  }
  if (L.peak) free(L.peak);
  if (L.filename) free(L.filename);
  L=EmptyNmrPeakList();

  return(L);
}

/*-------------------- EmptyNmrPeakList --------------------*/

nmrPeakList EmptyNmrPeakList(void)
{
  nmrPeakList L;

  L.filename=(char *)NULL;
  L.dim=0;
  L.npeaks=0;
  L.peak=(nmrPeak **)NULL;

  return(L);
}

/*-------------------- new_nmrPeak --------------------*/

nmrPeak *new_nmrPeak(int dim)
{
  nmrPeak *r;
  int i;

  r=(nmrPeak *)malloc(sizeof(nmrPeak));
  if (r) {
    r->nr=0;
    r->w=1.0;
    r->pt=(float *)malloc(dim*sizeof(float));
    r->p2=(float *)malloc(dim*sizeof(float));
    r->p3=(int *)malloc(dim*sizeof(int));
    r->asg=(char **)malloc(dim*sizeof(char *));
    if (r->pt&&r->p2&&r->p3&&r->asg) {
      for (i=0;i<dim;i++) {
	r->pt[i]=0.0;
	r->p2[i]=0.0;
	r->p3[i]=0;
	r->asg[i]=(char *)NULL;
      }
    }else{
      if (r->pt) free(r->pt);
      if (r->p2) free(r->p2);
      if (r->p3) free(r->p3);
      if (r->asg) free(r->asg);
      free(r);
      r=(nmrPeak *)NULL;
    }
  }
  return(r);
}
/*-------------------- free_nmrPeak --------------------*/

nmrPeak *free_nmrPeak(nmrPeak *p,int dim)
{
  int i;

  if (p) {
    if (p->pt) free(p->pt);
    if (p->p2) free(p->p2);
    if (p->p3) free(p->p3);
    if (p->asg) {
      for (i=0;i<dim;i++) if (p->asg[i]) free(p->asg[i]);
      free(p->asg);
    }
    free(p);
    p=(nmrPeak *)NULL;
  }
  return(p);
}
