#ifndef OBJECTDEF
#define OBJECTDEF

#include "wnpdef.h"
#include "gcalc.h"
#include "cptr.h"

#define MX_OBJ_CON 4
#define MX_OBJ_GRP 5
#define MX_OBJ_FFMNEM 6
#define MX_OBJ_CHS 3
#define MX_OBJ_GEOM 4
#define MX_OBJ_RATM 9

#define MX_XCN_FORMUL 20
#define MX_XCN_SYMOPS  192

#define BITS_PER_WORD (8*sizeof(unsigned long))
#define ATOMSET_WORDS 2
#define ATOMSET_BITS (BITS_PER_WORD*ATOMSET_WORDS)

/* flags for TreeCopyAtoms: */
#define OBJ_CPYINTERBNDS  (1<<0)
#define OBJ_CPYINTRABNDS  (1<<1)
#define OBJ_CPYALLBNDS    ((OBJ_CPYINTERBNDS|OBJ_CPYINTRABNDS))
#define OBJ_CPYMINTREE    (1<<2)
#define OBJ_CPYSETS       (1<<3)
#define OBJ_CPYXCN        (1<<4)
#define OBJ_CPYNOE        (1<<5)
#define OBJ_CPYLEV        (1<<16)

#define SVR_GLBLS  (DSP_ROOT|DSP_GLBL)
#define SVR_GRAPHS (DSP_MOL|DSP_CSO|DSP_CNT|DSP_TRCE|DSP_TRG|DSP_HHDIST)

#define ND_DISPOSE_DELETE 1
#define ND_DISPOSE_IGNORE 2
#define ND_DISPOSE_UNLINK 3

/*==================== definition of node types ====================*/

#define OBJ_N_TYPES 17

#define OBJ_ALL      (0xffffffff)
#define OBJ_ROOT     (1<<0)
#define OBJ_MOL      (1<<1)
#define OBJ_ATOM     (1<<2)
#define OBJ_LIB      (1<<3)
#define OBJ_ATOMDEF  (1<<4)
#define OBJ_SYMBOL   (1<<5)
#define OBJ_LINK     (1<<6)
#define OBJ_BONDINC  (1<<7)
#define OBJ_ATOMSET  (1<<8)
#define OBJ_XCN      (1<<9)
#define OBJ_XATOM    (1<<10)
#define OBJ_DSP      (1<<11)
#define OBJ_PRULE    (1<<12)
#define OBJ_NOELIST  (1<<13)
#define OBJ_CNF      (1<<14)
#define OBJ_FRULE    (1<<15)
#define OBJ_DATA     (1<<16)

/*==================== Abstract node definitions ====================*/

typedef struct Node {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
} Node;

#define in_chain(N,C) (mvld((N)->chain)||(Node *)(N)==(C)->e)

#define obj_is(x,t) \
  ( mvld(x) ? ((((Node *)(x))->type & (t)) == ((Node *)(x))->type) : 0)

#define ForObjs(o,r,t) \
        for(o=rfirst_type(r,t);o;o=rnext_type(r,o,t))

#define forObjs(o,r,t) \
       for(o=first_type(r,t);o;o=next_type(o,t))
 
#define is_root(x)      obj_is(x,OBJ_ROOT)

#define first_node(n) (Node *)first_type((Node *)n,OBJ_ALL)
#define next_node(a) (Node *)next_type((Node *)a,OBJ_ALL)

#define rfirst_node(n) (Node *)rfirst_type((Node *)n,OBJ_ALL)
#define rnext_node(n,a) (Node *)rnext_type((Node *)n,(Node *)a,OBJ_ALL)

#define forNodes(a,r) for(a=first_node(r);a;a=next_node(a))
#define ForNodes(a,r) for(a=rfirst_node(r);a;a=rnext_node(r,a))
#define NextNodes(a,r,s) for(a=rnext_node(r,s);a;a=rnext_node(r,a))

/*-------------------- Mol node definitions --------------------*/

typedef struct Mol {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char     *name;
  int       seqnr;
  void     *dsptr; 
  unsigned  long  local[ATOMSET_WORDS];
  unsigned  long  used[ATOMSET_WORDS];
  unsigned  short moltype;      /* residue, chain, ... */
  unsigned  keep:1;
} Mol;

/* subtypes : */

#define MOL_UNDEF     0
#define MOL_RESIDUE   1
#define MOL_CHAIN     2
#define MOL_FILE      3
#define MOL_SUBSTRUCT 4
#define MOL_MOLECULE  5
#define MOL_ROOT      6
#define MOL_NMRSET    7
#define MOL_MODEL     8

#define is_mol(x)      obj_is(x,OBJ_MOL)

#define first_mol(n) (Mol *)first_type((Node *)n,OBJ_MOL)
#define next_mol(a) (Mol *)next_type((Node *)a,OBJ_MOL)

#define rfirst_mol(n) (Mol *)rfirst_type((Node *)n,OBJ_MOL)
#define rnext_mol(n,a) (Mol *)rnext_type((Node *)n,(Node *)a,OBJ_MOL)

#define forMols(a,r) for(a=first_mol(r);a;a=next_mol(a))
#define ForMols(a,r) for(a=rfirst_mol(r);a;a=rnext_mol(r,a))
#define NextMols(a,r,s) for(a=rnext_mol(r,s);a;a=rnext_mol(r,a))

/*-------------------- structure independent cryst. info --------------------*/
typedef mat_4 Symop;

typedef struct Xcn {           /* structure independent cryst. data */
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char     *name;
  int       seqnr;
  char     *code;
  char     *title;
  struct    hisrec *history;
  char     *spgr;
  char     *scale;                     /* scale record, uninterpreted */
  struct    Dspobj *boxdsp;            /* pointer to box displ obj */
  mat_4     fta,atf;
  Symop     sym[MX_XCN_SYMOPS];
  char      forms[MX_XCN_FORMUL][3];
  float     formn[MX_XCN_FORMUL];
  float     lambda;
  float     cl[3],ca[3];
  float     esd_cl[3],esd_ca[3];
  float     asym[6];
  int       forml,lattice,centric,nsym,nsymx,box_color;
} Xcn;

typedef struct hisrec {
  struct hisrec *nxt;
  char   *t;
} hisrec;

#define is_xcn(x)       obj_is(x,OBJ_XCN)

#define first_xcn(n) (Xcn *)first_type((Node *)n,OBJ_XCN)
#define next_xcn(a) (Xcn *)next_type((Node *)a,OBJ_XCN)

#define rfirst_xcn(n) (Xcn *)rfirst_type((Node *)n,OBJ_XCN)
#define rnext_xcn(r,a) (Xcn *)rnext_type((Node *)r,(Node *)a,OBJ_XCN)

#define ForXcns(a,r) for(a=rfirst_xcn(r);a;a=rnext_xcn(r,a))
#define NextXcns(a,r,s) for(a=rnext_xcn(r,s);a;a=rnext_xcn(r,a))

#define forXcns(d,r) for(d=first_xcn(r);d;d=next_xcn(d))
#define nextXcns(d,s) for(d=next_xcn(s);d;d=next_xcn(d))


/*-------------------- NOE list --------------------*/

typedef struct NOElist {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char     *name;
  int       seqnr;
  struct NOEdata *first,*last;
  int avtype;
} NOElist;

#define NOE_AVTYPE_CENTER   1
#define NOE_AVTYPE_R6       2
#define NOE_AVTYPE_R3       3
#define NOE_AVTYPE_SUM      4

#define is_NOElist(x)       obj_is(x,OBJ_NOELIST)

#define first_NOElist(n) (NOElist *)first_type((Node *)n,OBJ_NOELIST)
#define next_NOElist(a) (NOElist *)next_type((Node *)a,OBJ_NOELIST)

#define rfirst_NOElist(n) (NOElist *)rfirst_type((Node *)n,OBJ_NOELIST)
#define rnext_NOElist(r,a) (NOElist *)rnext_type((Node *)r,(Node *)a,OBJ_NOELIST)

#define ForNOElists(a,r) for (a=rfirst_NOElist(r);a;a=rnext_NOElist(r,a))
#define NextNOElists(a,r,s) for(a=rnext_NOElist(r,s);a;a=rnext_NOElist(r,a))

#define forNOElists(d,r) for(d=first_NOElist(r);d;d=next_NOElist(d))
#define nextNOElists(d,s) for(d=next_NOElist(s);d;d=next_NOElist(d))

/*-------------------- display object definitions --------------------*/

typedef struct Dspobj {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char     *name;
  int       seqnr;
  int       dsptype;
  int       status;
  Node     *root;
  mat_4     mtrx;                 /* dsp transformation matrix (total) */
  mat_4     rot;                  /* local rotation */              
  float     tra[3],pivot[3];      /* local translation and pivot */
  cptr_t    cross;
} Dspobj;

/* subtypes : */

/* the following flags must match the corresponding definitions in the server */
#define DSP_ROOT      (1<<0)
#define DSP_GLBL      (1<<1)
#define DSP_MOL       (1<<2)
#define DSP_CSO       (1<<3)
#define DSP_CNT       (1<<4)      /* implemented as CSO/MDS */
#define DSP_TRCE      (1<<5)      /*      "      "     "    */
#define DSP_TRG       (1<<6)
#define DSP_HHDIST    (1<<7)      /* implemented as CSO/MDS */

/* the following flags must match the corresponding definitions in the server */
#define OBJ_FLG_DSP           (1<<0)
#define OBJ_FLG_RBLD          (1<<1)
#define OBJ_FLG_MOD           (1<<2)
#define OBJ_FLG_ACTIVE        (1<<3)

#define is_dsp(x)       obj_is(x,OBJ_DSP)

#define first_dspobj(n) (Dspobj *)first_type((Node *)n,OBJ_DSP)
#define next_dspobj(a) (Dspobj *)next_type((Node *)a,OBJ_DSP)

#define rfirst_dspobj(n) (Dspobj *)rfirst_type((Node *)n,OBJ_DSP)
#define rnext_dspobj(n,a) (Dspobj *)rnext_type((Node *)n,(Node *)a,OBJ_DSP)

#define forDspobjs(a,r) for(a=first_dspobj(r);a;a=next_dspobj(a))
#define ForDspobjs(a,r) for(a=rfirst_dspobj(r);a;a=rnext_dspobj(r,a))
#define NextDspobjs(a,r,s) for(a=rnext_dspobj(r,s);a;a=rnext_dspobj(r,a))

/*-------------------- Atom definitions --------------------*/

typedef struct Atom {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  struct    Atomdef *atp;
  char      *label;
  cptr_t    cross;
  struct    Bond **bnd;
  float     *x;                              /* (x,y,z,biso,sof)... */
  char     *alt;
  float     r,q,shift;
  unsigned  long set[ATOMSET_WORDS];
  int       color;
  unsigned  char nbnd,mbnd,nxyz,mxyz;        /* limits to 255 bonds etc... */
  unsigned  char flg;                        /* flags for dirty work       */
  char      grp[MX_OBJ_GRP];
  signed char lig,val,chrg,lcolor,ltype;     /* ligance,valence,charge,label color,label type */
  unsigned  stass:2;                         /* chem shift stereo assignment state */
  unsigned  arom:1;                          /* aromatic ?                         */
  unsigned  cosw:1,oop:2,dsp:1,xtal:1;       /* switching center? oop type, display? have xtal? */
  unsigned  ldsp:1,sdsp:1;                   /* display label? sphere? */
  unsigned  fe:1;                            /* fe is free-energy flag, never used !!! */
} Atom;

#define ATM_STASS_NO    0                    /* stereo assignment is irrelevant    */
#define ATM_STASS_YES   1                    /* shift is stereo-assigned           */
#define ATM_STASS_GUESS 2                    /* stereo assignment is guessed       */
#define ATM_STASS_UNK   3                    /* stereo assignment is unknown       */
#define ATM_LOCSZ       5                    /* number of floats pre location in x */
#define ATM_BISO        3                    /* offset to biso in x                */
#define ATM_SOF         4                    /* offset to sof in x                 */

#define is_atom(x)      obj_is(x,OBJ_ATOM)

#define first_atom(n) (Atom *)first_type((Node *)n,OBJ_ATOM)
#define next_atom(a) (Atom *)next_type((Node *)a,OBJ_ATOM)

#define rfirst_atom(n) (Atom *)rfirst_type((Node *)n,OBJ_ATOM)
#define rnext_atom(n,a) (Atom *)rnext_type((Node *)n,(Node *)a,OBJ_ATOM)

#define forAtoms(a,r) for(a=first_atom(r);a;a=next_atom(a))
#define ForAtoms(a,r) for(a=rfirst_atom(r);a;a=rnext_atom(r,a))
#define NextAtoms(a,r,s) for(a=rnext_atom(r,s);a;a=rnext_atom(r,a))

/*-------------------- crystallographic atom info --------------------*/

typedef struct Xatom {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char     *name;
  int       seqnr;
  struct    ADfix **adf;
  float     bfac[6],peak_height;   /* Note: order of temp. fact. is u11,22,33,23,13,12 !!! */
  float     pfix[10];
  float     esd_x[3],esd_sof,esd_bfac[6];
  int       nadf,madf,afix_root,block;
  unsigned  iso:1,peak:1;
} Xatom;

#define is_xatom(x)     obj_is(x,OBJ_XATOM)

#define first_xatom(n) (Xatom *)first_type((Node *)n,OBJ_XATOM)
#define next_xatom(a) (Xatom *)next_type((Node *)a,OBJ_XATOM)

#define rfirst_xatom(n) (Xatom *)rfirst_type((Node *)n,OBJ_XATOM)
#define rnext_xatom(n,a) (Xatom *)rnext_type((Node *)n,(Node *)a,OBJ_XATOM)

#define forXatoms(a,r) for(a=first_xatom(r);a;a=next_xatom(a))
#define ForXatoms(a,r) for(a=rfirst_xatom(r);a;a=rnext_xatom(r,a))
#define NextXatoms(a,r,s) for(a=rnext_xatom(r,s);a;a=rnext_xatom(r,a))

/*-------------------- potential type rule --------------------*/

#define PR_MXBND 8
#define PR_SOFT  1
#define PR_HARD  2
#define PR_SUBST 3

typedef struct PrAtom {
  struct Atom *match;          /* pointer to matching atom (after matchPrule) */
  float q;                     /* partial atomic charge for fragment charges */
  int nbnd;
  int bnd[PR_MXBND];           /* bond numbers */
  int nei[PR_MXBND];           /* atom numbers (neighbours) */
  char sym[3];                 /* chemical symbol */
  char val,lig,hc;             /* valence, ligance, hydrogen count */
  char sc,arom;                /* substituent count, aromatic ? */
  signed char chrg;            /* charge */
} PrAtom;

typedef struct PrBond {
  struct Bond *match;          /* pointer to matching bond (after matchPrule) */
  int from,to;                 /* atom numbers forming bond */
  char bo;                     /* formal bond order */
  unsigned arom:1;             /* aromatic ? */
  unsigned cover:1;            /* covering ? */
} PrBond;

typedef struct Prule {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  char      *def;              /* definition (swiles) */
  struct PrAtom    *atom;      /* list of atoms */
  struct PrBond    *bond;      /* list of bonds */
  int      interp;             /* type of rule ... */
  int      nat,nbnd;           /* number of atoms, numberof bonds */
} Prule;

#define is_prule(x)   obj_is(x,OBJ_PRULE)

#define first_prule(n) (Prule *)first_type((Node *)n,OBJ_PRULE)
#define next_prule(a) (Prule *)next_type((Node *)a,OBJ_PRULE)

#define rfirst_prule(n) (Prule *)rfirst_type((Node *)n,OBJ_PRULE)
#define rnext_prule(n,a) (Prule *)rnext_type((Node *)n,(Node *)a,OBJ_PRULE)

#define forPrules(a,r) for(a=first_prule(r);a;a=next_prule(a))
#define ForPrules(a,r) for(a=rfirst_prule(r);a;a=rnext_prule(r,a))
#define NextPrules(a,r,s) for(a=rnext_prule(r,s);a;a=rnext_prule(r,a))

/*-------------------- fix potential type rules --------------------*/

typedef struct Frule {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  struct    Atomdef **iatp,**oatp;    /* input types, output types     */
  struct    Atom **match;             /* matched atom                  */
  int       *nitp;                    /* number of types for i-th atom */
  int       *btp;                     /* bond types                    */
  int       *from,*to;                /* from atom, to atom            */
  int       nat;                      /* number of atoms               */
  int       nbnd;                     /* number of bonds               */
} Frule;

#define is_frule(x)   obj_is(x,OBJ_FRULE)

#define first_frule(n) (Frule *)first_type((Node *)n,OBJ_FRULE)
#define next_frule(a) (Frule *)next_type((Node *)a,OBJ_FRULE)

#define rfirst_frule(n) (Frule *)rfirst_type((Node *)n,OBJ_FRULE)
#define rnext_frule(n,a) (Frule *)rnext_type((Node *)n,(Node *)a,OBJ_FRULE)

#define forFrules(a,r) for(a=first_frule(r);a;a=next_frule(a))
#define ForFrules(a,r) for(a=rfirst_frule(r);a;a=rnext_frule(r,a))
#define NextFrules(a,r,s) for(a=rnext_frule(r,s);a;a=rnext_frule(r,a))

/*-------------------- Atomset definitions --------------------*/

typedef struct Atomset {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  int       bit;                /* bit number representing this set */
} Atomset;

#define is_atomset(x)   obj_is(x,OBJ_ATOMSET)

#define first_atomset(n) (Atomset *)first_type((Node *)n,OBJ_ATOMSET)
#define next_atomset(a) (Atomset *)next_type((Node *)a,OBJ_ATOMSET)

#define rfirst_atomset(n) (Atomset *)rfirst_type((Node *)n,OBJ_ATOMSET)
#define rnext_atomset(n,a) (Atomset *)rnext_type((Node *)n,(Node *)a,OBJ_ATOMSET)

#define ForAtomsets(a,r) for(a=rfirst_atomset(r);a;a=rnext_atomset(r,a))
#define NextAtomsets(a,r,s) for(a=rnext_atomset(r,s);a;a=rnext_atomset(r,a))

#define forAtomsets(d,r) for(d=first_atomset(r);d;d=next_atomset(d))
#define nextAtomsets(d,s) for(d=next_atomset(s);d;d=next_atomset(d))

/*-------------------- library node definitions --------------------*/

typedef struct Lib {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
} Lib;

#define is_lib(x)       obj_is(x,OBJ_LIB)

#define first_lib(n) (Lib *)first_type((Node *)n,OBJ_LIB)

/*-------------------- chemical symbol definitions --------------------*/

typedef struct Symbol {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  float     mass,rvdw,rcov;
  float     GBrad,GBscr;
  int       z,color,val;
} Symbol;

#define is_symbol(x)    obj_is(x,OBJ_SYMBOL);

#define first_symbol(n) (Symbol *)first_type((Node *)n,OBJ_SYMBOL)
#define next_symbol(a) (Symbol *)next_type((Node *)a,OBJ_SYMBOL)

#define forSymbols(d,r) for(d=first_symbol(r);d;d=next_symbol(d))
#define nextSymbols(d,s) for(d=next_symbol(s);d;d=next_symbol(d))

/*-------------------- atomtype definitions --------------------*/

typedef struct Atomdef {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  char      *bnd_equiv;                       /* equivalent name in bond-increment */
  float     mass,rvdw,rcov;
  float     GBrad,GBscr;
  int       z,nlnks;
  int       color,numtp;
  signed char  lig,val,chrg,orb;              /* ligance,valence,charge */
  char      chs[MX_OBJ_CHS],geom[MX_OBJ_GEOM];
  unsigned  oop:1,hbdon:1,hbacc:1,arom:1;     /* flag bits */
} Atomdef;

#define is_atomdef(x)   obj_is(x,OBJ_ATOMDEF)

#define first_atomdef(n) (Atomdef *)first_type((Node *)n,OBJ_ATOMDEF)
#define next_atomdef(a) (Atomdef *)next_type((Node *)a,OBJ_ATOMDEF)

#define rfirst_atomdef(n) (Atomdef *)rfirst_type((Node *)n,OBJ_ATOMDEF)
#define rnext_atomdef(n,a) (Atomdef *)rnext_type((Node *)n,(Node *)a,OBJ_ATOMDEF)

#define forAtomdefs(d,r) for(d=first_atomdef(r);d;d=next_atomdef(d))
#define nextAtomdefs(d,s) for(d=next_atomdef(s);d;d=next_atomdef(d))

#define ForAtomdefs(a,r) for(a=rfirst_atomdef(r);a;a=rnext_atomdef(r,a))
#define NextAtomdefs(a,r,s) for(a=rnext_atomdef(r,s);a;a=rnext_atomdef(r,a))

/*-------------------- general data definitions --------------------*/

#define OBJ_DATA_NONE     0       /* no data        */
#define OBJ_DATA_FLXMAT   1       /* Felix matrix   */
#define OBJ_DATA_INSANE   2       /* Insane status  */

typedef struct Data {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  struct Data  *(*copy)(struct Data *data);     /* Function called to copy Data */
  int (*delete)(struct Data *data);             /* Function called to delete Data */
  void      *data;                /* pointer to data */            
  int       dtype;                /* data type       */
} Data;

#define is_data(x)   obj_is(x,OBJ_DATA)

#define first_data(n) (Data *)first_type((Node *)n,OBJ_DATA)
#define next_data(a) (Data *)next_type((Node *)a,OBJ_DATA)

#define rfirst_data(n) (Data *)rfirst_type((Node *)n,OBJ_DATA)
#define rnext_data(n,a) (Data *)rnext_type((Node *)n,(Node *)a,OBJ_DATA)

#define forData(d,r) for(d=first_data(r);d;d=next_data(d))
#define nextData(d,s) for(d=next_data(s);d;d=next_data(d))

#define ForData(a,r) for(a=rfirst_data(r);a;a=rnext_data(r,a))
#define NextData(a,r,s) for(a=rnext_data(r,s);a;a=rnext_data(r,a))

/*-------------------- bond increment for charge calculation --------------------*/

typedef struct Bondinc {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  float     dij,dji;
} Bondinc;

#define is_bondinc(x)   obj_is(x,OBJ_BONDINC)

#define rfirst_bondinc(n) (Bondinc *)rfirst_type((Node *)n,OBJ_BONDINC)
#define rnext_bondinc(n,a) (Bondinc *)rnext_type((Node *)n,(Node *)a,OBJ_BONDINC)

#define first_bondinc(n) (Bondinc *)first_type((Node *)n,OBJ_BONDINC)
#define next_bondinc(a)  (Bondinc *)next_type((Node *)a,OBJ_BONDINC)

#define ForBondincs(a,r) for(a=rfirst_bondinc(r);a;a=rnext_bondinc(r,a))
#define NextBondincs(a,r,s) for(a=rnext_bondinc(r,s);a;a=rnext_bondinc(r,a))

#define forBondincs(d,r)  for(d=first_bondinc(r);d;d=next_bondinc(d))
#define nextBondincs(d,s) for(d=next_bondinc(s);d;d=next_bondinc(d))

/*-------------------- bondtype definitions --------------------*/

#define MX_OBJ_BTYPES 8

#define BT_DUMMY  0
#define BT_SINGLE 1
#define BT_DOUBLE 2
#define BT_TRIPLE 3
#define BT_AROMATIC 4
#define BT_AMIDE  5
#define BT_NOTCONNECTED 6
#define BT_UNKNOWN 7

/*-------------------- inter residue link definitions --------------------*/

typedef struct Link {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  char      *from;
  char      *to;
  float     dmin,dmax;
  int       bo;
  int       bt;
} Link;

#define is_link(x)      obj_is(x,OBJ_LINK)

#define first_link(n) (Link *)first_type((Node *)n,OBJ_LINK)
#define next_link(a) (Link *)next_type((Node *)a,OBJ_LINK)

#define forLinks(d,r) for(d=first_link(r);d;d=next_link(d))
#define nextLinks(d,s) for(d=next_link(s);d=next_link(d))

/*-------------------- conf node  --------------------*/

typedef struct CnfNode {
  long type;
  struct    Node *up;
  struct    Node *prv,*nxt;
  struct    Node *fst,*lst;
  struct    Node *chain;
  char      *name;
  int       seqnr;
  float     *x;
  float     energy;
  int       cseqnr;
  int       n;
} CnfNode;

#define is_cnf(x)      obj_is(x,OBJ_CNF)

#define first_cnf(n) (CnfNode *)first_type((Node *)n,OBJ_CNF)
#define last_cnf(n) (CnfNode *)last_type((Node *)n,OBJ_CNF)
#define next_cnf(a) (CnfNode *)next_type((Node *)a,OBJ_CNF)

#define forCnfs(d,r) for(d=first_cnf(r);d;d=next_cnf(d))
#define nextCnfs(d,s) for(d=next_cnf(s);d=next_cnf(d))

#define rfirst_cnf(n) (CnfNode *)rfirst_type((Node *)n,OBJ_CNF)
#define rnext_cnf(n,a) (CnfNode *)rnext_type((Node *)n,(Node *)a,OBJ_CNF)

#define ForCnfs(a,r) for(a=rfirst_cnf(r);a;a=rnext_cnf(r,a))
#define NextCnf(a,r,s) for(a=rnext_cnf(r,s);a;a=rnext_cnf(r,a))

/*======================================================================*/

typedef struct Bond {
  struct Bond *chain;
  cptr_t cross;
  struct Atom *f,*t;
  char *label;
  float r;                       /* cylinder radius */
  unsigned char bo36,bo,bt;      /* bo36 : partial bond order in 1/36ths */
  unsigned char flg;             /* flags for dirty work */
  unsigned char lw;              /* linewdith */
  unsigned arom:1,cyl:1,labt:4;  /* aromatic ? cylinder on ? label type */
} Bond;

#define ADFCLASS_AFIX 1
#define ADFCLASS_DFIX 2

typedef struct ADfix {
  long class;
  Xatom *f;
  Xatom *t;
  float dist;
  float esd;
} ADfix;

typedef struct Atmmatch {
  long count;
  Atom **a;
  Atom **b;
} Atmmatch;

typedef struct Nodechain{
  int n;
  Node *s;
  Node *e;
  Node *p;
} Nodechain;

typedef struct NOEdata {
  struct NOElist  *root;
  struct NOEdata *prv,*nxt;
  Atom *tptr,*sptr;
  char *def1,*def2;
  char *sseg,*tseg;
  char *cmnt;
  float dupper,weight;
  char sres[11],satom[11];
  char tres[11],tatom[11];
  char active,ps1,ps2;
} NOEdata;

#define PERSPECTIVE 1
#define ORTHO 2

#define STEREO_NO        0
#define STEREO_RELAXED   1
#define STEREO_LEFT      2
#define STEREO_RIGHT     3
#define STEREO_CROSSED   4
#define STEREO_CRYSTAL   5
#define STEREO_QUADBUFF  6 
#define STEREO_MAX_MODES 7

typedef struct PrjTS {      /* 3D projection info in graphics server  */
  int viewmode;             /* ortho, perspective, ...                */
  int stereomode;           /* mono, relaxed, left, right, ...        */ 
  float  ysize;             /* viewport's  y-size in A                */
  float  znear;             /* near z-clipping plane in A             */
  float  slabwidth;         /* slabwidth in A                         */
  float  viewdist;          /* viewing distance in screen y-sizes     */
  float  eyesep;            /* left/right eye separation in screen ys */
  float  aspect;            /* viewport's x:y aspect ratio            */
  float  scrysize;          /* screen y-size in A                     */
  float  sep;               /* viewport separation                    */
} PrjTS;
  
typedef struct Grid {
  char *name;               /* filename, etc...                         */
  Dspobj *Box;              /* box associated with grid                 */
  float xmin[3],xmax[3];    /* extent of grid                           */
  float dx[3];              /* delta x,y,z                              */
  float cl[3];              /* unit cell lengths                        */
  float ca[3];              /* unit cell angles                         */
  float xl[3],xh[3];        /* extent of grid to be used in contour add */
  int   n[3];               /* number of points along each dimension    */
  int   boxcolor,boxwidth,boxon; /* box graphics attributes             */
  float *val;               /* values at grid points                    */
} Grid;

typedef struct MPEOEopt {   /* options for the MPEOE method                                        */
  float tolref,dChi,dF;     /* refinement tolerance, init. delta Ch and delta f                    */
  float tolcomp;            /* tolerance during charge computation                                 */
  float minf,maxf;          /* allowed range for f-parameters (bond damping...)                    */
  int mxref;                /* max. number of  refinement steps                                    */
  int mxcomp;               /* max. number of iterations during MPEOE                              */
  int minocc;               /* minimum required number of occurances of parameter for optimiziation*/
  int repf;                 /* progress report frequency                                           */
} MPEOEopt;

typedef struct MPEOEinfo {  /* scratch pad to pass info between MPEOE routines                   */
  Atom **A;                 /* atoms for whcih charges are to be computed                        */
  Bond **B;                 /* list of bonds between atoms in A                                  */
  struct MPEOEatom **Mpa;   /* pointers to MPEO atom parameters for Atoms in A                   */
  struct MPEOEbond **Mpb;   /* pointers to MPEO bond parameters for Atoms in B                   */
  struct MPEOEatom *Latom;  /* pointer to Atomparameter library (head of linked list)            */
  struct MPEOEbond *Lbond;  /* pointer to Atomparameter library (head of linked list)            */
  float *Q,*iQ,*q,*p,*dp;   /* atom q, init. atom q, bond dq, parameters, dp for Simplex         */
  int nat,nbnd,npara,nparb; /* number od atoms, bonds, var. atom parameters, var. bond parameters*/
} MPEOEinfo;

typedef struct MPEOEatom {
  struct MPEOEatom *nxt;    /* pointer to next MPEOEatom in list                      */
  char *def;                /* pointer to fragment defintion                          */
  Prule *p;                 /* pointer to parsed swiles                               */
  float P[3];               /* electronegativity parameters (A,B,X)                   */
  int indX;                 /* true, iff X is an independent parameter                */
  int opt[3];               /* true iff parameter is to be optimized                  */
  int nu;                   /* number of times parameters are used in target function */
} MPEOEatom;

typedef struct MPEOEbond {
  struct MPEOEbond *nxt;    /* pointer to next MPEOEbond in list                      */
  char *def;                /* pointer to fragment defintion                          */
  Prule *p;                 /* pointer to parsed swiles                               */
  float f;                  /* damping factor for this type of bond                   */
  int opt;                  /* true iff parameter is to be optimized                  */
  int nu;                   /* number of times parameters are used in target function */
} MPEOEbond;

/*-------------------- Macros ... --------------------*/

typedef struct wnpmacro_t {
  struct wnpmacro_t *prv,*nxt;
  char *name;
  char **txt;
  int n_open;
  int n;
} wnpmacro_t;

/*==================== prototypes ====================*/

int (*data_delete_method(Data *d,int (*new)(Data *data)))(Data *data);
Data *(*data_copy_method(Data *d,Data *(*new)(Data *data)))(Data *data);

int   tree_addr(Node *a,Node *root,int *addr);
Node *locate_node(Node *root,int n,int *addr);
Node *locate_node_name(Node *r1,Node *r2,Node *n2);
char *data_typename(int type);
char *obj_typename(long type);
int   checknode(Node *n);
Node *make_node(int type);
Node *copy_node(Node *n);
Node *copy_parent(Node *n);
Node *add_node(Node *parent,char *name,int type);
Node *link_node(Node *new);
Node *unlink_node(Node *n);
Node *tree_copy(Node *r);
Mol  *TreeCopyAtoms(int nat,Atom **atom,int flg,int how,Node *target,Node *rtarget);
hisrec *make_history(void);
Node *rdel_obj(Node *n,long t);
Node *del_node(Node *node);
int   RDeleteNodesInChain(Nodechain *c);
void  RDeleteAtomsInArray(int n,Atom **atoms);
char *set_node_name(Node *node,char *name);
ADfix *make_adfix(void);
ADfix *add_adfix(Xatom *from,Xatom *to);
void del_adfix(ADfix *fix);
Bond *make_bond(void);
Bond *add_bond_raw(Atom *from,Atom *to);
Bond *add_bond_ng(Atom *from,Atom *to);
Bond *add_bond(Atom *from,Atom *to);
Bond *add_bond_copy(Atom *f,Atom *t,Bond *bo);
Bond *find_bond(Atom *from,Atom *to);
Bond *AtomToBondChain(Nodechain *c);
void  del_bond(Bond *bond);
void  del_bond_graph(Bond *b);
cptr_t add_bond_graph(Bond *b);
MPEOEatom *del_MPEOEatom(MPEOEatom *a);
MPEOEbond *del_MPEOEbond(MPEOEbond *b);
MPEOEatom *make_MPEOEatom(void);
MPEOEbond *make_MPEOEbond(void);
int MPEOEreport_missing(MPEOEinfo G,int prt);
void set_atom_type(Atom *a,Atomdef *type);
void set_atom_biso(Atom *a,float biso);
void set_atom_location(Atom *a,char alt);
Atmmatch *make_atmmatch(long count);
void  del_atmmatch(Atmmatch *m);
float *reall_atom_x(Atom *a,int size);
void  residue_closure(Nodechain *c);
Nodechain *make_node_chain(void);
int walk_order(Node *root,int n,Atom **atoms,int flg);
Nodechain *ptr_to_chain(Node **N,int n);
Nodechain *collect_by_flags(Node *n,int flags);
Nodechain tree2chain(Node *r,long type);
Node **tree2ptr(Node *r,long type,int *n);
void init_node_chain(Nodechain *c);
void del_node_chain(Nodechain *c);
void add_to_node_chain(Node *n,Nodechain *c);
Node *RemoveHeadOfChain(Nodechain *c);
void add_atomset_to_chain(Atomset *s,Nodechain *c);
Nodechain *free_node_chain(Nodechain *c);
void clear_bond_chain(Bond *b);
void clear_node_chain(Nodechain *c);
int  chain_reduce_subtree(Node *root,Nodechain *c);
Nodechain *walk_order_chain(Node *root,Nodechain *c,int flag);
int node_in_chain(Node *n,Nodechain *c);
Node **chain_to_ptr(Nodechain *c,int fr_chain);
void clear_atomset(Atomset *s);
void register_atomset(Atomset *s);
void remove_atomset_bit(Atomset *s);
unsigned long used_bits(Node *n,int bit);
Atomset *add_atomset(Mol *m,int bit,char *name);
Atomset *define_atomset(Mol *m,char *sname,Nodechain *c);
int in_atomset(Atom *a,Atomset *s);
void add_to_atomset(Atom *a,Atomset *s);
void remove_from_atomset(Atom *a,Atomset *s);
int  available_atomset_bit(int bit,Mol *m);
void renumber_atomsets(Node *root,Node *target);
void copy_set_bits(Node *source,Node *target);
void merge_atomsets(Node *t);
Node *newLib(char *name);

Node *find_node(Node *root,void *addr,long obj_type);
Node *find_name(Node *root,char *name,long obj_type);
Node *rfind_name(Node *root,char *name,long obj_type);
Node *rfirst_type(Node *root,long obj_type);
Node *rnext_type(Node *root,Node *now,long obj_type);
Node *first_type(Node *root,long obj_type);
Node *next_type(Node *now,long obj_type);
Node *last_type(Node *root,long obj_type);
Node *find_cousin(Node *now,long obj_type);
Node *find_cname(Node *root,char *name,long obj_type);

Atomdef *find_atomdef_name(Node *lib,char *name);
Atomdef *find_atomdef_number(Node *lib,int nr);

int     dsp_is_on(Dspobj *d);
void    del_assoc_dsp(Node *n);
void    del_graph(Dspobj *d);
Dspobj *assocobj(Node *n);
Dspobj *assocmdl(Node *n);
Dspobj *glblobj(Dspobj *d);
void    get_graph_xyz(int n,float *x,Atom **a);

NOEdata *make_noedata(void);
NOEdata *copy_noedata(NOEdata *noe);
NOEdata *link_noedata(NOEdata *noe);
NOEdata *add_noedata(NOElist *list);
void     del_noedata(NOEdata *noe);

Node    *common_root(Nodechain *c);
Node    *common_parent(Node *a,Node *b);

Bondinc *add_bondinc(Lib *lib,char *t1,char *t2,float d12,float d21);

void     transform_atom(mat_4 t,Atom *a);

float biso_equiv(Xatom *xa);

/* the following are for the KLUDGE */

int BT_number(char *mnem);
int BT_number(char *mnem);
char BT_order(int num);
int BT_arom(int num);
char *BT_mnem(int num);
int BT_count(int num);
int BT_el(int num);

wnpmacro_t *new_macro(void);
void set_macro_name(wnpmacro_t *m,char *name);
void add_macro_line(wnpmacro_t *m,char *line);
void insert_macro(wnpmacro_t *m,wnpmacro_t *list);
wnpmacro_t *unlink_macro(wnpmacro_t *m);
wnpmacro_t *delete_macro(wnpmacro_t *m);
wnpmacro_t *get_macro_by_name(wnpmacro_t *macros,char *name);
char *make_macro_menu(char *prfx,wnpmacro_t *macros);
char *read_macro(wnpmacro_t *m,int lnr);
void execute_macro(wnpmacro_t *m);
void open_macro(wnpmacro_t *m);
void close_macro(wnpmacro_t *m);

#endif
