#ifndef PAMBERDEF
#define PAMBERDEF

#include "object.h"

#define PRMTOP_FMT_OLD    1
#define PRMTOP_FMT_NEW    2

#define PRM_SLOPPY_NEVER  0
#define PRM_SLOPPY_ONCE   1
#define PRM_SLOPPY_ALWAYS 2
#define PRM_SLOPPY_TEST   3

#define PRM_MISSING_FIXED  0
#define PRM_MISSING_FLOPPY 1
#define PRM_MISSING_ZERO   2
#define PRM_MISSING_TEST   3

#define P_MAX_DIH_MULT 10  /* max multiplicity of dihedral term, should be ample. */

typedef struct P_BOND  {
  double REQ,RK;     /* parameters, obviously */
  int f,t;           /* numeric types */
  int fnr;           /* file number */
} P_BOND;

typedef struct P_ANGL  {
  double TEQ,TK;     /* parameters, obviously */
  int l,m,r;         /* numeric types */
  int lev;           /* wild-card level */
  int fnr;           /* file number */
} P_ANGL;

typedef struct P_DIHE  {
  double PK[P_MAX_DIH_MULT];
  double PN[P_MAX_DIH_MULT];
  double PHASE[P_MAX_DIH_MULT];   /* parameters, obviously */
  int lev[P_MAX_DIH_MULT];        /* wild-card level */
  int fnr[P_MAX_DIH_MULT];        /* file number in which coeff. is defined */
  int exact;                      /* 0: contains wild cards, >0 : file number with exact def. */
  int i,j,k,l ;                   /* numeric types */
  int m;                          /* multiplicity (dim. of parameter arrays) */   
  int bt;                         /* used to translate dihedral types to expanded types */
  int ok;                         /* defined ? */
} P_DIHE;

typedef struct P_IMPR  {
  double PK,PN,PHASE;/* parameters, obviously */
  int t[4];          /* numeric types */
  int o[4];          /* permutation of atom positions */
  int lev;           /* wild-card level */
  int fnr;
  int ok;            /* defined ? */
} P_IMPR;

typedef struct P_HBON  {
  double A,B;        /* parameters, obviously */
  int f,t;           /* numeric types */
  int fnr;           /* file number */
} P_HBON;

typedef struct PARMSET {
  double  *mass,*R,*E;
  char **atp;
  char ***equiv;
  int  *fnb,*fmass;
  struct P_BOND *bnd;
  struct P_ANGL *ang;
  struct P_DIHE *dih;
  struct P_IMPR *imp;
  struct P_HBON *hbo;
  int nmass,nbnd,nang,ndih,nimp,nhbo;
  int nequiv;
  int fnr,status;
} PARMSET;

int     oldParmFile(FILE *in);
PARMSET readParmSet(FILE *in,PARMSET p);
PARMSET emptyParmSet(void);
PARMSET clearParmSet(PARMSET p);
PARMSET readOldParmSet(FILE *in,PARMSET p);
PARMSET readFrcModSet(FILE *in,PARMSET p);
PARMSET initParmSetMasses(PARMSET p,Nodechain *c);
PARMSET readParmSetMasses(FILE *in,PARMSET p);
PARMSET readParmSetBonds(FILE *in,PARMSET p);
PARMSET readParmSetAngles(FILE *in,PARMSET p);
PARMSET readParmSetDihedrals(FILE *in,PARMSET p);
PARMSET readParmSetImpropers(FILE *in,PARMSET p);
PARMSET readParmSetHBonds(FILE *in,PARMSET p);
PARMSET readParmSetNBonds(FILE *in,PARMSET p);
PARMSET readParmSetEquivs(FILE *in,PARMSET p);
PARMSET clearParmSetEquivs(PARMSET p);
PARMSET reportParmSetMissing(PARMSET p);
int     checkParmSetMode(int mode);
int     missingParmSetMode(int mode);

#endif
