#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#ifndef mingw
#include <termios.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/param.h>
#include <dirent.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "streamio.h"
#include "bhtree.h"
#include "errhand.h"
#include "fileutil.h"
#include "wnpconst.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

extern Node *MolRoot,*DefaultNode;
extern Node *LibRoot;
extern Dspobj *DspRoot;

#define MAXTOKENSIZE 4096

#define REL_NOP     0
#define REL_EQ      1
#define REL_NE      2
#define REL_LT      3
#define REL_LE      4
#define REL_GT      5
#define REL_GE      6
#define REL_TEQ     7
#define REL_TNE     8

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

typedef struct ParseState {
  struct ParseState *nxt;
  int eoc,eol;
  char *prompt;
  char *deftoken;
}ParseState;

Stream *IOS;
int server;
float srvVersion=0.0;
char *srvString=(char *)NULL;
int graphix;

#define maxFailures 500
static int NFailures=0;
static int kbd_echo=1;
static ParseState *state; 
static char tok[MAXTOKENSIZE];
static Token ErrorToken = {
  (char *)NULL, (char *)NULL, (void *)NULL, 0, 0.0, TOK_FAIL, SRC_NULL
};

static Nodechain *mark_atoms=(Nodechain *)NULL;

static int FT[MAX_FT] = 
  { FT_CAR, FT_BIOC, FT_CIF, FT_CSD, FT_DIANA, FT_DOTS, FT_DUPPER, FT_WLB, FT_MACCS, FT_MAE , FT_MDL,
    FT_MMOD, FT_MOL, FT_MOL2, FT_PAR, FT_PDB, FT_SESS, FT_SHX, FT_XPLOR, FT_XPDB, FT_NOE,
    FT_CNF, FT_CNT};

static char *ft_menu_item[MAX_FT] =
  { "car/mdf", "Biocad", "CIF", "CSD", "diana", "dots", "dupper", "library", "maccs", "maestro", "mdl", 
    "mmod", "mol", "mol2", "par", "pdb", "session", "shelx", "xplor", "xpdb", "xnoe",
    "confa", "contour"
};

static char *DEFEXT[MAX_FT]= {
  ".cor|.car",".cpd",".cif",".dat",".cor",".dot",".upl",".wlb",".sdf",".mae",".sdf",".mmd",".mol",".mol2",".par",
  ".pdb",".wsv",".xcn",".pdb",".pdb",".tbl",".cnf",".cnt"
};

static char *SRCH[MAX_FT] =
{ "*.car|*.car.Z|*.cor|*.cor.Z",
  "*.cpd|*.cpd.Z",
  "*.cif|*.cif.Z",
  "*.dat|*.dat.Z",
  "*.cor|*.cor.Z",
  "*.dot|*.dot.Z",
  "*.upl|*.upl.Z",
  "*.wlb|*.wlb.Z",
  "*.sdf|*.sdf.Z",
  "*.mae|*.mae.Z",
  "*.sdf|*.sdf.Z",
  "*.mmd|*.mmd.Z",
  "*.mol|*.mol.Z|*.mol2|*.mol2.Z",
  "*.mol2|*.mol2.Z|*.mol|*.mol.Z",
  "*.par|*.par.Z",
  "*.pdb|*.pdb.Z",
  "*.wsv|*.wsv.Z",
  "*.xcn|*.xcn.Z",
  "*.pdb|*.pdb.Z",
  "*.pdb|*.pdb.Z",
  "*.tbl|*.tbl.Z",
  "*.cnf|*.cnf.Z",
  "*.cnt|*.cnt.Z",
};

static int fog_val[3] = { 0,0,0 };

static int col_val[WNP_MAXCOL][3] = {
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
  {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
  {0,0,0},{0,0,0}
};

static char *col_mnem[WNP_MAXCOL] = {
  (char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,
  (char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,
  (char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,
  (char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,
  (char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,
  (char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,(char *)NULL,
  (char *)NULL,(char *)NULL
};

static char *col_nmmnem[WNP_MAXCOL] = {
   "background", "1", "2", "3", "4",
   "5","6", "7", "8", "9",
   "10","11", "12", "13", "14",
   "15","16","17","18","19",
   "20","21","22","23","24",
   "25","26","27","28","29",
   "30","31"
};

FILE *redirFile=(FILE *)NULL;
static char fullname[MAXPATHLEN] = "";

#if defined(mingw)
char *realpath(const char *file_name, char *resolved_name)
{
  /* Mingw does not have this function. Should do a lot more than this... */

  strcpy(resolved_name,file_name);
  return(resolved_name);
}
#endif

static void mol_atoms(Nodechain *c);
static void up_atoms(Nodechain *c,int up);

/*-------------------- checkFailure --------------------*/

void uSleep(unsigned int);

static Token *checkFailure(Token *k)
{
  /* prevent runaway process if parser can't get good tokens. */

  if (k) {
    if (k->status==TOK_NULL||k->status==TOK_FAIL) NFailures++;
    else NFailures=0;
  }else NFailures++;

  if (NFailures>=maxFailures) {
    fprintf(stderr,"***************************************************\n");
    fprintf(stderr,"* Wit!P Parser: too many failures reading tokens. *\n");
    fprintf(stderr,"* Sorry, session aborted.                         *\n");
    fprintf(stderr,"***************************************************\n");
    send_header(graphix,"EXIT",0);
    exit(99);
  }else if (NFailures>=0.75*maxFailures) {
    fprintf(stderr,"* Wit!P Parser is starving: %d failures to get input token (%s).\n",
	    NFailures,(k->status==TOK_NULL?"TOK_NULL":"TOK_FAIL"));
    uSleep(5000000); /* take a brief nap */
  }

  return(k);
}

/*-------------------- CurrStream --------------------*/

Stream *CurrStream(void)
{
  return(IOS);
}

/*-------------------- SrvVersion --------------------*/

float SrvVersion(void)
{
  return(srvVersion); /* 0.0, if running w/o server */
}

int set_kbd_echo(int state)
{
  int r;
#ifndef mingw
  struct termios iostate;
#endif

  r=kbd_echo;
  kbd_echo=state;
  send_bheader(graphix,"KBEC",sizeof(int));
  send_int(graphix,kbd_echo);
#ifndef mingw
  if (isatty(STDIN_FILENO)) {
    tcgetattr(STDIN_FILENO,&iostate);
    if (kbd_echo) iostate.c_lflag |= ECHO;
    else iostate.c_lflag &= ~ECHO;
    tcsetattr(STDIN_FILENO,TCSADRAIN,&iostate);
  }
#endif
  return(r);
}

int get_kbd_echo(void)
{
  return(kbd_echo);
}
/*-------------------- Get_RawLine_MW --------------------*/

Token *Get_RawLine_MW(char *prompt,char *menu,int wait_flag)
{
  Token *k;
  
  k=make_Token();
  if (k->status==TOK_FAIL) {
    default_token((char *)NULL);
  }else{
    set_prompt(prompt);
    load_menu_buttons(0,menu);
  
    strm_pos(IOS)=strlen(strm_buff(IOS));
    read_stream(IOS);
    k->text=strdup(strm_buff(IOS));
		 
    clear_menu();
    k->source=strm_src(IOS);
    k->quoted=FALSE;
    k->ltok=k->text?strlen(k->text):0;
    state->eoc=FALSE;
    state->eol=TRUE;
    k->status=TOK_OK;  
    strm_pos(IOS)=strlen(strm_buff(IOS));
  }
  return(checkFailure(k));
}

/*-------------------- Get_String_MW --------------------*/

Token *Get_String_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  int nkeys,loop;
  
  k=make_Token();
  if (k->status==TOK_FAIL) {
    default_token((char *)NULL);
    return(checkFailure(k));
  }
  if (state->eoc) {
    if (state->deftoken) {
      k->status=TOK_OK;
      k->def=TRUE;
      k->text=state->deftoken;
      default_token((char *)NULL);
    }else{
      k->status=TOK_EOC;
    }
    default_token((char *)NULL);
    return(checkFailure(k));
  }

  set_prompt(prompt);
  k->status=TOK_NULL;
  if (state->deftoken) nkeys=load_menu_buttons(0,state->deftoken)+1;
  else nkeys=0;
  if (menu) nkeys+=load_menu_buttons(nkeys,menu)+1;
  loop=TRUE;
  
  do {
    if (token_stream(IOS,sep)) {
      k->text=Strdup(tok);
      if (streq(tok,"abort")) k->status=TOK_ABORT;
      else if (streq(tok,"done")) k->status=TOK_EOL;
      else if (k->text) k->status=TOK_OK;
      else k->status=TOK_NOK;
      loop=FALSE;
    }else if (strm_src(IOS)==SRC_NULL) {
      loop=FALSE;
      break;
    }
    if (strm_ltok(IOS)==0 && strm_term(IOS)=='\n') {
      if (strm_src(IOS)!=SRC_MENU && strm_src(IOS)!=SRC_PICK) {
	if (state->deftoken) {
	  if (state->eol) {
	    k->text=state->deftoken;
	    state->deftoken=(char *)NULL;
	    k->status=TOK_OK;
	    k->def=TRUE;
	    loop=FALSE;
	  }
	}else{
	  k->status=TOK_EOL;
	  loop=wait_flag;
	}
	state->eol=TRUE;
      }
    }
  } while(loop);
  clear_menu();
  k->source=strm_src(IOS);
  k->term=strm_term(IOS);
  k->quoted=strm_quoted(IOS);
  k->ltok=strm_ltok(IOS);
  state->eol=(k->term=='\n')||k->source==SRC_MENU||k->source==SRC_PICK;
  state->eoc=(k->term==';');
  default_token((char *)NULL);
  return(checkFailure(k));
}
/*-------------------- Get_File_MW --------------------*/

Token *Get_File_MW(char *dir,char *ext,char *prompt,char *sep,char *menu,int wait_flag)
{
  /* 25-Jan-2002: if ext==NULL -> don't start with dirlist */ 

  Token *k;
  char *the_menu,*srch,buf[4096],twd[MAXPATHLEN],ffn[MAXPATHLEN],*prefix,*p,*q,*r,*fext;
  int do_dir,loop,status;
  struct stat sbuff;
  UCpipe *ucp;
  
  the_menu=(char *)NULL;
  prefix=(char *)NULL;
  srch=(char *)NULL;

  if (dir) sprintf(twd,"%s",dir);
  else if (!getCWD(twd,sizeof(twd))) sprintf(twd,".");
  fext=(char *)NULL;
  if (ext) {
    p=ext;q=ext; r=buf; *r='\0';
    while ((p=strchr(q,'|'))) {
      *r++='*';
      strncpy(r,q,p-q); r+=p-q; *r++='|'; *r++='*';
      strncpy(r,q,p-q); r+=p-q;strcpy(r,".Z"); r+=2; *r='\0';
      q=p+1; if(strlen(q)>0) strcpy(r,"|");r++;*r='\0';
    }
    if (strlen(q)>0) {
      p=q+strlen(q); *r++='*';
      strncpy(r,q,p-q); r+=p-q; *r++='|'; *r++='*';
      strncpy(r,q,p-q); r+=p-q;strcpy(r,".Z"); r+=2; *r='\0';
      q=p+1; if(strlen(q)>0) strcpy(r,"|");r++;*r='\0';
    }
    fext=strdup(ext); if (strchr(fext,'|')) *strchr(fext,'|')='\0';
  }else sprintf(buf,"*");
  srch=Strdup(buf);
  if (menu) {
    prefix=(char *)malloc(strlen(menu)+strlen(" ' ' -searchmask -directories ' ' ")+1);
    if (prefix) sprintf(prefix,"%s ' ' -searchmask -directories ' ' ",menu);
  }else{
    prefix=(char *)malloc(strlen("-searchmask -directories ' '")+1);
    if (prefix) sprintf(prefix,"-searchmask -directories ' '");
  }
  do_dir=FALSE;
  the_menu=Make_DirFileNameList(prefix,twd,ext?srch:"",do_dir);
  
  k=(Token *)NULL;
  for(loop=TRUE;loop;) {
    loop=FALSE;
    k=free_Token(k);
    k=Get_String_MW(prompt,sep,the_menu,wait_flag);
    if (k->status==TOK_OK) {
      if (*k->text=='|') {
	ucp=UCpopen(k->text);
	k->ptr=(void *)ucp;
	if (!ucp) k->status=TOK_NOK;
	else k->status=TOK_OK;
      }else{
	strcpy(ffn,twd);
	newtwd(ffn,k->text);
	if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	status=stat(ffn,&sbuff);
	if (status<0) {
	  k=Prs_Kwd(k,the_menu);
	  if (k->status==TOK_OK) {
	    k=Prs_Kwd(k,menu);
	    if (k->status==TOK_OK) {
	      k->status=TOK_NOK;
	    }else{
	      k=Prs_Kwd(k,"-searchmask -directories");
	      if (k->status==TOK_OK) {
		if (streq(k->key,"-searchmask")) {
		  default_token(srch);
		  k=free_Token(k);
		  k=Get_String("search mask: ",";");
		  if (k->status==TOK_OK) {
		    if (srch) free(srch);
		    srch=k->text;
		    k->text=(char *)NULL;
		    if (the_menu) free(the_menu);
		    the_menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
		    loop=TRUE;
		  }
		}else if (streq(k->key,"-directories")) {
		  do_dir=!do_dir;
		  if (the_menu) free(the_menu);
		  the_menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
		  loop=TRUE;
		}else{
		  k->status=TOK_NOK;
		}
	      }else{
		if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
		ucp=UCpopen(extfil(buf,ffn,fext,FALSE));
		k->ptr=(void *)ucp;
		if (!ucp) k->status=TOK_NOK;
		else k->status=TOK_OK;
	      }
	    }
	  }else{
	    if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	    ucp=UCpopen(extfil(buf,ffn,fext,FALSE));
	    k->ptr=(void *)ucp;
	    if (!ucp) k->status=TOK_NOK;
	    else k->status=TOK_OK;
	  }
	}else{
	  if (sbuff.st_mode&S_IFDIR) {
	    strcpy(twd,ffn);
	    if (the_menu) free(the_menu);
	    the_menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
	    loop=TRUE;
	  }else{
	    if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	    ucp=UCpopen(extfil(buf,ffn,fext,FALSE));
	    k->ptr=(void *)ucp;
	    if (!ucp) k->status=TOK_NOK;
	    else k->status=TOK_OK;
	  }
	}
      }
    }
  }

  if (srch) free(srch);
  if (the_menu) free(the_menu);
  if (prefix) free(prefix);
  if (fext) free(fext);
  return(k);
}
/*-------------------- Get_Directory_MW --------------------*/

Token *Get_Directory_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  char *the_menu,twd[MAXPATHLEN],ffn[MAXPATHLEN],prmt[4096],*prefix;
  int it=0,loop,status;
  struct stat sbuff;
  
  the_menu=(char *)NULL;
  prefix=(char *)NULL;

  if (!getCWD(twd,sizeof(twd))) sprintf(twd,".");
  if (menu) {
    prefix=(char *)malloc(strlen(menu)+strlen(" '' ")+1);
    if (prefix) sprintf(prefix,"%s '' ",menu);
  }else{
    prefix=(char *)malloc(strlen("")+1);
    if (prefix) strcpy(prefix,"");
  }
  
  if (strm_src(IOS)==SRC_MENU) the_menu=Make_DirNameList(prefix,twd);

  k=(Token *)NULL;
  for(loop=TRUE;loop;) {
    loop=FALSE;
    k=free_Token(k);
    sprintf(prmt,"[%s] %s",twd,prompt);
    k=Get_String_MW(prmt,sep,the_menu,FALSE);
    if (k->status==TOK_EOL && wait_flag &&it==0) {
      if (the_menu) free(the_menu);
      the_menu=Make_DirNameList(prefix,twd);
      k=Get_String_MW(prmt,sep,the_menu,TRUE);
    }
    wait_flag=FALSE; it++;
    if (k->status==TOK_OK) {
      strcpy(ffn,twd);
      newtwd(ffn,k->text);
      status=stat(ffn,&sbuff);
      if (status<0) {
	k->status=TOK_NOK;
      }else{
	if (sbuff.st_mode&S_IFDIR) {
	  strcpy(twd,ffn);
	  if (k->source==SRC_MENU) {
	    if (the_menu) free(the_menu);
	    the_menu=Make_DirNameList(prefix,twd);
	  }
	  loop=TRUE;
	}else k->status=TOK_NOK;
      }
    }
  }

  if (k) {
    if (k->status==TOK_EOL) {
      if (k->text) free(k->text);
      k->text=strdup(twd);
      k->status=TOK_OK;
    }
  }

  if (the_menu) free(the_menu);
  if (prefix) free(prefix);

  return(k);
}
/*-------------------- Get_Fildes_MW --------------------*/

Token *Get_Fildes_MW(char *dir,char *ext,char *prompt,char *sep,char *menu,int wait_flag)
{
  /* use open(2) instead of UCpopen, can't read from pipe, 
     otherwise the same as Get_File_MW ... */
  /* 22-Jan-2002: obsolete ? */

  Token *k;
  char *the_menu,*srch,buf[4096],twd[MAXPATHLEN],ffn[MAXPATHLEN],*prefix,*p,*q,*r,*fext;
  int do_dir,loop,status;
  struct stat sbuff;
  
  the_menu=(char *)NULL;
  prefix=(char *)NULL;
  srch=(char *)NULL;

  if (dir) sprintf(twd,"%s",dir);
  else sprintf(twd,".");
  fext=(char *)NULL;
  if (ext) {
    p=ext;q=ext; r=buf; *r='\0';
    while ((p=strchr(q,'|'))) {
      *r++='*';
      strncpy(r,q,p-q); r+=p-q;
      q=p+1; if(strlen(q)>0) *r++='|';
      *r='\0';
    }
    if (strlen(q)>0) {
      p=q+strlen(q); *r++='*';
      strncpy(r,q,p-q); r+=p-q;
      *r='\0';
    }
    fext=strdup(ext); if (strchr(fext,'|')) *strchr(fext,'|')='\0';
  }else sprintf(buf,"*");
  srch=Strdup(buf);
  if (menu) {
    prefix=(char *)malloc(strlen(menu)+strlen(" ' ' -searchmask -directories ' ' ")+1);
    if (prefix) sprintf(prefix,"%s ' ' -searchmask -directories ' ' ",menu);
  }else{
    prefix=(char *)malloc(strlen("-searchmask -directories ' '")+1);
    if (prefix) sprintf(prefix,"-searchmask -directories ' '");
  }
  do_dir=FALSE;
  the_menu=Make_DirFileNameList(prefix,twd,ext?srch:"",do_dir);
  
  k=(Token *)NULL;
  for(loop=TRUE;loop;) {
    loop=FALSE;
    k=free_Token(k);
    k=Get_String_MW(prompt,sep,the_menu,wait_flag);
    if (k->status==TOK_OK) {
      strcpy(ffn,twd);
      newtwd(ffn,k->text);
      if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
      status=stat(ffn,&sbuff);
      if (status<0) {
	k=Prs_Kwd(k,the_menu);
	if (k->status==TOK_OK) {
	  k=Prs_Kwd(k,menu);
	  if (k->status==TOK_OK) {
	    k->status=TOK_NOK;
	  }else{
	    k=Prs_Kwd(k,"-searchmask -directories");
	    if (k->status==TOK_OK) {
	      if (streq(k->key,"-searchmask")) {
		default_token(srch);
		k=free_Token(k);
		k=Get_String("search mask: ",";");
		if (k->status==TOK_OK) {
		  if (srch) free(srch);
		  srch=k->text;
		  k->text=(char *)NULL;
		  if (the_menu) free(the_menu);
		  the_menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
		  loop=TRUE;
		}
	      }else if (streq(k->key,"-directories")) {
		do_dir=!do_dir;
		if (the_menu) free(the_menu);
		the_menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
		loop=TRUE;
	      }else{
		k->status=TOK_NOK;
	      }
	    }else{
	      if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	      k->n=open(extfil(buf,ffn,fext,FALSE),O_RDONLY);    
	      if (k->n!=-1) k->status=TOK_OK;
	      else k->status=TOK_NOK;
	    }
	  }
	}else{
	  if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	  k->n=open(extfil(buf,ffn,fext,FALSE),O_RDONLY);    
	  if (k->n!=-1) k->status=TOK_OK;
	  else k->status=TOK_NOK;
	}
      }else{
	if (sbuff.st_mode&S_IFDIR) {
	  strcpy(twd,ffn);
	  if (the_menu) free(the_menu);
	  the_menu=Make_DirFileNameList(prefix,twd,srch,do_dir);
	  loop=TRUE;
	}else{
	  if (strlen(ffn)>1 && ffn[strlen(ffn)-1]=='/') ffn[strlen(ffn)-1]='\0';
	  k->n=open(extfil(buf,ffn,fext,FALSE),O_RDONLY);    
	  if (k->n!=-1) k->status=TOK_OK;
	  else k->status=TOK_NOK;
	}
      }
    }
  }

  if (srch) free(srch);
  if (the_menu) free(the_menu);
  if (prefix) free(prefix);
  if (fext) free(fext);
  return(k);
}
/*-------------------- Get_AtomType --------------------*/

Token *Get_AtomType_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  char *the_menu;
  Atomdef *atp;
  Node *ResLib;

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  the_menu=Make_AtomTypeMenu(menu);
  k=Get_String_MW(prompt,sep,the_menu,wait_flag);
  if (k->status==TOK_OK) {
    ForAtomdefs(atp,ResLib) if (streq(atp->name,k->text)) break;
    k->ptr=(void *)atp;
    if (!is_atomdef(atp)) k->status=TOK_NOK;
  }
  if (the_menu) free(the_menu);
  return(k);
}

/*-------------------- Get_Color_MW --------------------*/

Token *Get_Color_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  char *lmenu,*cmenu,*colors;
  int i,ibase;
  Parse p;

  ibase=0;
  lmenu=(char *)malloc(((menu ? strlen(menu) : 0)+1024)*sizeof(char));
  if (lmenu) {
    if (menu) sprintf(lmenu,"%s ",menu);
    else *lmenu='\0';
    p.pos=0; p.sep=""; p.buf=lmenu; p.tok=(char *)NULL;
    while (get_token(&p)) ibase++;       /* number of menu_items */
    colors=lmenu+strlen(lmenu);
    for (i=1;i<WNP_MAXCOL;i++) {
      if (col_mnem[i]) sprintf(colors+strlen(colors),"%s ",col_mnem[i]);
      else sprintf(colors+strlen(colors),"%d ",i);
    }
    cmenu=lmenu;
  }else{
    cmenu=menu;
  }
    
  k=Get_Kwd_W(prompt,sep,cmenu,wait_flag);
  if (k->status==TOK_OK) {
    k->n=k->n-ibase+1;
  }else if (k->status==TOK_NOK) {
    k=Prs_Int(k);
    if (k->status==TOK_OK) {
      if (k->n<1||k->n>WNP_MAXCOL-1) k->status=TOK_NOK;
    }
  }

  if (lmenu) free(lmenu);
  return(k);
}
/*-------------------- Prs_Color --------------------*/

Token *Prs_Color(Token *k,int ibase) 
{
  int i;
  char colors[1024];

  if (k) {
    if (k->status==TOK_OK || k->status==TOK_NOK) {
      for(i=1,colors[0]='\0';i<WNP_MAXCOL;i++) {
	if (col_mnem[i]) sprintf(colors+strlen(colors),"%s ",col_mnem[i]);
	else sprintf(colors+strlen(colors),"%d ",i);
      }
      k=Prs_Kwd(k,colors);
      if (k->status==TOK_NOK) {
	k=Prs_Int(k);
	if (k->status==TOK_OK) {
	  if (k->n<1-ibase || k->n>=WNP_MAXCOL) k->status=TOK_NOK;
	}
      }else{
	k->n++;
      }
    }
  }
  return(k);
}
/*-------------------- Get_Float_MW --------------------*/

Token *Get_Float_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;

  k=Get_String_MW(prompt,sep,menu,wait_flag);
  return(Prs_Float(k));
}
/*-------------------- Prs_Float --------------------*/

Token *Prs_Float(Token *k)
{
/* Parse raw token k as float */

  int old;

  if (!k) return(k);
  if (k->status==TOK_OK || k->status==TOK_NOK) {
    old=err_seqnr();
    k->f=str2f(k->text);
    if (err_seqnr()!=old) {
      k->status=TOK_NOK;
      err_pop(1);
    }
    else k->status=TOK_OK;
  }
  return(k);
}
/*-------------------- Get_Float_MW --------------------*/

Token *Get_Int_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;

  k=Get_String_MW(prompt,sep,menu,wait_flag);
  return(Prs_Int(k));
}
/*-------------------- Prs_Int --------------------*/

Token *Prs_Int(Token *k)
{
/* Parse raw token k as int */

  int old;

  if (!k) return(k);
  if (k->status==TOK_OK || k->status==TOK_NOK) {
    old=err_seqnr();
    k->n=str2i(k->text);
    if (err_seqnr()!=old) {     /* very conservative, don't you think ? */
      k->status=TOK_NOK;
      err_pop(1);
    }
    else k->status=TOK_OK;
  }
  return(k);
}
/*-------------------- Get_XYZ_MW --------------------*/

Token *Get_XYZ_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;

  /*  get linr in 3D world space through picked point on screen... */

  pick_mode("PXYZ");
  k=Get_String_MW(prompt,sep,menu,wait_flag);
  pick_mode("POFF");
  return(Prs_XYZ(k));
}
/*-------------------- Prs_XYZ --------------------*/

Token *Prs_XYZ(Token *k) 
{
  Parse p;
  char buff[MAXTOKENSIZE],cf[MAXTOKENSIZE];
  float x[6];
  int i;

  if (!k) return(k);
  if (k->status==TOK_OK || k->status==TOK_NOK) {
    k->status=TOK_NOK;
    if (strlen(k->text)<=6) return(k);
    if (!strneq(k->text,"#PXYZ(",6)) return(k);
    Strcpy(buff,k->text+6);
    if (strrchr(buff,')')) *strrchr(buff,')')='\0';
    else return(k);
    errno=0;
    p.sep="";p.buf=buff;p.pos=0;p.mts=MAXTOKENSIZE;p.tok=cf;
    for (i=0;i<6&&token(&p);i++) {
      x[i]=str2f(p.tok);
      if (errno) return(k);
    }
    k->ptr=(void *)malloc(6*sizeof(float));
    if (!mvld(k->ptr)) return(k);
    for (i=0;i<6;i++) *((float *)(k->ptr)+i)=x[i];
    k->status=TOK_OK;
  }
  return(k);
}
/*-------------------- Get_Kwd_W --------------------*/

Token *Get_Kwd_W(char *prompt,char *sep,char *keylist,int wait_flag)
{
  Token *k;
  
  k=Get_String_MW(prompt,sep,keylist,wait_flag);
  return(Prs_Kwd(k,keylist));
}
/*-------------------- Prs_Kwd --------------------*/

Token *Prs_Kwd(Token *k,char *keylist)
{
/* Parse a keyword, given a raw token with status OK or NOK */

  char buffer[MAXTOKENSIZE];

  if (!k) return(k);
  if (k->status==TOK_OK || k->status==TOK_NOK) {
    k->n=kwd_match(k->text,keylist,buffer);
    if (k->n>=0) {
      if (mvld(k->key)) free(k->key);
      k->key=Strdup(buffer);
      if (!k->key) k->status=TOK_NOK;
      else k->status=TOK_OK;
    }else{
      k->status=TOK_NOK;
    }
    if (k->quoted) k->status=TOK_NOK;
  }
  return(k);
}
/*-------------------- Get_Mol_MW --------------------*/

Token *Get_Mol_MW(char *prompt,char *sep,char *menu,int wait_flag) 
{
  Token *k;
  Node *m;
  
  pick_mode("PATM");
  k=Get_String_MW(prompt,sep,menu,wait_flag);
  pick_mode("POFF");

  if (!k) return(k);
  if (k->status!=TOK_OK) return(k);
  
  k->status=TOK_NOK;
  m=name_to_nodeptr(k->text,strcasecmp);
  if (m) {
    if (m->type == OBJ_ATOM) {
      for (m=m->up;m;m=m->up) if (m->type==OBJ_MOL) break;
    }
  }
  if (m) {
    k->ptr=(void *)m;
    k->status=TOK_OK;
  }
  return(k);
}
/*-------------------- Get_Mols_DMW --------------------*/

Token *Get_Mols_DMW(char *prompt,char *sep,char *mmenu,int wait)
{
  /* get mols, with default menu (mols in default node, ops) */

  Token *k;
  char *menu;

  menu=Make_TopMolNameList((mmenu?mmenu:""),(mvld(DefaultNode) ? DefaultNode : MolRoot));
  k=Get_Mols_MW(prompt,sep,menu,wait);
  if (mvld(menu)) free(menu);
  send_bheader(graphix,"ARMK",0);
  return(k);
}

/*-------------------- Get_Mol_DMW --------------------*/

Token *Get_Mol_DMW(char *prompt,char *sep,char *amenu,int wait)
{
  /* get 1 mol, with default menu, prefixed by 'amenu'  (mols in default node, ops) */

  Token *k;
  char *menu;
  Nodechain *c;

  menu=Make_TopMolNameList((amenu?amenu:""),(mvld(DefaultNode) ? DefaultNode : MolRoot));
  k=Get_Mols_MW(prompt,sep,menu,wait);
  if (k->status==TOK_OK) {
    c=(Nodechain *)k->ptr;
    if (c->n==1) {
      k->ptr=(void *)c->s;
    }else{
      k->ptr=(void *)NULL;
      k->status=TOK_NOK;
    }
    clear_node_chain(c);
    free_node_chain(c);
  }
  if (mvld(menu)) free(menu);
  send_bheader(graphix,"ARMK",0);
  return(k);
}
/*-------------------- Get_Mols_MW --------------------*/

Token *Get_Mols_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  
  pick_mode("PATM");
  k=Get_String_MW(prompt,sep,menu,wait_flag);
  pick_mode("POFF");
  return(Prs_Mols_M(k,menu));
}
/*-------------------- Prs_Mol_M --------------------*/

Token *Prs_Mol_M(Token *k,char *menu)
{
  Nodechain *c;

  if (k) {
    k=Prs_Mols_M(k,menu);
    if (k->status==TOK_OK) {
      c=(Nodechain *)k->ptr;
      if (c->n==1) k->ptr=c->s;
      else k->status=TOK_NOK;
      clear_node_chain(c);
      free_node_chain(c);
    }
  }
  return(k);
}
/*-------------------- Prs_Mols_M --------------------*/

Token *Prs_Mols_M(Token *k,char *menu)
{
  Node *m,*s,*t,*T;
  Nodechain *c;
  int is_wild,lev,LEV;
  Nmatch *nmatch;
  Parse p;
  char nn[MAXTOKENSIZE];

  if (!k) return(k);
  if (k->status==TOK_OK || k->status==TOK_NOK ) {
    k->status=TOK_NOK;
    c=make_node_chain();
    if (mvld(c)) {
      is_wild=strpbrk(k->text,"*?+#%")!=(char *)NULL;
      m=pick_to_node(k->text);
      if (mvld(m)) {
	if (menu) {
	  p.sep="";p.buf=menu;p.pos=0;p.mts=MAXTOKENSIZE;p.tok=nn;
	  LEV=100000; T=(Node *)NULL;
	  while(get_token(&p)) {
	    s=name_to_nodeptr(p.tok,strcasecmp);
	    if (is_mol(s)) {
	      for(t=m,lev=0;mvld(t)&&lev<LEV;t=t->up,lev++) {
		if (t==s) {
		  T=t;
		  LEV=lev;
		  break;
		}
	      }
	    }
	  }
	  if (mvld(T)) add_to_node_chain(T,c);
	}
      }else if (is_wild) {
	nmatch=open_nmatch(k->text);
	for (m=next_nmatch(nmatch);mvld(m);m=next_nmatch(nmatch)) {
	  if (is_mol(m)) add_to_node_chain(m,c); 
	}
	close_nmatch(nmatch);	
      }else{
	m=name_to_nodeptr(k->text,strcasecmp);
	if (mvld(m)) {
	  if (is_mol(m)) add_to_node_chain(m,c);
	}
      }
      if (c->n<1) {
	free_node_chain(c);
	k->status=TOK_NOK;
      }else{
	k->ptr=c;
	k->status=TOK_OK;
      }
    }
  }
  return(k);
}

/*-------------------- Get_AtomCtxt_MW --------------------*/

Token *Get_AtomCtxt_MW(Node *ctxt,char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;

  pick_mode("PATM");
  k=Get_String_MW(prompt,sep,menu,wait_flag);
  pick_mode("POFF");
  
  if (!k) return(k);
  if (k->status!=TOK_OK) return(k);

  return(Prs_AtomCtxt(k,ctxt));

}
/*-------------------- Get_Atom_MW --------------------*/

Token *Get_Atom_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  Nodechain *atoms;

  pick_mode("PATM");
  k=Get_String_MW(prompt,sep,menu,wait_flag);
  pick_mode("POFF");
  
  if (!k) return(k);
  if (k->status!=TOK_OK) return(k);

  k=Prs_Kwd(k,"'-(sel)'");
  if (k->status==TOK_OK) {
    k=free_Token(k);
    k=Get_Atoms_D(prompt,sep);
    if (k->status==TOK_OK) {
      atoms=k->ptr;
      if (atoms->n==1) {
	k->ptr=atoms->s;
      }else{
	if (atoms->n>1) err_set(1,"atom selection is not unique.");
	else err_set(1,"atom selection is empty.");
	k->status=TOK_NOK;
	k->ptr=NULL;
      }
      clear_node_chain(atoms);
      free_node_chain(atoms);
    }
  }else{
    k=Prs_Atom(k);
  }

  return(k);
}
/*-------------------- Prs_AtomCtxt --------------------*/

Token *Prs_AtomCtxt(Token *k,Node *ctxt)
{
  Node *m;

  if (!k) return(k);
  if (k->status==TOK_OK || k->status==TOK_NOK ) {
    k->status=TOK_NOK;
    m=node_in_ctxt(k->text,ctxt,strcasecmp);
    if (m) {
      if (m->type == OBJ_ATOM) {
	k->ptr=(void *)m;
	k->status=TOK_OK;
      }
    }
  }
  return(k);
}
/*-------------------- Get_Bond_MW --------------------*/

Token *Get_Bond_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *r;
  char pr[256],token_string[512],tks1[200],tks2[200];
  Atom *f,*t;
  int i;

  sprintf(pr,"%s (1st atom) ",prompt);
  r=Get_Atom_MW(pr,sep,menu,wait_flag);
  if (r->status==TOK_OK) {
    Strcpy(tks1,r->text);
    f=(Atom *)r->ptr;
    free_Token(r);
    sprintf(pr,"%s (2nd atom) ",prompt);
    r=Get_Atom_MW(pr,sep,menu,wait_flag);
    r->n=1;
    if (r->status==TOK_OK) { /* should translate pick! */
      r->n=2;
      Strcpy(tks2,r->text); if (mvld(r->text)) free(r->text);
      sprintf(token_string,"%s %s",tks1,tks2);
      r->text=Strdup(token_string);
      t=(Atom *)r->ptr; i=f->nbnd;
      if (t!=f) for (i=0;i<f->nbnd;i++) if (t==f->bnd[i]->f||t==f->bnd[i]->t) break;
      if (i<f->nbnd) {
	r->ptr=(void *)f->bnd[i];
	if (f->bnd[i]->f!=f) {    /* swap (f,t) if necessary */
	  f->bnd[i]->f=f;
	  f->bnd[i]->t=t;
	}
      }else{
	r->ptr=(void *)NULL;
	r->status=TOK_NOK;
      }
    }
  }
  return(r);
}

/*-------------------- Get_nAtoms_MW --------------------*/
int Get_nAtoms_MW(int n,Token *atm[],char *pr,char *sp,char *mn,int wf)
{
  int got;
  char *suffix[4];
  char prompt[80];
  
  suffix[0]="st"; suffix[1]="nd";
  suffix[2]="rd"; suffix[3]="th";
  
  for (got=0;got<n;got++) atm[got]=(Token *)NULL;
  got=0;
  
  while (got<n) {
    if (n==1) {
      sprintf(prompt,"%s (atom) ",pr);
    }else{
      sprintf(prompt,"%s (%d%s atom) ",pr,got+1,suffix[got > 3 ? 3 : got]);
    }
    atm[got]=Get_Atom_MW(prompt,sp,mn,wf);
    if (!atm[got]) break;
    if (atm[got]->status) break;
    got++;
  }
  return(got);
}


/*-------------------- Get_Atoms_D --------------------*/

#define THEMENU " -done -and -not -or -atom -set -pzone -rzone -zone -connected -mol -up "

Token *Get_Atoms_D(char *prompt,char *sep)
{
  /* get atoms, with default menu (mols in default node, ops) */

  Token *k;
  char *menu=(char *)NULL;
  char *prefix = "* /* '' " THEMENU " ''";

  menu=Make_TopMolNameList(prefix,(mvld(DefaultNode) ? DefaultNode : MolRoot));
  k=Get_Atoms_M(prompt,sep,menu);

  if (mvld(menu)) free(menu);
  send_bheader(graphix,"ARMK",0);
  return(k);
}
/*-------------------- Get_Atoms_MW --------------------*/

Token *Get_Atoms_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  char *keys= 
   "-done -and -atom -connected -mol -up -not -or -set -pzone -rzone -zone";

  char *akeys="-done -and -atom -connected -mol -up -not -or -pzone -rzone -zone";

  char *smsmenu=
    "' ' ' ' ' ' " THEMENU " ''";

  char *cpsmenu=
    "' ' ' ' ' ' " THEMENU " ''"
    "t!comparison: 't! ' "
    "-lt -le '' -eq -neq '' ~eq ~neq '' -ge -gt";

  char *tpsmenu=
    "'' '' '' " THEMENU " ''"
    "t!type: 't! ' "
    "*.\\? unk Du du* '' "
    "C.\\? H.\\? N.\\? O.\\? P.\\? S.\\? '' "
    "C* H* N* O* P* S* '' "
    "*.m";

  char *chsmenu=
    "' ' ' ' ' ' " THEMENU " ''"
    "t!symbol: 't! ' "
    "C H N O S P ' ' F Cl Br I ' ' Du Lp ' ' "
    "Ac Ag Al Ar As At Au ' ' B Ba Be Bi Br ' ' "
    "Ca Cd Ce Cl Co Cr Cs Cu ' ' D Du Dy ' ' "
    "Er Eu ' ' F Fe Fr ' ' Ga Gd Ge ' ' H He Hf Hg Ho ' ' "
    "I In Ir ' ' K Kr ' ' La Li Lp Lu ' ' Mg Mn Mo ' ' "
    "N Na Nb Nd Ne Ni ' ' O Os ' ' P Pa Pb Pd Pm Po Pr Pt ' ' "
    "Ra Rb Re Rh Rn Ru ' ' S Sb Sc Se Si Sm Sn Sr ' ' "
    "Ta Tb Tc Te Th Ti Tl Tm ' ' U V W Xe Y Yb ' ' Zn Zr";

  char *anmenu=
    "' ' ' ' ' ' " THEMENU " ''"
    "t!name: 't! ' "
    "N,CA,C,O HOH*/ ' '  N CA C O ' ' C* H* N* O* S* ' ' ?A* ?B* ?G* ?D* ?E* ?Z* ?H*";

  char *apmenu = "biso charge ligance mass name nlocs q seqnr shift sof symbol type '' "  
    "fragment matches substructure";

  Token *atoms,*k,*sets;
  int i,j,n1=0,n2=0,loop,dtok,dofrg,wf,resclo,keep,op,ival,dispatch,imaxd,up=0;
  int amask;
  Nodechain *new,*old,*s;
  Node *n;
  Atom *a;
  float maxd,val=0,tol=0,pval;
  char Prompt[1024],*par=(char *)NULL,*prmt;
  Atom  **T1=NULL,**T2=NULL;
  Prule *pr;

  wf=FALSE; dtok=FALSE;
  atoms=make_Token();
  loop=(atoms->status!=TOK_FAIL);
  new=make_node_chain();
  atoms->ptr=(void *)new;
  if (!new) {
    free_Token(atoms);
    atoms= &ErrorToken;
    return(atoms);
  }
  k=(Token *)NULL;

  sprintf(Prompt,"%s",prompt);

  old=set_mark_atoms(new);

  do {
    keep=FALSE;
    if (!mvld(k)) {                /* get new token, if necessary... */
      err_dump();
      pick_mode("PATM");
      sprintf(Prompt,"[%d atoms selected] %s",new->n,prompt);
      k=Get_String_MW(Prompt,sep,menu,FALSE); 
      pick_mode("POFF");
    }
    switch (k->status) {
    case TOK_EOL:
      if (!dtok&&!wf&&(!wait_flag||atoms->status==TOK_OK)) {
	if (atoms->status!=TOK_OK) atoms->status=TOK_EOL;
	loop=FALSE;
      }
      wf=FALSE;
      dtok=FALSE;
      break;
    case TOK_EOC:
      if (atoms->status!=TOK_OK) atoms->status=k->status;
      loop=FALSE;
      break;
    case TOK_OK:
      dtok=FALSE;
      k=Prs_Kwd(k,keys);
      if (k->status==TOK_OK) {
	if (!streq(k->key,"-up")) up=0;

	/* parse -zone and -rzone */

	if (streq(k->key,"-zone")||streq(k->key,"-rzone")) {
	  resclo=streq(k->key,"-rzone");
	  k=free_Token(k);
	  k=Get_Float_M("radius = ",";",
			" 0 1 2 3 4 5 6 7 8 9 10 15 20 30 40 50 100");
	  if (k->status==TOK_OK) {
	    maxd=k->f;
	    T1=(Atom **)chain_to_ptr(new,FALSE); n1=new->n;
	    clear_node_chain(new);
	    free_node_chain(new); 
	    new=(Nodechain *)NULL;
	    set_mark_atoms(new);
	    k=free_Token(k);
	    send_bheader(graphix,"ARMK",0);
	    k=Get_Atoms_MW("atoms at center of zone: ",sep,menu,wait_flag);
	    send_bheader(graphix,"ARMK",0);
	    if (k->status==TOK_OK) {
	      new=(Nodechain *)k->ptr;
	      set_mark_atoms(new);
	      T2=(Atom **)chain_to_ptr(new,FALSE); n2=new->n;
	      clear_node_chain(new);
	      atoms->status=TOK_OK;
	      if (mark_close(n1,T1,n2,T2,maxd)) {
		for (i=0;i<n1;i++) {
		  if (T1[i]->flg&(1<<0)) {
		    add_to_node_chain((Node *)T1[i],new);
		  }
		  T1[i]->flg&=~(1<<0|1<<7);
		}
		if (resclo) residue_closure(new);
		resclo=FALSE;
	      }else{
		if (n1>0 && !mvld(T1)) {
		  err_set(1,"failed to generate BHtree(1)");
		  atoms->status=TOK_NOK;
		  loop=FALSE;
		}
		if (n2>0 && !mvld(T2)) {
		  err_set(1,"failed to generate BHtree(2)");
		  atoms->status=TOK_NOK;
		  loop=FALSE;
		}
	      }
	      if (mvld(T2)) free(T2);
	    }else{
	      atoms->status=k->status;
	      loop=FALSE;
	    }
	    if (mvld(T1)) free(T1);
	  }else{
	    atoms->status=k->status;
	    loop=FALSE;
	  }
	/* parse -pzone */

	}else if (streq(k->key,"-pzone")) {
	  k=free_Token(k);
	  k=Get_Int_M("radius = ",";",
			" 1 2 3 4 5 6 7 8 9 10 15 20 30 40 50 100");
	  if (k->status==TOK_OK) {
	    imaxd=k->n;
	    T1=(Atom **)chain_to_ptr(new,FALSE); n1=new->n;
	    clear_node_chain(new);
	    free_node_chain(new); 
	    new=(Nodechain *)NULL;
	    set_mark_atoms(new);
	    k=free_Token(k);
	    send_bheader(graphix,"ARMK",0);
	    k=Get_Atoms_MW("atoms at center of zone: ",sep,menu,wait_flag);
	    send_bheader(graphix,"ARMK",0);
	    if (k->status==TOK_OK) {
	      new=(Nodechain *)k->ptr;
	      set_mark_atoms(new);
	      pconnected_atoms(new,imaxd);
	      if (mvld(T1)) {
		for (j=0,i=0;i<n1;i++) {
		  T1[j]=T1[i];
		  if (node_in_chain((Node *)T1[i],new)) j++;
		}
		n1=j;
		clear_node_chain(new);
		for (i=0;i<n1;i++) add_to_node_chain((Node *)T1[i],new);
	      }else{
		if (n1>0) {
		  err_set(1,"memory allocation failure.");
		  atoms->status=TOK_NOK;
		  loop=FALSE;
		}else{
		  clear_node_chain(new);
		}
	      }
	    }else{
	      atoms->status=k->status;
	      loop=FALSE;
	    }
	    if (mvld(T1)) free(T1);
	  }else{
	    atoms->status=k->status;
	    loop=FALSE;
	  }
	}else if (streq(k->key,"-connected")) {

	  /* parse -connected */

	  connected_atoms(new);

	}else if (streq(k->key,"-atom")) {

	  /* parse -atom */

	  n1=new->n;
	  T1=(Atom **)chain_to_ptr(new,FALSE);
	  if (n1>0 && !T1) {
	    err_set(1,"parser error: memory allocation failure.");
	    atoms->status=TOK_NOK;
	    loop=FALSE;
	  }
	  clear_node_chain(new);
	  send_bheader(graphix,"ARMK",0);
	  k=free_Token(k);
	  k=Get_Kwd("atom property: ",";",apmenu);
	  if (k->status==TOK_OK) {

	    dtok=TRUE;
	    if (streq(k->key,"name")) {

	      /* -atom name */

	      dtok=FALSE;
	      for (wf=TRUE;loop;wf=FALSE) {
		k=free_Token(k);
		sprintf(Prompt,"[%d atoms selected] atom name: ",new->n);
		k=Get_String_MW(Prompt,";",anmenu,wf);
		if (k->status==TOK_OK) {
		  k=Prs_Kwd(k,akeys);
		  if (k->status==TOK_NOK) {
		    if (mvld(T1)) {
		      for (i=0;i<n1;i++) 
			if (MatchNodeName((Node *)T1[i],k->text)) 
			  add_to_node_chain((Node *)T1[i],new);
		    }
		    set_mark_atoms(new);
		  }else{
		    if (!streq(k->key,"-done")) keep=TRUE;
		    break;
		  }
		}else if (k->status==TOK_NOK) {
		  err_set(1,"bad token: %s",k->text);
		  atoms->status=TOK_NOK;
		  dtok=TRUE;
		}else if (k->status==TOK_EOL) {
		  break;
		}else if (k->status==TOK_EOC) {
		  break;
		}else {
		  clear_node_chain(new);
		  free_node_chain(new);
		  new=ptr_to_chain((Node **)T1,n1);
		  set_mark_atoms(new);
		  dtok=TRUE;
		  break;
		}
	      }
	    }else if (streq(k->key,"symbol")) {

	      /* -atom symbol */

	      dtok=FALSE;
	      for (wf=TRUE;loop;wf=FALSE) {
		k=free_Token(k);
		sprintf(Prompt,"[%d atoms selected] chemical symbol: ",new->n);
		k=Get_String_MW(Prompt,";",chsmenu,wf);
		if (k->status==TOK_OK) {
		  k=Prs_Kwd(k,akeys);
		  if (k->status==TOK_NOK) {
		    if (mvld(T1)) {
		      for (i=0;i<n1;i++) {
			if (mvld(T1[i]->atp)) {
			  if (mwild_xmatch(T1[i]->atp->chs,k->text)) 
			    add_to_node_chain((Node *)T1[i],new);
			}
		      }
		    }
		    set_mark_atoms(new);
		  }else{
		    if (!streq(k->key,"-done")) keep=TRUE;
		    break;
		  }
		}else if (k->status==TOK_NOK) {
		  err_set(1,"bad token: %s",k->text);
		  atoms->status=TOK_NOK;
		  dtok=TRUE;
		}else if (k->status==TOK_EOL) {
		  break;
		}else if (k->status==TOK_EOC) {
		  break;
		}else {
		  clear_node_chain(new);
		  free_node_chain(new);
		  new=ptr_to_chain((Node **)T1,n1);
		  set_mark_atoms(new);
		  dtok=TRUE;
		  break;
		}
	      }
	    }else if (streq(k->key,"type")) {

	      /* -atom type */

	      dtok=FALSE;
	      for (wf=TRUE;loop;wf=FALSE) {
		k=free_Token(k);
		sprintf(Prompt,"[%d atoms selected] atom type: ",new->n);
		k=Get_String_MW(Prompt,";",tpsmenu,wf);
		if (k->status==TOK_OK) {
		  k=Prs_Kwd(k,akeys);
		  if (k->status==TOK_NOK) {
		    if (mvld(T1)) {
		      for (i=0;i<n1;i++) {
			if (mvld(T1[i]->atp)) {
			  if (mwild_xmatch(T1[i]->atp->name,k->text)) 
			    add_to_node_chain((Node *)T1[i],new);
			}
		      }
		    }
		    set_mark_atoms(new);
		  }else{
		    if (!streq(k->key,"-done")) keep=TRUE;
		    break;
		  }
		}else if (k->status==TOK_NOK) {
		  err_set(1,"bad token: %s",k->text);
		  atoms->status=TOK_NOK;
		  dtok=TRUE;
		}else if (k->status==TOK_EOL) {
		  break;
		}else if (k->status==TOK_EOC) {
		  break;
		}else {
		  clear_node_chain(new);
		  free_node_chain(new);
		  new=ptr_to_chain((Node **)T1,n1);
		  set_mark_atoms(new);
		  dtok=TRUE;
		  break;
		}
	      }
	    }else if (streq(k->key,"matches")||
		      streq(k->key,"fragment")||
		      streq(k->key,"substructure")) {

	      /* -atom fragnment & -atom matches */

	      dtok=FALSE;
	      dofrg=streq(k->key,"fragment")||streq(k->key,"substructure");
	      amask=(streq(k->key,"substructure")?(1<<7):0);
	      for (i=0;i<n1;i++) T1[i]->flg|=amask;
	      for (wf=TRUE;loop;wf=FALSE) {
		k=free_Token(k);
		sprintf(Prompt,"[%d atoms selected] fragment: ",new->n);
		k=Get_String_MW(Prompt,";",smsmenu,wf);
		if (k->status==TOK_OK) {
		  k=Prs_Kwd(k,akeys);
		  if (k->status==TOK_NOK) {
		    if (mvld(T1)) {
		      pr=swilesPrule(k->text,PR_SUBST);
		      if (mvld(pr)) {
			for (i=0;i<n1;i++) {
			  if (matchPruleAtom(T1[i],pr,amask,0)) {  /* change here to implement strict rules */
			    add_to_node_chain((Node *)T1[i],new);
			    if (dofrg) {
			      for (j=1;j<pr->nat;j++) 
				add_to_node_chain((Node *)pr->atom[j].match,new);
			    }
			  }
			}
			del_node((Node *)pr);
		      }else err_set(1,"failed to parse fragment definition: %s",k->text);
		    }
		    
		    set_mark_atoms(new);
		  }else{
		    if (!streq(k->key,"-done")) keep=TRUE;
		    break;
		  }
		}else if (k->status==TOK_NOK) {
		  err_set(1,"bad token: %s",k->text);
		  atoms->status=TOK_NOK;
		  dtok=TRUE;
		}else if (k->status==TOK_EOL) {
		  break;
		}else if (k->status==TOK_EOC) {
		  break;
		}else {
		  clear_node_chain(new);
		  free_node_chain(new);
		  new=ptr_to_chain((Node **)T1,n1);
		  set_mark_atoms(new);
		  dtok=TRUE;
		  break;
		}
	      }
	      for (i=0;i<n1;i++) T1[i]->flg&=~amask;
	    }else if ((k=Prs_Kwd(k,"biso mass q shift sof"))->status==TOK_OK) {

	      /* -atom biso, mass, q, shift, sof ... (floating point values) */

	      dtok=FALSE;
	      par=strdup(k->key);
	      op=REL_EQ;
	      prmt="comparison operator or value: ";
	      for (dispatch=FALSE,wf=TRUE;loop;wf=FALSE) {
		k=free_Token(k);
		sprintf(Prompt,"[%d atoms selected] %s",new->n,prmt);
		k=Get_String_MW(Prompt,";",cpsmenu,wf);
		if (k->status==TOK_OK) {
		  if ((k=Prs_Kwd(k,"-lt -le '' -eq -neq '' ~eq ~neq '' -ge -gt"))->status==TOK_OK) {
		    op=(streq(k->key,"-lt")   ? REL_LT  :
			streq(k->key,"-le")   ? REL_LE  :
			streq(k->key,"-eq")   ? REL_EQ  :
			streq(k->key,"-neq")  ? REL_NE  :
			streq(k->key,"~eq")   ? REL_TEQ :
			streq(k->key,"~neq")  ? REL_TNE :
			streq(k->key,"-ge")   ? REL_GE  :
			streq(k->key,"-gt")   ? REL_GT  :
			REL_NOP);
		    prmt="value: ";
		  }else if ((k=Prs_Float(k))->status==TOK_OK) {
		    val=k->f;
		    dispatch=TRUE;
		    if (op==REL_TEQ||op==REL_TNE) {
		      k=free_Token(k);
		      k=Get_Float("tolerance: ",";");
		      if (k->status==TOK_OK) {
			tol=k->f;
			if (tol<0) tol=-tol;
		      }else dispatch=FALSE;
		    }
		    prmt="comparison operator or value: ";
		  }else if ((k=Prs_Kwd(k,akeys))->status==TOK_OK) {
		    keep=TRUE;
		    break;
		  }else break;
		  if (dispatch) {
		    dispatch=FALSE;
		    if (mvld(T1)) {
		      for (i=0;i<n1;i++) {
			if (streq(par,"biso")) pval=T1[i]->x[ATM_BISO];
			else if (streq(par,"q")) pval=T1[i]->q;
			else if (streq(par,"sof")) pval=T1[i]->x[ATM_SOF];
			else if (streq(par,"shift")) {
			  if (T1[i]->shift>0.9*WNP_NO_SHIFT) pval=T1[i]->shift;
			  else continue;                  /* compare with missing value is always false */
			}else if (streq(par,"mass")) {
			  if (mvld(T1[i]->atp)) pval=T1[i]->atp->mass;
			  else continue;
			}else continue;
			if (
			    (op==REL_EQ  && pval==val) ||
			    (op==REL_NE  && pval!=val) ||
			    (op==REL_LT  && pval<val)  ||
			    (op==REL_LE  && pval<=val) || 
			    (op==REL_GT  && pval>val)  ||   
			    (op==REL_GE  && pval>=val) ||
			    (op==REL_TEQ && pval>=val-tol&&pval<=val+tol) ||
			    (op==REL_TNE && (pval<val-tol||pval>val+tol)) )
			  add_to_node_chain((Node *)T1[i],new);
		      }
		    }
		    set_mark_atoms(new);
		  }
		}else if (k->status==TOK_EOL) {
		  break;
		}else if (k->status==TOK_EOC) {
		  break;
		}else break;
	      }
	      if (par) {free(par);par=(char *)NULL;}
	      if (k->status==TOK_NOK) {
		err_set(1,"bad token: %s",k->text);
		atoms->status=TOK_NOK;
		clear_node_chain(new);
		free_node_chain(new);
		new=ptr_to_chain((Node **)T1,n1);
		set_mark_atoms(new);
		dtok=TRUE;
	      }else if (k->status==TOK_EOL) {
	      }else if (k->status==TOK_EOC) {
	      }else if (k->status!=TOK_OK) {
		clear_node_chain(new);
		free_node_chain(new);
		new=ptr_to_chain((Node **)T1,n1);
		set_mark_atoms(new);
		dtok=TRUE;
		break;
	      }
	    }else if ((k=Prs_Kwd(k,"charge ligance seqnr nlocs"))->status==TOK_OK) {

	      /* -atom charge ligance seqnr... (integer values) */

	      dtok=FALSE;
	      par=strdup(k->key);
	      op=REL_EQ;
	      prmt="comparison operator or value: ";
	      for (dispatch=FALSE,wf=TRUE;loop;wf=FALSE) {
		k=free_Token(k);
		sprintf(Prompt,"[%d atoms selected] %s",new->n,prmt);
		k=Get_String_MW(Prompt,";",cpsmenu,wf);
		if (k->status==TOK_OK) {
		  if ((k=Prs_Kwd(k,"-lt -le '' -eq -neq '' ~eq ~neq '' -ge -gt"))->status==TOK_OK) {
		    op=(streq(k->key,"-lt")   ? REL_LT  :
			streq(k->key,"-le")   ? REL_LE  :
			streq(k->key,"-eq")   ? REL_EQ  :
			streq(k->key,"-neq")  ? REL_NE  :
			streq(k->key,"~eq")   ? REL_TEQ :
			streq(k->key,"~neq")  ? REL_TNE :
			streq(k->key,"-ge")   ? REL_GE  :
			streq(k->key,"-gt")   ? REL_GT  :
			REL_NOP);
		    prmt="value: ";
		  }else if ((k=Prs_Float(k))->status==TOK_OK) {
		    val=k->f;
		    dispatch=TRUE;
		    if (op==REL_TEQ||op==REL_TNE) {
		      k=free_Token(k);
		      k=Get_Float("tolerance: ",";");
		      if (k->status==TOK_OK) {
			tol=k->f;
			if (tol<0) tol=-tol;
		      }else dispatch=FALSE;
		    }
		    prmt="comparison operator or value: ";
		  }else if ((k=Prs_Kwd(k,akeys))->status==TOK_OK) {
		    keep=TRUE;
		    break;
		  }else break;
		  if (dispatch) {
		    dispatch=FALSE;
		    if (mvld(T1)) {
		      if (streq(par,"seqnr")) walk_order(MolRoot,n1,T1,(1<<0));
		      for (i=0;i<n1;i++) {
			if (streq(par,"charge")) ival=T1[i]->chrg;
			else if (streq(par,"ligance")) ival=T1[i]->lig;
			else if (streq(par,"nlocs")) ival=T1[i]->nxyz;
			else if (streq(par,"seqnr")) ival=i+1;
			else continue;
			if ((op==REL_EQ  && ival==val) ||
			    (op==REL_NE  && ival!=val) ||
			    (op==REL_LT  && ival<val)  ||
			    (op==REL_LE  && ival<=val) || 
			    (op==REL_GT  && ival>val)  ||   
			    (op==REL_GE  && ival>=val) ||
			    (op==REL_TEQ && ival>=val-tol&&ival<=val+tol) ||
			    (op==REL_TNE && (ival<val-tol||ival>val+tol)) )
			  add_to_node_chain((Node *)T1[i],new);
		      }
		      set_mark_atoms(new);
		    }
		  }
		}else if (k->status==TOK_EOL) {
		  break;
		}else if (k->status==TOK_EOC) {
		  break;
		}else break;
	      }
	      if (par) {free(par);par=(char *)NULL;}
	      if (k->status==TOK_NOK) {
		err_set(1,"bad token: %s",k->text);
		atoms->status=TOK_NOK;
		clear_node_chain(new);
		free_node_chain(new);
		new=ptr_to_chain((Node **)T1,n1);
		set_mark_atoms(new);
		dtok=TRUE;
	      }else if (k->status==TOK_EOL) {
	      }else if (k->status==TOK_EOC) {
	      }else if (k->status!=TOK_OK) {
		clear_node_chain(new);
		free_node_chain(new);
		new=ptr_to_chain((Node **)T1,n1);
		set_mark_atoms(new);
		dtok=TRUE;
		break;
	      }
	    }else{
	      
	      /* unimplemented atom property... */

	      err_set(1,"sorry, this is not implemented (work in progress...)");
	      clear_node_chain(new);
	      free_node_chain(new);
	      new=ptr_to_chain((Node **)T1,n1);
	      set_mark_atoms(new);
	    }
	  }else if (k->status==TOK_NOK) {

	    /* invalid atom property ... */

	    err_set(1,"unknown atom property: %s",k->text);
	    atoms->status=TOK_NOK;
	    loop=FALSE;
	  }else{

	    /* done with atom property... */

	    clear_node_chain(new);
	    free_node_chain(new);
	    new=ptr_to_chain((Node **)T1,n1);
	    set_mark_atoms(new);
	  }
	  if (mvld(T1)) free(T1);
	  T1=(Atom **)NULL;
	}else if (streq(k->key,"-and")) {

	  /* parse -and */

	  n1=new->n;
	  T1=(Atom **)chain_to_ptr(new,FALSE);
	  if (n1>0 && !T1) err_set(1,"parser error: memory allocation failure.");
	  clear_node_chain(new);
	  k=free_Token(k);
	  send_bheader(graphix,"ARMK",0);
	  k=Get_Atoms_MW("and atoms: ",sep,menu,wait_flag);
	  send_bheader(graphix,"ARMK",0);
	  if (k->status==TOK_OK) {
	    s=(Nodechain *)k->ptr;
	    if (mvld(T1)) {
	      for (i=0;i<n1;i++) T1[i]->flg|=(1<<0);
	      for (a=(Atom *)s->s;mvld(a);a=(Atom *)n) {
		n=a->chain;
		a->chain=(Node *)NULL;
		if (a->flg&(1<<0)) add_to_node_chain((Node *)a,new);
	      }
	      for (i=0;i<n1;i++) T1[i]->flg&=~(1<<0);
	    }else clear_node_chain(s);
	    free_node_chain(s);
	  }else{
	    atoms->status=TOK_NOK;
	    loop=FALSE;
	  }
	  if (mvld(T1)) free(T1);
	  T1=(Atom **)NULL;
	}else if (streq(k->key,"-or")) {

	  /* parse -or */

	  n1=new->n;
	  T1=(Atom **)chain_to_ptr(new,FALSE);
	  if (n1>0 && !T1) err_set(1,"parser error: memory allocation failure.");
	  clear_node_chain(new);
	  k=free_Token(k);
	  send_bheader(graphix,"ARMK",0);
	  k=Get_Atoms_MW("or atoms: ",sep,menu,wait_flag);
	  send_bheader(graphix,"ARMK",0);
	  if (k->status==TOK_OK) {
	    s=(Nodechain *)k->ptr;
	    *new=*s;
	    free_node_chain(s);
	  }else{
	    atoms->status=TOK_NOK;
	    loop=FALSE;
	  }
	  if (mvld(T1)) {
	    for (i=0;i<n1;i++) add_to_node_chain((Node *)T1[i],new);
	  }
	  if (mvld(T1)) free(T1);
	  T1=(Atom **)NULL;
	  
	}else if (streq(k->key,"-mol")) {
	  
	  mol_atoms(new);

	}else if (streq(k->key,"-up")) {
	  
	  up_atoms(new,++up);

	}else if (streq(k->key,"-not")) {

	  /* parse -not */

	  n1=new->n;
	  T1=(Atom **)chain_to_ptr(new,FALSE);
	  if (n1>0 && !T1) err_set(1,"parser error: memory allocation failure.");
	  clear_node_chain(new);
	  k=free_Token(k);
	  send_bheader(graphix,"ARMK",0);
	  k=Get_Atoms_MW("not atoms: ",sep,menu,wait_flag);
	  send_bheader(graphix,"ARMK",0);
	  if (k->status==TOK_OK) {
	    s=(Nodechain *)k->ptr;
	    if (mvld(T1)) {
	      for (i=0;i<n1;i++) {
		if (!node_in_chain((Node *)T1[i],s)) add_to_node_chain((Node *)T1[i],new);
	      }
	    }
	    clear_node_chain(s);
	    free_node_chain(s);
	  }else{
	    atoms->status=TOK_NOK;
	    loop=FALSE;
	  }
	  if (mvld(T1)) free(T1);
	  T1=(Atom **)NULL;
	}else if (streq(k->key,"-set")) {

	  /* parse -set */

	  sets=Get_Sets_D("atomset: ",";");
	  if (sets->status==TOK_OK) {
	    atoms->status=TOK_OK;
	    s=(Nodechain *)sets->ptr;
	    for (n=s->s;mvld(n);n=n->chain) add_atomset_to_chain((Atomset *)n,new);
	    clear_node_chain(s);
	    free_node_chain(s);
	  }else if (sets->status==TOK_NOK) {
	    err_set(1,"bad set name : %s",sets->text);
	  }else if (sets->status==TOK_EOC) loop=FALSE;
	  sets=free_Token(sets);
	}else if (streq(k->key,"-done")) {
	  loop=FALSE;
	}
      }else{

	/* if string does not start with '-', suppose it is an atom... */

	if (*k->text=='-') {
	  atoms->status=TOK_NOK;
	  err_set(1,"syntax error in atom expression: %s",k->text);
	  loop=FALSE;
	}else{
	  atoms->status=TOK_OK;
	  atom_chain(new,k->text);
	}
      }
      break;
    case TOK_NULL:
      break;
    case TOK_ABORT:
      atoms->status=TOK_ABORT;
      loop=FALSE;
      break;
    default: 
      break;
    }
    if (!keep) k=free_Token(k);
  }while (loop);

  if (atoms->status!=TOK_OK) {
    if (mvld(new)) {
      clear_node_chain(new);
      free_node_chain(new);
    }
    atoms->ptr=(void *)NULL;
  }else{
    atoms->n=new->n;
    atoms->ptr=(void *)new;
  }

  set_mark_atoms(old);
  return(atoms);
}
/*-------------------- Get_Bonds_D --------------------*/

Token *Get_Bonds_D(char *prompt,char *sep)
{
  /* zero-order approximation to bonds routine, does only intra bonds */

  Token *k;
  Bond *b;
  Nodechain *c;

  k=Get_Atoms_D(prompt,sep);
  send_bheader(graphix,"ARMK",0);
  if (k->status==TOK_OK) {
    c=(Nodechain *)k->ptr;
    k->ptr=AtomToBondChain(c);
    for (k->n=0,b=(Bond *)k->ptr;b;b=b->chain) k->n++;
    clear_node_chain(c);
    free_node_chain(c);
  }
  return(k);
}
/*-------------------- Get_Bonds_MW --------------------*/

Token *Get_Bonds_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  /* zero-order approximation to bonds routine, does only intra bonds */

  Token *k;
  Bond *b;
  Nodechain *c;

  k=Get_Atoms_MW(prompt,sep,menu,wait_flag);
  if (k->status==TOK_OK) {
    c=(Nodechain *)k->ptr;
    k->ptr=AtomToBondChain(c);
    for (k->n=0,b=(Bond *)k->ptr;b;b=b->chain) k->n++;
    clear_node_chain(c);
    free_node_chain(c);
  }
  return(k);
}
/*-------------------- Get_DspGlbl_D --------------------*/

Token *Get_DspGlbl_D(char *prompt,char *sep,char *prefix)
{
  char *glblnames=(char *)NULL;
  Token *tok=(Token *)NULL;

  glblnames=Make_GlblNameList(prefix,DspRoot);
  if (glblnames) tok=Get_DspGlbl_M(prompt,sep,glblnames);
  else tok=Get_DspGlbl_M(prompt,sep,prefix);
  if (glblnames) free(glblnames);
  return(tok);
}

/*-------------------- Get_DspGlbl_MW --------------------*/

Token *Get_DspGlbl_MW(char *prompt,char *sep,char *menu,int wait_flag)
{
  Token *k;
  Dspobj *m;
  Node *n;
  char buff[MAXTOKENSIZE];
  cptr_t cross;
  int i;
  unsigned int b[8];
  
  pick_mode("POBJ");
  k=Get_String_MW(prompt,sep,menu,wait_flag);
  pick_mode("POFF");
  
  if (!k) return(k);
  if (k->status!=TOK_OK) return(k);
  
  if (strneq(k->text,"#POBJ(",6)) {
    k->status=TOK_NOK;
    Strcpy(buff,k->text+6);
    if (strchr(buff,')')) *strchr(buff,')')='\0';
    else return(k);
    k->n=sscanf(buff,"%x %x %x %x %x %x %x %x",b,b+1,b+2,b+3,b+4,b+5,b+6,b+7);
    if (k->n>8) k->n=8;
    for (i=0;i<k->n;i++) cross.b[i]=b[i];
    for (;i<sizeof(cross);i++) cross.b[i]=0;
    ForDspobjs (m,DspRoot) {
      if (!memcmp(&((Dspobj *)m)->cross,&cross,sizeof(cross))) {
	k->ptr=(void *)m;
	k->status=TOK_OK;
	break;
      }
    }
  }else{
    n=name_to_nodeptr(k->text,strcasecmp);
    if (n) {
      k->ptr=(void *)NULL;
      if (n->type==OBJ_DSP) {
	/* k->ptr=((Dspobj *)n)->cross; */
	err_set(1,"internal error, parser.c(Get_DspGlbl_MW)");
      }else if (n->type==OBJ_MOL) {
	k->ptr=(void *)((Mol *)n)->dsptr;
      }
      if (k->ptr) k->status=TOK_OK;
      else k->status=TOK_NOK;
    }else{
      k->status=TOK_NOK;
      k->ptr=(void *)NULL;
    }
  }
  if (k->status==TOK_OK) k->ptr=(void *)glblobj((Dspobj *)k->ptr);
  return(k);
}  
/*-------------------- Get_Sets_D --------------------*/

Token *Get_Sets_D(char *prompt,char *sep)
{
  char menu[16392];
  Token *k;

  *menu='\0'; Make_SetMenu(MolRoot,menu);
  k=Get_String_M(prompt,sep,menu);
  if (k->status==TOK_OK) {
    k->ptr=make_node_chain();
    if (mvld(k->ptr)) {
      *(Nodechain *)k->ptr=atomset_chain(k->text);
    }else{
      k->status=TOK_NOK;
      err_set(1,"memory allocation failure.");
    }
  }
  return(k);
}
/*==================== Utilities ====================*/

int init_parser(void)
{
  if (srvString) {
    sscanf(srvString,"%f",&srvVersion);
  }
  state=(ParseState *)malloc(sizeof(ParseState));
  if (!state) return(0);
  IOS=open_stream((char *)NULL,MAX_Parse_BUFFER);
  state->nxt=(ParseState *)NULL;
  state->prompt=(char *)NULL;
  state->deftoken=(char *)NULL;
  state->eoc=FALSE;
  state->eol=FALSE;
  if (IOS) {
    set_stream(IOS,"redirection='<',continuation='\\',spctok=on,expvar=on");
    strm_parse(IOS)->mts=MAXTOKENSIZE;
    strm_parse(IOS)->tok=tok;
    return(1);
  }
  else return(0);
}

void start_parse(void)
{
  if (mvld(state)) {
    state->eoc=FALSE;
    state->eol=FALSE;
    if (mvld(state->deftoken)) free(state->deftoken);
    state->deftoken=(char *)NULL;
  }
}

void default_token(char *format,...)
{
  char buffer[1000];
  va_list args;

  if (!state) return;
  if (mvld(state->deftoken)) free(state->deftoken);
  state->deftoken=(char *)NULL;
  if (format) {
    buffer[0]='\0';
    va_start(args,format);
    vsprintf(buffer,format,args);
    va_end(args);
    if (strlen(buffer)>0) state->deftoken=Strdup(buffer);
  }
}

void prtGraphics(char *format,...)
{
  static char *kp=(char *)NULL;
  char buffer[65536],*p,*q,sv;
  va_list args;
  int l;

  if (kp) {                    /* left-over from last call */
    sprintf(buffer,"%s",kp);
    free(kp);
    kp=(char *)NULL;
  }else{
    buffer[0]='\0';
  } 
  va_start(args,format);
  vsprintf(buffer+strlen(buffer),format,args);
  va_end(args);
  p=buffer;
  while ((q=strchr(p,'\n'))) { /* chop into \n-terminated bits */
    sv=q[1]; q[1]='\0';
    wnpRPCprintResponse(0,p);
    if (graphix) send_message(graphix,"*MSG",strlen(p),p);
    else fprintf(stdout,"%s",p);
    if (redirFile) fprintf(redirFile,"%s",p);
    q[1]=sv;
    p=q+1;
  }
  if ((l=strlen(p))>0) {
    if (l<32768) kp=strdup(p); /* keep non-\n-terminated bit if small enough */
    else {                     /* else dump it... */
      wnpRPCprintResponse(0,p);
      if (graphix) send_message(graphix,"*MSG",strlen(p),p);
      else fprintf(stdout,"%s",p);
      if (redirFile) fprintf(redirFile,"%s",p);
    }
  }
}

void redirectPrtStream(int op,char *filename)
{
  if (redirFile) {
    fclose(redirFile);
    redirFile=(FILE *)NULL;
  }
  switch (op) {
  case RED_NEW:
    redirFile=fopen(filename,"w");
    if (!redirFile) {
      err_set(1,"redirect output, opening new file '%s' for write failed.\n",filename);
    }
    break;
  case RED_APP:
    redirFile=fopen(filename,"a");
    if (!redirFile) {
      err_set(1,"redirect output, opening new file '%s' for append failed.\n",filename);
    }
    break;
  case RED_OLD:
    if (strlen(fullname)>0) redirFile=fopen(fullname,"a");
    if (!redirFile) {
      err_set(1,"redirect output, opening old file '%s' for append failed.\n",fullname);
    }
    break;
  case RED_END:
    break;
  default:
    break;
  }
  if (op==RED_NEW||op==RED_APP) {
    if (!realpath(filename,fullname)) {
      err_set(1,"redirect output stream, failed to translate '%s' to full path name\n",
	      filename);
      fullname[0]='\0';
    }
  }
  err_dump();
}

int end_parse(void)
{
  int r;

  r=TOK_OK;
  if (!state) return(r);
  while(!state->eoc && !state->eol) {
    if (token_stream(IOS,";")) r=TOK_NOK;
    state->eol=(strm_term(IOS)=='\n');
    state->eoc=(strm_term(IOS)==';');
  }
  return(r);
}

int push_parse(char *filename)
{
  ParseState *new;
  Stream *ns;

  if (!state) return(0);
  new=(ParseState *)malloc(sizeof(ParseState));
  if (!new) return(0);
  ns=push_stream(IOS,filename);
  if (!ns) {
    if (mvld(new)) free(new);
    return(0);
  }
  new->nxt=state;
  new->prompt=(char *)NULL;
  new->deftoken=(char *)NULL;
  new->eoc=FALSE;
  new->eol=TRUE;
  state=new;
  return(1);
}
  

void pop_parse(void)
{
  ParseState *new;

  if (!state || !IOS) return;
  if (IOS->lev>0) pop_stream(IOS);
  if (state->nxt) {
    new=state->nxt;
    if (mvld(state->deftoken)) free(state->deftoken);
    if (mvld(state->prompt))  free(state->prompt);
    if (mvld(state)) free(state);
    state=new;
  }
}

int parser_level(void)
{
  return(IOS->lev);
}

char *get_prompt(char *buffer)
{
  /* edit current prompt with default into buffer, return ptr to buffer */
  /* note: sufficent size of buffer is caller's responsibility ! */

  if (state->deftoken) {
    if (state->prompt) {
      sprintf(buffer,"%s[%s] ",state->prompt,state->deftoken);
    }else{
      sprintf(buffer,"[%s] ",state->deftoken);
    }
  }else strcpy(buffer,(state->prompt?state->prompt:""));
  return(buffer);
}

void set_prompt(char *s)
{
  if (!state) return;
  if (mvld(state->prompt)) free(state->prompt);
  if (s) state->prompt=Strdup(s);
  else state->prompt=(char *)NULL;
}

void do_prompt(void)
{
  char buffer[4096];
  Atom *a;
  int n;

  do_pick("default");
  if (!state) return;
  if (graphix>0 && mark_atoms) {
    for (n=0,a=(Atom *)mark_atoms->s;mvld(a);a=(Atom *)a->chain) {
      if (cPtr(a->cross)&&a->dsp) n++;
    }
    if (n>0) {
      lock_evtq();
      send_bheader(graphix,"ASMK",n*cPtrSize());
      for (a=(Atom *)mark_atoms->s;mvld(a);a=(Atom *)a->chain) {
	if (cPtr(a->cross)&&a->dsp) send_bcptr(graphix,a->cross);
      }
      unlock_evtq();
    }
  }
  get_prompt(buffer);
  if (server) {
    send_bmessage(graphix,"PRMT",strlen(buffer),buffer);
  }else{
    fprintf(stdout,"%s",buffer);
  }
}
    
void handle_usap(cptr_t cross)
{
  Atom *x;
  
  char name[MAXTOKENSIZE];
  
  pick_mode("EUSA");                           /* re-enable usa-picking... */
  ForAtoms (x,MolRoot) {
    if (!memcmp(&(x->cross),&cross,sizeof(cross))) {
      nodeptr_to_name(name,(Node *)x);
      send_bmessage(graphix,"*MSG",strlen(name),name);
      break;
    }
  }
}

Token *make_Token(void)
{
  Token *r;

  r=(Token *)malloc(sizeof(Token));
  if (r) {
    r->text=(char *)NULL;
    r->key=(char *)NULL;
    r->ptr=(void *)NULL;
    r->n= -1;
    r->f=0.0;
    r->source=SRC_NULL;
    r->status=TOK_NULL;
    r->quoted=FALSE;
    r->def=FALSE;
  }else{
    r= &ErrorToken;
  }
  return(r);
}

Token *free_Token(Token *k)
{
/* Note: always returns pointer to null token */

  if (k==&ErrorToken) return((Token *)NULL);
  if (mvld(k)) {
    if (mvld(k->key)) free(k->key);
    if (mvld(k->text)) free(k->text);
    if (k->status==TOK_FAIL) return((Token *)NULL);
    free(k);
  }
  return((Token *)NULL);
}
/*-------------------- ft_defext --------------------*/

char *ft_defext(int ft)
{
  int i;

  for (i=0;i<MAX_FT;i++) if (ft==FT[i]) break;
  if (i<MAX_FT) return(DEFEXT[i]);
  else return((char *)NULL);
}

/*-------------------- ft_searchmask --------------------*/

char *ft_searchmask(int ft)
{
  int i;

  for (i=0;i<MAX_FT;i++) if (ft==FT[i]) break;
  if (i<MAX_FT) return(SRCH[i]);
  else return((char *)NULL);
}

int ft_number(char *ft_text)
{
  int i;

  for (i=0;i<MAX_FT;i++) if (streq(ft_text,ft_menu_item[i])) break;
  if (i<MAX_FT) return(FT[i]);
  else return(FT_NONE);
}

int ft_numberKW(char *ft_text)
{
  char *L;
  int r=-1;

  L=Make_FileTypeList("",FT_ALL_TYPES);
  if (L) {
    r=kwd_match(ft_text,L,NULL);
    free(L);
  }
  if (r>=0) r=FT[r];
  else r=FT_NONE;

  return(r);
}

/*-------------------- AtomsInNode --------------------*/

Nodechain *AtomsInNode(Nodechain *c,char *expr,Node *r)
{
  /* Build atom chain from expression, restricted to given root Node */
  /* Atoms are added to chain *c, returns pointer to updated chain.  */

  Node *s;

  s=DefaultNode;
  DefaultNode=r;
  atom_chain(c,expr);
  DefaultNode=s;
  return(c);
}

/*-------------------- atom_chain --------------------*/

Nodechain *atom_chain(Nodechain *c,char *expr)
{
  /* Build atom chain from expression */

  Nmatch *nmatch;
  Atom *a;
  Node *obj;

  if (c) {
    nmatch=open_nmatch(expr);
    if (nmatch) {
      obj=next_nmatch(nmatch);
      while (obj) {
	switch (obj->type) {
	case OBJ_MOL :
	  ForAtoms(a,obj) add_to_node_chain((Node *)a,c);
	  break;
	case OBJ_ATOM :
	  add_to_node_chain(obj,c);
	  break;
	case OBJ_ATOMSET :
	  add_atomset_to_chain((Atomset *)obj,c);
	  break;
	default: break;
	}
	obj=next_nmatch(nmatch);
      }
      close_nmatch(nmatch);
    }
  }
  return(c);
}
/*-------------------- atomset_chain --------------------*/

Nodechain atomset_chain(char *expr)
{
/* Build atomset chain from expression */

  Nodechain c;
  Nmatch *nmatch;
  Node *obj;

  c.n=0;
  c.s=(Node *)NULL;
  c.e=(Node *)NULL;
  c.p=(Node *)NULL;

  nmatch=open_nmatch(expr);
  if (!nmatch) {
    c.n= -1;
    return(c);
  }

  obj=next_nmatch(nmatch);
  while (obj) {
    switch (obj->type) {
    case OBJ_ATOMSET :
      add_to_node_chain(obj,&c);
      break;
    default: break;
    }
    obj=next_nmatch(nmatch);
  }
  close_nmatch(nmatch);
  return(c);
}
/*-------------------- Make_GlblNameList --------------------*/

char *Make_GlblNameList(char *prefix,Dspobj *root) 
{
  int l,n;
  Dspobj *d;
  Node *m;
  char buff[1024];
  char *r;

  l=0; n=0;
  ForDspobjs(d,root) {
    if (d->dsptype==DSP_GLBL) {
      m=d->root;
      if (mvld(m)) {
	l+=strlen(nodeptr_to_name(buff,m));
	n++;
      }
    }
  }

  l=l+n+strlen(prefix)+1;
  r=(char *)malloc((l+1)*sizeof(char));
  if (mvld(r)) {
    l=sprintf(r,"%s ",prefix);
    ForDspobjs(d,root) {
      if (d->dsptype==DSP_GLBL) {
	m=d->root;
	l+=sprintf(r+l,"%s ",nodeptr_to_name(buff,m));
      }
    }
  }

  return(r);
}

/*-------------------- Make_TopMolNameList --------------------*/

char *Make_TopMolNameList(char *prefix,Node *root)
{
  int l,n;
  Mol *m;
  char buff[1024];
  char *r;

  l=0;n=0;
  forMols(m,root) {
    l+=strlen(nodeptr_to_name(buff,(Node *)m));
    n++;
  }

  l=l+n+strlen(prefix)+1;
  r=(char *)malloc((l+1)*sizeof(char));
  l=sprintf(r,"%s ",prefix);
  if (mvld(r)) {
    l=sprintf(r,"%s ",prefix);
    forMols(m,root) {
      l+=sprintf(r+l,"%s ",nodeptr_to_name(buff,(Node *)m));
    }
  }
  return(r);
}
/*-------------------- Make_NodeNameList --------------------*/

char *Make_NodeNameList(char *prefix,Nodechain *c)
{
  int l,n;
  Node *m;
  char *r;

  l=0;n=0;
  for (m=c->s;mvld(m);m=m->chain) {
    l+=strlen(m->name);
    n++;
  }

  l=l+n+strlen(prefix)+1;
  r=(char *)malloc((l+1)*sizeof(char));
  if (mvld(r)) {
    l=sprintf(r,"%s ",prefix);
    for (m=c->s;mvld(m);m=m->chain) {
      l+=sprintf(r+l,"%s ",m->name);
    }
  }
  return(r);
}
/*-------------------- Make_FileNameList --------------------*/

char *Make_FileNameList(char *prefix,char *dir,char *sel)
{
  int n,i,l;
  char **names;
  char *r;

  r=(char *)NULL;
  n=DirList(dir,sel,S_IFREG,&names);

  for (l=1+strlen(prefix),i=0;i<n;i++) l+=strlen(names[i])+1;
  r=(char *)malloc((l+1)*sizeof(char));
  if (mvld(r)) {
    for (l=sprintf(r,"%s ",prefix),i=0;i<n;i++) {
      l+=sprintf(r+l,"%s ",names[i]);
    }
  }

  for (i=0;i<n;i++) free(names[i]);
  if (n>0) free(names);

  return(r);
}
/*-------------------- Make_DirNameList --------------------*/

char *Make_DirNameList(char *prefix,char *dir)
{
  int n,i,l;
  char **names;
  char *r,*sel;

  r=(char *)NULL; sel="*";
  n=DirList(dir,sel,S_IFDIR,&names);

  for (l=1+strlen(prefix),i=0;i<n;i++) l+=strlen(names[i])+2;
  r=(char *)malloc((l+1)*sizeof(char));
  if (mvld(r)) {
    for (l=sprintf(r,"%s ",prefix),i=0;i<n;i++) {
      l+=sprintf(r+l,"%s/ ",names[i]);
    }
  }

  for (i=0;i<n;i++) free(names[i]);
  if (n>0) free(names);

  return(r);
}
/*-------------------- Make_DirFileNameList --------------------*/

char *Make_DirFileNameList(char *prefix,char *dir,char *sel,int idir)
{
  char *r,*d;

  /* generate file name list, include directories, iff idir==TRUE */

  r=(char *)NULL;
  if (idir) {
    d=Make_DirNameList(prefix,dir);
    if (d) {
      r=Make_FileNameList(d,dir,sel);
      free(d);
    }
  }else{
    r=Make_FileNameList(prefix,dir,sel);
  }
  return(r);
}
/*-------------------- Make_FileTypeList --------------------*/

char *Make_FileTypeList(char *prefix,int types) 
{
  int i,l;
  char *r;

  l=0;
  for (i=0;i<MAX_FT;i++) if (types&FT[i]) l+=strlen(ft_menu_item[i])+1;
  l+=strlen(prefix)+1;
  r=(char *)malloc((l+1)*sizeof(char));
  if (mvld(r)) {
    for (l=sprintf(r,"%s ",prefix),i=0;i<MAX_FT;i++) {
      if (types&FT[i]) l+=sprintf(r+l,"%s ",ft_menu_item[i]); 
    }
  }
  return(r);
}
/*-------------------- set_mark_atoms --------------------*/

Nodechain *set_mark_atoms(Nodechain *atoms)
{
  Nodechain *old;

  old=mark_atoms;
  mark_atoms=atoms;
  return(old);
}
/*-------------------- ColMnem --------------------*/

char *ColMnem(int nr)
{
  if (nr>=0 && nr <WNP_MAXCOL) {
    if (mvld(col_mnem[nr])) return(col_mnem[nr]);
    else return(col_nmmnem[nr]);
  }else return(char *)NULL;
}

/*-------------------- ColName --------------------*/

char *ColName(int nr)
{
  if (nr>=0 && nr <WNP_MAXCOL) {
    if (mvld(col_mnem[nr])) return(col_mnem[nr]);
    else return(NULL);
  }else return(char *)NULL;
}



/*-------------------- ColVal --------------------*/

int ColVal(int nr,int comp)
{
  if (nr>=0 && nr<WNP_MAXCOL) {
    if (comp<0 || comp >2) return(0);
    else return(col_val[nr][comp]);
  }else if (nr==-1) {
    if (comp<0 || comp >2) return(0);
    else return(fog_val[comp]);
  }else return(0);
}

/*-------------------- DefineColor --------------------*/

void DefineColor(int nr,char *mnem,int R,int G,int B)
{
  int i;

  if (nr>=0 && nr<WNP_MAXCOL) {
    if (mvld(col_mnem[nr])) free(col_mnem[nr]);
    if (mvld(mnem)) col_mnem[nr]=Strdup(mnem);
    else col_mnem[nr]=(char *)NULL;
    col_val[nr][0]=R; col_val[nr][1]=G; col_val[nr][2]=B;
    if (graphix) {
      if (nr==0) {
	send_bheader(graphix,"BGRD",3*sizeof(int));
	for (i=0;i<3;i++) send_bint(graphix,col_val[nr][i]);
      }else{
	send_bheader(graphix,"CLDF",4*sizeof(int)+strlen(ColMnem(nr)));
	send_bint(graphix,nr);
	for (i=0;i<3;i++) send_bint(graphix,col_val[nr][i]);
	send_bdata(graphix,strlen(ColMnem(nr)),ColMnem(nr));
      }
    }
  }else if (nr==-1) {
    fog_val[0]=R; fog_val[1]=G; fog_val[2]=B; 
    if (graphix) {
      send_bheader(graphix,"FOGC",3*sizeof(int));
      for (i=0;i<3;i++) send_bint(graphix,fog_val[i]);
    }
  }
}
/*-------------------- Make_SetMenu --------------------*/

void Make_SetMenu(Node *root,char *menu)
{
  char buf[4096];
  int k,l;
  Mol *m;
  Atomset *s;

  l=0; k=0;
  forAtomsets(s,root) {
    sprintf(menu+l," %s",nodeptr_to_name(buf,(Node *)s));
    l+=strlen(buf)+1;
    k++;
  }
  if (k>1) {
    if (mvld(root->up)) {
      nodeptr_to_name(buf,root);
      sprintf(menu+l," %s/*",buf);
    }else sprintf(menu+l," /*");
  }
/*  if (k>0) sprintf(menu+strlen(menu)," ' '"); */
  forMols(m,root) Make_SetMenu((Node *)m,menu+strlen(menu));
}
/*-------------------- Make_AtomdefTypeMenu --------------------*/

char *Make_AtomTypeMenu(char *menu)
{
  char *out;
  Atomdef *atp;
  Node *ResLib;
  int l;

  if (menu) l=strlen(menu)+2;
  else l=1;

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  ForAtomdefs(atp,ResLib) l+=(mvld(atp->name) ? strlen(atp->name) + 1 : 0);
  out=(char *)malloc(l*sizeof(char));
  if (mvld(out)) {
    if (menu) l=sprintf(out,"%s ",menu);
    else l=0;
    ForAtomdefs(atp,ResLib) if (mvld(atp->name)) l+=sprintf(out+l,"%s ",atp->name);
  }
  return(out);
}
/*-------------------- Get_RuleLib --------------------*/

Token *Get_RuleLib_W(Node *root,char *prmt,char *sep,int wf)
{
  char buff[4096];
  Token *key;
  Node *ruleset;
  int i;

  if (!mvld(root)) {
    err_set(1,"Get_RuleLib: bad root.");
    return(&ErrorToken);
  }

  buff[0]='\0';
  for (i=0,ruleset=root->fst;mvld(ruleset);ruleset=ruleset->nxt) {
    if (is_lib(ruleset)) {
      if (mvld(ruleset->name)) {
	sprintf(buff+strlen(buff)," %s",ruleset->name);
	i++;
      }
    }
  }

  if (i==0) {
    err_set(1,"no automatic rules available.");
    return(&ErrorToken);
  }

  key=Get_Kwd_W(prmt,sep,buff,wf);
  if (key->status==TOK_OK) {
    ruleset=find_name(root,key->key,OBJ_LIB);   /* 25-Jan-2002: was key->text */
    if (!is_lib(ruleset)) {
      err_set(1,"ruleset %s not found.",key->text);
      key->status=TOK_NOK;
    }else key->ptr=ruleset;
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad token: %s\n",key->text);
  }
  return(key);
}
/*-------------------- Get_ObjectsR_MW --------------------*/

#define objnm(o) (mvld((o)->name)? (o)->name : (mvld((o)->root) ? (o)->root->name : (char *)NULL))

Token *Get_ObjectsR_MW(Dspobj *root,unsigned long type,char *prmt,char *sep,char *menu,int wf)
{
  Nodechain c,*f;
  Dspobj *o,*z,*m;
  Node *nxt;
  int l,i;
  unsigned int b[8];
  char *names,*nm,buff[1024];
  Token *key;
  cptr_t cross;

  key=make_Token(); key->status=TOK_NOK;
  names=(char *)NULL;
  l=4+(menu ? strlen(menu) : 0); 
  c.n=0;c.s=(Node *)NULL; c.e=c.s; c.p=c.s;

  root=(root?root:DspRoot);
  ForDspobjs(o,root) {
    if (o->dsptype&type) {
      nm=objnm(o);
      if (mvld(nm)) {
	l+=strlen(nm)+1;
	add_to_node_chain((Node *)o,&c);
      }
    }
  }
  if (c.n>0) {
    names=(char *)malloc(l*sizeof(char)+1);
    if (mvld(names)) {
      if (menu) sprintf(names,"%s * ",menu);
      else sprintf(names,"* ");
      for (z=(Dspobj *)c.s;mvld(z);z=(Dspobj *)z->chain) {
	if (mvld(objnm(z))) sprintf(names+strlen(names),"%s ",objnm(z));
      }
      key=free_Token(key);
      if (type&(DSP_MOL|DSP_GLBL)) pick_mode("POBJ");
      key=Get_String_MW(prmt,sep,names,wf);
      if (type&(DSP_MOL|DSP_GLBL)) pick_mode("POFF");
      if (key->status==TOK_OK) {
	if (strneq(key->text,"#POBJ(",6)) {
	  clear_node_chain(&c);
	  Strcpy(buff,key->text+6);
	  if (strchr(buff,')')) {
	    *strchr(buff,')')='\0';
	    key->n=sscanf(buff,"%x %x %x %x %x %x %x %x",b,b+1,b+2,b+3,b+4,b+5,b+6,b+7);
	    if (key->n>8) key->n=8;
	    for (i=0;i<key->n;i++) cross.b[i]=b[i];
	    for (;i<sizeof(cross);i++) cross.b[i]=0;
	  }else{
	    err_set(1,"bad token: %s",key->text);
	    key->n=-1;
	  }
	  f=(void *)make_node_chain();
	  key->ptr=f;
	  if (mvld(f)) {
	    ForDspobjs (m,root) {
	      if (!memcmp(&((Dspobj *)m)->cross,&cross,sizeof(cross))) {
		add_to_node_chain((Node *)m,f);
	      }
	    }
	    if (f->n==0) {
	      key->status=TOK_NOK;
	      key->ptr=(void *)NULL;
	      free_node_chain(f);
	    }
	  }else{
	    err_set(1,"memory allocation problem in 'get_objects'.");
	    key->status=TOK_NOK;
	  }
	}else{
	  f=(void *)make_node_chain();
	  key->ptr=f;
	  if (mvld(f)) {
	    for (z=(Dspobj *)c.s;mvld(z);z=(Dspobj *)nxt) {
	      nxt=z->chain; z->chain=(Node *)NULL;
	      if (wild_xmatch(objnm(z),key->text)) add_to_node_chain((Node *)z,f);
	    }
	    if (f->n==0) {
	      key->status=TOK_NOK;
	      key->ptr=(void *)NULL;
	      free_node_chain(f);
	    }
	  }else{
	    err_set(1,"memory allocation problem in 'get_objects'.");
	    clear_node_chain(&c);
	    key->status=TOK_NOK;
	  }
	}
      }else{
	clear_node_chain(&c);
      }
    }else{
      err_set(1,"memory allocation problem in 'get_objects'.");
      clear_node_chain(&c);
    }
  }else{
    err_set(1,"get_objects: no objects of specified type.");
    key->status=TOK_EOC;
  }

  if (mvld(names)) free(names);
  return(key);
}
/*-------------------- Get_ObjectR_MW --------------------*/

#define objnm(o) (mvld((o)->name)? (o)->name : (mvld((o)->root) ? (o)->root->name : (char *)NULL))

Token *Get_ObjectR_MW(Dspobj *root,unsigned long type,char *prmt,char *sep,char *menu,int wf)
{
  Nodechain c,*f;
  Dspobj *o,*z,*m;
  Node *nxt;
  int l,i;
  unsigned int b[8];
  char *names,*nm,buff[1024];
  Token *key;
  cptr_t cross;

  key=make_Token(); key->status=TOK_NOK;
  names=(char *)NULL;
  l=4+(menu ? strlen(menu) : 0); 
  c.n=0;c.s=(Node *)NULL; c.e=c.s; c.p=c.s;

  root=(root?root:DspRoot);
  ForDspobjs(o,root) {
    if (o->dsptype&type) {
      nm=objnm(o);
      if (mvld(nm)) {
	l+=strlen(nm)+1;
	add_to_node_chain((Node *)o,&c);
      }
    }
  }
  if (c.n>0) {
    names=(char *)malloc(l*sizeof(char)+1);
    if (mvld(names)) {
      if (menu) sprintf(names,"%s ",menu);
      else names[0]='\0';
      for (z=(Dspobj *)c.s;mvld(z);z=(Dspobj *)z->chain) {
	if (mvld(objnm(z))) sprintf(names+strlen(names),"%s ",objnm(z));
      }
      key=free_Token(key);
      if (type&(DSP_MOL|DSP_GLBL)) pick_mode("POBJ");
      key=Get_String_MW(prmt,sep,names,wf);
      if (type&(DSP_MOL|DSP_GLBL)) pick_mode("POFF");
      if (key->status==TOK_OK) {
	if (strneq(key->text,"#POBJ(",6)) {
	  clear_node_chain(&c);
	  Strcpy(buff,key->text+6);
	  if (strchr(buff,')')) {
	    *strchr(buff,')')='\0';
	    key->n=sscanf(buff,"%x %x %x %x %x %x %x %x",b,b+1,b+2,b+3,b+4,b+5,b+6,b+7);
	    if (key->n>8) key->n=8;
	    for (i=0;i<key->n;i++) cross.b[i]=b[i];
	    for (;i<sizeof(cross);i++) cross.b[i]=0;
	  }else{
	    err_set(1,"bad token: %s",key->text);
	    key->n=-1;
	  }
	  f=(void *)make_node_chain();
	  key->ptr=f;
	  if (mvld(f)) {
	    ForDspobjs (m,root) {
	      if (!memcmp(&((Dspobj *)m)->cross,&cross,sizeof(cross))) {
		add_to_node_chain((Node *)m,f);
	      }
	    }
	    if (f->n==1) {
	      key->ptr=(void *)f->s;
	    }else{
	      key->status=TOK_NOK;
	      key->ptr=(void *)NULL;
	    }
	    clear_node_chain(f);
	    free_node_chain(f);
	  }else{
	    err_set(1,"memory allocation problem in 'get_objects'.");
	    key->status=TOK_NOK;
	  }
	}else{
	  f=(void *)make_node_chain();
	  key->ptr=f;
	  if (mvld(f)) {
	    for (z=(Dspobj *)c.s;mvld(z);z=(Dspobj *)nxt) {
	      nxt=z->chain; z->chain=(Node *)NULL;
	      if (wild_xmatch(objnm(z),key->text)) add_to_node_chain((Node *)z,f);
	    }
	    if (f->n==1) {
	      key->ptr=(void *)f->s;
	    }else{
	      key->status=TOK_NOK;
	      key->ptr=(void *)NULL;
	    }
	    clear_node_chain(f);
	    free_node_chain(f);
	  }else{
	    err_set(1,"memory allocation problem in 'get_objects'.");
	    clear_node_chain(&c);
	    key->status=TOK_NOK;
	  }
	}
      }else{
	clear_node_chain(&c);
      }
    }else{
      err_set(1,"memory allocation problem in 'get_objects'.");
      clear_node_chain(&c);
    }
  }else{
    err_set(1,"get_object: no objects of specified type.");
    key->status=TOK_EOC;
  }

  if (mvld(names)) free(names);
  return(key);
}
/*-------------------- dotdotdot --------------------*/

void dotdotdot(char *pref,int k,int n,int m)
{
  /* write ... to entertain customer while CPU is grinding. */

  char buf[4096],*p;
  int i,a,b;

  if (server) {
    if (!(k==0&&n==0&&m==0)) {
      sprintf(buf,"%s",pref);
      if (k>0) {
	m=(m<1?n:m);
	m=(m>sizeof(buf)-strlen(pref)-1?sizeof(buf)-strlen(pref)-1:m);
	a=(k*m)/n;
	b=((k-1)*m)/n;
	if (a>b) {
	  for (p=buf+strlen(buf),i=0;i<a;i++) *p++='.';
	  *p='\0';
	  send_message(graphix,"PRMT",sizeof(buf),buf);
	}
      }else send_message(graphix,"PRMT",sizeof(buf),buf);
      
    }
  }else{
    if (k==0&&n==0&&m==0) fprintf(stdout,"\n");
    else {
      if (k==0) fprintf(stdout,"%s",pref);
      else{
	m=(m<1?n:m);
	a=(k*m)/n;
	b=((k-1)*m)/n;
	if (a>b) fprintf(stdout,".");
      }
    }
    fflush(stdout);
  }
}

/*-------------------- mol_atoms --------------------*/

static void mol_atoms(Nodechain *c)
{
  Nodechain R;
  Node *n;
  Dspobj *g;
  Atom *a;

  init_node_chain(&R);
  if (c) {
    for (n=c->s;n;n=n->chain) {
      if (is_atom(n)) {
	g=glblobj(assocobj(n));
	if (g) add_to_node_chain(g->root,&R);
      }
    }
    for (n=R.s;n;n=n->chain) ForAtoms(a,n) add_to_node_chain((Node *)a,c);
    clear_node_chain(&R);
  }
}
/*-------------------- up_atoms --------------------*/

static void up_atoms(Nodechain *c,int u)
{
  Nodechain R;
  Node *n,*m;
  int k;
  Atom *a;

  init_node_chain(&R);
  if (c) {
    for (m=c->s;m;m=m->chain) {
      if (is_atom(m)) {
	n=m;
	for (k=0;k<u;k++) {
	  if (n->up->up) {
	    n=n->up;
	    if (!n->up->up) break;
	  }
	}
	if (n) add_to_node_chain(n,&R);
      }
    }
    for (n=R.s;n;n=n->chain) ForAtoms(a,n) add_to_node_chain((Node *)a,c);
    clear_node_chain(&R);
  }
}
