/*   parser.h   */

#ifndef PARSERDEF
#define PARSERDEF
#include <stdio.h>
#include "object.h"
#include "token.h"
#include "strutil.h"
#include "streamio.h"

#define MAX_Parse_BUFFER 65535
#define TRUE 1

#define TOK_OK      0
#define TOK_EOL     1
#define TOK_EOC     2
#define TOK_NOK     3
#define TOK_ABORT   4
#define TOK_NULL    5
#define TOK_FAIL    6

#define ATM_UNION        1
#define ATM_INTERSECTION 2
#define ATM_DIFFERENCE   3

#define RED_OLD        1
#define RED_NEW        2
#define RED_APP        3
#define RED_END        4

typedef struct {
    char *text;
    char *key;
    void *ptr;
    double f;
    int n;
    int status;
    int source;
    int ltok;
    short quoted;
    short def;
    char term;
} Token;

/********************* Prototypes *********************/

Token  *make_Token(void);
Token  *Get_Kwd_W(char *prompt,char *sep,char *keylist,int wait_flag);
Token  *Prs_Kwd(Token *s,char *keys);
Token  *Get_RawLine_MW(char *prompt,char *menu,int wait_flag);
Token  *Get_String_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_AtomType_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Atom_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_AtomCtxt_MW(Node *ctxt,char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Bond_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Mol_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Mol_DMW(char *prompt,char *sep,char *menu,int wait);
Token  *Get_Mols_DMW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Mols_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Atoms_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Atoms_D(char *prompt,char *sep);
Token  *Get_Bonds_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Bonds_D(char *prompt,char *sep);
Token  *Get_DspGlbl_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_DspGlbl_D(char *prompt,char *sep,char *prefix);
Token  *Get_Sets_D(char *prompt,char *sep);
Token  *Get_Float_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Int_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_XYZ_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Color_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_File_MW(char *dir,char *ext,char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Directory_MW(char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_Fildes_MW(char *dir,char *ext,char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_ObjectsR_MW(Dspobj *r,unsigned long type,char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_ObjectR_MW(Dspobj *r,unsigned long type,char *prompt,char *sep,char *menu,int wait_flag);
Token  *Get_RuleLib_W(Node *root,char *prompt,char *sep,int wait_flag);
Token  *Prs_Float(Token *k);
Token  *Prs_Int(Token *k);
Token  *Prs_Mols_M(Token *k,char *menu);
Token  *Prs_Mol_M(Token *k,char *menu);
Token  *Prs_AtomCtxt(Token *k,Node *ctxt);
Token  *Prs_XYZ(Token *k);
Token  *Prs_Color(Token *k,int base);
Token  *free_Token(Token *s);
int     Get_nAtoms_MW(int n,Token *atm[],char *pr,char *sp,char *mn,int wf);
int     init_parser(void);
int     end_parse(void);
void    default_token(char *text,...);
void	set_prompt(char *s);
char   *get_prompt(char *buffer);
void    set_filename_select_string(char *sel);
void    do_prompt(void);
void    handle_usap(cptr_t a);
void    start_parse(void);
int     push_parse(char *file);
void    pop_parse(void);
int     parser_level(void);
Nodechain  *AtomsInNode(Nodechain *c,char *expr,Node *r);
Nodechain  *atom_chain(Nodechain *old,char *expr);
Nodechain  atomset_chain(char *expr);
char   *ft_defext(int ft);
char   *ft_searchmask(int ft);
int     ft_number(char *text);
int     ft_numberKW(char *ft_text);
char   *Make_GlblNameList(char *prefix,Dspobj *root);
char   *Make_TopMolNameList(char *prefix,Node *root);
char   *Make_NodeNameList(char *prefix,Nodechain *c);
char   *Make_FileNameList(char *prefix,char *dir,char *sel);
char   *Make_DirNameList(char *prefix,char *dir);
char   *Make_DirFileNameList(char *prefix,char *dir,char *sel,int idir);
char   *Make_FileTypeList(char *prefix,int types);
void    Make_SetMenu(Node *root,char *menu);
char   *Make_AtomTypeMenu(char *menu);
char   *ColMnem(int i);
char   *ColName(int i);
int     ColVal(int nr,int comp);
void    DefineColor(int nr,char *mnem,int R,int G,int B);
Stream *CurrStream(void);
void    redirectPrtStream(int op,char *filename);
int     set_kbd_echo(int state);

/********************* Macros **************************/

#define Prs_Atom(a) Prs_AtomCtxt(a,(Node *)(NULL))
#define Prs_Mols(m) Prs_Mols_M(m,(char *)(NULL))
#define Prs_Mol(m)  Prs_Mol_M(m,(char *)(NULL))

#define Get_Kwd(a,b,c) Get_Kwd_W(a,b,c,TRUE)

#define Get_RawLine(a) Get_RawLine_MW(a,NULL,TRUE)
#define Get_RawLine_W(a,b) Get_RawLine_MW(a,NULL,b)
#define Get_RawLine_M(a,b) Get_RawLine_MW(a,b,TRUE)

#define Get_String(a,b) Get_String_MW(a,b,(char *)NULL,TRUE)
#define Get_String_W(a,b,c) Get_String_MW(a,b,(char *)NULL,c)
#define Get_String_M(a,b,c) Get_String_MW(a,b,c,TRUE)

#define Get_AtomType(a,b) Get_AtomType_MW(a,b,(char *)NULL,TRUE);
#define Get_AtomType_M(a,b,c) Get_AtomType_MW(a,b,c,TRUE);
#define Get_AtomType_W(a,b,c) Get_AtomType_MW(a,b,(char *)NULL,c);

#define Get_Color(a,b) Get_Color_MW(a,b,(char *)NULL,TRUE);
#define Get_Color_W(a,b,c) Get_Color_MW(a,b,(char *)NULL,c);
#define Get_Color_M(a,b,c) Get_Color_MW(a,b,c,TRUE);

#define Get_DspGlbl(a,b) Get_DspGlbl_MW(a,b,(char *)NULL,TRUE)
#define Get_DspGlbl_W(a,b,c) Get_DspGlbl_MW(a,b,(char *)NULL,c)
#define Get_DspGlbl_M(a,b,c) Get_DspGlbl_MW(a,b,c,TRUE)

#define Get_Atom(a,b) Get_Atom_MW(a,b,(char *)NULL,TRUE)
#define Get_Atom_W(a,b,c) Get_Atom_MW(a,b,(char *)NULL,c)
#define Get_Atom_M(a,b,c) Get_Atom_MW(a,b,c,TRUE)

#define Get_Bond(a,b) Get_Bond_MW(a,b,(char *)NULL,TRUE)
#define Get_Bond_W(a,b,c) Get_Bond_MW(a,b,(char *)NULL,c)
#define Get_Bond_M(a,b,c) Get_Bond_MW(a,b,c,TRUE)

#define Get_AtomCtxt(a,b,c) Get_AtomCtxt_MW(a,b,c,(char *)NULL,TRUE)
#define Get_AtomCtxt_W(a,b,c,d) Get_AtomCtxt_MW(a,b,c,(char *)NULL,d)
#define Get_AtomCtxt_M(a,b,c,d) Get_AtomCtxt_MW(a,b,c,d,TRUE)

#define Get_nAtoms(a,b,c,d) Get_nAtoms_MW(a,b,c,d,(char *)NULL,TRUE)
#define Get_nAtoms_W(a,b,c,d,e) Get_nAtoms_MW(a,b,c,d,(char *)NULL,e)
#define Get_nAtoms_M(a,b,c,d,e) Get_nAtoms_MW(a,b,c,d,e,TRUE)

#define Get_Atoms(a,b) Get_Atoms_MW(a,b,(char *)NULL,TRUE)
#define Get_Atoms_W(a,b,c) Get_Atoms_MW(a,b,(char *)NULL,c)
#define Get_Atoms_M(a,b,c) Get_Atoms_MW(a,b,c,TRUE)

#define Get_Bonds(a,b) Get_Bonds_MW(a,b,(char *)NULL,TRUE)
#define Get_Bonds_W(a,b,c) Get_Bonds_MW(a,b,(char *)NULL,c)
#define Get_Bonds_M(a,b,c) Get_Bonds_MW(a,b,c,TRUE)

#define Get_Atomset(a,b) Get_Atomset_MW(a,b,(char *)NULL,TRUE)
#define Get_Atomset_W(a,b,c) Get_Atomset_MW(a,b,(char *)NULL,c)
#define Get_Atomset_M(a,b,c) Get_Atomset_MW(a,b,c,TRUE)

#define Get_Atomsetname(a,b) Get_Atomsetname_MW(a,b,(char *)NULL,TRUE)
#define Get_Atomsetname_W(a,b,c) Get_Atomsetname_MW(a,b,(char *)NULL,c)
#define Get_Atomsetname_M(a,b,c) Get_Atomsetname_MW(a,b,c,TRUE)

#define Get_Mol(a,b) Get_Mol_MW(a,b,(char *)NULL,TRUE)
#define Get_Mol_W(a,b,c) Get_Mol_MW(a,b,(char *)NULL,c)
#define Get_Mol_M(a,b,c) Get_Mol_MW(a,b,c,TRUE)

#define Get_Mol_D(a,b) Get_Mol_DMW(a,b,"/",TRUE);
#define Get_Mol_DM(a,b,c) Get_Mol_DMW(a,b,c,TRUE);
#define Get_Mol_DW(a,b,c) Get_Mol_DMW(a,b,"/",c);

#define Get_Mols(a,b) Get_Mols_MW(a,b,(char *)NULL,TRUE)
#define Get_Mols_W(a,b,c) Get_Mols_MW(a,b,(char *)NULL,c)
#define Get_Mols_M(a,b,c) Get_Mols_MW(a,b,c,TRUE)

#define Get_Mols_D(a,b) Get_Mols_DMW(a,b,(char *)NULL,TRUE);
#define Get_Mols_DM(a,b,c) Get_Mols_DMW(a,b,c,TRUE);
#define Get_Mols_DW(a,b,c) Get_Mols_DMW(a,b,(char *)NULL,c);

#define Get_Float(a,b) Get_Float_MW(a,b,(char *)NULL,TRUE)
#define Get_Float_W(a,b,c) Get_Float_MW(a,b,(char *)NULL,c)
#define Get_Float_M(a,b,c) Get_Float_MW(a,b,c,TRUE)

#define Get_Int(a,b) Get_Int_MW(a,b,(char *)NULL,TRUE)
#define Get_Int_W(a,b,c) Get_Int_MW(a,b,(char *)NULL,c)
#define Get_Int_M(a,b,c) Get_Int_MW(a,b,c,TRUE)

#define Get_XYZ(a,b) Get_XYZ_MW(a,b,(char *)NULL,TRUE)
#define Get_XYZ_W(a,b,c) Get_XYZ_MW(a,b,(char *)NULL,c)
#define Get_XYZ_M(a,b,c) Get_XYZ_MW(a,b,c,TRUE)

#define Get_File(dir,a,b,c) Get_File_MW(dir,a,b,c,(char *)NULL,TRUE)
#define Get_File_W(dir,a,b,c,d) Get_File_MW(dir,a,b,c,(char *)NULL,d)
#define Get_File_M(dir,a,b,c,d) Get_File_MW(dir,a,b,c,d,TRUE)

#define Get_Fildes(dir,a,b,c) Get_Fildes_MW(dir,a,b,c,(char *)NULL,TRUE)
#define Get_Fildes_W(dir,a,b,c,d) Get_Fildes_MW(dir,a,b,c,(char *)NULL,d)
#define Get_Fildes_M(dir,a,b,c,d) Get_Fildes_MW(dir,a,b,c,d,TRUE)

#define Get_Directory(a,b) Get_Directory_MW(a,b,(char *)NULL,TRUE)
#define Get_Directory_W(a,b,c) Get_Directory_MW(a,b,(char *)NULL,c)
#define Get_Directory_M(a,b,c) Get_Directory_MW(a,b,c,TRUE)

#define Get_Objects(type,a,b) Get_ObjectsR_MW((Dspobj *)NULL,type,a,b,(char *)NULL,TRUE)
#define Get_Objects_W(type,a,b,d) Get_ObjectsR_MW((Dspobj *)NULL,type,a,b,(char *)NULL,d)
#define Get_Objects_M(type,a,b,c) Get_ObjectsR_MW((Dspobj *)NULL,type,a,b,c,TRUE)
#define Get_Objects_MW(type,a,b,c,d) Get_ObjectsR_MW((Dspobj *)NULL,type,a,b,c,d)
#define Get_ObjectsR(r,type,a,b) Get_ObjectsR_MW(r,type,a,b,(char *)NULL,TRUE)
#define Get_ObjectsR_W(r,type,a,b,d) Get_ObjectsR_MW(r,type,a,b,(char *)NULL,d)
#define Get_ObjectsR_M(r,type,a,b,c) Get_ObjectsR_MW(r,type,a,b,c,TRUE)

#define Get_Object(type,a,b) Get_ObjectR_MW((Dspobj *)NULL,type,a,b,(char *)NULL,TRUE)
#define Get_Object_W(type,a,b,d) Get_ObjectR_MW((Dspobj *)NULL,type,a,b,(char *)NULL,d)
#define Get_Object_M(type,a,b,c) Get_ObjectR_MW((Dspobj *)NULL,type,a,b,c,TRUE)
#define Get_Object_MW(type,a,b,c,d) Get_ObjectR_MW((Dspobj *)NULL,type,a,b,c,d)
#define Get_ObjectR(r,type,a,b) Get_ObjectR_MW(r,type,a,b,(char *)NULL,TRUE)
#define Get_ObjectR_W(r,type,a,b,d) Get_ObjectR_MW(r,type,a,b,(char *)NULL,d)
#define Get_ObjectR_M(r,type,a,b,c) Get_ObjectR_MW(r,type,a,b,c,TRUE)

#define Get_RuleLib(r,a,b) Get_RuleLib_W(r,a,b,TRUE);

#endif

