#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "gcalc.h"
#include "bhtree.h"
#include "graphics.h"
#include "xtal.h"

#ifndef M_PI
#define M_PI  3.14159265358979323846
#endif

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Dspobj *DspRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define break_chain {if (!broken) fprintf(stderr,"chain broken\n");\
		       broken=TRUE; continue;}

#define MAX_DIST_CA 4.1

static void cmd_envir(void);
static void cmd_ca_stats(void);
static void cmd_ca_build(void);
static Atom *make_new_atom(Atom *a,Atom *b,Atom *c,char *name,float d,float v,float t);
static void envir_extract(Node *n,Node *r);
static void cmd_trace(void);
static Token *get_trace(char *prmt);
static void check_scale_cryst(void);

/*======================================================================*/

/****************  Miscellaneous ad-hockery around PDB ******************/

void cmd_pdb(void)
{
  Token *key;

  key=(Token *)NULL;
  key=Get_Kwd("PDB command> ",";","cellcheck CA-build CA-stats envir trace");
  if (key->status==TOK_OK) {
    if (key_is("envir"))           cmd_envir();
    else if (key_is("cellcheck"))  check_scale_cryst();
    else if (key_is("CA-stats"))   cmd_ca_stats();
    else if (key_is("CA-build"))   cmd_ca_build();
    else if (key_is("trace"))      cmd_trace();
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad PDB command: %s",key->text);
  }
  free_Token(key);
}

/*-------------------- cmd_ca_stats --------------------*/

#define MXRES 10000
#define NAT_STAT 3

static void cmd_ca_stats(void)
{
  Token *key;
  Node *root,*res;
  Atom *a,*ca[MXRES],*atm[MXRES][NAT_STAT];
  char nm[4][4];
  char *name[NAT_STAT]={"N","O","CB"};
  float alpha,tau,ang[NAT_STAT],tor[NAT_STAT];
  int i,k,l,nres,broken;
  FILE *out;
  char buf[2024];
  Nodechain r;

  key=(Token *)NULL; out=(FILE *)NULL;
  key=Get_String("output file (append): ",";");
  if (key->status==TOK_OK) {
    out=fopen(key->text,"a");
    if (!out) {
      err_set(1,"could not open %s for append.",key->text);
      key->status=TOK_ABORT;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad filename: %s",key->text);
    key->status=TOK_ABORT;
  }
  if (key->status==TOK_OK) key=Get_Mol_D("molecule: ",";");
  if (key->status==TOK_OK) {
    root=key->ptr;
    fprintf(out,"chain %s\n",nodeptr_to_name(buf,root));
    init_node_chain(&r);
    ForAtoms(a,root) add_to_node_chain(a->up,&r);
    if (r.n>MXRES) prtGraphics("Warning: %d residues (max allowed is %d)\n",r.n,MXRES);
    prtGraphics("%d residues in %s\n",r.n,nodeptr_to_name(buf,root));
    for (nres=0,res=r.s;mvld(res);res=res->chain) {
      if (nres==MXRES) break;
      ca[nres]=(Atom *)NULL;
      for (l=0;l<NAT_STAT;l++) atm[nres][l]=(Atom *)NULL;
      ForAtoms(a,res) {
	if (strceq(a->name,"CA")) ca[nres]=a;
	for (l=0;l<NAT_STAT;l++) if (strceq(a->name,name[l])) atm[nres][l]=a;
      }
      nres++;
    }
    clear_node_chain(&r);
    for (broken=TRUE,k=0;k<nres-3;k++) {
      if (!ca[k]||!ca[k+1]||!ca[k+2]||!ca[k+3]) break_chain;
      if (cdist(ca[k]->x,ca[k+1]->x)>MAX_DIST_CA) break_chain;
      if (cdist(ca[k+1]->x,ca[k+2]->x)>MAX_DIST_CA) break_chain;
      if (cdist(ca[k+2]->x,ca[k+3]->x)>MAX_DIST_CA) break_chain;
      alpha=cangled(ca[k]->x,ca[k+1]->x,ca[k+2]->x);
      tau=ctorsd(ca[k]->x,ca[k+1]->x,ca[k+2]->x,ca[k+3]->x);
      for (l=0;l<NAT_STAT;l++) {
	if (atm[k+1][l]) {
	  ang[l]=cangled(ca[k+2]->x,ca[k+1]->x,atm[k+1][l]->x);
	  tor[l]=ctorsd(ca[k]->x,ca[k+2]->x,ca[k+1]->x,atm[k+1][l]->x);
	}else{
	  ang[l]=360.0;
	  tor[l]=360.0;
	}
      }
      for (i=0;i<4;i++) Strcpy(nm[i],ca[k+i]->up->name);
      fprintf(out,"%s %s %s %s %6.2f %6.1f",nm[0],nm[1],nm[2],nm[3],alpha,tau);
      for (l=0;l<NAT_STAT;l++) fprintf(out," %6.2f %6.1f",ang[l],tor[l]);
      fprintf(out," %s\n",nodeptr_to_name(buf,ca[k+1]->up));
      broken=FALSE;
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad molecule token: %s",key->text);
  }
  if (out) fclose(out);
  free_Token(key);
}
/*------------------------------------------------------------------------*/

static void cmd_envir(void)
{
  Token *key;
  Node *n,*r;
  char filename[512];
  char old[512],code[512];
  int status;

  key=(Token *)NULL;
  old[0]='\0';
  n=(Node *)NULL;
  
  for (;;) {
    err_dump();
    key=free_Token(key);
    key=Get_String("residue : ",";");
    if (key->status==TOK_OK) {
      sprintf(filename,"/brookh%s/",key->text);
      *strchr(filename+8,'/')=0;
      sprintf(code,"%s",filename+8);
      if (!streq(old,filename)) {
	if (mvld(n)) del_node(n);
	n=add_node(MolRoot,code,OBJ_MOL);
	if (mvld(n)) {
	  status=read_pdb((UCpipe *)NULL,filename,n,FT_PDB);
	  strcpy(old,filename);
	  if (status) {
	    err_set(1,"could not read %s",filename);
	  }
	}
      }
      r=name_to_nodeptr(key->text,strcasecmp);
      if (mvld(r)) {
	envir_extract(n,r);
      }else{
	err_set(1,"could not find residue %s",key->text);
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad residue index: %s",key->text);
    }else break;
  }
  if (mvld(n)) del_node(n); 
  free_Token(key);
}
/*------------------------------------------------------------------------*/

static void envir_extract(Node *n,Node *r)
{
  Node *p;
  Atom *a,*b,*atm[3];
  char fullname[512];
  Nodechain c;
  float Dtol,x[3],X[3];
  int i,nbase,k;
  vec_3 dx,d1,d2,d3;
  mat_3 tf,tb;

  c.n=0; c.s=(Node *)NULL; c.e=(Node *)NULL; c.p=(Node *)NULL;

  if (!mvld(r->nxt)) return;

  /* add two residues to chain */

  ForAtoms(a,r) add_to_node_chain((Node *)a,&c);
  ForAtoms(a,r->nxt) add_to_node_chain((Node *)a,&c);

  if (c.n >0) {

    /* find enclosing box */

    for (i=0;i<3;i++) {
      x[i]=((Atom *)c.s)->x[i];
      X[i]=((Atom *)c.s)->x[i];
    }
    for (a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) {
      for (i=0;i<3;i++) {
	x[i]=(x[i]<a->x[i] ? x[i] : a->x[i]);
	X[i]=(X[i]>a->x[i] ? X[i] : a->x[i]);
      }
    }

    Dtol=5.0;
    for (i=0;i<3;i++) {
      x[i]=x[i]-Dtol;
      X[i]=X[i]+Dtol;
    }
    nbase=c.n;

    /* add atoms within Dtol to chain */

    ForAtoms(a,n) {
      if (a->x[0]<x[0] || a->x[1]<x[1] || a->x[2]<x[2]) continue;
      if (a->x[0]>X[0] || a->x[1]>X[1] || a->x[2]>X[2]) continue;
      for (k=0,b=(Atom *)c.s;k<nbase;b=(Atom *)b->chain,k++) {
	if (cdist(a->x,b->x)<=Dtol) break;
      }
      if (k<nbase) {
	p=a->up;
	ForAtoms(a,p) add_to_node_chain((Node *)a,&c);
	a=(Atom *)p->lst;
      }
    }

    /* find standard transformation */

    for (i=0;i<3;i++) atm[i]=(Atom *)NULL;
    ForAtoms(a,r) {
      if (streq("CA",a->name)) atm[0]=a;
      if (streq("C",a->name)) atm[1]=a;
      if (streq("O",a->name)) atm[2]=a;
    }

    for (i=0;i<3;i++) if (!mvld(atm[i])) break;
    if (i==3) {
      d2=v3n(v3mv3(vec3(atm[2]->x),vec3(atm[1]->x)));
      d3=v3n(v3xv3(d2,v3mv3(vec3(atm[0]->x),vec3(atm[1]->x))));
      d1=v3n(v3xv3(d2,d3));
      tf.r[0]=d1; tf.r[1]=d2; tf.r[2]=d3;
      tb=m3t(tf);
      dx=vec3(atm[1]->x);
    }else{
      tb=ident_3(); 
      tf=ident_3();
      if (mvld(atm[1])) dx=vec3(atm[1]->x);
      else for (i=0;i<3;i++) dx.c[i]=0.0;
    }

    /* transform to standard coordinate system */

    for (a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) {
      v32arr(a->x,m3xv3(tf,v3mv3(vec3(a->x),dx)));
      a->flg|=(1<<1);
    }

    /* write environment to file... */

    if (r->up==n) sprintf(fullname,"%s_%s.env",n->name,r->name);
    else if (r->up->up==n) sprintf(fullname,"%s_%s_%s.env",n->name,r->up->name,r->name);
    else if (r->up->up->up==n) sprintf(fullname,"%s_%s_%s_%s.env",n->name,r->up->up->name,r->up->name,r->name);
    else sprintf(fullname,"%s_T%s_%s.env",n->name,r->up->name,r->name);
    if (write_xplor_pdb(fullname,n,FALSE)) {
      err_set(1,"error writing %s",fullname);
    }else{
      prtGraphics("%d atoms written to %s\n",c.n,fullname);
    }
  }else{
    prtGraphics("no atoms.\n");
  }

  /* transform back to original system */

  for (a=(Atom *)c.s;mvld(a);a=(Atom *)a->chain) {
    v32arr(a->x,v3pv3(m3xv3(tb,vec3(a->x)),dx));
    a->flg&=~(1<<1);
  }

  clear_node_chain(&c);
}
/*-------------------- cmd_ca_build --------------------*/

static void cmd_ca_build(void)
{
  Nodechain *atms;
  float dis,tau,taua;
  Token *key;
  Atom *a,*b,*c,*d,*CB,*N,*O,*C;
  Nodechain new;
  char *nname;
  
  atms=(Nodechain *)NULL;
  key=Get_Atoms("atoms: ",";");
  if (key->status==TOK_OK) {
    atms=(Nodechain *)key->ptr;
    if (atms->n>3) {
      init_node_chain(&new);
      a=(Atom *)atms->s; b=(Atom *)a->chain; 
      c=(Atom *)b->chain; d=(Atom *)c->chain;
      for (;mvld(d);a=b,b=c,c=d,d=(Atom *)d->chain) {
	if (strneq(b->up->name,"GLY",3)) {
	  dis=1.08;
	  nname="HB1";
	}else{
	  dis=1.54;
	  nname="CB";
	}
	taua=ctorsd(a->x,b->x,c->x,d->x);
	if (strneq(c->up->name,"GLY",3)) {
	  if (cangled(a->x,b->x,c->x)<104) tau=-60;
	  else tau=+60;
	}else tau=260-taua;
	CB=make_new_atom(a,c,b,nname,dis,110.0,150-33*(180+taua)/230.0);
	if (!mvld(CB)) continue;
	CB->prv=(Node *)b; link_node((Node *)CB); 
	add_to_node_chain((Node *)CB,&new);
	add_bond_raw(CB,b);

	C=make_new_atom(a,c,b,"C",1.53,20.5,tau);
	if (!mvld(C)) continue;
	C->prv=(Node *)b; link_node((Node *)C); 
	add_to_node_chain((Node *)C,&new);
	add_bond_raw(C,b);

	O=make_new_atom(a,c,b,"O",2.42,47.0,tau);
	if (!mvld(O)) continue;
	O->prv=(Node *)C; link_node((Node *)O); 
	add_to_node_chain((Node *)O,&new);
	add_bond_raw(O,C);

	if (cdist(b->x,c->x) < MAX_DIST_CA) {
	  N=make_new_atom(b,O,C,"N",1.37,120,180);
	  if (!mvld(N)) continue;
	  N->nxt=(Node *)c; link_node((Node *)N); 
	  add_to_node_chain((Node *)N,&new);
	  add_bond_raw(N,c); 
	  add_bond_raw(N,C);
	}
      }
      send_atom_chain(&new);
      clear_node_chain(&new);
    }else{
      err_set(1,"not enough atoms...");
    }
  }else if (key->status==TOK_NOK) {
    err_set(1,"bad atomlist token: %s",key->text);
  }
  if (mvld(atms)) {
    clear_node_chain(atms);
    free_node_chain(atms);
  }
  free_Token(key);
}

/*----------------------------------------------------------------------*/

static Atom *make_new_atom(Atom *a,Atom *b,Atom *c,char *name,float d,float v,float t)
{
  Atom *atom;

  atom=(Atom *)make_node(OBJ_ATOM);
  if (mvld(atom)) {
    set_node_name((Node *)atom,name);
    if (!mvld(atom->atp)) set_atom_type(atom,guess_atom_type(name));
    cnstr(atom->x,a->x,b->x,c->x,d,v,t);
    atom->nxyz=1;
  }
  return(atom);
}
/*-------------------- cmd_trace --------------------*/

static void cmd_trace(void)
{
  char *menu = "add color delete on off width";
  char *name;
  Token *key;
  float maxd = MAX_DIST_CA;
  Nodechain *c;
  Node *m;
  Atom *a,*b;
  Gfxobj *T;
  int lw,on,newcolor,color = 2;
  Dspobj *t;

  key=Get_Kwd("trace option: ",";",menu);
  if (key->status==TOK_OK) {
    if (key_is("add")) {
      name=(char *)NULL;
      key=free_Token(key);
      key=Get_String("name: ",";");
      if (key->status==TOK_OK) {
	name=key->text;
	key->text=(char *)NULL;
	key=free_Token(key);
	key=Get_Atoms_D("atoms: ",";");
	if (key->status==TOK_OK) {
	  c=walk_order_chain((Node *)NULL,(Nodechain *)key->ptr,(1<<0));
	  T=OpenGfxObject(glblobj(assocmdl(common_root(c))),4096);
	  if (mvld(T)) {
	    t=T->Obj;
	    t->dsptype=DSP_TRCE;
	    b=(Atom *)NULL;
	    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	      if (is_atom(b)) {
		if (cdist(a->x,b->x)<=maxd) AddMD(T,b->x,a->x,color);
	      }
	      b=a;
	    }
	    prtGraphics("%d vector%s added to graphics object.\n",
			T->Nvec,(T->Nvec==1 ? "" : "s"));
	    set_node_name((Node *)t,name);
	    CloseGfxObject(T);
	  }else{
	    err_set(1,"failed to add graphics object.");
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad atom selection: %s",key->text);
	}
	if (mvld(name)) free(name);
      }else if (key->status==TOK_NOK) err_set(1,"bad name token: %s",key->text);
    }else if (key_is("color")) {
      key=free_Token(key);
      key=Get_Color("color: ",";");
      if (key->status==TOK_OK) {
	newcolor=key->n;
	key=free_Token(key);
	key=get_trace("trace: ");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (m=c->s;mvld(m);m=m->chain) {
	    if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	      send_bheader(graphix,"DSPC",sizeof(int)+sizeof(m));
	      send_bint(graphix,newcolor);
	      send_bptr(graphix,m);
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad color token: %s",key->text);
    }else if (key_is("width")) {
      key=free_Token(key);
      key=Get_Int_M("linewidth: ",";","1 2 3 4 5 6 7 8 9 10");
      if (key->status==TOK_OK) {
	lw=(key->n < 1 ? 1 : (key->n >33 ?  32 : key->n));
	key=free_Token(key);
	key=get_trace("trace: ");
	if (key->status==TOK_OK) {
	  c=(Nodechain *)key->ptr;
	  for (m=c->s;mvld(m);m=m->chain) {
	    if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	      send_bheader(graphix,"SOLW",sizeof(int)+sizeof(m));
	      send_bint(graphix,lw);
	      send_bptr(graphix,m);
	    }
	  }
	  clear_node_chain(c);
	  free_node_chain(c);
	}else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
      }else if (key->status==TOK_NOK) err_set(1,"bad linewidth token: %s",key->text);
    }else if (key_is("delete")) {
      key=free_Token(key);
      key=get_trace("trace: ");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	del_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
    }else if (key_is("off")||key_is("on")) {
      on=key_is("on");
      key=free_Token(key);
      key=get_trace("trace: ");
      if (key->status==TOK_OK) {
	c=(Nodechain *)key->ptr;
	for (m=c->s;mvld(m);m=m->chain) {
	  if (is_dsp(m)&&cPtr(((Dspobj *)m)->cross)) {
	    send_bheader(graphix,"DSPO",sizeof(int)+sizeof(m));
	    send_bint(graphix,on);
	    send_bptr(graphix,m);
	  }
	}
	clear_node_chain(c);
	free_node_chain(c);
      }else if (key->status==TOK_NOK) err_set(1,"bad trace selection: %s",key->text);
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad trace option %s",key->text);
  free_Token(key);
}
/*-------------------- get_trace --------------------*/

static Token *get_trace(char *prmt)
{
  Nodechain c,*f;
  Dspobj *o;
  Node *z,*nxt;
  int l;
  char *names;
  Token *key;

  key=make_Token(); key->status=TOK_NOK;
  names=(char *)NULL;
  l=4; c.n=0;c.s=(Node *)NULL; c.e=c.s; c.p=c.s;

  ForDspobjs(o,DspRoot) {
    if (o->dsptype==DSP_TRCE) {
      if (mvld(o->name)) {
	l+=strlen(o->name)+1;
	add_to_node_chain((Node *)o,&c);
      }
    }
  }
  if (c.n>0) {
    names=(char *)malloc(l*sizeof(char)+1);
    if (mvld(names)) {
      sprintf(names,"* ");
      for (z=c.s;mvld(z);z=z->chain) {
	if (mvld(z->name)) sprintf(names+strlen(names),"%s ",z->name);
      }
      free_Token(key);
      key=Get_String_M(prmt,";",names);
      if (key->status==TOK_OK) {
	f=(void *)make_node_chain();
	key->ptr=f;
	if (mvld(f)) {
	  for (z=c.s;mvld(z);z=nxt) {
	    nxt=z->chain; z->chain=(Node *)NULL;
	    if (wild_xmatch(z->name,key->text)) add_to_node_chain(z,f);
	  }
	  if (f->n==0) {
	    key->status=TOK_NOK;
	    key->ptr=(void *)NULL;
	    free_node_chain(f);
	  }
	}else{
	  err_set(1,"memory allocation problem in 'get_trace'.");
	  clear_node_chain(&c);
	  key->status=TOK_NOK;
	}
      }else{
	clear_node_chain(&c);
      }
    }else{
      err_set(1,"memory allocation problem in 'get_trace'.");
      clear_node_chain(&c);
    }
  }else{
    err_set(1,"no trace objects.");
    key->status=TOK_EOC;
  }

  if (mvld(names)) free(names);
  return(key);
}
/*-------------------- check_scale_crust --------------------*/

static void check_scale_cryst(void)
{
  /* check consistency of scale/cryst info of crystallographic node */

  Token *key;
  Node *r;
  Xcn *x;
  mat_4 M,C;
  float d,e,c[3],a[3],vs,vc;
  int i,j;
  
  key=(Token *)NULL;
  key=Get_Mol_D("molecule to check> ",";");
  if (key->status==TOK_OK) {
    r=(Node *)key->ptr;
    x=first_xcn(r);
    if (is_xcn(x)) {
      C=PDB_fta_matrix(x->cl,x->ca);
      vc=C.r[0].c[0]*C.r[1].c[1]*C.r[2].c[2]
	+C.r[0].c[1]*C.r[1].c[2]*C.r[2].c[0]
	+C.r[1].c[0]*C.r[2].c[1]*C.r[0].c[2]
	-C.r[2].c[0]*C.r[1].c[1]*C.r[0].c[2]
	-C.r[1].c[0]*C.r[0].c[1]*C.r[2].c[2]
	-C.r[0].c[0]*C.r[2].c[1]*C.r[1].c[2];

      for (i=0;i<3;i++) {
	for (c[i]=0,j=0;j<3;j++) c[i]+=x->fta.r[j].c[i]*x->fta.r[j].c[i];
	c[i]=sqrt(c[i]);
      }
      for (i=0;i<3;i++) {
	for (a[i]=0,j=0;j<3;j++) a[i]+=x->fta.r[j].c[(i+1)%3]*x->fta.r[j].c[(i+2)%3];
	a[i]=acos(a[i]/(c[(i+1)%3]*c[(i+2)%3]))*180.0/M_PI;
      }
      M=PDB_fta_matrix(c,a);
      vs=M.r[0].c[0]*M.r[1].c[1]*M.r[2].c[2]
	+M.r[0].c[1]*M.r[1].c[2]*M.r[2].c[0]
	+M.r[1].c[0]*M.r[2].c[1]*M.r[0].c[2]
	-M.r[2].c[0]*M.r[1].c[1]*M.r[0].c[2]
	-M.r[1].c[0]*M.r[0].c[1]*M.r[2].c[2]
	-M.r[0].c[0]*M.r[2].c[1]*M.r[1].c[2];
      for (d=0,i=0;i<3;i++) {
	for (e=0,j=0;j<3;j++) e+=(M.r[i].c[j]-C.r[i].c[j])*(M.r[i].c[j]-C.r[i].c[j]);
	e=sqrt(e); d=(d<e ? e : d);
      }

      prtGraphics("\n");
      prtGraphics("Entry %s : max. cryst/scale deviation= %0.3f\n",r->name,d);
      prtGraphics("   cryst: %8.3f%10.3f%10.3f%10.2f%7.2f%7.2f    V= %0.1f\n",
		  x->cl[0],x->cl[1],x->cl[2],x->ca[0],x->ca[1],x->ca[2],vc);
      prtGraphics("   scale: %8.3f%10.3f%10.3f%10.2f%7.2f%7.2f    V= %0.1f\n",
		  c[0],c[1],c[2],a[0],a[1],a[2],vs);
      prtGraphics("                scale                                   default\n");
      C=inv_4(C);
      for (i=0;i<3;i++) {
	prtGraphics("%11.6f%11.6f%11.6f        %11.6f%11.6f%11.6f\n",
		    x->atf.r[i].c[0],x->atf.r[i].c[1],x->atf.r[i].c[2],
		    C.r[i].c[0],C.r[i].c[1],C.r[i].c[2]);
      }	
      prtGraphics("\n");
    }else err_set(1,"no crystal info in %s",r->name);
  }else if (key->status==TOK_NOK) err_set(1,"bad molecule selection : %s",key->text);
  key=free_Token(key);
}
