#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <math.h>
#ifdef mingw
#include <time.h>
#endif

#define TRUE  1
#define FALSE 0

static int HW=1;
FILE *R=(FILE *)NULL;
static unsigned int S=0,O=0,P;
static unsigned int K;
static int giter=0;

#ifndef mingw
static void wakeup(int x);
#endif
static int KnotZero(void);

#include "isaac.h"

static ISAACctxt ctxt=NULL;

/* NOTE: the rseed routines are intended to be used to generate  */
/* random seeds for pseudo-random generators. They are not good  */
/* random number generators. */

#ifndef mingw
unsigned int rseedSlow(void)
{
  /* a rather slow method to generate a random unsigned int      */
  /* takes 1 second per byte, and should be pretty random on     */
  /* most time shared systems (relies on inaccuracy of alarm(2), */
  /* and on being interrupted by scheduler).                     */

  unsigned int r,b;
  int i;

  signal(SIGALRM,wakeup);
  for (b=0,r=0,i=0;i<8*sizeof(int);i+=8) {
    K=1;
    alarm(1);
    while(KnotZero()) {
      b++;
      b&=255;
    }
    r|=b<<i;
  }
  return(r);
}
#endif

unsigned int rseedFast(void)
{
#if defined(mingw)
#else
  struct timeval tv;
#endif
  unsigned int usec,sec;
  unsigned int r,r1,r2;
  int k;
  
  /* faster generator of random seed, relies on reasonable resolution */
  /* of timeofday. should not be used too often within one tick */

#if defined (mingw)
  usec=time((time_t *)NULL);
  sec=usec;
#else
  gettimeofday(&tv,NULL);
  usec=tv.tv_usec;
  sec=tv.tv_sec;
#endif

  if (S==0) S=usec&31;
  if (P) P+=19;
  else P=getpid();
  r1=sec^((P<<16)|(usec&(0xffff)));
  for (r2=0,k=0;k<31;k++) r2=r2<<k|((r1>>k)&1);
  r=r2^(((r1<<S))|(r1>>(S-32)))^usec^((O>>S)&1);
  O=(r>>S)^r2;
  return(r);
}

unsigned int randomHardware(void)
{
  unsigned int r,r2;

  /* try to get random number from /dev/random. If this fails, switch */
  /* to rseedFast method */

  if (HW) {
    if (!R) {
      R=fopen("/dev/urandom","r");
      if (R) {
	if (fread(&r,sizeof(unsigned int),1,R)<sizeof(unsigned int)) r=0;
	fclose(R);
	R=fopen("/dev/urandom","r");
	if (fread(&r2,sizeof(unsigned int),1,R)<sizeof(unsigned int)) r2=0;
	if (r==r2) {
	  fclose(R);
	  R=(FILE *)NULL;
	  HW=FALSE;
	  r=rseedFast();
	}
      }else{
	HW=FALSE;
	r=rseedFast();
      }
    }else if (fread(&r,sizeof(unsigned int),1,R)<sizeof(unsigned int)) r=rseedFast();
  }else r=rseedFast();
  return(r);
}

#ifndef mingw
static void wakeup(int sig)
{
  K=0;
  signal(SIGALRM,wakeup);
}
#endif

static int KnotZero(void)
{
  return(K!=0);
}

void setSeed(unsigned int seed)
{
  unsigned long s;

  /* Note: id seed==0, try to generate a random seed */

  if (seed) s=seed;
  else s=randomHardware();
  if (ctxt) ctxt=ISAACfree(ctxt);
  ctxt=ISAACseed(1,&s);
  giter=0;
}

double udrand(double a,double b)
{
  if (!ctxt) ctxt=ISAACseed(0,NULL);
  return(a+(b-a)*ISAACdrand(ctxt));
}

float ufrand(float a,float b)
{
  if (!ctxt) ctxt=ISAACseed(0,NULL);
  return(a+(b-a)*ISAACdrand(ctxt));
}

double gaussdrand(double mean,double stdev)
{
  double a1,a2,p,q;
  static double z1,z2;

  /* George Marsaglia's method for the genartion of Gaussian random mumbers */

  if (giter==0) {
    do {
      a1=udrand(-1.0,1.0);
      a2=udrand(-1.0,1.0);
      q=a1*a1+a2*a2;
    } while (!(0.0<q && q<=1.0));
    p=sqrt(-2.0*log(q)/q);
    z1=mean+stdev*a1*p;
    z2=mean+stdev*a2*p;
    giter=1;
  }else{
    giter=0;
    z1=z2;
  }
  return z1;
}

int uirand(int a,int b)
{
  if (!ctxt) ctxt=ISAACseed(0,NULL);
  return(a+(b-a+1)*ISAACdrand(ctxt));
}

/* FORTRAN interface to Gaussian random numbers */

double randg_()
{
  return gaussdrand(0.0,1.0);
}
