/*   readbio.c   */

#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "errhand.h"
#include "transpor.h"
#include "parser.h"

#define NO 0
#define YES 1
#define Return(x) {if (own_p) UCpclose(pipe);return(x);}
#define MAXTOKENSIZE 250

#define VX 0
#define V1 1
#define V2 2
#define V3 3

#define UNKNOWN 0
#define TOPOLOGY 1
#define SYMMETRY 2
#define ATOMSET 3
#define DONE 99999

#define NFLDEFS 12

#define FLD_CHS 2
#define FLD_MNEM 3
#define FLD_CHGRP 4
#define FLD_ISOTOPE 5
#define FLD_FCHRG 6
#define FLD_CHRG 7
#define FLD_SWITCH 8
#define FLD_OOPF 9
#define FLD_CHIRAL 10 
#define FLD_OCC 11
#define FLD_BFAC 12
#define FLD_CONN 13

#define FLD(f,t) (strfld(fld,rec,f,t))

extern Node *LibRoot;
extern int graphix;

/*********************************************************************/
int read_bio(UCpipe *pipe,char *file,Node *node)
{
     /*   Read BIOSYM file into node. Default extension is CAR, extension of
	  MDF-file  m u s t  be MDF. Returns non-zero status if read fails.  */

  char rec[256],cur_res[15],res[15],name[6],fld[50],fftype[9],elts[3];
  char hdr1[82],hdr2[82],*nm[4];
  char tok[MAXTOKENSIZE];
  char *fftypep,*t,*Lett,*plus,*minus;
  float x[3],charge=0,bo;
  Node *cur_node,*res_node=(Node *)NULL,*m,*n,nd[3];
  Node *ResLib;
  Atom *a=(Atom *)NULL,*b;
  Atomdef *tab,*default_type;
  Symbol *chs;
  Atomset *S;
  Bond *bnd;
  Parse p;
  int version,i,k,nbnd,recsz,ok,section,fldtype[NFLDEFS],nflds,
      sign,coff[3],symop,the_sign,wld[4],own_p,frmp;
  FILE *in;

  own_p=(pipe ? FALSE : TRUE);             /* do we own the pipe ? */
  
  /*  1. Read .CAR File   */
  
  if (own_p) {
    pipe=UCpopen(extfil(rec,file,".car",NO));
    if (!pipe) return(1);
  }
  in=pipe->fd;
  frmp=(pipe->pipe==UCP_PIPE&&pipe->filename[0]=='|');

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    Return(-1);
  }

  default_type=find_atomdef_name(ResLib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }
  
  /* try to find out what record format this file has */
  
  recsz=161;			
  if (!fgets(rec,recsz,in)) rec[0]='\0';	    /* if it is fixed 80 we have the two headers */
  if (strlen(rec)==recsz-1) {
    strncpy(hdr1,rec,80);hdr1[80]='\n';hdr1[81]='\0';
    strncpy(hdr2,rec+80,80);hdr2[80]='\n';hdr2[81]='\0';
    recsz=81;		 /* looks like fixed 80 */
  }else{
    strcpy(hdr1,rec);
    if (!fgets(rec,recsz,in)) rec[0]='0';      /* read second header line */       
    Strcpy(hdr2,rec);
  }
  prtGraphics("%s",hdr1);
  prtGraphics("%s",hdr2);
  if (strncasecmp(hdr1,"!BIOSYM ARCHIVE 1",17)==0) { /* Version 2.7 */
    version=V1;
    if (!fgets(rec,recsz,in)) rec[0]='\0';	/* Title */
    prtGraphics("%s",rec);
    if (!fgets(rec,recsz,in)) rec[0]='\0';	/* Date */
    prtGraphics("%s",rec);
    if (strncasecmp(hdr2,"PBC=ON",6)==0) {
      if (!fgets(rec,recsz,in)) rec[0]='\0';	/* PBC record ignored */
    }
  }else if (strncasecmp(hdr1,"!BIOSYM ARCHIVE 2",17)==0) { /* Version 2.10 */
    version=V2;
    if (!fgets(rec,recsz,in)) rec[0]='\0';	/* Title */
    prtGraphics("%s",rec);
    if (!fgets(rec,recsz,in)) rec[0]='\0';	/* Date */
    prtGraphics("%s",rec);
    if (strncasecmp(hdr2,"PBC=ON",6)==0) {
      if (!fgets(rec,recsz,in)) rec[0]='\0';	/* PBC record ignored */
    }
  }else if (strncasecmp(hdr1,"!BIOSYM ARCHIVE 3",17)==0) { /* Version 2.3 */
    version=V3;
    if (!fgets(rec,recsz,in)) rec[0]='\0';	/* Title */
    prtGraphics("%s",rec);
    if (!fgets(rec,recsz,in)) rec[0]='\0';	/* Date */
    prtGraphics("%s",rec);
    if (strncasecmp(hdr2,"PBC=ON",6)==0) {
      if (!fgets(rec,recsz,in)) rec[0]='\0';	/* PBC record ignored */
    }else if (strncasecmp(hdr2,"PBC=2D",6)==0) {
      if (!fgets(rec,recsz,in)) rec[0]='\0';	/* PBC record ignored */
    }else if (strncasecmp(hdr2,"HELIX",5)==0) {
      err_set(-1,"car-files containing helix information are not supported.");
      Return(-1);
    }
  }else{
    version=VX;
  }
  
  /* good luck... */
  
  if (!fgets(rec,recsz,in)) rec[0]='\0'; /* first atom record, I hope */
  while(strncmp(rec,"end",3)!=0 && !feof(in)) {
    strcpy(cur_res," ");
    cur_node=add_node(node,"m",OBJ_MOL);
    if (!cur_node) Return(2);
    ((Mol *)cur_node)->moltype=MOL_MOLECULE;
    while(strncmp(rec,"end",3)!=0 && !feof(in)) {
      switch (version) {
      case V1:;
      case VX:                /* unk. version assumed to be V1 compatible */
	if (strlen(rec)<74) Return(3);
	strsqz(name,FLD(1,4));
	for (i=0;i<3;i++) x[i]=str2f(FLD(6+15*i,20+15*i));
	strfld(res,rec,52,55);strcat(res,"_");
	strcat(res,FLD(56,60));strsqz(res,res);
	Strcpy(fftype,FLD(62,63));
	charge=str2f(FLD(66,74));
	break;
      case V2:
	if (strlen(rec)<75) Return(3);
	strsqz(name,FLD(1,4));
	for (i=0;i<3;i++) x[i]=str2f(FLD(6+15*i,20+15*i));
	strfld(res,rec,52,55);strcat(res,"_");
	strcat(res,FLD(56,60));strsqz(res,res);
	Strcpy(fftype,FLD(62,64));
	Strcpy(elts,FLD(65,66));
	charge=str2f(FLD(68,75));
	if (streq(fftype,"?")) sprintf(fftype,"%s.?",elts);
	break;
      case V3:
	if (strlen(rec)<80) Return(3);
	strsqz(name,FLD(1,5));
	x[0]=str2f(FLD(6,20)); x[1]=str2f(FLD(21,35)); x[2]=str2f(FLD(36,50));
	strfld(res,rec,52,55);strcat(res,"_");
	strcat(res,FLD(57,63));strsqz(res,res);
	Strcpy(fftype,FLD(64,70));
	Strcpy(elts,FLD(72,73));
	charge=str2f(FLD(75,80));
	if (streq(fftype,"?")) sprintf(fftype,"%s.?",elts);
	break;
      }
      if (strcmp(cur_res,res)!=0) {
	res_node=add_node(cur_node,res,OBJ_MOL);
	if (!res_node) Return(2);
	((Mol *)res_node)->moltype=MOL_RESIDUE;
	strcpy(cur_res,res);
      }
      a=(Atom *)add_node(res_node,(char *)NULL,OBJ_ATOM);
      if (!a) Return(2);
      set_node_name((Node *)a,name);
      for (i=0;i<3;i++) a->x[i]=x[i];
      a->nxyz=1;
      a->q=charge;
      tab=find_atomdef_name(ResLib,fftype);
      if (mvld(tab)) {
	set_atom_type(a,tab);
      }else{
	tab=(Atomdef *)copy_node((Node *)default_type);
	if (mvld(tab)) {
	  tab->up=ResLib;
	  link_node((Node *)tab);
	  set_node_name((Node *)tab,fftype);
	  strcpy(tab->chs,elts);
	  chs=find_symbol_name(ResLib,tab->chs);
	  if (mvld(chs)) {
	    tab->color=chs->color;
	    tab->mass=chs->mass;
	    tab->rcov=chs->rcov;
	    tab->rvdw=chs->rvdw;
	    tab->z=chs->z;
	    tab->val=chs->val;
	  }
	  set_atom_type(a,tab);
	}
      }
      if (!fgets(rec,recsz,in)) rec[0]='\0';
    }
    if (!fgets(rec,recsz,in)) rec[0]='\0';
  }
  
/* make molecule names unique */

  k=0;
  Lett="ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789";
  for (m=node->fst;m;m=m->nxt) {
    if (!is_mol(m)) continue;
    if (k<35) sprintf(name,"%c",Lett[k]);
    else {
      i=k/35-1;
      sprintf(name,"%c%c",Lett[i],Lett[k%35]);
    }
    set_node_name(m,name);
    k++;
  }

  if (frmp)   return(0);       /* if from pipe (except .Z) --> only read coordinates */
  if (own_p)  UCpclose(pipe);  /* if we own the pipe, close it... */
  else return(0);              /* if we don't own the pipe, we read only coordinates */

  /*   2. Read the corresponding .MDF file   */
  
  pipe=UCpopen(extfil(rec,file,".mdf",YES));
  if (!pipe) return(6);
  in=pipe->fd;
  
  p.buf=rec;p.tok=tok;p.mts=MAXTOKENSIZE;
  p.sep="";p.pos=0;

  for (k=0;k<3;k++) nd[k].chain=(Node *)NULL;

  if (!fgets(rec,255,in)) Return(701);
  get_token(&p);                          /* find out about what version */
  version=VX;
  if (strceq(p.tok,"!BIOSYM")) {
    get_token(&p);
    if (strceq(p.tok,"molecular_data")) {
      get_token(&p);
      if (streq(p.tok,"3")) version=V2;
      else if (streq(p.tok,"4")) version=V3;
    }
  }

  switch (version) {
  case VX:
    
    /* all old versions (before 2.10) */

    a=rfirst_atom(node);
    
    while (fgets(rec,255,in)) {
      p.pos=0;p.sep="";get_token(&p);
      if (strcmp(p.tok,"ATOM")==0) {
	if (!a) Return(702);
	fftypep=(mvld(a->atp) ? a->atp->name : "?");
	get_token(&p);if (strcmp(p.tok,a->name)!=0) Return(71);
	get_token(&p);
	if (!streq(p.tok,"?")&&!streq(p.tok,fftypep)) {
	  Return(72);
	}
	get_token(&p);Strcpy(fld,p.tok);
	get_token(&p);Strcpy(cur_res,p.tok);
	get_token(&p);strcat(cur_res,"_");strcat(cur_res,p.tok);
	if (strcmp(cur_res,a->up->name)!=0) {
	  Return(73);
	}
	strcpy(a->grp,fld);
	get_token(&p);                     /* ignore charge */
	get_token(&p);a->cosw=str2i(p.tok);
	get_token(&p);a->oop=str2i(p.tok);
	get_token(&p);a->fe=str2i(p.tok);
	get_token(&p);nbnd=str2i(p.tok);
	for (i=0;i<nbnd;i++) {
	  p.sep=":/";
	  Strcpy(cur_res,a->up->name);
	  get_token(&p);
	  if (p.term==':') {
	    if (p.ltok>9) Return(9);
	    Strcpy(cur_res,p.tok);
	    get_token(&p);
	  }
	  if (p.term!='/') {
	    Return(703);
	  }
	  if (p.ltok>4) Return(9);
	  Strcpy(name,p.tok);
	  p.sep="";
	  get_token(&p);bo=str2f(p.tok);
	  if (strcmp(cur_res,a->up->name)==0) {
	    for (b=(Atom *)a->nxt;b;b=(Atom *)b->nxt) {
	      if (strcmp(b->name,name)==0) break;
	    }
	  }else{
	    b=(Atom *)NULL;
	    for (n=a->up->nxt;n;n=n->nxt) {
	      if (strcmp(cur_res,n->name)==0) {
		for (b=(Atom *)n->fst;b;b=(Atom *)b->nxt) {
		  if (strcmp(b->name,name)==0) break;
		}
		break;
	      }
	    }
	  }
	  if (b) {
	    bnd=add_bond(a,b);
	    if (!bnd) Return(10);
	    bnd->bo36=bo*36+0.001;
	    bnd->bo=bo;
	  }
	}
	a=rnext_atom(node,a);
      }
    }
    break;                               /* done with old versions */
    
  case V2:                /* fall-through, almost the same as V3... */
  case V3:

    /* version 2.8, 2.10, 2.3  */

    section=UNKNOWN;
    nflds=0;
    S=(Atomset *)NULL;

    prtGraphics("%s",hdr1);
    for(a=rfirst_atom(node);section!=DONE;) {
      if (!fgets(rec,255,in)) {
	err_set(999,"unexpected eof or error reading mdf\n");
	break;
      }
      p.pos=0; p.sep=(section==TOPOLOGY ? ":" : ""); ok=get_token(&p);
      if (!ok||p.tok[0]=='!') continue; /* blank or comment */
      if (p.tok[0]=='#') {
	prtGraphics("%s",rec);
	if (streq(p.tok,"#topology")) section=TOPOLOGY;
	else if (strceq(p.tok,"#symmetry")) section=SYMMETRY; 
	else if (strceq(p.tok,"#atomset")) section=ATOMSET; 
	else if (strceq(p.tok,"#end")) section=DONE;
	else{
	  err_set(999,"unknown section header: %s",p.tok);
	  section=UNKNOWN;
	}
      }else{
	switch (section) {
	case TOPOLOGY:
	  if (p.tok[0]=='@') {
	    if (strceq(p.tok,"@column")) {
	      i=str2i(token(&p));
	      if (i!=nflds+1||nflds==NFLDEFS) {
		err_set(999,"out of sequence: %s",rec);
		section=DONE;
	      }else{
		get_token(&p);
		if (strceq(p.tok,"element")) k=FLD_CHS;
		else if (strceq(p.tok,"atom_type")) k=FLD_MNEM;
		else if (strceq(p.tok,"charge_group")) k=FLD_CHGRP;
		else if (strceq(p.tok,"isotope")) k=FLD_ISOTOPE;
		else if (strceq(p.tok,"formal_charge")) k=FLD_FCHRG;
		else if (strceq(p.tok,"charge")) k=FLD_CHRG;
		else if (strceq(p.tok,"switching_atom")) k=FLD_SWITCH;
		else if (strceq(p.tok,"oop_flag")) k=FLD_OOPF;
		else if (strceq(p.tok,"chirality_flag")) k=FLD_CHIRAL;
		else if (strceq(p.tok,"occupancy")) k=FLD_OCC;
		else if (strceq(p.tok,"xray_temp_factor")) k=FLD_BFAC;
		else if (strceq(p.tok,"connections")) k=FLD_CONN;
		else {
		  err_set(999,"bad rec: %s",rec);
		  section=DONE;
		}
		fldtype[nflds++]=k;
	      }
	    }else if (strceq(p.tok,"@molecule")) {
	      prtGraphics("%s",rec);
	      if (a) {
		if (mvld(a->up)) set_node_name(a->up->up,token(&p));
	      }
	    }else if (strceq(p.tok,"@assembly")) {
	      prtGraphics("assembly record ignored (no doc)\n");
	    }else{
	      err_set(999,"bad rec: %s",rec);
	      section=DONE;
	    }
	  }else{
	    if (a) {
	      Strcpy(cur_res,p.tok);
	      p.sep="";get_token(&p);
	      if (streq(cur_res,a->up->name)&&streq(p.tok,a->name)) {
		for (i=0;i<nflds;i++) {
		  switch (fldtype[i]) {
		  case FLD_CHS:
		    get_token(&p);
		    break;
		  case FLD_MNEM:
		    get_token(&p);
		    break;
		  case FLD_CHGRP:
		    Strcpy(a->grp,token(&p));
		    break;
		  case FLD_ISOTOPE:
		    get_token(&p);
		    break;
		  case FLD_FCHRG:
		    if (token(&p)) {
		      plus=strchr(p.tok,'+');
		      minus=strchr(p.tok,'-');
		      the_sign=(minus ? -1 : +1);
		      if (plus) *plus=' ';
		      if (minus) *minus=' ';
		      strsqz(p.tok,p.tok);
		      a->chrg=the_sign*str2i(p.tok);
		    }
		    break;
		  case FLD_CHRG:
		    a->q=str2f(token(&p));
		    break;
		  case FLD_SWITCH:
		    a->cosw=str2i(token(&p));
		    break;
		  case FLD_OOPF:
		    a->oop=str2i(token(&p));
		    break;
		  case FLD_CHIRAL:
		    get_token(&p);
		    break;
		  case FLD_OCC:
		    a->x[ATM_SOF]=str2f(token(&p));
		    break;
		  case FLD_BFAC:
		    a->x[ATM_BISO]=str2f(token(&p));
		    break;
		  case FLD_CONN:
		    if (i+1!=nflds) {
		      err_set(999,
			      "error: connections should come last "
			      "in atom record\n");
		      section=DONE;
		    }else{
		      while (get_token(&p)) {
			bo=1;
			for (k=0;k<3;k++) coff[k]=0;
			symop=1;
			if ((t=strrchr(p.tok,','))) {
			  *t='\0';   /* just ignore the wedge field */
			}
			if ((t=strrchr(p.tok,'/'))) {
			  bo=str2f(t+1);
			  *t='\0';
			}
			if ((t=strrchr(p.tok,'%'))) {
			  *t++ = '\0'; 
			  sign=1;
			  for (k=0;*t&&k<3;t++) {
			    if (*t=='-') sign=-1;
			    else if (*t=='+') sign=1;
			    else if (*t>='0'&&*t<='9') {
			      coff[k++]=sign*(*t-'0');
			      sign=1;
			    }
			  }
			  for (;k<3;k++) coff[k]=0;
			}
			if ((t=strrchr(p.tok,'#'))) {
			  symop=str2i(t+1);
			  *t='\0';
			}
			if ((t=strrchr(p.tok,':'))) {
			  Strcpy(name,t+1); *t='\0';
			  Strcpy(cur_res,p.tok);
			}else{
			  Strcpy(name,p.tok);
			  Strcpy(cur_res,a->up->name);
			}
			if (symop==1&&coff[0]==0&&coff[1]==0&&coff[2]==0) {
			  if (strcmp(cur_res,a->up->name)==0) {
			    for (b=(Atom *)a->nxt;b;b=(Atom *)b->nxt) {
			      if (strcmp(b->name,name)==0) break;
			    }
			  }else{
			    b=(Atom *)NULL;
			    for (n=a->up->nxt;n;n=n->nxt) {
			      if (strcmp(cur_res,n->name)==0) {
				for (b=(Atom *)n->fst;b;b=(Atom *)b->nxt) {
				  if (strcmp(b->name,name)==0) break;
				}
				break;
			      }
			    }
			  }
			  if (b) {
			    bnd=add_bond(a,b);
			    if (!bnd) Return(10);
			    bnd->bo36=bo*36+0.001;
			    bnd->bo=bo;
			  }
			}
		      }
		    }
		    break;
		  default:
		    err_set(999,"internal error(rdbio,mdf/atom decode fugu)\n");
		    Return(0);
		  }
		}
		a=rnext_atom(node,a);
	      }else{
		err_set(999,"error: %s:%s(car) should equal %s:%s(mdf)\n",
			    a->up->name,a->name,cur_res,p.tok);
		section=DONE;
	      }
	    }else{
	      err_set(999,"more atoms in mdf than in car\n");
	      section=DONE;;
	    }
	  }
	  break;
	case SYMMETRY:
	  break;
	case ATOMSET:
	  if (p.tok[0]=='@') {
	    if (is_atomset(S)) S=(Atomset *)NULL;
	    if (strceq(p.tok,"@list")) {
	      if (strceq(token(&p),"subset")) {
		get_token(&p);
		S=add_atomset((Mol *)node,-1,p.tok);
		for (k=1;k<4;k++) {
		  nm[k]="";
		  wld[k]=FALSE;
		}
		for (k=0;k<3;k++) {
		  for (n=nd[k].chain;mvld(n);n=m) {
		    m=n->chain;
		    n->chain=(Node *)NULL;
		  }
		  nd[k].chain=(Node *)NULL;
		}
		nd[0].chain=node;
	      }
	    }
	  }else if (is_atomset(S)) {
	    do {
	      for (k=3;k>0;k--) {
		if ((t=strrchr(p.tok,':'))) {
		  nm[k]=t+1;
		  if (strlen(nm[k])==0) nm[k]="*";
		  *t='\0';
		}else{
		  nm[k]=p.tok;
		  break;
		}
	      }
	      if (k==0) continue;   /* would be decent to report error !!! */
	      for (;k<3;k++) {
		wld[k]=strpbrk(nm[k],"*?")!=(char *)NULL;
		for (n=nd[k].chain;mvld(n);n=m) {
		  m=n->chain;
		  n->chain=(Node *)NULL;
		}
		nd[k].chain=(Node *)NULL;
		for (n=nd[k-1].chain;mvld(n);n=n->chain) {
		  if (is_mol(n)) {
		    for (m=n->fst;mvld(m);m=m->nxt) {
		      if (is_mol(m)&&mvld(m->name)) {
			if (wld[k]) {
			  if (wild_imatch(m->name,nm[k])) {
			    m->chain=nd[k].chain;
			    nd[k].chain=m;
			  }
			}else{
			  if (strceq(m->name,nm[k])) { 
			    m->chain=nd[k].chain;
			    nd[k].chain=m;
			  }
			}
		      }
		    }
		  }
		}
	      }
	      wld[3]=strpbrk(nm[3],"*?")!=(char *)NULL;
	      for (n=nd[2].chain;mvld(n);n=n->chain) {
		for (m=n->fst;mvld(m);m=m->nxt) {
		  if (is_atom(m)&&mvld(m->name)) {
		    if (wld[3]) {
		      if (wild_imatch(m->name,nm[3])) add_to_atomset((Atom *)m,S);
		    }else{
		      if (strceq(m->name,nm[3])) add_to_atomset((Atom *)m,S);
		    }
		  }
		}
	      }
	    }while (get_token(&p));
	  }
	  break;
	default:
	  break;
	}
      }
    }
    
    break;                               /* done with new version 2.10 */
  }

  UCpclose(pipe);
  for (k=0;k<3;k++) {
    for (n=nd[k].chain;mvld(n);n=m) {
      m=n->chain;
      n->chain=(Node *)NULL;
    }
  }

  if (a) return(7);


  /* set aromaticity flag and modify bond types where apropriate */

  ForAtoms(a,node) {
    if (mvld(a->atp)) a->arom=a->atp->arom;
    for (i=0;i<a->nbnd;i++) {
      if (a->bnd[i]->f==a) {
	bnd=a->bnd[i];
	bnd->bt=BT_UNKNOWN;
	if (bnd->bo36==54) {
	  if (mvld(bnd->t->atp)) bnd->t->arom=bnd->t->atp->arom;
	  if (a->arom && bnd->t->arom) {
	    set_bt(bnd,BT_AROMATIC);
	  }else{
	    set_bt(bnd,BT_SINGLE);   /* need a deloc. double bond type ... */
	  }
	}else{
	  if (bnd->bo==1) set_bt(bnd,BT_SINGLE);
	  else if (bnd->bo==2)  set_bt(bnd,BT_DOUBLE);
	  else if (bnd->bo==3) set_bt(bnd,BT_TRIPLE);
	}                            /* should also deal with amide bonds */
      }
    }
  }

  ForAtoms(a,node) {
    for (i=0;i<a->nbnd;i++) {
      if (a->bnd[i]->bo==2&&!a->bnd[i]->arom) mark_aromatic_bonds(a->bnd[i]);
    }
  }

  return(0);
}
