#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "streamio.h"
#include "errhand.h"
#include "transpor.h"

#define YES 1
#define NO  0
#define MAXTOKENSIZE 256

#define key_is(x) \
  (strcmp(key,x)==0)

extern int graphix;
extern Node *LibRoot;

/*========================================================================*/

int read_biocad (UCpipe *p,char *file,Node *node)
{
  Stream *in;
  char buf[MAXTOKENSIZE];
  int status;

  if (p) {
    in=open_stream((char *)NULL,MAXTOKENSIZE);
    set_stream_pipe(in,p);
  }else{
    in=open_stream(extfil(buf,file,".cpd",NO),MAXTOKENSIZE);
  }
  if (!mvld(in)) {
    err_set(1,"could not open stream on %s",extfil(buf,file,".cpd",NO));
    return(1);
  }

  status=0;
  if (read_biocad_mol(in,node)!=0) {
    status=1;
  }

  close_stream(in);
  return(status);

}

/*------------------------------------------------------------------------*/

int read_biocad_mol(Stream *in,Node *mol)
{
  char buf[MAXTOKENSIZE],tmp[MAXTOKENSIZE],*t;
  int status,curlev,n,bt=0;
  float *X;
  Node *nd,*ResLib;
  Mol *frag,*nfrag;
  Atom *a,*a1,*a2;
  Bond *b;
  Lib  *reslib;
  Symbol *S;
  Atomdef *def,*default_type;
  
  if (!mvld(in) || !mvld(mol)) {
    err_set(1,"error in args to read_biocad_mol...");
    return(1);
  }
  
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);

  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    return(-1);
  }
  reslib=(Lib *)ResLib;
  default_type=find_atomdef_name((Node *)reslib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  set_stream(in,"eol=no, redir=no, comment='!'");
  strm_tok(in)=buf; strm_mts(in)=MAXTOKENSIZE;
  
  status=0; curlev=0;
  if (streq(token_stream(in,"{}:();"),"COMPOUND")) {
    if (strm_term(in)==' ') {
      t=token_stream(in,"{}:();");
      if (t) set_node_name(mol,t);
      else{
	err_set(1,"compound name missing...");
	return(1);
      }
    }
    while(strm_term(in)!='{') {
      if (!token_stream(in,"{")) break;
    }
    if (strm_term(in)!='{') {
      err_set(1,"bad token separator after COMPOUND");
      return(1);
    }
    curlev++;
    while (curlev>0) {
      t=token_stream(in,"{}");
      if (strm_term(in)=='{') curlev++;
      if (strm_term(in)=='}') curlev--;
      if (t) {
	if (curlev==1 && streq(t,"FRAG")) {

	  /* parse FRAGMENT */

	  t=token_stream(in,"{");
	  frag=(Mol *)add_node(mol,t,OBJ_MOL);
	  if (!mvld(frag)) {
	    err_set(1,"memory allocation failed...");
	    return(1);
	  }
	  if (t) {
	    while(strm_term(in)!='{') {
	      if (!token_stream(in,"{")) break;
	    }
	    if (strm_term(in)!='{') {
	      err_set(1,"bad token sep. after fragment %s",t);
	      return(1);
	    }
	    curlev++;
	  }
	  while (curlev==2) {
	    t=token_stream(in,"}:");
	    if (!t) break;
	    if (strm_term(in)==':') {
	      if (streq(t,"TOPOLOGY")) {

		/* PARSE TOPOLOGY */

		for(;;) {
		  t=token_stream(in,"};:");
		  if (!t) {
		    err_set(1,"fugu-1 in topology");
		    return(1);
		  }
		  if (strm_term(in)=='}') {
		    curlev--;
		    break;
		  }
		  if (strm_term(in)==':') break;
		  if (streq(t,"ATOM")) {
		    t=token_stream(in,";");
		    if (t) {
		      a=(Atom *)add_node((Node *)frag,t,OBJ_ATOM);
		      if (!mvld(a)) {
			err_set(1,"mem. alloc. failed");
			return(1);
		      }
		    }else{
		      err_set(1,"fugu-2 in topology");
		      return(1);
		    }
		    while (strm_term(in)!=';') {
		      t=token_stream(in,"();");
		      if (!t) {
			err_set(1,"fugu-3 in topology");
			return(1);
		      }else{
			if (streq(t,"ELEMENTS")) {
			  t=token_stream(in,"();");
			  if (!t) {
			    err_set(1,"fugu-4 in topology"); 
			    return(1);
			  }else{
			    sprintf(tmp,"%s.?",t);
			    def=find_atomdef_name((Node *)reslib,tmp);
			    if (!mvld(def)) {
			      def=(Atomdef *)copy_node((Node *)default_type);
			      if (mvld(def)) {
				set_node_name((Node *)def,tmp);
				def->up=(Node *)reslib;
				link_node((Node *)def);
				Strcpy(def->chs,t);
				S=find_symbol_name((Node *)reslib,t);
				if (mvld(S)) {
				  def->mass=S->mass; def->rcov=S->rcov;
				  def->rvdw=S->rvdw;
				}
			      }
			    }
			    set_atom_type(a,def);
			  }
			}else if (streq(t,"AROMATIC")) {
			  a->arom=TRUE;
			}
		      }
		    }
		  }else if (streq(t,"BOND")) {
		    if (strm_term(in)!=' ') {
		      err_set(1,"bad sep after BOND");
		      return(1);
		    }
		    token_stream(in,"(,)"); /* discard name ... */
		    t=token_stream(in,"(,)");
		    if (strm_term(in)!=',') {
		      err_set(1,"bad bond def...");
		      return(1);
		    }
		    forAtoms(a1,frag) if (streq(a1->name,t)) break;
		    if (!mvld(a1)) {
		      err_set(1,"atom1 in bonddef not found.");
		      return(1);
		    }
		    t=token_stream(in,",)");
		    if (strm_term(in)!=')') {
		      err_set(1,"bad bond def (2)...");
		      return(1);
		    }
		    forAtoms(a2,frag) if (streq(a2->name,t)) break;
		    if (!mvld(a2)) {
		      err_set(1,"atom2 in bonddef not found.");
		      return(1);
		    }
		    b=add_bond_ng(a1,a2);
		    if (!mvld(b)) {
		      err_set(1,"failed to add bond...");
		      return(1);
		    }
		    while(strm_term(in)!=';') {
		      t=token_stream(in,";");
		      if (!t) {
			err_set(1,"missing ';' in bonddef...");
			return(1);
		      }
		      if (streq(t,"SINGLE")) bt=BT_SINGLE;
		      else if (streq(t,"DOUBLE")) bt=BT_DOUBLE;
		      else if (streq(t,"TRIPLE")) bt=BT_TRIPLE;
		      else if (streq(t,"AROMATIC")) bt=BT_AROMATIC;
		    }
		    set_bt(b,bt);
		  }else{
		    while (strm_term(in)!=';'&&strm_term(in)!='}') {
		      if (!token_stream(in,"};")) break;
		    }
		    if (strm_term(in)=='}') {
		      curlev--;
		      break;
		    }
		  }
		}
	      }
	    }else if (strm_term(in)=='}') curlev--;
	  }
	}else if (curlev==1 && streq(t,"CONF")) {

	  /* parse CONFORMATION */

	  token_stream(in,"(");
	  if (strm_term(in)!='(') {
	    err_set(1,"conf is fugu.");
	    return(1);
	  }
	  t=token_stream(in,")");
	  if (!t) {
	    err_set(1,"fugu 1 CONF");
	    return(1);
	  }
	  if (strm_term(in)!=')') {
	    err_set(1,"RPREN missing in conf frag name");
	    return(1);
	  }
	  forMols(frag,mol) if (streq(frag->name,t)) break;
	  if (!mvld(frag)) {
	    err_set(1,"CONF: fragment %s not found.",t);
	    return(1);
	  }
	  while (curlev==1) {
	    t=token_stream(in,"{");
	    if (!t) {
	      err_set(1,"even more fugu... (in CONF, of course)");
	      return(1);
	    }
	    if (strm_term(in)=='{') curlev++;
	  }
	  while(curlev==2) {
	    t=token_stream(in,"}:");
	    if (!t) {
	      err_set(1,"fugu the third in CONF");
	      return(1);
	    }
	    if (strm_term(in)=='}') curlev--;
	    else if (strm_term(in)==':') {
	      if (streq(t,"COORDS")) {
		
		/* now get coordinates ... */

		forAtoms(a,frag) {
		  t=token_stream(in,",}");
		  if (!t) {
		    err_set(1,"missing x coord.");
		    return(1);
		  }
		  if (a->nxyz>=a->mxyz) reall_atom_x(a,a->nxyz+1);
		  if (a->nxyz<a->mxyz) {
		    X=a->x+ATM_LOCSZ*a->nxyz;
		    if (strm_term(in)=='}') {curlev--;break;}
		    if (strm_term(in)==',') X[0]=str2f(t)*0.01;
		    t=token_stream(in,",}");
		    if (!t) {
		      err_set(1,"missing y coord.");
		      return(1);
		    }
		    if (strm_term(in)=='}') {curlev--;break;}
		    if (strm_term(in)==',') X[1]=str2f(t)*0.01;
		    t=token_stream(in,";}");
		    if (!t) {
		      err_set(1,"missing x coord.");
		      return(1);
		    }
		    if (strm_term(in)=='}') {curlev--;break;}
		    if (strm_term(in)==';') X[2]=str2f(t)*0.01;
		    a->nxyz++;
		  }
		}
	      }
	    }
	  }
	}
      }else{
	curlev=0;
      }
    }
      
  }else{
    err_set(1,"expected COMPOUND, not found...");
    status=1;
  }
  
  if (status==0) {

    /* delete atoms with unknown 3D coordinates */

    for(a=rfirst_atom(mol);mvld(a);a=a1) {
      a1=rnext_atom(mol,a);
      if (a->nxyz==0) del_node((Node *)a);
    }

    /* remove empty fragments... */

    for (frag=first_mol(mol);mvld(frag);frag=nfrag) {
      nfrag=rnext_mol(mol,frag);
      n=0;
      ForAtoms(a,frag) n++;
      if (n==0) del_node((Node *)frag);
    }

    /* if only one fragment, move up one level... */

    if (mol->fst==mol->lst) {
      if (mvld(mol->fst)) {
	frag=(Mol *)mol->fst;
	unlink_node((Node *)frag);
	for (nd=frag->fst;mvld(nd);nd=frag->fst) {
	  unlink_node(nd);
	  nd->up=mol;
	  link_node(nd);
	}
	del_node((Node *)frag);
      }
    }
  }

  return(status);
}
