#include "object.h"
#include "swdef.h"
#include "streamio.h"
#include "token.h"
#include "strutil.h"
#include "gcalc.h"
#include "xtal.h"
#include "errhand.h"
#include "mystdlib.h"
#include "gcalc.h"
#include <math.h>
#include <ctype.h>

#ifndef M_PI
#define M_PI  3.14159265358979323846
#endif

#define TRUE 1
#define FALSE 0
#define NO 0

#define MAXTOKENSIZE 256
#define MAXADFIXES 1000
#define MAX_ITEMS 512

#define Return(x,y) \
   {err_set(x,"read_cif: %s",y);\
    if (own_p) UCpclose(pipe);	\
    return(x);\
   }

#define Fgets(x,f) {if (!UCpgets(x,sizeof(x),f)) *x='\0';	\
                    while (strrchr(x,'\n')) *strrchr(x,'\n')='\0'; }

extern Node *LibRoot;

static int GetToken(UCpipe *in,Parse *p);
static void parse_formula(Xcn *xcn,char *formula);
static void parse_float_sigma(char *text,float *v,float *s);
static void add_history(Xcn *xcn,char *text);

/*******************************************************************/
/*                                                                 */
/* CIF reader: reads CIF files prodced by SHELX (and maybe others) */
/* For a descripton of CIF c.f. Acta. Cryst. (1991). A47, 655-685  */
/*                                                                 */
/*******************************************************************/


int read_cif(UCpipe *pipe,char *file,Node *node)
{
  char tok[MAXTOKENSIZE],rec[4096],*item[MAX_ITEMS],buf[4096];
  Parse p;
  FILE  *in;
  Atom *a;
  Atomdef *default_type;
  Xatom *xa;
  Xcn  *xcn;
  Node *asym;
  Node *ResLib;
  int in_text,reparse,old_err,i,j,status,nitems,niter,itm,own_p;
  float v,s,Z;

  own_p=(pipe ? FALSE : TRUE);

  if (!is_mol(node)) Return(2,"root is not a node");
  
  xcn=first_xcn(node);
  if (!xcn) {
    xcn=(Xcn *)add_node(node,"Xcn",OBJ_XCN);
  }
  if (!xcn) Return(3,"root has no XCN.");

  asym=add_node(node,"A",OBJ_MOL);
  if (!asym) Return(4,"memory allocation failure");

  if (own_p) pipe=UCpopen(extfil(tok,file,".cif",NO));
  if (pipe) {
    in=pipe->fd;
  }else{
    in=(FILE *)NULL;
  }
  if (!in) Return(1,"could not open input file");

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    if (own_p) UCpclose(pipe);
    return(-1);
  }
  default_type=find_atomdef_name(ResLib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  reparse=FALSE;
  status=0;
  Z=1;

  p.buf=rec; p.sep="#"; p.mts=sizeof(tok)-1; p.tok=tok;
  in_text=FALSE;

  for (;;) {
    if (!reparse) Fgets(rec,pipe); if (strlen(rec)==0&&UCpeof(pipe)) break;
    if (in_text) {
      if (*rec==';') in_text=!in_text;
      continue;
    }else if (*rec==';') {
      in_text=!in_text;
      continue;
    }
    if (!reparse) {
      p.pos=0; if (!get_token(&p)) continue;
    }else reparse=FALSE;
    
    if (strnceq("data_",tok,5)) {

      /* use data_ record for title... */

      xcn->title=Strdup(p.tks+5);

      /* parse cell geometry */

    }else if (strceq("_chemical_formula_sum",tok)) {
      if (GetToken(pipe,&p)) parse_formula(xcn,tok);
      else Return(1,"rdcif: missing chemical formula.");
    }else if (strceq("_cell_length_a",tok)) {
      v=1.0; s=0.0;
      if (GetToken(pipe,&p)) parse_float_sigma(tok,&v,&s);
      else Return(1,"rdcif: missing _cell_length_a value");
      xcn->cl[0]=v; xcn->esd_cl[0]=s;
    }else if (strceq("_cell_length_b",tok)) {
      v=1.0; s=0.0;
      if (GetToken(pipe,&p)) parse_float_sigma(tok,&v,&s);
      else Return(1,"rdcif: missing _cell_length_b value");
      xcn->cl[1]=v; xcn->esd_cl[1]=s;
    }else if (strceq("_cell_length_c",tok)) {
      v=1.0; s=0.0;
      if (GetToken(pipe,&p)) parse_float_sigma(tok,&v,&s);
      else Return(1,"rdcif: missing _cell_length_c value");
      xcn->cl[2]=v; xcn->esd_cl[2]=s;
    }else if (strceq("_cell_angle_alpha",tok)) {
      v=90.0; s=0.0;
      if (GetToken(pipe,&p)) parse_float_sigma(tok,&v,&s);
      else Return(1,"rdcif: missing _cell_angle_alpha value");
      xcn->ca[0]=v; xcn->esd_ca[0]=s;
    }else if (strceq("_cell_angle_beta",tok)) {
      v=90.0; s=0.0;
      if (GetToken(pipe,&p)) parse_float_sigma(tok,&v,&s);
      else Return(1,"rdcif: missing _cell_angle_beta value");
      xcn->ca[1]=v; xcn->esd_ca[1]=s;
    }else if (strceq("_cell_angle_gamma",tok)) {
      v=90.0; s=0.0;
      if (GetToken(pipe,&p)) parse_float_sigma(tok,&v,&s);
      else Return(1,"rdcif: missing _cell_angle_gamma value");
      xcn->ca[2]=v; xcn->esd_ca[2]=s;
    }else if (strceq("_cell_formula_units_Z",tok)) {
      if (GetToken(pipe,&p)) Z=str2f(tok);
      if (Z<1) Z=1.0;

    }else if (strceq("_symmetry_space_group_name_H-M",tok)) {
      if (mvld(xcn->spgr)) free(xcn->spgr);
      xcn->spgr=(char *)NULL;
      if (isprint((int)(p.buf[p.pos]))) {
	if (p.buf[p.pos]=='\'') {
	  get_token(&p); 
	  xcn->spgr=Strdup(tok);
	}else xcn->spgr=Strdup(p.buf+p.pos);
      }else{
	if (GetToken(pipe,&p)) xcn->spgr=Strdup(tok);
      }

    }else if (strceq("_symmetry_equiv_pos_as_xyz",tok)) {
      if (GetToken(pipe,&p)) {
	xcn->sym[0]=str2sym(tok);
	xcn->nsym=1;
      }

    }else if (strceq("_diffrn_radiation_wavelength",tok)) {
      if (GetToken(pipe,&p)) xcn->lambda=str2f(tok);
      else err_set(1,"rdcif: _diffrn_radiation_wavelength value");


    /* miscellaneous information which will be put into history records */

    }else if (strceq("_audit_creation_method",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"created by Wit!P from CIF, audit_creation_method: %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_computing_structure_solution",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"structure solution:   %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_computing_structure_refinement",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"structure refinement: %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_diffrn_reflns_theta_max",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"theta_max: %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_reflns_number_total",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"total number of reclections   : %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_reflns_number_observed",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"number of observed reclections: %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_reflns__observed_criterion",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"criterion                     : %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_refine_ls_R_factor_all",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"R(all): %s",tok);
	add_history(xcn,buf);
      }
    }else if (strceq("_refine_ls_R_factor_obs",tok)) {
      if (get_token(&p)) {
	sprintf(buf,"R(obs): %s",tok);
	add_history(xcn,buf);
      }


    /* parse a loop */

    }else if (strceq("loop_",tok)) {
      nitems=0;

      /* first establish list of loop items... */

      while (GetToken(pipe,&p)) {
	if (*tok=='_') {
	  if (nitems<MAX_ITEMS) {
	    item[nitems]=Strdup(tok);
	    if (item[nitems]) nitems++;
	    else {
	      for (i=0;i<nitems;i++) free(item[i]);
	      nitems=0;
	      err_set(1,"rdcif: memory allocation failed.");
	      Return(1,"rdcif: rest of file ignored.");
	    }
	  }else {
	    for (i=0;i<nitems;i++) free(item[i]);
	    err_set(1,"rdcif: max. number of items in loop exceeded: '%s'",tok);
	    Return(1,"rdcif: rest of file ignored.");
	  }
	}else break;
      }

      /* now read data in groups of 'nitems' items */

      for (niter=0;p.ltok>0;niter++) {
	a=(Atom *)NULL; xa=(Xatom *)NULL;
	for (itm=0;itm<nitems&&p.ltok>0;itm++,GetToken(pipe,&p)) {
	  
	  if (strceq("_symmetry_equiv_pos_as_xyz",item[itm])) {
	    if (niter>=0&&niter<MX_XCN_SYMOPS) {
	      xcn->sym[niter]=str2sym(tok);
	      xcn->nsym=niter+1;
	    }

	    /* atom sites */

	  }else if (strceq("_atom_site_label",item[itm])) { /* must be first in group */
	    if (find_name(asym,tok,OBJ_ATOM)) 
	      err_set(1,"rdcif: atom site label %s is not unique.",tok);
	    a=(Atom *)add_node(asym,tok,OBJ_ATOM);
	    if (is_atom(a)) {
	      set_atom_type(a,guess_atom_type(a->name));
	      xa=(Xatom *)add_node((Node *)a,(char *)NULL,OBJ_XATOM);
	      if (!xa) {
		err_set(1,"rdcif: failed to add cryst. info to %s\n",tok);
		break;
	      }
	      a->xtal=TRUE;
	      a->nxyz=1;
	    }else{
	      err_set(1,"rdcif: failed to add atom %s",tok);
	      break;
	    }

	  }else if (strceq("_atom_site_type_symbol",item[itm])) {
	    if (is_atom(a)) set_atom_type(a,guess_atom_type(tok));
	    
	  }else if (strceq("_atom_site_fract_x",item[itm])) {
	    if (is_atom(a)) parse_float_sigma(tok,a->x,xa->esd_x);

	  }else if (strceq("_atom_site_fract_y",item[itm])) {
	    if (is_atom(a)) parse_float_sigma(tok,a->x+1,xa->esd_x+1);

	  }else if (strceq("_atom_site_fract_z",item[itm])) {
	    if (is_atom(a)) parse_float_sigma(tok,a->x+2,xa->esd_x+2);

	  }else if (strceq("_atom_site_U_iso_or_equiv",item[itm])) {
	    if (is_atom(a)) {
	      parse_float_sigma(tok,xa->bfac,xa->esd_bfac);
	      a->x[ATM_BISO]=xa->bfac[0]*8*M_PI*M_PI;
	    }
	  }else if (strceq("_atom_site_occupancy",item[itm])) {
	    if (is_atom(a)) parse_float_sigma(tok,&a->x[ATM_SOF],&xa->esd_sof);


	    /* aniso temp. factors */

	  }else if (strceq("_atom_site_aniso_label",item[itm])) { /* must be first in group */
	    a=(Atom *)find_name(asym,tok,OBJ_ATOM);
	    if (!is_atom(a)) {
	      err_set(1,"rdcif: aniso atom not found, %s\n",tok);
	      break;
	    }else xa=first_xatom(a);
	    if (!xa) {
	      err_set(1,"rdcif: no xcn found for aniso atom, %s\n",tok);
	      break;
	    }else xa->iso=FALSE;
	  }else if (strceq("_atom_site_aniso_U_11",item[itm])) {
	    if (is_xatom(xa)) parse_float_sigma(tok,xa->bfac,xa->esd_bfac);
	  }else if (strceq("_atom_site_aniso_U_22",item[itm])) {
	    if (is_xatom(xa)) parse_float_sigma(tok,xa->bfac+1,xa->esd_bfac+1);
	  }else if (strceq("_atom_site_aniso_U_33",item[itm])) {
	    if (is_xatom(xa)) parse_float_sigma(tok,xa->bfac+2,xa->esd_bfac+2);
	  }else if (strceq("_atom_site_aniso_U_23",item[itm])) {
	    if (is_xatom(xa)) parse_float_sigma(tok,xa->bfac+3,xa->esd_bfac+3);
	  }else if (strceq("_atom_site_aniso_U_13",item[itm])) {
	    if (is_xatom(xa)) parse_float_sigma(tok,xa->bfac+4,xa->esd_bfac+4);
	  }else if (strceq("_atom_site_aniso_U_12",item[itm])) {
	    if (is_xatom(xa)) parse_float_sigma(tok,xa->bfac+5,xa->esd_bfac+5);

	  }

	}
	if (itm!=nitems) {
	  for (i=0;i<nitems;i++) free(item[i]);
	  nitems=0;
	  err_set(1,"rdcif: bad number of loop items");
	  Return(1,"rdcif: rest of file ignored.");
	}
	if (p.ltok>0&&(*tok=='_'||streq(p.tok,"loop_"))) {
	  break;
	}
      }
      reparse=p.ltok>0;
      for (i=0;i<nitems;i++) free(item[i]);
      nitems=0;
    }
  }

  if (own_p) UCpclose(pipe);

  /* clean up info in xcn */

  xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca); 
  old_err=err_seqnr();
  xcn->atf=inv_4(xcn->fta);
  if (old_err!=err_seqnr()) {
    Return(150,"rdcif: singular cell to angstrom matrix.");
  }


  /* move center of box to origin */

  for (i=0;i<3;i++) {
    xcn->fta.r[i].c[3]=0;
    for (j=0;j<3;j++) xcn->fta.r[i].c[3]-=xcn->fta.r[i].c[j]*0.5;
  }

  old_err=err_seqnr();
  xcn->atf=inv_4(xcn->fta);
  if (old_err!=err_seqnr()) {
    Return(1001,"singular cell to angstrom matrix.");
  }

  /* convert coordinates to cartesian */

  forAtoms(a,asym) v32arr(a->x,Axyz(xcn,vec3(a->x)));

  /* restore aromaticity flags lost due to fugu bonds */

  forAtoms(a,asym) if (mvld(a->atp)) {
    if (a->atp->arom) a->arom=TRUE;
  }

  /* compute equivalent Biso for anisotropic atoms and other cleanup */

  forAtoms(a,asym) {
    xa=first_xatom(a);
    if (xa) {
      if (!xa->iso) {
	a->x[ATM_BISO]=xa->bfac[0]*xa->bfac[1]*xa->bfac[2]
	  +2*xa->bfac[3]*xa->bfac[4]*xa->bfac[5]
	  -xa->bfac[3]*xa->bfac[3]*xa->bfac[0]
	  -xa->bfac[4]*xa->bfac[4]*xa->bfac[1]
	  -xa->bfac[5]*xa->bfac[5]*xa->bfac[2];
	a->x[ATM_BISO]=(a->x[ATM_BISO] > 0 ? pow(a->x[ATM_BISO],1.0/3.0) : 0.01);
	a->x[ATM_BISO]*=8*M_PI*M_PI;
      }
    }
  }

  /* expand symmetry operations */

  symopexp(xcn); 
  i=symopexp(xcn);
  if (i!=xcn->nsymx) {
    err_set(1002,"read_cif: more than %d expanded symops",MX_XCN_SYMOPS);
    xcn->nsymx=xcn->nsym;
  }else{
    symopreduce(xcn);        /* to get xcn->lattice amd xcn->centric right */
    symopexp(xcn); 
  }

  /* convert symmetry operators to angstrom/cartesian */

  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);

  /* and expand formula ... */

  for (i=0;i<xcn->forml;i++) xcn->formn[i]=(int)(Z*xcn->formn[i]+0.5);

  return(status);

}
/*-------------------- GetToken and other parsing stuff --------------------*/

static int GetToken(UCpipe *in,Parse *p)
{
  int r;

  if (p->term=='#') p->pos=strlen(p->buf); /* throw away comment */
  if (!(r=get_token(p))) {
    if (UCpgets(p->buf,p->mts,in)) {
      while (strrchr(p->buf,'\n')) *strrchr(p->buf,'\n')='\0';
      p->pos=0;
      r=get_token(p);
    }else r=0;
  }
  return(r);
}

static void parse_float_sigma(char *text,float *v,float *s)
{
  double sig;

  *s=0; *v=0;
  *v=(float)str2dsigma(text,&sig);
  *s=sig;
}

static void parse_formula(Xcn *x,char *formula)
{
  char *p;

  p=formula;

  for (;*p&&x->forml<MX_XCN_FORMUL;) {
    while (isspace((int)(*p))) p++;
    if (*p >= 'A' && *p <= 'Z') {
      x->formn[x->forml]=1;
      if ((*(p+1)>='A' && *(p+1) <= 'Z') || (*(p+1)>='a' && *(p+1) <= 'z')) {
	x->forms[x->forml][0]=*p; 
	x->forms[x->forml][1]=*(p+1);  
	x->forms[x->forml][2]='\0';
	p++;
      }else{
	x->forms[x->forml][0]=*p; 
	x->forms[x->forml][1]='\0';
      }
      p++;
      if (*p >= '0' && *p <= '9') {
	x->formn[x->forml]=atof(p);
	if (x->formn[x->forml]<0) x->formn[x->forml]=1;
      }
      x->forml++;
      while (*p&&!isspace((int)(*p))) p++;
    }else err_set(1,"bad formula: '%s'\n",formula);
  }
}

static void add_history(Xcn *xcn,char *text)
{
  hisrec *h,*nh;

  for (h=xcn->history;h;h=h->nxt) if (!h->nxt) break;
  nh=make_history();
  if (nh) {
    nh->t=Strdup(text);
    if (h) h->nxt=nh;
    else xcn->history=nh;
  }
}
