#include <stdio.h>
#include "mystdlib.h"
#include <stddef.h>
#include <math.h>
#include <errno.h>
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "errhand.h"
#include "transpor.h"
#include "parser.h"
#include "xtal.h"

#define NO 0
#define LOOP 1
#define DONE 0
#define ON  1
#define OFF 0

#define MAX_MACCS_BT 4
#define MAX_ELS 80

#define key_is(a) (streq(a,key->key))

#define Return(x,y) {err_set(x,y);\
                     return(x);}
#define Fgets(x,f) {if (!UCpgets(x,sizeof(x),f)) *x='\0';	\
                    while (strrchr(x,'\n')) *strrchr(x,'\n')='\0'; \
		    for (k_=strlen(x);k_<sizeof(x);k_++) x[k_]=' '; \
                    x[sizeof(x)-1]='\0';}

extern Node *MolRoot,*LibRoot,*DefaultNode;
extern Dspobj *DspRoot;
extern int graphix;
extern int autocenter_on_read;

static char *main_keys = "next previous goto accept reject save end ' ' "
       "command color crystal display delete label measure modify "
       "select set";

static int skip_csd(UCpipe *fpipe,FileCache *cache,int target,int now);
static Mol *previous(Node *root,Node *p);
static int sanity_check(Node *r);

#define CACHE_SIZE 250
#define CACHE_LIMIT 25

/**************************************************************************/

int read_csd(UCpipe *fpipe,char *file,Node *root)
     
/*   Read CSD file format. Default extension is .dat, NODE node must
	  be opened by caller. Returns non-zero status iff read fails.     */
     
{
  Token *key;
  Node *save_default,*mol,*prev_mol;
  Mol *m;
  Atom *a;
  Dspobj *global,*droot;
  char mol_name[80],buff[1024],*go_menu;
  int i,nat,nseq,next=0,status,read_it,reject,own_p;
  mat_4 C;
  Xcn *xcn;
  FileCache *cache;
  
  /*   Open file ...  */
  
  own_p=(fpipe ? FALSE : TRUE);
  if (own_p) fpipe=UCpopen(extfil(buff,file,".dat",NO));
  if (!fpipe) {
    err_set(1,"read_csd: could not open %s for read.\n",
	    extfil(buff,file,".dat",NO));
    return(1);
  }

  if (fpipe->pipe!=UCP_FILE) {
    go_menu="1 +1 +10 +100";
    cache=(FileCache *)NULL;
  }else{
    go_menu="1 +1 +10 +100 -1 -10 -100";
    cache=makeFileCache(CACHE_SIZE,CACHE_LIMIT);
  }

  save_default=DefaultNode;
  DefaultNode=root;
  nseq=1;
  status=LOOP;
  droot=glblobj(assocobj(root));
  mol=(Node *)NULL; prev_mol=mol;
  reject=TRUE;

  key=(Token *)NULL;
  while (status==LOOP) {
    read_it=FALSE;
    key=free_Token(key); err_dump();
/*    default_token("next"); */
    key=Get_Kwd("ReadCSD> ",";",main_keys);
    if (key->status==TOK_OK) {
      if (key_is("next")) {
	next=nseq;
	read_it=TRUE;
      }else if (key_is("previous")) {
	if (mvld(cache)) {
	  next=nseq-2;
	  read_it=TRUE;
	}else{
	  err_set(1,"sorry, can't backspace on a pipe.");
	}
      }else if (key_is("goto")) {
	key=free_Token(key);
	default_token("%d",nseq);
	key=Get_Int_M("molecule number: ",";",go_menu);
	if (key->status==TOK_OK) {
	  next=key->n;
	  if (*key->text=='+' || *key->text=='-') next+=nseq-1;
	  if (!mvld(cache) && next<nseq) {
	    err_set(1,"sorry, can't backspace on a pipe.");
	  }else{
	    read_it=TRUE;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad molecule number: %s",key->text);
	}
      }else if (key_is("accept")) {
	reject=FALSE;
      }else if (key_is("reject")) {
	reject=TRUE;
      }else if (key_is("save")) {
	if (is_mol(prev_mol)) {
	  cmd_write((Node *)prev_mol,FT_CAN_WRITE&FT_MOL_TYPES);
	}else{
	  err_set(1,"no molecule to save.");
	}
      }else if (key_is("end")) {
	status=DONE;
      }else{
	if (key_is("command")) {
	  command_loop("Wit!P(ReadCSD)> ");
	  if (!sanity_check(root)) status=DONE;
	}else if (key_is("color")) {
	  cmd_color();
	}else if (key_is("crystal")) {
	  m=previous(root,prev_mol);
	  if (is_mol(m)) {
	    xcn=first_xcn(m);
	    if (is_xcn(xcn)) {
	      cmd_crystal(xcn);
	      if (!sanity_check(root)) status=DONE;
	    }else{
	      err_set(1,"no crystal info for %s",m->name);
	    }
	  }else{
	    err_set(1,"no molecule.");
	  }
	}else if (key_is("delete")) {
	  cmd_delete();
	  if (!sanity_check(root)) status=DONE;
	}else if (key_is("display")) {
	  cmd_display();
	}else if (key_is("set")) {
	  cmd_set();
	}else if (key_is("select")) {
	  cmd_select();
	}else if (key_is("label")) {
	  cmd_label();
	}else if (key_is("measure")) {
	  cmd_measure();
	}else if (key_is("modify")) {
	  cmd_modify();
	}
      }
      if (next<1) {
	read_it=FALSE;
	next=1;
      }
      if (read_it) {
	nseq=skip_csd(fpipe,cache,next,nseq);
	if (UCpeof(fpipe)) {
	  err_set(1,"At end of file.");
	  if (!mvld(cache)) status=DONE;
	  read_it=FALSE;
	}
      }
      if (read_it) {
	sprintf(mol_name,"!MOL%d",next);
	prev_mol=mol;
	mol=add_node(root,mol_name,OBJ_MOL);
	if (mvld(mol)) {
	  if (read_csd_mol(fpipe,mol)>0) status=DONE;
	  lock_evtq();
	  if (UCpeof(fpipe)) {
	    if (!mvld(cache)) status=DONE;
	    del_node(mol); mol=(Node *)NULL;
	  }else{
	    nat=0;
	    for (C=ident_4(),i=0;i<3;i++) C.r[i].c[3]=0.0;
	    ForAtoms(a,mol) {
	      for (i=0;i<3;i++) C.r[i].c[3]-=a->x[i];
	      nat++;
	    }
	    prtGraphics("%d atoms read from %s\n",nat,fpipe->filename);
	    if (autocenter_on_read) {
	      if (nat>0) for (i=0;i<3;i++) C.r[i].c[3]/=nat;
	      ForAtoms(a,mol) transform_atom(C,a);
	      forXcns(xcn,mol) xcn_transform(C,xcn);
	    }
	    global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	    if (mvld(global)) {
	      send_mdl(global,(Mol *)mol);
	      global->root=mol;
	    }
	    ((Mol *)mol)->dsptr=global;
	    nseq=nseq+1;
	  }
	  if (reject) {
	    m=previous(root,prev_mol);
	    if (is_mol(m)) del_node((Node *)m);
	    prev_mol=(Node *)NULL;
	  }
	  unlock_evtq();
	  prev_mol=mol;
	}else{
	  err_set(1,"could not add mol node ");
	  status=DONE;
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad read-csd option: %s",key->text);
    }else if (key->status==TOK_EOC) status=DONE;
  }

  DefaultNode=save_default;
  free_Token(key);
  if (fpipe&&own_p) UCpclose(fpipe);
  if (mvld(cache)) freeFileCache(cache);
  return(0);
}
/*-------------------- read_csd_mol --------------------*/

int read_csd_mol(UCpipe *in,Node *node)
{
  char rec[86],fld[86],*p;
  Atom *a,*b;
  Atomdef *default_type;
  Node *au;
  Xcn *xcn;
  Bond *bnd,bl;
  Node *ResLib;
  float ctol,rad[MAX_ELS],r;
  char els[MAX_ELS][MX_OBJ_CHS],as,err,*errtxt;
  int k_;
  int skip,cell,atfor,n1,n2,n3,n4,nopr,nat,nel,i,k,
      cent;
  
  if (UCpeof(in)) return(-1);

  /*   Open default library */

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    Return(-1,"Failed to open default library");
  }
  default_type=find_atomdef_name(ResLib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  /* Skip to first record of entry... */

  do {
    Fgets(rec,in); if (strlen(rec)==0&&UCpeof(in)) break;
  } while(rec[0]!='#');

  if (UCpeof(in)) Return(0,"At end of file.");

  /* read and decode header line */

  set_node_name(node,strsqz(fld,strfld(fld,rec,2,9)));
  as=rec[66]; err=rec[60];
  p=rec+26;
  n1=str2i(strfld(fld,p,1,3));p+=3;
  n2=str2i(strfld(fld,p,1,3));p+=3;
  n3=str2i(strfld(fld,p,1,3));p+=3;
  n4=str2i(strfld(fld,p,1,3));p+=3;
  nopr=str2i(strfld(fld,p,1,3));p+=3;
  nel=str2i(strfld(fld,p,1,3));p+=3;
  nat=str2i(strfld(fld,p,1,3))+str2i(strfld(fld,p,4,6));p+=12;
  cell=str2i(strfld(fld,p,1,1));p+=2;
  atfor=str2i(strfld(fld,p,1,1));p+=1;
  cent=str2i(strfld(fld,p,1,1));p+=1;

  if (cell!=1) Return(0,"missing unit cell info.");
  if (nat<1) Return(0,"no atoms.");
  if (nel<1) Return(0,"no atom radii.");
  if (nel>MAX_ELS) Return(0,"too many element types");

  xcn=(Xcn *)add_node(node,"",OBJ_XCN);
  if (!mvld(xcn)) Return(1,"failed to generate xcn-node.");
  au=add_node(node,"A",OBJ_MOL);
  if (!mvld(au)) Return(1,"failed to allocate node for asymm. unit.");

  /* read cell and spacegroup */

  Fgets(rec,in); if (strlen(rec)==0&&UCpeof(in)) Return(1,"unexpected EOF.");
  if (mvld(xcn->spgr)) free(xcn->spgr);
  xcn->spgr=Strdup(strfld(fld,rec,64,71));
  for (i=0;i<3;i++) xcn->cl[i]=str2i(strfld(fld,rec+i*6,1,6));
  for (i=0;i<3;i++) xcn->ca[i]=str2i(strfld(fld,rec+18+i*6,1,6));
  p=rec+36;
  for (i=0;i<3;i++) for (k=*p++ - '0';k>0;k--) xcn->cl[i]/=10;
  for (i=0;i<3;i++) for (k=*p++ - '0';k>0;k--) xcn->ca[i]/=10;

  xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca);
  k=err_seqnr();
  xcn->atf=inv_4(xcn->fta);
  if (k!=err_seqnr()) Return(1,"singular cell-to-angstrom matrix.");

  ctol=str2f(strfld(fld,rec,75,76))/100.0;

  errtxt=( err != '0' ? "error" : "");
  prtGraphics("Ref.Code =  %s, Spacegroup = %s, esd = %c %s\n",
	      node->name,xcn->spgr,as,errtxt);
  prtGraphics("a,b,c =  %0.3f, %0.3f, %0.3f; "
	      "alpha,beta,gamma =  %0.2f, %0.2f, %0.2f\n",
	      xcn->cl[0],xcn->cl[1],xcn->cl[2],xcn->ca[0],xcn->ca[1],xcn->ca[2]);

  for (skip=(n1+n2+n3+n4+79)/80;skip>0;skip--) {
    if (!UCpgets(rec,85,in)) Return(1,"unexpected EOF.");
    if (strchr(rec,'\n')) *strchr(rec,'\n')='\0';
    prtGraphics("%s\n",rec);
  }
  if (nopr>0) xcn->nsym=0;
  for (skip=(nopr+4)/5;skip>0;skip--) {
    if (!UCpgets(rec,85,in)) Return(1,"unexpected EOF.");
    for (p=rec,i=0;i<5;i++) {
      if (xcn->nsym>=nopr) break;
      xcn->sym[xcn->nsym++]=csd2sym(p);
      p+=15;
    }
  }
  xcn->centric=(cent==1);
  symopexp(xcn); symopreduce(xcn);
  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);
  
  if (nel>0) {
    for (i=0;i<nel;i++) {
      if (i%16==0) {
	Fgets(rec,in); if (strlen(rec)==0&&UCpeof(in)) Return(1,"unexpected EOF.");
	p=rec;
      }
      strcpy(els[i],strsqz(fld,strfld(fld,p,1,2)));
      rad[i]=str2f(strfld(fld,p,3,5))/100; p+=5;
    }
  }

  if (atfor==1) {
    for (i=0;i<nat;i++) {
      if (i%4==0) {
	Fgets(rec,in); if (strlen(rec)==0&&UCpeof(in)) Return(1,"unexpected EOF.");
	p=rec;
      }
      a=(Atom *)add_node(au,strsqz(fld,strfld(fld,p,1,5)),OBJ_ATOM);
      if (mvld(a)) {
	p+=5; a->nxyz=1;
	for (k=0;k<3;k++) {
	  a->x[k]=str2f(strfld(fld,p,1,5))/10000;
	  p+=5;
	}
      }else{
	Return(1,"memory allocation failed.");
      }
    }
  }else if (atfor==2) {
    for (i=0;i<nat;i++) {
      if (i%3==0) {
	Fgets(rec,in); if (strlen(rec)==0&&UCpeof(in)) Return(1,"unexpected EOF.");
	p=rec;
      }
      a=(Atom *)add_node(au,strsqz(fld,strfld(fld,p,1,5)),OBJ_ATOM);
      if (mvld(a)) {
	p+=5; a->nxyz=1;
	for (k=0;k<3;k++) {
	  a->x[k]=str2f(strfld(fld,p,1,7))/100000;
	  p+=7;
	}
	p+=1;
      }else{
	Return(1,"memory allocation failed.");
      }
    }
  }else{
    Return(0,"No coordinates.");
  }

  /* convert to Angstrom/Cartesian coordinates, assign base atomtype... */
  /* find bonds with slow but for small molecules still useful method.  */

  bl.chain=(Bond *)NULL;
  if (nel==0) {
    strcpy(els[0],"Du");
    rad[0]=0.65;
    nel=1;
  }
  ForAtoms(a,au) {
    v32arr(a->x,Axyz(xcn,vec3(a->x)));
    set_atom_type(a,guess_atom_type(a->name));
    if (mvld(a->atp)) {
      for (k=0;k<nel;k++) if (!strcasecmp(a->atp->chs,els[k])) break;
      if (k<nel) {
	a->seqnr=k;
      }else{
	if (k<MAX_ELS) {
	  strcpy(els[k],a->atp->chs);
	  rad[k]=a->atp->rcov;
	  a->seqnr=k;
	  nel=k;
	}else{
	  a->seqnr=0;                /* emergency brake... */
	}
      }
    }else{
      a->seqnr=0;
    }
    ForAtoms(b,au) {
      if (b==a) break;
      r=cdist(a->x,b->x);
      if (r<=rad[a->seqnr]+rad[b->seqnr]+ctol) {
	bnd=add_bond_raw(a,b);
	if (mvld(bnd)) {
	  bnd->chain=bl.chain;
	  bl.chain=bnd;
	}
      } 
    }	  
  }

  /* try to assign bond types... */

  set_bt_automatic(bl.chain);
  clear_bond_chain(&bl);

  return(0);
}
/*-------------------- skip_csd --------------------*/	  

static int skip_csd(UCpipe *fpipe,FileCache *cache,int target,int now)
{
  FILE *in;
  int k,skipped,to_skip;
  char rec[82];

  in=fpipe->fd;

  if (mvld(cache)) {
    addFileCachePos(cache,now,ftell(in));
    k=findFileCachePos(cache,target);
    if (now>target || target-now>target-cache->nr[k]) {
      now=cache->nr[k];
      fseek(in,cache->pos[k],SEEK_SET);
    }
    to_skip=target-now;
    skipped=0;
    while(skipped<to_skip) {
      do {
	if (!UCpgets(rec,81,fpipe)) return(now+skipped);
      } while (rec[0]!='#');
      
      skipped++;
    }
    return(now+skipped);
  }else{
    to_skip=target-now;
    skipped=0;
    while(skipped<to_skip) {
      do {
	if (!UCpgets(rec,81,fpipe)) return(now+skipped);
      } while (rec[0]!='#');
      skipped++;
    }
    return(now+skipped);
  }
}
/*-------------------- previous --------------------*/

static Mol *previous(Node *root,Node *p)
{
  Mol *m;

  forMols(m,root) if (p==(Node *)m) break;
  return(m);
}
/*-------------------- sanity_check --------------------*/

static int sanity_check(Node *root)
{
  Mol *r;

  ForMols(r,MolRoot) if (root==(Node *)r) break;
  return(is_mol(r));
}
