#include <stdio.h>
#include <math.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "transpor.h"
#include "errhand.h"
#define NO 0
#define YES 1
#define Return(x) {if (own_p)UCpclose(in);return(x);}
#define MAXTOKENSIZE 133
#define RECSZ 256
#define TRUE 1
#define FALSE 0

#define key_is(s) (strncmp(rec,s,strlen(s))==0)

extern int graphix;
extern Node *LibRoot;

/*********************************************************************/
int read_diana(UCpipe *in,char *file,Node *node)
     
     /*   Read DIANA coordinate file, default extension is .cor,
	  Returns non-zero status if read fails.  */
{
  char rec[RECSZ],cur_res[20],resnam[20],resnr[20],name[20];
  char res[20],sernr[20],x[3][20];
  
  Node *res_node=(Node *)NULL,*cthis,*cnext,*r,*rn;
  Node *rthis,*rnext;
  Node *ResLib;
  Lib  *reslib;
  Link *L,*L2;
  Atom *a,*b,*t,*l1,*l2;
  Atmmatch *mthis,*mnext;
  Atomdef *default_type;
  Bond *bnd,*nb;
  int i,j,k,n,nat,maxsernr,ia,ib,recnr,own_p;
  float d;
  

  own_p=(in ? FALSE : TRUE);
  if (own_p) in=UCpopen(extfil(rec,file,".cor",NO));
  if (!in) return(1);
  
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    Return(-1);
  }
  reslib=(Lib *)ResLib;
  default_type=find_atomdef_name((Node *)reslib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }
  
  cur_res[0]='\0';
  cthis=node;
  nat=0;
  maxsernr=0;
  a=(Atom *)NULL;
  recnr=0;
  
  while(UCpgets(rec,RECSZ,in)) {
    if (strrchr(rec,'\n')) *strrchr(rec,'\n')='\0';

    if (recnr<3) {                 /* first three records are headers */
      prtGraphics("%s\n",rec);
      recnr++;
    }else{                         /* rest are atoms */
      recnr++;
      for (i=strlen(rec);i<80;i++) rec[i]=' ';
      rec[i]='\0';
      strfld(resnam,rec,19,23); strfld(resnr,rec,12,17);
      sprintf(res,"%s_%s",resnam,resnr);
      strsqz(res,res);
      if (!streq(cur_res,res)) {
	res_node=add_node(cthis,res,OBJ_MOL);
	if (!res_node) Return(3);
	strcpy(cur_res,res);
	((Mol *)res_node)->moltype=MOL_RESIDUE;
	a=(Atom *)NULL;
      }
      
      strfld(name,rec,7,11); strfld(sernr,rec,1,5); 
      strfld(x[0],rec,24,34); strfld(x[1],rec,35,45); strfld(x[2],rec,46,56);
      strsqz(name,name);
      
      a=(Atom *)add_node(res_node,(char *)NULL,OBJ_ATOM);
      if (!a) Return(4);
      n=err_seqnr();
      a->seqnr=str2i(sernr);
      for (i=0;i<3;i++) a->x[i]=str2f(x[i]);
      a->nxyz=1;
      n=err_seqnr()-n;
      if (n>0) {
	err_pop(n);
	err_set(1,"read diana: format error, line =");
	err_set(1,"%s",rec);
	Return(-1);
      }
      set_node_name((Node *)a,name);
      set_atom_type(a,default_type);
      maxsernr=(a->seqnr > maxsernr ? a->seqnr : maxsernr);
      nat++;
    }
  }
  
  /* Add info from residue library */
  
  mnext=(Atmmatch *)NULL;
  r=(Node *)NULL;
  cthis=(Node *)NULL;
  rthis=rfirst_type(node,OBJ_MOL);
  if (rthis) {
    if (((Mol *)rthis)->moltype==MOL_RESIDUE) {
      cthis=rthis->up;
      r=find_residue_name((Node *)reslib,rthis->name);
    }
  }
  mthis=(r ? match_residue_atoms(r,rthis) : (Atmmatch *)NULL);
  
  do { /* atomtypes, connectivity */
    if (mthis) {
      for (k=0;k<mthis->count;k++) {
	t=mthis->a[k]; a=mthis->b[k];
	if (a) {
	  if (mvld(t)) set_atom_type(a,t->atp);
	  for (i=0;i<t->nbnd;i++) {
	    bnd=t->bnd[i]; 
	    a=mthis->b[bnd->f->seqnr]; 
	    b=mthis->b[bnd->t->seqnr];
	    nb=add_bond(a,b);
	    set_bt(nb,bnd->bt);
	  }
	}
      }
    }
    
    /* inter residue links */
    
    cnext=(Node *)NULL;
    rn=(Node *)NULL;
    rnext=rnext_type(node,rthis,OBJ_MOL);
    if (rnext) {
      if (((Mol *)rnext)->moltype==MOL_RESIDUE) {
	cnext=rnext->up;
	rn=find_residue_name((Node *)reslib,rnext->name);
      }
    }
    mnext=(rn ? match_residue_atoms(rn,rnext) : (Atmmatch *)NULL);
    
    if (mthis && mnext && cthis==cnext) {
      forLinks(L,r) {
	if (!find_link_name(rn,L->name)) continue;
	forLinks(L2,reslib) {
	  if (streq(L->name,L2->name)) {
	    strcpy(rec,L2->from);
	    if (strchr(rec,'/')) {
	      *strchr(rec,'/')='\0';
	      if (strcmp(r->name,rec)) continue;
	      strcpy(rec,rec+strlen(rec)+1);
	    }
	    for (ia=0;ia<mthis->count;ia++) {
	      if (streq(mthis->a[ia]->name,rec)) break;
	    }
	    if (ia>=mthis->count) continue;
	    strcpy(rec,L2->to);
	    if (strchr(rec,'/')) {
	      *strchr(rec,'/')='\0';
	      if (strcmp(rn->name,rec)) continue;
	      strcpy(rec,rec+strlen(rec)+1);
	    }
	    for (ib=0;ib<mnext->count;ib++) {
	      if (streq(mnext->a[ib]->name,rec)) break;
	    }
	    if (ib>=mnext->count) continue;
	    l1=mthis->b[ia]; l2=mnext->b[ib];
	    for (d=0.0,j=0;j<3;j++) d+=(l1->x[j]-l2->x[j])*(l1->x[j]-l2->x[j]);
	    d=sqrt(d);
	    if (d>=L2->dmin && d<=L2->dmax) {
	      bnd=add_bond(l1,l2);
	      set_bt(bnd,L2->bt);
	    }
	    break;
	  }
	}
      }
    }
    del_atmmatch(mthis);
    r=rn;
    rthis=rnext;
    mthis=mnext;
    cthis=cnext;
  } while(rthis);
  del_atmmatch(mnext);
  
  if (own_p) UCpclose(in); 
  return(0);
}
