#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "streamio.h"
#include "errhand.h"

#define YES 1
#define NO  0
#define MAXTOKENSIZE 1024
#define MAXBND 512
#define WHITE 7

extern Node *LibRoot;
extern MPEOEatom MPEOEAtoms;
extern MPEOEbond MPEOEBonds;
extern int graphix;

#define key_is(x) \
  (strcmp(key,x)==0)

#define nextkwdmatch(x,s) \
  ((kwd_match(token_stream(in,s),x,key))>=0)


#define breakout(s) \
{status=(s);break;}

static char *mainkeys  = "link residue symbol type bond ffrule bond_equivalent "
                         " bond_increment charges MPEOE";
static char *typdkeys  = "symbol mass number valence geometry aromatic color " 
  " r_covalent vdwradius planar donor acceptor GBrad GBscr";
static char *symdkeys  = "mass z valence geometry color r_covalent vdwradius GBrad GBscr"
  " valence";
static char *linkkeys  = "from to bondtype dmin dmax";
static char *resdkeys  = "atom link";
static char *atmkeys   = "type bonds";

/****************************************************************************/

/*   Read WITNOTP library file    */

int read_lib(char *filename,Lib *lib)
{

  char buf[4096];
  Stream *in;

  in=open_stream(extfil(buf,filename,".wlb",NO),MAXTOKENSIZE);
  return(read_lib_stream(in,lib));

}

int read_lib_ucp(UCpipe *ucp,Lib *lib)
{
  Stream *in;

  in=open_stream((char *)NULL,MAXTOKENSIZE);
  set_stream_pipe(in,ucp);
  return(read_lib_stream(in,lib));
}

int read_lib_stream(Stream *in,Lib *lib)
{
  /* read library file from 'stream' in. 
     Stream will be closed at end-of-file or on error */

  char ffmnem[MAXTOKENSIZE],rsetname[MAXTOKENSIZE];
  char buf[MAXTOKENSIZE];
  char key[MAXTOKENSIZE];
  char msgbuff[MAXTOKENSIZE];
  char *q,*type1,*type2,*def;
  int status,i,k,elev;
  int Z,lev,typd_lev,atm_lev,bnd_lev,resd_lev,symd_lev,linkd_lev,count;
  int pr_interp=0;
  int nbnd,C,had_it;
  char n2[MAXBND][9];
  int bo[MAXBND],B;
  Atomdef *N,*default_type=(Atomdef *)NULL;
  Symbol *S;
  Link *L,*LD;
  Node *r,*node,*rset;
  Atom *a,*a1[MAXBND],*a2;
  Bond *bnd;
  Bondinc *Bi;
  Prule *pr;
  Frule *fr;
  float v1,v2,Q;
  MPEOEatom *mpeoea,*mpa;
  MPEOEbond *mpeoeb,*mpb;
  Node *ResLib=NULL,*PruleLib=NULL,*CruleLib=NULL,*FruleLib=NULL;

  if (!in) return(1);
  status=0;
  
  set_stream(in,"eol=no, redir=no, comment='#'");
  strm_tok(in)=buf; strm_mts(in)=MAXTOKENSIZE;
  
  if (!lib) {
    lib=(Lib *)LibRoot;
  }

  if (lib) {
    ResLib=find_name((Node *)lib,"ResLib",OBJ_LIB);
    PruleLib=find_name((Node *)lib,"PruleLib",OBJ_LIB);
    FruleLib=find_name((Node *)lib,"FruleLib",OBJ_LIB);
    CruleLib=find_name((Node *)lib,"CruleLib",OBJ_LIB);
    if (ResLib) {
      default_type=find_atomdef_name(ResLib,"unk");
      if (!mvld(default_type)) {
	default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
      }
    }else status=31;
  }else status=3;
  
  /* find last MPEOE atom and bond in current list */
  for (mpeoea=&MPEOEAtoms;mpeoea->nxt;) mpeoea=mpeoea->nxt;
  for (mpeoeb=&MPEOEBonds;mpeoeb->nxt;) mpeoeb=mpeoeb->nxt;
  had_it=FALSE;

  /* begin parsing */
  
  while(!status&&nextkwdmatch(mainkeys,"")) {
    lev=0;
    
    /* Atom type definition */
    
    if (key_is("type")) {   /* decode an atomtype definition */
      token_stream(in,"={");
      if (strm_ltok(in)==0) breakout(11);
      node=find_name(ResLib,buf,OBJ_ATOMDEF);
      if (!mvld(node)) {
	node=copy_node((Node *)default_type);
	if (mvld(node)) {
	  node->up=ResLib;
	  link_node(node);
	  set_node_name(node,buf);
	}
      }
      if (node) {
	N=(Atomdef *)node;
	typd_lev=lev; if (strm_term(in)=='{') lev++;
	do {
	  if (strm_eof(in)) breakout(2);
	  if (nextkwdmatch(typdkeys,"={,}")) {
	    if (key_is("symbol")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      Strcpy(N->chs,token_stream(in,",}"));
	      forSymbols(S,ResLib) {
		if (streq(S->name,N->chs)) {
		  N->mass=S->mass; N->rcov=S->rcov;
		  N->rvdw=S->rvdw; N->color=S->color;
		  N->val=S->val; N->z=S->z;
		  N->GBrad=S->GBrad;
		  N->GBscr=S->GBscr;
		  break;
		}
	      }
	      if (!S) breakout(15);
	    }else if (key_is("aromatic")) {
	      N->arom=1;
	    }else if (key_is("geometry")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      Strcpy(N->geom,token_stream(in,",}"));
	    }else if (key_is("mass")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->mass=str2f(token_stream(in,",}"));
	    }else if (key_is("number")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->numtp=str2i(token_stream(in,",}"));
	    }else if (key_is("color")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->color=str2i(token_stream(in,",}"));
	    }else if (key_is("vdwradius")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->rvdw=str2f(token_stream(in,",}"));
	    }else if (key_is("r_covalent")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->rcov=str2f(token_stream(in,",}"));
	    }else if (key_is("GBrad")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->GBrad=str2f(token_stream(in,",}"));
	    }else if (key_is("GBscr")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->GBscr=str2f(token_stream(in,",}"));
	    }else if (key_is("valence")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      N->val=str2i(token_stream(in,",}"));
	    }else if (key_is("planar")) {
	      N->oop=1;
	    }else if (key_is("acceptor")) {
	      N->hbacc=1;
	    }else if (key_is("donor")) {
	      N->hbdon=1;
	    }
	  }else if (strm_ltok(in)>0) breakout(13);
	  if (strm_term(in)=='{') lev++;
	  else if (strm_term(in)=='}') lev--;
	}while (!status && lev>typd_lev);
	if (status) del_node((Node *)N);
      }else{
	breakout(14);
      }
      
      /* definition of chemical symbol */
      
    }else if (key_is("symbol")) {
      token_stream(in,"={");
      if (strm_ltok(in)==0) breakout(101);
      S=find_symbol_name(ResLib,buf);
      if (!mvld(S)) S=(Symbol *)add_node(ResLib,buf,OBJ_SYMBOL);
      if (S) {
	Z=ptab_n(S->name);
	if (Z>=0) {
	  S->val=ptab_val(Z);
	  S->z=ptab_Z(Z);
	  S->mass=ptab_M(Z);
	}
	symd_lev=lev; if (strm_term(in)=='{') lev++;
	do {
	  if (strm_eof(in)) breakout(2);
	  if (nextkwdmatch(symdkeys,"={,}")) {
	    if (key_is("mass")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->mass=str2f(token_stream(in,",}"));
	    }else if (key_is("z")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->z=str2i(token_stream(in,",}"));
	    }else if (key_is("color")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->color=str2i(token_stream(in,",}"));
	    }else if (key_is("vdwradius")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->rvdw=str2f(token_stream(in,",}"));
	    }else if (key_is("r_covalent")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->rcov=str2f(token_stream(in,",}"));
	    }else if (key_is("GBrad")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->GBrad=str2f(token_stream(in,",}"));
	    }else if (key_is("GBscr")) {
	      if (!strchr(" =",strm_term(in))) breakout(102);
	      S->GBscr=str2f(token_stream(in,",}"));
	    }else if (key_is("valence")) {
	      if (!strchr(" =",strm_term(in))) breakout(12);
	      S->val=str2i(token_stream(in,",}"));
	    }
	  }else if (strm_ltok(in)>0) breakout(103);
	  if (strm_term(in)=='{') lev++;
	  else if (strm_term(in)=='}') lev--;
	}while (!status && lev>symd_lev);
	if (S->z>0 && S->val < 0) {
	  if (S->z<2) S->val=S->z;
	  else if (S->z<10) S->val=S->z - 2;
	  else if (S->z<18) S->val=S->z - 10;
	}
	if (status) del_node((Node *)S);
      }else{
	breakout(104);
      }
      
      /* Residue definition */
      
    }else if (key_is("residue")) {
      token_stream(in,"={");
      if (strm_ltok(in)==0) breakout(210);
      r=find_name(ResLib,buf,OBJ_MOL);
      if (mvld(r)) del_node(r);
      r=add_node(ResLib,buf,OBJ_MOL);
      if (!r) breakout(200);
      resd_lev=0; if (strm_term(in)=='{') lev++;
      nbnd=0;
      do {
	if (strm_eof(in)) breakout(2);
	if (nextkwdmatch(resdkeys,"={,}")) {
	  if (key_is("atom")) {
	    if (!strchr(" ",strm_term(in))) breakout(225);
	    a=(Atom *)add_node(r,(char *)NULL,OBJ_ATOM);
	    if (!a) breakout(230);
	    token_stream(in,"={");
	    if (strm_ltok(in)==0) breakout(240);
	    set_node_name((Node *)a,strm_tok(in));
	    atm_lev=lev; if (strm_term(in)=='{') lev++;
	    do {
	      if (strm_eof(in)) breakout(2);
	      if (nextkwdmatch(atmkeys,"={,}")) {
		if (key_is("type")) {
		  if (!strchr(" =",strm_term(in))) breakout(250);
		  Strcpy(ffmnem,token_stream(in,",}"));
		  N=find_atomdef_name(ResLib,ffmnem);
		  if (!N) breakout(260);
		  set_atom_type(a,N); a->color=N->color;
		  if (strm_term(in)=='}') lev--;   /* ??? */
		}else if (key_is("bonds")) {
		  if (!strchr(" ={",strm_term(in))) breakout(250);
		  bnd_lev=lev; if (strm_term(in)=='{') lev++;
		  do {
		    if (strm_eof(in)) breakout(2);
		    if (nbnd>MAXBND) breakout(270);
		    Strcpy(n2[nbnd],token_stream(in,",{[}"));
		    if (strm_ltok(in)>0) {
		      bo[nbnd]=BT_SINGLE;      /* default bond type */
		      a1[nbnd]=a;
		      if (strm_term(in)=='[') {
			bo[nbnd]=BT_number(token_stream(in,"]"));
			if (strm_term(in)!=']') breakout(280);
			if (bo[nbnd]<0) breakout(282);
		      }
		      nbnd++;
		    }
		    if (strm_term(in)=='{') lev++;
		    else if (strm_term(in)=='}') lev--;
		  }while (!status && lev>bnd_lev);
		}
	      }else if (strm_ltok(in)==0) {
		if (strm_term(in)=='{') lev++;
		else if (strm_term(in)=='}') lev--;
	      }else breakout(290);
	    }while (!status && lev>atm_lev);
	  }else if (key_is("link")) {
	    if (!strchr(" =",strm_term(in))) breakout(281);
	    L=(Link *)add_node((Node *)r,token_stream(in,",}"),OBJ_LINK);
	    if (L) {
	      if (strm_ltok(in)==0) breakout(282);
	      forLinks(LD,ResLib)if (!strcmp(LD->name,L->name)) break;
	      if (!LD) breakout(284);
	    }else{
	      breakout(282);
	    }
	    if (strm_term(in)=='}') lev--;
	  }
	}else if (strm_ltok(in)==0) {
	  if (strm_term(in)=='{') lev++; 
	  else if (strm_term(in)=='}') lev--;
	}else breakout(300);
      }while (!status && lev>resd_lev);
      
      if (!status) {
	count=0; ForAtoms(a2,r) a2->seqnr=count++;
	for (i=0;i<nbnd;i++) {
	  ForAtoms(a2,r) {
	    if (strcmp(a2->name,n2[i])==0) {
	      bnd=add_bond(a1[i],a2);
	      if (!bnd) {
		err_set(1,"%s:%s--%s",r->name,
			a1[i]->name,n2[i]);
		breakout(90001);
	      }
	      if (bo[i]) set_bt(bnd,bo[i]);
	      break;
	    }
	  }
	  if (!a2) {
	    err_set(1,"residue %s: bond %s -- %s, bad definition\n",
		    r->name,a1[i]->name,n2[i]);
	    breakout(90002);
	  }
	}
      }
      if (status) del_node(r);
      
      /* inter residue link */
      
    }else if (key_is("link")) {
      token_stream(in,"={");
      if (strm_ltok(in)==0) breakout(401);
      L=(Link *)find_name(ResLib,buf,OBJ_LINK);
      if (mvld(L)) del_node((Node *)L);
      L=(Link *)add_node(ResLib,buf,OBJ_LINK);
      if (L) {
	linkd_lev=lev; if (strm_term(in)=='{') lev++;
	L->bo=1; L->bt=BT_SINGLE;
	do {
	  if (strm_eof(in)) breakout(2);
	  if (nextkwdmatch(linkkeys,"={,}")) {
	    if (key_is("from")) {
	      if (!strchr(" =",strm_term(in))) breakout(402);
	      if (L->from) breakout(403);
	      L->from=Strdup(token_stream(in,",}"));
	      if (strm_ltok(in)==0) breakout(404);
	    }else if (key_is("to")) {
	      if (!strchr(" =",strm_term(in))) breakout(405);
	      if (L->to) breakout(406);
	      L->to=Strdup(token_stream(in,",}"));
	      if (strm_ltok(in)==0) breakout(407);
	    }else if (key_is("dmin")) {
	      if (!strchr(" =",strm_term(in))) breakout(405);
	      L->dmin=str2f(token_stream(in,",}"));
	    }else if (key_is("dmax")) {
	      if (!strchr(" =",strm_term(in))) breakout(405);
	      L->dmax=str2f(token_stream(in,",}"));
	    }else if (key_is("bondtype")) {
	      if (!strchr(" =",strm_term(in))) breakout(408);
	      B=BT_number(token_stream(in,",}"));
	      if (B>=0) {
		L->bt=B;
		L->bo=BT_order(B);
	      }else breakout(409);
	    }
	  }else if (strm_ltok(in)>0) breakout(410);
	  if (strm_term(in)=='{') lev++;
	  else if (strm_term(in)=='}') lev--;
	}while(!status && lev>linkd_lev);
	if (status) del_node((Node *)L);
      }else{
	breakout(400);
      }
      
      /* FF atomtype rule */

    }else if (key_is("ffrule")) {
      token_stream(in,"");             /* ruleset name */
      if (strm_ltok(in)==0) breakout(600);
      Strcpy(rsetname,strm_tok(in));
      token_stream(in,"=~?");
      if (strm_ltok(in)==0) breakout(601);
      Strcpy(ffmnem,strm_tok(in));
      if (strm_term(in)=='='||strm_term(in)=='~'||strm_term(in)=='?') {  /* parse a rule */
	if (strm_term(in)=='=') pr_interp=PR_HARD;
	else if (strm_term(in)=='~') pr_interp=PR_SOFT;
	else if (strm_term(in)=='?') pr_interp=PR_SUBST;
	do {
	  token_stream(in,",");
	  if (strm_ltok(in)==0) breakout(602);
	  pr=swilesPrule(strm_tok(in),pr_interp);
	  if (!mvld(pr)) breakout(603);
	  rset=find_name(PruleLib,rsetname,OBJ_LIB);
	  if (!is_lib(rset)) rset=add_node(PruleLib,rsetname,OBJ_LIB);
	  pr->up=rset;
	  if (!is_lib(pr->up)) {
	    del_node((Node *)pr);
	    breakout(604);
	  }else{
	    link_node((Node *)pr);
	  }
	  set_node_name((Node *)pr,ffmnem);
	} while(strm_term(in)==',');
      }else{
	switch (kwd_match(ffmnem,"reset fix",buf)) {
	case 0: /* delete all ffrules in specified ruleset */
	  rset=find_name(PruleLib,rsetname,OBJ_LIB);
	  if (is_lib(rset)) del_node(rset);
	  rset=find_name(FruleLib,rsetname,OBJ_LIB);
	  if (is_lib(rset)) del_node(rset);
	  break;
	case 1: /* parse and add type re-write rule */
	  token_stream(in,"{");
	  if (strm_ltok(in)==0) breakout(612);
	  fr=parseFrule(strm_tok(in),(Lib *)ResLib);
	  if (!mvld(fr)) breakout(613);
	  for (i=0;i<fr->nat;i++) {
	    token_stream(in,",}");
	    if (strm_ltok(in)==0) break;
	    if (streq(strm_tok(in),"*")) fr->oatp[i]=(Atomdef *)NULL;
	    else {
	      fr->oatp[i]=find_atomdef_name(ResLib,strm_tok(in));
	      if (is_atomdef(fr->oatp[i])) fr->oatp[i]->nlnks++;
	      else break;
	    }
	  }
	  if (i<fr->nat||strm_term(in)!='}') {
	    del_node((Node *)fr);
	    breakout(614);
	  }
	  rset=find_name(FruleLib,rsetname,OBJ_LIB);
	  if (!is_lib(rset)) rset=add_node(FruleLib,rsetname,OBJ_LIB);
	  fr->up=rset;
	  if (!is_lib(fr->up)) {
	    del_node((Node *)fr);
	    breakout(615);
	  }else{
	    link_node((Node *)fr);
	  }
	  break;
	default:
	  err_set(1,"bad ffrule keyword: %s",ffmnem);
	  breakout(605);
	}
	if (status) break;
      }
 
     /* equivalent name in bond_increment parameter */

    }else if (key_is("bond_equivalent")) {
      token_stream(in,"");
      if (strm_ltok(in)==0) breakout(701);
      N=find_atomdef_name(ResLib,strm_tok(in));
      token_stream(in,"");
      if (strm_ltok(in)==0) breakout(702);
      if (!mvld(N)) {
	breakout(703);
      }else {
	if (mvld(N->bnd_equiv)) free(N->bnd_equiv);
	N->bnd_equiv=(char *)NULL;
	if (!streq(N->name,strm_tok(in))) N->bnd_equiv=Strdup(strm_tok(in));
      }

    /* bond increment definitions... */

    }else if (key_is("bond_increment")) {
      token_stream(in,"");
      if (strm_ltok(in)==0) breakout(801);
      type1=Strdup(strm_tok(in));
      token_stream(in,"");
      if (strm_ltok(in)==0) breakout(802);
      type2=Strdup(strm_tok(in));
      v1=str2f(token_stream(in,""));
      v2=str2f(token_stream(in,""));
      Bi=add_bondinc((Lib *)ResLib,type1,type2,v1,v2);
      if (mvld(type1)) free(type1);
      type1=(char *)NULL;
      if (mvld(type2)) free(type2);
      type2=(char *)NULL;
      if (!mvld(Bi)) breakout(803);
 
    /* MPEOE parameter definition (atoms and bonds) */

    }else if (key_is("MPEOE")) {
      if (nextkwdmatch("reset atom bond init","")) {
	if (key_is("reset")) {
	  MPEOEresetInitRules();
	  while(MPEOEAtoms.nxt) {
	    mpeoea=MPEOEAtoms.nxt;
	    MPEOEAtoms.nxt=mpeoea->nxt;
	    del_MPEOEatom(mpeoea);
	  }
	  while(MPEOEBonds.nxt) {
	    mpeoeb=MPEOEBonds.nxt;
	    MPEOEBonds.nxt=mpeoeb->nxt;
	    del_MPEOEbond(mpeoeb);
	  }
	  mpeoea=&MPEOEAtoms;
	  mpeoeb=&MPEOEBonds;
	}else if (key_is("atom")) {
	  mpa=(MPEOEatom *)NULL;
	  strm_term(in)=',';
	  while (strm_term(in)==',') {
	    def=token_stream(in,",{");
	    if (!strchr(",{",strm_term(in))) breakout(852);
	    if (streq(def,"default")) {
	      if (mpa) breakout(853);
	      mpa=&MPEOEAtoms;
	      break;
	    }else{
	      if (!mpa) mpa=make_MPEOEatom();
	      if (mpa) {
		pr=swilesPrule(def,PR_HARD);
		if (pr) {
		  if (mpa->p) {
		    mpa->def=Strcat(mpa->def,", ");mpa->def=Strcat(mpa->def,def);
		    for (pr->prv=(Node *)mpa->p;mvld(pr->prv->nxt);pr->prv=(Node *)pr->prv->nxt);
		    pr->prv->nxt=(Node *)pr;
		  }else{
		    mpa->def=Strdup(def);
		    mpa->p=pr;
		  }
		}else{
		  mpa=del_MPEOEatom(mpa);
		  breakout(854);
		}
	      }else breakout(855);
	    }
	  }
	  if (strm_term(in)!='{') status=856;
	  if (status) {
	    if (mpa!=&MPEOEAtoms) mpa=del_MPEOEatom(mpa);
	    breakout(status);
	  }
	  if (mpa) {
	    mpa->P[0]=str2f(token_stream(in,",*"));
	    if (strm_term(in)=='*') {
	      mpa->opt[0]=TRUE;
	      token_stream(in,",");
	      if (strm_ltok(in)>0||strm_term(in)!=',') breakout(857);
	    }
	    mpa->P[1]=str2f(token_stream(in,",*}"));
	    if (strm_term(in)=='*') {
	      mpa->opt[1]=TRUE;
	      token_stream(in,",}");
	      if (strm_ltok(in)>0) breakout(858);
	    }
	    if (strm_term(in)==',') {
	      mpa->P[2]=str2f(token_stream(in,"*}"));
	      if (strm_term(in)=='*') {
		mpa->opt[2]=TRUE;
		token_stream(in,"}");
		if (strm_ltok(in)>0) breakout(859);
	      }
	      mpa->indX=TRUE;
	    }else mpa->P[2]=mpa->P[0]+mpa->P[1];
	    if (strm_term(in)!='}') {
	      if (mpa!=&MPEOEAtoms) del_MPEOEatom(mpa); 
	      breakout(860);
	    }
	    if (mpa!=&MPEOEAtoms) {
	      mpeoea->nxt=mpa;
	      mpeoea=mpa;
	    }
	  }else breakout(861);
	}else if (key_is("bond")) {
	  mpb=(MPEOEbond *)NULL;
	  strm_term(in)=',';
	  while (strm_term(in)==',') {
	    def=token_stream(in,",");
	    if (!strchr(" ,",strm_term(in))) breakout(871);
	    if (streq(def,"default")) {
	      if (mpb) breakout(872);
	      mpb=&MPEOEBonds;
	      break;
	    }else{
	      if (!mpb) mpb=make_MPEOEbond();
	      if (mpb) {
		pr=swilesPrule(def,PR_HARD);
		if (pr) {
		  if (mpb->p) {
		    mpb->def=Strcat(mpb->def,", ");mpb->def=Strcat(mpb->def,def);
		    for (pr->prv=(Node *)(Node *)mpb->p;mvld(pr->prv->nxt);pr->prv=(Node *)pr->prv->nxt);
		    pr->prv->nxt=(Node *)pr;
		  }else{
		    mpb->def=Strdup(def);
		    mpb->p=pr;
		  }
		}else{
		  mpb=del_MPEOEbond(mpb);
		  breakout(873);
		}
	      }else breakout(874);
	    }
	  }
	  if (strm_term(in)!=' ') status=875;
	  if (status) {
	    if (mpb!=&MPEOEBonds) mpb=del_MPEOEbond(mpb);
	    breakout(status);
	  }
	  if (mpb) {
	    mpb->f=str2f(token_stream(in,"*"));
	    if (mpb!=&MPEOEBonds) {
	      mpb->opt=strm_term(in)=='*';
	      mpeoeb->nxt=mpb;
	      mpeoeb=mpb;
	    }
	  }else breakout(876);
	}else if (key_is("init")) {  /* fragment definition for initial charge distribution */
	  token_stream(in,"{");
	  if (strm_ltok(in)==0) breakout(880);
	  if (strm_term(in)=='{') {        /* parse a rule */
	    Strcpy(msgbuff,strm_tok(in));  /* save for messges... */
	    pr=swilesPrule(strm_tok(in),PR_HARD);
	    if (!mvld(pr)) breakout(881);
	    pr->name=Strdup(strm_tok(in));
	    /*AWI 4-Dec-2001:  for (k=0;k<pr->nat;k++) if (pr->atom[k].chrg=='?') pr->atom[k].chrg=0; */
	    elev=err_seqnr();
	    for (k=0;k<pr->nat;k++) {
	      token_stream(in,",}");
	      if (strm_ltok(in)==0) break;
	      if (*(strm_tok(in)+strm_ltok(in)-1)=='*') {
		*(strm_tok(in)+strm_ltok(in)-1)='\0';
		if (!had_it) err_set(1,"Warning, non-covering flag (*) ignored in MPEOE init rule");
		had_it=TRUE;
	      }
	      pr->atom[k].q=str2f(strm_tok(in));
	      if (elev<err_seqnr()) breakout(882);
	      if (strm_term(in)=='}') break;
	    }
	    if (k!=pr->nat-1) breakout(883);
	    for (C=0,Q=0,k=0;k<pr->nat;k++) {
	      Q+=pr->atom[k].q;
	      if (pr->atom[k].chrg!='?') C+=pr->atom[k].chrg;
	    }
	    if (C-Q>0.001||C-Q<-0.001) 
	      err_set(1,"Warning, rule modifies net charge: %s (%d,%0.3f)",msgbuff,C,Q);
	    if (!MPEOEaddInitRule(pr)) {
	      del_node((Node *)pr);
	      err_set(1,"failed to add MPEOE initial charge distribution rule");
	      breakout(884);
	    }
	  }else breakout(885);
	}else breakout(898);
      }else breakout(899);

    /* fragment definition for fragment partial atomic charges */

    }else if (key_is("charges")) {
      token_stream(in,"");             /* ruleset name */
      if (strm_ltok(in)==0) breakout(900);
      Strcpy(rsetname,strm_tok(in));
      token_stream(in,"{");
      if (strm_ltok(in)==0) breakout(901);
      if (strm_term(in)=='{') {        /* parse a rule */
	Strcpy(msgbuff,strm_tok(in));  /* save for messges... */
	pr=swilesPrule(strm_tok(in),PR_HARD);
	if (!mvld(pr)) breakout(902);
	/*AWI 4-Dec-2001: for (k=0;k<pr->nat;k++) if (pr->atom[k].chrg=='?') pr->atom[k].chrg=0; */
	rset=find_name(CruleLib,rsetname,OBJ_LIB);
	if (!is_lib(rset)) rset=add_node(CruleLib,rsetname,OBJ_LIB);
	pr->up=rset;
	if (!is_lib(pr->up)) {
	  del_node((Node *)pr);
	  breakout(903);
	}else{
	  link_node((Node *)pr);
	}
	elev=err_seqnr();
	for (k=0;k<pr->nat;k++) {
	  token_stream(in,",}");
	  if (strm_ltok(in)==0) break;
	  if (*(strm_tok(in)+strm_ltok(in)-1)=='*') {
	    *(strm_tok(in)+strm_ltok(in)-1)='\0';
	    for (i=0;i<pr->atom[k].nbnd;i++) {
	      pr->bond[pr->atom[k].bnd[i]].cover=0;
	    }
	  }
	  pr->atom[k].q=str2f(strm_tok(in));
	  if (elev<err_seqnr()) breakout(904);
	  if (strm_term(in)=='}') break;
	}
	if (k!=pr->nat-1) breakout(905);
	if (pr->interp==PR_HARD) {
	  for (C=0,Q=0,k=0;k<pr->nat;k++) {
	    if (pr->atom[k].chrg!='?') C+=pr->atom[k].chrg;
	    Q+=pr->atom[k].q;
	  }
	  if (C-Q>0.00005||C-Q<-0.00005) 
	    err_set(1,"Warning, rule modifies net charge: %s (%d,%0.4f)",msgbuff,C,Q);
	}
      }else{
	switch (kwd_match(strm_tok(in),"reset",buf)) {
	case 0: /* delete all ffrules in specified ruleset */
	  rset=find_name(CruleLib,rsetname,OBJ_LIB);
	  if (is_lib(rset)) del_node(rset);
	  break;
	default:
	  err_set(1,"bad charges keyword: %s",strm_tok(in));
	  breakout(951);
	}
      }
    }

  }
  
  if (!status && !strm_eof(in)) status=3;
  if (status && status<90000) {
    q=strm_buff(in); 
    if (strlen(q)>0) {
      q=q+strlen(q)-1;
      if (*q=='\n') *q='\0';
    }
    err_set(1,"%s",strm_buff(in));
    for (q=msgbuff,i=0;i<strm_pos(in);i++) {
      if (strm_buff(in)[i]=='\t') *q++='\t';
      else *q++=' ';
    }
    *q++='^'; *q++='\0';
    err_set(1,msgbuff);
  }
  close_stream(in);
  return(status);
}

/*-------------------- find_link_name --------------------*/

Link *find_link_name(Node *lib,char *name)
{
  /* find link in library (non-recursive node search) by name */
  
  Link *l;
  
  forLinks(l,lib) if (streq(l->name,name)) break;
  return(l);
}
/*-------------------- find_symbol_name --------------------*/
Symbol *find_symbol_name(Node *lib,char *name)
{
  /* find symbol in library (non-recursive node search) by name */
  
  Symbol *s;
  
  forSymbols(s,lib) if (streq(s->name,name)) break;
  return(s);
}
/*-------------------- find_residue_name --------------------*/

Node *find_residue_name(Node *lib,char *name)
{
  /* find residue template with same name as r, match template atoms to 
     atoms of r. */
  
  Node *m;
  char tname[21],*u;
  
  /* strip off residue number from residue name */
  
  Strcpy(tname,name);
  u=strchr(tname,'_');
  if (u) *u='\0';
  
  /* search library for template with given name */
  
  if (is_lib(lib)) {
    for (m=first_type(lib,OBJ_MOL);m;m=next_type(m,OBJ_MOL)) {
      if (strcmp(tname,m->name)==0) break;
    }
  }else{
    m=(Node *)NULL;
  }
  return(m);
}
/*-------------------- match_residue_atoms --------------------*/

Atmmatch *match_residue_atoms(Node *a,Node *b)
{
  Atom *r,*l;
  long count;
  Atmmatch *m;
  
  count=0;
  ForAtoms(l,a) count++;
  m=make_atmmatch(count);
  if (m) {
    m->count=0;
    ForAtoms(l,a) {
      ForAtoms(r,b) {
	if (strcmp(r->name,l->name)==0) break;
      }
      m->a[m->count]=l;
      m->b[m->count]=r;
      m->count++;
    }
  }
  return(m);
}
