#include <stdio.h>
#include "mystdlib.h"
#include <stddef.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "errhand.h"
#include "transpor.h"
#include "parser.h"

#define NO 0
#define LOOP 1
#define DONE 0
#define ON  1
#define OFF 0

#define MAX_MACCS_BT 4
#define SHOW_NONE    0
#define SHOW_ALL     1
#define SHOW_SOME    2

static float MACCS_BT[MAX_MACCS_BT]  = 
   { BT_SINGLE, BT_DOUBLE, BT_TRIPLE, BT_AROMATIC} ;

#define key_is(a) (streq(a,key->key))
#define free_menu(x) if (mvld(x)) free(x),x=(char *)NULL

#define Return(x,y) {err_set(x,y);\
                     if(atom)free(atom);return(x);}

#define ReturnB(x,y) {err_set(1,y);\
                     if(atom)free(atom);del_node(x);return(NULL);}

extern Node *LibRoot,*DefaultNode,*MolRoot;
extern Dspobj *DspRoot;
extern int graphix;
extern int autocenter_on_read;

static int verbose = 1;
static int ntags = 0;
static char **tag = (char **)NULL;
static int data_mode=SHOW_SOME;

static int charge[8] = { 0,3,2,1,0,-1,-2,-3 };

static char *main_keys = "next previous goto accept reject save show end '' "
       "command color display delete label measure modify "
       "select set '' macros ''";

static char *Fgets(char *rec,int size,UCpipe *pipe);
static char *FGets(char *rec,int size,UCpipe *pipe);
static Mol *previous(Node *root,Node *p);
static void show_tags(void);
static int sanity_check(Node *r);
static int nsigch(char *rec);

#define CACHE_SIZE 250
#define CACHE_LIMIT 25

static wnpmacro_t *macros=NULL;

/**************************************************************************/

int read_maccs(UCpipe *fpipe,char *file,Node *root,int multi)
     
/*   Read MACCS MOL file format. Default extension is .sdf, NODE node must
	  be opened by caller. Returns non-zero status iff read fails.     */
     
{
  Token *key;
  Node *save_default,*mol,*prev_mol;
  Mol *m;
  Atom *a;
  Dspobj *global,*droot;
  char mol_name[80],buff[1024],*go_menu;
  char *main_menu=NULL;
  int i,j,nat,nseq,next=0,status,read_it,reject=TRUE,own_p;
  float c[3];
  FileCache *cache;
  wnpmacro_t *macro;
  
  /*  Open file ...  */
  
  own_p=(fpipe ? FALSE : TRUE);

  if (own_p) fpipe=UCpopen(extfil(buff,file,".sdf",NO));
  if (!fpipe) {
    err_set(1,"read_maccs: could not open %s for read.\n",
	    extfil(buff,file,".sdf",NO));
    return(1);
  }

  if (fpipe->pipe!=UCP_FILE) {
    go_menu="1 +1 +10 +100";
    cache=(FileCache *)NULL;
  }else{
    go_menu="1 +1 +10 +100 -1 -10 -100";
    cache=makeFileCache(CACHE_SIZE,CACHE_LIMIT);
  }

  save_default=DefaultNode;
  DefaultNode=root;
  nseq=1;
  status=LOOP;
  droot=glblobj(assocobj(root));
  mol=(Node *)NULL; prev_mol=mol;

  main_menu=make_macro_menu(main_keys,macros);

  key=(Token *)NULL;
  while (status==LOOP) {
    read_it=FALSE;
    if (multi) {
      key=free_Token(key); err_dump();
      key=Get_Kwd("ReadMaccs> ",";",main_menu);
    }else{
      status=DONE;
      key=free_Token(key);
      key=make_Token();
      key->status=TOK_OK;
      key->key=strdup("next");
    }
    if (key->status==TOK_OK) {
      if (key_is("next")) {
	next=nseq;
	read_it=TRUE;
      }else if (key_is("previous")) {
	next=nseq-2;
	read_it=TRUE;
      }else if (key_is("goto")) {
	key=free_Token(key);
	default_token("%d",nseq);
	key=Get_Int_M("molecule number: ",";",go_menu);
	if (key->status==TOK_OK) {
	  next=key->n;
	  if (*key->text=='+' || *key->text=='-') next+=nseq-1;
	  if (fpipe->pipe!=UCP_FILE && next<nseq) {
	    err_set(1,"sorry, can't backspace on a pipe.");
	    read_it=FALSE;
	  }else{
	    read_it=TRUE;
	  }
	}else if (key->status==TOK_NOK) {
	  err_set(1,"bad molecule number: %s",key->text);
	  read_it=FALSE;
	}
      }else if (key_is("accept")) {
	reject=FALSE;
      }else if (key_is("reject")) {
	reject=TRUE;
      }else if (key_is("save")) {
	if (is_mol(prev_mol)) {
	  cmd_write((Node *)prev_mol,FT_CAN_WRITE&FT_MOL_TYPES);
	}else{
	  err_set(1,"no molecule to save.");
	}
      }else if (key_is("show")) {
	show_tags();
      }else if (key_is("end")) {
	status=DONE;
	read_it=FALSE;
      }else if (key_is("macros")) {
	macros=getMACCSmacros();
	cmd_macro(macros);
	free_menu(main_menu);
	main_menu=make_macro_menu(main_keys,macros);
      }else{
	if (key_is("command")) {
	  command_loop("Wit!P(ReadMaccs)> ");
	  if (!sanity_check(root)) status=DONE;
	  free_menu(main_menu);
	  main_menu=make_macro_menu(main_keys,macros);
	}else if (key_is("color")) {
	  cmd_color();
	}else if (key_is("delete")) {
	  cmd_delete();
	  if (!sanity_check(root)) status=DONE;
	}else if (key_is("display")) {
	  cmd_display();
	}else if (key_is("set")) {
	  cmd_set();
	}else if (key_is("select")) {
	  cmd_select();
	}else if (key_is("label")) {
	  cmd_label();
	}else if (key_is("measure")) {
	  cmd_measure();
	}else if (key_is("modify")) {
	  cmd_modify();
	}else if (macros) {
	  macro=get_macro_by_name(macros,key->key);
	  if (macro) execute_macro(macro);
	}
      }
      if (next<1) {
	read_it=FALSE;
	next=1;
      }
      if (read_it) {
	nseq=skip_maccs(fpipe,cache,next,nseq);
	if (UCpeof(fpipe)) {
	  err_set(1,"At end of file.");
	  if (!mvld(cache)) status=DONE;
	  read_it=FALSE;
	}
      }
      if (read_it) {
	sprintf(mol_name,"!MOL%d",next);
	prev_mol=mol;
	mol=add_node(root,mol_name,OBJ_MOL);
	if (mvld(mol)) {
	  if (read_maccs_mol(fpipe,mol)>0) status=DONE;
	  lock_evtq();
	  if (UCpeof(fpipe)) {
	    if (!mvld(cache)) status=DONE;
	    if (!mvld(mol->fst)) {
	      del_node(mol); mol=(Node *)NULL;
	    }
	  }
	  if (mvld(mol)) {
	    nat=0;
	    for (i=0;i<3;i++) c[i]=0.0;
	    ForAtoms(a,mol) {
	      for (i=0;i<3;i++) c[i]+=a->x[i];
	      nat++;
	    }
	    prtGraphics("%d atoms read from %s\n",nat,fpipe->filename);
	    if (autocenter_on_read) {
	      if (nat>0) for (i=0;i<3;i++) c[i]/=nat;
	      ForAtoms(a,mol) {
		for (j=0;j<a->nxyz;j++) {
		  for (i=0;i<3;i++) a->x[i+3*j]-=c[i];
		}
	      }
	    }
	    if (multi) {
	      global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	      if (mvld(global)) {
		send_mdl(global,(Mol *)mol);
		global->root=mol;
	      }
	      ((Mol *)mol)->dsptr=global;
	    }else{
	      send_mdl(droot,(Mol *)mol);
	      ((Mol *)mol)->dsptr=droot;
	    }
	    nseq=nseq+1;
	  }
	  if (reject) {
	    m=previous(root,prev_mol);
	    if (is_mol(m)) del_node((Node *)m);
	    prev_mol=(Node *)NULL;
	  }
	  unlock_evtq();
	  prev_mol=mol;
	}else{
	  err_set(1,"could not add mol node ");
	  status=DONE;
	}
      }
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad read-maccs option: %s",key->text);
    }else if (key->status==TOK_EOC) status=DONE;
  }

  DefaultNode=save_default;
  free_Token(key);
  if (fpipe&&own_p) UCpclose(fpipe);
  if (mvld(cache)) freeFileCache(cache);
  free_menu(main_menu);
  return(0);
}
/*-------------------- skip_maccs --------------------*/	  

int skip_maccs(UCpipe *fpipe,FileCache *cache,int target,int now)
{
  FILE *in;
  int k,skipped,to_skip;
  char rec[82];

  in=fpipe->fd;

  if (mvld(cache)) {
    addFileCachePos(cache,now,ftell(in));
    k=findFileCachePos(cache,target);
    if (now>target || target-now>target-cache->nr[k]) {
      now=cache->nr[k];
      fseek(in,cache->pos[k],SEEK_SET);
    }
    to_skip=target-now;
    skipped=0;
    while(skipped<to_skip) {
      do {
	if (!Fgets(rec,81,fpipe)) return(now+skipped);
      } while (rec[0]!='$' || rec[0]!='$' || rec[0]!='$');
      skipped++;
    }
    return(now+skipped);
  }else{
    to_skip=target-now;
    skipped=0;
    while(skipped<to_skip) {
      do {
	if (!Fgets(rec,81,fpipe)) return(now+skipped);
      } while (rec[0]!='$' || rec[0]!='$' || rec[0]!='$');
      skipped++;
    }
    return(now+skipped);
  }
}
/*-------------------- read_maccs_mol --------------------*/

int read_maccs_mol(UCpipe *in,Node *node)
{
  char rec[82],fld[82],thistag[128],*p,*q;
  Atom *a,**atom;
  Atomdef *default_type;
  Bond *bnd;
  Node *ResLib;
  int i,j,k,from,to,bt,nat,nbnd,nal,nstxt,nprop;
  int ci,expchg,ne;
  
  atom=(Atom **)NULL;

  if (!in || UCpeof(in)) return(-1);
  if (!node) return(-1);

  /*   Open default library */

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    Return(-1,"Failed to open default library");
  }
  default_type=find_atomdef_name(ResLib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  /* Skip three header lines... */

  if (!Fgets(rec,81,in)) Return(-1,"At end of file."); 
  if (nsigch(rec)>0) {
    set_node_name(node,rec);
    if (verbose) prtGraphics("%s\n",rec);
  }

  for (i=0;i<2;i++) {
    if (!Fgets(rec,81,in)) break;
    if (verbose&&i>0&&nsigch(rec)>0) prtGraphics("%s\n",rec);
  }
  if (UCpeof(in)) Return(2,"unexpected EOF\n"); 
  
  /* read and decode count line */

  if (!FGets(rec,81,in)) Return(3,"no count line found\n");
  nat=str2i(strfld(fld,rec,1,3));
  nbnd=str2i(strfld(fld,rec+3,1,3));
  nal=str2i(strfld(fld,rec+6,1,3));
  nstxt=str2i(strfld(fld,rec+15,1,3));
  nprop=str2i(strfld(fld,rec+30,1,3));

  /* read atom block */

  for (i=0;i<nat;i++) {
    if (!FGets(rec,81,in)) break;
    a=(Atom *)add_node(node,(char *)NULL,OBJ_ATOM);
    if (!a) break;
    for (j=0;j<3;j++) a->x[j]=str2f(strfld(fld,rec+10*j,1,10));
    a->nxyz=1;
    strfld(fld,rec+31,1,3);strsqz(a->grp,fld);
    sprintf(fld,"%s%d",a->grp,i+1);
    set_node_name((Node *)a,fld); sprintf(a->grp,"??");
    set_atom_type(a,guess_atom_type(a->name));
    ci=str2i(strfld(fld,rec+36,1,3));
    if (ci>=0 && ci <=7) a->chrg=charge[ci];
    a->lig=-1;                                    /* unknown */
  }
  
  if (UCpeof(in)) Return(4,"unexpected EOF in atom section\n");
  if (i<nat) Return(5,"could not assign node for atom\n");

  /* atom pointer table for bond list */

  atom=(Atom **)malloc((nat+1)*sizeof(Atom *));
  if (!mvld(atom)) Return(6,"memory allocation failure\n");
  i=0; ForAtoms(a,node) atom[++i]=a;

  /* read bond list */

  for (j=0;j<nbnd;j++) {
    if (!FGets(rec,81,in)) Return(7,"early EOF in bond list\n");
    from=str2i(strfld(fld,rec,1,3)); to=str2i(strfld(fld,rec+3,1,3));
    if (from<1||from>nat||to<1||to>nat) Return(8,"bad bond record\n");
    bt=str2i(strfld(fld,rec+6,1,3));
    bnd=add_bond_ng(atom[from],atom[to]);
    if (!mvld(bnd)) Return(9,"bond addition failed\n");
    if (bt>=1 && bt <=MAX_MACCS_BT) {
      set_bt(bnd,MACCS_BT[bt-1]);
    }else{
      set_bt(bnd,BT_UNKNOWN);
    }
  }

  /* skip atom list records */

  for (i=0;i<nal;i++) {
    if (!Fgets(rec,81,in)) break;
  }
  if (UCpeof(in)) Return(10,"EOF encountered in atom lists section.\n");

  /* skip stext block */

  for (i=0;i<2*nstxt;i++) {
    if (!Fgets(rec,81,in)) break;
  }
  if (UCpeof(in)) Return(11,"EOF encountered in stext section.\n");

  /* read properties block */

  expchg=FALSE;

  for (i=0;i<nprop;i++) {
    if (!FGets(rec,81,in)) break;
    if (nprop==999&&strneq(rec,"M  END",6)) break;
    if (strneq(rec,"M  CHG",6)) {
      if (!expchg) {
	for (k=1;k<=nat;k++) atom[k]->chrg=0;
	expchg=TRUE;
      }
      ne=str2i(strfld(fld,rec+7,1,3));
      for (j=0;j<ne;j++) {
	k=str2i(strfld(fld,rec+10+j*8,1,4));
	if (k>=1&&k<=nat) {
	  atom[k]->chrg=str2f(strfld(fld,rec+10+j*8,5,8));
	}
      }
    }
  }

  if (UCpeof(in)) Return(12,"EOF encountered in properties section.\n");

  for (k=1;k<=nat;k++) atom[k]->lig=asg_formlig(atom[k]);

  /* search for end of molecule , set node name, if > record found */

  while(!strneq(rec,"$$$$",4)) {
    if (rec[0]=='>' && (q=strchr(rec,'<'))) {
      if ((p=strchr(q+1,'>'))) *p='\0';
      strncpy(thistag,q+1,sizeof(thistag));
      thistag[sizeof(thistag)-1]='\0';
      if (data_mode==SHOW_SOME) {
	for (i=0;i<ntags;i++) if (strceq(tag[i],q+1)) break;
	if (i<ntags) {
	  for(;;) {
	    if (!Fgets(rec,81,in)||nsigch(rec)==0||strneq(rec,"$$$$",4)) break;
	    prtGraphics("%s: %s\n",thistag,rec);
	  }
	}
      }else if (data_mode==SHOW_ALL) {
	for(;;) {
	  if (!Fgets(rec,81,in)||nsigch(rec)==0||strneq(rec,"$$$$",4)) break;
	  prtGraphics("%s: %s\n",thistag,rec);
	}
      }
    }
    if (!Fgets(rec,81,in)) break;
  }

  /* find and mark aromatic bonds... */

  ForAtoms(a,node) {
    for (j=0;j<a->nbnd;j++) {
      if (a->bnd[j]->f==a) asg_bondtype(a->bnd[j]);
    }
  }
  localize_aromatic_bonds(node);  /* ++TMP */

  /* remove some of the more common garbage from mol-name... */

  if (node->name) {
    for (p=q=node->name;*p;p++) {
      if (*p!=' ' && *p!='-' && *p!='!') *q++ = *p;
    }
    *q='\0';
  }

  /* all done */

  if (atom) free(atom);
  atom=(Atom **)NULL;

  return(0);
}
/*------------------------------------------------------------------------*/

wnpmacro_t *getMACCSmacros(void)
{
  if (!macros) macros=new_macro();
  return(macros);
}

/*-------------------- previous --------------------*/

static Mol *previous(Node *root,Node *p)
{
  Mol *m;

  forMols(m,root) if (p==(Node *)m) break;
  return(m);
}
/*-------------------- sanity_check --------------------*/

static int sanity_check(Node *root)
{
  Mol *r;

  ForMols(r,MolRoot) if (root==(Node *)r) break;
  return(is_mol(r));
}
/*-------------------- copied from 3ddb ----------------*/

static char *Fgets(char *rec,int size,UCpipe *pipe)
{
  /* removes trailing \n and \r */

  char *r;
  int c;

  r=UCpgets(rec,size,pipe);
  if (r) {
    if (strrchr(rec,'\n')) *strrchr(rec,'\n')='\0';
    else while ((c=UCpgetc(pipe))!=EOF) if (c=='\n') break;
    if (strrchr(rec,'\r')) *strrchr(rec,'\r')='\0';
    
  }
  return(r);
}

static char *FGets(char *rec,int size,UCpipe *pipe)
{
  /* pads with blank */

  char *r;
  int i;

  r=Fgets(rec,size,pipe);
  if (r) for (i=strlen(r);i<size;i++) rec[i]=' ';
  return(r);
}

/*-------------------- show_tags ----------------*/

static void show_tags(void)
{
  Token *key=(Token *)NULL;;

  key=free_Token(key);
  key=Get_String_M("data field: ",":","-all -clear -none -selected");
  if (key->status==TOK_OK) {
    key=Prs_Kwd(key,"-all -clear -none -selected");
    if (key->status==TOK_OK) {
      if (key_is("-all")) data_mode=SHOW_ALL;
      else if (key_is("-none")) data_mode=SHOW_NONE;
      else if (key_is("-selected")) data_mode=SHOW_SOME;
      else if (key_is("-clear")) {
	if (tag) free(tag);
	tag=(char **)NULL;
	ntags=0;
      }
      key->status=TOK_ABORT;
    }else{
      key->status=TOK_OK;
      tag=realloc(tag,(ntags+1)*sizeof(char *));
      if (tag) tag[ntags++]=strdup(key->text);
      else {
	ntags=0;
	err_set(1,"memory allocation failed, will show all data fields.");
	data_mode=SHOW_ALL;
	key->status=TOK_ABORT;
      }
    }
  }else if (key->status==TOK_NOK) err_set(1,"bad token: %s",key->text);
  free_Token(key);
}

/*-------------------- nsigch --------------------*/

static int nsigch(char *rec)
{
  /* count significant characters in record */

  int r=0;
  char *p;

  if (rec) {
    for (p=rec;*p;p++) if (isgraph(*p)) r+=1;
  }
  return(r);
}
