#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "transpor.h"
#include "errhand.h"

#define NO 0
#define YES 1
#define MAXTOKENSIZE 133
#define RECSZ 4196
#define TRUE 1
#define FALSE 0

extern int graphix;

/*********************************************************************/
int read_diana_noe(char *file,NOElist *noes)
     
     /*   Read DIANA distance constraint file, default extension is .upl,
	  Returns non-zero status if read fails.  */
{
  char rec[RECSZ],*fld[7],cur_res[11],new_res[11];
  Stream *in=NULL;
  int i,status=0;
  float d;
  NOEdata *noe;
  
  if (!is_NOElist(noes)) {
    err_set(1,"bad argument(noes) in read_diana_noe.\n");
    return(1);
  }

  in=open_stream(extfil(rec,file,".upl",NO),RECSZ);
  if (!in) {
    err_set(1,"read_diana_noe: could not open %s\n",extfil(rec,file,".upl",NO));
    return(1);
  }
  strm_tok(in)=rec; strm_mts(in)=sizeof(rec);
  set_stream(in,"eol=yes,comment='#'");
  
  cur_res[0]='\0';
  
  strm_flush(in);
  for (i=0;i<7;i++) fld[i]=NULL;
  while (!strm_eof(in)) {
    for (i=0;i<7;i++) {
      if (token_stream(in,"")) {
	fld[i]=Strdup(strm_tok(in));
      }else{
	break;
      }
    }
    if (i==7) {
      sprintf(cur_res,"%s_%s",fld[1],fld[0]);
      sprintf(new_res,"%s_%s",fld[4],fld[3]);
      d=str2d(fld[6]);
      noe=add_noedata(noes);
      if (mvld(noe)) {
	Strcpy(noe->sres,cur_res);
	Strcpy(noe->tres,new_res);
	Strcpy(noe->satom,fld[2]); noe->def1=Strdup(fld[2]);
	Strcpy(noe->tatom,fld[5]); noe->def2=Strdup(fld[5]);
	noe->dupper=d; noe->weight=1.0;
      }else{
	err_set(1,"read_diana_noe: memory alllocation failure\n");
	status=1;
      }
    }  
    for (i=0;i<7;i++) {
      if (fld[i]) free(fld[i]);
      fld[i]=NULL;
    }
    strm_flush(in);
  }
  
  close_stream(in);
  return(status);
}
/*-------------------- read_xplor_noe --------------------*/

int read_xplor_noe(char *file,NOElist *noes)
     
     /*   Read XPLOR distance constraint file, default extension is .tbl,
	  Returns non-zero status if read fails.  */
{
  Stream *in;
  int status;
  char fld[RECSZ];

  if (!is_NOElist(noes)) {
    err_set(1,"bad argument(noes) in read_xplor_noe.\n");
    return(1);
  }

  in=open_stream(extfil(fld,file,".tbl",NO),RECSZ);
  if (!in) {
    err_set(1,"read_xplor_noe: could not open %s\n",extfil(fld,file,".upl",NO));
    return(1);
  }
  strm_tok(in)=fld; strm_mts(in)=sizeof(fld);
  status=read_xplor_noe_stream(in,noes);
  close_stream(in);

  return(status);
}

/*-------------------- read_xplor_noe_stream --------------------*/

int read_xplor_noe_stream(Stream *in,NOElist *noes)
{
  char sresi[2][5],tresi[2][5],satom[6],tatom[6],sres[11],tres[11];
  char nm1[RECSZ],nm2[RECSZ],fld1[RECSZ],*p,*nl,*ssegid,*tsegid,*cmnt;
  int i,wld1,wld2,plev,l;
  float d,dplus;
  NOEdata *noe;

  set_stream(in,"eol=no,redir='@',comment='!'"); /* set parser to XPLOR style */
  ssegid=(char *)NULL; tsegid=(char *)NULL;

  cmnt=(char *)NULL;

  while (token_stream(in,"()>")) {     /* simplified parse... */
    if (streq(strm_tok(in),"EONOE")||(streq(strm_tok(in),"</N")&&strm_term(in)=='>')) break;
    if (streq(strm_tok(in),"{*")) {   /* useful comment ? */
      l=strrchr(strm_buff(in)+strm_pos(in),'}')-(strm_buff(in)+strm_pos(in))+1;
      if (l>=0) {
	if (cmnt) free(cmnt);
	cmnt=(char *)malloc((l+4)*sizeof(char));
	sprintf(cmnt,"{* ");
	strncpy(cmnt+3,strm_buff(in)+strm_pos(in),l);
	cmnt[l+3]='\0';
	strm_pos(in)+=l;
      }
    }
    if (strm_ltok(in)<4 ||
	kwd_match(strm_tok(in),"assign",(char *)NULL)<0) continue;
    plev=0; if (strm_term(in)=='(') plev++;
    if (plev!=1) {
      err_set(1,"bad NOE syntax (first selection)");
      continue;
    }
    nl=nm1;nl[0]='\0';
    Strcpy(sresi[0],"*"); Strcpy(sresi[1],"*");
    if (ssegid) free(ssegid);
    if (tsegid) free(tsegid);
    ssegid=(char *)NULL; tsegid=(char *)NULL;

    while (plev>0) {                  /* parse the first selection... */
      if (!token_stream(in,"=()")) break;
      plev+=strm_term(in)=='('; plev-=strm_term(in)==')';
      if (strm_ltok(in)>3 &&
	  kwd_match(strm_tok(in),"resname residue segid",fld1)>=0) {
	if (token_stream(in,"=()")) {
	  plev+=strm_term(in)=='('; plev-=strm_term(in)==')';
	  if (streq(fld1,"resname")) Strcpy(sresi[0],strm_tok(in));
	  if (streq(fld1,"residue")) Strcpy(sresi[1],strm_tok(in));
	  if (streq(fld1,"segid")) {
	    if (ssegid) {
	      free(ssegid);
	      ssegid=(char *)NULL;
	    }
	    ssegid=Strdup(strm_tok(in));
	  }
	}else{
	  break;
	}
      }else if (strm_ltok(in)>3 &&
		kwd_match(strm_tok(in),"name",fld1)>=0) {
	if (token_stream(in,"=()")) {
	  plev+=strm_term(in)=='('; plev-=strm_term(in)==')';
	  if (strm_ltok(in)>0) {
	    if (strlen(nm1)>0) *nl++ = ',';
	    strcpy(nl,strm_tok(in));
	    nl+=strlen(nl);
	  }
	}else{
	  break;
	}
      }
    }

    if (plev!=0) {
      err_set(1,"bad syntax in NOE (1st select)");/* should show the line... */
      continue;
    }
    if (strlen(nm1)<1) strcpy(nm1,"*");

    if (!token_stream(in,"()")) {
      err_set(1,"bad NOE syntax");
      continue;
    }
	
    plev=0; if (strm_term(in)=='(') plev++;
    if (plev!=1) {
      err_set(1,"bad NOE syntax (second selection)");
      continue;
    }
    nl=nm2;nl[0]='\0';
    Strcpy(tresi[0],"*"); Strcpy(tresi[1],"*");

    while (plev>0) {                  /* parse the second selection... */
      if (!token_stream(in,"=()")) break;
      plev+=strm_term(in)=='('; plev-=strm_term(in)==')';
      if (strm_ltok(in)>3 &&
	  kwd_match(strm_tok(in),"resname residue segid",fld1)>=0) {
	if (token_stream(in,"=()")) {
	  plev+=strm_term(in)=='('; plev-=strm_term(in)==')';
	  if (streq(fld1,"resname")) Strcpy(tresi[0],strm_tok(in));
	  if (streq(fld1,"residue")) Strcpy(tresi[1],strm_tok(in));
	  if (streq(fld1,"segid")) {
	    if (tsegid) {
	      free(tsegid);
	      tsegid=(char *)NULL;
	    }
	    tsegid=Strdup(strm_tok(in));
	  }
	}else{
	  break;
	}
      }else if (strm_ltok(in)>3 &&
		kwd_match(strm_tok(in),"name",fld1)>=0) {
	if (token_stream(in,"=()")) {
	  plev+=strm_term(in)=='('; plev-=strm_term(in)==')';
	  if (strm_ltok(in)>0) {
	    if (strlen(nm2)>0) *nl++ = ',';
	    strcpy(nl,strm_tok(in));
	    nl+=strlen(nl);
	  }
	}else{
	  break;
	}
      }
    }

    if (plev!=0) {
      err_set(1,"bad syntax in NOE (2nd) select");/* should show the line... */
      continue;
    }
    if (strlen(nm2)<1) strcpy(nm2,"*");

    i=err_since();
    d=str2f(token_stream(in,""));    /* get three floats... */
    dplus=str2f(token_stream(in,""));
    dplus=str2f(token_stream(in,""));
    if (i!=err_since()) {
      err_set(1,"bad float in NOE definition.");
      continue;
    }
    
    for (wld1=FALSE,nl=nm1;*nl;nl++) {
      wld1|=(*nl=='*')||(*nl==',')||(*nl=='#')||(*nl=='+')||(*nl=='%');
    }
    for (wld2=FALSE,nl=nm2;*nl;nl++) {
      wld2|=(*nl=='*')||(*nl==',')||(*nl=='#')||(*nl=='+')||(*nl=='%');
    }

    sprintf(sres,"%s_%s",sresi[0],sresi[1]);
    sprintf(tres,"%s_%s",tresi[0],tresi[1]);

    if (wld1) {                     /* generate pseudo atom name */
      satom[0]='Q';
      for (p=satom+1,nl=nm1+1;*nl;nl++) {
	if (*nl==',') break;
	if (*nl!='*'&&*nl!='#'&&*nl!='%'&&*nl!='+') *p++ = *nl;
      }
      *p='\0';
    }else{                          /* copy atom name */
      Strcpy(satom,nm1);
    }

    if (wld2) {
      tatom[0]='Q';
      for (p=tatom+1,nl=nm2+1;*nl;nl++) {
	if (*nl==',') break;
	if (*nl!='*'&&*nl!='#'&&*nl!='%'&&*nl!='+') *p++ = *nl;
      }
      *p='\0';
    }else{
      Strcpy(tatom,nm2);
    }

    noe=add_noedata(noes);
    if (mvld(noe)) {
      Strcpy(noe->sres,sres);
      Strcpy(noe->tres,tres);
      Strcpy(noe->satom,satom);
      Strcpy(noe->tatom,tatom);
      if (wld1) noe->ps1=TRUE;
      noe->def1=Strdup(nm1);
      if (wld2) noe->ps2=TRUE;
      noe->def2=Strdup(nm2);
      noe->dupper=d+dplus; noe->weight=1.0;
      noe->sseg=ssegid; ssegid=(char *)NULL;
      noe->tseg=tsegid; tsegid=(char *)NULL;
      noe->cmnt=cmnt;
      cmnt=(char *)NULL;
    }else{
      err_set(1,"read_xplor_noe: memory alllocation failure\n");
      if (ssegid) free(ssegid);
      if (tsegid) free(tsegid);
      if (cmnt) free(cmnt);
      return(1);
    }

  }
  
  if (ssegid) free(ssegid);
  if (tsegid) free(tsegid);
  if (cmnt) free(cmnt);
  return(0);
}
/*-------------------- write_xplor(&discover)_noe --------------------*/

static void pseudo_diana_to_xplor(char *xplor,char *diana)
{
  if (diana[0]=='Q') {
    if (diana[1]=='Q') {
      sprintf(xplor,"H%s*",diana+2);
    }else if (diana[1]=='R') {
      strcpy(xplor,diana);    /* wrong, but don't know how to handle this */
    }else if (streq(diana,"QCN")) {
      sprintf(xplor,"HN*");
    }else{
      sprintf(xplor,"H%s*",diana+1);
    }
  }else{
    strcpy(xplor,diana);
  }
}

static void residue_diana_to_xplor(char *xplor,char *diana)
{
  char *n;
  if ((n=strchr(diana,'_'))) {
    if (*diana=='*') {
      strcpy(xplor,n+1);
    }else{
      *n='\0';
      sprintf(xplor,"%s and resnam %s",n+1,diana);
      *n='_';
    }
  }else strcpy(xplor,diana);
}

int write_xplor_noe(char *file,NOElist *noes)
{
  /* write NOE lists as XPLOR NOE table (KLUDGE) */

  FILE *out;
  int status=0;
  char rec[256];

  extfil(rec,file,".tbl",NO);
  out=fopen(rec,"w");
  if (out) {
    write_xplor_noelist(out,noes);
    fclose(out);
  }else {
    err_set(1,"write_xplor_noe: could not open %s for write.\n",rec);
    status=1;
  }
  return(status);
}

void write_xplor_noelist(FILE *out,NOElist *noes)
{
  char satom[80],tatom[80],sres[80],tres[80];
  NOEdata *noe;
  float d,dplus,dminus;

  for (noe=noes->first;mvld(noe);noe=noe->nxt) {
    pseudo_diana_to_xplor(satom,noe->satom);
    pseudo_diana_to_xplor(tatom,noe->tatom);
    residue_diana_to_xplor(sres,noe->sres);
    residue_diana_to_xplor(tres,noe->tres);
    dplus=0.0;                                /* a good choice ? */
    d=noe->dupper-dplus;
    dminus=d-0.0;                             /* a goos choice ? */
    if (dminus<0.0) dminus=0;
    if (noe->cmnt) fprintf(out,"       {* %s *}\n",noe->cmnt);
    if (noe->sseg) {
      if (noe->tseg) {
	fprintf(out,"assign (segid %s and resid %4s and name %s) \n"
		"       (segid %s and resid %4s and name %s) %5.2f %5.2f %5.2f\n",
		noe->sseg,sres,satom,noe->tseg,tres,tatom,d,dminus,dplus);
      }else{
	fprintf(out,"assign (segid %s and resid %4s and name %s) \n"
		"       (resid %4s and name %s) %5.2f %5.2f %5.2f\n",
		noe->sseg,sres,satom,tres,tatom,d,dminus,dplus);
      }
    }else{
      if (noe->tseg) {
	fprintf(out,"assign (resid %4s and name %s) \n"
		"       (segid %s and resid %4s and name %s) %5.2f %5.2f %5.2f\n",
		sres,satom,noe->tseg,tres,tatom,d,dminus,dplus);
      }else{
	fprintf(out,"assign (resid %4s and name %s) \n"
		"       (resid %4s and name %s) %5.2f %5.2f %5.2f\n",
		sres,satom,tres,tatom,d,dminus,dplus);
      }
    }
  }
}

/*--------------------  write_discover_noe --------------------*/

int write_discover_noe(char *file,NOElist *noes)
{
  /* write NOE lists as DISCOVER distance restraint table (KLUDGE) */

  FILE *out;
  char rec[256];
  NOEdata *noe;

  extfil(rec,file,".rstrnt",NO);
  out=fopen(rec,"w");
  if (!out) {
    err_set(1,"write_discover_noe: could not open %s for write.\n",rec);
    return(1);
  }

  fprintf(out,"!BIOSYM restraint 1\n");
  fprintf(out,"#distance\n");

  for (noe=noes->first;mvld(noe);noe=noe->nxt) {
    *rec='\0';
    if (noe->sseg) {
      if (*noe->sseg >= 'A' && *noe->sseg <= 'I') {
	sprintf(rec+strlen(rec),"%c:",*noe->sseg-'A'+'1');
      }
      else{
	sprintf(rec+strlen(rec),"1:");
      }
    }else{
      sprintf(rec+strlen(rec),"1:");
    }
    sprintf(rec+strlen(rec),"%s:%s  ",noe->sres,noe->satom);
    if (noe->tseg) {
      if (*noe->tseg >= 'A' && *noe->tseg <= 'I') {
	sprintf(rec+strlen(rec),"%c:",*noe->tseg-'A'+'1');
      }else{
	sprintf(rec+strlen(rec),"1:");
      }
    }else{
      sprintf(rec+strlen(rec),"1:");
    }
    sprintf(rec+strlen(rec),"%s:%s  ",noe->tres,noe->tatom);
    sprintf(rec+strlen(rec)," 0.0   %3.3f  5.0 5.0 1000.0",noe->dupper);
    fprintf(out,"%s\n",rec);
  }
  fclose(out);
  return(0);
}
      
/*--------------------  write_sodgeom_noe --------------------*/

int write_sodgeom_noe(char *file,NOElist *noes)
{
  FILE *out;
  char rec[256];
  NOEdata *noe;

  extfil(rec,file,".rstrnt",NO);
  out=fopen(rec,"w");
  if (!out) {
    err_set(1,"write_sodgeom_noe: could not open %s for write.\n",rec);
    return(1);
  }

  for (noe=noes->first;mvld(noe);noe=noe->nxt) {
    if (noe->sptr&&noe->tptr) {
      fprintf(out,"%4d %4d %8.3f %8.3f\n",noe->sptr->seqnr,noe->tptr->seqnr,0.0,noe->dupper);
    }
  }
  fclose(out);
  return(0);
}
