#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "strutil.h"
#include "streamio.h"
#include "rpcmethods.h"

/*** Notes: ***/
/*  1) when handleRPC feeds parser, we should be honest about the source SRC_XXX  
    2) we should handle RPC's only if read is to feed parser...
*/


void do_prompt(void);

static int NAPTIME=1;        /* in micro-seconds */
static int MAXNAPTIME=256;

#ifdef USEREADLINE
#define AReadlineFunction(X) readline(X)
#define AHistoryFunction(X)  add_history(X)
#define StifleHistory(X) stifle_history(X)
static char *the_line=(NULL);
static int usereadline=0;
#else
static int usereadline=0;
#endif

char *AReadlineFunction(char *prmt);
void  AHistoryFunction(char *line);
void  StifleHistory(int n);

static void wait_read(int ch,int w);
static void addLF(char *buff,int size);
static char *nwfgets(char *s, int size, FILE *stream);

/*==================== reading via TCP/QIO ====================*/

#include "object.h"
#include "swdef.h"
#include "transpor.h"
#include "parser.h"
#include <memory.h>
#include <errno.h>

extern int server;
extern int graphix;

#define hdr_is(x) (strncmp((x),hdr,4)==0)

int read_null(char *buff,int nch)
{
  char *hdr,hdr4[5];
#ifdef USEREADLINE
  char *s,*new_line,bfr[2048];
#endif
  float xw[6];
  int i,n,got=0;
  Qmsg *Msg;
  char *msgmask;
  cptr_t cross;
  int wait;

  if (server) {
    do_prompt();
    msgmask="KEYB MENU SCCN USAP PATM POBJ PXYZ";
    i=0;
    wait=wnpRPCisOn()?MSGQNOWAIT:MSGQWAIT;
    while(!(Msg=nextqmsg(msgmask))) {
      if (!fillmsgq(wait)) {
	if (!handleRPC(NULL,0)) {
	  NAPTIME=(NAPTIME<MAXNAPTIME ? 2*NAPTIME : NAPTIME);
	  uSleep(NAPTIME);
	}
      }
    }
    NAPTIME=1;

    hdr=Msg->type;
    n=Msg->size;
    if (hdr_is("KEYB")) {
      if (n>=nch) n=nch-1;
      if (n>0) memcpy(buff,Msg->msg,n);
      delqmsg(Msg);
      buff[n]='\0';got=SRC_KEYB;
    }else if (hdr_is("MENU")) {
      memcpy(buff,Msg->msg,n);
      delqmsg(Msg);
      buff[n]='\0';got=SRC_MENU;
    }else if (hdr_is("SCCN")) {
      delqmsg(Msg);		/* pretty dirty stuff here... */
      graphics_down("Server closed connection...");
      got=read_null(buff,nch);
    }else if (hdr_is("USAP")) {
      if (n==cPtrSize()) {
	cross=cPtrNULL();
	memcpy(&cross,Msg->msg,n);    
	handle_usap(cross);
	delqmsg(Msg);
	got=read_null(buff,nch);
      }
    }else if (hdr_is("PATM") || hdr_is("POBJ")) {
      if (n==cPtrSize()) {
	cross=cPtrNULL();
	memcpy(&cross,Msg->msg,n);    
	Strcpy(hdr4,hdr);
	sprintf(buff,"'#%s(",hdr4);
	for (i=0;i<n;i++) sprintf(buff+strlen(buff)," %1x",cross.b[i]);
	sprintf(buff+strlen(buff)," )'");
	delqmsg(Msg);got=SRC_PICK;
      }
    }else if (hdr_is("PXYZ")) {
      if (n==6*sizeof(int)) {
	for (i=0;i<6;i++) {
	  xw[i]=MY_f(*(int *)(Msg->msg+i*sizeof(int)));
	}
	sprintf(buff,"'#PXYZ(%f %f %f %f %f %f)'",
		xw[0],xw[1],xw[2],xw[3],xw[4],xw[5]);
      }else{
	sprintf(buff,"'#PXYZ'");
      }
      delqmsg(Msg);got=SRC_PICK;
    }
  }else{
    err_dump();
    if (usereadline&&!wnpRPCisOn()) {
#ifdef USEREADLINE
      s=get_prompt(bfr);
      new_line=AReadlineFunction((s?s:""));
      if (new_line) {
	got=SRC_STDIN;
	strncpy(buff,new_line,nch);
	if (!streq(new_line,the_line)) AHistoryFunction(new_line);
	if (the_line) free(the_line);
	the_line=new_line;
      }else{
	got=SRC_NULL;
	if (the_line) free(the_line);
	the_line=(char *)NULL;
      }
#endif
    }else{
      do_prompt();
#ifdef mingw
      /* why is this necessary for mingw ??? */
      fflush(stderr); fflush(stdout);
#endif
      for (;;) {
	if (nwfgets(buff,nch,stdin)) break;
	if ((ferror(stdin)&&errno!=EAGAIN)||feof(stdin)) break;
	if (handleRPC(buff,nch-1)) break;
	uSleep(NAPTIME);
      }
      if ((ferror(stdin)&&errno!=EAGAIN)||feof(stdin)) got=SRC_NULL;
      else got=SRC_STDIN;
    }
  }
  if (got) addLF(buff,nch);
  return(got);
}

int read_file(char *buff,int nch,UCpipe *in)
{
  int got=SRC_NULL;

  *buff='\0';
  for (;;) {
    if (in->pipe==UCP_BLOCK) {
      if (UCpgets(buff,nch,in)) break;
    }else{
      if (nwfgets(buff,nch,in->fd)) break;
    }
    if ((UCperror(in)&&errno!=EAGAIN)||(strlen(buff)==0&&UCpeof(in))) break;
    if (handleRPC(buff,nch)) break;
    uSleep(NAPTIME);
  }
  if ((UCperror(in)&&errno!=EAGAIN&&errno!=0)||(strlen(buff)==0&&UCpeof(in))) got=SRC_NULL;
  else got=SRC_FILE;

  if (got) addLF(buff,nch);
  return(got);
}

int read_xmlrpc(char *buff,int nch,FILE *fd)
{
  int got=SRC_NULL;

  if (wnpRPCisOn()) {
    for (;;) {
      if (handleRPC(buff,nch-1)) break;
      uSleep(NAPTIME);
    }
    got=SRC_XMLRPC;
  }
  
  if (got) addLF(buff,nch);
  return(got);
}

/*------------------------------------------------------------------------*/

int use_readline(int s)
{
  int r;

  r=usereadline;
#ifdef USEREADLINE
  StifleHistory(4096);
  usereadline=s;
#endif
  return(r);
}

#ifndef AReadlineFunction
char *AReadlineFunction(char *prompt)
{
  char buff[32000];
  fprintf(stderr,"%s",prompt);
  if (!fgets(buff,32000,stdin)) buff[0]='\0';
  return(strdup(buff));
}
#endif

#ifndef AHistoryFunction
void AHistoryFunction(char *line)
{

}
#endif

#ifndef StifleHistory
void StifleHistory(int n)
{

}
#endif

static char *nwfgets(char *s, int size, FILE *stream)
{
  char *r=NULL,*q;
  int l;

  /*  No-wait fgets function. Don't wait for the first byte, once the first 
      byte is in, wait for the rest of the line, up to the first newline.
      This is necessary if stream is a pipe. */

  wait_read(fileno(stream),0);

  if ((r=fgets(s,size,stream))) {
    l=strlen(s); q=s+l-1;
    while (*q!='\n'&&size>1) {
      size-=l;
      if (!fgets(q+1,size,stream)) q[1]='\0';
      l=strlen(q+1); q=q+l;
      if (feof(stream)) break;
    }
  }
  if (ferror(stream)&&errno==EAGAIN) clearerr(stream); /* ignore this 'error'! */
  wait_read(fileno(stream),1);

  return(r);
}

static void wait_read(int ch,int w)
{
#ifndef mingw
  int flags;

  flags=fcntl(ch,F_GETFL,0);
  if (w) fcntl(ch,F_SETFL,~FNDELAY & flags);
  else fcntl(ch,F_SETFL,FNDELAY | flags);
#endif
}

static void addLF(char *buff,int size)
{
  int l;
  l=strlen(buff);
  if (l==0&&size>1) {
    buff[0]='\n';
    buff[1]='\0';
  }else{
    if (l<size-1) {
      if (buff[l-1]!='\n') {
	buff[l]='\n';
	buff[l+1]='\0';
      }
    }
  }
}
