#include <stdio.h>
#include "object.h"
#include "swdef.h"
#include "fileutil.h"
#include "token.h"
#include "strutil.h"
#include "gcalc.h"
#include "xtal.h"
#include "errhand.h"
#include "mystdlib.h"
#include "gcalc.h"
#include "streamio.h"
#include <math.h>

#define TRUE 1
#define FALSE 0
#define NO 0

#define MAXTOKENSIZE 256
#define MAXATOMKIND 50

#define Fgets(x,f) {if (!UCpgets(x,sizeof(x),f)) *x='\0'; \
                    else{ \
                      while (strrchr(x,'\n')) *strrchr(x,'\n')='\0'; \
		      for (k_=strlen(x)+1;k_<sizeof(x);k_++) x[k_]=' '; \
                      x[sizeof(x)-1]='\0';} \
                    }

#define Return(x,y) {err_set(x,"read_par: %s",y);if (own_p)UCpclose(fpipe);return(x);}

#define api2 78.95683

extern Node *LibRoot;

static int perm_bf[6]={0,1,2,5,4,3};  /* bring RSA to Shelx order */

int read_par(  UCpipe *fpipe,char *file,Node *node)
{
  char rec[MAXTOKENSIZE];
  char chs[3];
  Atomdef *atp[MAXATOMKIND],*default_type,*def=(Atomdef *)NULL,*smtp;
  Symbol *S;
  char tf_type='\0',fld[40];
  Xatom *xa=(Xatom *)NULL;
  Xcn  *xcn;
  Node *asym;
  Node *ResLib;
  Symop sym_n;
  Atom *a;
  float ra[3],rl[3],fc[6],rcov,rvdw;
  int dupl,old_err,k_,i,j,status,rtype,ob,ob5,own_p;

  own_p=(fpipe ? FALSE : TRUE);
  if (own_p) fpipe=UCpopen(extfil(rec,file,".par",NO));
  if (!fpipe) {
    err_set(1,"read_par: could not open %s",extfil(rec,file,".par",NO));
    return(1);
  }

  if (!is_mol(node)) Return(2,"root is not a node");
  
  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    Return(-1,"Failed to open default library");
  }

  default_type=find_atomdef_name(ResLib,"unk");    
  if (!default_type) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  xcn=first_xcn(node);
  if (!xcn) {
    xcn=(Xcn *)add_node(node,(char *)NULL,OBJ_XCN);
  }
  if (!xcn) Return(3,"root has no XCN.");

  asym=add_node(node,"A",OBJ_MOL);
  if (!asym) Return(4,"memory allocation failure");

  status=0;
  a=(Atom *)NULL;
  for (i=0;i<MAXATOMKIND;i++) atp[i]=(Atomdef *)NULL;
  
  ob5=0;
  Fgets(rec,fpipe);
  while(*rec) {
    rtype=str2i(strfld(fld,rec,1,3));
    switch (rtype) {

    case 1:      /* title  */
      if (mvld(xcn->title)) free(xcn->title);
      xcn->title=Strdup(rec+4);
      break;

    case 2:      /* unit cell geometry */
      for (i=0;i<3;i++) xcn->cl[i]=str2f(strfld(fld,rec+3,i*9+1,(i+1)*9));
      for (i=0;i<3;i++) xcn->ca[i]=str2f(strfld(fld,rec+30,i*9+1,(i+1)*9));
      xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca); 
      old_err=err_seqnr();
      xcn->atf=inv_4(xcn->fta);
      if (old_err!=err_seqnr()) {
        Return(150,"singular cell to angstrom matrix.");
      }
      break;

    case 3:      /* esd cell */
      for (i=0;i<3;i++) xcn->esd_cl[i]=str2f(strfld(fld,rec+3,i*9+1,(i+1)*9));
      for (i=0;i<3;i++) xcn->esd_ca[i]=str2f(strfld(fld,rec+30,i*9+1,(i+1)*9));
      break;

    case 4:      /* Formula */
      ob=xcn->forml;
      xcn->forml+=str2i(strfld(fld,rec,4,8));
      for (i=ob;i<xcn->forml;i++) {
	Strcpy(xcn->forms[i],strfld(fld,rec+10,(i-ob)*8+1,(i-ob)*8+3));
	if (xcn->forms[i][1]>='A'&&xcn->forms[i][1]<='Z') xcn->forms[i][1]+=32;
	xcn->formn[i]=str2i(strfld(fld,rec+10,(i-ob)*8+4,(i-ob+1)*8));
	def=find_atomdef_name(ResLib,xcn->forms[i]);
	if (!mvld(def)) {
	  smtp=guess_atom_type(xcn->forms[i]);
	  if (mvld(smtp)) def=(Atomdef *)copy_node((Node *)smtp);
	  else def=(Atomdef *)copy_node((Node *)default_type);
	  if (mvld(def)) {
	    set_node_name((Node *)def,xcn->forms[i]);
	    def->up=ResLib;
	    link_node((Node *)def);
	    S=find_symbol_name(ResLib,xcn->forms[i]);
	    if (mvld(S)) {
	      def->rvdw=S->rvdw; def->rcov=S->rcov;
	      def->mass=S->mass; def->color=S->color;
	    }
	    Strcpy(def->chs,xcn->forms[i]);
	  }
	}
	atp[i]=def;
      }
      break;

    case 5:      /* covalent and vdW radii */
      if (xcn->forml<=0) {
	err_set(509,"read_par: record type 5 out of order.");
      }else{
	ob=(ob5+5 > xcn->forml ?  xcn->forml : ob5+5);
	for (i=ob5;i<ob;i++) {
	  if (atp[i]) {
	    rcov=str2f(strfld(fld,rec+3,(i-ob5)*12+1,(i-ob5)*12+6));
	    rvdw=str2f(strfld(fld,rec+3,(i-ob5)*12+7,(i-ob5)*12+12));
	    atp[i]->rcov=rcov; atp[i]->rvdw=rvdw;
	  }
	}
	ob5=ob;
      }
      break;

    case 6:      /* space group definition */
      if (mvld(xcn->spgr)) free(xcn->spgr);
      xcn->spgr=Strdup(strfld(fld,rec,5,25));
      xcn->centric=streq(strfld(fld,rec,33,34),"-1");
      if (rec[38]=='P') xcn->lattice=1;
      else if (rec[38]=='I') xcn->lattice=2;
      else if (rec[38]=='R') xcn->lattice=3;
      else if (rec[38]=='F') xcn->lattice=4;
      else if (rec[38]=='A') xcn->lattice=5;
      else if (rec[38]=='B') xcn->lattice=6;
      else if (rec[38]=='C') xcn->lattice=7;
      else err_set(101,"read_par: unknown lattice type: %c");
      break;

    case 7:      /* symmetry operator */
      sym_n=zero_4(); sym_n.r[3].c[3]=1.0;
      for (i=0;i<3;i++) {
	sym_n.r[i].c[3]=str2f(strfld(fld,rec+5,20*i+10,20*i+18));
	for (j=0;j<3;j++) {
	  sym_n.r[i].c[j]=str2f(strfld(fld,rec+5,20*i+j*3+1,20*i+(j+1)*3));
	}
      }
      for (dupl=FALSE,i=0; !dupl&&i<xcn->nsym; i++) {
        dupl=symopequiv(sym_n,xcn->sym[i]);
      }
      if (!dupl) {
	if (xcn->nsym>=MX_XCN_SYMOPS) {
	  Return(601,"too many symmetry operations.");
	}else{
	  xcn->sym[xcn->nsym++]=sym_n;
	}
      }
      break;
      
    case 8:      /* first atom record */
      a=(Atom *)add_node(asym,strfld(fld,rec,6,14),OBJ_ATOM);
      if (!mvld(a)) Return(501,"add atom failure.");
      if (a->mxyz<1) {
	if (!reall_atom_x(a,1)) {
	  Return(503,"xyz reallocation failure.");
	}
      }
      for (i=0;i<3;i++) a->x[i]=str2f(strfld(fld,rec+13,i*9+1,i*9+9));
      a->nxyz=1;
      xa=(Xatom *)add_node((Node *)a,(char *)NULL,OBJ_XATOM);
      if (!xa) {
	Return(504,"Xatom creation failure.");
      }
      a->xtal=1;
      a->x[ATM_SOF]=str2f(strfld(fld,rec,50,58));
      xa->bfac[0]=str2f(strfld(fld,rec,41,49));
      xa->iso=TRUE;
      tf_type=rec[59];

      Strcpy(chs,a->name);
      if (chs[1]>='A'&&chs[1]<='Z') chs[1]+=32;
      if (chs[1]<'a'||chs[1]>'z') chs[1]='\0';
      for (i=0;i<xcn->forml;i++) {
	if (mvld(atp[i])) {
	  if (streq(atp[i]->name,chs)) set_atom_type(a,atp[i]);
	}
      }
      if (!mvld(a->atp)) set_atom_type(a,def);
      break;

    case 9:     /* sigmas of type 8 parameters */
      if (!streq(a->name,strfld(fld,rec,6,14))) {
	Return(901,"record type 9 out of order.");
      }
      for (i=0;i<3;i++) xa->esd_x[i]=str2f(strfld(fld,rec+13,i*9+1,i*9+9));
      xa->esd_sof=str2f(strfld(fld,rec,50,58));
      xa->esd_bfac[0]=str2f(strfld(fld,rec,41,49));
      break;

    case 10:    /* parameter fix */
      if (!streq(a->name,strfld(fld,rec,6,14))) {
	Return(1001,"record type 10 out of order.");
      }
      for (i=0;i<3;i++) {
	if (str2f(strfld(fld,rec+13,i*9+1,i*9+9))<0.5) xa->pfix[i]=10.0;
      }
      if (str2f(strfld(fld,rec,50,58))<0.5) xa->pfix[3]=10.0;
      if (str2f(strfld(fld,rec,41,49))<0.5) xa->pfix[4]=10.0;
      break;

    case 11:    /* aniso temperature factor */
      if (!streq(a->name,strfld(fld,rec,6,14))) {
	Return(1101,"record type 11 out of order.");
      }
      if (tf_type!='2'&&tf_type!='5') break;
      for (i=0;i<6;i++) {
	xa->bfac[perm_bf[i]]=str2f(strfld(fld,rec+13,i*9+1,i*9+9));
	xa->esd_bfac[perm_bf[i]]=0.0;
      }
      xa->iso=FALSE;
      break;

    case 12:    /* sigmas of aniso temperature factor */
      if (!streq(a->name,strfld(fld,rec,6,14))) {
	Return(1201,"record type 12 out of order.");
      }
      if (tf_type!='2'&&tf_type!='5') break;
      if (!xa->iso) {
	for (i=0;i<6;i++) {
	  xa->esd_bfac[perm_bf[i]]=str2f(strfld(fld,rec+13,i*9+1,i*9+9));
	}
      }
      break;

    case 21:     /* wave length */
      xcn->lambda=str2f(strfld(fld,rec,4,10));
      break;

    default:
      break;

    }

    Fgets(rec,fpipe);
  }

  if (own_p) UCpclose(fpipe);

/* move center of box to origin */

  for (i=0;i<3;i++) {
    xcn->fta.r[i].c[3]=0;
    for (j=0;j<3;j++) xcn->fta.r[i].c[3]-=xcn->fta.r[i].c[j]*0.5;
  } 

  old_err=err_seqnr();
  xcn->atf=inv_4(xcn->fta);
  if (old_err!=err_seqnr()) {
    err_set(1001,"singular cell to angstrom matrix.");
    return(1001);
  }

/* convert coordinates to cartesian */

  forAtoms(a,asym) v32arr(a->x,Axyz(xcn,vec3(a->x)));

/* convert temperature factors RSA-->SHELX */

  recell(xcn->cl,xcn->ca,rl,ra);
  fc[0]=1.0/(rl[0]*rl[0]*api2*0.25);
  fc[1]=1.0/(rl[1]*rl[1]*api2*0.25);
  fc[2]=1.0/(rl[2]*rl[2]*api2*0.25);
  fc[3]=1.0/(rl[1]*rl[0]*api2*0.25);
  fc[4]=1.0/(rl[0]*rl[2]*api2*0.25);
  fc[5]=1.0/(rl[0]*rl[1]*api2*0.25);
    
  ForAtoms(a,asym) {
    xa=first_xatom(a);
    if (mvld(xa)) {
      if (xa->iso) {
	xa->bfac[0]/=api2;
	xa->esd_bfac[0]/=api2;
      }else{
	for (i=0;i<6;i++) {
	  xa->bfac[i]*=fc[i];
	  xa->esd_bfac[i]*=fc[i];
	}
      }
      a->x[ATM_BISO]=biso_equiv(xa);
    }
  }

/* expand symmetry operations */

  i=symopexp(xcn);
  if (i!=xcn->nsymx) {
    err_set(1002,"read_shelx: more than %d expanded symops",MX_XCN_SYMOPS);
    xcn->nsymx=xcn->nsym;
  }

/* convert symmetry operators to angstrom/cartesian */

  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);

  return(status);
}
