#include "object.h"
#include "swdef.h"
#include "streamio.h"
#include "token.h"
#include "strutil.h"
#include "gcalc.h"
#include "xtal.h"
#include "errhand.h"
#include "mystdlib.h"
#include "gcalc.h"
#include <math.h>

#define TRUE 1
#define FALSE 0
#define NO 0

#define ATM_COORD   1
#define ATM_TOPOL   2
#define ATM_REC3    3
#define ATM_ESD     4
#define ATM_PARFIX  5
#define ATM_REC6    6

#define ATOM 1
#define PEAK 2

#define MAXTOKENSIZE 256
#define MAXADFIXES 1000

#define Return(x,y) \
   {err_set(x,"read_shelx: %s",y);\
    for (i=0;i<nadfrec;i++) \
      if (mvld(adfrec[i].to)) free(adfrec[i].to);\
    return(x);\
   }

typedef struct Adfrecord {
  long class;
  Xatom *from;
  char *to;
  float dist;
  float esd;
} Adfrecord;

extern Node *LibRoot;

int read_shelx(UCpipe *pipe,char *file,Node *node)
{
  char tok[MAXTOKENSIZE],ffmnem[MAXTOKENSIZE];
  Parse *p;
  Stream *in;
  Atom *a,*b;
  Atomset *s;
  Adfrecord adfrec[MAXADFIXES];
  ADfix *adf;
  Atomdef *default_type,*tab;
  Bond *bnd;
  Xatom *xa=(Xatom *)NULL;
  Xcn  *xcn;
  Node *asym,*ResLib;
  char *c;
  Symop sym_n;
  hisrec *h,*nh;
  int dupl,old_err,i,j,n,type,status,has_bt,atype=0,nadfrec;
  int found,have_cell,have_X2A;
  mat_4 X2A;
  vec_3 ax0={{0,0,0}},axa={{1,0,0}},axb={{0,1,0}},axc={{0,0,1}};
  float a0[3],aa[3],ab[3],ac[3];

  nadfrec=0; have_cell=FALSE; have_X2A=FALSE;
  for (i=0;i<nadfrec;i++) adfrec[i].to=(char *)NULL;

  if (!is_mol(node)) Return(2,"root is not a node");
  
  xcn=first_xcn(node);
  if (!xcn) {
    xcn=(Xcn *)add_node(node,"Xcn",OBJ_XCN);
  }
  if (!xcn) Return(3,"root has no XCN.");

  asym=add_node(node,"A",OBJ_MOL);
  if (!asym) Return(4,"memory allocation failure");

  if (pipe) {
    in=open_stream((char *)NULL,MAXTOKENSIZE);
    set_stream_pipe(in,pipe);
  }else{
    in=open_stream(extfil(tok,file,".xcn",NO),MAXTOKENSIZE);
  }
  if (!in) Return(1,"could not open input file");
  set_stream(in,"eol=no,redir=no,comment=no");

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    close_stream(in);
    return(-1);
  }
  default_type=find_atomdef_name(ResLib,"unk");
  if (!mvld(default_type)) {
    default_type=(Atomdef *)add_node(ResLib,"unk",OBJ_ATOMDEF);
  }

  status=0;
  a=(Atom *)NULL;


  in->p->tok=tok; in->p->mts=MAXTOKENSIZE;

  while(token_stream(in,"")) {

/*  replace last newline (if any) by null */

    p=strm_parse(in);
    c=strrchr(p->buf,'\n');
    if (c) *c='\0';

/*  TITLE record */

    if (strcasecmp("TITLE",tok)==0) {
      p=strm_parse(in);
      if (mvld(xcn->title)) free(xcn->title);
      xcn->title=Strdup(p->buf+p->pos);
      if (!xcn->title) {
        close_stream(in);
        Return(101,"bad title record");
      }

/*  CELL record */

    }else if (strcasecmp("CELL",tok)==0) {
      p=strm_parse(in);
      xcn->lambda=str2f(token(p));
      for (i=0;i<3;i++) xcn->cl[i]=str2f(token(p));
      for (i=0;i<3;i++) xcn->ca[i]=str2f(token(p));
      xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca); 
      old_err=err_seqnr();
      xcn->atf=inv_4(xcn->fta);
      if (old_err!=err_seqnr()) {
	close_stream(in);
        Return(150,"singular cell to angstrom matrix.");
      }else have_cell=TRUE;;

/*  CELL_ESD record */

    }else if (strcasecmp("CELL_ESD",tok)==0) {
      p=strm_parse(in);
      for (i=0;i<3;i++) xcn->esd_cl[i]=str2f(token(p));
      for (i=0;i<3;i++) xcn->esd_ca[i]=str2f(token(p));

/*  FORMULA record */

    }else if (strcasecmp("FORMULA",tok)==0) {
      p=strm_parse(in);
      xcn->forml=0;
      p=strm_parse(in);
      while (get_token(p)) {
        if (p->term=='\n' || p->term=='\0') break;   /* ?? needed ?? */
        if (p->ltok<1 || p->ltok>2) {
          close_stream(in);
          Return(201,"bad formula record.");
        }else{
          if (xcn->forml>=MX_XCN_FORMUL) {
            close_stream(in);
            Return(202,"too many items in formula");
          }else{
            Strcpy(xcn->forms[xcn->forml],p->tok);
            xcn->formn[xcn->forml]=1;
          }
        }
        if (strchr("0123456789",p->buf[p->pos])) {
          xcn->formn[xcn->forml]=str2i(token(p));
          if (xcn->formn[xcn->forml]<1) {
            close_stream(in);
            Return(203,"element count in formula not a number");
          }
        }
        xcn->forml++;
      }
        
/*  SPGR-record */

    }else if (strcasecmp("SPGR",tok)==0) {
      p=strm_parse(in);
      xcn->lattice=str2i(token(p));
      xcn->centric=(xcn->lattice > 0 ? TRUE : FALSE);
      if (!xcn->centric) xcn->lattice = -xcn->lattice;
      if (xcn->lattice<1 || xcn->lattice>7) {
        close_stream(in);
        Return(301,"bad lattice parameter in sgpr record.");
      }
      if (mvld(xcn->spgr)) free(xcn->spgr);
      xcn->spgr=Strdup(p->buf+p->pos);
      if (!xcn->spgr) {
        close_stream(in);
        Return(302,"memory allocation failure.");
      }

/*  ASYMM-record */

    }else if (strcasecmp("ASYMM",tok)==0) {
      p=strm_parse(in);
      for (i=0;i<6;i++) xcn->asym[i]=str2f(token(p));
    
/*  SCALE-record */

    }else if (strcasecmp("SCALE",tok)==0) {
      xcn->scale=Strdup(p->buf+p->pos);
      if (mvld(xcn->scale)) free(xcn->scale);
      xcn->scale=Strdup(p->buf+p->pos);
      if (!xcn->scale) {
        close_stream(in);
        Return(303,"memory allocation failure.");
      }

/*  Wit!P additional info */

    }else if (streq("wit!p",tok)) {
      while (get_token(p)) {
	if (streq(tok,"sets")) {    /* definition of atomsets */
	  while (get_token(p)) {
	    forAtomsets(s,node) if (streq(tok,s->name)) break;
	    if (!mvld(s)) s=add_atomset((Mol *)node,-1,tok);
	    if (!is_atomset(s)) err_set(1,"failed to add set %s",tok);
	  }
	}else if (streq("XtoA",tok)) {  /* definition of transformation from cryst. to A-cart */
	  if (have_cell) {
	    for (i=0;i<3;i++) for (j=0;j<4;j++) X2A.r[i].c[j]=str2f(token(p));
	    for (j=0;j<3;j++) X2A.r[3].c[j]=0.0;
	    X2A.r[3].c[3]=1.0;
	    v32arr(a0,m4xv3(X2A,ax0)); v32arr(aa,m4xv3(X2A,axa)); 
	    v32arr(ab,m4xv3(X2A,axb)); v32arr(ac,m4xv3(X2A,axc));
	    if (fabs(cdist(a0,aa)-xcn->cl[0]) < 0.005      &&
		fabs(cdist(a0,ab)-xcn->cl[1]) <  0.005     &&
		fabs(cdist(a0,ac)-xcn->cl[2]) <  0.005     &&
		fabs(cangled(ab,a0,ac)-xcn->ca[0]) < 0.05  &&  
		fabs(cangled(ac,a0,aa)-xcn->ca[1]) < 0.05  &&
		fabs(cangled(aa,a0,ab)-xcn->ca[2]) < 0.05) {
	      have_X2A=TRUE;
	      xcn->fta=X2A;
	      old_err=err_seqnr();
	      xcn->atf=inv_4(xcn->fta);
	      if (old_err!=err_seqnr()) {
		close_stream(in);
		Return(150,"singular cell to angstrom matrix.");
	      }
	    }else err_set(1,"rdshlx: X2A not consistent with 'cell'");
	  }else err_set(1,"rdshlx: wit!p XtoA before 'cell' record, ignored.");
	}
      }

/*  HISTORY-record */

    }else if (strcasecmp("HISTORY",tok)==0) {
      p=strm_parse(in);
      for (h=xcn->history;h;h=h->nxt) if (!h->nxt) break;
      nh=make_history();
      if (nh) {
        nh->t=Strdup(p->buf+p->pos);
        if (h) h->nxt=nh;
        else xcn->history=nh;
      }else{
        close_stream(in);
        Return(401,"memory allocation failure.");
      }

/*  ATOM-record  */

    }else if (strcasecmp("ATOM",tok)==0 || strcasecmp("PEAK",tok)==0) {
      if (strcasecmp("ATOM",tok)==0) atype=ATOM;
      else if (strcasecmp("PEAK",tok)==0) atype=PEAK;
      p=strm_parse(in);
      type=str2i(token(p));
      switch (type) {
      case ATM_COORD:
        a=(Atom *)add_node(asym,(char *)NULL,OBJ_ATOM);
        if (!a) {
          close_stream(in);
          Return(501,"add atom failure.");
        }
        set_node_name((Node *)a,token(p));
        if (!a->name) {
          close_stream(in);
          Return(502,"set atom name failure.");
        }
	set_atom_type(a,guess_atom_type(a->name));
        if (a->mxyz<1) {
          if (!reall_atom_x(a,1)) {
            close_stream(in);
            Return(503,"xyz reallocation failure.");
          }
        }
        for (i=0;i<3;i++) a->x[i]=str2f(token(p));
        a->nxyz=1;
        xa=(Xatom *)add_node((Node *)a,(char *)NULL,OBJ_XATOM);
        if (!xa) {
          close_stream(in);
          Return(504,"Xatom creation failure.");
        }
        a->xtal=1;
        a->x[ATM_SOF]=str2f(token(p));
        if (get_token(p)) {
          if (atype==PEAK) {
            xa->peak_height=str2f(tok);
          }else{
            xa->bfac[0]=str2f(tok);
          }
        }
        for (i=1;i<6;i++) {
          if (!get_token(p)) break;
          xa->bfac[i]=str2f(tok);
        }
        xa->iso=(i!=6);
        if (i!=1&&i!=6) {
          close_stream(in);
          Return(505,"bad bfactor");
        }
        xa->peak=(atype==PEAK);
        if (xa->peak) xa->iso=TRUE;
	a->x[ATM_BISO]=biso_equiv(xa);
        break;

      case ATM_TOPOL:
        if (!a || !xa) {
          close_stream(in);
          Return(506,"topology before coord");
        }
        Strcpy(ffmnem,token(p));
	tab=find_atomdef_name(ResLib,ffmnem);
	if (!mvld(tab)) {
	  tab=(Atomdef *)copy_node((Node *)default_type);
	  if (mvld(tab)) {
	    tab->up=ResLib;
	    link_node((Node *)tab);
	    set_node_name((Node *)tab,ffmnem);
	  }
	}
	set_atom_type(a,tab);
        p->sep="[]";
        while(get_token(p)) {
          forAtoms(b,asym) if (a!=b&&streq(tok,b->name)) break;
          has_bt=FALSE;
          if (p->term=='[') {
            if (!get_token(p)||p->term!=']') {
              close_stream(in);
              Return(507,"bad bondorder specification");
            }
            has_bt=TRUE;
          }
          if (b) {
            bnd=add_bond(a,b);
            if (!bnd) {
              close_stream(in);
              Return(508,"add bond failure");
            }
	    set_bt(bnd,BT_SINGLE);
            if (has_bt) {
	      if (BT_number(tok)<0) {
		err_set(1,"read_shelx: bond type %s not found.\n",tok);
		set_bt(bnd,BT_UNKNOWN);
	      }else{
		set_bt(bnd,BT_number(tok));
	      }
	    }
          }
        } 
        break;

      case ATM_ESD:
        if (!a || !xa) {
          close_stream(in);
          Return(509,"atom esd before coord.");
        }
        for (i=0;i<3;i++) xa->esd_x[i]=str2f(token(p));
        xa->esd_sof=str2f(token(p));
        xa->esd_bfac[0]=str2f(token(p));
        if (!xa->iso) {
          for (i=1;i<6;i++) xa->esd_bfac[i]=str2f(token(p));
        }
        break;

      case ATM_REC3:          /* additional info used by Wit!P */
        if (!a || !xa) {
          close_stream(in);
          Return(510,"atom record type 3 before coord.");
        }
	while (get_token(p)) {
	  if (streq(tok,"chrg")) {
	    if (get_token(p)) {
	      if (streq(tok,"?")) a->chrg='?';
	      else a->chrg=str2i(tok);
	    }
	  }else if (streq(tok,"q")) {
	    a->q=str2f(token(p));
	  }else if (streq(tok,"sets")) {
	    while (get_token(p)) {  /* rest of the record are set names */
	      forAtomsets(s,node) if (streq(tok,s->name)) break;
	      if (is_atomset(s)) add_to_atomset(a,s);
	    }
	  }else{
	    err_set(1,"bad atom 3 record : %s",p->buf);
	    break;
	  }
	}
        break;

      case ATM_PARFIX:
        if (!a || !xa) {
          close_stream(in);
          Return(510,"atom record type 5 before coord.");
        }
	n=(xa->iso ? 5 : 10);
	for (i=0;i<n;i++) xa->pfix[i]=str2f(token(p));
        break;

      case ATM_REC6:
        if (!a || !xa) {
          close_stream(in);
          Return(510,"atom record type 6 before coord.");
        }
	p->sep="";
	while (get_token(p)) {
	  if (strcasecmp("AFIX_ROOT",tok)==0) {
	    xa->afix_root=str2i(token(p));
	    if (xa->afix_root<=0) {
	      err_set(551,"read_shelx: atom %s, bad afix_root",a->name);
	    }
	  }else if (strcasecmp("AFIX_TO",tok)==0) {
	    if (nadfrec<MAXADFIXES) {
	      adfrec[nadfrec].class=ADFCLASS_AFIX;
	      adfrec[nadfrec].to=Strdup(token(p));
	      adfrec[nadfrec].from=xa;
	      if (!mvld(adfrec[nadfrec].to)) {
		err_set(552,"read_shelx: atom %s, bad afix_to",a->name);
	      }else{
		nadfrec++;
	      }
	    }else{
	      err_set(561,"read_shelx: afix/dfix memory overflow");
	    }
	  }else if (strcasecmp("BLOCK",tok)==0) {
	    xa->block=str2i(token(p));
	    if (xa->block<=0) {
	      err_set(553,"read_shelx: atom %s, bad block",a->name);
	    }
	  }else if (strcasecmp("DFIX_TO",tok)==0) {
	    if (nadfrec<MAXADFIXES) {
	      adfrec[nadfrec].class=ADFCLASS_DFIX;
	      adfrec[nadfrec].from=xa;
	      adfrec[nadfrec].to=Strdup(token(p));
	      adfrec[nadfrec].dist=str2f(token(p));
	      adfrec[nadfrec].esd=str2f(token(p));
	      if (!mvld(adfrec[nadfrec].to)) {
		err_set(552,"read_shelx: atom %s, bad afix_to",a->name);
	      }else{
		nadfrec++;
	      }
	    }else{
	      err_set(561,"read_shelx: afix/dfix memory overflow");
	    }
	  }else{
	    err_set(555,"read_shelx: atom %s, bad token(%s)",a->name,tok);
	  }
	}
        break;
      }

    }else if (strcasecmp("SYMM",tok)==0) {
      sym_n=str2sym(p->buf+p->pos);
      for (dupl=FALSE,i=0; !dupl&&i<xcn->nsym; i++) {
        dupl=symopequiv(sym_n,xcn->sym[i]);
      }
      if (dupl) {
	err_set(600,"read_shelx: duplicate symm op(%s)",p->buf+p->pos);
      }else if (xcn->nsym>=MX_XCN_SYMOPS) {
        close_stream(in);
        Return(601,"too many symmetry operations.");
      }else{
	xcn->sym[xcn->nsym++]=sym_n;
      }

/*  unrecognized record */

    }else{
      close_stream(in);
      Return(9999,"unrecognized record type");
    }

/* force new read */
  
    flush_record(in);
  }
  close_stream(in);

/* convert AFIX/DFIX records ... */

  for (i=0;i<nadfrec;i++) {
    found=FALSE;
    forAtoms(a,asym) {
      if (strcasecmp(a->name,adfrec[i].to)==0) {
	found=TRUE;
	xa=first_xatom(a);
	if (mvld(xa)) {
	  adf=add_adfix(adfrec[i].from,xa);
	  if (mvld(adf)) {
	    adf->class=adfrec[i].class;
	    switch (adfrec[i].class) {
	    case ADFCLASS_DFIX:
	      adf->dist=adfrec[i].dist;
	      adf->esd=adfrec[i].esd;
	      break;
	    case ADFCLASS_AFIX:
	      break;
	    default:
	      break;
	    }
	  }else{
	    err_set(1000,"memalloc failed...");
	  }
	}else{
	  err_set(1000,"afix/dfix root not found: %s --> %s",
		  adfrec[i].from->up->name,a->name);
	}
	break;
      }
    }
    if (!found) {
      err_set(1000,"afix/dfix root not found: %s --> %s",
	      adfrec[i].from->up->name,adfrec[i].to);
    }
  }
		  
  for (i=0;i<nadfrec;i++) if (mvld(adfrec[i].to)) free(adfrec[i].to);
  nadfrec=0;

/* move center of box to origin */

  if (!have_X2A) {
    for (i=0;i<3;i++) {
      xcn->fta.r[i].c[3]=0;
      for (j=0;j<3;j++) xcn->fta.r[i].c[3]-=xcn->fta.r[i].c[j]*0.5;
    }
  }

  old_err=err_seqnr();
  xcn->atf=inv_4(xcn->fta);
  if (old_err!=err_seqnr()) {
    Return(1001,"singular cell to angstrom matrix.");
  }

/* convert coordinates to cartesian */

  forAtoms(a,asym) v32arr(a->x,Axyz(xcn,vec3(a->x)));

/* restore aromaticity flags lost due to fugu bonds */

  forAtoms(a,asym) if (mvld(a->atp)) {
    if (a->atp->arom) a->arom=TRUE;
  }

/* expand symmetry operations */

  i=symopexp(xcn);
  if (i!=xcn->nsymx) {
    err_set(1002,"read_shelx: more than %d expanded symops",MX_XCN_SYMOPS);
    xcn->nsymx=xcn->nsym;
  }

/* convert symmetry operators to angstrom/cartesian */

  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);

  return(status);
}
