#include <stdio.h>
#include <math.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "xtal.h"
#include <ctype.h>

int read_spgrp(Xcn *xcn,char *name)
{
  char *spgdir,buffer[512],tkn[256],*O,*P;
  char spgnam[256];
  UCpipe *Spg;
  Parse p;
  int r,i,j,ers,nsym,lattice=0,centric,lp;
  mat_4 sv_fta,sv_atf,sym[MX_XCN_SYMOPS],m;
  float s,det,min_det,min_len,min_err,max_det,max_len;

  if (!is_xcn(xcn)) {
    err_set(1,"bad argument(xcn) to read_spgrp.");
    return(1);
  }
  if (!name) {
    err_set(1,"bad argument(name) to read_spgrp.");
    return(1);
  }
  
  /* convert spacegroup name to canonical form */

  for (O=spgnam,P=name;*P;P++) {
    if (!isspace(*P)) {
      *O++=(*P>='A'&&*P<='Z'?*P+32:(*P=='/'?'_':*P));
    }
  }
  *O='\0'; *spgnam-=(*spgnam>='a'&&*spgnam<='z'?32:0);
  if (strlen(spgnam)==0) sprintf(spgnam,"%s","P1");

  /* edit pathname to spacegroup definiton file */

  spgdir=getenv("WITNOTP_SPACEGROUPS");
  if (spgdir) {
    if (strlen(spgdir)>0 && spgdir[strlen(spgdir)-1]=='/') {
      sprintf(buffer,"%s%s",spgdir,spgnam);
      lp=strlen(spgdir);
    }else{
      sprintf(buffer,"%s/%s",spgdir,spgnam);
      lp=strlen(spgdir)+1;
    }
  }else{
    sprintf(buffer,"%s",spgnam);
    lp=0;
  }  

  Spg = UCpopen(buffer);
  if (!Spg) {
    err_pop(1);
    for (P=buffer+lp;*P;P++) if (*P=='_') *P=':';
    Spg = UCpopen(buffer);
  }
  if (!Spg) {
    err_pop(1);
    for (P=buffer+lp;*P;P++) if (*P==':') *P='_';
    Spg = UCpopen(buffer);
  }
  if (!Spg) {
    err_set(1,"Could not open spacegroup file %s for read.",buffer);
    return(1);
  }

  /* read definition of spacegroup, with some checks... */

  p.sep=""; p.tok=tkn; p.mts=250; p.buf=buffer; r=1;
  buffer[0]='\0';

  /* ... use cell geometry for testing... */

  sv_fta=xcn->fta; sv_atf=xcn->atf;
  xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca);
  xcn->atf=inv_4(xcn->fta);
  min_det=1e30; min_len=1e30; min_err=1e30;

  while (r) {
    if (!strneq(buffer,">>> ",4)) if (!fgets(buffer,510,Spg->fd)) buffer[0]='\0';
    if (feof(Spg->fd)) break;
    if (!strneq(buffer,">>> ",4)) continue;
    for (p.pos=0,i=0;i<4;i++) get_token(&p);
    switch(*tkn) {
    case 'A':
      lattice=A_latt;
      break;
    case 'B':
      lattice=B_latt;
      break;
    case 'C':
      lattice=C_latt;
      break;
    case 'I':
      lattice=I_latt;
      break;
    case 'F':
      lattice=F_latt;
      break;
    case 'P':
      lattice=P_latt;
      break;
    case 'H':
      lattice=H_latt;
      break;
    case 'R':
      lattice=R_latt;
      break;
    }
    get_token(&p); centric=(*tkn=='c');
    nsym=0; max_len=0; max_det=0;
    for (;fgets(buffer,510,Spg->fd);) {
      if (feof(Spg->fd)) break;
      if (strneq(buffer,">>> ",4)) break;
      ers=err_since();
      sym[nsym]=str2sym(buffer);
      m=symA(xcn,sym[nsym]);
      if (ers==err_since()) {
	for (i=0;i<3;i++) {
	  for (s=0,j=0;j<3;j++) s+=m.r[j].c[i]*m.r[j].c[i];
	  if (fabs(sqrt(s)-1.0)>max_len) max_len=fabs(sqrt(s)-1.0);
	}
	det=m.r[0].c[0]*m.r[1].c[1]*m.r[2].c[2]
	  +m.r[0].c[1]*m.r[1].c[2]*m.r[2].c[0]
	    +m.r[1].c[0]*m.r[2].c[1]*m.r[0].c[2]
	      -m.r[2].c[0]*m.r[1].c[1]*m.r[0].c[2]
		-m.r[1].c[0]*m.r[0].c[1]*m.r[2].c[2]
		  -m.r[0].c[0]*m.r[2].c[1]*m.r[1].c[2];
	det=fabs(det);
	if (fabs(det-1.0)>max_det) max_det=fabs(det-1.0);
	nsym++;
      }
    }
    if (max_det<min_err && max_len<min_err) {
      for (i=0;i<nsym;i++) xcn->sym[i]=sym[i];
      xcn->centric=centric; xcn->lattice=lattice;
      min_det=max_det; min_len=max_len;
      min_err=( min_det > min_len ? min_det : min_len);
      xcn->nsym=nsym;
    }
  }
  r=(min_err>0.0005);
  if (!r) {
    symopexp(xcn); if (mvld(xcn->spgr)) free(xcn->spgr);
    xcn->spgr=Strdup(name);
    if (xcn->spgr) {
      /* substitutute '/' for '_' and ':' which were used because of OS restrictions */
      for (P=xcn->spgr;*P;P++) {
	if (*P=='_'||*P==':') *P='/';
      }
    }
  }else{
    err_set(1,"spacegroup %s is incompatible with cell.",name);
    err_set(1,"determinant error = %0.5f, length error = %0.5f%",min_det,min_len);
  }
  UCpclose(Spg);
  xcn->atf=sv_atf; xcn->fta=sv_fta;
  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);
  return(r);
}
