#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "transpor.h"
#include "errhand.h"
#include "parser.h"

#define NO 0
#define YES 1
#define Return(x) {UCpclose(in);return(x);}
#define MAXTOKENSIZE 133
#define RECSZ 256
#define TRUE 1
#define FALSE 0
#define Fgets(x,f) {if (!fgets(x,sizeof(x),f)) *x='\0';if (strchr(x,'\n')) *strrchr(x,'\n')='\0';}

#define CHUNK 1000

static int *Mob=(int *)NULL;
static int NMob=0;

/*********************************************************************/
int read_cnf(char *file,Node *root)
{
  /* read multi conformation file from conformational analysis */

  Parse p;
  int n,s,maxn,i,j,k,*mob,nmob,nerr;
  char buf[256],rec[249];
  FILE *in;
  UCpipe *fpipe;
  CnfNode *new;
  float *X,*NX;

  p.mts=256; p.buf=rec; p.tok=buf; p.sep="";

  
  fpipe=UCpopen(extfil(rec,file,".cnf",NO));
  if (!mvld(fpipe)) {
    err_set(1,"could not open %s for read.",extfil(rec,file,".cnf",NO));
    return(1);
  }

  in=fpipe->fd;

  s=0; X=(float *)NULL; maxn=0;

  for(;;) {
    if (s==0) Fgets(rec,in);
    s=0;
    if (feof(in)) break;
    p.pos=0; p.sep="";
    if (!get_token(&p)) continue;
    if (strnceq(p.tok,"Mobile",strlen("Mobile"))) {
      if (!get_token(&p)) {
	err_set(1,"read_cnf: bad format of Mobile record:");
	err_set(1,"%s",rec);
	s=1;
	break;
      }
      nerr=err_seqnr();
      nmob=str2d(p.tok);
      if (nerr!=err_seqnr()) {
	err_set(1,"read_cnf: could not read number of mobile atoms:");
	err_set(1,"%s",rec);
	s=1;
	break;
      }
      mob=(int *)malloc(nmob*sizeof(int));
      if (!mvld(mob)) {
	err_set(1,"read_cnf: failed to allocate memory for mobile atom list.");
	s=1;
	break;
      }
      for (k=0;k<nmob;) {
	if (!get_token(&p)) {
	  p.pos=0;
	  Fgets(rec,in);
	  if (feof(in)) {
	    err_set(1,"read_cnf: premature EOF in Mobile atom section.");
	    break;
	  }
	}else{
	  mob[k]=str2d(p.tok)-1;
	  if (nerr!=err_seqnr()) {
	    err_set(1,"read_cnf: could not read mobile atoms:");
	    err_set(1,"%s",rec);
	    break;
	  }
	  if (mob[k]<0) {
	    err_set(1,"read_cnf: mobile atom numbers must be > 0:");
	    err_set(1,"%s",rec);
	  }	    
	  k++;
	}
      }
      if (k<nmob) {
	err_set(1,"read_cnf: expected %d mobile atoms, but got %d.",nmob,k);
	if (mob) free(mob);
	mob=(int *)NULL;
	s=1;
	break;
      }
      if (mob) {
	if (Mob) {
	  if (NMob!=nmob) {
	    err_set(1,"read_cnf: mobile section inconsistent with previous definition");
	    err_set(1,"using first definition");
	  }
	  free(mob);
	}else{
	  NMob=nmob;
	  Mob=mob;
	}
	mob=(int *)NULL;
	nmob=0;
      }

    }else if (strnceq(p.tok,"Conformation",strlen("Conformation"))) {

      if (!get_token(&p)) continue;
    
      new=(CnfNode *)add_node(root,p.tok,OBJ_CNF);
      if (!mvld(new)) {
	err_set(1,"failed to create Cnf node.");
	s=1;
	break;
      }

      p.sep="=";
      if (get_token(&p)) {
	if (streq(p.tok,"Energy")) {
	  if (get_token(&p)) {
	    new->energy=str2f(p.tok);
	  }
	}
      }

      for (n=0;;n++) {
	Fgets(rec,in); 
	if (feof(in)) break;
	if ((rec[0]>='A'&&rec[0]<='Z')||(rec[0]>='a'&&rec[0]<='z')) {
	  s=-1;
	  break;
	}
	p.pos=0; p.sep="";
	for (i=0;i<3;i++) {
	  if (!get_token(&p)) break;
	  if (n==maxn) {
	    NX=(float *)malloc(3*(maxn+CHUNK)*sizeof(float));
	    if (!mvld(NX)) {
	      err_set(1,"memory allocation failed in read_cnf.");
	      break;
	    }
	    maxn+=CHUNK;
	    if (mvld(X)) {
	      for (j=0;j<3*n;j++) NX[j]=X[j];
	      free(X);
	    }
	    X=NX;
	  }
	  X[i+3*n]=str2f(p.tok);
	}
	if (s==-1) break;
	if (i<3) {
	  err_set(1,"bad format of cnf file...");
	  s=1;
	  break;
	}
      }
      if (s<=0) {
	new->x=(float *)malloc(3*n*sizeof(float));
	if (mvld(new)) {
	  for (j=0;j<3*n;j++) new->x[j]=X[j];
	  new->n=n;
	}
      }else{
	del_node((Node *)new);
	break;
      }

    }else continue;

  }

  if (mvld(X)) free(X);
  UCpclose(fpipe);
  return(s);
}
/*-------------------- get_mobile_cnf --------------------*/

int *get_mobile_cnf(int *nmob)
{
  /* Return mobile atom definition to caller. It is the caller's */
  /* responsibility to eventually free the mobile atom list.     */

  int *r;

  r=Mob;
  *nmob=NMob;
  Mob=(int *)NULL;
  NMob=0;
  return(r);
}
