#include <stdio.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "fileutil.h"
#include "transpor.h"
#include "errhand.h"
#include "parser.h"
#include "bhtree.h"

#define NO 0
#define YES 1
#define Return(x) {UCpclose(in);return(x);}
#define MAXTOKENSIZE 133
#define RECSZ 160
#define TRUE 1
#define FALSE 0

#define MXLIST 128
#define MXTREE 2048
#define MAX_DOTS 8192

extern int graphix;
extern Node *DefaultNode;

/*********************************************************************/
int read_dots(char *file,Nodechain *c)
     
     /*   Read dots file  */
{
  char rec[RECSZ];
  UCpipe *in;
  Token *key;
  int status=0;
  BHtree *tree;
  Atom *a,*candidate[MXLIST];
  float X[3],x[MAX_DOTS][4],cutoff,dist[MXLIST],d,pmin,pmax;
  int i,k,n,m,grain,my_chain=0,ndots;
  
  in=UCpopen(extfil(rec,file,".dot",NO));
  if (!in) {
    err_set(1,"read_dots: could not open %s\n",extfil(rec,file,".dot",NO));
    return(1);
  }
  
  key=(Token *)NULL;

  if (!c) {
    err_dump();
    status=1;
    key=Get_Atoms_D("read dots for atoms: ",";");
    send_bheader(graphix,"ARMK",0);
    if (key->status==TOK_OK) {
      c=(Nodechain *)key->ptr;
      my_chain=TRUE;
    }else if (key->status==TOK_NOK) {
      err_set(1,"bad atom selection: %s",key->text);
      my_chain=FALSE;
    }
    key=free_Token(key);
  }else{
    my_chain=FALSE;
  }
  if (c) {
    if (c->n > 0) {
      grain = ( c->n < MXTREE ? 1 : c->n / MXTREE );
      cutoff=0.1;
      tree=generateBHtree(c,grain);
      if (my_chain) clear_node_chain(c);
      if (mvld(tree)) {
	n=0;a=(Atom *)NULL;
	pmin=1.e32; pmax=-pmin;
	ndots=0;
	while(fgets(rec,RECSZ,in->fd)) {
	  if (strneq(rec,"ATOM",4)) {
	    send_dots(a,n,(float *)x);
	    sscanf(rec+4,"%f %f %f",X,X+1,X+2);
	    m=findBHcloseAtoms(tree,X,cutoff,candidate,dist,MXLIST);
	    if (m>0) {
	      d=dist[0]; 
	      a=candidate[0];
	      for (i=1;i<m;i++) {
		if (dist[i]<d) {
		  d=dist[i];
		  a=candidate[i];
		}
	      }
	    }else{
	      a=(Atom *)NULL;
	    }
	    n=0;
	  }else if (strneq(rec,"DOT ",4)) {
	    if (n<MAX_DOTS) {
	      k=sscanf(rec+4,"%f %f %f %f",x[n],x[n]+1,x[n]+2,x[n]+3);
	      if (k<4) x[n][3]=0.0;
	      pmin=(pmin > x[n][3] ? x[n][3] : pmin);
	      pmax=(pmax < x[n][3] ? x[n][3] : pmax);
	      n++; ndots++;
	    }
	  }
	}
	send_dots(a,n,(float *)x);
	freeBHtree(tree);
	status=0;
	if (ndots==0) {
	  pmin=0;
	  pmax=0;
	}
	if (pmin<pmax) {
	  prtGraphics("%d dots, min. prop. = %0.5f, max. prop = %0.5f\n",
		      ndots,pmin,pmax);
	}else{
	  prtGraphics("%d dots, property value = %0.5f\n",ndots,pmin);
	}
      }else{
	err_set(1,"failed to generate BHtree (internal error.)");
      }
    }else{
      err_set(1,"no atoms in specified molecule.");
    }
  }

  free_Token(key);
  err_dump();
  UCpclose(in);
  if (my_chain) {
    clear_node_chain(c);
    free_node_chain(c);
  }
  return(status);
}
/*-------------------- send_dots --------------------*/

void send_dots(Atom *a,int n,float *x)
{
  int size;

  if (n<1 || !mvld(a) || !cPtr(a->cross)) return;
  size=4*n*sizeof(float)+cPtrSize();
  send_bheader(graphix,"DOTA",size);
  send_bcptr(graphix,a->cross);
  send_bfloata(graphix,4*n,x);
}

void send_dots_np(Atom *a,int n,float *x)
{
  int size,i;

  /* send dots, no property in x */

  if (n<1 || !mvld(a) || !cPtr(a->cross)) return;
  size=4*n*sizeof(float)+cPtrSize();
  send_bheader(graphix,"DOTA",size);
  send_bcptr(graphix,a->cross);
  for (i=0;i<n;i++) {
    send_bfloata(graphix,3,x+3*i);
    send_bfloat(graphix,0.0);
  }
}

  
		  
