#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "errhand.h"
#include "xtal.h"

extern int graphix;
extern Node *LibRoot;
extern Node *MolRoot;

int autocenter_on_read;
int autobond_on_read;
int send_to_graphix;

#define TRUE 1
#define FALSE 0

static int maccs_sanity_check(Node *root);
static int pipenr=0;

/*-------------------- maccs_sanity_check --------------------*/

static int maccs_sanity_check(Node *root)
{
  Mol *r;

  ForMols(r,MolRoot) if (root==(Node *)r) break;
  return(is_mol(r));
}

/*-------------------- ReadNode --------------------*/

Node *ReadNode(void *vroot,char *filn,int type)
{
  Node *mol,*model,*r,*root;
  Mol *nm;
  Dspobj *global,*droot;
  NOElist *noes;
  NOEdata *noe;
  Atom *a;
  int status,nat,count;
  char *fn,pipename[30];
  
  root=(Node *)vroot;
  fn=strrchr(filn,'/');
  fn = (fn == (char *)NULL ? filn : fn+1);
  if (*filn=='|') {
    sprintf(pipename,"pipe%d",++pipenr);
    fn=pipename;
  }

  if (type==FT_WLB) {
    prtGraphics("reading library from file %s ...\n",filn);
    status=read_lib(filn,(Lib *)root);
    if (status) {
      err_set(1,"read lib failed: status=%d",status);
    }
    r=(Node *)NULL;

  }else if (type==FT_DOTS) {
    read_dots(filn,(Nodechain *)NULL);
    r=(Node *)NULL;

  }else if (type==FT_CNF) {
    read_cnf(filn,root);
    r=(Node *)NULL;

  }else if (type==FT_CNT) {
    *((Grid **)vroot) = read_grid(filn); 
    r=(Node *)NULL;

  }else if (type==FT_NOE || type==FT_DUPPER) {
    if (!mvld(root)) root=MolRoot;
    if (!is_mol(root)) {
      err_set(1,"read_node: can't read noe list into %s.",root->name);
    }else if (((Mol *)root)->moltype!=MOL_NMRSET) {
      err_set(1,"%s is not an NMR set.",root->name);
    }else{
      noes=(NOElist *)make_node(OBJ_NOELIST);
      if (is_NOElist(noes)) {
	if (type==FT_DUPPER) status=read_diana_noe(filn,noes);
	else if (type==FT_NOE) status=read_xplor_noe(filn,noes);
	else {
	  err_set(1,"internal error in readnode...");
	  status=1;
	}
	if (status) {
	  err_set(1,"NMR NOE read: failed to read file.");
	  del_node((Node *)noes);
	}else{
	  for(count=0,noe=noes->first;mvld(noe);noe=noe->nxt) count++;
	  set_node_name((Node *)noes,fn);
	  noes->up=root; link_node((Node *)noes);
	  prtGraphics("NOE list %s contains %d NOEs\n",fn,count);
	}
      }else{
	err_set(1,"cmd_nmr_noe: memory allocation failure.");
      }
    }
    r=(Node *)NULL;

  }else if (type&FT_MOL_TYPES) {

    if (type!=FT_SESS) {
      if (!mvld(root)) root=MolRoot;
      if (!is_mol(root)) {
	err_set(1,"read_node: cant read molecule into %s.",root->name);
	mol=(Node *)NULL;
      }else{
	mol=add_node(root,fn,OBJ_MOL);
	if (!mol) {
	  err_set(1,"readnode: memory allocation failed...\n");
	}
      }
    }
    if (type==FT_SESS) {
      restoreSession(NULL,filn);
      mol=NULL; /* to make the rest of this function happy */
    }else if (is_mol(mol)) {
      ((Mol *)mol)->moltype=MOL_FILE;
      send_to_graphix=TRUE;
      prtGraphics("reading molecule from file %s ... \n",filn);
      switch(type) {
      case FT_CAR :
	status=read_bio((UCpipe *)NULL,filn,mol);
	break;
      case FT_BIOC:
	status=read_biocad((UCpipe *)NULL,filn,mol);
	send_to_graphix=TRUE;       
	break;
      case FT_MACCS:
      case FT_MDL:
	droot=glblobj(assocobj(root));
	global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	if (mvld(global)) global->root=mol;
	((Mol *)mol)->dsptr=global;
	((Mol *)mol)->keep=1;	
	status=read_maccs((UCpipe *)NULL,filn,mol,type==FT_MACCS);
	send_to_graphix=FALSE;       /* graphics objects sent by read_maccs */
	if (!maccs_sanity_check(mol)) {
	  mol=(Node *)NULL;
	  err_set(1,"readmaccs lost the root.");
	}else{
	  ((Mol *)mol)->keep=0;
	  if (!mvld(mol->fst)) {
	    prtGraphics("no molecules in %s, node deleted.\n",mol->name);
	    del_node(mol); mol=(Node *)NULL;
	  }
	}
	break;
      case FT_CSD:
	droot=glblobj(assocobj(root));
	global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	if (mvld(global)) global->root=mol;
	((Mol *)mol)->dsptr=global;
	((Mol *)mol)->keep=1;	
	status=read_csd((UCpipe *)NULL,filn,mol);
	send_to_graphix=FALSE;       /* graphics objects sent by read_csd */
	if (!maccs_sanity_check(mol)) {
	  mol=(Node *)NULL;
	  err_set(1,"readcsd lost the root.");
	}else{
	  ((Mol *)mol)->keep=0;	
	  if (!mvld(mol->fst)) {
	    prtGraphics("no molecules in %s, node deleted.\n",mol->name);
	    del_node(mol); mol=(Node *)NULL;
	  }
	}
	break;
      case FT_MMOD:
	status=read_mmod((UCpipe *)NULL,filn,mol);
	break;
      case FT_MOL:
      case FT_MOL2:
	status=read_syb((UCpipe *)NULL,filn,mol,type);
	break;
      case FT_PDB:
      case FT_XPDB:
	status=read_pdb((UCpipe *)NULL,filn,mol,type);
	break;
      case FT_SHX:
	status=read_shelx((UCpipe *)NULL,filn,mol);
	break;
      case FT_CIF:
	status=read_cif((UCpipe *)NULL,filn,mol);
	break;
      case FT_DIANA:
	status=read_diana((UCpipe *)NULL,filn,mol);
	break;
      case FT_PAR:
	status=read_par((UCpipe *)NULL,filn,mol);
	break;
      case FT_MAE:
	status=read_mae((UCpipe *)NULL,filn,mol);
	break;
      default:
	status = -1;
	break;
      }
      if (status!=0) {
	err_set(1,"Status is %d",status);
      }
      if (is_mol(mol)) {
	nat=0;
	ForAtoms(a,mol) {
	  nat++;
	}
	if ((nat==0 && status)||status<0) {
	  del_node(mol);
	  mol=(Node *)NULL;
	}else if (send_to_graphix) { 
	  prtGraphics("%d atoms read from %s\n",nat,filn);
	  condAutocenterMol(mol);
	  if (((Mol *)mol)->moltype==MOL_NMRSET) {
	    droot=glblobj(assocobj(root));
	    global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	    if (mvld(global)) {
	      ((Mol *)mol)->dsptr=global;
	      global->root=mol;
	      droot=global;
	      prtGraphics("sending vectorlists...\n");
	      lock_evtq();
	      forObjs(model,mol,OBJ_MOL) {
		global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
		if (mvld(global)) {
		  send_mdl(global,(Mol *)model);
		  ((Mol *)model)->dsptr=global;
		  global->root=model;
		}
	      } 
	      unlock_evtq();
	    }
	  }else{
	    droot=glblobj(assocobj(root));
	    global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
	    if (mvld(global)) {
	      prtGraphics("sending vectorlist...\n");
	      send_mdl(global,(Mol *)mol);
	      ((Mol *)mol)->dsptr=global;
	      global->root=mol;
	    }
	  }
	  condAutobondMol(mol);
	}
      }
    }
    r=mol;
  }else{
    err_set(1,"cmd_read: internal error, filetype (%s) is fugu.");
    r=(Node *)NULL;
  }
  if (is_mol(r)) {
    forMols(nm,root) if (r==(Node *)nm) break;
  }else nm=(Mol *)NULL;
  return((Node *)nm);
}
/*-------------------- readmoltypes --------------------*/

Node *readmoltypes(UCpipe *p,char *fname,int type)
{
  Node *r;
  int status=0;

  r=make_node(OBJ_MOL);
  if (r) {
    ((Mol *)r)->moltype=MOL_FILE;
    switch (type) {
    case FT_CAR:
      status=read_bio(p,fname,r);
      break;
    case FT_BIOC:
      status=read_biocad(p,fname,r);
      break;
    case FT_MACCS:
      status=read_maccs(p,fname,r,TRUE);
      break;
    case FT_MDL:
      status=read_maccs(p,fname,r,FALSE);
      break;
    case FT_CSD:
      status=read_csd(p,fname,r);
      break;
    case FT_MMOD:
      status=read_mmod(p,fname,r);
      break;
    case FT_MOL:
    case FT_MOL2:
      status=read_syb(p,fname,r,type);
      break;
    case FT_PDB:
    case FT_XPDB:
      status=read_pdb(p,fname,r,type);
      break;
    case FT_SHX:
      status=read_shelx(p,fname,r);
      break;
    case FT_CIF:
      status=read_cif(p,fname,r);
      break;
    case FT_DIANA:
      status=read_diana(p,fname,r);
      break;
    case FT_PAR:
      status=read_par(p,fname,r);
      break;
    default:
      err_set(1,"readmoltypes: type %d not implemented.",type);
      r=del_node(r);
      break;
    }
  }
  if (r) {
    if (status) {
      err_set(1,"read status is %d\n",status);
      if (!rfirst_atom(r)) r=del_node(r);
    }
  }
  return(r);
}

void autocenterMol(Node *mol)
{
  float c[3];
  mat_4 tra;
  Xcn *xcn;
  int nat,i;
  Atom *a;
  
  if (is_mol(mol)) {
    xcn=first_xcn(mol);
    nat=0;
    for (i=0;i<3;i++) c[i]=0.0;
    ForAtoms(a,mol) {
      for (i=0;i<3;i++) c[i]+=a->x[i];
      nat++;
    }
    tra=ident_4();
    if (nat>0) {
      for (i=0;i<3;i++) {
	c[i]/=nat;
	tra.r[i].c[3]=-c[i];
      }
    }
    ForAtoms(a,mol) transform_atom(tra,a);
    if (mvld(xcn)) {
      for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symF(xcn,xcn->sym[i]);
      for (i=0;i<3;i++) xcn->fta.r[i].c[3]-=c[i];
      xcn->atf=inv_4(xcn->fta);
      for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);
    }
  }
}

void condAutocenterMol(Node *mol)
{
  if (autocenter_on_read) {
    autocenterMol(mol);
  }
}

void autobondMol(Node *mol)
{
  Nodechain C ;
  int nbnd;

  if (is_mol(mol)) {
    C=tree2chain(mol,OBJ_ATOM);
    nbnd=bond_auto_make(&C,autobond_mode());
    if (nbnd>0) {
      prtGraphics("%d automatic bond%s added to %s\n",
		  nbnd,nbnd==1?"":"s",(mvld(mol->name) ? mol->name : "molecule"));
    }
    clear_node_chain(&C);
  }
}

void condAutobondMol(Node *mol)
{
  if (autobond_on_read) {
    autobondMol(mol);
  }
}
