#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "xtal.h"
#include "surface.h"
#include "streamio.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Node *DspRoot;
extern Node *LibRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))

#define BREAK(x) {err_set(1,x); break;}
#define chomp(x) if (strrchr(x,'\n')) *strrchr(x,'\n')='\0'; if (strrchr(x,'\r')) *strrchr(x,'\r')='\0'

#define CHUNK 4000

#define TRUE 1
#define FALSE 0

/* nodes table for MDT */
static Node **NTAB=NULL;
static int ndone=0;
static int maxNTAB=0;

/* nodes table for DDT */
static Node **MTAB=NULL;
static int mdone=0;
static int maxMTAB=0;

static Node *ResLib=NULL;
static float oldr;
static float obr=-1;
static int oblabt=-1;
static int oblw=-1;
static int obcyl=-1;
static int vers=0; 

static void xmlspclunquote(char *line);
static int restoreAtomDots(Stream *s);
static int restoreMDTNodes(Node *root,Stream *s);
static int restoreDDTNodes(Dspobj *root,Stream *s);
static int parseAtomData(Atom *a,Stream *s);
static int parseXCNData(Xcn *xcn,UCpipe *in);
static void postDDT(void);

/*========================================================================*/

static int parseAtomData(Atom *a,Stream *s)
{
  Bond *b;
  Atomdef *atp;
  int iup,nr,havexyz=1,seqnr,nc,n,off,l;
  unsigned long bits;
  int status=0;

  if (!is_atom(a)) {
    err_set(1,"parseAtomData: no atom.");
    status=9000;
  }

  seqnr=err_seqnr();
  /* 1st record */
  strm_flush(s);
  for (nr=0;!status&&token_stream(s,"");nr++) {
    switch (nr) {

    case 0:   /* FF type mnemonic */
      atp=find_atomdef_name(ResLib,strm_tok(s));
      if (!atp) {
	err_set(0,"could not find atom type %s in current library, guessing...",strm_tok(s));
	atp=guess_atom_type(strm_tok(s));
	if (atp) {
	  atp=(Atomdef *)copy_node((Node *)atp);
	  if (atp) {
	    set_node_name((Node *)atp,strm_tok(s));
	    atp->up=ResLib;
	    link_node((Node *)atp);
	  }
	}
      }
      if (!atp) {
	err_set(1,"failed, session restore aborted");
	status=9001;
      }
      set_atom_type(a,atp);
      break;
	  
    case 1:   /* color */
      a->color=str2i(strm_tok(s));
      break;

    case 2:   /* x */
    case 3:   /* y */
    case 4:   /* z */
      if (!streq(strm_tok(s),".")) a->x[nr-2]=str2f(strm_tok(s));
      else havexyz=0;
      break;

    case 5:   /* part. charge q */
      if (!streq(strm_tok(s),".")) a->q=str2f(strm_tok(s));
      break;

    case 6:   /* isotropic B factor */
      if (!streq(strm_tok(s),".")) a->x[ATM_BISO]=str2f(strm_tok(s));
      break;

    case 7:   /* site occupancy */
      if (!streq(strm_tok(s),".")) a->x[ATM_SOF]=str2f(strm_tok(s));
      break;

    case 8:   /* chemical shift */
      if (!streq(strm_tok(s),".")) a->shift=str2f(strm_tok(s));
      break;

    default:
      break;  /* ignore extra data */
    }
  }
  if (havexyz&&nr>4) a->nxyz=1;
  else a->nxyz=0;
      
  if (err_seqnr()>seqnr) {
    err_set(1,"format error, restore aborted");
    status=9002;
  }

  /* 2nd record: radius, label color, label type, label text */
  strm_flush(s);
  a->r=oldr;
  for (nr=0;!status&&token_stream(s,"");nr++) {
    switch (nr) {
    case 0:  /* radius */
      if (streq(strm_tok(s),".")) a->r=oldr;
      else if (streq(strm_tok(s),"v")) a->r=(a->atp?a->atp->rvdw:0.1);
      else {
	a->r=str2f(strm_tok(s));
	oldr=a->r;
      }
      break;
	
    case 1:  /* label color */
      if (streq(strm_tok(s),".")) a->lcolor=WNP_COLOR_BYATOM;
      else a->lcolor=str2i(strm_tok(s));
      break;

    case 2:  /* labeltype */
      a->ltype=str2i(strm_tok(s));
      if (a->ltype!=ATM_LAB_NONE) {
	if (a->label) free(a->label);
	a->label=Strdup(strm_buff(s)+strm_pos(s));
	if (strchr(a->label,'\n')) *strchr(a->label,'\n')='\0';
	strm_pos(s)+=strlen(a->label);
	if (strlen(a->label)==0) {
	  free(a->label);
	  a->label=NULL;
	}else{
	  a->ltype=ATM_LAB_TEXT;
	}
      }
      break;

    default:
      break; /* ignore extra data */
    }
  }

  /* 3rd record  */

  strm_flush(s);
  for (nr=0;!status&&token_stream(s,"");nr++) {
    switch (nr) {
    case 0:  /* ligance */
      if (strlen(strm_tok(s))==1&&(strm_tok(s)[0]=='?'||strm_tok(s)[0]=='.')) a->lig='?';
      else a->lig=str2i(strm_tok(s));
      break;

    case 1:  /* formal charge */
      if (strlen(strm_tok(s))==1&&(strm_tok(s)[0]=='?'||strm_tok(s)[0]=='.')) a->chrg='?';
      else a->chrg=str2i(strm_tok(s));
      break;

    case 2:  /* bits packed into integer */
      sscanf(strm_tok(s),"%lx%n",&bits,&nc);
      if (nc!=strlen(strm_tok(s))) {
	err_set(1,"bad bits: %s",strm_tok(s));
	err_set(1,"format error, restore aborted");
	status=9003;
      }
      a->dsp=bits%2;    bits/=2;
      a->ldsp=bits%2;   bits/=2;
      a->sdsp=bits%2;   bits/=2;
      a->arom=bits%2;   bits/=2;
      a->oop=bits%4;    bits/=4;
      a->cosw=bits%2;   bits/=2;
      a->stass=bits%4;  bits/=4;
      a->fe=bits%2;     bits/=2;
      break;

    default: /* atomset words */
      if (nr-3<ATOMSET_WORDS) {
	sscanf(strm_tok(s),"%lx%n",a->set+(nr-3),&nc);
	if (nc!=strlen(strm_tok(s))) err_set(1,"bad atomset word: %s",strm_tok(s));
      }else {
	/*** this is a problem that we currently ignore...
	err_set(1,"too much data in 2nd line of atom data");
	status=9004;
	***/
      }
    }
  }
  if (!status&&err_seqnr()>seqnr) {
    err_set(1,"format error, restore aborted");
    status=9005;
  }


  /* 4th record: charge group and bond list  */
  
  strm_flush(s);
  if (!status&&token_stream(s,"")) {
    if (strm_tok(s)[0]!='.') StrCpy(a->grp,strm_tok(s),sizeof(a->grp));
    else StrCpy(a->grp,"??",sizeof(a->grp));
    while (!status&&token_stream(s,",")) {
      iup=str2i(strm_tok(s));
      if (iup>0&&iup<ndone&&is_atom(NTAB[iup])) {
	b=add_bond_ng((Atom *)NTAB[iup],a);
	b->cyl=obcyl; b->lw=oblw; b->r=obr; b->labt=oblabt;
	set_bt(b,1);
	for (n=0;strm_term(s)==',';n++) {
	  if (!token_stream(s,",")) {
	    err_set(1,"%s\n",strm_buff(s));
	    err_set(1,"atom data has invalid bond list");
	    status=9006;
	  }
	  switch (n) {
	  case 0:
	    if (strm_ltok(s)>0)  set_bt(b,BT_number(strm_tok(s)));
	    break;
	  case 1:
	    if (strm_ltok(s)>0) {
	      oblw=str2i(strm_tok(s));
	      b->lw=oblw;
	    }
	    break;
	  case 2:
	    if (strm_tok(s)[0]=='Y'||strm_tok(s)[0]=='N') {
	      obcyl=strm_tok(s)[0]=='Y'?1:0;
	      b->cyl=obcyl;
	      off=1;
	    }else off=0;
	    if (strm_tok(s)[off]!='\0') {
	      obr=str2f(strm_tok(s)+off);
	      b->r=obr;
	    }
	    break;
	  case 3:
	    if (strm_ltok(s)>0) {
	      if (strm_parse(s)->quoted) {
		b->labt=BND_LAB_TEXT;
		b->label=strdup(strm_tok(s));
	      }else{
		oblabt=str2i(strm_tok(s));
		b->labt=oblabt;
	      }
	    }
	    break;
	  default:
	    err_set(1,"%s\n",strm_buff(s));
	    err_set(1,"atom data has invalid bond list");
	    status=9007;
	    break;
	  }
	}
      }else{
	err_set(1,"invalid node number in bond list");
	status=9008;
      }
    }
  }
  
  /* alternate atom sites, if a->mxyz > 1 */

  for (l=1;l<a->mxyz;l++) {
    strm_flush(s);
    for (nr=0;nr<ATM_LOCSZ;nr++) a->x[l*ATM_LOCSZ+nr]=0.0;
    a->x[l*ATM_LOCSZ+ATM_SOF]=1.0;
    for (nr=0;!status&&token_stream(s,"");nr++) {
      switch (nr) {
      case 0:  /* x */
      case 1:  /* y */
      case 2:  /* z */
	if (!streq(strm_tok(s),".")) a->x[l*ATM_LOCSZ+nr]=str2f(strm_tok(s));
	else a->x[l*ATM_LOCSZ+nr]=0.0;
	break;
      case 3:  /* biso */
	if (!streq(strm_tok(s),".")) a->x[l*ATM_LOCSZ+ATM_BISO]=str2f(strm_tok(s));
	else a->x[l*ATM_LOCSZ+ATM_BISO]=0.0;
	break;
      case 4:  /* sof */
	if (!streq(strm_tok(s),".")) a->x[l*ATM_LOCSZ+ATM_SOF]=str2f(strm_tok(s));
	else a->x[l*ATM_LOCSZ+ATM_SOF]=1.0;
	break;
      }
    }
  }
  a->nxyz=a->mxyz;

  return(status);
}
/*------------------------------------------------------------------------*/

static int parseXCNData(Xcn *xcn,UCpipe *in)
{
  char buff[8192],fld[1024];
  Parse p;
  hisrec *h,*nh;
  int i,j,have_cell=0,have_xtoa=0,seqnr,old_err,loop=TRUE,dupl,status=0;
  float err;
  Symop sym_n;
  mat_4 X2A;

  p.buf=buff;
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);

  seqnr=err_seqnr();

  if (!is_xcn(xcn)) {
    err_set(1,"no XCN to hold data.");
    status=9001;
  }

  while (loop&&fgets(buff,sizeof(buff),in->fd)) {
    chomp(buff); p.pos=0;
    if (get_token(&p)) {

      if (streq(p.tok,"T")) {
	if (mvld(xcn->title)) free(xcn->title);
	xcn->title=NULL;
	if (p.buf[p.pos]) xcn->title=Strdup(p.buf+p.pos);

      }else if (streq(p.tok,"L")) {
	xcn->lambda=str2f(token(&p));

      }else if (streq(p.tok,"C")) {
	for (i=0;i<3;i++) xcn->cl[i]=str2f(token(&p));
	for (i=0;i<3;i++) xcn->ca[i]=str2f(token(&p));
	xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca); 
	xcn->atf=inv_4(xcn->fta);
	have_cell=1;

      }else if (streq(p.tok,"c")) {
	for (i=0;i<3;i++) xcn->esd_cl[i]=str2f(token(&p));
	for (i=0;i<3;i++) xcn->esd_ca[i]=str2f(token(&p));

      }else if (streq(p.tok,"F")) {
	xcn->forml=0;
	while (get_token(&p)) {
	  if (p.term=='\0') break;   /* ?? needed ?? */
	  if (p.ltok<1 || p.ltok>2) BREAK("bad formula record in XCN.");
	  if (xcn->forml>=MX_XCN_FORMUL) BREAK("too many items in formula in XCN");
	  Strcpy(xcn->forms[xcn->forml],p.tok);
	  xcn->formn[xcn->forml]=1;
	  if (strchr("0123456789",p.buf[p.pos])) {
	    xcn->formn[xcn->forml]=str2i(token(&p));
	    if (xcn->formn[xcn->forml]<1) BREAK("element count in formula in XCN is not a number");
	  }
	  xcn->forml++;
	}

      }else if (streq(p.tok,"S")) {
	xcn->lattice=str2i(token(&p));
	xcn->centric=(xcn->lattice > 0 ? TRUE : FALSE);
	if (!xcn->centric) xcn->lattice = -xcn->lattice;
	if (xcn->lattice<1 || xcn->lattice>7) BREAK("bad lattice parameter in spgrp record.");
	if (mvld(xcn->spgr)) free(xcn->spgr);
	xcn->spgr=Strdup(p.buf+p.pos);
	if (!xcn->spgr) BREAK("memory allocation failure.");

      }else if (streq(p.tok,"s")) {
	sym_n=str2sym(p.buf+p.pos);
	for (dupl=FALSE,i=0; !dupl&&i<xcn->nsym; i++) dupl=symopequiv(sym_n,xcn->sym[i]);
	if (!dupl) {
	  if (xcn->nsym<MX_XCN_SYMOPS) {
	    xcn->sym[xcn->nsym]=sym_n;
	    xcn->nsym++;
	  }else BREAK("too many symmetry operations.");
	}
      }else if (streq(p.tok,"A")) {
	for (i=0;i<6;i++) xcn->asym[i]=str2f(token(&p));

      }else if (streq(p.tok,"scale")) {
	xcn->scale=Strdup(p.buf+p.pos);
	if (mvld(xcn->scale)) free(xcn->scale);
	xcn->scale=Strdup(p.buf+p.pos);
    
      }else if (streq(p.tok,"H")) {
	for (h=xcn->history;h;h=h->nxt) if (!h->nxt) break;
	nh=make_history();
	if (nh) {
	  nh->t=Strdup(p.buf+p.pos);
	  if (h) h->nxt=nh;
	  else xcn->history=nh;
	}else BREAK("memory allocation failure.");

      }else if (streq(p.tok,"M")) {
	have_xtoa=1+have_cell;
	if (have_cell) {
	  for (i=0;i<3;i++) for (j=0;j<4;j++) X2A.r[i].c[j]=str2f(token(&p));
	  for (j=0;j<3;j++) X2A.r[3].c[j]=0.0;
	  X2A.r[3].c[3]=1.0;
	  xcn->fta=X2A;
	  old_err=err_seqnr();
	  xcn->atf=inv_4(xcn->fta);
	  if (old_err!=err_seqnr()) BREAK("singular cell to angstrom matrix.");
	}
	loop=FALSE; /* this is always the last record */
      }
    }
  }
  if (seqnr==err_seqnr()) {
    /* expand symmetry operators */
    i=symopexp(xcn);
    if (i!=xcn->nsymx) {
      err_set(1002,"rstore XCN: more than %d expanded symops",MX_XCN_SYMOPS);
      xcn->nsymx=xcn->nsym;
    }

    /* convert symmetry operators to angstrom/cartesian */
    for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);
  
    if (have_xtoa==1) err_set(1,"restoreXCNode: XtoA, but no cell, XtoA ignored");

    err=check_fta_matrix(xcn->fta,xcn->cl,xcn->ca);
    if (err>0.1) err_set(0,"restoreXCNode: XCN cell and XtoA inconsistent, error= %0.3f A.",err);
  }
  
  if (err_seqnr()>seqnr) {
    err_set(1,"restoreXCNode: format error, restore aborted");
    status=9000;
  }

  return(status);
}

/*========================================================================*/      

static int restoreMolNode(Node *root,Stream *s)
{
  int status=0,ns,i,nc;
  Mol *new=NULL;

  /* create and register new node */
  if (!is_mol(root)) {
    err_set(1,"restoreMolNode: no mol root for node");
    status=9000;
  }else   if (ndone<maxNTAB) {
    new=(Mol *)make_node(OBJ_MOL);
    NTAB[ndone++]=(Node *)new;
    if (!new) {
      err_set(1,"restoreMolNode: failed to add new mol node.");
      status=9001;
    }else{
      new->up=root;
      link_node((Node *)new);
    }
  }else{
    err_set(1,"restoreMolNode: node table overflow");
    err_set(1,"%s",strm_buff(s));
    status=9002;
  }    

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  /* parse content */
  new->moltype=str2i(token_stream(s,""));
  new->keep=str2i(token_stream(s,""));
  ns=str2i(token_stream(s,""));
  for (i=0;i<ns&&!status;i++) {
    if (token_stream(s,"")) {
      if (i<ATOMSET_WORDS) {
	sscanf(strm_tok(s),"%lx%n",new->used+i,&nc);
	if (nc!=strm_ltok(s)) {
	  err_set(1,"bad atomset word in mol data: %s.",strm_tok(s));
	  status=9003;
	}
      }
    }
  }
  for (i=0;i<ns&&!status;i++) {
    if (token_stream(s,"")) {
      if (i<ATOMSET_WORDS) {
	sscanf(strm_tok(s),"%lx%n",new->local+i,&nc);
	if (nc!=strm_ltok(s)) {
	  err_set(1,"bad atomset word in mol data: %s.",strm_tok(s));
	  status=9003;
	}
      }
    }
  }

  if (!status) status=restoreMDTNodes((Node *)new,s);
  
  if (!status&&(!streq("</M",strm_tok(s))||strm_term(s)!='>')) {
    err_set(1,"badly nested XML elements, expected </M>");
    err_set(1,"%s",strm_buff(s));
    status=9004;
  }

  return(status);
}

static int restoreNOEList(Node *root,Stream *s)
{
  int status=0;
  NOElist *new=NULL;

  /* create and register new node */
  
  if (!is_mol(root)) {
    err_set(1,"restoreNOEList: no mol root for list");
    status=9000;
  }else if (ndone<maxNTAB) {
    new=(NOElist *)make_node(OBJ_NOELIST);
    NTAB[ndone++]=(Node *)new;
    if (!new) {
      err_set(1,"restoreNOEList: failed to add new NOElist node.");
      status=9001;
    }else{
      new->up=root;
      link_node((Node *)new);
    }
  }else{
    err_set(1,"restoreNOEList: node table overflow");
    err_set(1,"%s",strm_buff(s));
    status=9002;
  }    

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else if (streq("average",strm_tok(s))) {
      if (strm_term(s)!='>') {
	new->avtype=str2i(token_stream(s,">"));
	if (new->avtype==0) {
	  err_set(1,"invalid AV type in noe list");
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  if (!status) {
    set_stream(s,"eol=no,redirection='@',comment='!'"); /* set parser to XPLOR style */
    status=read_xplor_noe_stream(s,new);
    set_stream(s,"eol=no,redirection=off, comment=off");
  }
  
  if (!status&&(!streq("</N",strm_tok(s))||strm_term(s)!='>')) {
    err_set(1,"badly nested XML elements, expected </N>");
    err_set(1,"%s",strm_buff(s));
    status=9004;
  }

  return(status);
}

static int restoreAtomNode(Node *root,Stream *s)
{
  int status=0,l;
  Atom *new=NULL;

  /* create and register new node */
  if (!is_mol(root)) {
    err_set(1,"restoreAtomNode: no mol root for atom");
    status=9000;
  }else if (ndone<maxNTAB) {
    new=(Atom *)make_node(OBJ_ATOM);
    NTAB[ndone++]=(Node *)new;
    if (!new) {
      err_set(1,"restoreAtomNode: failed to add new atom node.");
      status=9001;
    }else{
      new->up=root;
      link_node((Node *)new);
    }
  }else{
    err_set(1,"restoreAtomNode: node table overflow");
    err_set(1,"%s",strm_buff(s));
    status=9002;
  }    

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else if (streq("alt",strm_tok(s))&&strm_term(s)!='>') {           /* alternate site indicators */
      if (token_stream(s,">")) {
	l=strlen(strm_tok(s));
	if (l>1) {
	  reall_atom_x(new,l);
	  strncpy(new->alt,strm_tok(s),l);
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  if (!status) {
    set_stream(s,"eol=yes");
    status=parseAtomData(new,s);
    set_stream(s,"eol=no");
  }
  if (!status) status=restoreMDTNodes((Node *)new,s);

  if (!status&&(!streq("</A",strm_tok(s))||strm_term(s)!='>')) {
    err_set(1,"badly nested XML elements, expected </A>");
    err_set(1,"%s",strm_buff(s));
    status=9004;
  }
  
  return(status);
}  

int restoreAtomsetNode(Node *root,Stream *s)
{
  Atomset *new=NULL;
  int bit=-1,done=FALSE,status=0,seqnr;

  /* create and register new node */
  if (!is_mol(root)) {
    err_set(1,"restoreAtomsetNode: no mol root for atomset");
    status=9000;
  }else if (ndone<maxNTAB) {
    new=(Atomset *)make_node(OBJ_ATOMSET);
    NTAB[ndone++]=(Node *)new;
    if (!new) {
      err_set(1,"restoreAtomsetNode: failed to add new atomset node.");
      status=9001;
    }else{
      new->up=root;
      link_node((Node *)new);
    }
  }else{
    err_set(1,"restoreAtomsetNode: node table overflow");
    err_set(1,"%s",strm_buff(s));
    status=9002;
  }    

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    else if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else if (streq("bit",strm_tok(s))) {
      if (strm_term(s)!='>') {
	seqnr=err_seqnr();
	bit=str2i(token_stream(s,">"));
	if (seqnr<err_seqnr()) {
	  err_set(1,"invalid bit number in set definition");
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  if (!status) {
    if (bit<0||bit>=ATOMSET_BITS) {
      err_set(1,"invalid atom set definition");
      err_set(1,"%s",strm_buff(s));
      status=9004;
    }else{
      new->bit=bit;
    }
  }

  if (!status&&!done) {
    status=restoreMDTNodes(root,s);
    if (!status&&(!streq("</S",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </S>");
      err_set(1,"%s",strm_buff(s));
      status=9005;
    }
  }

  return(status);
}

static int restoreXCNode(Node *root,Stream *s)
{
  Xcn *new=NULL;
  int done=FALSE,status=0;

  /* create and register new node */
  if (!is_mol(root)) {
    err_set(1,"restoreXCNode: no mol root for atomset");
    status=9000;
  }else if (ndone<maxNTAB) {
    new=(Xcn *)make_node(OBJ_XCN);
    NTAB[ndone++]=(Node *)new;
    if (!new) {
      err_set(1,"restoreXCNode: failed to add XCN atomset node.");
      status=9001;
    }else{
      new->up=root;
      link_node((Node *)new);
    }
  }else{
    err_set(1,"restoreXCNode: node table overflow");
    err_set(1,"%s",strm_buff(s));
    status=9002;
  }    

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    else if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  if (!done) {
    strm_flush(s);
    if (!status) status=parseXCNData(new,(UCpipe *)(s->in));
    if (!status) status=restoreMDTNodes((Node *)new,s);
  
    if (!status&&(!streq("</X",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </X>");
      err_set(1,"%s",strm_buff(s));
      status=9004;
    }
  }
  
  return(status);
}

static int restoreMDTNodes(Node *root,Stream *s)
{
  int status=0;

  while (!status&&!strm_eof(s)) {
    if (token_stream(s,">")) {
      if (streq("<M",strm_tok(s))) status=restoreMolNode(root,s);
      else if (streq("<A",strm_tok(s))) status=restoreAtomNode(root,s);
      else if (streq("<N",strm_tok(s))) status=restoreNOEList(root,s);
      else if (streq("<S",strm_tok(s))) status=restoreAtomsetNode(root,s);
      else if (streq("<X",strm_tok(s))) status=restoreXCNode(root,s);
      else if (strneq("</",strm_tok(s),2)) break;
      else ignoreXMLToken(s); /* just ignore */
    }
  }
  return(status);
}
/*========================================================================*/

int restoreMDT(Stream *s)
{
  int done=FALSE,status=0,seqnr;

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    if (streq("nodes",strm_tok(s))) {
      if (strm_term(s)!='>') {
	seqnr=err_seqnr();
	maxNTAB=str2i(token_stream(s,">"));
	if (seqnr<err_seqnr()) {
	  err_set(1,"invalid value of nodes attribute");
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }

  if (maxNTAB>0) {
    NTAB=malloc(maxNTAB*sizeof(Node *));
    if (!NTAB) {
      err_set(1,"failed to allocate node table of size %d in restore MDT",maxNTAB);
      status=9004;
    }else{
      NTAB[0]=MolRoot; ndone=1;
    }
  }
  
  if (!status&&!done) {
    status=restoreMDTNodes(MolRoot,s);
    if (!status&&(!streq("</MDT",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </MDT>");
      err_set(1,"%s",strm_buff(s));
      status=9004;
    }
  }

  if (!status&&ndone!=maxNTAB) {
    err_set(0,"number of nodes in MDT section (%d) does not match value of MDT nodes attribute (%d.",ndone,maxNTAB);
    /* should this be a fatal error? */
  }

  return(status);
}

/*========================================================================*/      

int initRestore(Stream *s)
{
  if (token_stream(s,"=,>")) {
    if (streq(strm_tok(s),"WSV")) {
      vers=str2i(token_stream(s,""));
    }else{
      while (!streq(strm_tok(s),"<WSV")&&!strm_eof(s)) token_stream(s,"=,>");
      if (streq(strm_tok(s),"<WSV")) {
	while (token_stream(s,"=,>")) {
	  if (streq(strm_tok(s),"version")) {
	    vers=str2i(token_stream(s,",>"));
	  }else{
	    err_set(1,"initRestore: invalid WSV attribute: %s",strm_tok(s));
	    vers=-1;
	  }
	}
      }else{
	err_set(1,"initRestore: no <WSV> element found in session file.");
	vers=-1;
      }
    }
  }else{
    vers=-2;
  }

  if (vers==1) initRestoreV1();
  else if (vers>1) {
    ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
    oldr=-1;
    obr=-1;
    oblabt=-1;
    oblw=-1;
    obcyl=-1;
    maxNTAB=0;
    maxMTAB=0;
  }    
  return(vers);
}

/*================================================================================*/

static int restoreMDLNode(Dspobj *root,Stream *s)
{
 int status=0,r=-1,dspstat=1,done=FALSE,seqnr;
  Dspobj *new=NULL;
  Mol *mol;

  if (!is_dsp(root)||root->dsptype!=DSP_GLBL) {
    err_set(1,"restoreMDLNode: attempted to link GLBL to non-GLBL");
    status=9000;
  }

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    else if (streq("dsp",strm_tok(s))) {
      if (strm_term(s)!='>') {
	dspstat=kwd_match(token_stream(s,">"),"off on",NULL);
	if (dspstat<0||dspstat>1) {
	  err_set(1,"restoreGlobalNode: invalid display status: %d",dspstat);
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else if (streq("m",strm_tok(s))) {
      if (strm_term(s)!='>') {
	seqnr=err_seqnr();
	r=str2i(token_stream(s,">"));
	if (seqnr<err_seqnr()) {
	  err_set(1,"restoreGlobalNode: format error");
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }

  if (!status) {
    if (r>=0&&r<ndone) {
      mol=(Mol *)NTAB[r];
      if (is_mol(mol)) {
	new=send_mdl(root,mol);
	new->root=(Node *)mol;
	if (new) {
	  set_dsp_status(new,dspstat);
	}else{
	  err_set(1,"failed to create MDL shadow structure");
	  status=9000;
	}
	mol->dsptr=root;
	root->root=(Node *)mol;
      }else{
	err_set(1,"bad pointer to MOL tree: %d",r);
	err_set(1,"%s",strm_buff(s));
	status=9004;
      }
    }else{
      err_set(1,"bad pointer to MOL tree: %d",r);
      status=9004;
    }
  }

  if (!status&&!done) {
    status=restoreMDTNodes((Node *)root,s);
    if (!status&&(!streq("</MDL",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </MDL>");
      err_set(1,"%s",strm_buff(s));
      status=9005;
    }
  }

  return(status);
}

static int restoreCSONode(Dspobj *root,char *subt,Stream *s)
{
  int i,status=0,dspstat=1,seqnr,lw,mds_type,color,n,size;
  float tpos[3],*x;
  char *expct;
  Dspobj *new=NULL;

  if (!is_dsp(root)||root->dsptype!=DSP_GLBL) {
    err_set(1,"restoreCSONode: attempted to link %s to non-GLBL",subt);
    status=9000;
  }else{
    new=new_dsp_obj(root,NULL,DSP_CSO);                  /*   always a CSO to the server... */
    if (!new) {
      err_set(1,"restoreCSONode: failed to add new display node");
      status=9001;
    }
  }

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else if (streq("dsp",strm_tok(s))) {
      if (strm_term(s)!='>') {
	dspstat=kwd_match(token_stream(s,">"),"off on",NULL);
	if (dspstat<0||dspstat>1) {
	  err_set(1,"restoreCSONode: invalid display status: %d",dspstat);
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }

  if (!status) {
    set_dsp_status(new,dspstat);
    if (streq(subt,"CNT")) new->dsptype=DSP_CNT;
    else if (streq(subt,"TRCE")) new->dsptype=DSP_TRCE;
    seqnr=err_seqnr();
    lw=str2i(token_stream(s,""));
    mds_type=str2i(token_stream(s,""));
    if (seqnr==err_seqnr()) {
      send_bheader(graphix,"SOLW",sizeof(new)+sizeof(int)); send_bint(graphix,lw); send_bptr(graphix,new);
      send_bheader(graphix,"MDST",sizeof(new)+sizeof(int)); send_bint(graphix,mds_type); send_bptr(graphix,new);
    }else{
      err_set(1,"CSO format error.");
      err_set(1,"%s",strm_buff(s));
      status=9000;
    }
    
    while (!status&&token_stream(s,">")) {
     if (strm_term(s)=='>') break;

      if (streq("L",strm_tok(s))) {
	color=str2i(token_stream(s,""));
	n=str2i(token_stream(s,""));
	if (seqnr<err_seqnr()) break;
	if (n>0) {
	  x=malloc(n*sizeof(float));
	  if (!x) {
	    err_set(1,"CSO: failed to allocate scratch array to hold temporary data.");
	    status=9011;
	    n=0;
	  }
	}else x=NULL;
	for (i=0;i<n;i++) {
	  x[i]=str2f(token_stream(s,""));
	  if (seqnr<err_seqnr()) {
	    err_set(1,"CSO format error.");
	    err_set(1,"%s",strm_buff(s));
	    status=9011;
	  }
	}
	if (!status) {
	  n=n/6;
	  if (n>0) {
	    size=sizeof(new)+sizeof(int)+6*n*sizeof(int);
	    send_bheader(graphix,"AMDS",size);
	    send_bptr(graphix,new);
	    send_bint(graphix,color);
	    send_bfloata(graphix,6*n,x);
	  }
	}
	if (x) free(x);
	x=NULL;

      }else if (streq("T",strm_tok(s))) {
	color=str2i(token_stream(s,""));
	for(i=0;i<3;i++) tpos[i]=str2f(token_stream(s,""));
	if (seqnr<err_seqnr()) {
	  err_set(1,"CSO: format error.");
	  err_set(1,"%s",strm_buff(s));
	  status=9011;
	}
	if (token_stream(s,"")&&!status) {
	  size=sizeof(new)+sizeof(int)+3*sizeof(int)+strlen(strm_tok(s));
	  send_bheader(graphix,"ATOB",size);
	  send_bptr(graphix,new);
	  send_bint(graphix,color);
	  send_bfloata(graphix,3,tpos);
	  send_bdata(graphix,strlen(strm_tok(s)),strm_tok(s));
	}
      }

    }
  }
  
  if (!status) {
    expct=new->dsptype==DSP_TRCE?"</TRCE":new->dsptype==DSP_CNT?"</CNT":"</CSO";
    if (!status&&(!streq(expct,strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected %s>",expct);
      err_set(1,"%s",strm_buff(s));
      status=9005;
    }
  }

  return(status);
}

static int restoreTRGNode(Dspobj *root,Stream *s)
{
  int status=0,dspstat=1;
  Surface *S;
  Dspobj *new=NULL;
  char *name=NULL;

  if (!is_dsp(root)||root->dsptype!=DSP_GLBL) {
    err_set(1,"restoreTRGNode: attempted to link TRG to non-GLBL");
    status=9000;
  }

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("n",strm_tok(s))) {
      name=Strdup(token_stream(s,">"));
    }else if (streq("dsp",strm_tok(s))) {
      if (strm_term(s)!='>') {
	dspstat=kwd_match(token_stream(s,">"),"off on",NULL);
	if (dspstat<0||dspstat>1) {
	  err_set(1,"restoreTRGNode: invalid display status: %d",dspstat);
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }

  if (!status) {
    strm_flush(s);
    if (s->type==STRM_PIPE) S=readSurface(((UCpipe *)s->in)->fd);
    else S=NULL;
    new=addSurfaceToDisplay(S,root,name);
    S=freeSurface(S);
    set_dsp_status(new,dspstat);
    set_node_name((Node *)new,name);
  }

  if (!status) {
    if (!streq("</TRG",token_stream(s,">"))||strm_term(s)!='>') {
      err_set(1,"badly nested XML elements, expected </TRG>");
      err_set(1,"%s",strm_buff(s));
      status=9005;
    }
  }

  if (name) free(name);

  return(status);
}

static int restoreGlobalNode(Dspobj *root,Stream *s)
{
  int status=0,r=-1,dspstat=1,i,j,done=FALSE,seqnr;
  Dspobj *new=NULL;
  Mol *mol;

  /* create and register new node */
  if (!root) {
    new=(Dspobj *)DspRoot;
    MTAB[mdone++]=(Node *)new;
  }else if (!is_dsp(root)||root->dsptype!=DSP_GLBL) {
    err_set(1,"restoreGlblNode: attempted to link GLBL to non-GLBL");
    status=9000;
  }else if (mdone<maxMTAB) {
    new=new_dsp_obj(root,NULL,DSP_GLBL);
    MTAB[mdone++]=(Node *)new;
    if (!new) {
      err_set(1,"restoreGlblNode: failed to add new display node");
      status=9001;
    }
  }else{
    err_set(1,"restoreGlblNode: node table overflow");
    err_set(1,"%s",strm_buff(s));
    status=9002;
  }    

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    else if (streq("n",strm_tok(s))) {
      if (strm_term(s)!='>') set_node_name((Node *)new,token_stream(s,">"));
    }else if (streq("dsp",strm_tok(s))) {
      if (strm_term(s)!='>') {
	dspstat=kwd_match(token_stream(s,">"),"off on",NULL);
	if (dspstat<0||dspstat>1) {
	  err_set(1,"restoreGlobalNode: invalid display status: %d",dspstat);
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else if (streq("m",strm_tok(s))) {
      if (strm_term(s)!='>') {
	seqnr=err_seqnr();
	r=str2i(token_stream(s,">"));
	if (seqnr<err_seqnr()) {
	  err_set(1,"restoreGlobalNode: format error");
	  err_set(1,"%s",strm_buff(s));
	  status=9003;
	}
      }
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  if (!status) {
    seqnr=err_seqnr();
    for (i=0;i<4;i++) {
      for (j=0;j<4&&seqnr==err_seqnr();j++) {
	new->mtrx.r[i].c[j]=str2d(token_stream(s,""));
      }
    }
    if (seqnr<err_seqnr()) {
      err_set(1,"restoreGlobalNode: format error in display matrix.");
      err_set(1,"%s",strm_buff(s));
      status=9004;
    }
  }

  if (!status) {
    if (r>=0&&r<ndone) {
      mol=(Mol *)NTAB[r];
      if (is_mol(mol)) {
	mol->dsptr=new;
	new->root=(Node *)mol;
      }else{
	err_set(1,"bad pointer to MOL tree");
	status=9103;
      }
    }else{
      err_set(1,"bad pointer to MOL tree: %d",r);
      status=9104;
    }
    new->status=dspstat;
    put_RTP(new,&new->mtrx);
  }

  if (!status&&!done) {
    status=restoreDDTNodes(new,s);
    if (!status&&(!streq("</GLBL",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </GLBL>");
      err_set(1,"%s",strm_buff(s));
      status=9009;
    }
  }

  return(status);
  
}

static int restoreDDTNodes(Dspobj *root,Stream *s)
{
  int status=0;

  while (!status&&!strm_eof(s)) {
    if (token_stream(s,">")) {
      if (streq("<GLBL",strm_tok(s)))      status=restoreGlobalNode(root,s);
      else if (streq("<MDL",strm_tok(s)))  status=restoreMDLNode(root,s);
      else if (streq("<CSO",strm_tok(s)))  status=restoreCSONode(root,"CSO",s);
      else if (streq("<CNT",strm_tok(s)))  status=restoreCSONode(root,"CNT",s);
      else if (streq("<TRCE",strm_tok(s))) status=restoreCSONode(root,"TRCE",s);
      else if (streq("<TRG",strm_tok(s)))  status=restoreTRGNode(root,s);
      else if (strneq("</",strm_tok(s),2)) break;
      else continue; /* just ignore */
    }
  }

  return(status);
}

int restoreDDT(Stream *s)
{
  int seqnr,status=0;

  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq(strm_tok(s),"nodes")) {
      seqnr=err_seqnr();
      maxMTAB=str2i(token_stream(s,">"));
      if (seqnr!=err_seqnr()||maxMTAB<0) {
	err_set(1,"restoreDDT: invalid DDT nodes attribute value: %s",strm_tok(s));
	status=9001;
      }
    }else{
      err_set(1,"restoreDDT: invalid DDT attribute: %s",strm_tok(s));
      status=9001;
      break;
    }
  }

  if (!status) {
    if (maxMTAB>0) {
      MTAB=malloc(maxMTAB*sizeof(Node *));
      if (!MTAB) {
	err_set(1,"failed to allocate node table of size %d in restore DDT",maxMTAB);
	status=9004;
      }
    }
  }

  if (!status) {
    status=restoreDDTNodes(NULL,s);
    if (!status&&(!streq("</DDT",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </DDT>");
      err_set(1,"%s",strm_buff(s));
      status=9004;
    }
  }

  if (!status) postDDT();

  return(status);
}

static void postDDT(void)
{
  Nodechain atoms;
  Atom *a;
  Bond *B,*bn;
  int i,j;

  /* deal with non-standard atom and bond attributes... */

  /* atoms */

  init_node_chain(&atoms);
  for (i=0;i<ndone;i++) {
    if (is_atom(NTAB[i])&&cPtr(((Atom *)NTAB[i])->cross)) add_to_node_chain(NTAB[i],&atoms);
  }
  send_atom_attributes(&atoms);
  clear_node_chain(&atoms);
    
  /* bonds */

  for (B=NULL,i=0;i<ndone;i++) {
    if (is_atom(NTAB[i])) {
      a=(Atom *)NTAB[i];
      for (j=0;j<a->nbnd;j++) {
	if (a==a->bnd[j]->t) {
	  a->bnd[j]->chain=B;
	  B=a->bnd[j];
	}
      }
    }
  }
  send_bond_attributes(B);
  for (;B;B=bn) {         /* clear bond chain  */
    bn=B->chain;
    B->chain=NULL;
  }
}

int restoreDots(Stream *s)
{
  int status=0,done=FALSE;

  /* parse element attributes */
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    /* no known attributes... */
  }

  if (!done) {
    while (!status&&token_stream(s,">")) {
      if (streq("<D",strm_tok(s)))      status=restoreAtomDots(s);
      else if (strneq("</",strm_tok(s),2)&&strm_term(s)=='>') break;
      else continue; /* just ignore */
    }
    if (!status&&(!streq("</DOTS",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </DOTS>");
      err_set(1,"%s",strm_buff(s));
      status=9001;
    }
  }

  return(status);
}

static int restoreAtomDots(Stream *s)
{
  Atom *a=NULL;
  int status=0,seqnr,n,color=-999,dsp=-999,style=-999,done=FALSE,nx=0;
  float *xyz=NULL,*new=NULL;

  /* parse element attributes */
  seqnr=err_seqnr();
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    if (streq("a",strm_tok(s))) {
      if (strm_term(s)!='>') {
	n=str2i(token_stream(s,">"));
	if (seqnr<err_seqnr()||n<0||n>maxNTAB||!is_atom(NTAB[n])) {
	  err_set(1,"restoreAtomDots: invalid atom number");
	  err_set(1,"%s",strm_buff(s));
	  status=9001;
	}else a=(Atom *)NTAB[n];
      }
    }else if (streq("col",strm_tok(s))) {
      if (strm_term(s)!='>') color=str2i(token_stream(s,">"));
    }else if (streq("dsp",strm_tok(s))) {
      if (strm_term(s)!='>') dsp=str2i(token_stream(s,">"));
    }else if (streq("style",strm_tok(s))) {
      if (strm_term(s)!='>') style=str2i(token_stream(s,">"));
    }else{ /* unknown attribute */
      if (strm_term(s)!='>') token_stream(s,">"); /* ignore value of unknown attribute */
    }
  }
  
  if (!done&&!status) {
    n=0; nx=0;
    while (!status&&token_stream(s,">")) {
      if (strneq("</",strm_tok(s),2)&&strm_term(s)=='>') break;
      if (n==nx) {
	new=realloc(xyz,(nx+CHUNK)*sizeof(float));
	if (new) {
	  xyz=new;
	  nx+=CHUNK;
	}else{
	  err_set(0,"restoreAtomDots: memory allocation error.");
	  status=9100;
	}
      }
      if (n<nx) xyz[n++]=str2f(strm_tok(s));
      if (seqnr<err_seqnr()) {
	err_set(1,"restoreAtomDots: format error.");
	err_set(1,"%s",strm_buff(s));
	status=9002;
      }
    }
    if (!status&&cPtr(a->cross)&&n/4>0) send_dots(a,n/4,xyz);
  }
  if (!status&&cPtr(a->cross)) {
    if (dsp!=-999) {
      send_bheader(graphix,"DOTD",cPtrSize()+sizeof(int));
      send_bint(graphix,dsp); send_bcptr(graphix,a->cross);
    }
    if (style!=-999) {
      send_bheader(graphix,"DOTS",cPtrSize()+sizeof(int));
      send_bint(graphix,style); send_bcptr(graphix,a->cross);
    }
    if (color!=-999) {
      send_bheader(graphix,"DOTC",cPtrSize()+sizeof(int));
      send_bint(graphix,color); send_bcptr(graphix,a->cross);
    }
  }

  if (!status&&(!streq("</D",strm_tok(s))||strm_term(s)!='>')) {
    err_set(1,"badly nested XML elements, expected </D>");
    err_set(1,"%s",strm_buff(s));
    status=9003;
  }

  if (xyz) free(xyz);
  return(status);
}

int restoreSRV(Stream *s)
{
  PrjTS Prj;
  int *ramp,i,l,nr,R,G,B,status=0,seqnr,done=FALSE,m,n;
  float lwm[] = { 1.5, 0.5, 10.0, 10.0 };
  float pmin,pmax;

  /* parse element attributes */
  seqnr=err_seqnr();
  while (!status&&strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    /* no known attributes... */
  }

  if (!done) {
    while (!status&&token_stream(s,">")) {
      if (strneq("</",strm_tok(s),2)&&strm_term(s)=='>') break;
      
      else if (streq(strm_tok(s),"P")) {
	Prj.viewmode=str2i(token_stream(s,""));
	Prj.stereomode=str2i(token_stream(s,""));
	Prj.ysize=str2f(token_stream(s,""));
	Prj.znear=str2f(token_stream(s,""));
	Prj.slabwidth=str2f(token_stream(s,""));
	Prj.viewdist=str2f(token_stream(s,""));
	Prj.eyesep=str2f(token_stream(s,""));
	Prj.aspect=str2f(token_stream(s,""));
	Prj.scrysize=str2f(token_stream(s,""));
	Prj.sep=str2f(token_stream(s,""));
	if (seqnr==err_seqnr()) {
	  setThreeDPrj(&Prj);
	}else{
	  err_set(1,"%s",strm_buff(s));
	  err_set(1,"restoreSRV: bad definition of projection");
	  status=9000;
	}

      }else if (streq(strm_tok(s),"A")) {
	if (token_stream(s,"")) setAttribute("SAAL",str2i(strm_tok(s)));  /* c.f. save.c for comments */
	if (token_stream(s,"")) setAttribute("SDPC",str2i(strm_tok(s)));
	if (token_stream(s,"")) setAttribute("SDTH",str2i(strm_tok(s)));
	if (token_stream(s,"")) setAttribute("SLTM",str2i(strm_tok(s)));

      }else if (streq(strm_tok(s),"L")) {
	if (token_stream(s,"")) lwm[0]=str2f(strm_tok(s));  /* c.f. save.c for comments */
	if (token_stream(s,"")) lwm[1]=str2f(strm_tok(s));
	if (token_stream(s,"")) lwm[2]=str2f(strm_tok(s));
	if (token_stream(s,"")) lwm[3]=str2f(strm_tok(s));
	if (seqnr==err_seqnr()) setAttributeFA("SLWM",lwm);

      }else if (streq(strm_tok(s),"CR")) {
	if ((l=str2i(token_stream(s,"")))>0) {
	  ramp=malloc(l*sizeof(int));
	  pmin=str2f(token_stream(s,""));
	  pmax=str2f(token_stream(s,""));
	  if (ramp) {
	    for (i=0;i<l;i++) ramp[i]=str2i(token_stream(s,""));
	    DefineColorRamp(l,ramp);
	    setPropertyMap(pmin,pmax);
	    free(ramp);
	  }else err_set(1,"error reading color ramp, ignored.");
	}

      }else if (streq(strm_tok(s),"C")) {
	nr=str2i(token_stream(s,""));
	R=str2i(token_stream(s,""));
	G=str2i(token_stream(s,""));
	B=str2i(token_stream(s,""));
	if (seqnr==err_seqnr()) {
	  if (nr>=-1&&nr<WNP_MAXCOL) {
	    DefineColor(nr,token_stream(s,""),R,G,B);
	  }else{
	    err_set(1,"%s",strm_buff(s));
	    err_set(1,"restoreSRV: bad color definition in session fille");
	    status=9000;
	  }
	}else{
	  err_set(1,"restoreSRV: bad definition of projection");
	  err_set(1,"%s",strm_buff(s));
	  status=9000;
	}

      }else if (streq(strm_tok(s),"SCYM")) {
	m=str2i(token_stream(s,""));
	n=str2i(token_stream(s,""));
	if (seqnr==err_seqnr()) {
	  set_cyl_mode(m,n);
	}else{
	  err_set(1,"%s",strm_buff(s));
	  err_set(1,"restoreSRV: bad cylinder definition in session fille");
	  status=9000;
	}

      }else if (streq(strm_tok(s),"SPHM")) {
	m=str2i(token_stream(s,""));
	n=str2i(token_stream(s,""));
	if (seqnr==err_seqnr()) {
	  set_sph_mode(m,n);
	}else{
	  err_set(1,"%s",strm_buff(s));
	  err_set(1,"restoreSRV: bad sphere definition in session fille");
	  status=9000;
	}

      }else if (strm_ltok(s)>1&&strm_tok(s)[0]=='<'&&isalpha(strm_tok(s)[1])) {
	ignoreXMLToken(s);
      }else{
	err_set(0,"restoreSRV: unknown token in session file");
	err_set(0,"%s",strm_buff(s));
	strm_flush(s);
	err_set(0,"");
      }
      if (!status&&seqnr<err_seqnr()) {
	err_set(1,"restoreSRV: format error.");
	err_set(1,"%s",strm_buff(s));
	status=9100;
      }
    }
    if (!status&&(!streq("</SRV",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </SRV>");
      err_set(1,"%s",strm_buff(s));
      status=9003;
    }
  }

  return(status);
}

int restoreDMON(Stream *s)
{
  int seqnr,status=0,done=FALSE;
  Atom *a[CHUNK];
  int i,k,n;

  while (strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq("/",strm_tok(s))&&strm_term(s)=='>') done=TRUE;
    if (streq(strm_tok(s),"status")) {
      seqnr=err_seqnr();
      n=str2i(token_stream(s,">"));
      if (seqnr!=err_seqnr()) {
	err_set(1,"restoreDDT: invalid DMON status attribute value: %s",strm_tok(s));
      }else{
	send_bheader(graphix,"MONS",sizeof(int));
	send_bint(graphix,n);
      }
    }else if (streq(strm_tok(s),"color")) {
      seqnr=err_seqnr();
      n=str2i(token_stream(s,">"));
      if (seqnr!=err_seqnr()) {
	err_set(1,"restoreDDT: invalid DMON color attribute value: %s",strm_tok(s));
      }else{
	send_bheader(graphix,"MONC",sizeof(int));
	send_bint(graphix,n);
      }
    }else if (streq(strm_tok(s),"label")) {
      seqnr=err_seqnr();
      n=str2i(token_stream(s,">"));
      if (seqnr!=err_seqnr()) {
	err_set(1,"restoreDDT: invalid DMON label attribute value: %s",strm_tok(s));
      }else{
	send_bheader(graphix,"MONL",sizeof(int));
	send_bint(graphix,n);
      }
    }else if (streq(strm_tok(s),"width")) {
      seqnr=err_seqnr();
      n=str2i(token_stream(s,",>"));
      if (seqnr!=err_seqnr()) {
	err_set(1,"restoreDDT: invalid DMON width attribute value: %s",strm_tok(s));
      }else{
	send_bheader(graphix,"MONW",sizeof(int));
	send_bint(graphix,n);
      }
    }      
  }

  if (!done) {
    n=0; seqnr=err_seqnr();
    while (!status&&token_stream(s,">")) {
      if (streq("</DMON",strm_tok(s))&&strm_term(s)=='>') break;
      k=str2i(strm_tok(s));
      if (seqnr<err_seqnr()||k<0||k>maxNTAB||!is_atom(NTAB[k])) {
	err_set(1,"restoreDMON: invalid atom number.");
	err_set(1,"%s",strm_buff(s));
	status=9001;
      }else{
	a[n++]=(Atom *)NTAB[k];
	if (n==CHUNK) {
	  send_bheader(graphix,"AMON",n*cPtrSize());
	  for (i=0;i<n;i++) send_bcptr(graphix,a[i]?a[i]->cross:cPtrNULL());
	  n=0;
	}
      }
    }
    if (n>0) {
      send_bheader(graphix,"AMON",n*cPtrSize()); 
      for (i=0;i<n;i++) send_bcptr(graphix,a[i]?a[i]->cross:cPtrNULL());
    }
    if (!status&&(!streq("</DMON",strm_tok(s))||strm_term(s)!='>')) {
      err_set(1,"badly nested XML elements, expected </DMON>");
      err_set(1,"%s",strm_buff(s));
      status=9003;
    }
  }

  return(status);
}

void freeRestoreData(void)
{
  if (vers==1) freeRestoreDataV1();
  else{
    if (NTAB) free(NTAB);
    NTAB=0;
    ndone=0;

    if (MTAB) free(MTAB);
    MTAB=0;
    mdone=0;
  }
}

/*========================================================================*/

int restoreMacros(Stream *s)
{
  wnpmacro_t *macros,*m;
  int status=0,xmlspcl;

  if (!s) return(9001);

  macros=NULL;
  while (strm_term(s)!='>'&&token_stream(s,"=>")) {
    if (streq(strm_tok(s),"type")) {
      if (token_stream(s,">")) {
	macros=getMacros(strm_tok(s));
	if (!macros) {
	  err_set(0,"could not find macro table for macro type %s",strm_tok(s));
	  err_set(0,"%s",strm_buff(s));
	}
      }
    }else err_set(0,"unknown MACROS attribute: %s",strm_tok(s));
  }
  if (!macros) err_set(0,"invalid or undefined macro type. %s",strm_buff(s));
  while (!strm_eof(s)&&!status) {
    if (token_stream(s,">")) {
      if (strneq(strm_tok(s),"</",2)&&strm_term(s)=='>') break;
      else if (streq(strm_tok(s),"<MACRO")) {
	m=NULL;
	while (strm_term(s)!='>'&&token_stream(s,"=>")) {
	  if (streq(strm_tok(s),"name")) {
	    if (token_stream(s,">")&&!m&&macros) {
	      m=new_macro();
	      if (m) {
		set_macro_name(m,strm_tok(s));
		insert_macro(m,macros);
	      }else err_set(0,"failed to allocate data structure for macro %s",strm_tok(s));
	    }
	  }else err_set(0,"unknown MACRO attribute: %s",strm_tok(s));
	}

	set_stream(s,"eol=yes");
	if (token_stream(s,">")) xmlspcl=!streq("<![CDATA[",strm_tok(s));
	else xmlspcl=TRUE;

	do {
	  strm_flush(s); 
	  if (token_stream(s,">")&&xmlspcl&&streq("</MACRO",strm_tok(s))&&strm_term(s)=='>') break;
	  if (streq("<![CDATA[",strm_tok(s))) xmlspcl=FALSE;
	  else if (streq("]]",strm_tok(s))&&strm_term(s)=='>') xmlspcl=TRUE;
	  else{
	    if (xmlspcl) xmlspclunquote(strm_buff(s));
	    add_macro_line(m,strm_buff(s));
	  }
	} while (!strm_eof(s));
	set_stream(s,"eol=no");

	if (!streq("</MACRO",strm_tok(s))||strm_term(s)!='>') {
	  err_set(1,"badly nested XML elements, expected </MACRO>");
	  err_set(1,"%s",strm_buff(s));
	  status=9004;
	}
      }else status=ignoreXMLToken(s);
    }
  }

  if (!status&&(!streq("</MACROS",strm_tok(s))||strm_term(s)!='>')) {
    err_set(1,"badly nested XML elements, expected </MACROS>");
    err_set(1,"%s",strm_buff(s));
    status=9005;
  }

  return(status);
}

static void xmlspclunquote(char *line)
{
  char *p,*q,*s;

  if (line) {
    for (q=line,p=line;*p;p++) {
      if (*p=='&') {
	if ((s=strchr(p,';'))) {
	  if (strneq(p,"&amp;",5))        *q++='&';
	  else if (strneq(p,"&lt;",4))    *q++='<';
	  else if (strneq(p,"&gt;",4))    *q++='>';
	  else if (strneq(p,"&apos;",6))  *q++='\'';
	  else if (strneq(p,"&quot;",6))  *q++='"';
	  else while(p<=s) *q++=*p++;
	  p=s;
	}
      }else *q++=*p;
    }
    return;
  }
}

int ignoreXMLToken(Stream *s)
{
  char *tok=NULL;
  int l,done=FALSE,status=0;
  
  /* Note: should deal properly with <![CDATA[ and ]]> tags. */

  if (strneq("<!--",strm_tok(s),4)) {     /* comment */
    strm_pos(s)=strm_parse(s)->tks-strm_buff(s)+4;
    while(!status&&!strm_eof(s)&&token_stream(s,">")) {
      l=strm_ltok(s);
      if (l>1 && strm_parse(s)->tks[l]=='>' && strm_tok(s)[l-1]=='-' && strm_tok(s)[l-2]=='-') break;
      if (l>=4 && strneq("<!--",strm_tok(s),4)) status=ignoreXMLToken(s); /* XML doe not allow this */
    }
    if (strm_eof(s)) {
      err_set(0,"ignoreXMLToken: encountered EOF in comment element");
      status=1;
    }

  }else if (strneq("<?",strm_tok(s),2)) { /* <? ... ?> */
    strm_pos(s)=strm_parse(s)->tks-strm_buff(s)+2;
    while(!status&&!strm_eof(s)&&token_stream(s,">")) {
      l=strm_ltok(s);
      if (l>0 && strm_parse(s)->tks[l]=='>' && strm_tok(s)[l-1]=='?') break;
    }
    if (strm_eof(s)) {
      err_set(0,"ignoreXMLToken: encountered EOF in <?.. element");
      status=1;
    }
    
  }else if (strm_tok(s)[0]=='<'&&(isalpha(strm_tok(s)[1])||strm_tok(s)[1]=='_')) { /* elements */
    strm_pos(s)=strm_parse(s)->tks-strm_buff(s);
    token_stream(s,">"); tok=strdup(strm_tok(s)+1); 
    if (strrchr(tok,'/')) {
      done=TRUE;
      *strrchr(tok,'/')='\0';
    }
    while( !done && !status && !strm_eof(s) &&strm_term(s)!='>' && token_stream(s,">")) {
      if (streq(strm_tok(s),"/")&&strm_term(s)=='>') done=TRUE;
    }
    if (!done) {
      while(!status&&!strm_eof(s)&&token_stream(s,">")) {
	if (strneq("</",strm_tok(s),2)&&strm_term(s)=='>') {
	  if (streq(tok,strm_tok(s)+2)) break;
	  else {
	    err_set(0,"badly nested XML elements, expected </%s>",tok);
	    err_set(0,"%s",strm_buff(s));
	    status=1;
	  }
	}else ignoreXMLToken(s);
      }
    }

  }else{ /* everything else: simply ignored (should not do this for <![CDATA[ and ]]>) */
  }

  if (tok) free(tok);
  return(status);
}
