#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "xtal.h"
#include "surface.h"
#include "streamio.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Node *DspRoot;
extern Node *LibRoot;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))

#define ABEND(x) {err_set(1,x); new=del_node(new); return(new);}
#define chomp(x) if (strrchr(x,'\n')) *strrchr(x,'\n')='\0'


#define TRUE 1
#define FALSE 0

/* nodes table for MDT */
static Node **NTAB=NULL;
static int ndone=0;

/* nodes table for DDT */
static Node **MTAB=NULL;
static int mdone=0;

static float oldr;
static float obr=-1;
static int oblabt=-1;
static int oblw=-1;
static int obcyl=-1;

/*-------------------- OBJ_MOL --------------------*/

static Node *restoreMolNode(UCpipe *in)
{
  Node *new;
  Mol *m;
  Parse p;
  char buff[8192],fld[1024];
  int i,ns,nc;

  new=make_node(OBJ_MOL);
  if (new) {
    p.buf=buff;
    p.tok=fld;
    p.sep="";
    p.mts=sizeof(fld);
    m=(Mol *)new;
    if (fgets(buff,sizeof(buff),in->fd)) {
      p.pos=0;
      m->moltype=str2i(token(&p));
      m->keep=str2i(token(&p));
      ns=str2i(token(&p)); if (ns>ATOMSET_WORDS) ns=ATOMSET_WORDS;
      for (i=0;i<ns;i++) {
	if (token(&p)) {
	  sscanf(fld,"%lx%n",m->used+i,&nc);
	  if (nc!=strlen(fld)) {
	    err_set(1,"bad atomset word in moldata: %s",fld);
	    ABEND("format error, restore aborted");
	  }
	}
      }
      for (i=0;i<ns;i++) {
	if (token(&p)) {
	  sscanf(fld,"%lx%n",m->local+i,&nc);
	  if (nc!=strlen(fld)) {
	    err_set(1,"bad atomset word in moldata: %s",fld);
	    ABEND("format error, restore aborted");
	  }
	}
      }
    }else ABEND("failed to read molecule data");
  }else err_set(1,"failed to create mol node.");
  return(new);
}
/*-------------------- OBJ_NOELIST --------------------*/

static Node *restoreNOEList(UCpipe *in)
{
  Node *new;
  NOElist *n;
  char buff[8192],fld[1024];
  int status;
  Parse p;
  Stream *s;

  new=make_node(OBJ_NOELIST);
  if (new) {
    p.buf=buff;
    p.tok=fld;
    p.sep="";
    p.mts=sizeof(fld);
    n=(NOElist *)new;
    if (fgets(buff,sizeof(buff),in->fd)) {
      p.pos=1;
      n->avtype=str2i(token(&p));
      if (n->avtype==0) {
	err_set(1,"invalid AV type in noelist:");
	err_set(1,"%s",p.buf);
      }
      s=open_stream(NULL,8192);
      set_stream_pipe(s,in);
      s->p->tok=fld;
      s->p->mts=sizeof(fld);
      status=read_xplor_noe_stream(s,n);
      s=close_stream(s);
      if (status) err_set(1,"restoreNOEList: failed.");
    }
  }else err_set(1,"failed to create NOE list node.");
  return(new);
}

/*-------------------- OBJ_ATOM --------------------*/

static Node *restoreAtomNode(UCpipe *in)
{
  Node *ResLib,*new;
  Atom *a;
  Bond *b;
  Atomdef *atp;
  Parse p;
  char buff[8192],fld[1024];
  int iup,nr,havexyz=1,seqnr,nc,n,off;
  unsigned long bits;

  p.buf=buff;
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  new=make_node(OBJ_ATOM);
  if (new) {

    a=(Atom *)new;
    seqnr=err_seqnr();

    /* 1st record */
    if (fgets(buff,sizeof(buff),in->fd)) {
      p.pos=0;
      for (p.pos=0,nr=0;token(&p);nr++) {
	switch (nr) {

	case 0:   /* FF type mnemonic */
	  atp=find_atomdef_name(ResLib,fld);
	  if (!atp) {
	    err_set(1,"could not find atom type %s in current library, guessing...",fld);
	    atp=guess_atom_type(fld);
	    if (atp) {
	      atp=(Atomdef *)copy_node((Node *)atp);
	      if (atp) {
		set_node_name((Node *)atp,fld);
		atp->up=ResLib;
		link_node((Node *)atp);
	      }
	    }
	  }
	  if (!atp) ABEND("failed, session restore aborted");
	  set_atom_type(a,atp);
	  break;
	  
	case 1:   /* color */
	  a->color=str2i(fld);
	  break;

	case 2:   /* x */
	case 3:   /* y */
	case 4:   /* z */
	  if (!streq(fld,".")) a->x[nr-2]=str2f(fld);
	  else havexyz=0;
	  break;

	case 5:   /* part. charge q */
	  if (!streq(fld,".")) a->q=str2f(fld);
	  break;

	case 6:   /* isotropic B factor */
	  if (!streq(fld,".")) a->x[ATM_BISO]=str2f(fld);
	  break;

	case 7:   /* site occupancy */
	  if (!streq(fld,".")) a->x[ATM_SOF]=str2f(fld);
	  break;

	case 8:   /* chemical shift */
	  if (!streq(fld,".")) a->shift=str2f(fld);
	  break;

	default:
	  ABEND("extra data in first line of atom data.");
	}
      }
      if (havexyz&&nr>4) a->nxyz=1;
      else a->nxyz=0;
      
      if (nr<2) ABEND("data in first line of atom data");
      if (err_seqnr()>seqnr) ABEND("format error, restore aborted");

    } else ABEND("failed to read atom data");

    /* 2nd record: radius, label color, label type, label text */
    if (fgets(buff,sizeof(buff),in->fd)) {
      a->r=oldr;
      for (p.pos=0,nr=0;token(&p);nr++) {
	switch (nr) {
	case 0:  /* radius */
	  if (streq(p.tok,".")) a->r=oldr;
	  else if (streq(p.tok,"v")) a->r=(a->atp?a->atp->rvdw:0.1);
	  else {
	    a->r=str2f(p.tok);
	    oldr=a->r;
	  }
	  break;

	case 1:  /* label color */
	  if (streq(p.tok,".")) a->lcolor=WNP_COLOR_BYATOM;
	  else a->lcolor=str2i(p.tok);
	  break;

	case 2:  /* labeltype */
	  a->ltype=str2i(p.tok);
	  if (a->ltype==ATM_LAB_TEXT) {
	    if (a->label) free(a->label);
	    if (strchr(p.buf,'\n')) *strchr(p.buf,'\n')='\0';
	    a->label=Strdup(p.buf+p.pos);
	    p.pos=strlen(p.buf);
	  }
	  break;

	default:
	  break;
	}
      }
    }

    /* 3rd record  */
    if (fgets(buff,sizeof(buff),in->fd)) {
      for (p.pos=0,nr=0;token(&p);nr++) {
	switch (nr) {
	case 0:  /* ligance */
	  if (strlen(fld)==1&&(fld[0]=='?'||fld[0]=='.')) a->lig='?';
	  else a->lig=str2i(fld);
	  break;

	case 1:  /* formal charge */
	  if (strlen(fld)==1&&(fld[0]=='?'||fld[0]=='.')) a->chrg='?';
	  else a->chrg=str2i(fld);
	  break;

	case 2:  /* bits packed into integer */
	  sscanf(fld,"%lx%n",&bits,&nc);
	  if (nc!=strlen(fld)) {
	    err_set(1,"bad bits: %s",fld);
	    ABEND("format error, restore aborted");
	  }
	  a->dsp=bits%2;    bits/=2;
	  a->ldsp=bits%2;   bits/=2;
	  a->sdsp=bits%2;   bits/=2;
	  a->arom=bits%2;   bits/=2;
	  a->oop=bits%4;    bits/=4;
	  a->cosw=bits%2;   bits/=2;
	  a->stass=bits%4;  bits/=4;
	  a->fe=bits%2;     bits/=2;
	  break;

	default: /* atomset words */
	  if (nr-3<ATOMSET_WORDS) {
	    sscanf(fld,"%lx%n",a->set+(nr-3),&nc);
	    if (nc!=strlen(fld)) err_set(1,"bad atomset word: %s",fld);
	  }else ABEND("too much data in 2nd line of atom data");
	}
      }
      if (err_seqnr()>seqnr) ABEND("format error, restore aborted");

    }else ABEND("failed to read atom data");

    /* 4th record: charge group and bond list  */
    if (fgets(buff,sizeof(buff),in->fd)) {
      p.pos=0;
      if (get_token(&p)) {
	if (fld[0]!='.') StrCpy(a->grp,fld,sizeof(a->grp));
	else StrCpy(a->grp,"??",sizeof(a->grp));
      }
      p.sep=",";
      while (get_token(&p)) {
	iup=str2i(fld);
	if (iup>0&&iup<ndone&&is_atom(NTAB[iup])) {
	  b=add_bond_ng((Atom *)NTAB[iup],a);
	  b->cyl=obcyl; b->lw=oblw; b->r=obr; b->labt=oblabt;
	  set_bt(b,1);
	  for (n=0;p.term==',';n++) {
	    if (!get_token(&p)) {
	      err_set(1,"%s\n",buff);
	      ABEND("atom data has invalid bond list");
	    }
	    switch (n) {
	    case 0:
	      if (p.tok[0]!='\0')  set_bt(b,BT_number(p.tok));
	      break;
	    case 1:
	      if (p.tok[0]!='\0') {
		oblw=str2i(p.tok);
		b->lw=oblw;
	      }
	      break;
	    case 2:
	      if (p.tok[0]=='Y'||p.tok[0]=='N') {
		obcyl=p.tok[0]=='Y'?1:0;
		b->cyl=obcyl;
		off=1;
	      }else off=0;
	      if (p.tok[off]!='\0') {
		obr=str2f(p.tok+off);
		b->r=obr;
	      }
	      break;
	    case 3:
	      if (p.tok[0]!='\0') {
		if (p.quoted) {
		  b->labt=BND_LAB_TEXT;
		  b->label=strdup(p.tok);
		}else{
		  oblabt=str2i(p.tok);
		  b->labt=oblabt;
		}
	      }
	      break;
	    default:
	      err_set(1,"%s\n",buff);
	      ABEND("atom data has invalid bond list");
	    }
	  }
	}else ABEND("invalid node number in bond list");
      }
    }else ABEND("failed to read atom data");

  }else {
    ABEND("failed to create atom node.");
  }
  return(new);
}

/*-------------------- OBJ_ATOMSET --------------------*/

static Node *restoreAtomsetNode(UCpipe *in)
{
  Node *new;
  Atomset *s;
  char buff[8192];
  int n;

  new=make_node(OBJ_ATOMSET);
  if (new) {
    s=(Atomset *)new;
    if (fgets(buff,sizeof(buff),in->fd)) {
      n=sscanf(buff,"%d",&(s->bit));
      if (n<1) ABEND("invalid atom set definition");
      if (s->bit<0||s->bit>=ATOMSET_BITS) ABEND("invalid atom set definition");
    }else ABEND("invalid atomset definition");
  }else ABEND("failed to create atomset node.");
  return(new);
}

/*-------------------- OBJ_XCN --------------------*/

static Node *restoreXCNode(UCpipe *in)
{
  Node *new;
  Xcn *xcn;
  char buff[8192],fld[1024];
  Parse p;
  hisrec *h,*nh;
  int i,j,have_cell=0,have_xtoa=0,seqnr,old_err,loop=TRUE,dupl;
  float err;
  Symop sym_n;
  mat_4 X2A;

  p.buf=buff;
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);
  
  new=make_node(OBJ_XCN);
  if (new) {
    xcn=(Xcn *)new;
    seqnr=err_seqnr();
    while (loop&&fgets(buff,sizeof(buff),in->fd)) {
      chomp(buff); p.pos=0;
      if (get_token(&p)) {

	if (streq(p.tok,"T")) {
	  if (mvld(xcn->title)) free(xcn->title);
	  xcn->title=NULL;
	  if (p.buf[p.pos]) xcn->title=Strdup(p.buf+p.pos);

	}else if (streq(p.tok,"L")) {
	  xcn->lambda=str2f(token(&p));

	}else if (streq(p.tok,"C")) {
	  for (i=0;i<3;i++) xcn->cl[i]=str2f(token(&p));
	  for (i=0;i<3;i++) xcn->ca[i]=str2f(token(&p));
	  xcn->fta=fract_to_angstrom(xcn->cl,xcn->ca); 
	  xcn->atf=inv_4(xcn->fta);
	  have_cell=1;

	}else if (streq(p.tok,"c")) {
	  for (i=0;i<3;i++) xcn->esd_cl[i]=str2f(token(&p));
	  for (i=0;i<3;i++) xcn->esd_ca[i]=str2f(token(&p));

	}else if (streq(p.tok,"F")) {
	  xcn->forml=0;
	  while (get_token(&p)) {
	    if (p.term=='\0') break;   /* ?? needed ?? */
	    if (p.ltok<1 || p.ltok>2) ABEND("bad formula record in XCN.");
	    if (xcn->forml>=MX_XCN_FORMUL) ABEND("too many items in formula in XCN");
	    Strcpy(xcn->forms[xcn->forml],p.tok);
	    xcn->formn[xcn->forml]=1;
	    if (strchr("0123456789",p.buf[p.pos])) {
	      xcn->formn[xcn->forml]=str2i(token(&p));
	      if (xcn->formn[xcn->forml]<1) ABEND("element count in formula in XCN is not a number");
	    }
	    xcn->forml++;
	  }

	}else if (streq(p.tok,"S")) {
	  xcn->lattice=str2i(token(&p));
	  xcn->centric=(xcn->lattice > 0 ? TRUE : FALSE);
	  if (!xcn->centric) xcn->lattice = -xcn->lattice;
	  if (xcn->lattice<1 || xcn->lattice>7) ABEND("bad lattice parameter in spgrp record.");
	  if (mvld(xcn->spgr)) free(xcn->spgr);
	  xcn->spgr=Strdup(p.buf+p.pos);
	  if (!xcn->spgr) ABEND("memory allocation failure.");

	}else if (streq(p.tok,"s")) {
	  sym_n=str2sym(p.buf+p.pos);
	  for (dupl=FALSE,i=0; !dupl&&i<xcn->nsym; i++) dupl=symopequiv(sym_n,xcn->sym[i]);
	  if (!dupl) {
	    if (xcn->nsym<MX_XCN_SYMOPS) {
	      xcn->sym[xcn->nsym]=sym_n;
	      xcn->nsym++;
	    }else ABEND("too many symmetry operations.");
	  }
	}else if (streq(p.tok,"A")) {
	  for (i=0;i<6;i++) xcn->asym[i]=str2f(token(&p));

	}else if (streq(p.tok,"scale")) {
	  xcn->scale=Strdup(p.buf+p.pos);
	  if (mvld(xcn->scale)) free(xcn->scale);
	  xcn->scale=Strdup(p.buf+p.pos);
    
	}else if (streq(p.tok,"H")) {
	  for (h=xcn->history;h;h=h->nxt) if (!h->nxt) break;
	  nh=make_history();
	  if (nh) {
	    nh->t=Strdup(p.buf+p.pos);
	    if (h) h->nxt=nh;
	    else xcn->history=nh;
	  }else ABEND("memory allocation failure.");

	}else if (streq(p.tok,"M")) {
	  if (get_token(&p)) {
	    if (streq("XtoA",p.tok)) {  /* definition of transformation from cryst. to A-cart */
	      have_xtoa=1+have_cell;
	      if (have_cell) {
		for (i=0;i<3;i++) for (j=0;j<4;j++) X2A.r[i].c[j]=str2f(token(&p));
		for (j=0;j<3;j++) X2A.r[3].c[j]=0.0;
		X2A.r[3].c[3]=1.0;
		xcn->fta=X2A;
		old_err=err_seqnr();
		xcn->atf=inv_4(xcn->fta);
		if (old_err!=err_seqnr()) ABEND("singular cell to angstrom matrix.");
	      }
	    }
	  }
	  loop=FALSE; /* this is always the last record */
	}
      }
    }
  }else ABEND("failed to rstore XCN");
  if (err_seqnr()>seqnr) ABEND("format error, restore aborted");

  /* expand symmetry operators */
  i=symopexp(xcn);
  if (i!=xcn->nsymx) {
    err_set(1002,"rstore XCN: more than %d expanded symops",MX_XCN_SYMOPS);
    xcn->nsymx=xcn->nsym;
  }

  /* convert symmetry operators to angstrom/cartesian */
  for (i=0;i<xcn->nsymx;i++) xcn->sym[i]=symA(xcn,xcn->sym[i]);
  
  if (have_xtoa==1) err_set(1,"restore XCN: XtoA, but no cell, XtoA ignored");

  err=check_fta_matrix(xcn->fta,xcn->cl,xcn->ca);
  if (err>0.1) err_set(1,"restore XCN: XCN cell and XtoA inconsistent, error= %0.3f A.",err);

  return(new);
}

/*========================================================================*/      

static int restoreGLBLNode(UCpipe *in,Dspobj *D,int root,int dspstat)
{
  char buff[8192];
  int i,n,status=0;
  Mol *mol;
  
  for (i=0;i<4;i++) {
    if (fgets(buff,sizeof(buff),in->fd)) {
    }else{
      err_set(1,"EOF in matrix of OBJ_GLB");
      status=9101;
      break;
    }
    n=sscanf(buff,"%f %f %f %f",D->mtrx.r[i].c,D->mtrx.r[i].c+1,D->mtrx.r[i].c+2,D->mtrx.r[i].c+3);
    if (n!=4) {
      err_set(1,"bad format in matrix of OBJ_GLB");
      status=9102;
    }
  }
  if (!status) {
    if (root>=0&&root<ndone) {
      mol=(Mol *)NTAB[root];
      if (is_mol(mol)) {
	mol->dsptr=D;
	D->root=(Node *)mol;
      }else{
	err_set(1,"bad pointer to MOL tree");
	status=9103;
      }
    }else{
      err_set(1,"bad pointer to MOL tree");
      status=9104;
    }
    D->status=dspstat;
    put_RTP(D,&D->mtrx);
  }
  return(status);
}
      
/*========================================================================*/      

int initRestoreV1(void)
{
  oldr=-1;
  obr=-1;
  oblabt=-1;
  oblw=-1;
  obcyl=-1;
  ndone=0;
  mdone=0;

  return(0);
}

int restoreMDTV1(Stream *s)
{
  Node *new=NULL;
  char buff[8192],fld[1024];
  int iup,nc,n,N,status=0;
  UCpipe *in;

  n=sscanf(strm_buff(s)+strm_pos(s),"%d",&N);
  strm_flush(s);
  if (n!=1) {
    err_set(1,"incorrect session file format");
    return(2);
  }
  in=(UCpipe *)s->in;

  if (N<0) {
    err_set(1,"invalid MDT tag in session file.");
    return(3);
  }

  if (N==0) return(0);   /* it was easy ! */

  
  NTAB=malloc(N*sizeof(Node *));
  if (!NTAB) {
    err_set(1,"failed to allocate node table of size %d in restore MDT",N);
    return(4);
  }
  
  NTAB[0]=MolRoot; ndone=1; 
  while (fgets(buff,sizeof(buff),in->fd)) {
    chomp(buff);
    if (*buff=='-') break;
    n=sscanf(buff,"%s %d%n",fld,&iup,&nc);
    if (n<2||iup<0||iup>=ndone) {
      err_set(1,"invalid Node tag found in session file:");
      err_set(1,"%s",buff);
      break;
    }
    switch (fld[0]) {
    case 'M':
      new=restoreMolNode(in);
      break;
    case 'N':
      new=restoreNOEList(in);
      break;
    case 'A':
      new=restoreAtomNode(in);
      break;
    case 'S':
      new=restoreAtomsetNode(in);
      break;
    case 'x':   /* Xatom node, not saved */
      break;
    case 'X':
      new=restoreXCNode(in);
      break;
    default:
      err_set(1,"invalid Node tag found in session file:");
      err_set(1,"%s",buff);
      new=NULL;
      break;
    }
    if (new) {
      set_node_name(new,buff+nc+1);
      new->up=NTAB[iup];
      link_node(new);
      NTAB[ndone]=new;
      ndone++;
    }else break;
    
  }

  if (ferror(in->fd)) {
    err_set(1,"I/O error while reading restore file.");
    status=901;
  }

  return(status);
}

int restoreDDTV1(Stream *s)
{
  Nodechain c;
  Node *lst;
  Mol *mol;
  Atom *a;
  Bond *B,*bs,*bn;
  Surface *S;
  Dspobj *D,*C,*global,*mdl;
  char buff[8192],fld[8192],*name;
  int i,j,color,size,root,iup,n,N,status=0,dspstat,mds_type,lw;
  float tpos[3],*x;
  Parse p;
  UCpipe *in;

  n=sscanf(strm_buff(s)+strm_pos(s),"%d",&N);
  strm_flush(s);
  if (n!=1) {
    err_set(1,"incorrect session file format");
    return(2);
  }
  in=(UCpipe *)s->in;

  if (N<0) {
    err_set(1,"invalid DDT tag in session file.");
    return(3);
  }

  if (N==0) return(0);   /* it was easy ! */

  MTAB=malloc(N*sizeof(Node *));
  if (!MTAB) {
    err_set(1,"failed to allocate node table of size %d in restore DDT",N);
    return(4);
  }
  
  mdone=0; 
  while (status==0&&fgets(buff,sizeof(buff),in->fd)) {
    chomp(buff);
    if (*buff=='-') break;
  
    n=sscanf(buff,"%*s %d %d %d %s",&iup,&dspstat,&root,fld);
    if (n<3||iup>=mdone) {
      err_set(1,"invalid Display tag found in session file:");
      err_set(1,"%s",buff);
      status=9000;
      break;
    }
    name=(n>=4?fld:NULL);
    
    if (strneq(buff,"ROOT",4)||strneq(buff,"GLBL",4)) {
      if (iup>=0) {
	if (iup<mdone) D=new_dsp_obj((Dspobj *)MTAB[iup],name,DSP_GLBL);
	else{
	  err_set(1,"invalid parent node pointer in DDT");
	  status=9001;
	  break;
	}
      }else D=(Dspobj *)DspRoot;
      status=restoreGLBLNode(in,D,root,dspstat);
      MTAB[mdone++]=(Node *)D;


    }else if (strneq(buff,"MDL",3)) {
      if (iup>=0&&iup<mdone) {
	global=(Dspobj *)MTAB[iup];
	if (root>=0&&root<ndone) {
	  mol=(Mol *)NTAB[root];
	  mdl=send_mdl(global,mol);
	  if (mdl) {
	    set_dsp_status(mdl,dspstat);
	  }else{
	    err_set(1,"failed to create MDL shadow structure");
	    status=9000;
	  }
	  mol->dsptr=global;
	  global->root=(Node *)mol;
	}else mdl=NULL;
      }else{
	err_set(1,"invalid parent node pointer in DDT");
	status=9001;
      }
     

    }else if (strneq(buff,"CSO",3)||strneq(buff,"CNT",3)||strneq(buff,"TRCE",4)) {
      if (iup>=0&&iup<mdone) {
	global=(Dspobj *)MTAB[iup];
	C=new_dsp_obj(global,name,DSP_CSO);                      /*   always a CSO to the server... */
	if (C) {
	  set_dsp_status(C,dspstat);
	  p.buf=buff;
	  p.tok=fld;
	  p.sep="";
	  p.mts=sizeof(fld);
	  if (strneq(buff,"CNT",3)) C->dsptype=DSP_CNT;          /* ..but something more specific to me */
	  else if (strneq(buff,"TRCE",4)) C->dsptype=DSP_TRCE;
	  if (fgets(buff,sizeof(buff),in->fd)) {
	    p.pos=0;
	    lw=str2i(token(&p));
	    send_bheader(graphix,"SOLW",sizeof(C)+sizeof(int));
	    send_bint(graphix,lw);
	    send_bptr(graphix,C);
	    mds_type=str2i(token(&p));
	    send_bheader(graphix,"MDST",sizeof(C)+sizeof(int));
	    send_bint(graphix,mds_type);
	    send_bptr(graphix,C);

	    while (!status&&fgets(buff,sizeof(buff),in->fd)) {
	      p.pos=0;
	      if (streq("---",token(&p))) break;     /* end of list of graphics elements in this CSO */

	      if (streq("L",p.tok)) {                /* move-draw sequence */
		color=str2i(token(&p));
		n=str2i(token(&p));
		x=(float *)malloc(n*sizeof(float));
		if (x) {
		  for (i=0;i<n;i++) {
		    if (i%6==0) {
		      if (!fgets(buff,sizeof(buff),in->fd)) break;
		      p.pos=0;
		    }
		    x[i]=str2f(token(&p));
		  }
		  if (i<n) {
		    err_set(1,"restoreDDT: failed to read vector data, expected %s, but got %d elements",n,i);
		    status=9000;
		  }else{
		    n=n/6;     /* each vector has two points, each point has three coordinates */
		    if (n>0) {
		      size=sizeof(C)+sizeof(int)+6*n*sizeof(int);
		      send_bheader(graphix,"AMDS",size);
		      send_bptr(graphix,C);
		      send_bint(graphix,color);
		      send_bfloata(graphix,6*n,x);
		    }
		  }
		  free(x); x=NULL;
		}else{
		  err_set(1,"restoreDDT: failed to allocate memory to hold %d vectors",n/6);
		  status=9000;
		}
	      }else if (streq("T",p.tok)) {          /* text object */
		color=str2i(token(&p));
		tpos[0]=str2f(token(&p)); tpos[1]=str2f(token(&p)); tpos[2]=str2f(token(&p));
		if (token(&p)) {
		  size=sizeof(C)+sizeof(int)+3*sizeof(int)+strlen(p.tok);
		  send_bheader(graphix,"ATOB",size);
		  send_bptr(graphix,C);
		  send_bint(graphix,color);
		  send_bfloata(graphix,3,tpos);
		  send_bdata(graphix,strlen(p.tok),p.tok);    /* w/o '\0' for end of string */
		}

	      }else{
		err_set(1,"restoreDDT: bad list element in CSO data");
		status=9000;
		break;
	      }
	    }

	  }else{
	    err_set(1,"restoreDDT: invalid CSO display node data");
	    status=9000;
	  }
	}else{
	  err_set(1,"restoreDDT: failed to create CSO display node");
	  status=9000;
	}

      }else{
	err_set(1,"invalid parent node pointer in DDT");
	status=9000;
      }

    }else if (strneq(buff,"TRG",3)) {
      if (iup>=0&&iup<mdone) {
	global=(Dspobj *)MTAB[iup];
	S=readSurface(in->fd);
	lst=global->lst;
	if (S) {
	  addSurfaceToDisplay(S,global,name);
	  S=freeSurface(S);
	} /* simply ignore, if this failed... */
	C=(Dspobj *)global->lst;
	if ((Node *)C!=lst) set_dsp_status(C,dspstat);
      }else{
	err_set(1,"invalid parent node pointer in DDT");
	status=9001;
      }


    }else if (strneq(buff,"HHDIST",6)) {
    
    }else{  
      err_set(1,"%s",buff);
      err_set(1,"bad node type, not restored");
      status=9002;
    }
  }

  /* deal with non-standard atom and bond attributes... */

  /* atoms */

  init_node_chain(&c);
  for (i=0;i<ndone;i++) {
    if (is_atom(NTAB[i])&&cPtr(((Atom *)NTAB[i])->cross)) {
      a=(Atom *)NTAB[i];
      if (!a->ldsp) {
	send_bheader(graphix,"ALST",cPtrSize()+sizeof(int));
	send_bint(graphix,0);
	send_bcptr(graphix,a->cross);
      }
      if (a->lcolor!=WNP_COLOR_BYATOM) {
	send_bheader(graphix,"ALCL",cPtrSize()+sizeof(int));
	send_bint(graphix,a->lcolor);
	send_bcptr(graphix,a->cross);
      }
      if (a->ltype!=ATM_LAB_NONE) {
	set_atom_label(a,a->ltype,a->label,FALSE);
      }
      if (a->sdsp) add_to_node_chain(NTAB[i],&c);
    }
  }
  if (c.n>0) {
    send_bheader(graphix,"ARST",sizeof(int)+c.n*cPtrSize());
    send_bint(graphix,1);
    for (a=(Atom *)c.s;a;a=(Atom *)a->chain) send_bcptr(graphix,a->cross);
    clear_node_chain(&c);
  }
    
  /* bonds */

  for (B=NULL,i=0;i<ndone;i++) {
    if (is_atom(NTAB[i])) {
      a=(Atom *)NTAB[i];
      for (j=0;j<a->nbnd;j++) {
	if (a==a->bnd[j]->t) {
	  a->bnd[j]->chain=B;
	  B=a->bnd[j];
	  if (B->labt!=BND_LAB_NONE) set_bond_label(B,B->labt,B->label);
	}
      }
    }
  }

  for (n=0,bs=B,bn=B;bn;bn=bn->chain) {   /* bond linewidth */
    if (cPtr(bn->cross)) n++;
    if (!bn->chain||bn->chain->lw!=bs->lw) {
      if (bs->lw!=DefaultBondWidth()) {
	send_bheader(graphix,"SBLW",sizeof(int)+n*cPtrSize());
	send_bint(graphix,bs->lw);
	for (;bs!=bn->chain;bs=bs->chain) if (cPtr(bs->cross)) send_bcptr(graphix,bs->cross);
      }else bs=bn->chain;
      n=0;
    }
  }

  for (n=0,bs=B,bn=B;bn;bn=bn->chain) {   /* bond radius */
    if (cPtr(bn->cross)) n++;
    if (!bn->chain||bn->chain->r!=bs->r) {
      if (bs->r!=0.1) {
	send_bheader(graphix,"BRAD",sizeof(int)+n*cPtrSize());
	send_bfloat(graphix,bs->r);
	for (;bs!=bn->chain;bs=bs->chain) if (cPtr(bs->cross)) send_bcptr(graphix,bs->cross);
      }else bs=bn->chain;
      n=0;
    }
  }

  for (n=0,bs=B,bn=B;bn;bn=bn->chain) {   /* bond cylinder status */
    if (cPtr(bn->cross)) n++;
    if (!bn->chain||bn->chain->cyl!=bs->cyl) {
      if (bs->cyl) {
	send_bheader(graphix,"BSOL",sizeof(int)+n*cPtrSize());
	send_bfloat(graphix,bs->cyl);
	for (;bs!=bn->chain;bs=bs->chain) if (cPtr(bs->cross)) send_bcptr(graphix,bs->cross);
      }else bs=bn->chain;
      n=0;
    }
  }

  for (;B;B=bn) {         /* clear bond chain  */
    bn=B->chain;
    B->chain=NULL;
  }

  return(status);
}

int restoreDotsV1(Stream *s)
{
  Atom *a;
  Parse p;
  char buff[8192],fld[4096];
  int status=0,seqnr,aseqnr,n,i,j,color,dsp,style;
  float *xyz;
  UCpipe *in;

  strm_flush(s);
  in=(UCpipe *)s->in;
  
  p.buf=buff;
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);

  while (!status&&fgets(buff,sizeof(buff),in->fd)) {
    p.pos=0; 
    if (get_token(&p)) {
      if (streq(p.tok,"---")) break;
      
      seqnr=err_seqnr();
      aseqnr=str2i(p.tok);
      n=str2i(token(&p));
      color=str2i(token(&p));
      dsp=str2i(token(&p));
      style=str2i(token(&p));
      if (aseqnr<0||aseqnr>=ndone||!is_atom(NTAB[aseqnr])) 
	err_set(1,"restoreDots: invalid atom number.");
      if (err_seqnr()==seqnr) {
	a=(Atom *)NTAB[aseqnr];
	xyz=malloc(4*n*sizeof(float));
	if (xyz) {
	  for (i=0;i<n&&!status;i++) {
	    if (fgets(buff,sizeof(buff),in->fd)) {
	      p.pos=0;
	      for (j=0;j<4;j++) xyz[4*i+j]=str2f(token(&p));
	    }else{
	      err_set(1,"restoreDots: I/O error, failed to read dots.");
	      status=9000;
	    }
	  }
	  if (i==n) {
	    if (cPtr(a->cross)) {
	      send_dots(a,n,xyz);
	      if (dsp!=DOT_IFATOM) {
		send_bheader(graphix,"DOTD",cPtrSize()+sizeof(int));
		send_bint(graphix,dsp); send_bcptr(graphix,a->cross);
	      }
	      if (color!=DOT_BYATOM) {
		send_bheader(graphix,"DOTC",cPtrSize()+sizeof(int));
		send_bint(graphix,color); send_bcptr(graphix,a->cross);
	      }
	      if (style!=DOT_SMALL) {
		send_bheader(graphix,"DOTS",cPtrSize()+sizeof(int));
		send_bint(graphix,style); send_bcptr(graphix,a->cross);
	      }
	    }
	  }
	  free(xyz);
	}else{
	  err_set(1,"restoreDots: failed to allocate memory for dots.");
	  status=9000;
	}
      }else{
	err_set(1,"restoreDots: format error in dot records.");
	status=9000;
      }
    }
  }
  return(status);
}

int restoreSRVV1(Stream *s)
{
  PrjTS Prj;
  Parse p;
  char buff[8192],fld[4096];
  int *ramp,i,l,nr,R,G,B,status=0,seqnr;
  float lwm[] = { 1.5, 0.5, 10.0, 10.0 };
  float pmin,pmax;
  UCpipe *in;

  in=(UCpipe *)s->in;
  strm_flush(s);

  p.buf=buff;
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);
  seqnr=err_seqnr();

  while (!status&&fgets(buff,sizeof(buff),in->fd)) {
    p.pos=0; 
    if (get_token(&p)) {
      if (streq(p.tok,"---")) break;

      if (streq(p.tok,"P")) {
	Prj.viewmode=str2i(token(&p));
	Prj.stereomode=str2i(token(&p));
	Prj.ysize=str2f(token(&p));
	Prj.znear=str2f(token(&p));
	Prj.slabwidth=str2f(token(&p));
	Prj.viewdist=str2f(token(&p));
	Prj.eyesep=str2f(token(&p));
	Prj.aspect=str2f(token(&p));
	Prj.scrysize=str2f(token(&p));
	Prj.sep=str2f(token(&p));
	if (seqnr==err_seqnr()) {
	  setThreeDPrj(&Prj);
	}else{
	  err_set(1,"%s",buff);
	  err_set(1,"restoreSRV: bad definition of projection");
	  status=9000;
	}

      }else if (streq(p.tok,"A")) {
	if (get_token(&p)) setAttribute("SAAL",str2i(p.tok));  /* c.f. save.c for comments */
	if (get_token(&p)) setAttribute("SDPC",str2i(p.tok));
	if (get_token(&p)) setAttribute("SDTH",str2i(p.tok));
	if (get_token(&p)) setAttribute("SLTM",str2i(p.tok));

      }else if (streq(p.tok,"L")) {
	seqnr=err_seqnr();
	if (get_token(&p)) lwm[0]=str2f(p.tok);  /* c.f. save.c for comments */
	if (get_token(&p)) lwm[1]=str2f(p.tok);
	if (get_token(&p)) lwm[2]=str2f(p.tok);
	if (get_token(&p)) lwm[3]=str2f(p.tok);
	if (seqnr==err_seqnr()) setAttributeFA("SLWM",lwm);

      }else if (streq(p.tok,"CR")) {
	if ((l=str2i(token(&p)))>0) {
	  ramp=malloc(l*sizeof(int));
	  pmin=str2f(token(&p));
	  pmax=str2f(token(&p));
	  if (ramp) {
	    for (i=0;i<l;i++) if (get_token(&p)) ramp[i]=str2i(p.tok);
	    DefineColorRamp(l,ramp);
	    setPropertyMap(pmin,pmax);
	    free(ramp);
	  }else err_set(1,"error reading color ramp, ignored.");
	}

      }else if (streq(p.tok,"C")) {
	nr=str2i(token(&p));
	R=str2i(token(&p));
	G=str2i(token(&p));
	B=str2i(token(&p));
	get_token(&p);
	if (seqnr==err_seqnr()) {
	  if (nr>=-1&&nr<WNP_MAXCOL) {
	    DefineColor(nr,p.tok,R,G,B);
	  }else{
	    err_set(1,"%s",buff);
	    err_set(1,"restoreSRV: bad color definition in session fille");
	    status=9000;
	  }
	}else{
	  err_set(1,"%s",buff);
	  err_set(1,"restoreSRV: bad definition of projection");
	  status=9000;
	}
      }else{
	err_set(1,"restoreSRV: bad token in session file");
	status=9000;
      }
    }
  }

  return(status);
}

int restoreDMONV1(Stream *s)
{
  int seqnr,status=0;
  char buff[4096],fld[4096];
  Atom *a[500];
  Parse p;
  int i,k,n;
  UCpipe *in;

  in=(UCpipe *)s->in;

  p.buf=strm_buff(s);
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);
  p.pos=strm_pos(s);

  seqnr=err_seqnr();
  if (get_token(&p)) {
    n=str2i(p.tok); if (seqnr!=err_seqnr()) status=9000;
    send_bheader(graphix,"MONS",sizeof(int));
    send_bint(graphix,n);
  }
  if (get_token(&p)) {
    n=str2i(p.tok); if (seqnr!=err_seqnr()) status=9000;
    send_bheader(graphix,"MONC",sizeof(int));
    send_bint(graphix,n);
  }
  if (get_token(&p)) {
    n=str2i(p.tok); if (seqnr!=err_seqnr()) status=9000;
    send_bheader(graphix,"MONL",sizeof(int));
    send_bint(graphix,n);
  }
  if (get_token(&p)) {
    n=str2i(p.tok); if (seqnr!=err_seqnr()) status=9000;
    send_bheader(graphix,"MONW",sizeof(int));
    send_bint(graphix,n);
  }

  strm_flush(s);
  p.buf=buff;

  if (!status) {
    for (n=0;fgets(buff,sizeof(buff),in->fd);) {
      p.pos=0;
      if (strneq(buff,"---",3)) break;
      while (get_token(&p)) {
	k=str2i(p.tok);
	if (k>-0&&k<ndone&&is_atom(NTAB[k])) a[n++]=(Atom *)NTAB[k];
	else a[n++]=NULL;
	if (n==500) {
	  send_bheader(graphix,"AMON",n*cPtrSize());
	  for (i=0;i<n;i++) send_bcptr(graphix,a[i]?a[i]->cross:cPtrNULL());
	  n=0;
	}
      }
    }
    if (n>0) {
      send_bheader(graphix,"AMON",n*cPtrSize()); 
      for (i=0;i<n;i++) send_bcptr(graphix,a[i]?a[i]->cross:cPtrNULL());
    }
  }else err_set(1,"restoreDMON: invalid DMON header found in session file.");

  return(status);
}

void freeRestoreDataV1(void)
{
  if (NTAB) free(NTAB);
  NTAB=0;
  ndone=0;

  if (MTAB) free(MTAB);
  MTAB=0;
  mdone=0;
}

/*========================================================================*/

typedef struct {
  char *name;
  wnpmacro_t *(*get)(void);
} macNamTab_t;

static macNamTab_t macTab[] = {
  { "maccs" , getMACCSmacros }
};


int restoreMacrosV1(Stream *s)
{
  wnpmacro_t *macros,*m;
  int status=0,i,n;
  char buff[4096],fld[4096];
  Parse p;
  UCpipe *in;

  in=(UCpipe *)s->in;
  strm_flush(s);

  p.buf=buff;
  p.tok=fld;
  p.sep="";
  p.mts=sizeof(fld);

  for (;fgets(buff,sizeof(buff),in->fd);) {
    if (strneq(buff,"---",3)) break;
    p.pos=0;
    if (get_token(&p)) {
      n=sizeof(macTab)/sizeof(macNamTab_t);
      for (i=0;i<n;i++) if (streq(p.tok,macTab[i].name)) break;
      if (i<n) macros=macTab[i].get();
      else macros=NULL;
      if (macros) {
	m=new_macro();
	set_macro_name(m,token(&p));
	for (;fgets(buff,sizeof(buff),in->fd);) {
	  if (strneq(buff,"ORCAM",5)) break;
	  else add_macro_line(m,buff);
	}
	insert_macro(m,macros);
      }else{
	err_set(1,"could not find macro table for macro type %s",p.tok);
	err_set(1,"%s",p.buf);
	status=9001;
      }
    }
  }
  return(status);
}
