#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "salign.h"


/*-------------------- prtAlignment --------------------*/

void  prtAlignment(FILE *out,int m,SAlires *A,int n, SAlires *B)
{
  /* Q&D printout of Alignment... */

  int p,q;

  for (p=0,q=0;p<m&&q<n;) {
    for (;p<m&&A[p].pos<q;p++) fprintf(out,"%4d  --\n",p+1);
    for (;q<n&&B[q].pos<p;q++) fprintf(out," --  %4d\n",q+1);
    for (;p<m&&A[p].pos==q;p++,q++) fprintf(out,"%4d %4d\n",p+1,q+1);
  }
  for (;p<m;p++) fprintf(out,"%4d  --\n",p+1); 
  for (;q<n;q++) fprintf(out," --  %4d\n",q+1); 
}

/*========================================================================*/

float ResCompare5(int i,int j,float *dA,float *dB)
{
  int n=0,k;
  float r=0.0,*da,*db;

  for (da=dA+i*6,db=dB+j*6,k=0;k<6;k++) {
    if (da[k]>0&&db[k]>0) {
      r+=(da[k]-db[k])*(da[k]-db[k]);
      n++;
    }
  }
  if (n>0) r=sqrt(r/n);
  else r=3.0;

  return(r);
}

void ResCompareSetup5(int m,SAlires *A,int n,SAlires *B,float **dA,float **dB)
{
  *dA=Res5Distances(m,A);
  *dB=Res5Distances(n,B);

  if (!*dA||!*dB) {
    if (*dA) free(*dA);
    *dA=NULL;
    if (*dB) free(*dB);
    *dB=NULL; 
  }
}

float *Res5Distances(int m,SAlires *A)
{
  float *x,*y,d,*dA=NULL;
  int i,k,*C=NULL;
  int s[6][2] = { {-2,+2}, {-2, 0}, {-2,+1}, {-1,+1}, {-1,+2}, {0,+2}};

  if (m>0) dA=(float *)malloc(6*m*sizeof(float));
  else dA=(float *)NULL;

  if (dA) {
    C=malloc(m*sizeof(int));

    if (C) {
      C[0]=0;
      for (i=1;i<m;i++) {
	d=(A[i].x[0]-A[i-1].x[0])*(A[i].x[0]-A[i-1].x[0])
	  +(A[i].x[1]-A[i-1].x[1])*(A[i].x[1]-A[i-1].x[1])
	  +(A[i].x[2]-A[i-1].x[2])*(A[i].x[2]-A[i-1].x[2]);
	if (d<4.5*4.5) C[i]=C[i-1];
	else C[i]=C[i-1]+1;
      }
    }

    for (i=0;i<m;i++) {
      for (k=0;k<6;k++) {
	if (C&&(C[i+s[k][0]]!=C[i]||C[i+s[k][1]]!=C[i])) dA[6*i+k]=0;
	else if (i+s[k][0]>=0&&i+s[k][1]<m) {
	  x=A[i+s[k][0]].x; y=A[i+s[k][1]].x;
	  dA[6*i+k]=sqrt((x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]));
	}else dA[6*i+k]=0.0;
      }
    }

    if (C) free(C);
  
  }
  return(dA);
}

float ResCompare6(int i,int j,float *dA,float *dB)
{
  int n=0,k;
  float r=0.0;
  int s[4]={0,1,2,4};

  for (k=0;k<4;k++) {
    if (dA[2*i+s[k]]>0&&dB[2*j+s[k]]>0) {
      r+=(dA[2*i+s[k]]-dB[2*j+s[k]])*(dA[2*i+s[k]]-dB[2*j+s[k]]);
      n++;
    }
  }
  if (n>0) r=sqrt(r/n);
  else r=3.0;

  return(r);
}

void ResCompareSetup6(int m,SAlires *A,int n,SAlires *B,float **dA,float **dB)
{
  float *x,*y;
  int i;

  if (m>0) *dA=(float *)malloc(2*(m+2)*sizeof(float));
  else *dA=(float *)NULL;

  if (n>0) *dB=(float *)malloc(2*(n+2)*sizeof(float));
  else *dB=(float *)NULL;  

  if (!*dA||!*dB) {
    if (*dA) free(*dA);
    *dA=(float *)NULL;
    if (*dB) free(*dB);
    *dB=(float *)NULL;
    return;
  }

  for (i=0;i<m;i++) {
    if (i<m-3) {
      x=A[i].x; y=A[i+3].x;
      (*dA)[i*2]=sqrt((x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]));
    }else (*dA)[i*2]=0.0;
    if (i<m-5) {
      x=A[i].x; y=A[i+5].x;
      (*dA)[i*2+1]=sqrt((x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]));
    }else (*dA)[i*2+1]=0.0;
  }
    
  for (i=0;i<n;i++) {
    if (i<n-3) {
      x=B[i].x; y=B[i+3].x;
      (*dB)[i*2]=sqrt((x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]));
    }else (*dB)[i*2]=0.0;
    if (i<n-5) {
      x=B[i].x; y=B[i+5].x;
      (*dB)[i*2+1]=sqrt((x[0]-y[0])*(x[0]-y[0])+(x[1]-y[1])*(x[1]-y[1])+(x[2]-y[2])*(x[2]-y[2]));
    }else (*dB)[i*2+1]=0.0;
  }
  
  for (i=0;i<2;i++) {
    (*dA)[2*(m+i)]=(*dA)[2*(m-1)];
    (*dA)[2*(m+i)+1]=(*dA)[2*(m-1)+1];
    (*dB)[2*(n+i)]=(*dB)[2*(n-1)];
    (*dB)[2*(n+i)+1]=(*dB)[2*(n-1)+1];
  }
    
}
/*-------------------- algCopy -----------------------------*/

SAlires *SAliresCopy(int n,SAlires *s)
{
  SAlires *c;
  int i;

  if (n>0&&s) {
    c=(SAlires *)malloc(n*sizeof(SAlires));
    for (i=0;i<n;i++) c[i]=s[i];
  }else c=(SAlires *)NULL;
  return(c);
}
