#include <stdio.h>
#include <math.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"
#include "wnpconst.h"
#include "xtal.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Node *LibRoot;

/* Macro: print float with n decimals, remove trailing 0's */
#define FLTOUT(x,n) {float _x; _x=x;\
    fprintf(out,"%s",prfx);\
    sprintf(prfx,"%0.*f",(n),(_x));\
    while (prfx[strlen(prfx)-1]=='0') prfx[strlen(prfx)-1]='\0';\
    fprintf(out," %s",prfx);\
    prfx[0]='\0';\
  }

/* Macro: print float with n decimal, or '.' if float is within t of d */
#define FLTmOUT(x,n,d,t) {\
    if ( (x)-(d)<(t)&&(d)-(x)<(t) ) {\
      sprintf(prfx+strlen(prfx)," .");\
    }else{ \
      fprintf(out,"%s",prfx);\
      sprintf(prfx,"%0.*f",(n),(x));\
      while (prfx[strlen(prfx)-1]=='0') prfx[strlen(prfx)-1]='\0';\
      fprintf(out," %s",prfx);\
      prfx[0]='\0';\
    }\
  }

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))

#define TRUE 1
#define FALSE 0


#define SV_TYPES (OBJ_MOL|OBJ_ATOM|OBJ_ATOMSET|OBJ_XCN|OBJ_NOELIST)

static int findseqnr(Atom *a);
static float oldr=-1;
static float obr=-1;
static int oblabt=-1;
static int oblw=-1;
static int obcyl=-1;

/*-------------------- OBJ_MOL --------------------*/

static void saveMolNode(FILE *out,Node *n)
{
  Node *q;
  Mol *m;
  int i;

  m=(Mol *)n;

  fprintf(out,"<M n=\"%s\">\n",n->name);
  fprintf(out,"\t%d %d %d",m->moltype,m->keep,ATOMSET_WORDS);
  for (i=0;i<ATOMSET_WORDS;i++) fprintf(out," %lx",m->used[i]);
  for (i=0;i<ATOMSET_WORDS;i++) fprintf(out," %lx",m->local[i]);
  fprintf(out,"\n");
  
  forNodes(q,n) saveMDTNode(out,q);

  fprintf(out,"</M>\n");
}

/*-------------------- OBJ_ATOM --------------------*/

static void saveAtomNode(FILE *out,Node *n)
{
  Node *q;
  Atom *a,*b;
  int i,j;
  char prfx[1024],Q;
  unsigned long bits;

  a=(Atom *)n;

  /* node type, index to up-node, node name */
  if (a->nxyz<=1) {
    fprintf(out,"<A n=\"%s\">\n",n->name);
  }else{
    fprintf(out,"<A n=\"%s\" alt=\"",n->name);
    for (i=0;i<a->nxyz;i++) fprintf(out,"%c",a->alt[i]);
    fprintf(out,"\">\n");
  }

  /* FF type, color, x, y, z, q, biso, sof, shift ( '.' = default value ) */
  *prfx=0;
  fprintf(out,"\t%s %d",a->atp?a->atp->name:"?",a->color);
  if (a->nxyz>0&&a->x) {
    for (i=0;i<3;i++) FLTmOUT(a->x[i],5,0,0);
  }
  FLTmOUT(a->q,6,0,1.e-7);
  FLTmOUT(a->x[ATM_BISO],2,0,1.e-3);
  FLTmOUT(a->x[ATM_SOF],3,1,1.e-4);
  FLTmOUT(a->shift,4,1.1*WNP_NO_SHIFT,-0.01*WNP_NO_SHIFT);
  fprintf(out,"\n");

  /* radius, label color, label type, label text */

  *prfx=0;
  fprintf(out,"\t");
  if (a->atp&&a->atp->rvdw==a->r) fprintf(out,"v");
  else if (a->r==oldr) sprintf(prfx,".");
  else {
    FLTOUT(a->r,3);
    oldr=a->r;
  }
  if (a->lcolor==WNP_COLOR_BYATOM) sprintf(prfx+strlen(prfx)," .");
  else {
    fprintf(out,"%s %d",prfx,a->lcolor);
    *prfx=0;
  }
  if (a->ltype!=ATM_LAB_NONE) {
    fprintf(out,"%s %d",prfx,a->ltype);
    if (a->label) fprintf(out," %s",a->label);
  }
  fprintf(out,"\n");

  /* ligance, formal charge, bits(fe,cosw,oop,stass,arom,dsp), atomset words  */
  bits=a->fe;
  bits=(bits<<2)|a->stass;
  bits=(bits<<1)|a->cosw;
  bits=(bits<<2)|a->oop;
  bits=(bits<<1)|a->arom;
  bits=(bits<<1)|a->sdsp;
  bits=(bits<<1)|a->ldsp;
  bits=(bits<<1)|a->dsp;
  
  if (a->lig!='?') fprintf(out,"\t%d",a->lig); else fprintf(out,"\t?");
  if (a->chrg!='?') fprintf(out," %d",a->chrg); else fprintf(out," ?");
  fprintf(out," %lx",bits);
  for (i=0;i<ATOMSET_WORDS;i++) fprintf(out," %lx",a->set[i]);
  fprintf(out,"\n");

  /* charge group, bond list (non redundant, may be blank line): index to neighbour, type mnemonic */
  *prfx=0;
  if (!streq(a->grp,"??")) {
    fprintf(out,"\t%s",a->grp);
    *prfx=0;
  }else sprintf(prfx,"\t.");
  for (i=0;i<a->nbnd;i++) {
    b=neighbour(i,a);
    if (b->seqnr<a->seqnr) {
      fprintf(out,"%s %d",prfx,b->seqnr);
      if (a->bnd[i]->bt!=1) {
	fprintf(out,",%s",BT_mnem(a->bnd[i]->bt));
	*prfx=0;
      }else sprintf(prfx,",");
      if (a->bnd[i]->lw!=oblw) {
	fprintf(out,"%s,%d",prfx,a->bnd[i]->lw);oblw=a->bnd[i]->lw;
	*prfx=0;
	oblw=a->bnd[i]->lw;
      }else sprintf(prfx+strlen(prfx),",");
      if (a->bnd[i]->cyl!=obcyl) {
	fprintf(out,"%s,%s",prfx,a->bnd[i]->cyl?"Y":"N");
	*prfx=0;
	obcyl=a->bnd[i]->cyl;
      }else sprintf(prfx+strlen(prfx),",");
      if (a->bnd[i]->r!=obr) {
	fprintf(out,"%s",prfx);			\
	sprintf(prfx,"%0.4f",a->bnd[i]->r);				\
	while (prfx[strlen(prfx)-1]=='0') prfx[strlen(prfx)-1]='\0';	\
	fprintf(out,"%s",prfx);					\
	prfx[0]='\0';							\
	obr=a->bnd[i]->r;
      }
      if (a->bnd[i]->label&&a->bnd[i]->labt==BND_LAB_TEXT) {
	if (strchr(a->bnd[i]->label,'"')) Q='\'';
	else Q='\"';
	fprintf(out,"%s,%c%s%c",prfx,Q,a->bnd[i]->label,Q);
	*prfx='\0';
      }else{
	if (a->bnd[i]->labt!=oblabt) {
	  fprintf(out,"%s,%d",prfx,a->bnd[i]->labt);
	  *prfx=0;
	  oblabt=a->bnd[i]->labt;
	}
      }
      *prfx='\0';
    }
  }
  fprintf(out,"\n");

  for (i=1;i<a->nxyz;i++) {
    sprintf(prfx,"\t");
    for (j=0;j<3;j++) FLTmOUT(a->x[i*ATM_LOCSZ+j],5,0,0);
    FLTmOUT(a->x[i*ATM_LOCSZ+ATM_BISO],2,0,1.e-3);
    FLTmOUT(a->x[i*ATM_LOCSZ+ATM_SOF],3,1,1.e-4);
    fprintf(out,"\n");
  }
  
  forNodes(q,n) saveMDTNode(out,q);

  fprintf(out,"</A>\n");
  
}

/*-------------------- OBJ_ATOMSET --------------------*/

static void saveAtomsetNode(FILE *out,Node *n)
{
  Atomset *s;

  s=(Atomset *)n;

  fprintf(out,"<S n=\"%s\" bit=\"%d\"/>\n",n->name,s->bit);
}

/*-------------------- OBJ_XCN --------------------*/

static void saveXCNode(FILE *out,Node *n)
{
  Node *q;
  Xcn *xcn;
  int i,j;
  Symop sym;
  hisrec *h;
  char buf[256],prfx[256];
  float err;

  fprintf(out,"<X n=\"%s\">\n",n->name);
  if (is_xcn(n)) {
    xcn=(Xcn *)n;

    *prfx=0;

    /* write title */
    if (xcn->title) fprintf(out,"\tT %s\n",xcn->title);

    /* write wave length and cell parameters */
    fprintf(out,"\tC");
    for (i=0;i<3;i++) FLTOUT(xcn->cl[i],4);
    for (i=0;i<3;i++) FLTOUT(xcn->ca[i],4);
    fprintf(out,"\n");

    /* write estimated standard devitions for cell parameters */
    fprintf(out,"\tc");
    for (i=0;i<3;i++) FLTOUT(xcn->esd_cl[i],4);
    for (i=0;i<3;i++) FLTOUT(xcn->esd_ca[i],4);
    fprintf(out,"\n");

    if (fabs(xcn->lambda-1.0)>1e-5) {
      fprintf(out,"\tL");
      FLTOUT(xcn->lambda,4);
      fprintf(out,"\n");
    }

    /* write mol. formula */
    if (xcn->forml>0) {
      fprintf(out,"\tF");
      for (i=0;i<xcn->forml;i++) fprintf(out," %s %d",xcn->forms[i],(int)xcn->formn[i]);
      fprintf(out,"\n");
    }

    /* write lattice type and space group name */
    i=xcn->lattice; if (!xcn->centric) i = -i;
    fprintf(out,"\tS %d %s\n",i,xcn->spgr);
    
    /* write symmetry operators */
    for (i=0;i<xcn->nsym;i++) {
      sym=symF(xcn,xcn->sym[i]);sym2str(buf,sym);
      if (!streq("X,Y,Z",buf)) fprintf(out,"\ts %s\n",buf);
    }
    
    /* write asymmetric unit definition */
    for (i=0;i<3;i++) err=fabs(xcn->asym[i])+fabs(xcn->asym[3+i]-1.0);
    if (err>1.e-5) {
      fprintf(out,"\tA");
      for (i=0;i<6;i++) FLTOUT(xcn->asym[i],5);
      fprintf(out,"\n");
    }

    /* scale record (very rare...) */
    if (xcn->scale) fprintf(out,"scale %s\n",xcn->scale);

    /* write history records */
    for (h=xcn->history;h;h=h->nxt) {
      fprintf(out,"\tH %s\n",h->t);
    }

    /* write cryst. to A-cart. transformation matrix into file */
    fprintf(out,"\tM");
    for (i=0;i<3;i++) for(j=0;j<4;j++) FLTOUT(xcn->fta.r[i].c[j],5);
    fprintf(out,"\n");
  }

  
  forNodes(q,n) saveMDTNode(out,q);
  fprintf(out,"</X>\n");

}

/*-------------------- OBJ_DSP(DSP_GLBL) --------------------*/

static void saveGLBLNode(FILE *out,Node *n)
{
  int i,j;
  Node *q;

  Dspobj *D;
  char prfx[256];

  D=(Dspobj *)n;
  
  if (is_dsp(n)&&D->dsptype==DSP_GLBL) {
    fprintf(out,"<GLBL dsp=\"%s\" m=\"%d\">\n",D->status?"on":"off",D->root->seqnr);

    *prfx=0;
    for (i=0;i<4;i++) {
      fprintf(out,"\t");
      for (j=0;j<4;j++) FLTOUT(D->mtrx.r[i].c[j],5);
      fprintf(out,"\n");
    }
  
    for (q=n->fst;is_dsp(q);q=q->nxt) saveDDTNodes(out,q);
    fprintf(out,"</GLBL>\n");
  }
}

static void saveMDLNode(FILE *out,Dspobj *D)
{
  if (D->dsptype==DSP_MOL) {
    fprintf(out,"<MDL dsp=\"%s\" m=\"%d\"/>\n",D->status?"on":"off",D->root->seqnr);
  }
}

/*-------------------- save OBJ_DSP(DSP_CSO,DSP_TRACE,DSP_CNT) --------------------*/

static void saveCSONode(FILE *out,Dspobj *D)
{
  Qmsg *Msg=NULL;
  int i,n,size,type,lw,mds_type,ltype,color;
  int *msg;
  char prfx[1024];

  if (graphix&&SrvVersion()>2.19999) {
    send_bheader(graphix,"GGOB",sizeof(D));
    send_ptr(graphix,D);
    Msg=waitqmsg("GGOB SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }else{
	if (Msg->size>=4*sizeof(int)) {
	  msg=(int *)Msg->msg;
	  type=MY_i(*msg++);
	  MY_i(*msg++);   /* flags, not used */
	  switch (type) {
	  case DSP_CSO:
	    lw=MY_i(*msg++);
	    mds_type=MY_i(*msg++);
	    fprintf(out,"<%s n=\"%s\" dsp=\"%s\">\n",
		    D->dsptype==DSP_TRCE?"TRCE":D->dsptype==DSP_CNT?"CNT":"CSO",
		    D->name,D->status?"on":"off");
	    fprintf(out,"\t%d %d\n",lw,mds_type);
	    Msg=delqmsg(Msg);
	    while ((Msg=waitqmsg("DATA SCCN"))) {
	      if (strneq(Msg->type,"SCCN",4)) {
		Msg=delqmsg(Msg);
		graphics_down("Server closed connection...");
		err_set(1,"saveCSO: data transfer from server is incomplete.");
		break;
	      }else if (Msg->size>0) {
		msg=(int *)Msg->msg;
		size=Msg->size/sizeof(int);
		if (size>2) {
		  ltype=MY_i(*msg++);
		  color=MY_i(*msg++);

		  switch (ltype) {
		  case WNP_LST_MDS:
		    n=(Msg->size-2*sizeof(int))/sizeof(int);
		    if (n>0) {
		      fprintf(out,"\tL %d %d",color,n);
		      for (i=0;i<n;i++) {
			if (i%6==0) sprintf(prfx,"\n\t");
			FLTOUT(MY_f(*msg++),4);
		      }
		      fprintf(out,"\n");
		    }
		    break;
		  case WNP_LST_TOBJ:
		    if (size>=5) {
		      fprintf(out,"\tT %d",color);
		      *prfx=0;
		      FLTOUT(MY_f(*msg++),4); FLTOUT(MY_f(*msg++),4); FLTOUT(MY_f(*msg++),4); /* x,y,z */
		      fprintf(out," %s\n",Msg->msg+5*sizeof(int));
		    }
		    break;
		  default:
		    err_set(1,"saveCSO: invalid DATA respons from server to GGOB request");
		  }
		}
	      }else break;	    
	      Msg=delqmsg(Msg);
	    }
	    fprintf(out,"</%s>\n",
		    D->dsptype==DSP_TRCE?"TRCE":D->dsptype==DSP_CNT?"CNT":"CSO");
	    Msg=delqmsg(Msg);
	    break;
	  default:
	    err_set(1,"saveCSO: invalid response from server to GGOB request");
	    break;
	  }
	}
      }
    }
    delqmsg(Msg);
  }
}

static void saveTRGNode(FILE *out,Dspobj *D)
{
  Surface *S;
  int flags;

  if (graphix&&SrvVersion()>2.19999) {
    S=getSurfaceFromServer(D,&flags);
    if (S) {
      fprintf(out,"<TRG n=\"%s\" dsp=\"%s\">\n",D->name,D->status?"on":"off");
      writeSurface(out,S);
      S=freeSurface(S);
      fprintf(out,"</TRG>\n");
    }
  }
}

static void saveNOEListNode(FILE *out,Node *n)
{
  NOElist *m;
  
  m=(NOElist *)n;
  
  fprintf(out,"<N n=\"%s\" average=\"%d\">\n",n->name,m->avtype);
  write_xplor_noelist(out,m);
  fprintf(out,"</N>\n");
}

/*========================================================================*/

int initSave(FILE *out)
{
  int status=0;

  fprintf(out,"<?xml version=\"1.0\"?>\n\n<WSV version=\"2\">\n");
  oldr=-1;
  obr=-1;
  oblabt=-1;
  oblw=-1;
  obcyl=-1;

  return(status);
}

int numberMDTNodes(int *N,Node *r) 
{
  Node *n;

  /* N: total number of nodes visited so far */

  if ((r->type&SV_TYPES)==0) return(*N);

  r->seqnr=*N;
  *N+=1;

  forNodes(n,r) {
    numberMDTNodes(N,n);
  }

  return(*N);
}

void saveMDT(FILE *out,Node *root)
{
  int N=0;
  Node *n;

  if (root->up) N=1;
  N=numberMDTNodes(&N,root);
  fprintf(out,"\n<MDT nodes=\"%d\">\n",N);
  if (root->up) saveMDTNode(out,root);
  else {
    forNodes(n,root) saveMDTNode(out,n);
  }
  fprintf(out,"</MDT>");
}

void saveMDTNode(FILE *out,Node *n)
{
  if ((n->type&SV_TYPES)==0) return;

  switch (n->type) {
  case OBJ_MOL:
    saveMolNode(out,n);
    break;
  case OBJ_ATOM:
    saveAtomNode(out,n);
    break;
  case OBJ_XCN:
    saveXCNode(out,n);
    break;
  case OBJ_ATOMSET:
    saveAtomsetNode(out,n);
    break;
  case OBJ_NOELIST:
    saveNOEListNode(out,n);
    break;
  default:
    prtGraphics("node type %d not dumped\n",n->type);
    break;
  }

  
}
/*========================================================================*/

int numberDDTNodes(int *N,Node *r,int type) 
{
  Node *n;

  /* N: total number of nodes visited so far */

  if (is_dsp(r)&&(((Dspobj *)r)->dsptype&type)) {
    r->seqnr=*N;
    *N+=1;
  }

  forNodes(n,r) {
    numberDDTNodes(N,n,type);
  }
  return(*N);
}

void saveDDT(FILE *out,Node *root)
{
  int M=0;

  numberDDTNodes(&M,root,SVR_GLBLS);
  fprintf(out,"\n<DDT nodes=\"%d\">\n",M);
  saveDDTNodes(out,root);
  fprintf(out,"</DDT>\n");
}

void saveDDTNodes(FILE *out,Node *n)
{
  Dspobj *D;

  D=(Dspobj *)n;
  if (is_dsp(n)) {
    switch (D->dsptype) {
    case DSP_ROOT:
      saveGLBLNode(out,n);
      break;
    case DSP_GLBL:
      saveGLBLNode(out,n);
      break;
    case DSP_MOL:
      saveMDLNode(out,D);
      break;
    case DSP_CNT:         /* implemented as CSO */
    case DSP_TRCE:        /* implemented as CSO */
    case DSP_CSO:
      saveCSONode(out,D);
      break;
    case DSP_TRG:
      saveTRGNode(out,D);
      break;
    case DSP_HHDIST:
      break;
    }
  }
}

void saveDots(FILE *out)
{
  Qmsg *Msg=NULL;
  Atom *a;
  int i,j,*msg,dsp,style,color,n,seqnr,tag=TRUE;
  char prfx[256];
  float p;

  if (SrvVersion()>=2.1999) {

    send_header(graphix,"GDOT",0);
    while ((Msg=waitqmsg("GDOT SCCN"))) {
      if (strneq(Msg->type,"SCCN",4)) {
	Msg=delqmsg(Msg);
	graphics_down("Server closed connection...");
      }else if (Msg->size==0) {
	Msg=delqmsg(Msg);
	break;
      }else{
	if (Msg->size>=sizeof(a)+4*sizeof(int)) {
	  memcpy(&a,Msg->msg,sizeof(a));
	  msg=(int *)(Msg->msg+sizeof(a));
	  n=MY_i(*msg++);
	  color=MY_i(*msg++);
	  dsp=MY_i(*msg++);
	  style=MY_i(*msg++);
	  if (Msg->size>=sizeof(a)+4*sizeof(int)+4*n*sizeof(int)) {
	    seqnr=findseqnr(a);
	    if (seqnr>=0) {
	      if (tag) {
		fprintf(out,"\n<DOTS>\n");
		tag=FALSE;
	      }
	      fprintf(out,"<D a=\"%d\" col=\"%d\" dsp=\"%d\" style=\"%d\"%s>\n",seqnr,color,dsp,style,n>0?"":"/");
	      for (i=0;i<n;i++) {
		*prfx='\0';
		for (j=0;j<3;j++) FLTOUT(MY_f(*msg++),4);
		p=MY_f(*msg++);
		if (p==0) fprintf(out," 0");
		else fprintf(out," %0.4g",p);
		fprintf(out,"\n");
	      }
	      if (n>0) fprintf(out,"</D>\n");
	    }
	  }else{
	    err_set(1,"saveDots: bad dots message received from server...");
	  }
	}else{
	  err_set(1,"saveDots: bad dots message received from server...");
	}
      }
      Msg=delqmsg(Msg);
    }
  }
  if (!tag) fprintf(out,"</DOTS>\n");
}

void saveDMON(FILE *out)
{
  Atom *a,*b;
  int i,n,m,*ip,sa,sb;
  char *p;
  Qmsg *Msg;
  char *txt,*attr[] = {"status","color","label","width",NULL};

  if (SrvVersion()>=2.1999) {
    send_bheader(graphix,"GMON",0);
    if ((Msg=waitqmsg("GMON SCCN"))) {
      if (strneq(Msg->type,"SCCN",4)) {
	graphics_down("Server closed connection...");
      }else{
	fprintf(out,"\n<DMON");
	m=Msg->size>=4*sizeof(int)?4:Msg->size/sizeof(int);
	n=(Msg->size-4*sizeof(int))/(2*sizeof(a));
	for (txt=attr[0],ip=(int *)Msg->msg,i=0;i<m;i++) {
	  if (txt) {
	    fprintf(out," %s=\"%d\"",txt,MY_i(*ip++));
	    txt=attr[i+1];
	  }else ip++;
	}
	fprintf(out,">\n");
	for (p=Msg->msg+m*sizeof(int),i=0;i<n;i++) {
	  memcpy(&a,p,sizeof(a)); p+=sizeof(a); sa=findseqnr(a);
	  memcpy(&b,p,sizeof(b)); p+=sizeof(b); sb=findseqnr(b);
	  if (sa>=0&&sb>=0) {
	    if (i%5==0) fprintf(out,"\t%d %d",sa,sb);
	    else fprintf(out," %d %d%s",sa,sb,i%5==4?"\n":"");
	  }
	}
	fprintf(out,"\n</DMON>\n");
      }
      Msg=delqmsg(Msg);
    }
  }
}

void saveSRV(FILE *out)
{
  PrjTS Prj;
  int i,l;
  float *floats,pmin,pmax;
  int *ints;

  fprintf(out,"\n<SRV>\n");

  Prj=getThreeDPrj();
  fprintf(out,"\tP %d %d %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f %0.4f\n",
	  Prj.viewmode,Prj.stereomode,Prj.ysize,Prj.znear,Prj.slabwidth,
	  Prj.viewdist,Prj.eyesep,Prj.aspect,Prj.scrysize,Prj.sep);
  
  fprintf(out,"\tA %d %d %d %d\n",
	  getAttribute("SAAL"),  /* anti aliasing mode */
	  getAttribute("SDPC"),  /* depth-cue mode */
	  getAttribute("SDTH"),  /* dithering mode */
	  getAttribute("SLTM")); /* 3D text mode */

  
  if ((floats=getAttributeFA("SLWM"))) {
    fprintf(out,"\tL %0.1f %0.1f %0.1f %0.1f\n",
	    floats[0],floats[1],floats[2],floats[3]);
  }

  l=NRampColors();
  if (l>0) {
    ints=ColorRamp();
    getPropertyMap(&pmin,&pmax);
    fprintf(out,"\tCR %d %0.6f %0.6f",l,pmin,pmax);
    for (i=0;i<l;i++) fprintf(out," %d",ints[i]);
    fprintf(out,"\n");
  }    

  fprintf(out,"\tC %2d %d %d %d fog\n",-1,ColVal(-1,0),ColVal(-1,1),ColVal(-1,2));
  fprintf(out,"\tC %2d %d %d %d background \n",0,ColVal(0,0),ColVal(0,1),ColVal(0,2));
  for (i=1;i<WNP_MAXCOL;i++) {
    if (ColName(i)) {
      fprintf(out,"\tC %2d %d %d %d %s\n",
	      i,ColVal(i,0),ColVal(i,1),ColVal(i,2),ColName(i));
    }
  }
  
  if (get_I_parm("cyl_mode")>-1 || get_I_parm("cyl_poly")>-1) {  /* cylinder quality */
    fprintf(out,"\tSCYM %d %d\n",
	    get_I_parm("cyl_mode"),
	    get_I_parm("cyl_poly"));
  }

  if (get_I_parm("sph_mode")>-1 || get_I_parm("sph_poly")>-1) {  /* sphere quality */
    fprintf(out,"\tSPHM %d %d\n",
	    get_I_parm("sph_mode"),
	    get_I_parm("sph_poly"));
  }

  fprintf(out,"</SRV>\n");
}


static int findseqnr(Atom *a)
{
  Atom *b;

  /* could (should?) be optimize by using binary search on sorted list, or hash table */

  ForAtoms(b,MolRoot) if (a==b) break;
  if (a) return(a->seqnr);
  else return(-1);
}

void saveMacros(FILE *out)
{
  wnpmacro_t *macros,*m;
  int i,n,tag=TRUE;
  char *type;

  n=nMacroTypes();
  for (i=0;i<n;i++) {
    tag=TRUE;
    type=getMacroType(i);
    if ((macros=getMacros(type))) {
      for (m=macros->nxt;m;m=m->nxt) {
	if (m->name&&m->txt) {
	  if (tag) {
	    fprintf(out,"\n<MACROS type=\"%s\">\n",type);
	    tag=FALSE;
	  }
	  fprintf(out,"\n<MACRO name=\"%s\">\n<![CDATA[\n",m->name);
	  for (i=0;i<m->n;i++) fprintf(out,"%s",m->txt[i]);
	  fprintf(out,"]]>\n</MACRO>\n");
	}
      }
    }
    if (!tag) fprintf(out,"\n</MACROS>\n");
  }
}
