#include <stdio.h>
#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "utlsock.h"
#include "transpor.h"
#include "strutil.h"
#include "gcalc.h"
#include "errhand.h"
#include <unistd.h>

extern int graphix,server;

static PrjTS ThreeDPrj;

#define TRUE 1
#define FALSE 0

#define ADTSZ (sizeof(void *)+4*sizeof(int)+4)

static void send_batom(Atom *a){

  unsigned char trail[4];

  trail[0]=(a->color > 0 ? a->color : -a->color);
  trail[1]=a->dsp;
  trail[2]=0;         /* padding for compatibility with old servers... */
  trail[3]=0;

  send_bptr(graphix,a);
  send_bfloat(graphix,a->x[0]);
  send_bfloat(graphix,a->x[1]);
  send_bfloat(graphix,a->x[2]);
  send_bfloat(graphix,a->r);
  send_bdata(graphix,4,(char *)trail);
}

static void send_bcatom(Atom *a){

  unsigned char trail[4];

  trail[0]=(a->color > 0 ? a->color : -a->color);
  trail[1]=a->dsp;
  trail[2]=0;         /* padding for compatibility with old servers... */
  trail[3]=0;

  send_bcptr(graphix,a->cross);
  send_bfloat(graphix,a->x[0]);
  send_bfloat(graphix,a->x[1]);
  send_bfloat(graphix,a->x[2]);
  send_bfloat(graphix,a->r);
  send_bdata(graphix,4,(char *)trail);
}

Dspobj *send_mdl(Dspobj *global,Mol *m)
{
  Dspobj *mdl=NULL;
  Nodechain A;
  
  /*   check arguments */

  if (!is_dsp(global)) return(mdl);
  if (!is_mol(m)) return(mdl);
  if (global->dsptype!=DSP_GLBL) return(mdl);

  /*   create molecular display object in graphics box */

  mdl=new_dsp_obj(global,(char *)NULL,DSP_MOL);
  if (!mdl) return(mdl);
  m->dsptr=mdl;
  mdl->root=(Node *)m;
  if (!cPtr(mdl->cross)) return(mdl);      /* rejected, or graphics not running */

  /*   collect atoms into chain  */

  A=tree2chain((Node *)m,OBJ_ATOM);

  /*   send to graphics and clear chain   */  

  mdl_add_atom_chain(mdl,&A);
  clear_node_chain(&A);
  return(mdl);
}

/*-------------------- send_one_atom --------------------*/

void send_one_atom(Atom *a)
{
  Bond *b,*bf,*bl;
  int i,size,nbnd;
  Qmsg *Msg;
  Dspobj *mdl;

  if (!graphix) return;
  if (!mvld(a)) return;
  if (!is_atom(a)) return;
  if (cPtr(a->cross)) return;
  
  mdl=assocmdl((Node *)a);
  if (!is_dsp(mdl)) return;

  size=sizeof(mdl)+ADTSZ;
  send_bheader(graphix,"AATM",size);
  send_bptr(graphix,mdl);
  send_batom(a);
  trflush(graphix);

  nbnd=0; bf=(Bond *)NULL; bl=(Bond *)NULL; 

  for (i=0;i<a->nbnd;i++) {
    if (!a->bnd[i]->chain) {
      if (!bf) bf=a->bnd[i];
      if (bl) bl->chain=a->bnd[i];
      bl=a->bnd[i];
      a->bnd[i]->chain=bl;
      nbnd++;
    }
  }

  if (bl) bl->chain=(Bond *)NULL;

  /*  read back cross atom addresses from graphics box */

  a->cross=cPtrNULL();
  size=cPtrSize();
  Msg=waitqmsg("ATM* SCCN");
  if (Msg) {
    if (strneq(Msg->type,"SCCN",4)) {
      delqmsg(Msg);
      Msg=(Qmsg *)NULL;
      graphics_down("Server closed connection...");
    }
  }else{
    graphics_down("no response from server.");
  }
  if (Msg) {
    if (Msg->size>=size) memcpy(&(a->cross),Msg->msg,size);
    delqmsg(Msg);
  }

  send_bond_chain(bf);

  /* reset bond chain */

  for (b=bf; b; b=bf) {bf=b->chain;b->chain=(Bond *)NULL;}
}

/*-------------------- mdl_add_atoms --------------------*/

void mdl_add_atoms(Dspobj *mdl,Node *root)
{
  /* add all atoms in tree(root) to MDL mdl...     */
  /* only atoms with a->cross==NULL will be added. */

  Nodechain A;
  Atom *a;

  A=tree2chain(root,OBJ_ATOM);
  for (A.n=0,A.e=(Node *)NULL,a=(Atom *)A.s;mvld(a);a=(Atom *)A.e->chain) {
    if (cPtr(a->cross)) {
      if (A.e) A.e->chain=(Node *)a->chain;
      a->chain=(Node *)NULL;
    }else{
      A.e=(Node *)a;
      A.n++;
    }
  }
  mdl_add_atom_chain(mdl,&A);
  clear_node_chain(&A);
}

/*-------------------- mdl_add_atom_chain --------------------*/

void mdl_add_atom_chain(Dspobj *mdl,Nodechain *A)
{
/*   add atoms to molecular display object,
     chain A should contain only atoms with cross==NULL */

  Atom *a;
  Bond *b,*bf,*bl;
  int i,size,nbnd,ps;
  Qmsg *Msg;

  if (graphix<1) return;

  if (!is_dsp(mdl)) return;
  if (mdl->dsptype!=DSP_MOL) return;

  if (A->n<1) return;

  lock_evtq();

  /*  first send atoms, count new bonds... */

  size=sizeof(mdl)+A->n*ADTSZ;
  send_bheader(graphix,"AATM",size);
  send_bptr(graphix,mdl);

  nbnd=0; bf=(Bond *)NULL; bl=(Bond *)NULL; 
  for (a=(Atom *)A->s;a;a=(Atom *)a->chain) {
    send_batom(a);
    for (i=0;i<a->nbnd;i++) {
      if (!a->bnd[i]->chain) {
	if (!bf) bf=a->bnd[i];
	if (bl) bl->chain=a->bnd[i];
	bl=a->bnd[i];
	a->bnd[i]->chain=bl;
	nbnd++;
      }
    }
  }
  trflush(graphix);
  if (bl) bl->chain=(Bond *)NULL;

  /*  read back cross atom addresses from graphics box */

  ps=cPtrSize();
  Msg=waitqmsg("ATM* SCCN");    /*todo: should also read SCCN */
  if (Msg) {
    if (strneq(Msg->type,"SCCN",4)) {
      delqmsg(Msg);
      Msg=(Qmsg *)NULL;
      graphics_down("Server closed connection...");
    }
  }else{
    graphics_down("no response from server.");
  }
  a=(Atom *)A->s;
  if (Msg) {
    size=(Msg->size)/ps;
    for (i=0; size>0&&a; a=(Atom *)a->chain,size--) {
      memcpy(&(a->cross),Msg->msg+i*ps,ps);
      i++;
    }
    delqmsg(Msg);
  }

  send_atom_attributes(A);

  send_bond_chain(bf);
  send_bond_attributes(bf);

  /* reset bond chain */

  for (b=bf; b; b=bf) {bf=b->chain;b->chain=(Bond *)NULL;}
  unlock_evtq();
}

/*-------------------- send_atom_attributes --------------------*/

void send_atom_attributes(Nodechain *atoms)
{
  Atom *a;
  int n=0;

  /* deal with non-standard atom attributes... */

  for (a=(Atom *)atoms->s;is_atom(a);a=(Atom *)a->chain) {
    if (cPtr(a->cross)) {
      if (!a->ldsp) {
	send_bheader(graphix,"ALST",cPtrSize()+sizeof(int));
	send_bint(graphix,0);
	send_bcptr(graphix,a->cross);
      }
      if (a->lcolor!=WNP_COLOR_BYATOM) {
	send_bheader(graphix,"ALCL",cPtrSize()+sizeof(int));
	send_bint(graphix,a->lcolor);
	send_bcptr(graphix,a->cross);
      }
      if (a->ltype!=ATM_LAB_NONE) {
	set_atom_label(a,a->ltype,a->label,FALSE);
      }
      if (a->sdsp) n++;
    }
  }
  if (n>0) {
    send_bheader(graphix,"ARST",sizeof(int)+n*cPtrSize());
    send_bint(graphix,1);
    for (a=(Atom *)atoms->s;a;a=(Atom *)a->chain) {
      if (a->sdsp) send_bcptr(graphix,a->cross);
    }
  }
}
/*-------------------- send_bond_attributes --------------------*/

void send_bond_attributes(Bond *bf)
{
  Bond *bs,*bn;
  int n;

  for (bn=bf;bn;bn=bn->chain) {
    if (bn->labt!=BND_LAB_NONE) set_bond_label(bn,bn->labt,bn->label);
  }

  for (n=0,bs=bf,bn=bf;bn;bn=bn->chain) {   /* bond linewidth */
    if (cPtr(bn->cross)) n++;
    if (!bn->chain||bn->chain->lw!=bs->lw) {
      if (bs->lw!=DefaultBondWidth()) {
	send_bheader(graphix,"SBLW",sizeof(int)+n*cPtrSize());
	send_bint(graphix,bs->lw);
	for (;bs!=bn->chain;bs=bs->chain) if (cPtr(bs->cross)) send_bcptr(graphix,bs->cross);
      }else bs=bn->chain;
      n=0;
    }
  }

  for (n=0,bs=bf,bn=bf;bn;bn=bn->chain) {   /* bond radius */
    if (cPtr(bn->cross)) n++;
    if (!bn->chain||bn->chain->r!=bs->r) {
      if (bs->r!=0.1) {
	send_bheader(graphix,"BRAD",sizeof(int)+n*cPtrSize());
	send_bfloat(graphix,bs->r);
	for (;bs!=bn->chain;bs=bs->chain) if (cPtr(bs->cross)) send_bcptr(graphix,bs->cross);
      }else bs=bn->chain;
      n=0;
    }
  }

  for (n=0,bs=bf,bn=bf;bn;bn=bn->chain) {   /* bond cylinder status */
    if (cPtr(bn->cross)) n++;
    if (!bn->chain||bn->chain->cyl!=bs->cyl) {
      if (bs->cyl) {
	send_bheader(graphix,"BSOL",sizeof(int)+n*cPtrSize());
	send_bfloat(graphix,bs->cyl);
	for (;bs!=bn->chain;bs=bs->chain) if (cPtr(bs->cross)) send_bcptr(graphix,bs->cross);
      }else bs=bn->chain;
      n=0;
    }
  }
}

/*-------------------- send_bond_chain --------------------*/

void send_bond_chain(Bond *bs)
{
  /* send bonds starting in chain starting at bs to graphics box */

  Bond *b;
  int i,size,nb,ps;
  Qmsg *Msg;

  /* count bonds that are not in graphics, but have both atoms in gr. */

  nb=0;
  for (nb=0,b=bs;mvld(b);b=b->chain) {
    if (cPtr(b->f->cross) && cPtr(b->t->cross) && !cPtr(b->cross)) nb++;
  }

  if (nb>0) {

    /* send atom pointer pair list */
    send_bheader(graphix,"ABND",2*nb*cPtrSize());
    for (b=bs;b;b=b->chain) {
      if (cPtr(b->f->cross) && cPtr(b->t->cross) && !cPtr(b->cross)) {
	send_bcptr(graphix,b->f->cross);
	send_bcptr(graphix,b->t->cross);
      }
    }
    
    /* now read back cross bond pointers */

    Msg=waitqmsg("BND* SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }
    }else{
      graphics_down("no response from server.");
    }
    for (b=bs;b;b=b->chain) memset(&(b->cross),0,8);
    ps=cPtrSize();
    if (Msg) {
      size=Msg->size/ps;
      for (i=0,b=bs; i<size&&b; b=b->chain) {
	if (cPtr(b->f->cross) && cPtr(b->t->cross) && !cPtr(b->cross)) {
	  memcpy(&(b->cross),Msg->msg+i*ps,ps);
	  i++;
	}
      }
      delqmsg(Msg);
    }
  }
}
/*-------------------- mdl_repl_atom_chain --------------------*/

void mdl_repl_atom_chain(Nodechain *A)
{
/*   repl atom data in molecular display objects,
     chain A should contain only atoms with cross!=NULL */

  Atom *a;
  int size;

  if (graphix<1) return;

  if (A->n<1) return;

  lock_evtq();

  size=A->n*(cPtrSize()+4*sizeof(int)+4);
  send_bheader(graphix,"RATM",size);
  for (a=(Atom *)A->s;a;a=(Atom *)a->chain) send_bcatom(a);

  unlock_evtq();
}
/*-------------------- send_atom_chain --------------------*/

void send_atom_chain(Nodechain *c)
{
  Atom *a;
  int i,nat,nbnd,size,ps;
  Bond *b,*bf,*bl;
  Qmsg *Msg;

  if (graphix<1) return;
  if (!c) return;
  if (c->n<1) return;

  for (nat=0,a=(Atom *)c->s;a;a=(Atom *)a->chain) {
    if (!cPtr(a->cross)) {
      if (assocmdl(a->up)) {
	nat++;
	a->flg|=(1<<1);
      }
    }
  }
  if (nat>0) {
    lock_evtq();
    nbnd=0; bf=(Bond *)NULL; bl=(Bond *)NULL; 
    send_bheader(graphix,"AACH",nat*(sizeof(Dspobj *)+ADTSZ));
    for (a=(Atom *)c->s;a;a=(Atom *)a->chain) {
      if (a->flg&(1<<1)) {
	send_bptr(graphix,assocmdl(a->up));
	send_batom(a);
	for (i=0;i<a->nbnd;i++) {
	  if (!a->bnd[i]->chain) {
	    if (!bf) bf=a->bnd[i];
	    if (bl) bl->chain=a->bnd[i];
	    bl=a->bnd[i];
	    a->bnd[i]->chain=bl;
	    nbnd++;
	  }
	}
      }
    }
    if (bl) bl->chain=(Bond *)NULL;
	
    /*  read back cross atom addresses from graphics box */

    Msg=waitqmsg("ATM* SCCN");    /*todo: should also read SCCN */
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }
    }else{
      graphics_down("no response from server.");
    }
    ps=cPtrSize();
    a=(Atom *)c->s;
    if (Msg) {
      size=(Msg->size)/ps;
      for (i=0,a=(Atom *)c->s; size>0&&a; a=(Atom *)a->chain) {
	if (a->flg&(1<<1)) {
	  a->flg&=~(1<<1);
	  memcpy(&(a->cross),Msg->msg+i*ps,ps);
	  size--;
	  i++;
	}
      }
      delqmsg(Msg);
    }
    for (;a;a=(Atom *)a->chain) {
      if (a->flg&(1<<1)) {
	a->flg&=~(1<<1);
	memset(&(a->cross),0,8);
      }
    }

    send_atom_attributes(c);
    
    send_bond_chain(bf);
    send_bond_attributes(bf);
    
    /* reset bond chain */
    
    for (b=bf; b; b=bf) {bf=b->chain;b->chain=(Bond *)NULL;}
    unlock_evtq();
  }
}
/*-------------------- obj_transform --------------------*/

mat_4 obj_transform(Node *n)
{
/*  get current modelling matrix for display object associated with n */
/*  return identity, if n is not associated with a display object. */

  return(dsp_transform(assocobj(n)));
}
/*---------------------get_tree_matrices --------------------*/

void get_tree_matrices(Dspobj *root,char *type)
{
/* get current modelling matrix for all global display objects in 
   tree rooted at 'root'. Type may be "GMMx" to get MVT's, or
   "GRTP", to retrieve Rotation, translation and pivot. 

   NOTE: GMMx matrices are transposed MVT's off the OpenGL Stack (OpenGL
   uses row vectors, which are multiplied from the my MVT to give transformed
   row vectors:   x' = x * M). The last column in a GMMx matrix in the
   client contains the translation vector, 0 in the last row, 1 in M[3][3].
   GMMx matrices contain the complete transformation matrix for the GLBL nodes.
   In contrast, GRTP matrices contain the partial ('local') transformation
   matrix, with the rotation matrix as in a GMMx, but with the translation in the
   last row (not column), and the pivot in the last column. [Not nice, but it
   can't be fixed without breaking compatibility with old server or old
   session files].

   If the server version is < 2.2, GRTP requests will set the matrices
   to ident_4;

*/

  int i,j,n,m,size;
  Dspobj *d;
  Qmsg *Msg;
  int *immx;

  if (!is_dsp(root)) return;
  if (!streq(type,"GMMx")&&!streq(type,"GRTP")) {
    err_set(1,"get_tree_matrices, invalid matrix type: %s",type);
    return;
  }

  /* count globals in tree */

  n=0;
  if (root->dsptype==DSP_GLBL) {
    n++;
    root->mtrx=ident_4();
  }
  ForDspobjs(d,root) {
    if (d->dsptype==DSP_GLBL) {
      n++;
      d->mtrx=ident_4();
    }
  }

  if (n>0&&streq(type,"GRTP")&&graphix>0&&SrvVersion()<2.2) {
    err_set(1,"get_tree_matrices, matrix type %s not supported by server version %0.2f",type,SrvVersion());
    return;
  }

  if (n>0) {
    size=n*sizeof(d);
    send_bheader(graphix,type,size);
    if (root->dsptype==DSP_GLBL) send_bptr(graphix,root);
    ForDspobjs(d,root) {
      if (d->dsptype==DSP_GLBL) send_bptr(graphix,d);
    }
    trflush(graphix);
    Msg=waitqmsg("GMMx GRTP SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection.");
      }
    }else{
      graphics_down("no response from server.");
    }
    m=(Msg ? Msg->size/(16*sizeof(int)) : 0);
    if (m>n) m=n;
    if (m>0) {
      immx=(int *)Msg->msg;
      if (immx) {
	if (m>0 && root->dsptype==DSP_GLBL) {
	  for (i=0;i<4;i++) {     /* note: transposition takes place */
	    for (j=0;j<4;j++) root->mtrx.r[j].c[i]=MY_f(*immx++);
	  }
	  m--;
	}
	ForDspobjs(d,root) {      /* note: transposition takes place */
	  if (m == 0) break;
	  if (d->dsptype==DSP_GLBL) {
	    for (i=0;i<4;i++) {
	      for (j=0;j<4;j++) d->mtrx.r[j].c[i]=MY_f(*immx++);
	    }
	    m--;
	  }
	}
      }
    }
    delqmsg(Msg);
  }
}

/*-------------------- dsp_transform_grfx --------------------*/

mat_4 dsp_transform_grfx(Dspobj *obj)
{
/*  get current modelling matrix for display object n */

  mat_4 mmx;
  int i,j,immx[4][4];
  Qmsg *Msg;

  if (is_dsp(obj)) {
    send_bheader(graphix,"GMMx",sizeof(obj));
    send_ptr(graphix,obj);
    Msg=waitqmsg("GMMx SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }
    }else{
      graphics_down("no response from server.");
    }
    if (Msg) {
      if (Msg->size==16*sizeof(int)) {
	readqmsg(Msg,(char *)immx);
	for (i=0;i<4;i++) {
	  for (j=0;j<4;j++) {
	    mmx.r[j].c[i]=MY_f(immx[i][j]);
	  }
	}
      }else{
	delqmsg(Msg);
	mmx=ident_4();
      }
    }else{
      mmx=ident_4();
    }
  }else{
    mmx=ident_4();
  }
  return(mmx);
}
/*-------------------- getThreeDprj --------------------*/

PrjTS getThreeDPrj(void)
{
  PrjTS r;
  Qmsg *Msg;
  int *repl;

  if (graphix>0) {
    send_header(graphix,"GPRJ",0);
    Msg=waitqmsg("GPRJ SCCN");
    if (Msg) {
      if (strneq(Msg->type,"SCCN",4)) {
	delqmsg(Msg);
	Msg=(Qmsg *)NULL;
	graphics_down("Server closed connection...");
      }
    }else{
      graphics_down("no response from server.");
    }
  }else Msg=(Qmsg *)NULL;
  if (Msg) {
    if (Msg->size>=sizeof(PrjTS)) {
      repl=(int *)(Msg->msg);
      r.viewmode=MY_i(repl[0]);
      r.stereomode=MY_i(repl[1]);
      r.ysize=MY_f(repl[2]);
      r.znear=MY_f(repl[3]);
      r.slabwidth=MY_f(repl[4]);
      r.viewdist=MY_f(repl[5]);
      r.eyesep=MY_f(repl[6]);
      r.aspect=MY_f(repl[7]);
      r.scrysize=MY_f(repl[8]);
      r.sep=MY_f(repl[9]);
    }else{
      err_set(1,"Error: wrong size of GPRJ reply.");
      r=ThreeDPrj;
    } 
    delqmsg(Msg);
  }else{
    r=ThreeDPrj;
  }
  ThreeDPrj=r;
  return(r);
}
/*-------------------- setThreeDPrj --------------------*/

void setThreeDPrj(PrjTS *prj)
{
  send_bheader(graphix,"SPRJ",10*sizeof(int));
  send_bint(graphix,prj->viewmode);
  send_bint(graphix,prj->stereomode);
  send_bfloat(graphix,prj->ysize);
  send_bfloat(graphix,prj->znear);
  send_bfloat(graphix,prj->slabwidth);
  send_bfloat(graphix,prj->viewdist);
  send_bfloat(graphix,prj->eyesep);
  send_bfloat(graphix,prj->aspect);
  send_bfloat(graphix,prj->scrysize);
  send_bfloat(graphix,prj->sep);
  trflush(graphix);

  ThreeDPrj= *prj;
}

/*-------------------- new_dsp_obj --------------------*/

Dspobj *new_dsp_obj(Dspobj *parent,char *name,int type)
{
  Dspobj *son;
  Qmsg *Msg;

  if (parent) {
    if (type!=DSP_ROOT) {
      son=(Dspobj *)add_node((Node *)parent,name,OBJ_DSP);
      if (son) son->dsptype=type;
    }else{
      son=parent;
    }
    if (son) {
      if (graphix) {
	send_bheader(graphix,"AOBJ",2*sizeof(Dspobj *)+sizeof(type));
	send_bptr(graphix,parent);
	send_bptr(graphix,son); 
	send_bint(graphix,type);
	trflush(graphix);
	Msg=waitqmsg("OBJ* SCCN"); 
	if (Msg) {
	  if (strneq(Msg->type,"SCCN",4)) {
	    son->cross=cPtrNULL();;
	    delqmsg(Msg);
	    Msg=(Qmsg *)NULL;
	    graphics_down("Server closed connection...");
	  }
	  if (Msg) {
	    son->cross=cPtrNULL();
	    if (Msg->size>=cPtrSize()) memcpy(&(son->cross),Msg->msg,cPtrSize());
	    if (!cPtr(son->cross)) {
	      err_set(1,"new_dsp_obj: server rejected %s",(name ? name : "(no name)"));
	    }
	    delqmsg(Msg);
	    Msg=(Qmsg *)NULL;
	  }
	}else{
	  son->cross=cPtrNULL();
	  graphics_down("no response from server.");
	}
      }else{
	son->cross=cPtrNULL();
      }
    }else{
      err_set(1,"failed to add display object...");
    }
  }else{
    son=(Dspobj *)NULL;
  }
  return(son);
}

void set_dsp_status(Dspobj *obj,int status)
{
  if (is_dsp(obj)&&(obj->status!=status)) {
    obj->status=status;
    send_bheader(graphix,"DSPO",sizeof(int)+sizeof(Dspobj *));
    send_bint(graphix,obj->status&OBJ_FLG_DSP?1:0);
    send_bptr(graphix,obj);
  }
}

/*-------------------- send_newx --------------------*/

void send_newx(Nodechain *c)
{
  int n;
  Atom *a;

  if (graphix<1) return;
  if (!mvld(c)) return;

  for (n=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) if (cPtr(a->cross)) n++;
  if (n>0) {
    send_bheader(graphix,"NEWX",n*(3*sizeof(float)+cPtrSize()));
    for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
      if (cPtr(a->cross)) {
	send_bcptr(graphix,a->cross);
	send_bfloata(graphix,3,a->x);
      }
    }
    trflush(graphix);
  }
}
/*-------------------- send_matrix --------------------*/

void send_dsp_matrix_type(Dspobj *o,mat_4 *mtrx,char *type)
{
  if (graphix<1) return;
  if (!is_dsp(o)) return;
  if (!cPtr(o->cross)) return;
  if (!mtrx) return;

  if (streq("SRTP",type)) send_bheader(graphix,type,sizeof(o)+16*sizeof(int));
  else {
    err_set(1,"send_matrix: invalid matix type: %s",type);
    return;
  }
  send_bptr(graphix,o);
  send_bmat4(graphix,mtrx);
}

/*-------------------- transform_object --------------------*/

void transform_object(Dspobj *obj,mat_4 *tra)
{
  /* send transformation to graphics object ... */

  if (graphix<1) return;
  if (!is_dsp(obj)) return;
  if (!cPtr(obj->cross)) return;
  if (!tra) return;

  send_bheader(graphix,"TOBJ",sizeof(obj)+16*sizeof(int));
  send_bptr(graphix,obj);
  send_bmat4(graphix,tra);
}

/*-------------------- lock_evtq / unlock_evtq --------------------*/

static int evtq_state=0;

void lock_evtq(void)
{
  if (evtq_state==0) send_message(graphix,"LOCK",8,"EVTQ=ON ");
  evtq_state++;
}

void unlock_evtq(void)
{
  evtq_state=(evtq_state > 0 ? evtq_state-1 : 0 );
  if (evtq_state==0) send_message(graphix,"LOCK",8,"EVTQ=OFF");  
}
/*-------------------- mark_atom_chain --------------------*/

void mark_atom_chain(Nodechain *c)
{
  int n;
  Atom *a;

  if (graphix>0) {
    for (n=0,a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) if (cPtr(a->cross)) n++;
    if (n>0) {
      send_bheader(graphix,"ASMK",n*cPtrSize());
      for (a=(Atom *)c->s;mvld(a);a=(Atom *)a->chain) {
	if (cPtr(a->cross)) send_bcptr(graphix,a->cross);
      }
      trflush(graphix);
    }
  }
}

/*-------------------- send_molecule --------------------*/

void send_molecule(Node *root,Mol *mol)
{
  /* send molecule, creating display nodes (global and mdl) in root */

  Dspobj *droot,*global;

  droot=glblobj(assocobj(root));
  global=new_dsp_obj(droot,(char *)NULL,DSP_GLBL);
  if (mvld(global)) {
    send_mdl(global,mol);
    mol->dsptr=global;
    global->root=(Node *)mol;
  }
}
/*-------------------- graphics_down --------------------*/

void graphics_down(char *txt)
{
  if (graphix>0 && server !=0) {
    if (graphix>0) {
      close_transport(graphix);
      graphix = 0;
    }
    if (server!=0) {
      close_service(server);
      server=0;
    }
    if (txt) err_set(1,"%s",txt);
  }
}
/*-------------------- clear_atom_mark --------------------*/

void clear_atom_mark(Atom *a)
{
  if (cPtr(a->cross)) {
    send_bheader(graphix,"ARMK",cPtrSize());
    send_bcptr(graphix,a->cross);
  }
}

/*-------------------- send_bmat4 --------------------*/

void send_bmat4(int chan,mat_4 *mtrx)
{
  int i,j;

  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) {
      send_bfloat(chan,mtrx->r[j].c[i]);
    }
  }
}

