#include <stdlib.h>
#include "mystdlib.h"
#include "wnpconst.h"
#include "insane.h"
#include "token.h"

#define FALSE 0

#define N_RESDEFS 20

/* Table for 20 natural amino acid residues: */

static InsaneResDef Res[N_RESDEFS] =
{ 
  {"A","ALA", 0,
   {"HN","N","HA","CA","HB#","CB"},
   {  1 , 1 ,  1 ,  1 ,   1 ,  1 , 0}
  },

  {"C","CYS", 0,
   {"HN","N","HA","CA","HB1","HB2","CB"},
   {  1 , 1 ,  1 ,  1 ,     2     ,  1 , 0}
  },

  {"D","ASP", 0,
   {"HN","N","HA","CA","HB1","HB2","CB"},
   {  1 , 1 ,  1 ,  1 ,     2     ,  1 , 0}
  },

  {"E","GLU", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HG1","HG2","CG"}, 
   {  1 , 1 ,  1 ,  1 ,     2     ,  1 ,     2     ,  1 , 0}
  },

  {"F","PHE", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HD1","HD2","CD1","CD2","HE1","HE2","CE1","CE2","HZ","CZ"},
   {  1 , 1 ,  1 ,  1 ,     2     ,  1 ,     2     ,     2     ,     2     ,     2     ,  1 ,  1 , 0}
  },

  {"G", "GLY", 0,
   {"HN","N","HA1","HA2","CA"},
   {  1 , 1 ,     2     ,  1 , 0}
  },

  {"H","HIS", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HD2","CD2","HE1","CE1"},
   {  1,  1,   1,   1,      2     ,  1,    1 ,   1 ,   1,    1 , 0}
  },

  {"I","ILE", 0,
   {"HN","N","HA","CA","HB","CB","HG11","HG12","CG1","HD1#","CD1","HG2#","CG2"},
   {  1,  1 ,  1 ,  1 ,  1 ,  1 ,      2      ,   1 ,    1,    1 ,    1 ,   1 , 0}
  },

  {"K", "LYS", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HG1","HG2","CG","HD1","HD2","CD","HE1","HE2","CE"},
   {  1 , 1,   1 ,  1 ,     2     ,  1 ,     2     ,  1 ,     2     ,  1 ,     2     ,  1 , 0}
  },

  {"L","LEU", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HG","CG","HD1#","HD2#","CD1","CD2"},
   {  1,  1,   1,   1,      2     ,  1,   1,   1,       2      ,     2     , 0}
  },  

  {"M","MET", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HG1","HG2","CG","HE#","CE"},
   {  1,  1,   1,   1,      2     ,  1,      2     ,  1 ,   1 ,  1 , 0}
  },

  {"N","ASN", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HD21","HD22","ND2"},
   {  1,  1,   1,   1,      2     ,  1 ,      2      ,  1 , 0}
  },

  {"P","PRO", 0,
   {"HA","CA","HB1","HB2","CB","HG1","HG2","CG","HD1","HD2","CD"},
   {  1,   1,      2     ,  1,      2     ,  1,      2     ,  1 , 0}
  },

  {"Q","GLN", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HG1","HG2","CG","HE21","HE22","NE2"}, 
   {  1 , 1 ,  1 ,  1 ,     2     ,  1 ,     2     ,  1 ,      2      ,  1 , 0}
  },

  {"R","ARG", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HG1","HG2","CG","HD1","HD2","CD"},
   {  1 , 1,   1 ,  1 ,     2     ,  1 ,     2     ,  1 ,     2     ,  1 , 0}
  },

  {"S","SER", 0,
   {"HN","N","HA","CA","HB1","HB2","CB"},
   {  1 , 1 ,  1 ,  1 ,     2     ,  1 , 0}
  },

  {"T","THR", 0,
   {"HN","N","HA","CA","HB","CB","HG2#","CG2"},
   {  1 , 1 ,  1 ,  1 ,  1 ,  1 ,    1 ,   1 , 0}
  },

  {"V", "VAL", 0,
   {"HN","N","HA","CA","HB","CB","HG1#","HG2#","CG1","CG2"},
   {  1 , 1,   1,   1,   1,   1,       2      ,     2     , 0}
  },

  {"W","TRP", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HD1","CD1","HE1","NE1","HZ2","CZ2","HH2","CH2","HZ3","CZ3","HE3","CE3"},
   {  1,  1,   1,   1,      2     ,  1 ,   1 ,   1 ,   1,    1,    1 ,   1 ,   1 ,   1 ,   1 ,   1,    1 ,   1 , 0}
  },

  {"Y", "TYR", 0,
   {"HN","N","HA","CA","HB1","HB2","CB","HD1","HD2","CD1","CD2","HE1","HE2","CE1","CE2"},
   {  1,  1,   1,   1,      2     ,  1,      2     ,     2     ,     2     ,     2     , 0}
  }

};
/*========================================================================*/

InsaneResDef *parseShifts(char *rec,int N,InsaneResDef *Res,int *status)
{
  /* parse shift record for a whole residue, return pointer to
     structure with parsed information */

  Parse p;
  InsaneResDef *R=(InsaneResDef *)NULL;
  char tok[256],*endptr;
  int m,n,r=0,i,j,k,l,jj;
  float shifts[MX_RESDEF_GROUPS];
  int nshifts[MX_RESDEF_GROUPS],sa[MX_RESDEF_GROUPS];

  *status=0;

  p.buf=rec; p.pos=0; p.sep=",/"; 
  p.tok=tok; p.mts=sizeof(tok);

  if (get_token(&p)) {
    for (r=0;r<N;r++) if (!strcmp(Res[r].res,p.tok)) break;
    if (r<N) {
      if (get_token(&p)) {
	Res[r].nr=strtol(p.tok,&endptr,10);
	if (*endptr!='\0'||p.ltok==0) *status=p.pos;
      }
      m=0; n=0; nshifts[m]=0;
      while (*status==0&&get_token(&p)) {
	if (strcmp("na",p.tok)) {
	  shifts[n]=strtod(p.tok,&endptr);
	  if (*endptr!='\0'||p.ltok==0) *status=p.pos;
	}else{
	  shifts[n]=WNP_NO_SHIFT;
	  sa[n]=ATM_STASS_NO;
	}
	nshifts[m]++;
	if (nshifts[m]==1) {
	  if (p.term=='/') sa[n]=ATM_STASS_YES ;  /* is assignment stereo specific ? */
	  else if (p.term=='?') sa[n]=ATM_STASS_GUESS;
	  else if (p.term==',') sa[n]=ATM_STASS_UNK;
	  else sa[n]=ATM_STASS_NO;
	}else sa[n]=sa[n-1];
	n++;
	if (p.term==' ') {
	  m++;
	  nshifts[m]=0;
	}
      }
    }else *status=-1;
    if (!*status) {                   /* finished parsing... store info. */
      for (i=0;i<MX_RESDEF_ATMS;i++) {
	Res[r].ppm[i]=WNP_NO_SHIFT;
	Res[r].sa[i]=ATM_STASS_NO;
      }
      for (k=0,l=0,m=0;Res[r].natm[m]>0;m++) {
	if (nshifts[m]==0) break;
	for (j=0,jj=0;j<Res[r].natm[m];j++) {
	  Res[r].ppm[k+j]=shifts[l+jj];
	  Res[r].sa[k+j]=sa[l+jj];
	  if (jj<nshifts[m]-1) jj++;
	}
	k+=Res[r].natm[m];
	l+=nshifts[m];
      }
      if (nshifts[m]>0) *status=p.pos;
    }
  }else *status=-1;
  R=(*status?NULL:Res+r);
  return(R);
}
/*-------------------- parsedResDef --------------------*/

InsaneResDef parsedResDef(char *rec,int *status)
{
  InsaneResDef *R;

  R=parseShifts(rec,N_RESDEFS,Res,status);
  if (*status) R=Res+0;
  return(*R);
}
/*-------------------- storeResDef --------------------*/

int storeResDef(Mol *root,InsaneResDef *R)
{
  int k,n,ass,chg;
  Mol *m,*r;
  Atom *a;
  char rname[512];

  /* store parsed shift info in molecular structure */
  /* return number of changed shifts (-1 if no atom matched) */

  ass=0; chg=0;
  sprintf(rname,"%s_%d",R->res3,R->nr);
  for (n=0,k=0;R->natm[k]>0;k++) n+=R->natm[k];
  forMols(m,root) {
    ForMols(r,m) {
      if (strceq(r->name,rname)) {
	for (k=0;k<n;k++) {
	  if (R->ppm[k]>0.9*WNP_NO_SHIFT) {
	    forAtoms(a,r) {
	      if (wild_xmatch(a->name,R->atm[k])) {
		if (!TOLEQ(a->shift,R->ppm[k],5.e-6)||a->stass!=R->sa[k]) chg++;
		a->shift=R->ppm[k];
		a->stass=R->sa[k];
		ass++;
	      }
	    }
	  }
	}
      }
    }
  }
  return(ass>0?chg:-1);
}
