#include <string.h>
#include "streamio.h"
#include "strutil.h"
#include "mystdlib.h"
#include "parser.h"
#include "errhand.h"
#include <ctype.h>

#define MAXBUF 4096
#define MAX_LEVEL 64
#define YES 1
#define NO 0

void  prtGraphics(char *text,...);
int   read_file(char *buff,int buffsz,UCpipe *in);
int   read_xmlrpc(char *buff,int buffsz);
int   wnpRPCisOn(void);
void  releaseRPC(void);
void  blockRPC(void);

static void expvar_stream(Stream *s);
static void ignore_to_eol(Stream *s);
static char *strmkey = "eol expvar redirection comment continuation spctok";

Stream *open_stream(char *file,int buffsz)
{
  Stream *s;

  s=new_stream(buffsz);
  if (s) {
    set_stream(s,"eol=yes, redirection=off, comment=off, spctok=off, expvar=off");
    if (s->buff) {
      s->p=(Parse *)malloc(sizeof(Parse));
      if (s->p) {
        s->p->sep=(char *)NULL;
        s->p->buf=s->buff;
        s->p->mts=0;
        s->p->tok=(char *)NULL;
        s->p->pos=0;
        if (file) {
          s->in=UCpopen(file); s->own_p=TRUE;
	  s->type=STRM_PIPE;
          if (!s->in) s=pop_stream(s);
        }
      }else{
        s=pop_stream(s);
      }
    }else{
      s=pop_stream(s);
    }
  }
  return(s);
}

int set_stream_pipe(Stream *s,UCpipe *p)
{
  int error=FALSE;

  /* assign UC pipe to stream     */
  /* return TRUE operation failed */

  if (s->type==STRM_NULL) {
    s->in=p;
    s->type=STRM_PIPE;
    s->own_p=FALSE;
    s->line=0;
  }else error=TRUE;
  return(error);
}

int set_stream_macro(Stream *s,void *m)
{
  int error=FALSE;

  /* assign macro to stream       */
  /* return TRUE operation failed */

  if (s->type==STRM_NULL) {
    open_macro(m);
    s->in=m;
    s->type=STRM_MACRO;
    s->own_p=FALSE;
    s->line=0;
  }else error=TRUE;
  return(error);
}

Stream *del_stream(Stream *s)
{
  /* to be used only if pipe is closed, and stream not in linked list */

  if (s) {
    if (s->buff) free(s->buff);
    if (s->p&&s->own_p) free(s->p);
    if (s->CNT) free(s->CNT);
    free(s); s=NULL;
  }
  return(s);
}

Stream *new_stream(int buffsz)
{
  Stream *s=NULL;

  s=malloc(sizeof(Stream));
  s->in=NULL;
  s->type=STRM_NULL;
  s->line=0;
  s->eof=FALSE;
  s->prv=NULL;
  s->nxt=NULL;
  s->p=NULL;
  s->own_p=FALSE;
  s->lev=0;
  s->src=SRC_NULL;
  s->buffsz=buffsz;
  if (buffsz>=0) s->buff=malloc(buffsz+1);
  else s->buff=NULL;
  if (s->buff) s->buff[0]='\0';
  s->spctok=FALSE;
  s->expvar=FALSE;
  s->CNT=NULL;
  s->EOL='\0';
  s->COM='\0';
  s->NWS='\0';

  return(s);
}

void set_stream_props(Stream *new,Stream *old)
{
  if (new&&old) {
    new->COM=old->COM;
    new->NWS=old->NWS;
    new->EOL=old->EOL;
    if (new->CNT) free(new->CNT);
    new->CNT=Strdup(old->CNT); 
    new->spctok=old->spctok;
    new->expvar=old->expvar;
  }
}

Stream *push_stream(Stream *s,char *file)
{
  Stream *new,old;

  if (s) {
    old = *s;
    if (old.lev<MAX_LEVEL) {
      new=open_stream(file,old.buffsz);
      if (new) {                         /* grab old attributes */
	*(new->p)= *(old.p);
	new->p->buf=new->buff;
	new->p->pos=0;
	new->lev=old.lev+1;
        set_stream_props(new,s);
	*s = *new;                       /* pointer square dance... */
	*new = old;
	if (old.prv) old.prv->nxt=new;
	new->nxt=s;
	s->prv=new;
	new=s;
      }
    }else{
      err_set(1,"Maximum level of indirection reached: %s\n",file);
      new=(Stream *)NULL;
    }
  }else{
    new=s;
  }
  return(new);
}

int set_stream(Stream *s,char *cmd)
{
/* set stream attributes, return 0 if bad */

  int status;
  Parse p;
  Stream *t;
  char tok[120];

  status=0;
  if (s && cmd) {
    status=1;
    p.buf=cmd; p.pos=0; p.sep="=,";
    p.mts=120; p.tok=tok;
    while(get_token(&p)) {
      if (kwd_match(tok,strmkey,tok)<0) {
	status=0;
	break;
      }
      if (strcmp(tok,"eol")==0) {
	if (p.term!='=') {
	  for (t=s;t;t=t->nxt) t->EOL=NO;
	}else{
	  if (!get_token(&p)) break;
	  if (kwd_match(tok,"on yes",tok)>=0) {
	    for (t=s;t;t=t->nxt) t->EOL=YES;
	  }else if (kwd_match(tok,"off no",tok)>=0) {
	    for (t=s;t;t=t->nxt) t->EOL=NO;
	  }else{
	    status=0;
	    break;
	  }
	}
      }else if (strcmp(tok,"continuation")==0) {
	if (p.term!='=') {
	  if (s->CNT) free(s->CNT);
	  s->CNT=(char *)NULL; 
	}else{
	  if (!get_token(&p)) break;
	  if (s->CNT) free(s->CNT);
	  s->CNT=(char *)malloc(p.ltok+2);
	  if (s->CNT) sprintf(s->CNT,"%s\n",tok); 
	  else{
	    status=0;
	    break;
	  }
	}
      }else if (strcmp(tok,"redirection")==0) {
	if (p.term!='=') {
	  status=0;
	  break;
	}else{
	  if (!get_token(&p)) break;
	  if (p.quoted) {
	    if (p.ltok==1) {
	      for (t=s;t;t=t->nxt) t->NWS=tok[0];
	    }else{
	      status=0;
	      break;
	    }
	  }else{
	    if (kwd_match(tok,"no off",tok)>=0) {
	      for (t=s;t;t=t->nxt) t->NWS=NO;
	    }else{
	      status=0;
	      break;
	    }
	  }
	}
      }else if (strcmp(tok,"comment")==0) {
	if (p.term!='=') {
	  status=0;
	  break;
	}else{
	  if (!get_token(&p)) break;
	  if (p.quoted) {
	    if (p.ltok==1) {
	      for (t=s;t;t=t->nxt) t->COM=tok[0];
	    }else{
	      status=0;
	      break;
	    }
	  }else{
	    if (kwd_match(tok,"no off",tok)>=0) {
	      for (t=s;t;t=t->nxt) t->COM=NO;
	    }else{
	      status=0;
	      break;
	    }
	  }
	}
      }else if (strcmp(tok,"spctok")==0) {
	if (p.term!='=') {
	  status=0;
	  break;
	}else{
	  if (!get_token(&p)) break;
	  if (kwd_match(tok,"no off",tok)>=0) {
	    for (t=s;t;t=t->nxt) t->spctok=NO;
	  }else if (kwd_match(tok,"yes on",tok)>=0) {
	    for (t=s;t;t=t->nxt) t->spctok=YES;
	  }else{
	    status=0;
	    break;
	  }
	}
      }else if (strcmp(tok,"expvar")==0) {
	if (p.term!='=') {
	  status=0;
	  break;
	}else{
	  if (!get_token(&p)) break;
	  if (kwd_match(tok,"no off",tok)>=0) {
	    for (t=s;t;t=t->nxt) t->expvar=NO;
	  }else if (kwd_match(tok,"yes on",tok)>=0) {
	    for (t=s;t;t=t->nxt) t->expvar=YES;
	  }else{
	    status=0;
	    break;
	  }
	}
      }else{
	status=0;
	break;
      }
    }
  }
  return(status);
}

Stream *pop_stream(Stream *s)
{
  Stream *p;
  int status;
  char *fn;

  if (mvld(s)) {
    if (s->nxt) pop_stream(s->nxt);
    p=s->prv;
    if (p) {
      p->p->sep=s->p->sep;
      p->p->mts=s->p->mts;
      p->p->tok=s->p->tok;
      p->nxt=(Stream *)NULL;
      if (p->CNT) free (p->CNT);
      p->CNT=s->CNT; s->CNT=(char *)NULL;  
      p->spctok=s->spctok;
      p->expvar=s->expvar;
    }
    if (mvld(s->buff)) free(s->buff);
    if (mvld(s->CNT))  free(s->CNT); 
    if (mvld(s->p)) free(s->p);
    if (s->in&&s->type==STRM_PIPE&&s->own_p) {
      fn=((UCpipe *)s->in)->filename;
      ((UCpipe *)s->in)->filename=(char *)NULL;
      status=UCpclose(((UCpipe *)s->in));
      if (status) {
	err_set(1,"pop_stream: %s\n",fn);
	err_set(1,"pop_stream: error status (%d,%d)\n",status>>8,status&0xff);
      }
      if (fn) free(fn);
    }else if (s->type==STRM_MACRO) {
      close_macro(s->in);
    }
    if (mvld(p)) {
      *s = *p;
      free(p);
      s->nxt=(Stream *)NULL;
      if (mvld(s->prv)) s->prv->nxt=s;
      p=s;
    }
  }else{
    p=s;
  }
  return(p);
}

Stream *close_stream(Stream *s)
{
  if (s) {
    while(mvld(s->prv)) s=s->prv;
    pop_stream(s);
    free(s);
    s=NULL;
  }
  return(s);
}

char *token_stream(Stream *s,char *sep)
{
  int op;
  char *r;
  char filename[4096],fullname[4096];
  Parse p,*q;
  
  r=(char *)NULL;
  do {
    s->src=read_stream(s);
    if (!s->src) break;
    s->p->sep=sep;
    if ((r=token(s->p))==(char *)NULL) {  /* handle end-of-line */
      if (s->EOL) break;
      if (s->p->term!='\n') break;
    }else{
      if (s->COM && s->COM==r[0]) {       /* handle in-line comment */
	flush_record(s);
	r=(char *)NULL;
      }else if (s->NWS && s->NWS==r[0]){  /* handle redirection */
	if (!s->p->quoted) {
	  p = *s->p;
	  p.sep=""; p.tok=filename; p.mts=sizeof(filename); 
	  p.pos=s->p->tks-s->p->buf+1;
	  q=s->p;
	  if (get_token(&p)) {
	    if (FindFileName("$WNP_MACPATH",p.tok,fullname,sizeof(fullname))) {
	      if (push_stream(s,fullname)) {
		if (p.buf[p.pos]=='\n') p.pos++;
		q->pos=p.pos;
		r=(char *)NULL;
	      }
	    }
	  }
	}
      }else if (!s->p->quoted&&s->spctok) {
	if (streq(r,"-->")) {               /* comment */
	  if (s->p->buf==s->p->tks) flush_record(s);
	  else ignore_to_eol(s);
	  r=(char *)NULL;
	}else if (streq(r,"==>")) {         /* message */
	  if (strlen(s->p->tks+4)>0) prtGraphics("%s",s->p->tks+4);
	  if (s->p->buf==s->p->tks) flush_record(s);
	  else ignore_to_eol(s);
	  r=(char *)NULL;
	}else if (r[0]=='>') {             /* output redirection */
	  if (r[1]=='>') op=RED_APP;       /* append to named file */
	  else if (r[1]=='+') op=RED_OLD;  /* append to previous file */
	  else if (r[1]=='-') op=RED_END;  /* close file */
	  else op=RED_NEW;                 /* start new file */
	  if (op==RED_NEW||op==RED_APP) {
	    p = *s->p;
	    p.sep=""; p.tok=filename; p.mts=sizeof(filename); 
	    p.pos=s->p->tks-s->p->buf+(op==RED_APP?2:1);
	    if (!get_token(&p)) filename[0]='\0';
	    s->p->pos=p.pos;
	  }
	  redirectPrtStream(op,filename);
	  r=(char *)NULL;
	}
      }
    }
  }while(!r);
  return(r);
}
    
int read_stream(Stream *s)
{
  int n,l,k;
  int exp=FALSE;
  char *line;

  if (s) {
    if (!(s->p->buf[s->p->pos])) {
      s->p->buf[0]='\0'; s->p->pos=0;
      if (s->type==STRM_PIPE) {
	n=read_file(s->buff,s->buffsz,((UCpipe *)s->in));
	if (n==SRC_NULL) {
	  if (!UCpeof(((UCpipe *)s->in))) err_set(1,"read_strean error, on file %s",((UCpipe *)s->in)->filename); 
	  s->eof=TRUE;                 /* should we pop all levels on error? */
	  if (s->lev > 0) {
	    s=pop_stream(s);
	    n=read_stream(s);
	    exp=TRUE;                  /* variables expanded, if any */
	  }
	}else{
	  s->line++;
	}
      }else if (s->type==STRM_MACRO) {
	line=read_macro(s->in,s->line);
	if (line) {
	  s->line++;
	  StrCpy(s->buff,line,s->buffsz);
	  if (strlen(line)>strlen(s->buff)) err_set(1,"read_stream: error, macro line truncated.\n");
	  n=SRC_MACRO;
	}else {
	  s->eof=TRUE;
	  if (s->lev > 0) {
	    s=pop_stream(s);
	    n=read_stream(s);
	    exp=TRUE;                  /* variables expanded, if any */
	  }else{
	    n=SRC_NULL;
	  }
	}	  
      }else if (s->type==STRM_NULL) {
	err_dump();
	n=read_null(s->buff,s->buffsz);
	if (n!=SRC_NULL) s->line++;
	else if (wnpRPCisOn()) {
	  s->type=STRM_XMLRPC;
	  n=read_xmlrpc(s->buff,s->buffsz);
	  if (n!=SRC_NULL) s->line++;
	}
      }else if (s->type==STRM_XMLRPC) {
	err_dump();
	if (!wnpRPCisOn()) s->type=STRM_NULL;
	n=read_xmlrpc(s->buff,s->buffsz);
	if (n!=SRC_NULL) s->line++;
      }else{
	n=SRC_NULL;
      }
      if (s->CNT&&s->src!=SRC_NULL) 
	if ((l=strlen(s->buff))>=(k=strlen(s->CNT)))
	  if (streq(s->buff+l-k,s->CNT)) s->buff[l-k]='\0';
      if (!exp&&s->expvar) expvar_stream(s);
    }else{
      n=s->src;
    }
  }else{
    n=SRC_NULL;
  }
  return(n);
}

static void expvar_stream(Stream *s)
{
  char *spec="\\$";
  char *p,*cpy,*v,*e,q;
  char buffer[MAXBUF];
  int l;

  /* Variable substitution: replace ${name} by value of env. variable name */
  /* $< will read 1 line from stdin (graphix) and insert it into stream    */

  cpy=strdup(s->buff);
  if (cpy) {
    p=cpy; l=0;
    for (;*p;p++) {
      if (*p=='\\') {
	if (strchr(spec,p[1])) p++;     /* quoted special char */
	if (l<s->buffsz) s->buff[l++]=*p;
      }else if (*p=='$') {
	if (p[1]=='{') {
	  v=p+2;
	  p=strchr(v,'}');
	  if (p) {
	    *p='\0';
	    e=getenv(v);
	    if (e) {       /* if found, copy definition into buffer */
	      for (;*e;e++) if (l<s->buffsz) s->buff[l++]=*e;
	    }
	  }else{
	    p=v-2;
	    if (l<s->buffsz) s->buff[l++]=*p;
	  }
	}else{
	  for (v=p+1;*v;v++) if (!isalnum(*v)&&*v!='_'&&*v!='<') break;
	  if (v>p+1) {
	    q=*v; *v='\0';
	    if (streq(p+1,"<")) {
	      *buffer='\0';
	      blockRPC();
	      read_null(buffer,MAXBUF);
	      releaseRPC();
	      if (strlen(buffer)>0) {
		e=buffer;
		if (*(e+strlen(e)-1)=='\n') *(e+strlen(e)-1)='\0';
	      	if (*(e+strlen(e)-1)=='\r') *(e+strlen(e)-1)='\0';
	      }else e=(char *)NULL;
	    }else{
	      e=getenv(p+1);
	    }
	    if (e) {       /* if found, copy definition into buffer */
	      for (;*e;e++) if (l<s->buffsz) s->buff[l++]=*e;
	      p=v-1;
	    }else{
	      if (l<s->buffsz) s->buff[l++]=*p;
	    }
	    *v=q;
	  }else{
	    if (l<s->buffsz) s->buff[l++]=*p;
	  }
	}
      }else{
	if (l<s->buffsz) s->buff[l++]=*p; 
      }
    }
    s->buff[l]='\0';
  }
  if (cpy) free(cpy);
}

void flush_record(Stream *s)
{
  if (s) {
    s->p->pos=strlen(s->p->buf);
  }
}

static void ignore_to_eol(Stream *s)
{
  if (s) {
    if (strlen(s->p->buf)>0) s->p->pos=strlen(s->p->buf)-1;
    else s->p->pos=0;
  }
}
