/*   streamio.h   */

#ifndef STREAMIODEF
#define STREAMIODEF

#include <stdio.h>
#include "strutil.h"
#include "fileutil.h"

/* Source of last token parsed from input buffer of stream:                           */
#define SRC_NULL    0        /*  nothing in buffer                                    */
#define SRC_FILE    1        /*  read from a file  by nwfgets                         */
#define SRC_PICK    2        /*  PATM, POBJ or PXYZ token from graphics server        */
#define SRC_MENU    3        /*  MENU token from graphics server                      */
#define SRC_STDIN   4        /*  stdin (via nwfgets, AReadlineFunction), or XML-RPC   */
#define SRC_KEYB    5        /*  KEYB token from graphics server                      */
#define SRC_MACRO   6        /*  from macro buffer                                    */
#define SRC_XMLRPC  7        /*  from XML-RPC, through STRM_XMLRPC stream             */

/* type of stream :                                                                   */
#define STRM_NULL   0        /*  read_null function (stdin, graphics server, XML-RPC) */
#define STRM_PIPE   1        /*  read_file function (files through UCpipes)           */
#define STRM_MACRO  2        /*  read_macro function (macro buffer)                   */
#define STRM_XMLRPC 3        /*  read_xmlrpc function (input from XML-RPC server)     */

typedef struct Stream {
  struct Stream *prv;        /*  pointer to previous level    */
  struct Stream *nxt;        /*  pointer to next level        */
  void   *in;                /*  source of stream             */
  Parse  *p;                 /*  associated parsing structure */
  char   *buff;              /*  I/O buffer                   */
  char   *CNT;               /*  line continuation mark       */
  int    type;               /*  typ of source of stream      */
  int    line;               /*  line number (next line)      */
  int    eof;                /*  have we reached EOF?         */
  int    own_p;              /*  does stream own in pipe ?    */ 
  int    lev;                /*  level of I/O stream          */
  int    src;                /*  source of last token         */
  int    buffsz;             /*  buffer size                  */
  int    spctok;             /*  handle special tokens ?      */
  int    expvar;             /*  expand variables?            */
  char   EOL;                /*  report EOLs ?                */
  char   COM;                /*  in-line comment character    */
  char   NWS;                /*  new source character         */
} Stream;

/********************* Prototypes *********************/

Stream  *new_stream(int buffsz);
Stream  *del_stream(Stream *s);
Stream	*open_stream(char *file,int buffsz);
Stream  *push_stream(Stream *s,char *filename);
Stream  *close_stream(Stream *s);
Stream  *pop_stream(Stream *s);
char    *token_stream(Stream *s,char *sep);
int      read_stream(Stream *s);
int      read_null(char *buffer,int nch);
void     flush_record(Stream *s);
int      set_stream(Stream *s,char *cmd);
int      set_stream_pipe(Stream *s,UCpipe *p);
int      set_stream_macro(Stream *s,void *m);
void     set_stream_props(Stream *new,Stream *old);

/********************* Macros *************************/

#define strm_parse(x)  (x->p)
#define strm_tok(x)    (x->p->tok)
#define strm_mts(x)    (x->p->mts)
#define strm_term(x)   (x->p->term)
#define strm_ltok(x)   (x->p->ltok)
#define strm_quoted(x) (x->p->quoted )
#define strm_eof(x)    (x->eof)
#define strm_buff(x)   (x->p->buf)
#define strm_pos(x)    (x->p->pos)
#define strm_src(x)    (x->src)
#define strm_level(x)  (x->lev)
#define strm_char(x)   (x->p->buf[x->p->pos])
#define strm_flush(x)  x->p->buf[x->p->pos]='\0'

#endif
