#include <math.h>
#include "mystdlib.h"
#include "swdef.h"
#include "strutil.h"
#include "structal.h"
#include "xbhtree.h"

static char *mode=NULL;
static float gapo=3.0,gape=0.333333;
static int minbsz=0;
static float (*compare)(int i,int j,float *dA,float *dB)=ResCompare5;

static int getBlocks(int *BL,int m,SAlires *TA);
static int algAccept(int m,SAlires *TA,int n,SAlires *SA,mat_4 *Tra,int *npb,float *rmsb,
		     SAlires *ta,SAlires *sa,mat_4 tra,int *np,float rms,int nm);

/*-------------------- algBestStructal --------------------*/

int algBestStructal(int m,Atom **T,SAlires *TA,int n,Atom **S,SAlires *SA,
		       float cutoff,mat_4 *Tra,int *npb,float *rmsb)
{
  float *X,rms,rmss;
  SAlires *sa,*ta;
  mat_4 tra;
  int nm,*np,*BL=NULL,nmb,i,j,k,nbl=0;

  /* June 27 2006: new Strategy. 

     1) algStructal (alignment using local backbone
     geometry) is used only to generate starting alignment of source with
     target. The initial alignemnt is never accepted, no rmsd based reduction
     will be applied to the initial alignment.

     2) each consequtive block of aligned residues from the initial alignment
     will be used to set up an initial alignemt for the extended Gerstein
     method.

   */

  sa=SAliresCopy(n,SA);
  ta=SAliresCopy(m,TA);
  X=(float *)malloc(3*n*sizeof(float));
  if (mvld(npb)) np=(int *)malloc(n*sizeof(int));
  else np=(int *)NULL;
  if (mode) BL=malloc(3*m*sizeof(int));
 
  *Tra=ident_4();
  nmb=-1;
  rmss=0;

  if (mvld(sa)&&mvld(ta)&&mvld(X)&&(!mvld(npb)||mvld(np))) {
    for (i=0;i<m;i++) ta[i].x=T[i]->x;
    for (i=0;i<n;i++) {
      sa[i].x=X+3*i;
      for (j=0;j<3;j++) sa[i].x[j]=S[i]->x[j];
    }
    if (mode) { /* "partial" == "full" right now... */
      if ((nm=algStructal(m,T,ta,n,S,sa,cutoff,&tra,np,&rms))>0) {
	nbl=getBlocks(BL,m,ta);
	nm=algReduce(m,T,ta,n,S,sa,cutoff,&tra,np,&rms);
	nmb=algAccept(m,TA,n,SA,Tra,npb,rmsb,ta,sa,tra,np,rms,nm);
	rmss=normalizedRMS(nm,rms);	
	if (nm>0) {
	  for (i=0;i<n;i++) v32arr(sa[i].x,m4xv3(tra,vec3(S[i]->x)));
	  if ((nm=algGerstein(m,T,ta,n,S,sa,cutoff,&tra,np,&rms))>0) {
	    if (rmss>normalizedRMS(nm,rms)) {
	      nmb=algAccept(m,TA,n,SA,Tra,npb,rmsb,ta,sa,tra,np,rms,nm);
	      rmss=normalizedRMS(nm,rms);
	    }
	  }
	}
	for (k=0;k<nbl;k++) {
	  if (BL[3*k+2]>=minbsz) {
	    rms=rms_best_mat(BL[3*k+2],S+BL[3*k+1],T+BL[3*k],&tra);
	    for (i=0;i<n;i++) v32arr(sa[i].x,m4xv3(tra,vec3(S[i]->x)));
	    if ((nm=algGerstein(m,T,ta,n,S,sa,cutoff,&tra,np,&rms))>0) {
	      if (rmss>normalizedRMS(nm,rms)) {
		nmb=algAccept(m,TA,n,SA,Tra,npb,rmsb,ta,sa,tra,np,rms,nm);
		rmss=normalizedRMS(nm,rms);
	      }
	    }
	  }
	}
      }
    }else{                     /* mode is "simple" */
      if ((nm=algStructal(m,T,ta,n,S,sa,cutoff,&tra,np,&rms))>0) {
	nmb=algAccept(m,TA,n,SA,Tra,npb,rmsb,ta,sa,tra,np,rms,nm);
	rmss=normalizedRMS(nm,rms);
	for (i=0;i<n;i++) v32arr(sa[i].x,m4xv3(*Tra,vec3(S[i]->x)));
	if ((nm=algGerstein(m,T,ta,n,S,sa,cutoff,&tra,np,&rms))>0) {
	  if (rmss>normalizedRMS(nm,rms)) {
	    nmb=algAccept(m,TA,n,SA,Tra,npb,rmsb,ta,sa,tra,np,rms,nm);
	    rmss=normalizedRMS(nm,rms);
	  }
	}
      }
    }
  }
  if (mvld(sa)) free(sa);
  if (mvld(ta)) free(ta);
  if (mvld(X))  free(X);
  if (mvld(np)) free(np);
  if (mvld(BL)) free(BL);
  return(nmb);
}
/*-------------------- algStructal --------------------*/

int algStructal(int m,Atom **T,SAlires *TA,int n,Atom **S,SAlires *SA,
		float cutoff,mat_4 *Tra,int *np,float *rmsd)
{
  int i,nm;
  Atom **s,**t;
  float score;
  
  s=(Atom **)malloc(n*sizeof(Atom *));
  t=(Atom **)malloc(m*sizeof(Atom *));

  *Tra=ident_4();
  if (mvld(s)&&mvld(t)) {
    if (!salign(m,TA,n,SA,compare,gapo,gape,&score)) {
      for (nm=0,i=0;i<m;i++) {
	if (TA[i].pos>=0) {
	  t[nm]=T[i];
	  s[nm]=S[TA[i].pos];
	  nm++;
	}
      }
      *rmsd=rms_best_mat(nm,s,t,Tra);
      if (!mode) {
	nm=algReduce(m,T,TA,n,S,SA,cutoff,Tra,np,rmsd);
      }
    }else nm=-1;
  }else nm=-1;
  if (mvld(t)) free(t);
  if (mvld(s)) free(s);
  return(nm);
}
/*-------------------- algGerstein --------------------*/

int algGerstein(int m,Atom **T,SAlires *TA,int n,Atom **S,SAlires *SA,
		float cutoff,mat_4 *Tra,int *np,float *rmsd)
{
  int i,j,k,c,eb,nm,*ba;
  float score,rms,rmsb=0,rmss,rmssb=0;
  Atom **s,**t;
  mat_4 tra,trab;
  
  s=(Atom **)malloc(n*sizeof(Atom *));
  t=(Atom **)malloc(m*sizeof(Atom *));
  ba=(int *)malloc(m*sizeof(int));

  nm=-1;
  *Tra=ident_4();
  trab=ident_4();
  if (mvld(s)&&mvld(t)&&mvld(ba)) {
    for (i=0;i<m;i++) ba[i]=-1;
    for (eb=0,c=0;c<4&&eb<500;eb++) {
      if (dynGersteinX(m,TA,n,SA,&score)) break;
      for (nm=0,i=0;i<m;i++) {
	if (TA[i].pos>=0) {
	  t[nm]=T[i];
	  s[nm]=S[TA[i].pos];
	  nm++;
	}
      }
      rms=rms_best_mat(nm,s,t,&tra);
      rmss=normalizedRMS(nm,rms);
      for (i=0;i<n;i++) {
	for (j=0;j<3;j++) {
	  for (SA[i].x[j]=tra.r[j].c[3],k=0;k<3;k++) {
	    SA[i].x[j]+=tra.r[j].c[k]*S[i]->x[k];
	  }
	}
      }
      if (c==0||rmss<rmssb-0.000001) {
	for (i=0;i<m;i++) ba[i]=TA[i].pos;
	rmssb=rmss;
	rmsb=rms;
	trab=tra;
	c=1;
      }else c++;
    }
    if (c>0) {
      for (i=0;i<n;i++) SA[i].pos=-1;
      for (i=0;i<m;i++) {
	TA[i].pos=ba[i];
	if (ba[i]>=0) SA[ba[i]].pos=i;
      }
      *Tra=trab;
      *rmsd=rmsb;
      nm=algReduce(m,T,TA,n,S,SA,cutoff,Tra,np,rmsd);
    }else nm=-1;
  }else nm=-1;
  if (mvld(ba)) free(ba);
  if (mvld(t)) free(t);
  if (mvld(s)) free(s);
  return(nm);
}
/*-------------------- algReduce --------------------*/

int algReduce(int m,Atom **T,SAlires *TA,int n,Atom **S,SAlires *SA,
	      float cutoff,mat_4 *Tra,int *ba,float *rmsd)
{
  int nm=0,i,j;
  float score;
  Atom **s=(Atom **)NULL,**t=(Atom **)NULL;;

  if (ba) {
    s=(Atom **)malloc(n*sizeof(Atom *));
    t=(Atom **)malloc(m*sizeof(Atom *));
    nm=0;
    if (mvld(s)&&mvld(t)) {
      for (i=0;i<n;i++) if (SA[i].pos>=0) ba[nm++]=i;
      if (*rmsd>0) {
	for (j=0;j<3;j++) {
	  for (nm=0,i=0;i<m;i++) {
	    if (TA[i].pos>=0) {
	      score=v3l(v3mv3(m4xv3(*Tra,vec3(S[TA[i].pos]->x)),vec3(T[i]->x)));
	      if (score<=(1+0.5*j)*(*rmsd)) {
		t[nm]=T[i];
		s[nm]=S[TA[i].pos];
		ba[nm]=TA[i].pos;
		nm++;
	      }
	    }
	  }
	  *rmsd=rms_best_mat(nm,s,t,Tra);
	}
      }
      if (*rmsd>0&&cutoff<0) cutoff=-cutoff*(*rmsd);
      if (cutoff>0) {
	for (nm=0,i=0;i<m;i++) {
	  if (TA[i].pos>=0) {
	    score=v3l(v3mv3(m4xv3(*Tra,vec3(S[TA[i].pos]->x)),vec3(T[i]->x)));
	    if (score<=cutoff) {
	      t[nm]=T[i];
	      s[nm]=S[TA[i].pos];
	      ba[nm]=TA[i].pos;
	      nm++;
	    }
	  }
	}
	*rmsd=rms_best_mat(nm,s,t,Tra);
      }
    }else nm=-1;
  }else{
    for (i=0;i<n;i++) if (SA[i].pos>=0) nm++;
  }
  if (mvld(s)) free(s);
  if (mvld(t)) free(t);
  return(nm);
}
/*-------------------- algExpand --------------------*/

int algExpand(int m,Atom **T,SAlires *TA,int n,Atom **S,SAlires *SA,
	      float cutoff,mat_4 *Tra,int *np,float *rmsd)
{
  float score;
  SAlires *sa,*ta;
  int nm=0,i,j,bn,*bs,msz;

  msz=4;

  sa=SAliresCopy(n,SA);
  ta=SAliresCopy(m,TA);
  bs=(int *)malloc((n+1)*sizeof(int));
  if (mvld(sa)&&mvld(ta)&&mvld(bs)) {
    for (i=0;i<n;i++) sa[i].pos=-1;
    for (i=0;i<m;i++) ta[i].pos=-1;
    /* find alignment blocks in input alignment: 
       maxdist = cutoff, min number of consequtive pairs: msz.
    */
    for (bn=0,i=0;i<n;bn=bs[i],i++) {
      if (SA[i].pos>=0) {
	if (bn==0||SA[i-1].pos==SA[i].pos-1) {
	  score=v3l(v3mv3(m4xv3(*Tra,vec3(S[i]->x)),vec3(T[SA[i].pos]->x)));
	  bs[i]=(score < cutoff ? bn+1 : 0);
	}else bs[i]=0;
      }else bs[i]=0;
    }
    
    if (mvld(np)) {
      nm=0;
      for (i=0;i<n;i++) {
	if (bs[i]>msz) np[nm++]=i;
	else if (bs[i]==msz) for (j=msz-1;j>=0;j--) np[nm++]=i-j;
      }
    }
  }else nm=-1;
  if (mvld(bs)) free(bs);
  if (mvld(sa)) free(sa);
  if (mvld(ta)) free(ta);
  return(nm);
}
/*-------------------- normalizedRMS --------------------*/

float normalizedRMS(int n, float rms)
{
  return(225*rms/(n+135)+2.0*(n>200?0:pow(1.5,10-n)));
}

/*-------------------- SetStructalResCompare --------------------*/

void SetStructalResCompare(char *how)
{
  if (streq(how,"ResCompare5")) {
    compare=ResCompare5;
  }else if (streq(how,"ResCompare6")) {
    compare=ResCompare6;
  }
}

/*-------------------- SetStructalGapPenalty --------------------*/

void SetStructalGapPenalty(float o,float e)
{
  gapo=o;
  gape=e;
}

/*-------------------- SetStructalMode --------------------*/

void SetStructalMode(char *new,int n)
{
  if (mode) {free(mode);} mode=NULL;
  if (new) {
    if (!streq(new,"simple")) mode=strdup(new);
    if (streq(new,"partial")) minbsz=n;
    if (streq(new,"full")) minbsz=0;
  }
}

/*-------------------- GetStructalGapPenalty --------------------*/

void GetStructalGapPenalty(float *o,float *e)
{
  *o=gapo;
  *e=gape;
}

/*-------------------- algAccept --------------------*/

static int algAccept(int m,SAlires *TA,int n,SAlires *SA,mat_4 *Tra,int *npb,float *rmsb,
		     SAlires *ta,SAlires *sa,mat_4 tra,int *np,float rms,int nm)
{
  int i;

  if (mvld(npb)) for (i=0;i<nm;i++) npb[i]=np[i];
  for (i=0;i<n;i++) SA[i].pos=sa[i].pos;
  for (i=0;i<m;i++) TA[i].pos=ta[i].pos;
  *Tra=tra;
  *rmsb=rms;
  return(nm);
}

/*-------------------- getBlocks --------------------*/

static int getBlocks(int *BL,int m,SAlires *TA)
{
  int i;
  int nbl=0;

  if (BL) {
    for (i=0;i<m;i++) {
      if (TA[i].pos>=0) {
	BL[3*nbl]=i;
	BL[3*nbl+1]=TA[i].pos;
	BL[3*nbl+2]=1;
	for (i++;i<m;i++) {
	  if (TA[i].pos==BL[3*nbl+1]+BL[3*nbl+2]) BL[3*nbl+2]++;
	  else break;
	}
	nbl++;
	i--;
      }
    }
  }else nbl=0;
  return(nbl);
}

