#include <stdio.h>
#include <stdlib.h>
#include "object.h"
#include "swdef.h"
#include "strutil.h"

#define FLG1 (1<<1)
#define FLG2 (1<<2)

#define neighb(a,i) (a->bnd[i]->f==a?a->bnd[i]->t:a->bnd[i]->f)

static int candList(Atom *a,int nb, Atom **B,int *cand)
{
  int j,k,l,nt,nc;
  Atom *t;
  char chs[256][MX_OBJ_CHS];
  int cnt[256],cntb[256];

  nt=0; nc=0;
  for (k=0;k<a->nbnd;k++) {
    t=neighb(a,k);
    if (t->flg&FLG1) {
      for (l=0;l<nt;l++) if (streq(t->atp->chs,chs[l])) break;
      if (l<nt) cnt[l]+=1;
      else {
	cnt[nt]=1;
	strcpy(chs[nt],t->atp->chs);
	nt+=1;
      }
    }
  }
  for (j=0;j<nb;j++) {
    if (streq(a->atp->chs,B[j]->atp->chs)&&!B[j]->chain) {
      for (l=0;l<nt;l++) cntb[l]=cnt[l];
      for (k=0;k<B[j]->nbnd;k++) {
	t=neighb(B[j],k);
	if (t->flg&FLG1) {
	  for (l=0;l<nt;l++) if (streq(t->atp->chs,chs[l])) break;
	  if (l<nt) cntb[l]-=1;
	}
      }
      for (l=0;l<nt;l++) if (cntb[l]>0) break;
      if (l==nt) {
	cand[nc]=j;
	nc++;
      }
    }
  }
  if (nc==1) {
    a->chain=(Node *)B[cand[0]];
    B[cand[0]]->chain=(Node *)a;
  }
  return(nc);
}

static int refineLists(int na,Atom **A,int nb, Atom **B,int *nc, int **lists)
{
  int i,j,k,l,m,nm,n,D=0;
  Atom *a=NULL,*b,*c;

  do {
    nm=nb;
    a=NULL;
    for (i=0;i<na;i++) {
      if (nc[i]>0 && !A[i]->chain) {
	if (nc[i]==0) A[i]->flg|=FLG2;
	n=0;
	for (l=0;l<nc[i];l++) {
	  if (!B[lists[i][l]]->chain||B[lists[i][l]]->chain==(Node *)A[i]) {
	    b=B[lists[i][l]];
	    for (k=0;k<b->nbnd;k++) {
	      c=neighb(b,k);
	      if (c->chain && !find_bond(A[i],(Atom *)c->chain)) break;
	    }
	    if (k==b->nbnd) {
	      lists[i][n]=lists[i][l];
	      n++;
	    }
	  }
	}
	if (n<nc[i]) {
	  D+=nc[i]-n;
	  A[i]->flg&=~FLG2;
	  nc[i]=n;
	}
	if (n==1) {
	  A[i]->chain=(Node *)B[lists[i][0]];
	  B[lists[i][0]]->chain=(Node *)A[i];
	}
	if (!(A[i]->flg&FLG2)&&nc[i]<nm) {
	  nm=nc[i];
	  a=A[i];
	}
      }
    }

    if (a) {
      a->flg|=FLG2;
      i=a->seqnr;
      for (k=0;k<a->nbnd;k++) {
	b=neighb(a,k);
	if (b->flg&FLG1) {
	  if (b->chain) {
	    n=0;
	    for (l=0;l<nc[i];l++) {
	      if (find_bond(B[lists[i][l]],(Atom *)b->chain)) {
		lists[i][n]=lists[i][l];
		n++;
	      }
	    }
	    if (n<nc[i]) {
	      D+=nc[i]-n;
	      a->flg&=~FLG2;
	      nc[i]=n;
	      if (nc[i]==1) {
		a->chain=(Node *)B[lists[i][0]];
		B[lists[i][0]]->chain=(Node *)a;
	      }
	    }
	  }else{
	    n=0;
	    j=b->seqnr;
	    for (l=0;l<nc[j];l++) {
	      for (m=0;m<nc[i];m++) if (!B[lists[j][l]]->chain&&find_bond(B[lists[j][l]],B[lists[i][m]])) break;
	      if (m<nc[i]) {
		lists[j][n]=lists[j][l];
		n++;
	      }
	    }
	    if (n<nc[j]) {
	      D+=nc[j]-n;
	      b->flg&=~FLG2;
	      nc[j]=n;
	      if (nc[j]==1) {
		b->chain=(Node *)B[lists[j][0]];
		B[lists[j][0]]->chain=(Node *)b;
	      }
	    }
	  }
	}
      }
    }
  } while(a);
  return(D);
}

static int candLists(int na,Atom **A,int nb, Atom **B,int *nc, int **lists)
{
  int i,l,*cand=NULL;

  cand=malloc(nb*sizeof(int));

  if (cand) {
    for (i=0;i<na;i++) {
      nc[i]=candList(A[i],nb,B,cand);
      if (nc[i]>0) {
	lists[i]=malloc(nc[i]*sizeof(int));
	for (l=0;l<nc[i];l++) lists[i][l]=cand[l];
      }else lists[i]=(int *)NULL;
    }
    refineLists(na,A,nb,B,nc,lists);
  }else{
    for (i=0;i<na;i++) {
      nc[i]=0;
      lists[i]=(int *)NULL;
    }
  }
  
  if (cand) free(cand);

  for (i=0;i<na;i++) if (nc[i]==0) break;
  return (i==na);
}

static void save_state(int na,Atom **A,int nb, Atom **B,int *nc, int **lists,int **ncs,int ***listss)
{
  int i,k;

  if (*ncs) free(*ncs);
  if (*listss) {
    for (i=0;i<na;i++) if ((*listss)[i]) free((*listss)[i]);
  }
  *ncs=malloc(na*sizeof(int));
  *listss=malloc(na*sizeof(int *));
  for (i=0;i<na;i++) {
    if (nc[i]>0) {
      (*listss)[i]=malloc(nc[i]*sizeof(int));
      for (k=0;k<nc[i];k++) (*listss)[i][k]=lists[i][k];
      A[i]->flg&=~FLG2;
      if (A[i]->chain) {
	A[i]->chain->chain=NULL;
	A[i]->chain=NULL;
      }
    }else{
      (*listss)[i]=NULL;
    }
    (*ncs)[i]=nc[i];
  }
}

static void restore_state(int na,Atom **A,int nb, Atom **B,int *nc, int **lists,int *ncs,int **listss)
{
  int i,k;

  for (i=0;i<na;i++) {
    if (ncs[i]>0) {
      for (k=0;k<ncs[i];k++) lists[i][k]=listss[i][k];
      A[i]->flg&=~FLG2;
      if (A[i]->chain) {
	A[i]->chain->chain=NULL;
	A[i]->chain=NULL;
      }
    }
    nc[i]=ncs[i];
  }
}


static int backtracker(int na,Atom **A,int nb, Atom **B,int *nc, int **lists)
{
  int i,im,k,nm,state;
  int *ncs=NULL,**listss=NULL;

  im=0;
  nm=nb+1;
  for (i=0;i<na;i++) {
    if (nc[i]>1&&nc[i]<nm) {
      nm=nc[i];
      im=i;
    }else if (nc[i]<1) {
      return FALSE;
    }
  }
  if (nm<=nb) {
    save_state(na,A,nb,B,nc,lists,&ncs,&listss);
    for (k=0;k<ncs[im];k++) {
      nc[im]=1;
      lists[im][0]=listss[im][k];
      refineLists(na,A,nb,B,nc,lists);
      if (backtracker(na,A,nb,B,nc,lists)) break;
      restore_state(na,A,nb,B,nc,lists,ncs,listss);
    }
  }

  if (ncs) free(ncs);
  if (listss) {
    for (i=0;i<na;i++) if (listss[i]) free(listss[i]);
    free(listss);
  }

  state=TRUE;
  for (i=0;i<na;i++) state=state&&nc[i]==1;
  
  return(state);
}

/*------------------------------------------------------------------------*/

int findSubstructure(int na,Atom ** A,int nb, Atom **B)
{
  int i,status=0;
  int **cand=NULL,*nc=NULL;

  if (na>nb||na==0) return 0;

  for (i=0;i<na;i++) {
    A[i]->flg|=FLG1;
    A[i]->flg&=~FLG2;
    A[i]->seqnr=i;
  }
  for (i=0;i<nb;i++) {
    B[i]->flg|=FLG1;
  }
  
  cand=malloc(na*sizeof(int *));
  nc=malloc(na*sizeof(int));

  if (cand&&nc) {
    if (candLists(na,A,nb,B,nc,cand)) {
      if (backtracker(na,A,nb,B,nc,cand)) {
	status=na;
      }else{
	status=0;
      }
    }
  }

  if (status==0) {
    for (i=0;i<na;i++) {
      if (A[i]->chain) {
	A[i]->chain->chain=NULL;
	A[i]->chain=NULL;
      }
    }
  }

  for (i=0;i<na;i++) {
    A[i]->flg&=~FLG1;
    A[i]->flg&=~FLG2;
    if (cand&&cand[i]) free(cand[i]);
  }
  for (i=0;i<nb;i++) {
    B[i]->flg&=~FLG1;
  }

  if (cand) free(cand);
  if (nc) free(nc);

  return(status);
}
/*------------------------------------------------------------------------*/

int applySubstructureMatch(int na,Atom **A)
{
  int i,k,nomatch=0;
  Atom *a,*b;
  Bond *bnd;

  for (i=0;i<na;i++) {
    if (A[i]->chain) {
      a=A[i];
      b=(Atom *)A[i]->chain;
      set_atom_type(a,b->atp);
      a->chrg=b->chrg;
      for (k=0;k<a->nbnd;k++) {
	if (a==a->bnd[k]->f) {
	  b=a->bnd[k]->t;
	  if (b->chain) {
	    bnd=find_bond((Atom *)a->chain,(Atom *)b->chain);
	    if (bnd) set_bt(a->bnd[k],bnd->bt);
	    else nomatch++;
	  }
	}
      }
    }
  }
  return(nomatch);
}
/*------------------------------------------------------------------------*/


void clearSubstructureMatch(int na,Atom **A)
{
  int i;
  
  for (i=0;i<na;i++) {
    if (A[i]->chain) {
      A[i]->chain->chain=NULL;
      A[i]->chain=NULL;
    }
  }
}
