#ifndef SURFACEDEF
#define SURFACEDEF

#include <stdio.h>
#include "object.h"

#define SURF_TYPE_UNK 0    /* undefined surface type              */
#define SURF_TYPE_TRG 1    /* triangulated surface                */
#define SURF_TYPE_MSH 2    /* mesh of (x,y),(y,z),(z,x) countours */

typedef struct SPoint {
  float x[3],n[3];
} SPoint;

typedef struct STriangle {
  int v[3];
} STriangle;

typedef struct Surface {
  struct SPoint *v;
  struct STriangle *t;
  int type,color,lw,style,pattern,bckclr;
  int nv,nt;
} Surface;

Surface *newSurface(int type);
Surface *freeSurface(Surface *S);
Surface *readSurface(FILE *f);
int      writeSurface(FILE *f,Surface *S);
void     SurfAndVol(Surface *S,float *surf,float *vol);
Dspobj  *addSurfaceToDisplay(Surface *S,Dspobj *glob,char *name);
void     surfaceNormals(Surface *S,Grid *G);
void     smoothSurface(Surface *S,float bias,int iter);
void     surface_info(Surface *S);
int      AddTMDtoS(Surface *S,mat_4 T,float *x,float *y);
Surface *getTRGSurfaceFromServer(int *msg,int size);
Surface *getSurfaceFromServer(Dspobj *obj,int *flags);
void     clipSurfaceAtoms(Surface *S,Nodechain *c,float cutoff);
void     clipSurfaceLargest(Surface *S);
void     replaceSurface(Dspobj *obj,Surface *S);

#endif
