#include <stdio.h>

#include "object.h"
#include "swdef.h"
#include "mystdlib.h"
#include "errhand.h"
#include "parser.h"
#include "transpor.h"
#include "strutil.h"

extern int graphix;
extern Node *DefaultNode,*MolRoot;
extern Node *DspRoot;
extern Node *LibRoot;

static int current_version=2;

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)
#define key_is(a) (streq(a,key->key))

#define TRUE 1
#define FALSE 0

static void clearSession(void)
{
  /* clear everything in current session */

  while (MolRoot->fst) del_node(MolRoot->fst);
  while (DspRoot->fst) del_node(DspRoot->fst);
  clearMacros();

}

void restoreSession(UCpipe *in,char *filename)
{
  char buff[4096];
  int status=0,version=0;
  Stream *s;

  s=open_stream(in?NULL:extfil(buff,filename,".wsv",FALSE),8192);
  if (!s) {
    err_set(1,"restoreSession: failed to open file %s for read",filename);
    return;
  }
  if (in) set_stream_pipe(s,in);
  s->p->tok=buff;
  s->p->mts=sizeof(buff);

  clearSession();

  lock_evtq();

  set_stream(s,"redirection=off, comment=off, spctok=off, expvar=off");
  version=initRestore(s);
  if (version==1) set_stream(s,"eol=yes");
  else set_stream(s,"eol=no");
  if (version<1) {
    err_set(1,"invalid WSV version:");
    err_set(1,"%s",s->buff);
    status=1;
  }
  if (version>current_version) 
    prtGraphics("Format of session file is %d, expected version is %d.\n",version,current_version);

  while (!status && !strm_eof(s)) {
    if (token_stream(s,">")) {
      if (version==1) {        /* old format */
	if (streq("SRV",strm_tok(s))) status=restoreSRVV1(s);
	else if (streq("MDT",strm_tok(s))) status=restoreMDTV1(s);
	else if (streq("DDT",strm_tok(s))) {
	  status=restoreDDTV1(s);
	  if (status) clearSession();
	}else if (streq("DMON",strm_tok(s)))  status=restoreDMONV1(s);
	else if (streq("DOTS",strm_tok(s)))   status=restoreDotsV1(s);
	else if (streq("Macros",strm_tok(s))) status=restoreMacrosV1(s);
	else {
	  err_set(1,"invalid section in session file: %s\n",strm_buff(s));
	  status=1;
	}
      }else{                    /* modern (version>1) format */
	if (streq("<SRV",strm_tok(s))) status=restoreSRV(s);
	else if (streq("<MDT",strm_tok(s))) status=restoreMDT(s);
	else if (streq("<DDT",strm_tok(s))) {
	  status=restoreDDT(s);
	  if (status) clearSession();
	}else if (streq("<DMON",strm_tok(s))) status=restoreDMON(s);
	else if (streq("<DOTS",strm_tok(s))) status=restoreDots(s);
	else if (streq("<MACROS",strm_tok(s))) status=restoreMacros(s);
	else if (streq("</WSV",strm_tok(s))) break;
	else if (*strm_tok(s)=='<') {
	  ignoreXMLToken(s);
	}else{
	  err_set(1,"invalid data in session file: %s\n",strm_buff(s));
	  status=1;
	}
      }
    }
  }

  if (version>1) {
    if (strm_eof(s)) err_set(1,"missing </WSV> tag in session file %s.",filename);
  }else if (!strm_eof(s)) {
    err_set(1,"session restore incomplete.");
  }

  close_stream(s);

  unlock_evtq();

  freeRestoreData();

}

void saveSession(char *filename)
{
  char file[4096];
  FILE *out;
 
  extfil(file,filename,".wsv",FALSE);
  if ((out=fopen(file,"w"))==NULL) {
    err_set(1,"saveSession: failed to open %s for write\n",file);
    return;
  }

  initSave(out);
  saveSRV(out);

  /* no comment */
  get_RTP((Dspobj *)DspRoot);

  saveMDT(out,MolRoot);
  saveDDT(out,DspRoot);
  saveDMON(out);
  saveDots(out);
  saveMacros(out);

  /* ... and do not forget to close the file. */

  fprintf(out,"\n</WSV>\n"); 
  if (ferror(out)) err_set(1,"error writing session save file %s",file);
  if (fclose(out)) {
    err_set(1,"error writing session save file %s",file);
  }else{
    prtGraphics("Session saved to file %s\n",file);
  }
}

