#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "object.h"
#include "swdef.h"

#ifdef mingw
#include <winsock.h>
#else
#include <sys/mman.h>
#include <unistd.h>
#endif

#if defined(linux)
#define HAVE_SHM
#endif

#ifdef USESOCKETS

#include <signal.h>
void broken_pipe(int);

extern int server;

#endif

Node *MolRoot=(Node *)NULL;
Node *DefaultNode=(Node *)NULL;
Node *LibRoot=(Node *)NULL;
Dspobj *DspRoot=(Dspobj *)NULL;
Dspobj *PickRoot=(Dspobj *)NULL;

int main(int argc,char **argv)
{
  /* initialize everything, then enter command loop */

  int status,*RPCstatus=0,fd;
  int i,pid,do_fork=FALSE,do_rpc=FALSE;
  char shmemreg[256];

#ifdef mingw
  int err;
  WSADATA wsaData; 
#endif

#if defined(USESOCKETS) && !defined(mingw)
  signal(SIGPIPE,broken_pipe);
#endif

#ifdef mingw
  err = WSAStartup( 0x0101, &wsaData ); 
  if ( err != 0 ) { 
    fprintf(stderr,"couldn't find a useable winsock.dll.\n");
    return(1); 
  } 
#endif

  for (i=1;i<argc;i++) {
    if (strcmp(argv[i],"-B")==0) do_fork=TRUE;
    else if (strcmp(argv[i],"-rpc")==0) do_rpc=TRUE;
  }

#ifdef mingw
  if (do_fork) {
    fprintf(stderr,"note: -B option (\"background\") is not supported on MinGW\n");
    do_fork=FALSE;
  }
#endif

#ifndef mingw
  if (do_fork && do_rpc) {
    pid=getpid();
#ifdef HAVE_SHM
    sprintf(shmemreg,"/wnp%dshmr1",pid);
    fd=shm_open(shmemreg,O_RDWR|O_CREAT,S_IRWXU);
#else
    sprintf(shmemreg,"/tmp/wnp%dshmr1.shm",pid);
    fd=open(shmemreg,O_RDWR|O_CREAT,S_IRWXU);
#endif    
    if (fd>=0) {
      if (ftruncate(fd,4096)==0) RPCstatus=(int *)mmap(0,4096,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
      else RPCstatus=NULL;
#ifndef HAVE_SHM
      close(fd);
#endif
      if (RPCstatus==(int *)MAP_FAILED) {
	fprintf(stderr,"failed to map shared memory region %s\n",shmemreg);
	return(1);
      }
    }else{
      fprintf(stderr,"failed to open shared memory region %s\n",shmemreg);
      return(1);
    }
#ifdef HAVE_SHM
    shm_unlink(shmemreg);
#else
    unlink(shmemreg);
#endif
  }
#endif

  if (do_fork) {
    if ((pid=fork()>0)) {
      if (do_rpc) {
	for (i=0;*RPCstatus==0;i++) {
	  if (i>500) break;
	  usleep(20000);
	}
	return(*RPCstatus-1);
      }else return(0);
    }else if (pid<0) {
      fprintf(stderr," failed to fork client...\n");
      return (1);
    }
  }

  status=init_all(argc,argv);
  if (do_fork && do_rpc) *RPCstatus=status+1;

  if (status) return(status);
  command_loop("Wit!P> ");
  graphics_down(NULL);
#ifdef mingw
  WSACleanup();
#endif
  return(0);
}

#if defined(USESOCKETS) && !defined(mingw)
void broken_pipe(int n)
{
  signal(SIGPIPE,broken_pipe);
  if (n==SIGPIPE) {
    fprintf(stderr,"broken pipe...\n");
    fflush(stderr);    
  }
}
#endif
