#ifndef SWDEF
#define SWDEF

#include "wnpdef.h"
#include "object.h"
#include "streamio.h"
#include "surface.h"

#define CHRM_PSF_CHARMM  1
#define CHRM_PSF_XPLOR   2

#define CHKDAT_FIX    1
#define CHKDAT_ASK    2
#define CHKDAT_CHECK  3
#define CHKDAT_REPORT 4

#define ATM_DSP_ON    0
#define ATM_DSP_ONLY  1
#define ATM_DSP_OFF   2

#define AUTO_PAIR_NEAR     1
#define AUTO_PAIR_STABLE   2
#define AUTO_PAIR_STRUCTAL 3

#define AUTOBOND_ALL    0
#define AUTOBOND_RESID  1    /* atoms must be in same residue for autobond */
#define AUTOBOND_CHAIN  2    /* atoms must be in same chain for autobond */
#define AUTOBOND_OBJECT 3    /* atoms must be in same display object for autobond */

#define MSGQWAIT 1
#define MSGQNOWAIT 0
#define MAX_NLEV 25
#define MAX_NPRIM 50

#define AUTONM_HHEAV   1
#define AUTONM_HSEQ    2
#define AUTONM_SEQ     3
#define AUTONM_ALARM   4

#define MAX_FT    23

#define FT_CAR    (1<<0)
#define FT_MOL    (1<<1)
#define FT_PDB    (1<<2)
#define FT_SHX    (1<<3)
#define FT_WLB    (1<<4)
#define FT_PAR    (1<<5)
#define FT_MACCS  (1<<6)
#define FT_MAE    (1<<7)
#define FT_DIANA  (1<<8)
#define FT_DUPPER (1<<9)
#define FT_XPLOR  (1<<10)
#define FT_XPDB   (1<<11)
#define FT_DOTS   (1<<12)
#define FT_MOL2   (1<<13)
#define FT_BIOC   (1<<14)
#define FT_NOE    (1<<15)
#define FT_CNF    (1<<16)
#define FT_CNT    (1<<17)
#define FT_CSD    (1<<18)
#define FT_CIF    (1<<19)
#define FT_MMOD   (1<<20)
#define FT_SESS   (1<<21)
#define FT_MDL    (1<<22)

#define FT_NONE   0

#define FT_MOL_TYPES \
   (FT_CAR|FT_MOL|FT_MOL2|FT_PDB|FT_SHX|FT_PAR|FT_MACCS|FT_DIANA|FT_XPLOR|FT_XPDB|FT_BIOC|FT_CSD|FT_CIF|FT_MMOD|FT_SESS|FT_MDL|FT_MAE)

#define FT_ALL_TYPES (~FT_NONE)

#define FT_CAN_READ \
   (FT_CAR|FT_MOL|FT_MOL2|FT_PDB|FT_XPDB|FT_SHX|FT_PAR|FT_MACCS|FT_DIANA|FT_WLB|FT_DUPPER|FT_DOTS|FT_BIOC|FT_NOE|FT_CNF|FT_CNT|FT_CSD|FT_CIF|FT_MMOD|FT_SESS|FT_MDL|FT_MAE)

#define FT_CAN_WRITE \
   (FT_CAR|FT_MOL|FT_MOL2|FT_SHX|FT_XPLOR|FT_PDB|FT_XPDB|FT_MACCS|FT_MMOD|FT_SESS)

#define FT_CAN_APPEND \
   (FT_MOL|FT_MOL2|FT_XPDB|FT_MACCS|FT_MMOD)

typedef struct {
  int nlev;
  int level;
  Node *node,*up;
  int start[MAX_NLEV];
  int nprim[MAX_NLEV];
  int type[MAX_NPRIM];
  int active[MAX_NPRIM];
  char *str1[MAX_NPRIM];
  char *str2[MAX_NPRIM];
  Node *first[MAX_NPRIM];
  Node *last[MAX_NPRIM];
} Nmatch;


typedef struct Qmsg {
  struct Qmsg *nxt;
  char type[4];
  int size;
  char *msg;    
} Qmsg;

typedef struct dot_template {
  float R;
  long n;
  float *x;
} dot_template;

#include "gcalc.h"

/* ------------------- Prototypes ------------------------*/

int     InternalDataCheck(int fix);
int     read_amber_crd(UCpipe *in,Node *root);
int     write_amber_prmtop(char *fname,Node *root,int format);
mat_4   GetFitCoTransform(void);
int     WriteMolecule(Node *node,char *filename,int type,int append);
int     display_atom_chain(Nodechain *c,int op); /* op: ATM_DSP_ON or ATM_DSP_OFF ,return: n changed */

void    command_loop(char *prompt);
Dspobj     *cmd_add_surface(Surface *S);
void        cmd_amber(void);
void        cmd_bond(void);
void        cmd_build(void);
void        cmd_caveat(void);
void        cmd_charmm(void);
void        cmd_color(void);
void        cmd_confa(void);
void        cmd_construct(void);
void        cmd_contour(Grid *g);
void        cmd_copy(int move);
void        cmd_cpk(void);
void        cmd_crystal(Xcn *xcn);
void        cmd_cyana(void);
void        cmd_define(void);
wnpmacro_t *cmd_define_macro(wnpmacro_t *macros);
void        cmd_delete(void);
void        cmd_discover(void);
void        cmd_display(void);
void        cmd_display_obj(char *typ);
void        cmd_display_reset(void);
void        cmd_display_set(void);
void        cmd_divan(void);
void        cmd_dock(void);
void        cmd_dots(void);
int         cmd_exit(void);
void        cmd_extract(void);
void        cmd_fit(char *op);
void        cmd_fragment(void);
void        cmd_freeze(void);
void        cmd_genlig(void);
void        cmd_global(void);
void        cmd_invert(void);
void        cmd_invert_center(void);
void        cmd_invert_molecule(void);
void        cmd_join(void);
void        cmd_label(void);
void        cmd_list(void);
void        cmd_loop(void);
void        cmd_ludi(void);
void        cmd_maccs(void);
void        cmd_macro(wnpmacro_t *m);
void        cmd_match(void);
void        cmd_measure(void);
void        cmd_merge(void);
void        cmd_minimax(void);
void        cmd_modify(void);
void        cmd_mopac(void);
void        cmd_mpeoe(void);
void        cmd_mutate(void);
void        cmd_nmr(void);
void        cmd_ortep(void);
void        cmd_pause(void);
void        cmd_pdb(void);
void        cmd_piff(void);
void        cmd_plot(void);
void        cmd_pluto(void);
void        cmd_pwd(void);
void        cmd_quest(void);
void        cmd_radii(void);
void        cmd_read(void *vroot,int types);
void        cmd_read_amber_crd(void);
int         cmd_read_coordinates(Nodechain *atoms);
Surface    *cmd_read_surface(void);
void        cmd_rename(void);
void        cmd_render(void);
void        cmd_reset(void);
void        cmd_select(void);
void        cmd_set(void);
void        cmd_setdefault(void);
void        cmd_setenv(void);
void        cmd_sketch(void);
void        cmd_split(void);
void        cmd_surface(void);
void        cmd_surface_clip(void);
void        cmd_surface_color(void);
void        cmd_surface_delete(void);
void        cmd_surface_linewidth(void);
void        cmd_surface_status(char *status);
void        cmd_surface_style(void);
void        cmd_surface_transparency(void);
void        cmd_write(Node *root,int types);
void        cmd_write_amber_crd(void);
void        cmd_write_amber_prep(void);
void        cmd_write_amber_prmtop(void);
int         cmd_write_surface(Surface *S);
void        cmd_xray(void);

int     nMacroTypes(void);
char   *getMacroType(int i);
wnpmacro_t *getMacros(char *type);
wnpmacro_t *getMACCSmacros(void);
void    replace_macro(wnpmacro_t *old, wnpmacro_t *new);

void    set_bond_label(Bond *b,int type,char *text);
void    set_atom_label(Atom *a,int type,char *text,int forceon);
int     bond_auto_make(Nodechain *atoms,int mode);
int     SetDefaultBondWidth(int w);
int     DefaultBondWidth(void);
int     SetDefaultAtomColor(int color);
int     DefaultAtomColor(void);
void    autobondMol(Node *mol);
void    autocenterMol(Node *mol);
void    condAutobondMol(Node *mol);
void    condAutocenterMol(Node *mol);
int     set_autobond_mode(int mode);
int     autobond_mode(void);
int     autobond_status(void);
Link    *find_link_name(Node *lib,char *name);
Symbol  *find_symbol_name(Node *lib,char *name);
Node    *find_residue_name(Node *lib,char *name);
Atmmatch *match_residue_atoms(Node *template,Node *res);
int	init_all(int argc,char **argv);
void	init_defaults(void);
int	init_parser(void);
int	init_root(void);
void    initThreeDPrj(void);
void	parse_command(void);
Node   *MatchNodeName(Node *a,char *name);
Node   *name_to_nodeptr(char *name,int (*compare)(const char *reg,const char *wild));
Node   *node_in_ctxt(char *name,Node *ctxt,\
			  int (*compare)(const char *reg,const char *wild));
Node   *node_in_root(char *name,Node *root,\
			  int (*compare)(const char *reg,const char *wild));
char   *nodeptr_to_name(char *name,Node *n);
char   *nodeptr_to_name_ctxt(char *name,Node *n,Node *ctxt);
Node   *first_nmatch(char *name,Nmatch *s);
Nmatch *open_nmatch(char *name);
Node   *next_nmatch(Nmatch *s);
void    close_nmatch(Nmatch *s);
void	initmsgq(void);
Qmsg   *qmsg(char *type,int size,char *msg);
void	sendqmsg(Qmsg *msg);
void    send_dots(Atom *a,int n,float *x);
void    send_dots_np(Atom *a,int n,float *x);
Qmsg   *delqmsg(Qmsg *msg);
void    del_dsp(int o,Node *obj);
mat_4   obj_transform(Node *obj);
mat_4   dsp_transform(Dspobj *d);
mat_4   dsp_transform_grfx(Dspobj *d);
void    get_tree_matrices(Dspobj *root,char *type);
void    reset_object(Dspobj *o);
void    rotate_object(Dspobj *o,float *axis,float angle);
void    set_pivot(Dspobj *obj,float *pivot);
void    translate_object(Dspobj *obj,float *dx);
void    get_MVT(Dspobj *root);
void    get_RTP(Dspobj *root);
void    put_RTP(Dspobj *obj,mat_4 *matrix);
void    transform_object(Dspobj *obj,mat_4 *tra);
PrjTS   getThreeDPrj(void);
void    setThreeDPrj(PrjTS *prj);
Qmsg   *nextqmsg(char *types);
Qmsg   *fillmsgq(int wait_opt);
Qmsg   *waitqmsg(char *types);
void    syncqmsg(char *discard);
void	readqmsg(Qmsg *qm,char *msg);
void    cltoan(float m[3][3],float l[3],float a[3]);
void    clear_menu(void);
void    clear_menu_button(int button);
void    refresh_menu(void);
int     load_menu_buttons(int first,char *text);
int     read_maccs_mol(UCpipe *in,Node *node);
int     skip_maccs(UCpipe *fpipe,FileCache *cache,int target,int now);
int     read_csd_mol(UCpipe *in,Node *node);
int     read_lib(char *file,Lib *lib);
int     read_lib_stream(Stream *in,Lib *lib);
int     read_lib_ucp(UCpipe *ucp,Lib *lib);
int     read_bio(UCpipe *p,char *file,Node *node);
int     read_mae(UCpipe *p,char *hdr1,Node *node);
int     read_mmod(UCpipe *p,char *hdr1,Node *node);
int     read_mmod1(UCpipe *in,Node *node,int *nat,Atom ***Atms);
CnfNode *read_mmodcnf(UCpipe *in,int nat,Atom **Atms,Atom ***atms);
int     read_mol(UCpipe *in,char *hdr1,Node *node);
int     read_par(UCpipe *p,char *file,Node *node);
int     read_pdb(UCpipe *p,char *file,Node *node,int type);
int     read_syb(UCpipe *p,char *file,Node *node,int type);
int     read_shelx(UCpipe *p,char *file,Node *node);
int     read_cif(UCpipe *p,char *file,Node *node);
int     read_diana(UCpipe *p,char *file,Node *node);
int     read_diana_noe(char *file,NOElist *noes);
int     read_xplor_noe(char *file,NOElist *noes);
int     read_xplor_noe_stream(Stream *in,NOElist *noes);
int     read_biocad(UCpipe *p,char *file,Node *node);
int     read_biocad_mol(Stream *in,Node *node);
Grid   *read_grid(char *filename);
int     read_maccs(UCpipe *fpipe,char *file,Node *node,int multi);
int     read_csd(UCpipe *p,char *file,Node *node);
int     write_bio(char *file,char *hdr1,char *hdr2,Node *m);
int     write_charmm_rtf(char *file,Node *mol);
int     write_charmm_psf(char *file,Node *mol,int mode);
int     write_amber_parm(char *file,Node *mol,char *prmfile,char *modfile,int format);
int     write_amber_prep(char *file,Node *mol);
int     write_amber_crd(char *file,Node *mol);
int     write_mmod(char *file,Node *mol,int append);
int     write_maccs(char *file,Node *mol,int append);
int     write_mol(char *file,Node *mol,int append);
int     write_shelx(char *file,Node *mol);
void    write_xcn(FILE *out,Xcn *xcn);
int     write_xplor(char *file,Node *mol);
int     write_xplor_pdb(char *file,Node *mol,int append);
int     write_xplor_psf(char *file,Node *mol);
int     write_xplor_kludge(char *file,Node *mol);
int     write_xplor_noe(char *file,NOElist *noes);
void    write_xplor_noelist(FILE *out,NOElist *noes);
int     write_pdb(char *file,Node *mol);
int     write_taff_xplor(char *file,Node *mol);
void    setTAFFweight(char *term,float f);
void    setTAFFformat(char *format);
char   *xplr_ff_type(char *mnem);
char   *xplr_hb_type(char *mnem);
char   *xpl_sel(char *buf,Atom *a,Node *m);
void    make_xpl_resnam(char *name,char *xplnm,char *xplnu);
Node   *readmoltypes(UCpipe *p,char *filn,int type);
Node   *ReadNode(void *vroot,char *filn,int type);
void    mark_atom_chain(Nodechain *c);
void    send_bmat4(int chan,mat_4 *mtrx);
void    send_dsp_matrix_type(Dspobj *o,mat_4 *mtrx,char *type);
void    send_molecule(Node *root,Mol *m);
Dspobj *send_mdl(Dspobj *global,Mol *m);
void    send_bond_chain(Bond *b);
void    send_bond_attributes(Bond *bf);
void    send_one_atom(Atom *a);
void    mdl_add_atoms(Dspobj *mdl,Node *root);
void    mdl_add_atom_chain(Dspobj *mdl,Nodechain *atoms);
void    mdl_repl_atom_chain(Nodechain *atoms);
void    send_atom_chain(Nodechain *c);
void    send_atom_attributes(Nodechain *atoms);
void    lock_evtq(void);
void    unlock_evtq(void);
void    send_newx(Nodechain *c);
void    do_pick(char *mode);
Dspobj *pick_root(Dspobj *new_root);
void    pick_mode(char *mode);
Node   *pick_to_node(char *s);
Dspobj *new_dsp_obj(Dspobj *root,char *name,int type);
void    set_dsp_status(Dspobj *obj,int status);
int     set_auto_label(int mode);
Nodechain *set_mark_atoms(Nodechain *atoms);
int     auto_label(void);
void    pconnected_atoms(Nodechain *c,int dist);
void    connected_atoms(Nodechain *c);
Nodechain *connected_to_right(Atom *left,Atom *right);
int     ringsz(Bond *b,int maxsz);
Atomdef *guess_atom_type(char *name);
int     mark_aromatic_bonds(Bond *b);
int     asg_bondtype(Bond *b);
int     asg_bondtype2(Bond *b);
int     eff_lig(Atom *a);
int     asg_formlig(Atom *a);
int     sane_formlig(Atom *a);
int     asg_formchrg(Atom *a,int lig);
void    clear_atom_mark(Atom *a);
void    displayFind(Dspobj *obj,Nodechain *c);

Prule  *swilesPrule(char *swiles,int interp);
int     matchPruleAtom(Atom *a,Prule *pr,int amask,int bmask);
int     matchPruleBond(Bond *b,Prule *pr,int amask,int bmask);
int     matchPrule(Atom *a,Bond *b,Prule *pr,int amask,int bmask);
void    add_hydrogens(Atom *a,Nodechain *h);
int     automatic_types(Nodechain *atms);
int     write_mol2(char *file,Node *m,int append);
void    make_bio_resnam(char *name,char *bionm,char *bionu);
int     findSubstructure(int na,Atom ** A,int nb, Atom **B);
int     applySubstructureMatch(int na,Atom **A);
void    clearSubstructureMatch(int na,Atom **A);

int     ptab_val(int i);
float   ptab_M(int i);
int     ptab_Z(int i);
int     ptab_n(char *chs);
char   *ptab_chs(int z);

int     auto_pairs(int m,Atom **A,int n,Atom **B,float d,int method);
int     auto_near_pairs(int m,Atom **A,int n,Atom **B,float d);
int     auto_stable_pairs(int m,Atom **A,int n,Atom **B,float d);
int     auto_structal_pairs(int m,Atom **A,int n,Atom **B,float maxd);

float   RmsFit(int nat,Node *smol,Atom **s,Atom **t,mat_4 *Tra);
float   RmsMultiFit(int nat,int nmol,Node **mols,Atom ***atoms);
float   rms_best_mat(int nat,Atom **s,Atom **t,mat_4 *tra);
float   rms_best_mat_x(int nat,Atom **s,float *x,mat_4 *tra);
float   rms_best_mat_yx(int nat,float *y,float *x,mat_4 *tra);
float   rms_dist(int nat,Atom **s,Atom **t);
float   rms_dist_x(int nat,Atom **s,float *x);
void    cotransform(mat_4 Tra,Node *mol);
float   get_bondinc(Lib *lib,Atom *a,Atom *b);
mat_4   PriaxMat(int n,float **x);
mat_4   PMofI(int n,float **x,float *m);
void    ApplyTransf(mat_4 R,int n,float **x,float **y);

int     DirList(char *dir,char *sel,int mode,char **names[]);
int     isRegularFile(char *name);
int     isDirectory(char *name);

void    err_dump(void);
char   *newtwd(char *buffer,char *update);
char   *getCWD(char *buffer,int size);

int     read_spgrp(Xcn *xcn,char *name);
Dspobj *crys_make_box_dsp(Node *root,Xcn *xcn);

int near_neighbours(Atom *a,Atom *b,int nbonds);
int  read_cnf(char *filn,Node *root);
int *get_mobile_cnf(int *nmob);
int  read_cnt(char *filn,Node *root);
int  read_dots(char *filn,Nodechain *c);
void compute_dots_vdw(Nodechain *c);
int  compute_dots_connolly(Nodechain *c,char *filename,int LR,int BU);
void delete_dots(Nodechain *c);
void set_dots_color(Nodechain *c,int col);
void set_dots_dsp(Nodechain *c,int dsp);
void set_dots_style(Nodechain *c,int style);
int  set_bt(Bond *b,int type);
void set_biosym_version(char *v);
void setBiosymFF(char *menm);
void set_bt_automatic(Bond *B);
float est_pi_bondorder(Bond *bnd);
float default_bndlength(Atom *a,Atom *b,int bt);

Bond *aromatic_cluster(Atom *a,int flag);
int  localize_aromatic_bonds(Node *mol);
int  localize_bond_cluster(Bond *bnd);
void clear_bond_cluster(Atom *a,int flag);

int atm_geom_class(Atom *a);
int all_charges(Nodechain *c,int label_flg,float q);
int bondinc_charges(Nodechain *c,int label_flg);
int fragment_charges(Nodechain *c,Lib *lib,int label_flg);

MPEOEinfo MPEOEsetup(Nodechain *A);
MPEOEinfo MPEOEsetup_refinement(Nodechain *A);
MPEOEinfo MPEOEinit_scratch(void);
MPEOEopt  MPEOEgetOpt(void);
int       MPEOEcharges(Nodechain *c,int label_flg);
float     MPEOEiterate(MPEOEinfo G,int nit,float dq);
void      MPEOEfree_scratch(MPEOEinfo G);
float     MPEOErefine(Nodechain *A);
float     MPEOEqrms(int n,float *p,MPEOEinfo G);
void      MPEOEupdate(int n,float *p,MPEOEinfo G);
void      MPEOEsetOpt(MPEOEopt opt);
void      MPEOEdump(FILE *out);
int       MPEOEaddInitRule(Prule *pr);
void      MPEOEresetInitRules(void);
void      MPEOEsetGranularity(int n);

void set_altobj_key(char *key);
void init_mousemap(void);
char *auto_atom_name(Atom *a,char *name,int mode);
void  auto_atom_group(Node *red);
void  auto_atom_group_wooly(Node *red);
void  auto_atom_group_greedy(Node *red);
void  auto_atom_group_hetero(Node *red);
int  checkabort(void);
int  startCmd(char *command,char *logfile,char *errfile,int wait,char *priority);

void free_dot_template(dot_template *t);
dot_template *find_best_template(float R,int n,dot_template *t[]);
dot_template *make_dot_template(float R,float dens);
int clip_dots(int n,float *x,Atom *a);
void cmd_compute_dots(void);
int RampColor(float val,float mn,float mx);
void DefineColorRamp(int ncol,int *cols);
int *ColorRamp(void);
int  NRampColors(void);
char *mouse_func_mnems(void);
char *get_mouse_keys(int func);
void show_mouse_map(void);
void set_mouse_function(int func,char *keys);
void set_mouse(char *left_or_right);
void graphics_down(char *reason);
float SrvVersion(void);

void define_fragment_lib(char *name,UCpipe *lib);
int  del_fragment(Node *l);
Node *fragment_root(void);

Frule *parseFrule(char *text,Lib *lib);
int    matchFrule(Atom *a,Frule *p);

float SimPlexMin(int n,float *x,float *dx,
                 float (*F)(int n,float *z,void *info),void *data,
                 float tolf,int miter,
		 int (*report)(int it,int n,float **p,float *v,int hi,int lo),
		 int *status);

void dotdotdot(char *pref,int k,int n,int m);

void  grid_dump(Grid *g);
float grid_value(Grid *g,vec_3 pt);
vec_3 grid_grad(Grid *g,vec_3 pt);
float grid_valuex(Grid *g,float *pt);
vec_3 grid_gradx(Grid *g,float *pt);
Grid *emptyGrid(void);
Grid *freeGrid(Grid *g);
Dspobj* grid_box_delete(Grid *G);
int   write_grid(FILE *f,Grid *g);
int   grid_FN_dotdist(Grid *G);
int   grid_FN_Contact(Grid *G);
int   grid_FN_SaSDist(Grid *G);
int   grid_FN_EDFDist(Grid *G);
int   grid_FN_ConnollyDist(Grid *G);
int   grid_FN_Coulomb(Grid *G);
int   grid_FN_truncate(Grid *G);
int   grid_FN_vdwdist(Grid *G);
int   grid_FN_mult(Grid *G);
int   grid_fn_mult(Grid *g,float f);
int   grid_fn_mult(Grid *g,float f);
int   grid_fn_zero(Grid *g);
int   grid_fn_dotdist(Grid *G,int ndot,float **dot);
int   grid_fn_truncate(Grid *g,float b,float a);
int   grid_fn_Contact(Grid *g,Nodechain *c,float r);
int   grid_fn_Coulomb(Grid *g,Nodechain *c,int dielt,float eps);
int   grid_fn_vdwdist(Grid *g,Nodechain *c);
int   grid_fn_SaSDist(Grid *g,Nodechain *c,float r);
int   grid_fn_EDFDist(Grid *g,Nodechain *c,float d);
int   grid_fn_ConnollyDist(Grid *g,Nodechain *c,float r);
int   LRinit(Atom **A,int n,float r,float d,int buried);
void  LRclose(void);
dot_template *LRdots(Atom *a);
Grid *grid_create(Grid *oldGrid);
Grid *cond_del_grid(Grid *oldGrid);
int   grid_compute(Grid *G);
void  grid_report(Grid *G);
Grid *grid_delete(Grid *oldGrid);
Grid *grid_read(Grid *oldGrid);
void  grid_write(Grid *G);
void  grid_box_make(Grid *G);
void  prtGraphics(char *text,...);

void restoreSession(UCpipe *in,char *filename);
void saveSession(char *filename);

int  initSave(FILE *out);
int  numberMDTNodes(int *N,Node *r) ;
void saveMDTNode(FILE *out,Node *n) ;
void saveMDT(FILE *out,Node *root);
int  numberDDTNodes(int *N,Node *r,int type) ;
void saveDDT(FILE *out,Node *root);
void saveDDTNodes(FILE *out,Node *n);
void saveSRV(FILE *out);
void saveDots(FILE *out);
void saveDMON(FILE *out);
void saveMacros(FILE *out);

int  ignoreXMLToken(Stream *s);
int  initRestoreV1(void);
int  restoreMDTV1(Stream *s);
int  restoreDDTV1(Stream *s);
int  restoreSRVV1(Stream *s);
int  restoreDotsV1(Stream *s);
int  restoreDMONV1(Stream *s);
int  restoreMacrosV1(Stream *s);
void freeRestoreDataV1(void);

int  initRestore(Stream *s);
int  restoreMDT(Stream *s);
int  restoreDDT(Stream *s);
int  restoreSRV(Stream *s);
int  restoreDots(Stream *s);
int  restoreDMON(Stream *s);
int  restoreMacros(Stream *s);
void freeRestoreData(void);
void clearMacros(void);

int setAttribute(char *which,int what);
int getAttribute(char *which);

void   setAttributeFA(char *which, float *what);
float *getAttributeFA(char *which);
void   setPropertyMap(float min,float max);
void   getPropertyMap(float *min,float *max);
int    get_I_parm(char *what);
void   set_cyl_mode(int mode,int npoly);
void   set_sph_mode(int mode,int tessdepth);

int Setenv(char *name,char *value);
int inBackground(void);

void cmd_xmlrpc(void);
void wnpRPCshutdown(void);
void wnpRPCconnect(int port);
int  wnpRPCisOn(void);
int  wnpRPCisOff(void);
int  wnpRPCstart(int port);
int  wnpRPCport(void);
int  wnpHaveRPC(void);
int  wnpRPCmaxCalls(void);
void wnpRPCprintResponse(int mode,char *msg);
int  wnpRPCmutexLock(void);
int  wnpRPCmutexUnlock(void);

int  handleRPC(char *buff,int buffsz);

int  use_readline(int status);
void uSleep(unsigned int usec);

void cmd_smcsa(void);

#endif
