#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "token.h"

#define TRUE 1
#define FALSE 0
#define NOW p->buf[p->pos]
#define NUL '\0'
#define NC (char *)NULL
#define LF '\n'
  
  /* low level, general purpose token parsing */
  
int get_token(Parse *p)
{
  char quote,*sep;
  int  start,end;
  
  if (!p) return(FALSE);
  if (!p->buf) return(FALSE);

  sep=(p->sep ? p->sep : "");
  p->ltok=0;
  p->quoted=FALSE;
  
  /* skip leading whitespace, if any */
  
  while(isspace(NOW) && NOW != NUL && NOW != LF) p->pos=p->pos+1;
  if (NOW == NUL) {
    p->ltok = 0; p->term = NUL;
    if (p->tok != NC) p->tok[0] = NUL;
    p->tks = NC; return(FALSE);
  }else if (NOW == LF) {
    p->pos=p->pos+1;
    p->ltok = 0; p->term = LF;
    if (p->tok) p->tok[0]='\0';
    p->tks=NC; return(FALSE);
  }
  
  /* if next char is sep, token is NIL  */
  
  if (strchr(sep,NOW) != NC) {
    p->term = NOW; p->pos=p->pos+1;
    if (p->tok != NC) p->tok[0]=NUL;
    p->ltok=0; p->tks = NC;
    while(isspace(NOW) && NOW != NUL && NOW != LF) p->pos=p->pos+1;
    return(TRUE);
  }
  
  /* if next char is one of the quote chars, the token is quoted */
  
  if (NOW == '\"' || NOW == '\'') {
    quote=NOW; p->quoted = TRUE; p->pos=p->pos+1;}
  else {
    quote='\0'; p->quoted = FALSE;
  }
  
  /* at start of token, enter parse loop */
  
  start=p->pos; p->tks = p->buf + start;
  while ((p->quoted || (strchr(sep,NOW) == NC && !isspace(NOW)))
	 && NOW != NUL && NOW != quote) {
    p->ltok++; p->pos=p->pos+1;
  }
  
  /* parse complete, copy to token buffer, if requested */
  
  if (p->tok != NC) {
    end=(p->ltok >= p->mts ? p->mts-1 : p->ltok);
    if (end<1) end=p->ltok;
    if (end>0) strncpy(p->tok,&(p->buf[start]),end);
    p->tok[end]=NUL;
  }
  
  /* if token was quoted, skip over right-quote */
  
  if (NOW == quote && NOW != NUL) p->pos=p->pos+1;
  
  /* skip over trailing whitespace and consume separator  */
  
  while(isspace(NOW) && NOW != NUL && NOW != LF) p->pos=p->pos+1;
  if (strchr(sep,NOW) == NC || NOW == NUL || NOW == LF) p->term = ' ';
  else {p->term=NOW; p->pos=p->pos+1;}
  while(isspace(NOW) && NOW != NUL && NOW != LF) p->pos=p->pos+1;
  return(TRUE);
}

char *token(Parse *p)
{
  /* extract token and return pointer (NULL, if no token) 
     if p->tok is NULL, the pointer is into p->buf, and the
     token is not necessarily ASCIZ */
  
  char *r;
  
  if (get_token(p)) {
    if (p->tok) {
      r=p->tok;
    }else{
      r=p->tks;
    }
  }else{
    r=(char *)NULL;
  }
  return(r);
}
