/*todo: better error handling */

/* Message transport system */

#ifdef mingw
#include <windows.h>
#include <winsock.h>
#include <time.h>
#else
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif
#include <errno.h>
#include <stdio.h>
#include "transpor.h"
#ifdef sgi
#include <limits.h>
#endif
#ifndef EWOULDBLOCK
#if defined(mingw)
#define EWOULDBLOCK EAGAIN
#define ECONNRESET  104
#endif
#endif

/*===============================================================*/

/*    Routines to convert between MY format and SGI              */
/*    NOTE: SGI_f is an integer to avoid problems with NAN's ... */

static union {
  int i;
  float f;
  unsigned char c[4];
} u_swap;

/*********************************************************/

#if defined(sgi) || defined(darwin)

float MY_f(int x)    /* convert SGI float to MY float */
{
  u_swap.i=x;
  return(u_swap.f);
}

int SGI_f(float x)   /* convert MY float to SGI float */
{
  u_swap.f=x;
  return(u_swap.i);
}

int MY_i(int i)      /* convert SGI int to MY int */
{
  return(i);
}

int SGI_i(int i)     /* convert MY int to SGI int */
{
  return(i);
}

#endif

float OWN_f(int swap,float x)
{
  unsigned char c;
  
  u_swap.f=x;
  if (swap) {
    c=u_swap.c[0];
    u_swap.c[0]=u_swap.c[3];
    u_swap.c[3]=c;
    c=u_swap.c[1];
    u_swap.c[1]=u_swap.c[2];
    u_swap.c[2]=c;
  }
  return(u_swap.f);
}

int OWN_i(int swap,int x)
{
  unsigned char c;
  
  u_swap.i=x;
  if (swap) {
    c=u_swap.c[0];
    u_swap.c[0]=u_swap.c[3];
    u_swap.c[3]=c;
    c=u_swap.c[1];
    u_swap.c[1]=u_swap.c[2];
    u_swap.c[2]=c;
  }
  return(u_swap.i);
}

/******************************************************************/

#if defined(linux) || defined(cygwin) || defined(mingw)

float MY_f(int x)
{
  unsigned char c;
  
  u_swap.i=x;
  c=u_swap.c[0];
  u_swap.c[0]=u_swap.c[3];
  u_swap.c[3]=c;
  c=u_swap.c[1];
  u_swap.c[1]=u_swap.c[2];
  u_swap.c[2]=c;
  return(u_swap.f);
}

int SGI_f(float x)
{
  unsigned char c;
  
  u_swap.f=x;
  c=u_swap.c[0];
  u_swap.c[0]=u_swap.c[3];
  u_swap.c[3]=c;
  c=u_swap.c[1];
  u_swap.c[1]=u_swap.c[2];
  u_swap.c[2]=c;
  return(u_swap.i);
}

int MY_i(int x)
{
  unsigned char c;
  
  u_swap.i=x;
  c=u_swap.c[0];
  u_swap.c[0]=u_swap.c[3];
  u_swap.c[3]=c;
  c=u_swap.c[1];
  u_swap.c[1]=u_swap.c[2];
  u_swap.c[2]=c;
  return(u_swap.i);
}


int SGI_i(int x)
{
  unsigned char c;
  
  u_swap.i=x;
  c=u_swap.c[0];
  u_swap.c[0]=u_swap.c[3];
  u_swap.c[3]=c;
  c=u_swap.c[1];
  u_swap.c[1]=u_swap.c[2];
  u_swap.c[2]=c;
  return(u_swap.i);
}

#endif

/*==================== end of host format routines ====================*/

#if defined (sgi)
void uSleep(unsigned int usec)
{
  long ticks;
  
  ticks=(long)(usec/1000000.0)*CLK_TCK;
  sginap(ticks);
}
#else
#if defined (mingw)
void uSleep(unsigned int usec)
{
  int ticks;
  
  ticks=(usec/1000000.0)*CLOCKS_PER_SEC;
  sleep(ticks);
}
#else
void uSleep(unsigned int usec)
{
  usleep(usec);
}
#endif
#endif

#if defined (linux) || defined (mingw) || defined(sgi) || defined (darwin)
#define SendRecv
#endif 

/* MAXC  is the maximum number of transport channels */

#define MAXC 10
#define TRUE 1
#define FALSE 0
#define nogood(x) (x<0 || x>=MAXC)

static char CHCKSM[4] = {(char)0x5C,(char)0x36,(char)0xA9,(char)0xEB};

static int nchasg=0;		/* number od assigned channels */
static int file[MAXC];		/* channel's file handle */
static int assigned[MAXC];	/* true iff channel is assigned */
static char *bufi[MAXC];	/* pointer to channel's input buffer */
static char *bufo[MAXC];	/* pointer to channel's output buffer */
static int ibs[MAXC];		/* channel's input buffer size */
static int obs[MAXC];		/* channel's output buffer size */
static int bi[MAXC];		/* number of bytes in channels input buffer */
static int bin[MAXC];		/* next input byte (offset into bufi) */
static int bon[MAXC];		/* next output bytes (offset into bufo) */
static char hdr[MAXC][12];	/* 12 character header of input message */
static int sz[MAXC];		/* input message size (excl. header) */
static int rd[MAXC];		/* number of bytes read (excl. header) */
static int ig[MAXC];		/* ignore message if true */
static int waitmode[MAXC];      /* are we in wait or in nowait mode ? */
#ifdef SendRecv
static int flgs[MAXC];          /* flagword for send/recv */     
#endif

static int extr_int(char *b)	/* extract int from buffer b */
{
  int i;
  for (i=0;i<sizeof(int);i++) u_swap.c[i]= *b++;
  return(SGI_i(u_swap.i));
}

static void insert_int(char *b,int i)
{
  u_swap.i=SGI_i(i);
  for(i=0;i<sizeof(int);i++) *b++ = u_swap.c[i];
}

static int get(int ch,char *b,int s)
{
  int i,n,m;
  
  while (ig[ch]>0) {
    n=(ig[ch]<(bi[ch]-bin[ch]) ? ig[ch] : (bi[ch]-bin[ch]));
    ig[ch]-=n;
    bin[ch]+=n;
    if (bin[ch]==bi[ch]) {
#ifdef SendRecv
      n=recv(file[ch],bufi[ch],ibs[ch],flgs[ch]);
#else
      n=read(file[ch],bufi[ch],ibs[ch]);
#endif
      if (n<0 && errno!=EWOULDBLOCK && errno!=0 ) {
	fprintf(stderr,"errno: %d, ",errno);
	perror("receive");
      }
      bin[ch]=0;
      bi[ch]=(n>0 ? n : 0);	    
    }
    if (!bi[ch]) break;
  }
  if (ig[ch]>0) return(0);
  m=0;
  while (m<s) {
    n=((s-m)>bi[ch]-bin[ch] ? bi[ch]-bin[ch] : (s-m));
    for (i=0;i<n;i++) *b++ = bufi[ch][bin[ch]+i];
    bin[ch]+=n;
    m+=n;
    if (m<s && bin[ch]==bi[ch]) {
#ifdef SendRecv
      n=recv(file[ch],bufi[ch],ibs[ch],flgs[ch]);
#else
      n=read(file[ch],bufi[ch],ibs[ch]);
#endif
      if (n<0 && errno!=EWOULDBLOCK && errno!=0 ) {
	fprintf(stderr,"errno: %d, ",errno);
	perror("receive");
      }
      bin[ch]=0;
      bi[ch]=(n>0 ? n : 0);
    }
    if (!bi[ch]) break;
  }
  return(m);
}

static int get_w(int ch,char *b,int s)
{
  int m,n;
  
  m=0;
  while(m<s) {
    n=get(ch,b+m,s-m);
    m+=n;
    if (n<=0) uSleep(10000); 		/* release CPU for a tick */
  }
  return(m);
}

static void flush(int ch)
{
  int n,i,nretry=60;
  
  while (bon[ch]>0) {
    errno=0;
#if defined(SendRecv)
    n=send(file[ch],bufo[ch],bon[ch],flgs[ch]);
#else
    n=write(file[ch],bufo[ch],bon[ch]);
#endif
    if (n>0) {
      nretry=60;
      for (i=n;i<bon[ch];i++)  bufo[ch][i-n]=bufo[ch][i];
      bon[ch]-=n;
    }
    if (errno==EWOULDBLOCK) {
      uSleep(10000);	    /* should maybe wait even longer... */
    }else if (errno||n<0) {
      perror("transport(send)");
      if (errno==ECONNRESET||errno==EPIPE) nretry=1;
      else uSleep(1000000);
      nretry--;
      if (nretry==0) break;
    }
  }
  if (bon[ch]>0) {
    fprintf(stderr,"transport (send), fatal error.\n");
    bon[ch]=0;
  }
}

static void put_header(int ch,char *hd,int size)
{
  int i;

  /* Note: Since messages in the buffer may be unaligned with respect
     to int, the checksum must be computed byte-wise . */

  if (obs[ch]-bon[ch]<12) flush(ch); 
  strncpy(bufo[ch]+bon[ch],hd,4);
  insert_int(bufo[ch]+bon[ch]+4,size);

  /* insert four byte header checksum : xor of hd and size */
  for (i=0;i<4;i++) {
    bufo[ch][bon[ch]+i+8]=bufo[ch][bon[ch]+i]^bufo[ch][bon[ch]+i+4]^CHCKSM[i];
  }
  bon[ch]+=12;
}  

static int check_header(int ch)
{
  /* check header checksum of hdr[ch], if wrong, shift left one byte,
     decrement rd[ch], return FALSE, else return TRUE ,
     if rd[ch] is < 12 return FALSE regardless                      */

  /* Note: doing the checksum by int is O.K., since hdr[ch] is always
     aligned by the compiler (true for SGI, FX-2800, ...) */

  int i,chk,r;
  
  if (rd[ch]<12) {
    r=FALSE;
  }else{
    for (i=0,chk=FALSE;i<4&&!chk;i++) chk=hdr[ch][i]^hdr[ch][i+4]^hdr[ch][i+8]^CHCKSM[i];
    if (chk) {
      r=FALSE;
      for (i=1;i<rd[ch];i++) hdr[ch][i-1]=hdr[ch][i];
      rd[ch]--;
    }else{
      r=TRUE;
    }
  }
  return(r);
}  

/*========================================================================*/

/*------------------------------------------------------------------------*/
int open_transport(int fildes,int bsi,int bso)
{
  int ch;
  
  if (nchasg>=MAXC) return(0);
  for (ch=0;ch<MAXC;ch++) if (!assigned[ch]) break;
  if (nogood(ch)) return(0);
  
  ibs[ch]=(bsi < 64 ? 64 : bsi);
  bufi[ch]=(char *)malloc(ibs[ch]);
  if (!bufi[ch]) return(0);
  obs[ch]=(bso < 64 ? 64 : bso);
  bufo[ch]=(char *)malloc(obs[ch]);
  if (!bufo[ch]) {
    free(bufi[ch]);
    return(0);
  }
  file[ch]=fildes;
  bi[ch]=0;
  bin[ch]=0;
  bon[ch]=0;
  sz[ch]= -1;
  rd[ch]=0;
  ig[ch]=0;
  waitmode[ch]=TRUE;
#ifdef SendRecv
  flgs[ch]=0;
#endif 
  
  nchasg++;
  assigned[ch]=TRUE;
  return(ch+1);
}

void close_transport(int ch)
{
  ch--;
  if (nogood(ch)) return;
  if (!assigned[ch]) return;
  
  flush(ch);
  free(bufi[ch]);
  free(bufo[ch]);
  assigned[ch]=FALSE;
  nchasg--;
}

int transport_wait(int ch,int state)
{
  int old;
#if defined(mingw)
  unsigned long dontblock;
#endif
#if !defined(SendRecv)
  int flags;
#endif

  ch--;
  if (nogood(ch)) return(-1);
  if (!assigned[ch]) return(-1);

  old=(waitmode[ch]?TRUE:FALSE);
  waitmode[ch]=(state?TRUE:FALSE);

#if defined(SendRecv)

#if defined(mingw)
  /* WIN32 does not seem to have MSG_DONTWAIT flag */
  if (old!=waitmode[ch]) {
    dontblock=!waitmode[ch];
    ioctlsocket(file[ch],FIONBIO,&dontblock);
  }
#else
  if (waitmode[ch]) flgs[ch]&=~MSG_DONTWAIT;
  else flgs[ch]|=MSG_DONTWAIT;
#endif

#else
  if (old!=waitmode[ch]) {
    flags=fcntl(file[ch],F_GETFL,0);
    if (waitmode[ch]) {
      fcntl(file[ch],F_SETFL,~FNDELAY & flags);
    }else{
      fcntl(file[ch],F_SETFL,FNDELAY | flags);
    }
  }
#endif
  
  return(old);
}

int read_message(int ch,char *hd,char *ms)
{
  int n;
  
  ch--;
  if (nogood(ch)) return(-1);
  if (!assigned[ch]) return(-1);
  
  if (sz[ch]>=0 && rd[ch] >0) {
    ig[ch]=sz[ch]-rd[ch];
    sz[ch]= -1;
  }
  if (sz[ch]<0) {
    for(;;) if (got_message(ch+1)) break; 
    sz[ch]=extr_int(hdr[ch]+4);
    rd[ch]=0;
  }
  
  strncpy(hd,hdr[ch],4);
  n=get_w(ch,ms,sz[ch]);
  sz[ch]= -1;
  rd[ch]=0;
  return(n);
}

char *got_message(int ch)
{
  int n;
  
  ch--;
  if (nogood(ch)) return((char *)NULL);
  if (!assigned[ch]) return((char *)NULL);
  
  if (sz[ch]>=0) return(hdr[ch]);
  do {
    n=get(ch,hdr[ch]+rd[ch],12-rd[ch]);
    rd[ch]+=n;
    if (rd[ch]<12) return((char *)NULL);
  } while (!check_header(ch));
  sz[ch]=extr_int(hdr[ch]+4);
  rd[ch]=0;
  return(hdr[ch]);
}

int message_size(int ch)
{
  ch--;
  if (nogood(ch)) return(0);
  if (!assigned[ch]) return(0);
  
  if (sz[ch]<0) return(0);
  return(sz[ch]);
}

int read_data(int ch,int size,char *ms)
{
  int m;
  
  ch--;
  if (nogood(ch)) return(0);
  if (!assigned[ch]) return(0);
  if (size>sz[ch]-rd[ch]) return(0);
  
  m=get_w(ch,ms,size);
  rd[ch]+=m;
  if (sz[ch]<=rd[ch]) {
    sz[ch]= -1;
    rd[ch]=0;
  }
  return(m);
}

int read_data_raw(int ch,int size,char *ms)
{
  int m;

  ch--;
  if (nogood(ch)) return(0);
  if (!assigned[ch]) return(0);
  m=get_w(ch,ms,size);
  return(m);
}

int discard_message(int ch)
{
  /* discard current message on channel ch */
  
  int r;

  ch--;
  if (nogood(ch)) return(0);
  if (!assigned[ch]) return(0);
  
  if (sz[ch]>=0) {
    ig[ch]=sz[ch]-rd[ch];
    sz[ch]= -1;
    rd[ch]=  0;
    r=ig[ch];
  }else r=0;
  return(r);
}

int remaining_size(int ch)
{
  /* report number of bytes remaining in message */

  int r;

  ch--;
  if (nogood(ch)) return(0);
  if (!assigned[ch]) return(0);
  
  if (sz[ch]>=0) r=sz[ch]-rd[ch];
  else r=-1;
  return(r);
}

int send_bmessage(int ch,char *hd,int size,char *ms)
{
  char *t;
  int i,m;
  
  ch--;
  if (nogood(ch)) return(-1);
  if (!assigned[ch]) return(-1);
  
  put_header(ch,hd,size);

  while (size>0) {
    t=bufo[ch]+bon[ch];
    m=(obs[ch]-bon[ch] < size ? obs[ch]-bon[ch] : size);
    for (i=0;i<m;i++) *(t++) = *(ms++);
    size -= m; bon[ch] += m;
    if (size>0) flush(ch);
  }
  return(1);
}

int send_bheader(int ch,char *hd,int size)
{
  ch--;
  if (nogood(ch)) return(-1);
  if (!assigned[ch]) return(-1);
  
  put_header(ch,hd,size);
  return(1);
}

int send_bdata(int ch,int size,char *data)
{
  int i,m;
  char *t;
  
  ch--;
  if (nogood(ch)) return(-1);
  if (!assigned[ch]) return(-1);
  
  while (size>0) {
    t=bufo[ch]+bon[ch];
    m=(obs[ch]-bon[ch] < size ? obs[ch]-bon[ch] : size);
    for (i=0;i<m;i++) *(t++) = *(data++);
    size -= m; bon[ch] += m;
    if (size>0) flush(ch);
  }
  return(1);
}

int send_data(int ch,int size,char *data)
{
  int s;
  s=send_bdata(ch,size,data);
  trflush(ch);
  return(s);
}

int send_message(int ch,char *hd,int size,char *data)
{
  int s;
  s=send_bmessage(ch,hd,size,data);
  trflush(ch);
  return(s);
}

int send_header(int ch,char *hd,int size)
{
  int s;
  s=send_bheader(ch,hd,size);
  trflush(ch);
  return(s);
}

void trflush(int ch)
{
  ch--;
  if (nogood(ch)) return;
  if (!assigned[ch]) return;
  
  flush(ch);
}

/*-------------------- pr_message --------------------*/

void prt_message(int ch,char *format,...)
{
  va_list args;
  char buffer[4046]; /* enough, I hope */
  
  /*  print message: if ch > 0 use send_message(ch,"*MSG",...
      else print to standard output.               */
  
  if (ch>0) {
    va_start(args,format);
    vsprintf(buffer,format,args);
    va_end(args);
    send_message(ch,"*MSG",strlen(buffer),buffer);
  }else{
    va_start(args,format);
    vprintf(format,args);
    va_end(args);
  }
}
/*-------------------- send_bint --------------------*/
int send_bint(int ch,int n)
{
  int k,s;
  
  if (ch>0) {
    k=SGI_i(n);
    s=send_bdata(ch,sizeof(int),(char *)&k);
    return(s);    
  }else{
    return(1);
  }
}
/*-------------------- send_int --------------------*/
int send_int(int ch,int n)
{
  int k,s;
  
  if (ch>0) {
    k=SGI_i(n);
    s=send_data(ch,sizeof(int),(char *)&k);
    return(s);    
  }else{
    return(1);
  }
}
/*-------------------- send_bptr --------------------*/

int send_bptr(int ch,void *p)
{
  int s;
  
  if (ch>0) {
    s=send_bdata(ch,sizeof(p),(char *)&p);
    return(s);    
  }else{
    return(1);
  }
}
/*-------------------- send_ptr --------------------*/

int send_ptr(int ch,void *p)
{
  int s;
  
  if (ch>0) {
    s=send_data(ch,sizeof(p),(char *)&p);
    return(s);    
  }else{
    return(1);
  }
}

/*-------------------- send_bfloat --------------------*/

int send_bfloat(int ch,float x)
{
  int s;
  int ix;
  
  if (ch>0) {
    ix=SGI_f(x);
    s=send_bdata(ch,sizeof(int),(char *)&ix);
    return(s);    
  }else{
    return(1);
  }
}
/*-------------------- send_float --------------------*/

int send_float(int ch,float x)
{
  int s;
  
  s=send_bfloat(ch,x);
  trflush(ch);
  return(s);
}
/*-------------------- send_bfloata --------------------*/

int send_bfloata(int ch,int n,float *x)
{
  int s=1;
  int i,ix;
  
  if (ch>0) {
    for (i=0;i<n;i++) {
      ix=SGI_f(x[i]);
      s=send_bdata(ch,sizeof(int),(char *)&ix);
    }
  }else{
    s=1;
  }
  return(s);
}
/*-------------------- send_floata --------------------*/

int send_floata(int ch,int n,float *x)
{
  int s;
  
  s=send_bfloata(ch,n,x);
  trflush(ch);
  return(s);
}

