#ifdef withRPC

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <xmlrpc.h>
#include <xmlrpc_client.h>

#define NAME "Xmlrpc-c Witnotp Client"
#define VERSION "1.0"

#define TRUE  1
#define FALSE 0
#define MAXSZ 600000

static char msg[600000];
static int verbose=TRUE,prompt=FALSE,multi=FALSE,keep=FALSE;

#define NAPTIME 100000
static int Ntry=1;

static void start_client(xmlrpc_env *env,char *name, char *version);
static void stop_client(void);
static int send_message(xmlrpc_env *env,char *url,char *message);
static void uSleep(unsigned int usec);

void wnpcmdHelp(void);

int main(int argc, char **argv) 
{
  xmlrpc_env env;
  char serverURL[2048];
  char *srvr=NULL;
  char *o,*p;
  int sz,i,status=0;
  float w;

  if (getenv("WNP_RPC_PORT")) sprintf(serverURL,"http://localhost:%s/RPC2",getenv("WNP_RPC_PORT"));
  else sprintf(serverURL,"http://localhost:1917/RPC2");

  verbose=TRUE;
  prompt=FALSE;
  multi=FALSE;
  keep=FALSE;
  p=msg; sz=MAXSZ;

  /* handle command line arguments */

  for (i=1;argv[i];i++) {
    if (!strcmp(argv[i],"--help")) {
      wnpcmdHelp();
      exit(0);
    }else if (argv[i][0]!='-'||strcmp(argv[i],"--")==0) {  /* rest of the command line is Wit!P command */
      if (strcmp(argv[i],"--")==0) i++;
      for (;i<argc;i++) {
	if (p!=msg) *p++=' ';
	strcpy(p,argv[i]);
	p+=strlen(p);
      }
    }else{                              /* single letter options */
      for (o=argv[i]+1;*o;o++) {
	if (*o=='q') verbose=FALSE;
	else if (*o=='p') prompt=TRUE;
	else if (*o=='l') multi=TRUE;
	else if (*o=='k') keep=TRUE;    /* keep server alive after 129 response (only in -l mode) */
	else if (*o=='w') {
	  o=o[1]?o+1:argv[++i];
	  if (o) {
	    sscanf(o,"%f%n",&w,&sz);
	    if (sz==strlen(o)) {
	      Ntry=(w>0?1+1000000*w/NAPTIME:1);
	    }else o=NULL;
	  }
	}else if (*o=='s') {
	  srvr=o[1]?o+1:argv[++i];
	  if (srvr) {
	    serverURL[0]='\0';
	    if (strlen(srvr)<7||strncmp(srvr,"http://",7)) sprintf(serverURL,"http://");
	    if (!strchr(srvr,':')) sprintf(serverURL+strlen(serverURL),"localhost:");
	    sprintf(serverURL+strlen(serverURL),"%s",srvr);
	    if (strlen(srvr)<5||strcmp(srvr+strlen(srvr)-5,"/RPC2")) {
	      sprintf(serverURL+strlen(serverURL),"/RPC2");
	    }
	  }
	  break;
	}
      }
    }
  }

  start_client(&env,NAME,VERSION);

  if (multi) {
    if (p!=msg) status=send_message(&env,serverURL,msg);
    while (!env.fault_occurred&&fgets(msg,MAXSZ,stdin)) {
      status=send_message(&env,serverURL,msg);
      if (status==129&&!keep) break;
    }

  }else if (!env.fault_occurred) {
    if (p==msg) {
      while (fgets(p,sz,stdin)) {
	sz-=strlen(p);
	p+=strlen(p);
	if (sz<1000) break;
      }
    }
    status=send_message(&env,serverURL,msg);
  }
    
  stop_client();

  if (env.fault_occurred) {
    fprintf(stderr, "XML-RPC Fault: %s (%d)\n",
	    env.fault_string, env.fault_code);
    status=130;
  }
  return(status);

}

static void start_client(xmlrpc_env *env,char *name, char *version)
{
  xmlrpc_env_init(env);

  /* Initialize our error-handling environment. */
  xmlrpc_env_init(env);
  xmlrpc_client_init2(env, XMLRPC_CLIENT_NO_FLAGS, name, version, NULL, 0);
}

static void stop_client(void)
{
  xmlrpc_client_cleanup();
}


static int send_message(xmlrpc_env *env,char *serverURL,char *message)
{
  char *line=NULL,*cmdPrompt=NULL;
  xmlrpc_value *result=NULL,*cmdResponse,*string;
  int i,rpcStatus,nl;

  /* Make the remote procedure call, 
     retry on "conncetion refused" (error -504) on first call. */

  for (i=0;i<Ntry;i++) {
    result = xmlrpc_client_call(env, serverURL, "wnp.cmd", "(s)", message);
    if (!env->fault_occurred) break;
    if (i<Ntry-1 &&
	env->fault_code==-504 &&
	(strstr(env->fault_string,"Connection refused")||
	 strstr(env->fault_string,"failed to perform HTTP POST request")||
	 strstr(env->fault_string,"failed to execute the HTTP POST transaction")) ) {
      xmlrpc_env_clean(env);
      xmlrpc_env_init(env);
      uSleep(NAPTIME);
    }
  }
  Ntry=1;

  /* Parse the result */
  cmdPrompt=NULL; cmdResponse=NULL;
  if (!env->fault_occurred) {
    if (xmlrpc_value_type(result)==XMLRPC_TYPE_ARRAY) {
      if (xmlrpc_array_size(env,result)>=3) {
	xmlrpc_decompose_value(env,result,"(isA)",&rpcStatus,&cmdPrompt,&cmdResponse);
	if (!env->fault_occurred) {
	  if (rpcStatus>=0) {
	    if (xmlrpc_value_type(cmdResponse)==XMLRPC_TYPE_ARRAY) {
	      if (verbose) {
		nl=xmlrpc_array_size(env,cmdResponse);
		for (i=0;i<nl;i++) {
		  xmlrpc_array_read_item(env,cmdResponse,i,&string);
		  if (!env->fault_occurred) {
		    if (xmlrpc_value_type(string)==XMLRPC_TYPE_STRING) {
		      xmlrpc_read_string(env,string,(const char **)&line);
		      if (!env->fault_occurred) {
			fprintf(stdout,"%s%s",line,(strlen(line)==0||line[strlen(line)-1]!='\n')?"\n":"");
			free(line); line=NULL;
		      }
		    }
		    xmlrpc_DECREF(string);
		  }
		}
	      }
	      if (prompt&&cmdPrompt) fprintf(stdout,"%s",cmdPrompt);
	    }
	  }
	  if (cmdPrompt) free(cmdPrompt);
	  if (cmdResponse) xmlrpc_DECREF(cmdResponse);
	}
      }
    }
    xmlrpc_DECREF(result);
  }
  if (env->fault_occurred) rpcStatus=-3;
  
  if (rpcStatus>127) rpcStatus=127;
  if (rpcStatus<0) rpcStatus=127-rpcStatus;

  return(rpcStatus);
}

#if defined (sgi)
static void uSleep(unsigned int usec)
{
  long ticks;
  
  ticks=(long)(usec/1000000.0)*CLK_TCK;
  sginap(ticks);
}
#else
#if defined (mingw)
static void uSleep(unsigned int usec)
{
  int ticks;
  
  ticks=(usec/1000000.0)*CLOCKS_PER_SEC;
  sleep(ticks);
}
#else
static void uSleep(unsigned int usec)
{
  usleep(usec);
}
#endif
#endif

#else

#include <stdlib.h>
#include <stdio.h>

int main(int argc, char **argv) 
{
  fprintf(stderr,"**** non-functional wnpcmd compiled without XML/RPC support\n"
	  "compile with -DwithRPC flag to generate a functional version.\n");
  return(1);
}

#endif
