#include <stdio.h>
#include <stdlib.h>
#include "wnprandom.h"
#include "random.h"

int WnpRandomKSubset(int K,int N,int *a)
{
  /*

    Return a random K-subset of {0,1,...,N-1}.

    K: size of desired random subset (input)
    N: size of set from which selection is made (input)

    a: indices [0...N-1] of subset members in ascending order.

    Adapted from an algorithm in "Combinatorial Algorithms : An Update 
    (CBMS-NSF Regional Conference Series in Applied Mathematics) 
    by Herbert S. Wilf."

  */
		 
  int *b,p,r=0,s,t,c,i,l,m=0,x,status;
  double n,k;

  n=N;
  k=K;
  status=0;

  if (K==N) {

    /* select all */

    for (i=0;i<k;i++) a[i]=i;

  }else if (K>N||K<0) {
    
    /* no valid selection possible */

    status=1;

  }else if (K==1) {

    /* select 1   */

    a[0]=N*WnpRandU();

  }else {

    status=1;
    if (K>N-K) {                           
      b=(int *)malloc((N-K)*sizeof(int));
      if (b) {
	if ((status=WnpRandomKSubset(N-K,N,b))==0) { /* K<N/2 : select complement */
	  for (i=0,r=0,l=0;i<N;i++) {
	    if (r<N-K&&b[r]==i) r++;                 /* pick if not in complement */
	    else a[l++]=i;
	  }
	}
	free(b);
      }
    }

    if (status) {
      status=0;

      /* set up lower bounds for k equal sized bins */

      for (i=0;i<k;i++) a[i]=i*n/k;

      /* randomly select bin occupancy number, total occupancy = k */

      for (c=0;c<k;) {
	x=n*WnpRandU();
	l=x*k/n;
	t=(l+1)*n/k;
	if (t<=x) l++;
	if(x>=a[l]) {
	  a[l]=a[l]+1;
	  c++;
	}
      }

      /* collect augmented lower bounds of occupied bins */

      for (p=0,i=0;i<k;i++) if (a[i]!=(int)(i*n/k)) a[p++]=a[i]-1;
      for (i=p;i<k;i++) a[i]=-1;
      p--;

      /* convert augmented lower bounds to bin numbers */

      for(s=k-1;p>=0;p--) {
	l=a[p]*k/n;
	t=(l+1)*n/k;
	if (t<=a[p]) l++;
	t=l*n/k; t=a[p]-t+1;
	a[p]=-1;
	a[s]=l;
	s-=t;
      }

      /* fill each bin with set members */

      for (l=k-1;l>=0;l--) {
	if (a[l]>=0) {
	  m=a[l]*n/k;                       /* lower bound */
	  s=(a[l]+1)*n/k-m;                 /* bin size    */
	  r=l;
	}
	x=m+s*WnpRandU();
	for (i=l+1;r>=0;i++) {
	  if (i>r||x<a[i]) {
	    a[i-1]=x;
	    s--;
	    break;
	  }else{
	    x++;
	    a[i-1]=a[i];
	  }
	}
      }
    }
  }
  return(status);
}

/*-------------------- WnpRandomSeed --------------------*/

void WnpRandomSeed(int seed)
{
  setSeed((unsigned int)seed);
}

/*-------------------- WnpRandU --------------------*/

double WnpRandU(void)
{
  return(ufrand(0,1));
}
