#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include "mystdlib.h"
#include "object.h"
#include "swdef.h"
#include "strutil.h"
#include "errhand.h"

#define VX 0
#define V1 1
#define V2 2
#define V3 3

#define NO 0
#define YES 1
#define Return(x) {fclose(out);return(x);}

static char *FF=(char *)NULL;

#define MAXBUF 2048

/*========================================================================*/

/*   Write molecule to file in BIOSYM .CAR/.MDF format   */

int write_bio(char *file,char *hdr1,char *hdr2,Node *m)
{
  Atom *a,*b;
  Atomset *S;
  Node *oldres,*oldmol,*res,*mol;
  FILE *out;
  char filename[MAXBUF],rnam[5],rnum[6],brnam[5],brnum[6],ffmnem[4],elts[3];
  char *mol_name,buf[MAXBUF];
  int status,nat,i,l,pend,version,chrg,nsets;
  time_t now;
  float bo;
  
  /*   1. Write .CAR File   */

  status=0;
  extfil(filename,file,".car",NO);
  if ((out=fopen(filename,"w"))==NULL) return(1);
  
  if (strncmp(hdr1,"!BIOSYM archive 1",strlen("!BIOSYM archive 1"))==0) {
    version=V1;
  }else if (strncmp(hdr1,"!BIOSYM archive 2",strlen("!BIOSYM archive 2"))==0) {
    version=V2;
  }else if (strncmp(hdr1,"!BIOSYM archive 3",strlen("!BIOSYM archive 3"))==0) {
    version=V3;
  }else{
    version=VX;
  }

  switch (version) {
  case V1:;
  case V2:;
  case V3:
    status=fprintf(out,"%-80s\n",hdr1);
    status=fprintf(out,"%s\n","PBC=OFF");
    status=fprintf(out,"%-80s\n",hdr2);
    time(&now);
    status=fprintf(out,"!DATE %s",asctime(localtime(&now)));
    break;
  case VX:
    status=fprintf(out,"%-80s\n",hdr1);
    time(&now);
    status=fprintf(out,"!DATE %s",asctime(localtime(&now)));
    break;
  }
  if (status==EOF) Return(2);
  
  nat=0;
  oldres=(Node *)NULL;
  oldmol=(Node *)NULL;
  pend=FALSE;
  
  ForAtoms(a,m) {
    res=a->up;
    if (oldres!=res) {
      mol=(m==res ? res : res->up);
      make_bio_resnam(res->name,rnam,rnum);
      if (oldmol!=mol) {
	if (pend) fprintf(out,"end\n");
	pend=FALSE;
	oldmol=mol;
      }
      oldres=res;
    }
    nat++;
    if (mvld(a->atp)) {
      if ((i=strlen(a->atp->name)-2)>=0) {
	if (streq(a->atp->name+i,".?")) {
	  Strcpy(ffmnem,"?");
	}else if (strlen(a->atp->name)>3) {
	  Strcpy(ffmnem,"?");
	}else {
	  Strcpy(ffmnem,a->atp->name); 
	}
      }else{
	Strcpy(ffmnem,a->atp->name);
      }
      Strcpy(elts,a->atp->chs);
    }else{
      Strcpy(ffmnem,"?");
      Strcpy(elts,"?");
    }
    switch (version) {
    case VX:;
    case V1:
      status=fprintf(out,
		     "%-4s %15.9f%15.9f%15.9f %-4s %4s %-4s %8.4f %5d\n",
		     a->name,a->x[0],a->x[1],a->x[2],
		     rnam,rnum,ffmnem,a->q,nat);

      break;
    case V2:
      status=fprintf(out,
		     "%-4s %15.9f%15.9f%15.9f %-4s%4s  %-3s%-2s %8.4f%5d\n",
		     a->name,a->x[0],a->x[1],a->x[2],
		     rnam,rnum,ffmnem,elts,a->q,nat);
      break;
    case V3:
      status=fprintf(out,
		     "%-5s%15.9f%15.9f%15.9f %-4s %-7s%-7s %-2s %6.3f\n",
		     a->name,a->x[0],a->x[1],a->x[2],
		     rnam,rnum,ffmnem,elts,a->q);
    }
    if (status==EOF) Return(2);
    pend=TRUE;
  }
  if (pend) if (fprintf(out,"end\n")==EOF) Return(2);

  if (fprintf(out,"end\n")==EOF) Return(2);
  
  status=0;
  if (ferror(out)) {
    err_set(1,"wrcar: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrcar: error closing file %s",filename);
    status=5;
  }

  if (status) return(status);

  /*   2. And now for that MDF file...   */
  
  extfil(filename,file,".mdf",YES);
  if ((out=fopen(filename,"w"))==NULL) return(3);
  
  switch (version) {
  case VX:
  case V1:
    if (fprintf(out,"VERSION   2.2\n")==EOF) Return(4);
    if (fprintf(out,"!%s\n",hdr1)==EOF) Return(4);
    if (fprintf(out,"!%s\n",hdr2)==EOF) Return(4);
    break;
  case V2:
    if (fprintf(out,"!BIOSYM molecular_data 3\n")==EOF) Return(4);
    if (fprintf(out,"!%s\n",hdr1)==EOF) Return(4);
    if (fprintf(out,"!%s\n",hdr2)==EOF) Return(4);
    if (fprintf(out,"\n#topology\n\n")==EOF) Return(4);
    if (fprintf(out,"\n")==EOF) Return(4);
    if (fprintf(out,"@column 1 element\n")==EOF) Return(4);
    if (fprintf(out,"@column 2 atom_type %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 3 charge_group %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 4 isotope\n")==EOF) Return(4);
    if (fprintf(out,"@column 5 formal_charge\n")==EOF) Return(4);
    if (fprintf(out,"@column 6 charge %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 7 switching_atom %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 8 oop_flag %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 9 chirality_flag \n")==EOF) Return(4);
    if (fprintf(out,"@column 10 occupancy \n")==EOF) Return(4);
    if (fprintf(out,"@column 11 xray_temp_factor\n")==EOF) Return(4);
    if (fprintf(out,"@column 12 connections \n")==EOF) Return(4);
    break;
  case V3:
    if (fprintf(out,"!BIOSYM molecular_data 4\n")==EOF) Return(4);
    if (fprintf(out,"!%s\n",hdr1)==EOF) Return(4);
    if (fprintf(out,"!%s\n",hdr2)==EOF) Return(4);
    if (fprintf(out,"\n#topology\n\n")==EOF) Return(4);
    if (fprintf(out,"\n")==EOF) Return(4);
    if (fprintf(out,"@column 1 element\n")==EOF) Return(4);
    if (fprintf(out,"@column 2 atom_type %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 3 charge_group %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 4 isotope\n")==EOF) Return(4);
    if (fprintf(out,"@column 5 formal_charge\n")==EOF) Return(4);
    if (fprintf(out,"@column 6 charge %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 7 switching_atom %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 8 oop_flag %s\n",(FF?FF:"cvff"))==EOF) Return(4);
    if (fprintf(out,"@column 9 chirality_flag \n")==EOF) Return(4);
    if (fprintf(out,"@column 10 occupancy \n")==EOF) Return(4);
    if (fprintf(out,"@column 11 xray_temp_factor\n")==EOF) Return(4);
    if (fprintf(out,"@column 12 connections \n")==EOF) Return(4);
    break;
  }    
  
  nat=0;
  oldres=(Node *)NULL;
  oldmol=(Node *)NULL;
  pend=FALSE;
  
  ForAtoms(a,m) {
    res=a->up;
    if (oldres!=res) {
      mol=(m==res ? res : res->up);
      make_bio_resnam(res->name,rnam,rnum);
      if (oldmol!=mol) {
	mol_name = (mvld(mol->name) ? mol->name : "");
	switch (version) {
	case VX:
	case V1:
	  if (pend) fprintf(out,"END\n");
	  break;
	case V2:
	case V3:
	  fprintf(out,"\n@molecule %s\n\n",mol_name);
	  break;
	}
	pend=FALSE;
	oldmol=mol;
      }
      oldres=res;
    }
    nat++;
    if (mvld(a->atp)) {
      if ((i=strlen(a->atp->name)-2)>=0) {
	if (streq(a->atp->name+i,".?")) {
	  Strcpy(ffmnem,"?");
	}else if (strlen(a->atp->name)>3) {
	  Strcpy(ffmnem,"?");
	}else {
	  Strcpy(ffmnem,a->atp->name); 
	}
      }else{
	Strcpy(ffmnem,a->atp->name);
      }
      Strcpy(elts,a->atp->chs);
    }else{
      Strcpy(ffmnem,"?");
      Strcpy(elts,"?");
    }
    chrg=(a->chrg=='?' ? 0 : a->chrg);
    switch (version) {
    case VX:
    case V1:
      status=fprintf(out,"ATOM  %-4s %-4s%-4s %-4s %-4s %7.4f %2d%2d%2d%2d ",
		     a->name,ffmnem,a->grp,rnam,rnum,a->q,a->cosw,
		     (a->nbnd==3 ? a->oop : 0),a->fe,a->nbnd); 
      break;
    case V2:
      sprintf(buf,"%s_%s:%s",rnam,rnum,a->name);
      status=fprintf(out,
		     "%-20s %-2s %-4s %-4s 0 %3d %7.4f %d %d 8 %7.4f %7.4f ",
		     buf,elts,ffmnem,a->grp,chrg,a->q,a->cosw,
		     (a->nbnd==3 ? a->oop : 0),a->x[ATM_SOF],a->x[ATM_BISO]);
      break;
    case V3:
      sprintf(buf,"%s_%s:%s",rnam,rnum,a->name);
      status=fprintf(out,
		     "%-20s%-2s %-7s %-4s 0 %3d %7.4f %d %d 8 %7.4f %7.4f",
		     buf,elts,ffmnem,a->grp,chrg,a->q,a->cosw,
		     (a->nbnd==3 ? a->oop : 0),a->x[ATM_SOF],a->x[ATM_BISO]);
      break;
    }
    if (status==EOF) Return(4);
    for(i=0;i<a->nbnd;i++) {
      b = (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f);
      if (b->flg&(1<<1)) {
	if (b->up!=res) {
	  make_bio_resnam(b->up->name,brnam,brnum);
	  status=fprintf(out," %s_%s:%s",brnam,brnum,b->name);
	  if (status==EOF) Return(4);
	}else{
	  status=fprintf(out," %s",b->name);
	  if (status==EOF) Return(4);
	}
	bo=(a->bnd[i]->bt==BT_AROMATIC ? 1.5 : a->bnd[i]->bo); 
	if (bo!=1.0) if (fprintf(out,"/%3.1f",bo)==EOF) Return(4);
      }
    }
    pend=TRUE;
    if (fprintf(out,"\n")==EOF) Return(4);
  }
  
  switch (version) {
  case VX:
  case V1:
    if (fprintf(out,"END\n")==EOF) Return(4);
    break;
  case V2:;    /* fall-through */
  case V3:
    nsets=0;
    forAtomsets(S,m) {
      l=0;
      nsets++; oldmol=(Node *)NULL;
      if (nsets==1) fprintf(out,"\n#atomset\n");
      fprintf(out,"\n@list subset %s\n",S->name);
      ForAtoms(a,m) {
	if (in_atomset(a,S)) {
	  res=a->up; 
	  mol=(m==res ? res : res->up);
	  make_bio_resnam(res->name,rnam,rnum);
	  if (!mvld(a->prv)) {
	    forAtoms(b,a->up) if (!in_atomset(b,S)) break;
	  }else b=a;
	  if (mvld(b)) {
	    if (oldmol==mol) {
	      if (l>0) fprintf(out,"%s\n",buf);
	      l=sprintf(buf,"%s_%s:%s ",rnam,rnum,a->name);
	    }else {
	      if (l>0) fprintf(out,"%s\n",buf);
	      l=sprintf(buf,"%s:%s_%s:%s ",mol->name,rnam,rnum,a->name); 
	    }
	    if (l>70) {
	      fprintf(out,"%s\n",buf);
	      l=0;
	    }
	    for (a=(Atom *)a->nxt;mvld(a);a=(Atom *)a->nxt) {
	      if (is_atom(a)) {
		if (in_atomset(a,S)) {
		  l+=sprintf(buf+l,"%s ",a->name);
		  if (l>70) {
		    fprintf(out,"%s\n",buf);
		    l=0;
		  }
		}
	      }
	    }
	  }else{
	    if (oldmol==mol) {
	      if (l>0) fprintf(out,"%s\n",buf);
	      l=sprintf(buf,"\n%s_%s:* ",rnam,rnum);
	    }else {
	      if (l>0) fprintf(out,"%s\n",buf);
	      l=sprintf(buf,"\n%s:%s_%s:* ",mol->name,rnam,rnum); 
	    }
	    if (l>70) {
	      fprintf(out,"%s\n",buf);
	      l=0;
	    }
	  }
	  a=(Atom *)res->lst;
	  oldmol=mol;
	}
      }
      if (l>0) fprintf(out,"%s\n",buf);
    }
    if (fprintf(out,"\n#end\n")==EOF) Return(4);
    break;
  }
  status=0;
  if (ferror(out)) {
    err_set(1,"wrmdf: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrmdf: error closing file %s",filename);
    status=5;
  }

  return(status);
}
/*-------------------- make_bio_resnam --------------------*/

void make_bio_resnam(char *name,char *bionm,char *bionu)
{
  /* generate BIOSYM residue name and number from internal residue name */

  char tmp[20],*p;
  int i;

  Strcpy(tmp,name);
  for (i=0,p=tmp;i<4&&*p;) {
    if (*p=='_'||*p==' ') break;
    bionm[i++]=*p++;
  }
  if (*p=='_'||*p==' ') p++;
  bionm[i]='\0';

  for (i=0;i<4&&*p;) {
    if (*p=='_'||*p==' ') break;
    bionu[i++]=*p++;
  }
  bionu[i]='\0';

  if (strlen(bionm)==0) sprintf(bionm,"UNK");
  if (strlen(bionu)==0) sprintf(bionu,"0");

}  
/*-------------------- setBiosymFF --------------------*/

void setBiosymFF(char *mnem)
{
  if (mvld(FF)) free(FF);
  FF=strdup(mnem);
}
