#include <stdio.h>
#include <stdarg.h>
#include "mystdlib.h"
#include "strutil.h"
#include "object.h"
#include "swdef.h"
#include "errhand.h"
  
#define Return(x) {if (out) fclose(out);return(x);}        
#define CHG(x) (x<0 ? 10000*x - 0.001 : 10000*x + 0.001)
#define YES 1
#define NO 0
#define MXBT 8

static int output(FILE *out,char *buffer)
{
  int r=0;

  if (out) {
    r=fprintf(out,"%s",buffer);
  }else{
    prtGraphics("%s",buffer);
    r=strlen(buffer);
  }
  return r;
}

static char *outl(char *buff,char *format,...)
{
  va_list args;
  va_start(args,format);
  vsprintf(buff,format,args);
  va_end(args);
  return buff;
}

/***************************************************************************/
int write_maccs(char *file,Node *mol,int append)
{
  FILE *out;
  Atom *a,*b;
  Bond *bnd;
  int i,nat,k,nbnd,status,stereo,chrg,bord;
  char buff[4096];

  char filename[4096];
  
  status=0;
  if (!is_mol(mol)) {
    err_set(1,"bad mol-pointer passed to write_maccs\n");
    return(2);
  }

  /*  Count number of atoms and bonds  */
  
  nat=0;nbnd=0; 
  ForAtoms(a,mol) {
    if (a->flg&(1<<1)) {
      nat++;
      for (i=0;i<a->nbnd;i++) {
	if (a!=a->bnd[i]->t && a->bnd[i]->t->flg&(1<<1)) nbnd++;
      }
      a->seqnr=nat;
    }
  }

  if (nat>255) {
   err_set(1,"%s: too many atoms (%d), max. allowed is 255\n",(nat>999 ? "Error":"Warning"),nat);
    if (nat>999) return(2);
  }
  if (nbnd>255) {
    err_set(1,"%s: too many bonds (%d), max. allowed is 255\n",(nbnd>999 ? "Error":"Warning"),nbnd);
    if (nbnd>999) return(2);
  }
  
  /* Open outputfile ...  */
  
  if (!streq(file,"-")) {
    extfil(filename,file,".sdf",NO);
    if ((out=fopen(filename,append?"a":"w"))==NULL) return(1);
  }else{
    out=NULL;
  }

  /* Write the header records and counts line...  */
  
  if (mvld(mol->name)) output(out,outl(buff,"%s\n",mol->name));
  else output(out,outl(buff,"NoName\n"));
  output(out,outl(buff,"  WitNotP           3D     1.00000\n"));
  output(out,outl(buff,"MACCS Structure Data File generated by Wit!P\n"));
  output(out,outl(buff,"%3d%3d  0  0  0  0            %3d V2000\n",nat,nbnd,999));
  
  /* Write the atom block... */

  ForAtoms(a,mol) {
    if (a->flg&(1<<1)) {
      stereo=0; 
      if (a->chrg=='?'||a->chrg==0) chrg=0;
      else chrg=4-a->chrg;
      status=output(out,outl(buff,"%10.4f%10.4f%10.4f %-2s  0%3d%3d\n",
		     a->x[0],a->x[1],a->x[2],
			     (mvld(a->atp)?a->atp->chs:"Xx"),
			     chrg,stereo));
      if (status==EOF) Return(3);
    }
  }
  
  /* Bonds are next... */
  
  k=localize_aromatic_bonds(mol);
  if (k>0) err_set(1,"warning: failed to localize aromatic %d bonds.\n",k);

  ForAtoms(a,mol) {
    for (k=0;k<a->nbnd;k++) {
      bnd=a->bnd[k];
      b=(bnd->f == a ? bnd->t : bnd->f);
      if (b->seqnr > a->seqnr && b->flg&(1<<1)) {
	bord=bnd->bo;
	if (bord=='?'&&b->arom) bord=4;
	if (bord<1||bord>4) bord=0;
	status=output(out,outl(buff,"%3d%3d%3d\n",
			       a->seqnr,b->seqnr,bord));
	if (status==EOF) Return(3);
      }
    }
  }

  /* Properties block... */

  output(out,outl(buff,"M  END\n"));

  /* done... */

  status=output(out,outl(buff,"$$$$\n"));
  if (status==EOF) Return(3);

  if (out) fclose(out);
  return(0);
}
