#include <stdio.h>
#include "mystdlib.h"
#include "strutil.h"
#include "object.h"
#include "swdef.h"
#include "errhand.h"
  
#define Return(x) {fclose(out);return(x);}        
#define CHG(x) (x<0 ? 10000*x - 0.001 : 10000*x + 0.001)
#define YES 1
#define NO 0
#define MXBT 8

typedef struct TAFFtype {  /* hard-wired mnem/numtype table */
  char *mnem;
  int  numtp;
} TAFFtype;

static TAFFtype taff_type[] =
{
 {"C.3",   1},{"C.2",   2},{"C.1",   4},{"C.ar",  3},{"C.cat",33},
 {"H",    13},{"N.4",  31},{"N.3",   5},{"N.2",   6},{"N.1",   7},
 {"N.ar", 11},{"N.pl3",19},{"N.am", 28},{"O.3",   8},{"O.2",   9},
 {"O.co2",32},{"S.3",  10},{"S.2",  18},{"S.o",  29},{"S.o2", 30},
 {"P.3",  12},{"Br",   14},{"Cl",   15},{"F",    16},{"I",    17},
 {"Lp",   20},{"Na",   21},{"K",    22},{"Ca",   23},{"Li",   24},
 {"Al",   25},{"Du",   26},{"Si",   27},{"Zn",   95},{"B.pl3",96},
 {"B.3",  97},{"Se",   98},{"D",    99},
 {(char *)NULL,0}
};

/***************************************************************************/
int write_mol(char *file,Node *mol,int append)
{
  FILE *out;
  Atom *a,*b;
  Bond *bnd;
  int i,nat,k,nbnd,chg_val,status,nfeat,numtp;
  static int btt[MXBT] = {8,1,2,3,5,4,10,9};
  char filename[4096];
  
  /* Open outputfile ...  */
  
  extfil(filename,file,".mol",NO);
  if ((out=fopen(filename,append?"a":"w"))==NULL) return(1);
  
  /*  Count number of atoms and bonds  */
  
  nat=0;nbnd=0;chg_val=NO;
  ForAtoms(a,mol) {
    if (a->flg&(1<<1)) {
      nat++;
      for (i=0;i<a->nbnd;i++) {
	if (a!=a->bnd[i]->t && a->bnd[i]->t->flg&(1<<1)) nbnd++;
      }
      a->seqnr=nat;
      chg_val |= a->q!=0.0;
    }
  }
  
  /*   And now we write the first header record and all the atoms too...  */
  
  if (fprintf(out,"%4d MOL          %-20s%15d\n",nat,mol->name,0)==EOF) Return(3);
  
  ForAtoms(a,mol) {
    if (a->flg&(1<<1)) {
      if (mvld(a->atp)) {
	for (i=0;taff_type[i].mnem;i++) if (streq(a->atp->name,taff_type[i].mnem)) break;
	numtp=(taff_type[i].mnem ? taff_type[i].numtp : a->atp->numtp);
      }else numtp=0;
      status=fprintf(out,"%4d%4d%9.4f%9.4f%9.4f%s\n",
		     a->seqnr,numtp,a->x[0],a->x[1],a->x[2],a->name);
      if (status==EOF) Return(3);
    }
  }
  
  /*   Bonds are next... */
  
  if (fprintf(out,"%4d MOL\n",nbnd)==EOF) Return(3);
  
  i=0;
  ForAtoms(a,mol) {
    for (k=0;k<a->nbnd;k++) {
      bnd=a->bnd[k];
      b=(bnd->f == a ? bnd->t : bnd->f);
      if (b->seqnr > a->seqnr && b->flg&(1<<1)) {
	status=fprintf(out,"%4d%4d%4d         %4d\n",
		       ++i,a->seqnr,b->seqnr,btt[bnd->bt]);
	if (status==EOF) Return(3);
      }
    }
  }
  
  /*   and now do the charges...  */
  
  nfeat=0;
  if (chg_val) {nfeat+=(nat+4)/5;if (nat<=5) nfeat++;}
  if (fprintf(out,"%4d MOL\n",nfeat)==EOF) Return(3);
  if (chg_val) {
    i=0;
    ForAtoms(a,mol) {
      if (i%5 == 0) if (fprintf(out,"   4")==EOF) Return(3);
      if (fprintf(out,"%6ld",(long)CHG(a->q))==EOF) Return(3);
      i++;
      if (i%5 == 0) {
	if (i==5) {
	  if (fprintf(out,"%8.1f",(float)nat)==EOF) Return(3);
	  if (fprintf(out,"     0.0\n")==EOF) Return(3);
	}else if (i==10) {
	  if (fprintf(out,"     9.0     0.0\n")==EOF) Return(3);
	}else{
	  if (fprintf(out,"     0.0     0.0\n")==EOF) Return(3);
	}
      }
    }
    if (nat<5) {
      do {
	if (fprintf(out,"     0")==EOF) Return(3);
	i++;
      }while(i<5);
      if (fprintf(out,"%8.1f     0.0\n",(float)nat)==EOF) Return(3);
    }
    if (nat<10) {
      if (fprintf(out,"   4")==EOF) Return(3);
      do {
	if (fprintf(out,"     0")==EOF) Return(3);
	i++;
      }while(i<10);
      if (fprintf(out,"     9.0     0.0\n")==EOF) Return(3);
    }
    if (i%5 != 0) {
      while(i%5 != 0) {
	if (fprintf(out,"     0")==EOF) Return(3);
	i++;
      }
      if (fprintf(out,"     0.0     0.0\n")==EOF) Return(3);
    }
  }
  
  status=0;
  if (ferror(out)) {
    err_set(1,"wrmol: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrmol: error closing file %s",filename);
    status=5;
  }

  return(status);
}
