#include <stdio.h>
#include <stddef.h>
#include <ctype.h>
#include <stdarg.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "strutil.h"
#include "errhand.h"

/* EXP1: write substructure fields in ATOM section, 
   don't collect level 1 atoms into substructure.      19-MAY-2003 --> breaks SYBYL (Nov 2003)*/

#define EXP1

#define YES 1
#define NO  0

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static int output(FILE *out,char *buffer)
{
  int r=0;

  if (out) {
    r=fprintf(out,"%s",buffer);
  }else{
    prtGraphics("%s",buffer);
    r=strlen(buffer);
  }
  return r;
}

static char *outl(char *buff,char *format,...)
{
  va_list args;
  va_start(args,format);
  vsprintf(buff,format,args);
  va_end(args);
  return buff;
}

/*-------------------- write_mol2 --------------------*/

int write_mol2(char *file,Node *m,int append)
{
  /* write MOL2 file */
  
  /* QUICK&DIRTY, works for simple cases only... */

  FILE *out;
  Atom *a,*b,S;
  Atomset *s;
  Node *n,sroot,*sub;
  Bond broot,*bnd;
  int nib,status,i,j,k,l=0,nat,nbnd,nsub,nsets,nmemb,r,fn,ft,fs=0,fc;
  int ia,ib,nb;
#ifndef EXP1
  int nsubm;
#endif
  char filename[256],buff[512],buf2[512],*fill=(char *)NULL,*txtn,*txtt,
    *txts=(char *)NULL,*txtc,*bflg;

  /* open output file */

  if (!streq(file,"-")) {
    extfil(filename,file,".mol",NO);
    if ((out=fopen(filename,append?"a":"w"))==NULL) {
      err_set(1,"write_mol2: could not open %s for write\n",filename);
      return(1);
    }
  }else{
    out=NULL;
  }

  /* write header records */

  output(out,outl(buff,"# TRIPOS MOL2 file generated by WITNOTP\n\n"));

  /* count atoms, bonds, substructures, sets, put bonds in chain... */

  status=0;
  nat=0; nbnd=0; nsub=0; nsets=0; n=(Node *)NULL;
  bnd= &broot; bnd->chain=(Bond *)NULL;
  sub= &sroot; sub->chain=(Node *)NULL;
#ifndef EXP1
  nsubm=0;
#endif

  forAtomsets(s,m) nsets++;               /* only top level sets... */
  ForAtoms(a,m) {
    nat++; a->seqnr=nat;
    for (i=0;i<a->nbnd;i++) {
      if (a->bnd[i]->f==a && a->bnd[i]->t->flg&(1<<1)) {
	nbnd++;
	bnd->chain=a->bnd[i];
	bnd=a->bnd[i];
      }
    }
    if (a->up != n) {
      n=a->up;
      if (n!=m) {
	nsub++;
	sub->chain=a->up;
	sub=a->up;
      }else{
#ifndef EXP1
	if (!m->chain&&sub!=m) {
	  nsub++;nsubm=nsub;
	  sub->chain=m;
	  sub=m;
	}
#endif
      }
    }
  }
  bnd->chain=(Bond *)NULL;

  /* write MOLECULE section */

  output(out,outl(buff,"@<TRIPOS>MOLECULE\n"));
  output(out,outl(buff,"%s\n",(mvld(m->name) ? m->name : "****")));
  output(out,outl(buff," %d %d %d 0 %d\n",nat,nbnd,nsub,nsets));
  output(out,outl(buff,"****\n"));                                /* dictionary */
  output(out,outl(buff,"USER_CHARGES\n"));
  output(out,outl(buff,"INVALID_CHARGES\n\n"));


  /* write ATOM section */


  if (nat>0) {
    output(out,outl(buff,"@<TRIPOS>ATOM\n"));
    
    fill="                              ";
    fill=fill+strlen(fill);
    
    i=0;r=0;n=(Node *)NULL;
    ForAtoms(a,m) {
      if (a->up != n) {
	n=(a->up);
	if (n != m) {
	  r++;
	  l=r;
	  sprintf(buf2,"%s",(mvld(a->up->name) ? a->up->name : "****"));
	}else{
	  l=0;
#ifndef EXP1
	  l=nsubm;
#endif
	  sprintf(buf2,"****");
	}
	txts=buf2;
#ifndef EXP1
	txts="****";
#endif
	fs=strlen(txts); fs=(fs >8 ? 0 : 8-fs);
      }
      txtn=(mvld(a->name) ? a->name : "****");
      fn=strlen(txtn);
      for (k=0;k<fn;k++) if (!isspace(txtn[k])) break;
      if (k==fn) txtn="****";
      fn=strlen(txtn); fn=(fn >8 ? 0 : 8-fn);
      txtt=(mvld(a->atp) ? a->atp->name : "****");
      ft=strlen(txtt); ft=(ft >6 ? 0 : 6-ft);
      output(out,outl(buff,"%7d %s%s %9.5f %9.5f %9.5f %s%s %d %s%s %11.6f \n",
		      ++i,txtn,fill-fn,
		      a->x[0],a->x[1],a->x[2],
		      txtt,fill-ft,
		      l,txts,fill-fs,
		      a->q));
    }
  }

  /* formal charges (Wit!P extension of MOL2 file format) */

  ia=0;ib=0;nb=0;
  output(out,outl(buff,"\n#<Wit!P>CHARGES (formal)"));
  for(a=rfirst_atom(m);a;) {
    ++ia;
    if (a->chrg!='?') {
      output(out,outl(buff,"%s %+3d: %d",(nb%6==0?"\n#":""),a->chrg,ia));
      ib=ia;
      NextAtoms(b,m,a) {
  	if (b->chrg==a->chrg) ib++;
  	else break;
      }
      if (ib>ia) output(out,outl(buff,"-%d;",ib));
      else output(out,";");
      nb++; a=b; ia=ib;
    }else a=rnext_atom(m,a);
  }
  output(out,outl(buff,"\n"));

  /* ... and now the bonds ... */

  if (nbnd>0) {
    output(out,outl(buff,"\n@<TRIPOS>BOND\n"));
    for (i=1,bnd=broot.chain;mvld(bnd);bnd=bnd->chain,i++) {
      bflg=(bnd->f->up != bnd->t->up ? "INTERRES " : " ");
      output(out,outl(buff,"%7d %7d %7d %s %s\n",
      		      i,bnd->f->seqnr,bnd->t->seqnr,BT_mnem(bnd->bt),bflg));
    }
  }

  /* reset the bond list */

  for (bnd=broot.chain;mvld(bnd);bnd=broot.chain) {
    broot.chain=bnd->chain;
    bnd->chain=(Bond *)NULL;
  }

  /* ... and write substructures, if there are any... */

  if (nsub>0) {
    output(out,outl(buff,"\n@<TRIPOS>SUBSTRUCTURE\n"));
    for (i=1,sub=sroot.chain;mvld(sub);sub=sub->chain,i++) {
      a=rfirst_atom(sub); r=(mvld(a) ? a->seqnr : 0);
      nib=0;
      forAtoms(a,sub) {
  	for (j=0;j<a->nbnd;j++) {
  	  b=neighbour(j,a);
  	  if (b->flg&(1<<1)&&b->up!=sub) nib++;
  	}
      }
      sprintf(buf2,"%s",(mvld(sub->name) ? sub->name : "****"));
      txts=buf2;
      fs=strlen(txts); fs=(fs >8 ? 0 : 8-fs);
      if (sub->up != m) txtc=(mvld(sub->up->name) ? sub->up->name : "****");
      else txtc="****";
      fc=strlen(txtc); fc=(fc >12 ? 0 : 12-fc);
      output(out,outl(buff,"%7d %s%s %7d GROUP **** %s%s **** %d \n",
  		      i,txts,fill-fs,r,txtc,fill-fc,nib));
    }
  }

  /* ... next write top level atomsets, if any... */

  if (nsets>0) {
    output(out,outl(buff,"\n@<TRIPOS>SET\n"));
    forAtomsets(s,m) {
      nmemb=0; b=&S;
      ForAtoms(a,m) {                  /* count and chain members */
	if (in_atomset(a,s)) {
	  nmemb++;
	  b->chain=(Node *)a;
	  b=a;
	}
      }
      b->chain=(Node *)NULL;
      output(out,outl(buff,"%s STATIC ATOMS <user>\n",s->name));
      output(out,outl(buff,"%d ",nmemb)); i=0;
      for (b=(Atom *)S.chain;mvld(b);b=(Atom *)S.chain) {
	S.chain=b->chain;
	i++;
	if (i%10 == 1) output(out,outl(buff," \\\n"));
	output(out,outl(buff," %d",b->seqnr));
	b->chain=(Node *)NULL;
      }
      output(out,outl(buff,"\n"));
    }
  }

  /* clean up... */

  for (sub=sroot.chain;mvld(sub);sub=sroot.chain) {
    sroot.chain=sub->chain;
    sub->chain=(Node *)NULL;
  }

  if (out&&ferror(out)) {
    err_set(1,"wrmol2: error writing to %s",filename);
    status=2;
  }
  if (out&&fclose(out)) {
    err_set(1,"wrmol2: error closing file %sn",filename);
    status=3;
  }

  return(status);
}  
