#include <stdio.h>
#include <stddef.h>
#include <ctype.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "strutil.h"
#include "errhand.h"

/* Finton's first try with Wit!p coding
   Write PDB file, with CONECT field, and hopefully following PDB guidelines
   http://www.rcsb.org/pdb/docs/format/pdbguide2.2/guide2.2_frame.html
   February 2005 */

#define YES 1
#define NO  0

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

extern Node *LibRoot;

static int need_conect(Bond *b);
static  Node *ResLib;
static  Lib  *reslib;

/*-------------------- write_pdb --------------------*/

int write_pdb(char *file, Node *m)
{
  /* write a PDB file */
  FILE * out;
  int status,nat,lev,l,i,no,maxlev;
  Atom *a,*b;
  Node *n;
  char filename[256],aname[5],rname[5],rnum[6],segid[5],chs[3],fchrg[3];

  ResLib=find_name(LibRoot,"ResLib",OBJ_LIB);
  if (!mvld(ResLib)) {
    err_set(-1,"Failed to open default library");
    return(-1);
  }
  reslib=(Lib *)ResLib;

  /* open output file */

  extfil(filename,file,".pdb",NO);
  if ((out=fopen(filename,"w"))==NULL) {
    err_set(1,"write_pdb: could not open %s for write\n",filename);
    return(1);
  }
  
  /* write header records */

  fprintf(out,"REMARK PDB coordinate file generated by WITNOTP\n");

  status=0; nat=0;
  maxlev=0;
  ForAtoms(a,m) {
    for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
    if (lev>maxlev) maxlev=lev;
  }
  ForAtoms(a,m) {
    for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
    if (lev>2&&lev==maxlev) {
      Strcpy(segid,a->up->up->name);
      n=a->up;
    }else{
      sprintf(segid," ");
      n=a->up;
    }
    make_xpl_resnam(a->up->name,rname,rnum);
    ForAtoms(b,n) {
      a=b;
      if (a->flg&(1<<1)) {
	nat++; 
	a->seqnr=nat;
	fprintf(out,"ATOM  %5d ",nat);
	if (mvld(a->name)){
	  Strcpy(aname,a->name);
	  if (strlen(aname)>3) fprintf(out,"%4s ",aname);
	  else if (mvld(a->atp)) {
	    if (a->atp->chs[1]=='\0') fprintf(out," %-3s ",aname);
	    else fprintf(out,"%-3s  ",aname);
	  }else fprintf(out," %-3s ",aname);
	}else fprintf(out,"     ");
	if (mvld(a->atp)) {
	  Strcpy(chs,a->atp->chs);
	  for (l=0;l<strlen(chs);l++) chs[l]=toupper(chs[l]);
	}else{
	  chs[0]='\0';
	}
	if (strlen(rname)==4) {
	  fprintf(out,"%-4s%c",rname,*segid);   /* this breaks PDB format 3.2 */
	}else{
	  fprintf(out,"%3s %c",rname,*segid);   /* reside name right-justified cols. 18-20 */
	}
	l=strlen(rnum);
	if ((l>0&&(rnum[l-1]>='A'))||l==5) fprintf(out,"%5s   ",rnum);
	else fprintf(out,"%4s    ",rnum);
	for (l=0;l<3;l++) fprintf(out,"%8.3f",a->x[l]);
	if (a->chrg>0&&a->chrg<10) {
	  fchrg[0]='0'+a->chrg;
	  fchrg[1]='+';
	}else if (a->chrg<0&&a->chrg>-10) {
	  fchrg[0]='0'-a->chrg;
	  fchrg[1]='-';
	}else{
	  fchrg[0]=' ';
	  fchrg[1]=' ';
	}
	fchrg[2]='\0';
	if (a->x[ATM_BISO]>999.99) {
	  fprintf(out,"%6.2f%6.1f      %-4s%2s%2s\n",
		  a->x[ATM_SOF],(a->x[ATM_BISO]>9999.9 ? 9999.9 : a->x[ATM_BISO]),segid,chs,fchrg);
	}else{	  
	  fprintf(out,"%6.2f%6.2f      %-4s%2s%2s\n",a->x[ATM_SOF],a->x[ATM_BISO],segid,chs,fchrg);
	}
      }
    }
  }

  /* write CONECT records */

  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (i=0;i<a->nbnd;i++) {
	b=neighbour(i,a);
	if (a<b&&b->flg&(1<<1)) {
	  if (need_conect(a->bnd[i])) {
	    a->flg|=(1<<2);
	    b->flg|=(1<<2);
	  }
	}
      }
    }
  }

  ForAtoms(a,m) {
    if (a->flg&(1<<2)) {
      a->flg&=~(1<<2);
      for (i=0,no=0;i<a->nbnd;i++) {
	b=neighbour(i,a);
	if (b->flg&(1<<1)) {
	  if (no==0) fprintf(out,"CONECT%5d",a->seqnr);
	  fprintf(out,"%5d",b->seqnr);
	  no++;
	  if (no==4) {
	    fprintf(out,"\n");
	    no=0;
	  }
	}
      }
      if (no>0) fprintf(out,"\n");
    }
  }
	
  fprintf(out,"END\n");
  if (ferror(out)) {
    err_set(1,"wrxpdb: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrxpdb: error closing file %s",filename);
    status=5;
  }
  return(status);
}
/*-------------------- need_conect --------------------*/

static int need_conect(Bond *b)
{
  int r,i;
  Node *fr,*tr,*lr,*lr2;
  Atom *a,*c;
  Link *L,*L2;
  float d;

  r=1;

  fr=b->f->up;
  tr=b->t->up;

  if (fr==tr) {
    if ((lr =find_residue_name((Node *)reslib,fr->name))) {
      ForAtoms(a,lr) if (strceq(a->name,b->f->name)) break;
      if (a) {
	for (i=0;i<a->nbnd;i++) {
	  c=neighbour(i,a);
	  if (strceq(c->name,b->t->name)) break;
	}
	if (i<a->nbnd) r=0;
      }
    }
  }else if (fr->nxt==tr) {
    if ((lr=find_residue_name((Node *)reslib,fr->name))&&
	(lr2=find_residue_name((Node *)reslib,tr->name))) {
      forLinks(L,lr) {
	if (find_link_name(lr2,L->name)) {
	  forLinks(L2,reslib) {
	    if (streq(L->name,L2->name)) {
	      if ((strceq(b->f->name,L2->from)&&strceq(b->t->name,L2->to))||
		  (strceq(b->t->name,L2->from)&&strceq(b->f->name,L2->to))) {
		d=cdist(b->f->x,b->t->x);
		if (d>=L2->dmin && d<=L2->dmax) {
		  r=0;
		  break;
		}
	      }
	    }
	  }
	}
	if (r==0) break;
      }
    }
  }
  return(r);
}
