#include <stdio.h>
#include "mystdlib.h"
#include "strutil.h"
#include "object.h"
#include "swdef.h"
#include "errhand.h"
#include "xtal.h"
  
#define TRUE 1
#define FALSE 0
#define YES 1
#define NO 0

#define DEFAULT_BISO 0.05

/***************************************************************************/
int write_shelx(char *file,Node *mol)
{
  FILE *out;
  Atom *a,*b;
  Node *n;
  Bond *bnd;
  Xcn  *xcn;
  Xatom *xa,*default_xa;
  ADfix *adf;
  int m,i,j,rec5,nafix,status;
  unsigned long chk;
  float x[3],bf0,sd;
  char *rt,buf[256];
  Atomset *s;

  char filename[256];
  
  status=0;
  if (!is_mol(mol)) {
    err_set(1,"write shelx: bad root.");
    return(1);
  }

  xcn=(Xcn *)NULL;
  for (n=(Node *)mol;n;n=n->up) {
    if (!is_mol(n)) continue;
    xcn=first_xcn(n);
    if (xcn) break;
  }

  if (!xcn) {
    err_set(2,"write shelx: no crytallographic info.");
    return(2);
  }
  default_xa=(Xatom *)NULL;

  /* convert peaks with chem symbol from formula to atoms... */

  ForAtoms(a,mol) {
    xa=first_xatom(a);
    if (mvld(xa)) {
      if (xa->peak) {
	if (mvld(a->atp)) {
	  for (i=0;i<xcn->forml;i++) {
	    if (strcasecmp(a->atp->chs,xcn->forms[i])==0) {
	      xa->peak=FALSE;
	      xa->iso=TRUE;
	      xa->bfac[0]=DEFAULT_BISO;
	      break;
	    }
	  }
	}
      }
    }
  }

  /* Open outputfile ...  */
  
  extfil(filename,file,".xcn",NO);
  if ((out=fopen(filename,"w"))==NULL) {
    err_set(3,"write shelx: could not open outputfile.");
    return(3);
  }

/* write structure independent info (XCN) */

  write_xcn(out,xcn);

  /* definition of atomsets defined at top level */

  sprintf(buf,"wit!p sets");
  forAtomsets(s,mol) {
    if (strlen(buf)+strlen(s->name)<250) {
      sprintf(buf+strlen(buf)," %s",s->name);
    }else{
      if (strlen(buf)>strlen("wit!p sets")) fprintf(out,"%s\n",buf);
      sprintf(buf,"wit!p sets");
      if (strlen(buf)+strlen(s->name)<250) sprintf(buf+strlen(buf)," set %s",s->name);
    }
  }
  if (strlen(buf)>strlen("wit!p sets")) fprintf(out,"%s\n",buf);

  /* write atom and peak info */

  default_xa=(Xatom *)make_node(OBJ_XATOM);

  ForAtoms(a,mol) {
    xa=first_xatom(a);
    if (!mvld(xa)) {
      xa=default_xa;
      if (xa) {
	if (a->x[ATM_BISO]>0.00001) xa->bfac[0]=a->x[ATM_BISO]/(8*3.141592653*3.141592653);
      }
    }
    if (!mvld(xa)) {
      err_set(101,"crystallographic atom info missing.");
      fclose(out);
      return(101);
    }
    for (i=0;i<3;i++) {
      x[i]=xcn->atf.r[i].c[3];
      for (j=0;j<3;j++) x[i]+=xcn->atf.r[i].c[j]*a->x[j];
    }
    rt=(xa->peak ? "peak" : "atom");
    fprintf(out,"%s 1 %s %f %f %f ",rt,a->name,x[0],x[1],x[2]);
    bf0=(xa->peak ? xa->peak_height : xa->bfac[0]);
    fprintf(out,"%f %f",a->x[ATM_SOF],bf0);
    if (xa->iso) fprintf(out,"\n");
    else fprintf(out," %f %f %f %f %f\n",
		 xa->bfac[1],xa->bfac[2],xa->bfac[3], xa->bfac[4],xa->bfac[5]);
    if (mvld(a->atp)) {
      Strcpy(buf,a->atp->name);
    }else{
      Strcpy(buf,"unk");
    }
    fprintf(out,"%s 2 %s",rt,buf);
    for (i=0;i<a->nbnd;i++) {
      bnd=a->bnd[i];
      b=(bnd->f==a ? bnd->t : bnd->f);
      if (b->flg&(1<<1)) fprintf(out," %s[%s]",b->name,BT_mnem(bnd->bt));
    }
    fprintf(out,"\n");
 
    /* atom 3 records have additional info for Wit!P */
    sprintf(buf,"%s 3",rt);
    if (a->chrg!='?') sprintf(buf,"%s 3 chrg %d q %0.4f",rt,a->chrg,a->q);
    else sprintf(buf,"%s 3 chrg ? q %0.4f",rt,a->q);
    for (chk=0,i=0;i<ATOMSET_WORDS;i++) chk|=((Mol *)mol)->local[i]&a->set[i];
    if (chk) {
      sprintf(buf+strlen(buf)," sets");
      forAtomsets(s,mol) {
	if (in_atomset(a,s)) {
	  if (strlen(buf)+strlen(s->name)<250) {
	    sprintf(buf+strlen(buf)," %s",s->name);
	  }else{
	    if (strlen(buf)>strlen("atom 3 sets")) fprintf(out,"%s\n",buf);
	    sprintf(buf,"%s 3 sets",rt);
	    if (strlen(buf)+strlen(s->name)<250) sprintf(buf+strlen(buf)," %s",s->name);
	  }
	}
      }
    }
    if (strlen(buf)>strlen("atom 3 sets")) fprintf(out,"%s\n",buf);

    /* the following record types are output only for atoms (not peaks) */
    if (!xa->peak) {
      sd=xa->esd_sof;
      for (i=0;i<3;i++) sd+=xa->esd_x[0];
      for (i=0;i<(xa->iso? 1 : 6);i++) sd+=xa->esd_bfac[i];
      if (sd>0) {
       fprintf(out,"%s 4 %f %f %f ",rt,xa->esd_x[0],xa->esd_x[1],xa->esd_x[2]);
       fprintf(out,"%f %f",xa->esd_sof,xa->esd_bfac[0]);
       if (!xa->iso) for (i=1;i<6;i++) fprintf(out," %f",xa->esd_bfac[i]);
       fprintf(out,"\n");
      }
      m=(xa->iso ? 5 : 10);
      rec5=FALSE;
      for (i=0;i<m;i++) if (xa->pfix[i]!=0.0) rec5=TRUE;
      if (rec5) {
	fprintf(out,"atom 5 ");
	for (j=0,i=0;i<3;i++,j++) {
	  if (xa->pfix[j]!=0.0) fprintf(out,"%f ",x[i]+10.0);
	  else fprintf(out,"0 ");
	}
	if (xa->pfix[j]!=0.0) fprintf(out,"%f ",a->x[ATM_SOF]+10.0);
	else fprintf(out,"0 ");
	j++;
	for (i=0;i<m-4;i++,j++) {
	  if (xa->pfix[j]!=0.0) fprintf(out,"%f ",xa->bfac[i]+10.0);
	  else fprintf(out,"0 ");
	}
	fprintf(out,"\n");
      }
    }

    sprintf(buf,"atom 6 ");
    if (xa->block>0) sprintf(buf+strlen(buf),"block %d ",xa->block);
    nafix=0;
    for (i=0;i<xa->nadf;i++) {
      adf=xa->adf[i];
      if (adf->f==xa) {
	switch(adf->class) {
	case ADFCLASS_DFIX:
	  sprintf(buf+strlen(buf),"dfix_to %s %9.4f %9.4f ",
		  adf->t->up->name,adf->dist,adf->esd);
	  break;
	case ADFCLASS_AFIX:
	  sprintf(buf+strlen(buf),"afix_to %s ",adf->t->up->name);
	  break;
	default:
	  break;
	}
      }else{
	if (adf->class==ADFCLASS_AFIX) nafix++;
      }
    }
    if (xa->afix_root&&nafix>0) {
      sprintf(buf+strlen(buf),"afix_root %d ",xa->afix_root);
    }
    if (strlen(buf)>strlen("atom 6 ")) fprintf(out,"%s\n",buf);
  }

  if (ferror(out)) {
    err_set(1,"wrcar: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrcar: error closing file %s",filename);
    status=5;
  }

  if (mvld(default_xa)) del_node((Node *)default_xa);
  return(status);
}

void write_xcn(FILE *out,Xcn *xcn)
{
  int i,j;
  Symop sym;
  hisrec *h;
  char buf[256];

/* write structure independent info (XCN) */

  if (out&&is_xcn(xcn)) {
    fprintf(out,"title %s\n",xcn->title);
    fprintf(out,"cell     %f %f %f %f %f %f %f\n",xcn->lambda,
	    xcn->cl[0],xcn->cl[1],xcn->cl[2],xcn->ca[0],xcn->ca[1],xcn->ca[2]);
    fprintf(out,"cell_esd %f %f %f %f %f %f\n",
	    xcn->esd_cl[0],xcn->esd_cl[1],xcn->esd_cl[2],
	    xcn->esd_ca[0],xcn->esd_ca[1],xcn->esd_ca[2]);
    fprintf(out,"formula");
    for (i=0;i<xcn->forml;i++) fprintf(out," %s %d",xcn->forms[i],(int)xcn->formn[i]);
    fprintf(out,"\n");
    i=xcn->lattice; if (!xcn->centric) i = -i;
    fprintf(out,"spgr %d %s\n",i,xcn->spgr);
    for (i=0;i<xcn->nsym;i++) {
      sym=symF(xcn,xcn->sym[i]);sym2str(buf,sym);
      if (!streq("X,Y,Z",buf)) fprintf(out,"symm %s\n",buf);
    }
    fprintf(out,"asymm %f %f %f %f %f %f\n",
	    xcn->asym[0],xcn->asym[1],xcn->asym[2],
	    xcn->asym[3],xcn->asym[4],xcn->asym[5]);
    if (xcn->scale) fprintf(out,"scale %s\n",xcn->scale);
    for (h=xcn->history;h;h=h->nxt) {
      fprintf(out,"history %s\n",h->t);
    }

    /* write cryst. to A-cart. transformation matrix into file */

    fprintf(out,"wit!p XtoA");
    for (i=0;i<3;i++) for(j=0;j<4;j++) fprintf(out," %0.6f",xcn->fta.r[i].c[j]);
    fprintf(out,"\n");
  }
}
