#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include "swdef.h"
#include "mystdlib.h"
#include "object.h"
#include "strutil.h"
#include "errhand.h"

#define YES 1
#define NO  0

#define MAXBUF 256

#define neighbour(i,a) (a->bnd[i]->f == a ? a->bnd[i]->t : a->bnd[i]->f)

static int linAtom(Atom *a,Bond *b,Bond *c);


/*-------------------- write_xplor --------------------*/

int write_xplor(char *file,Node *m)
{
  int s;

/* Write XPLOR PDB/PSF file pair (only "selected" atoms (i.e. flg_1 = 1) )*/

  s=write_xplor_kludge(file,m);   /* Note: this is for the NMR/DG kludge only... */
  return(s);
}
/*-------------------- write_pdb --------------------*/

int write_xplor_pdb(char *file,Node *m,int append)
{
  /* write XPLOR PDB file (only "selected" atoms (i.e. flg_1 = 1) )*/
  
  FILE *out;
  Atom *a,*b;
  Node *n;
  int status,l,nat,lev;
  char filename[256],aname[5],rname[5],rnum[6],segid[5];

  /* open output file */

  extfil(filename,file,".pdb",NO);
  if ((out=fopen(filename,append?"a":"w"))==NULL) {
    err_set(1,"write_xplor_pdb: could not open %s for write\n",filename);
    return(1);
  }

  /* write header records */

  fprintf(out,"REMARK XPLOR coordinate file generated by WITNOTP\n");

  status=0; nat=0;
  ForAtoms(a,m) {
    for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
    if (lev>2) {
      Strcpy(segid,a->up->up->name);
      n=a->up;
    }else{
      sprintf(segid," ");
      n=a->up;
    }
    ForAtoms(b,n) {
      a=b;
      if (a->flg&(1<<1)) {
	nat++; 
	fprintf(out,"ATOM  %5d ",nat);
	if (mvld(a->name)){
	  Strcpy(aname,a->name);
	  if (strlen(aname)>3) fprintf(out,"%4s ",aname);
	  else if (mvld(a->atp)) {
	    if (a->atp->chs[1]=='\0') fprintf(out," %-3s ",aname);
	    else fprintf(out,"%-3s  ",aname);
	  }else fprintf(out," %-3s ",aname);
	}else fprintf(out,"     ");
	make_xpl_resnam(a->up->name,rname,rnum);
	fprintf(out,"%-4s ",rname);
	l=strlen(rnum);
	if ((l>0&&(rnum[l-1]>='A'))||l==5) fprintf(out,"%5s   ",rnum);
	else fprintf(out,"%4s    ",rnum);
	for (l=0;l<3;l++) fprintf(out,"%8.3f",a->x[l]);
	if (a->x[ATM_BISO]>999.9) {
	  fprintf(out,"%6.2f%6.1f      %-4s\n",
		  a->x[ATM_SOF],(a->x[ATM_BISO]>9999.9 ? 9999.9 : a->x[ATM_BISO]),segid);
	}else{	  
	  fprintf(out,"%6.2f%6.2f      %-4s\n",a->x[ATM_SOF],a->x[ATM_BISO],segid);
	}
      }
    }
  }
  fprintf(out,"END\n");
  if (ferror(out)) {
    err_set(1,"wrxpdb: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrxpdb: error closing file %s",filename);
    status=5;
  }
  return(status);
}  
/*-------------------- write_xplor_psf --------------------*/

int write_xplor_psf(char *file,Node *m)
{
  /* write XPLOR/PSF file (only "selected" atoms (i.e. flg_1 = 1) )*/
  /* atoms with flg_2 set have special treatment (hydrogens in H-Bonds) */

  FILE *out;
  int status,lev;
  long where;
  int i,j,k,l,ntit,nat,nbnd,nth,nphi,nimphi,ndon,nacc,nnb,ngrp;
  char filename[256],chain_id[5],res_nam[5],res_nr[6],aname[5],atype[5];
  Atom *a,*b,*n1,*n2,*t1,*t2,*t3,*t4;
  Node *n;

  status=0;
  nat=0; nbnd=0; nth=0; nphi=0; nimphi=0; ndon=0; nacc=0; nnb=0;
  ngrp=0; ntit=0;

  /* count atoms, check atomnames, potential names etc... */

  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      nat++;
      a->seqnr=nat;
    }
  }

  /* open output file */

  extfil(filename,file,".psf",YES);
  if ((out=fopen(filename,"w"))==NULL) {
    err_set(1,"write_xplor_psf: could not open %s for write\n",filename);
    return(1);
  }

  /* write header records */

  fprintf(out,"PSF\n");
  ntit=1;
  fprintf(out,"\n%8d !NTITLE title records\n",ntit);
  fprintf(out," REMARKS XPLOR structure file generated by WITNOTP\n");
  if (ntit==0) fprintf(out,"\n");

  /* write atom records */

  fprintf(out,"\n%8d !NATOM atoms\n",nat);

  ForAtoms(a,m) {
    for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
    if (lev>2) {
      Strcpy(chain_id,a->up->up->name);
      n=a->up;
    }else{
      sprintf(chain_id," ");
      n=a->up;
    }
    ForAtoms(b,n) {
      a=b;
      if (a->flg&(1<<1)) {
	make_xpl_resnam(a->up->name,res_nam,res_nr);
	Strcpy(aname,a->name); 
	if (a->flg&(1<<2)) {
	  Strcpy(atype,xplr_hb_type(a->atp->name));
	}else{
	  Strcpy(atype,xplr_ff_type(a->atp->name));
	}
	fprintf(out,"%8d %-4s %-4s %-4s %-4s %-4s %14.6f %9.4f %11d\n",
		a->seqnr,chain_id,res_nr,res_nam,aname,atype,a->q,a->atp->mass,0);
	for (k=0,j=0;j<a->nbnd;j++) {
	  if (a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) nbnd++;
	  if (a->bnd[j]->f->flg&(1<<1)&&a->bnd[j]->t->flg&(1<<1)) k++;
	}
	nth+=(k*(k-1))/2;
      }
    }
  }
  if (nat==0) fprintf(out,"\n");

  /* bonds... */

  fprintf(out,"\n%8d !NBOND bonds\n",nbnd); k=0;
  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (j=0;j<a->nbnd;j++) {
	if (a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) {
	  fprintf(out,"%8d%8d",a->seqnr,a->bnd[j]->t->seqnr);
	  if ((++k)%4==0) fprintf(out,"\n");
	}
      }
    }
  }
  if (k%4!=0) fprintf(out,"\n");
  if (nbnd==0) fprintf(out,"\n");

  /* angles... */

  fprintf(out,"\n%8d !NTHETA angles\n",nth); k=0;
  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (j=1;j<a->nbnd;j++) {
	n1=neighbour(j-1,a);
	if (n1->flg&(1<<1)) {
	  for (l=j;l<a->nbnd;l++) {
	    n2=neighbour(l,a);
	    if (n2->flg&(1<<1)) {
	      fprintf(out,"%8d%8d%8d",n1->seqnr,a->seqnr,n2->seqnr);
	      if ((++k)%3==0) fprintf(out,"\n");
	    }
	  }
	}
      }
    }
  }

  if (k%3!=0) fprintf(out,"\n");
  if (nth==0) fprintf(out,"\n");

  /* torsions, play dirty trick with ftell/fseek ... */

  where=ftell(out);
  fprintf(out,"\n%8d !NPHI torsions\n",nphi);  /* dymmy only */
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1))||a->nbnd<2) continue;
    for (j=0;j<a->nbnd;j++) {
      if (a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) {
	if (a->bnd[j]->t->nbnd<2) continue;
	t2=a; t3=a->bnd[j]->t;
	for (k=0;k<t2->nbnd;k++) {
	  if (t2->bnd[k]==a->bnd[j]) continue;
	  if (linAtom(a,a->bnd[j],a->bnd[k])) continue;    /* 1.12.97: avoid ill-defined dihedrals... */
	  t1=neighbour(k,t2); if (!(t1->flg&(1<<1))) continue;
	  for (l=0;l<t3->nbnd;l++) {
	    if (t3->bnd[l]==a->bnd[j]) continue;
	    if (linAtom(t3,a->bnd[j],t3->bnd[l])) continue;/* 1.12.97: avoid ill-defined dihedrals... */
	    t4=neighbour(l,t3);
	    if (!(t4->flg&(1<<1)) || t1==t4) continue;
	    fprintf(out,"%8d%8d%8d%8d",
		    t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr);
	    if ((++nphi)%2==0) fprintf(out,"\n");
	  }
	}
      }
    }
  }
  if (nphi%2 != 0) fprintf(out,"\n");
  if (nphi==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NPHI torsions\n",nphi);  /* the real number */
  fseek(out,0,2);
	    
  /* improper torsions (oop and chiral) */

  where=ftell(out);
  fprintf(out,"\n%8d !NIMPHI improper torsions\n",nimphi); /* only a dummy... */
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1)) || a->oop==0 || a->nbnd!=3) continue;
    t1=a;
    t2=neighbour(0,a); if (!(t2->flg&(1<<1))) continue;
    t3=neighbour(1,a); if (!(t3->flg&(1<<1))) continue;
    t4=neighbour(2,a); if (!(t4->flg&(1<<1))) continue;
    fprintf(out,"%8d%8d%8d%8d",
	    t2->seqnr,t1->seqnr,t3->seqnr,t4->seqnr);
    if ((++nimphi)%2==0) fprintf(out,"\n");
    fprintf(out,"%8d%8d%8d%8d",
	    t3->seqnr,t1->seqnr,t4->seqnr,t2->seqnr);
    if ((++nimphi)%2==0) fprintf(out,"\n");
    fprintf(out,"%8d%8d%8d%8d",
	    t4->seqnr,t1->seqnr,t2->seqnr,t3->seqnr);
    if ((++nimphi)%2==0) fprintf(out,"\n");
  }   
  if (nimphi%2 !=0 ) fprintf(out,"\n");
  if (nimphi==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NIMPHI improper torsions\n",nimphi);  /* the real number */
  fseek(out,0,2);

  /* hydrogen bond donors */

  where=ftell(out);
  fprintf(out,"\n%8d !NDON hb donors\n",ndon);
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1)) || !a->atp->hbdon) continue;
    k=0;
    for (j=0;j<a->nbnd;j++) {
      t2=neighbour(j,a); if (!(t2->flg&(1<<1))) continue;
      if (t2->atp->z==1) {
	fprintf(out,"%8d%8d",a->seqnr,t2->seqnr);
	if ((++ndon)%4==0) fprintf(out,"\n");
	k++;
      }
    }
    if (k==0) {
      fprintf(out,"%8d%8d",a->seqnr,k);
      if ((++ndon)%4==0) fprintf(out,"\n");
    }
  }
  if (ndon%4 !=0 ) fprintf(out,"\n");
  if (ndon==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NDON hb donors\n",ndon);  /* the real number */
  fseek(out,0,2);

  /* hydrogen bond acceptors */

  where=ftell(out);
  fprintf(out,"\n%8d !NACC hb acceptors\n",nacc);
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1)) || !a->atp->hbacc) continue;
    k=0;
    for (j=0;j<a->nbnd;j++) {
      t2=neighbour(j,a); if (!(t2->flg&(1<<1))) continue;
      if (t2->atp->z==1) continue;
      if (k!=0) break;
      k=t2->seqnr;
    }
    if (j<a->nbnd) k=0;
    fprintf(out,"%8d%8d",a->seqnr,k);
    if ((++nacc)%4==0) fprintf(out,"\n");
  }
  if (nacc%4 !=0 ) fprintf(out,"\n");
  if (nacc==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NACC hb acceptors\n",nacc);  /* the real number */
  fseek(out,0,2);

  /* etc... */

  fprintf(out,"\n%8d !NNB non-bonded exclusions\n",nnb);
  if (nnb==0) fprintf(out,"\n");

  for (i=0;i<nat;i++) {                     /* no idea what this might be */
    fprintf(out,"%8d",0);
    if ((i+1)%8==0) fprintf(out,"\n");
  }
  if (i%8!=0&&i>0) fprintf(out,"\n");

  ngrp=1; 
  b=(Atom *)NULL;
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1))) continue;
    if (mvld(b)) {
      if (a->up!=b->up) ngrp++;   /* groups = residues */
    }
    b=a;
  }
  fprintf(out,"\n%8d%8d !NGRP charge groups\n",ngrp,0); 
  if (ngrp==1) {
    fprintf(out,"\n");
  }else{
    fprintf(out,"%8d%8d%8d",0,0,0);
    ngrp=1;
    b=(Atom *)NULL;
    k=0;
    ForAtoms(a,m) {
      if (!(a->flg&(1<<1))) continue;
      if (mvld(b)) {
	if (a->up!=b->up) {   /* groups = residues */
	  if (ngrp%3 == 0) fprintf(out,"\n");
	  fprintf(out,"%8d%8d%8d",k,0,0);
	  ngrp++;
	}
      }
      b=a;
      k++;
    }
    fprintf(out,"\n");
  }
  if (ferror(out)) {
    err_set(1,"wrxpsf: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrxpsf: error closing file %s",filename);
    status=5;
  }
  return(status);
}
/*-------------------- write_kludge --------------------*/

int write_xplor_kludge(char *file,Node *m)
{
  /* write XPLOR/PSF and PRM file (only "selected" atoms (i.e. flg_1 = 1) )*/

  FILE *out,*outp;
  int status,lev;
  long where;
  int i,j,k,l,ntit,nat,nbnd,nth,nphi,nimphi,ndon,nacc,nnb,ngrp;
  char filename[256],chain_id[5],res_nam[5],res_nr[6],aname[5],atype[5];
  char buf[MAXBUF],*fcon=(char *)NULL;
  Atom *a,*b,*n1,*n2,*t1,*t2=(Atom *)NULL,*t3=(Atom *)NULL,*t4=(Atom *)NULL;
  float d,rv,eps;
  Node *n;

  status=0;
  nat=0; nbnd=0; nth=0; nphi=0; nimphi=0; ndon=0; nacc=0; nnb=0;
  ngrp=0; ntit=0;

  /* count atoms, check atomnames, potential names etc... */

  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      nat++;
      a->seqnr=nat;
    }
  }

  /* open output files */

  extfil(filename,file,".psf",YES);
  if ((out=fopen(filename,"w"))==NULL) {
    err_set(1,"write_xplor_kludge: could not open %s for write\n",filename);
    return(1);
  }

  extfil(filename,file,".prm",YES);
  if ((outp=fopen(filename,"w"))==NULL) {
    err_set(1,"write_xplor_kludge: could not open %s for write\n",filename);
    fclose(out);
    return(1);
  }

  /* write header records */

  fprintf(out,"PSF\n");
  ntit=1;
  fprintf(out,"\n%8d !NTITLE title records\n",ntit);
  fprintf(out," REMARKS XPLOR structure file generated by WITNOTP\n");
  if (ntit==0) fprintf(out,"\n");
  fprintf(outp,"set echo=false end\n\n");
  fprintf(outp,"evaluate ($tff_bon=500)\n");
  fprintf(outp,"evaluate ($tff_ang=100)\n");
  fprintf(outp,"evaluate ($tff_chi=100)\n");
  fprintf(outp,"evaluate ($tff_pla=100)\n");
  fprintf(outp,"evaluate ($tff_tor=100)\n");
  fprintf(outp,"\nparameter\n\n");
  /* write atom records */

  fprintf(out,"\n%8d !NATOM atoms\n",nat);
	  
  ForAtoms(a,m) {
    for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
    if (lev>2) {
      Strcpy(chain_id,a->up->up->name);
      n=a->up;
    }else{
      sprintf(chain_id," ");
      n=a->up;
    }
    ForAtoms(b,n) {
      a=b;
      if (a->flg&(1<<1)) {
	make_xpl_resnam(a->up->name,res_nam,res_nr);
	Strcpy(aname,a->name);
	sprintf(buf,"T%d",a->seqnr); Strcpy(atype,buf);
	fprintf(out,"%8d %-4s %-4s %-4s %-4s %-4s %14.7f %9.4f %11d\n",
		a->seqnr,chain_id,res_nr,res_nam,aname,atype,a->q,a->atp->mass,0);
	rv=2*a->atp->rvdw/1.122462;   /* X-Plor wants sigma, not R */
	eps=0.125;
	fprintf(outp," NONBONDED %-4s %5.2f %5.2f   %5.2f %5.2f\n",
		atype,eps,rv,eps,0.85*rv);
	for (k=0,j=0;j<a->nbnd;j++) {
	  if (a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) nbnd++;
	  if (a->bnd[j]->f->flg&(1<<1)&&a->bnd[j]->t->flg&(1<<1)) k++;
	}
	nth+=(k*(k-1))/2;
      }
    }
  }
  if (nat==0) fprintf(out,"\n");
  fprintf(outp,"\n");

  /* bonds... */

  fprintf(out,"\n%8d !NBOND bonds\n",nbnd); k=0;
  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (j=0;j<a->nbnd;j++) {
	if (a==a->bnd[j]->f && a->bnd[j]->t->flg&(1<<1)) {
	  d=cdist(a->x,a->bnd[j]->t->x);
	  fprintf(outp," BOND T%d\t T%d\t $tff_bon %6.3f\n",
		  a->seqnr,a->bnd[j]->t->seqnr,d);
	  fprintf(out,"%8d%8d",a->seqnr,a->bnd[j]->t->seqnr);
	  if ((++k)%4==0) fprintf(out,"\n");
	}
      }
    }
  }
  if (k%4!=0) fprintf(out,"\n");
  if (nbnd==0) fprintf(out,"\n");
  fprintf(outp,"\n");

  /* angles... */

  fprintf(out,"\n%8d !NTHETA angles\n",nth); k=0;
  ForAtoms(a,m) {
    if (a->flg&(1<<1)) {
      for (j=1;j<a->nbnd;j++) {
	n1=neighbour(j-1,a);
	if (n1->flg&(1<<1)) {
	  for (l=j;l<a->nbnd;l++) {
	    n2=neighbour(l,a);
	    if (n2->flg&(1<<1)) {
	      fprintf(out,"%8d%8d%8d",n1->seqnr,a->seqnr,n2->seqnr);
	      if ((++k)%3==0) fprintf(out,"\n");
	      d=cangled(n1->x,a->x,n2->x);
	      fprintf(outp," ANGLE T%d\t T%d\t T%d\t  $tff_ang %6.2f\n",
		      n1->seqnr,a->seqnr,n2->seqnr,d);
	    }
	  }
	}
      }
    }
  }

  if (k%3!=0) fprintf(out,"\n");
  if (nth==0) fprintf(out,"\n");
  fprintf(outp,"\n");

  /* torsions, play dirty trick with ftell/fseek ... */
  where=ftell(out);
  fprintf(out,"\n%8d !NPHI torsions\n",nphi);  /* dymmy only */
  if (nphi%2 != 0) fprintf(out,"\n");
  if (nphi==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NPHI torsions\n",nphi);  /* the real number */
  fseek(out,0,2);

  /* improper torsions  */

  where=ftell(out);
  fprintf(out,"\n%8d !NIMPHI improper torsions\n",nimphi); /* only a dummy... */
  ForAtoms(a,m) {
    if (!(a->flg&(1<<1))) continue;
    t1=(Atom *)NULL;
    if (a->oop!=0 && a->nbnd==3) {
      t1=a;
      t2=neighbour(0,a); if (!(t2->flg&(1<<1))) continue;
      t3=neighbour(1,a); if (!(t3->flg&(1<<1))) continue;
      t4=neighbour(2,a); if (!(t4->flg&(1<<1))) continue;
      fprintf(out,"%8d%8d%8d%8d",
	      t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr);
      if ((++nimphi)%2==0) fprintf(out,"\n");
      fcon="$tff_pla";
    }else if (a->nbnd==4) {
      t1=neighbour(0,a); if (!(t1->flg&(1<<1))) continue;
      t2=neighbour(1,a); if (!(t2->flg&(1<<1))) continue;
      t3=neighbour(2,a); if (!(t3->flg&(1<<1))) continue;
      t4=neighbour(3,a); if (!(t4->flg&(1<<1))) continue;
      fprintf(out,"%8d%8d%8d%8d",
	      t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr);
      if ((++nimphi)%2==0) fprintf(out,"\n");
      fcon="$tff_chi";
    }
    if (mvld(t1)) {
      d=ctorsd(t1->x,t2->x,t3->x,t4->x);
      if (streq(fcon,"$tff_pla")) {
	if (fabs(d)<25) d=0.0;
	if (fabs(fabs(d)-180)<25) d=180.0;
      }
      fprintf(outp," IMPROPER T%d\t T%d\t T%d\t T%d\t %s  0  %5.0f\n",
	      t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr,fcon,d);
    }
    if (/* Jan 26 2001 a->oop!=0 && */ a->nbnd<=3) {  /* Jan 26 2001: = --> <= */
      for (j=0;j<a->nbnd;j++) {
	if (a==a->bnd[j]->t) continue;
	if (BT_order(a->bnd[j]->bt)==2||BT_arom(a->bnd[j]->bt)
	    ||BT_number("am")==a->bnd[j]->bt) {
	  t2=a; t3=a->bnd[j]->t;
	  if (t2->nbnd<=3) {        /* Jan 26 2001: = --> <= */
	    for (k=0;k<t2->nbnd;k++) {
	      t1=neighbour(k,t2); if (t1==t3) continue;
	      for (l=0;l<t3->nbnd;l++) {
		t4=neighbour(l,t3); if (t4==t2||t4==t1) continue;
		if (!(t1->flg&(1<<1))) continue;
		if (!(t2->flg&(1<<1))) continue;
		if (!(t3->flg&(1<<1))) continue;
		if (!(t4->flg&(1<<1))) continue;
		fprintf(out,"%8d%8d%8d%8d",
			t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr);
		if ((++nimphi)%2==0) fprintf(out,"\n");
		d=ctorsd(t1->x,t2->x,t3->x,t4->x); fcon="$tff_tor";
		if (fabs(d)<25) d=0.0;
		if (fabs(fabs(d)-180)<25) d=180.0;
		fprintf(outp," IMPROPER T%d\t T%d\t T%d\t T%d\t %s  0  %5.0f\n",
			t1->seqnr,t2->seqnr,t3->seqnr,t4->seqnr,fcon,d);

	      }
	    }
	  }
	}
      }
    }
  }
  if (nimphi%2 !=0 ) fprintf(out,"\n");
  if (nimphi==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NIMPHI improper torsions\n",nimphi);  /* the real number */
  fseek(out,0,2);

  /* hydrogen bond donors */

  where=ftell(out);
  fprintf(out,"\n%8d !NDON hb donors\n",ndon);
  if (ndon%4 !=0 ) fprintf(out,"\n");
  if (ndon==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NDON hb donors\n",ndon);  /* the real number */
  fseek(out,0,2);

  /* hydrogen bond acceptors */

  where=ftell(out);
  fprintf(out,"\n%8d !NACC hb acceptors\n",nacc);
  if (nacc%4 !=0 ) fprintf(out,"\n");
  if (nacc==0) fprintf(out,"\n");
  fseek(out,where,0);
  fprintf(out,"\n%8d !NACC hb acceptors\n",nacc);  /* the real number */
  fseek(out,0,2);

  /* etc... */

  fprintf(out,"\n%8d !NNB non-bonded exclusions\n",nnb);
  if (nnb==0) fprintf(out,"\n");

  for (i=0;i<nat;i++) {                     /* no idea what this might be */
    fprintf(out,"%8d",0);
    if ((i+1)%8==0) fprintf(out,"\n");
  }
  if (i%8!=0&&i>0) fprintf(out,"\n");

  ngrp=0;
  fprintf(out,"\n%8d%8d !NGRP charge groups\n",1,0);
  if (ngrp==0) fprintf(out,"\n");
  fprintf(outp,"\nend\n");

  if (ferror(out)) {
    err_set(1,"wrxpsf: error writing to %s",filename);
    status=4;
  }
  if (fclose(out)) {
    err_set(1,"wrxpsf: error closing file %s",filename);
    status=5;
  }
  if (ferror(outp)) {
    err_set(1,"wrxprm: error writing to %s",filename);
    status=4;
  }
  if (fclose(outp)) {
    err_set(1,"wrxprm: error closing file %s",filename);
    status=5;
  }
  return(status);
}
/*-------------------- make_xpl_resnam --------------------*/

void make_xpl_resnam(char *name,char *xplnm,char *xplnu)
{
  /* generate XPLOLR residue name and number from internal residue name */

  char tmp[20],*p;
  int i;

  Strcpy(tmp,name);
  for (i=0,p=tmp;i<4&&*p;) {
    if ( *p >= 'a' && *p <= 'z' ) *p+='A'-'a';
    if (*p=='_'||*p==' ') break;
    xplnm[i++]=*p++;
  }
  if (*p=='_'||*p==' ') p++;
  xplnm[i]='\0';

  for (i=0;i<5&&*p;) {
    if ( *p >= 'a' && *p <= 'z' ) *p+='A'-'a';
    if (*p=='_'||*p==' ') break;
    xplnu[i++]=*p++;
  }
  xplnu[i]='\0';

  if (strlen(xplnm)==0) sprintf(xplnm,"UNK");
  if (strlen(xplnu)==0) sprintf(xplnu,"0");
}  
/*-------------------- xplr_ff_type --------------------*/

static char xplr_ff_type_buff[5];

char *xplr_ff_type(char *mnem)
{
  int n;
  char *i,*o=(char *)NULL;

  xplr_ff_type_buff[0]='\0';
  if (mnem) {
    for (n=0,i=mnem,o=xplr_ff_type_buff;*i&&(n<4);) {
      *o = *i++;
      if ( *o != '.' ) {
	if ( *o >= 'a' && *o <= 'z' ) *o+='A'-'a';
	o++;
	n++;
      }
    }
  }
  *o='\0';
  return(xplr_ff_type_buff);
}
/*-------------------- xplr_hb_type --------------------*/

static char xplr_hb_type_buff[5];

char *xplr_hb_type(char *mnem)
{
  int n;
  char *i,*o=(char *)NULL;

  xplr_hb_type_buff[0]='\0';
  if (mnem) {
    for (n=0,i=mnem,o=xplr_hb_type_buff;*i&&(n<4);) {
      *o = *i++;
      if ( *o != '.' ) {
	if ( *o >= 'a' && *o <= 'z' ) *o+='A'-'a';
	o++;
	n++;
	if (n==1) {     /* mark as hb-hyd */
	  *o++='*';
	  n++;
	}
      }
    }
  }
  *o='\0';
  return(xplr_hb_type_buff);
}
/*-------------------- xpl_full_name --------------------*/

char *xpl_sel(char *buf,Atom *a,Node *m)
{
  int lev;
  Node *n;
  char chain_id[5],res_id[6],res_nam[5],aname[5];

  make_xpl_resnam(a->up->name,res_nam,res_id);
  sprintf(aname,"%-4s",a->name);
  for (lev=0,n=(Node *)a;n!=m;n=n->up) lev++;
  if (lev>2) {
    Strcpy(chain_id,a->up->up->name);
    sprintf(buf,"(segid %s and resid %s and name %s)",
	    chain_id,res_id,aname);
  }else{
    sprintf(buf,"(resid %s and name %s)",
	    res_id,aname);
  }
  return(buf);
}
/*-------------------- linAtom -------------------- */

static int linAtom(Atom *a,Bond *b,Bond *c)
{
  return(a->nbnd==2&&
	 (b->bt==BT_TRIPLE||c->bt==BT_TRIPLE
	  ||(b->bt==BT_DOUBLE&&c->bt==BT_DOUBLE)
	  )
	 );
}
