#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#if defined(linux) || defined(darwin)
#include <unistd.h>
#include <sys/types.h>
#include <sys/times.h>
#include <limits.h>
#include <time.h>
#endif

#define MAXUNIT 100

static FILE *file[MAXUNIT];

void cioinit_(void)
{
  int i;
  for (i=0;i<100;i++) {
    file[i]=(FILE *)NULL;
  }
}

char *ztrimm_(char *string,int *l)
{
  char *p;
  for (p=string + *l - 1;p>=string && *p==' ';p--) *p='\0';
  return(string);
}

int cioftell_(int *unit)
{
  int pos;

  pos=0;
  if (*unit>=0 && *unit<MAXUNIT) {
    if (file[*unit]) pos=ftell(file[*unit]);
  }
  return(pos);
}

int ciofseek_(int *unit,int *pos)
{
  int r;

  r=-1;
  if (*unit>=0 && *unit<MAXUNIT) {
    if (file[*unit]) r=fseek(file[*unit],*pos,SEEK_SET);
  }
  return(r);
}

int ciofopen_(char *name,char *mode)
{
  int i;

  for (i=0;i<MAXUNIT;i++) if (!file[i]) break;
  if (i<MAXUNIT) {
    if (*mode=='w'||*mode=='W') file[i]=fopen(name,"w");
    else if (*mode=='r'||*mode=='R') file[i]=fopen(name,"r");
    else if (*mode=='u'||*mode=='U') file[i]=fopen(name,"r+");
    else if (*mode=='a'||*mode=='A') file[i]=fopen(name,"a");
    else i=0;
    if (!file[i]) i=-1;
  }else{
    i=-1;
  }
  return(i);
}

void ciofclose_(int *unit)
{
  if (*unit<0||*unit>=MAXUNIT) return;
  if (file[*unit]) fclose(file[*unit]);
  file[*unit]=(FILE *)NULL;
}

int ciogetrec_(int *unit,char *record,int *l)
{
  int r;
  char *c;

  if (*unit<0 || *unit>=MAXUNIT) return(0);
  if (!file[*unit]) return(0);
  if (feof(file[*unit])) return(0);

  c=fgets(record,*l-1,file[*unit]);
  if (!c) {
    r=0;
  }else{
    r=strlen(record);
  }
  return(r);
}

void cioputrec_(int *unit,char *record)
{
  if (*unit<0 || *unit>=MAXUNIT) return;
  if (!file[*unit]) return;

  fputs(record,file[*unit]);
}
  
int ciofgets_(int *unit,char *record,int *l)
{
  int n;
  int c;

  if (*unit<0 || *unit>=MAXUNIT) return(0);
  if (!file[*unit]) return(0);
  if (feof(file[*unit])) return(0);

  for (n=0;n<*l;n++) {
    c=fgetc(file[*unit]);
    if (c==EOF||c=='\n') break;
    record[n]=c;
  }
  if (c==EOF&&n==0) return(0);
  for (;n<*l;n++) record[n]=' ';
  while(c!=EOF&&c!='\n') c=fgetc(file[*unit]);
  return(1);
}

int ciofskip_(int *unit,int *n)
{
  int s;
  int c;

  s=0;

  if (*unit<0 || *unit>=MAXUNIT) return(0);
  if (!file[*unit]) return(0);
  if (feof(file[*unit])) return(0);

  while(s<*n) {
    c=fgetc(file[*unit]);
    if (c==EOF) break;
    if (c=='\n') s++;
    if (c=='\\') {
      c=fgetc(file[*unit]);
      if (c==EOF) break;
    }
  }
  return(s);
}

int ciogetfld_(int *unit,char *fld,int *l)
{
  int r,i;
  int c;

  r=-1;

  if (*unit<0 || *unit>=MAXUNIT) return(r);
  if (!file[*unit]) return(r);
  if (feof(file[*unit])) return(r);

  /* skip whitespace */

  for (;;) {
    c=fgetc(file[*unit]);
    if (c==EOF) break;
    if (c=='\n') break;
    if (c=='\\') {
      c=fgetc(file[*unit]);
      if (c==EOF) break;
      if (c=='\n') continue;
    }
    if (!isspace(c)) break;
  } 

  if (c==EOF) return(r);
  ungetc(c,file[*unit]);

  /* extract space separated field */

  for (r=0;r<*l;r++) {
    c=fgetc(file[*unit]);
    if (isspace(c) || c=='\\' || c==EOF) break;
    fld[r]=c;
  }

  /* pad return-string with spaces */

  for (i=r;i<*l;i++) fld[i]=' ';

  /* ignore rest of field */

  while(!isspace(c)&&c!='\\'&&c!=EOF) c=fgetc(file[*unit]);
  if (c=='\n'||c=='\\') ungetc(c,file[*unit]);

  return(r);
}

void ciofputs_(int *unit,char *record,int *l)
{
  int i;

  if (*unit<0 || *unit>=MAXUNIT) return;
  if (!file[*unit]) return;
  
  for (i=0;i<*l;i++) fputc(record[i],file[*unit]);
}

#ifdef linux

void rename_(char *old,char *new)
{
  char bo[1024],bn[1024],*p;
  int i;

  for (i=0,p=old;*p!=' '&&*p!='\0';i++) bo[i]=*p++; bo[i]='\0';
  for (i=0,p=new;*p!=' '&&*p!='\0';i++) bn[i]=*p++; bn[i]='\0';
  rename(bo,bn);
}

void unlink_(char *name)
{
  char bn[1024],*p;
  int i;

  for (i=0,p=name;*p!=' '&&*p!='\0';i++) bn[i]=*p++; bn[i]='\0';
  unlink_(bn);
}

void exit_(int *s)
{
  exit(*s);
}

float secnds_(float *t)
{
  return(((double)clock())/CLOCKS_PER_SEC-*t);
}

float etime_(float t[2])
{
  struct tms T;

  times(&T);
  t[0]=((float)T.tms_utime)/CLOCKS_PER_SEC;
  t[1]=((float)T.tms_stime)/CLOCKS_PER_SEC;
  return(t[0]+t[1]);
}

void qsort_(void *R,int *n,int *s,int (*comp)(const void *a,const void *b))
{
  qsort(R,*n,*s,comp);
}

#endif
