      integer function dlbfgs(func,n,m,x,f,g,maxrmsg,maxfun,W)
c
c  Driver for lbfgs minimizer.
c  --------------------------
c
c     double precision func(n,x,g) : evaluates objective and gradient at x (dimension n)
c     n: number of parameters
c     m: number of corfrections in BFGS step (>0)
c     x: parameter values (in/out)
c     f: value of objective at x
c     g: gradient at x
c     maxrmsg: convergence criterion (rms gradient)
c     maxfun: maximum number of function calls allowed before forced termination
c     W: workspace (double, dimension at least xxx)
c
c     Return: <0 error return from lbfgs,
c              0 converged,
c              1 maxfun exceeded.
c
      integer MP,LP
      double precision GTOL,STPMIN,STPMAX
      COMMON /LB3/MP,LP,GTOL,STPMIN,STPMAX
c
      integer n,m,maxfun,iprint(2),ncalls,i,op,conjgradtest
      double precision func,x(n),f,t,g(n),maxrmsg,W(*),ssq,ssqb
      external func
c
      ncalls=0
      iprint(1)=-1   ! no output by lbfgs
      LP=-1          ! no error messages by lbfgs
      GTOL=0.9       ! 
c
      f=func(n,x,g)
      ncalls=ncalls+1
      ssqb=0.0
      do i=1,n
         ssq=ssq+g(i)*g(i)
         W(i)=x(i)
         W(i+n)=g(i)
      end do
      ssqb=ssq
      t=f
      op=0
c
      do while ( 
     .     op.ge.0 .and. op.lt.2 .and.
     .     ncalls.lt.maxfun .and.
     .     ssq.gt.n*maxrmsg*maxrmsg .and.
     .     .not. conjgradtest().gt.0
     .     ) 
         call lbfgs(n,m,W,t,W(1+n),
     .        .FALSE.,W(1+2*n),iprint,0.0D1,1.0D-16,W(1+3*n),op)
         if (op.eq.1) then
            t=func(n,W,W(1+n))
            ncalls=ncalls+1
            ssq=0.0
            do i=1,n
               ssq=ssq+W(i+n)*W(i+n)
            end do
            if (t.lt.f .or. (t.eq.f .and. ssq.lt.ssqb)) then
               do i=1,n
                  x(i)=W(i)
                  g(i)=W(i+n)
               end do
               ssqb=ssq
               f=t
            endif
         end if
      end do
c
      if (ssq.le.n*maxrmsg*maxrmsg) then
         dlbfgs=0
      else if (ncalls.ge.maxfun) then
         dlbfgs=1
      else if (op.ge.0) then
         dlbfgs=2
      else
         dlbfgs=op
      endif
c
      return
      end

c===============================================================================
      integer function dcgfam(func,n,m,x,f,g,maxrmsg,maxfun,W)
c
c  Driver for cgfam minimizer.
c  --------------------------
c
c     double precision func(n,x,g) : evaluates objective and gradient at x (dimension n)
c     n: number of parameters
c     m: CG method (1: Fletcher-Reeves, 2: Polak-Ribiere, 3: positive Polak-Ribiere)
c        m<0: same as -m, with restart every n steps
c     x: parameter values (in/out)
c     f: value of objective at x
c     g: gradient at x
c     maxrmsg: convergence criterion (rms gradient)
c     maxfun: maximum number of function calls allowed before forced termination
c     W: workspace (double, dimension at least xxx)
c
c     Return: <0 error return from cgfam,
c              0 converged,
c              1 maxfun exceeded.
c
c
      integer n,m,maxfun,iprint(2),ncalls,i,op,method,conjgradtest
      logical ir
      double precision func,x(n),f,t,g(n),maxrmsg,W(*),ssq,ssqb
      external func
c
      integer MP,LP
      COMMON /CGDD/CGMP,CGLP
c
      ncalls=0
      iprint(1)=-1   ! no output by lbfgs
      LP=-1          ! no error messages, please.
c
      f=func(n,x,g)
      ncalls=ncalls+1
      ssqb=0.0
      do i=1,n
         ssq=ssq+g(i)*g(i)
         W(i)=x(i)
         W(i+n)=g(i)
      end do
      ssqb=ssq
      t=f
      op=0
      if (m.lt.0) then
         method=-m
         ir=.true.
      else
         method=m
         ir=.false.
      end if
c
      do while ( 
     .     op.ge.0 .and. op.lt.3 .and.
     .     ncalls.lt.maxfun .and.
     .     ssq.gt.n*maxrmsg*maxrmsg .and.
     .     .not. conjgradtest().gt.0
     .     ) 
         call cgfam(n,W,t,W(1+n),W(1+2*n),W(1+3*n),
     .        iprint,0.0D1,W(1+4*n),op,ir,method,.false.)
         if (op.eq.1) then
            t=func(n,W,W(1+n))
            ncalls=ncalls+1
            ssq=0.0
            do i=1,n
               ssq=ssq+W(i+n)*W(i+n)
            end do
            if (t.lt.f .or. (t.eq.f .and. ssq.lt.ssqb)) then
               do i=1,n
                  x(i)=W(i)
                  g(i)=W(i+n)
               end do
               ssqb=ssq
               f=t
            endif
         end if
      end do
c
      if (ssq.le.n*maxrmsg*maxrmsg) then
         dcgfam=0
      else if (ncalls.ge.maxfun) then
         dcgfam=1
      else if (op.ge.0) then
         dcgfam=2
      else
         dcgfam=op
      endif
c
      return
      end

