      recursive subroutine egnbnd_1p(p,w,n,x,evg,ecg,gr)
c
c
c   Compute van der Waals energy 'e' for the molecule whose coordinates
c   are stored in x (n atoms). At the same time the gradient resulting from
c   the van der Waals potentials will be added to 'gr'.
c
c   multiplicative switching function (used for VSHIFT and ESHIFT):
c   ..............................................................
c
c               2    2  2     2    2         2    2
c            ( R  - x  )  * (R  - x  - 3 * (r  - x  ) )
c    sw(x) =  -----------------------------------------
c                               2    2  3
c                            ( R  - r  )
c
c    applied between r and R (cut-on, cut-off)
c
c
c    additive shifting function (VSHIFT):
c    ...................................
c               6
c    sh(x) = C*R  +  D
c
c    applied below R (cut-off). C and D are type-dependent, chosen such
c    that the modified Energy and Force vanish at x=R.
c
c
c    multiplicative shifting function (ESHIFT):
c    .........................................
c                    2
c                   x    2
c    sh(x) = ( 1 - ---  )
c                    2
c                   R
c
c    applied below R (cut-off)
c
      implicit none
#include "sizes.h"
#include "energy.h"
#include "atom.h"
#include "apara.h"
#include "option.h"

      real*8  evg,ecg
      real*8  dec,hbs6,hbs12,ex,rc,eel,e12,e6
      real*8  f,r6,r12,dx,dy,dz,rs,A,B,C,D,s,se
      real*8  x1,x2,x3,gr1,gr2,gr3,dis,cutoff2,cuton,cuton2
      real*8  de,sw,fsw,rswd,rswd12,ru,rl,csw1,csw2,deci
      integer i,j,w,n,ni,nj,in,p,case
      integer nblist,nsat,sat(matot)
      logical coul

c     local copies of x,gr improve cache usage on FX-2800

      real*8 gr(3,n),x(3,n),ev,ec

c     
c     Initialize energy
c     
      ev=0.0
      ec=0.0
      hbs6=hbscale**6
      hbs12=hbs6*hbs6
      cutoff2=cutoff*cutoff
      cuton=cutoff-cutswr
      cuton2=cuton*cuton
      if (cutoff2-cuton2.gt.1.e-6) then
         rswd=1/(cutoff2-cuton2)**3
      else
         rswd=0.0
      end if
      rswd12=12*rswd
      csw1=-1.0/cutoff2
      csw2=4.0*csw1
      coul=(chrgok.and.diel.ge.1.e-6)
      if (coul) then
         dec=FPEPS0/diel
         ex=1
         if (.not.diel_const) ex=2
      end if

      case=0
      if (coul) case=2
      if (cutoff.lt.1.e30) case=case+1

      goto (100,200,300,400) case+1

c     
c     Compute van der Waals energy for each pair of non-bonded atoms.
c     

c     no Cutoff, no Coulomb...

 100  continue
      do i=p+1,nat,w
         ni=LJ(i)
         x1=x(1,i)
         x2=x(2,i)
         x3=x(3,i)
         gr1=gr(1,i)
         gr2=gr(2,i)
         gr3=gr(3,i)
         nsat=nblist(i,sat)
         do in=1,nsat
            j=sat(in)
            if (j.lt.0) then
               j=-j
               nj=LJ(j)
               if (fftype.eq.'TAFF') then
                  A=vaij(ni,nj)*hbs12
                  B=vbij(ni,nj)*hbs6
               else if (fftype.eq.'CHARMM') then
                  A=vaij14(ni,nj)
                  B=vbij14(ni,nj)
               end if
            else
               nj=LJ(j)
               A=vaij(ni,nj)
               B=vbij(ni,nj)
            end if
            dx=x1-x(1,j)
            dy=x2-x(2,j)
            dz=x3-x(3,j)
            dis=dx*dx+dy*dy+dz*dz
            if (dis.lt.0.01) dis=0.01
            rs=1/(dis)
            r6=rs*rs*rs
            r12=r6*r6
            e12=A*r12
            e6=B*r6
            de=e12+e6
            f=-6*rs*(de+e12)
            ev=ev+de
            dx=f*dx
            dy=f*dy
            dz=f*dz
            gr1=gr1+dx
            gr2=gr2+dy
            gr3=gr3+dz
            gr(1,j)=gr(1,j)-dx
            gr(2,j)=gr(2,j)-dy
            gr(3,j)=gr(3,j)-dz
         end do
         gr(1,i)=gr1
         gr(2,i)=gr2
         gr(3,i)=gr3
      end do
      go to 1000

c     Cutoff, no Coulomb

 200  continue
      do i=p+1,nat,w
         ni=LJ(i)
         x1=x(1,i)
         x2=x(2,i)
         x3=x(3,i)
         gr1=gr(1,i)
         gr2=gr(2,i)
         gr3=gr(3,i)
         nsat=nblist(i,sat)
         do in=1,nsat
            j=sat(in)
            if (j.lt.0) then
               j=-j
               nj=LJ(j)
               if (fftype.eq.'TAFF') then
                  A=vaij(ni,nj)*hbs12
                  B=vbij(ni,nj)*hbs6
               else if (fftype.eq.'CHARMM') then
                  A=vaij14(ni,nj)
                  B=vbij14(ni,nj)
                  C=vcij14(ni,nj)
                  D=vdij14(ni,nj)
               end if
            else
               nj=LJ(j)
               A=vaij(ni,nj)
               B=vbij(ni,nj)
               C=vcij(ni,nj)
               D=vdij(ni,nj)
            end if
            dx=x1-x(1,j)
            dy=x2-x(2,j)
            dz=x3-x(3,j)
            dis=dx*dx+dy*dy+dz*dz
            if (dis.lt.0.01) dis=0.01
            if (dis.lt.cutoff2) then
               rs=1/(dis)
               r6=rs*rs*rs
               r12=r6*r6
               e12=A*r12
               e6=B*r6
               de=e12+e6
               f=-6*rs*(de+e12)
               if (vswitch) then
                  if (dis.gt.cuton2) then
                     ru=cutoff2-dis   ! switching for vdW...
                     rl=cuton2-dis
                     sw=rswd*ru*ru*(ru-3*rl)
                     fsw=rswd12*rl*ru
                     f=f*sw+de*fsw
                     de=de*sw
                  end if
               else if (vshift) then  ! vdW shifting...
                  s=dis*dis*dis*C
                  de=de+s+D
                  f=f+6*rs*s
               end if
               ev=ev+de
               dx=f*dx
               dy=f*dy
               dz=f*dz
               gr1=gr1+dx
               gr2=gr2+dy
               gr3=gr3+dz
               gr(1,j)=gr(1,j)-dx
               gr(2,j)=gr(2,j)-dy
               gr(3,j)=gr(3,j)-dz
            end if
         end do
         gr(1,i)=gr1
         gr(2,i)=gr2
         gr(3,i)=gr3
      end do
      go to 1000

c     no Cutoff, Coulomb

 300  continue
      do i=p+1,nat,w
         ni=LJ(i)
         x1=x(1,i)
         x2=x(2,i)
         x3=x(3,i)
         gr1=gr(1,i)
         gr2=gr(2,i)
         gr3=gr(3,i)
         deci=dec*chrg(i)
         nsat=nblist(i,sat)
         do in=1,nsat
            j=sat(in)
            if (j.lt.0) then
               j=-j
               nj=LJ(j)
               if (fftype.eq.'TAFF') then
                  A=vaij(ni,nj)*hbs12
                  B=vbij(ni,nj)*hbs6
                  eel=deci*chrg(j)
               else if (fftype.eq.'CHARMM') then
                  A=vaij14(ni,nj)
                  B=vbij14(ni,nj)
                  eel=e14fac*deci*chrg(j)
               end if
            else
               nj=LJ(j)
               A=vaij(ni,nj)
               B=vbij(ni,nj)
               eel=deci*chrg(j)
            end if
            dx=x1-x(1,j)
            dy=x2-x(2,j)
            dz=x3-x(3,j)
            dis=dx*dx+dy*dy+dz*dz
            if (dis.lt.0.01) dis=0.01
            rs=1/(dis)
            r6=rs*rs*rs
            r12=r6*r6
            e12=A*r12
            e6=B*r6
            de=e12+e6
            f=-6*rs*(de+e12)
            rc=rs
            if (diel_const) rc=sqrt(rs)
            eel=eel*rc
            f=f-ex*eel*rs
            ev=ev+de
            ec=ec+eel
            dx=f*dx
            dy=f*dy
            dz=f*dz
            gr1=gr1+dx
            gr2=gr2+dy
            gr3=gr3+dz
            gr(1,j)=gr(1,j)-dx
            gr(2,j)=gr(2,j)-dy
            gr(3,j)=gr(3,j)-dz
         end do
         gr(1,i)=gr1
         gr(2,i)=gr2
         gr(3,i)=gr3
      end do
      go to 1000

c     Cutoff, Coulomb...

 400  continue
      do i=p+1,nat,w
         ni=LJ(i)
         x1=x(1,i)
         x2=x(2,i)
         x3=x(3,i)
         gr1=gr(1,i)
         gr2=gr(2,i)
         gr3=gr(3,i)
         deci=dec*chrg(i)
         nsat=nblist(i,sat)
         do in=1,nsat
            j=sat(in)
            if (j.lt.0) then
               j=-j
               nj=LJ(j)
               if (fftype.eq.'TAFF') then
                  A=vaij(ni,nj)*hbs12
                  B=vbij(ni,nj)*hbs6
                  eel=deci*chrg(j)
               else if (fftype.eq.'CHARMM') then
                  A=vaij14(ni,nj)
                  B=vbij14(ni,nj)
                  C=vcij14(ni,nj)
                  D=vdij14(ni,nj)
                  eel=e14fac*deci*chrg(j)
               end if
            else
               nj=LJ(j)
               A=vaij(ni,nj)
               B=vbij(ni,nj)
               C=vcij(ni,nj)
               D=vdij(ni,nj)
               eel=deci*chrg(j)
            end if
            dx=x1-x(1,j)
            dy=x2-x(2,j)
            dz=x3-x(3,j)
            dis=dx*dx+dy*dy+dz*dz
            if (dis.lt.0.01) dis=0.01
            if (dis.lt.cutoff2) then
               rs=1/(dis)
               rc=rs
               if (diel_const) rc=sqrt(rc)
               r6=rs*rs*rs
               r12=r6*r6
               e12=A*r12
               e6=B*r6
               de=e12+e6
               eel=eel*rc
               if (vswitch.or.eswitch) then  ! comput switching fuc. + deriv.
                  ru=cutoff2-dis 
                  rl=cuton2-dis
                  sw=rswd*ru*ru*(ru-3*rl)
                  fsw=rswd12*rl*ru
               end if
               f=-6*rs*(de+e12)
               if (vswitch) then
                  if (dis.gt.cuton2) then
                     f=f*sw+de*fsw
                     de=de*sw
                  end if
               else if (vshift) then
                  s=dis*dis*dis*C
                  de=de+s+D
                  f=f+6*rs*s
               end if
               if (eswitch) then
                  if (dis.gt.cuton2) then
                     f=f-ex*eel*rs*sw+eel*fsw
                     eel=eel*sw
                  else
                     f=f-ex*eel*rs
                  end if
               else if (eshift) then
                  se=1.0+csw1*dis
                  eel=eel*se
                  f=f+csw2*eel
                  eel=eel*se
                  f=f-ex*eel*rs
               end if
               ev=ev+de
               ec=ec+eel
               dx=f*dx
               dy=f*dy
               dz=f*dz
               gr1=gr1+dx
               gr2=gr2+dy
               gr3=gr3+dz
               gr(1,j)=gr(1,j)-dx
               gr(2,j)=gr(2,j)-dy
               gr(3,j)=gr(3,j)-dz
            end if
         end do
         gr(1,i)=gr1
         gr(2,i)=gr2
         gr(3,i)=gr3
      end do
      go to 1000

c
c     for AMBER: have to do 1-4 non-bonded
c
 1000 continue
      if (fftype.ne.'AMBER') goto 2000

      goto (1100,1100,1200,1200) case+1

c     1-4, no Coulomb

 1100 continue
      do i=p+1,nat,w
         ni=LJ(i)
         x1=x(1,i)
         x2=x(2,i)
         x3=x(3,i)
         gr1=gr(1,i)
         gr2=gr(2,i)
         gr3=gr(3,i)
         do in=hbpt(i-1)+1,hbpt(i)
            j=hb14(in)
            nj=LJ(j)
            A=scnb*vaij(ni,nj)
            B=scnb*vbij(ni,nj)
            dx=x1-x(1,j)
            dy=x2-x(2,j)
            dz=x3-x(3,j)
            dis=dx*dx+dy*dy+dz*dz
            rs=1/(dis)
            r6=rs*rs*rs
            r12=r6*r6
            e12=A*r12
            e6=B*r6
            de=e12+e6
            f=-6*rs*(de+e12)
            ev=ev+de
            dx=f*dx
            dy=f*dy
            dz=f*dz
            gr1=gr1+dx
            gr2=gr2+dy
            gr3=gr3+dz
            gr(1,j)=gr(1,j)-dx
            gr(2,j)=gr(2,j)-dy
            gr(3,j)=gr(3,j)-dz
         end do
         gr(1,i)=gr1
         gr(2,i)=gr2
         gr(3,i)=gr3
      end do
      go to 2000

c     1-4, Coulomb

 1200 continue
      do i=p+1,nat,w
         ni=LJ(i)
         x1=x(1,i)
         x2=x(2,i)
         x3=x(3,i)
         gr1=gr(1,i)
         gr2=gr(2,i)
         gr3=gr(3,i)
         deci=scee*dec*chrg(i)
         do in=hbpt(i-1)+1,hbpt(i)
            j=hb14(in)
            nj=LJ(j)
            A=scnb*vaij(ni,nj)
            B=scnb*vbij(ni,nj)
            dx=x1-x(1,j)
            dy=x2-x(2,j)
            dz=x3-x(3,j)
            dis=dx*dx+dy*dy+dz*dz
            rs=1/(dis)
            r6=rs*rs*rs
            r12=r6*r6
            e12=A*r12
            e6=B*r6
            de=e12+e6
            f=-6*rs*(de+e12)
            rc=rs
            if (diel_const) rc=sqrt(rs)
            eel=deci*chrg(j)*rc
            f=f-ex*eel*rs
            ev=ev+de
            ec=ec+eel
            dx=f*dx
            dy=f*dy
            dz=f*dz
            gr1=gr1+dx
            gr2=gr2+dy
            gr3=gr3+dz
            gr(1,j)=gr(1,j)-dx
            gr(2,j)=gr(2,j)-dy
            gr(3,j)=gr(3,j)-dz
         end do
         gr(1,i)=gr1
         gr(2,i)=gr2
         gr(3,i)=gr3
      end do
      go to 2000
c     
c     done...
c     

 2000 continue

      evg=ev
      ecg=ec

      return
      end
