      logical function gnblst(nx,x)
c     
c     Generate non bonded interaction list
c     
      implicit none
c     
c     
      integer nx
      real*8 x(3,*)
c     
#include "sizes.h"
#include "energy.h"
#include "option.h"
#include "atom.h"
#include "apara.h"
c     
      integer i,j,k,n,l,s,bs,p,nt,omp_get_num_threads,omp_get_thread_num
      integer nblistinit
      logical nblv(mxproc),nb_list_valid,gnblst_1p
      real etime,tnow
c     
      if (nx.ne.nat) then
         write(6,'(a)') 
     .        ' *** gnblst: internal error (nx.ne.nat)'
         write(6,'(a)') 
     .        ' *** gnblst: internal error (nx.ne.nat)'
         gnblst=.FALSE.
         return
      end if

      tnow=etime(tarray)
      if (nbl_valid) then
         if (cutoff.lt.1.e30) then
c     
#ifdef USEOMP
C$OMP PARALLEL PRIVATE(p), SHARED(nx,nt)
            nt=omp_get_num_threads()
            p=omp_get_thread_num()+1    
#else 
            nt=1
            p=1
#endif
            nblv(p)=nb_list_valid(p,nt,nx,x)
#ifdef USEOMP
C$OMP END PARALLEL
#endif
            do p=1,nt
               nbl_valid=nbl_valid.and.nblv(p)
            end do
         end if
      end if
c     
      if (.not.nbl_valid) then
         do i=1,nx
            xnbl(1,i)=x(1,i)
            xnbl(2,i)=x(2,i)
            xnbl(3,i)=x(3,i)
         end do
         if (nat.ne.nblistinit(nat)) then
            write(6,'(a)') 
     .           ' *** Failed to generate nonbonded list.'
            write(21,'(a)') 
     .           ' *** Failed to generate nonbonded list.'
            gnblst=.FALSE.
         else
c     
#ifdef USEOMP
C$OMP PARALLEL PRIVATE(p), SHARED(nt)
            nt=omp_get_num_threads()
            p=omp_get_thread_num()+1       
#else 
            nt=1
            p=1
#endif
            nblv(p)=gnblst_1p(p,nt)
#ifdef USEOMP
C$OMP END PARALLEL
#endif
            gnblst=.TRUE.
            do p=1,nt
               gnblst=gnblst.and.nblv(p)
            end do
c     
            if (.not.gnblst) then
               write(6,'(a/a)') 
     .              ' *** Too many nonbonded interactions.',
     .              ' *** Use a shorter cutoff.'
               write(21,'(a/a)') 
     .              ' *** Too many nonbonded interactions.',
     .              ' *** Use a shorter cutoff.'
            end if
         end if
      else
         gnblst=.TRUE.
      end if
      nbl_valid=gnblst
      tgnblst=tgnblst+etime(tarray)-tnow
c     
      return
      end
c------------------------------------------------------------------------

      logical function nb_list_valid(p,w,nx,x)

c     
c     conservative check of validity of current non-bonded list
c     
      implicit none
c     
c     
      integer nx
      real*8 x(3,*)
c     
#include "sizes.h"
#include "energy.h"
#include "option.h"
#include "atom.h"
#include "apara.h"
c     
      integer i,p,i1,i2,s,w
      real*8 tol,d
c     
      tol=0.25*cutmrg*cutmrg
      s=(nx+w-1)/w
      i1=(p-1)*s+1
      i2=min(i1+s-1,nx)
      nb_list_valid=.true.
      do i=i1,i2
         nb_list_valid=nb_list_valid.and.tol.ge.
     .        (xnbl(1,i)-x(1,i))*(xnbl(1,i)-x(1,i))
     .        +(xnbl(2,i)-x(2,i))*(xnbl(2,i)-x(2,i))
     .        +(xnbl(3,i)-x(3,i))*(xnbl(3,i)-x(3,i))
      end do
c     
      return
      end
