      logical function input()
c     
c     Read molecule for MINIMAX                  
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
#include "option.h"
#include "energy.h"
#if defined(sgi)
#define READONLY ,readonly
#else 
#define READONLY
#endif
c     
      integer*4    status,k,iargc,i,argc,np,nin,km,ks
      integer*4    number_of_processors,cl_size,ciofopen
      integer*4    lng
      integer str_trim
      character   extfil*132,cmdlin*250
      logical getmol,readcon,rdabtabs
      logical coulomb,constr,flg,dynamic
c     
c     Setup options
c     
      call getenv('MINIMAX_FF_TABLES',ff_tab_dir)
      if (ff_tab_dir.eq.' ') ff_tab_dir='./'
c     
      FPEPS0=332.0716   ! FF dependent 'constant'
      input=.TRUE.
      dynamic=.TRUE.
      maxrestart=5
      negr=-1
      agr=0.00035
      nenout=10
      xout=.FALSE.
      pres_chiral=.false.
      deexp=0.01
      method=2          ! try limited memory BFGS, whenever possible
      molfile=' '
      coulomb=.true.
      diel=1.0
      diel_const=.false.
      save=.false.
      argc=iargc()
      constr=.false.
      cutoff=1.e30
      cutmrg=1.0
      cutswr=1.0
      vshift=.false.
      vswitch=.false.
      eshift=.false.
      eswitch=.false.
      cl_size=number_of_processors()
      nproc=1
      hbscale=0.70
      fftype='TAFF'
      psf_file=' '
      e14fac=0.5
      xrq_file=' '
      prm_file=''       ! in minimax FF directory...
      nbxmod=5
      scee=1.0D0/1.2D0
      scnb=1.0D0/2.0D0
c     
c     decode command line arguments, UNIX style...
c     
      i=0
      do while(i.lt.argc)
         i=i+1
         call getarg(i,cmdlin)
         if (cmdlin(1:1).eq.'-') then
            if (cmdlin.eq.'-taff'.or.cmdlin.eq.'-TAFF') then
               fftype='TAFF'
            else if (cmdlin.eq.'-charmm'.or.cmdlin.eq.'-CHARMM') then
               fftype='CHARMM'
            else if (cmdlin.eq.'-amber'.or.
     .              cmdlin.eq.'-AMBER'.or.cmdlin.eq.'-Amber') then
               fftype='AMBER'
            else if (cmdlin(2:4).eq.'psf') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,'(a)') psf_file
            else if (cmdlin(2:4).eq.'prm') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,'(a)') prm_file
            else if (cmdlin(2:4).eq.'e14') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) e14fac
            else if (cmdlin(2:5).eq.'scnb') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) scnb
               if (scnb.ne.0.0D0) scnb=1.0D0/scnb
            else if (cmdlin(2:5).eq.'scee') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) scee
               if (scee.ne.0.0D0) scee=1.0D0/scnb
            else if (cmdlin(2:4).eq.'nbx') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) nbxmod
            else if (cmdlin(2:4).eq.'esw') then
               eswitch=.true.
               eshift=.false.
            else if (cmdlin(2:4).eq.'esh') then
               eshift=.true.
               eswitch=.false.
            else if (cmdlin(2:4).eq.'vsw') then
               vswitch=.true.
               vshift=.false.
            else if (cmdlin(2:4).eq.'vsh') then
               vshift=.true.
               vswitch=.false.
            else if (cmdlin(2:4).eq.'xrq') then
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,'(a)') xrq_file
            else if (cmdlin(2:2).eq.'i') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,*) negr
            else if (cmdlin(2:2).eq.'R') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,*) maxrestart
            else if (cmdlin(2:2).eq.'n') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               km=index(cmdlin,'+')
               ks=index(cmdlin,'-')
               if (km.gt.0) cmdlin(km:km)=' '
               if (ks.gt.0) cmdlin(ks:ks)=' '
               if (km.eq.ks) then
                  read(cmdlin,*) cutoff
               else if (km.lt.ks) then
                  if (km.eq.0) then
                     read(cmdlin,*) cutoff,cutswr
                  else
                     read(cmdlin,*) cutoff,cutmrg,cutswr
                  end if
               else
                  if (ks.eq.0) then
                     read(cmdlin,*) cutoff,cutmrg
                  else
                     read(cmdlin,*) cutoff,cutswr,cutmrg
                  end if
               end if
               cutswr=min(cutswr,cutoff)
               if (cutoff.eq.0) cutoff=1.e30
            else if (cmdlin(2:2).eq.'h') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,*) hbscale
            else if (cmdlin(2:2).eq.'p'.or.cmdlin(2:2).eq.'P') then
               dynamic=(cmdlin(2:2).eq.'p')
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,*) nproc
            else if (cmdlin(2:2).eq.'r') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,*) nenout
            else if (cmdlin(2:2).eq.'g') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               write(*,'(a,a,a)')
               read (cmdlin,*) agr
            else if (cmdlin(2:2).eq.'c') then
               coulomb=cmdlin(3:3).ne.'-'
            else if (cmdlin(2:2).eq.'C') then
               if (cmdlin(3:3).eq.'f') then
                  i=i+1
                  call getarg(i,constr_file)
                  constr=.true.
               else if (cmdlin(3:3).eq.'s') then
                  pres_chiral=.TRUE.
               else
                  constr_file=' '
                  constr=.true.
               end if
            else if (cmdlin(2:2).eq.'s') then
               save=cmdlin(3:3).ne.'-'
            else if (cmdlin(2:2).eq.'b') then
               k=2
               do while (cmdlin(k:k).eq.'b')
                  deexp=deexp*100
                  k=k+1
               end do
            else if (cmdlin(2:2).eq.'m') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               if (cmdlin(1:1).eq.'c') then        ! Conj. gradient in C
                  method=1
               else if (cmdlin(1:1).eq.'b') then   ! lim. memory BFGS (J. Nocedal)
                  method=2
               else if (cmdlin(1:1).eq.'o') then   ! CGFAM (J. Nocedal)
                  method=3
               else
                  write(*,'(2a)') 
     .                 ' ERROR: unknown method ',cmdlin(1:10)
               end if
            else if (cmdlin(2:2).eq.'t') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               ff_tab_dir=cmdlin
            else if (cmdlin(2:2).eq.'d') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,'(f20.0)',err=8400) diel
               diel_const=.true.
            else if (cmdlin(2:2).eq.'D') then
               if (cmdlin(3:3).ne.' ') then
                  cmdlin=cmdlin(3:)
               else
                  i=i+1
                  call getarg(i,cmdlin)
               end if
               read(cmdlin,'(f20.0)',err=8400) diel
               diel_const=.false.
            else if (cmdlin(2:2).eq.'X') then
               xout=.TRUE.
            else
               write(*,'(a,a)') 
     .              ' ERROR : unknown option ',cmdlin(1:10)
               input=.FALSE.
               return
            end if
         else
            read(cmdlin,'(a)') molfile
         end if
      end do
c     
      if (fftype.eq.'TAFF') then
         vshift=.false.         ! always use vswitch
         vswitch=.not.vshift
         eshift=.not.diel_const ! use eshift iff distance dep. dielectric
         eswitch=.not.eshift
         FPEPS0=332.17
      end if
c     
      if (fftype.eq.'CHARMM') then
         if (nbxmod.lt.-5.or.nbxmod.gt.5) then
            write(*,'(a)') ' ERROR : bad value of nbxmod.'
            input=.false.
         end if
         if (.not.eshift.and..not.eswitch) then
            eshift=.true.       ! use e-shift, by default
            eswitch=.not.eshift
         end if
         if (.not.vshift.and..not.vswitch) then
            vshift=.false.      ! use v-switch, by default
            vswitch=.not.vshift
         end if
         if (prm_file.eq.' ') prm_file='PARM.PRM'  ! in the default FF directory 
         FPEPS0=332.0716
      end if
c     
      if (fftype.eq.'AMBER') then
         if (.not.eshift.and..not.eswitch) then
            eshift=.false.      ! use e-switch, by default
            eswitch=.not.eshift
         end if
         if (.not.vshift.and..not.vswitch) then
            vshift=.false.      ! use v-switch, by default
            vswitch=.not.vshift
         end if
         if (prm_file.eq.' ') prm_file=extfil(molfile,'.prm',.true.)
         FPEPS0=18.2223D0**2
      end if
c     
      if (.not.input) return
c     
      if (nproc.lt.1) nproc=1
      if (nproc.gt.mxproc) nproc=mxproc
      if (nproc.gt.cl_size) nproc=cl_size
#ifdef USEOMP
      call omp_set_num_threads(nproc)
      call omp_set_dynamic(dynamic)
#endif
c     
c     get mol-filename
c     
      if (molfile.eq.' ') then
         write(*,'(a,$)') ' Molecule: '
         read(*,'(a)') molfile
      end if
      if (psf_file.eq.' ') psf_file=extfil(molfile,'.psf',.true.)
c     
c     Open .MOL file and output file...
c     
      call unlink(extfil(molfile,'.dat',.true.))
      open(unit=21,file=extfil(molfile,'.dat',.true.),
     *     status='NEW')
      cmdlin=extfil(molfile,'.mol',.false.)
      call ztrimm(cmdlin,len(cmdlin))
      nin=ciofopen(cmdlin,'r')
      if (nin.lt.0) then
         cmdlin=extfil(molfile,'.mol2',.false.)
         call ztrimm(cmdlin,len(cmdlin))
         nin=ciofopen(cmdlin,'r')
      end if
      molfile=cmdlin

      if (nin.lt.0) go to 8100
c     
c     Read atom and bond definition tables...
c     
      lng=str_trim(ff_tab_dir,ff_tab_dir)
      if (ff_tab_dir(lng:lng).ne.'/'.and.ff_tab_dir(lng:lng).ne.']') 
     *     ff_tab_dir(lng+1:lng+1)='/'
      if (.not.rdabtabs()) go to 8000
c     
c     Retrieve molecular description
c     
      call molrd(nin,status)
      call ciofclose(nin)
      if (status.ne.0) then
         write(6,'(a,i3)') ' *** ERROR reading .MOL file.',status
         write(21,'(a,i3)') ' *** ERROR reading .MOL file.',status
         input=.FALSE.
         return
      end if
      call renum
      if (.not.getmol(nat,name,xyz,nres,respt,ntype,chrg,chrgok,
     *     nbnd,bnd,bndt))
     *     go to 8200
      k=0
      do i=1,nat
         segid(i)='----'
         resid(i)='----'
         resnr(i)='----'
      end do
      if (nat.lt.2) then
         write(6,'(a)') ' Less than two non-dummy atoms.'
         input=.FALSE.
         return
      end if
      chrgok=chrgok.and.coulomb
      if (chrgok) then
         flg=.false.
         do i=1,nat
            flg=flg.or.abs(chrg(i)).gt.0
         end do
         chrgok=chrgok.and.flg
      end if
      if (negr.eq.0.or.nenout.eq.0) save=.false.
c     
c     Read constraints...
c     
      ncon=0
      ncona=0
      nconp=0
      ncond=0
      nconc=0
      nconx=0
      nfix=0
      ntot=nat
      do i=1,nat
         agg(i)=0
      end do
      if (constr) then
         if (constr_file.eq.' ') then
            open(unit=1,file=extfil(molfile,'.con',.true.),
     .           status='old',err=8300 READONLY)
            if (.not.readcon()) go to 8200
            close(unit=1)
         else
            open(unit=1,file=constr_file,
     .           status='old',err=8350 READONLY)
            if (.not.readcon()) go to 8200
            close(unit=1)
         end if
      end if
c     
      input=.TRUE.
      return
c     
c     Errors:
c     
 8000 continue
      input=.FALSE.
      return
 8100 continue
      lng=str_trim(molfile,extfil(molfile,'.mol',.false.))
      write(6,'(a,a)') ' Can''t find input molecule ',molfile(1:lng)
      input=.FALSE.
      return
 8200 continue
      input=.FALSE.
      return
 8300 write(6,'(a,a)') 
     .     ' Could not open ',extfil(molfile,'.con',.true.)
      input=.FALSE.
      return
 8350 write(6,'(a,a)') ' Could not open ',constr_file
      input=.FALSE.
      return
 8400 write(6,'(a)') ' Bad format of DIELECTRICITY option.'
      input=.FALSE.
      return
      end
c     -------------------------------------------------------------------------
c     
      integer function number_of_processors()
c     
      integer mp_numthreads,omp_get_num_procs
c     
      number_of_processors=1
#ifdef USEOMP
      number_of_processors=omp_get_num_procs()
#endif
      return
      end
