      Program MINIMAX
c     
c     Force Field Program. Conjugate gradient minimizer applied to
c     SYBYL force field. 
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
#include "option.h"
#include "worksp.h"
c     
      integer nrestart
      integer status,npar,xyz2par
      integer len
      integer str_len,conjgradtest
      character*4 dieltxt
      character*6 ecut,vcut
      character*7 nbcut
      character*35 meth(3)
      logical input,topol,readff,gnblst,add_chiral
      real*8 etot,egf,eff
      real*4 etime,ttot,topt,te1,te2
      real*4 secnds,elapsed
      real*8 rmsgrad,svdeexp,fact,onestep,gmax
      data meth /
     .     'Conjugate Gradient (Polak-Ribiere) ',
     .     'lim. memory BFGS (J. Nocedal)      ',
     .     'Conjugate Gradient (J. Nocedal)    ' /
c     
c     Initialize timers
c     
      call conjgradclear
      elapsed=secnds(0.0)
      ttot=etime(tarray)
      tetot=0.0
      tebond=0.0
      tebend=0.0
      teang=0.0
      tetors=0.0
      tecou=0.0
      tevdw=0.0
      tecon=0.0
      call cioinit()
c     
c     Read molecule
c     
      status=1
      if (.not.input()) go to 8000
      status=2
      if (nat.eq.0) go to 8000
      status=3
      if (.not.topol()) go to 8000
      status=4
c     
c     Set up forcefield for subsequent MM optimization
c     
      if (.not.readff()) go to 8000
      status=5
c
c   Add chirality constraints...
c
      if (pres_chiral) then
         if (.not.add_chiral()) go to 8000
      end if
c
c   Remove redundant terms from Energy if there are fixed atoms...
c
      if (nfix.gt.0) call reduce
      nbl_valid=.false.
      status=6
c     
c     Dump initial status and options...
c     
      npar=xyz2par(xyz,work(3*nat+1))     ! parameters in work(3*nat+1)
      if (negr.eq.-1) negr=5*npar
      ncalls=-1
      len=str_len(molfile)
      write(6,'(/a,a)') 
     .	' MINIMAX Molecular Mechanics Calculation for ',molfile(1:len)
      write(6,'(a,i2,a/)')  ' Running on ',nproc,' processor(s)'
      write(21,'(/a,a)') 
     .	' MINIMAX Molecular Mechanics Calculation for ',molfile(1:len)
      write(21,'(a,i2,a/)')  ' Running on ',nproc,' processor(s)'
      dieltxt=' * r'
      if (diel_const) dieltxt=' '
      nbcut='none'
      if (cutoff.lt.1.0e30) write(nbcut,'(f7.1)') cutoff
      if (fftype.eq.'CHARMM') then
         len=str_len(prm_file)
         write(21,'(2a)') 
     .     ' FF parameters read from ',prm_file(1:len)
      else if (fftype.eq.'AMBER') then
         len=str_len(prm_file)
         write(21,'(2a)') 
     .     ' Topology and FF parameters read from ',prm_file(1:len)
      else
         len=str_len(ff_tab_dir)
         write(21,'(2a)') 
     .     ' FF parameter files read from ',ff_tab_dir(1:len)
      end if
      write(21,'(2a/a,i6/a,f9.5/a,f5.1,a)') 
     .     ' Minimizer :                   ',meth(method),
     .     ' max. number of iterations:       ' ,negr,
     .     ' required accuracy :           '    ,agr,
     .     ' dielectric function :             ',diel,dieltxt
      if (fftype.eq.'TAFF') then
         write(21,'(a,f5.2)') 
     .     ' TAFF HB radius scale :            ',hbscale
      end if
      if (cutoff.lt.1.e30) then
         write(21,'(a,f7.2/a,f7.2/a,f7.2)') 
     .        ' cutoff switching on :           ',cutoff-cutswr,
     .        ' cutoff switching off:           ',cutoff,
     .        ' non-bonded list range :         ',cutoff+cutmrg
      else
         write(21,'(a)') ' No cutoff'
      end if
      if (fftype.eq.'CHARMM') then
         ecut="trunc."
         vcut="trunc."
         if (vswitch) vcut='switch'
         if (vshift) vcut='shift'
         if (eswitch) ecut='switch'
         if (eshift) ecut='shift'
         write(21,'(a,i2,a,f5.2,4a)')
     .        ' CHARMM: nbxmod = ',nbxmod,', e14fac = ',e14fac,
     .        ', cutoff: elec = ',ecut,
     .        ', vdW = ',vcut
      end if
      if (fftype.eq.'AMBER') then
         ecut="trunc."
         vcut="trunc."
         if (vswitch) vcut='switch'
         if (vshift) vcut='shift'
         if (eswitch) ecut='switch'
         if (eshift) ecut='shift'
         write(21,'(a,f6.4,a,f6.4,4a)')
     .        ' AMBER: scnb = ',1/scnb,', scee = ',1/scee,
     .        ', cutoff: elec = ',ecut,
     .        ', vdW = ',vcut
      end if
c     
      if (nenout.gt.0) then
         write(6,'(2a)') 
     .        ' Iter     Total      Bond     Angle   Torsion',
     .        '     Plane   vdWaals   Coulomb'
      end if
      if (negr.eq.0) ncalls=0
      call trapint                 ! catch ^C signal...
      etot=egf(npar,work(3*nat+1),work(6*nat+1))
      call dpar2dxyz(work(6*nat+1),work) ! moldmp wants dE/dx in work
      rms_g=rmsgrad(nat,work,gmax)
      call moldmp
      ebest=etot
      if (save) call putmol(nat,xyz)            
      write(21,'(///,6(/a,f11.3,a),/a/a,f11.3,a////)')
     .     '      Bond stretching : ',ebnd,' Kcal/mole',
     .     '      Angle bending   : ',eang,' Kcal/mole',
     .     '      Plane bending   : ',epla,' Kcal/mole',
     .     '      Torsion energy  : ',etors,' Kcal/mole',
     .     '      van der Waals   : ',evdw,' Kcal/mole',
     .     '      Coulomb energy  : ',ecou,' Kcal/mole',
     .     '      =======================================',
     .     '      INITIAL ENERGY  : ',etot,' Kcal/mole'
      if (ncon+ncona+ncond+nconc+nconp+nconx.gt.0) then
         eff=etot-econ
         write(21,'(/a,f11.3,a/a,f11.3,a)')
     .        '      Constraints     : ',econ,' Kcal/mole',
     .        '      Mol. Mech. Energ: ',eff ,' Kcal/mole'
      end if
      if (negr.eq.0) then
         write(6,'(/a,f10.5/a,f10.5)')
     .        '   Gradient r.m.s (mdyn) : ',rms_g,
     .        '      "     max.  (mdyn) : ',fmax_g
         status=7
         go to 8000
      end if
c     
c     Do MM
c     
      if (nenout.gt.0) then
         write(21,'(a,a,a)') char(12),
     .        ' Iter     Total      Bond     Angle   Torsion',
     .        '     Plane   vdWaals   Coulomb'
      end if
      mmcon=.false.
      topt=etime(tarray)
      te1=tetot
      nrestart=0
      svdeexp=deexp
      do while(.not.mmcon.and..not.conjgradtest().eq.1.and.
     .     nrestart.le.maxrestart.and.ncalls.lt.negr)
         if (gmax.gt.1.0.or..not.(gmax.gt.-1)) then
            deexp=onestep(nat,xyz)
         end if
         if (deexp.eq.0)  deexp=svdeexp
         call mmcalc
         te2=tetot-te1
         topt=etime(tarray)-topt
         npar=xyz2par(xyz,work(3*nat+1))
         etot=egf(npar,work(3*nat+1),work(6*nat+1))
         call dpar2dxyz(work(6*nat+1),work)
         rms_g=rmsgrad(nat,work,gmax)
         mmcon=rms_g.le.agr
         nrestart=nrestart+1
      end do
      deexp=svdeexp
c     
c     Report status of final cycle
c     
      write(6,'(i5,7(f10.2))') ncalls,
     .     etot,
     .     ebnd,
     .     eang,
     .     etors,
     .     epla,
     .     evdw,
     .     ecou
      write(6,'(a)') ' (Energies are in Kcal/mole)'
      write(21,'(i5,7(f10.2))') ncalls,
     .     etot,
     .     ebnd,
     .     eang,
     .     etors,
     .     epla,
     .     evdw,
     .     ecou
      write(21,'(a)') ' (Energies are in Kcal/mole)'
c     
c     Write new coordinates to file
      call moldmp
      mmcon = (rms_g.le.agr)
c     
      write(21,'(///,6(/a,f11.3,a),/a/a,f11.3,a////)')
     .     '      Bond stretching : ',ebnd,' Kcal/mole',
     .     '      Angle bending   : ',eang,' Kcal/mole',
     .     '      Plane bending   : ',epla,' Kcal/mole',
     .     '      Torsion energy  : ',etors,' Kcal/mole',
     .     '      van der Waals   : ',evdw,' Kcal/mole',
     .     '      Coulomb energy  : ',ecou,' Kcal/mole',
     .     '      =======================================',
     .     '      TOTAL ENERGY    : ',etot,' Kcal/mole'
      eff=etot-econ
      write(6,'(/a,f10.5/a,f10.5)')
     .     '   Gradient r.m.s (mdyn) : ',rms_g,
     .     '      "     max.  (mdyn) : ',fmax_g
      if (ncon+ncona+ncond+nconc+nconp+nconx.gt.0) then
         write(6,'(/a,f11.3,a/a,f11.3,a)')
     .        '      Constraints     : ',econ,' Kcal/mole',
     .        '      Mol. Mech. Energ: ',eff ,' Kcal/mole'
         write(21,'(/a,f11.3,a/a,f11.3,a//)')
     .        '      Constraints     : ',econ,' Kcal/mole',
     .        '      Mol. Mech. Energ: ',eff ,' Kcal/mole'
      end if
      if (mmcon) then
         write(6,'(a)') ' Molecular Mechanics converged.'
         write(21,'(a)') ' Molecular Mechanics converged.'
C	 26.4.2005: changed normal exit status from 99 to 0 (prevent requeueing loop in SGE)
         status=0
      else if (ncalls.lt.negr) then
         if (.not.xout) then
            write(6,'(a)') ' *** Molecular Mechanics not converged ***'
         end if
         write(21,'(a)') ' *** Molecular Mechanics not converged ***'
         status=7
      else
         if (.not.xout) then
            write(6,'(a)')' *** Max. number of iterations exceeded. ***'
         write(21,'(a)')' *** Max. number of iterations exceeded. ***'
         end if
         status=8
      end if
      if (.not.xout) then
         write( 6,'(/a,a)') ' Algorithm used : ',meth(method)
      end if
      write(21,'(/a,a)') ' Algorithm used : ',meth(method)
      ttot=etime(tarray)-ttot
      elapsed=secnds(elapsed)
      if (.not.xout) then
         write(*,'(/a,f12.1)') ' CPU Time (seconds) : ',ttot
         write(*,'(a,f12.1)')   ' Elapsed Time (sec) : ',elapsed
      end if
      write(21,'(/a,f12.1)') ' CPU Time (seconds) : ',ttot
      write(21,'(a,f12.1)')   ' Elapsed Time (sec) : ',elapsed
c     
c     Output timer statistics:
c     
      if (negr.eq.0) ncalls=1
      if (.not.xout) then
         write(6,'(/a/a/)') ' Timer statistics : ',
     .        ' ----------------'
         write(6,'(a,f10.4)') ' Time spent in opt. loop      : ',topt
         write(6,'(a,f10.4)') ' Energy & Gradient in opt.loop: ',te2
         write(6,'(a,f10.4)') ' Energy & Gradient (total)    : ',tetot
         write(6,'(a,f10.4)') ' Bond Stretch                 : ',tebond
         write(6,'(a,f10.4)') ' Angle Bend                   : ',teang
         write(6,'(a,f10.4)') ' Torsion                      : ',tetors
         write(6,'(a,f10.4)') ' Out of Plane                 : ',tebend
         write(6,'(a,f10.4)') ' van der Waals + Coulomb      : ',tevdw
         write(6,'(a,f10.4)') ' Generation of non-bonded list: ',tgnblst
         if (ncon+nconp+ncond+nconc+ncona+nconx.gt.0) 
     .        write(6,'(a,f10.4)') ' Constraints                  : ',
     .        tecon
         write(6,'(/a,i4/)')  ' Sample size : ',ncalls+1
      end if
c     
      write(21,'(/a/a/)') ' Timer statistics : ',
     .     ' ----------------'
      write(21,'(a,f10.4)') ' Time spent in opt. loop      : ',topt
      write(21,'(a,f10.4)') ' Energy & Gradient in opt.loop: ',te2
      write(21,'(a,f10.4)') ' Energy & Gradient (total)    : ',tetot
      write(21,'(a,f10.4)') ' Bond Stretch                 : ',tebond
      write(21,'(a,f10.4)') ' Angle Bend                   : ',teang
      write(21,'(a,f10.4)') ' Torsion                      : ',tetors
      write(21,'(a,f10.4)') ' Out of Plane                 : ',tebend
      write(21,'(a,f10.4)') ' van der Waals + Coulomb      : ',tevdw
      write(21,'(a,f10.4)') ' Generation of non-bonded list: ',tgnblst
      if (ncon+nconp+ncond+nconc+ncona+nconx.gt.0) 
     .     write(21,'(a,f10.4)') ' Constraints                  : ',
     .     tecon
      write(21,'(/a,i4/)')  ' Sample size : ',ncalls+1
c     
      close(unit=21)	
      call putmol(nat,xyz)
      if (xout) call xputmol(nat,xyz)
c     
c     That is it, folks...
c     
 8000 continue
      call write_xrq(xrq_file)
      if (status.eq.1) then
         write(6,'(a)') ' Exit status: input error.'
      else if (status.eq.2) then
         write(6,'(a)') ' Exit status: no atoms.'
      else if (status.eq.3) then
         write(6,'(a)') ' Exit status: topology error.'
      else if (status.eq.4) then
         write(6,'(a)') ' Exit status: missing FF parameters.'
      else if (status.eq.5) then
         write(6,'(a)') ' Exit status: non-bonded list error.'
      else if (status.eq.6) then
         write(6,'(a)') ' unknown error, check coredump...'
      else if (status.eq.7) then
         if (negr.gt.0) then
            if (method.eq.1) then
               write(6,'(a)') ' *** Not converged, try "-bbb" option.'
            else if (method.eq.2) then
               write(6,'(2a)') 
     .              ' *** Not converged, try to restart from final',
     .              ' structure'
            else
               write(6,'(a)') ' *** Not converged.'
            end if
         end if
      else if (status.eq.8) then
         write(6,'(a)') ' *** Maximum number of iterations exceeded.'
      else if (status.eq.99.or.status.eq.0) then
         write(6,'(a)') ' Normal completion, converged.'
      else
         write(6,'(a)') ' *** Exit status unknown.'
      end if
      if (.not.xout) write(6,'(/)')
      call exit(status)
      end
