      subroutine mmcalc
c     
c     Molecular mechanics part of Minimax
c     
      implicit none
#include "sizes.h"
#include "atom.h"
#include "energy.h"
#include "worksp.h"
#include "option.h"
c     
      integer status
      real*8  ef,p(3,matot),egf
      integer np,xyz2par,conjgrad,dlbfgs,dcgfam
      integer iprint(2),m,op
c     
      external egf
c
      np=xyz2par(xyz,p)
c     
c     Clear interrupt flag and selectminimization technique:
c     
      call conjgradclear

      if (method.eq.1) then          ! Conjugate Gradient in C (Polak Ribiere)
         status=conjgrad(egf,np,p,ef,work,deexp,agr/6.95E-3,negr)
         if (status.eq.-3) then
            write(6,'(A)') ' Conjugate gradient ran out of memory.'
         end if
         mmcon=status.ge.0

      else if (method.eq.2) then     ! limited memory BFGS (Jorge Nocedal)
         m=(mwork-5*np)/(2*np+2)
         if (m.gt.1) then
            status=dlbfgs(egf,np,m,p,ef,work,agr/6.95E-3,negr,
     .           work(1+np))
            if (status.eq.-999) then    ! we don't have the LBFGS minimzer
               method=1
               status=conjgrad(egf,np,p,ef,work,deexp,agr/6.95E-3,negr)
            end if
         else
            method=1
            write(6,'(A)') ' System is too big for LBFGS, using '
     .           //'simple Conjugate gradient instead'
            status=conjgrad(egf,np,p,ef,work,deexp,agr/6.95E-3,negr)
            if (status.eq.-3) then
               write(6,'(A)') ' Conjugate gradient ran out of memory.'
            end if
         end if
         mmcon=status.eq.0
         
      else if (method.eq.3) then     ! J. Nocedal's Polak Ribiere CG
         m=-3                        ! select PR with restart
         if (5*np.le.mwork) then
            status=dcgfam(egf,np,m,p,ef,work,agr/6.95E-3,negr,
     .           work(1+np))
            if (status.eq.-999) then    ! we don't have the CGFAM minimzer
               method=1
               status=conjgrad(egf,np,p,ef,work,deexp,agr/6.95E-3,negr)
            end if
         else
            method=1
            write(6,'(A)') ' System is too big for CGFAM, using '
     .           //'simple Conjugate gradient instead'
            status=conjgrad(egf,np,p,ef,work,deexp,agr/6.95E-3,negr)
            if (status.eq.-3) then
               write(6,'(A)') ' Conjugate gradient ran out of memory.'
            end if
         end if
         mmcon=status.eq.0
      end if

      call par2xyz(p,xyz)
c     
c     done...
c     
      return
      end
c------------------------------------------------------------------------
      subroutine pleasestop
c
#include "option.h"
c
c      Called by SIGINT signal handler
c
      call conjgradstop
      write(6,'(a)') '*** STOP scheduled at end of next iteration.'
c
      return
      end
