      subroutine molrd(nin,status)
c     ****************************
c
c     the subroutine reads the sybyl.mol-file into the common
c     block 'sybcom'.
c
c     status = 0: no errors.
c            = 1: file has not mol-file format.
c            = 2: error in mol-file.
c            = 3: too many atoms.
c
      implicit none
#include "sizes.h"
#include "sybcom.h"
c
      integer NONE,ATOM,BOND,MOLECULE
      parameter (NONE=0,ATOM=1,BOND=2,MOLECULE=3)
      character line*160,idmol*4,rec*80
      integer*4 nin,status,ichrg(5),i,j,k,nft,nftch,nchrg,nftyp
      real*8 chflag,val1,val2
      integer*4 cioftell,pos,ciofgets,ciogetfld,in,
     .     doanatom,doabond,section,nmolr,nat_x,nbnd_x
c
      status = 0
      pos=cioftell(nin)
      in=ciofgets(nin,rec,80)
      if (in.ne.0.and.rec(5:8).eq.' MOL') then
         ismol2=.false.
      else 
         ismol2=.true.
      end if
      if (ismol2) go to 2000

c############################## OLD FORMAT ##############################
c     read atom records:
c
      read(rec,'(i4,a,10x,a,1x,a,1x,i4)',end=2000,err=2000) 
     .     nat1,idmol,molnam,filnam,iroot
      ismol2=.FALSE.
      if(nat1.gt.matot) then
         status=3
         return 
      end if
      do i = 1,nat1
         if (ciofgets(nin,rec,80).eq.0) go to 100
         read(rec,1010,end=100,err=100) 
     .        natp(i),(xyz(k,i),k=1,3),natlab(i)
1010     format(4x,1i4,3f9.4,a)
         subid(i)=0
      end do
c
c     read bond records:
c
      if (ciofgets(nin,rec,80).eq.0) go to 100
      read(rec,'(i4)',end=100,err=100) nbond
      do i = 1,nbond
         if (ciofgets(nin,rec,80).eq.0) go to 100
         read(rec,1020,end=100,err=100) nbnd(1,i),nbnd(2,i),nbndtp(i)
1020     format(4x,2i4,9x,i4)
      end do
c
c     read feature records:
c
      chfl = .false.
      do i = 1,matot
         chrg(i) = 0.0
      end do
      if (ciofgets(nin,rec,80).eq.0) go to 90
      read(rec,'(i4)',end=90,err=100) nft
      if(nft.eq.0) go to 90
c
      nfta = 0
      nftch = 0
      nchrg = 0
      chflag = 1.0
      do i =1,nft
         if (ciofgets(nin,rec,80).eq.0) go to 100
         read(rec,'(i4)',end=100,err=100) nftyp
         if(nftyp.eq.4) then
            nftch = nftch + 1
            read(rec,1025,end=100,err=100) (ichrg(j),j=1,5),val1,val2
1025        format(4x,5i6,2f8.4)
            do 20 j = 1,5
               nchrg = nchrg + 1
               chrg(nchrg) = float(ichrg(j))/10000.
20          continue
            if(nftch.eq.1) chflag = val2
            if(nftch.eq.2) chgmet = val1
         else
            nfta = nfta + 1
            ftlin(nfta) = line
         end if
      end do
      chfl = (chflag.lt.0.1).and.nftch.gt.0
c
      if(.not.chfl) then
         nft = nfta
         chgmet = 0.0
         do i = 1,nat1
            chrg(i) = 0.0
         end do
      end if
 90   continue
      return
c
100   status = 2
      return 
c
c############################## NEW FORMAT ##############################
c
 2000 continue
      ismol2=.TRUE.
      call ciofseek(nin,pos)
      nat1=0
      nbond=0
      filnam=' '
      iroot=0
      section=NONE
      status=0
      chfl=.true.
      in=0
      nmolr=0
      nat_x=-1
      nbnd_x=-1
c
c   skip to molecule section
c
      do while(status.eq.0.and.in.ge.0)
         in=ciogetfld(nin,rec,80)
         if (in.gt.0) then
            if (rec(1:1).eq.'@') then
               call str_upcase(rec(1:in),rec(1:in))
               if (rec(1:13).eq.'@<TRIPOS>ATOM') then
                  section=ATOM
               else if (rec(1:13).eq.'@<TRIPOS>BOND') then
                  section=BOND
               else if (rec(1:17).eq.'@<TRIPOS>MOLECULE') then
                  section=MOLECULE
               else
                  section=NONE
               end if
            else if (rec(1:1).ne.'#') then
               if (section.eq.ATOM) then
                  status=doanatom(nin,rec)
               else if (section.eq.BOND) then
                  status=doabond(nin)
               else if (section.eq.MOLECULE) then
                  if (nmolr.eq.0) then
                     molnam=rec(1:in)
                  else if (nmolr.eq.1) then
                     read(rec,'(bn,i20)',end=2999,err=2999) nat_x
                     in=ciogetfld(nin,rec,20)
                     if (in.gt.0) then
                        read(rec,'(bn,i20)',end=2999,err=2999) nbnd_x
                     end if
                  else if (nmolr.eq.3) then
                     call str_upcase(rec(1:in),rec(1:in))
                     if (in.eq.10.and.rec(1:10).eq.'NO_CHARGES') then
                        chfl=.false.
                     end if
                  end if
                  nmolr=nmolr+1
               end if
            end if
         end if
         call ciofskip(nin,1)
      end do
c               
c a few sanity checks...
c
      if (status.eq.1002) then
         write(6,'(a,i6)') ' too many atoms, max allowed:',matot
         status=1001
      else if (status.eq.2002) then
         write(6,'(a,i6)') ' too many bonds, max allowed:',mbtot
         status=1001
      else if (status.eq.1003) then
         write(6,'(a)') ' format error in atom section.'
         status=1001
      else if (status.eq.2003) then
         write(6,'(a)') ' format error in bond section.'
         status=1001
      else if (nat1.ne.nat_x.or.nbond.ne.nbnd_x) then
         write(6,'(a,i6,i6)') ' number of atoms/bonds = ',nat1,nbond
         write(6,'(a,i6,i6)') ' expected              = ',nat_x,nbnd_x
         status=1002
      end if
      return
c
 2999 continue
      do i=1,10
         if (ciofgets(nin,rec,80).gt.0) write(6,'(1x,a)') rec
      end do
      status=1001
      return
      end
c
c------------------------------------------------------------------------
c
      integer function old_ff_type(mnem)
c
      implicit none
#include "sizes.h"
#include "apara.h"
#include "option.h"
      character*8 mnem
      integer i
c
      old_ff_type=0
      do i=1,n_atype
         if (amnem(i).eq.mnem) then
            old_ff_type=old_type(i)
            return
         end if
      end do
      if (fftype.eq.'CHARMM'.or.fftype.eq.'AMBER') then
         if (n_atype.lt.matype) then
            n_atype=n_atype+1
            amnem(n_atype)=mnem
            old_type(n_atype)=n_atype
            old_ff_type=n_atype
         end if
      end if
      return
      end
c------------------------------------------------------------------------
c
      integer function old_bnd_type(mnem)
c
      implicit none
#include "sizes.h"
#include "apara.h"
#include "option.h"
      character*8 mnem
      character*2 bt(8)
      integer i
      integer bid(8),chm(8)
      logical chem(8)
      data bt  / '1','2','3','am','ar','du','un','nc' /
      data bid /  1,  2,  3,  4,   5,   8,   9,   10  /
      data chm/   1,  1,  1,  1,   1,   0,   0,    0  /
c
      old_bnd_type=0
      do i=1,n_btype
         if (bmnem(i).eq.mnem) then
            old_bnd_type=bnd_id(i)
            return
         end if
      end do
      if (fftype.eq.'CHARMM'.or.fftype.eq.'AMBER') then
         if (n_btype.lt.mbtype) then
            do i=1,8
               if (mnem(1:2).eq.bt(i)) then
                  n_btype=n_btype+1
                  bmnem(n_btype)=mnem
                  bnd_id(n_btype)=bid(i)
                  chem_bnd(n_btype)=chm(i).ne.0
                  old_bnd_type=bnd_id(n_btype)
               end if
            end do
         end if
      end if
      return
      end
c------------------------------------------------------------------------
c
      integer*4 function doanatom(nin,fld)
c
      implicit none
      integer nin
      character*(*) fld
#include "sizes.h"
#include "sybcom.h"
c
      real*8 x(3),ch
      character fft*8,nm*8,num*20
      integer*4 i,in,old_ff_type,ciogetfld,sub
c
      doanatom=1001
      if (ciogetfld(nin,nm,8).le.0) return
      do i=1,3
         if (ciogetfld(nin,num,20).le.0) return
         read(num,'(f20.0)',end=100,err=100) x(i)
      end do
      if (ciogetfld(nin,fft,8).le.0) return
      in=ciogetfld(nin,num,20)
      if (in.le.0.or.num(1:4).eq.'****') then
         sub=0
      else
         read(num,'(bn,i20)',end=100,err=100) sub
      end if
      do i=1,2
         in=ciogetfld(nin,num,20)
      end do
      if (in.le.0.or.num(1:4).eq.'****') then
         chfl=.false.
         ch=0.0
      else
         read(num,'(f20.0)',end=100,err=100) ch
      end if
      doanatom=1002
      nat1=nat1+1
      if (nat1.gt.matot) return
      natlab(nat1)=nm
      do i=1,3
         xyz(i,nat1)=x(i)
      end do
      natp(nat1)=old_ff_type(fft)
      chrg(nat1)=ch
      subid(nat1)=sub
      doanatom=0
      return
c
 100  continue
      doanatom=1003
      return
      end
c------------------------------------------------------------------------
c
      integer*4 function doabond(nin)
c
      implicit none
      integer nin
#include "sizes.h"
#include "sybcom.h"
      character from*20,to*20,btp*8
      integer old_bnd_type,ciogetfld
c
      doabond=2001
      if (ciogetfld(nin,from,20).le.0) return
      if (ciogetfld(nin,to,20).le.0) return
      if (ciogetfld(nin,btp,8).le.0) return
c
      nbond=nbond+1
      doabond=2002
      if (nbond.gt.mbtot) return
      doabond=2003
      read(from,'(bn,i20)',end=100,err=100) nbnd(1,nbond)
      read(to,'(bn,i20)',end=100,err=100) nbnd(2,nbond)
      nbndtp(nbond)=old_bnd_type(btp)
      doabond=0
      return
c      
 100  continue
      return
      end
