#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

/* Dynamic memory nonbonded list                              */
/* -----------------------------                              */
/* sat[i] is a pointer to the neighborlist for atom i,        */
/* nnb[i] is the number of neighbors in the list for atom i,  */
/* ssat[i] is an alias for sat[i], used when i is <= mxshrt.  */
/* nat is the number of atoms (size of sat and nnb arrays),   */
/* mxshrt is the maximum value that can be stored in a short. */
/* Note: sat[i][k] < i, and stored as a short if i <= mxshrt. */

#ifdef SHRT_MAX
static int mxshrt=SHRT_MAX;
#else
static int mxshrt=0;
#endif

static int **sat=(int **)NULL;
static short **ssat=(short **)NULL; /* an alias for sat */
static int *nnb=(int *)NULL;
static int nat=0;

int nblistinit_(int *n)
{
  /* set up of nonbonded list arena for n atoms */
  /* return 0 if set up failed, nat otherwise    */

  int i;
  unsigned short m;

  /* try to determine mxshrt, if no #definition was available */
  if (mxshrt==0) {
    m=1; mxshrt=-1;
    while (m!=0&&mxshrt<128) { /* <128 as an emergency break */
      mxshrt++;
      m=m<<1;
    }
    if (m==0) mxshrt=(1<<mxshrt)-1;
    else mxshrt=-1;  /* if determination failed. */
  }

  /* free old data structures */
  if (sat) {
    for (i=0;i<nat;i++) {
      if (sat[i]) {
	free (sat[i]);
	sat[i]=(int *)NULL;
      }
    }
    free(sat);
    sat=(int **)NULL;
    ssat=(short **)NULL;
  }
  if (nnb) free(nnb);
  nnb=(int *)NULL;

  /* set up new data structures */
  nat=*n;
  if (nat<0) nat=0;

  if (nat>0) {
    sat=(int **)malloc(nat*sizeof(int *));
    ssat=(short **)sat;            /* just an alias... */
    nnb=(int *)malloc(nat*sizeof(int));
    if (sat) for (i=0;i<nat;i++) sat[i]=(int *)NULL;
    else {
      if (nnb) {
	free(nnb);
	nnb=(int *)NULL;
      }
    }
    if (nnb) for(i=0;i<nat;i++) nnb[i]=0;
    else {
      if (sat) {
	free(sat);
	sat=(int **)NULL;
      }
    }
    if (!nnb) nat=0;
  }
  return(nat);
}
/*------------------------------------------------------------------------*/

int nblist_(int *i,int *N)
{
  /* retrieve neighborlist for atom i, return number of neighbors */

  int j,n;

  n=nnb[*i-1];
  if (*i-1<=mxshrt) for (j=0;j<n;j++) N[j]=ssat[*i-1][j];
  else for (j=0;j<n;j++) N[j]=sat[*i-1][j];
  return(n);
}
/*------------------------------------------------------------------------*/

int nbliststore_(int *i,int *lsat)
{
  /* store neighborlist for atom i, return number of neighbors stored  */

  int k,j,n;

  k=*i-1;
  if (0<=k&&k<nat) {
    if (sat[k]) free(sat[k]);
    sat[k]=(int *)NULL;
    nnb[k]=0;
    for (n=0,j=0;j<k;j++) if (lsat[j]!=0) n++;
    if (n>0) {
      if (k<=mxshrt) {
	ssat[k]=malloc(n*sizeof(short));
	if (ssat[k]) {
	  nnb[k]=n;
	  for (n=0,j=0;j<k;j++) {
	    if (lsat[j]==1) ssat[k][n++]=j+1;
	    else if (lsat[j]==2) ssat[k][n++]=-(j+1);
	  }
	}else n=0;
      }else{
	sat[k]=malloc(n*sizeof(int));
	if (sat[k]) {
	  nnb[k]=n;
	  for (n=0,j=0;j<k;j++) {
	    if (lsat[j]==1) sat[k][n++]=j+1;
	    else if (lsat[j]==2) sat[k][n++]=-(j+1);
	  }
	}else n=0;
      }
    }
  }
  return(n);
}
