      logical function rdabtabs()
c     
c     Read atom and bond definition tables...
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "option.h"
#if defined(sgi)
#define READONLY ,readonly
#else 
#define READONLY
#endif
c     
      character cmd*250,term*1,oopmnem*8
      character fld*20
      integer str_trim
      integer pos,start,end,i,j
      integer l
      logical iqstat

      rdabtabs=.TRUE.

C
C     If the FF is CHARMM or AMBER, then we don't really need this...
C
      if (fftype.eq.'CHARMM'.or.fftype.eq.'AMBER') then
         n_atype=0
         n_btype=0
         return
      endif
c     
c     First the atom table ...
c     
      l=str_trim(cmd,ff_tab_dir)
      cmd(l+1:)='ATOM_DEF.'
      inquire(file=cmd,exist=iqstat)
      if (.not.iqstat) cmd(l+1:)='ATOM_DEF'
      open(unit=22,file=cmd,status='old',err=8000 READONLY)
c     
      start=0
      end=0
      pos=1
 10   continue
      read (22,'(a)',end=90) cmd
      if (cmd(1:1).eq.'#') go to 10
      if (n_atype.eq.matype) then
         write(6,*) 'Too many atomtypes in ATOM_DEF file.'
         rdabtabs=.FALSE.
         return
      end if
      n_atype=n_atype+1
      oopcent(n_atype)=.FALSE.
      pos=1
      call token(cmd,250,pos,0,' ',start,end,term)
      amnem(n_atype)=cmd(start:end)
      call token(cmd,250,pos,0,' ',start,end,term)
      hets(n_atype)=cmd(start:end)
      do i=1,2
         call token(cmd,250,pos,0,' ',start,end,term)
      end do
      call token(cmd,250,pos,0,' ',start,end,term)
      equiv(n_atype)=cmd(start:end)
      call token(cmd,250,pos,0,' ',start,end,term)
      fld=cmd(start:end)
      read(fld,'(bn,i20)',end=8001,err=8001) atomic_n(n_atype)
      do i=1,5
         call token(cmd,250,pos,0,' ',start,end,term)
      end do
      hbd(n_atype)=cmd(start:start).eq.'Y'
      call token(cmd,250,pos,0,' ',start,end,term)
      hba(n_atype)=cmd(start:start).eq.'Y'
      do i=1,4
         call token(cmd,250,pos,0,' ',start,end,term)
      end do
      fld=cmd(start:end)
      read(fld,'(bn,i20)',end=8001,err=8001) old_type(n_atype)
      go to 10
c     
 90   continue
      close(unit=22)
      do i=1,n_atype
         etype(i)=i
         if (equiv(i).ne.amnem(i)) then
            do j=1,n_atype
               if (equiv(i).eq.amnem(j)) etype(i)=j
            end do
         end if
      end do
c     
c     And now for the bond tables...
c     
      l=str_trim(cmd,ff_tab_dir)
      cmd(l+1:)='BOND_DEF.'
      inquire(file=cmd,exist=iqstat)
      if (.not.iqstat) cmd(l+1:)='BOND_DEF'
      open(unit=22,file=cmd,status='old',err=8100 READONLY)
c     
 110  continue
      read(22,'(a)',end=190) cmd
      if (cmd(1:1).eq.'#') go to 110   ! to allow comment (Minimax extension)
      if (n_btype.eq.mbtype) then
         write(6,*) 'Too many definitions in BOND_DEF.'
         rdabtabs=.FALSE.
         return
      end if
      n_btype=n_btype+1
      pos=1
      call token(cmd,250,pos,0,' ',start,end,term)
      bmnem(n_btype)=cmd(start:end)
      call token(cmd,250,pos,0,' ',start,end,term)
      fld=cmd(start:end)
      read(fld,'(bn,i20)',end=8101,err=8101) bnd_id(n_btype)
      call token(cmd,250,pos,0,' ',start,end,term)
      call token(cmd,250,pos,0,' ',start,end,term)
      chem_bnd(n_btype)=cmd(start:start).eq.'Y'
      go to 110
c     
 190  continue
      close(unit=22)
c     
c     Read oop parameters...
c     
      l=str_trim(cmd,ff_tab_dir)
      cmd(l+1:)='TAFF_OOP_BEND.'
      inquire(file=cmd,exist=iqstat)
      if (.not.iqstat) cmd(l+1:)='TAFF_OOP_BEND'
      open(unit=22,file=cmd,status='old',err=8200 READONLY)
c     
 210  continue
      read(22,'(a)',end=290) cmd
      if (cmd(1:1).eq.'#') goto 210
      pos=1
      call token(cmd,250,pos,0,' ',start,end,term)
      oopmnem=cmd(start:end)
      do l=1,n_atype
         if (amnem(l).eq.oopmnem) then
            call token(cmd,250,pos,0,' ',start,end,term)
            fld=cmd(start:end)
            read(fld,'(bn,f20.0)',end=8201,err=8201) koop(l)
            oopcent(l)=.TRUE.
            go to 210
         end if
      end do
      write(6,*) 'Bad atom type mnemonic in TAFF_OOP_BEND : ',oopmnem
      rdabtabs=.FALSE.
      return
c     
 290  continue
      close(unit=22)
      rdabtabs=.TRUE.
      return
c     
 8000 continue
      write(6,*) 'Error: could not open ATOM_DEF file.'
      rdabtabs=.FALSE.
      return
 8001 continue
      write(6,*) 'Error: format in line ',n_atype,' of ATOM_DEF.'
      rdabtabs=.FALSE.
      return
 8100 continue
      write(6,*) 'Error: could not open BOND_DEF file.'
      rdabtabs=.FALSE.
      return
 8101 continue
      write(6,*) 'Error: format in line ',n_btype,' of BOND_DEF.'
      rdabtabs=.FALSE.
      return
 8200 continue
      write(6,*) 'Error: could not open TAFF_OOP_BEND.'
      rdabtabs=.FALSE.
      return
 8201 continue
      write(6,*) 'Error: format error in TAFF_OOP_BEND for ',oopmnem
      rdabtabs=.FALSE.
      return
      end     
