      logical function readff_charmm()
c     
c     Read force field parameters from FF tables, CHARMM version
c     
      implicit none
c     
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
#include "option.h"
#include "worksp.h"
#if defined(sgi)
#define READONLY ,readonly
#else 
#define READONLY
#endif
c     
c     
      integer i,l,status,k,mode,nundef,impptr
      character*132 rec
      integer str_trim,charmm_dihprm,charmm_impprm,charmm_nboprm,
     .     charmm_bndprm,charmm_angprm,charmm_nboopt,charmm_nbfprm,
     .     str_len
      logical iqstat
c     
      readff_charmm=.FALSE.
c     
      l=index(prm_file,'/')
      if (l.gt.0) then
         rec=prm_file
         open(unit=22,file=rec,status='old',err=8000 READONLY)
         prm_file=rec
      else
         l=str_len(ff_tab_dir)
         rec=ff_tab_dir(1:l)//prm_file
         inquire(file=rec,exist=iqstat)
         if (.not.iqstat) rec=prm_file
         open(unit=22,file=rec,status='old',err=8000 READONLY)
         prm_file=rec
      end if
c     
      status=0
      impptr=1+ntor
      do i=1,ntor+npla
         iwork(i)=-1
      end do
c     
      do while (status.eq.0) 
         call charmm_getprm(22,rec,k,*7000,*8010)
         if (k.gt.0) then
            mode=k
            if (mode.eq.5) status=charmm_nboopt(rec)
         else if (mode.gt.0) then
            if (mode.eq.1) status=charmm_bndprm(rec)
            if (mode.eq.2) status=charmm_angprm(rec)
            if (mode.eq.3) status=charmm_dihprm(rec,iwork)
            if (mode.eq.4) status=charmm_impprm(rec,iwork(impptr))
            if (mode.eq.5) status=charmm_nboprm(rec)
            if (mode.eq.6) status=charmm_nbfprm(rec)
         end if
      end do
      if (status.ne.0) goto 9000
c     
c     check if all parameters have been found...
c     
 7000 continue
      nundef=0
c     Bonds...
      do i=1,nbnd
         if (bndl0(i).lt.0.0) then
            write(6,'(a,a,i5,a,i5,a)') 
     .           ' Missing bond stretch parameter, ',
     .           amnem(ntype(bnd(1,i)))//'(',bnd(1,i),
     .           ')   '//amnem(ntype(bnd(2,i)))//'(',bnd(2,i),')'
            nundef=nundef+1
         end if
      end do
c     Angles...
      do i=1,nang
         if (angl0(i).lt.0.0) then
            write(6,'(a,a,i5,a,i5,a,i5,a)') 
     .           ' Missing angle bend parameter, ',
     .           amnem(ntype(ang(1,i)))//'(',ang(1,i),
     .           ')   '//amnem(ntype(ang(2,i)))//'(',ang(2,i),
     .           ')   '//amnem(ntype(ang(3,i)))//'(',ang(3,i),')'
            nundef=nundef+1
         end if
      end do
C     Dihedrals...
      do i=1,ntor
         tor0(i)=-360
         if (ntortrm(i).eq.0) then
            write(6,'(a,a,i5,a,i5,a,i5,a,i5,a)') 
     .           ' Missing torsion parameter, ',
     .           amnem(ntype(tor(1,i)))//'(',tor(1,i),
     .           ')   '//amnem(ntype(tor(2,i)))//'(',tor(2,i),
     .           ')   '//amnem(ntype(tor(3,i)))//'(',tor(3,i),
     .           ')   '//amnem(ntype(tor(4,i)))//'(',tor(4,i),')'
            nundef=nundef+1
         end if
      end do
C     Impropers...
      do i=1,npla
         if (kpla(i).lt.0) then
            write(6,'(a,a,i5,a,i5,a,i5,a,i5,a)') 
     .           ' Missing improper torsion parameter, ',
     .           amnem(ntype(pla(1,i)))//'(',pla(1,i),
     .           ')   '//amnem(ntype(pla(2,i)))//'(',pla(2,i),
     .           ')   '//amnem(ntype(pla(3,i)))//'(',pla(3,i),
     .           ')   '//amnem(ntype(pla(4,i)))//'(',pla(4,i),')'
            nundef=nundef+1
         end if
      end do
C     Van der Waals...
      do i=1,n_atype
         if (rvdw(i).lt.0) then
            write(6,'(a,a,i5,a,i5,a)') 
     .           ' Missing vdW parameter, ',amnem(i)
            nundef=nundef+1
         end if
      end do
c     
      readff_charmm=nundef.eq.0
      goto 9000
c     
c     Errors:
c     
 8000 continue
      l=str_len(rec)
      write(6,'(a)') ' Could not open parameter file '
     .     //rec(1:l)//' for read.'
      return
c     
 8010 continue
      write(6,'(a)') ' Error reading FF parameter file.'
      goto 9000
c     
 9000 continue
      close(unit=22)
      return
      end
c------------------------------------------------------------------------

      subroutine charmm_getprm(u,rec,k,*,*)
c     
      implicit none
      integer u
      character rec*(*)
      integer l,f,t,k,pos
      character*1 term
      character*4 key(12),test
      integer nk(12)
      data key / 'BOND','THET','ANGL','PHI ','DIHE','IMPR','IMPH',
     .           'NONB','NBON','NBFI','HBON','END ' /
      data nk /     1  ,  2   ,  2   ,  3   ,  3   ,  4   ,  4   ,
     .              5  ,  5   ,  6   , 99   , 99 /
c     
      k=0
      rec=' '
      do while (rec.eq.' ')
         read(22,'(a)',end=8001,err=8002) rec
         if (rec(1:1).eq.'*') then 
            rec=' '
         else
            l=index(rec,'!')
            if (l.gt.0) rec(l:)=' '
         end if
      end do
      pos=1
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (t.gt.f+3) t=f+3
      test=rec(f:t)
      call str_upcase(test,test)
      do l=1,11
         if (test.eq.key(l)) k=nk(l)
      end do
      return
c     
 8001 return 1
 8002 return 2
      end
c------------------------------------------------------------------------

      integer function charmm_bndprm(rec)
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
      character*(*) rec
c     
      integer i,pos,f,t,tp1,tp2,num_atype,str_len,l
      character term*1
      real k,b
c     
      pos=1
      call token(rec,len(rec),pos,0,' ',f,t,term)
      tp1=num_atype(rec(f:t))
      if (tp1.gt.0) then
         call token(rec,len(rec),pos,0,' ',f,t,term)
         tp2=num_atype(rec(f:t))
         if (tp2.gt.0) then
            call token(rec,len(rec),pos,0,' ',f,t,term)
            read(rec(f:t),*,end=8001,err=8001) k
            call token(rec,len(rec),pos,0,' ',f,t,term)
            read(rec(f:t),*,end=8001,err=8001) b
            do i=1,nbnd
               if ((tp1.eq.ntype(bnd(1,i)).and.
     .              tp2.eq.ntype(bnd(2,i))).or.
     .              (tp1.eq.ntype(bnd(2,i)).and.
     .              tp2.eq.ntype(bnd(1,i)))) then
                  bndl0(i)=b
                  kbnd(i)=2*k   ! PRM file has k/2
               end if
            end do
         end if
      end if
      charmm_bndprm=0
      return
c     
 8001 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR , BOND record with bad format:',
     .     rec(1:l)
      charmm_bndprm=1
      return
      end
c------------------------------------------------------------------------

      integer function charmm_angprm(rec)
c     
      implicit none
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
      character*(*) rec
c     
      integer i,l,pos,f,t,tp1,tp2,tp3,num_atype
      integer str_len
      character term*1
      real k,a
c     
      pos=1
      call token(rec,len(rec),pos,0,' ',f,t,term)
      tp1=num_atype(rec(f:t))
      if (tp1.gt.0) then
         call token(rec,len(rec),pos,0,' ',f,t,term)
         tp2=num_atype(rec(f:t))
         if (tp2.gt.0) then
            call token(rec,len(rec),pos,0,' ',f,t,term)
            tp3=num_atype(rec(f:t))
            if (tp3.gt.0) then
               call token(rec,len(rec),pos,0,' ',f,t,term)
               read(rec(f:t),*,end=8001,err=8001) k
               k=k*DEG2RAD*DEG2RAD ! File: Kcal/Mol*rad*rad
               call token(rec,len(rec),pos,0,' ',f,t,term)
               read(rec(f:t),*,end=8001,err=8001) a
               do i=1,nang
                  if (angl0(i).lt.0.and.
     .                 tp2.eq.ntype(ang(2,i)).and.
     .                 ((tp1.eq.ntype(ang(1,i)).and.
     .                 tp3.eq.ntype(ang(3,i))).or.
     .                 (tp1.eq.ntype(ang(3,i)).and.
     .                 tp3.eq.ntype(ang(1,i))))) then
                     angl0(i)=a
                     kang(i)=2*k ! PRM files has k/2
                  end if
               end do
            end if
         end if
      end if
c     
      charmm_angprm=0
      return
c     
 8001 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR , ANGLE record with bad format:',
     .     rec(1:l)
      charmm_angprm=1
      return
      end
c------------------------------------------------------------------------

      integer function charmm_impmat(t,p)
c     
      implicit none
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
      integer t(4),p(4),lev,islev(16),i
      integer typeq
c     
      data islev / 6,0,0,0,0,0,5,0,4,3,0,0,2,0,0,0 /
c     
      charmm_impmat=-1
      lev=0
      do i=1,4
         lev=2*lev
         if (t(i).eq.0) lev=lev+1
      end do
      lev=islev(lev+1)
      if (typeq(t(1),ntype(p(1)))+
     .     typeq(t(2),ntype(p(2)))+
     .     typeq(t(3),ntype(p(3)))+
     .     typeq(t(4),ntype(p(4))).gt.0) then
         charmm_impmat=lev
      else if (typeq(t(4),ntype(p(1)))+
     .        typeq(t(3),ntype(p(2)))+
     .        typeq(t(2),ntype(p(3)))+
     .        typeq(t(1),ntype(p(4))).gt.0) then
         charmm_impmat=lev
      end if
      return
      end
c------------------------------------------------------------------------

      integer function charmm_impprm(rec,implev)
c     
      implicit none
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
      character*(*) rec
      integer implev(*)
c     
      integer i,k,l,pos,f,t,tp(4),per,lev
      integer str_len,charmm_impmat,num_atype
      character term*1
      real*8 fc,shift
c     
      charmm_impprm=0
      pos=1
      do i=1,4
         call token(rec,len(rec),pos,0,' ',f,t,term)
         tp(i)=num_atype(rec(f:t))
         if (tp(i).lt.0) go to 7000
      end do
      call token(rec,len(rec),pos,0,' ',f,t,term)
      read(rec(f:t),*,end=8001,err=8001) fc
      call token(rec,len(rec),pos,0,' ',f,t,term)
      read(rec(f:t),*,end=8001,err=8001) per
      call token(rec,len(rec),pos,0,' ',f,t,term)
      read(rec(f:t),*,end=8001,err=8001) shift
c     
c     Search matching energy terms in energy-common
c     
c     Rules: 
c     - wild-card precedence determined in charmm_impmat
c     - multiple definitions: first definition will be used
c     - only terms with zero periodicity are supported
c     
      do i=1,npla
         lev=charmm_impmat(tp,pla(1,i))
         if (lev.gt.implev(i)) then
            if (per.ne.0) goto 8101
            implev(i)=lev
            pla0(i)=shift*DEG2RAD
            kpla(i)=fc
         end if
      end do
c     
 7000 continue
      return
c     
 8001 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR, IMPROPER record with bad format:',
     .     rec(1:l)
      charmm_impprm=1
      return
 8101 continue
      write(6,'(a/a)') 
     .     '*** ERROR: IMPROPER terms with per>0 not supported: ',
     .     rec(1:50)
      charmm_impprm=1
      return
      end
c------------------------------------------------------------------------

      integer function charmm_dihmat(tp,tor)
c     
      implicit none
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
      integer tp(4),tor(4),lev
c     
      lev=-1
      if (tp(1).eq.0.and.tp(4).eq.0) then
         if ((tp(2).eq.ntype(tor(2)).and.tp(3).eq.ntype(tor(3))).or.
     .        (tp(2).eq.ntype(tor(3)).and.tp(3).eq.ntype(tor(2)))) then
            lev=2
         end if
      else
         if (tp(1).eq.ntype(tor(1)).and.tp(2).eq.ntype(tor(2)).and.
     .        tp(3).eq.ntype(tor(3)).and.tp(4).eq.ntype(tor(4))) lev=4
         if (tp(1).eq.ntype(tor(4)).and.tp(2).eq.ntype(tor(3)).and.
     .        tp(3).eq.ntype(tor(2)).and.tp(4).eq.ntype(tor(1))) lev=4
         
      end if
      charmm_dihmat=lev
c     
      return
      end
c------------------------------------------------------------------------

      integer function charmm_dihprm(rec,torlev)
c     
      implicit none
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
      character*(*) rec
      integer torlev(*)
c     
      integer i,k,l,pos,f,t,tp(4),per,lev,nf,bt
      integer str_len,num_atype,typeq,charmm_dihmat
      character term*1
      real*8 fc,shift
c     
      charmm_dihprm=0
      bt=1
      pos=1
      do i=1,4
         call token(rec,len(rec),pos,0,' ',f,t,term)
         tp(i)=num_atype(rec(f:t))
         if (tp(i).lt.0) go to 7000
      end do
      call token(rec,len(rec),pos,0,' ',f,t,term)
      read(rec(f:t),*,end=8001,err=8001) fc
      call token(rec,len(rec),pos,0,' ',f,t,term)
      read(rec(f:t),*,end=8001,err=8001) per
      call token(rec,len(rec),pos,0,' ',f,t,term)
      read(rec(f:t),*,end=8001,err=8001) shift
c     
c     Search matching energy terms in energy-common
c     
c     Rules: 
c     - only terms of type A-B-C-D and X-B-C-X are allowd
c     - terms of type A-B-C-D have precedence over X-B-C-X
c     - multiple Fourier terms: all terms must be of the same type
c     - multiple definitions: last definition is used.
c     - zero periodicity terms are constant (different from CHARMM)
c     
      if ((tp(1).eq.0.and.tp(4).ne.0).or.
     .     (tp(4).eq.0.and.tp(1).ne.0).or.
     .     tp(2).eq.0.or.tp(3).eq.0) goto 8306
      nf=0
      do i=1,ntor
         if (bt.eq.tort(i)) then
            lev=charmm_dihmat(tp,tor(1,i))
            if (lev.gt.0) then
               if (ntortrm(i).eq.0) torlev(i)=0
               if (lev.gt.torlev(i)) ntortrm(i)=0
               k=0
               do l=1,ntortrm(i)
                  if (kzf(tortp(l,i)).eq.per) k=l
               end do
               if (k.eq.0) then
                  if (ntortrm(i).ge.mxft) goto 8305
                  ntortrm(i)=ntortrm(i)+1
                  k=ntortrm(i)
               end if
               if (lev.ge.torlev(i)) then
                  if (nf.eq.0) then
                     ntortp=ntortp+1
                     if (ntortp.gt.mtotyp) goto 8304
                     kzt(ntortp)=fc
                     kzs(ntortp)=shift*DEG2RAD
                     kzf(ntortp)=per
                  end if
                  nf=nf+1
                  torlev(i)=lev
                  tortp(k,i)=ntortp
               end if
            end if
         end if
      end do
c     
 7000 continue
      return
c     
 8001 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR, PRM record with bad format:',
     .     rec(1:l)
      charmm_dihprm=1
      return
 8304 continue
      write(6,'(a)') ' Too many torsion parameter types,'//
     .     ' increase program dimensions.'
      charmm_dihprm=1
      return
 8305 continue
      write(6,'(a,i2)') ' Too many fourier terms (torsion), '//
     .     'max. allowed is ',mxft
      write(6,'(a)') ' Unknown atomtype in TAFF_VDW :'
      charmm_dihprm=1
      return
 8306 continue
      write(6,'(a)') '*** ERROR: invalid dihedral term: '//rec(1:45)
      charmm_dihprm=1
      return
      end
c------------------------------------------------------------------------

      integer function charmm_nboopt(rec)
c     
c     Function that could be used to parse NBOND options.
c     Current version of MINIMAX ignores options in parametr
c     files, uses command line options instead.
c     
      implicit none
      character*(*) rec
      integer i
c     
c     simply handle continuation lines...
c     
      charmm_nboopt=1
      do while (charmm_nboopt.ne.0)
         i=len(rec)
         do while(rec(i:i).eq.' ')
            i=i-1
         end do
         if (rec(i:i).eq.'-') then
            call charmm_getprm(22,rec,i,*8001,*8001)
            if (i.gt.0) goto 8001
         else
            charmm_nboopt=0
         end if
      end do
      return
c     
 8001 continue
      write(6,'(a)') '*** ERROR parsing NBOND options.'
      charmm_nboopt=1
      return
      end

c------------------------------------------------------------------------

      integer function charmm_nboprm(rec)
c     
c     Parse NBOND parameter definition.
c     <type> <polarizability> <emin> <rmin> [ <plo14> [ <e14min> [ <r14min> ] ] ]
c     If 1-4 paramters are missing they default to the normal non-bonded
c     parameters. <type> may contain X-Plor style wildcards.
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
      character rec*(*)
c
      real*8 alpha,emin,rmin,alpha14,emin14,rmin14,e,r,
     .     alp,alp14
      integer pos,f,t,str_len,l,i,j
      character*1 term
      character*20 wldtype
      logical wldmatch
c     
c     parse parameter record
c
      pos=1
      call token(rec,len(rec),pos,0,' ',f,t,term)
      wldtype=rec(f:t)
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) then
         read(rec(f:t),*,end=8001,err=8001) alp
      else
         goto 8001
      end if
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) then
         read(rec(f:t),*,end=8001,err=8001) emin
      else
         goto 8001
      end if
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) then
         read(rec(f:t),*,end=8001,err=8001) rmin
      else
         goto 8001
      end if
      alp14=alp
      emin14=emin
      if (rmin.lt.0.000001) rmin=0.000001
      rmin14=rmin
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) read(rec(f:t),*,end=8001,err=8001) alp14
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) read(rec(f:t),*,end=8001,err=8001) emin14
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) read(rec(f:t),*,end=8001,err=8001) rmin14
      if (rmin14.lt.0.000001) rmin14=0.000001
c
c     store parameters, if applicable
c
      do i=1,n_atype
         if (wldmatch(amnem(i),wldtype)) then
            rvdw(i)=rmin
            epsil(i)=-emin
            apol(i)=max(0.0D0,alp)
            rvdw14(i)=rmin14
            epsil14(i)=-emin14
            apol14(i)=max(0.0D0,alp14)
            do j=1,n_atype
               if (rvdw(j).gt.0) call charmm_storeLJ(i,j)
            end do
         end if
      end do
c
c     done, normal return
c
      charmm_nboprm=0
      return
c
 8001 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR, NBOND record with bad format:',
     .     rec(1:l)
      charmm_nboprm=1
      return
c
      end
c------------------------------------------------------------------------

      integer function charmm_nbfprm(rec)
c     
c     Parse NBFIX parameter definition.
c     <type1> <type2> <emin> <r1min+r2min> [ [ <e14min> [ <r14min> ] ]
c     If 1-4 paramters are missing they default to the normal non-bonded
c     parameters. <typei> may contain X-Plor style wildcards.
c     
      implicit none
#include "sizes.h"
#include "apara.h"
      character rec*(*)
      integer pos,f,t,str_len,l,i,j,nf
      character*20 wld1,wld2
      real*8 emin,rmin,emin14,rmin14
      character*1 term
      logical wldmatch
c
c     Parse parameter record
c
      pos=1
      call token(rec,len(rec),pos,0,' ',f,t,term)
      wld1=rec(f:t)
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) then
         wld2=rec(f:t)
      else
         goto 8001
      end if
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) then
         read(rec(f:t),*,end=8001,err=8001) emin
      else
         goto 8001
      end if
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) then
         read(rec(f:t),*,end=8001,err=8001) rmin
      else
         goto 8001
      end if
      emin14=emin
      rmin14=rmin
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) read(rec(f:t),*,end=8001,err=8001) emin14
      call token(rec,len(rec),pos,0,' ',f,t,term)
      if (f.le.t) read(rec(f:t),*,end=8001,err=8001) rmin14
c
c     Store parameters, if applicable
c
      nf=0
      i=0
      do while (i.le.n_atype)
         i=i+1
         if (wldmatch(amnem(i),wld1)) then
            j=0
            do while (j.le.n_atype)
               j=j+1
               if (wldmatch(amnem(j),wld2)) then 
                  if (emin.gt.0) goto 8002
                  vaij(i,j)=-emin*rmin**12
                  vbij(i,j)=2*emin*rmin**6
                  vaij(j,i)=vaij(i,j)
                  vbij(j,i)=vbij(i,j)
                  if (emin14.gt.0) goto 8002
                  vaij14(i,j)=-emin14*rmin14**12
                  vbij14(i,j)=2*emin14*rmin14**6
                  vaij14(j,i)=vaij14(i,j)
                  vbij14(j,i)=vbij14(i,j)
                  nf=nf+1
               end if
            end do
            if (nf.eq.0) i=n_atype   ! to break out of outer loop
         end if
      end do
      charmm_nbfprm=0
      return
c
 8001 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR, NBFIX record with bad format:',
     .     rec(1:l)
      charmm_nbfprm=1
      return
 8002 continue
      l=str_len(rec)
      write(6,'(a/a)') 
     .     '*** ERROR, NBFIX - positive emin not supported:',
     .     rec(1:l)
      charmm_nbfprm=1
      return
c
      end
c------------------------------------------------------------------------
      subroutine charmm_storeLJ(i,j)
c     
      implicit none
#include "sizes.h"
#include "apara.h"
      integer i,j
c
      real*8 c,r6,a,b,epsi,epsj,alpi,alpj,ri,rj
      integer k
c
      data c /362.3461/
c
c     Convert (r,eps,alp)(i,j) to Lenard-Jones (A,B) and store result
c     in vaij/vbij. If eps is negative, -eps is used as the effective
c     number of electrons, and the Slater-Kirkwood formula is used
c     to determine the 'real' eps.
c
c     NOTE: Evdw = A/r**12 + B/r**6  ---> B < 0 !

      do k=1,2
         if (k.eq.1) then      ! fetch regular parameters
            epsi=epsil(i)
            epsj=epsil(j)
            alpi=apol(i)
            alpj=apol(j)
            ri=rvdw(i)
            rj=rvdw(j)
         else                  ! fetch regular parameters
            epsi=epsil14(i)
            epsj=epsil14(j)
            alpi=apol14(i)
            alpj=apol14(j)
            ri=rvdw14(i)
            rj=rvdw14(j)
         end if
         r6=max(0.000001,(ri+rj)**6)
         if (epsi.gt.0.and.epsj.lt.0) then
            epsj=0.25*c*alpj*sqrt(-alpj*epsj)/(2*rj)**6
         else if (epsi.lt.0.and.epsj.gt.0) then
            epsi=0.25*c*alpi*sqrt(-alpi*epsi)/(2*ri)**6
         end if
         if (epsi.gt.0.0.and.epsj.gt.0.0) then   ! convert from (eps,r)
            b=2*sqrt(epsi*epsj)*r6
         else if (epsi.lt.0.and.epsj.lt.0) then  ! use Slater-Kirwood
            b=c*alpi*alpj/(sqrt(-alpi/epsi)+sqrt(-alpj/epsj))
         else
            b=0
         end if
         if (k.eq.1) then
            vaij(i,j)=0.5*b*r6
            vbij(i,j)=-b   ! -b, since E=A/r**12+B/r**6
         else
            vaij14(i,j)=0.5*b*r6
            vbij14(i,j)=-b ! -b, since E=A/r**12+B/r**6
         end if
      end do
c
      vaij(j,i)=vaij(i,j)
      vbij(j,i)=vbij(i,j)
      vaij14(j,i)=vaij14(i,j)
      vbij14(j,i)=vbij14(i,j)
c
      return
      end
