      logical function readff_taff()
c     
c     Read force field parameters from FF tables (TAFF version)
c     
      implicit none
c     
#include "sizes.h"
#include "constants.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
#include "option.h"
#include "worksp.h"
#include "renumber.h"
#if defined(sgi)
#define READONLY ,readonly
#else 
#define READONLY
#endif
c     
      integer bndlev(mbtot),anglev(mantot),torlev(mtotot),lev
      integer i,j,k,l,nf,pos,f,t,tp(4),bt,per,sig
      integer num_atype,num_btype,nundef,typeq
      integer str_trim
      real*8  fc,b0,a0
      character*250 rec
      character*20 fld
      character*1 term
      logical iqstat
c     
      equivalence (iwork,bndlev),(iwork,anglev),(iwork,torlev)
c     
      readff_taff=.TRUE.
      nundef=0
c     
c     Bondstretch first...
c     
      l=str_trim(rec,ff_tab_dir)
      rec(l+1:)='TAFF_BOND_STRETCH.'
      inquire(file=rec,exist=iqstat)
      if (.not.iqstat) rec(l+1:)='TAFF_BOND_STRETCH'
      open(unit=22,file=rec,status='old',err=8000 READONLY)
c     
      do i=1,nbnd
         bndlev(i)=-1
      end do
c     
c     Read and decode next parameter definition...
c     
      f=0
      t=0
      j=0
      bt=0
 10   continue
      read(22,'(a)',end=90) rec
      if (rec(1:1).eq.'#') go to 10
      pos=1
      call token(rec,250,pos,0,' ',f,t,term)                  
      tp(1)=num_atype(rec(f:t))
      call token(rec,250,pos,0,' ',f,t,term)
      tp(2)=num_atype(rec(f:t))
      if (tp(1).lt.0.or.tp(2).lt.0) go to 8002
      call token(rec,250,pos,0,' ',f,t,term)
      bt=num_btype(rec(f:t))
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8001,err=8001) b0
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8001,err=8001) fc
c     
c     Search matching energy terms in energy-common
c     
      do i=1,nbnd
         if (bndlev(i).lt.200.and.bndt(i).eq.bt) then
            lev=max(typeq(tp(1),ntype(bnd(1,i)))+
     .           typeq(tp(2),ntype(bnd(2,i))),
     .           typeq(tp(2),ntype(bnd(1,i)))+
     .           typeq(tp(1),ntype(bnd(2,i))))
            if (lev.gt.bndlev(i)) then
               bndlev(i)=lev
               bndl0(i)=b0
               kbnd(i)=fc
            end if
         end if
      end do
c     
      go to 10
c     
c     done with bond stretch...
c     
 90   continue
      close(unit=22)
      do i=1,nbnd
         if (bndlev(i).lt.0) then
            nundef=nundef+1
            write(6,'(2a,2i5)') ' Missing bond stretch parameter,',
     .           ' atoms ',old(bnd(1,i)),old(bnd(2,i))
         end if
      end do
c     
c     And now for angle bending...
c     
      l=str_trim(rec,ff_tab_dir)
      rec(l+1:)='TAFF_ANGLE_BEND.'
      inquire(file=rec,exist=iqstat)
      if (.not.iqstat) rec(l+1:)='TAFF_ANGLE_BEND'
      open(unit=22,file=rec,status='old',err=8100 READONLY)
c     
      do i=1,nang
         anglev(i)=-1
      end do
c     
c     Read and decode next parameter definition
c     
 110  continue
      read(22,'(a)',end=190) rec
      if (rec(1:1).eq.'#') go to 110
      pos=1
      do i=1,3
         call token(rec,250,pos,0,' ',f,t,term)
         tp(i)=num_atype(rec(f:t))
      end do
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8101,err=8101) a0
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8101,err=8101) fc
c     
c     Search matching energy terms in energy-common
c     
      do i=1,nang
         if (anglev(i).lt.300) then
            lev=max(typeq(tp(1),ntype(ang(1,i)))+
     .           typeq(tp(2),ntype(ang(2,i)))+
     .           typeq(tp(3),ntype(ang(3,i))),
     .           typeq(tp(3),ntype(ang(1,i)))+
     .           typeq(tp(2),ntype(ang(2,i)))+
     .           typeq(tp(1),ntype(ang(3,i))))
            if (lev.gt.anglev(i)) then
               anglev(i)=lev
               angl0(i)=a0
               kang(i)=fc
            end if
         end if
      end do
c     
      go to 110
c     
c     done with angle bend...
c     
 190  continue
      do i=1,nang
         if (anglev(i).lt.0) then
            nundef=nundef+1
            write(6,'(2a,3i5)') ' Missing angle bending parameter, ',
     .           'atoms ',old(ang(1,i)),old(ang(2,i)),old(ang(3,i))
         end if
      end do
      close(unit=22)
c     
c     and now for the torsion potentials...
c     
      l=str_trim(rec,ff_tab_dir)
      rec(l+1:)='TAFF_TORS.'
      inquire(file=rec,exist=iqstat)
      if (.not.iqstat) rec(l+1:)='TAFF_TORS'
      open(unit=22,file=rec,status='old',err=8200 READONLY)
c     
      ntortp=0
      do i=1,ntor
         ntortrm(i)=0
         torlev(i)=-1
      end do
c     
c     Read and decode next parameter definition
c     
 210  continue
      read(22,'(a)',end=290) rec
      if (rec(1:1).eq.'#') go to 210
      pos=1
      do i=1,4
         call token(rec,250,pos,0,' ',f,t,term)
         tp(i)=num_atype(rec(f:t))
      end do
      call token(rec,250,pos,0,' ',f,t,term)
      bt=num_btype(rec(f:t))
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8201,err=8201) fc
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(i20)',end=8201,err=8201) per
      sig=sign(1,per)
      per=abs(per)
c     
c     Search matching energy terms in energy-common
c     
      nf=0
      do i=1,ntor
         if (bt.eq.tort(i)) then
            lev=max(typeq(tp(1),ntype(tor(1,i)))+
     .           typeq(tp(2),ntype(tor(2,i)))+
     .           typeq(tp(3),ntype(tor(3,i)))+
     .           typeq(tp(4),ntype(tor(4,i))),
     .           typeq(tp(4),ntype(tor(1,i)))+
     .           typeq(tp(3),ntype(tor(2,i)))+
     .           typeq(tp(2),ntype(tor(3,i)))+
     .           typeq(tp(1),ntype(tor(4,i))))
            if (lev.gt.0) then
               if (lev.ge.torlev(i)) then
                  if (lev.gt.torlev(i)) ntortrm(i)=0  ! forget all lower level parameters
                  k=0
                  do l=1,ntortrm(i)
                     if (kzf(tortp(l,i)).eq.per) k=l
                  end do
                  if (k.eq.0) then
                     if (ntortrm(i).ge.mxft) goto 8305
                     ntortrm(i)=ntortrm(i)+1
                     k=ntortrm(i)
                     torlev(i)=0
                  end if
                  if (lev.gt.torlev(i)) then
                     if (nf.eq.0) then
                        ntortp=ntortp+1
                        if (ntortp.gt.mtotyp) goto 8304
                        kzt(ntortp)=fc/2
                        kzs(ntortp)=0
                        if (sig.lt.0) kzs(ntortp)=PI
                        kzf(ntortp)=per
                     end if
                     nf=nf+1
                     torlev(i)=lev
                     tortp(k,i)=ntortp
                  end if
               end if
            end if
         end if
      end do
c     
      go to 210
c     
c     done with torsions
c     
 290  continue
      close(unit=22)
      do i=1,ntor
         if (ntortrm(i).eq.0) then
            nundef=nundef+1
            write(6,'(2a,4i5)') ' Missing torsion potential, ',
     .           'atoms ',(old(tor(j,i)),j=1,4)
         end if
      end do
c     
c     Van der Waals...
c     
      l=str_trim(rec,ff_tab_dir)
      rec(l+1:)='TAFF_VDW.'
      inquire(file=rec,exist=iqstat)
      if (.not.iqstat) rec(l+1:)='TAFF_VDW'
      open(unit=22,file=rec,status='old',err=8300 READONLY)
c     
 310  continue
      read(22,'(a)',end=390) rec
      if (rec(1:1).eq.'#') go to 310
      pos=1
      call token(rec,250,pos,0,' ',f,t,term)
      j=num_atype(rec(f:t))
      if (j.le.0) go to 8302
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8301,err=8301) rvdw(j)
      call token(rec,250,pos,0,' ',f,t,term)
      fld=rec(f:t)
      read(fld,'(f20.0)',end=8301,err=8301) epsil(j)
      if (rvdw(j).eq.0) then
         epsil(j)=0.0
         rvdw(j)=0.01
      end if
      go to 310
c     
c     Done with vdw...
c     
 390  continue
      close(unit=22)
      do j=1,n_atype
         do i=j,n_atype
            vaij(i,j)=sqrt(epsil(i)*epsil(j))*(rvdw(i)+rvdw(j))**12
            vbij(i,j)=-2*sqrt(epsil(i)*epsil(j))*(rvdw(i)+rvdw(j))**6
         end do
      end do
      do j=2,n_atype
         do i=1,j-1
            vaij(i,j)=vaij(j,i)
            vbij(i,j)=vbij(j,i)
         end do
      end do
c     
c     Out of plane bending potentials...
c     
      do i=1,npla
         kpla(i)=koop(ntype(pla(1,i)))
      end do
c     
c     Done...
c     
      if (nundef.gt.0) go to 8900
c     
      readff_taff=.TRUE.
      R E T U R N
c     
c     Errors...
c     
 8000 continue
      write(6,'(a)') ' Could not open TAFF_BOND_STRETCH.'
      readff_taff=.FALSE.
      return
 8001 continue
      write(6,'(a)') ' Format error in TAFF_BOND_STRETCH : '
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8002 continue
      write(6,'(a)') ' Unknown atomtype in TAFF_BOND_STRETCH :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8100 continue
      write(6,'(a)') ' Could not open TAFF_ANGLE_BEND.'
      readff_taff=.FALSE.
      return
 8101 continue
      write(6,'(a)') ' Format error in TAFF_ANGLE_BEND :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8102 continue
      write(6,'(a)') ' Unknown atomtype in TAFF_ANGLE_BEND :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8200 continue
      write(6,'(a)') ' Could not open TAFF_TORS.'
      readff_taff=.FALSE.
      return
 8201 continue
      write(6,'(a)') ' Format error in TAFF_TORS :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8202 continue
      write(6,'(a)') ' Unknown atomtype in TAFF_TORS :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8300 continue
      write(6,'(a)') ' Could not open TAFF_VDW.'
      readff_taff=.FALSE.
      return
 8301 continue
      write(6,'(a)') ' Format error in TAFF_VDW :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8302 continue
      write(6,'(a)') ' Unknown atomtype in TAFF_VDW :'
      write(6,'(2a)') ' ',rec(1:75)
      readff_taff=.FALSE.
      return
 8304 continue
      write(6,'(a)') ' Too many torsion parameter types,'//
     .     ' increase program dimensions.'
      readff_taff=.FALSE.
      return
 8305 continue
      write(6,'(a,i2)') ' Too many fourier terms (torsion), '//
     .     'max. allowed is ',mxft
      write(6,'(a)') ' Unknown atomtype in TAFF_VDW :'
      readff_taff=.FALSE.
      return
 8900 continue
      write(6,'(a)') ' *** ERROR : missing force field parameters.'
      readff_taff=.FALSE.
      return
c     
      end
