      logical function topol_amber()
c     
c     Setup of energy function, AMBER version.
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
#include "worksp.h"
#include "option.h"
#include "constants.h"
#if defined(sgi)
#define READONLY ,readonly
#else 
#define READONLY
#endif
c     
c  13.12.2006: added support for 'new' (V00001.000) file format.
c
      integer MXPAR,MXVXPAR
      parameter (MXPAR=500)
      parameter (MXVXPAR=((MATYPE+1)*MATYPE)/2)
c
      character rec*250,sect*20,fmt*250
      real*8 dummy
      real*8 RK(MXPAR),REQ(MXPAR),TK(MXPAR),TEQ(MXPAR),
     .     VX(MXVXPAR)
      integer i,j,l,M,n
      integer
     .     NATOM,NTYPES,NBONH,MBONA,NTHETH,MTHETA,
     .     NPHIH,MPHIA,NHPARM,NPARM,NNB,ANRES,
     .     NBONA,NTHETA,NPHIA,NUMBND,NUMANG,NPTRA,
     .     NATYP,NPHB,IFPERT,NBPER,NGPER,NDPER,
     .     MBPER,MGPER,MDPER,IFBOX,NMXRS,IFCAP
      integer ICO(matype,matype),angt(mantot)
      integer str_trim,insert_excl
      logical newfmt,chkflag
c
      equivalence (iwork,angt)
c
      topol_amber=.FALSE. ! assume guilty until proven innocent...
      newfmt=.FALSE.      ! assume old file format.
c
c     open the parameter/topology file
c
      open(unit=22,file=prm_file,status='old',err=8000 READONLY)
c
c     read (and ignore) title line
c
      sect='title'
      read(22,'(a)',end=8001,err=8001) rec
      call str_upcase(rec,rec)
      newfmt=rec(1:8).eq.'%VERSION'
      if (newfmt) then
         fmt='(a)'
         if (.not.chkflag('TITLE',fmt)) go to 8001
         read(22,fmt,end=8001,err=8001) rec
      end if
c
c     read counts
c
      sect='counts'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag("POINTERS",fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     NATOM,NTYPES,NBONH,MBONA,NTHETH,MTHETA,
     .     NPHIH,MPHIA,NHPARM,NPARM,NNB,ANRES,
     .     NBONA,NTHETA,NPHIA,NUMBND,NUMANG,NPTRA,
     .     NATYP,NPHB,IFPERT,NBPER,NGPER,NDPER,
     .     MBPER,MGPER,MDPER,IFBOX,NMXRS,IFCAP
c
c     do a few checks (size consistent with mol file...)
c
      if (NPARM.ne.0) then
         write(6,'(a)') 'Parameter/Topology: '//
     .        'NPARM.ne.0, not acceptable in MINIMAX.'
         goto 9000
      end if
      if (nat.ne.NATOM) then
         write(6,'(a)') 'Parameter/Topology: number of atoms does not '
     .        //'match number in mol file.'
         goto 9000
      end if
      if (mbtot.lt.NBONH+NBONA) then
         write(6,'(a)') 'Parameter/Topology: number of bonds '//
     .        'exceeded.'
         goto 9000
      end if
      if (mantot.lt.NTHETH+NTHETA) then
         write(6,'(a)') 'Parameter/Topology: number of angles '//
     .        'exceeded.'
         goto 9000
      end if
      if (mwork2.lt.NTHETH+NTHETA) then
         write(6,'(a)') 'Parameter/Topology: '//
     .        'IWORK overflow (angles).'
         goto 9000
      end if
      if (mtotot.lt.NPHIH+NPHIA) then
         write(6,'(a)') 'Parameter/Topology: number of dihedrals '//
     .        'exceeded.'
         goto 9000
      end if
      if (matype.lt.NTYPES) then
         write(6,'(a)') 'Parameter/Topology: number of LJ types '//
     .        'exceeded.'
         goto 9000
      end if
      if (NUMBND.gt.MXPAR) then
         write(6,'(a)') 'Parameter/Topology: too many '//
     .        'bond stretch types'
         goto 9000
      end if
      if (NUMANG.gt.MXPAR) then
         write(6,'(a)') 'Parameter/Topology: too many '//
     .        'angle bend types'
         goto 9000
      end if
      if (NPTRA.gt.mtotyp.or.NPTRA.gt.mwork) then
         write(6,'(a)') 'Parameter/Topology: too many '//
     .        'dihedral types'
         goto 9000
      end if
      if (NNB.gt.mxexcl) then
         write(6,'(a)') 'Parameter/Topology: too many '//
     .        'non-bonded exclusions'
         goto 9000
      end if
c
c     hbs = 0, except for TAFF
c
      do i=1,nat
         hbs(i)=0
      end do
c         
c     read (and ignore) atom names
c
      sect='atom names'
      fmt='(20a4)'
      if (newfmt) then
         if (.not.chkflag('ATOM_NAME',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (rec,i=1,NATOM)
c
c     read (and ignore) partial atomic charges
c
      sect='partial charges'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('CHARGE',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (chrg(i),i=1,NATOM)
      do i=1,NATOM
         chrg(i)=chrg(i)/sqrt(332.0522D0)
      end do
      chrgok=.True.
c
c     read (and ignore) atom masses
c
      sect='masses'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('MASS',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (dummy,i=1,NATOM)
c
c     read LJ types
c
      sect='LJ types'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('ATOM_TYPE_INDEX',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (LJ(i),i=1,NATOM)
      do i=1,NATOM
         if (LJ(i).lt.1.or.LJ(i).gt.NTYPES) then
            write(6,'(a)') ' LJ type out of valid range'
            goto 8001
         end if
      end do
      n_ljtype=NTYPES
c
c     read NUMEX array (number of exclusions per atom)
c
      sect='NUMEX array'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('NUMBER_EXCLUDED_ATOMS',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (expt(i),i=1,NATOM)
      expt(0)=0
      do i=1,NATOM
         if (expt(i).lt.0.or.expt(i).gt.NAT) then
            write(6,'(a)') ' NUMEX entry out of valid range' 
            goto 8001
         end if
         expt(i)=expt(i)+expt(i-1)
      end do
      if (expt(NATOM).gt.mxexcl) then
         write(6,'(a)') ' Too many non-bonded exclusions.' 
         goto 8001
      end if
c
c     read ICO pointers to LJ parameters
c
      sect='ICO pointers'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('NONBONDED_PARM_INDEX',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     ((ICO(i,j),j=1,NTYPES),i=1,NTYPES)
      M=((NTYPES+1)*NTYPES)/2
      do i=1,NTYPES
         do j=1,NTYPES
            if (ICO(i,j).lt.1.or.ICO(i,j).gt.M) then
               write(6,'(a)') ' ICO pointer out of valid range'
               if (ICO(i,j).lt.1) then
                  write(6,'(a)') ' 10-12 potential not implemented' 
               end if
               goto 8001
            end if
         end do
      end do
c
c     read (and ignore) residue labels, residue pointers
c
      sect='residue labels'
      fmt='(20a4)'
      if (newfmt) then
         if (.not.chkflag('RESIDUE_LABEL',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (rec,i=1,ANRES)
      sect='residue pointers'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('RESIDUE_POINTER',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (j,i=1,ANRES)
c
c     read bond stretch parameters
c
      sect='bond stretch K'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('BOND_FORCE_CONSTANT',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (RK(i),i=1,NUMBND)
      sect='bond stretch Req'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('BOND_EQUIL_VALUE',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (REQ(i),i=1,NUMBND)
c
c     read angle bend parameters
c
      sect='angle bend K'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('ANGLE_FORCE_CONSTANT',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (TK(i),i=1,NUMANG)
      sect='angle bend Aeq'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('ANGLE_EQUIL_VALUE',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (TEQ(i),i=1,NUMANG)
c
c     read torsion and improper parameters
c
      sect='dihedral K'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('DIHEDRAL_FORCE_CONSTANT',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (kzt(i),i=1,NPTRA)
      sect='dihedral n'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('DIHEDRAL_PERIODICITY',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (work(i),i=1,NPTRA)
      do i=1,NPTRA
         kzf(i)=nint(work(i))
      end do
      sect='dihedral phase'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('DIHEDRAL_PHASE',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (kzs(i),i=1,NPTRA)
c
c     read (and ignore) SOLTY parameters
c
      sect='SOLTY'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('SOLTY',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (dummy,i=1,NATYP)
c
c     read LJ parameters (A,B)
c
      sect='LJ parameters A'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('LENNARD_JONES_ACOEF',fmt)) go to 8001
      end if
      M=((NTYPES+1)*NTYPES)/2
      read(22,fmt,end=8001,err=8001) (vx(i),i=1,M)
      do i=1,NTYPES
         do j=1,NTYPES
            vaij(i,j)=vx(ICO(i,j))
         end do
      end do
      sect='LJ parameters B'
      fmt='(5e16.8)'
      if (newfmt) then
         if (.not.chkflag('LENNARD_JONES_BCOEF',fmt)) go to 8001
      end if
      M=((NTYPES+1)*NTYPES)/2
      read(22,fmt,end=8001,err=8001) (vx(i),i=1,M)
      do i=1,NTYPES
         do j=1,NTYPES
            vbij(i,j)=-vx(ICO(i,j)) ! since we have E = A/r^12+B/r^6
         end do
      end do
c
c     read bond stretch terms
c
      sect='bonds (H)'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('BONDS_INC_HYDROGEN',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     (bnd(1,i),bnd(2,i),bndt(i),i=1,NBONH)
      sect='bonds (non-H)'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('BONDS_WITHOUT_HYDROGEN',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     (bnd(1,i),bnd(2,i),bndt(i),i=NBONH+1,NBONH+NBONA)
      nbnd=NBONH+NBONA
      do i=1,nbnd
         bnd(1,i)=bnd(1,i)/3+1
         bnd(2,i)=bnd(2,i)/3+1
         if (bnd(1,i).lt.1.or.bnd(1,i).gt.nat.or.
     .       bnd(2,i).lt.1.or.bnd(2,i).gt.nat) then
            write(6,'(a)') ' atom index out of range'
            go to 8001
         end if
         if (bndt(i).lt.1.or.bndt(i).gt.NUMBND) then
            write(6,'(a)') ' type index out of range'
            go to 8001
         end if
         bndl0(i)=REQ(bndt(i))
         kbnd(i)=2*RK(bndt(i))
      end do
c
c     read angle bend terms
c
      sect='angles (H)'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('ANGLES_INC_HYDROGEN',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     (ang(1,i),ang(2,i),ang(3,i),angt(i),i=1,NTHETH)
      sect='bonds (non-H)'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('ANGLES_WITHOUT_HYDROGEN',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     (ang(1,i),ang(2,i),ang(3,i),angt(i),i=NTHETH+1,NTHETH+NTHETA)
      nang=NTHETH+NTHETA
      do i=1,nang
         ang(1,i)=ang(1,i)/3+1
         ang(2,i)=ang(2,i)/3+1
         ang(3,i)=ang(3,i)/3+1
         if (ang(1,i).lt.1.or.ang(1,i).gt.nat.or.
     .       ang(2,i).lt.1.or.ang(2,i).gt.nat.or.
     .       ang(3,i).lt.1.or.ang(3,i).gt.nat) then
            write(6,'(a)') ' atom index out of range'
            go to 8001
         end if
         if (angt(i).lt.1.or.angt(i).gt.NUMANG) then
            write(6,'(a)') ' type index out of range'
            go to 8001
         end if
         angl0(i)=TEQ(angt(i))*RAD2DEG
         kang(i)=2*TK(angt(i))*DEG2RAD*DEG2RAD
      end do
c
c     dihedral terms
c
      sect='dihedrals (H)'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('DIHEDRALS_INC_HYDROGEN',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     (tor(1,i),tor(2,i),tor(3,i),tor(4,i),tort(i),i=1,NPHIH)
      sect='dihedrals (non-H)'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('DIHEDRALS_WITHOUT_HYDROGEN',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) 
     .     (tor(1,i),tor(2,i),tor(3,i),tor(4,i),tort(i),
     .     i=NPHIH+1,NPHIH+NPHIA)
      ntor=NPHIH+NPHIA
      do i=1,ntor
         M=0
         if (tor(3,i).lt.0) then
            tor(3,i)=-tor(3,i)
            M=M+1
         end if
         if (tor(4,i).lt.0) then
            tor(4,i)=-tor(4,i)
            M=M+2
         end if
         tor(1,i)=tor(1,i)/3+1
         tor(2,i)=tor(2,i)/3+1
         tor(3,i)=tor(3,i)/3+1
         tor(4,i)=tor(4,i)/3+1
         if (tor(1,i).lt.1.or.tor(1,i).gt.nat.or.
     .       tor(2,i).lt.1.or.tor(2,i).gt.nat.or.
     .       tor(3,i).gt.nat.or.tor(4,i).gt.nat) then
            write(6,'(a)') ' atom index out of range'
            go to 8001
         end if
         if (tort(i).lt.1.or.tort(i).gt.NPTRA) then
            write(6,'(a)') ' type index out of range'
            go to 8001
         end if
         ntortrm(i)=1
         tortp(1,i)=tort(i)
         tor0(i)=-360
         tort(i)=M
      end do
c
c     collect 1-4 nonbonded list
c
      n=0
      do i=1,ntor
         if (mod(tort(i),2).eq.0) then
            iwork(n+1)=tor(1,i)
            iwork(n+2)=tor(4,i)
            n=n+2
         end if
      end do
      nhb14=insert_excl(n/2,iwork,nat,hbpt,hb14)
c
c     pack dihedral list
c
      n=1
      do i=2,ntor
         if (ntortrm(n).lt.mxft.and.
     .        tor(1,n).eq.tor(1,i).and.tor(2,n).eq.tor(2,i).and.
     .        tor(3,n).eq.tor(3,i).and.tor(4,n).eq.tor(4,i)) then
            ntortrm(n)=ntortrm(n)+1
            tortp(ntortrm(n),n)=tortp(1,i)
         else
            n=n+1
            tort(n)=tort(i)
            do j=1,4
               tor(j,n)=tor(j,i)
            end do
            tortp(1,n)=tortp(1,i)
            ntortrm(n)=1
         end if
      end do
      ntor=min(ntor,n)
c
c     read non-bonded exclusion lists, convert to internal format
c
      sect='exclusions'
      fmt='(12i6)'
      if (newfmt) then
         if (.not.chkflag('EXCLUDED_ATOMS_LIST',fmt)) go to 8001
      end if
      read(22,fmt,end=8001,err=8001) (iij(i),i=1,NNB)
      n=0
      do i=1,natom
         do j=expt(i-1)+1,expt(i)
            if (iij(j).ge.1.and.iij(j).le.nat) then
               iwork(n+1)=i
               iwork(n+2)=iij(j)
               n=n+2
            end if
         end do
      end do
      nexcl=insert_excl(n/2,iwork,nat,expt,iij)
c
c     ignore the rest of the file
c
      topol_amber=.TRUE.
      goto 9000
c
c     Error handling...
c
 8000 continue
      l=str_trim(rec,prm_file)
      write(6,'(a)') ' Could not open parameter/topology file '
     .     //rec(1:l)//' for read.'
      return
c
 8001 continue
      write(6,'(a)') ' Parameter/Topology file: failed to read '//sect
      goto 9000
c
c     Close file and return
c
 9000 continue
      close(unit=22)
      return
      end
c------------------------------------------------------------------------
      logical function chkflag(flag,fmt)
c
      character fmt*(*),flag*(*),rec*250,flg*250
      integer l,str_trim

      chkflag=.FALSE.
      read(22,'(a)',end=8000,err=8000) rec
      call str_upcase(rec,rec)
      l=str_trim(flg,rec(7:))
      if (rec(1:6).ne.'%FLAG '.or.flg(1:l).ne.flag) go to 8000
      read(22,'(a)',end=8000,err=8000) rec
      call str_upcase(rec,rec)
      if (rec(1:7).ne.'%FORMAT') go to 8000
      l=str_trim(fmt,rec(8:))
      chkflag=.TRUE.
      return 
c
 8000 continue
      return
      end
