      logical function topol_charmm()
c
c     Setup of energy function, CHARMM version.
c
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
#include "worksp.h"
#include "option.h"
#if defined(sgi)
#define READONLY ,readonly
#else
#define READONLY
#endif
c
      integer i,j,k,k1,l,n,nl,nbx,mm,nn,kn,jn,knn,jnn
      character*250 rec
      integer str_trim,insert_excl
c     
      topol_charmm=.FALSE.
      open(unit=22,file=psf_file,status='old',err=8000 READONLY)
c
c     check PSF tag in 1st header record:
c
      nl=1
      read(22,'(a)',end=8001,err=8001) rec
      if (rec(1:4).ne.'PSF ') then
         write(6,'(a)') ' Error reading topology file, bad tag in 1st'//
     .        ' header record'
         goto 9000
      end if
c
c     skip title lines
c
      nl=3
      read(22,'(/i8)',end=8002,err=8003) n
      if (n.lt.1) n=1     ! for compatibility with CHARMM
      do i=1,n
         nl=nl+1
         read(22,'(a)',end=8002,err=8003) rec
      end do
c
c     atom records..., only read reid, resnr, segid, for the rest
c     we use what was read from molfile
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) n
      if (n.ne.nat) then
         write(6,'(a)') ' Incorrect number of atoms in topology file.'
         write(6,'(a,i6,a,i6)') ' Molecule: ',nbnd,', topology file: ',n
         goto 9000
      end if
      if (n.lt.1) n=1     ! for compatibility with CHARMM
      do i=1,n
         nl=nl+1
         read(22,'(9x,a4,1x,a4,1x,a4)',end=8002,err=8003) 
     .        segid(i),resnr(i),resid(i)
      end do
c
c     read bonds (this will overwrite bonds from molfile)
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) n
      if (n.ne.nbnd) then
         write(6,'(a)') ' Incorrect number of bonds in topology file.'
         write(6,'(a,i6,a,i6)') ' Molecule: ',nbnd,', topology file: ',n
         goto 9000
      end if
      read(22,'(8i8)',end=8002,err=8003) 
     .     ((bnd(j,l),j=1,2),l=1,nbnd)
      do k=1,nbnd
         bndl0(k)=-1000   !  bugfix 11-Jan-02: was kbnd(k)=-1000 
         do j=1,2
            if (bnd(j,k).lt.1.or.bnd(j,k).gt.nat) then
               write(6,'(a,i6)') ' invalid atom number (bonds):',
     .              bnd(j,k)
               goto 9000
            end if
         end do
      end do
c
c     bond angles...
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) nang
      if (nang.gt.mantot) then
         write(6,'(a,i5,a,i5)') ' Too many bond angles(',
     .        nang,', max. is ',mantot
         goto 9000
      end if
      read(22,'(9i8)',end=8002,err=8003) 
     .     ((ang(j,l),j=1,3),l=1,nang)
      do k=1,nang
         angl0(k)=-1000
         do j=1,3
            if (ang(j,k).lt.1.or.ang(j,k).gt.nat) then
               write(6,'(a,i6)') ' invalid atom number (angles):',
     .              ang(j,k)
               goto 9000
            end if
         end do
      end do
c
c     dihedrals
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) ntor
      if (ntor.gt.mtotot) then
         write(6,'(a,i5,a,i5)') ' Too many dihedral angles(',
     .        ntor,', max. is ',mtotot
         goto 9000
      end if
      read(22,'(8i8)',end=8002,err=8003) 
     .     ((tor(j,l),j=1,4),l=1,ntor)
      ntortp=0
      do k=1,ntor
         tort(k)=1              ! all bonds are type 1 ...
         ntortrm(k)=0
         do j=1,3
            if (tor(j,k).lt.1.or.tor(j,k).gt.nat) then
               write(6,'(a,i6)') ' invalid atom number (dihedrals):',
     .              tor(j,k)
               goto 9000
            end if
         end do
      end do
c
c     'improper' torsions (stored in /plane/)
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) npla
      if (npla.gt.mplane) then
         write(6,'(a,i5,a,i5)') ' Too many improper torsions(',
     .        npla,', max. is ',mplane
         goto 9000
      end if
      read(22,'(8i8)',end=8002,err=8003) 
     .     ((pla(j,l),j=1,4),l=1,npla)
      do k=1,npla
         kpla(k)=-1.0
         do j=1,3
            if (pla(j,k).lt.1.or.pla(j,k).gt.nat) then
               write(6,'(a,i6)') ' invalid atom number (impropers):',
     .              pla(j,k)
               goto 9000
            end if
         end do
      end do
c
c     HB donors... ignored.
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) n
      read(22,'(8i8)',end=8002,err=8003) (i,j,l=1,n)

c
c     HB acceptors... ignored.
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) n
      read(22,'(8i8)',end=8002,err=8003) (i,j,l=1,n)
c
c     Non-bonded exclusions... (always 0 from Wit!P)
c
      nl=nl+2
      read(22,'(/i8)',end=8002,err=8003) nexcl
      if (nexcl.gt.mxexcl.and.nbxmod.gt.0) then
         write(6,'(a)') 
     .        ' ERROR: exclusion list in PSF produced overflow.'
         goto 9000
      end if
      if (nbxmod.gt.0) then
         read(22,'(8i8)',end=8002,err=8003) (iij(i),i=1,nexcl)
      else
         read(22,'(8i8)',end=8002,err=8003) (iij(1),i=1,nexcl)
      end if
      read(22,'(8i8)',end=8002,err=8003) (expt(i),i=1,nat)
      if (nbxmod.lt.0) nexcl=0
      expt(0)=0
c
c     Make non-bonded exclusion list
c
      nbx=nat+1+2*nbnd
c     First use explicit exclusions from psf file...
      if (nexcl.gt.0) then
         nexcl=0
         do i=1,nat
            do k=expt(i-1)+1,expt(i)
               iwork(nbx+nexcl+1)=i
               iwork(nbx+nexcl+2)=iij(k)
               nexcl=nexcl+2
            end do
         end do
      end if
c     Then insert automatic exclusions depending on nbxmod
      if (abs(nbxmod).gt.1) then    ! exclude bonded interactions
         do k=1,nbnd
            iwork(nbx+nexcl+1)=bnd(1,k)
            iwork(nbx+nexcl+2)=bnd(2,k)
            nexcl=nexcl+2
         end do
      end if
      if (abs(nbxmod).gt.2) then    ! exclude 1-3 interactions
         call charmm_generateCT     ! need connection table for this...
         do i=1,nat
            nn=iwork(i+1)-iwork(i)
            do j=1,nn-1
               do k=j+1,nn
                  iwork(nbx+nexcl+1)=iwork(iwork(i)+j)
                  iwork(nbx+nexcl+2)=iwork(iwork(i)+k)
                  nexcl=nexcl+2
               end do
            end do
         end do
      end if
      if (abs(nbxmod).eq.4) then
         do i=1,nbnd
            j=bnd(1,i)
            k=bnd(2,i)
            mm=iwork(j+1)-iwork(j)
            nn=iwork(k+1)-iwork(k)
            if (mm.gt.1.and.nn.gt.1) then
               do jn=1,mm
                  jnn=iwork(iwork(j)+jn)
                  if (jnn.ne.k) then
                     do kn=1,nn
                        knn=iwork(iwork(k)+kn)
                        if (knn.ne.j.and.knn.ne.jnn) then
                           iwork(nbx+nexcl+1)=jnn
                           iwork(nbx+nexcl+2)=knn
                           nexcl=nexcl+2
                        end if
                     end do
                  end if
               end do
            end if
         end do
      end if
c     insert exclusions into internal list
      nexcl=insert_excl(nexcl/2,iwork(nbx+1),nat,expt,iij)
c     no special scaling for 1-4 HBs:
      do k=1,nat                
         hbs(k)=0
      end do
c     use hb14 to mark 1-4 interactions if abs(nbxmod).eq.5
      nhb14=0
      if (abs(nbxmod).eq.5) then
         do i=1,nbnd
            j=bnd(1,i)
            k=bnd(2,i)
            mm=iwork(j+1)-iwork(j)
            nn=iwork(k+1)-iwork(k)
            if (mm.gt.1.and.nn.gt.1) then
               do jn=1,mm
                  jnn=iwork(iwork(j)+jn)
                  if (jnn.ne.k) then
                     do kn=1,nn
                        knn=iwork(iwork(k)+kn)
                        if (knn.ne.j.and.knn.ne.jnn) then
                           iwork(nbx+nhb14+1)=jnn
                           iwork(nbx+nhb14+2)=knn
                           nhb14=nhb14+2
                        end if
                     end do
                  end if
               end do
            end if
         end do
      end if
      nhb14=insert_excl(nhb14/2,iwork(nbx+1),nat,hbpt,hb14)
c
c
c     rest of file (charge groups) ignored...
c     done. if we have come this far, everything is o.k...
c
      topol_charmm=.TRUE.
      goto 9000
c
c Errors:
c
 8000 continue
      l=str_trim(rec,psf_file)
      write(6,'(a)') ' Could not open topology file '
     .     //rec(1:l)//' for read.'
      return
c
 8001 continue
      write(6,'(a)') ' Error reading topology file: '//
     .     'failed to read 1st header recoord'
      goto 9000
c
 8002 continue
      write(6,'(a,i7,a)') ' Premature EOF (topology file, line:',nl,')'
      goto 9000
c
 8003 continue
      write(6,'(a,i7,a)') ' I/O error (topology file, line:',nl,')'
      goto 9000
c
c     all done...
c
 9000 continue
      close(unit=22)
      return
      end
c------------------------------------------------------------------------

      subroutine charmm_generateCT
c
c     generate temporary connection table in lower part of IWORK
c     List of k-th neighbour of atom i will be in iwork(iwork(i)+k)),
c     atom i has iwork(i+1)-iwork(i) neighbours.
c
      implicit none
#include "sizes.h"
#include "atom.h"
#include "energy.h"
#include "worksp.h"
#include "option.h"
      integer i,j,k,nbl
c
      nbl=nat+1
      do k=1,nbl
         iwork(k)=0
      end do
      do k=1,nbnd             ! counting loop
         i=bnd(1,k)
         j=bnd(2,k)
         if (i.le.j) then
            bnd(1,k)=j
            bnd(2,k)=i
         end if
         iwork(i)=iwork(i)+1  ! count neighbours of i
         iwork(j)=iwork(j)+1  ! count neighbours of j
      end do
      do i=1,nat
         iwork(i+1)=iwork(i+1)+iwork(i) ! accumlated number of neighbours
      end do
      do k=1,nbnd             ! store loop
         i=bnd(1,k)
         j=bnd(2,k)
         iwork(nbl+iwork(i))=j 
         iwork(nbl+iwork(j))=i
         iwork(i)=iwork(i)-1  ! undo accumulation
         iwork(j)=iwork(j)-1
      end do
      do k=1,nbl
         iwork(k)=iwork(k)+nbl
      end do
      return
      end
