      logical function topol_taff()
c     
c     Analysis of topology of initial structure, TAFF version
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "atom.h"
#include "energy.h"
#include "worksp.h"
c     
      integer i,j,k,nbl
      integer jn,jnn,kn,knn,mm,nn
      integer insert_excl
c     
      topol_taff=.TRUE.
c     
c     Generate neighbourlist in iwork, swap i,j in bond if i.le.j
c     
      nbl=nat+1
      do k=1,nat+1
         iwork(k)=0
      end do
      do k=1,nbnd
         i=bnd(1,k)
         j=bnd(2,k)
         if (i.le.j) then
            bnd(1,k)=j
            bnd(2,k)=i
         end if
         iwork(i)=iwork(i)+1
         iwork(j)=iwork(j)+1
      end do
      do i=1,nat
         iwork(i+1)=iwork(i+1)+iwork(i)
      end do
      do k=1,nbnd
         i=bnd(1,k)
         j=bnd(2,k)
         iwork(nbl+iwork(i))=j
         iwork(nbl+iwork(j))=i
         iwork(i)=iwork(i)-1
         iwork(j)=iwork(j)-1
      end do
c     
c     Create list of bondangles
c     
      nang=0
      do i=1,nat
         nn=iwork(i+1)-iwork(i)
         do j=1,nn-1
            do k=j+1,nn
               nang=nang+1
               if (iwork(nbl+iwork(i)+j).gt.iwork(nbl+iwork(i)+k)) then
                  ang(1,nang)=iwork(nbl+iwork(i)+j)
                  ang(2,nang)=i
                  ang(3,nang)=iwork(nbl+iwork(i)+k)
               else
                  ang(3,nang)=iwork(nbl+iwork(i)+j)
                  ang(2,nang)=i
                  ang(1,nang)=iwork(nbl+iwork(i)+k)
               end if
            end do
         end do
      end do
c     
      if (nang.gt.mantot) then
         write(6,'(a)') ' Too many bond angles.'
         topol_taff=.FALSE.
         return
      end if
c     
c     Create list of torsion angles
c     
      ntor=0
      do i=1,nbnd
         j=bnd(1,i)
         k=bnd(2,i)
         mm=iwork(j+1)-iwork(j)
         nn=iwork(k+1)-iwork(k)
         if (mm.gt.1.and.nn.gt.1) then
            do jn=1,mm
               jnn=iwork(nbl+iwork(j)+jn)
               if (jnn.ne.k) then
                  do kn=1,nn
                     knn=iwork(nbl+iwork(k)+kn)
                     if (knn.ne.j.and.knn.ne.jnn) then
                        ntor=ntor+1
                        tor(1,ntor)=jnn
                        tor(2,ntor)=j
                        tor(3,ntor)=k
                        tor(4,ntor)=knn
                        tort(ntor)=bndt(i)
                     end if
                  end do
               end if
            end do
         end if
      end do
c     
      if (ntor.gt.mtotot) then
         write(6,'(a)') ' Too many torsion angles.'
         topol_taff=.FALSE.
         return
      end if
c     
c     Create list of out of plane centers
c     
      npla=0
      do 400 i=1,nat
         if (oopcent(ntype(i))) then
            npla=npla+1
            nn=1
            pla(nn,npla)=i
            do 310 j=1,nbnd
               if (i.eq.bnd(1,j)) then
                  nn=nn+1
                  if (nn.le.4) pla(nn,npla)=bnd(2,j)
               else if (i.eq.bnd(2,j)) then
                  nn=nn+1
                  if (nn.le.4) pla(nn,npla)=bnd(1,j)
               end if
 310        continue
            if (nn.ne.4) npla=npla-1
         end if
 400  continue
c     
c     Mark atoms involved in H-bonds...
c     
      do i=1,nat
         if (hba(ntype(i))) then
            hbs(i)=-1
         else
            hbs(i)=0
         end if
      end do
c     
      do i=1,nbnd
         if (hbd(ntype(bnd(1,i))).and.atomic_n(ntype(bnd(2,i))).eq.1)
     .        hbs(bnd(2,i))=1
         if (hbd(ntype(bnd(2,i))).and.atomic_n(ntype(bnd(1,i))).eq.1)
     .        hbs(bnd(1,i))=1
      end do
c     
c     Make list of 1-4 hydrogen bonds...
c     
      nhb14=0
      do k=1,ntor
         i=tor(1,k)
         j=tor(4,k)
         if ((hbs(i).lt.0 .and. hbs(j).gt.0) .or.
     .        (hbs(i).gt.0 .and. hbs(j).lt.0)) then
            if (i.lt.j) then
               i=tor(4,k)
               j=tor(1,k)
            end if
            iwork(nhb14+1)=i
            iwork(nhb14+2)=j
            nhb14=nhb14+2
         end if
      end do
      nhb14=insert_excl(nhb14/2,iwork,nat,hbpt,hb14)
c
c     Make non-bonded exclusion list
c
      nexcl=0
      do k=1,nbnd
         iwork(nexcl+1)=bnd(1,k)
         iwork(nexcl+2)=bnd(2,k)
         nexcl=nexcl+2
      end do
      do k=1,nang
         iwork(nexcl+1)=ang(1,k)
         iwork(nexcl+2)=ang(3,k)
         nexcl=nexcl+2
      end do
      nexcl=insert_excl(nexcl/2,iwork,nat,expt,iij)
c
c     done...
c     
      topol_taff=.TRUE.
      return
      end
