#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(alliant)
int strcasecmp(char *a,char *b);
int strncasecmp(char *a,char *b,int l);
#endif /* (alliant) */

#define TRUE  1
#define FALSE 0

#define V1  5
#define V2  5
#define V3  6

/* Filter to convert BIOSYM .arc file into Wit!P .cnf file */


int main() {
  
  char rec[256],hdr1[82],hdr2[82],*X;
  int recsz,nat,n,end_mode;
  FILE *in,*out;
  float x,y,z,energy;

  in=stdin;
  out=stdout;
  recsz=161;

  if (!fgets(rec,recsz,in)) {	      /* if it is fixed 80 we have the two headers */
    fprintf(stderr,"failed to read header\n");
    return(1);
  }
  if (strlen(rec)==recsz-1) {
    recsz=81;		      /* looks like fixed 80 */
    strncpy(hdr1,rec,80);hdr1[80]='\n';hdr1[81]='\0';
    strncpy(hdr2,rec+80,80);hdr2[80]='\n';hdr2[81]='\0';
  }else{
    strcpy(hdr1,rec);
    if (fgets(rec,recsz,in)) {    /* read second header line */       
      strncpy(hdr2,rec,80); hdr2[80]='\n';hdr2[81]='\0';
    }else{
      fprintf(stderr,"failed to read header\n");
      return(1);
    }
  }
  if (strncasecmp(rec,"!BIOSYM ARCHIVE 1",17)==0) { /* Version 2.7 */
    X=rec+V1;
    if (strncasecmp(hdr2,"PBC=ON",6)==0 && !fgets(rec,recsz,in)) fprintf(stderr,"PBC format error\n"); /* PBC record */
  }else if (strncasecmp(hdr1,"!BIOSYM ARCHIVE 2",17)==0) { /* Version 2.10 */
    X=rec+V2;
    if (strncasecmp(hdr2,"PBC=ON",6)==0 && !fgets(rec,recsz,in)) fprintf(stderr,"PBC format error\n"); /* PBC record */
  }else if (strncasecmp(hdr1,"!BIOSYM ARCHIVE 3",17)==0) { /* Version 2.3 */
    X=rec+V3;
    if (strncasecmp(hdr2,"PBC=ON",6)==0 && !fgets(rec,recsz,in)) fprintf(stderr,"PBC format error\n"); /* PBC record */
    else if (strncasecmp(hdr2,"PBC=2D",6)==0 && !fgets(rec,recsz,in)) fprintf(stderr,"PBC format error\n");
    else if (strncasecmp(hdr2,"HELIX",5)==0) {
      fprintf(stderr,"arc-files containing helix information are not supported.");
      return(1);
    }
  }else{
    fprintf(stderr,"Unknown Archive Version.\n");
    return(1);
  }

  n=0;
  for (;;) {
    if (!fgets(rec,recsz,in) || feof(in)) break; /* Title and Energy */
    energy=atof(rec+60);
    if (!fgets(rec,recsz,in) || feof(in)) break; /* Date */
    fprintf(out,"Conformation Conf_%d Energy = %0.3f Status = Archived\n",
	    ++n,energy);
    nat=0; end_mode=FALSE;
    for (;;) {
      if (!fgets(rec,recsz,in) || feof(in)) {
	fprintf(stderr,"unexpected eof.\n");
	return(1);
      }
      if (strncasecmp(rec,"end",3)==0) {
	if (end_mode) break;
	else end_mode=TRUE;
      }else{
	end_mode=FALSE;
	sscanf(X,"%f %f %f",&x,&y,&z);
	fprintf(out,"%0.3f %0.3f %0.3f\n",x,y,z);
	nat++;
      }
    }
    fprintf(stderr,"Conformation %d, Energy = %0.3f, %d atoms.\n",
	    n,energy,nat);
    
  }
  fprintf(stderr,"%d molecules converted.\n",n);
  return(0);
}
