#include <stdio.h>
#include <stdlib.h>
#include "cmaplib.h"

/* Convert CCP4 Map file to Wit!P CNT file.           */
/* Requires tools from library cmaplib                */
/* Only mode 2 map files (data is REAL) are supported */
/* A.W., August 2007                                  */

static void usage(void);

int main(int argc,char **argv)
{
  int i,j,k,mode,nx[3],mapdim[3],order[3],origin[3],grd[3],I[3],size,success;
  FILE *out;
  char *infile=NULL,*outfile=NULL;
  void *map;
  float cell[6],xmin[3],dx[3],*rho,*section;
  

  for (i=1;i<argc;i++) {
    if (argv[i][0]!='-') {
      if (!infile) infile=argv[i];
      else if (!outfile) outfile=argv[i];
      else {
	fprintf(stderr,"usage: ccp42wnp ccp4infile [wnpoutfile]\n");
	return(1);
      }
    }else if (!strcmp(argv[i],"--help")||!strcmp(argv[i],"-h")) {
      usage();
      return(0);
    }
  }
  if (!infile) {
    fprintf(stderr,"usage: ccp42wnp ccp4infile [wnpoutfile]\n");
    return(1);
  }
  
  map=ccp4_cmap_open(infile,O_RDONLY);
  if (map) {
    mode=ccp4_cmap_get_datamode(map);
    if (mode!=2) {
      fprintf(stderr,"sorry, map is mode %d, but we support only mode 2 maps.\n",mode);
      ccp4_cmap_close(map);
      return(3);
    }
    ccp4_cmap_get_cell(map,cell);
    ccp4_cmap_get_dim(map,mapdim);
    ccp4_cmap_get_grid(map,grd);           /* note: this is in XYZ order */
    ccp4_cmap_get_order(map,order);
    ccp4_cmap_get_origin(map,origin);
    size=mapdim[0]*mapdim[1]*mapdim[2];
    for (i=0;i<3;i++) {
      order[i]-=1;
      xmin[order[i]]=(float)origin[i]/grd[order[i]];
      dx[order[i]]=1.0/grd[order[i]];
      nx[order[i]]=mapdim[i];
    }

    rho=malloc(size*sizeof(int));
    section=malloc(mapdim[0]*mapdim[1]*sizeof(float));
    if (rho&&section) {
      ccp4_cmap_seek_section(map,0,SEEK_SET);
      for (k=0;k<mapdim[2];k++) {
	success=ccp4_cmap_read_section(map,section);
	if (!success) {
	  fprintf(stderr,"failed to read section %d from %s\n",k,infile);
	  break;
	}
	I[order[2]]=k;
	for (j=0;j<mapdim[1];j++) {
	  I[order[1]]=j;
	  for (i=0;i<mapdim[0];i++) {
	    I[order[0]]=i;
	    rho[(I[2]*nx[1]+I[1])*nx[0]+I[0]]=section[mapdim[0]*j+i];
	  }
	}
      }
      ccp4_cmap_close(map);
      if (k==mapdim[2]) {
	if (outfile) out=fopen(outfile,"w");
	else out=stdout;
	if (out) {
	  fwrite(xmin,sizeof(float),3,out);
	  fwrite(dx,sizeof(float),3,out);
	  fwrite(nx,sizeof(int),3,out);
	  fwrite(rho,sizeof(float),size,out);
	  fwrite(cell,sizeof(float),6,out);
	  fclose(out);
	}else{
	  fprintf(stderr,"failed to open output map file %s\n",outfile);
	  free(section); free(rho); 
	  return(4);
	}
      }else{
	fprintf(stderr,"failed to read complete map file %s\n",infile);
	free(section); free(rho);
 	return(5);
      }
      free(section); free(rho);
    }else{
      fprintf(stderr,"failed to allocate space for %dx%dx%d map.\n",mapdim[0],mapdim[1],mapdim[2]);
      ccp4_cmap_close(map);
      return(2);
    }
  }else{
    fprintf(stderr,"failed to open CCP4 map file %s\n",infile);
    return(2);
  }
  return(0);
}

static void usage(void)
{
  fprintf(stderr,
	  "Usage: ccp42wnp ccp4mapfile [witnotGridFile]\n"
	  "\n"
	  "Description:\n"
	  "\n"
	  "Converts a CCP4 map file to a Wit!P grid file. The CCP4 map file must be\n"
	  "a data mode 2 file (i.e. the data is stored as floating point numbers).\n"
	  "The axes in the input file may have any order. The output file is a binary\n"
	  "file with axes order XYZ. The output will be written to stdout, if no filename\n"
	  "is specified for the output file. The grid file may be contoured in Wit!P\n"
	  "with the Contour and with the Surface/Grid commands.\n"
	  "\n");
}
