#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


static void swapper(int n,void *data);

#define V(k) (p0+(k&1?1:0)+(k&2?nx[1]:0)+(k&4?nx[0]*nx[1]:0))
#define TRUE 1
#define FALSE 0

/* Assign property value to dots. Usage: grd2dot grid_file_name <dots_in >dots_out 

   Property is interpolated grid value at dot position.
*/

int main(int argc,char **argv)
{
  FILE *in;
  float xmin[3],dx[3],*val,*v,vmin,vmax,dot[4],c[3];
  float s,wt,x0[3];
  int nx[3],n,j,k,nat,ndot,ix[3],swap;
  char rec[257];
  
  if (argc!=2) {
    fprintf(stdout,"usage: grid2cnt grid.cnt <dots_in >dots.out\n");
    return(1);
  }

  /* read grid file into memory */

  in=fopen(argv[1],"r");
  if (in) {
    if (3!=fread(xmin,sizeof(float),3,in)) {
      fprintf(stderr,"could not read 'xmin'\n");
      return(3);
    }
    if (3!=fread(dx,sizeof(float),3,in)) {
      fprintf(stderr,"could not read 'dx'\n");
      return(3);
    }
    if (3!=fread(nx,sizeof(int),3,in)) {
      fprintf(stderr,"could not read 'nx'\n");
      return(3);
    }

    /* plausibility check to determine byte sex */

    if (nx[0]<1||nx[1]<1||nx[2]<1||
	((double)nx[0]*(double)nx[1]*(double)nx[2])>1.e8) {
      fprintf(stderr,"wrong byte sex? swapping bytes.\n");
      swap=TRUE;
    }else swap=FALSE;

    if (swap) { 
      swapper(3,nx);
      swapper(3,xmin);
      swapper(3,dx);
    }

    if (nx[0]<1||nx[1]<1||nx[2]<1||
	((double)nx[0]*(double)nx[1]*(double)nx[2])>1.e8) {
      fprintf(stderr,"sorry. this does not look like a good grid file.\n");
      exit(2);
    }

    n=nx[0]*nx[1]*nx[2];
    fprintf(stderr,"grid base at %0.3f,%0.3f,%0.3f\n",xmin[0],xmin[1],xmin[2]);
    fprintf(stderr,"step size %0.3f,%0.3f,%0.3f\n",dx[0],dx[1],dx[2]);
    fprintf(stderr,"number of points %d = %d x %d x %d\n",n,nx[0],nx[1],nx[2]);
    val=(float *)malloc(n*sizeof(float));
    if (!val) {
      fprintf(stderr,"could not allocate memory for grid\n");
      return(4);
    }
    if (n!=fread(val,sizeof(float),n,in)) {
      fprintf(stderr,"could not read contour file\n");
      return(5);
    }
    if (swap) swapper(n,val);
  }else{
    fprintf(stderr,"could not read input contour file %s\n",argv[1]);
    return(2);
  }


  /*determine min/max grid values */

  for (vmin=*val,vmax=*val,v=val,k=0;k<n;k++,v++) {
    if (*v>vmax) vmax=*v;
    if (vmin>*v) vmin=*v;
  }

  fprintf(stderr,"min/max in grid: %0.6f / %0.6f\n",vmin,vmax);

  /* read atoms and dots... */

  nat=0;
  ndot=0;

  s=vmax;
  vmax=vmin;
  vmin=s;

  while(fgets(rec,256,stdin)) {
    if (strncmp("ATOM",rec,4)==0) {
      fprintf(stdout,"%s",rec);
      nat++;
    }else if (strncmp("DOT",rec,3)==0) {
      sscanf(rec+3,"%f %f %f",&dot[0],&dot[1],&dot[2]);
      dot[3]=9999.9;     /* should handle case where dots extend beyond grid... */
      for (k=0;k<3;k++) ix[k]=(dot[k]-xmin[k])/dx[k]; 
      for (k=0;k<3;k++) if (ix[k]<0 || ix[k]>nx[k]-2) break;
      if (k==3) {
	dot[3]=0.0;
	v=val+ix[2]*(nx[0]*nx[1])+ix[1]*nx[0]+ix[0];
	for (k=0;k<3;k++) {
	  x0[k]=xmin[k]+ix[k]*dx[k];
	  c[k]=(dot[k]-x0[k])/dx[k];
	}
	for (j=0;j<8;j++) {
	  for (wt=1.0,k=0;k<3;k++) wt*=(j&(1<<k) ? (c[k]) : 1-c[k]);
	  dot[3]+=wt*v[(j&1?1:0)+(j&2?nx[0]:0)+(j&4?nx[0]*nx[1]:0)];
	}
      }
      fprintf(stdout,"DOT %0.3f %0.3f %0.3f %0.6f\n",dot[0],dot[1],dot[2],dot[3]);
      if (dot[3]<9999.9) {
	vmin=(vmin > dot[3] ? dot[3] : vmin);
	vmax=(vmax < dot[3] ? dot[3] : vmax);
      }
      ndot++;
    }
  }
  fprintf(stderr,"%d atoms, %d dots\n",nat,ndot);
  fprintf(stderr,"min/max on surface: %0.6f / %0.6f\n",vmin,vmax);

  return(0);
}

static void swapper(int n,void *data)
{
  char *p,c;
  int i;

  p=data;
  for (i=0;i<n;i++) {
    c=p[0]; p[0]=p[3]; p[3]=c;
    c=p[1]; p[1]=p[2]; p[2]=c;
    p+=4;
  }
}
