      program his2cnf
c.......................................................................
c Convert Discover History file to Wit!P .cnf File.
c
c Usage: his2cnf infile [ outfile ] [ options ]
c
c   infile:    inputfile in .his format
c   outfile:   outputfile n .cnf format, default: - = stdout,
c   options:
c             -f #  : first frame
c             -t #  : last frame
c             -s #  : output every #-th frame only
c             -m    : output only mobile atoms (default)
c             -a    : output all atoms (negates -m)
c             -p    : peek only (list header record for each selected frame)
c             -h    : print this help text
c.......................................................................

      implicit none

      integer mxnat
      parameter (mxnat=30000)

      real*8 vers,d,dt,et,ek,ep
      real*4 x(mxnat),y(mxnat),z(mxnat)
      character*4 cv(20),titl(20),stitl(20)
      character out*60,str*60,cmd*132
      integer kntr,nat,i,j,k,nmov,mov(mxnat),nmol,nres,freq,start,
     .     nframe,nrd,nbonds,wrt(mxnat),nout

      integer iargc,narg,nf,first,last,step
      logical mobile,peek
c.......................................................................
      
      mobile=.true.
      peek=.false.
      nf=0
      nframe=0
      first=1
      last=2000000000
      step=1
      narg=iargc()
      open(unit=5,form='unformatted')

      i=1
      do while (i.le.narg) 
         call getarg(i,cmd)
         i=i+1
         if (cmd(1:2).eq.'-m') then
            mobile=.true.
         else if (cmd(1:2).eq.'-a') then
            mobile=.false.
         else if (cmd(1:2).eq.'-p') then
            peek=.true.
         else if (cmd(1:2).eq.'-f') then
            cmd=cmd(3:)
            if (cmd.eq.' ') then
               if (i.le.narg) then
                  call getarg(i,cmd)
                  i=i+1
               end if
            end if
            read(cmd,*) first
         else if (cmd(1:2).eq.'-t') then
            cmd=cmd(3:)
            if (cmd.eq.' ') then
               if (i.le.narg) then
                  call getarg(i,cmd)
                  i=i+1
               end if
            end if
            read(cmd,*) last
            if (last.lt.1) last=2000000000
         else if (cmd(1:2).eq.'-s') then
            cmd=cmd(3:)
            if (cmd.eq.' ') then
               if (i.le.narg) then
                  call getarg(i,cmd)
                  i=i+1
               end if
            end if
            read(cmd,*) step
            if (step.lt.1) step=1
         else if (cmd.eq.'-') then
            nf=nf+1
         else if (cmd(1:1).ne.'-') then
            if (nf.eq.0) then
               close(unit=5)
               open(unit=5,form='unformatted',file=cmd,
     .              status='old',err=3000)
               write(0,'(a)') ' in = '//cmd(1:72)
            else if (nf.eq.1) then
               close(unit=6)
               open(unit=6,file=cmd,status='unknown',err=4000)
               write(0,'(a)') ' out= '//cmd(1:72)
            endif
            nf=nf+1
         else
            if (cmd(1:2).ne.'-h') then
               write(0,'(a)') ' unknown option: '//cmd(1:20)
            endif
            write(0,'(a)')
     .        'Convert Discover History file to Wit!P .cnf File.',
     .        ' ',
     .        'Usage: his2cnf infile [ outfile ] [ options ]',
     .        ' ',
     .        '  infile:    inputfile in .his format',
     .        '  outfile:   outputfile n .cnf format,'
     .           //'default: - = stdout,',
     .        '  options:',
     .        '            -f #  : first frame',
     .        '            -t #  : last frame',
     .        '            -s #  : output every #-th frame only',
     .        '            -m    : output only mobile atoms (default)',
     .        '            -a    : output all atoms (negates -m)',
     .        '            -p    : peek only (list header record'
     .           //'for each selected frame)',
     .        '            -h    : print this help text'
            call exit(3)
         end if
      end do

      read(5,end=1000,err=1000) kntr
      read(5,end=1000,err=1000) cv,vers
      read(5,end=1000,err=1000) titl
      read(5,end=1000,err=1000) stitl

      read(5,end=1000,err=1000) i
      read(5,end=1000,err=1000) i
      read(5,end=1000,err=1000) nat
      if (nat.gt.mxnat) go to 2000

      read(5,end=1000,err=1000) i,nmov,(mov(k),k=1,min(mxnat,nmov))
      if (nmov.gt.mxnat) go to 2000
      read(5,end=1000,err=1000) nmol
      read(5,end=1000,err=1000) i,nres

      read(5,end=1000,err=1000) nbonds
      read(5,end=1000,err=1000) d
      read(5,end=1000,err=1000) d   ! unk rec ???
      read(5,end=1000,err=1000) i,dt,freq,start

      if (mobile) then
         nout=nmov
         do i=1,nmov
            wrt(i)=mov(i)
         end do
      else
         nout=nat
         do i=1,nat
            wrt(i)=i
         end do
      end if

      nrd=nat
      if (vers.gt.2.55) nrd=nmov

c now for the first frame...

      read(5,end=200,err=1000) et,ep,ek
      read(5,end=200,err=1000) (x(i),y(i),z(i),i=1,nat)
      read(5,end=200,err=1000) d

      nframe=nframe+1
      if (nframe.lt.first.or.mod(nframe-first,step).ne.0) go to 100

      write(6,'(a,i5,a,f10.3,a,i7)') 'Conformation ',nframe,
     .     ' Energy = ',et,' Status = History Time = ',
     .     freq*(nframe+start-1)


      if (.not.peek) then
         do j=1,nout
            write(str,'(3f20.3)') x(wrt(j)),y(wrt(j)),z(wrt(j))
            out=' '
            k=1
            do i=1,60
               if (str(i:i).ne.' ') then
                  k=k+1
                  out(k:k)=str(i:i)
               else
                  if (out(k:k).ne.' ') then
                     k=k+1
                     out(k:k)=str(i:i)
                  end if
               end if
            end do
            write(6,'(a)') out(1:k)
         end do
      endif

c ...and the rest of the frames...

 100  continue

      if (nframe.ge.last) go to 200

      read(5,end=200,err=200) kntr
      if (kntr.eq.0) go to 200
      read(5,end=200,err=200) et,ep,ek
      read(5,end=200,err=200) d
      read(5,end=200,err=200) (x(mov(i)),y(mov(i)),z(mov(i)),i=1,nrd)
      read(5,end=200,err=200) d

      nframe=nframe+1
      if (nframe.lt.first.or.mod(nframe-first,step).ne.0) go to 100

      write(6,'(a,i5,a,f10.3,a,i7)') 'Conformation ',nframe,
     .     ' Energy = ',ep,' Status = History Time = ',
     .     freq*(nframe+start-1)


      if (.not.peek) then
         do j=1,nout
            write(str,'(3f20.3)') x(wrt(j)),y(wrt(j)),z(wrt(j))
            out=' '
            k=1
            do i=1,60
               if (str(i:i).ne.' ') then
                  k=k+1
                  out(k:k)=str(i:i)
               else
                  if (out(k:k).ne.' ') then
                     k=k+1
                     out(k:k)=str(i:i)
                  end if
               end if
            end do
            write(6,'(a)') out(1:k)
         end do
      end if
      go to 100

 200  call exit(0)

 1000 continue
      write(0,'(a)') ' Unexpected EOF or mishandled error condition.'
      call exit(1)

 2000 continue
      write(0,'(a)') ' Too many atoms/mobile atoms.'
      call exit(2)

 3000 continue
      write(0,'(a)') ' Error: could not open input file for read:'
      write(0,'(a)') cmd(1:60)
      call exit(4)

 4000 continue
      write(0,'(a)') ' Error: could not open output file for write:'
      write(0,'(a)') cmd(1:60)
      call exit(5)
      
      end
