#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Converts Amber OWFEG grid file Sybyl MOL2 file */

int SGI_i(int i);
int SGI_f(float f);

int main(int argc,char **argv)
{
  FILE *in,*out;
  float xmin[3],dx[3],*V,*val,miss=1.e31,y,z;
  double v;
  int i,j,k,nx[3],n,ng,status,nat;
  char *end,buf[4096],fld[20];

  if (argc<3) {
    fprintf(stderr,"usage: owfeg2mol2 in_file out_file\n");
    return(1);
  }

  in=fopen(argv[1],"r");
  if (!in) {
    fprintf(stderr,"could not open contour file %s for read\n",argv[1]);
    return (1);
  }

  status=0; *buf='\0';
  fgets(buf,sizeof(buf),in);
  if (strlen(buf)>=63) {
    for (i=0;i<3;i++) {
      strncpy(fld,buf+i*8,8); fld[8]='\0';
      xmin[i]=strtod(fld,&end);
      status|=(end!=fld+8);
      strncpy(fld,buf+32+i*10,10); fld[10]='\0';
      nx[i]=strtol(fld,&end,10);
      status|=(end!=fld+10);
    }
    strncpy(fld,buf+24,8); fld[8]='\0';
    dx[0]=strtod(fld,&end);
    status|=(end!=fld+8);
  }
  if (status) {
    fprintf(stderr,"owfeg2mol2: 1st header, format error.\n");
    return(1);
  }
  dx[1]=dx[0]; dx[2]=dx[0];
  nx[0]++; nx[1]++; nx[2]++; 
  n=nx[0]*nx[1]*nx[2];

  val=(float *)malloc(n*sizeof(float));
  if (!val) {
    fprintf(stderr,"owfeg2mol2: could not allocate grid of size %d\n",n);
    return(2);
  }

  ng=0;
  while (!feof(in)) {
    for (;!status&&fgets(buf,sizeof(buf),in);) {
      if (strncasecmp(buf,"Start",5)) {
	fprintf(stderr,"owfeg2mol2: input format error %s\n",buf);
	return(3);
      }else{
	fprintf(stderr,"%s",buf);
	for (k=0,j=0;k<n&&!status;k++,j=(j+1)%nx[0]) {
	  if (j%10==0) {    /* 10 values per line */
	    *buf='\0'; fgets(buf,sizeof(buf),in);
	  }
	  strncpy(fld,buf+(j%10)*8,8);
	  v=strtod(fld,&end);
	  status|=(end!=fld+8);
	  if (v-888.888<0.001&&v-888.888>-0.001) v=miss;
	  val[k]=v;
	}
	if (status) {
	  fprintf(stderr,"owfeg2mol2: format error, grid %d, point %d\n",ng+1,k+1);
	}
	sprintf(buf,"%s_%d.mol2",argv[2],ng+1);
	out=fopen(buf,"w");
	if (out) {
	  for (nat=0,k=0;k<n;k++) if (val[k]!=miss) nat++;
	  fprintf(out,"# TRIPOS MOL2 file generated by owfeg2mol2\n"
		  "\n"
		  "@<TRIPOS>MOLECULE\n"
		  "%s_%d\n"
		  "%d 0 0 0 0\n"
		  "****\n"
		  "USER_CHARGES\n"
		  "INVALIS_CHARGES\n"
		  "\n"
		  "@<TRIPOS>ATOM\n",
		  argv[2],ng+1,nat);
	  for (nat=0,k=0,V=val;k<nx[2];k++) {
	    z=k*dx[2]+xmin[2];
	    for (j=0;j<nx[1];j++) {
	      y=j*dx[1]+xmin[1];
	      for (i=0;i<nx[0];i++,V++) {
		if (*V==miss) continue;
		fprintf(out,"%7d C   %10.4f %10.4f %10.4f C.3 0 **** %10.4f\n",
			++nat,i*dx[0]+xmin[0],y,z,*V);
	      }
	    }
	  }
	  fclose(out);
	  ng++;
	}else{
	  fprintf(stderr,"owfeg2mol2: failed to open output file %s\n",buf);
	  status=7;
	}
      }
    }
  }
  if (val) free(val);
  return(0);
}
