#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Converts Amber OWFEG grid file to Wit!P grid file */

int SGI_i(int i);
int SGI_f(float f);

int main(int argc,char **argv)
{
  FILE *in,*out;
  float xmin[3],dx[3],miss;
  double v;
  int i,j,k,nx[3],n,ng,*val,h[9],status;
  char *end,buf[4096],fld[20];

  if (argc<3) {
    fprintf(stderr,"usage: owfeg2witcnt in_file out_file [wit_miss_val]\n");
    return(1);
  }
  if (argc>3) miss=atof(argv[3]);
  else miss=1.e31;
  if (miss<0) miss=-miss;

  in=fopen(argv[1],"r");
  if (!in) {
    fprintf(stderr,"could not open contour file %s for read\n",argv[1]);
    return (1);
  }

  status=0; *buf='\0';
  fgets(buf,sizeof(buf),in);
  if (strlen(buf)>=63) {
    for (i=0;i<3;i++) {
      strncpy(fld,buf+i*8,8); fld[8]='\0';
      xmin[i]=strtod(fld,&end);
      status|=(end!=fld+8);
      strncpy(fld,buf+32+i*10,10); fld[10]='\0';
      nx[i]=strtol(fld,&end,10);
      status|=(end!=fld+10);
    }
    strncpy(fld,buf+24,8); fld[8]='\0';
    dx[0]=strtod(fld,&end);
    status|=(end!=fld+8);
  }
  if (status) {
    fprintf(stderr,"owfeg2witcnt: 1st header, format error.\n");
    return(1);
  }
  dx[1]=dx[0]; dx[2]=dx[0];
  nx[0]++; nx[1]++; nx[2]++; 
  n=nx[0]*nx[1]*nx[2];

  val=(int *)malloc(n*sizeof(float));
  if (!val) {
    fprintf(stderr,"owfeg2witcnt: could not allocate grid of size %d\n",n);
    return(2);
  }

  ng=0;
  while (!feof(in)) {
    for (;!status&&fgets(buf,sizeof(buf),in);) {
      if (strncasecmp(buf,"Start",5)) {
	fprintf(stderr,"owfeg2witcnt: input format error %s\n",buf);
	return(3);
      }else{
	for (k=0,j=0;k<n&&!status;k++,j=(j+1)%nx[0]) {
	  if (j%10==0) {    /* 10 values per line */
	    *buf='\0'; fgets(buf,sizeof(buf),in);
	  }
	  strncpy(fld,buf+(j%10)*8,8);
	  v=strtod(fld,&end);
	  status|=(end!=fld+8);
	  if (v-888.888<0.001&&v-888.888>-0.001) v=miss;
	  val[k]=SGI_f(v);
	}
	if (status) {
	  fprintf(stderr,"owfeg2witcnt: format error, grid %d, point %d\n",ng+1,k+1);
	  return(3);
	}
	sprintf(buf,"%s_%d.cnt",argv[2],ng+1);
	out=fopen(buf,"w");
	h[0]=SGI_f(xmin[0]);h[1]=SGI_f(xmin[1]);h[2]=SGI_f(xmin[2]);
	h[3]=SGI_f(dx[0]);  h[4]=SGI_f(dx[1]);  h[5]=SGI_f(dx[2]);
	h[6]=SGI_i(nx[0]);  h[7]=SGI_i(nx[1]);  h[8]=SGI_i(nx[2]);
	status=fwrite(h,sizeof(int),9,out); 
	status+=fwrite(val,sizeof(int),n,out);
	status+=fclose(out);
	ng++;
	if (status!=n+9) {
	  fprintf(stderr,"owfeg2witcnt: error writing output file %s, status %d\n",buf,status);
	}else status=0;
      }
    }
  }
  if (val) free(val);
  return(0);
}
/*===============================================================*/

/*    Routines to convert between MY format and SGI              */
/*    NOTE: SGI_f is an integer to avoid problems with NAN's ... */

static union {
  int i;
  float f;
  unsigned char c[4];
} u_swap;

/*********************************************************/

#if defined(sgi) || defined(darwin)

int SGI_f(float x)   /* convert MY float to SGI float */
{
  u_swap.f=x;
  return(u_swap.i);
}

int SGI_i(int i)     /* convert MY int to SGI int */
{
  return(i);
}

#endif

/******************************************************************/

#if defined(linux) || defined(cygwin) || defined(mingw)

int SGI_f(float x)
{
  unsigned char c;
  
  u_swap.f=x;
  c=u_swap.c[0];
  u_swap.c[0]=u_swap.c[3];
  u_swap.c[3]=c;
  c=u_swap.c[1];
  u_swap.c[1]=u_swap.c[2];
  u_swap.c[2]=c;
  return(u_swap.i);
}

int SGI_i(int x)
{
  unsigned char c;
  
  u_swap.i=x;
  c=u_swap.c[0];
  u_swap.c[0]=u_swap.c[3];
  u_swap.c[3]=c;
  c=u_swap.c[1];
  u_swap.c[1]=u_swap.c[2];
  u_swap.c[2]=c;
  return(u_swap.i);
}

#endif
