#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#define TRUE 1
#define FALSE 0
#define mm2pts 2.835  /* points per mm */

/* Filter to convert SGP Postscript file to enacapsulated postscript 

   -L   : lansdscape 
   -P   : portrait
   -o x y : offset of origin (mm)
   -xs  s : x-szie (mm)
   -ys  s : y-size (mm)

*/

int main(int argc,char **argv)
{
  char rec[512];
  float xd,yd,xs,ys,xo,yo;
  float PLLx,PLLy,PURx,PURy;
  int i,landscape,LLx,LLy,URx,URy;

  /* set up defaults... */

  landscape=TRUE;
  xs=-1; ys=-1; xo=-1; yo=-1;

  /* interpret command line arguments ... */

  for (i=1;i<argc;i++) {
    if (strcmp(argv[i],"-L")==0) landscape=TRUE;
    else if (strcmp(argv[i],"-P")==0) landscape=FALSE;
    else if (strcmp(argv[i],"-o")==0) {
      if (i<argc+2) {
	sscanf(argv[++i],"%f",&xo);
	sscanf(argv[++i],"%f",&yo);
      }else{
	fprintf(stderr,"bad -o option\n");
	return(1);
      }
    }else if (strcmp(argv[i],"-xs")==0) {
      if (i<argc+1) {
	sscanf(argv[++i],"%f",&xs);
      }else{
	fprintf(stderr,"bad -xs option\n");
	return(1);
      }
    }else if (strcmp(argv[i],"-ys")==0) {
      if (i<argc+1) {
	sscanf(argv[++i],"%f",&ys);
      }else{
	fprintf(stderr,"bad -ys option\n");
	return(1);
      }
    }else{
      if (strcmp(argv[i],"-h")!=0) fprintf(stderr,"bad option: %s\n",argv[i]);
      fprintf(stderr,"valid options are:\n");
      fprintf(stderr,"-L : landscape\n");
      fprintf(stderr,"-P : portrait\n");
      fprintf(stderr,"-o xoffset yoffset (mm)\n");
      fprintf(stderr,"-xs xsize (mm)\n");
      fprintf(stderr,"-ys ysize (mm)\n");
      fprintf(stderr,"(default: -L -o 25 10 -xs 247 -ys 190)\n");
      return(1);
    }
  }
  
  /* get header records */
    
  rec[0]='\0';
  fgets(rec,sizeof(rec),stdin); /* EPS header */

  fgets(rec,sizeof(rec),stdin); /* bounding box */
  if (strncmp(rec,"%%BoundingBox",strlen("%%BoundingBox"))!=0) {
    fprintf(stderr,"input file is not in SGP EPS format\n");
    return (1);
  }

  fgets(rec,sizeof(rec),stdin); /* SGP header */
  if (strncmp(rec,"%SGP Postscript",strlen("%SGP Postscript"))!=0) {
    fprintf(stderr,"input file is not in SGP EPS format\n");
    return (1);
  }
  sscanf(rec+strlen("%SGP Postscript"),"%f %f",&xd,&yd);
  
  fgets(rec,sizeof(rec),stdin); /* initial CTM, should check really ...*/
  if (feof(stdin)) {
    fprintf(stderr,"premature EOF\n");
    return(1);
  }

  /* determine bounding box */
  
  if (xo<=0) xo=(landscape ? 25.0 : 10.0 );
  if (yo<=0) yo=(landscape ? 10.0 : 25.0 );
  
  if (xs<0 && ys <0) {
    if (landscape) {
      if (xd/yd>247/190) xs=247;
      else ys=190;
    }else{
      if (xd/yd<190/247) ys=247;
      else xs=190; 
    }
  }

  if (xs<=0) xs = ( ys > 0 ? ys*xd/yd : (landscape ? 247 : 190));
  if (ys<=0) ys = xs*yd/xd;
  
  if (landscape) {
    PLLx = 210-yo; PLLy = xo;
    PURx = PLLx-ys; PURy = PLLy+xs;
  }else{
    PLLx = xo; PLLy = yo;
    PURx = PLLx+xs; PURy = PLLy+ys;
  }
  
  PLLx *= mm2pts; PLLy *= mm2pts; PURx *= mm2pts; PURy *= mm2pts;
  LLx = (PLLx < PURx ? PLLx : PURx);
  LLy = (PLLy < PURy ? PLLy : PURy);
  URx = (PURx > PLLx ? PURx : PLLx) + 0.99999;
  URy = (PURy > PLLy ? PURy : PLLy) + 0.99999;
  
  fprintf(stdout,"%%!PS-Adobe-2.0 EPSF-2.0\n");
  fprintf(stdout,"%%%%BoundingBox: %d %d %d %d\n",LLx,LLy,URx,URy);
  fprintf(stdout,"%%SGP Postscript %f %f\n",xd,yd);
    
  if (landscape) {
    fprintf(stdout,"[ %7.1f %7.1f %7.1f %7.1f %7.1f %7.1f ] concat\n",
	    0.0,(PURy-PLLy)/xd,
	    (PURx-PLLx)/yd,0.0,
	    PLLx,PLLy);
  }else{
    fprintf(stdout,"[ %7.1f %7.1f %7.1f %7.1f %7.1f %7.1f ] concat\n",
	    (PURx-PLLx)/xd,0.0,
	    0.0,(PURy-PLLy)/yd,
	    PLLx,PLLy);
  }
  
  /* just copy the rest of the file ... */

  fgets(rec,sizeof(rec),stdin);
  while (!feof(stdin)) {
    fputs(rec,stdout);
    fgets(rec,sizeof(rec),stdin);
  }

  return(0);
}
