#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <sgpdef.h>
/*
  
  
  Usage: sw_plot [-d <device>] [<file>]

  where: <device> one of HP7500 (default), HP7200, EPS, cEPS
         <file> name of device independent plot file (default stdin).


  Format of the device independent file is as follows:
  
  
  byte 0,1      :  magic number (19937)
  byte 2,3,4,5  :  x-size of window  (long)
  byte 6,7,8,9  :  y-size of window  (long)
  
  bytes 10 to end : commands + parameters
  
  
  commands:
  
  V x[3],y[3]   : vector from x to y (float)
  M x[3]	 : move to x (float)
  D x[3]	 : draw to y (float)
  C color       : set color (short)
  l		 : begin line
  e		 : end line
  3 x[3]        : 3-D point
  P x1,x2 y1,y2 : viewport, fraction of window.
  v v[4][4]     : new model/viewing matrix (float)
  p p[4][4]     : new projection matrix (float)
  z near,far    : position of z-clip planes (far < near < 0) in world coord.
  L width       : linewidth
  s style       : set linestyle 
  X x[3]        : draw a dot at x
  E             : end
  
  
  The program always does a page throw at the end.
  
  */


/* ------------------------------------------------------------------------- */

FILE *Plot = (FILE *)NULL;
FILE *dbg = (FILE *)NULL;

typedef float Matrix[4][4];

Matrix vm,pm,vpm;
float near,far;

#define MAXCOL 9
#define MAXDP 9

int coltab[MAXCOL] = {
  1,2,3,7,4,6,5,1,8
  };

int dp_tab[MAXDP] = {
  0,2,1,3,4,5,6,7,8
  };

/* black red green yellow blue magenta cyan white  8 in WITNOTP go to
       1   2     3      7   4        6    5     1  8 on the HP Plotter */    

#define PlRead(rec,size) fread((char *)rec,1,(int)(size),Plot)

void Vec3D(float *x,float *y);
void M4mult(Matrix a,Matrix b,Matrix c);
void V4mult(float *y,float *x,Matrix a);

int main(int argc,char **argv)
{
  int i,j,n;
  int dev;
  int LineMode;
  int xs,ys;
  short col,magic,lw,dp;
  float xsz,ysz,xl,xh,yl,yh,vec[6],xold[3],xnew[3],vwp[4];
  float nvx1,nvx2,nvy1,nvy2;
  char cmd,*arg,*filename;
  
  dbg=(FILE *)NULL;;
  dev=HP7500;	        	    /* default plotter */
  filename=(char *)NULL;

  for (n=1;n<argc;n++) {
    arg=argv[n];
    if (strncmp(arg,"-d",2)==0) {
      if (strlen(arg)==2) {
	if (++n<argc) arg=argv[n];
	else {
	  fprintf(stderr,"bad -d option (missing argument)\n");
	  return(1);
	}
      }else{
	arg=arg+2;
      }
      if (strncmp(arg,"HP75",4)==0||strncmp(arg,"hp75",4)==0) {
	dev=HP7500;
      }else if (strncmp(arg,"HP72",4)==0||strncmp(arg,"hp72",4)==0) {
	dev=HP7200;
      }else if (strncmp(arg,"EPS",3)==0||strncmp(arg,"eps",3)==0) {
	dev=EPSCRIPT;
      }else if (strncmp(arg,"cEPS",4)==0||strncmp(arg,"ceps",4)==0) {
	dev=EPSCOLOR;
      }else{
	fprintf(stderr,"bad argument of -d option\n");
	return(2);
      }
    }else if (strncmp(arg,"-D",2)==0) {
      if (strlen(arg)>2) {
	dbg=fopen(arg+2,"w");
	if (!dbg) {
	  fprintf(stderr,"plot: sorry, could not open %s for write.\n",arg+2);
	  dbg=stderr;
	}
      }else dbg=stderr;
    }else{
      filename=arg;
    }
  }

  if (filename) {
    Plot=fopen(filename,"r");
    if (!Plot) {
      fprintf(stderr,"Could not open %s for read.\n",filename);
      return(3);
    }  
  }else{
    Plot=stdin;
  }

  PlRead(&magic,sizeof(short));
  if (magic!=19937) {
    fprintf(stderr,"wrong magic number...\n");
    return(4);
  }
  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) vpm[i][j]=vm[i][j]=pm[i][j]=0.0;
    vpm[i][i]=vm[i][i]=pm[i][i]=1.0;
  }
  PlRead(&xs,sizeof(int));
  PlRead(&ys,sizeof(int));
  if (dbg) fprintf(dbg,"Window size = %d,%d\n",xs,ys);
  if (xs<=0 || ys<=0) {
    fprintf(stderr,"Bad size of window (x,s) = %d,%d \n",xs,ys);
  }
  xsz=xs,ysz=ys;
  if (xsz>1.3*ysz) {
    xl=0.1*130.0;
    xh=130.0;
    yl=0.55*100-0.45*100*1.3*ysz/xsz;
    yh=0.55*100+0.45*100*1.3*ysz/xsz;
  }else{
    yl=0.1*100;
    yh=100.0;
    xl=0.55*130-0.45*130*xsz/(1.3*ysz);
    xh=0.55*130+0.45*130*xsz/(1.3*ysz);
  }
  near = -1.0;
  far = -2.0;
  
  grstrt(dev);
  
  vwport(xl,xh,yl,yh);
  window(-1,1,-1,1);
  LineMode = 0;
  
  PlRead(&cmd,sizeof(char));
  
  while(cmd!='E') {
    if (feof(Plot)) break;
    switch (cmd) {
    case 'P' :
      PlRead(vwp,4*sizeof(float));
      nvx1=xl+(xh-xl)*vwp[0];
      nvx2=xl+(xh-xl)*vwp[1];
      nvy1=yl+(yh-yl)*vwp[2];
      nvy2=yl+(yh-yl)*vwp[3];
      if (dbg) fprintf(dbg,"viewport %0.3f,%0.3f  %0.3f,%0.3f\n",nvx1,nvy1,nvx2,nvy2);
      vwport(nvx1,nvx2,nvy1,nvy2);
      break;
    case 'V' :
      PlRead(vec,6*sizeof(float));
      Vec3D(vec,vec+3);
      if (dbg) fprintf(dbg,"Vector %0.3f,%0.3f,%0.3f  %0.3f,%0.3f,%0.3f \n",
		       vec[0],vec[1],vec[2],vec[3],vec[4],vec[5]);
      for (i=0;i<3;i++) xold[i]=vec[3+i];
      break;
    case 'M' :
      PlRead(xold,3*sizeof(float));
      if (dbg) fprintf(dbg,"Move   %0.3f,%0.3f,%0.3f\n",xold[0],xold[1],xold[2]);
      break;
    case 'D' :
      PlRead(xnew,3*sizeof(float));
      Vec3D(xold,xnew);
      for (i=0;i<3;i++) xold[i]=xnew[i];
      if (dbg) fprintf(dbg,"Draw   %0.3f,%0.3f,%0.3f\n",xold[0],xold[1],xold[2]);
      break;
    case 'X' :
      PlRead(xold,3*sizeof(float));
      Vec3D(xold,xold);
      if (dbg) fprintf(dbg,"Point  %0.3f,%0.3f,%0.3f\n",xold[0],xold[1],xold[2]);
      break;
    case 'l' :
      LineMode=1;
      if (dbg) fprintf(dbg,"BeginLine\n");
      break;
    case 'e' :
      LineMode=0;
      if (dbg) fprintf(dbg,"EndLine\n");
      break;
    case 'L' :
      PlRead(&lw,sizeof(short));
      linewidth(lw);
      if (dbg) fprintf(dbg,"Linewidth %d\n",lw);
      break;
    case 's' :
      PlRead(&dp,sizeof(short));
      dp = (dp < 0 || dp >= MAXDP ? 0 : dp );
      dashpt(dp);
      if (dbg) fprintf(dbg,"Dashpattern %d\n",dp);
      break;
    case '3' :
      if (LineMode == 1) {
	PlRead(xold,3*sizeof(float));
	LineMode = 2;
      }else if (LineMode == 2) {
	PlRead(xnew,3*sizeof(float));
	Vec3D(xold,xnew);
	for (i=0;i<3;i++) xold[i]=xnew[i];
      }else{
	PlRead(xold,3*sizeof(float));
      }
      if (dbg) fprintf(dbg,"Vertex %0.3f,%0.3f,%0.3f\n",xold[0],xold[1],xold[2]);
      break;
    case 'C' :
      PlRead(&col,sizeof(short));
      if (dbg) fprintf(dbg,"Color %d\n",col);
      col=(col<1 || col >=MAXCOL ? 1 : col);
      linclr(coltab[col]);		    
      break;
    case 'v' :
      PlRead(vm,16*sizeof(float));
      if (dbg) {
	fprintf(dbg,"ViewMatrix\n");
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",vm[0][0],vm[1][0],vm[2][0],vm[3][0]);
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",vm[0][1],vm[1][1],vm[2][1],vm[3][1]);
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",vm[0][2],vm[1][2],vm[2][2],vm[3][2]);
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",vm[0][3],vm[1][3],vm[2][3],vm[3][3]);
      }
      M4mult(vpm,vm,pm);
      break;
    case 'p' :
      PlRead(pm,16*sizeof(float));
      if (dbg) {
	fprintf(dbg,"ProjectionMatrix\n");
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",pm[0][0],pm[1][0],pm[2][0],pm[3][0]);
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",pm[0][1],pm[1][1],pm[2][1],pm[3][1]);
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",pm[0][2],pm[1][2],pm[2][2],pm[3][2]);
	fprintf(dbg,"  %9.4f %9.4f %9.4f %9.4f \n",pm[0][3],pm[1][3],pm[2][3],pm[3][3]);
      }
      M4mult(vpm,vm,pm);
      break;
    case 'z' :
      PlRead(&near,sizeof(float));
      PlRead(&far,sizeof(float));
      if (dbg) fprintf(dbg,"Z-clip %0.3f, %0.3f\n",near,far);
      break;
    default: break;
    }
    PlRead(&cmd,sizeof(char));
  }
  /* why this ?   unlink(argv[n]); */
  if (dev==HP7200) newpag();
  grstop();
  return(0);
}

void Vec3D(float *x,float *y)
{
  short i,clipx,clipy,visible;
  
  float xw[4],yw[4],xs[4],ys[4],xc[4],yc[4];
  float t;
  
  for (i=0;i<3;i++) {
    xw[i]=x[i];
    yw[i]=y[i];
  }
  xw[3]=1.0;
  yw[3]=1.0;
  V4mult(xc,xw,vm);
  V4mult(yc,yw,vm);
  
  clipx=(xc[2]>near?1:0); clipx|=((xc[2]<far?1:0)<<1); /* in two steps, to make MipsPro7.2 happy */
  clipy=(yc[2]>near?1:0); clipy|=((yc[2]<far?1:0)<<1);
  visible = ( !(clipx||clipy) || !(clipx&clipy));
  if (visible) {
    switch (clipx) {
    case 1 :
      t=(near-xc[2])/(yc[2]-xc[2]);
      xc[0]=xc[0]+t*(yc[0]-xc[0]);
      xc[1]=xc[1]+t*(yc[1]-xc[1]);
      xc[2]=near;
      break;
    case 2 :
      t=(far-xc[2])/(yc[2]-xc[2]);
      xc[0]=xc[0]+t*(yc[0]-xc[0]);
      xc[1]=xc[1]+t*(yc[1]-xc[1]);
      xc[2]=far;
      break;
    default: 
      break;
    }
    switch (clipy) {
    case 1 :
      t=(near-yc[2])/(xc[2]-yc[2]);
      yc[0]=yc[0]+t*(xc[0]-yc[0]);
      yc[1]=yc[1]+t*(xc[1]-yc[1]);
      yc[2]=near;
      break;
    case 2 :
      t=(far-yc[2])/(xc[2]-yc[2]);
      yc[0]=yc[0]+t*(xc[0]-yc[0]);
      yc[1]=yc[1]+t*(xc[1]-yc[1]);
      yc[2]=far;
      break;
    default: 
      break;
    }
    V4mult(xs,xc,pm);
    V4mult(ys,yc,pm);
    move(xs[0]/xs[3],xs[1]/xs[3]);
    draw(ys[0]/ys[3],ys[1]/ys[3]);
  }
}

void M4mult(Matrix a,Matrix b,Matrix c)
{
  int i,j,k;    
  
  for (i=0;i<4;i++) {
    for (k=0;k<4;k++) {
      a[i][k]=0.0;
      for (j=0;j<4;j++) {
	a[i][k]+=b[i][j]*c[j][k];
      }
    }
  }
}

void V4mult(float *y,float *x,Matrix a)
{
  int i,j;
  
  for (j=0;j<4;j++) {
    y[j]=0.0;
    for (i=0;i<4;i++) y[j]+=x[i]*a[i][j];	
  }
}
