#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Converts TA grid file to Wit!P grid file, but works for
   cell = 1 1 1 90 90 90 only !!!
*/

FILE *SYB_CFILE_OPEN_READ(char *file);

void SYB_CFILE_GET_PARAMETERS(FILE *file,char *source,double cell[6],
			     double delta[3],double limits[6],
			     double dnsl[2],int *nc,int *nr,int *ns,int *ss);
int SYB_CFILE_READ(FILE *file,int shtnr,double *sheet);
int SYB_CFILE_CLOSE(FILE *file);

int main(int argc,char **argv)
{
  FILE *in,*out;
  float xmin[3],dx[3],*val,*v,miss,c[6];
  int nx[3],n,i,j,k,nc,nr,ns,ss;
  char source[512];
  double cell[6],delta[3],limits[6],dnsl[2];
  double *sheet,*s;

  if (argc<3) {
    fprintf(stderr,"usage: ta2witcnt in_file out_file [wit_miss_val]\n");
    return(1);
  }
  if (argc>3) miss=atof(argv[3]);
  else miss=1.e31;
  if (miss<0) miss=-miss;

  in=SYB_CFILE_OPEN_READ(argv[1]);
  if (!in) {
    fprintf(stderr,"could not open contour file %s for read\n",argv[1]);
    return (1);
  }

  SYB_CFILE_GET_PARAMETERS(in,source,cell,delta,limits,dnsl,&nc,&nr,&ns,&ss);
  sheet=(double *)malloc(ss*sizeof(double));
  if (!sheet) {
    fprintf(stderr,"could not sheet of size %d\n",ss);
    return(2);
  }
  val=(float *)malloc(ns*ss*sizeof(float));
  if (!val) {
    fprintf(stderr,"could not allocate grid of size %d\n",ns*ss);
    return(2);
  }

  for (i=0;i<3;i++) {
    xmin[i]=limits[i];
    dx[i]=delta[i];
  }
  nx[0]=nc; nx[1]=nr; nx[2]=ns;

  if (strcmp(argv[2],"-")==0) out=stdout;
  else out=fopen(argv[2],"w");
  if (!out) {
    fprintf(stderr,"could not open output file %s\n",argv[2]);
    return(3);
  }
  fwrite(xmin,sizeof(float),3,out);
  fwrite(dx,sizeof(float),3,out);
  fwrite(nx,sizeof(int),3,out);
  n=nx[0]*nx[1]*nx[2];

  for (v=val,i=0;i<nx[2];i++) {
    if (!SYB_CFILE_READ(in,i,sheet)) break;
    for (s=sheet,j=0;j<nx[1];j++) {
      for (k=0;k<nx[0];k++) {
	if (*s>miss || *s<-miss) *v++ = miss;
	else *v++=*s;
	s++;
      }
    }
  }
  if (i<nx[2]) {
    fprintf(stderr,"failed to read sheet #%d\n",i);
    SYB_CFILE_CLOSE(in);
    return(7);
  }
  
  fwrite(val,sizeof(float),n,out);
  for (i=0;i<6;i++) c[i]=cell[i];
  fwrite(c,sizeof(float),6,out);

  SYB_CFILE_CLOSE(in);
  return(0);

}
