#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

FILE *SYB_CFILE_CREATE(char *file,char *source,double cell[6],
		       double delta[3],double limits[6]);

int SYB_CFILE_WRITE(FILE *file,int sheetno,double *sheet);
int SYB_CFILE_CLOSE(FILE *file);

int main(int argc,char **argv)
{
  FILE *in,*out;
  float xmin[3],dx[3],*val,*v,miss,c[6];
  double MISS=1.e40;
  int nx[3],n,i,j,k;
  
  double cell[6],delta[3],limits[6];
  double *sheet,*s;

  if (argc<3) {
    fprintf(stderr,"usage: wit2tacnt in_file out_file [wit_miss_val]\n");
    return(1);
  }

  if (strcmp(argv[1],"-")==0) in=stdin;
  else in=fopen(argv[1],"r");
  if (argc>3) miss=atof(argv[3]);
  else miss=1.e30;
  if (miss<0) miss=-miss;

  if (in) {
    if (3!=fread(xmin,sizeof(float),3,in)) {
      fprintf(stderr,"could not read 'xmin'\n");
      return(3);
    }
    if (3!=fread(dx,sizeof(float),3,in)) {
      fprintf(stderr,"could not read 'dx'\n");
      return(3);
    }
    if (3!=fread(nx,sizeof(int),3,in)) {
      fprintf(stderr,"could not read 'nx'\n");
      return(3);
    }
    n=nx[0]*nx[1]*nx[2];
    val=(float *)malloc(n*sizeof(float));
    if (!val) {
      fprintf(stderr,"could not allocate memory for grid\n");
      return(4);
    }
    if (n!=fread(val,sizeof(float),n,in)) {
      fprintf(stderr,"could not read contour file\n");
      return(5);
    }

    if (fread(c,sizeof(float),6,in)==6) {   /* try to read the unit cell */
      for (i=0;i<3;i++) {
	cell[i]=c[i]; cell[i+3]=c[i+3];
      }
    }else{
      for (i=0;i<3;i++) {
	cell[i]=1.0; cell[i+3]=90.0;
      }
    }

    sheet=(double *)malloc(nx[0]*nx[1]*sizeof(double));
    if (!sheet) {
      fprintf(stderr,"could not allocate memory for sheet\n");
      return(4);
    }

    for (i=0;i<3;i++) {
      delta[i]=dx[i];
      limits[i]=xmin[i]; limits[i+3]=xmin[i]+(nx[i]-0.999)*dx[i];
    }

    out=SYB_CFILE_CREATE(argv[2],"argv[1]",cell,delta,limits);
    if (!out) {
      fprintf(stderr,"could not open outfile %s\n",argv[2]);
      return(6);
    }

    for (v=val,i=0;i<nx[2];i++) {
      for (s=sheet,j=0;j<nx[1];j++) {
	for (k=0;k<nx[0];k++) {
	  *s = *v++;
	  if (*s < -miss || *s > miss) *s=MISS;
	  s++;
	}
      }
      if (!SYB_CFILE_WRITE(out,i,sheet)) break;
    }
    if (i<nx[2]) {
      fprintf(stderr,"failed to write sheet #%d\n",i);
      SYB_CFILE_CLOSE(out);
      return(7);
    }
    
    SYB_CFILE_CLOSE(out);
    return(0);

  }else{
    fprintf(stderr,"could not read input contour file %s\n",argv[1]);
    return(2);
  }
}
