/*  Modified from xfont (OpenGl Book),
    mfont <x_11_fontname>                                                      

    reads font and dumps character 32-126 in Wit!P font file format to stdout
*/

#include <stddef.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#if defined(darwin)
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <OpenGL/glx.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/Xaw/Form.h>
#include <X11/Shell.h>
#include "GLwDrawA.h"

GLuint base;

#define APPL_CLASS  "WnpFont"
#define GOOD	0
#define BAD	1

static XtAppContext appContext;
static Widget appShell;

XFontStruct *XF;
static GLXContext  GlBoxContext;
static Display    *GlBoxDisplay;
static Widget      GlBoxWidget;
static int         GlBoxWidth;
static int         GlBoxHeight;
static char       *font;

void makeRasterFont(char *font);
void printString(char *s);
Widget initGlBox(Widget parent);
Widget initWidgetTree(Widget shell);
static XVisualInfo *chooseVisual(Widget w);

/*==================== Callbacks ====================*/

void GlBoxGinitCB(Widget w,XtPointer clientData,XtPointer callbackData)
{
  Arg args[1];
  XVisualInfo *visinf;

  XtSetArg(args[0],GLwNvisualInfo,&visinf);
  XtGetValues(w,args,1);
  
  GlBoxDisplay=XtDisplay(w);
  GlBoxContext=glXCreateContext(GlBoxDisplay,visinf,0,TRUE);
  GlBoxWidget=w;
  GLwDrawingAreaMakeCurrent(GlBoxWidget,GlBoxContext);
  GlBoxWidth=((GLwDrawingAreaCallbackStruct*)callbackData)->width;
  GlBoxHeight=((GLwDrawingAreaCallbackStruct*)callbackData)->height;

  makeRasterFont (font);
  glShadeModel (GL_FLAT);  
}

void GlBoxResizeCB(Widget widg,XtPointer clientData,XtPointer callbackData)
{
  GlBoxWidth=((GLwDrawingAreaCallbackStruct*)callbackData)->width;
  GlBoxHeight=((GLwDrawingAreaCallbackStruct*)callbackData)->height;
}

void GlBoxExposeCB(Widget widg,XtPointer clientData,XtPointer callbackData)
{
  GLfloat white[3] = { 1.0, 1.0, 1.0};
  GLfloat fpix[10000];
  GLubyte fmap[10];
  int i,k,l;
  int W,w,h,px,py;
  char teststring[33];
  XCharStruct *XC;

  glViewport(0,0,GlBoxWidth,GlBoxHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0,GlBoxWidth,0.0,GlBoxHeight,-1,1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glReadBuffer(GL_FRONT);
  glClear(GL_COLOR_BUFFER_BIT);
  glClearColor(0,0,0,1);
  XC = XF->per_char - XF->min_char_or_byte2 + 32;

  for (i = 32; i < 127; i ++) {
    glClear(GL_COLOR_BUFFER_BIT);
    glColor3fv(white);
    glRasterPos2i(GlBoxWidth/2,GlBoxHeight/2);
    teststring[0] = (char) (i);
    teststring[1] = 0;
    printString(teststring);
    h=XC->ascent+XC->descent; w=XC->rbearing-XC->lbearing; W=(w+7)/8;
    px=GlBoxWidth/2+XC->lbearing; py=GlBoxHeight/2-XC->descent;
#ifdef GLXUSEFONTBUG
    py++;
#endif
    if (w+h>0) glReadPixels(px,py,w,h,GL_RED,GL_FLOAT,fpix);
    else {
      w=1;
      h=1;
      fpix[0]=0;
      W=1;
    }
    fprintf(stdout,"CHAR %3d %3d %3d %3d %3d %3d %3d\n",
	    i,w,h,-XC->lbearing,XC->descent,XC->width,0);
    for (k=0;k<h;k++) {
      for (l=0;l<W;l++) fmap[l]=0;
      for(l=0;l<w;l++) fmap[l/8]|=(fpix[k*w+l] < 0.5 ? 0 : (1<<(7-l%8)));
      for (l=0;l<W;l++) fprintf(stdout," %02x",fmap[l]);
    }
    fprintf(stdout,"\n");
    XC++;
  }
  glFlush ();
  exit(0);
}

/*==================== Functions ====================*/

void makeRasterFont(char *font)
{
    XFontStruct *fontInfo;
    Font id;
    unsigned int first, last;

    fontInfo = XLoadQueryFont(GlBoxDisplay,font);    
    if (fontInfo == NULL) {
        fprintf (stderr,"no font found\n");
	exit (0);
    }
    XF=fontInfo;
    fprintf(stdout,"FONT %3d %3d %3d %3d %s\n",
	    XF->ascent,XF->descent,32,126,font);

    id = fontInfo->fid;
    first = fontInfo->min_char_or_byte2;
    last = fontInfo->max_char_or_byte2;
    base = glGenLists((GLuint) last+1);
    if (base == 0) {
        printf ("out of display lists\n");
	exit (0);
    }
    glXUseXFont(id, first, last-first+1, base+first);
/*    *height = fontInfo->ascent + fontInfo->descent;
    *width = fontInfo->max_bounds.width;  */
}

void printString(char *s)
{
    glPushAttrib (GL_LIST_BIT);
    glListBase(base);
    glCallLists(strlen(s), GL_UNSIGNED_BYTE, (GLubyte *)s);
    glPopAttrib ();
}

/*-------------------- initWidgetTree --------------------*/

Widget initWidgetTree(Widget shell)
{
  Arg args[50];
  int n=0;
  Widget rootW;

  XtSetArg(args[n],XtNdefaultDistance,0); n++;
  rootW=XtCreateManagedWidget("main",formWidgetClass,shell,args,n);
  XtManageChild(rootW);
 
  initGlBox(rootW);
  fprintf(stderr,"initialized.\n");
  XtRealizeWidget(shell);
  fprintf(stderr,"realized.\n");
  
  return(rootW);
}
/*-------------------- initGlBox --------------------*/

Widget initGlBox(Widget parent)
{
  int n=0;
  Arg glbox_args[50];
  Widget box;
  XVisualInfo *vinf;

  XtSetArg(glbox_args[n],XtNtop,(XtArgVal)XtChainTop); n++;
  XtSetArg(glbox_args[n],XtNleft,(XtArgVal)XtChainLeft); n++;
  XtSetArg(glbox_args[n],XtNbottom,(XtArgVal)XtChainBottom); n++;
  XtSetArg(glbox_args[n],XtNright,(XtArgVal)XtChainRight); n++;

  vinf=chooseVisual(parent);
  if (vinf) {
    XtSetArg(glbox_args[n],GLwNvisualInfo,vinf); n++;
  }else{
    fprintf(stderr,"*** Warning *** could not determine appropriate visual.\n");
  }

  box=XtCreateManagedWidget("ThreeD",glwDrawingAreaWidgetClass,parent,
			    glbox_args,n);

  XtAddCallback(box,GLwNginitCallback,GlBoxGinitCB,(XtPointer)NULL);
  XtAddCallback(box,GLwNexposeCallback,GlBoxExposeCB,(XtPointer)NULL);
  XtAddCallback(box,GLwNresizeCallback,GlBoxResizeCB,(XtPointer)NULL);

  XtManageChild(box);

  return(box);
}
/*====================== Main =====================*/

int main(int argc, char** argv)
{
  static String FallBackResources[] = {
    "WnpFont*geometry: 500x500+50+70",
    "*ThreeD*borderWidth:  0",
    "*ThreeD*rgba: TRUE",
    "*ThreeD*doublebuffer: FALSE",
    "*ThreeD*allocateBackground: TRUE",
    NULL
    };

  font = argv[1];

  appShell = XtAppInitialize(&appContext,APPL_CLASS,(XrmOptionDescRec*)NULL,
			     0,&argc,argv,FallBackResources,NULL,0);
  
  if (!appShell) {
    fprintf(stderr,"Could not open application shell.\n");
    return(BAD);
  }
  
  initWidgetTree(appShell);
  fprintf(stderr,"tree.\n");
  XtAppMainLoop(appContext);
  return(GOOD); /* will never happen... */
}
/*-------------------- chooseVisual --------------------*/
static XVisualInfo *chooseVisual(Widget w)
{
  XVisualInfo *v;
  Display *dsp;
  int scr;

  int attr1[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4, GLX_BLUE_SIZE, 4, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, GLX_ALPHA_SIZE, 0, None
  };
    
  int attr2[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4, GLX_BLUE_SIZE, 4, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, None
  };

  int attr3[] = {
    GLX_RGBA, GLX_RED_SIZE, 2, GLX_GREEN_SIZE, 2, GLX_BLUE_SIZE, 2, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, None
  };

  int attr4[] = {
    GLX_RGBA, GLX_DEPTH_SIZE, 1, GLX_DOUBLEBUFFER, None
  };

  dsp=XtDisplay(w);
  scr=DefaultScreen(dsp);

  v=glXChooseVisual(dsp,scr,attr1);
  if (!v) v=glXChooseVisual(dsp,scr,attr2);
  if (!v) v=glXChooseVisual(dsp,scr,attr3);
  if (!v) v=glXChooseVisual(dsp,scr,attr4);

  return(v);
}
