#include "wnp.h"
#include "wnpwin.h"
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#if defined(darwin)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

static int haveCB=FALSE;  /* input or expose callback since last graphics upd ? */
static int leftMouse=FALSE;
static int modstate=0;    /* the six mouse buttons... */
static int txpos=0,typos=0;
static int dx=0,dy=0;
static int calibratePosition=FALSE;
static int glBox = 0;
static int win_fsgeom[4]; /* size and position of full-screen window */
static int mtrapxmin,mtrapxmax,mtrapymin,mtrapymax;
static int mtrap=FALSE;
static int firstReshape=TRUE;
static int win_width,win_height;

/* ======================================================================== */

void WnpRegisterGlBox(int w)
{
  glBox=w;
}

void WnpGlBoxWinpos(int pos[4])
{
  int w;

  w=glutGetWindow();
  if (w!=glBox) glutSetWindow(glBox);
  glutPositionWindow(pos[0]-dx,pos[1]-dy);
  txpos=pos[0]; typos=pos[1];
  calibratePosition=TRUE;
  glutReshapeWindow(pos[2],pos[3]);
  if (w!=glBox) glutSetWindow(w);
}

void WnpGlBoxFullScreen(int pos[4])
{
  WnpGlBoxWinpos(win_fsgeom);
}

void WnpCalibratePosition(void)
{
  int x,y;

  if (calibratePosition) {
    x=glutGet(GLUT_WINDOW_X);
    y=glutGet(GLUT_WINDOW_Y);
    if (x!=txpos||y!=typos) {
      dx+=x-txpos;
      dy+=y-typos;
      glutPositionWindow(txpos-dx,typos-dy);
    }
    calibratePosition=FALSE;
  }
}

void WnpGetGlBoxWinpos(int pos[4])
{
  int w;

  w=glutGetWindow();
  if (w!=glBox) glutSetWindow(glBox);
  pos[0]=glutGet(GLUT_WINDOW_X);
  pos[1]=glutGet(GLUT_WINDOW_Y);
  pos[2]=glutGet(GLUT_WINDOW_WIDTH);
  pos[3]=glutGet(GLUT_WINDOW_HEIGHT);
  if (w!=glBox) glutSetWindow(w);
}

void WnpRingBell(void)
{
  fprintf(stderr,"\007"); /* Let's hope, it works... */
}

void WnpSetMouseLeft(int right)
{
  if (right) leftMouse=FALSE;
  else leftMouse=TRUE;
}

int WnpButton(unsigned int xb)
{
  if (leftMouse) 
    return (xb==GLUT_LEFT_BUTTON ? WNP_MOD_RIGHT
	    :(xb==GLUT_MIDDLE_BUTTON ? WNP_MOD_MIDDLE
	      :(xb==GLUT_RIGHT_BUTTON ? WNP_MOD_LEFT:0)
	      )
	    ); 
  else
    return (xb==GLUT_LEFT_BUTTON ? WNP_MOD_LEFT
	    :(xb==GLUT_MIDDLE_BUTTON ? WNP_MOD_MIDDLE
	      :(xb==GLUT_RIGHT_BUTTON ? WNP_MOD_RIGHT:0)
	      )
	    ); 
}

int WnpModState()
{
  /* return state of modifier keys (CASLMR) */
  /* get mouse keys from own book-keeping, others from system */

  int mod;

  mod=glutGetModifiers();
  modstate&=~(WNP_MOD_SHIFT|WNP_MOD_ALT|WNP_MOD_CTRL);
  if (mod&GLUT_ACTIVE_ALT)  modstate|=WNP_MOD_ALT;
  if (mod&GLUT_ACTIVE_CTRL) modstate|=WNP_MOD_CTRL;
  if (mod&GLUT_ACTIVE_SHIFT) modstate|=WNP_MOD_SHIFT;
  return(modstate);
}

void WnpMouseMotionMask(int state)
{
  int w;
  
  w=glutGetWindow();
  if (w!=glBox) glutSetWindow(glBox);
  if (state) {
    glutMotionFunc(WnpMotionCB);
    glutPassiveMotionFunc(WnpPassiveMotionCB);
  }else{
    glutMotionFunc(NULL);
    glutPassiveMotionFunc(NULL);
  }
  if (w!=glBox) glutSetWindow(w);
}

void WnpSetMouseTrap(int xmin,int ymin,int width, int height)
{
  /* trap pointer within box described by X11-window coords x,y,h,w */

  mtrap=TRUE;
  mtrapxmin=xmin;
  mtrapymin=ymin;
  mtrapxmax=xmin+width-1;
  mtrapymax=ymin+height-1;
}

void WnpRemoveMouseTrap(void)
{
  mtrap=FALSE;
}

void WnpTrapMouse(int *x,int *y)
{
  int dy,dx;
  
  /* note x,y are in GLUT coordinates, i.e. origin is upper-left */

  if (mtrap) {
    dx=clamp(*x,mtrapxmin,mtrapxmax);
    dy=clamp(*y,mtrapymin,mtrapymax);
    if (dx!=*x||dy!=*y) {
      glutWarpPointer(dx,dy);
      *x=dx; *y=dy;
    }
  }
}

void WnpMakeGlBoxCurrent(void)
{
  glutSetWindow(glBox);
}

void WnpSwapGlBuffers(void)
{
  int w;
  
  w=glutGetWindow();
  if (w!=glBox) glutSetWindow(glBox);
  glutSwapBuffers();
  if (w!=glBox) glutSetWindow(w);
}

int WnpDisplayWidth(void)
{
  return(win_fsgeom[2]);
}

int WnpDisplayHeight(void)
{
  return(win_fsgeom[3]);
}

int WnpGlBoxWindow(void)
{
  return(glBox);
}

void WnpLookupKey(int key,int *class,int *sym)
{
  switch (key) {
  case GLUT_KEY_F1:        *class=WNP_FUNC_KEY;   *sym=WNP_F1_KEY;     break;
  case GLUT_KEY_F2:        *class=WNP_FUNC_KEY;   *sym=WNP_F2_KEY;     break;
  case GLUT_KEY_F3:        *class=WNP_FUNC_KEY;   *sym=WNP_F3_KEY;     break;
  case GLUT_KEY_F4:        *class=WNP_FUNC_KEY;   *sym=WNP_F4_KEY;     break;
  case GLUT_KEY_F5:        *class=WNP_FUNC_KEY;   *sym=WNP_F5_KEY;     break;
  case GLUT_KEY_F6:        *class=WNP_FUNC_KEY;   *sym=WNP_F6_KEY;     break;
  case GLUT_KEY_F7:        *class=WNP_FUNC_KEY;   *sym=WNP_F7_KEY;     break;
  case GLUT_KEY_F8:        *class=WNP_FUNC_KEY;   *sym=WNP_F8_KEY;     break;
  case GLUT_KEY_F9:        *class=WNP_FUNC_KEY;   *sym=WNP_F9_KEY;     break;
  case GLUT_KEY_F10:       *class=WNP_FUNC_KEY;   *sym=WNP_F10_KEY;    break;
  case GLUT_KEY_F11:       *class=WNP_FUNC_KEY;   *sym=WNP_F11_KEY;    break;
  case GLUT_KEY_F12:       *class=WNP_FUNC_KEY;   *sym=WNP_F12_KEY;    break;
  case GLUT_KEY_UP:        *class=WNP_ARROW_KEY;  *sym=WNP_UP_KEY;     break;
  case GLUT_KEY_DOWN:      *class=WNP_ARROW_KEY;  *sym=WNP_DOWN_KEY;   break;
  case GLUT_KEY_LEFT:      *class=WNP_ARROW_KEY;  *sym=WNP_LEFT_KEY;   break;
  case GLUT_KEY_RIGHT:     *class=WNP_ARROW_KEY;  *sym=WNP_RIGHT_KEY;  break;
  case GLUT_KEY_PAGE_UP:   *class=WNP_FUNC_KEY;   *sym=WNP_PGUP_KEY;   break;
  case GLUT_KEY_PAGE_DOWN: *class=WNP_FUNC_KEY;   *sym=WNP_PGDOWN_KEY; break;
  default:                 *class=WNP_IGNORE_KEY; break;
  }
}

int WnpResetCBFlag(void)
{
  int r;

  r=haveCB;
  haveCB=FALSE;
  return(r);
}

void WnpCloseGraphics(void)
{
  glutSetWindow(glBox);
  WnpSetStereoMode(WNP_STEREO_NO);
#if defined(cygwin) || defined(mingw)
  /* wait for mouse-button-up to avoid pop-up of background menu [or worse] */
  uSleep(250000);
#endif
  glutDestroyWindow(glBox);
}

/*==================== Callbacks ====================*/

void WnpReshapeCB(int width, int height)
{
  if (firstReshape) {
    firstReshape=FALSE;
    win_fsgeom[0]=0; win_fsgeom[1]=0;
    win_fsgeom[2]=glutGet(GLUT_SCREEN_WIDTH);
    win_fsgeom[3]=glutGet(GLUT_SCREEN_HEIGHT);
    glutIdleFunc(WnpWorkProcCB); /* dont try this before GLBOX is initialized... */
    WnpInitGlBox(width,height);
    WnpInitCursor();
  }
  win_width=width; win_height=height;
  WnpGlBoxViewport(0,0,width,height);
}

void WnpWorkProcCB(void)
{
  WnpWorkProc();
}

void WnpDisplayCB(void)
{
  WnpCalibratePosition();
  WnpSetGlBoxRedrawFlag(TRUE);
  haveCB=TRUE;
}

void WnpKeyboardCB(unsigned char key,int x,int y)
{
  WnpAddKey(key,WnpModState());
  haveCB=TRUE;
}

void WnpSpecialCB(int key,int x,int y)
{
  int keyclass,keysym;

  WnpLookupKey(key,&keyclass,&keysym);
  switch (keyclass) {
  case WNP_FUNC_KEY:
    WnpFunctionKey(keysym,WnpModState());
    break;
  case WNP_ARROW_KEY:
    WnpArrowKey(keysym,WnpModState());
    break;
  default:
    break;
  }
  haveCB=TRUE;
}

void WnpMouseCB(int button,int state,int x,int y)
{
  switch (state) {
  case GLUT_DOWN:
    if (x<0||y<0||x>=win_width||y>=win_height) modstate=0;
    else modstate=MouseButtonPress(WnpButton(button),WnpModState(),x,y);
    break;
  case GLUT_UP:
    if (x<0||y<0||x>=win_width||y>=win_height) modstate=0;
    else modstate=MouseButtonRelease(WnpButton(button),WnpModState(),x,y);
    break;
  }
}

void WnpMotionCB(int x,int y)
{
  WnpTrapMouse(&x,&y);
  WnpGlBoxShowCursor();
  WnpGlMoveCursor(x,y);
  MouseMoved(modstate,x,y); /* not WnpModState(), since not in core CB */
  haveCB=TRUE;
}

void WnpPassiveMotionCB(int x,int y)
{
  WnpMotionCB(x,y);
}

void WnpEntryCB(int state)
{
  switch (state) {
  case GLUT_ENTERED:
    break;
  case GLUT_LEFT:
    WnpGlMoveCursor(-100,-100);
    break;
  }
}
