#include <stdlib.h>
#include <stddef.h>
#include "wnp.h"

static char *cursColorName[WNP_MAXCCOL] = {
  "black","red","green","blue","yellow","white"
};
static int cursColor[WNP_MAXCCOL];           /* cursor color table            */

static char noCursor[] = {0x0};              /* this will turn cursor off     */
static char BoxCursor[] = {
  0xff,0xff, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x81,0x81,
  0x81,0x81, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0xff,0xff
};
static char boxCursor[] = {
  0xff,0xff, 0x01,0x80, 0x01,0x80, 0x01,0x80, 0x81,0x81, 0x01,0x80, 0x01,0x80, 0x01,0x80,
  0xff,0xff
};
static char ArrowCursor[] = {
  0x01,0x03,0x07,0x0f,0x1f,0x3f,0x7f,0xff,0x7f,0x1f,0x3b,0x30,0x70,0x60,0xe0,0x40
};
static char arrowCursor[] = {
  0x01,0x07,0x1f,0x7f,0x7f,0x3b,0x70,0xe0
};

typedef struct WnpCursor{
  int cursor;                    /* X11 cursor id, (0 if not defined)   */
  int mono,cryst;                /* cursor in mono/cryst mode           */
  int *color;                    /* cursor, foreground=background       */   
  char *map;                     /* pointer to raw bitmap data          */
  int width,height;              /* width and height in pixels          */
  int hotx,hoty;                 /* hotspot                             */
} WnpCursor;

static WnpCursor curtab[WNP_MAX_CURSOR]= {   /* cursor table                     */
  {0, WNP_CURSOR_NONE,  WNP_CURSOR_NONE,  cursColor+WNP_CURSOR_RED,   noCursor,     1, 1, 0,0},
  {0, WNP_CURSOR_DOTBOX,WNP_CURSOR_dotbox,cursColor+WNP_CURSOR_RED,   BoxCursor,   16,16, 7,7},
  {0, WNP_CURSOR_DOTBOX,WNP_CURSOR_dotbox,cursColor+WNP_CURSOR_RED,   boxCursor,   16, 9, 7,4},
  {0, WNP_CURSOR_ARROW, WNP_CURSOR_arrow ,cursColor+WNP_CURSOR_GREEN, ArrowCursor,  8,16, 0,0},
  {0, WNP_CURSOR_ARROW, WNP_CURSOR_arrow ,cursColor+WNP_CURSOR_GREEN, arrowCursor,  8, 8, 0,0}
};

static int hide_cursor = FALSE;        /* should we hide cursor in glbox ? */
static int current_cursor = -1;        /* current index into cursor tab    */
static int inCursorMode=FALSE;

/* ======================================================================== */

/* This module contains mostly stubs for non-implemented cursor functions */

void WnpInitCursor(void)
{
  /*stub*/ /*should define graphics and menu cursor (mono and stereo */

  int i;
  for (i=0;i<WNP_MAX_CURSOR;i++) {
    curtab[i].cursor=i;
  }
}

void WnpDefineCursor(int id)
{
  /*stub*/ /* should set current cursor to cursor[id] */
}

void WnpSetGlBoxCursorColor(int cursor,int color)
{
  /*stub*/ /* should redefine cursor color */
}

int WnpGlBoxShowCursor(void)
{
  int r,o;

  r=hide_cursor;
  hide_cursor=FALSE;
  if (r) {
    o=current_cursor; 
    current_cursor=WNP_CURSOR_NONE;
    WnpSetGlBoxCursor(o);
  }
  return(r);
}

int WnpGlBoxHideCursor(void)
{
  int r,o;
  
  r=hide_cursor;
  if (!r) WnpDefineCursor(WNP_CURSOR_NONE);
  hide_cursor=TRUE;
  return(r);
}

int WnpSetGlBoxCursor(int new)
{
  int old_cursor,crystaleyes;

  crystaleyes=WnpCrystalEyesMode();
  old_cursor=current_cursor;
  if (new>=0 && new < WNP_MAX_CURSOR) {
    new=(crystaleyes ? curtab[new].cryst : curtab[new].mono);
    if (current_cursor!=new) {
      if (!hide_cursor) WnpDefineCursor(new);
      current_cursor=new;
    }
  }
  return(old_cursor);
}

void WnpGlDrawCursor(int swapped)
{
  /* cursor drawn by system */
}

void WnpGlMoveCursor(int x,int y)
{
}

void WnpGlRestoreCursorPatch(void)
{
  /* never have a patch to restore */
}

void WnpGlSetCursorRedrawFlag(int state)
{
}
