#include <windows.h>
#include <stdio.h>
#include "wnp.h"
#include "wnpwin.h"

static HWND  wnpGlBoxWindow=NULL;    /* Window handle for                         */
static HDC   wnpGlBoxDC=NULL;        /* Device context handle (owned)             */
static HGLRC wnpGlBoxGLRC=NULL;      /* Rendering context handle                  */

static int win_fsgeom[4];            /* geometry of display (x,y,cx,cy)           */
static int win_geom[4];              /* current client area geometry              */
static int track_mouse=FALSE;        /* do we handle mouse motion events ?        */
static int mtrapxmin,mtrapxmax,mtrapymin,mtrapymax;
static int mtrap=FALSE;
static int leftMouse=FALSE;
static int haveCB=FALSE;

/* Note on accelerated pixel formats (from MSDN)
PFD_GENERIC_ACCELERATED 	
   The pixel format is supported by a device driver that accelerates the generic implementation. 
   If this flag is clear and the PFD_GENERIC_FORMAT flag is set, the pixel format is supported
   by the generic implementation only.

PFD_GENERIC_FORMAT 	
   The pixel format is supported by the GDI software implementation, which is also known as the 
   generic implementation. If this bit is clear, the pixel format is supported by a device driver 
   or hardware.
*/

HGLRC WnpGlBoxWin32Setup( HDC hDC )
{
  PIXELFORMATDESCRIPTOR pfd = 
  {
    sizeof( PIXELFORMATDESCRIPTOR ),
    1,
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER | PFD_GENERIC_ACCELERATED,
    PFD_TYPE_RGBA,
    24,                 /* bits per pixel, will be redefined... */
    0, 0, 0, 0, 0, 0,
    0,
    0,
    0,
    0, 0, 0, 0,
    16,                 /* 16-bit depth buffer */
    0,                  /* no stencil buffer   */
    0,                  /* no aux buffers      */
    PFD_MAIN_PLANE,	/* main layer          */
    0,	
    0, 0, 0
  };
  int  selected_pf;
  HGLRC hGLRC;
   
  pfd.cColorBits=GetDeviceCaps( hDC, BITSPIXEL );

  if ( ( selected_pf = ChoosePixelFormat( hDC, &pfd ) ) == 0 ) {
    MessageBox( 0, "Failed to find acceptable pixel format", "Setup Error", MB_ICONERROR | MB_OK );
    return NULL;
  }
  
  if ( !SetPixelFormat( hDC, selected_pf, &pfd) ) {
    fprintf(stderr,"pixel format NOT set\n");
    MessageBox( 0, "Failed to SetPixelFormat", "Setup Error", MB_ICONERROR | MB_OK );
    return NULL;
  }

  hGLRC = wglCreateContext(hDC);
  return hGLRC;
}

void WnpMakeGlBoxCurrent(void)
{
  wglMakeCurrent(wnpGlBoxDC,wnpGlBoxGLRC);
}

void WnpCloseGraphics()
{
  WnpSetStereoMode(WNP_STEREO_NO);

  if (wnpGlBoxGLRC) {
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(wnpGlBoxGLRC);
    wnpGlBoxGLRC=NULL;
  }
  if (wnpGlBoxWindow && wnpGlBoxDC) {
    ReleaseDC(wnpGlBoxWindow,wnpGlBoxDC);
    wnpGlBoxDC=NULL;
  }
}

void WnpMouseMotionMask(int state)
{
  track_mouse=state;
}

void WnpSwapGlBuffers(void)
{
  SwapBuffers(wnpGlBoxDC);
}

void WnpGetGlBoxWinpos(int geom[4])
{
  geom[0]=win_geom[0];   geom[1]=win_geom[1]; 
  geom[2]=win_geom[2];   geom[3]=win_geom[3]; 
}

void WnpGlBoxWinpos(int geom[4])
{
  /*stub*/ /* should move/resize window for new geometry */
}

void WnpGlBoxFullScreen(int geom[4])
{
  /*stub*/ /* should move/resize window for new geometry */
}

int WnpDisplayWidth()
{
  return(win_fsgeom[2]);
}

int WnpDisplayHeight()
{
  return(win_fsgeom[3]);
}

void WnpRingBell(void)
{
  fprintf(stderr,"\007");
}

void WnpSetMouseTrap(int xmin,int ymin,int width, int height)
{
  /* trap pointer within box described by client area coords x,y,h,w */

  mtrap=TRUE;
  mtrapxmin=xmin;
  mtrapymin=ymin;
  mtrapxmax=xmin+width-1;
  mtrapymax=ymin+height-1;
}

void WnpRemoveMouseTrap(void)
{
  mtrap=FALSE;
}

void WnpTrapMouse(int *x,int *y)
{
  int dy,dx;
  
  /* note x,y are in client coordinates, i.e. origin is upper-left */

  if (mtrap) {
    dx=clamp(*x,mtrapxmin,mtrapxmax);
    dy=clamp(*y,mtrapymin,mtrapymax);
    if (dx!=*x||dy!=*y) {
  /*stub*/ /* should move the mouse back inside trap rectangle. how ? */
      *x=dx; *y=dy;
    }
  }
}

void WnpSetMouseLeft(int right)
{
  if (right) leftMouse=FALSE;
  else leftMouse=TRUE;
}

int WnpModState(void)
{
  /* sync. Windows  modifier key state in internal format... */
 
  int r;

  r=0;
  if (leftMouse) {
    if (GetKeyState(VK_LBUTTON)<0) r|=WNP_MOD_RIGHT;
    if (GetKeyState(VK_MBUTTON)<0) r|=WNP_MOD_MIDDLE;
    if (GetKeyState(VK_RBUTTON)<0) r|=WNP_MOD_LEFT;
    if (GetKeyState(VK_SHIFT)<0)   r|=WNP_MOD_SHIFT;
    if (GetKeyState(VK_CONTROL)<0) r|=WNP_MOD_CTRL;
    if (GetKeyState(VK_MENU)<0)    r|=WNP_MOD_ALT;
  }else{
    if (GetKeyState(VK_RBUTTON)<0) r|=WNP_MOD_RIGHT;
    if (GetKeyState(VK_MBUTTON)<0) r|=WNP_MOD_MIDDLE;
    if (GetKeyState(VK_LBUTTON)<0) r|=WNP_MOD_LEFT;
    if (GetKeyState(VK_SHIFT)<0)   r|=WNP_MOD_SHIFT;
    if (GetKeyState(VK_CONTROL)<0) r|=WNP_MOD_CTRL;
    if (GetKeyState(VK_MENU)<0)    r|=WNP_MOD_ALT;
  }
  return(r);
}

void WnpLookupKey(WPARAM wParam,int *keyclass,int *keysym)
{
  switch (wParam) {
  case VK_LEFT:  *keyclass=WNP_ARROW_KEY; *keysym=WNP_LEFT_KEY;     break;
  case VK_RIGHT: *keyclass=WNP_ARROW_KEY; *keysym=WNP_RIGHT_KEY;    break;
  case VK_UP:    *keyclass=WNP_ARROW_KEY; *keysym=WNP_UP_KEY;       break;
  case VK_DOWN:  *keyclass=WNP_ARROW_KEY; *keysym=WNP_DOWN_KEY;     break;
  case VK_F1:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F1_KEY;       break;
  case VK_F2:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F2_KEY;       break;
  case VK_F3:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F3_KEY;       break;
  case VK_F4:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F4_KEY;       break;
  case VK_F5:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F5_KEY;       break;
  case VK_F6:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F6_KEY;       break;
  case VK_F7:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F7_KEY;       break;
  case VK_F8:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F8_KEY;       break;
  case VK_F9:    *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F9_KEY;       break;
  case VK_F10:   *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F10_KEY;      break;
  case VK_F11:   *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F11_KEY;      break;
  case VK_F12:   *keyclass=WNP_FUNC_KEY;  *keysym=WNP_F12_KEY;      break;
  case VK_PRIOR: *keyclass=WNP_FUNC_KEY;  *keysym=WNP_PGUP_KEY;     break;
  case VK_NEXT:  *keyclass=WNP_FUNC_KEY;  *keysym=WNP_PGDOWN_KEY;   break;
  default:       *keyclass=WNP_IGNORE_KEY;                          break;
  }
}

int WnpButton(UINT b)
{
  if (leftMouse) 
    return (b==WM_LBUTTONDOWN||b==WM_LBUTTONUP ? WNP_MOD_RIGHT
	    :(b==WM_MBUTTONDOWN||b==WM_MBUTTONUP ? WNP_MOD_MIDDLE
	      :(b==WM_RBUTTONDOWN||b==WM_RBUTTONUP ? WNP_MOD_LEFT:0)
	      )
	    ); 
  else
    return (b==WM_LBUTTONDOWN||b==WM_LBUTTONUP ? WNP_MOD_LEFT
	    :(b==WM_MBUTTONDOWN||b==WM_MBUTTONUP ? WNP_MOD_MIDDLE
	      :(b==WM_RBUTTONDOWN||b==WM_RBUTTONUP ? WNP_MOD_RIGHT:0)
	      )
	    ); 
}

HWND WnpGlBoxWindow(void)
{
  return(wnpGlBoxWindow);
}

/*==================== Callbacks ====================*/

int WnpGlBoxInitCB(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
  int status=TRUE;
  RECT rect;

  wnpGlBoxWindow=hwnd;
  wnpGlBoxDC=GetDC(wnpGlBoxWindow);
  wnpGlBoxGLRC=WnpGlBoxWin32Setup(wnpGlBoxDC);
  
  if (wnpGlBoxGLRC) {
    GetWindowRect(wnpGlBoxWindow,&rect);
    win_geom[0]=rect.left;  win_geom[1]=rect.top;
    win_geom[2]=rect.right; win_geom[3]=rect.bottom;
    win_fsgeom[0]=0; win_fsgeom[1]=0;
    win_fsgeom[2]=GetSystemMetrics(SM_CXSCREEN);
    win_fsgeom[3]=GetSystemMetrics(SM_CYSCREEN);
    WnpMakeGlBoxCurrent();
    GetClientRect(wnpGlBoxWindow,&rect);
    WnpInitGlBox(rect.right,rect.bottom);
  }else status=FALSE;
  return(status);
}

LRESULT CALLBACK WndProc(HWND hwnd,UINT msgid,WPARAM wParam,LPARAM lParam)
{
  LRESULT r;
  int keysym,keyclass;
  int x,y;

  r=0;
  switch (msgid) {

  case WM_CREATE:
    if (!WnpGlBoxInitCB(hwnd,wParam,lParam)) PostQuitMessage(0);
    break;

  case WM_PAINT:
    haveCB=TRUE;
    WnpSetGlBoxRedrawFlag(TRUE);
    ValidateRect(hwnd,NULL);
    break;

  case WM_SIZE:
    WnpGlBoxViewport(0,0,LOWORD(lParam),HIWORD(lParam));
    break;

  case WM_CHAR:
    WnpAddKey(wParam,0);
    break;

  case WM_SYSCHAR:
    WnpAddKey(wParam,WNP_MOD_ALT);
    break;

  case WM_SYSKEYDOWN:
  case WM_KEYDOWN:
    haveCB=TRUE;
    WnpLookupKey(wParam,&keyclass,&keysym);
    switch (keyclass) {
    case WNP_FUNC_KEY:
      WnpFunctionKey(keysym,WnpModState());
      break;
    case WNP_ARROW_KEY:
      WnpArrowKey(keysym,WnpModState());
      break;
    default:
      break;
    }
    break;

  case WM_LBUTTONDOWN:
  case WM_MBUTTONDOWN:
  case WM_RBUTTONDOWN:
    MouseButtonPress(WnpButton(msgid),WnpModState(),LOWORD(lParam),HIWORD(lParam));
    break;

  case WM_LBUTTONUP:
  case WM_MBUTTONUP:
  case WM_RBUTTONUP:
    MouseButtonRelease(WnpButton(msgid),WnpModState(),LOWORD(lParam),HIWORD(lParam));
    break;

  case WM_MOUSEMOVE:
    haveCB=TRUE;
    x=LOWORD(lParam);
    y=HIWORD(lParam);
    WnpTrapMouse(&x,&y);
    WnpGlBoxShowCursor();
    MouseMoved(WnpModState(),x,y);
    break;

  case WM_DESTROY:
    WnpExit();
    break;

  default:
    r=DefWindowProc(hwnd,msgid,wParam,lParam);
    break;
  }
  return(r);
}

/* ----------------------------------------------- */
/* test and reset CB flag (used by work procedure) */

int WnpResetCBFlag(void)
{
  int r;

  r=haveCB;
  haveCB=FALSE;
  return(r);
}
