#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include "wnp.h"
#include "wnpwin.h"
#include <X11/Xaw/Form.h>
#include <X11/Shell.h>
#include "GLwDrawA.h"

#define FALSE 0
#define GOOD	0
#define BAD	1

#define APPL_CLASS  "WitNotP"

static Widget initWidgetTree(Widget appShell);
static Widget initGlBox(Widget parent);
static XtAppContext appContext;
static Widget appShell;

static XVisualInfo *WnpChooseVisual(Widget w);

/* ======================================================================== */
/* X11 API Entry Point */

int main(int argc,char **argv)
{
  int i,k,background,pid,applic;

  char *service_name;
  static String FallBackResources[] = {
    "*main.geometry: 640x480+20+20",
    "*ThreeD*borderWidth:  0",
    "*ThreeD*allocateBackground: TRUE",
    "*ThreeD*width: 640",
    "*ThreeD*height: 480",
    NULL
  };

  /* Test if WnpAtomPicker in draw.c can handle this CPU architecture... */

  k=sizeof(void *)/sizeof(GLuint);
  if (k*sizeof(GLuint)!=sizeof(void *)||k>2) {
    if (k>2) fprintf(stderr,"FATAL: sizeof(void *) exceeds 2*sizeof(GLunit)\n");
    if (k*sizeof(GLuint)!=sizeof(void *)) fprintf(stderr,"FATAL: sizeof(void *) is not an integer multiple of sizeof(GLunit)\n");
    fprintf(stderr,"WITNOTP graphics server failed to start: unsupported CPU architecture.\n");
    return(0);
  }

  background=TRUE;           /* default is to run detached... */
  service_name="witnotp";    /* default service name          */
  
  /* process (and remove) some of the command line options */

  for (i=1;i<argc;) {
    if (strcmp("foreground",argv[i])==0) {
      background=FALSE;
      for (k=i;k<argc-1;k++) argv[k]=argv[k+1];
      argv[k]=(char *)NULL; argc--;
    }else if (strcmp("service",argv[i])==0) {
      service_name=strdup(argv[i+1]);
      for (k=i;k<argc-2;k++) argv[k]=argv[k+2];
      argv[k]=(char *)NULL; argc-=2;
    }else i++;
  }
  
  /* push it into the background... */
  
  if (background) {
    pid=fork();
    if (pid>0) return(0);
    else if (pid<0) {
      fprintf(stderr," failed to fork server...\n");
      return (1);
    }
  }

  /* wait for request from client and go... */

  applic=WnpStartComm(service_name);
  if (!applic) {
    fprintf(stderr,"WITNOTP: Could not open message transport to client...\n");
    return(0);
  }

  /* ready to go X11 */

  appShell = XtOpenApplication(&appContext,APPL_CLASS,(XrmOptionDescRec*)NULL,
			     0,&argc,argv,FallBackResources,applicationShellWidgetClass,NULL,0);
  
  if (!appShell) {
    fprintf(stderr,"Could not open application shell.\n");
    return(BAD);
  }
  
  initWidgetTree(appShell);
  XtAppAddWorkProc(appContext,WnpWorkProcCB,(XtPointer)NULL);
  XtAppMainLoop(appContext);

  return(0); /* this will never happen... */
}
/*-------------------- initWidgetTree --------------------*/

static Widget initWidgetTree(Widget shell)
{
  Arg args[50];
  int n;
  
  Widget rootW;

  n=0;
  XtSetArg(args[n],XtNdefaultDistance,0); n++;
  
  rootW=XtCreateManagedWidget("main",formWidgetClass,shell,args,n);
  XtManageChild(rootW);
  
  initGlBox(rootW);
  XtRealizeWidget(shell);
  
  return(rootW);
}
/*-------------------- initGlBox --------------------*/

static Widget initGlBox(Widget parent)
{
  XVisualInfo *vinf;
  Arg glbox_args[50];
  Widget box;
  int n;

  n=0;

  XtSetArg(glbox_args[n],XtNtop,(XtArgVal)XtChainTop); n++;
  XtSetArg(glbox_args[n],XtNleft,(XtArgVal)XtChainLeft); n++;
  XtSetArg(glbox_args[n],XtNbottom,(XtArgVal)XtChainBottom); n++;
  XtSetArg(glbox_args[n],XtNright,(XtArgVal)XtChainRight); n++;

  XtSetArg(glbox_args[n],GLwNrgba,TRUE); n++;
  XtSetArg(glbox_args[n],GLwNdoublebuffer,TRUE); n++;
  XtSetArg(glbox_args[n],GLwNdepthSize,1); n++;

  vinf=WnpChooseVisual(parent);
  if (vinf) {
    XtSetArg(glbox_args[n],GLwNvisualInfo,vinf); n++;
  }else{
    fprintf(stderr,"*** Warning *** could not determine appropriate visual.\n");
  }

  box=XtCreateManagedWidget("ThreeD",glwDrawingAreaWidgetClass,parent,
			    glbox_args,n);

  XtAddCallback(box,GLwNginitCallback,WnpGlBoxGinitCB,(XtPointer)NULL);
  XtAddCallback(box,GLwNexposeCallback,WnpGlBoxExposeCB,(XtPointer)NULL);
  XtAddCallback(box,GLwNresizeCallback,WnpGlBoxResizeCB,(XtPointer)NULL);
  XtAddCallback(box,GLwNinputCallback,WnpGlBoxInputCB,(XtPointer)NULL);
  
  XtManageChild(box);
  return(box);
}

static XVisualInfo *WnpChooseVisual(Widget w)
{
  XVisualInfo *v;
  Display *dsp;
  int scr;

  int attr1q[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4, GLX_BLUE_SIZE, 4, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, GLX_ALPHA_SIZE, 0, GLX_STEREO, None
  };
    
  int attr2q[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4, GLX_BLUE_SIZE, 4, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, GLX_STEREO, None
  };

  int attr1[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4, GLX_BLUE_SIZE, 4, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, GLX_ALPHA_SIZE, 0, None
  };
    
  int attr2[] = {
    GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4, GLX_BLUE_SIZE, 4, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, None
  };

  int attr3[] = {
    GLX_RGBA, GLX_RED_SIZE, 2, GLX_GREEN_SIZE, 2, GLX_BLUE_SIZE, 2, GLX_DEPTH_SIZE, 1,
    GLX_DOUBLEBUFFER, None
  };

  int attr4[] = {
    GLX_RGBA, GLX_DEPTH_SIZE, 1, GLX_DOUBLEBUFFER, None
  };

  dsp=XtDisplay(w);
  scr=DefaultScreen(dsp);

  v=glXChooseVisual(dsp,scr,attr1q);
  if (!v) v=glXChooseVisual(dsp,scr,attr2q);
  if (!v) v=glXChooseVisual(dsp,scr,attr1);
  if (!v) v=glXChooseVisual(dsp,scr,attr2);
  if (!v) v=glXChooseVisual(dsp,scr,attr3);
  if (!v) v=glXChooseVisual(dsp,scr,attr4);

  return(v);
}
