#include "wnp.h"
#include "wnpwin.h"
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "GLwDrawA.h"
#include <X11/X.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>

static GLXContext  wnpGlBoxContext;
static Window      wnpGlBoxWindow;
static Display    *wnpGlBoxDisplay;
static Widget      wnpGlBoxWidget;
static int         wnpGlBoxScreenNr;
static Colormap    wnpGlBoxCmap;

static int win_fsgeom[4]; /* size and position of full-screen window */

static int txpos=0,typos=0;
static int dx=0,dy=0;
static int calibratePosition=FALSE;
static int mtrapxmin,mtrapxmax,mtrapymin,mtrapymax;
static int mtrap=FALSE;
static int leftMouse=FALSE;
static int haveCB=FALSE;  /* input or expose callback since last graphics upd ? */

/* ======================================================================== */

void WnpCloseGraphics(void)
{
  WnpSetStereoMode(WNP_STEREO_NO);
}

void WnpLookupKey(XKeyEvent *K,int *class,int *sym,char *text)
{
  /* translate XKey event to internal class... */

  int n;
  KeySym k;

  n=XLookupString(K,text,WNP_MAXLINE,&k,(XComposeStatus *)NULL);
  if (n==1) {
    *class=WNP_ASCII_KEY;
    *sym=0;
  }else if (n==0) {
    switch (k) {
    case XK_Left:  *class=WNP_ARROW_KEY; *sym=WNP_LEFT_KEY;     break;
    case XK_Right: *class=WNP_ARROW_KEY; *sym=WNP_RIGHT_KEY;    break;
    case XK_Up:    *class=WNP_ARROW_KEY; *sym=WNP_UP_KEY;       break;
    case XK_Down:  *class=WNP_ARROW_KEY; *sym=WNP_DOWN_KEY;     break;
    case XK_F1:    *class=WNP_FUNC_KEY;  *sym=WNP_F1_KEY;       break;
    case XK_F2:    *class=WNP_FUNC_KEY;  *sym=WNP_F2_KEY;       break;
    case XK_F3:    *class=WNP_FUNC_KEY;  *sym=WNP_F3_KEY;       break;
    case XK_F4:    *class=WNP_FUNC_KEY;  *sym=WNP_F4_KEY;       break;
    case XK_F5:    *class=WNP_FUNC_KEY;  *sym=WNP_F5_KEY;       break;
    case XK_F6:    *class=WNP_FUNC_KEY;  *sym=WNP_F6_KEY;       break;
    case XK_F7:    *class=WNP_FUNC_KEY;  *sym=WNP_F7_KEY;       break;
    case XK_F8:    *class=WNP_FUNC_KEY;  *sym=WNP_F8_KEY;       break;
    case XK_F9:    *class=WNP_FUNC_KEY;  *sym=WNP_F9_KEY;       break;
    case XK_F10:   *class=WNP_FUNC_KEY;  *sym=WNP_F10_KEY;      break;
    case XK_F11:   *class=WNP_FUNC_KEY;  *sym=WNP_F11_KEY;      break;
    case XK_F12:   *class=WNP_FUNC_KEY;  *sym=WNP_F12_KEY;      break;
      /*some systems have shift-F1 xmodmap'ed to shift-F13... */
    case XK_F13:   *class=WNP_FUNC_KEY;  *sym=WNP_F1_KEY;       break;
    case XK_F14:   *class=WNP_FUNC_KEY;  *sym=WNP_F2_KEY;       break;
    case XK_F15:   *class=WNP_FUNC_KEY;  *sym=WNP_F3_KEY;       break;
    case XK_F16:   *class=WNP_FUNC_KEY;  *sym=WNP_F4_KEY;       break;
    case XK_F17:   *class=WNP_FUNC_KEY;  *sym=WNP_F5_KEY;       break;
    case XK_F18:   *class=WNP_FUNC_KEY;  *sym=WNP_F6_KEY;       break;
    case XK_F19:   *class=WNP_FUNC_KEY;  *sym=WNP_F7_KEY;       break;
    case XK_F20:   *class=WNP_FUNC_KEY;  *sym=WNP_F8_KEY;       break;
    case XK_Prior: *class=WNP_FUNC_KEY;  *sym=WNP_PGUP_KEY;     break;
    case XK_Next:  *class=WNP_FUNC_KEY;  *sym=WNP_PGDOWN_KEY;   break;
    default:       *class=WNP_IGNORE_KEY;                       break;
    }
  }else *class=WNP_IGNORE_KEY;    
}

void WnpSetMouseLeft(int right)
{
  if (right) leftMouse=FALSE;
  else leftMouse=TRUE;
}

int WnpModState(unsigned int state)
{
  /* translate X modifier key state into internal format... */
 
  int r;

  r=0;
  if (leftMouse) {
    if (state&Button1Mask) r|=WNP_MOD_RIGHT;
    if (state&Button2Mask) r|=WNP_MOD_MIDDLE;
    if (state&Button3Mask) r|=WNP_MOD_LEFT;
    if (state&ShiftMask)   r|=WNP_MOD_SHIFT;
    if (state&ControlMask) r|=WNP_MOD_CTRL;
    if (state&Mod1Mask)    r|=WNP_MOD_ALT;
  }else{
    if (state&Button1Mask) r|=WNP_MOD_LEFT;
    if (state&Button2Mask) r|=WNP_MOD_MIDDLE;
    if (state&Button3Mask) r|=WNP_MOD_RIGHT;
    if (state&ShiftMask)   r|=WNP_MOD_SHIFT;
    if (state&ControlMask) r|=WNP_MOD_CTRL;
    if (state&Mod1Mask)    r|=WNP_MOD_ALT;
  }
  return(r);
}

void WnpGlBoxWinpos(int pos[4])
{
  Arg args[20];
  Position geom[4];
  Cardinal n;
  Widget w;

  geom[0]=pos[0]-dx; geom[1]=pos[1]-dy;
  geom[2]=pos[2]; geom[3]=pos[3];
  w=XtParent(XtParent(WnpGlBoxWidget()));    /* only granpa allows geom change... */
  n=0;
  XtSetArg(args[n],XtNx,geom[0]);n++;
  XtSetArg(args[n],XtNy,geom[1]);n++;
  XtSetArg(args[n],XtNwidth,geom[2]);n++;
  XtSetArg(args[n],XtNheight,geom[3]);n++;
  XtSetValues(w,args,n);
  txpos=pos[0]; typos=pos[1];
  calibratePosition=TRUE;
}

void WnpCalibratePosition(void)
{
  int pos[4];

  if (calibratePosition) {
    WnpGetGlBoxWinpos(pos);
    if (pos[0]!=txpos||pos[1]!=typos) {
      dx+=pos[0]-txpos;
      dy+=pos[1]-typos;
      pos[0]-=dx; pos[1]-=dy;
      WnpGlBoxWinpos(pos);
    }
    calibratePosition=FALSE;
  }
}

void WnpGlBoxFullScreen(int pos[4])
{
  WnpGlBoxWinpos(pos);  /* works only if WM cooperates... */
}

void WnpGetGlBoxWinpos(int pos[4])
{
  Arg args[20];
  Position geom[4];
  Cardinal n;
  Widget w;

  w=XtParent(XtParent(WnpGlBoxWidget()));    /* only granpa knows position... */
  n=0;
  XtSetArg(args[n],XtNx,geom);n++;
  XtSetArg(args[n],XtNy,geom+1);n++;
  XtSetArg(args[n],XtNwidth,geom+2);n++;
  XtSetArg(args[n],XtNheight,geom+3);n++;
  XtGetValues(w,args,n);
  pos[0]=geom[0]; pos[1]=geom[1]; 
  pos[2]=geom[2]; pos[3]=geom[3];
}

void WnpRingBell(void)
{
  XBell(WnpGlBoxDisplay(),0);
}

void WnpSetEventMask(long mask)
{
  XSetWindowAttributes set;
  XWindowAttributes get;

  /* note: 0x7fffffffL masks out sign-bit in your_event_mask... */

  XGetWindowAttributes(WnpGlBoxDisplay(),WnpGlBoxWindow(),&get);
  set.event_mask=(get.your_event_mask|mask)&0x7fffffffL;
  XChangeWindowAttributes(WnpGlBoxDisplay(),WnpGlBoxWindow(),CWEventMask,&set);
}

void WnpResetEventMask(long mask)
{
  XSetWindowAttributes set;
  XWindowAttributes get;

  /* note: 0x7fffffffL masks out sign-bit in your_event_mask... */

  XGetWindowAttributes(WnpGlBoxDisplay(),WnpGlBoxWindow(),&get);
  set.event_mask=(get.your_event_mask&(~mask))&0x7fffffffL;
  XChangeWindowAttributes(WnpGlBoxDisplay(),WnpGlBoxWindow(),CWEventMask,&set);
}

long WnpGetEventMask(void)
{
  XWindowAttributes get;

  XGetWindowAttributes(WnpGlBoxDisplay(),WnpGlBoxWindow(),&get);
  return(get.your_event_mask);
}

void WnpMouseMotionMask(int state)
{
  if (state) WnpSetEventMask(PointerMotionMask);
  else WnpResetEventMask(PointerMotionMask);
}

void WnpSetMouseTrap(int xmin,int ymin,int width, int height)
{
  /* trap pointer within box described by X11-window coords x,y,h,w */

  mtrap=TRUE;
  mtrapxmin=xmin;
  mtrapymin=ymin;
  mtrapxmax=xmin+width-1;
  mtrapymax=ymin+height-1;
}

void WnpRemoveMouseTrap(void)
{
  mtrap=FALSE;
}

void WnpTrapMouse(int *x,int *y)
{
  int dy,dx;
  
  /* note x,y are in X11-window coordinates, i.e. origin is upper-left */

  if (mtrap) {
    dx=clamp(*x,mtrapxmin,mtrapxmax);
    dy=clamp(*y,mtrapymin,mtrapymax);
    if (dx!=*x||dy!=*y) {
      XWarpPointer(WnpGlBoxDisplay(),WnpGlBoxWindow(),WnpGlBoxWindow(),0,0,0,0,dx,dy);
      *x=dx; *y=dy;
    }
  }
}

void WnpMakeGlBoxCurrent(void)
{
  GLwDrawingAreaMakeCurrent(wnpGlBoxWidget,wnpGlBoxContext);
}

void WnpSwapGlBuffers(void)
{
  glXSwapBuffers(wnpGlBoxDisplay,wnpGlBoxWindow);
}

int WnpDisplayWidth(void)
{
  return(win_fsgeom[2]);
}

int WnpDisplayHeight(void)
{
  return(win_fsgeom[3]);
}

Window WnpGlBoxWindow(void)
{
  return(wnpGlBoxWindow);
}

Display *WnpGlBoxDisplay(void)
{
  return(wnpGlBoxDisplay);
}

Widget WnpGlBoxWidget(void)
{
  return(wnpGlBoxWidget);
}

Colormap WnpGlBoxColormap(void)
{
  return(wnpGlBoxCmap);
}

int WnpButton(unsigned int xb)
{
  if (leftMouse) 
    return (xb==Button1 ? WNP_MOD_RIGHT
	    :(xb==Button2 ? WNP_MOD_MIDDLE
	      :(xb==Button3 ? WNP_MOD_LEFT:0)
	      )
	    ); 
  else
    return (xb==Button1 ? WNP_MOD_LEFT
	    :(xb==Button2 ? WNP_MOD_MIDDLE
	      :(xb==Button3 ? WNP_MOD_RIGHT:0)
	      )
	    ); 
}

/*==================== Callbacks ====================*/

void WnpGlBoxGinitCB(Widget w,XtPointer clientData,XtPointer callbackData)
{
  Dimension width,height;
  Arg args[1];
  XVisualInfo *visinf;

  XtSetArg(args[0],GLwNvisualInfo,&visinf);
  XtGetValues(w,args,1);
  
  wnpGlBoxWindow=XtWindow(w);
  wnpGlBoxDisplay=XtDisplay(w);
  wnpGlBoxContext=glXCreateContext(wnpGlBoxDisplay,visinf,0,TRUE);
  wnpGlBoxWidget=w;
  wnpGlBoxScreenNr=DefaultScreen(wnpGlBoxDisplay);
  wnpGlBoxCmap=DefaultColormap(wnpGlBoxDisplay,wnpGlBoxScreenNr);

  win_fsgeom[0]=0; win_fsgeom[1]=0;
  win_fsgeom[2]=DisplayWidth(wnpGlBoxDisplay,wnpGlBoxScreenNr);
  win_fsgeom[3]=DisplayHeight(wnpGlBoxDisplay,wnpGlBoxScreenNr);

  width=((GLwDrawingAreaCallbackStruct*)callbackData)->width;
  height=((GLwDrawingAreaCallbackStruct*)callbackData)->height;
  
  WnpMakeGlBoxCurrent();
  WnpInitGlBox(width,height);
  WnpInitCursor();
  WnpSetGlBoxCursor(WNP_CURSOR_ARROW);
  XStoreName(wnpGlBoxDisplay,wnpGlBoxWindow,"Wit!P Graphics Server");
  XStoreName(wnpGlBoxDisplay,XtWindow(XtParent(XtParent(w))),"Wit!P Graphics Server");
}

Boolean WnpWorkProcCB(XtPointer clientData)
{
  WnpWorkProc();
  return(FALSE);
}

void WnpGlBoxExposeCB(Widget w,XtPointer clientData,XtPointer callbackData)
{
  WnpCalibratePosition();
  WnpSetGlBoxRedrawFlag(TRUE);
  haveCB=TRUE;
}

void WnpGlBoxResizeCB(Widget w,XtPointer clientData,XtPointer callbackData)
{
  Dimension width,height;

  width=((GLwDrawingAreaCallbackStruct*)callbackData)->width;
  height=((GLwDrawingAreaCallbackStruct*)callbackData)->height;
  
  WnpGlBoxViewport(0,0,width,height);
}

void WnpGlBoxInputCB(Widget w,XtPointer clientData,XtPointer callbackData)
{
  XEvent *E;
  XButtonEvent *B;
  XMotionEvent *M;
  XKeyEvent *K;
  int  keysym,keyclass,x,y;
  char keytext[WNP_MAXLINE];

  /* NOTE: in Irix5.3 F1 and RETURN do not generate KeyPress event !!!
           ---> turn Press/Release handling around for this inferior OS */

  E=((GLwDrawingAreaCallbackStruct*)callbackData)->event;
  switch (E->type) {
  case KeyPress:
    K=(XKeyEvent *)E;
    WnpLookupKey(K,&keyclass,&keysym,keytext);
    switch (keyclass) {
    case WNP_ASCII_KEY:
      haveCB=TRUE;
      WnpAddKey(keytext[0],WnpModState(K->state));
      break;
    case WNP_IGNORE_KEY:
      /* simply ignored */
      break;
    case WNP_FUNC_KEY:
      haveCB=TRUE;
      WnpFunctionKey(keysym,WnpModState(K->state));
      break;
    case WNP_ARROW_KEY:
      haveCB=TRUE;
      WnpArrowKey(keysym,WnpModState(K->state));
      break;
    default:
      break;
    }
    break;
  case KeyRelease:
    /* ignored... */
    break;
  case ButtonPress:
    B=(XButtonEvent *)E;
    MouseButtonPress(WnpButton(B->button),WnpModState(B->state),B->x,B->y);
    break;
  case ButtonRelease:
    B=(XButtonEvent *)E;
    MouseButtonRelease(WnpButton(B->button),WnpModState(B->state),B->x,B->y);
    break;
  case MotionNotify:
    haveCB=TRUE;
    M=(XMotionEvent *)E;
    x=M->x; y=M->y;
    WnpTrapMouse(&x,&y);
    WnpGlBoxShowCursor();
    MouseMoved(WnpModState(M->state),x,y);
    break;
  default:
    fprintf(stderr,"other input event, ");
    break;
  }
}

int WnpResetCBFlag(void)
{
  int r;

  r=haveCB;
  haveCB=FALSE;
  return(r);
}
