#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include "wnp.h"
#include "wnpgl.h"

static float color_tab[WNP_MAXCOL][3] = {           /* colors for 3D graphics */
  { 0.0, 0.0, 0.0 },               /* black      */
  { 1.0, 0.0, 0.0 },               /* red        */
  { 0.0, 1.0, 0.0 },               /* green      */
  { 1.0, 1.0, 0.0 },               /* yellow     */
  { 0.0, 0.0, 1.0 },               /* blue       */
  { 1.0, 0.0, 1.0 },               /* magenta    */
  { 0.0, 1.0, 1.0 },               /* cyan       */
  { 1.0, 1.0, 1.0 },               /* white      */
  { 0.0, 0.0, 0.0 },               /* black      */
  { 1.0, 0.0, 0.0 },               /* red        */
  { 0.0, 1.0, 0.0 },               /* green      */
  { 1.0, 1.0, 0.0 },               /* yellow     */
  { 0.0, 0.0, 1.0 },               /* blue       */
  { 1.0, 0.0, 1.0 },               /* magenta    */
  { 0.0, 0.188235, 0.313726},      /* MenuBack   */
  { 1.0, 1.0, 1.0 },               /* MenuFront  */
  { 1.0, 1.0, 1.0 },               /* 16 */   
  { 1.0, 1.0, 1.0 },               /* 17 */     
  { 1.0, 1.0, 1.0 },               /* 18 */     
  { 1.0, 1.0, 1.0 },               /* 19 */     
  { 1.0, 1.0, 1.0 },               /* 20 */     
  { 1.0, 1.0, 1.0 },               /* 21 */     
  { 1.0, 1.0, 1.0 },               /* 22 */     
  { 1.0, 1.0, 1.0 },               /* 23 */     
  { 1.0, 1.0, 1.0 },               /* 24 */    
  { 1.0, 1.0, 1.0 },               /* 25 */    
  { 1.0, 1.0, 1.0 },               /* 26 */  
  { 1.0, 1.0, 1.0 },               /* 27 */         
  { 1.0, 1.0, 1.0 },               /* 28 */     
  { 1.0, 1.0, 1.0 },               /* 29 */     
  { 1.0, 1.0, 1.0 },               /* 30 */     
  { 1.0, 1.0, 1.0 },               /* 31 */   
};

/* Lightsource 0 : */

static float light_diffuse[] = { 1.0,1.0,1.0,1.0 };
static float light_specular[]= { 1.0,1.0,1.0,1.0 };
static float light_position[]= { 50.0,50.0,100.0,1.0};
static float ambient_light[] = { 0.1,0.1,0.1,1.0};

static float spclr_tab[WNP_MAXSCOL][3] = {          /* colors for menu and message box */
  { 0.0, 0.188235, 0.313726},                       /* MenuBack                        */
  { 1.0, 1.0, 1.0 },                                /* MenuText                        */
  { 0.8, 0.8, 0.8 },                                /* MenuHBack                       */
  { 0.0, 0.188235, 0.313726},                       /* MenuHTtxt                       */
  { 1.0, 0.0, 0.0 },                                /* MenuCursor (not used)           */
};

static int color = WNP_MAXCOL;                      /* index of current color          */
static int ccolor = -1;                             /* index of current clear color    */
static GLfloat fogColor[4] = {0.0, 0.0, 0.0, 1.0};  /* current fog color color         */     
static GLfloat clearColor[4] = {0.0, 0.0, 0.0, 1.0};/* current clear color             */ 
static float fogRange[2];

static int    dotN = 0;                             /* number of entries in old style dot-prop map */
static int   *dotMap = (int *)NULL;                 /* color indices of old style map              */
static float *dotLim = (float *)NULL;               /* limits of old style map                     */

int WnpNoColor(void)
{
  int r;
  
  r=color;
  color=WNP_MAXCOL;
  return(r);
}

int WnpColor(int col)
{
  int r;

  r=color;

  if (color!=col) {
    if (col>=0 && col <WNP_MAXCOL) {
      glColor3fv(color_tab[col]);
    }else if (col<0 && col >= -WNP_MAXSCOL) {
      glColor3fv(spclr_tab[-col-1]);
    }else{
      glColor3fv(color_tab[WNP_DEFAULT_COLOR]);
    }
    color=col; 
  }
  return(r);
}

void WnpMaterialColor(int face,int mode,int col,float alpha,float inte)
{
  float v[4];

  col=clamp(col,0,WNP_MAXCOL);
  if (col==WNP_MAXCOL) col=WNP_DEFAULT_COLOR;
  v[0]=inte*color_tab[col][0];
  v[1]=inte*color_tab[col][1];
  v[2]=inte*color_tab[col][2];
  v[3]=alpha;
  glMaterialfv(face,mode,v);
}

void WnpDotColorMap(int n,int *map,float *lim)
{
  int i;

  /* setup old style property map for dots... */

  if (dotMap) free(dotMap); dotMap=(int *)NULL;
  if (dotLim) free(dotLim); dotLim=(float *)NULL;
  dotN=0;

  if (n>1) {
    dotMap=malloc(n*sizeof(int));
    dotLim=malloc((n-1)*sizeof(float));
    if (dotMap && dotLim) {
      for (i=0;i<n;i++) dotMap[i]=map[i];
      for (i=0;i<n-1;i++) dotLim[i]=lim[i];
      dotN=n;
    }else{
      if (dotMap) free(dotMap); dotMap=(int *)NULL;
      if (dotLim) free(dotLim); dotLim=(float *)NULL;
    }
  }
}

int WnpColorDef(int idx,float R,float G, float B)
{
  int r;

  /* the name is lost... */

  if (idx>=0 && idx < WNP_MAXCOL) {
    color_tab[idx][0]=R;
    color_tab[idx][1]=G;
    color_tab[idx][2]=B;
    r=1;
  }else r=0;
  return(r);
}

void WnpDotColor(float val)
{
  int i;

  if (dotN>1) {
    for (i=0;i<dotN-1;i++) if (val<dotLim[i]) break;
    WnpColor(dotMap[i]);
  }else WnpColor(1);
}

void WnpColorInit(void)               /* here's some dirty work... */
{
  /* the following is for depth-cueing */

  /* NOTE: Brookdale-G on ENGINE/XP can't do GL_LINEAR Fog for bitmaps!!! */
  /*       14-Jan-2003                                                    */

  glEnable(GL_FOG);
  glFogi (GL_FOG_MODE, GL_LINEAR);
  /* Mesa: incorrect depthcue in ortho for GL_NICEST ( 12-Jul-2003) */
  glHint (GL_FOG_HINT, GL_FASTEST); 
  glFogf (GL_FOG_START, 0.0);
  glFogf (GL_FOG_END, 200.0);
  glFogfv (GL_FOG_COLOR, fogColor);
  glDepthFunc(GL_LEQUAL);
  glDepthRange(0.0,1.0);
  glEnable(GL_DEPTH_TEST);
  glShadeModel(GL_FLAT);

  /* and now for smooth lines... */
  
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glHint(GL_LINE_SMOOTH_HINT,GL_FASTEST);
  WnpVLineWidth(1);        /* this will enable smooth lines... */

  /* light definitions ... */

  glLightfv(GL_LIGHT0,GL_DIFFUSE,light_diffuse);
  glLightfv(GL_LIGHT0,GL_SPECULAR,light_specular);
  glLightfv(GL_LIGHT0,GL_POSITION,light_position);

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,ambient_light);

  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHTING);
}

void WnpFogRange(float znear,float zfar)
{
  fogRange[0]=znear;fogRange[1]=zfar;
  glFogf(GL_FOG_START,fogRange[0]);
  glFogf(GL_FOG_END,fogRange[1]);
}

void WnpSetFogColor(float *rgb)
{
  fogColor[0]=rgb[0];
  fogColor[1]=rgb[1];
  fogColor[2]=rgb[2];
  glFogfv(GL_FOG_COLOR,fogColor);
}

void WnpSetBackgroundColor(float *rgb)
{
  clearColor[0]=rgb[0];
  clearColor[1]=rgb[1];
  clearColor[2]=rgb[2];
  if (ccolor==WNP_THREED_BACKGROUND) ccolor=-1;
}

void WnpClear(int color)
{
  if (ccolor!=color) {
    if (color==WNP_THREED_BACKGROUND) glClearColor(clearColor[0],clearColor[1],clearColor[2],clearColor[3]);
    else glClearColor(spclr_tab[color-1][0],spclr_tab[color-1][1],spclr_tab[color-1][2],1.0);
  }
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  ccolor=color;
}

void WnpMoveLight(float *dx)
{
  light_position[0]+=dx[0];
  light_position[1]+=dx[1];
  light_position[2]+=dx[2];
  glLightfv(GL_LIGHT0,GL_POSITION,light_position);
}

void WnpRotateLight(float a)
{
  float x[3];
  WnpMatrix m;
  int i,j;

  glPushMatrix();
  glLoadIdentity();
  glRotatef(a,0,0,1);
  glGetFloatv(GL_MODELVIEW_MATRIX,(float *)m);
  glPopMatrix();
  for (i=0;i<3;i++) for(j=0,x[i]=0;j<3;j++) x[i]+=light_position[j]*m[j][i];
  for (i=0;i<3;i++) light_position[i]=x[i];
  glLightfv(GL_LIGHT0,GL_POSITION,light_position);
}

void WnpToggleLightSourceLocal(void)
{
  if (light_position[3]>0.5) {
    light_position[3]=0.0;
    WnpSendToCom("light source is local");
  }else  {
    light_position[3]=1.0;
    WnpSendToCom("light source is at infinity");
  }
  glLightfv(GL_LIGHT0,GL_POSITION,light_position);
  WnpSetThreeDRedrawFlag(TRUE);
}

void WnpAdjustLightIntensity(int scl)
{
  float s;
  int i;

  s=1.0+0.002*scl;
  for (i=0;i<3;i++) light_diffuse[i]=clamp(s*light_diffuse[i],0.01,1.0);
  for (i=0;i<3;i++) light_specular[i]=clamp(s*s*light_specular[i],0.01,1.0);
  glLightfv(GL_LIGHT0,GL_SPECULAR,light_specular);
  glLightfv(GL_LIGHT0,GL_DIFFUSE,light_diffuse);
  WnpSetThreeDRedrawFlag(TRUE);
}

void WnpAdjustAmbientLightIntensity(int scl)
{
  float s;
  int i;

  s=1.0+0.002*scl;
  for (i=0;i<3;i++) ambient_light[i]=clamp(s*ambient_light[i],0.01,1.0);
  glLightfv(GL_LIGHT0,GL_AMBIENT,ambient_light);
  WnpSetThreeDRedrawFlag(TRUE);
}
