#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include "wnp.h"
#include "../client/transpor.h"

#define hdr_is(x) (!strncmp(x,hdr,4))

static int debug=FALSE;
static int chan=0;                  /* communication channel to client... */
static int ptrsize=4;               /* my partner's pointer size */
static int adtsz,artsz;

static void         *read_ptr(void);
static cptr_t        read_cptr(void);
static void        **read_ptra(int n);
static cptr_t       *read_cptra(int n);
static int           read_int(void);
static void          read_ints(int *inta,int m,int n);
static int          *read_inta(int n);
static float         read_float(void);
static void          read_floats(float *fbuf,int m,int n);
static float        *read_floata(int n);
static void          read_matrix(WnpMatrix m);
static void          read_text(char *buf,int max,int n);
static char         *read_texta(int n);
static WnpPrj        read_prj(void);
static WnpAtom      *read_atom(void);
static WnpAtom      *read_atomd(void);
static WnpBond      *read_bond(void);
static WnpMDS       *read_mdsa(void);
static WnpTOBJ      *read_tobja(void);
static WnpTRG       *read_trgvna(void);
static WnpNOE       *read_noea(void);
static void          read_trg(TRGObj *o);
static void          write_ptr(char *hdr,void *ptr);
static void          write_cptr(cptr_t cptr);
static void          write_prj(WnpPrj prj);
static void          write_float(float x);
static void          write_floata(int n,float *x);
static void          write_floats(char *hdr,int n,float *val);
static void          write_matrix(WnpMatrix m);
static void          sendObject(DspObj *obj);
static void          sendAtomDots(int n,WnpAtom **A);
static void          send1AtomDots(WnpDots *d);
static void          sendMonitorPairs(WnpMonPair *m);

int WnpOpenTransport(int client)
{
  /* open buffered communication channel on socket */

  chan=open_transport(client,8192,8192);
  transport_wait(chan,FALSE);             /* we always use non-blocking I/O on this channel */
  if (sizeof(void *)!=ptrsize) ptrsize=1; /* trigger error, if PTSZ is not set by client... */
  adtsz=ptrsize+4*sizeof(int)+4;
  artsz=sizeof(void *)+4*sizeof(int)+4;
  return(chan);
}

int WnpCommDebug(void)
{
  debug=!debug;
  return(debug);
}

int WnpCommQueue(int *back)
{
  int busy,size,i,n,m,typ,R,G,B;
  char *hdr;
  char buf[WNP_MAXREC+1];
  int  ibuf[WNP_MAXREC],pmap[WNP_MAXCOL];
  float fbuf[WNP_MAXREC],fval,plim[WNP_MAXCOL],rgb[3];
  void *obj;
  cptr_t cobj,cson,*ptr;
  DspObj *dsp;
  MdlObj *mdl;
  GlblObj *glbl;
  WnpNOE *noe;
  WnpAtom *a,*pa,**ptra;
  WnpBond *b,*pb;
  WnpPrj prj;
  WnpMatrix mat;

  busy=FALSE; *back=FALSE;
  while ((hdr=got_message(chan))) {

    size=message_size(chan); busy=TRUE;
    if (debug) fprintf(stderr,"%4s %d\n",hdr,size);

    if (hdr_is("EVTQ")) {                    /* turn attention to event-q           */
      *back=TRUE;                            /* return and hand over to X11         */
      discard_message(chan);
      break;

    }else if (hdr_is("PTSZ")) {              /* my partner's pointer size           */
      ptrsize=read_int();
      send_bheader(chan,"PTSZ",sizeof(int));
      send_int(chan,sizeof(void *));
      adtsz=ptrsize+4*sizeof(int)+4;
      artsz=sizeof(void *)+4*sizeof(int)+4;
      
    }else if (hdr_is("DUSA")) {              /* disable usa-picking                 */
      read_cptr();                           /* the pick root... ignored.           */
      WnpSetPickMode(WNP_PICK_USA_OFF);
 
    }else if (hdr_is("EUSA")) {              /* enable usa-picking                  */
      read_cptr();
      WnpSetPickMode(WNP_PICK_USA_ON);

    }else if (hdr_is("POFF")) {              /* atom/object/xyz pick off            */
      read_cptr();
      WnpSetPickMode(WNP_PICK_OFF);

    }else if (hdr_is("PATM")) {              /* atom pick mode                      */
      WnpSetAtomPickRoot(MyObjPtr(read_cptr()));
      WnpSetPickMode(WNP_PICK_ATOM);

    }else if (hdr_is("POBJ")) {              /* object pick mode                    */
      WnpSetObjectPickRoot(MyObjPtr(read_cptr()));
      WnpSetPickMode(WNP_PICK_OBJECT);

    }else if (hdr_is("PXYZ")) {              /* point pick mode                     */
      read_cptr();
      WnpSetPickMode(WNP_PICK_XYZ);

    }else if (hdr_is("PXYA")) {              /* point/atom pick mode                */
      WnpSetAtomPickRoot(MyObjPtr(read_cptr()));
      WnpSetPickMode(WNP_PICK_XYZ_ATOM);

    }else if (hdr_is("AOBJ")) {              /* add object to graphics              */
      if (size==2*ptrsize+sizeof(int)) {
	cobj=read_cptr(); cson=read_cptr(); typ=read_int();
	obj=WnpAddDspObj(cobj,cson,typ);
      }else obj=(DspObj *)(NULL);
      write_ptr("OBJ*",obj);

    }else if (hdr_is("AMDS")) {              /* add move-draw sequence              */
      dsp=MyObjPtr(read_cptr());
      WnpAddMDS((CSObj *)dsp,read_mdsa());

    }else if (hdr_is("ATOB")) {              /* add text object                     */
      dsp=MyObjPtr(read_cptr());
      WnpAddTOBJ((CSObj *)dsp,read_tobja());

    }else if (hdr_is("ATRV")) {              /* add vertices & normals to surface   */
      dsp=MyObjPtr(read_cptr());
      WnpAddTRG((TRGObj *)dsp,read_trgvna());
      
    }else if (hdr_is("ATRF")) {              /* add triangle list to surface        */
      dsp=MyObjPtr(read_cptr());
      read_trg((TRGObj *)dsp);

    }else if (hdr_is("TRGS")) {              /* rendering style for surfaces        */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      for (i=0;i<n;i++) WnpSetDisplayStyle(MyObjPtr(read_cptr()),m);

    }else if (hdr_is("TRGT")) {              /* set transparency level of surface   */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      for (i=0;i<n;i++) WnpSetTransparency(MyObjPtr(read_cptr()),m);

    }else if (hdr_is("GGOB")) {              /* get graphics objects                */
      n=size/ptrsize;
      for (i=0;i<n;i++) sendObject(MyObjPtr(read_cptr()));

    }else if (hdr_is("SALT")) {              /* select alternate object             */
      if (size==ptrsize) glbl=(GlblObj *)MyObjPtr(read_cptr());
      else glbl=(GlblObj *)NULL;
      WnpSetAltObj(glbl);
      
    }else if (hdr_is("DSPO")) {              /* set display status of object        */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      if (m==-1) {
	for (i=0;i<n;i++) WnpToggleObjectFlags(MyObjPtr(read_cptr()),OBJ_FLG_DSP);
      }else {
	for (i=0;i<n;i++) WnpNewObjectFlags(MyObjPtr(read_cptr()),OBJ_FLG_DSP,m);
      }

    }else if (hdr_is("SOLW")) {              /* set object linewidth                */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      for (i=0;i<n;i++) WnpSetObjectLW(MyObjPtr(read_cptr()),m);

    }else if (hdr_is("MDST")) {              /* set mds_type of CSO                 */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      for (i=0;i<n;i++) WnpSetObjectMDST(MyObjPtr(read_cptr()),m);

    }else if (hdr_is("DSPC")) {              /* set display color of object         */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      for (i=0;i<n;i++) WnpSetObjectColor(MyObjPtr(read_cptr()),m);

    }else if (hdr_is("SBFC")) {              /* set backface  color of object       */
      n=(size-sizeof(int))/ptrsize;
      m=read_int();
      for (i=0;i<n;i++) WnpSetBackfaceColor(MyObjPtr(read_cptr()),m);

    }else if (hdr_is("DOBJ")) {              /* delete display object               */
      if (size==ptrsize) WnpDeleteDspObj(MyObjPtr(read_cptr()));
      
    }else if (hdr_is("ROBT")) {              /* reset object transformation         */
      while (size>=ptrsize) {
	WnpResetMVT(MyObjPtr(read_cptr()));
	size-=ptrsize;
      }

    }else if (hdr_is("SCDR")) {              /* set display scale relative          */
      if (size==sizeof(float)) {
	WnpThreeDScaleRel(read_float());
      }

    }else if (hdr_is("TRDO")) {              /* translate display object            */
      if (size==ptrsize+3*sizeof(float)) {
	glbl=WnpAssocGlbl(MyObjPtr(read_cptr()));
	read_floats(fbuf,WNP_MAXREC,3);
	WnpThreeDTranslate(glbl,fbuf);
      }

    }else if (hdr_is("RODO")) {              /* rotate display object               */
      if (size==ptrsize+4*sizeof(float)) {
	glbl=WnpAssocGlbl(MyObjPtr(read_cptr()));
	fval=read_float();
	read_floats(fbuf,WNP_MAXREC,3);
	WnpThreeDRotate(glbl,fbuf,fval);
      }

    }else if (hdr_is("GRTP")) {              /* get rotation, translation and pivot of global */
      n=size/ptrsize;                        /* response is packed into WnpMatrix (R,T), last row=pivot */
      ptr=read_cptra(n);
      if (ptr) {
	send_bheader(chan,"GRTP",n*sizeof(WnpMatrix));
	for (i=0;i<n;i++) {
	  WnpGetRTP(mat,MyObjPtr(ptr[i]));
	  write_matrix(mat);
	}
	free(ptr);
      }else send_header(chan,"GRTP",0);

    }else if (hdr_is("SRTP")) {              /* set rotation, translation and pivot of global */  
      n=size/(ptrsize+sizeof(WnpMatrix));    /* data is packed into WnpMatrix (R,T), last row=pivot */
      for (i=0;i<n;i++) {
	cobj=read_cptr();
	read_matrix(mat);
	WnpSetRTP(mat,MyObjPtr(cobj));
      }

    }else if (hdr_is("NPIV")) {              /* set pivot                           */
      if (size==ptrsize+3*sizeof(float)) {
	glbl=WnpAssocGlbl(MyObjPtr(read_cptr()));
	read_floats(fbuf,WNP_MAXREC,3);
	WnpSetPivot(glbl,fbuf);
      }
      
    }else if (hdr_is("GMMx")) {              /* get 3D modelling/viewing matrix     */
      n=size/ptrsize;
      ptr=read_cptra(n);
      if (ptr) {
	send_bheader(chan,"GMMx",n*sizeof(WnpMatrix));
	if (n>0) WnpUpdMVT();
	for (i=0;i<n;i++) {
	  WnpGetMVT(mat,MyObjPtr(ptr[i]));
	  write_matrix(mat);
	}
	free(ptr);
      }else send_header(chan,"GMMx",0);

    }else if (hdr_is("FRZE")) {              /* freeze object transformation        */
      while (size>=ptrsize) {
	WnpFreezeMVT((DspObj *)WnpAssocGlbl(MyObjPtr(read_cptr())),mat);
	size-=sizeof(obj);
      }

    }else if (hdr_is("TOBJ")) {              /* read and apply transformation       */
      if (size==sizeof(WnpMatrix)+ptrsize) {
	dsp=MyObjPtr(read_cptr()); read_matrix(mat);
	WnpTransformObject(dsp,mat);
      }else WnpSendToCom("ERROR: bad modelling/viewing matrix.");

    }else if (hdr_is("AATM")) {              /* add atoms to MDL                    */
      if (size>=ptrsize) mdl=(MdlObj *)MyObjPtr(read_cptr());
      else mdl=(MdlObj *)NULL;
      if (mdl&&mdl->type==DSP_MDL) n=(size-ptrsize)/adtsz;
      else n=0;
      pa=(n>0 ? read_atom() : (WnpAtom *)NULL);
      for (a=pa,i=1;i<n&&a;i++,a=a->nxt) a->nxt=read_atom();
      send_bheader(chan,"ATM*",i*sizeof(a));
      for (a=pa;a;a=pa) {
	pa=a->nxt;
	WnpAddAtomToMdl(a,mdl);
	send_bptr(chan,a);
      }

    }else if (hdr_is("AACH")) {              /* add atomchain                       */
      n=size/(ptrsize+adtsz);
      if (n>0) {
	mdl=(MdlObj *)MyObjPtr(read_cptr());
	pa=read_atom(); if (pa) pa->up=mdl;
      }else pa=(WnpAtom *)NULL;
      for (a=pa,i=1;i<n&&a;i++,a=a->nxt) {
	mdl=(MdlObj *)MyObjPtr(read_cptr());
	a->nxt=read_atom(); if (a->nxt) a->nxt->up=mdl;
      }
      send_bheader(chan,"ATM*",i*sizeof(a));
      for (a=pa;a;a=pa) {
	pa=a->nxt;
	WnpAddAtomToMdl(a,a->up);
	send_bptr(chan,a);
      }

    }else if (hdr_is("NEWX")) {              /* new atom coordinates                */
      n=size/(sizeof(a)+3*sizeof(float));
      for (i=0;i<n;i++) {
	a=read_ptr(); read_floats(fbuf,3,3);
	WnpNewAtomXYZ(a,fbuf);
      }

    }else if (hdr_is("ATCO")) {              /* set atom color                      */
      m=read_int();
      n=(size-sizeof(int))/sizeof(a);
      for (i=0;i<n;i++) WnpNewAtomColor(read_ptr(),m);

    }else if (hdr_is("ATDS")) {              /* set atom display status             */
      m=read_int();
      n=(size-sizeof(int))/sizeof(a);
      for (i=0;i<n;i++) WnpNewAtomFlags(read_ptr(),ATM_FLG_DSP,m);

    }else if (hdr_is("ARST")) {		     /* atom render status (sphere)         */
      m=read_int();
      n=(size-sizeof(int))/sizeof(a);
      for (i=0;i<n;i++) WnpNewAtomFlags(read_ptr(),ATM_FLG_SPHERE,m);

    }else if (hdr_is("ASTR")) {              /* atom-set-radius                     */
      n=size/(sizeof(a)+sizeof(float));
      for (i=0;i<n;i++) {
	a=read_ptr(); WnpNewAtomRadius(a,read_float());
      }

    }else if (hdr_is("ASCR")) {              /* atom-scale-radius                   */
      fval=read_float();
      n=(size-sizeof(float))/(sizeof(a));
      for (i=0;i<n;i++) WnpScaleAtomRadius(read_ptr(),fval);

    }else if (hdr_is("ALCL")) {              /* set atom label color                */
      n=(size-sizeof(int))/(sizeof(a));
      m=read_int();
      for (i=0;i<n;i++) WnpSetAtomLabelColor(read_ptr(),m);

    }else if (hdr_is("SLDC")) {              /* set default atom label color        */
      n=size/sizeof(int);
      if (n>0) WnpSetDefaultAtomLabelColor(read_int());
    }else if (hdr_is("ALSZ")) {              /* set atom label size (A)             */
      WnpSetAtomLabelSize(read_float());

    }else if (hdr_is("ALTX")||               /* new atom label text                 */
	      hdr_is("ALTN")) {              /* new atom label                      */
      n=size-sizeof(a); a=read_ptr();
      WnpSetAtomLabel(a,read_texta(n));
      if (!strncmp(hdr,"ALTN",4)) WnpNewAtomFlags(a,ATM_FLG_LDSP,TRUE);

    }else if (hdr_is("ALRM")) {              /* remove atom label                   */
      n=size/sizeof(a);
      for (i=0;i<n;i++) WnpRemoveAtomLabel(read_ptr());

    }else if (hdr_is("ALST")) {              /* set atom label status               */
      n=(size-sizeof(int))/(sizeof(a));
      m=read_int();
      for (i=0;i<n;i++) WnpNewAtomFlags(read_ptr(),ATM_FLG_LDSP,m);

    }else if (hdr_is("DOTA")) {              /* add dots structure for one atom     */
      n=(size-sizeof(a))/(4*sizeof(float));
      a=read_ptr();
      WnpReplaceDots(a,n,read_floata(4*n));

    }else if (hdr_is("CDOT")) {              /* copy dots                           */
      n=size/2*sizeof(a);
      for (i=0;i<n;i=i+2) {
	a=read_ptr();
	WnpCopyDots(read_ptr(),a);
      }

    }else if (hdr_is("DOTR")) {              /* remove dot structures for atoms     */
      n=size/sizeof(a);
      for (i=0;i<n;i++) WnpDeleteDots(read_ptr());

    }else if (hdr_is("DOTC")) {              /* change color of dots                */
      n=(size-sizeof(int))/sizeof(a);
      m=read_int();
      for (i=0;i<n;i++) WnpColorDots(read_ptr(),m);

    }else if (hdr_is("DOTS")) {              /* change style of dots                */
      n=(size-sizeof(int))/sizeof(a);
      m=read_int();
      for (i=0;i<n;i++) WnpDotsStyle(read_ptr(),m);

    }else if (hdr_is("DOTD")) {              /* change dots display status for atoms*/
      n=(size-sizeof(int))/sizeof(a);
      m=read_int();
      for (i=0;i<n;i++) WnpDisplayDots(read_ptr(),m);

    }else if (hdr_is("DOTM")) {              /* Dots property/color map (old style) */
      n=((size-sizeof(int))/(sizeof(int)+sizeof(float)));
      n=clamp(n,0,WNP_MAXCOL-1);
      if (n>0&&n<WNP_MAXCOL) {
	pmap[0]=read_int();
	for (i=0;i<n;i++) {
	  plim[i]=read_float();
	  pmap[i+1]=read_int();
	}
	WnpDotColorMap(n+1,pmap,plim);
      }

    }else if (hdr_is("GDOT")) {              /* get dots */
      n=size/sizeof(WnpAtom *);
      ptra=(WnpAtom **)read_ptra(n);
      sendAtomDots(n,ptra);
      if (ptra) free(ptra);
      ptra=NULL;

    }else if (hdr_is("RATM")) {              /* replace atom in mdl, keep XTRA data */
      n=size/artsz;
      for (i=0;i<n; i++) {
	a=read_ptr();
	WnpReplaceAtom(a,read_atomd());
      }
      
    }else if (hdr_is("DATM")) {              /* delete atom from mdl                */
      n=size/sizeof(a);
      for (i=0;i<n;i++) WnpRemoveAtom(read_ptr());

    }else if (hdr_is("ASMK")) {              /* set atom mark                       */
      n=size/sizeof(void *);
      for (i=0;i<n;i++) WnpSetAtomMark(read_ptr(),TRUE);

    }else if (hdr_is("ARMK")) {              /* remove atom mark                    */
      n=size/sizeof(void *);
      if (n>0) for (i=0;i<n;i++) WnpSetAtomMark(read_ptr(),FALSE);
      else WnpClearAllAtomMarks();

    }else if (hdr_is("AMON")) {              /* add distance monitor pair           */
      n=size/(2*sizeof(a));
      for (i=0;i<n;i++) {
	a=read_ptr();
	WnpAddMonitorPair(a,read_ptr());
      }

    }else if (hdr_is("GMON")) {              /* get distance monitor pair list      */
      sendMonitorPairs(WnpMonitorPairs());

    }else if (hdr_is("CMON")) {              /* clear distance monitor pair list    */
      WnpClearMonitorPairs();

    }else if (hdr_is("MONL")) {              /* set mon-label status                */
      WnpMonitorLabel(read_int());

    }else if (hdr_is("MONS")) {              /* set monitor status                  */
      WnpMonitorStatus(read_int());
      
    }else if (hdr_is("MONC")) {              /* set monitor colour                  */
      WnpMonitorColor(read_int());

    }else if (hdr_is("MONW")) {              /* set monitor linewidth               */
      WnpMonitorWidth(read_int());

    }else if (hdr_is("ANOE")) {              /* add NOEs                            */
      while (noe=read_noea()) WnpAddNOE(noe);

    }else if (hdr_is("DNOE")) {              /* delete NOEs                         */
      n=size/sizeof(dsp);
      if (n==0) WnpRemoveNOEs((DspObj *)WnpDisplayRoot());
      else for (i=0;i<n;i++) WnpRemoveNOEs(read_ptr());

    }else if (hdr_is("SNOE")) {              /* set NOE display attributes          */
      n=size/(sizeof(obj)+3*(sizeof(int)));
      for (i=0;i<n;i++) {
	dsp=read_ptr(); read_ints(ibuf,3,3);
	WnpSetNOEAttributes(dsp,ibuf[0],ibuf[1],ibuf[2]);
      }

    }else if (hdr_is("ABND")) {              /* add bonds                           */
      n=size/(2*sizeof(void *));
      pb=(n>0 ? read_bond() : (WnpBond *)NULL);
      for (b=pb,i=1;i<n&&b;i++,b=b->nxt) b->nxt=read_bond();
      send_bheader(chan,"BND*",i*sizeof(b));
      for (b=pb;b;b=pb) {
	pb=b->nxt;
	WnpAddBond(b);
	send_bptr(chan,b);
      }

    }else if (hdr_is("DBND")) {              /* delete bond from mdl                */
      n=size/sizeof(b);
      for (i=0;i<n;i++) WnpRemoveBond(read_ptr());

    }else if (hdr_is("BSOL")) {              /* bond rendering flag                 */
      n=(size-sizeof(int))/sizeof(b);
      m=read_int();
      for (i=0;i <n; i++) WnpNewBondFlags(read_ptr(),BND_FLG_CYLINDER,m);

    }else if (hdr_is("BRAD")) {              /* modify cylinder radius of bonds     */
      n=(size-sizeof(float))/sizeof(b);
      fval=read_float();
      for (i=0;i <n; i++) WnpSetBondRadius(read_ptr(),fval);

    }else if (hdr_is("BLST")) {              /* set bond label status [never used]  */
      n=(size-sizeof(int))/(sizeof(b));
      m=read_int();
      for (i=0;i<n;i++) WnpNewBondFlags(read_ptr(),BND_FLG_LDSP,m);

    }else if (hdr_is("BLTN")) {              /* new bond label                      */
      n=size-sizeof(b); b=read_ptr();
      WnpSetBondLabel(b,read_texta(n));

    }else if (hdr_is("BLRM")) {              /* remove bond label                   */
      n=size/sizeof(b);
      for (i=0;i<n;i++) WnpRemoveBondLabel(read_ptr());

    }else if (hdr_is("BLCL")) {              /* set bond label color                */
      n=(size-sizeof(int))/(sizeof(b));
      m=read_int();
      for (i=0;i<n;i++) WnpSetBondLabelColor(read_ptr(),m);

    }else if (hdr_is("SBLW")) {              /* set bond linewidth                  */
      n=(size-sizeof(int))/(sizeof(b));
      m=read_int();
      for (i=0;i <n; i++) WnpSetBondLW(read_ptr(),m);

    }else if (hdr_is("SBDW")) {              /* set default bond width              */
      n=(size)/(sizeof(int));
      if (n>0) WnpSetDefaultBondLW(read_int());

    }else if (hdr_is("SLWM")) {              /* set virtual linewidth map           */
      n=size/sizeof(float);
      if (n==4) {
	read_floats(fbuf,4,4);
	WnpSetLWMap(fbuf);
      }

    }else if (hdr_is("*MSG")) {              /* post a message                      */
      read_text(buf,WNP_MAXREC,size);
      WnpSendToCom("%s",buf);

    }else if (hdr_is("LDBT")) {              /* load menu button                    */
      if (size>=sizeof(int)) {
	m=read_int(); read_text(buf,WNP_MAXREC,size-sizeof(int));
	WnpLoadMenuButtons(WnpMainMenu(),m,buf);
      }

    }else if (hdr_is("SPHM")) {	             /* set sphere mode                     */
      m=read_int(); n=read_int();
      WnpSetSphereMode(m,n);

    }else if (hdr_is("SCYM")) {	             /* set cylinder mode                   */
      m=read_int(); n=read_int();
      WnpSetCylinderMode(m,n);

    }else if (hdr_is("CLMN")) {              /* clear menu                          */
      WnpClearMenu(WnpMainMenu());

    }else if (hdr_is("CLBT")) {              /* clear menu button                   */
      WnpClearMenuButton(WnpMainMenu(),read_int());

    }else if (hdr_is("RFMN")) {              /* refresh menu                        */
      WnpRefreshMenu(WnpActiveMenu());

    }else if (hdr_is("PRMT")) {              /* set the prompt string               */
      read_text(buf,WNP_MAXREC,size);
      WnpRefreshMenu(WnpActiveMenu());
      WnpSetPrompt(buf);

    }else if (hdr_is("SENV")) {              /* set environment variable            */
      read_text(buf,WNP_MAXREC,size);
      WnpSetenv(buf);

    }else if (hdr_is("DRAX")) {              /* define axis of rotation             */
      n=size/sizeof(a);
      WnpNewRotAxis(n,(WnpAtom **)read_ptra(n));
	
    }else if (hdr_is("SRAX")) {              /* modify angle of rotation            */
      n=size/sizeof(float);
      if (n>0) WnpIncRotAxis(read_float());
      else WnpResetRotAxis();

    }else if (hdr_is("FRAX")) {              /* freeze angle of rotation            */
      fval=WnpFreezeRotAxis();
      write_floats("SRAX",1,&fval);

    }else if (hdr_is("SPRJ")) {              /* set 3D projection                   */
      if (size>=sizeof(prj)) {
	prj=read_prj();
	prj.viewmode=WnpGetViewMode();
	prj.stereomode=WnpGetStereoMode();
	WnpSetThreeDPrj(prj);
      }else WnpSendToCom("ERROR: bad projection structure.");

    }else if (hdr_is("GPRJ")) {              /* get 3D projection                   */
      write_prj(WnpGetThreeDPrj());

    }else if (hdr_is("SSTM")) {              /* set stereo mode                     */
      WnpSetStereoMode(read_int());

    }else if (hdr_is("SVWM")) {              /* set view mode                       */
      read_text(buf,WNP_MAXREC,size);
      if (strncmp(buf,"PERS",4)==0) WnpSetViewMode(WNP_VIEW_PERSPECTIVE);
      else if (strncmp(buf,"ORTH",4)==0) WnpSetViewMode(WNP_VIEW_ORTHO);

    }else if (hdr_is("SMEN")) {              /* set menu mode                       */
      WnpMenuCommState(!read_int());

    }else if (hdr_is("CLDF")) {              /* color definition                    */
      m=read_int(); R=read_int(); G=read_int(); B=read_int();
      read_text(buf,WNP_MAXREC,size-4*sizeof(int)); /* mnemonic, ignored... */
      WnpColorDef(m,R/255.0,G/255.0,B/255.0);
      WnpSetGlBoxRedrawFlag(TRUE);

    }else if (hdr_is("MMAP")) {              /* set mouse key map                   */
      if (size==(WNP_MAX_MOUSE_FUNC+1)*sizeof(int)) {
	m=read_int(); 
	read_ints(ibuf,WNP_MAXREC,WNP_MAX_MOUSE_FUNC);
	WnpSetMouseMap(ibuf); WnpSetAltObjKey(m);
      }else WnpSendToCom("ERROR: bad mouse map");

    }else if (hdr_is("SMFU")) {	             /* set mouse function                  */
      n=read_int(); m=read_int();
      MouseMapFunction(m,n);

    }else if (hdr_is("SMLR")) {              /* set mouse left-or-right             */
      WnpSetMouseLeft(read_int());

    }else if (hdr_is("SRGB")) {              /* set RGB mode, always on in oGL      */
      discard_message(chan);

    }else if (hdr_is("FOGC")) {              /* set fog color                       */
      rgb[0]=read_int()/255.0; rgb[1]=read_int()/255.0; rgb[2]=read_int()/255.0;
      WnpSetFogColor(rgb);

    }else if (hdr_is("BGRD")) {              /* set background color                */
      rgb[0]=read_int()/255.0; rgb[1]=read_int()/255.0; rgb[2]=read_int()/255.0;
      WnpSetBackgroundColor(rgb);
      WnpSetGlBoxRedrawFlag(TRUE);

    }else if (hdr_is("SAAL")) {              /* set antialiasing mode               */
      ViewportSetModeFlags("ThreeD",WNP_MODE_SMOOTH,read_int());

    }else if (hdr_is("SDPC")) {              /* set depthcue mode                   */
      ViewportSetModeFlags("ThreeD",WNP_MODE_FOG,read_int());

    }else if (hdr_is("SDTH")) {              /* set dither mode                     */
      ViewportSetModeFlags("ThreeD",WNP_MODE_DITHER,read_int());

    }else if (hdr_is("SLTM")) {	             /* set 3D text mode                    */
      WnpTextQuality(read_int());

    }else if (hdr_is("LOCK")) {              /* lock event-Q                        */
      read_text(buf,WNP_MAXREC,size);
      if (!strcmp(buf,"EVTQ=ON "))   WnpLockEvtQ(1);
      else if (!strcmp(buf,"EVTQ=OFF")) WnpLockEvtQ(-1);
      else WnpSendToCom("bad LOCK message: %s",buf);

    }else if (hdr_is("ROCK")) {              /* rock around the clock               */
      read_floats(fbuf,2,2);
      WnpSetRocking(fbuf);

    }else if (hdr_is("SCDP")) {              /* take a screen dump                  */
      read_text(buf,WNP_MAXREC,size);
      WnpGlBoxRedraw();
      WnpScreenDump(buf);

    }else if (hdr_is("PLOT")) {              /* plot                                */
      WnpPlotThreeD();

    }else if (hdr_is("SYNC")) {              /* syncronize comms                    */
      send_header(chan,"SYNC",0); 

    }else if (hdr_is("KBEC")) {              /* keyboard echo status (1=0n, 0=off)  */
      WnpKeyboardEcho(read_int());

    }else if (hdr_is("EXIT")) {              /* E X I T                             */
      WnpExit();

    }else {
      WnpSendToCom("ERROR: unknown message (%4s)",hdr);
      discard_message(chan);
    }
    n=discard_message(chan);
    if (n>0) WnpSendToCom("WARNING: %d chars ignored, (%4s, %d)",n,hdr,size); 
  }
  trflush(chan);
  return(busy);
}

void WnpSendTextToClient(char *hdr,int n,char *text)
{
  send_bmessage(chan,hdr,n,text);
  trflush(chan);
}

void WnpReportUSAPick(int n,WnpAtom **atoms)
{
  int i;

  if (n>0) {
    for (i=0;i<n;i++) write_ptr("USAP",(void *)atoms[i]);
    WnpSetPickMode(WNP_PICK_USA_OFF);
    trflush(chan);
  }
}

void WnpReportAtomPick(WnpAtom *atom)
{
  write_ptr("PATM",(void *)atom);
  WnpSetAtomMark(atom,TRUE);
  WnpSetPickMode(WNP_PICK_OFF);
  trflush(chan);
}

void WnpReportObjectPick(DspObj *object)
{
  write_ptr("POBJ",(void *)object);
  WnpSetPickMode(WNP_PICK_OFF);
  trflush(chan);
}

void WnpReportPickLine(float *wl)
{
  write_floats("PXYZ",6,wl);
  WnpSetPickMode(WNP_PICK_OFF);
  trflush(chan);
}

/*------------------------------------------------------------------------*/

static void *read_ptr(void)
{
  void *ptr;

  if (read_data(chan,sizeof(void *),(char *)(&ptr))<sizeof(void *)) ptr=(void *)NULL;
  return(ptr);
}

static cptr_t read_cptr(void)
{
  cptr_t ptr;

  memset(&ptr,0,8);
  if (read_data(chan,ptrsize,(char *)(&ptr))<ptrsize) memset(&ptr,0,8);
  return(ptr);
}

static void **read_ptra(int n)
{
  void **r;
  int i;

  if (n>0) r=(void **)malloc(n*sizeof(void *));
  else r=(void **)NULL;
  if (r) for (i=0;i<n;i++) r[i]=read_ptr();
  return(r);
}

static cptr_t *read_cptra(int n)
{
  cptr_t *r;
  int i;

  if (n>0) r=(cptr_t *)malloc(n*sizeof(cptr_t));
  else r=(cptr_t *)NULL;
  if (r) for (i=0;i<n;i++) r[i]=read_cptr();
  return(r);
}

static int read_int(void)
{
  int n;

  if (read_data(chan,sizeof(int),(char *)(&n))<sizeof(int)) n=0;
  return(MY_i(n));
}

static float read_float(void)
{
  int n;

  if (read_data(chan,sizeof(float),(char *)(&n))<sizeof(float)) n=0.0;
  return(MY_f(n));
}

static int *read_inta(int n)
{
  int *r;

  if (n>0) {
    r=(int *)malloc(n*sizeof(int));
    if (r) read_ints(r,n,n);
  }else r=(int *)NULL;
  return(r);
}

static void read_ints(int *ibuf,int m,int n)
{
  int i;

  if (n<0) n=0;
  if (m && n>m ) n=m;
  for (i=0;i<n;i++) ibuf[i]=read_int();
}

static float *read_floata(int n)
{
  float *r;

  if (n>0) {
    r=(float *)malloc(n*sizeof(float));
    if (r) read_floats(r,n,n);
  }else r=(float *)NULL;
  return(r);
}

static void read_floats(float *fbuf,int m,int n)
{
  int i;

  if (n<0) n=0;
  if (m && n>m ) n=m;
  for (i=0;i<n;i++) fbuf[i]=read_float();
}

static void read_text(char *buf,int m,int n)
{
  if (n<0) n=0;
  if (m && n>m ) n=m;
  read_data(chan,n,buf);
  buf[n]='\0';
}

static char *read_texta(int n)
{
  char *r;

  if (n>=0) r=malloc((n+1)*sizeof(char));
  if (r) read_text(r,n,n);
  return(r);
}

static WnpPrj read_prj(void)
{
  WnpPrj prj;

  prj.viewmode=read_int();
  prj.stereomode=read_int();
  prj.ysize=read_float();
  prj.znear=read_float();
  prj.slabwidth=read_float();
  prj.viewdist=read_float();
  prj.eyesep=read_float();
  prj.aspect=read_float();
  prj.scrysize=read_float();
  prj.sep=read_float();

  return(prj);
}

static WnpAtom *read_atom(void)
{
  cptr_t cross;
  WnpAtom *a;

  cross=read_cptr();
  if ((a=read_atomd())) a->cross=cross;
  return(a);
}

static WnpAtom *read_atomd(void)
{
  WnpAtom *a;
  unsigned char trail[4]; /* color, dsp, padding, padding */

  /* read atom attributes... */

  a=WnpMakeAtom();
  if (a) {
    read_floats(a->x,3,3);
    a->r=read_float();
    read_data(chan,4,(char *)trail);
    a->color=trail[0];
    if (trail[1]) {
      a->flags|=ATM_FLG_DSP;
      a->flags|=ATM_FLG_LDSP;
    }else{
      a->flags&=~ATM_FLG_DSP;
      a->flags&=~ATM_FLG_LDSP;
    }
  }
  return(a);
}

static WnpBond *read_bond(void)
{
  WnpBond *b;
  WnpAtom *f,*t;

  f=read_ptr(); t=read_ptr();
  if (f&&t) {
    b=WnpMakeBond();
    if (b) {
      b->f=f;
      b->t=t;
    }
  }else b=(WnpBond *)NULL;
  return(b);
}

static void read_matrix(WnpMatrix m)
{
  int i,j;

  for(i=0;i<4;i++) for(j=0;j<4;j++) m[i][j]=read_float();
}

static WnpMDS *read_mdsa(void)
{
  int n,col;
  float *x;
  WnpMDS *mds;

  col=read_int();
  n=remaining_size(chan)/(6*sizeof(float));
  x=read_floata(6*n);
  if (x) {
    mds=WnpMakeMDS();
    if (mds) {
      mds->n=2*n;
      mds->color=col;
      mds->x=x;
    }else free(x);
  }else mds=(WnpMDS *)NULL;
  return(mds);
}

static WnpTOBJ *read_tobja(void)
{
  int n,col;
  char *t;
  float x[3];
  WnpTOBJ *tob;

  col=read_int();
  read_floats(x,3,3);
  n=remaining_size(chan);
  t=read_texta(n);
  if (t) {
    tob=WnpMakeTOBJ();
    if (tob) {
      tob->color=col;
      tob->text=t;
      tob->x[0]=x[0];
      tob->x[1]=x[1];
      tob->x[2]=x[2];
    }else free(t);
  }else tob=(WnpTOBJ *)NULL;
  return(tob);
}

static WnpTRG *read_trgvna(void)
{
  WnpTRG *trg;
  float *vn;
  int n;

  n=remaining_size(chan)/(6*sizeof(float));
  vn=read_floata(6*n);
  if (vn) {
    trg=WnpMakeTRG();
    if (trg) {
      trg->nv=n;
      trg->v=vn;
    }else free(vn);
  }else trg=(WnpTRG *)NULL;
  return(trg);
}

static void read_trg(TRGObj *obj)
{
  WnpTRG *trg;
  int i,n,*t;

  if (obj&&obj->type==DSP_TRG) {
    trg=(WnpTRG *)obj->fst;         /* always goes into first mesh */
    if (trg&&trg->type==WNP_LST_TRG) {
      if (trg->nt==0&&trg->nv>0) {  /* vertices must be defined */
	n=remaining_size(chan)/(3*sizeof(int));
	t=read_inta(3*n);
	if (t) {                    /* indices must be valid */
	  for (i=0;i<3*n;i++) if (t[i]<0||t[i]>trg->nv-1) break;
	  if (i<3*n) {
	    free(t);
	    t=(int *)NULL;
	  }
	}
	if (t) {
	  for (i=0;i<3*n;i++) t[i]*=6;  /* convert from vertex# to offset */
	  trg->nt=n;
	  trg->t=t;
	}
      }
    }
  }
}

static WnpNOE *read_noea(void)
{
  WnpNOE *n;

  if (remaining_size(chan)<=0) return((WnpNOE *)NULL);
  if (remaining_size(chan)>=(2*sizeof(void *)+2*sizeof(float)+sizeof(int))) {
    n=WnpMakeNOE();
    if (n) {
      n->f=read_ptr(); n->t=read_ptr();
      n->D=read_float(); n->color=read_int();
      n->viol=read_float();         /* new since server version 2.01 */
    }
  }else n=(WnpNOE *)NULL;
  return(n);
}

static void write_floats(char *hdr,int n,float *val)
{
  send_bheader(chan,hdr,n*sizeof(float));
  send_floata(chan,n,val);
}

static void write_ptr(char *hdr,void *ptr)
{
  send_bmessage(chan,hdr,sizeof(void *),(char *)&ptr);
}

static void write_cptr(cptr_t p)
{
  send_bdata(chan,ptrsize,(char *)&p);     /* this will use the cross pointer size in V2.1 */
}

static void write_prj(WnpPrj prj)
{
  send_bheader(chan,"GPRJ",2*sizeof(int)+8*sizeof(float));
  send_bint(chan,prj.viewmode);
  send_bint(chan,prj.stereomode);
  send_bfloat(chan,prj.ysize);
  send_bfloat(chan,prj.znear);
  send_bfloat(chan,prj.slabwidth);
  send_bfloat(chan,prj.viewdist);
  send_bfloat(chan,prj.eyesep);
  send_bfloat(chan,prj.aspect);
  send_bfloat(chan,prj.scrysize);
  send_bfloat(chan,prj.sep);
}

static void write_float(float x)
{
  int y;
  y=SGI_f(x);
  send_bdata(chan,sizeof(int),(char *)&y);
}

static void write_floata(int n,float *x)
{
  int i,y;

  for (i=0;i<n;i++) {
    y=SGI_f(x[i]);
    send_bdata(chan,sizeof(int),(char *)&y);
  }
}

static void write_matrix(WnpMatrix m)
{
  int i,j;

  for (i=0;i<4;i++) for (j=0;j<4;j++) write_float(m[i][j]);
}

static void sendObject(DspObj *obj)
{
  TRGObj *T;
  CSObj *C;
  WnpLst *l;
  WnpTRG *trg;
  WnpMDS *mds;
  WnpTOBJ *tob;
  int i,N,nv,nt,size;

  if (obj) {
    switch (obj->type) {
    case DSP_TRG:
      T=(TRGObj *)obj;
      nv=0;nt=0;
      for (l=T->fst;l;l=l->nxt) {             /* just counting... */
	if (l->type==WNP_LST_TRG) {
	  trg=(WnpTRG *)l;
	  if (trg->v&&trg->t&&trg->nt>0) {
	    nv+=trg->nv;
	    nt+=trg->nt;
	  }
	}
      }
      size=9*sizeof(int)+6*nv*sizeof(int)+3*nt*sizeof(int);
      send_bheader(chan,"GGOB",size);
      send_bint(chan,T->type);        /* 9 int */
      send_bint(chan,T->flags);
      send_bint(chan,T->color);
      send_bint(chan,T->lw);
      send_bint(chan,T->style);
      send_bint(chan,T->pat);
      send_bint(chan,T->bckclr);
      send_bint(chan,nv);
      send_bint(chan,nt);
      for (l=T->fst;l;l=l->nxt) {             /* write vertices with normals... */
	if (l->type==WNP_LST_TRG) {
	  trg=(WnpTRG *)l;
	  if (trg->v&&trg->t&&trg->nt>0) write_floata(6*trg->nv,trg->v);
	}
      }
      for (l=T->fst;l;l=l->nxt) {             /* write triangles... */
	if (l->type==WNP_LST_TRG) {
	  trg=(WnpTRG *)l; N=0;
	  if (trg->v&&trg->t&&trg->nt>0) {
	    for(i=0;i<3*nt;i++) send_bint(chan,N+trg->t[i]/6); /* /6: convert offset to vertex # */
	    N+=trg->nv;
	  }
	}
      }
      break;
    case DSP_CSO:
      C=(CSObj *)obj;
      size=4*sizeof(int);
      send_bheader(chan,"GGOB",size);
      send_bint(chan,C->type);        /* 4 int */
      send_bint(chan,C->flags);
      send_bint(chan,C->lw);
      send_bint(chan,C->mds_type);
      for (l=C->fst;l;l=l->nxt) {
	switch (l->type) {
	case WNP_LST_MDS:
	  mds=(WnpMDS *)l;
	  if (mds->n>0&&mds->x) {
	    size=(2+3*mds->n)*sizeof(int);
	    send_bheader(chan,"DATA",size);
	    send_bint(chan,l->type);
	    send_bint(chan,mds->color);
	    write_floata(3*mds->n,mds->x);
	  }
	  break;
	case WNP_LST_TOBJ:
	  tob=(WnpTOBJ *)l;
	  if (tob->text&&strlen(tob->text)>0) {
	    size=5*sizeof(int)+strlen(tob->text)+1;
	    send_bheader(chan,"DATA",size);
	    send_bint(chan,l->type);
	    send_bint(chan,tob->color);
	    write_floata(3,tob->x);
	    send_bdata(chan,strlen(tob->text)+1,tob->text);   /* including '\0' at end of string! */
	  }
	}
      }
      send_bmessage(chan,"DATA",0,NULL);   /* end of data blocks */
      break;
    default:
      send_bmessage(chan,"GGOB",0,NULL);
      break;
    }
  }
}

static void sendAtomDots(int n,WnpAtom **A)
{
  GlblObj *r;
  WnpAtom *a;
  WnpDots *d;
  int i;

  if (A) {
    for (i=0;i<n;i++) {
      if (A[i]) {
	d=WnpFindDots(a);
	if (d) send1AtomDots(d);
      }
    }
  }else{
    r=WnpDisplayRoot();
    ForAtoms(a,r) {
      d=WnpFindDots(a);
      if (d) send1AtomDots(d);
    }
  }
  send_bheader(chan,"GDOT",0);  /* end of transmission */
}

static void send1AtomDots(WnpDots *d)
{
  int size;

  if (d&&d->n>0&&d->xyz) {
    size=ptrsize+4*sizeof(int)+4*d->n*sizeof(int);
    send_bheader(chan,"GDOT",size);
    write_cptr(d->up->cross);
    send_bint(chan,d->n);
    send_bint(chan,d->color);
    send_bint(chan,d->dsp);
    send_bint(chan,d->style);
    write_floata(4*d->n,d->xyz);
  }
}

static void sendMonitorPairs(WnpMonPair *m)
{
  int n;
  WnpMonPair *p;

  if (m) for (n=0,p=m;p;p=(WnpMonPair *)p->nxt) n++;
  else n=0;

  send_bheader(chan,"GMON",4*sizeof(int)+2*n*ptrsize);
  send_bint(chan,WnpDefaultMonDsp());
  send_bint(chan,WnpDefaultMonColor());
  send_bint(chan,WnpDefaultMonLabel());
  send_bint(chan,WnpDefaultMonWidth());

  if (n>0) {
    for (p=m;p;p=(WnpMonPair *)p->nxt) {
      write_cptr(p->f->cross);
      write_cptr(p->t->cross);
    }
  }
}
