#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "wnp.h"
#include "wnpgl.h"

typedef struct WnpFont {
  char *name;                    /* font name (with wild cards)                                */
  GLubyte **map;                 /* per character bitmaps                                      */
  int *w,*h,*xo,*yo,*xm,*ym;     /* character metric                                           */
  GLuint base;                   /* display list id of GL bit-map font                         */
  int first,last,asc,desc,width; /* first/last character defined in font, ascent,descent,width */
  int class;                     /* font class (e.g. WNP_FNT_COURIER,...                       */
} WnpFont;

static WnpFont WnpReadFont(char *name,int class);
static WnpFont WnpScaleFont(WnpFont *Font,float sx,float sy,char *name,int class);

#define initFont(f) \
  f.name=(char *)NULL; \
  f.class=0; \
  f.map=(GLubyte **)NULL; \
  f.w=(int *)NULL; \
  f.h=(int *)NULL; \
  f.xo=(int *)NULL; \
  f.yo=(int *)NULL; \
  f.xm=(int *)NULL; \
  f.ym=(int *)NULL; \
  f.base=0; \
  f.first=0; \
  f.last=0; \
  f.asc=0; \
  f.desc=0; \
  f.width=0

int     nFonts=0;
int     activeFont=-1;
WnpFont FontTab[WNP_MAX_FONTS];

void WnpLoadFonts(void)
{
  int i,n;

  n=0;
  
  FontTab[n]=WnpReadFont("hard",WNP_FNT_HARD);        if (FontTab[n].map) n++;

  FontTab[n]=WnpReadFont("courier8",WNP_FNT_COURIER) ;if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("courier10",WNP_FNT_COURIER);if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("courier12",WNP_FNT_COURIER);if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("courier14",WNP_FNT_COURIER);if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("courier18",WNP_FNT_COURIER);if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("courier24",WNP_FNT_COURIER);if (FontTab[n].map) n++;
  if (n>0&&!strcmp(FontTab[n-1].name,"courier24")) {
    i=n-1;
    FontTab[n]=WnpScaleFont(FontTab+i,0.125,0.125,"3",WNP_FNT_COURIER); if (FontTab[n].map) n++;
    FontTab[n]=WnpScaleFont(FontTab+i,0.16667,0.16667,"4",WNP_FNT_COURIER); if (FontTab[n].map) n++;
    FontTab[n]=WnpScaleFont(FontTab+i,0.25,0.25,"6",WNP_FNT_COURIER); if (FontTab[n].map) n++;
    FontTab[n]=WnpScaleFont(FontTab+i-4,2.0,1.0,"stereo",WNP_FNT_STEREO); if (FontTab[n].map) n++;
  }

  FontTab[n]=WnpReadFont("helv8",WNP_FNT_HELV);  if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv10",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv12",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv14",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv16",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv20",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv25",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv30",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv40",WNP_FNT_HELV); if (FontTab[n].map) n++;
  FontTab[n]=WnpReadFont("helv60",WNP_FNT_HELV); if (FontTab[n].map) n++;

  nFonts=n;
  
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
#ifdef FONTSINLIST
  for (i=0;i<nFonts;i++) {
    FontTab[i].base = glGenLists((GLuint) (FontTab[i].last+1));
    if (FontTab[i].base==0) fprintf(stderr,"%s : out of display lists\n",FontTab[i].name);
    else{
      for(n=0;n<FontTab[i].last-FontTab[i].first+1;n++) {
	glNewList(FontTab[i].base+FontTab[i].first+n,GL_COMPILE);
	glBitmap(
	  FontTab[i].w[n],FontTab[i].h[n],
	  FontTab[i].xo[n],FontTab[i].yo[n],
	  FontTab[i].xm[n],FontTab[i].ym[n],
	  FontTab[i].map[n]);
	glEndList();
      }
    }
  }
#endif
}

/*-------------------- WnpReadFont --------------------*/

static WnpFont WnpReadFont(char *name,int class)
{
  char fn[512],rec[1024],buf[1024],*fp;
  WnpFont font;
  FILE *ff;
  int b;
  int i,j,n,delmap,nb,w,h,xm,ym,xo,yo,c;

  initFont(font);
  font.name=strdup(name);
  font.class=class;
  delmap=FALSE;

  fp=getenv("WNP_FONTPATH");
  if (!fp) fp=".";
  sprintf(fn,"%s/%s.fnt",fp,name);
  ff=fopen(fn,"r");
  if (ff) {
    if (fgets(rec,1024,ff)) sscanf(rec,"%s %d %d %d %d %s",buf,&font.asc,&font.desc,&font.first,&font.last,buf+10);
    else buf[0]='\0';
    if (!strcmp(buf,"FONT")) {
      font.map=(GLubyte **)malloc((font.last-font.first+1)*sizeof(GLubyte *));
      font.w= (int *)malloc((font.last-font.first+1)*sizeof(int));
      font.h= (int *)malloc((font.last-font.first+1)*sizeof(int));
      font.xo=(int *)malloc((font.last-font.first+1)*sizeof(int));
      font.yo=(int *)malloc((font.last-font.first+1)*sizeof(int));
      font.xm=(int *)malloc((font.last-font.first+1)*sizeof(int));
      font.ym=(int *)malloc((font.last-font.first+1)*sizeof(int));
      if (font.map&&font.w&&font.h&&font.xo&&font.yo&&font.xm&&font.ym) {
	for (j=0;j<=font.last-font.first;j++) font.map[j]=(GLubyte *)NULL;
	for (j=0;j<=font.last-font.first;j++) {
	  n=fscanf(ff,"%s %d %d %d %d %d %d %d",buf,&c,&w,&h,&xo,&yo,&xm,&ym);
	  if (n==8&&(!strcmp(buf,"CHAR")||c!=font.first+j)) {
	    font.w[j]=w; font.h[j]=h;
	    font.xo[j]=xo; font.yo[j]=yo;
	    font.xm[j]=xm; font.ym[j]=ym;
	    nb=(w+7)/8; nb*=h;
	    if (nb>0) font.map[j]=(GLubyte *)malloc(nb*sizeof(GLubyte));
	    if (font.map[j]) {
	      for (i=0;i<nb;i++) {
		if (fscanf(ff,"%x",&b)==1) font.map[j][i]=b;
		else font.map[j][i]=0;
	      }
	    }else{
	      fprintf(stderr,"memory allocation failed (%d), font:  %s\n",j+font.first,fn);
	      delmap=TRUE;
	    }
	  }else{
	    fprintf(stderr,"bad character (%d), font:  %s\n",j+font.first,fn);
	    delmap=TRUE;
	    break;
	  }
	}
      }else{
	fprintf(stderr,"memory allocation failed, font:  %s\n",fn);
	delmap=TRUE;
      }
    }else{
      fprintf(stderr,"bad font:  %s\n",fn);
      delmap=TRUE;
    }
  }else{
    fprintf(stderr,"could not load font %s\n",fn);
    delmap=TRUE;
  }

  if (delmap) {
    if (font.map) {
      for(j=0;j<=font.last-font.first;j++) if (font.map[j]) free(font.map[j]);
      free(font.map);font.map=(GLubyte **)NULL;
    }    
    if (font.w ) free(font.map);font.w =(int *)NULL;
    if (font.h ) free(font.map);font.h =(int *)NULL;
    if (font.xo) free(font.map);font.xo=(int *)NULL;
    if (font.yo) free(font.map);font.yo=(int *)NULL;
    if (font.xm) free(font.map);font.xm=(int *)NULL;
    if (font.ym) free(font.map);font.ym=(int *)NULL;
  }else{
    font.width=font.xm[0];
    for (j=1;j<=font.last-font.first;j++) if (font.width!=font.xm[j]) font.width=0;
  }

  return(font);
}

/*-------------------- WnpScaleFont --------------------*/

#define Byte(u,i,j) (u[j*Nb+(i/8)])
#define byte(u,i,j) (u[j*nb+(i/8)])
#define bit(l)      (1<<(7-(l%8)))

static WnpFont WnpScaleFont(WnpFont *Font,float sx,float sy,char *name,int class)
{
  WnpFont F,f;
  int n,delmap,j,k,l,K,L,nb,Nb;
  float dx,dy;
  
  F=*Font;
  initFont(f);
  f.name=strdup(name);
  f.class=class;
  delmap=FALSE;

  n=F.last-F.first+1;
  if (n>0) {
    f.first=F.first; f.last=F.last;
    f.asc=clamp(sy*F.asc,1,sy*F.asc); 
    f.desc=clamp(sy*F.desc,1,sy*F.desc);
    f.map=(GLubyte **)malloc(n*sizeof(GLubyte *));
    f.w= (int *)malloc(n*sizeof(int));
    f.h= (int *)malloc(n*sizeof(int));
    f.xo=(int *)malloc(n*sizeof(int));
    f.yo=(int *)malloc(n*sizeof(int));
    f.xm=(int *)malloc(n*sizeof(int));
    f.ym=(int *)malloc(n*sizeof(int));
    f.width=rint(sx*F.width); f.asc=rint(sy*F.asc); f.desc=rint(sy*F.desc);
    if (f.map&&f.w&&f.h&&f.xo&&f.yo&&f.xm&&f.ym) {
      for (j=0;j<=f.last-f.first;j++) f.map[j]=(GLubyte *)NULL;
      for (j=0;j<=f.last-f.first;j++) {
	f.w[j]=clamp(rint(sx*F.w[j]),1,rint(sx*F.w[j])+0.4999);
	f.h[j]=clamp(rint(sy*F.h[j]),1,rint(sy*F.h[j])+0.4999);
	f.xm[j]=rint(sx*F.xm[j]); if (F.xm[j]>0&&f.xm[j]==0) f.xm[j]=1;
	f.ym[j]=rint(sy*F.ym[j]); if (F.ym[j]>0&&f.ym[j]==0) f.ym[j]=1;
	f.xo[j]=rint(sx*F.xo[j]); f.yo[j]=rint(sy*F.yo[j]);
	nb=(f.w[j]+7)/8; Nb=(F.w[j]+7)/8;
	dx=((float)f.w[j])/F.w[j]; dy=((float)f.h[j])/F.h[j];
	f.map[j]=(GLubyte *)malloc(nb*f.h[j]*sizeof(GLubyte));
	for (k=0;k<nb*f.h[j];k++) f.map[j][k]=0;
	if (F.h[j]>f.h[j]) {
	  if (F.w[j]>f.w[j]) {
	    for (K=0;K<F.h[j];K++) {
	      k=dy*K;
	      for (L=0;L<F.w[j];L++) {
		l=dx*L;
		if (Byte(F.map[j],L,K)&bit(L)) byte(f.map[j],l,k)|=bit(l);
	      }
	    }
	  }else{
	    for (K=0;K<F.h[j];K++) {
	      k=dy*K;
	      for (l=0;l<f.w[j];l++) {
		L=l/dx; L=clamp(L,0,F.w[j]-1);
		if (Byte(F.map[j],L,K)&bit(L)) byte(f.map[j],l,k)|=bit(l);
	      }
	    }
	  }
	}else{
	  if (F.w[j]>f.w[j]) {
	    for (k=0;k<f.h[j];k++) {
	      K=k/dy;  K=clamp(K,0,F.h[j]-1);
	      for (L=0;L<F.w[j];L++) {
		l=dx*L;
		if (Byte(F.map[j],L,K)&bit(L)) byte(f.map[j],l,k)|=bit(l);
	      }
	    }
	  }else{
	    for (k=0;k<f.h[j];k++) {
	      K=k/dy; K=clamp(K,0,F.h[j]-1);
	      for (l=0;l<f.w[j];l++) {
		L=l/dx; L=clamp(L,0,F.w[j]-1);
		if (Byte(F.map[j],L,K)&bit(L)) byte(f.map[j],l,k)|=bit(l);
	      }
	    }
	  }
	}
      }
    }else delmap=TRUE;
  }  
  if (delmap) {
    if (f.map) {
      for(j=0;j<=f.last-f.first;j++) if (f.map[j]) free(f.map[j]);
      free(f.map);f.map=(GLubyte **)NULL;
    }    
    if (f.w ) free(f.map);f.w =(int *)NULL;
    if (f.h ) free(f.map);f.h =(int *)NULL;
    if (f.xo) free(f.map);f.xo=(int *)NULL;
    if (f.yo) free(f.map);f.yo=(int *)NULL;
    if (f.xm) free(f.map);f.xm=(int *)NULL;
    if (f.ym) free(f.map);f.ym=(int *)NULL;
  }else{
    f.width=f.xm[0];
    for (j=1;j<=f.last-f.first;j++) if (f.width!=f.xm[j]) f.width=0;
  }
  return(f);
}

/*-------------------- WnpSelFont --------------------*/

int WnpSelFont(int class,int height)
{
  /* find font in given class with appropriate height... */

  int r,i,h;

  for (r=-1,h=-1,i=0;i<nFonts;i++) {
    if (FontTab[i].class&class) {
      if (height>=FontTab[i].desc+FontTab[i].asc) {
	if (FontTab[i].desc+FontTab[i].asc>h||h>height) {
	  h=FontTab[i].desc+FontTab[i].asc;
	  r=i;
	}
      }else{
	if (h>height) {
	  if (FontTab[i].desc+FontTab[i].asc<h) {
	    h=FontTab[i].desc+FontTab[i].asc;
	    r=i;
	  }
	}else if (h<0) {
	  h=FontTab[i].desc+FontTab[i].asc;
	  r=i;
	}
      }
    }
  }
  return(r);
}
/*-------------------- WnpSetFont... etc. --------------------*/

int WnpSetFont(int font)
{
  int r;

  r=activeFont;
  if (font>=0&&font<nFonts) activeFont=font;
  return(r);
}

int WnpActiveFontNr(void)
{
  return(activeFont);
}

/*-------------------- WnpTextWidth --------------------*/

int WnpTextWidth(char *text,int nch)
{
  /* return text width (pixels) of first nch characters of text in active font */

  int w,i,c;
  char *p;

  if (activeFont>=0) {
    nch=(nch > strlen(text) ? strlen(text) : (nch < 0 ? strlen(text) : nch));
    if (FontTab[activeFont].width>0) {
      w=nch*FontTab[activeFont].width;
    }else{
      for(w=0,p=text,i=0;i<nch;i++,p++) {
	c=*p-FontTab[activeFont].first;
	if (c>=0&&c<=FontTab[activeFont].last-FontTab[activeFont].first) {
	  w+=FontTab[activeFont].xm[c];
	}
      }
    }
  }else w=0;
  return(w);
}
/*-------------------- WnpTextPos3f --------------------*/

void WnpTextPos3f(float x,float y,float z)
{
  glRasterPos3f(x,y,z);
}
/*-------------------- WnpTextDraw --------------------*/

void WnpTextDraw(char *text)
{
  char *p;
  int k;
  float rcol[4];

  if (activeFont>=0&&text) {
    glGetFloatv(GL_CURRENT_RASTER_COLOR,rcol);/* for pre-Impact graphics under 6.2, sad ! */
    if (FontTab[activeFont].base>0) {
      glListBase(FontTab[activeFont].base);
      glCallLists(strlen(text), GL_UNSIGNED_BYTE, (GLubyte *)text);
    }else{
      for (p=text;*p;p++) {
	if (FontTab[activeFont].first<=*p&&*p<=FontTab[activeFont].last) {
	  k=*p-FontTab[activeFont].first;
	  glBitmap(FontTab[activeFont].w[k],FontTab[activeFont].h[k],
		   FontTab[activeFont].xo[k],FontTab[activeFont].yo[k],
		   FontTab[activeFont].xm[k],FontTab[activeFont].ym[k],
		   FontTab[activeFont].map[k]);
	}
      }
    }
  }
}
