#include "wnp.h"
#include "wnpgl.h"
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#define nCEmodes 3

static const char *GLVendor="";              /* oGL Vendor string                          */
static const char *GLRenderer="";            /* oGL Renderor                               */
static const char *GLVersion="";             /* oGL Library Version                        */
static int  GLType=WNP_GL_UNKNOWN;
static int  TxtBuffers;                      /* Framebuffers to update in msg/menu ports   */
static int  DoubleBufferMode=TRUE;           /* is double buffer mode enabled ?            */
static int  allowDBswitching=FALSE;          /* is changing double buffer mode allowed ?   */
static int  wnpGlBoxMode = 0;

/* Crystal Eye stereo mode: 1 SGIs, 2 O2, 3 Linux PC 768x1024 */
/* Alt-S may be used to step through modes...                 */

static int CEmode = 1;     /* sgi O2  PC  */ /* Crystal Eyes mode (initial)                */
static int yleft[nCEmodes]=  {532,512,403};  /* YLEFT in mode k                            */
static int ystereo[nCEmodes]={491,511,363};  /* YSTEREO in mode k                          */
static int YLEFT;                            /* offset for left-eye stereo, mode 'crystal' */
static int YSTEREO;                          /* fs height in crystal stereo mode           */
static int vwp_orgx=0;                       /* viewport origin x                          */
static int vwp_orgy=0;                       /* viewport origin y                          */
static int vwp_width=0;                      /* viewport width                             */
static int vwp_height=0;                     /* viewort height                             */
static int redraw_flag=FALSE;                /* need to redraw the lot ?                   */
static int menu_and_comm=TRUE;               /* are menu and comm boxes open ?             */
static int threed=TRUE;                      /* is the 3D box open ?                       */
static int win_fsgeom[4];                    /* window position and size in FS mode        */
static int win_geom[4];                      /* window position and size                   */
static int win_full;                         /* full screen mode ?                         */
static float TDx2,TDy1;                      /* 3D/menu/comm division                      */
static int crystaleyes=FALSE;                /* true if monitor in cryst.eyes mode         */
static float lineWidth=-1;                   /* current linewidth                          */
static float lwmin=1.5;                      /* pixels in virtual lw=1                     */
static float lwmax=10.0;                     /* pixels in maximum lw                       */
static float lwstp=0.5;                      /* granularity                                */
static float maxSmoothWidth=4.0;             /* maximum width of smooth lines              */
static float lwr[2];                         /* allowd anti-aliased linewidth range...     */
static float lgr;                            /* ...and granularity.                        */

static int drawBuffer[16] = {
  GL_NONE,
  GL_FRONT,
  GL_BACK,
  GL_FRONT_AND_BACK,
  GL_LEFT,
  GL_FRONT_LEFT,
  GL_BACK_LEFT,
  GL_LEFT,
  GL_RIGHT,
  GL_FRONT_RIGHT,
  GL_BACK_RIGHT,
  GL_RIGHT,
  GL_FRONT_AND_BACK,
  GL_FRONT,
  GL_BACK,
  GL_FRONT_AND_BACK
};

static Viewport GlViewports = { "GlBox",(Viewport *)NULL, 0.0,1.0,0.0,1.0, TRUE };

/*-------------------- GlBox callback definitions --------------------*/

void WnpInitGlBox(int width,int height)
{
  char buf[WNP_MAXREC];
  
  GLboolean flag;
  GLint drwbuf;
  FILE *f;

  GLVendor=(char *)glGetString(GL_VENDOR);
  GLRenderer=(char *)glGetString(GL_RENDERER);
  GLVersion=(char *)glGetString(GL_VERSION);
  if (!strcmp(GLVendor,"SGI")) {
    if (!strncmp(GLRenderer,"NEWPORT",7)||
	!strncmp(GLRenderer,"GR1.",4)||
	!strncmp(GLRenderer,"VGX",3)||
	!strncmp(GLRenderer,"LIGHT",5)||
	!strncmp(GLRenderer,"GU1",3)||
	!strncmp(GLRenderer,"GR2",3)||
	!strncmp(GLRenderer,"GR3",3)||
	!strncmp(GLRenderer,"VG",2)||
        !strncmp(GLRenderer,"VPRO",4)||
	!strncmp(GLRenderer,"RE",2))              GLType=WNP_GL_OLDSGI;
    else if (!strncmp(GLRenderer,"IMPACT",6)||
	!strncmp(GLRenderer,"IR",2)||
	!strncmp(GLRenderer,"CRIME",5))              GLType=WNP_GL_NEWSGI;
    else if (!strncmp(GLRenderer,"Generic/MMX",11))  GLType=WNP_GL_MMXSGI;
    else GLType=WNP_GL_UNKNOWN;
  }else if (!strncmp(GLRenderer,"Mesa X11",8)) {
    GLType=WNP_GL_MESAX11;
  }else if (!strncmp(GLRenderer,"Mesa GLX Indirect",17)) {
    GLType=WNP_GL_MESAX11;
  }else if (!strncmp(GLRenderer,"SavageMX",8)) {
    GLType=WNP_GL_SAVAGEMX;
  }else if (strstr(GLVendor,"NVIDIA")&&strstr(GLRenderer,"Quadro")) {
    GLType=WNP_GL_NVIDIAQUADRO;
  }else if (!strncmp(GLRenderer,"GeForce",7)) {    /* just a guess, really... */
    GLType=WNP_GL_GEFORCE;
  }else if (!strncmp(GLRenderer,"GDI Generic",11))  {
    GLType=WNP_GL_GDIGEN;
  }else{
    sprintf(buf,"%s %s, version %s",GLVendor,GLRenderer,GLVersion);
    if (strstr(buf,"Radeon 7500 DDR x86/SSE2")) GLType=WNP_GL_RADEON7500;
    else if (strstr(buf,"Intel Brookdale-G")) GLType=WNP_GL_BROOKDALE_G;
    else if (strstr(buf,"Intel 945G")) GLType=WNP_GL_INTEL945G;
    else if (!strcmp("nouveau",GLVendor)&&!strcmp("Gallium 0.4 on NVD9",GLRenderer)) GLType=WNP_GL_NV_GALLIUM;
    else GLType=WNP_GL_UNKNOWN;
  }

  if (GLType==WNP_GL_NEWSGI) TxtBuffers=WNP_MODE_BANDF;
  else if (GLType==WNP_GL_NVIDIAQUADRO) TxtBuffers=WNP_MODE_FRONT;/* some drivers cant't handle BANDF */
  else if (GLType==WNP_GL_GEFORCE) TxtBuffers=WNP_MODE_BANDF;     /* hope that's right... */
  else if (GLType==WNP_GL_RADEON7500) TxtBuffers=WNP_MODE_FRONT;  /* faster on WJ's and HM's box */
  else if (GLType==WNP_GL_BROOKDALE_G) TxtBuffers=WNP_MODE_FRONT; /* needed, why ??? */
  else TxtBuffers=WNP_MODE_FRONT;

  if (!strcmp(GLVendor,"SGI")&&strncmp(GLRenderer,"Generic/MMX",11)) {
    lwmin=1.0;
    lwmax=10.0;
    lwstp=1.0;
    maxSmoothWidth=1.0;    /* most SGI's are slow to render wide anti-aliased lines */
  }

  WnpGlBoxViewport(0,0,width,height);
  WnpColorInit();
  glEnable(GL_SCISSOR_TEST);
  wnpGlBoxMode=0;
  glGetBooleanv(GL_FOG,&flag); if (flag) wnpGlBoxMode|=WNP_MODE_FOG;
  glGetBooleanv(GL_DEPTH_TEST,&flag); if (flag) wnpGlBoxMode|=WNP_MODE_ZBUFF;
  glGetBooleanv(GL_BLEND,&flag); if (flag) wnpGlBoxMode|=WNP_MODE_BLEND;
  glGetBooleanv(GL_LINE_SMOOTH,&flag); if (flag) wnpGlBoxMode|=WNP_MODE_SMOOTH;
  glGetBooleanv(GL_DITHER,&flag); if (flag) wnpGlBoxMode|=WNP_MODE_DITHER;
  glGetIntegerv(GL_DRAW_BUFFER,&drwbuf);
  if (drwbuf==GL_FRONT) wnpGlBoxMode|=WNP_MODE_FRONT;
  else if (drwbuf==GL_BACK) wnpGlBoxMode|=WNP_MODE_BACK;
  else if (drwbuf==GL_FRONT_AND_BACK) wnpGlBoxMode|=WNP_MODE_BACK|WNP_MODE_FRONT;

  glGetFloatv(GL_LINE_WIDTH_RANGE,lwr);
  glGetFloatv(GL_LINE_WIDTH_GRANULARITY,&lgr);

  TDx2=0.85; TDy1=6.0/WNP_MENU_LINES;

  win_full=FALSE;
  win_fsgeom[0]=0; win_fsgeom[1]=0;
  win_fsgeom[2]=WnpDisplayWidth();
  win_fsgeom[3]=WnpDisplayHeight();
  WnpGetGlBoxWinpos(win_geom);

  if(!strncmp(GLRenderer,"CRIME",5)) CEmode=1;
  else if (win_fsgeom[3]==768) CEmode=2;
  else CEmode=0;
  YLEFT=yleft[CEmode];
  YSTEREO=ystereo[CEmode];

  VwpCreate("ThreeD",
	    0,TDy1,TDx2,1,
	    WNP_MODE_BACK|WNP_MODE_FOG|WNP_MODE_ZBUFF|
	    WNP_MODE_BLEND|WNP_MODE_SMOOTH|WNP_MODE_DITHER);
  VwpCreate("Menu",TDx2,0,1,1,TxtBuffers);
  VwpCreate("Comm",0,0,TDx2,TDy1,TxtBuffers);
  VwpCreate("Cursor",0,0,1,1,WNP_MODE_FRONT|WNP_MODE_RDFRONT);

  WnpLoadFonts();
  WnpTextInit();
  WnpMessageBoxInit();
  WnpMenuBoxInit();
  WnpKeyboardInit();

  WnpMouseMotionMask(TRUE);

  sprintf(buf,"Graphics server running on ");
  gethostname(buf+strlen(buf),WNP_MAXREC-strlen(buf)-1);
  sprintf(buf+strlen(buf),", renderer is %s %s, version %s",GLVendor,GLRenderer,GLVersion);
  WnpSendToCom("%s",buf);
  if (GLType==WNP_GL_UNKNOWN) WnpSendToCom("Renderer is unknown, good luck...");
  else if (GLType==WNP_GL_NV_GALLIUM) {
    WnpSendToCom("Renderer is known to have problems in double buffer mode on some platforms.");
    WnpSendToCom("Use ALT-D to toggle between single / double buffer mode.");
    DoubleBufferMode=FALSE;
    allowDBswitching=TRUE;
  }
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  WnpSetGlBoxRedrawFlag(TRUE);

}


/*------------------------------------------------------------------------*/

void WnpGetGlBoxGeom(int geom[4])
{
  geom[0]=vwp_orgx;
  geom[1]=vwp_orgy;
  geom[2]=vwp_width;
  geom[3]=vwp_height;
}

int WnpYStereo(void)
{
  return(YSTEREO);
}

int WnpGlBoxRedrawFlag(void)
{
  return(redraw_flag);
}

void WnpSetGlBoxRedrawFlag(int s)
{
  redraw_flag=s;
}

int WnpGetGlBoxMode(void)
{
  return(wnpGlBoxMode);
}

int WnpSetGlBoxMode(int mode)
{
  int r,d;

  r=wnpGlBoxMode;
  if (mode!=wnpGlBoxMode) {
    d=mode^wnpGlBoxMode;
    wnpGlBoxMode=mode;
    if (d&WNP_MODE_FOG) {
      if (mode&WNP_MODE_FOG) glEnable(GL_FOG);
      else glDisable(GL_FOG);
    }
    if (d&WNP_MODE_ZBUFF) {
      if (mode&WNP_MODE_ZBUFF) glEnable(GL_DEPTH_TEST);
      else glDisable(GL_DEPTH_TEST);
    }
    if (d&WNP_MODE_BLEND) {
      if (mode&WNP_MODE_BLEND) glEnable(GL_BLEND);
      else glDisable(GL_BLEND);
    }
    if (d&WNP_MODE_SMOOTH) {
      if (mode&WNP_MODE_SMOOTH) glEnable(GL_LINE_SMOOTH);
      else glDisable(GL_LINE_SMOOTH);
      lineWidth=-1; WnpVLineWidth(1);
    }
    if (d&WNP_MODE_BUFFER) {
      if (DoubleBufferMode) {
	glDrawBuffer(drawBuffer[mode&WNP_MODE_BUFFER]);
      }else{
	glDrawBuffer(GL_FRONT);
      }
    }
    if (d&WNP_MODE_RDFRONT) {
      if (mode&WNP_MODE_RDFRONT) glReadBuffer(GL_FRONT);
    }
    if (d&WNP_MODE_RDBACK) {
      if (mode&WNP_MODE_RDBACK) glReadBuffer(GL_BACK);
    }
    if (d&WNP_MODE_DITHER) {
      if (mode&WNP_MODE_DITHER) glEnable(GL_DITHER);
      else glDisable(GL_DITHER);
    }
  }
  return(r);
}

void WnpGlBoxViewport(int orgx,int orgy,int width,int height)
{
  redraw_flag=vwp_orgx!=orgx || vwp_orgy!=orgy;
  redraw_flag |= vwp_width!=width || vwp_height!=height;
  vwp_orgx=orgx; vwp_orgy=orgy;
  vwp_width=width; vwp_height=height;
}

int WnpGlBoxRedraw(void)
{
  int r;

  r=FALSE;
  if (redraw_flag) {
    r=TRUE;
    WnpSetThreeDRedrawFlag(TRUE);
    WnpDrawMenuBox(TxtBuffers);
    WnpDrawMessageBox(TxtBuffers);
  }
  WnpDrawPendingLines();
  if (WnpAxisRotation()) WnpSetThreeDRedrawFlag(TRUE);
  if (redraw_flag||(WnpThreeDRedrawFlag()&&!WnpEvtQLocked())) {
    r=TRUE;
    WnpDrawMessageBoxBB();
    WnpDrawMenuBoxBB();
    if (ViewportIsOpen("ThreeD")) {
      WnpDrawThreeDBox();
    }
    if (DoubleBufferMode) WnpSwapGlBuffers();
    WnpGlDrawCursor(TRUE);
    WnpSetThreeDRedrawFlag(/*redraw_flag*/ FALSE);
  }else{
    WnpGlDrawCursor(FALSE);
  }
  redraw_flag=FALSE;
  glFlush();

  /* the following avoids cursor flicker on inferior systems: */
  WnpSetGlBoxMode(WNP_MODE_BACK);

  return(r);
}

void WnpChange3DSize(int dx,int dy)
{
  Viewport *v;
  float nTDx2,nTDy1;

  if (dy<0&&!threed) {
    TDy1=1;
    WnpToggleThreeD();
  }

  nTDx2=TDx2+dx*0.015; if (nTDx2>1) nTDx2=1; if (nTDx2<0) nTDx2=0;
  nTDy1=TDy1+(1.0*dy)/WNP_MENU_LINES; if (nTDy1>1) nTDy1=1; if (nTDy1<0) nTDy1=0;
  if (nTDx2!=TDx2||nTDy1!=TDy1) {
    TDx2=nTDx2; TDy1=nTDy1;
    if (menu_and_comm) {
      WnpSetGlBoxRedrawFlag(TRUE);
      if (v=VwpGet("Menu"))    v->x1=TDx2;
      if (v=VwpGet("ThreeD")) {v->x2=TDx2; if (threed) v->y1=TDy1;}
      if (v=VwpGet("Comm"))   {v->x2=TDx2; if (threed) v->y2=TDy1;}
    }
  }
}

void WnpGlBoxToggleFullScreen(void)
{
  if (!crystaleyes) {
    if (win_full) {
      WnpGlBoxWinpos(win_geom);
      win_full=FALSE;
    }else{
      WnpGetGlBoxWinpos(win_geom);
      WnpGlBoxFullScreen(win_fsgeom);
      win_full=TRUE;
    }
  }
}

void WnpMenuCommState(int on)
{
  Viewport *v;

  if (menu_and_comm&&!on) {
    if (!threed) WnpToggleThreeD();
    if (v=VwpGet("ThreeD")) {v->x2=1; v->y1=0;}
    ViewportClose("Menu");
    ViewportClose("Comm");
    WnpSetGlBoxRedrawFlag(TRUE);
  }else if (!menu_and_comm&&on) {
    ViewportOpen("Menu");
    ViewportOpen("Comm");
    if (threed) {
      if (v=VwpGet("Menu"))    v->x1=TDx2;
      if (v=VwpGet("ThreeD")) {v->x2=TDx2; v->y1=TDy1;}
      if (v=VwpGet("Comm"))   {v->x2=TDx2; v->y2=TDy1;}
      WnpSetGlBoxRedrawFlag(TRUE);
    }else{
      if (v=VwpGet("Menu"))    v->x1=TDx2;
      if (v=VwpGet("Comm"))   {v->x2=TDx2; v->y2=1;}
    }
  }
  menu_and_comm=on;
}

void WnpToggleMenuComm(void)
{
  WnpMenuCommState(!menu_and_comm);
}

void WnpThreeDState(int on)
{
  Viewport *v;

  if (!threed&&on) {
    ViewportOpen("ThreeD");
    if (v=VwpGet("Menu"))    v->x1=TDx2;
    if (v=VwpGet("ThreeD")) {v->x2=TDx2; v->y1=TDy1;}
    if (v=VwpGet("Comm"))   {v->x2=TDx2; v->y2=TDy1;}
    WnpSetGlBoxRedrawFlag(TRUE);
  }else if (threed&&!on) {
    ViewportClose("ThreeD");
    if (v=VwpGet("ThreeD")) {v->x2=TDx2; v->y1=1;}
    if (v=VwpGet("Comm"))   {v->x2=TDx2; v->y2=1;}
    threed=FALSE;
    WnpMenuCommState(TRUE);
    WnpSetGlBoxRedrawFlag(TRUE);
  }
  threed=on;
}

void WnpToggleThreeD(void)
{
  WnpThreeDState(!threed);
}

Viewport *WnpScreen2FractVwpXY(int x,int y,float *xf,float *yf)
{
  float xp,yp,xo,yo,s;
  Viewport *p;

  s=(crystaleyes ? ((float )YSTEREO)/win_fsgeom[3] : 1.0);
  xo=vwp_orgx; yo=vwp_orgy;    /* origin in GL-Screen */
  xp=x; yp=vwp_height-1-y;
  if (crystaleyes) yp-=YLEFT;
  *xf=(xp-xo)/vwp_width; *yf=(yp-yo)/(s*(vwp_height-1));
  for (p=GlViewports.nxt;p;p=p->nxt) {
    if (p->open) {
      if (p->x1 <= *xf && *xf < p->x2 && p->y1 <= *yf && *yf <p ->y2) {
	*xf=(*xf-p->x1)/(p->x2-p->x1);
	*yf=(*yf-p->y1)/(p->y2-p->y1);
	break;
      }
    }
  }
  return(p);
}

void WnpFractVwpXY2Window(Viewport *v,float xf,float yf,int *x,int *y)
{
  float s;

  /* convert fractional viewport coordinates to Window coordinates (pixels,
     relative to (LL) corner of oGL widget */

  s=(crystaleyes ? ((float )YSTEREO)/win_fsgeom[3] : 1.0);
  *x=vwp_orgx+(vwp_width -1)*(v->x1+xf*(v->x2-v->x1))+0.1;
  *y=vwp_orgy+s*(vwp_height-1)*(v->y1+yf*(v->y2-v->y1))+0.1;
  if (crystaleyes) *y+=YLEFT;
}

/*-------------------- Viewport handling --------------------*/

Viewport *VwpCreate(char *name,float xl,float yl,float xh,float yh,int mode)
{
  Viewport *v,*p;

  v=(Viewport *)malloc(sizeof(Viewport));
  if (v) {
    v->mode=mode;
    v->name=strdup(name);
    v->x1=xl; v->x2=xh;
    v->y1=yl; v->y2=yh;
    v->open=TRUE; v->nxt=(Viewport *)NULL;
    for (p=&GlViewports;p->nxt;p=p->nxt);
    p->nxt=v;
  }
  return(v);
}

int WnpGetVwpDimensions(char *name,float *width,float *height)
{
  Viewport *v;
  int r;

  if (v=VwpGet(name)) {
    if (v->open&&v->y2>v->y1) {
      r=TRUE;
      *width=vwp_width*(v->x2-v->x1)+0.5;
      *height=vwp_height*(v->y2-v->y1)+0.5;
    }else{
      r=FALSE;
      *width=win_fsgeom[2];
      *height=win_fsgeom[3];
    }
  }else{
    r=FALSE;
    *width=win_fsgeom[2];
    *height=win_fsgeom[3];
  }
  return(r);
}

Viewport *VwpGet(char *name)
{
  Viewport *p;

  for (p=GlViewports.nxt;p;p=p->nxt) if (!strcmp(p->name,name)) break;
  return(p);
}

void ViewportOpen(char *name)
{
  Viewport *p;

  p=VwpGet(name);
  if (p) {
    if (!p->open) WnpSetGlBoxRedrawFlag(TRUE);
    p->open=TRUE;
  }
}

void ViewportClose(char *name)
{
  Viewport *p;

  p=VwpGet(name);
  if (p) {
    if (p->open) WnpSetGlBoxRedrawFlag(TRUE);
    p->open=FALSE;
  }
}

int ViewportIsOpen(char *name)
{
  Viewport *p;

  p=VwpGet(name);
  return(p ? p->open && p->x2>p->x1+0.0005 && p->y2>p->y1+0.0005  : FALSE);
}

void ViewportSet(char *name,int port)
{
  Viewport *v;
  int ox,oy,wx,wy;
  float s;

  v=VwpGet(name);
  if (v) {
    switch (port) {
    case WNP_VWP_FULL:
      ox=vwp_orgx+vwp_width*(v->x1)+0.5;
      oy=vwp_orgy+vwp_height*(v->y1)+0.5;
      wx=vwp_width*(v->x2-v->x1)+0.5;
      wy=vwp_height*(v->y2-v->y1)+0.5;
      break;
    case WNP_VWP_LEFT:
      ox=vwp_orgx+vwp_width*(v->x1)+0.5;
      oy=vwp_orgy+vwp_height*(v->y1)+0.5;
      wx=0.5*vwp_width*(v->x2-v->x1)+0.5;
      wy=vwp_height*(v->y2-v->y1)+0.5;
      break;
    case WNP_VWP_RIGHT:
      wx=0.5*vwp_width*(v->x2-v->x1)+0.5;
      wy=vwp_height*(v->y2-v->y1)+0.5;
      ox=vwp_orgx+wx+vwp_width*(v->x1)+0.5;
      oy=vwp_orgy+vwp_height*(v->y1)+0.5;
      break;
    case WNP_VWP_BOTTOM:
      s=(crystaleyes ? (YSTEREO+1.0)/win_fsgeom[3] : 1.0);
      ox=vwp_orgx+vwp_width*(v->x1)+0.5;
      oy=vwp_orgy+s*vwp_height*(v->y1)+0.5;
      wx=vwp_width*(v->x2-v->x1)+0.5;
      wy=s*vwp_height*(v->y2-v->y1)+0.5;
      break;
    case WNP_VWP_TOP:
      s=(crystaleyes ? (YSTEREO+1.0)/win_fsgeom[3] : 1.0);
      ox=vwp_orgx+vwp_width*(v->x1)+0.5;
      oy=vwp_orgy+s*vwp_height*(v->y1)+0.5+YLEFT;
      wx=vwp_width*(v->x2-v->x1)+0.5;
      wy=s*vwp_height*(v->y2-v->y1)+0.5;
      break;
    }
    glViewport(ox,oy,wx,wy);
    glScissor(ox,oy,wx,wy);
    if (v->mode!=wnpGlBoxMode) WnpSetGlBoxMode(v->mode);
  }
}

int ViewportSetMode(char *name,int mode)
{
  Viewport *v;
  int r;

  if (v=VwpGet(name)) {
    r=v->mode;
    v->mode=mode;
  }else r=0;
  return(r);
}

int ViewportSetModeFlags(char *name,int flags,int mode)
{
  Viewport *v;
  int r;

  if (v=VwpGet(name)) {
    r=v->mode;
    if (mode) v->mode|=flags;
    else v->mode&=~flags;
  }else r=0;
  return(r);
}

int ViewportLocate(char *name,int *xo,int *yo,int *width,int *height)
{
  int r;
  Viewport *v;
  int X[4];
  float s;

  /* Note: this gives coordinates on screen (not relative to glBox) */
  /* in crystaleyes mode, this locates the bottom (right) viewport. */

  if (v=VwpGet(name)) {
    r=TRUE;
    s=(crystaleyes ? (YSTEREO+1.0)/win_fsgeom[3] : 1.0);
    *xo=vwp_orgx+vwp_width*(v->x1)+0.5;
    *yo=vwp_orgy+s*vwp_height*(v->y1)+0.5;
    *width=vwp_width*(v->x2-v->x1)+0.5;
    *height=s*vwp_height*(v->y2-v->y1)+0.5;
    WnpGetGlBoxWinpos(X);
    *xo+=X[0]; *yo+=win_fsgeom[3]-X[1]-X[3];    /* convert to screen, X11 has y=0 at top */
  }else r=FALSE;
  return(r);
}

int WnpPixelsPerLine(void)
{
  float s;
  int r;

  s=(crystaleyes ? (YSTEREO+1.0)/win_fsgeom[3] : 1.0);
  r=s*((vwp_height+WNP_MENU_LINES-1)/WNP_MENU_LINES);
  return(r);
}
/*------------------------------------------------------------------------*/

void WnpCrystalEyes(int mode)
{
  if (mode&&!crystaleyes) {
    WnpSetMonitor(WNP_STEREO_CRYSTAL);
    if (!win_full) WnpGetGlBoxWinpos(win_geom);
    WnpGlBoxWinpos(win_fsgeom);
    WnpSetMouseTrap(0,0,win_fsgeom[2],YSTEREO+1);
  }else if (!mode&&crystaleyes) {
    WnpSetMonitor(WNP_STEREO_NO);
    if (!win_full) WnpGlBoxWinpos(win_geom);
    WnpRemoveMouseTrap();
  }
  crystaleyes=mode;
  WnpSetGlBoxCursor(WnpSetGlBoxCursor(WNP_CURSOR_NONE));
}

int WnpCrystalEyesMode(void)
{
  return(crystaleyes);
}

void WnpSetMonitor(int mode)
{
#if defined (sgi) || defined (linux) || defined (cygwin) || defined (mingw)
  char cmd[4096];
  int s;

  switch (mode) {
  case WNP_STEREO_CRYSTAL:
    sprintf(cmd,"wnp_stereo AB on '%s' '%s' '%s'",GLVendor,GLRenderer,GLVersion);
    s=system(cmd);
    if (s) WnpSendToCom("wnp_stereo failed, status=%d",s);
    break;
  case WNP_STEREO_NO:
    sprintf(cmd,"wnp_stereo AB off '%s' '%s' '%s'",GLVendor,GLRenderer,GLVersion);
    s=system(cmd);
    if (s) WnpSendToCom("wnp_stereo failed, status=%d",s);
    break;
  }
#endif
}

void WnpGlLineWidth(float w)
{
  if (lineWidth!=w) {
    if (w>maxSmoothWidth) {
      if (wnpGlBoxMode&WNP_MODE_SMOOTH) glDisable(GL_LINE_SMOOTH);
    }
    glLineWidth(w);
    if (w<=maxSmoothWidth) {
      if (wnpGlBoxMode&WNP_MODE_SMOOTH) glEnable(GL_LINE_SMOOTH);
    }
  }
  lineWidth=w;
}

void WnpGlPointSize(float w)
{
  glPointSize(w);
}

void WnpSetLWMap(float map[4])
{
  lwmin=map[0];
  lwmax=map[1];
  lwstp=map[2];
  maxSmoothWidth=map[3];

  if (lwmin<lwr[0]) lwmin=lwr[0];
  /*  if (lwstp<lgr) lwstp=lgr; */
  if (maxSmoothWidth>lwr[1]) maxSmoothWidth=lwr[1];

  lineWidth=-1;
  WnpVLineWidth(1);
  WnpSetThreeDRedrawFlag(TRUE);
}

float WnpRealLineWidth(int w)
{
  return(clamp(lwmin+(w-1)*lwstp,lwmin,lwmax));
}

void WnpToggleTxtBuffMode(void)
{
  if (TxtBuffers==WNP_MODE_BANDF) {
    ViewportSetModeFlags("Menu",TxtBuffers,FALSE);
    ViewportSetModeFlags("Comm",TxtBuffers,FALSE);
    TxtBuffers=WNP_MODE_FRONT;
    ViewportSetModeFlags("Menu",TxtBuffers,TRUE);
    ViewportSetModeFlags("Comm",TxtBuffers,TRUE);
    WnpSendToCom("Front");
  }else{
    ViewportSetModeFlags("Menu",TxtBuffers,FALSE);
    ViewportSetModeFlags("Comm",TxtBuffers,FALSE);
    TxtBuffers=WNP_MODE_BANDF;
    ViewportSetModeFlags("Menu",TxtBuffers,TRUE);
    ViewportSetModeFlags("Comm",TxtBuffers,TRUE);
    WnpSendToCom("Front&Back");
  }
}

void WnpToggleDoubleBufferMode(void)
{
  if (allowDBswitching) {
    if (DoubleBufferMode) WnpSendToCom("double buffer mode disabled.");
    else WnpSendToCom("double buffer mode enabled.");
    DoubleBufferMode=!DoubleBufferMode;
  }
}

void WnpStereoPar(int op,int dir) {

  /* Adjust YLEFT and YSTEREO parameters for above/below stereo */
  /* May be connected to keys in keyboard handler to allow interactive */
  /* tuning. New parameters will be written stdout */
  
  switch (op) {
  case 1:
    YLEFT+=dir;
    break;
  case 2:
    YSTEREO+=dir;
    break;
  }
  WnpSendToCom("YLEFT %4d YSTEREO %d , full=%d",YLEFT,YSTEREO, win_fsgeom[3]);
}

void WnpCEmode(int mode)
{
  /* Set crystal eyes stereo mode. If mode is < 0, try next mode */

  if (mode<0) CEmode=(++CEmode)%nCEmodes;
  else CEmode=clamp(mode,0,nCEmodes-1);
  YLEFT=yleft[CEmode];
  YSTEREO=ystereo[CEmode];
  WnpSetGlBoxRedrawFlag(TRUE);
}
