#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wnp.h"
#include "wnpgl.h"

static float TR[3]={0.0,0.0,0.0};      /* view shift for current projection  */
static void ShiftView(float x,float y,float z);
int gaussj(int n,float *a,float *b);   /* no prototype in wnp.h...           */

/*------------------------------------------------------------------------*/
static void ShiftView(float x,float y,float z)
{
  TR[0]=x; TR[1]=y;TR[2]=z;
}
/*-------------------- Perspective Projection  --------------------*/

void WnpStereoPersp(float xmin,float xmax,float ymin,float ymax,float znear,
		float zfar,float zzps,float dist,float eye,float sep)
{
  float xmid,ymid,clip_near,clip_far,top,bottom,left,right,dx,dy,n_over_d;
  WnpMatrix pm;

  dx=xmax-xmin;
  dy=ymax-ymin;
  xmid=0.5*(xmax+xmin);
  ymid=0.5*(ymax+ymin);
  clip_near=dist+zzps-znear;
  clip_far=dist+zzps-zfar;
  n_over_d=clip_near/dist;
  top=n_over_d*dy*0.5;
  bottom=-top;
  right=n_over_d*(dx*0.5-eye);
  left=n_over_d*(-dx*0.5-eye);
  if (eye>0) {
    right+=sep;
    left+=sep;
  }else if (eye<0) {
    right-=sep;
    left-=sep;
  }
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(left,right,bottom,top,clip_near,clip_far);
  glTranslatef(-xmid-eye,-ymid,0.0);
  glGetFloatv(GL_PROJECTION_MATRIX,(float *)pm);
  ShiftView(0,0,-zzps-dist);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(TR[0],TR[1],TR[2]);
  WnpFogRange(clip_near,clip_far);
  WnpPlotProjection(pm);
  WnpPlotClipPlanes(-clip_near,-clip_far);
}

/*-------------------- Orthogonal Projection  --------------------*/

void WnpStereoOrtho(float xmin,float xmax,float ymin,float ymax,float znear,
		float zfar,float dist,float eye,float sep)
{
  float xmid,ymid,clip_near,clip_far,top,bottom,left,right,dx,dy,pseudo;
  WnpMatrix pm;

  if (dist>0) pseudo=atan(eye/dist)*180.0/3.14159265;
  else if (eye<0) pseudo=-3;
  else pseudo=+3;
  dx=xmax-xmin;
  dy=ymax-ymin;
  xmid=0.5*(xmax+xmin);
  ymid=0.5*(ymax+ymin);
  clip_near=0;
  clip_far=znear-zfar;
  top=dy*0.5;
  bottom=-top;
  right=dx*0.5;
  left=-dx*0.5;
  if (eye>0) {
    right+=sep;
    left+=sep;
  }else if (eye<0) {
    right-=sep;
    left-=sep;
  }
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(left,right,bottom,top,clip_near,clip_far);
  glRotatef(-pseudo,0,1,0);
  glTranslatef(-xmid,-ymid,0);
  glGetFloatv(GL_PROJECTION_MATRIX,(float *)pm);
  ShiftView(0,0,-znear);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(TR[0],TR[1],TR[2]);
  WnpFogRange(clip_near,clip_far);
  WnpPlotProjection(pm);
  WnpPlotClipPlanes(-clip_near,-clip_far);
}

void WnpZeroViewShift(void)
{
  TR[0]=0;
  TR[1]=0;
  TR[2]=0;  
}

void WnpViewShift(WnpMatrix m)
{
  /* undo view shift in MVT matrix m */

  m[3][0]-=TR[0]; m[3][1]-=TR[1]; m[3][2]-=TR[2];
}

void WnpCleanMVT(WnpMatrix a)
{
  /* "clean" MVT matrix a, i.e. scale = 1 , rotation orthogonal,... */

  float s;
  int i;
  
  for (i=0;i<3;i++) a[i][3]=0.0;
  a[3][3]=1.0;
  
  for (i=0,s=0.0;i<3;i++) s+=a[0][i]*a[0][i];
  s=sqrt(s);
  for (i=0;i<3;i++) a[0][i]/=s;	/* first row normalized */
  
  a[2][0]=a[0][1]*a[1][2]-a[0][2]*a[1][1];
  a[2][1]=a[0][2]*a[1][0]-a[0][0]*a[1][2];
  a[2][2]=a[0][0]*a[1][1]-a[0][1]*a[1][0];
  for (i=0,s=0.0;i<3;i++) s+=a[2][i]*a[2][i];
  s=sqrt(s);
  for (i=0;i<3;i++) a[2][i]/=s;	/* third row orthonormal to first */
  
  a[1][0]=a[2][1]*a[0][2]-a[2][2]*a[0][1];
  a[1][1]=a[2][2]*a[0][0]-a[2][0]*a[0][2];
  a[1][2]=a[2][0]*a[0][1]-a[2][1]*a[0][0];
  for (i=0,s=0.0;i<3;i++) s+=a[1][i]*a[1][i];
  s=sqrt(s);
  for (i=0;i<3;i++) a[1][i]/=s;	/* second row orthonormal to 1,3 */
}

int WnpPickPoint(int xw,int yw,float wl[6])
{
  /* translates point (xw,yw) in window coordinates into line wl in world
     coordinates (i.e. two points at near and far clippingplanes with projection
     (xw,yw)). Uses current viewport, projection and modelling/viewing matrix  */

  int r,i,j,k;
  GLint vwp[4];
  float xfs[4],xfw[4];
  WnpMatrix pm,vm,tm;

  /* get viewport, projection and modelling/viewing matrices */

  glGetIntegerv(GL_VIEWPORT,vwp);
  glGetFloatv(GL_PROJECTION_MATRIX,(float *)pm);
  glGetFloatv(GL_MODELVIEW_MATRIX,(float *)vm);

  /* get product of projection and viewing matrices, and invert  */

  for (i=0;i<4;i++) 
    for (k=0;k<4;k++) 
      for (tm[i][k]=0.0,j=0;j<4;j++) tm[i][k]+=vm[i][j]*pm[j][k];
  r=gaussj(4,(float *)tm,(float *)NULL);

  if (!r) {

    /* convert to normalized window coordinates, at near clipping plane  */

    xfs[0]=(2.0*(xw-vwp[0])+1.0)/vwp[2]-1.0;
    xfs[1]=(2.0*(yw-vwp[1])+1.0)/vwp[3]-1.0;
    xfs[2]=-1.0; xfs[3]=1.0;

    /* backtransform normalized point... */

    for (i=0;i<4;i++) for (xfw[i]=0.0,j=0;j<4;j++) xfw[i]+=tm[j][i]*xfs[j];
    if (xfw[3] < 0.000000001 && xfw[3] > -0.000000001) xfw[3]=1.0;
    wl[0]=xfw[0]/xfw[3]; wl[1]=xfw[1]/xfw[3]; wl[2]=xfw[2]/xfw[3]; 

    /* backtransform normalized point far point... */

    xfs[2]=1.0;
    for (i=0;i<4;i++) for (xfw[i]=0.0,j=0;j<4;j++) xfw[i]+=tm[j][i]*xfs[j];
    if (xfw[3] < 0.000000001 && xfw[3] > -0.000000001) xfw[3]=1.0;
    wl[3]=xfw[0]/xfw[3]; wl[4]=xfw[1]/xfw[3]; wl[5]=xfw[2]/xfw[3]; 
  }else{
    for (i=0;i<6;i++) wl[i]=0.0;
    wl[2]=1.0;
  }
  return(r);
}

/*-------------------- MakeStipple --------------------*/

int *WnpMakePTStipple(int level)
{
  int i;
  int *s;

  /* stipple patterns for pseudo-transparent surfaces */

  level=clamp(level,0,3);
  if (level>0) {
    s=(int *)malloc(32*sizeof(int));
    if (s) {
      switch (level) {
      case 3:
	for (i=0;i<32;i+=4) {    /* 75% transparent */
	  s[i]  =0x88888888;
	  s[i+1]=0x44444444;
	  s[i+2]=0x22222222;
	  s[i+3]=0x11111111;
	}
	break;
      case 2:
	for (i=0;i<32;i+=2) {    /* 50 % transparent */
	  s[i]  =0xaaaaaaaa;
	  s[i+1]=0x55555555;
	}
	break;
      case 1:
	for (i=0;i<32;i+=4) {    /* 25 % transparent */
	  s[i]  =0xeeeeeeee;
	  s[i+1]=0xdddddddd;
	  s[i+2]=0xbbbbbbbb;
	  s[i+3]=0x77777777;
	}
	break;
      }
    }
  }else s=(int *)NULL;
  return(s);
}
