#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include "wnp.h"

#define DEL 191		    /* DEL key + 64 */
#define ctrl(x) ((x)-64)

typedef struct h_rec {			/* history record definition */
  struct h_rec *prv;
  struct h_rec *nxt;
  int seqnr;
  char *text;
} h_rec;

static h_rec *last_h;			/* pointer to last history rec */
static h_rec *first_h;			/* pointer to first history rec */
static h_rec *history;			/* pointer to current hist rec */
static int hsz;				/* number of history recs */
static int mxhsz;			/* max size of history file */
static short pos;			/* cursor position */
static char kbuf[WNP_MAXLINE];          /* line buffer */
static int nc;				/* number of characters incl '\0' */
static int KeyboardMode;                /* local or remote */
static int kbd_echo = 1;                /* echo keyboard input in command line buffer */

static void line_init(void);
static void add_kbd_history(char *s);
static void prv_kbd_history(void);
static void nxt_kbd_history(void);

/*-------------------- KEYBOARD Handler --------------------*/

void WnpKeyboardInit(void)
{
  /* initialize keyboard subsystem */
  
  line_init();
  hsz=0;
  mxhsz=256;
  first_h=(h_rec *)NULL;
  last_h=(h_rec *)NULL;
  history=last_h;
  kbd_echo=1;
  WnpSetKeyboardMode(WNP_KBD_REMOTE);
}

void WnpAddKey(unsigned int key,unsigned int state)
{
  int i;
  
  if (KeyboardMode==WNP_KBD_IGNORE) return;
  
  if (state==WNP_MOD_CTRL+WNP_MOD_SHIFT) { /* for MAC users */
    key=key+64;                            /* uncontrolify key */
    state=WNP_MOD_ALT;                     /* press ALT, release SHIFT+CTRL */
  }

  if (state==WNP_MOD_ALT) {                /* may be an "accelerator" */
    switch (key) {
    case 'P':
    case 'p':
      WnpSetViewMode(WNP_VIEW_PERSPECTIVE);
      break;
    case 'O':
    case 'o':
      WnpSetViewMode(WNP_VIEW_ORTHO);
      break;
    case 'L':
    case 'l':
      WnpToggleLightSourceLocal();
      break;
    case 'T':
    case 't':
      WnpToggleTxtBuffMode();
      break;
    case 'D':
    case 'd':
      WnpToggleDoubleBufferMode();
      break;
    case 'S':                       /* change crystal eye stereo mode */
    case 's':
      WnpCEmode(-1);
      break;
    case ',':                       /* adjust YLEFT down */ /* dead in DOS */
      WnpStereoPar(1,-1);           
      WnpSetGlBoxRedrawFlag(TRUE);
      break;
    case '.':                       /* adjust YLEFT up */    /* dead in DOS */
      WnpStereoPar(1,+1);
      WnpSetGlBoxRedrawFlag(TRUE);
      break;
    case ';':                       /* adjust YSTEREO down */  /* dead in DOS */
      WnpStereoPar(2,-1);
      WnpSetGlBoxRedrawFlag(TRUE);
      break;
    case '\'':                      /* adjust STERO up */  /* dead in DOS */
      WnpStereoPar(2,+1);
      WnpSetGlBoxRedrawFlag(TRUE);
      break;
    default:
      WnpRingBell();
      break;
    }
  }else if (isprint(key)) {	          /* insert printable characters */
    if (nc==WNP_MAXLINE) {
      WnpRingBell();
    }else{
      for (i=nc;i>pos;i--) kbuf[i]=kbuf[i-1];
      kbuf[pos++]=(char)key;
      nc++;
    }
  }else{
    switch (key+64) {		/* +64 = uncontrolify */
    case 'U' :			/* erase to start of line */
      for (i=pos;i<nc;i++) kbuf[i-pos]=kbuf[i];
      nc-=pos;
      pos=0;
      break;
    case 'A' :			                    /* go to start of line */
      pos=0;
      break;
    case 'B' :			                    /* back 1 character */
      if (pos>0) pos--;
      break;
    case 'F' :			                    /* forward 1 character */
      if (pos<nc-1) pos++;
      break;
    case 'E' :			                    /* go to end of line */
      pos=nc-1;
      break;
    case 'D' :		                            /* delete next character */
      if (pos<nc-1) {
	for (i=pos+1;i<nc;i++) kbuf[i-1]=kbuf[i];
	nc--;
      }
      break;
    case DEL :					    /* just in case... */
    case 'H' :					    /* delete prev. character */
      if (pos>0) {
	pos=pos-1;
	for (i=pos;i<nc-1;i++) kbuf[i]=kbuf[i+1];
	nc--;
      }
      break;
    case 'K' :			       /* kill to end of line */
      kbuf[pos]='\0';
      nc=pos+1;
      break;		
    case 'M' :			       /* carriage return */
      kbuf[nc++]='\n';kbuf[nc]='\0';
      if (kbd_echo) WnpSendToCom("%s",kbuf);         /* send it to message area... */
      if (KeyboardMode==WNP_KBD_REMOTE) {
	WnpSendTextToClient("KEYB",nc,kbuf); /* send it to message queue */
	WnpSetPrompt((char *)NULL);
      }else if (KeyboardMode==WNP_KBD_LOCAL) {
	WnpSetLocalString(kbuf);
	WnpSetLocalPrompt((char *)NULL);
      }
      if (kbd_echo) add_kbd_history(kbuf);
      history=(h_rec *)NULL;
      line_init();
      break;
    case 'P' :			       /* previous history line      */
      prv_kbd_history();
      break;
    case 'N' :			       /* next history line          */
      nxt_kbd_history();
      break;
    default: break;		       /* ignore all others          */
    }
  }
  if (kbd_echo) WnpSendKeyboardToDisplay(pos,kbuf);  /* send it to command box     */
}

void WnpKeyboardEcho(int state)
{
  kbd_echo=state;
}

int WnpSetKeyboardMode(int mode)
{
  int r;

  r=KeyboardMode;
  KeyboardMode=mode;
  WnpDrawKeyLine();
  return(r);
}

int WnpKeyboardMode(void)
{
  return(KeyboardMode);
}
/*========================================================================*/

/* Local functions for history file handling... */

static void line_init(void)
{
  /* initialize line buffer */
  
  pos=0;
  kbuf[pos]='\0';
  nc=1;
}

static void add_kbd_history(char *s)
{
  /* add string s to keyboard history */
  
  char *nl;
  h_rec new,*old_last,*old_first;
  
  if (mxhsz<1) return;
  
  new.prv=last_h;
  new.nxt=(h_rec *)NULL;
  if (new.prv) new.seqnr=(new.prv)->seqnr+1;
  else new.seqnr=1;
  new.text=strdup(s);
  if (!new.text) return;
  nl=strrchr(new.text,'\n'); if (nl) *nl='\0';  /* remove new-line, if any */
  
  if (hsz==mxhsz) {		/* replace old first with new */
    if (first_h->text) free(first_h->text);
    old_last=last_h;
    old_first=first_h;
    first_h=old_first->nxt;
    first_h->prv=(h_rec *)NULL;
    last_h=old_first;
    *last_h=new;
    old_last->nxt=last_h;
  }else{			
    if (last_h) {		/* insert new after last */
      last_h->nxt=(h_rec *)malloc(sizeof(h_rec));
      if (last_h->nxt) {
	*(last_h->nxt)=new;
	last_h=last_h->nxt;
	hsz++;
      }else free(new.text); 
    }else{			/* start history file */
      last_h=(h_rec *)malloc(sizeof(h_rec));
      first_h=last_h;
      if (last_h) {
	*last_h=new;
	hsz++;
      }else free(new.text);	    
    }
  }
}

static void prv_kbd_history(void)
{
  /* make previous history record the current one */
  
  if (!history) history=last_h;
  else if (history->prv) history=history->prv;
  if (history) {
    if (history->text) {
      strncpy(kbuf,history->text,WNP_MAXLINE-1);
      kbuf[WNP_MAXLINE-1]='\0';
      pos=strlen(kbuf);
      nc=pos+1;
    }else{
      line_init();
    }
  }
}

static void nxt_kbd_history(void)
{
  /* make next history record the current one */
  
  if (history) {
    if (history->nxt) {
      history=history->nxt;
      if (history->text) {
	strncpy(kbuf,history->text,WNP_MAXLINE-1);
	kbuf[WNP_MAXLINE-1]='\0';
	pos=strlen(kbuf);
	nc=pos+1;
      }else{
	kbuf[0]='\0';
	pos=0;
	nc=1;
      }
    }else{
      line_init();
      history=(h_rec *)NULL;
    }
  }
}
