#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "../client/token.h"
#include "wnp.h"
#include "wnpgl.h"

#define abortButton -2
#define doneButton  -3
#define prevButton  -6
#define nextButton  -5

static WnpMenu *wnpMainMenu=(WnpMenu *)NULL;
static WnpMenu *wnpActiveMenu=(WnpMenu *)NULL;
static int      menuFont = 0;
static int      redraw=TRUE;
static int      clear=FALSE;

void WnpMenuBoxInit(void)
{
  wnpMainMenu=WnpCreateMenu("Menu",WNP_MENU_BUTTONS,WNP_MENU_LINES,7);
  wnpActiveMenu=wnpMainMenu;
  WnpSetMenuText(wnpMainMenu,abortButton,"abort");
  WnpSetMenuText(wnpMainMenu,doneButton,"done");
  WnpSetMenuText(wnpMainMenu,prevButton,"previous");
  WnpSetMenuText(wnpMainMenu,nextButton,"next");
  WnpSetMenuAction(wnpMainMenu,prevButton,WnpScrollMenuDown);
  WnpSetMenuAction(wnpMainMenu,nextButton,WnpScrollMenuUp);
  wnpMainMenu->dirty=TRUE;
  redraw=TRUE;
}

void WnpDrawMenuBox(int mode)
{
  int n,m;
  Viewport *v;

  if (mode==WNP_MODE_BACK&&!redraw) return;
  n=(WnpGetStereoMode()==WNP_STEREO_CRYSTAL ? WNP_FNT_STEREO : WNP_FNT_COURIER);
  menuFont=WnpSelFont(n,0.95*WnpPixelsPerLine());
  v=VwpGet("Menu");
  if (v&&v->open) {
    m=(mode==WNP_MODE_BACK&&clear ? WNP_MODE_BANDF : mode);
    v->mode=(v->mode&~WNP_MODE_BANDF)|m;
    WnpDrawMenu(wnpActiveMenu);
    if (mode==WNP_MODE_BACK) v->mode=(v->mode&~WNP_MODE_BANDF)|WNP_MODE_FRONT; 
    redraw=!(mode&WNP_MODE_BACK);
  }
}

void WnpDrawMenuBoxBB(void)
{
  /* redraw Menubox back buffer, if necessary */
  
  if (redraw) WnpDrawMenuBox(WNP_MODE_BACK);
}

void WnpDrawMenu(WnpMenu *m)
{
  if (m) m->dirty=TRUE;
  WnpRefreshMenu(m);
}

void WnpRefreshMenu(WnpMenu *m)
{
  int i,stereomode,view,port,r;
  GLint vwp[4];
  float yl,yh;
  Viewport *v;

  v=VwpGet("Menu"); r=(v? !(v->mode&WNP_MODE_BACK) : FALSE);
  if ((!m||m->dirty)&&ViewportIsOpen("Menu")) {
    WnpGlRestoreCursorPatch();
    WnpGlSetCursorRedrawFlag(TRUE);
    if (m) {
      /* for (m->ndef=0,i=0;i<m->mbtn;i++) if (m->btn[i].active) m->ndef=i+1; */
     WnpActivateButton(m,abortButton,TRUE);
     WnpActivateButton(m,doneButton,TRUE);
     WnpActivateButton(m,nextButton,(m->ndef>0)&&(m->top+m->nbtn-m->nres<m->ndef));
     WnpActivateButton(m,prevButton,(m->ndef>0)&&(m->top>0));
    }
    WnpSetFont(menuFont);
    stereomode=WnpGetStereoMode();
    redraw=redraw||r;
    clear=FALSE;
    for (view=WNP_LEFT_VIEW;view<=WNP_RIGHT_VIEW;view++) {
      port=WnpSelectTextport(view,stereomode);
      if (port!=WNP_VWP_NONE) {
	ViewportSet("Menu",port);
	glGetIntegerv(GL_VIEWPORT,vwp);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0,vwp[2],0,WNP_MENU_LINES);
	glMatrixMode(GL_MODELVIEW);
	WnpMenuColor(WNP_MENU_COLOR_BCKG);
	glRectf(0,0,vwp[2],WNP_MENU_LINES);
	if (m) {
	  for (i = -m->nres;i<0;i++) {                        /* special buttons  */
	    yl=-i-1; yh=yl+0.9;
	    if (m->btn[i].background!=WNP_MENU_COLOR_BCKG) {
	      WnpMenuColor(m->btn[i].background);
	      glRectf(0,yl,vwp[2],yh);
	    }
	    WnpMenuColor(m->btn[i].color);
	    WnpTextPos3f(2,yl+0.2,0);
	    WnpTextDraw(m->btn[i].text);
	  }
	  for (i=0;i<m->nbtn-m->nres;i++) {                   /* scrolling region */
	    if (i+m->top>=m->ndef) break;
	    yl=(WNP_MENU_LINES-i-1); yh=yl+0.9;
	    if (m->btn[i+m->top].background!=WNP_MENU_COLOR_BCKG) {
	      WnpMenuColor(m->btn[i+m->top].background);
	      glRectf(0,yl,vwp[2],yh);
	    }
	    WnpMenuColor(m->btn[i+m->top].color);
	    WnpTextPos3f(2,yl+0.2,0);
	    WnpTextDraw(m->btn[i+m->top].text);
	  }
	}
      }
    }
  }
}

void WnpDrawMenuButton(WnpMenu *m,int i)
{
  float yl,yh;
  int stereomode,view,port;
  GLint vwp[4];
  Viewport *v;

  if (m&&(v=VwpGet("Menu"))&&v->open&& i>= -m->nres && i<m->mbtn) {
    stereomode=WnpGetStereoMode();
    for (view=WNP_LEFT_VIEW;view<=WNP_RIGHT_VIEW;view++) {
      port=WnpSelectTextport(view,stereomode);
      if (port!=WNP_VWP_NONE) {
	ViewportSet("Menu",port);
	glGetIntegerv(GL_VIEWPORT,vwp);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0,vwp[2],0,WNP_MENU_LINES);
	glMatrixMode(GL_MODELVIEW);
	yl=(i<0 ? -i-1 : WNP_MENU_LINES-i-1+m->top);
	yh=yl+0.9;
	WnpMenuColor(m->btn[i].background);
	glRectf(0,yl,vwp[2],yh);
	WnpMenuColor(m->btn[i].color);
	WnpTextPos3f(2,yl+0.2,0);
	WnpTextDraw(m->btn[i].text);
      }
    }
  }
}

int WnpActivateButton(WnpMenu *m,int i,int new)
{
  int s;

  if (m&& i>= -m->nres && i<m->mbtn) {
    s=m->btn[i].active;
    m->btn[i].active=new;
    m->dirty=new!=s;
  }else s=FALSE;

  return(s);
}

WnpMenu *WnpCreateMenu(char *vwp,int mbtn,int nbtn,int nres)
{
  WnpMenu *m;
  int i;

  m=(WnpMenu *)malloc(sizeof(WnpMenu));
  if (m) {
    m->btn=(WnpMenuButton *)malloc((mbtn+nres)*sizeof(WnpMenuButton));
    if (m->btn) {
      for (i=0;i<mbtn+nres;i++) m->btn[i].text=(char *)NULL;
      m->btn+=nres;
      m->mbtn=mbtn;
      m->nbtn=nbtn;
      m->nres=nres;
      m->host = strdup(vwp);
      m->dirty=TRUE;
      WnpClearMenu(m);
      for (i = -m->nres;i<0;i++) WnpClearMenuButton(m,i);
    }else{
      free(m->btn);
    }
  }
  return(m);
}    

void WnpClearMenu(WnpMenu *m)
{
  int i;
  if (m) {
    m->ndef=0;
    WnpResizeMenu(m,WNP_MENU_BUTTONS);
    for (i=0;i<m->mbtn;i++) WnpClearMenuButton(m,i);
    m->top=0; 
  }
}

void WnpLoadMenuButtons(WnpMenu *m,int nxt,char *text)
{
  Parse prs;
  char buf[WNP_MAXREC];

  if (m&&text&&nxt>=0) {
    prs.buf=text; 
    prs.tok=buf; prs.mts=sizeof(buf);
    prs.sep=""; prs.pos=0;
    while (get_token(&prs)) {
      if (nxt>=m->mbtn) WnpResizeMenu(m,m->mbtn+4096);
      if (nxt>=m->mbtn) break;
      WnpClearMenuButton(m,nxt);
      if (prs.ltok>2&&((!strncmp(buf,"t!",2)||(!strncmp(buf,"T!",2))))) {
	m->btn[nxt].text=strdup(buf+2); 
	m->btn[nxt].active=FALSE;      /* just a title... */
      }else{
	m->btn[nxt].text=strdup(buf);
	m->btn[nxt].active=TRUE;
      }
      nxt++;
      if (nxt>m->ndef) m->ndef=nxt;
    }
    //menu    m->top=0;
  }
}

void WnpResizeMenu(WnpMenu *m,int newsize)
{
  WnpMenuButton *new;
  int i;

  if (newsize<m->ndef) newsize=m->ndef;
  if (newsize!=m->mbtn) {
    if (newsize<m->mbtn) {
      for (i=newsize;i<m->mbtn;i++) WnpClearMenuButton(m,i);
      m->mbtn=newsize;
    }
    new=realloc(m->btn-m->nres,(newsize+m->nres)*sizeof(WnpMenuButton));
    if (new) {
      m->btn=new+=m->nres;
      for (i=m->mbtn;i<newsize;i++) {
	m->btn[i].text=(char *)NULL;
	WnpClearMenuButton(m,i);
      }
      m->mbtn=newsize;
    }
  }
}

void WnpHighlightMenuButton(Viewport *focus,float y)
{
  /* update menu button highlight, mouse is in 'focus', at position (x,y) */
  /* position is given in fractional GL box coordinates.                 */

  int n,i,k;
  WnpMenu *m;

  if (wnpActiveMenu) {
    WnpGlRestoreCursorPatch();
    WnpGlSetCursorRedrawFlag(TRUE);
    m=wnpActiveMenu;
    n=WNP_MENU_LINES;
    if (focus&&strcmp(focus->name,"Menu")==0) {
      n=WNP_MENU_LINES*y;
      if (n<m->nres) n=-n-1;
      else n=m->top+WNP_MENU_LINES-n-1;
    }else n=m->mbtn;
    for (k=0;k<WNP_MENU_LINES-m->nres;k++) {   /* off in normal menu */
      i=k+m->top;
      if (i>=m->ndef) break;
      if (i!=n&&m->btn[i].color==WNP_MENU_COLOR_HTXT
	  &&m->btn[i].background==WNP_MENU_COLOR_HBCK) {
	m->btn[i].color=WNP_MENU_COLOR_TEXT;
	m->btn[i].background=WNP_MENU_COLOR_BCKG;
	WnpDrawMenuButton(m,i);
	redraw=TRUE;
      }
    }
    for (i=-m->nres;i<0;i++) {                 /* off in special menu */
      if (i!=n&&m->btn[i].color==WNP_MENU_COLOR_HTXT
	  &&m->btn[i].background==WNP_MENU_COLOR_HBCK) {
	m->btn[i].color=WNP_MENU_COLOR_TEXT;
	m->btn[i].background=WNP_MENU_COLOR_BCKG;
	WnpDrawMenuButton(m,i);
	redraw=TRUE;
      }
    }
    if (-m->nres <= n && n < m->mbtn) {
      if (m->btn[n].active) {
	if (m->btn[n].color!=WNP_MENU_COLOR_HTXT||m->btn[n].background!=WNP_MENU_COLOR_HBCK) {
	  m->btn[n].color=WNP_MENU_COLOR_HTXT;
	  m->btn[n].background=WNP_MENU_COLOR_HBCK;
	  WnpDrawMenuButton(m,n);
	}
      }
    }
  }
}

void WnpPressMenuButton(Viewport *focus,float y)
{
  /* press menu button , mouse is in 'focus', at position (x,y) */
  /* position is given in fractional viewport  coordinates.     */

  int n;
  WnpMenu *m;

  WnpGlRestoreCursorPatch();
  WnpGlSetCursorRedrawFlag(TRUE);
  if (wnpActiveMenu) {
    m=wnpActiveMenu;
    if (focus&&strcmp(focus->name,"Menu")==0) {
      n=WNP_MENU_LINES*y;
      if (n<m->nres) n=-n-1;
      else n=m->top+WNP_MENU_LINES-n-1;
      if (n >= -m->nres && n < m->mbtn) {
	if (m->btn[n].active) {
	  if (m->btn[n].action) m->btn[n].action(m,n);
	  else if (m->btn[n].text) {
	    WnpSendTextToClient("MENU",strlen(m->btn[n].text),m->btn[n].text);
	    WnpSendToCom("%s",m->btn[n].text);
	  }
	}
      }
    }
  }
}

void WnpClearMenuButton(WnpMenu *m,int i)
{
  int k;

  if (m) {
    if (i>= -m->nres && i<m->mbtn) {
      m->btn[i].color=WNP_MENU_COLOR_TEXT;
      m->btn[i].background=WNP_MENU_COLOR_BCKG;
      m->btn[i].active=FALSE;
      if (m->btn[i].text) free(m->btn[i].text); m->btn[i].text=(char *)NULL;
      m->btn[i].action=(WnpMenuAction)NULL;
      m->dirty=TRUE;
      clear=TRUE;
      for (k=m->ndef-1;k>=0;k--) if (m->btn[i].text&&strlen(m->btn[i].text)>0) break;
      m->ndef=k+1;
    }
  }
}

void WnpScrollMenuDown(void *menu,int dummy)
{
  WnpMenu *m;
  
  dummy=0;
  m=(WnpMenu *)menu;
  if (m) {
    if (m->top>0) {
      m->top-=m->nbtn-m->nres;
      if (m->top<0) m->top=dummy;
      WnpDrawMenu(m);
    }
  }
}

void WnpScrollMenuUp(void *menu,int dummy)
{
  WnpMenu *m;
  
  dummy=dummy;
  m=(WnpMenu *)menu;
  if (m) {
    if (m->top+m->nbtn-m->nres<=m->ndef) {
      m->top+=m->nbtn-m->nres;
      WnpDrawMenu(m);
    }
  }
}

void WnpSetMenuText(WnpMenu *m,int i,char *text)
{
  if (m && i>= -m->nres && i<m->mbtn) {
    if (m->btn[i].text) free(m->btn[i].text);
    m->btn[i].text = (text ? strdup(text) : (char *)NULL);
    m->dirty=TRUE;
  }
}

void WnpSetMenuAction(WnpMenu *m,int i,WnpMenuAction action)
{
  if (m && i>= -m->nres && i<m->mbtn) m->btn[i].action=action;
}

int WnpMenuColor(int col)
{
  /* access special color table (index < 0) for menus and messages... */

  return(WnpColor(-col));
}

WnpMenu *WnpSetActiveMenu(WnpMenu *new)
{
  WnpMenu *old;

  old=wnpActiveMenu;
  if (old!=new) {
    wnpActiveMenu=new;
    WnpDrawMenu(new);
  }
  return(old);
}

WnpMenu *WnpActiveMenu(void)
{
  return(wnpActiveMenu);
}

WnpMenu *WnpMainMenu(void)
{
  return(wnpMainMenu);
}
