#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "wnp.h"
#include "wnpgl.h"

static char **line;		     /* pointers to lines of msg buffer        */
static int Lpt,Nmsgs=0;		     /* highest message number, Buffer size    */
static int Nvmsgs;                   /* size of message buffer in lines        */
static int win_top;		     /* msg# of first window line              */
static char *prompt;		     /* prompt string                          */
static char *local_prompt;	     /* prompt string in local KBD mode        */
static char key_line[WNP_MAXLINE];   /* keybord echo area                      */
static int cpt;			     /* cursor position in key_line            */

static int messageFont=0;            /* font to be used in command/message box */ 
static int redraw=TRUE;              /* need redraw of background buffer       */
static int pending=0;                /* number of pending message lines        */
static int cpending=0;               /* pending changes in command line        */

typedef struct kbmn_t {              /* save active menu and kbd mode          */
  WnpMenu *menu;                     /* used in dumpMessageBuffer CB           */
  int      kbd;
} kbmn_t;

static void dumpMessageBufferCB(char *file,void *data);
static void dumpMessageBuffer(char *file);

/*-------------------- MessageBoxInit --------------------*/

void WnpMessageBoxInit(void)
{
  int i;
  
  Lpt = -1;
  win_top=0;
  redraw=TRUE;
  pending=0;
  cpending=0;
  key_line[0]='\0';
  cpt=0;
  Nmsgs=2048;			/* save up to Nmsgs messages         */
  Nvmsgs=0;                     /* will be set by redraw function... */
  line=(char **) malloc(Nmsgs*sizeof(char *));
  if (!line) {
    fprintf(stderr,"***FATAL ERROR (WITNOTP, message_box_init)***\n");
    exit(1);
  }
  for (i=0;i<Nmsgs;i++) line[i]=(char *)NULL;
  prompt=(char *)NULL;
  local_prompt=(char *)NULL;
}
/*-------------------- DrawMessageBox --------------------*/

void WnpDrawMessageBox(int mode)
{
  int n;
  Viewport *v;

  if (mode==WNP_MODE_BACK&&!redraw) return;
  if ((v=VwpGet("Comm"))&&v->open) {
    v->mode=(v->mode&~WNP_MODE_BANDF)|mode;
    n=(WnpGetStereoMode()==WNP_STEREO_CRYSTAL ? WNP_FNT_STEREO : WNP_FNT_COURIER);
    messageFont=WnpSelFont(n,0.95*WnpPixelsPerLine());
    WnpSetFont(messageFont);
    pending=1; WnpDrawPendingLines();
    if (mode==WNP_MODE_BACK) v->mode=(v->mode&~WNP_MODE_BANDF)|WNP_MODE_FRONT; 
    redraw=!(mode&WNP_MODE_BACK);
    pending=0;
  }
}

void WnpDrawPendingLines(void)
{
  int jscr,kbdm,len,i,iw,n,stereomode,view,port;
  GLint vwp[4];
  Viewport *v;
  char *ps;

  if (pending||cpending) {
    WnpGlRestoreCursorPatch();
    WnpGlSetCursorRedrawFlag(TRUE);
    if ((v=VwpGet("Comm"))&&v->open) {
      stereomode=WnpGetStereoMode();
      if (pending) {
	n=Nvmsgs; 
	Nvmsgs=(v->y2)*WNP_MENU_LINES-1+0.1; if (Nvmsgs<0) Nvmsgs=0;
	win_top+=n-Nvmsgs;
	if (win_top + Nvmsgs > Lpt) win_top=Lpt-Nvmsgs+1;
	if (win_top<0) win_top=0;
	jscr=clamp(Nvmsgs/4,5,100);
      }
      kbdm=WnpKeyboardMode();
      ps=(kbdm == WNP_KBD_LOCAL ? local_prompt :(kbdm == WNP_KBD_REMOTE ? prompt : (char *)NULL));
      for (view=WNP_LEFT_VIEW;view<=WNP_RIGHT_VIEW;view++) {
	port=WnpSelectTextport(view,stereomode);
	if (port!=WNP_VWP_NONE) {
	  ViewportSet("Comm",port);
	  glGetIntegerv(GL_VIEWPORT,vwp);
	  glMatrixMode(GL_PROJECTION);
	  glLoadIdentity();
	  gluOrtho2D(0,vwp[2],0,Nvmsgs+1); /* x coords are pixels, y are lines */
	  glMatrixMode(GL_MODELVIEW);
	  if (pending) {                                /* deal with pending message lines */
	    for(i=0,iw=win_top;i<Nvmsgs;i++,iw++) {
	      if (i%jscr==0) {
		WnpMenuColor(WNP_MENU_COLOR_BCKG);
		glRectf(0,Nvmsgs-i-(jscr-1),vwp[2],Nvmsgs-i+1);
		WnpMenuColor(WNP_MENU_COLOR_TEXT);
	      }
	      if (line[iw%Nmsgs]) {
		WnpTextPos3f(2,Nvmsgs-i+0.25,0);
		WnpTextDraw(line[iw%Nmsgs]);
	      }
	    }
	  }
	  WnpMenuColor(WNP_MENU_COLOR_BCKG);
	  glRectf(0,0,vwp[2],1);
	  WnpMenuColor(WNP_MENU_COLOR_TEXT);
	  len=2;
	  if (ps) {                            /* do prompt */
	    WnpTextPos3f(len,0.25,0);
	    WnpTextDraw(ps);
	    len+=WnpTextWidth(ps,-1);
	  }
	  WnpTextPos3f(len,0.25,0);            /* draw the text */
	  WnpTextDraw(key_line);
	  if (cpt<WNP_MAXLINE-1) {             /* draw the text cursor  */
	    len+=WnpTextWidth(key_line,cpt);
	    WnpMenuColor(WNP_MENU_COLOR_CURSOR);
	    WnpTextPos3f(len,0.25,0);
	    WnpTextDraw("_");
	  }
	}
      }
    }
    redraw=!(v->mode&WNP_MODE_BACK);
    pending=0; cpending=0;
  }
}

void WnpDrawMessageBoxBB(void)
{
  /* redraw Messagebox back buffer, if necessary */
  
  if (redraw) WnpDrawMessageBox(WNP_MODE_BACK);
}

void WnpDrawKeyLine(void)
{
  cpending++;
}

void WnpSendKeyboardToDisplay(int cpos,char *msg)
{
  /* copy kbd line to display buffer */

  cpt=cpos;
  strncpy(key_line,msg,WNP_MAXLINE-1);
  key_line[WNP_MAXLINE-1]='\0';
  WnpDrawKeyLine();
}

void WnpSendToCom(char *msg,...)
{
  int i,lpt;
  va_list args;
  char *nl,buffer[WNP_MAXREC];

  if (msg[0]!='\0') {
    Lpt++;
    lpt=Lpt % Nmsgs;
    if (line[lpt]) free(line[lpt]);
    va_start(args,msg);
    vsprintf(buffer,msg,args);
    va_end(args);
    for (i=strlen(buffer)-1;i>=0;i--) if (buffer[i]!=' ') break;
    buffer[i+1]='\0';
    line[lpt]=strdup(buffer);
    if (line[lpt]) {
      nl=strrchr(line[lpt],'\n'); if (nl) *nl='\0';
      if (win_top+Nvmsgs <= Lpt) {
	win_top=Lpt-Nvmsgs+1;
	pending++;
      }else pending++;
    }
  }
}

void WnpSetLocalPrompt(char *s)
{
  if (local_prompt) free(local_prompt);
  if (s) local_prompt=strdup(s);
  else local_prompt=(char *)NULL;
  if (WnpKeyboardMode()==WNP_KBD_LOCAL) {
    WnpDrawPendingLines();
    WnpDrawKeyLine();
  }
}

void WnpSetPrompt(char *s)
{
  if (prompt) free(prompt);
  if (s) prompt=strdup(s);
  else prompt=(char *)NULL;
  if (WnpKeyboardMode()==WNP_KBD_REMOTE) WnpDrawKeyLine();
}

void WnpMsgScroll(int n)
{
  int o,fst_line;
  
  o=win_top;
  if (n!=0) {
    fst_line=(Lpt+1-Nmsgs <0 ? 0 : Lpt+1-Nmsgs);
    win_top+=n;
    if (win_top + Nvmsgs > Lpt) win_top=Lpt-Nvmsgs+1;
    if (win_top < fst_line) win_top=fst_line;
    if (win_top!=o) pending=TRUE;
  }
}

void WnpMsgPgUp(void)
{
  int o,fst_line;
  
  o=win_top;
  fst_line=(Lpt+1-Nmsgs <0 ? 0 : Lpt+1-Nmsgs);
  win_top-=Nvmsgs;
  if (win_top < fst_line) win_top=fst_line;
  if (win_top!=o) pending=TRUE;
}

void WnpMsgPgDown(void)
{
  int o,fst_line;
  
  o=win_top;
  fst_line=(Lpt+1-Nmsgs <0 ? 0 : Lpt+1-Nmsgs);
  win_top+=Nvmsgs;
  if (win_top + Nvmsgs >= Lpt) win_top=Lpt-Nvmsgs+1;
  if (win_top<fst_line) win_top=fst_line;
  if (win_top!=o) pending=TRUE;
}

void WnpMessageDump(char *file)
{
  WnpMenu *save;
  int old;
  kbmn_t *kbmn_state;

  if (!file) {
    if (!WnpLocalStringDispatcher()) {
      kbmn_state=(kbmn_t *)malloc(sizeof(kbmn_t));
      save=WnpSetActiveMenu((WnpMenu *)NULL);             /* register local-string CB */
      WnpSetLocalPrompt("name of file for text dump: ");
      old=WnpSetKeyboardMode(WNP_KBD_LOCAL);
      if (kbmn_state) {
	kbmn_state->menu=save;
	kbmn_state->kbd=old;
      }
      WnpGetLocalString(dumpMessageBufferCB,(void *)kbmn_state);
    }
  }else dumpMessageBuffer(file);
}

static void dumpMessageBufferCB(char *file,void *kbmn_state)
{
  dumpMessageBuffer(file);
  if (kbmn_state) {                                     /* restore keyboard and menu */
    WnpSetKeyboardMode(((kbmn_t *)kbmn_state)->kbd);
    WnpSetActiveMenu(((kbmn_t *)kbmn_state)->menu);
    free(kbmn_state);
  }
}

static void dumpMessageBuffer(char *name)
{
  FILE *out;
  int i,k,fst_line;

  if (name&&strlen(name)>0&&*name!=' ') {
    if (out=fopen(name,"w")) {
      fst_line=(Lpt-Nmsgs <0 ? 0 : Lpt-Nmsgs);
      for (k=0,i=fst_line;i<Lpt;i++) {                  /* not: last line not dumped. */
	if (line[i%Nmsgs]) {
	  fprintf(out,"%s\n",line[i%Nmsgs]);
	  k++;
	}
      }      
      WnpSendToCom("%d lines dumped to %s",k,name);
      fclose(out);
    }else WnpSendToCom("failed to open file %s for text dump.",name);
  }
}
