#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wnp.h"

#define activeDspObj(state) ((state)&altobj_key ? WnpAltDspObj() : WnpDisplayRoot())

static int mouse_function[WNP_MAX_MOUSE_FUNC];
static int mouse_x,mouse_y;
static int wnpPickMode = WNP_PICK_OFF;
static int usaPick = FALSE;
static int altobj_key = WNP_MOD_ALT;

void MouseReset(void)
{
  int i;

  for (i=0;i<WNP_MAX_MOUSE_FUNC;i++) mouse_function[i]=WNP_MOUSE_TRACK;
}

void MouseMapFunction(int function,int state)
{
  if (state>=0&&state<WNP_MAX_MOUSE_FUNC) mouse_function[state]=function;
}

int MouseButtonPress(int button,int state,int x,int y)
{
  float xf,yf;
  Viewport *focus;
  
  state|=button;

  if (0<=state&&state<WNP_MAX_MOUSE_FUNC) {
    if (focus=WnpScreen2FractVwpXY(x,y,&xf,&yf)) {
      if (strcmp(focus->name,"ThreeD")==0) {
	switch (mouse_function[state]) {
	case WNP_MOUSE_PICK:              WnpThreeDPick(focus,xf,yf,wnpPickMode);       break;
	case WNP_MOUSE_USAP: if (usaPick) WnpThreeDPick(focus,xf,yf,WNP_PICK_USA);      break;
	case WNP_MOUSE_SETPIVOT:          WnpThreeDPick(focus,xf,yf,WNP_PICK_PIVOT);    break;
	case WNP_MOUSE_SETALTPIVOT:       WnpThreeDPick(focus,xf,yf,WNP_PICK_ALTPIVOT); break;
	case WNP_MOUSE_SELALT:            WnpThreeDPick(focus,xf,yf,WNP_PICK_SELALT);   break;
	case WNP_MOUSE_SELMALT:           WnpThreeDPick(focus,xf,yf,WNP_PICK_SELMALT);  break;
	} 
      }else if (strcmp(focus->name,"Menu")==0) {
	switch (mouse_function[state]) {
	case WNP_MOUSE_PICK:              WnpPressMenuButton(focus,yf);                 break;
	}
      }
    }
  }
  mouse_x=x; mouse_y=y;
  return(state);
}

int MouseButtonRelease(unsigned int button,int state,int x,int y)
{
  state&=~button;
  mouse_x=x; mouse_y=y;
  return(state);
}

void MouseMoved(int state,int x,int y)
{
  int dx,dy,major;
  float xf,yf;
  Viewport *focus;

  dx=x-mouse_x;
  dy=mouse_y-y;
  major = ( abs(dx) < abs(dy) ? dy : dx);

  if (0<=state&&state<WNP_MAX_MOUSE_FUNC) {
    switch (mouse_function[state]) {
    case WNP_MOUSE_TRACK:
      focus=WnpScreen2FractVwpXY(x,y,&xf,&yf);
      WnpHighlightMenuButton(focus,yf);
      if (focus&&strcmp(focus->name,"ThreeD")==0) WnpSetGlBoxCursor(WNP_CURSOR_DOTBOX);
      else WnpSetGlBoxCursor(WNP_CURSOR_ARROW);
      break;
    case WNP_MOUSE_SCALE:
      WnpThreeDScale(major);
      break;
    case WNP_MOUSE_XYTRANS:
      WnpThreeDTranslatePx(activeDspObj(state),dx,dy,0);
      break;
    case WNP_MOUSE_ZTRANS:
      WnpThreeDTranslatePx(activeDspObj(state),0,0,-major);
      break;
    case WNP_MOUSE_XYROT:
      WnpThreeDXYrot(activeDspObj(state),dx,dy);
      break;
    case WNP_MOUSE_ZROT:
      WnpThreeDZrot(activeDspObj(state),x,y,dx,dy);
      break;
    case WNP_MOUSE_AXROT:
      WnpIncRotAxis(0.25*major);
      break;
    case WNP_MOUSE_SLABWIDTH:
      WnpThreeDSlabWidth(major);
      break;
    case WNP_MOUSE_SLABPOS:
      WnpThreeDSlabPosition(major);
      break;
    case WNP_MOUSE_STEREOSEP:
      WnpThreeDStereoSeparation(major);
      break;
    case WNP_MOUSE_VIEWDIST:
      WnpThreeDViewingDistance(major);
      break;
    case WNP_MOUSE_LIGHTXY:
      WnpAdjustLightPosition(dx,dy,0);
      break;
    case WNP_MOUSE_LIGHTZ:
      WnpAdjustLightPosition(0,0,-major);
      break;
    case WNP_MOUSE_LIGHTROT:
      WnpAdjustLightDirection(x,y,dx,dy);
      break;
    case WNP_MOUSE_LIGHTINT:
      WnpAdjustLightIntensity(major);
      break;
    default:
      break;
    }
  }
  mouse_x=x; mouse_y=y;
}

void WnpSetMouseMap(int *map)
{
  int i;
  
  for (i=0;i<WNP_MAX_MOUSE_FUNC;i++) MouseMapFunction(map[i],i);
}

void WnpSetAltObjKey(int key)
{
  if (key==WNP_MOD_ALT||WNP_MOD_SHIFT||WNP_MOD_CTRL) altobj_key=key;
}

int WnpSetPickMode(int new)
{
  int r;
  
  r=wnpPickMode;
  switch (new) {
  case WNP_PICK_USA_ON:
    usaPick=TRUE;
    break;
  case WNP_PICK_USA_OFF:
    usaPick=FALSE;
    break;
  case WNP_PICK_OBJECT:
  case WNP_PICK_ATOM:
  case WNP_PICK_XYZ:
  case WNP_PICK_XYZ_ATOM:
  case WNP_PICK_TWOD_TEXT:
    WnpSetGlBoxCursorColor(WNP_CURSOR_DOTBOX,WNP_CURSOR_GREEN);
    wnpPickMode=new;
    break;
  case WNP_PICK_OFF:
    WnpSetGlBoxCursorColor(WNP_CURSOR_DOTBOX,WNP_CURSOR_RED);
    wnpPickMode=new;
    break;
  default:
    break;
  }
  return(r);
}
