#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "wnp.h"

static GlblObj *Root = (GlblObj *)NULL;           /* root of diaplay tree    */
static GlblObj *AltObj =  (GlblObj *)NULL;        /* alt-Object              */
static DspObj *AtomPickRoot = (DspObj *)NULL;     /* root for atom picks...  */
static DspObj *ObjectPickRoot = (DspObj *)NULL;   /* root for object picks...*/

static int wnpDeletions = FALSE;                  /* are there pending atom deletions ? */

static int wnpMonColor = 2;                       /* global monitor color    */
static int wnpMonLabel = TRUE;                    /* global label state      */
static int wnpMonWidth = 1;                       /* global monitor width    */
static int wnpMonDsp   = TRUE;                    /* global display status   */
static WnpMonPair *monPairs = (WnpMonPair *)NULL; /* distance monitor pairs  */
static WnpBond *interBonds =  (WnpBond *)NULL;    /* inter-MDL bond list     */

static int defaultBondLW = 1;                     /* default bond linewidth  */
static int defaultAtomLabelColor = WNP_COLOR_BYATOM; /* default atom label color */
static DspObj *myObjPtr(cptr_t his,DspObj *root);

/*========================================================================*/

GlblObj *WnpDisplayRoot(void)
{
  return(Root);
}

GlblObj *WnpAltDspObj(void)
{
  return(AltObj);
}

GlblObj *WnpSetAltObj(GlblObj *o)
{
  GlblObj *r;

  r=AltObj;
  AltObj=o;
  if (AltObj&&AltObj->type!=DSP_GLBL) AltObj=(GlblObj *)NULL;
  return(r);
}

DspObj *WnpAtomPickRoot(void)
{
  return(AtomPickRoot);
}

DspObj *WnpSetAtomPickRoot(DspObj *new)
{
  DspObj *r;

  r=AtomPickRoot;
  AtomPickRoot=new;
  return(r);
}

DspObj *WnpObjectPickRoot(void)
{
  return(ObjectPickRoot);
}

DspObj *WnpSetObjectPickRoot(DspObj *new)
{
  DspObj *r;

  r=ObjectPickRoot;
  ObjectPickRoot=new;
  return(r);
}

WnpMonPair *WnpMonitorPairs(void)
{
  return(monPairs);
}

WnpBond *WnpInterBonds(void)
{
  return(interBonds?(interBonds->nxt?interBonds:(WnpBond *)NULL):(WnpBond *)NULL);
}

void *WnpAddDspObj(cptr_t p,cptr_t s,int type)
{
  GlblObj *parent;
  DspObj *r;
  int i,j;

  switch (type) {
  case DSP_ROOT:
    if (!Root) Root=WnpMakeGlblObj();
    r=(DspObj *)Root;
    parent=(GlblObj *)NULL;
    break;
  case DSP_GLBL:
    parent=(GlblObj *)MyObjPtr(p);
    if (parent&&parent->type==DSP_GLBL) r=(DspObj *)WnpMakeGlblObj();
    else r=(DspObj *)NULL;
    if (r) {
      for (i=0;i<4;i++) 
	for (j=0;j<4;j++) ((GlblObj *)r)->mvt[i][j]=parent->mvt[i][j];
    }
    break;
  case DSP_MDL:
    parent=(GlblObj *)MyObjPtr(p);
    if (parent&&parent->type==DSP_GLBL) r=(DspObj *)WnpMakeMdlObj();
    else r=(DspObj *)NULL;
    break;
  case DSP_CSO:
    parent=(GlblObj *)MyObjPtr(p);
    if (parent&&parent->type==DSP_GLBL) r=(DspObj *)WnpMakeCSObj();
    else r=(DspObj *)NULL;
    break;
  case DSP_TRACE:
    parent=(GlblObj *)MyObjPtr(p);
    if (parent&&parent->type==DSP_GLBL) r=(DspObj *)NULL;
    else r=(DspObj *)NULL;
    break;
  case DSP_CNT:
    parent=(GlblObj *)MyObjPtr(p);
    if (parent&&parent->type==DSP_GLBL) r=(DspObj *)NULL;
    else r=(DspObj *)NULL;
    break;
  case DSP_TRG:
    parent=(GlblObj *)MyObjPtr(p);
    if (parent&&parent->type==DSP_GLBL) r=(DspObj *)WnpMakeTRGObj();
    else r=(DspObj *)NULL;
    break;
  default:
    r=(DspObj *)NULL;
    break;
  }
  if (r) {
    r->cross=s;
    r->up=parent;
    if (parent) {
      if (parent->lst) {
	parent->lst->nxt=r;
	r->prv=parent->lst;
      }else parent->fst=r;
      parent->lst=r;
    }
    r->flags|=OBJ_FLG_DSP;
  }
  return((void *)r);
}

void WnpTransformObject(DspObj *o,WnpMatrix mmv)
{
  DspObj *son;

  if (!o) return;
  switch (o->type) {
  case DSP_GLBL:                          
    for (son=((GlblObj *)o)->fst;son;son=son->nxt) WnpTransformObject(son,mmv);
    break;
  case DSP_MDL:
    WnpTransformMdl((MdlObj *)o,mmv);
    break;
  case DSP_CSO:
    WnpTransformCSO((CSObj *)o,mmv);
    break;    
  case DSP_TRG:
    WnpTransformTRG((TRGObj *)o,mmv);
    break;
  }
}

void WnpNewObjectFlags(DspObj *o,int flags,int state)
{
  int old,diff;
  
  if (!o) return;
  old=o->flags;
  if (state) o->flags|=flags;
  else o->flags&=~flags;
  diff=(old^o->flags);
  if (diff) WnpSetModifiedFlag(o);
}

void WnpToggleObjectFlags(DspObj *o,int flags)
{
  int old,diff;
  
  if (!o) return;
  old=o->flags;
  o->flags=(flags&(flags^old))|((~flags)&old);
  diff=(old^o->flags);
  if (diff) WnpSetModifiedFlag(o);
}

void WnpSetObjectColor(DspObj *o,int color)
{
  DspObj *s;
  WnpLst *l;
  
  if (o) {
    switch (o->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)o)->fst;s;s=s->nxt) WnpSetObjectColor(s,color);
      break;
    case DSP_CSO:
      for (l=((CSObj *)o)->fst;l;l=l->nxt) {
	switch (l->type) {
	case WNP_LST_MDS:
	  if (((WnpMDS *)l)->color!=color) {
	    WnpSetModifiedFlag(o);
	    ((WnpMDS *)l)->color=color;
	  }
	  break;
	case WNP_LST_TOBJ:
	  if (((WnpTOBJ *)l)->color!=color) {
	    WnpSetModifiedFlag(o);
	    ((WnpTOBJ *)l)->color=color;
	  }
	  break;
	}
      }
      break;
    case DSP_TRG:
      if (((TRGObj *)o)->color!=color) {
	WnpSetModifiedFlag(o);
	((TRGObj *)o)->color=color;
      }
      break;
    }
  }
}

void WnpSetBackfaceColor(DspObj *o,int color)
{
  DspObj *s;
  
  if (o) {
    switch (o->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)o)->fst;s;s=s->nxt) WnpSetBackfaceColor(s,color);
      break;
    case DSP_TRG:
      if (((TRGObj *)o)->bckclr!=color) {
	WnpSetModifiedFlag(o);
	((TRGObj *)o)->bckclr=color;
      }
      break;
    }
  }
}

void WnpSetObjectLW(DspObj *o,int lw)
{
  DspObj *s;
  
  if (o) {
    switch (o->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)o)->fst;s;s=s->nxt) WnpSetObjectLW(s,lw);
      break;
    case DSP_CSO:
      if (((CSObj *)o)->lw!=lw) {
	WnpSetModifiedFlag(o);
	((CSObj *)o)->lw=lw;
      }
      break;
    case DSP_TRG:
      if (((TRGObj *)o)->lw!=lw) {
	WnpSetModifiedFlag(o);
	((TRGObj *)o)->lw=lw;
      }	
      break;
    }
  }
}

void WnpSetObjectMDST(DspObj *o,int mdstype)
{
  DspObj *s;
  
  if (o) {
    switch (o->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)o)->fst;s;s=s->nxt) WnpSetObjectMDST(s,mdstype);
      break;
    case DSP_CSO:
      if (((CSObj *)o)->mds_type!=mdstype) {
	WnpSetModifiedFlag(o);
	((CSObj *)o)->mds_type=mdstype;
      }
      break;
    }
  }
}

void WnpSetDisplayStyle(DspObj *obj,int style)
{
  DspObj *s;

  if (obj) {
    switch (obj->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)obj)->fst;s;s=s->nxt) WnpSetDisplayStyle(s,style);
      break;
    case DSP_TRG:
      if (((TRGObj *)obj)->style!=style) {
	WnpSetModifiedFlag(obj);
	((TRGObj *)obj)->style=style;
      }
      break;
    }
  }
}

void WnpSetTransparency(DspObj *obj,int level)
{
  DspObj *s;
  TRGObj *t;

  if (obj) {
    switch (obj->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)obj)->fst;s;s=s->nxt) WnpSetTransparency(s,level);
      break;
    case DSP_TRG:
      t=(TRGObj *)obj;
      if (t->stipple) {
	free(t->stipple);
      }
      t->stipple=WnpMakePTStipple(level);
      t->pat=level;
      WnpSetModifiedFlag(obj);
      break;
    }
  }
}

void WnpRemoveNOEs(DspObj *obj)
{
  DspObj *s;
  MdlObj *m;

  if (obj) {
    switch (obj->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)obj)->fst;s;s=s->nxt) WnpRemoveNOEs(s);
      break;
    case DSP_MDL:
      m=(MdlObj *)obj;
      if (m->fstNOE->nxt) {
	WnpSetRebuildFlag(obj);
	m->fstNOE->nxt=WnpDeleteLst(m->fstNOE->nxt);
      }
      break;
    }
  }
}

void WnpSetNOEAttributes(DspObj *obj,int label,int style,int width)
{
  DspObj *s;
  MdlObj *m;

  if (obj) {
    switch (obj->type) {
    case DSP_GLBL:
      for (s=((GlblObj *)obj)->fst;s;s=s->nxt) WnpSetNOEAttributes(s,label,style,width);
      break;
    case DSP_MDL:
      m=(MdlObj *)obj;
      if (m->fstNOE->nxt) WnpSetModifiedFlag(obj);
      m->NOElabt=label;
      m->NOEstyle=style;
      m->NOEwidth=width;
      break;
    }
  }
}

DspObj *WnpDeleteDspObj(DspObj *obj)
{
  DspObj *r;

  if (obj) {
    WnpSetModifiedFlag(obj);
    switch (obj->type) {                         /* delete data, depending on type */
    case DSP_GLBL:
      WnpDeleteGlblObj((GlblObj *)obj);
      break;
    case DSP_MDL:
      WnpDeleteMdlObj((MdlObj *)obj);
      break;
    case DSP_CSO:
      WnpDeleteCSObj((CSObj *)obj);
      break;
    case DSP_CNT:
      break;
    case DSP_TRG:
      WnpDeleteTRGObj((TRGObj *)obj);
      break;
    case DSP_TRACE:
      break;
    }
  }else r=((DspObj *)NULL);
  return(r);
}

void WnpUnlinkDspObj(DspObj *obj)
{
  GlblObj *u;

  u=obj->up; 
  if (u&&u->type==DSP_GLBL) {
    if (u->fst==obj) u->fst=obj->nxt;
    if (u->lst==obj) u->lst=obj->prv;
    obj->up=(GlblObj *)NULL;
  }
  if (obj->prv) obj->prv->nxt=obj->nxt;
  if (obj->nxt) obj->nxt->prv=obj->prv;
  obj->nxt=(DspObj *)NULL; 
  obj->prv=(DspObj *)NULL; 
}

DspObj *WnpFirstObj(DspObj *r,int types)
{
  DspObj *o,*s;

  o=(DspObj *)NULL;
  if (r) {
    if (r->type==DSP_GLBL) {
      for (s=((GlblObj *)r)->fst;!o&&s;s=s->nxt) o=WnpFirstObj(s,types);
      if (!o) if (r->type&types) o=r;
    }else if (r->type&types) o=r;
    else o=(DspObj *)NULL;
  }
  return(o);
}

DspObj *WnpNextObj(DspObj *root,DspObj *now,int obj_type)
{
/* recursive (DFS) find next object of given type in subtree root */

  DspObj *n,*s;

  n=now;
  if (!n) {
    s=(DspObj*)NULL;
  }else{
    for (s=(DspObj *)NULL,n=n->nxt;!s && n;n=n->nxt) {
      if (n->type==DSP_GLBL) s=WnpFirstObj(n,obj_type);
      if (!s) if (n->type&obj_type) s=n;
    }
    if (!s) {
      n=now;
      while (n->up && (DspObj *)n->up != root && !s) {
	n=(DspObj *)n->up;
	if (n->type&obj_type) s=n;
	else s=WnpNextObj(root,n,obj_type);
      }
    }
  }
  return(s);
}

void WnpSetPivot(GlblObj *g,float *piv)
{
  int i,j;

  if (g&&g->type==DSP_GLBL) {
    for (j=0;j<3;j++) {
      g->tra[j]+=g->pivot[j]-piv[j];
      for (i=0;i<3;i++) g->tra[j]+=(piv[i]-g->pivot[i])*g->rot[i][j];
    }
    for (j=0;j<3;j++) g->pivot[j]=piv[j];
  }
}

void WnpSetPivotAtom(GlblObj *o,WnpAtom *a)
{
  GlblObj *oa;
  float x[3],y[3];
  int i,j;

  if (a&&o&&o->type==DSP_GLBL) {
    oa=a->up->up;
    if (oa!=o) {
      for (i=0;i<3;i++) {
	x[i]=oa->mvt[3][i]-o->mvt[3][i];
	for (j=0;j<3;j++) x[i]+=a->x[j]*oa->mvt[j][i];
      }
      for (i=0;i<3;i++) 
	for (y[i]=0,j=0;j<3;j++) y[i]+=x[j]*o->mvt[i][j];
      WnpSetPivot(o,y);
    }else WnpSetPivot(o,a->x);
  }
}

void WnpGetRTP(WnpMatrix m,DspObj *o)
{
  GlblObj *g;
  int i,j;

  g=AssocGlbl(o);
  if (g) {
    for (i=0;i<3;i++) {
      for (j=0;j<3;j++) m[i][j]=g->rot[i][j];
    }
    for (j=0;j<3;j++) {
      m[j][3]=g->tra[j];
      m[3][j]=g->pivot[j];
    }
    m[3][3]=1.0;
  }
}

void WnpSetRTP(WnpMatrix m,DspObj *o)
{
  GlblObj *g;
  int i,j;

  g=AssocGlbl(o);
  if (g) {
    for (i=0;i<3;i++) {
      for (j=0;j<3;j++) g->rot[i][j]=m[i][j];
    }
    for (j=0;j<3;j++) {
      g->rot[3][j]=0.0; g->rot[j][3]=0.0;
      g->tra[j]=m[j][3];
      g->pivot[j]=m[3][j];
    }
    g->rot[3][3]=1.0;
    WnpCleanMVT(g->rot);
    WnpSetModifiedFlag(o);
    WnpSetMVTFlag(FALSE);
  }
}

void WnpGetMVT(WnpMatrix m,DspObj *o)
{
  GlblObj *g;
  int i,j;

  g=AssocGlbl(o);
  if (g) {
    for (i=0;i<4;i++) for (j=0;j<4;j++) m[i][j]=g->mvt[i][j];
    WnpViewShift(m);
  }else{
    for (i=0;i<4;i++) for (j=0;j<4;j++) m[i][j]=0.0;
    for (i=0;i<4;i++) m[i][i]=1.0;
  }
}

void WnpResetMVT(DspObj *o)
{
  GlblObj *g;
  int i,j;

  g=AssocGlbl(o);
  if (g) {
    WnpSetModifiedFlag(o);
    for (i=0;i<4;i++) for (j=0;j<4;j++) g->rot[i][j]=0.0;
    for (i=0;i<4;i++) g->rot[i][i]=1.0;
    for (i=0;i<3;i++) g->tra[i]=0.0;
    WnpSetMVTFlag(FALSE);
  }  
}

GlblObj *WnpAssocGlbl(DspObj *o) 
{
  return(o?(o->type==DSP_GLBL?(GlblObj *)o:o->up):(GlblObj *)NULL);
}

GlblObj *WnpMakeGlblObj(void)
{
  GlblObj *r;
  int i,j;

  r=(GlblObj *)malloc(sizeof(GlblObj));
  if (r) {
    r->type=DSP_GLBL;
    r->up=(GlblObj *)NULL;
    r->prv=(DspObj *)NULL;
    r->nxt=(DspObj *)NULL;
    r->fst=(DspObj *)NULL;
    r->lst=(DspObj *)NULL;
    r->flags=0;
    memset(&(r->cross),0,8);
    for (i=0;i<3;i++) {
      r->pivot[i]=0;
      r->tra[i]=0;
    }
    for (i=0;i<4;i++) for(j=0;j<4;j++) r->rot[i][j]=0.0;
    for (i=0;i<4;i++) r->rot[i][i]=1.0;
    for (i=0;i<4;i++) for(j=0;j<4;j++) r->mvt[i][j]=r->rot[i][j];
  }
  return(r);
}

GlblObj *WnpDeleteGlblObj(GlblObj *obj)
{
  if (obj&&obj->type==DSP_GLBL) {
    if (obj==AltObj) AltObj=(GlblObj *)NULL;
    WnpUnlinkDspObj((DspObj *)obj);
    while (obj->fst) WnpDeleteDspObj(obj->fst);
    if (obj==Root) Root=(GlblObj *)NULL;
    free(obj);
  }
  return((GlblObj *)NULL);
}

MdlObj *WnpMakeMdlObj(void)
{
  MdlObj *r;
  WnpAtom *a;
  WnpBond *b;
  WnpNOE  *n;
  WnpDots *d;

  r=(MdlObj *)malloc(sizeof(MdlObj));
  a=WnpMakeAtom(); b=WnpMakeBond(); n=WnpMakeNOE();
  d=WnpMakeDots();
  if (r&&a&&b&&n&&d) {
    r->type=DSP_MDL;
    r->up=(GlblObj *)NULL;
    r->prv=(DspObj *)NULL;
    r->nxt=(DspObj *)NULL;
    r->flags=0;
    memset(&(r->cross),0,8);
    r->fstAtom=a;
    r->fstBond=b;
    r->fstNOE=n;
    r->fstDot=d;
    r->NOEstyle=(WNP_NOE_DOTTED|WNP_NOE_BROKEN);
    r->NOElabt=(WNP_NOE_LAB_NONE);
    r->NOEwidth=1;
  }else{
    WnpDeleteAtom(a);
    WnpDeleteBond(b);
    WnpDeleteLstEntry((WnpLst *)n);
    WnpDeleteLstEntry((WnpLst *)d);
    if (r) free(r); 
    r=(MdlObj *)NULL;
  }
  return(r);
}

void WnpTransformMdl(MdlObj *obj,WnpMatrix mmv)
{
  WnpAtom *a;
  int i,j;
  float n[3];

  for (a=obj->fstAtom->nxt;a;a=a->nxt) {
    for (j=0;j<3;j++) for (n[j]=mmv[3][j],i=0;i<3;i++) n[j]+=a->x[i]*mmv[i][j];
    for (j=0;j<3;j++) a->x[j]=n[j];
    WnpTransformDots(WnpFindDots(a),mmv);
  }
}

void WnpTransformDots(WnpDots *d,WnpMatrix mvt)
{
  float x[3],*p;
  int i,j,k;

  if (d&&d->xyz) {
    for (p=d->xyz,k=0;k<d->n;k++,p+=4) {
      for (i=0;i<3;i++) for (x[i]=mvt[3][i],j=0;j<3;j++) x[i]+=p[j]*mvt[j][i];
      for (i=0;i<3;i++) p[i]=x[i];
    }
  }
}

void WnpTransformCSO(CSObj *o,WnpMatrix mvt)
{
  WnpLst *l;
  WnpMDS *mds;
  WnpTOBJ *tob;
  float x[3],*p;
  int i,j,k;

  for (l=o->fst;l;l=l->nxt) {
    switch (l->type) {
    case WNP_LST_MDS:
      mds=(WnpMDS *)l;
      if (mds->x) {
	for (p=mds->x,k=0;k<mds->n;k++,p+=3) {
	  for (i=0;i<3;i++) for (x[i]=mvt[3][i],j=0;j<3;j++) x[i]+=p[j]*mvt[j][i];
	  for (i=0;i<3;i++) p[i]=x[i];
	}
      }
      break;
    case WNP_LST_TOBJ:
      tob=(WnpTOBJ *)l;
      if (tob->x) {
	for (p=tob->x,i=0;i<3;i++) for (x[i]=mvt[3][i],j=0;j<3;j++) x[i]+=p[j]*mvt[j][i];
	for (i=0;i<3;i++) p[i]=x[i];
      }
      break;
    }
  }
}

void WnpTransformTRG(TRGObj *o,WnpMatrix mvt)
{
  WnpLst *l;
  WnpTRG *trg;
  float x[3],*p;
  int i,j,k;

  for (l=o->fst;l;l=l->nxt) {
    switch (l->type) {
    case WNP_LST_TRG:
      trg=(WnpTRG *)l;
      if (trg->v) {
	for (p=trg->v,k=0;k<trg->nv;k++,p+=3) {
	  /* position */
	  for (i=0;i<3;i++) for (x[i]=mvt[3][i],j=0;j<3;j++) x[i]+=p[j]*mvt[j][i];
	  for (i=0;i<3;i++) p[i]=x[i];
	  /* normal */
	  p+=3;
	  for (i=0;i<3;i++) for (x[i]=0,j=0;j<3;j++) x[i]+=p[j]*mvt[j][i];
	  for (i=0;i<3;i++) p[i]=x[i];
	}
      }
      break;
    }
  }
}

MdlObj *WnpDeleteMdlObj(MdlObj *obj)
{
  WnpAtom *a;
  WnpBond *b;

  if (obj&&obj->type==DSP_MDL) {
    WnpUnlinkDspObj((DspObj *)obj);
    if (obj->fstBond) {
      while(b=obj->fstBond) {
	obj->fstBond=b->nxt;
	WnpDeleteBond(b);
      }
    }
    if (obj->fstAtom) {
      for (a=obj->fstAtom->nxt;a;a=a->nxt) a->flags|=ATM_FLG_REMOVE;
      wnpDeletions=TRUE; WnpRebuildInterLists();
      while(a=obj->fstAtom) {
	obj->fstAtom=a->nxt;
	WnpDeleteAtom(a);
      }
    }
    obj->fstNOE=(WnpNOE *)WnpDeleteLst((WnpLst *)obj->fstNOE);
  }
  return((MdlObj *)NULL);
}

CSObj *WnpMakeCSObj(void)
{
  CSObj *r;

  r=malloc(sizeof(CSObj));
  if (r) {
    r->type=DSP_CSO;
    r->up=(GlblObj *)NULL;
    r->prv=(DspObj *)NULL;
    r->nxt=(DspObj *)NULL;
    memset(&(r->cross),0,8);
    r->flags=0;
    r->lw=1;
    r->mds_type=WNP_CSO_LINES;
    r->fst=(WnpLst *)NULL;
  }
  return(r);
}

TRGObj *WnpMakeTRGObj(void)
{
  TRGObj *r;

  r=malloc(sizeof(TRGObj));
  if (r) {
    r->type=DSP_TRG;
    r->up=(GlblObj *)NULL;
    r->prv=(DspObj *)NULL;
    r->nxt=(DspObj *)NULL;
    memset(&(r->cross),0,8);
    r->flags=0;
    r->lw=1; r->color=3;
    r->style=WNP_TRG_RENDERED;
    r->stipple=(int *)NULL;
    r->pat=0; r->bckclr=0;
    r->fst=(WnpLst *)NULL;
  }
  return(r);
}

void WnpAddTRG(TRGObj *obj,WnpTRG *trg)
{
  if (obj&&obj->type==DSP_TRG) {
    if (trg&&trg->type==WNP_LST_TRG) {
      trg->up=obj;
      trg->nxt=obj->fst;
      if (trg->nxt) trg->nxt->prv=(WnpLst *)trg;
      trg->prv=(WnpLst *)NULL;
      obj->fst=(WnpLst *)trg;
      WnpSetModifiedFlag((DspObj *)obj);
    }
  }
}

void WnpAddMDS(CSObj *cso,WnpMDS *mds)
{
  if (cso&&cso->type==DSP_CSO) {
    if (mds&&mds->type==WNP_LST_MDS) {
      mds->up=cso;
      mds->nxt=cso->fst;
      if (mds->nxt) mds->nxt->prv=(WnpLst *)mds;
      mds->prv=(WnpLst *)NULL;
      cso->fst=(WnpLst *)mds;
      WnpSetModifiedFlag((DspObj *)cso);
    }
  }
}

void WnpAddTOBJ(CSObj *cso,WnpTOBJ *tob)
{
  if (cso&&cso->type==DSP_CSO) {
    if (tob&&tob->type==WNP_LST_TOBJ) {
      tob->up=cso;
      tob->nxt=cso->fst;
      if (tob->nxt) tob->nxt->prv=(WnpLst *)tob;
      tob->prv=(WnpLst *)NULL;
      cso->fst=(WnpLst *)tob;
      WnpSetModifiedFlag((DspObj *)cso);
    }
  }
}

CSObj *WnpDeleteCSObj(CSObj *obj)
{
  if (obj&&obj->type==DSP_CSO) {
    WnpUnlinkDspObj((DspObj *)obj);
    WnpDeleteLst(obj->fst);
  }
  return((CSObj *)NULL);
}

TRGObj *WnpDeleteTRGObj(TRGObj *obj)
{
  if (obj&&obj->type==DSP_TRG) {
    WnpUnlinkDspObj((DspObj *)obj);
    WnpDeleteLst(obj->fst);
    if (obj->stipple) free(obj->stipple);
  }
  return((TRGObj *)NULL);
}

WnpAtom *WnpMakeAtom(void)
{
  WnpAtom *r;

  r=(WnpAtom *)malloc(sizeof(WnpAtom));
  if (r) {
    r->up=(MdlObj *)NULL;
    r->nxt=(WnpAtom *)NULL;
    memset(&(r->cross),0,8);
    r->sph=(WnpAtom *)NULL;
    r->lab=(WnpAtom *)NULL;
    r->mid=(WnpAtom *)NULL;
    r->flags=0;
    r->xtra=(WnpLst *)NULL;
    r->x[0]=0; r->x[1]=0; r->x[2]=0;
    r->r=0;
    r->color=1; r->lcolor=defaultAtomLabelColor;
    r->label=(char *)NULL;
  }
  return(r);
}

WnpAtom *WnpFirstAtom(DspObj *r)
{
  WnpAtom *a;
  MdlObj *mdl;

  ForMdl(mdl,r) if (mdl&&mdl->fstAtom&&mdl->fstAtom->nxt) break;
  if (mdl) a=mdl->fstAtom->nxt;
  else a=(WnpAtom *)NULL;
  return(a);
}

WnpAtom *WnpNextAtom(DspObj *r,WnpAtom *a)
{
  WnpAtom *o;
  MdlObj *mdl;

  if (a) {
    if (a->nxt) o=a->nxt;
    else {
      for(mdl=WnpNextMdl(r,a->up);mdl;mdl=WnpNextMdl(r,mdl)) 
	if (mdl&&mdl->fstAtom&&mdl->fstAtom->nxt) break;
      if (mdl) o=mdl->fstAtom->nxt;
      else o=(WnpAtom *)NULL;
    }
  }else o=(WnpAtom *)NULL;
  return(o);
}

void WnpAddAtomToMdl(WnpAtom *a,MdlObj *mdl)
{
  if (a&&mdl&&mdl->type==DSP_MDL) {
    a->up=mdl;                                  /* atom is added at head of list */
    a->nxt=mdl->fstAtom->nxt;
    mdl->fstAtom->nxt=a;
    WnpSetRebuildFlag((DspObj *)mdl);
  }
}

void WnpSetAtomMark(WnpAtom *a,int mark)
{
  int mask;

  mask=a->flags&(ATM_FLG_MARK|ATM_FLG_DSP);
  if (mark) {
    if (mask==ATM_FLG_DSP) {
      a->flags|=ATM_FLG_MARK;
      WnpSetRebuildFlag((DspObj *)a->up);
    }else if (!mask) a->flags|=ATM_FLG_MARK;
  }else {
    if (mask==(ATM_FLG_MARK|ATM_FLG_DSP)) WnpSetRebuildFlag((DspObj *)a->up);
    a->flags&=~(ATM_FLG_MARK);
  }
}

void WnpClearAllAtomMarks(void)
{
  WnpAtom *a;

  ForAtoms(a,Root) WnpSetAtomMark(a,FALSE);
}

void WnpRemoveAtom(WnpAtom *a)
{
  /* remove atom from display ... */

  a->flags|=ATM_FLG_REMOVE;
  WnpSetRebuildFlag((DspObj *)a->up);
  wnpDeletions=TRUE;
}

void WnpRemoveAtomLabel(WnpAtom *a)
{
  if (a) {
    if (a->label) {
      free(a->label);
      if (AllFlags(a->flags,ATM_FLG_LDSP|ATM_FLG_DSP)) WnpSetRebuildFlag((DspObj *)a->up);
      a->label=(char *)NULL;
    }
  }
  
}

void WnpSetAtomLabel(WnpAtom *a,char *text)
{
  if (a) {
    if (a->label) free(a->label);
    a->label=text;
    if (AllFlags(a->flags,ATM_FLG_LDSP|ATM_FLG_DSP)) WnpSetRebuildFlag((DspObj *)a->up);
  }
}

void WnpSetAtomLabelColor(WnpAtom *a,int color)
{
  if (a) {
    if (a->lcolor!=color &&
	AllFlags(a->flags,ATM_FLG_LDSP|ATM_FLG_DSP)) WnpSetRebuildFlag((DspObj *)a->up);
    a->lcolor=color;
  }
}

void WnpReplaceAtom(WnpAtom *old,WnpAtom *new)
{
  if (new) {
    if (old) {
      old->color=new->color;
      if (new->flags&ATM_FLG_DSP) old->flags|=ATM_FLG_DSP;
      else old->flags&=~ATM_FLG_DSP;
      old->x[0]=new->x[0]; old->x[1]=new->x[1]; old->x[2]=new->x[2];
      old->r=new->r;
      if (new->flags&ATM_FLG_DSP) WnpSetModifiedFlag((DspObj *)old->up);
    }
    WnpDeleteAtom(new);  /* note: new atom is not in mon-pair list... */
  }
}

WnpAtom *WnpDeleteAtom(WnpAtom *a)
{
  WnpAtom *r;

  /* delete atom , return pointer to next atom ... */

  if (a) {
    r=a->nxt;
    while (a->xtra) a->xtra=WnpDeleteLst(a->xtra);
    if (a->label) free(a->label);
    free(a);
  }else r=(WnpAtom *)NULL;
  return(r);
}

void WnpNewAtomXYZ(WnpAtom *a,float *x)
{
  int i;

  for (i=0;i<3;i++) a->x[i]=x[i];
  WnpSetModifiedFlag((DspObj *)a->up);
}

void WnpNewAtomFlags(WnpAtom *a,int flags,int state)
{
  int old,diff;
  
  old=a->flags;
  if (state) a->flags|=flags;
  else a->flags&=~flags;
  diff=(old^a->flags);
  if (diff) WnpSetRebuildFlag((DspObj *)a->up);
}

void WnpNewAtomColor(WnpAtom *a,int color)
{
  if ((a->color!=color)&&a->flags&ATM_FLG_DSP) WnpSetModifiedFlag((DspObj *)a->up);
  a->color=color;
}

void WnpNewAtomRadius(WnpAtom *a,float r)
{
  if (a->r!=r&&AllFlags(a->flags,ATM_FLG_DSP|ATM_FLG_SPHERE)) WnpSetModifiedFlag((DspObj *)a->up);
  a->r=r;
}

void WnpScaleAtomRadius(WnpAtom *a,float s)
{
  if (s!=1.0&&AllFlags(a->flags,ATM_FLG_DSP|ATM_FLG_SPHERE)) WnpSetModifiedFlag((DspObj *)a->up);
  a->r*=clamp(s,0,1.e4);
}

void WnpCopyDots(WnpAtom *trg,WnpAtom *src)
{
  WnpDots *d,*o;

  if (trg&&src) {
    WnpDeleteDots(trg);
    WnpSetRebuildFlag((DspObj *)trg->up);
    o=WnpFindDots(src);
    if (o) {
      d=WnpMakeDots();
      if (d) {
	d->up=trg;
	d->nxt=trg->xtra;
	trg->xtra=(WnpLst *)d;
	if (d->nxt) d->nxt->prv=trg->xtra;
	d->n=o->n;
	if (d->n>0) {
	  d->xyz=malloc(4*d->n*sizeof(float));
	  if (d->xyz) memcpy(d->xyz,o->xyz,4*d->n*sizeof(float));
	  else d->n=0;
	}
	if (d->n>0) {
	  d->style=o->style;
	  d->color=o->color;
	  d->dsp=o->dsp;
	}else WnpDeleteDots(trg);
      }
    }
  }
}

void WnpReplaceDots(WnpAtom *a,int n,float *xyz)
{
  WnpDots *d;

  if (a) {
    d=WnpFindDots(a);
    if (d) {
      if (d->xyz) free(d->xyz); 
      d->xyz=(float *)NULL;
      d->n=0;
    }else{
      d=WnpMakeDots();
      if (d) {
	d->up=a;
	d->nxt=a->xtra;
	a->xtra=(WnpLst *)d;
	if (d->nxt) d->nxt->prv=a->xtra;
      }
    }
    if (d) {
      d->xyz=xyz;
      d->n=(d->xyz ? n : 0);
      if (d->n > 0 && (d->dsp==DOT_ON||
	  ((a->flags&ATM_FLG_DSP)&&(d->dsp==DOT_IFATOM)))) WnpSetRebuildFlag((DspObj *)a->up);
    }else{
      free(xyz);
    }
  }else free(xyz);
}

void WnpDeleteDots(WnpAtom *a)
{
  WnpDots *d;
  
  d=WnpFindDots(a);
  if (d) {
    if (d->dsp==DOT_ON||
	((a->flags&ATM_FLG_DSP)&&(d->dsp==DOT_IFATOM))) WnpSetRebuildFlag((DspObj *)a->up);
    if (d==(WnpDots *)a->xtra) a->xtra=WnpDeleteLstEntry((WnpLst *)d);
    else WnpDeleteLstEntry((WnpLst *)d);
  }
}

void WnpDotsStyle(WnpAtom *a,int  style)
{
  WnpDots *d;

  d=WnpFindDots(a);
  if (d&&d->style!=style) {
    d->style=style;
    if (d->dsp==DOT_ON||
	((a->flags&ATM_FLG_DSP)&&(d->dsp==DOT_IFATOM))) WnpSetModifiedFlag((DspObj *)a->up);
  }
}

void WnpColorDots(WnpAtom *a,int color)
{
  WnpDots *d;

  d=WnpFindDots(a);
  if (d&&d->color!=color) {
    d->color=color;
    if (d->dsp==DOT_ON||
	((a->flags&ATM_FLG_DSP)&&(d->dsp==DOT_IFATOM))) WnpSetModifiedFlag((DspObj *)a->up);
  }
}

void WnpDisplayDots(WnpAtom *a,int dsp)
{
  WnpDots *d;

  d=WnpFindDots(a);
  if (d&&d->dsp!=dsp) {
    d->dsp=dsp;
    WnpSetRebuildFlag((DspObj *)a->up);
  }
}

WnpDots *WnpFindDots(WnpAtom *a)
{
  WnpLst *lst;

  if (a) for (lst=a->xtra;lst;lst=lst->nxt) if (lst->type==WNP_LST_DOTS) break;
  else lst=(WnpLst *)NULL;
  return((WnpDots *)lst);
}

int WnpDefaultMonColor(void)
{
  return(wnpMonColor);
}

int WnpDefaultMonLabel(void)
{
  return(wnpMonLabel);
}

int WnpDefaultMonWidth(void)
{
  return(wnpMonWidth);
}

int WnpDefaultMonDsp(void)
{
  return(wnpMonDsp);
}


void WnpAddMonitorPair(WnpAtom *f,WnpAtom *t)
{
  WnpMonPair *m;

  if (f&&t) {
    m=malloc(sizeof(WnpMonPair));
    if (m) {
      m->f=f;
      m->t=t;
      m->type=WNP_LST_MON;
      m->dsp=wnpMonDsp;
      m->color=wnpMonColor;
      m->label=wnpMonLabel;
      m->width=wnpMonWidth;
      m->nxt=(WnpLst *)monPairs;
      if (m->nxt) m->nxt->prv=(WnpLst *)m;
      m->prv=(WnpLst *)NULL;
      monPairs=m;
      if (f->flags&t->flags&ATM_FLG_DSP) WnpSetThreeDRedrawFlag(TRUE);
    }
  }
}

void WnpClearMonitorPairs(void)
{
  if (monPairs) WnpSetThreeDRedrawFlag(TRUE); 
  monPairs=(WnpMonPair *)WnpDeleteLst((WnpLst *)monPairs);
}

void WnpMonitorLabel(int state)
{
  WnpMonPair *m;

  if (wnpMonLabel!=state) {
    if (monPairs) WnpSetThreeDRedrawFlag(TRUE);
    wnpMonLabel=state;
    for (m=monPairs;m;m=(WnpMonPair *)m->nxt) if (m->type==WNP_LST_MON) m->label=wnpMonLabel;
  }
}

void WnpMonitorColor(int color)
{
  WnpMonPair *m;

  if (wnpMonColor!=color) {
    if (monPairs) WnpSetThreeDRedrawFlag(TRUE);
    wnpMonColor=color;
    for (m=monPairs;m;m=(WnpMonPair *)m->nxt) if (m->type==WNP_LST_MON) m->color=wnpMonColor;
  }
}

void WnpMonitorWidth(int width)
{
  WnpMonPair *m;

  if (wnpMonWidth!=width) {
    if (monPairs) WnpSetThreeDRedrawFlag(TRUE);
    wnpMonWidth=width;
    for (m=monPairs;m;m=(WnpMonPair *)m->nxt) if (m->type==WNP_LST_MON) m->width=wnpMonWidth;
  }
}

void WnpMonitorStatus(int status)
{
  WnpMonPair *m;

  if (wnpMonDsp!=status) {
    if (monPairs) WnpSetThreeDRedrawFlag(TRUE);
    wnpMonDsp=status;
    for (m=monPairs;m;m=(WnpMonPair *)m->nxt) if (m->type==WNP_LST_MON) m->dsp=wnpMonDsp;
  }
}

void WnpAddNOE(WnpNOE *noe)
{
  MdlObj *mdl;

  if (noe&&noe->type==WNP_LST_NOE) {
    if (noe->f->up==noe->t->up) {
      mdl=noe->f->up;
      if (mdl&&mdl->type==DSP_MDL) {
	noe->nxt=mdl->fstNOE->nxt;
	if (noe->nxt) noe->nxt->prv=(WnpLst *)noe;
	mdl->fstNOE->nxt=(WnpLst *)noe;
	noe->prv=(WnpLst *)mdl->fstNOE;
	if (noe->f->flags&noe->t->flags&ATM_FLG_DSP) WnpSetRebuildFlag((DspObj *)mdl);
      }
    }
  }
}

WnpNOE *WnpMakeNOE(void)
{
  WnpNOE *r;

  r=malloc(sizeof(WnpNOE));
  if (r) {
    r->prv=0;
    r->nxt=0;
    r->D=0;
    r->viol=0;
    r->color=2;
    r->f=(WnpAtom *)NULL;
    r->t=(WnpAtom *)NULL;
    r->type=WNP_LST_NOE;
    r->dsp=(WnpNOE *)NULL;
  }
  return(r);
}

WnpLst *WnpDeleteLst(WnpLst *lst)
{
  while (lst) lst=WnpDeleteLstEntry(lst);
  return(lst);
}

WnpLst *WnpDeleteLstEntry(WnpLst *lst)
{
  WnpLst *r;

  /* delete list entry, return pointer to next entry */

  if (lst) {
    r=lst->nxt;
    if (lst->prv) lst->prv->nxt=lst->nxt;
    if (lst->nxt) lst->nxt->prv=lst->prv;
    switch (lst->type) {
    case WNP_LST_DOTS:
      if (((WnpDots *)lst)->xyz) free(((WnpDots *)lst)->xyz);
      break;
    case WNP_LST_MDS:
      if (((WnpMDS *)lst)->x) free(((WnpMDS *)lst)->x);
      break;
    case WNP_LST_TOBJ:
      if (((WnpTOBJ *)lst)->text) free(((WnpTOBJ *)lst)->text);
      break;
    case WNP_LST_TRG:
      if (((WnpTRG *)lst)->v) free(((WnpTRG *)lst)->v);
      if (((WnpTRG *)lst)->t) free(((WnpTRG *)lst)->t);
      break;
    case WNP_LST_MON:
      if (monPairs==(WnpMonPair *)lst) monPairs=(WnpMonPair *)lst->nxt;
      break;
    }
    free(lst);
  }else r=(WnpLst *)NULL;
  return(r);
}

WnpBond *WnpMakeBond(void)
{
  WnpBond *r;

  r=(WnpBond *)malloc(sizeof(WnpBond));
  if (r) {
    r->up=(MdlObj *)NULL;
    r->nxt=(WnpBond *)NULL;
    r->flags=BND_FLG_LDSP;
    r->f=(WnpAtom *)NULL;
    r->t=(WnpAtom *)NULL;
    r->dsp=(WnpBond *)NULL;
    r->lab=(WnpBond *)NULL;
    r->label=(char *)NULL;
    r->lcolor=3;
    r->lw=1;
    r->r=0.1;
  }
  return(r);
}    

void WnpAddBond(WnpBond *b)
{
  MdlObj *mdl;

  if (b) {
    b->lw=defaultBondLW;
    if (b->f&&b->t) {
      if (b->f->up==b->t->up) {                /* intra-MDL bond */
	mdl=b->f->up;
	b->up=mdl;
	b->nxt=mdl->fstBond->nxt;
	mdl->fstBond->nxt=b;
	WnpSetRebuildFlag((DspObj *)mdl);
      }else {                                  /* inter-MDL bond */
	if (!interBonds) interBonds=WnpMakeBond();
	if (interBonds) {
	  b->nxt=interBonds->nxt;
	  interBonds->nxt=b;
	}
      }
    }
  }
}

void WnpNewBondFlags(WnpBond *b,int flags,int state)
{
  int diff,old;

  old=b->flags;
  if (state) b->flags|=flags;
  else b->flags&=~flags;
  diff=(old^b->flags);
  if (diff) WnpSetRebuildFlag((DspObj *)b->up);
}

void WnpSetBondRadius(WnpBond *b,float r)
{
  if (r>0) {
    if (b->r!=r) {
      b->r=r;
      if ((b->f->flags&b->t->flags&ATM_FLG_DSP)
	  &&b->flags&BND_FLG_CYLINDER)  WnpSetModifiedFlag((DspObj *)b->up);
    }
  }else if (r!=-1.0) {
    b->r*=clamp(-r,0,1.e4);
    if ((b->f->flags&b->t->flags&ATM_FLG_DSP)
	&&b->flags&BND_FLG_CYLINDER)  WnpSetModifiedFlag((DspObj *)b->up);
  }
}

void WnpSetBondLW(WnpBond *b,int lw)
{
  if (b->lw!=clamp(lw,1,20)) {
    b->lw=clamp(lw,1,20);
    if ((b->f->flags&b->t->flags&ATM_FLG_DSP)) WnpSetRebuildFlag((DspObj *)b->up);
  }
}

void WnpSetDefaultAtomLabelColor(int color)
{
  defaultAtomLabelColor=color;
}

void WnpSetDefaultBondLW(int lw)
{
  defaultBondLW=clamp(lw,1,50);
}

void WnpSetBondLabel(WnpBond *b,char *text)
{
  if (b) {
    if (b->label) free(b->label);
    b->label=text;
    if ((b->f->flags&b->t->flags&ATM_FLG_DSP)
	     &&b->flags&BND_FLG_LDSP) WnpSetRebuildFlag((DspObj *)b->up);
  }
}

void WnpSetBondLabelColor(WnpBond *b,int color)
{
  if (b&&b->lcolor!=color&&b->label) {
    b->lcolor=color;
    if ((b->f->flags&b->t->flags&ATM_FLG_DSP)
	&&b->flags&BND_FLG_LDSP)  WnpSetModifiedFlag((DspObj *)b->up);
  }
}

void WnpRemoveBondLabel(WnpBond *b)
{
  if (b&&b->label) {
    free(b->label);
    b->label=(char *)NULL;
    if ((b->f->flags&b->t->flags&ATM_FLG_DSP)
	&&b->flags&BND_FLG_LDSP)  WnpSetRebuildFlag((DspObj *)b->up);
  }
}

void WnpRemoveBond(WnpBond *b)
{
  /* remove bond from display ... */

  b->flags|=BND_FLG_REMOVE;
  WnpSetRebuildFlag((DspObj *)b->up);
}

WnpBond *WnpDeleteBond(WnpBond *b)
{
  WnpBond *r;

  /* delete bond, return pointer to next bond in list */

  if (b) {
    r=b->nxt;
    WnpSetRebuildFlag((DspObj *)b->up);
    if (b->label) free(b->label);
    free(b);
  }else r=(WnpBond *)NULL;
  return(r);
}

WnpDots *WnpMakeDots(void)
{
  WnpDots *r;

  r=(WnpDots *)malloc(sizeof(WnpDots));
  if (r) {
    r->prv=(WnpLst *)NULL;
    r->nxt=(WnpLst *)NULL;
    r->dot=(WnpDots *)NULL;
    r->type=WNP_LST_DOTS;
    r->xyz=(float *)NULL;
    r->n=0;
    r->color=DOT_BYATOM;
    r->style=DOT_SMALL;
    r->dsp=DOT_IFATOM;
    r->up=(WnpAtom *)NULL;
  }
  return(r);
}

WnpTRG *WnpMakeTRG(void)
{
  WnpTRG *r;

  r=(WnpTRG *)malloc(sizeof(WnpTRG));
  if (r) {
    r->prv=(WnpLst *)NULL;
    r->nxt=(WnpLst *)NULL;
    r->type=WNP_LST_TRG;
    r->up=(TRGObj *)NULL;
    r->v=(float *)NULL;
    r->t=(int *)NULL;
    r->nv=0; r->nt=0;
  }
  return(r);
}

WnpMDS *WnpMakeMDS(void)
{
  WnpMDS *r;

  r=(WnpMDS *)malloc(sizeof(WnpMDS));
  if (r) {
    r->prv=(WnpLst *)NULL;
    r->nxt=(WnpLst *)NULL;
    r->type=WNP_LST_MDS;
    r->up=(CSObj *)NULL;
    r->x=(float *)NULL;
    r->color=3;
    r->n=0;
  }
  return(r);
}

WnpTOBJ *WnpMakeTOBJ(void)
{
  WnpTOBJ *r;

  r=(WnpTOBJ *)malloc(sizeof(WnpTOBJ));
  if (r) {
    r->prv=(WnpLst *)NULL;
    r->nxt=(WnpLst *)NULL;
    r->type=WNP_LST_TOBJ;
    r->up=(CSObj *)NULL;
    r->text=(char *)NULL;
    r->color=3;
  }
  return(r);
}

DspObj *MyObjPtr(cptr_t his)
{
  DspObj *r;

  r=myObjPtr(his,(DspObj *)Root);
  return(r);
}

void WnpSetRebuildFlag(DspObj *obj)
{
  if (obj&&!(obj->flags&OBJ_FLG_RBLD)) {
    obj->flags|=OBJ_FLG_RBLD;
    WnpSetModifiedFlag((DspObj *)obj->up);
  }else if (!obj) WnpSetThreeDRedrawFlag(TRUE);
}

void WnpSetModifiedFlag(DspObj *obj)
{
  if (obj&&!(obj->flags&OBJ_FLG_MOD)) {
    obj->flags|=OBJ_FLG_MOD;
    WnpSetModifiedFlag((DspObj *)obj->up);
  }else if (!obj) WnpSetThreeDRedrawFlag(TRUE);
}

void WnpRebuildInterLists(void)
{
  WnpLst *m;
  WnpBond *b,*n;
  WnpMonPair *p;

  if (wnpDeletions) {
    if (monPairs) {
      for (m=(WnpLst *)monPairs;m;) {
	if (m->type==WNP_LST_MON) {
	  p=(WnpMonPair *)m;
	  if ((p->f->flags|p->t->flags)&ATM_FLG_REMOVE) m=WnpDeleteLstEntry(m);
	  else m=m->nxt;
	}
      }
    }
    if (interBonds) {
      for (b=interBonds;n=b->nxt;) {
	if ((n->f->flags|n->t->flags)&ATM_FLG_REMOVE) {
	  b->nxt=n->nxt;
	  WnpDeleteBond(n);
	}else b=n;
      }
    }
  }
  wnpDeletions=FALSE;
}
  
#define MX_BND_TYPE 20

void WnpRebuildInterBondList(void)
{
  if (!interBonds) interBonds=WnpMakeBond();
  if (interBonds) interBonds->lab=(WnpBond *)NULL;
}

void WnpRebuildMdl(MdlObj *obj)
{
  WnpAtom *a,*n,*am,*ai,*ad,*amt,*ait;
  WnpBond *b,*m,*bh[MX_BND_TYPE],*bt[MX_BND_TYPE];
  WnpDots *dot;
  WnpNOE  *noe;
  int i,l,active;

  if (obj&&obj->type==DSP_MDL) {

    /* deal with pending deletions of atoms and bonds */

    for (b=obj->fstBond;m=b->nxt;) {
      if ((m->flags&BND_FLG_REMOVE)||(m->f->flags|m->t->flags)&ATM_FLG_REMOVE) {
	b->nxt=m->nxt;
	WnpDeleteBond(m);
      }else b=m;
    }
    WnpRebuildInterLists();      /* just in case... */
    for (a=obj->fstAtom;n=a->nxt;) {
      n->flags|=ATM_FLG_ISO;
      if (n->flags&ATM_FLG_REMOVE) {
	a->nxt=n->nxt;
	WnpDeleteAtom(n);
      }else a=n;
    }

    /* generate list of bonds, first cylinders, then lines, in ascending order of lw */

    obj->fstBond->lab=(WnpBond *)NULL;
    for (i=0;i<MX_BND_TYPE;i++) bh[i]=bt[i]=(WnpBond *)NULL;
    for (b=obj->fstBond->nxt;b;b=b->nxt) {
      if (b->f->flags&b->t->flags&ATM_FLG_DSP) {
	b->f->flags&=~ATM_FLG_ISO;
	b->t->flags&=~ATM_FLG_ISO;
	if (b->flags&BND_FLG_CYLINDER) l=0;
	else l=clamp(b->lw,1,MX_BND_TYPE-1);
	if (!bt[l])  bt[l]=b;
	b->dsp=bh[l];
	bh[l]=b;
	if (b->label) {
	  b->lab=obj->fstBond->lab;
	  obj->fstBond->lab=b;
	}
      }
    }
    obj->fstBond->dsp=(WnpBond *)NULL;
    for (i=MX_BND_TYPE;i>0;) {
      if (bt[--i]) {
	bt[i]->dsp=obj->fstBond->dsp;
	obj->fstBond->dsp=bh[i];
      }
    }

    /* generate list of atom labels, spheres and isolated atoms, dots... */

    obj->fstAtom->sph=(WnpAtom *)NULL;
    obj->fstAtom->lab=(WnpAtom *)NULL;
    obj->fstAtom->mid=(WnpAtom *)NULL;
    am=ai=ad=amt=ait=(WnpAtom *)NULL;
    obj->fstDot->dot=(WnpDots *)NULL;
    for (a=obj->fstAtom->nxt;a;a=a->nxt) {
      if (a->flags&ATM_FLG_DSP) {
	if (a->flags&ATM_FLG_SPHERE) {
	  a->sph=obj->fstAtom->sph;
	  obj->fstAtom->sph=a;
	}
	if ((a->flags&ATM_FLG_LDSP)&&a->label) {
	  a->lab=obj->fstAtom->lab;
	  obj->fstAtom->lab=a;
	}
	if (a->flags&ATM_FLG_MARK) {
	  if (!amt) amt=a;
	  a->mid=am;
	  am=a;
	}else if (a->flags&ATM_FLG_ISO) {
	  if (!ait) ait=a;
	  a->mid=ai;
	  ai=a;
	}else{
	  a->mid=ad;
	  ad=a;
	}
      }
      dot=WnpFindDots(a);
      if (dot&&(dot->dsp==DOT_ON||
		((a->flags&ATM_FLG_DSP)&&(dot->dsp==DOT_IFATOM)))
	  &&dot->n>0&&dot->xyz) {
	dot->dot=obj->fstDot->dot;
	obj->fstDot->dot=dot;
      }
    }
    obj->fstAtom->mid=(am ? am : (ai ? ai : ad));
    if (amt) amt->mid=(ai ? ai : ad);
    if (ait) ait->mid=ad;

    /* generate NOE display list */

    obj->fstNOE->dsp=(WnpNOE *)NULL;
    for (noe=(WnpNOE *)obj->fstNOE->nxt;noe;noe=(WnpNOE *)noe->nxt) {
      if (noe->f->flags&ATM_FLG_DSP&&noe->t->flags&ATM_FLG_DSP) {
	noe->dsp=obj->fstNOE->dsp;
	obj->fstNOE->dsp=noe;
      }
    }

    obj->flags&=~OBJ_FLG_RBLD;
    active=obj->fstNOE->dsp||obj->fstDot->dot||obj->fstBond->dsp||obj->fstBond->lab||
      obj->fstAtom->sph||obj->fstAtom->lab||obj->fstAtom->mid;
    if (active) obj->flags|=OBJ_FLG_ACTIVE;
    else obj->flags&=~OBJ_FLG_ACTIVE;
  }
}

/*------------------------------------------------------------------------*/

static DspObj *myObjPtr(cptr_t his,DspObj *root)
{
  DspObj *r,*s;

  r=(DspObj *)NULL;
  if (root) {
    if (!memcmp(&his,&(root->cross),8)) r=root;
    else if (root->type==DSP_GLBL) {
      for (s=((GlblObj *)root)->fst; !r&&s; s=s->nxt) r=myObjPtr(his,s);
    }
  }
  return(r);
}
