#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#ifndef mingw
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#endif

/*------------------------------------------------------------------------*/

/* Replacements for open and pclose functions */
/* PipeOpen finds shell in /bin, or in $PATH  */


#ifndef mingw
static pid_t *openpid = NULL;
static int maxopen;
#endif

FILE *WnpPipeOpen(char *cmd,char *type)
{

#ifdef mingw
  return(popen(cmd,type));
#else

  int	i,fd[2];
  pid_t	pid;
  FILE	*fp=(FILE *)NULL;

  if (*type=='r' || *type=='w') {
    if (!openpid) {            /* need to allocate memory for pids ... */
      maxopen=sysconf(_SC_OPEN_MAX);
      if (maxopen<=0) maxopen=512;
      openpid=(pid_t *)calloc(maxopen,sizeof(pid_t));
    }
    if (openpid) {
      if (!pipe(fd)) {

	pid=fork();

	if (pid==0) {          /* subprocess */
	  if (*type=='r') {
	    close(fd[0]);
	    if (fd[1] != STDOUT_FILENO) {
	      dup2(fd[1], STDOUT_FILENO);
	      close(fd[1]);
	    }
	  }else if (*type=='w') {
	    close(fd[1]);
	    if (fd[0] != STDIN_FILENO) {
	      dup2(fd[0], STDIN_FILENO);
	      close(fd[0]);
	    }
	  }
	  /* close open pipes in child */
	  for (i=0;i<maxopen;i++) if (openpid[i]>0) close(i);
	  /* execute shell from /bin, or whatever is found in path */
	  if (!access("/bin/sh",X_OK)) execl("/bin/sh","sh","-c",cmd,NULL);
	  else if (!access("/bin/sh.exe",X_OK)) execl("/bin/sh","sh","-c",cmd,NULL);
	  else execlp("sh","sh","-c",cmd,NULL);
	  exit(127);           /* here only if execl failed */

	}else if (pid>0){      /* parent */
	  if (*type=='r') {
	    close(fd[1]);
	    fp=fdopen(fd[0],type);
	  }else if (*type=='w') {
	    close(fd[0]);
	    fp=fdopen(fd[1],type);
	  }
	  if (fp) openpid[fileno(fp)]=pid;  /* store pid of child for use by PipeClose */

	}

      }
    }
  }
  return(fp);
#endif
}

int WnpPipeClose(FILE *fp)
{
#ifdef mingw
  return(pclose(fp));
#else

  int	fd,status=-1;
  pid_t	pid;

  if (openpid) {
    fd=fileno(fp);
    if ((fd=fileno(fp))>=0 && (pid=openpid[fd])!=0) {
      openpid[fd]=0;
      if (!fclose(fp)) {
	while (waitpid(pid, &status,0)<0) {
	  if (errno!=EINTR) {
	    status=-1;
	    break;
	  }
	}
      }
    }
  }
  return(status);
#endif
}
