#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include "wnp.h"
#include "wnpgl.h"

/* module to handle switching between OpenGL render and Plot mode */

static short magic=19937;		    /* plot file magic number */
static int renderMode = WNP_NORMAL_DRAW;
static FILE *Plot=(FILE *)NULL;

static int pltWidth=0;          /* what we have set for SGP */
static int pltLineWidth=1;      
static int pltPointSize=1;
static int vcount=0;
static int loopMode = 0;
static float firstX[3];

#define PlWrite(rec,size) fwrite((char *)(rec),1,(int)(size),Plot)

void WnpPlotThreeD(void)
{
  char *cmd;

  cmd=getenv("PLOTserver");
  if (cmd) {
    if (WnpPlotThreeDBox(cmd)) WnpRingBell();
    else WnpSendToCom("sorry, could not plot to %s.",cmd);
  }else{
    WnpSendToCom("sorry, environment variable PLOTserver is not defined.");
  }
}

void WnpRenderMode(int mode)
{
  switch (mode) {
  case WNP_NORMAL_DRAW:
    renderMode=mode;
    break;
  case WNP_PLOT_MODE:
    renderMode=mode;
    break;
  }
}

void WnpSetColor(int color)
{
  short clr;

  if (renderMode==WNP_PLOT_MODE) {
    if (Plot) {
      clr=color;
      PlWrite("C",1);
      PlWrite(&clr,sizeof(short));
    }
  }else WnpColor(color);
}

void WnpVertex(float *x)
{
  if (renderMode==WNP_PLOT_MODE) {
    if (Plot) {
      switch (loopMode) {
      case GL_LINES:
	if (vcount%2) PlWrite("D",1);
	else PlWrite("M",1);
	PlWrite(x,3*sizeof(float));
	vcount=1-vcount;
	break;
      case GL_POINTS:
	PlWrite("X",1);
	PlWrite(x,3*sizeof(float));
	break;
      case GL_LINE_LOOP:
	if (!vcount) {
	  firstX[0]=x[0]; firstX[1]=x[2]; firstX[0]=x[3];
	  PlWrite("l",1);	 
	  vcount++;
	}
	PlWrite("3",1); PlWrite(x,3*sizeof(float));
	break;
      case GL_LINE_STRIP:
	if (!vcount) {
	  PlWrite("l",1);	 
	  vcount++;
	}
	PlWrite("3",1); PlWrite(x,3*sizeof(float));
	break;
      }
    }
  }else glVertex3fv(x);
}

void WnpGlBegin(int mode)
{
  if (renderMode==WNP_PLOT_MODE) {
    loopMode=mode;
    switch (loopMode) {
    case GL_LINES:
    case GL_LINE_LOOP:
    case GL_LINE_STRIP:
      WnpLineWidth(pltLineWidth);
      break;
    case GL_POINTS:
      WnpPointSize(pltPointSize);
      break;
    }
    vcount=0;
  }else glBegin(mode);
}

void WnpGlEnd(void)
{
  if (renderMode==WNP_PLOT_MODE) {
    switch (loopMode) {
    case GL_LINE_LOOP:
      if (Plot&&vcount>1) {
	PlWrite("3",1);
	PlWrite(firstX,3*sizeof(float));
	PlWrite("e",1);
      }
      break;
    case GL_LINE_STRIP:
      PlWrite("e",1);
      break;
    }
    loopMode=0;
  }else glEnd();
}

void WnpVLineWidth(int width)
{
  float rw;

  rw=WnpRealLineWidth(width);
  WnpLineWidth(rw);
}

void WnpLineWidth(float width)
{
  short w;

  if (renderMode==WNP_PLOT_MODE) {
    if (Plot) {
      w=width+0.5;
      if (w!=pltWidth) {
	PlWrite("L",1);
	PlWrite(&w,sizeof(w));
	pltWidth=w;
      }
      pltLineWidth=w;
    }
  }else{
    WnpGlLineWidth(width);
  }
}

void WnpVPointSize(int width)
{
  float rw;

  rw=WnpRealLineWidth(width);
  WnpPointSize(rw);
}

void WnpPointSize(float width)
{
  short w;

  if (renderMode==WNP_PLOT_MODE) {
    if (Plot) {
      w=width+0.5;
      if (w!=pltWidth) {
	PlWrite("L",1);
	PlWrite(&w,sizeof(w));
	pltWidth=w;
      }
      pltPointSize=w;
    }
  }else WnpGlPointSize(width);
}

void *WnpInitPlot(char *file,int xs,int ys)
{
  if (!Plot) {
    if (Plot=WnpPipeOpen(file,"w")) {
      PlWrite(&magic,sizeof(short));
      PlWrite(&xs,sizeof(ys));    /* current window size */
      PlWrite(&ys,sizeof(xs));
      pltWidth=0; pltPointSize=1;
      WnpLineWidth(0);
    }
  }
  return((void *)Plot);
}

void WnpPlotProjection(WnpMatrix  pm)
{
  if (Plot) {
    PlWrite("p",1);
    PlWrite(pm,sizeof(WnpMatrix));
  }
}

void WnpPlotMVT(WnpMatrix  mvt)
{
  if (Plot) {
    PlWrite("v",1);
    PlWrite(mvt,sizeof(WnpMatrix));
  }
}

void WnpPlotCurrentMVT(void)
{
  WnpMatrix mvt;

  if (Plot) {
    glGetFloatv(GL_MODELVIEW_MATRIX,(float *)mvt);
    PlWrite("v",1);
    PlWrite(mvt,sizeof(WnpMatrix));
  }
}

void WnpLineStipple(int num)
{
  short pat;

  if (renderMode==WNP_PLOT_MODE) {
    if (Plot) {
      pat=num;
      PlWrite("s",1);
      PlWrite(&pat,sizeof(pat));
    }
  }else{
    switch(num) {
    case WNP_SOLID_LINE:
      glDisable(GL_LINE_STIPPLE);
      break;
    case WNP_DOTTED_LINE:
      glEnable(GL_LINE_STIPPLE);
      glLineStipple(3,0xaaaa); 
      break;
    }
  }
}

void WnpPlotClipPlanes(float znear,float zfar)
{
  if (Plot) {
    PlWrite("z",1);
    PlWrite(&znear,sizeof(float));
    PlWrite(&zfar,sizeof(float));
  }
}

void WnpPlotViewport(float xl,float yl,float w,float h)
{
  float xh,yh;

  if (Plot) {
    xh=xl+w;
    yh=yl+h;
    PlWrite("P",1);
    PlWrite(&xl,sizeof(float));
    PlWrite(&xh,sizeof(float));
    PlWrite(&yl,sizeof(float));
    PlWrite(&yh,sizeof(float));
  }
}

void WnpDrawCylinder(float *x,float rx,float *y,float ry)
{
  int old;

  if (renderMode==WNP_PLOT_MODE) {
    if (Plot) {
      old=pltLineWidth;
      WnpLineWidth(5);
      WnpGlBegin(GL_LINES);
      WnpVertex(x);
      WnpVertex(y);
      WnpGlEnd();
      WnpLineWidth(old);
    }
  }else  WnpDrawGlCylinder(x,rx,y,ry);
}

void WnpEndPlot(void)
{
  if (Plot) {
    PlWrite("E",1);
    WnpPipeClose(Plot);
    Plot=(FILE *)NULL;
  }
}

