#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wnp.h"

static WnpAtom **atoms=(WnpAtom **)NULL;
static int nat=0;
static float angle,delta;

void WnpNewRotAxis(int n,WnpAtom **mob)
{
  WnpResetRotAxis();
  if (atoms) {
    free(atoms); atoms=(WnpAtom **)NULL;
    nat=0;
    angle=0;
    delta=0;
  }
  if (mob) {
    if (n>2) {         /* first two atoms do not move... */
      nat=n;
      atoms=mob;
      angle=0.0;
      delta=0.0;
    }else{
      free(mob);
    }
  }
}

void WnpIncRotAxis(float d)
{
  if (atoms) delta+=d;
}

void WnpResetRotAxis(void)
{
  if (atoms) {
    delta=-angle; 
    if (WnpAxisRotation()) WnpSetThreeDRedrawFlag(TRUE);
  }
}

float WnpFreezeRotAxis(void)
{
  float r;

  if (atoms) {
    if (WnpAxisRotation()) WnpSetThreeDRedrawFlag(TRUE);
    r=angle;
    angle=0;
  }else{
    r=0;
  }
  return(r);
}

int WnpAxisRotation(void)
{
  WnpMatrix r;
  float e[3],s,ct,st;
  int i,j,k,redraw;

  if (atoms && nat>=2 && delta!=0.0) {
    
    redraw=TRUE;

    /* set up 4x4 homog. rot matrix in r */
    
    for (s=0,i=0;i<3;i++) {
      e[i]=atoms[1]->x[i]-atoms[0]->x[i];
      s+=e[i]*e[i];
    }
    if (s>0) {
      for (i=0;i<3;i++) e[i]/=sqrt(s);
      ct=cos(delta*M_PI/180.0);
      st=sin(delta*M_PI/180.0);
      r[0][0]=e[0]*e[0]+(1.0-e[0]*e[0])*ct;
      r[1][1]=e[1]*e[1]+(1.0-e[1]*e[1])*ct;
      r[2][2]=e[2]*e[2]+(1.0-e[2]*e[2])*ct;
      r[0][1]=e[0]*e[1]*(1-ct)-e[2]*st;
      r[1][2]=e[1]*e[2]*(1-ct)-e[0]*st;
      r[2][0]=e[2]*e[0]*(1-ct)-e[1]*st;
      r[1][0]=e[0]*e[1]*(1-ct)+e[2]*st;
      r[2][1]=e[1]*e[2]*(1-ct)+e[0]*st;
      r[0][2]=e[2]*e[0]*(1-ct)+e[1]*st;
      for (i=0;i<3;i++) {
	r[i][3]=atoms[0]->x[i];
	for (j=0;j<3;j++) r[i][3]-=r[i][j]*atoms[0]->x[j];
	r[3][i]=0.0;
      }
      r[3][3]=1.0;
    }else{
      for (i=0;i<4;i++) {
	for (j=0;j<4;j++) r[i][j]=0.0;
	r[i][i]=1.0;
      }
    }
    
    for (i=2;i<nat;i++) {
      for (j=0;j<3;j++) {
	e[j]=r[j][3];
	for (k=0;k<3;k++) e[j]+=r[j][k]*atoms[i]->x[k];
      }
      for (j=0;j<3;j++) atoms[i]->x[j]=e[j];
    }
    
    angle+=delta;
    delta=0;
  }else redraw=FALSE;
  return(redraw);
}
