#include <stdio.h>
#include <string.h>
#include "wnp.h"
#include "wnpgl.h"
#include "stroke.font"

/*    Definition of global variables   */

typedef struct TEXT_ENV {
	float sx,sy;		        /* x,y char size in world units          */
	int   icur,fcur;		/* initial final cursor                  */
	float pos[3];			/* x,y text position                     */
	int   qual;                     /* text 'quality'                        */
} TEXT_ENV;

static TEXT_ENV t = {
  1.0,1.0,
  1,3,
  {0.0,0.0,0.0},
  WNP_TXT_STROKE
};
  
static int libFont = -1;

static float xic[9] = { 0.0, 0.5, 1.0, 0.0, 0.5, 1.0, 0.0, 0.5, 1.0 };
static float yic[9] = { 0.0, 0.0, 0.0, 0.5, 0.5, 0.5, 1.0, 1.0, 1.0 };

/*   Initialize text environment   */

void WnpTextInit(void)
{
  int i;
  t.sx=1.0; t.sy=1.0;
  t.icur=1;t.fcur=3;
  for (i=0;i<3;i++) t.pos[i]=0.0;
  t.qual = WNP_TXT_HARD;
}

/*   Set text position */

void WnpTextPos3fv(float x[3])
{
  int i;

  for (i=0;i<3;i++) t.pos[i]=x[i];
}

/*   Set text size in world units  */

void WnpTextSize(float sx,float sy)
{
  t.sx=sx;t.sy=sy;
  libFont=-1;
}

/*   Set text initial cursor position */

void WnpTextIniCur(int icur)
{
  t.icur=icur;
}

/*   Set text final cursor position */

void WnpTextFinCur(int fcur)
{
  t.fcur=fcur;
}

void WnpTextQuality(short qual)
{
  t.qual = qual;
}

/*-----------------------------------------------------------------------*/

void WnpDrawText(float *x,WnpMatrix mvt,char *text)
{
  float xp[3],xo[3],*v,vm[3];
  int nch,i,j,n,N,n_out,sv;
  
  if (!text) return;
  nch=strlen(text);
  if (nch==0) return;
  
  switch (t.qual) {
  case WNP_TXT_HARD:
    sv=WnpSetFont(0);
    WnpTextPos3f(x[0],x[1],x[2]);
    WnpTextDraw(text);
    WnpSetFont(sv);
    break;
  case WNP_TXT_LIB:
    if (libFont<0) libFont=WnpSelFont(WNP_FNT_HELV,t.sy*WnpPixelPerA());
    sv=WnpSetFont(libFont);
    WnpTextPos3f(x[0],x[1],x[2]);
    WnpTextDraw(text);
    WnpSetFont(sv);
    break;
  case WNP_TXT_STROKE:

    /* Setup transformation depending on icur, textpos, etc... */
    
    glPushMatrix(); glLoadIdentity();
    for (i=0;i<3;i++) for (xp[i]=mvt[3][i],j=0;j<3;j++) xp[i]+=x[j]*mvt[j][i];
    glTranslatef(xp[0],xp[1],xp[2]);
    glScalef(t.sx,t.sy,1.0);
    xo[0] = -xic[t.icur-1]*nch; xo[1] = -yic[t.icur-1]; xo[2] = 0.0;
    glTranslatef(xo[0],xo[1],xo[2]);
    WnpPlotCurrentMVT();
    
    /* output characters... */
    
    for (i=0;i<nch;i++) {
      n=(int)text[i]-32;
      if (n>=0 && n<96) {
	v=STD_stroke+STD_P[n];
	N=STD_N[n];
	for (n_out=0,j=0;j<N;j++) {
	  if (v[3]) {
	    if (n_out>0) WnpGlEnd();
	    WnpGlBegin(GL_LINE_STRIP); n_out=0;
	  }
	  vm[0]=i+v[0];
	  vm[1]=v[1];
	  vm[2]=0;
	  WnpVertex(vm);
	  v+=4;
	  n_out++;
	}
	if (n_out>0) WnpGlEnd();
      }
    }
    glPopMatrix();
    WnpPlotCurrentMVT();
    break;
  }
}
