#if defined(cygwin) || defined(mingw)
#include <windows.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "wnp.h"
#include "wnpgl.h"
#include "wnpwin.h"

static int redraw_flag=FALSE;                      /* 3D window needs redraw        */

static WnpPrj threeDPrj = {
  WNP_VIEW_PERSPECTIVE,                    /* ortho, perspective, ...                */
  WNP_STEREO_NO,                           /* mono, relaxed, left, right, ...        */ 
  15.0,                                    /* viewport's  y-size in A                */
  10.0,                                    /* near z-clipping plane in A             */
  20.0,                                    /* slabwidth in A                         */
  2.0,                                     /* viewing distance in screen y-sizes     */
  0.2,                                     /* left/right eye separation in screen ys */
  1.3,                                     /* viewport's x:y aspect ratio            */
  15.0,                                    /* screen y-size in A                     */
  0.0                                      /* viewport separation                    */
};
  
static float scaleSpeed  = 0.004;                   /* "gain" on scale               */
static float rotateSpeed = 0.333;                   /* "gain" on rotate              */
static float zrotBreak = 30.0;                      /* catch z-rot singularity       */
static float zrotSpeed = 0.333;                     /* "gain" on z-rot               */

static int   vwp_width=0;                           /* viewport width                */
static int   vwp_height=0;                          /* viewort height                */

static int   pickXsize = 15;                        /* size of picking window        */
static int   pickYsize = 15;                        /* in pixels                     */

static int prev_stereo_mode=WNP_STEREO_RELAXED;     /* stereo mode eneterd by enter-stereo*/

static void dumpScreenCB(char *file,void *data);
static void dumpScreen(char *file);

typedef struct kbmn_t {                             /* save active menu and kbd mode */
  WnpMenu *menu;                                    /* used in dumpMessageBuffer CB  */
  int      kbd;
} kbmn_t;


/*-------------------- temporary stuff --------------------*/

void checkmatrix(char *txt,WnpMatrix m)
{
  int i,j,k;
  float f;

  fprintf(stderr,"%s\n",txt);
  for (i=0;i<4;i++) {
    for (j=0;j<4;j++) fprintf(stderr,"%10.5f ",m[j][i]);
    fprintf(stderr,"\n");
  }
  fprintf (stderr,"\n");
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      for (f=0.0,k=0;k<3;k++) f+=m[i][k]*m[j][k];
      fprintf(stderr,"%10.5f ",f);
    }
    fprintf(stderr,"\n");
  }
  fprintf (stderr,"\n");
}

/*----------------------------------------------------------------------------*/
void WnpDrawThreeDBox(void)
{
  GLint vwp[4],port,view;

  if (ViewportIsOpen("ThreeD")) {
    for (view=WNP_LEFT_VIEW;view<=WNP_RIGHT_VIEW;view++) {
      port=WnpSelectViewport(view,threeDPrj.stereomode);
      if (port!=WNP_VWP_NONE) {
	WnpRebuildInterLists();
	glPushMatrix();
	if (threeDPrj.stereomode==WNP_STEREO_QUADBUFF) {
	  if (view==WNP_LEFT_VIEW) ViewportSetModeFlags("ThreeD",WNP_MODE_LEFT,1);
	  else if (view==WNP_RIGHT_VIEW) ViewportSetModeFlags("ThreeD",WNP_MODE_RIGHT,1);
	}
	ViewportSet("ThreeD",port);
	glGetIntegerv(GL_VIEWPORT,vwp);
	vwp_width=vwp[2];vwp_height=vwp[3];
	WnpSetThreeDView(view);
	WnpClear(WNP_THREED_BACKGROUND);
	WnpDrawThreeDScene();
	glPopMatrix();
	WnpDrawInterBonds(WnpInterBonds());
	WnpDrawMonitorList(WnpMonitorPairs());
	if (threeDPrj.stereomode==WNP_STEREO_QUADBUFF) {
	  ViewportSetModeFlags("ThreeD",WNP_MODE_QUAD,0);
	}
      }
    }
  }
  WnpSetMVTFlag(TRUE);
  redraw_flag=FALSE;
}

int WnpPlotThreeDBox(char *cmd)
{
  GLint vwp[4],port,view,r;
  float xo,yo,w,h;

  r=FALSE;
  if (ViewportIsOpen("ThreeD")) {
    WnpRenderMode(WNP_PLOT_MODE);
    ViewportSet("ThreeD",WNP_VWP_FULL);
    glGetIntegerv(GL_VIEWPORT,vwp);
    xo=vwp[0]; yo=vwp[1]; w=vwp[2]; h=vwp[3];
    if (WnpInitPlot(cmd,w,h)) r=TRUE;;
    if (r) {
      for (view=WNP_LEFT_VIEW;view<=WNP_RIGHT_VIEW;view++) {
	port=WnpSelectViewport(view,threeDPrj.stereomode);
	if (port!=WNP_VWP_NONE) {
	  WnpRebuildInterLists();
	  glPushMatrix();
	  ViewportSet("ThreeD",port);
	  glGetIntegerv(GL_VIEWPORT,vwp);
	  WnpPlotViewport((vwp[0]-xo)/w,(vwp[1]-yo)/h,vwp[2]/w,vwp[3]/h);
	  vwp_width=vwp[2];vwp_height=vwp[3];
	  WnpSetThreeDView(view);
	  WnpDrawThreeDScene();
	  glPopMatrix();
	  WnpDrawInterBonds(WnpInterBonds());
	  WnpDrawMonitorList(WnpMonitorPairs());
	}
      }
      WnpEndPlot();
    }
    WnpRenderMode(WNP_NORMAL_DRAW); 
  }
  return(r);
}

void WnpThreeDPick(Viewport *focus,float xf,float yf,int mode)
{
  int view,port,n,xw,yw;
  WnpAtom *atoms[WNP_MAX_USA_ATOMS];
  DspObj *objects;
  float line[6];
  
  if (focus&&!strcmp(focus->name,"ThreeD")&&focus->open) {
    view=WnpFindThreeDView(xf,threeDPrj.stereomode);
    port=WnpSelectViewport(view,threeDPrj.stereomode);
    if (port!=WNP_VWP_NONE) {
      glPushMatrix();
      ViewportSet("ThreeD",port);
      WnpSetThreeDView(view);
      WnpFractVwpXY2Window(focus,xf,yf,&xw,&yw);
      WnpPickPoint(xw,yw,line);
      WnpThreeDPickMatrix(xw,yw,threeDPrj.stereomode);
      switch (mode) {
      case WNP_PICK_USA:
	n=WnpPickAtoms((DspObj *)WnpDisplayRoot(),atoms,WNP_MAX_USA_ATOMS);
	WnpReportUSAPick(clamp(n,0,WNP_MAX_USA_ATOMS),atoms);
	break;
      case WNP_PICK_ATOM:
	n=WnpPickAtoms(WnpAtomPickRoot(),atoms,1);
	if (n==1) WnpReportAtomPick(atoms[0]);        /* atom must be unique ! */
	break;
      case WNP_PICK_OBJECT:
	n=WnpPickObjects(WnpObjectPickRoot(),&objects,1);
	if (n==1) WnpReportObjectPick(objects);       /* object must be unique ! */
	break;
      case WNP_PICK_XYZ:
	WnpReportPickLine(line);
	break;
      case WNP_PICK_XYZ_ATOM:
	n=WnpPickAtoms(WnpAtomPickRoot(),atoms,1);
	if (n==1) WnpReportAtomPick(atoms[0]);        /* atom must be unique ! */
	else WnpReportPickLine(line);
	break;
      case WNP_PICK_PIVOT:
	n=WnpPickAtoms((DspObj *)WnpDisplayRoot(),atoms,1);
	if (n==1) WnpSetPivotAtom(WnpDisplayRoot(),atoms[0]);
	break;
      case WNP_PICK_ALTPIVOT:
	n=WnpPickAtoms((DspObj *)WnpDisplayRoot(),atoms,1);
	if (n==1) WnpSetPivotAtom(WnpAltDspObj(),atoms[0]);
	break;
      case WNP_PICK_SELALT:
	n=WnpPickObjects((DspObj *)WnpDisplayRoot(),&objects,1);
	if (n==1) WnpSetAltObj(AssocGlbl(objects)); 
	break;
      case WNP_PICK_SELMALT:
	n=WnpPickObjects((DspObj *)WnpDisplayRoot(),&objects,1);
	if (n==1) WnpSetAltObj(AssocGlbl(objects)->up);
	break;
      default:
	break;
      }
      glPopMatrix();
    }
  }
}

void WnpThreeDPickMatrix(int x,int y,int mode)
{
  /* pre-multiply current projection matrix with pickmatrix at (x,y) */
  /* (x,y) are in window coordinates, stereo mode is mode.           */

  GLint vwp[4];
  int pxs,pys;
  WnpMatrix prj;

  pxs=pickXsize; pys=(mode==WNP_STEREO_CRYSTAL ? 0.5*pickYsize+0.5 : pickYsize);
  glGetIntegerv(GL_VIEWPORT,vwp);
  glGetFloatv(GL_PROJECTION_MATRIX,(float *)prj);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix(x,y,pxs,pys,vwp);
  glMultMatrixf((float *)prj);
  glMatrixMode(GL_MODELVIEW);
}

void WnpSetThreeDView(int view)
{
  float xs,ys,dist,znear,zfar,eyepos;
  WnpPrj prj;
  
  prj=WnpGetThreeDPrj();
  ys=0.5*prj.ysize;
  xs=prj.aspect*ys;

  znear=prj.znear;
  zfar=prj.znear-prj.slabwidth;
  dist = prj.scrysize*prj.viewdist;
  eyepos = view*0.5*prj.scrysize*prj.eyesep;

  switch (prj.viewmode) {
  case WNP_VIEW_PERSPECTIVE:
    WnpStereoPersp(-xs,xs,-ys,ys,znear,zfar,znear,dist,eyepos,prj.sep);
    break;
  case WNP_VIEW_ORTHO:
    WnpStereoOrtho(-xs,xs,-ys,ys,znear,zfar,dist,eyepos,prj.sep);
    break;
  }
}

WnpPrj WnpGetThreeDPrj(void)
{
  WnpPrj r;
  float TDxs,TDys;
  
  /* need to update screensize in A and x/y aspect ratio ... */

  r=threeDPrj;
  WnpGetVwpDimensions("ThreeD",&TDxs,&TDys);
  r.aspect=TDxs/TDys;
  if (r.stereomode==WNP_STEREO_CROSSED||r.stereomode==WNP_STEREO_RELAXED) r.aspect*=0.5;
  r.scrysize=r.ysize*WnpDisplayHeight()/TDys;

  return(r);
}

int WnpSelectViewport(int view,int stereomode)
{
  int vwp;

  switch (stereomode) {
  case WNP_STEREO_NO:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_LEFT:
    vwp=(view==WNP_LEFT_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_RIGHT:
    vwp=(view==WNP_RIGHT_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_CROSSED:
    vwp=(view==WNP_RIGHT_VIEW ? WNP_VWP_LEFT :
	 (view==WNP_LEFT_VIEW ? WNP_VWP_RIGHT :
	  WNP_VWP_NONE)
	 );
    break;
  case WNP_STEREO_RELAXED:
    vwp=(view==WNP_RIGHT_VIEW ? WNP_VWP_RIGHT :
	 (view==WNP_LEFT_VIEW ? WNP_VWP_LEFT :
	  WNP_VWP_NONE)
	 );
    break;
  case WNP_STEREO_CRYSTAL:
    vwp=(view==WNP_RIGHT_VIEW ? WNP_VWP_BOTTOM :
	 (view==WNP_LEFT_VIEW ? WNP_VWP_TOP:
	  WNP_VWP_NONE)
	 );
    break;
  case WNP_STEREO_QUADBUFF:
    vwp=(view==WNP_RIGHT_VIEW ? WNP_VWP_FULL :
	 (view==WNP_LEFT_VIEW ? WNP_VWP_FULL :
	  WNP_VWP_NONE)
	 );
    break;
  default:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  }
  return(vwp);
}

int WnpSelectTextport(int view,int stereomode)
{
  int vwp;

  switch (stereomode) {
  case WNP_STEREO_NO:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_LEFT:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_RIGHT:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_CROSSED:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_RELAXED:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  case WNP_STEREO_CRYSTAL:
    vwp=(view==WNP_RIGHT_VIEW ? WNP_VWP_BOTTOM:
	 (view==WNP_LEFT_VIEW ? WNP_VWP_TOP:
	  WNP_VWP_NONE)
	 );
    break;
  case WNP_STEREO_QUADBUFF:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  default:
    vwp=(view==WNP_MIDDLE_VIEW ? WNP_VWP_FULL : WNP_VWP_NONE);
    break;
  }
  return(vwp);
}

int WnpFindThreeDView(float xf,int stereomode)
{
  int view;

  switch (stereomode) {
  case WNP_STEREO_NO:
    view=WNP_MIDDLE_VIEW;
    break;
  case WNP_STEREO_LEFT:
    view=WNP_LEFT_VIEW;
    break;
  case WNP_STEREO_RIGHT:
    view=WNP_RIGHT_VIEW;
    break;
  case WNP_STEREO_CROSSED:
    view=(xf <=0.5 ? WNP_RIGHT_VIEW : WNP_LEFT_VIEW);
    break;
  case WNP_STEREO_RELAXED:
    view=(xf <=0.5 ? WNP_LEFT_VIEW : WNP_RIGHT_VIEW);
    break;
  case WNP_STEREO_CRYSTAL:
    view=WNP_LEFT_VIEW;
    break;
  case WNP_STEREO_QUADBUFF:
    view=WNP_LEFT_VIEW;
    break;
  default:
    view=WNP_MIDDLE_VIEW;
    break;
  }
  return(view);
}

void WnpSetThreeDPrj(WnpPrj prj)
{
  threeDPrj=prj;
  redraw_flag=TRUE;
}

void WnpDrawThreeDScene(void)
{
  GlblObj *top;

  top=WnpDisplayRoot();
  if (top) {
    glPushAttrib(GL_LIGHTING_BIT);   /* set GL_SMOOTH to work around a bug in       */
    glShadeModel(GL_SMOOTH);         /* versions of OpenGL                          */
    WnpDrawObject((DspObj *)top);
    glPopAttrib();                   /* reset Lighting attributes (end of bug fix ) */
  }
}

void WnpThreeDScale(int upd)
{
  if (upd!=0) threeDPrj.ysize*=(1-upd*scaleSpeed);
  if (threeDPrj.ysize < 0.1) threeDPrj.ysize=0.1;
  threeDPrj.scrysize=threeDPrj.ysize*vwp_height/WnpDisplayHeight();
  WnpSetThreeDRedrawFlag(TRUE);
}

void WnpThreeDScaleRel(float scale)
{
  if (scale>0) {
    threeDPrj.ysize*=scale;
    if (threeDPrj.ysize < 0.1) threeDPrj.ysize=0.1;
    threeDPrj.scrysize=threeDPrj.ysize*vwp_height/WnpDisplayHeight();
    WnpSetThreeDRedrawFlag(TRUE);
  }
}

void WnpThreeDSlabPosition(int upd)
{
  float s,d;
  WnpPrj prj;
 
  if (upd!=0) {
    prj=WnpGetThreeDPrj();
    d=-upd*prj.scrysize/WnpDisplayHeight();
    switch (prj.viewmode) {
    case WNP_VIEW_ORTHO:
      s=1.0;
      break;
    case WNP_VIEW_PERSPECTIVE:
      s=1.0-d/(prj.viewdist*prj.scrysize);
      break;
    default:
      s=1.0;
      break;
    }
    if (s>0) {
      prj.ysize*=s;
      prj.znear+=d;
    }
    WnpSetThreeDPrj(prj);
  }
}

void WnpThreeDSlabWidth(int upd)
{
  WnpPrj prj;
  
  if (upd!=0) {
    prj=WnpGetThreeDPrj();
    prj.slabwidth+=upd*prj.scrysize/WnpDisplayHeight();
    if (prj.slabwidth<0.1) prj.slabwidth=0.1;
    WnpSetThreeDPrj(prj);
  }
}

void WnpThreeDStereoSeparation(int upd)
{
  WnpPrj prj;

  if (upd!=0) {
    prj=WnpGetThreeDPrj();
    prj.sep-=upd*prj.scrysize/WnpDisplayHeight();
    WnpSetThreeDPrj(prj);
  }
}    

void WnpThreeDViewingDistance(int upd)
{
  WnpPrj prj;

  if (upd!=0) {
    prj=WnpGetThreeDPrj();
    prj.viewdist+=(float)upd/WnpDisplayHeight();
    if (prj.viewdist<0.1) prj.viewdist=0.1;
    WnpSetThreeDPrj(prj);
  }
}

void WnpThreeDTranslatePx(GlblObj *o,int dx,int dy,int dz)
{
  float txyz[3];

  txyz[0]=dx*threeDPrj.ysize/vwp_height;
  txyz[1]=dy*threeDPrj.ysize/vwp_height;
  txyz[2]=dz*threeDPrj.ysize/vwp_height;

  WnpThreeDTranslate(o,txyz);
}

void WnpThreeDTranslate(GlblObj *o,float *txyz)
{
  int i,j;

  if (o&&o->type==DSP_GLBL) {
    if (o->up) {
      for (i=0;i<3;i++) {
	for (j=0;j<3;j++) o->tra[i]+=o->up->mvt[i][j]*txyz[j];
      }
    }else{
      o->tra[0]+=txyz[0];
      o->tra[1]+=txyz[1];
      o->tra[2]+=txyz[2];
    }
    WnpSetThreeDRedrawFlag(TRUE);
    WnpSetMVTFlag(FALSE);
  }
}

void WnpThreeDRotate(GlblObj *o,float *axis,float angle)
{
  WnpMatrix ui,u;
  int i,j;

  if (o&&o->type==DSP_GLBL) {
    glPushMatrix();
    glLoadIdentity();
    if (o->up) {
      ui[3][3]=u[3][3]=1.0;
      for (i=0;i<3;i++) {
	for (j=0;j<3;j++) ui[j][i]=u[i][j]=o->up->mvt[i][j];
	ui[i][3]=u[3][i]=0.0; ui[3][i]=u[i][3]=0.0;
      }
      glMultMatrixf((float *)ui);
      glRotatef(angle,axis[0],axis[1],axis[2]);
      glMultMatrixf((float *)u);
    }else{
      glRotatef(angle,axis[0],axis[1],axis[2]);
    }
    glMultMatrixf((float *)o->rot);
    glGetFloatv(GL_MODELVIEW_MATRIX,(float *)o->rot);
    WnpCleanMVT(o->rot);
    glPopMatrix();
    WnpSetThreeDRedrawFlag(TRUE);
    WnpSetMVTFlag(FALSE);
  }
}

void WnpThreeDXYrot(GlblObj *o,int dx,int dy)
{
  float axis[3],angle;

  axis[0]=-dy; axis[1]=dx; axis[2]=0;
  angle=rotateSpeed*sqrt(dx*dx+dy*dy);
  WnpThreeDRotate(o,axis,angle);
}

void WnpThreeDZrot(GlblObj *o,int x,int y,int dx,int dy)
{
  int vx1,vx2,vy1,vy2;
  float a,axis[3];

  vx1=x-0.5*vwp_width;
  vx2=vx1+dx;
  
  vy1=0.5*vwp_height-y;
  vy2=vy1+dy;

  if (o&&o->type==DSP_GLBL) {
    if ((vx1||vy1) && (vx2||vy2)) {
      a=(atan2(vy2,vx2)-atan2(vy1,vx1))*180.0;
      if (a<-180.0) a+=180.0;
      if (a>+180.0) a-=180.0;
      a*=zrotSpeed;
      axis[0]=0; axis[1]=0; axis[2]=1.0;
      if (abs(a)<zrotBreak) WnpThreeDRotate(o,axis,a);
    }
  }
}

void WnpAdjustLightPosition(int dx,int dy,int dz)
{
  float t[3];

  t[0]=dx*threeDPrj.ysize/vwp_height;
  t[1]=dy*threeDPrj.ysize/vwp_height;
  t[2]=dz*threeDPrj.ysize/vwp_height;
  WnpMoveLight(t);
  redraw_flag=TRUE;
}

void WnpAdjustLightDirection(int x,int y,int dx,int dy)
{
  int vx1,vx2,vy1,vy2;
  float a;

  vx1=x-0.5*vwp_width;
  vx2=vx1+dx;
  
  vy1=0.5*vwp_height-y;
  vy2=vy1+dy;

  if ((vx1||vy1) && (vx2||vy2)) {
    a=(atan2(vy2,vx2)-atan2(vy1,vx1))*180.0;
    if (a<-180.0) a+=180.0;
    if (a>+180.0) a-=180.0;
    a*=zrotSpeed;
    WnpRotateLight(a);
    redraw_flag=TRUE;
  }
}

int WnpThreeDRedrawFlag(void)
{
  GlblObj *R;

  R=WnpDisplayRoot();
  return(redraw_flag||(R ? R->flags&OBJ_FLG_MOD : 0));
}

void WnpSetThreeDRedrawFlag(int s)
{
  redraw_flag=s;
}

void WnpScreenDump(char *file)
{
  WnpMenu *save;
  int old;
  kbmn_t *kbmn_state;

  if (!file) {
    if (!WnpLocalStringDispatcher()) {
      kbmn_state=(kbmn_t *)malloc(sizeof(kbmn_t));
      save=WnpSetActiveMenu((WnpMenu *)NULL);             /* register local-string CB */
      WnpSetLocalPrompt("name of file for screen dump: ");
      old=WnpSetKeyboardMode(WNP_KBD_LOCAL);
      if (kbmn_state) {
	kbmn_state->menu=save;
	kbmn_state->kbd=old;
      }
      WnpGetLocalString(dumpScreenCB,(void *)kbmn_state);
    }
  }else dumpScreen(file);
}

int WnpSetStereoMode(int mode)
{
  WnpPrj prj;
  int r;
  GLboolean qbs;

  prj=WnpGetThreeDPrj();
  r=prj.stereomode;
  if (mode==WNP_STEREO_PREVIOUS) mode=prev_stereo_mode;
  if (mode==WNP_STEREO_QUADBUFF) {
    glGetBooleanv(GL_STEREO,&qbs);
    if (!qbs) mode=r;
  }
  if (mode!=r) {
    prj.stereomode=mode;
    WnpSetThreeDPrj(prj);
    if (mode==WNP_STEREO_CRYSTAL||r==WNP_STEREO_CRYSTAL) WnpSetGlBoxRedrawFlag(TRUE);
    if (mode==WNP_STEREO_QUADBUFF||r==WNP_STEREO_QUADBUFF) WnpSetGlBoxRedrawFlag(TRUE);
  }
  if (mode==WNP_STEREO_CRYSTAL||
      mode==WNP_STEREO_CROSSED||
      mode==WNP_STEREO_QUADBUFF||
      mode==WNP_STEREO_RELAXED) prev_stereo_mode=mode;

  WnpCrystalEyes(mode==WNP_STEREO_CRYSTAL);

  return(r);
}

int WnpGetStereoMode(void)
{
  WnpPrj prj;
  prj=WnpGetThreeDPrj();
  return(prj.stereomode);
}

int WnpSetViewMode(int mode)
{
  int r;

  r=threeDPrj.viewmode;
  redraw_flag|=(mode!=threeDPrj.viewmode);
  threeDPrj.viewmode=mode;

  return(r);
}
int WnpGetViewMode(void)
{
  return(threeDPrj.viewmode);
}

float WnpPixelPerA(void)
{
  return(WnpDisplayHeight()/clamp(threeDPrj.scrysize,0.001,threeDPrj.scrysize));
}

/*------------------------------------------------------------------------*/
static void dumpScreenCB(char *file,void *kbmn_state)
{
  WnpSetLocalPrompt("dumping ");
  if (kbmn_state) WnpSetActiveMenu(((kbmn_t*)kbmn_state)->menu);
  WnpWorkProc();
  dumpScreen(file);
  if (kbmn_state) {
    WnpSetKeyboardMode(((kbmn_t*)kbmn_state)->kbd);
    free(kbmn_state);
  }
}

static void dumpScreen(char *name)
{
  char cmd[WNP_MAXREC],buf[WNP_MAXREC];
  int xo,yo,width,height;
  int x1,x2,y1,y2,xmax,ymax;
  FILE *srv;

  if (name&&ViewportIsOpen("ThreeD")) {
    if (strlen(name)>0) {
      ViewportLocate("ThreeD",&xo,&yo,&width,&height);
      xmax=WnpDisplayWidth()-1; ymax=WnpDisplayHeight()-1;
      x1=clamp(xo,0,xmax); x2=clamp(xo+width-1,0,xmax);
      y1=clamp(yo,0,ymax); y2=clamp(yo+height-1,0,ymax);
      sprintf(cmd,"%s %s 0x%08x %d %d %d %d %d %d %d %d %d %d",
	      "wnp_windmp",name,(unsigned int)WnpGlBoxWindow(),
	      0,0,width,height,x1,x2,y1,y2,xmax,ymax);
      WnpGlRestoreCursorPatch();
      glFlush();
      srv=WnpPipeOpen(cmd,"r");
      if (srv) {
	while(fgets(buf,sizeof(buf),srv)) WnpSendToCom("%s",buf);
	WnpPipeClose(srv);
      }else{
	WnpSendToCom("screen dump: failed to connect to wnp_windmp tool");
      }
      WnpGlDrawCursor(TRUE);
      WnpRingBell();
    }
  }
}
