#ifndef WNPTYPESDEF
#define WNPTYPESDEF

typedef float WnpMatrix[4][4];

typedef void (*WnpLocalStringCB)(char *locstring,void *data);

typedef struct Viewport {
  char *name;                    /* unique name of viewport             */
  struct Viewport *nxt;          /* pointer to next viewport in list    */
  float x1,x2,y1,y2;             /* fractional coordinates of LL and UR */
  int open;                      /* open/closed flag                    */
  int mode;                      /* modes for this viewport             */
} Viewport;

typedef struct WnpPrj{
  int viewmode;                  /* ortho, perspective, ...                */
  int stereomode;                /* mono, relaxed, left, right, ...        */ 
  float ysize;                   /* viewport's  y-size in A                */
  float znear;                   /* near z-clipping plane in A             */
  float slabwidth;               /* slabwidth in A                         */
  float viewdist;                /* viewing distance in screen y-sizes     */
  float eyesep;                  /* left/right eye separation in screen ys */
  float aspect;                  /* viewport's x:y aspect ratio            */
  float scrysize;                /* screen y-size in A                     */
  float sep;                     /* viewport separation                    */
} WnpPrj;

typedef struct WnpMenu {
  char *host;                    /* name of host viewport                        */
  struct WnpMenuButton *btn;     /* menu button list (reserved have neg. offset) */
  int nbtn,mbtn,nres,top,ndef;   /* nbr of buttons, max nbtn, res. buttons,...   */
  int dirty;                     /* flag, true if menu needs redraw.             */
} WnpMenu;

typedef void (*WnpMenuAction)(void *data,int button);

typedef struct WnpMenuButton {
  WnpMenuAction action;          /* action on press-button */
  char *text;                    /* button label           */
  int color;                     /* label color            */
  int background;                /* background color       */
  int active;                    /* button-is-active flag  */
} WnpMenuButton;

/*-------------------- Display object types ... --------------------*/

typedef struct {
  unsigned char b[8];
} cptr_t;

typedef struct DspObj {          /* abstract display object structure */
  struct GlblObj *up;
  struct DspObj *prv,*nxt;
  cptr_t cross;
  int type,flags;
} DspObj;

typedef struct GlblObj {         /* global object */
  struct GlblObj *up;
  struct DspObj *prv,*nxt;
  cptr_t cross;
  int type,flags;
  struct DspObj *fst,*lst;
  WnpMatrix mvt,rot;
  float tra[3],pivot[3];
} GlblObj;

typedef struct MdlObj {          /* molecular display list */
  struct GlblObj *up;
  struct DspObj *prv,*nxt;
  cptr_t cross;
  int type,flags;
  struct WnpAtom *fstAtom;
  struct WnpBond *fstBond;
  struct WnpNOE  *fstNOE;
  struct WnpDots *fstDot;
  int NOElabt,NOEstyle,NOEwidth;
} MdlObj;

typedef struct CSObj {           /* color sorted object    */
  struct GlblObj *up;
  struct DspObj *prv,*nxt;
  cptr_t cross;
  int type,flags;
  int lw,mds_type;
  struct WnpLst *fst;
} CSObj;

typedef struct TRGObj {           /* triangulated surface object  */
  struct GlblObj *up;
  struct DspObj *prv,*nxt;
  cptr_t cross;
  int type,flags;
  int *stipple;                   /* stipple pattern for pseudo transparency */
  int  color,lw,style,pat,bckclr; /* rendering attributes                    */
  struct WnpLst *fst;             /* linked list of triangle-meshes          */
} TRGObj;

/*-------------------- Atoms and Bonds... --------------------*/

typedef struct WnpAtom {
  struct MdlObj *up;
  struct WnpAtom *nxt;
  cptr_t cross;
  struct WnpAtom *sph,*lab,*mid;
  struct WnpLst *xtra;
  char *label;
  float x[3],r;
  int flags;
  int color,lcolor;
} WnpAtom;

typedef struct WnpBond {
  struct MdlObj *up;
  struct WnpBond *nxt;
  struct WnpAtom *f,*t;
  struct WnpBond *dsp;
  struct WnpBond *lab;
  char *label;
  float r;
  int lcolor,lw;
  int flags;
} WnpBond;

/*-------------------- linked list element types --------------------*/

typedef struct WnpLst {          /* abstract doubly linked list element */
  struct WnpLst *prv,*nxt;
  int type;
} WnpLst;

typedef struct WnpDots {
  struct WnpLst *prv,*nxt;
  int type;
  struct WnpAtom *up;
  struct WnpDots *dot;           /* pointer to next dot in display        */
  float *xyz;                    /* note: four per point (x,y,z,property) */
  int n;
  int color,dsp,style;
} WnpDots;

typedef struct WnpMonPair {
  struct WnpLst *prv,*nxt;
  int type;
  WnpAtom *f,*t;
  int dsp,color,label,width;
} WnpMonPair;

typedef struct WnpNOE {
  struct WnpLst *prv,*nxt;
  int type;
  struct WnpNOE *dsp;
  struct WnpAtom *f,*t;
  int color;
  float D,viol;
} WnpNOE;

typedef struct WnpMDS {          /* move/draw sequence...                */
  struct WnpLst *prv,*nxt;
  int type;
  struct CSObj *up;
  int color;
  int n;
  float *x;
} WnpMDS;

typedef struct WnpTOBJ {          /* Text object                          */
  struct WnpLst *prv,*nxt;
  int type;
  struct CSObj *up;
  int color;
  char *text;
  float x[3];
} WnpTOBJ;

typedef struct WnpTRG{            /* Triangle mesh                               */
  struct WnpLst *prv,*nxt;
  int type;
  struct TRGObj *up;
  float *v;                       /* position and normal (6 floats) per vertex   */
  int   *t;                       /* triangles (three pointers into vertex list) */
  int   nv,nt;                    /* number of vertices and triangles            */
} WnpTRG;

#endif

