#include <stdio.h>
#include <sys/types.h>
#ifndef mingw
#include <sys/times.h>
#endif
#include <sys/param.h>
#include <unistd.h>
#include "wnp.h"
#include "wnpgl.h"

void uSleep(long microseconds);

#if defined(darwin)
#define HZ 100
#endif

#define GRAPHICS 1
#define CHAT     2

static int mode=GRAPHICS;
static int nidle=0;                      /* number of idle calls                     */
static int thresh = 25;                  /* number of idle loops before sleep        */
static int Tinner = 2;                   /* number of idle Comm calls in inner loop  */
static int curs_loops = 200;             /* number of cursor-on loops in photo mode  */

/* Benchmarking */

static int benchmark = 0;                /* benchmark loop count                     */
static int Loop = 0;                     /* number of benchmark loops                */
#ifndef mingw
static int E1,U1,S1;                     /* timer setting at start of benchmark      */
static int oldKBmode;                    /* keyboard and ...                         */
static WnpMenu *oldActiveMenu;           /* ... menu state saved during local string */
static int bmstmode=FALSE;               /* are we waiting for benchmark param CB ?  */
#endif

/* Locking and Rocking  */

static int lock = 0;                     /* lock level count                         */
static int rock = FALSE;                 /* rocking flag                             */
static float rockampl=30.0;              /* default rocking amplitude                */
static float rockdelta=1.0;              /* default increment of rock angle          */
static int rockdir=1;                    /* rocking direction                        */
static float rockangle=0;                /* current rock angle                       */
static float rockax[3]={0,1,0};          /* axis of rocking                          */

static int win_photo;                    /* in photo mode ?                          */
static int curs_cnt=0;                   /* counter to cursor-off in photo-mode      */

static void StartBenchMark(char *msg,void *dummy);

/*------------------------------------------------------------------------*/

void WnpWorkProc(void)
{
  int busy,back;

  /* Workprocedure called when X has nothing else to do... */

  do {
    if (lock<=0) {
      if (WnpResetCBFlag()) mode=GRAPHICS;
      if (benchmark < Loop) {
	mode=GRAPHICS;
	WnpThreeDXYrot(WnpDisplayRoot(),2,1);
	benchmark++;
      }else if (Loop>0&&benchmark==Loop) {
	WnpSetBenchmarkCounter();
      }else if (rock&&mode==GRAPHICS) {
	WnpThreeDRotate(WnpDisplayRoot(),rockax,rockdelta*rockdir);
	rockangle+=rockdelta*rockdir;
	if (rockangle>=rockampl)  rockdir=-1;
	if (rockangle<=-rockampl) rockdir=+1;
      }
    }
    switch (mode) {
    case GRAPHICS:
      busy=WnpGlBoxRedraw();
      mode=CHAT;
      back=TRUE;
      break;
    case CHAT:
      busy=WnpCommQueue(&back);
      if (back) mode=GRAPHICS;
      break;
    }
    nidle=(busy ? 0 : nidle+1);
  } while (!(back||rock||Loop)&&nidle<Tinner);
  if (nidle>thresh) {
    /* WnpReportGlError("WnpWorkProc - GL error"); */ /* turn on for debugging */
    uSleep(10000);
    curs_cnt=(curs_cnt+1)%curs_loops;
    if (win_photo&&curs_cnt==0) WnpGlBoxHideCursor();
    mode=(mode==CHAT&&!back ? GRAPHICS : CHAT) ;
  }
}

/*-------------------- Benchmark support functions --------------------*/

void WnpSetBenchmarkCounter(void)
{
#ifndef mingw
  int e2;
  struct tms tms;  

  if (Loop>0) {
    WnpRingBell();
    e2=times(&tms);
    WnpSendToCom("Times for %d loops: Elapsed = %d,   User = %0.1f,   System = %0.1f\n",
		 benchmark,(e2-E1)/HZ,((float)tms.tms_utime-U1)/HZ,
		 ((float)tms.tms_stime-S1)/HZ);
    Loop=0; benchmark=0;
  }else {
    if (!bmstmode) {
      oldActiveMenu=WnpSetActiveMenu((WnpMenu *)NULL);
      WnpSetLocalPrompt("number of benchmark loops: ");
      oldKBmode=WnpSetKeyboardMode(WNP_KBD_LOCAL);
      WnpGetLocalString(StartBenchMark,(void *)(NULL));
      bmstmode=TRUE;
    }
  }
#endif
}

static void StartBenchMark(char *msg,void *dummy)
{
#ifndef mingw
  int n;
  struct tms tms;  

  /* so here we are, back from local interaction. Restore keyboard and menu */

  dummy=dummy;
  WnpSetKeyboardMode(oldKBmode);
  WnpSetActiveMenu(oldActiveMenu);
  bmstmode=FALSE;

  /* decode the message, which should tell us the number of loops... */

  if (msg) {
    if (sscanf(msg,"%d",&n)==1) Loop=(n>=0 ? n : 0);
    E1=times(&tms);
    U1=tms.tms_utime; S1=tms.tms_stime;
    benchmark=0;
    if (Loop>0) WnpRingBell();
  }
#endif
}
/*-------------------- Locking and Rocking --------------------*/

void WnpLockEvtQ(int inc)
{
  lock+=inc;
}

int WnpEvtQLocked(void)
{
  return(lock>0);
}

void WnpSetRocking(float *pars)
{
  rockampl=pars[0];
  if (rockampl<0) rockampl=-rockampl; 
  rockdelta=pars[1];
  rockangle=clamp(rockangle,-rockampl,rockampl);
}

void WnpToggleRocking(void)
{
  rock=!rock;
}

void WnpTogglePhotoMode(int ticks)
{
  win_photo=!win_photo;
  if (!win_photo) WnpGlBoxShowCursor();
  if (ticks>0) curs_loops=ticks;
}

