      subroutine betas(*)
c     
c     Compute resonance integrals beta and twocenter repulsion integrals gamma.
c     The original Lindner method (PIMM) is used for beta. The offdiagonal terms
c     of the gamma matrix are computed from the diagonals (parameter tcri) using
c     the Ohno-Klopman formula (QC-3 form. 13b p 183).
c     
c     BUGFIX 2-Jul-1987 by AWI : discontinuity in sigma overlap of
c     pi-orbitals when pi-overlap of same orbitals changes sign. 
c     Lindner: force S-pi to be positive by sign change in S
c     Widmer : force S to be positive.
c     
      implicit none
      include 'sizes.h'
      include 'atom.h'
      include 'energy.h'
      include 'apara.h'
      include 'bpara.h'
      include 'fock.h'
      include 'worksp.h'
c     
c     Declarations of local variables
c     
      integer ne(10)            ! max. 10 neighbours per atom
      integer l,k,i,j,ia,nn,ib,m,nmel
      real    nrml(3,mpiat)     ! normal vectors on pi-centers
      real    d12(3),d13(3)
      real    eab(3),ca(3),cb(3),rot(3,3)
      real    fnor,rab,cost,sint,cosp,sinp,spi,ovlp,ssi,ovlps 
      real    ok,x,g1,g2
      equivalence(work,nrml)
c     
c     Definition of Ohno-Klopman function
c     
      ok(x,g1,g2)=14.397/sqrt(x**2+0.25*(14.397/g1+14.397/g2)**2)
c     
c     Mark bonds in beta matrix
c     
      ib=0
      nmel=(natpi*(natpi+1))/2
      do 6 l=1,nmel
         beta(l)=0.0
 6    continue
      do 8 k=1,nbnd
         if (pinr(bnd(1,k)).ne.0.and.pinr(bnd(2,k)).ne.0) then
            i=max0(pinr(bnd(1,k)),pinr(bnd(2,k)))
            j=min0(pinr(bnd(1,k)),pinr(bnd(2,k)))
            l=(i*(i-1))/2+j
            beta(l)=1.0
         end if
 8    continue
c     
c     Compute normals to pi-centers
c     
      do 100 i=1,natpi
         ia=atnr(i)
c     
c     Search immediate neighbours of atom ia
c     
         nn=0
         do 10 k=1,nbnd
            if (ia.eq.bnd(1,k)) then
               nn=nn+1
               ne(nn)=bnd(2,k)
            else if (ia.eq.bnd(2,k)) then
               nn=nn+1
               ne(nn)=bnd(1,k)
            end if
 10      continue
         if (nn.eq.1) then
c     
c     'ia' has only one neighbour. Search 'second neighbourhood'.
c     
            nn=0
            ib=ne(1)
            do 20 k=1,nbnd
               if (ib.eq.bnd(1,k)) then
                  if (bnd(2,k).ne.ia) then
                     nn=nn+1
                     ne(nn)=bnd(2,k)
                  end if
               else if (ib.eq.bnd(2,k)) then
                  if (bnd(1,k).ne.ia) then
                     nn=nn+1
                     ne(nn)=bnd(1,k)
                  end if
               end if
 20         continue
         end if
c     
c     If necessary include 'ia' and 'ib' to define plane
c     
         if (nn.eq.0) then
            write(5,'(a)') ' Can''t compute normal of '//name(ia)
            return 1
         else if (nn.eq.1) then
            nn=2
            ne(2)=ib
         end if
         if (nn.eq.2) then
            ne(3)=ia
            nn=3
         end if
c     
c     Compute normal to plane defined by first three neighbours (first or second)
c     
         do 30 k=1,3
            d12(k)=xyz(k,ne(1))-xyz(k,ne(2))
            d13(k)=xyz(k,ne(1))-xyz(k,ne(3))
 30      continue
         nrml(1,i)=d13(2)*d12(3)-d13(3)*d12(2)
         nrml(2,i)=d13(3)*d12(1)-d13(1)*d12(3)
         nrml(3,i)=d13(1)*d12(2)-d13(2)*d12(1)
         fnor=1.0/sqrt(nrml(1,i)**2+nrml(2,i)**2+nrml(3,i)**2)
         do 40 k=1,3
            nrml(k,i)=fnor*nrml(k,i)
 40      continue
 100  continue
c     
c     The beta-matrix is stored in symmetric storage mode
c     
      do 200 i=2,natpi
         ia=atnr(i)
         l=(i*(i-1))/2
         do 190 j=1,i-1
            ib=atnr(j)
            l=l+1
c     
c     Compute 'bond'-direction e, rotationmatrix rot and rotated normals ca,cb
c     
            do 110 k=1,3
               eab(k)=xyz(k,ia)-xyz(k,ib)
 110        continue
            rab=sqrt(eab(1)**2+eab(2)**2+eab(3)**2)
            do 120 k=1,3
               eab(k)=eab(k)/rab
 120        continue
            cost=eab(3)
            sint=sqrt(1.0-amin1(1.0,cost**2))
            if (sint.lt.1.e-7) then
               cosp=1.0
               sinp=0.0
            else
               cosp=eab(1)/sint
               sinp=eab(2)/sint
            end if
            rot(1,1)=sint*cosp
            rot(1,2)=sint*sinp
            rot(1,3)=cost
            rot(2,1)=cost*cosp
            rot(2,2)=cost*sinp
            rot(2,3)=-sint
            rot(3,1)=-sinp
            rot(3,2)=cosp
            rot(3,3)=0.0
            do 130 k=1,3
               ca(k)=0.0
               cb(k)=0.0
               do 125 m=1,3
                  ca(k)=ca(k)+rot(k,m)*nrml(m,i)
                  cb(k)=cb(k)+rot(k,m)*nrml(m,j)
 125           continue
 130        continue
c     
c     Evaluate overlapintegral (Slater-Zener orbitals), pi-pi and sigma-sigma.
c     Total overlap should be positive (sign change is equivalent to inverting
c     direction of normal at one of the two centers...).
c     
            spi=(ca(2)*cb(2)+ca(3)*cb(3))*ovlp(rab,zsz(ntype(ia)),
     *           zsz(ntype(ib)))
            ssi=-ca(1)*cb(1)*ovlps(rab,zsz(ntype(ia)),zsz(ntype(ib)))
            if (spi+ssi.lt.0.0) then
               spi=-spi
               ssi=-ssi
            end if
c     
c     Calculate beta from overlap and parameter bk. If no pi-bond exists between
c     'ia' and 'ib' only the sigma-sigma part will be used to account for homo-
c     conjugative effects.
c     
            if (beta(l).gt.0.5) then
               beta(l)=-bk(ntype(ia),ntype(ib))*(spi+ssi)
            else
               beta(l)=-bk(ntype(ia),ntype(ib))*ssi
            end if
            gamma(l)=ok(rab,tcri(ntype(ia)),tcri(ntype(ib)))
 190     continue
 200  continue
c     
c     Diagonal terms of beta and gamma are easy:
c     
      do 210 i=1,natpi
         l=(i*(i+1))/2
         k=ntype(atnr(i))
         beta(l)=uip(k)
         gamma(l)=tcri(k)
 210  continue
c     
c     done...
c     
      return
      end
