      subroutine egbend(x,e,gr)
c     
c     Compute energy due to out of plane deformations for neighbours of
c     pi-centers and update gradient.
c     
      implicit none
      real pihalf
      parameter (pihalf=3.141592653/2.0)
      include 'sizes.h'
      include 'energy.h'
      real    x(3,*),gr(3,*),xj(3),xk(3),xl(3)

      integer i,j,i1,i2,i3,i4
      real ril,t1,t2,t3,s,b,e,f,tn,fs,dj1,dj2,dj3,dk1,dk2,
     .     dk3,dl1,dl2,dl3,rd

c     
c     The following procedure returns the interatomic distance between atom i
c     and atom j (retrieved from symm. matrix r)
c     
      rd(i,j)=abs(r((max0(i,j)*(max0(i,j)-1))/2+min0(i,j)))
c     
c     Initialize energy
c     
      e=0.0
c     
c     Compute out of plane deformations for third neighbour. Deformation is
c     defined as the angle between the plane of the first three atoms and the
c     bond formed by the first and fourth atom.
c     
      do 100 i=1,npla
         plad(i)=0.0
         i1=pla(1,i)
         i2=pla(2,i)
         i3=pla(3,i)
         i4=pla(4,i)
         do 10 j=1,3
            xj(j)=x(j,i2)-x(j,i1)
            xk(j)=x(j,i3)-x(j,i1)
            xl(j)=x(j,i4)-x(j,i1)
 10      continue
         ril=rd(i1,i4)
         t1=xj(2)*xk(3)-xj(3)*xk(2)
         t2=xj(3)*xk(1)-xj(1)*xk(3)
         t3=xj(1)*xk(2)-xj(2)*xk(1)
         tn=sqrt(t1*t1+t2*t2+t3*t3)
         if (tn.lt.1.e-6) go to 100
         t1=t1/tn
         t2=t2/tn
         t3=t3/tn
         s=(t1*xl(1)+t2*xl(2)+t3*xl(3))/ril
         s=amax1(-1.0,amin1(1.0,s))
         b=pihalf-acos(s)
         plad(i)=b
c     
c     Energy
c     
         eibend(i)=0.5*kpla(i)*b**2
         e=e+eibend(i)
         f=kpla(i)*b
c     
c     Gradient
c     
         if (1-s**2.gt.1.e-6) then
            fs=f/sqrt(1-s**2)
            dj1=fs*((xk(2)*xl(3)-xk(3)*xl(2))/ril
     *           -(xk(2)*t3-xk(3)*t2)*s)/tn
            dj2=fs*((xk(3)*xl(1)-xk(1)*xl(3))/ril
     *           -(xk(3)*t1-xk(1)*t3)*s)/tn
            dj3=fs*((xk(1)*xl(2)-xk(2)*xl(1))/ril
     *           -(xk(1)*t2-xk(2)*t1)*s)/tn
            dk1=fs*((xl(2)*xj(3)-xl(3)*xj(2))/ril
     *           -(t2*xj(3)-t3*xj(2))*s)/tn
            dk2=fs*((xl(3)*xj(1)-xl(1)*xj(3))/ril
     *           -(t3*xj(1)-t1*xj(3))*s)/tn
            dk3=fs*((xl(1)*xj(2)-xl(2)*xj(1))/ril
     *           -(t1*xj(2)-t2*xj(1))*s)/tn
            dl1=fs*(t1-s*xl(1)/ril)/ril
            dl2=fs*(t2-s*xl(2)/ril)/ril
            dl3=fs*(t3-s*xl(3)/ril)/ril
            gr(1,i1)=gr(1,i1)-dj1-dk1-dl1
            gr(2,i1)=gr(2,i1)-dj2-dk2-dl2
            gr(3,i1)=gr(3,i1)-dj3-dk3-dl3
            gr(1,i2)=gr(1,i2)+dj1
            gr(2,i2)=gr(2,i2)+dj2
            gr(3,i2)=gr(3,i2)+dj3
            gr(1,i3)=gr(1,i3)+dk1
            gr(2,i3)=gr(2,i3)+dk2
            gr(3,i3)=gr(3,i3)+dk3
            gr(1,i4)=gr(1,i4)+dl1
            gr(2,i4)=gr(2,i4)+dl2
            gr(3,i4)=gr(3,i4)+dl3
         end if
 100  continue
c     
c     done...
c     
      return
      end
