      subroutine egconto(x,e,gr)
c     
c     Compute torsional constraint energy and update gradient 'gr'. 
c     
      implicit none
      include 'sizes.h'
      include 'energy.h'
      real    x(3,*),gr(3,*),sixty
      parameter (sixty=3.141592653/3)

      integer i,j,i1,i2,i3,i4,l
      real r32,ta1,ta2,ta3,tb1,tb2,tb3,tan,tbn,sigma,sp,t,
     .     dl,dh,dmid,dt,e,f,s2,s3,f1,f2,f3,f4,rd
c     
c     The following procedure returns the interatomic distance between atom i
c     and atom j (retrieved from symm. matrix r)
c     
      rd(i,j)=abs(r((max0(i,j)*(max0(i,j)-1))/2+min0(i,j)))
c     
c     Compute torsion angles and directions of forces 'ta' and 'tb'
c     
      do 100 i=1,ncond
         i1=atcond(1,i)
         i2=atcond(2,i)
         i3=atcond(3,i)
         i4=atcond(4,i)
         r32=abs(rd(i3,i2))
         ta1=(x(2,i2)-x(2,i3))*(x(3,i2)-x(3,i1))-
     *        (x(3,i2)-x(3,i3))*(x(2,i2)-x(2,i1))
c     
         ta2=(x(3,i2)-x(3,i3))*(x(1,i2)-x(1,i1))-
     *        (x(1,i2)-x(1,i3))*(x(3,i2)-x(3,i1))
c     
         ta3=(x(1,i2)-x(1,i3))*(x(2,i2)-x(2,i1))-
     *        (x(2,i2)-x(2,i3))*(x(1,i2)-x(1,i1))
c     
         tb1=(x(2,i2)-x(2,i3))*(x(3,i3)-x(3,i4))-
     *        (x(3,i2)-x(3,i3))*(x(2,i3)-x(2,i4))
c     
         tb2=(x(3,i2)-x(3,i3))*(x(1,i3)-x(1,i4))-
     *        (x(1,i2)-x(1,i3))*(x(3,i3)-x(3,i4))
c     
         tb3=(x(1,i2)-x(1,i3))*(x(2,i3)-x(2,i4))-
     *        (x(2,i2)-x(2,i3))*(x(1,i3)-x(1,i4))
c     
         tan=sqrt(ta1*ta1+ta2*ta2+ta3*ta3)
         tbn=sqrt(tb1*tb1+tb2*tb2+tb3*tb3)
         if (tan.lt.1.e-6.or.tbn.lt.1.e-6) then
            dcond(i)=0.0
         else
            ta1=ta1/tan
            ta2=ta2/tan
            ta3=ta3/tan
            tb1=tb1/tbn
            tb2=tb2/tbn
            tb3=tb3/tbn
            sigma=sign(1.0,(ta2*tb3-ta3*tb2)*(x(1,i2)-x(1,i3))+
     *           (ta3*tb1-ta1*tb3)*(x(2,i2)-x(2,i3))+
     *           (ta1*tb2-ta2*tb1)*(x(3,i2)-x(3,i3)))
            sp=amax1(-1.0,amin1(1.0,ta1*tb1+ta2*tb2+ta3*tb3))
            t=-sigma*acos(sp)
            dcond(i)=t*57.29578
c     
c     normalize constraint...
c     
            if (dlowd(i).gt.0.0) then
               dl=dlowd(i)-360.0
               dh=dhighd(i)-360.0
            else
               dl=dlowd(i)
               dh=dhighd(i)
            end if
            dmid=(dl+dh+360.0)*0.5
            dt=0.0
            if (dcond(i).gt.dmid) then
               if (dcond(i).lt.dl+360.0) dt=dcond(i)-dl-360.0
               if (dcond(i).gt.dh+360.0) dt=dcond(i)-dh-360.0
            else
               if (dcond(i).lt.dl)       dt=dcond(i)-dl
               if (dcond(i).gt.dh)       dt=dcond(i)-dh
            end if
            dt=dt/57.29578
c     
c     Energy
c     
            epend(i)=0.5*kcond(i)*dt*dt
            e=e+epend(i)
            f=kcond(i)*dt
c     
c     Gradient
c     
            if (f.ne.0.0) then
               s2=0.0
               s3=0.0
               do 90 l=1,3
                  s2=s2+(x(l,i2)-x(l,i3))*(x(l,i2)-x(l,i1))
                  s3=s3+(x(l,i2)-x(l,i3))*(x(l,i3)-x(l,i4))
 90            continue
               f1=-f*r32/tan
               f2=f*s2/(tan*r32)
               f3=f*s3/(tbn*r32)
               f4=-f*r32/tbn
               gr(1,i1)=gr(1,i1)+f1*ta1
               gr(1,i2)=gr(1,i2)-(f1+f2)*ta1+f3*tb1
               gr(2,i1)=gr(2,i1)+f1*ta2
               gr(2,i2)=gr(2,i2)-(f1+f2)*ta2+f3*tb2
               gr(3,i1)=gr(3,i1)+f1*ta3
               gr(3,i2)=gr(3,i2)-(f1+f2)*ta3+f3*tb3
               gr(1,i3)=gr(1,i3)+(f4-f3)*tb1+f2*ta1
               gr(1,i4)=gr(1,i4)-f4*tb1
               gr(2,i3)=gr(2,i3)+(f4-f3)*tb2+f2*ta2
               gr(2,i4)=gr(2,i4)-f4*tb2
               gr(3,i3)=gr(3,i3)+(f4-f3)*tb3+f2*ta3
               gr(3,i4)=gr(3,i4)-f4*tb3
            end if
         end if
 100  continue
c     
c     done...
c     
      return
      end
