      subroutine egtors(x,ed,es,gr)
c     
c     Compute torsional energy and update gradient 'gr'. The torsional
c     energy is the sum of 'ed' (deformation) and 'es' (strain).
c     
      implicit none
      include 'sizes.h'
      include 'energy.h'
      real    x(3,*),gr(3,*)
      real sixty
      parameter (sixty=3.141592653/3)

      integer i1,i2,i3,i4,l,ik,i,j
      real es,r32,ta1,ta2,ta3,tb1,tb2,tb3,tan,tbn,sigma,sp,t,
     .     sint,cost,v,f,sin2t,cos2t,fd,s2,s3,f1,f2,f3,f4,ed,rd

c     
c     The following procedure returns the interatomic distance between atom i
c     and atom j (retrieved from symm. matrix r)
c     
      rd(i,j)=abs(r((max0(i,j)*(max0(i,j)-1))/2+min0(i,j)))
c     
c     Initialize energy
c     
      ed=0.0
      es=0.0
c     
c     Compute torsion angles and directions of forces 'ta' and 'tb'
c     
      do 100 i=1,ntor
         i1=tor(1,i)
         i2=tor(2,i)
         i3=tor(3,i)
         i4=tor(4,i)
         r32=rd(i3,i2)
         ta1=(x(2,i2)-x(2,i3))*(x(3,i2)-x(3,i1))-
     *        (x(3,i2)-x(3,i3))*(x(2,i2)-x(2,i1))
c     
         ta2=(x(3,i2)-x(3,i3))*(x(1,i2)-x(1,i1))-
     *        (x(1,i2)-x(1,i3))*(x(3,i2)-x(3,i1))
c     
         ta3=(x(1,i2)-x(1,i3))*(x(2,i2)-x(2,i1))-
     *        (x(2,i2)-x(2,i3))*(x(1,i2)-x(1,i1))
c     
         tb1=(x(2,i2)-x(2,i3))*(x(3,i3)-x(3,i4))-
     *        (x(3,i2)-x(3,i3))*(x(2,i3)-x(2,i4))
c     
         tb2=(x(3,i2)-x(3,i3))*(x(1,i3)-x(1,i4))-
     *        (x(1,i2)-x(1,i3))*(x(3,i3)-x(3,i4))
c     
         tb3=(x(1,i2)-x(1,i3))*(x(2,i3)-x(2,i4))-
     *        (x(2,i2)-x(2,i3))*(x(1,i3)-x(1,i4))
c     
         tan=sqrt(ta1*ta1+ta2*ta2+ta3*ta3)
         tbn=sqrt(tb1*tb1+tb2*tb2+tb3*tb3)
         if (tan.lt.1.e-6.or.tbn.lt.1.e-6) then
            tora(i)=0.0
         else
            ta1=ta1/tan
            ta2=ta2/tan
            ta3=ta3/tan
            tb1=tb1/tbn
            tb2=tb2/tbn
            tb3=tb3/tbn
            sigma=sign(1.0,(ta2*tb3-ta3*tb2)*(x(1,i2)-x(1,i3))+
     *           (ta3*tb1-ta1*tb3)*(x(2,i2)-x(2,i3))+
     *           (ta1*tb2-ta2*tb1)*(x(3,i2)-x(3,i3)))
            sp=amax1(-1.0,amin1(1.0,ta1*tb1+ta2*tb2+ta3*tb3))
            t=-sigma*acos(sp)
            tora(i)=t
c     
c     Energy
c     
            eitord(i)=0
            eitors(i)=0
            if (tortyp(i).eq.1.or.tortyp(i).eq.4) then
               sint=sin(t)
               cost=cos(t)
               v=0.75E-4*(kvor(i)+ktt(i)*sint**2)
               eitors(i)=v*sint**2
               f=sint*cost*(1.5E-4*ktt(i)*sint**2+2*v)
               if (tortyp(i).eq.1) then
                  sin2t=2*sint*cost
                  cos2t=cost*cost-sint*sint
                  eitord(i)=-0.25*kzt(i)*(1-cos2t)
                  fd=-0.5*kzt(i)*sin2t
                  f=f+fd
               end if
            else if (tortyp(i).eq.2) then
               eitord(i)=0.5*kzt(i)*(1+cos(3*t))
               f=-1.5*kzt(i)*sin(3*t)
            else if (tortyp(i).eq.3) then
               eitord(i)=0.5*kzt(i)*(1+cos(6*t))
               f=-3.0*kzt(i)*sin(6*t)
            else if (tortyp(i).eq.5) then
               eitord(i)=0.5*kzt(i)*(1+cos(2*t))
     .              +0.5*ktt(i)*(1+cos(3*t))
               f=-1.0*kzt(i)*sin(2*t)-1.5*ktt(i)*sin(3*t)
            else
               f=0.0
            end if
            ed=ed+eitord(i)
            es=es+eitors(i)
c     
c     Gradient
c     
            if (f.ne.0.0) then
               s2=0.0
               s3=0.0
               do 90 l=1,3
                  s2=s2+(x(l,i2)-x(l,i3))*(x(l,i2)-x(l,i1))
                  s3=s3+(x(l,i2)-x(l,i3))*(x(l,i3)-x(l,i4))
 90            continue
               f1=-f*r32/tan
               f2=f*s2/(tan*r32)
               f3=f*s3/(tbn*r32)
               f4=-f*r32/tbn
               gr(1,i1)=gr(1,i1)+f1*ta1
               gr(1,i2)=gr(1,i2)-(f1+f2)*ta1+f3*tb1
               gr(2,i1)=gr(2,i1)+f1*ta2
               gr(2,i2)=gr(2,i2)-(f1+f2)*ta2+f3*tb2
               gr(3,i1)=gr(3,i1)+f1*ta3
               gr(3,i2)=gr(3,i2)-(f1+f2)*ta3+f3*tb3
               gr(1,i3)=gr(1,i3)+(f4-f3)*tb1+f2*ta1
               gr(1,i4)=gr(1,i4)-f4*tb1
               gr(2,i3)=gr(2,i3)+(f4-f3)*tb2+f2*ta2
               gr(2,i4)=gr(2,i4)-f4*tb2
               gr(3,i3)=gr(3,i3)+(f4-f3)*tb3+f2*ta3
               gr(3,i4)=gr(3,i4)-f4*tb3
            end if
         end if
         if (tortyp(i).eq.1) then
            ik=(max0(i1,i4)*(max0(i1,i4)-1))/2+min0(i1,i4)
            r(ik)=-abs(r(ik))
         end if
 100  continue
c     
c     done...
c     
      return
      end
