      subroutine egvdw(x,e,gr)
c     
c     Compute van der Waals energy 'e' for the molecule whose coordinates
c     are stored in x (n atoms). At the same time the gradient resulting from
c     the van der Waals potentials will be added to 'gr'.
c     
      implicit none
      include 'sizes.h'
      include 'energy.h'
      include 'atom.h'
      include 'bpara.h'
      real    x(3,*),gr(3,*),e
      integer i,j,ij

      real rij,ref,dr,a,c,v,f
      integer ni,nj
      
c     
c     Initialize energy
c     
      e=0.0
c     
c     Compute van der Waals energy for each pair of non-bonded atoms.
c     r(i,j)<0 is interpreted as a flag to exclude this pair from the
c     energy- and gradient calculation (1-3, 1-4, hydrogenbonds ...).
c     
      do 100 i=2,nat
         ij=(i*(i-1))/2
         do 90 j=1,i-1
            ij=ij+1
            rij=r(ij)
            if (rij.gt.0.0) then 
               ref=amax1(1.0,rij) ! If atoms are very close,
               dr=rij-ref       ! a linear extrapolation will
               ni=ntype(i)      ! be used.
               nj=ntype(j)
               a=avdw(ni,nj)*exp(-bvdw(ni,nj)*ref)
               c=cvdw(ni,nj)/ref**6
               v=ref**(-dvdw(ni,nj))
               f=0.6943E-5*(-bvdw(ni,nj)*a*v-dvdw(ni,nj)*a*v/ref
     *              +6*c/ref)
               e=e+0.6943E-5*(a*v-c)+f*dr
c     
c     Gradient
c     
               f=f/rij
               gr(1,i)=gr(1,i)+f*(x(1,i)-x(1,j))
               gr(1,j)=gr(1,j)-f*(x(1,i)-x(1,j))
               gr(2,i)=gr(2,i)+f*(x(2,i)-x(2,j))
               gr(2,j)=gr(2,j)-f*(x(2,i)-x(2,j))
               gr(3,i)=gr(3,i)+f*(x(3,i)-x(3,j))
               gr(3,j)=gr(3,j)-f*(x(3,i)-x(3,j))
            end if
 90      continue
 100  continue
c     
c     done...
c     
      return
      end
