      subroutine flexigeom(t,x,y)
c     
c     Compute new geometry of flexible molecule. 
c     
c     t   : angle of rotation about flexible bonds, real(*), input
c     x   : conformation corresponding to t=0, real(3,*), input
c     y   : new conformation, after rotation, real(3,*), output
c     
c     The dimensions of t, x, and y must be at least nrotb, nat, and
c     nat (resp.). x and y may be equivalenced.
c     
      implicit none
      include 'sizes.h'
      include 'atom.h'
      include 'tree.h'
c     
      integer i,j,k,l,tr,ptr,i1,i2,i3
      real    t(*),x(3,*),y(3,*),rm(4,4),yk(3)
c     
c     Copy coordinates if there are no flexible bonds.
c     
      if (nrotb.eq.0) then
         do j=1,nat
            do i=1,3
               y(i,j)=x(i,j)
            end do
         end do
         return
      end if
c     
c     Mark all transforms invalid
c     
      do i=1,nrotb
         vflag(i)=.false.
      end do
c     
c     Transform atoms in the sequence defined by their depth first numbers
c     
      do i=1,nat
         k=dfn(i)
         if (trnr(k).ne.0) then
            tr=trnr(k)
            if (vflag(tr)) then
c     
c     Old transform is valid, use it.
c     
               do j=1,3
                  yk(j)=brotm(j,4,tr)
                  do l=1,3
                     yk(j)=yk(j)+brotm(j,l,tr)*x(l,k)
                  end do
               end do
               do j=1,3
                  y(j,k)=yk(j)
               end do
            else
c     
c     New transform. Build homogenous transformation matrix.
c     First apply parent-transformation to second atom of bond (k).
c     
               if (ptra(tr).ne.0) then
                  ptr=ptra(tr)
                  do i1=1,3
                     yk(i1)=brotm(i1,4,ptr)
                     do i2=1,3
                        yk(i1)=yk(i1)+brotm(i1,i2,ptr)*x(i2,k)
                     end do
                  end do
                  do i1=1,3
                     y(i1,k)=yk(i1)
                  end do
               else
                  do i1=1,3
                     y(i1,k)=x(i1,k)
                  end do
               end if
c     
c     Now determine new matrix.
c     
               call rotax(y(1,axis(1,tr)),y(1,k),t(tr),rm)
               if (ptra(tr).ne.0) then
                  do i1=1,4
                     do i2=1,4
                        brotm(i1,i2,tr)=0.0
                        do i3=1,4
                           brotm(i1,i2,tr)=brotm(i1,i2,tr)+
     *                          rm(i1,i3)*brotm(i3,i2,ptra(tr))
                        end do
                     end do
                  end do
               else
                  do i1=1,4
                     do i2=1,4
                        brotm(i1,i2,tr)=rm(i1,i2)
                     end do
                  end do
               end if
               vflag(tr)=.true.
            end if
         else
c     
c     Transform is 0 (=identity)
c     
            do j=1,3
               y(j,k)=x(j,k)
            end do
         end if
      end do
c     
c     done...
c     
      return
      end     
