      subroutine flexitree(*)
c     
c     Build tree structure for flexible bond optimization. Also count the number
c     of connected components and assign the component number of each atom to
c     fragnr(i). Take alternate return if the flexible bonds contain ring bonds.
c     
      implicit none
      include 'sizes.h'
c     
      integer mimat,mscratch,iaf,iar,iai,ias
      parameter (mimat=2*mbtot+matot)
      parameter (mscratch=mbtot+matot)
      parameter (iaf=1)
      parameter (iar=iaf+matot)
      parameter (iai=iar+matot)
      parameter (ias=iai+mimat)
c     
      include 'atom.h'
      include 'energy.h'
      include 'tree.h'
      include 'worksp.h'
      include 'fragment.h'
c     
      integer i,kfrag,k,father(matot),rp(matot),imat(mimat),
     *     scratch(mscratch),from,to,tr
      logical found
      integer ia
c     
      equivalence (father,iwork(iaf)),(rp,iwork(iar)),
     *     (imat,iwork(iai)),(scratch,iwork(ias))
c     
c     Build incidence matrix from bondlist and perform depth first search
c     
      call ge2imat(nbnd,bnd,nat,rp,imat)
      call gdfs(nat,rp,imat,nbnd,bnd,anchor,dfn,father,scratch)
c     
c     Sort atomnumbers by dfn (ascending)
c     
      do i=1,nat
         scratch(i)=dfn(i)
         dfn(i)=i
      end do
      call shellsort(nat,scratch,dfn)
c     
c     Determine transformation numbers and orient rotatable bonds
c     
      kfrag=0
      do k=1,nrotb
         ptra(k)=0
      end do
      from=0
      tr=0
      i=0
      do while (i.lt.nat)
         i=i+1
         if (from.eq.0) kfrag=kfrag+1
         to=dfn(i)            
         fragnr(to)=kfrag
         if (from.eq.father(to)) then
            trnr(to)=tr
            k=0
            found=.false.
            do while(k.lt.nrotb.and..not.found)
               k=k+1
               if (from.eq.axis(1,k).and.to.eq.axis(2,k)) then
                  tr=k
                  ptra(tr)=trnr(to)
                  trnr(to)=tr
                  found=.true.
               else if (from.eq.axis(2,k).and.to.eq.axis(1,k)) then
                  tr=k
                  ia=axis(2,k)
                  axis(2,k)=axis(1,k)
                  axis(1,k)=ia
                  ptra(tr)=trnr(to)
                  trnr(to)=tr
                  found=.true.
               end if
            end do
            from=to
         else
            do while (from.ne.father(to))
               from=father(from)
            end do
            if (from.ne.0) then
               tr=trnr(from)
            else
               tr=0
            end if
            i=i-1
         end if
      end do
c     
c     Check against ring bonds in flexible optimization
c     
      do i=1,nrotb
         vflag(i)=.false.
      end do
      do i=1,nat
         if (trnr(i).ne.0) vflag(trnr(i))=.true.
      end do
      do i=1,nrotb
         if (.not.vflag(i)) then
            write(6,'(a)') ' *** ERROR *** Ring bond in '
     *           //'/FLEXIBLE optimization.'
            write(21,'(a)') ' *** ERROR *** Ring bond in '
     *           //'/FLEXIBLE optimization.'
            return 1
         end if
      end do
c     
c     done...
c     
      return
      end
