      subroutine gdfs(n,rp,imat,m,e,root,dfn,father,work)
c     
c     Generation of depth first forest of an undirected graph.
c     (Swamy&Thulasiraman, Graphs, Networks and Alogorithms, p.452, modified)
c     
c     n     : number of vertices, integer, input
c     rp    : row pointers of incidence matrix (row-packed), integer (n), input
c     imat  : rowise packed incidence matrix, integer (m+n), input
c     m     : number of edges, integer, input
c     e     : edges, integer (2,m), input
c     root  : root vertex of tree to generate, integer, input
c     dfn   : depth first number, integer (n), output
c     father: father of node in dft, integer (n), output
c     work  : workspace for marks of visited verices and examined edges.
c     
      implicit none
      integer rp(*),imat(*),e(2,*),dfn(*),father(*),work(*)
      integer r,i,k,n,m,root,v,l,s,w
c     
c     Initialize
c     
      do k=1,n                  ! label all vertices "unvisited"
         father(k)=0
         work(k)=0
      end do
      do i=n+1,n+m              ! label all edges "unexamined"
         work(i)=0    
      end do
      r=max0(1,min0(n,root))
      i=1
c     
c     Start on new connected component: Find unmarked vertex r.
c     
 200  continue
      k=0
      do while (k.lt.n.and.r.eq.0)
         k=k+1
         if (work(k).eq.0) r=k
      end do
      if (r.eq.0) go to 1000    ! Done. No more unmarked vertices.
      dfn(r)=i
      work(r)=1
      v=r
      r=0
c     
c     Find unexamined edge incident on v
c     
 300  continue
      l=rp(v)+1
      k=0
      do s=l,l+imat(rp(v))-1
         if (work(imat(s)+n).eq.0) k=imat(s)
      end do
      if (k.eq.0) go to 500
      work(n+k)=1
      if (e(1,k).eq.v) then
         w=e(2,k)
      else
         w=e(1,k)
      end if
c     
c     Update DFN and visit son.
c     
      if (work(w).eq.0) then
         i=i+1
         dfn(w)=i
         father(w)=v
         work(w)=1
         v=w
      end if
      go to 300
c     
c     Are we done with this connected component ?
c     
 500  continue
      if (father(v).ne.0) then
         v=father(v)
         go to 300
      else
         go to 200
      end if
c     
c     Done (for all connected component)
c     
 1000 continue
      return
      end
