      subroutine ge2imat(m,e,n,rp,imat)
c     
c     Generate incidence matrix (packed form) from edgelist of an undirected
c     graph.
c     
c     m   : number of edges, integer, input
c     e   : edges, integer (2,m), input
c     n   : number of nodes, integer, input
c     rp  : row pointer, integer (n), output
c     imat: incidenxe matrix, integer (n+2*m), output
c     
c     The resulting incidence matrix is stored in rowise (=nodewise) packed form.
c     rp(i) is a pointer to the first entry of node i in imat. The entries of
c     imat consist of the number of edges incident to this node (first entry),
c     followed by the edgenumbers of these edges.
c     
      implicit none
      integer e(2,*),rp(*),imat(*)
      integer n,m,i,j,k,l
c     
c     First pass over list of edges to count number of edges incident to each
c     node:
c     
      do i=1,n
         rp(i)=0
      end do
      do j=1,m
         rp(e(1,j))=rp(e(1,j))+1
         rp(e(2,j))=rp(e(2,j))+1
      end do
c     
c     Initialize the incidencematrix imat and set up rowpointers rp
c     
      k=1
      do i=1,n
         imat(k)=0
         l=k
         k=k+rp(i)+1
         rp(i)=l
      end do
c     
c     Second pass over edgelist to build the incidence matrix
c     
      do j=1,m
         do k=1,2
            l=e(k,j)
            imat(rp(l))=imat(rp(l))+1
            imat(rp(l)+imat(rp(l)))=j
         end do
      end do
c     
c     Done...
c     
      return
      end
