      subroutine getfrag(how)
c     
c     Get definition of rigid fragments (provisional version).
c     The method of fragmentation is given by how:
c     how='FLEX' : fragments are connected components,
c     ='FRAG' : fragments are defined explicitly (nfrag>0) or
c     implicitly through flexible bond transformations (nfrag=0).
c     
      implicit none
      include 'sizes.h'
      include 'atom.h'
      include 'fragment.h'
      include 'tree.h'
      include 'worksp.h'
c     
      character*8 t
      character*4 how
      integer i,nt
c     
      if (how.eq.'FLEX') then
c     
c     Use component number as fragment number
c     
         do i=1,nat
            iwork(i)=fragnr(i)
            fragnr(i)=i
         end do
      else if (how.eq.'FRAG'.and.nfrag.eq.0) then
c     
c     Use flexi-bond transform number as fragment number
c     
         do i=1,nat
            iwork(i)=trnr(i)+1000*fragnr(i)
            fragnr(i)=i
         end do
c     
      else
c     
c     Use atomlabels to identify fragments
c     
         do i=1,nat
            iwork(i)=0
            fragnr(i)=i
            call strupcase(t,name(i))
            nt=0
            do while (nt.lt.nfrag.and.iwork(i).eq.0)
               nt=nt+1
               if (t.eq.lfrag(nt)) iwork(i)=nt
            end do
         end do
      end if
      call shellsort(nat,iwork,fragnr)
c     
c     Assign fragment number 0 to single atoms
c     
      do i=2,nat-1
         if (iwork(i-1).ne.iwork(i).and.
     *        iwork(i).ne.iwork(i+1)     ) iwork(i)=0
      end do
      if (iwork(1).ne.iwork(2)) iwork(1)=0
      if (iwork(nat-1).ne.iwork(nat)) iwork(nat)=0
      call shellsort(nat,iwork,fragnr)
c     
c     Setup fragment pointers
c     
      if (iwork(1).eq.0) then
         nfrag=0
      else
         nfrag=1
      end if
      fragpt(0)=0
      do i=1,nat-1
         if (iwork(i).ne.iwork(i+1)) then
            fragpt(nfrag)=i
            nfrag=nfrag+1
         end if
      end do
      fragpt(nfrag)=nat
c     
c     Determine fragment sizes from fragment pointers
c     
      nsingle=fragpt(0)
      fragsize(0)=nsingle
      do i=1,nfrag
         fragsize(i)=fragpt(i)-fragpt(i-1)
      end do     
c     
c     Remove trivial fragment coordinates
c     
      if (nsingle.eq.0.and.nfrag.eq.1) nfrag=0
c     
c     done...
c     
      return
      end
