      subroutine input(*)
c     
c     Read Inputdata for PIFF
c     
      implicit none
#include "sizes.h"
#include "apara.h"
#include "bpara.h"
#include "atom.h"
#include "option.h"
#include "energy.h"
#include "tree.h"
#include "fragment.h"
c     
      character*300 cmdlin
      logical chrgok,constr,coulomb
      integer*4  status,i,iargc,argc,nq
c     
      character   extfil*132,string*80,edit*80
c     
c     Setup options
c     
      molfile=' '
      interact=.FALSE.
      check_bonds=.true.
      negr=-1
      mxscf=-1
      mcyc=10
      agr=0.00035
      ascf=1.0e-4
      nenout=10
      allcnv=0.01
      edit=' '
      dielc=1.0
      coulomb=.true.
      do_scf=.true.
      flexible=.false.
      nfrag=0
      nsingle=0
      deexp=1.0e-5
      scf_out=.false.
      no_output=.false.
      constr=.false.
c       
c     Parse command line
c     
      argc=iargc()
      i=0
      do while(i.lt.argc)
         i=i+1
         call getarg(i,cmdlin)
         if (cmdlin(1:1).eq.'-') then
            if (cmdlin(2:2).eq.'B') then
               check_bonds=.false.
            else if (cmdlin(2:2).eq.'i')  then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) negr
            else if (cmdlin(2:2).eq.'S') then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) mxscf
            else if (cmdlin(2:2).eq.'I') then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) mcyc
            else if (cmdlin(2:2).eq.'X') then
               interact=.TRUE.
            else if (cmdlin(2:2).eq.'g') then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) agr
            else if (cmdlin(2:2).eq.'r')  then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) nenout
            else if (cmdlin(2:2).eq.'G') then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               read(cmdlin,*) allcnv
            else if (cmdlin(2:2).eq.'E') then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               edit=cmdlin(1:len(edit))
            else if (cmdlin(2:2).eq.'c') then
               coulomb=cmdlin(3:3).ne.'-'
            else if (cmdlin(2:2).eq.'H') then
               do_scf=.false.
            else if (cmdlin(2:3).eq.'Os') then
               scf_out=.true.
            else if (cmdlin(2:3).eq.'O0') then
               no_output=.true.
            else if (cmdlin(2:2).eq.'C') then
               constr=.true.
            else if (cmdlin(2:2).eq.'f') then
               if (i.eq.argc) go to 8400
               i=i+1
               call getarg(i,cmdlin)
               nfrag=nfrag+1
               lfrag(nfrag)=cmdlin(1:len(lfrag(nfrag)))
            else if (cmdlin(2:2).eq.'F') then
               flexible=.true.
            else
               write(*,*) 'Error: unknown option '//cmdlin(1:10)
               return 1
            end if
         else
            molfile=cmdlin(1:len(molfile))
         end if
      end do
      if (molfile.eq.' ') then
         write(*,'(a,$)') 'Molecule: '
         read(*,'(a)') molfile
      end if
c     
c     Retrieve molecular description
c     
      if (interact) then
         open(unit=1,file=molfile,status='old',err=8100)
         call read_int(1,nat,name,xyz,ntype,chrg,chrgok,
     .        nbnd,bnd,*8200)
      else
         open(unit=1,file=extfil(molfile,'.min',.false.),
     .        status='old',err=8100)
         call molrd(1,status)
         if (status.ne.0) then
            write(6,'(a,i3)') ' *** ERROR reading .MOL file.',status
            return 1
         end if
         call getmol(nat,name,xyz,ntype,chrg,chrgok,nbnd,bnd,*8200)
      end if
      if (nat.lt.2) then
         write(6,'(a)') ' Less than two non-dummy atoms.'
         return 1
      end if
      call getrbond(nrotb,axis,anchor)
      do i=1,nat
         ctype(i)=type(ntype(i))
      end do
      close(unit=1)
c     
c     Set structure dependent options
c     
      agr=agr*1.0e-3
      if (negr.eq.-1) negr=3*nat*5
      if (mxscf.eq.-1) mxscf=3*nat
      if (edit.ne.' ') call edtpar(string,*8200)
      if (flexible) then
         if (anchor.eq.0) anchor=axis(1,1)
         if (nrotb.gt.mrotb) then
            write(6,'(a)') ' *** ERROR *** Too many flexible bonds.'
            return 1
         else if (nrotb.eq.0) then
            write(6,'(a)') ' *** ERROR *** No flexible bonds '
     .           //'for /FLEXIBLE optimization.'
            return 1
         end if
         call flexitree(*8200)
         call getfrag('FLEX')
         flexible=.true.
      end if
      fragment=nfrag.gt.0
      if (fragment) then
         call getfrag('FRAG')
      end if
      if (.not.flexible) nrotb=0
      nq=nrotb+3*nsingle+6*nfrag
      if (nq.gt.3*matot) then
         write(6,'(a)') ' *** ERROR : too many gen. coordinates.'
         return 1
      end if
      if (nq.lt.1.and.(fragment.or.flexible)) then
         mcyc=0
         write(6,'(a)') ' *** Strange fragment definition,'//
     .        ' nothing left to optimize.'
      end if
      if (.not.chrgok) coulomb=.false.
      if (.not.coulomb) dielc=0.0
      if (constr) then
         open(unit=1,file=extfil(molfile,'.con',.true.),status='old',
     .        err=8300)
         call readcon(*8200)
      end if
c     
c     Consistency check.
c     
      if (fragment.and.flexible) then
         write(6,'(a)')
     .        ' *** Conflicting qualifiers: /Flexible /Fragment'
         return 1
      end if
c     
c     Open outputfile
c     
      if (interact) no_output=.TRUE.
      if (no_output) then
#ifdef mingw
         open(unit=21,file='nul',status='UNKNOWN')
#else
         open(unit=21,file='/dev/null',status='UNKNOWN')
#endif
      else
         open(unit=21,file=extfil(molfile,'.dat',.true.),
     .        status='UNKNOWN')
      end if
c     
      return
c     
c     Errors:
c     
 8100 write(6,'(a)') ' Can''t find molecule '//molfile
      return 1
 8200 return 1
 8300 write(6,'(a)') ' Can''t find constraints...'
      return 1
 8400 write(6,'(a)') ' Missing argument in option: '//cmdlin(1:10)
      return 1
      end
      
