c     
c     Functions to evaluate overlap integrals. These functions are copies
c     of the ones used in PIMM.
c     
      real function ovlps(rr,za,zb)
c     
c     This function evaluates overlap integrals (2p sigma/2p sigma)
c     using functions oa and ob
c     
      implicit none
      real a(5),b(5),rr,za,zb,sa,sb,alpha,beta,rab,w,oa,ob
      integer i,n

      sa=0.5*za
      sb=0.5*zb
      rab=rr/0.52917706
      alpha=0.5*rab*(sa+sb)
      beta=0.5*rab*(sb-sa)
      do 3 i=1,5
         n=i-1
         a(i)=oa(alpha,n)
         b(i)=ob(beta,n)
 3    continue
      w=-sqrt((sa*sb)**5)*(rab**5)*0.0625
      ovlps=w*(b(3)*(a(5)+a(1))-a(3)*(b(5)+b(1)))
      return
      end
c     
      real function ovlp(rr,za,zb)
c     
c     This function evaluates overlap integrals (2p pi/2p pi) using
c     functions oa and ob
c     
      implicit none
      real a(5),b(5),rr,za,zb,sa,sb,rab,alpha,beta,oa,ob,w
      integer n,i

      sa=0.5*za
      sb=0.5*zb
      rab=rr/0.52917706
      alpha=0.5*rab*(sa+sb)
      beta=0.5*rab*(sb-sa)
      do 3 i=1,5
         n=i-1
         a(i)=oa(alpha,n)
         b(i)=ob(beta,n)
 3    continue
      w=sqrt((sa*sb)**5)*(rab**5)*0.0625
      ovlp=0.5*w*(a(5)*(b(1)-b(3))-b(5)*(a(1)-a(3))-a(3)*b(1)+b(3)*a(1))
      return
      end
c     
      real function oa(a,k)
c     
c     This function calculates the  a  integrals associated with the
c     overlap routine listed above.
c     
      integer k,l,m
      real s,b,a

      b=1.0/a
      s=1.0
      oa=1.0
      do 1 m=1,k
         l=k-m+1
         s=l*s*b
         oa=oa+s
 1    continue
      oa=oa*b*exp(-a)
      return
c     
      end
      real function ob(beta,n)
c     
c     This function calculates the b integrals associated with the
c     overlap routine listed above.
c     
      implicit none
      integer n
      real beta,b,fn,fnumer,sum,factor,fi,denom,t,tsum

      b=beta**2
      fn=n
      if (mod(n,2).ne.0) then
         fnumer=fn+2.0
         sum=beta/fnumer
         factor=-2.0
         fi=3.0
      else
         fnumer=fn+1.
         sum=1./fnumer
         factor=2.0
         fi=2.0
      end if
      t=sum
 10   continue
      denom=fnumer+2.0
      t=((((t/fi)*b)/(fi-1.0))*fnumer)/denom
      tsum=sum+t
      if (sum-tsum.eq.0) go to 20
      sum=tsum
      fi=fi+2.0
      fnumer=denom
      go to 10
 20   continue
      ob=factor*sum
      return
      end
