      Program PIFF
c     
c     Pi-SCF / Force Field Program
c     
      implicit none
      include 'sizes.h'
      include 'apara.h'
      include 'bpara.h'
      include 'atom.h'
      include 'energy.h'
      include 'fock.h'
      include 'option.h'
      include 'fragment.h'
      include 'tree.h'
c     
      logical cnv
      real hofold,emmo,emm,etot
      integer niter
c     
c     Read molecule
c     
      call input(*8000)
      if (nat.eq.0) go to 8000
      call chkpar(*8000)
c     
c     Analysis of initial structure
c     
      call topol(*8000)
c     
c     Select method
c     
      if (nrotb+3*nsingle+6*nfrag.gt.0) then
         method=2               ! use generalized coordinates
      else
         method=1               ! use cartesian coordinates
      end if
      do_scf=do_scf.and.natpi.gt.0
         if (natpi.eq.0) go to 50 ! no pi-SCF needed.
c     
c     Hueckel analysis to obtain first bondordermatrix for subsequent Pople-SCF
c     
         call hueckl(*8000)
         if (.not.do_scf) go to 50
c     
c     Pariser-Parr-Pople SCF (initial)
c     
         call betas(*8000)
         call pople(*8000)
c     
c     Set up forcefield for subsequent MM optimization
c     
 50      continue
         if (nenout.gt.0) 
     *        write(6,'(/a,a//a)') ' Pi-SCF Force Field for ',molfile,
     *        ' Iter     Total      Bond     Angle   Torsion'//
     *        '     Plane   vdWaals   Coulomb'
         call esetup
         ncalls=0
         call eout
         call moldmp
         if (scf_out.and.do_scf) call scfdmp
         niter=0
         hofold=hofm
         emmo=ebnd+eang+etord+etors+epla+evdw+ecou
         cnv=.false.
         if (.not.do_scf) mcyc=min0(mcyc,1)
c     
c     Do MM
c     
         do while (niter.lt.mcyc.and..not.cnv)
            call mmcalc
            if (do_scf) then
               call betas(*8000)
               call pople(*8000)
               call esetup
            end if
            niter=niter+1
            if (nenout.gt.0.and.do_scf)
     *           write(6,'(a,i3,a)')
     *           ' ---------------------------- End of cycle',niter,
     *           ' -----------------------------'
            call eout
            emm=ebnd+eang+etord+etors+epla+evdw+ecou
            cnv=abs(emm-emmo).le.allcnv
     *           .and.abs(hofold-hofm).lt.5*allcnv
            emmo=emm
            hofold=hofm
         end do
c     
c     Report energy of final cycle
c     
         if (nenout.gt.0.and.mcyc.gt.0) then
            etot=ebnd+eang+etord+etors+epla+evdw+ecou+econ
            write(6,'(i5,7(f10.2))') ncalls,
     *           etot,
     *           ebnd,
     *           eang,
     *           (etord+etors),
     *           epla,
     *           evdw,
     *           ecou
            write(6,'(a)') ' (Energies are in kJ/mole)'
         end if
         if (nenout.gt.0.and..not.interact.and.do_scf) then
            write(6,'(/a,f10.2,a,f10.2,a/)') ' MM/SCF Energy: ',
     .           hofm,' kJ/mole, ',hofm/4.186,' kcal/mole.'
         end if
c     
c     Write new coordinates to file
c     
         if (mcyc.gt.0) then
            call moldmp
            if (scf_out.and.do_scf) call scfdmp
            if (interact) call xputmol(nat,xyz)
            if (.not.interact) call putmol(nat,xyz)
         end if
c     
c     Report status
c     
         if (do_scf) then
            if (scfcon) then
               write(21,'(a)') '   Final SCF converged.'
               write(6,'(a)')  '   Final SCF converged.'
            else
               write(21,'(a)') '   Final SCF not converged.'
               write(6,'(a)')  '   Final SCF not converged.'
            end if
            if (mmcon) then
               write(21,'(a)') '   Final MM  converged.'
               write(6,'(a,f10.6)')  
     .              '   Final MM  converged, r.m.s. gradient = ',rmsG
            else
               write(21,'(a)') '   Final MM  not converged.'
               write(6,'(a,f10.6)')  
     .              'Final MM  not converged, r.m.s. gradient = ',rmsG
            end if
            if (cnv) then
               write(21,'(a)') '   MM/pi-SCF Selfconsistency.'
               write( 6,'(a)') '   MM/pi-SCF Selfconsistency.'
            else
               write(21,'(a)') '   No MM/pi-SCF Selfconsistency.'
               write( 6,'(a)') '   No MM/pi-SCF Selfconsistency.'
            end if
         else
            if (mmcon) then
               write(21,'(a)') '   Molecular Mechanics converged.'
               write( 6,'(a,f10.6)') 
     .              '   Molecular Mechanics converged, '
     .              //'r.m.s. gradient = ',rmsG
            else
               write(21,'(a)') '   Molecular Mechanics not converged.'
               write( 6,'(a,f10.6)') 
     .              '   Molecular Mechanics not converged, '
     .              //'r.m.s. gradient = ',rmsG
            end if
         end if
         if (.not.interact) write(6,'(/)')
         if (nenout.gt.0.and.interact.and.do_scf) then
            write(6,'(/a,f10.2,a,f10.2,a)') '   MM/SCF Energy: ',
     .           hofm,' kJ/mole, ',hofm/4.186,' kcal/mole.'
         end if
c     
c     That's as far as this version of PIFF will go...
c     
 8000    continue
         call exit(0)
         end
