      subroutine rotax(a,b,tau,rot)
c     
c     Build 4x4 matrix of clockwise rotation about axis a-->b
c     by angle tau (radians). Result is a homogenous transformation
c     matrix stored in rot.
c     
      implicit none
      real a(3),b(3),e(3),tau,rot(4,4),ct,st,s
      integer i,j
c     
      ct=cos(tau)
      st=sin(tau)
c     
c     Compute unit vector e in the direction of a-->b. If a-->b has length
c     zero, assume e = (1,1,1)/sqrt(3).
c     
      s=0.0
      do i=1,3
         e(i)=b(i)-a(i)
         s=s+e(i)*e(i)
      end do
      if (s.gt.0) then
         do i=1,3
            e(i)=e(i)/sqrt(s)
         end do
      else
         do i=1,3
            e(i)=sqrt(1/3.0)
         end do
      end if
c     
c     Compute 3x3 rotation matrix
c     
      rot(1,1)=e(1)*e(1)+(1.0-e(1)*e(1))*ct
      rot(2,2)=e(2)*e(2)+(1.0-e(2)*e(2))*ct
      rot(3,3)=e(3)*e(3)+(1.0-e(3)*e(3))*ct
c     
      rot(1,2)=e(1)*e(2)*(1-ct)-e(3)*st
      rot(2,3)=e(2)*e(3)*(1-ct)-e(1)*st
      rot(3,1)=e(3)*e(1)*(1-ct)-e(2)*st
      rot(2,1)=e(1)*e(2)*(1-ct)+e(3)*st
      rot(3,2)=e(2)*e(3)*(1-ct)+e(1)*st
      rot(1,3)=e(3)*e(1)*(1-ct)+e(2)*st
c     
c     Add translation
c     
      do i=1,3
         rot(i,4)=a(i)
         do j=1,3
            rot(i,4)=rot(i,4)-rot(i,j)*a(j)
         end do
         rot(4,i)=0.0
      end do
c     
c     Add scale (always 1)
c     
      rot(4,4)=1.0
c     
c     done...
c     
      return
      end
