      subroutine topol(*)
c     
c     Analysis of topology and pi-electron system of initial structure
c     
      implicit none
      include 'sizes.h'
      include 'apara.h'
      include 'bpara.h'
      include 'atom.h'
      include 'energy.h'

      integer i,j,k,l,nn,nbad
c     
c     Count number of atoms in pi-system, number of pi-electrons
c     
      nelpi=0
      natpi=0
      do 10 i=1,nat
         if (npiel(ntype(i)).gt.0) then
            natpi=natpi+1
            if (natpi.gt.mpiat) then
               write(6,'(a)') ' Too many atoms in pi-system'
               return 1
            end if
            atnr(natpi)=i
            pinr(i)=natpi
            nelpi=nelpi+npiel(ntype(i))
         else
            pinr(i)=0
         end if
 10   continue
c     
c     Abort if this is not a closed shell job.
c     
      if (mod(nelpi,2).ne.0) then
         write(6,'('' Not closed shell. Aborted...'')')
         return 1
      end if
c     
c     Check bondlist and compute bondlengths
c     
      nbad=0
      do j=1,nbnd
         k=bnd(1,j)
         l=bnd(2,j)
         if (al(ntype(k),ntype(l)).lt.0.001) then
            nbad=nbad+1
            write(6,'(a,2i4)') ' No parameters for bond',k,l
         else
            bndl(j)=sqrt((xyz(1,k)-xyz(1,l))**2+
     *           (xyz(2,k)-xyz(2,l))**2+(xyz(3,k)-xyz(3,l))**2)
         end if
      end do
      if (nbad.gt.0) return 1
c     
c     Create list of bondangles
c     
      nang=0
      do 200 i=1,nbnd-1
         do 190 j=i+1,nbnd
            if (bnd(1,i).eq.bnd(1,j)) then
               nang=nang+1
               if (nang.le.mantot) then
                  ang(1,nang)=bnd(2,i)
                  ang(2,nang)=bnd(1,i)
                  ang(3,nang)=bnd(2,j)
               end if
            else if (bnd(1,i).eq.bnd(2,j)) then
               nang=nang+1
               if (nang.le.mantot) then
                  ang(1,nang)=bnd(2,i)
                  ang(2,nang)=bnd(1,i)
                  ang(3,nang)=bnd(1,j)
               end if
            else if (bnd(2,i).eq.bnd(1,j)) then
               nang=nang+1
               if (nang.le.mantot) then
                  ang(1,nang)=bnd(1,i)
                  ang(2,nang)=bnd(2,i)
                  ang(3,nang)=bnd(2,j)
               end if
            else if (bnd(2,i).eq.bnd(2,j)) then
               nang=nang+1
               if (nang.le.mantot) then
                  ang(1,nang)=bnd(1,i)
                  ang(2,nang)=bnd(2,i)
                  ang(3,nang)=bnd(1,j)
               end if
            end if
 190     continue
 200  continue
      if (nang.gt.mantot) then
         write(6,'(a)') ' Too many bond angles.'
         return 1
      end if
c     
c     Create list of torsion angles
c     
      ntor=0
      do 300 i=1,nang-1
         do 290 j=i+1,nang
            if (ang(1,i).eq.ang(2,j)) then
               if (ang(2,i).eq.ang(3,j)) then
                  if (ang(1,j).ne.ang(3,i)) then
                     ntor=ntor+1
                     if (ntor.le.mtotot) then
                        tor(1,ntor)=ang(1,j)
                        tor(2,ntor)=ang(1,i)
                        tor(3,ntor)=ang(2,i)
                        tor(4,ntor)=ang(3,i)
                     end if
                  end if
               else if (ang(2,i).eq.ang(1,j)) then
                  if (ang(3,j).ne.ang(3,i)) then
                     ntor=ntor+1
                     if (ntor.le.mtotot) then
                        tor(1,ntor)=ang(3,j)
                        tor(2,ntor)=ang(1,i)
                        tor(3,ntor)=ang(2,i)
                        tor(4,ntor)=ang(3,i)
                     end if
                  end if
               end if
            else if (ang(3,i).eq.ang(2,j)) then
               if (ang(2,i).eq.ang(1,j)) then
                  if (ang(1,i).ne.ang(3,j)) then
                     ntor=ntor+1
                     if (ntor.le.mtotot) then
                        tor(1,ntor)=ang(1,i)
                        tor(2,ntor)=ang(2,i)
                        tor(3,ntor)=ang(3,i)
                        tor(4,ntor)=ang(3,j)
                     end if
                  end if
               else if (ang(2,i).eq.ang(3,j)) then
                  if (ang(1,i).ne.ang(1,j)) then
                     ntor=ntor+1
                     if (ntor.le.mtotot) then
                        tor(1,ntor)=ang(1,i)
                        tor(2,ntor)=ang(2,i)
                        tor(3,ntor)=ang(3,i)
                        tor(4,ntor)=ang(1,j)
                     end if
                  end if
               end if
            end if
 290     continue
 300  continue
      if (ntor.gt.mtotot) then
         write(6,'(a)') ' Too many torsion angles.'
         return 1
      end if
c     
c     Create list of pi-center planes
c     
      npla=0
      do 400 i=1,natpi
         npla=npla+1
         nn=1
         pla(nn,npla)=atnr(i)
         do 310 j=1,nbnd
            if (atnr(i).eq.bnd(1,j)) then
               nn=nn+1
               if (nn.le.4) pla(nn,npla)=bnd(2,j)
            else if (atnr(i).eq.bnd(2,j)) then
               nn=nn+1
               if (nn.le.4) pla(nn,npla)=bnd(1,j)
            end if
 310     continue
         if (nn.ne.4) then
            npla=npla-1
         else
            do 390 j=1,2
               npla=npla+1
               pla(1,npla)=pla(1,npla-1)
               pla(2,npla)=pla(4,npla-1)
               pla(3,npla)=pla(2,npla-1)
               pla(4,npla)=pla(3,npla-1)
 390        continue  
         end if
 400  continue

c     
c     done...
c     
      return
      end
