#ifdef mingw
#include <windows.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>

#define TRUE  1
#define FALSE 0

static int HW=1;
FILE *R=(FILE *)NULL;
static unsigned int S=0,O=0,P;

#ifndef mingw
static unsigned int K;
static void wakeup(int x);
static int KnotZero(void);
#endif

#include "isaac.h"

static ISAACctxt ctxt=NULL;

/* NOTE: the rseed routines are intended to be used to generate  */
/* random seeds for pseudo-random generators. They are not good  */
/* random number generators. */


#ifndef mingw
unsigned int rseedSlow(void)
{
  /* a rather slow method to generate a random unsigned int      */
  /* takes 1 second per byte, and should be pretty random on     */
  /* most time shared systems (relies on inaccuracy of alarm(2), */
  /* and on being interrupted by scheduler).                     */

  unsigned int r,b;
  int i;

  signal(SIGALRM,wakeup);
  for (b=0,r=0,i=0;i<8*sizeof(int);i+=8) {
    K=1;
    alarm(1);
    while(KnotZero()) {
      b++;
      b&=255;
    }
    r|=b<<i;
  }
  return(r);
}
#endif

unsigned int rseedFast(void)
{
  struct timeval tv;
  unsigned int r,r1,r2;
  int k;
  
  /* faster generator of random seed, relies on reasonable resolution */
  /* of timeofday. should not be used too often within one tick */

#ifndef mingw
  gettimeofday(&tv,NULL);
#else
  tv.tv_usec=GetTickCount();
  tv.tv_sec=GetTickCount()^(0x132547fa);
#endif
  if (S==0) S=tv.tv_usec&31;
  if (P) P+=19;
  else P=getpid();
  r1=tv.tv_sec^((P<<16)|(tv.tv_usec&(0xffff)));
  for (r2=0,k=0;k<31;k++) r2=r2<<k|((r1>>k)&1);
  r=r2^(((r1<<S))|(r1>>(S-32)))^tv.tv_usec^((O>>S)&1);
  O=(r>>S)^r2;
  return(r);
}

unsigned int randomHardware(void)
{
  unsigned int r,r2;

  /* try to get random number from /dev/random. If this fails, switch */
  /* to rseedFast method */

  if (HW) {
    if (!R) {
      R=fopen("/dev/urandom","r");
      if (R) {
	if (fread(&r,sizeof(unsigned int),1,R)<=0) r=0;
	fclose(R);
	R=fopen("/dev/urandom","r");
	if (fread(&r2,sizeof(unsigned int),1,R)<=0)r=0 ;
	if (r==r2) {
	  fclose(R);
	  R=(FILE *)NULL;
	  HW=FALSE;
	  r=rseedFast();
	}
      }else{
	HW=FALSE;
	r=rseedFast();
      }
    }else if (fread(&r,sizeof(unsigned int),1,R)<=0) r=rseedFast();
  }else r=rseedFast();
  return(r);
}

#ifndef mingw
static void wakeup(int sig)
{
  K=0;
  signal(SIGALRM,wakeup);
}
#endif

#ifndef mingw
static int KnotZero(void)
{
  return(K!=0);
}
#endif

void setSeed(unsigned int seed)
{
  unsigned long s;

  /* Note: id seed==0, try to generate a random seed */

  if (seed) s=seed;
  else s=randomHardware();
  if (ctxt) ctxt=ISAACfree(ctxt);
  ctxt=ISAACseed(1,&s);
}

float ufrand(float a,float b)
{
  if (!ctxt) ctxt=ISAACseed(0,NULL);
  return(a+(b-a)*ISAACdrand(ctxt));
}

int uirand(int a,int b)
{
  int r;

  if (!ctxt) ctxt=ISAACseed(0,NULL);
  r=a+(b-a+1)*ISAACdrand(ctxt);
  if (r>b) r=b;
  else if (r<a) r=a;
  return(r);
}

/* FORTRAN interface... */

void setrand_(unsigned int *s)
{
  if (*s!=0) setSeed(rseedFast());
  else setSeed(*s);
}

float randomf_(void)
{
  return(ufrand(0.0,1.0));
}
